/* Automatically @generated by tree-sitter v0.25.8 */

#include "tree_sitter/parser.h"

#if defined(__GNUC__) || defined(__clang__)
#pragma GCC diagnostic ignored "-Wmissing-field-initializers"
#endif

#define LANGUAGE_VERSION 15
#define STATE_COUNT 1870
#define LARGE_STATE_COUNT 387
#define SYMBOL_COUNT 261
#define ALIAS_COUNT 4
#define TOKEN_COUNT 134
#define EXTERNAL_TOKEN_COUNT 8
#define FIELD_COUNT 36
#define MAX_ALIAS_SEQUENCE_LENGTH 9
#define MAX_RESERVED_WORD_SET_SIZE 35
#define PRODUCTION_ID_COUNT 121
#define SUPERTYPE_COUNT 5

enum ts_symbol_identifiers {
  sym_identifier = 1,
  sym_hash_bang_line = 2,
  anon_sym_export = 3,
  anon_sym_STAR = 4,
  anon_sym_default = 5,
  anon_sym_as = 6,
  anon_sym_LBRACE = 7,
  anon_sym_COMMA = 8,
  anon_sym_RBRACE = 9,
  anon_sym_import = 10,
  anon_sym_from = 11,
  anon_sym_with = 12,
  anon_sym_var = 13,
  anon_sym_let = 14,
  anon_sym_const = 15,
  anon_sym_using = 16,
  anon_sym_await = 17,
  anon_sym_of = 18,
  anon_sym_else = 19,
  anon_sym_if = 20,
  anon_sym_switch = 21,
  anon_sym_for = 22,
  anon_sym_LPAREN = 23,
  anon_sym_SEMI = 24,
  anon_sym_RPAREN = 25,
  anon_sym_in = 26,
  anon_sym_while = 27,
  anon_sym_do = 28,
  anon_sym_try = 29,
  anon_sym_break = 30,
  anon_sym_continue = 31,
  anon_sym_debugger = 32,
  anon_sym_return = 33,
  anon_sym_throw = 34,
  anon_sym_COLON = 35,
  anon_sym_case = 36,
  anon_sym_catch = 37,
  anon_sym_finally = 38,
  anon_sym_yield = 39,
  anon_sym_EQ = 40,
  anon_sym_LBRACK = 41,
  anon_sym_RBRACK = 42,
  sym_html_character_reference = 43,
  anon_sym_LT = 44,
  anon_sym_GT = 45,
  sym_jsx_identifier = 46,
  anon_sym_DOT = 47,
  anon_sym_LT_SLASH = 48,
  anon_sym_SLASH_GT = 49,
  anon_sym_DQUOTE = 50,
  anon_sym_SQUOTE = 51,
  sym_unescaped_double_jsx_string_fragment = 52,
  sym_unescaped_single_jsx_string_fragment = 53,
  anon_sym_class = 54,
  anon_sym_extends = 55,
  anon_sym_async = 56,
  anon_sym_function = 57,
  anon_sym_EQ_GT = 58,
  sym_optional_chain = 59,
  anon_sym_new = 60,
  anon_sym_PLUS_EQ = 61,
  anon_sym_DASH_EQ = 62,
  anon_sym_STAR_EQ = 63,
  anon_sym_SLASH_EQ = 64,
  anon_sym_PERCENT_EQ = 65,
  anon_sym_CARET_EQ = 66,
  anon_sym_AMP_EQ = 67,
  anon_sym_PIPE_EQ = 68,
  anon_sym_GT_GT_EQ = 69,
  anon_sym_GT_GT_GT_EQ = 70,
  anon_sym_LT_LT_EQ = 71,
  anon_sym_STAR_STAR_EQ = 72,
  anon_sym_AMP_AMP_EQ = 73,
  anon_sym_PIPE_PIPE_EQ = 74,
  anon_sym_QMARK_QMARK_EQ = 75,
  anon_sym_DOT_DOT_DOT = 76,
  anon_sym_AMP_AMP = 77,
  anon_sym_PIPE_PIPE = 78,
  anon_sym_GT_GT = 79,
  anon_sym_GT_GT_GT = 80,
  anon_sym_LT_LT = 81,
  anon_sym_AMP = 82,
  anon_sym_CARET = 83,
  anon_sym_PIPE = 84,
  anon_sym_PLUS = 85,
  anon_sym_DASH = 86,
  anon_sym_SLASH = 87,
  anon_sym_PERCENT = 88,
  anon_sym_STAR_STAR = 89,
  anon_sym_LT_EQ = 90,
  anon_sym_EQ_EQ = 91,
  anon_sym_EQ_EQ_EQ = 92,
  anon_sym_BANG_EQ = 93,
  anon_sym_BANG_EQ_EQ = 94,
  anon_sym_GT_EQ = 95,
  anon_sym_QMARK_QMARK = 96,
  anon_sym_instanceof = 97,
  anon_sym_BANG = 98,
  anon_sym_TILDE = 99,
  anon_sym_typeof = 100,
  anon_sym_void = 101,
  anon_sym_delete = 102,
  anon_sym_PLUS_PLUS = 103,
  anon_sym_DASH_DASH = 104,
  sym_unescaped_double_string_fragment = 105,
  sym_unescaped_single_string_fragment = 106,
  sym_escape_sequence = 107,
  sym_comment = 108,
  anon_sym_BQUOTE = 109,
  anon_sym_DOLLAR_LBRACE = 110,
  anon_sym_SLASH2 = 111,
  sym_regex_pattern = 112,
  sym_regex_flags = 113,
  sym_number = 114,
  sym_private_property_identifier = 115,
  anon_sym_target = 116,
  anon_sym_meta = 117,
  sym_this = 118,
  sym_super = 119,
  sym_true = 120,
  sym_false = 121,
  sym_null = 122,
  sym_undefined = 123,
  anon_sym_AT = 124,
  anon_sym_static = 125,
  aux_sym_method_definition_token1 = 126,
  anon_sym_get = 127,
  anon_sym_set = 128,
  sym__automatic_semicolon = 129,
  sym__template_chars = 130,
  sym__ternary_qmark = 131,
  sym_html_comment = 132,
  sym_jsx_text = 133,
  sym_program = 134,
  sym_export_statement = 135,
  sym_namespace_export = 136,
  sym_export_clause = 137,
  sym_export_specifier = 138,
  sym__module_export_name = 139,
  sym_declaration = 140,
  sym_import = 141,
  sym_import_statement = 142,
  sym_import_clause = 143,
  sym__from_clause = 144,
  sym_namespace_import = 145,
  sym_named_imports = 146,
  sym_import_specifier = 147,
  sym_import_attribute = 148,
  sym_statement = 149,
  sym_expression_statement = 150,
  sym_variable_declaration = 151,
  sym_lexical_declaration = 152,
  sym_using_declaration = 153,
  sym_variable_declarator = 154,
  sym_statement_block = 155,
  sym_else_clause = 156,
  sym_if_statement = 157,
  sym_switch_statement = 158,
  sym_for_statement = 159,
  sym_for_in_statement = 160,
  sym__for_header = 161,
  sym_while_statement = 162,
  sym_do_statement = 163,
  sym_try_statement = 164,
  sym_with_statement = 165,
  sym_break_statement = 166,
  sym_continue_statement = 167,
  sym_debugger_statement = 168,
  sym_return_statement = 169,
  sym_throw_statement = 170,
  sym_empty_statement = 171,
  sym_labeled_statement = 172,
  sym_switch_body = 173,
  sym_switch_case = 174,
  sym_switch_default = 175,
  sym_catch_clause = 176,
  sym_finally_clause = 177,
  sym_parenthesized_expression = 178,
  sym_expression = 179,
  sym_primary_expression = 180,
  sym_yield_expression = 181,
  sym_object = 182,
  sym_object_pattern = 183,
  sym_assignment_pattern = 184,
  sym_object_assignment_pattern = 185,
  sym_array = 186,
  sym_array_pattern = 187,
  sym_jsx_element = 188,
  sym_jsx_expression = 189,
  sym_jsx_opening_element = 190,
  sym_nested_identifier = 191,
  sym_jsx_namespace_name = 192,
  sym_jsx_closing_element = 193,
  sym_jsx_self_closing_element = 194,
  sym_jsx_attribute = 195,
  sym__jsx_string = 196,
  sym_class = 197,
  sym_class_declaration = 198,
  sym_class_heritage = 199,
  sym_function_expression = 200,
  sym_function_declaration = 201,
  sym_generator_function = 202,
  sym_generator_function_declaration = 203,
  sym_arrow_function = 204,
  sym_call_expression = 205,
  sym_new_expression = 206,
  sym_await_expression = 207,
  sym_member_expression = 208,
  sym_subscript_expression = 209,
  sym_assignment_expression = 210,
  sym__augmented_assignment_lhs = 211,
  sym_augmented_assignment_expression = 212,
  sym__initializer = 213,
  sym__destructuring_pattern = 214,
  sym_spread_element = 215,
  sym_ternary_expression = 216,
  sym_binary_expression = 217,
  sym_unary_expression = 218,
  sym_update_expression = 219,
  sym_sequence_expression = 220,
  sym_string = 221,
  sym_template_string = 222,
  sym_template_substitution = 223,
  sym_regex = 224,
  sym_meta_property = 225,
  sym_arguments = 226,
  sym_decorator = 227,
  sym_decorator_member_expression = 228,
  sym_decorator_call_expression = 229,
  sym_class_body = 230,
  sym_field_definition = 231,
  sym_formal_parameters = 232,
  sym_class_static_block = 233,
  sym_pattern = 234,
  sym_rest_pattern = 235,
  sym_method_definition = 236,
  sym_pair = 237,
  sym_pair_pattern = 238,
  sym__property_name = 239,
  sym_computed_property_name = 240,
  aux_sym_program_repeat1 = 241,
  aux_sym_export_statement_repeat1 = 242,
  aux_sym_export_clause_repeat1 = 243,
  aux_sym_named_imports_repeat1 = 244,
  aux_sym_variable_declaration_repeat1 = 245,
  aux_sym_switch_body_repeat1 = 246,
  aux_sym_object_repeat1 = 247,
  aux_sym_object_pattern_repeat1 = 248,
  aux_sym_array_repeat1 = 249,
  aux_sym_array_pattern_repeat1 = 250,
  aux_sym_jsx_element_repeat1 = 251,
  aux_sym_jsx_opening_element_repeat1 = 252,
  aux_sym__jsx_string_repeat1 = 253,
  aux_sym__jsx_string_repeat2 = 254,
  aux_sym_sequence_expression_repeat1 = 255,
  aux_sym_string_repeat1 = 256,
  aux_sym_string_repeat2 = 257,
  aux_sym_template_string_repeat1 = 258,
  aux_sym_class_body_repeat1 = 259,
  aux_sym_formal_parameters_repeat1 = 260,
  alias_sym_property_identifier = 261,
  alias_sym_shorthand_property_identifier = 262,
  alias_sym_shorthand_property_identifier_pattern = 263,
  alias_sym_statement_identifier = 264,
};

static const char * const ts_symbol_names[] = {
  [ts_builtin_sym_end] = "end",
  [sym_identifier] = "identifier",
  [sym_hash_bang_line] = "hash_bang_line",
  [anon_sym_export] = "export",
  [anon_sym_STAR] = "*",
  [anon_sym_default] = "default",
  [anon_sym_as] = "as",
  [anon_sym_LBRACE] = "{",
  [anon_sym_COMMA] = ",",
  [anon_sym_RBRACE] = "}",
  [anon_sym_import] = "import",
  [anon_sym_from] = "from",
  [anon_sym_with] = "with",
  [anon_sym_var] = "var",
  [anon_sym_let] = "let",
  [anon_sym_const] = "const",
  [anon_sym_using] = "using",
  [anon_sym_await] = "await",
  [anon_sym_of] = "of",
  [anon_sym_else] = "else",
  [anon_sym_if] = "if",
  [anon_sym_switch] = "switch",
  [anon_sym_for] = "for",
  [anon_sym_LPAREN] = "(",
  [anon_sym_SEMI] = ";",
  [anon_sym_RPAREN] = ")",
  [anon_sym_in] = "in",
  [anon_sym_while] = "while",
  [anon_sym_do] = "do",
  [anon_sym_try] = "try",
  [anon_sym_break] = "break",
  [anon_sym_continue] = "continue",
  [anon_sym_debugger] = "debugger",
  [anon_sym_return] = "return",
  [anon_sym_throw] = "throw",
  [anon_sym_COLON] = ":",
  [anon_sym_case] = "case",
  [anon_sym_catch] = "catch",
  [anon_sym_finally] = "finally",
  [anon_sym_yield] = "yield",
  [anon_sym_EQ] = "=",
  [anon_sym_LBRACK] = "[",
  [anon_sym_RBRACK] = "]",
  [sym_html_character_reference] = "html_character_reference",
  [anon_sym_LT] = "<",
  [anon_sym_GT] = ">",
  [sym_jsx_identifier] = "identifier",
  [anon_sym_DOT] = ".",
  [anon_sym_LT_SLASH] = "</",
  [anon_sym_SLASH_GT] = "/>",
  [anon_sym_DQUOTE] = "\"",
  [anon_sym_SQUOTE] = "'",
  [sym_unescaped_double_jsx_string_fragment] = "string_fragment",
  [sym_unescaped_single_jsx_string_fragment] = "string_fragment",
  [anon_sym_class] = "class",
  [anon_sym_extends] = "extends",
  [anon_sym_async] = "async",
  [anon_sym_function] = "function",
  [anon_sym_EQ_GT] = "=>",
  [sym_optional_chain] = "optional_chain",
  [anon_sym_new] = "new",
  [anon_sym_PLUS_EQ] = "+=",
  [anon_sym_DASH_EQ] = "-=",
  [anon_sym_STAR_EQ] = "*=",
  [anon_sym_SLASH_EQ] = "/=",
  [anon_sym_PERCENT_EQ] = "%=",
  [anon_sym_CARET_EQ] = "^=",
  [anon_sym_AMP_EQ] = "&=",
  [anon_sym_PIPE_EQ] = "|=",
  [anon_sym_GT_GT_EQ] = ">>=",
  [anon_sym_GT_GT_GT_EQ] = ">>>=",
  [anon_sym_LT_LT_EQ] = "<<=",
  [anon_sym_STAR_STAR_EQ] = "**=",
  [anon_sym_AMP_AMP_EQ] = "&&=",
  [anon_sym_PIPE_PIPE_EQ] = "||=",
  [anon_sym_QMARK_QMARK_EQ] = "\?\?=",
  [anon_sym_DOT_DOT_DOT] = "...",
  [anon_sym_AMP_AMP] = "&&",
  [anon_sym_PIPE_PIPE] = "||",
  [anon_sym_GT_GT] = ">>",
  [anon_sym_GT_GT_GT] = ">>>",
  [anon_sym_LT_LT] = "<<",
  [anon_sym_AMP] = "&",
  [anon_sym_CARET] = "^",
  [anon_sym_PIPE] = "|",
  [anon_sym_PLUS] = "+",
  [anon_sym_DASH] = "-",
  [anon_sym_SLASH] = "/",
  [anon_sym_PERCENT] = "%",
  [anon_sym_STAR_STAR] = "**",
  [anon_sym_LT_EQ] = "<=",
  [anon_sym_EQ_EQ] = "==",
  [anon_sym_EQ_EQ_EQ] = "===",
  [anon_sym_BANG_EQ] = "!=",
  [anon_sym_BANG_EQ_EQ] = "!==",
  [anon_sym_GT_EQ] = ">=",
  [anon_sym_QMARK_QMARK] = "\?\?",
  [anon_sym_instanceof] = "instanceof",
  [anon_sym_BANG] = "!",
  [anon_sym_TILDE] = "~",
  [anon_sym_typeof] = "typeof",
  [anon_sym_void] = "void",
  [anon_sym_delete] = "delete",
  [anon_sym_PLUS_PLUS] = "++",
  [anon_sym_DASH_DASH] = "--",
  [sym_unescaped_double_string_fragment] = "string_fragment",
  [sym_unescaped_single_string_fragment] = "string_fragment",
  [sym_escape_sequence] = "escape_sequence",
  [sym_comment] = "comment",
  [anon_sym_BQUOTE] = "`",
  [anon_sym_DOLLAR_LBRACE] = "${",
  [anon_sym_SLASH2] = "/",
  [sym_regex_pattern] = "regex_pattern",
  [sym_regex_flags] = "regex_flags",
  [sym_number] = "number",
  [sym_private_property_identifier] = "private_property_identifier",
  [anon_sym_target] = "target",
  [anon_sym_meta] = "meta",
  [sym_this] = "this",
  [sym_super] = "super",
  [sym_true] = "true",
  [sym_false] = "false",
  [sym_null] = "null",
  [sym_undefined] = "undefined",
  [anon_sym_AT] = "@",
  [anon_sym_static] = "static",
  [aux_sym_method_definition_token1] = "static get",
  [anon_sym_get] = "get",
  [anon_sym_set] = "set",
  [sym__automatic_semicolon] = "_automatic_semicolon",
  [sym__template_chars] = "string_fragment",
  [sym__ternary_qmark] = "\?",
  [sym_html_comment] = "html_comment",
  [sym_jsx_text] = "jsx_text",
  [sym_program] = "program",
  [sym_export_statement] = "export_statement",
  [sym_namespace_export] = "namespace_export",
  [sym_export_clause] = "export_clause",
  [sym_export_specifier] = "export_specifier",
  [sym__module_export_name] = "_module_export_name",
  [sym_declaration] = "declaration",
  [sym_import] = "import",
  [sym_import_statement] = "import_statement",
  [sym_import_clause] = "import_clause",
  [sym__from_clause] = "_from_clause",
  [sym_namespace_import] = "namespace_import",
  [sym_named_imports] = "named_imports",
  [sym_import_specifier] = "import_specifier",
  [sym_import_attribute] = "import_attribute",
  [sym_statement] = "statement",
  [sym_expression_statement] = "expression_statement",
  [sym_variable_declaration] = "variable_declaration",
  [sym_lexical_declaration] = "lexical_declaration",
  [sym_using_declaration] = "using_declaration",
  [sym_variable_declarator] = "variable_declarator",
  [sym_statement_block] = "statement_block",
  [sym_else_clause] = "else_clause",
  [sym_if_statement] = "if_statement",
  [sym_switch_statement] = "switch_statement",
  [sym_for_statement] = "for_statement",
  [sym_for_in_statement] = "for_in_statement",
  [sym__for_header] = "_for_header",
  [sym_while_statement] = "while_statement",
  [sym_do_statement] = "do_statement",
  [sym_try_statement] = "try_statement",
  [sym_with_statement] = "with_statement",
  [sym_break_statement] = "break_statement",
  [sym_continue_statement] = "continue_statement",
  [sym_debugger_statement] = "debugger_statement",
  [sym_return_statement] = "return_statement",
  [sym_throw_statement] = "throw_statement",
  [sym_empty_statement] = "empty_statement",
  [sym_labeled_statement] = "labeled_statement",
  [sym_switch_body] = "switch_body",
  [sym_switch_case] = "switch_case",
  [sym_switch_default] = "switch_default",
  [sym_catch_clause] = "catch_clause",
  [sym_finally_clause] = "finally_clause",
  [sym_parenthesized_expression] = "parenthesized_expression",
  [sym_expression] = "expression",
  [sym_primary_expression] = "primary_expression",
  [sym_yield_expression] = "yield_expression",
  [sym_object] = "object",
  [sym_object_pattern] = "object_pattern",
  [sym_assignment_pattern] = "assignment_pattern",
  [sym_object_assignment_pattern] = "object_assignment_pattern",
  [sym_array] = "array",
  [sym_array_pattern] = "array_pattern",
  [sym_jsx_element] = "jsx_element",
  [sym_jsx_expression] = "jsx_expression",
  [sym_jsx_opening_element] = "jsx_opening_element",
  [sym_nested_identifier] = "member_expression",
  [sym_jsx_namespace_name] = "jsx_namespace_name",
  [sym_jsx_closing_element] = "jsx_closing_element",
  [sym_jsx_self_closing_element] = "jsx_self_closing_element",
  [sym_jsx_attribute] = "jsx_attribute",
  [sym__jsx_string] = "string",
  [sym_class] = "class",
  [sym_class_declaration] = "class_declaration",
  [sym_class_heritage] = "class_heritage",
  [sym_function_expression] = "function_expression",
  [sym_function_declaration] = "function_declaration",
  [sym_generator_function] = "generator_function",
  [sym_generator_function_declaration] = "generator_function_declaration",
  [sym_arrow_function] = "arrow_function",
  [sym_call_expression] = "call_expression",
  [sym_new_expression] = "new_expression",
  [sym_await_expression] = "await_expression",
  [sym_member_expression] = "member_expression",
  [sym_subscript_expression] = "subscript_expression",
  [sym_assignment_expression] = "assignment_expression",
  [sym__augmented_assignment_lhs] = "_augmented_assignment_lhs",
  [sym_augmented_assignment_expression] = "augmented_assignment_expression",
  [sym__initializer] = "_initializer",
  [sym__destructuring_pattern] = "_destructuring_pattern",
  [sym_spread_element] = "spread_element",
  [sym_ternary_expression] = "ternary_expression",
  [sym_binary_expression] = "binary_expression",
  [sym_unary_expression] = "unary_expression",
  [sym_update_expression] = "update_expression",
  [sym_sequence_expression] = "sequence_expression",
  [sym_string] = "string",
  [sym_template_string] = "template_string",
  [sym_template_substitution] = "template_substitution",
  [sym_regex] = "regex",
  [sym_meta_property] = "meta_property",
  [sym_arguments] = "arguments",
  [sym_decorator] = "decorator",
  [sym_decorator_member_expression] = "member_expression",
  [sym_decorator_call_expression] = "call_expression",
  [sym_class_body] = "class_body",
  [sym_field_definition] = "field_definition",
  [sym_formal_parameters] = "formal_parameters",
  [sym_class_static_block] = "class_static_block",
  [sym_pattern] = "pattern",
  [sym_rest_pattern] = "rest_pattern",
  [sym_method_definition] = "method_definition",
  [sym_pair] = "pair",
  [sym_pair_pattern] = "pair_pattern",
  [sym__property_name] = "_property_name",
  [sym_computed_property_name] = "computed_property_name",
  [aux_sym_program_repeat1] = "program_repeat1",
  [aux_sym_export_statement_repeat1] = "export_statement_repeat1",
  [aux_sym_export_clause_repeat1] = "export_clause_repeat1",
  [aux_sym_named_imports_repeat1] = "named_imports_repeat1",
  [aux_sym_variable_declaration_repeat1] = "variable_declaration_repeat1",
  [aux_sym_switch_body_repeat1] = "switch_body_repeat1",
  [aux_sym_object_repeat1] = "object_repeat1",
  [aux_sym_object_pattern_repeat1] = "object_pattern_repeat1",
  [aux_sym_array_repeat1] = "array_repeat1",
  [aux_sym_array_pattern_repeat1] = "array_pattern_repeat1",
  [aux_sym_jsx_element_repeat1] = "jsx_element_repeat1",
  [aux_sym_jsx_opening_element_repeat1] = "jsx_opening_element_repeat1",
  [aux_sym__jsx_string_repeat1] = "_jsx_string_repeat1",
  [aux_sym__jsx_string_repeat2] = "_jsx_string_repeat2",
  [aux_sym_sequence_expression_repeat1] = "sequence_expression_repeat1",
  [aux_sym_string_repeat1] = "string_repeat1",
  [aux_sym_string_repeat2] = "string_repeat2",
  [aux_sym_template_string_repeat1] = "template_string_repeat1",
  [aux_sym_class_body_repeat1] = "class_body_repeat1",
  [aux_sym_formal_parameters_repeat1] = "formal_parameters_repeat1",
  [alias_sym_property_identifier] = "property_identifier",
  [alias_sym_shorthand_property_identifier] = "shorthand_property_identifier",
  [alias_sym_shorthand_property_identifier_pattern] = "shorthand_property_identifier_pattern",
  [alias_sym_statement_identifier] = "statement_identifier",
};

static const TSSymbol ts_symbol_map[] = {
  [ts_builtin_sym_end] = ts_builtin_sym_end,
  [sym_identifier] = sym_identifier,
  [sym_hash_bang_line] = sym_hash_bang_line,
  [anon_sym_export] = anon_sym_export,
  [anon_sym_STAR] = anon_sym_STAR,
  [anon_sym_default] = anon_sym_default,
  [anon_sym_as] = anon_sym_as,
  [anon_sym_LBRACE] = anon_sym_LBRACE,
  [anon_sym_COMMA] = anon_sym_COMMA,
  [anon_sym_RBRACE] = anon_sym_RBRACE,
  [anon_sym_import] = anon_sym_import,
  [anon_sym_from] = anon_sym_from,
  [anon_sym_with] = anon_sym_with,
  [anon_sym_var] = anon_sym_var,
  [anon_sym_let] = anon_sym_let,
  [anon_sym_const] = anon_sym_const,
  [anon_sym_using] = anon_sym_using,
  [anon_sym_await] = anon_sym_await,
  [anon_sym_of] = anon_sym_of,
  [anon_sym_else] = anon_sym_else,
  [anon_sym_if] = anon_sym_if,
  [anon_sym_switch] = anon_sym_switch,
  [anon_sym_for] = anon_sym_for,
  [anon_sym_LPAREN] = anon_sym_LPAREN,
  [anon_sym_SEMI] = anon_sym_SEMI,
  [anon_sym_RPAREN] = anon_sym_RPAREN,
  [anon_sym_in] = anon_sym_in,
  [anon_sym_while] = anon_sym_while,
  [anon_sym_do] = anon_sym_do,
  [anon_sym_try] = anon_sym_try,
  [anon_sym_break] = anon_sym_break,
  [anon_sym_continue] = anon_sym_continue,
  [anon_sym_debugger] = anon_sym_debugger,
  [anon_sym_return] = anon_sym_return,
  [anon_sym_throw] = anon_sym_throw,
  [anon_sym_COLON] = anon_sym_COLON,
  [anon_sym_case] = anon_sym_case,
  [anon_sym_catch] = anon_sym_catch,
  [anon_sym_finally] = anon_sym_finally,
  [anon_sym_yield] = anon_sym_yield,
  [anon_sym_EQ] = anon_sym_EQ,
  [anon_sym_LBRACK] = anon_sym_LBRACK,
  [anon_sym_RBRACK] = anon_sym_RBRACK,
  [sym_html_character_reference] = sym_html_character_reference,
  [anon_sym_LT] = anon_sym_LT,
  [anon_sym_GT] = anon_sym_GT,
  [sym_jsx_identifier] = sym_identifier,
  [anon_sym_DOT] = anon_sym_DOT,
  [anon_sym_LT_SLASH] = anon_sym_LT_SLASH,
  [anon_sym_SLASH_GT] = anon_sym_SLASH_GT,
  [anon_sym_DQUOTE] = anon_sym_DQUOTE,
  [anon_sym_SQUOTE] = anon_sym_SQUOTE,
  [sym_unescaped_double_jsx_string_fragment] = sym__template_chars,
  [sym_unescaped_single_jsx_string_fragment] = sym__template_chars,
  [anon_sym_class] = anon_sym_class,
  [anon_sym_extends] = anon_sym_extends,
  [anon_sym_async] = anon_sym_async,
  [anon_sym_function] = anon_sym_function,
  [anon_sym_EQ_GT] = anon_sym_EQ_GT,
  [sym_optional_chain] = sym_optional_chain,
  [anon_sym_new] = anon_sym_new,
  [anon_sym_PLUS_EQ] = anon_sym_PLUS_EQ,
  [anon_sym_DASH_EQ] = anon_sym_DASH_EQ,
  [anon_sym_STAR_EQ] = anon_sym_STAR_EQ,
  [anon_sym_SLASH_EQ] = anon_sym_SLASH_EQ,
  [anon_sym_PERCENT_EQ] = anon_sym_PERCENT_EQ,
  [anon_sym_CARET_EQ] = anon_sym_CARET_EQ,
  [anon_sym_AMP_EQ] = anon_sym_AMP_EQ,
  [anon_sym_PIPE_EQ] = anon_sym_PIPE_EQ,
  [anon_sym_GT_GT_EQ] = anon_sym_GT_GT_EQ,
  [anon_sym_GT_GT_GT_EQ] = anon_sym_GT_GT_GT_EQ,
  [anon_sym_LT_LT_EQ] = anon_sym_LT_LT_EQ,
  [anon_sym_STAR_STAR_EQ] = anon_sym_STAR_STAR_EQ,
  [anon_sym_AMP_AMP_EQ] = anon_sym_AMP_AMP_EQ,
  [anon_sym_PIPE_PIPE_EQ] = anon_sym_PIPE_PIPE_EQ,
  [anon_sym_QMARK_QMARK_EQ] = anon_sym_QMARK_QMARK_EQ,
  [anon_sym_DOT_DOT_DOT] = anon_sym_DOT_DOT_DOT,
  [anon_sym_AMP_AMP] = anon_sym_AMP_AMP,
  [anon_sym_PIPE_PIPE] = anon_sym_PIPE_PIPE,
  [anon_sym_GT_GT] = anon_sym_GT_GT,
  [anon_sym_GT_GT_GT] = anon_sym_GT_GT_GT,
  [anon_sym_LT_LT] = anon_sym_LT_LT,
  [anon_sym_AMP] = anon_sym_AMP,
  [anon_sym_CARET] = anon_sym_CARET,
  [anon_sym_PIPE] = anon_sym_PIPE,
  [anon_sym_PLUS] = anon_sym_PLUS,
  [anon_sym_DASH] = anon_sym_DASH,
  [anon_sym_SLASH] = anon_sym_SLASH,
  [anon_sym_PERCENT] = anon_sym_PERCENT,
  [anon_sym_STAR_STAR] = anon_sym_STAR_STAR,
  [anon_sym_LT_EQ] = anon_sym_LT_EQ,
  [anon_sym_EQ_EQ] = anon_sym_EQ_EQ,
  [anon_sym_EQ_EQ_EQ] = anon_sym_EQ_EQ_EQ,
  [anon_sym_BANG_EQ] = anon_sym_BANG_EQ,
  [anon_sym_BANG_EQ_EQ] = anon_sym_BANG_EQ_EQ,
  [anon_sym_GT_EQ] = anon_sym_GT_EQ,
  [anon_sym_QMARK_QMARK] = anon_sym_QMARK_QMARK,
  [anon_sym_instanceof] = anon_sym_instanceof,
  [anon_sym_BANG] = anon_sym_BANG,
  [anon_sym_TILDE] = anon_sym_TILDE,
  [anon_sym_typeof] = anon_sym_typeof,
  [anon_sym_void] = anon_sym_void,
  [anon_sym_delete] = anon_sym_delete,
  [anon_sym_PLUS_PLUS] = anon_sym_PLUS_PLUS,
  [anon_sym_DASH_DASH] = anon_sym_DASH_DASH,
  [sym_unescaped_double_string_fragment] = sym__template_chars,
  [sym_unescaped_single_string_fragment] = sym__template_chars,
  [sym_escape_sequence] = sym_escape_sequence,
  [sym_comment] = sym_comment,
  [anon_sym_BQUOTE] = anon_sym_BQUOTE,
  [anon_sym_DOLLAR_LBRACE] = anon_sym_DOLLAR_LBRACE,
  [anon_sym_SLASH2] = anon_sym_SLASH,
  [sym_regex_pattern] = sym_regex_pattern,
  [sym_regex_flags] = sym_regex_flags,
  [sym_number] = sym_number,
  [sym_private_property_identifier] = sym_private_property_identifier,
  [anon_sym_target] = anon_sym_target,
  [anon_sym_meta] = anon_sym_meta,
  [sym_this] = sym_this,
  [sym_super] = sym_super,
  [sym_true] = sym_true,
  [sym_false] = sym_false,
  [sym_null] = sym_null,
  [sym_undefined] = sym_undefined,
  [anon_sym_AT] = anon_sym_AT,
  [anon_sym_static] = anon_sym_static,
  [aux_sym_method_definition_token1] = aux_sym_method_definition_token1,
  [anon_sym_get] = anon_sym_get,
  [anon_sym_set] = anon_sym_set,
  [sym__automatic_semicolon] = sym__automatic_semicolon,
  [sym__template_chars] = sym__template_chars,
  [sym__ternary_qmark] = sym__ternary_qmark,
  [sym_html_comment] = sym_html_comment,
  [sym_jsx_text] = sym_jsx_text,
  [sym_program] = sym_program,
  [sym_export_statement] = sym_export_statement,
  [sym_namespace_export] = sym_namespace_export,
  [sym_export_clause] = sym_export_clause,
  [sym_export_specifier] = sym_export_specifier,
  [sym__module_export_name] = sym__module_export_name,
  [sym_declaration] = sym_declaration,
  [sym_import] = sym_import,
  [sym_import_statement] = sym_import_statement,
  [sym_import_clause] = sym_import_clause,
  [sym__from_clause] = sym__from_clause,
  [sym_namespace_import] = sym_namespace_import,
  [sym_named_imports] = sym_named_imports,
  [sym_import_specifier] = sym_import_specifier,
  [sym_import_attribute] = sym_import_attribute,
  [sym_statement] = sym_statement,
  [sym_expression_statement] = sym_expression_statement,
  [sym_variable_declaration] = sym_variable_declaration,
  [sym_lexical_declaration] = sym_lexical_declaration,
  [sym_using_declaration] = sym_using_declaration,
  [sym_variable_declarator] = sym_variable_declarator,
  [sym_statement_block] = sym_statement_block,
  [sym_else_clause] = sym_else_clause,
  [sym_if_statement] = sym_if_statement,
  [sym_switch_statement] = sym_switch_statement,
  [sym_for_statement] = sym_for_statement,
  [sym_for_in_statement] = sym_for_in_statement,
  [sym__for_header] = sym__for_header,
  [sym_while_statement] = sym_while_statement,
  [sym_do_statement] = sym_do_statement,
  [sym_try_statement] = sym_try_statement,
  [sym_with_statement] = sym_with_statement,
  [sym_break_statement] = sym_break_statement,
  [sym_continue_statement] = sym_continue_statement,
  [sym_debugger_statement] = sym_debugger_statement,
  [sym_return_statement] = sym_return_statement,
  [sym_throw_statement] = sym_throw_statement,
  [sym_empty_statement] = sym_empty_statement,
  [sym_labeled_statement] = sym_labeled_statement,
  [sym_switch_body] = sym_switch_body,
  [sym_switch_case] = sym_switch_case,
  [sym_switch_default] = sym_switch_default,
  [sym_catch_clause] = sym_catch_clause,
  [sym_finally_clause] = sym_finally_clause,
  [sym_parenthesized_expression] = sym_parenthesized_expression,
  [sym_expression] = sym_expression,
  [sym_primary_expression] = sym_primary_expression,
  [sym_yield_expression] = sym_yield_expression,
  [sym_object] = sym_object,
  [sym_object_pattern] = sym_object_pattern,
  [sym_assignment_pattern] = sym_assignment_pattern,
  [sym_object_assignment_pattern] = sym_object_assignment_pattern,
  [sym_array] = sym_array,
  [sym_array_pattern] = sym_array_pattern,
  [sym_jsx_element] = sym_jsx_element,
  [sym_jsx_expression] = sym_jsx_expression,
  [sym_jsx_opening_element] = sym_jsx_opening_element,
  [sym_nested_identifier] = sym_member_expression,
  [sym_jsx_namespace_name] = sym_jsx_namespace_name,
  [sym_jsx_closing_element] = sym_jsx_closing_element,
  [sym_jsx_self_closing_element] = sym_jsx_self_closing_element,
  [sym_jsx_attribute] = sym_jsx_attribute,
  [sym__jsx_string] = sym_string,
  [sym_class] = sym_class,
  [sym_class_declaration] = sym_class_declaration,
  [sym_class_heritage] = sym_class_heritage,
  [sym_function_expression] = sym_function_expression,
  [sym_function_declaration] = sym_function_declaration,
  [sym_generator_function] = sym_generator_function,
  [sym_generator_function_declaration] = sym_generator_function_declaration,
  [sym_arrow_function] = sym_arrow_function,
  [sym_call_expression] = sym_call_expression,
  [sym_new_expression] = sym_new_expression,
  [sym_await_expression] = sym_await_expression,
  [sym_member_expression] = sym_member_expression,
  [sym_subscript_expression] = sym_subscript_expression,
  [sym_assignment_expression] = sym_assignment_expression,
  [sym__augmented_assignment_lhs] = sym__augmented_assignment_lhs,
  [sym_augmented_assignment_expression] = sym_augmented_assignment_expression,
  [sym__initializer] = sym__initializer,
  [sym__destructuring_pattern] = sym__destructuring_pattern,
  [sym_spread_element] = sym_spread_element,
  [sym_ternary_expression] = sym_ternary_expression,
  [sym_binary_expression] = sym_binary_expression,
  [sym_unary_expression] = sym_unary_expression,
  [sym_update_expression] = sym_update_expression,
  [sym_sequence_expression] = sym_sequence_expression,
  [sym_string] = sym_string,
  [sym_template_string] = sym_template_string,
  [sym_template_substitution] = sym_template_substitution,
  [sym_regex] = sym_regex,
  [sym_meta_property] = sym_meta_property,
  [sym_arguments] = sym_arguments,
  [sym_decorator] = sym_decorator,
  [sym_decorator_member_expression] = sym_member_expression,
  [sym_decorator_call_expression] = sym_call_expression,
  [sym_class_body] = sym_class_body,
  [sym_field_definition] = sym_field_definition,
  [sym_formal_parameters] = sym_formal_parameters,
  [sym_class_static_block] = sym_class_static_block,
  [sym_pattern] = sym_pattern,
  [sym_rest_pattern] = sym_rest_pattern,
  [sym_method_definition] = sym_method_definition,
  [sym_pair] = sym_pair,
  [sym_pair_pattern] = sym_pair_pattern,
  [sym__property_name] = sym__property_name,
  [sym_computed_property_name] = sym_computed_property_name,
  [aux_sym_program_repeat1] = aux_sym_program_repeat1,
  [aux_sym_export_statement_repeat1] = aux_sym_export_statement_repeat1,
  [aux_sym_export_clause_repeat1] = aux_sym_export_clause_repeat1,
  [aux_sym_named_imports_repeat1] = aux_sym_named_imports_repeat1,
  [aux_sym_variable_declaration_repeat1] = aux_sym_variable_declaration_repeat1,
  [aux_sym_switch_body_repeat1] = aux_sym_switch_body_repeat1,
  [aux_sym_object_repeat1] = aux_sym_object_repeat1,
  [aux_sym_object_pattern_repeat1] = aux_sym_object_pattern_repeat1,
  [aux_sym_array_repeat1] = aux_sym_array_repeat1,
  [aux_sym_array_pattern_repeat1] = aux_sym_array_pattern_repeat1,
  [aux_sym_jsx_element_repeat1] = aux_sym_jsx_element_repeat1,
  [aux_sym_jsx_opening_element_repeat1] = aux_sym_jsx_opening_element_repeat1,
  [aux_sym__jsx_string_repeat1] = aux_sym__jsx_string_repeat1,
  [aux_sym__jsx_string_repeat2] = aux_sym__jsx_string_repeat2,
  [aux_sym_sequence_expression_repeat1] = aux_sym_sequence_expression_repeat1,
  [aux_sym_string_repeat1] = aux_sym_string_repeat1,
  [aux_sym_string_repeat2] = aux_sym_string_repeat2,
  [aux_sym_template_string_repeat1] = aux_sym_template_string_repeat1,
  [aux_sym_class_body_repeat1] = aux_sym_class_body_repeat1,
  [aux_sym_formal_parameters_repeat1] = aux_sym_formal_parameters_repeat1,
  [alias_sym_property_identifier] = alias_sym_property_identifier,
  [alias_sym_shorthand_property_identifier] = alias_sym_shorthand_property_identifier,
  [alias_sym_shorthand_property_identifier_pattern] = alias_sym_shorthand_property_identifier_pattern,
  [alias_sym_statement_identifier] = alias_sym_statement_identifier,
};

static const TSSymbolMetadata ts_symbol_metadata[] = {
  [ts_builtin_sym_end] = {
    .visible = false,
    .named = true,
  },
  [sym_identifier] = {
    .visible = true,
    .named = true,
  },
  [sym_hash_bang_line] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_export] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_STAR] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_default] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_as] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LBRACE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_COMMA] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_RBRACE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_import] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_from] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_with] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_var] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_let] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_const] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_using] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_await] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_of] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_else] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_if] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_switch] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_for] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LPAREN] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_SEMI] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_RPAREN] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_in] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_while] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_do] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_try] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_break] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_continue] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_debugger] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_return] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_throw] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_COLON] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_case] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_catch] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_finally] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_yield] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LBRACK] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_RBRACK] = {
    .visible = true,
    .named = false,
  },
  [sym_html_character_reference] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_LT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_GT] = {
    .visible = true,
    .named = false,
  },
  [sym_jsx_identifier] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_DOT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LT_SLASH] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_SLASH_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DQUOTE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_SQUOTE] = {
    .visible = true,
    .named = false,
  },
  [sym_unescaped_double_jsx_string_fragment] = {
    .visible = true,
    .named = true,
  },
  [sym_unescaped_single_jsx_string_fragment] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_class] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_extends] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_async] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_function] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_EQ_GT] = {
    .visible = true,
    .named = false,
  },
  [sym_optional_chain] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_new] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PLUS_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DASH_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_STAR_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_SLASH_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PERCENT_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_CARET_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_AMP_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PIPE_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_GT_GT_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_GT_GT_GT_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LT_LT_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_STAR_STAR_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_AMP_AMP_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PIPE_PIPE_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_QMARK_QMARK_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DOT_DOT_DOT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_AMP_AMP] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PIPE_PIPE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_GT_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_GT_GT_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LT_LT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_AMP] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_CARET] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PIPE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PLUS] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DASH] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_SLASH] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PERCENT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_STAR_STAR] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LT_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_EQ_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_EQ_EQ_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_BANG_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_BANG_EQ_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_GT_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_QMARK_QMARK] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_instanceof] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_BANG] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_TILDE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_typeof] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_void] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_delete] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PLUS_PLUS] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DASH_DASH] = {
    .visible = true,
    .named = false,
  },
  [sym_unescaped_double_string_fragment] = {
    .visible = true,
    .named = true,
  },
  [sym_unescaped_single_string_fragment] = {
    .visible = true,
    .named = true,
  },
  [sym_escape_sequence] = {
    .visible = true,
    .named = true,
  },
  [sym_comment] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_BQUOTE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DOLLAR_LBRACE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_SLASH2] = {
    .visible = true,
    .named = false,
  },
  [sym_regex_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_regex_flags] = {
    .visible = true,
    .named = true,
  },
  [sym_number] = {
    .visible = true,
    .named = true,
  },
  [sym_private_property_identifier] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_target] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_meta] = {
    .visible = true,
    .named = false,
  },
  [sym_this] = {
    .visible = true,
    .named = true,
  },
  [sym_super] = {
    .visible = true,
    .named = true,
  },
  [sym_true] = {
    .visible = true,
    .named = true,
  },
  [sym_false] = {
    .visible = true,
    .named = true,
  },
  [sym_null] = {
    .visible = true,
    .named = true,
  },
  [sym_undefined] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_AT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_static] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_method_definition_token1] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_get] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_set] = {
    .visible = true,
    .named = false,
  },
  [sym__automatic_semicolon] = {
    .visible = false,
    .named = true,
  },
  [sym__template_chars] = {
    .visible = true,
    .named = true,
  },
  [sym__ternary_qmark] = {
    .visible = true,
    .named = false,
  },
  [sym_html_comment] = {
    .visible = true,
    .named = true,
  },
  [sym_jsx_text] = {
    .visible = true,
    .named = true,
  },
  [sym_program] = {
    .visible = true,
    .named = true,
  },
  [sym_export_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_namespace_export] = {
    .visible = true,
    .named = true,
  },
  [sym_export_clause] = {
    .visible = true,
    .named = true,
  },
  [sym_export_specifier] = {
    .visible = true,
    .named = true,
  },
  [sym__module_export_name] = {
    .visible = false,
    .named = true,
  },
  [sym_declaration] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym_import] = {
    .visible = true,
    .named = true,
  },
  [sym_import_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_import_clause] = {
    .visible = true,
    .named = true,
  },
  [sym__from_clause] = {
    .visible = false,
    .named = true,
  },
  [sym_namespace_import] = {
    .visible = true,
    .named = true,
  },
  [sym_named_imports] = {
    .visible = true,
    .named = true,
  },
  [sym_import_specifier] = {
    .visible = true,
    .named = true,
  },
  [sym_import_attribute] = {
    .visible = true,
    .named = true,
  },
  [sym_statement] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym_expression_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_variable_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_lexical_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_using_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_variable_declarator] = {
    .visible = true,
    .named = true,
  },
  [sym_statement_block] = {
    .visible = true,
    .named = true,
  },
  [sym_else_clause] = {
    .visible = true,
    .named = true,
  },
  [sym_if_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_switch_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_for_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_for_in_statement] = {
    .visible = true,
    .named = true,
  },
  [sym__for_header] = {
    .visible = false,
    .named = true,
  },
  [sym_while_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_do_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_try_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_with_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_break_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_continue_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_debugger_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_return_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_throw_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_empty_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_labeled_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_switch_body] = {
    .visible = true,
    .named = true,
  },
  [sym_switch_case] = {
    .visible = true,
    .named = true,
  },
  [sym_switch_default] = {
    .visible = true,
    .named = true,
  },
  [sym_catch_clause] = {
    .visible = true,
    .named = true,
  },
  [sym_finally_clause] = {
    .visible = true,
    .named = true,
  },
  [sym_parenthesized_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_expression] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym_primary_expression] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym_yield_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_object] = {
    .visible = true,
    .named = true,
  },
  [sym_object_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_assignment_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_object_assignment_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_array] = {
    .visible = true,
    .named = true,
  },
  [sym_array_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_jsx_element] = {
    .visible = true,
    .named = true,
  },
  [sym_jsx_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_jsx_opening_element] = {
    .visible = true,
    .named = true,
  },
  [sym_nested_identifier] = {
    .visible = true,
    .named = true,
  },
  [sym_jsx_namespace_name] = {
    .visible = true,
    .named = true,
  },
  [sym_jsx_closing_element] = {
    .visible = true,
    .named = true,
  },
  [sym_jsx_self_closing_element] = {
    .visible = true,
    .named = true,
  },
  [sym_jsx_attribute] = {
    .visible = true,
    .named = true,
  },
  [sym__jsx_string] = {
    .visible = true,
    .named = true,
  },
  [sym_class] = {
    .visible = true,
    .named = true,
  },
  [sym_class_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_class_heritage] = {
    .visible = true,
    .named = true,
  },
  [sym_function_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_function_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_generator_function] = {
    .visible = true,
    .named = true,
  },
  [sym_generator_function_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_arrow_function] = {
    .visible = true,
    .named = true,
  },
  [sym_call_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_new_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_await_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_member_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_subscript_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_assignment_expression] = {
    .visible = true,
    .named = true,
  },
  [sym__augmented_assignment_lhs] = {
    .visible = false,
    .named = true,
  },
  [sym_augmented_assignment_expression] = {
    .visible = true,
    .named = true,
  },
  [sym__initializer] = {
    .visible = false,
    .named = true,
  },
  [sym__destructuring_pattern] = {
    .visible = false,
    .named = true,
  },
  [sym_spread_element] = {
    .visible = true,
    .named = true,
  },
  [sym_ternary_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_binary_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_unary_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_update_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_sequence_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_string] = {
    .visible = true,
    .named = true,
  },
  [sym_template_string] = {
    .visible = true,
    .named = true,
  },
  [sym_template_substitution] = {
    .visible = true,
    .named = true,
  },
  [sym_regex] = {
    .visible = true,
    .named = true,
  },
  [sym_meta_property] = {
    .visible = true,
    .named = true,
  },
  [sym_arguments] = {
    .visible = true,
    .named = true,
  },
  [sym_decorator] = {
    .visible = true,
    .named = true,
  },
  [sym_decorator_member_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_decorator_call_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_class_body] = {
    .visible = true,
    .named = true,
  },
  [sym_field_definition] = {
    .visible = true,
    .named = true,
  },
  [sym_formal_parameters] = {
    .visible = true,
    .named = true,
  },
  [sym_class_static_block] = {
    .visible = true,
    .named = true,
  },
  [sym_pattern] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym_rest_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_method_definition] = {
    .visible = true,
    .named = true,
  },
  [sym_pair] = {
    .visible = true,
    .named = true,
  },
  [sym_pair_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym__property_name] = {
    .visible = false,
    .named = true,
  },
  [sym_computed_property_name] = {
    .visible = true,
    .named = true,
  },
  [aux_sym_program_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_export_statement_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_export_clause_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_named_imports_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_variable_declaration_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_switch_body_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_object_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_object_pattern_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_array_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_array_pattern_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_jsx_element_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_jsx_opening_element_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__jsx_string_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym__jsx_string_repeat2] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_sequence_expression_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_string_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_string_repeat2] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_template_string_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_class_body_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_formal_parameters_repeat1] = {
    .visible = false,
    .named = false,
  },
  [alias_sym_property_identifier] = {
    .visible = true,
    .named = true,
  },
  [alias_sym_shorthand_property_identifier] = {
    .visible = true,
    .named = true,
  },
  [alias_sym_shorthand_property_identifier_pattern] = {
    .visible = true,
    .named = true,
  },
  [alias_sym_statement_identifier] = {
    .visible = true,
    .named = true,
  },
};

enum ts_field_identifiers {
  field_alias = 1,
  field_alternative = 2,
  field_argument = 3,
  field_arguments = 4,
  field_attribute = 5,
  field_body = 6,
  field_close_tag = 7,
  field_condition = 8,
  field_consequence = 9,
  field_constructor = 10,
  field_declaration = 11,
  field_decorator = 12,
  field_finalizer = 13,
  field_flags = 14,
  field_function = 15,
  field_handler = 16,
  field_increment = 17,
  field_index = 18,
  field_initializer = 19,
  field_key = 20,
  field_kind = 21,
  field_label = 22,
  field_left = 23,
  field_member = 24,
  field_name = 25,
  field_object = 26,
  field_open_tag = 27,
  field_operator = 28,
  field_optional_chain = 29,
  field_parameter = 30,
  field_parameters = 31,
  field_pattern = 32,
  field_property = 33,
  field_right = 34,
  field_source = 35,
  field_value = 36,
};

static const char * const ts_field_names[] = {
  [0] = NULL,
  [field_alias] = "alias",
  [field_alternative] = "alternative",
  [field_argument] = "argument",
  [field_arguments] = "arguments",
  [field_attribute] = "attribute",
  [field_body] = "body",
  [field_close_tag] = "close_tag",
  [field_condition] = "condition",
  [field_consequence] = "consequence",
  [field_constructor] = "constructor",
  [field_declaration] = "declaration",
  [field_decorator] = "decorator",
  [field_finalizer] = "finalizer",
  [field_flags] = "flags",
  [field_function] = "function",
  [field_handler] = "handler",
  [field_increment] = "increment",
  [field_index] = "index",
  [field_initializer] = "initializer",
  [field_key] = "key",
  [field_kind] = "kind",
  [field_label] = "label",
  [field_left] = "left",
  [field_member] = "member",
  [field_name] = "name",
  [field_object] = "object",
  [field_open_tag] = "open_tag",
  [field_operator] = "operator",
  [field_optional_chain] = "optional_chain",
  [field_parameter] = "parameter",
  [field_parameters] = "parameters",
  [field_pattern] = "pattern",
  [field_property] = "property",
  [field_right] = "right",
  [field_source] = "source",
  [field_value] = "value",
};

static const TSMapSlice ts_field_map_slices[PRODUCTION_ID_COUNT] = {
  [2] = {.index = 0, .length = 1},
  [3] = {.index = 1, .length = 1},
  [5] = {.index = 2, .length = 1},
  [6] = {.index = 2, .length = 1},
  [7] = {.index = 3, .length = 1},
  [8] = {.index = 4, .length = 1},
  [9] = {.index = 5, .length = 2},
  [10] = {.index = 7, .length = 2},
  [11] = {.index = 9, .length = 2},
  [12] = {.index = 11, .length = 2},
  [13] = {.index = 13, .length = 2},
  [14] = {.index = 15, .length = 1},
  [15] = {.index = 16, .length = 2},
  [16] = {.index = 18, .length = 2},
  [17] = {.index = 20, .length = 2},
  [21] = {.index = 22, .length = 1},
  [22] = {.index = 23, .length = 2},
  [23] = {.index = 25, .length = 2},
  [24] = {.index = 25, .length = 2},
  [25] = {.index = 27, .length = 1},
  [26] = {.index = 28, .length = 2},
  [27] = {.index = 30, .length = 2},
  [28] = {.index = 32, .length = 6},
  [29] = {.index = 38, .length = 2},
  [30] = {.index = 40, .length = 2},
  [31] = {.index = 42, .length = 2},
  [32] = {.index = 44, .length = 1},
  [33] = {.index = 45, .length = 1},
  [34] = {.index = 46, .length = 1},
  [35] = {.index = 47, .length = 1},
  [36] = {.index = 48, .length = 1},
  [37] = {.index = 49, .length = 2},
  [38] = {.index = 51, .length = 1},
  [39] = {.index = 52, .length = 2},
  [40] = {.index = 54, .length = 2},
  [41] = {.index = 56, .length = 1},
  [42] = {.index = 18, .length = 2},
  [43] = {.index = 20, .length = 2},
  [44] = {.index = 57, .length = 3},
  [45] = {.index = 60, .length = 2},
  [46] = {.index = 60, .length = 2},
  [47] = {.index = 62, .length = 3},
  [48] = {.index = 62, .length = 3},
  [49] = {.index = 65, .length = 3},
  [50] = {.index = 68, .length = 2},
  [51] = {.index = 70, .length = 2},
  [52] = {.index = 72, .length = 2},
  [53] = {.index = 74, .length = 2},
  [54] = {.index = 76, .length = 1},
  [55] = {.index = 77, .length = 1},
  [56] = {.index = 18, .length = 2},
  [57] = {.index = 78, .length = 2},
  [58] = {.index = 80, .length = 3},
  [59] = {.index = 83, .length = 1},
  [60] = {.index = 84, .length = 2},
  [61] = {.index = 86, .length = 3},
  [62] = {.index = 89, .length = 6},
  [63] = {.index = 95, .length = 2},
  [64] = {.index = 97, .length = 3},
  [65] = {.index = 100, .length = 2},
  [66] = {.index = 102, .length = 2},
  [67] = {.index = 104, .length = 1},
  [68] = {.index = 105, .length = 2},
  [69] = {.index = 107, .length = 2},
  [70] = {.index = 109, .length = 1},
  [71] = {.index = 110, .length = 2},
  [72] = {.index = 112, .length = 2},
  [73] = {.index = 114, .length = 2},
  [74] = {.index = 116, .length = 2},
  [75] = {.index = 114, .length = 2},
  [76] = {.index = 118, .length = 2},
  [77] = {.index = 120, .length = 3},
  [78] = {.index = 123, .length = 2},
  [79] = {.index = 125, .length = 2},
  [80] = {.index = 127, .length = 2},
  [81] = {.index = 129, .length = 3},
  [82] = {.index = 132, .length = 2},
  [83] = {.index = 134, .length = 2},
  [84] = {.index = 136, .length = 4},
  [85] = {.index = 140, .length = 2},
  [86] = {.index = 142, .length = 2},
  [87] = {.index = 144, .length = 3},
  [88] = {.index = 147, .length = 2},
  [89] = {.index = 149, .length = 3},
  [90] = {.index = 152, .length = 3},
  [91] = {.index = 155, .length = 3},
  [92] = {.index = 158, .length = 2},
  [93] = {.index = 160, .length = 3},
  [94] = {.index = 163, .length = 4},
  [95] = {.index = 167, .length = 3},
  [96] = {.index = 167, .length = 3},
  [97] = {.index = 170, .length = 3},
  [98] = {.index = 173, .length = 3},
  [99] = {.index = 176, .length = 3},
  [100] = {.index = 179, .length = 4},
  [101] = {.index = 183, .length = 2},
  [102] = {.index = 185, .length = 4},
  [103] = {.index = 185, .length = 4},
  [104] = {.index = 189, .length = 4},
  [105] = {.index = 193, .length = 4},
  [106] = {.index = 197, .length = 3},
  [107] = {.index = 200, .length = 2},
  [108] = {.index = 202, .length = 4},
  [109] = {.index = 206, .length = 5},
  [110] = {.index = 206, .length = 5},
  [111] = {.index = 211, .length = 4},
  [112] = {.index = 211, .length = 4},
  [113] = {.index = 215, .length = 5},
  [114] = {.index = 215, .length = 5},
  [115] = {.index = 220, .length = 5},
  [116] = {.index = 225, .length = 4},
  [117] = {.index = 229, .length = 4},
  [118] = {.index = 233, .length = 5},
  [119] = {.index = 233, .length = 5},
  [120] = {.index = 238, .length = 5},
};

static const TSFieldMapEntry ts_field_map_entries[] = {
  [0] =
    {field_decorator, 0},
  [1] =
    {field_declaration, 1},
  [2] =
    {field_name, 0},
  [3] =
    {field_body, 1},
  [4] =
    {field_constructor, 1},
  [5] =
    {field_argument, 1},
    {field_operator, 0},
  [7] =
    {field_arguments, 1},
    {field_function, 0},
  [9] =
    {field_argument, 0},
    {field_operator, 1},
  [11] =
    {field_close_tag, 1},
    {field_open_tag, 0},
  [13] =
    {field_decorator, 0, .inherited = true},
    {field_decorator, 1, .inherited = true},
  [15] =
    {field_declaration, 2},
  [16] =
    {field_body, 2},
    {field_label, 0},
  [18] =
    {field_left, 0},
    {field_right, 2},
  [20] =
    {field_body, 2},
    {field_parameter, 0},
  [22] =
    {field_source, 1},
  [23] =
    {field_body, 2},
    {field_object, 1},
  [25] =
    {field_name, 0},
    {field_value, 1, .inherited = true},
  [27] =
    {field_kind, 0},
  [28] =
    {field_condition, 1},
    {field_consequence, 2},
  [30] =
    {field_body, 2},
    {field_value, 1},
  [32] =
    {field_body, 2},
    {field_kind, 1, .inherited = true},
    {field_left, 1, .inherited = true},
    {field_operator, 1, .inherited = true},
    {field_right, 1, .inherited = true},
    {field_value, 1, .inherited = true},
  [38] =
    {field_body, 2},
    {field_condition, 1},
  [40] =
    {field_body, 1},
    {field_handler, 2},
  [42] =
    {field_body, 1},
    {field_finalizer, 2},
  [44] =
    {field_label, 1},
  [45] =
    {field_name, 1},
  [46] =
    {field_attribute, 0},
  [47] =
    {field_member, 0},
  [48] =
    {field_property, 0},
  [49] =
    {field_body, 2},
    {field_name, 1},
  [51] =
    {field_body, 2},
  [52] =
    {field_body, 2},
    {field_parameters, 1},
  [54] =
    {field_arguments, 2},
    {field_constructor, 1},
  [56] =
    {field_pattern, 1},
  [57] =
    {field_left, 0},
    {field_operator, 1},
    {field_right, 2},
  [60] =
    {field_object, 0},
    {field_property, 2},
  [62] =
    {field_object, 0},
    {field_optional_chain, 1},
    {field_property, 2},
  [65] =
    {field_arguments, 2},
    {field_function, 0},
    {field_optional_chain, 1},
  [68] =
    {field_close_tag, 2},
    {field_open_tag, 0},
  [70] =
    {field_body, 2},
    {field_parameters, 0},
  [72] =
    {field_declaration, 2},
    {field_decorator, 0, .inherited = true},
  [74] =
    {field_body, 2},
    {field_decorator, 0, .inherited = true},
  [76] =
    {field_source, 2, .inherited = true},
  [77] =
    {field_value, 2},
  [78] =
    {field_key, 0},
    {field_value, 2},
  [80] =
    {field_body, 2},
    {field_name, 0},
    {field_parameters, 1},
  [83] =
    {field_value, 1},
  [84] =
    {field_kind, 0},
    {field_kind, 1},
  [86] =
    {field_alternative, 3},
    {field_condition, 1},
    {field_consequence, 2},
  [89] =
    {field_body, 3},
    {field_kind, 2, .inherited = true},
    {field_left, 2, .inherited = true},
    {field_operator, 2, .inherited = true},
    {field_right, 2, .inherited = true},
    {field_value, 2, .inherited = true},
  [95] =
    {field_body, 1},
    {field_condition, 3},
  [97] =
    {field_body, 1},
    {field_finalizer, 3},
    {field_handler, 2},
  [100] =
    {field_attribute, 2, .inherited = true},
    {field_name, 1},
  [102] =
    {field_attribute, 0, .inherited = true},
    {field_attribute, 1, .inherited = true},
  [104] =
    {field_property, 1},
  [105] =
    {field_property, 0},
    {field_value, 1, .inherited = true},
  [107] =
    {field_decorator, 0, .inherited = true},
    {field_property, 1},
  [109] =
    {field_member, 1, .inherited = true},
  [110] =
    {field_member, 0, .inherited = true},
    {field_member, 1, .inherited = true},
  [112] =
    {field_body, 3},
    {field_name, 1},
  [114] =
    {field_body, 3},
    {field_parameter, 1},
  [116] =
    {field_body, 3},
    {field_parameters, 2},
  [118] =
    {field_body, 3},
    {field_parameters, 1},
  [120] =
    {field_body, 3},
    {field_name, 1},
    {field_parameters, 2},
  [123] =
    {field_flags, 3},
    {field_pattern, 1},
  [125] =
    {field_index, 2},
    {field_object, 0},
  [127] =
    {field_declaration, 3},
    {field_decorator, 0, .inherited = true},
  [129] =
    {field_body, 3},
    {field_decorator, 0, .inherited = true},
    {field_name, 2},
  [132] =
    {field_body, 3},
    {field_decorator, 0, .inherited = true},
  [134] =
    {field_alias, 2},
    {field_name, 0},
  [136] =
    {field_body, 3},
    {field_decorator, 0, .inherited = true},
    {field_name, 1},
    {field_parameters, 2},
  [140] =
    {field_property, 1},
    {field_value, 2, .inherited = true},
  [142] =
    {field_decorator, 0, .inherited = true},
    {field_property, 2},
  [144] =
    {field_decorator, 0, .inherited = true},
    {field_property, 1},
    {field_value, 2, .inherited = true},
  [147] =
    {field_body, 4},
    {field_parameters, 3},
  [149] =
    {field_body, 4},
    {field_name, 2},
    {field_parameters, 3},
  [152] =
    {field_alternative, 4},
    {field_condition, 0},
    {field_consequence, 2},
  [155] =
    {field_index, 3},
    {field_object, 0},
    {field_optional_chain, 1},
  [158] =
    {field_decorator, 0, .inherited = true},
    {field_value, 3},
  [160] =
    {field_body, 4},
    {field_decorator, 0, .inherited = true},
    {field_name, 2},
  [163] =
    {field_body, 4},
    {field_decorator, 0, .inherited = true},
    {field_name, 2},
    {field_parameters, 3},
  [167] =
    {field_left, 1},
    {field_operator, 2},
    {field_right, 3},
  [170] =
    {field_body, 5},
    {field_condition, 3},
    {field_initializer, 2},
  [173] =
    {field_decorator, 0, .inherited = true},
    {field_property, 2},
    {field_value, 3, .inherited = true},
  [176] =
    {field_body, 5},
    {field_name, 3},
    {field_parameters, 4},
  [179] =
    {field_body, 5},
    {field_decorator, 0, .inherited = true},
    {field_name, 3},
    {field_parameters, 4},
  [183] =
    {field_body, 3},
    {field_value, 1},
  [185] =
    {field_kind, 1},
    {field_left, 2},
    {field_operator, 3},
    {field_right, 4},
  [189] =
    {field_body, 6},
    {field_condition, 3},
    {field_increment, 4},
    {field_initializer, 2},
  [193] =
    {field_body, 6},
    {field_condition, 3},
    {field_condition, 4},
    {field_initializer, 2},
  [197] =
    {field_body, 6},
    {field_condition, 4},
    {field_initializer, 2},
  [200] =
    {field_body, 4},
    {field_parameter, 2},
  [202] =
    {field_body, 6},
    {field_decorator, 0, .inherited = true},
    {field_name, 4},
    {field_parameters, 5},
  [206] =
    {field_kind, 1},
    {field_left, 2},
    {field_operator, 4},
    {field_right, 5},
    {field_value, 3, .inherited = true},
  [211] =
    {field_kind, 1},
    {field_left, 2},
    {field_operator, 4},
    {field_right, 5},
  [215] =
    {field_kind, 1},
    {field_kind, 2},
    {field_left, 3},
    {field_operator, 4},
    {field_right, 5},
  [220] =
    {field_body, 7},
    {field_condition, 3},
    {field_condition, 4},
    {field_increment, 5},
    {field_initializer, 2},
  [225] =
    {field_body, 7},
    {field_condition, 4},
    {field_increment, 5},
    {field_initializer, 2},
  [229] =
    {field_body, 7},
    {field_condition, 4},
    {field_condition, 5},
    {field_initializer, 2},
  [233] =
    {field_kind, 1},
    {field_kind, 2},
    {field_left, 3},
    {field_operator, 5},
    {field_right, 6},
  [238] =
    {field_body, 8},
    {field_condition, 4},
    {field_condition, 5},
    {field_increment, 6},
    {field_initializer, 2},
};

static const TSSymbol ts_alias_sequences[PRODUCTION_ID_COUNT][MAX_ALIAS_SEQUENCE_LENGTH] = {
  [0] = {0},
  [1] = {
    [0] = sym_identifier,
  },
  [4] = {
    [0] = alias_sym_property_identifier,
  },
  [5] = {
    [0] = sym_identifier,
  },
  [15] = {
    [0] = alias_sym_statement_identifier,
  },
  [16] = {
    [0] = sym_identifier,
  },
  [17] = {
    [0] = sym_identifier,
  },
  [18] = {
    [1] = alias_sym_shorthand_property_identifier,
  },
  [19] = {
    [1] = alias_sym_shorthand_property_identifier_pattern,
  },
  [20] = {
    [1] = sym_identifier,
  },
  [23] = {
    [0] = sym_identifier,
  },
  [32] = {
    [1] = alias_sym_statement_identifier,
  },
  [45] = {
    [2] = alias_sym_property_identifier,
  },
  [47] = {
    [2] = alias_sym_property_identifier,
  },
  [56] = {
    [0] = alias_sym_shorthand_property_identifier_pattern,
  },
  [73] = {
    [1] = sym_identifier,
  },
  [95] = {
    [1] = sym_identifier,
  },
  [102] = {
    [2] = sym_identifier,
  },
  [109] = {
    [2] = sym_identifier,
  },
  [111] = {
    [2] = sym_identifier,
  },
  [113] = {
    [3] = sym_identifier,
  },
  [118] = {
    [3] = sym_identifier,
  },
};

static const uint16_t ts_non_terminal_alias_map[] = {
  0,
};

static const TSStateId ts_primary_state_ids[STATE_COUNT] = {
  [0] = 0,
  [1] = 1,
  [2] = 2,
  [3] = 2,
  [4] = 2,
  [5] = 2,
  [6] = 2,
  [7] = 7,
  [8] = 7,
  [9] = 9,
  [10] = 7,
  [11] = 11,
  [12] = 12,
  [13] = 13,
  [14] = 14,
  [15] = 15,
  [16] = 15,
  [17] = 17,
  [18] = 18,
  [19] = 19,
  [20] = 20,
  [21] = 21,
  [22] = 18,
  [23] = 19,
  [24] = 18,
  [25] = 19,
  [26] = 18,
  [27] = 19,
  [28] = 19,
  [29] = 29,
  [30] = 18,
  [31] = 29,
  [32] = 32,
  [33] = 15,
  [34] = 34,
  [35] = 15,
  [36] = 36,
  [37] = 15,
  [38] = 36,
  [39] = 15,
  [40] = 40,
  [41] = 41,
  [42] = 42,
  [43] = 43,
  [44] = 44,
  [45] = 45,
  [46] = 46,
  [47] = 47,
  [48] = 48,
  [49] = 49,
  [50] = 50,
  [51] = 32,
  [52] = 52,
  [53] = 53,
  [54] = 54,
  [55] = 55,
  [56] = 56,
  [57] = 40,
  [58] = 53,
  [59] = 42,
  [60] = 43,
  [61] = 44,
  [62] = 45,
  [63] = 46,
  [64] = 47,
  [65] = 48,
  [66] = 49,
  [67] = 50,
  [68] = 52,
  [69] = 55,
  [70] = 56,
  [71] = 54,
  [72] = 72,
  [73] = 34,
  [74] = 36,
  [75] = 15,
  [76] = 15,
  [77] = 15,
  [78] = 15,
  [79] = 15,
  [80] = 80,
  [81] = 80,
  [82] = 80,
  [83] = 80,
  [84] = 80,
  [85] = 80,
  [86] = 86,
  [87] = 87,
  [88] = 88,
  [89] = 89,
  [90] = 90,
  [91] = 91,
  [92] = 92,
  [93] = 93,
  [94] = 94,
  [95] = 95,
  [96] = 96,
  [97] = 97,
  [98] = 98,
  [99] = 99,
  [100] = 100,
  [101] = 101,
  [102] = 102,
  [103] = 102,
  [104] = 101,
  [105] = 105,
  [106] = 105,
  [107] = 107,
  [108] = 105,
  [109] = 105,
  [110] = 105,
  [111] = 105,
  [112] = 112,
  [113] = 113,
  [114] = 112,
  [115] = 115,
  [116] = 116,
  [117] = 117,
  [118] = 115,
  [119] = 119,
  [120] = 119,
  [121] = 121,
  [122] = 122,
  [123] = 122,
  [124] = 124,
  [125] = 122,
  [126] = 126,
  [127] = 127,
  [128] = 128,
  [129] = 129,
  [130] = 130,
  [131] = 129,
  [132] = 132,
  [133] = 133,
  [134] = 133,
  [135] = 135,
  [136] = 136,
  [137] = 133,
  [138] = 138,
  [139] = 138,
  [140] = 140,
  [141] = 141,
  [142] = 142,
  [143] = 141,
  [144] = 144,
  [145] = 145,
  [146] = 146,
  [147] = 144,
  [148] = 145,
  [149] = 146,
  [150] = 150,
  [151] = 150,
  [152] = 152,
  [153] = 152,
  [154] = 154,
  [155] = 155,
  [156] = 156,
  [157] = 157,
  [158] = 158,
  [159] = 159,
  [160] = 160,
  [161] = 161,
  [162] = 162,
  [163] = 163,
  [164] = 159,
  [165] = 157,
  [166] = 166,
  [167] = 167,
  [168] = 168,
  [169] = 167,
  [170] = 170,
  [171] = 171,
  [172] = 166,
  [173] = 168,
  [174] = 174,
  [175] = 175,
  [176] = 159,
  [177] = 152,
  [178] = 178,
  [179] = 154,
  [180] = 168,
  [181] = 181,
  [182] = 167,
  [183] = 183,
  [184] = 175,
  [185] = 152,
  [186] = 154,
  [187] = 161,
  [188] = 154,
  [189] = 181,
  [190] = 190,
  [191] = 191,
  [192] = 157,
  [193] = 193,
  [194] = 159,
  [195] = 178,
  [196] = 168,
  [197] = 167,
  [198] = 175,
  [199] = 152,
  [200] = 154,
  [201] = 201,
  [202] = 202,
  [203] = 159,
  [204] = 168,
  [205] = 167,
  [206] = 175,
  [207] = 152,
  [208] = 154,
  [209] = 159,
  [210] = 157,
  [211] = 211,
  [212] = 168,
  [213] = 157,
  [214] = 167,
  [215] = 175,
  [216] = 157,
  [217] = 175,
  [218] = 218,
  [219] = 219,
  [220] = 220,
  [221] = 221,
  [222] = 222,
  [223] = 223,
  [224] = 220,
  [225] = 225,
  [226] = 226,
  [227] = 227,
  [228] = 228,
  [229] = 229,
  [230] = 230,
  [231] = 231,
  [232] = 232,
  [233] = 233,
  [234] = 234,
  [235] = 235,
  [236] = 236,
  [237] = 237,
  [238] = 238,
  [239] = 239,
  [240] = 240,
  [241] = 241,
  [242] = 242,
  [243] = 243,
  [244] = 220,
  [245] = 229,
  [246] = 225,
  [247] = 247,
  [248] = 228,
  [249] = 231,
  [250] = 219,
  [251] = 233,
  [252] = 234,
  [253] = 235,
  [254] = 236,
  [255] = 237,
  [256] = 238,
  [257] = 239,
  [258] = 240,
  [259] = 241,
  [260] = 242,
  [261] = 243,
  [262] = 247,
  [263] = 263,
  [264] = 263,
  [265] = 265,
  [266] = 266,
  [267] = 226,
  [268] = 268,
  [269] = 269,
  [270] = 270,
  [271] = 271,
  [272] = 266,
  [273] = 266,
  [274] = 220,
  [275] = 229,
  [276] = 225,
  [277] = 232,
  [278] = 231,
  [279] = 219,
  [280] = 233,
  [281] = 235,
  [282] = 236,
  [283] = 237,
  [284] = 238,
  [285] = 239,
  [286] = 240,
  [287] = 241,
  [288] = 242,
  [289] = 243,
  [290] = 247,
  [291] = 263,
  [292] = 226,
  [293] = 293,
  [294] = 229,
  [295] = 225,
  [296] = 228,
  [297] = 231,
  [298] = 219,
  [299] = 232,
  [300] = 233,
  [301] = 235,
  [302] = 236,
  [303] = 237,
  [304] = 238,
  [305] = 239,
  [306] = 240,
  [307] = 241,
  [308] = 242,
  [309] = 243,
  [310] = 247,
  [311] = 263,
  [312] = 269,
  [313] = 226,
  [314] = 314,
  [315] = 220,
  [316] = 229,
  [317] = 225,
  [318] = 228,
  [319] = 231,
  [320] = 219,
  [321] = 233,
  [322] = 235,
  [323] = 236,
  [324] = 237,
  [325] = 238,
  [326] = 239,
  [327] = 240,
  [328] = 241,
  [329] = 242,
  [330] = 243,
  [331] = 247,
  [332] = 263,
  [333] = 226,
  [334] = 232,
  [335] = 220,
  [336] = 229,
  [337] = 225,
  [338] = 228,
  [339] = 231,
  [340] = 219,
  [341] = 233,
  [342] = 235,
  [343] = 236,
  [344] = 237,
  [345] = 238,
  [346] = 239,
  [347] = 240,
  [348] = 241,
  [349] = 242,
  [350] = 243,
  [351] = 247,
  [352] = 263,
  [353] = 226,
  [354] = 232,
  [355] = 232,
  [356] = 228,
  [357] = 357,
  [358] = 357,
  [359] = 357,
  [360] = 360,
  [361] = 361,
  [362] = 360,
  [363] = 361,
  [364] = 361,
  [365] = 360,
  [366] = 366,
  [367] = 366,
  [368] = 368,
  [369] = 88,
  [370] = 368,
  [371] = 371,
  [372] = 366,
  [373] = 366,
  [374] = 91,
  [375] = 366,
  [376] = 371,
  [377] = 366,
  [378] = 378,
  [379] = 378,
  [380] = 368,
  [381] = 381,
  [382] = 382,
  [383] = 383,
  [384] = 89,
  [385] = 87,
  [386] = 382,
  [387] = 378,
  [388] = 388,
  [389] = 389,
  [390] = 388,
  [391] = 87,
  [392] = 392,
  [393] = 366,
  [394] = 91,
  [395] = 395,
  [396] = 396,
  [397] = 396,
  [398] = 398,
  [399] = 90,
  [400] = 89,
  [401] = 401,
  [402] = 86,
  [403] = 403,
  [404] = 404,
  [405] = 405,
  [406] = 366,
  [407] = 407,
  [408] = 408,
  [409] = 409,
  [410] = 366,
  [411] = 411,
  [412] = 366,
  [413] = 413,
  [414] = 366,
  [415] = 415,
  [416] = 88,
  [417] = 383,
  [418] = 418,
  [419] = 419,
  [420] = 420,
  [421] = 421,
  [422] = 422,
  [423] = 423,
  [424] = 424,
  [425] = 425,
  [426] = 426,
  [427] = 427,
  [428] = 428,
  [429] = 429,
  [430] = 430,
  [431] = 431,
  [432] = 432,
  [433] = 433,
  [434] = 434,
  [435] = 435,
  [436] = 436,
  [437] = 437,
  [438] = 438,
  [439] = 439,
  [440] = 440,
  [441] = 441,
  [442] = 442,
  [443] = 443,
  [444] = 444,
  [445] = 445,
  [446] = 446,
  [447] = 447,
  [448] = 448,
  [449] = 449,
  [450] = 450,
  [451] = 451,
  [452] = 452,
  [453] = 453,
  [454] = 454,
  [455] = 455,
  [456] = 456,
  [457] = 457,
  [458] = 458,
  [459] = 459,
  [460] = 460,
  [461] = 461,
  [462] = 462,
  [463] = 463,
  [464] = 464,
  [465] = 465,
  [466] = 466,
  [467] = 467,
  [468] = 468,
  [469] = 469,
  [470] = 470,
  [471] = 471,
  [472] = 472,
  [473] = 473,
  [474] = 474,
  [475] = 475,
  [476] = 476,
  [477] = 477,
  [478] = 478,
  [479] = 479,
  [480] = 480,
  [481] = 481,
  [482] = 482,
  [483] = 483,
  [484] = 484,
  [485] = 485,
  [486] = 486,
  [487] = 487,
  [488] = 488,
  [489] = 489,
  [490] = 490,
  [491] = 404,
  [492] = 490,
  [493] = 493,
  [494] = 494,
  [495] = 495,
  [496] = 496,
  [497] = 497,
  [498] = 498,
  [499] = 490,
  [500] = 500,
  [501] = 501,
  [502] = 502,
  [503] = 501,
  [504] = 502,
  [505] = 505,
  [506] = 506,
  [507] = 498,
  [508] = 508,
  [509] = 509,
  [510] = 510,
  [511] = 511,
  [512] = 512,
  [513] = 513,
  [514] = 514,
  [515] = 515,
  [516] = 516,
  [517] = 517,
  [518] = 518,
  [519] = 519,
  [520] = 520,
  [521] = 521,
  [522] = 522,
  [523] = 523,
  [524] = 404,
  [525] = 525,
  [526] = 497,
  [527] = 501,
  [528] = 498,
  [529] = 501,
  [530] = 502,
  [531] = 531,
  [532] = 532,
  [533] = 533,
  [534] = 500,
  [535] = 535,
  [536] = 533,
  [537] = 502,
  [538] = 404,
  [539] = 539,
  [540] = 540,
  [541] = 531,
  [542] = 535,
  [543] = 501,
  [544] = 502,
  [545] = 494,
  [546] = 502,
  [547] = 495,
  [548] = 496,
  [549] = 532,
  [550] = 501,
  [551] = 493,
  [552] = 498,
  [553] = 553,
  [554] = 540,
  [555] = 555,
  [556] = 556,
  [557] = 557,
  [558] = 557,
  [559] = 539,
  [560] = 498,
  [561] = 532,
  [562] = 556,
  [563] = 533,
  [564] = 564,
  [565] = 502,
  [566] = 501,
  [567] = 502,
  [568] = 501,
  [569] = 502,
  [570] = 501,
  [571] = 502,
  [572] = 557,
  [573] = 502,
  [574] = 553,
  [575] = 555,
  [576] = 501,
  [577] = 501,
  [578] = 498,
  [579] = 498,
  [580] = 564,
  [581] = 581,
  [582] = 92,
  [583] = 583,
  [584] = 91,
  [585] = 585,
  [586] = 88,
  [587] = 587,
  [588] = 588,
  [589] = 589,
  [590] = 590,
  [591] = 591,
  [592] = 592,
  [593] = 593,
  [594] = 90,
  [595] = 595,
  [596] = 596,
  [597] = 597,
  [598] = 598,
  [599] = 599,
  [600] = 600,
  [601] = 601,
  [602] = 602,
  [603] = 603,
  [604] = 604,
  [605] = 605,
  [606] = 606,
  [607] = 607,
  [608] = 608,
  [609] = 609,
  [610] = 610,
  [611] = 611,
  [612] = 612,
  [613] = 613,
  [614] = 614,
  [615] = 615,
  [616] = 616,
  [617] = 617,
  [618] = 89,
  [619] = 619,
  [620] = 620,
  [621] = 621,
  [622] = 622,
  [623] = 623,
  [624] = 624,
  [625] = 86,
  [626] = 626,
  [627] = 627,
  [628] = 628,
  [629] = 629,
  [630] = 630,
  [631] = 631,
  [632] = 632,
  [633] = 633,
  [634] = 634,
  [635] = 635,
  [636] = 636,
  [637] = 637,
  [638] = 638,
  [639] = 639,
  [640] = 640,
  [641] = 641,
  [642] = 642,
  [643] = 643,
  [644] = 644,
  [645] = 645,
  [646] = 646,
  [647] = 87,
  [648] = 648,
  [649] = 649,
  [650] = 650,
  [651] = 651,
  [652] = 632,
  [653] = 653,
  [654] = 654,
  [655] = 655,
  [656] = 656,
  [657] = 581,
  [658] = 658,
  [659] = 590,
  [660] = 660,
  [661] = 661,
  [662] = 662,
  [663] = 663,
  [664] = 664,
  [665] = 648,
  [666] = 666,
  [667] = 667,
  [668] = 668,
  [669] = 669,
  [670] = 603,
  [671] = 671,
  [672] = 672,
  [673] = 673,
  [674] = 674,
  [675] = 675,
  [676] = 676,
  [677] = 677,
  [678] = 678,
  [679] = 679,
  [680] = 680,
  [681] = 583,
  [682] = 585,
  [683] = 603,
  [684] = 92,
  [685] = 632,
  [686] = 589,
  [687] = 614,
  [688] = 590,
  [689] = 640,
  [690] = 690,
  [691] = 591,
  [692] = 587,
  [693] = 92,
  [694] = 648,
  [695] = 588,
  [696] = 660,
  [697] = 617,
  [698] = 653,
  [699] = 596,
  [700] = 597,
  [701] = 598,
  [702] = 94,
  [703] = 661,
  [704] = 662,
  [705] = 669,
  [706] = 599,
  [707] = 671,
  [708] = 672,
  [709] = 673,
  [710] = 674,
  [711] = 675,
  [712] = 676,
  [713] = 677,
  [714] = 678,
  [715] = 679,
  [716] = 680,
  [717] = 667,
  [718] = 654,
  [719] = 660,
  [720] = 600,
  [721] = 664,
  [722] = 666,
  [723] = 668,
  [724] = 93,
  [725] = 661,
  [726] = 662,
  [727] = 655,
  [728] = 601,
  [729] = 658,
  [730] = 603,
  [731] = 632,
  [732] = 604,
  [733] = 648,
  [734] = 616,
  [735] = 651,
  [736] = 636,
  [737] = 630,
  [738] = 621,
  [739] = 635,
  [740] = 646,
  [741] = 639,
  [742] = 629,
  [743] = 644,
  [744] = 606,
  [745] = 608,
  [746] = 602,
  [747] = 634,
  [748] = 615,
  [749] = 669,
  [750] = 642,
  [751] = 671,
  [752] = 95,
  [753] = 672,
  [754] = 645,
  [755] = 650,
  [756] = 673,
  [757] = 622,
  [758] = 607,
  [759] = 626,
  [760] = 674,
  [761] = 761,
  [762] = 653,
  [763] = 675,
  [764] = 631,
  [765] = 765,
  [766] = 619,
  [767] = 649,
  [768] = 98,
  [769] = 655,
  [770] = 593,
  [771] = 590,
  [772] = 620,
  [773] = 605,
  [774] = 96,
  [775] = 676,
  [776] = 677,
  [777] = 678,
  [778] = 679,
  [779] = 627,
  [780] = 680,
  [781] = 623,
  [782] = 99,
  [783] = 595,
  [784] = 609,
  [785] = 624,
  [786] = 612,
  [787] = 656,
  [788] = 610,
  [789] = 641,
  [790] = 611,
  [791] = 667,
  [792] = 664,
  [793] = 590,
  [794] = 628,
  [795] = 598,
  [796] = 663,
  [797] = 637,
  [798] = 97,
  [799] = 100,
  [800] = 638,
  [801] = 613,
  [802] = 663,
  [803] = 654,
  [804] = 765,
  [805] = 592,
  [806] = 666,
  [807] = 643,
  [808] = 656,
  [809] = 658,
  [810] = 633,
  [811] = 668,
  [812] = 648,
  [813] = 813,
  [814] = 814,
  [815] = 815,
  [816] = 816,
  [817] = 603,
  [818] = 818,
  [819] = 761,
  [820] = 632,
  [821] = 821,
  [822] = 822,
  [823] = 823,
  [824] = 824,
  [825] = 825,
  [826] = 826,
  [827] = 827,
  [828] = 828,
  [829] = 829,
  [830] = 830,
  [831] = 831,
  [832] = 832,
  [833] = 833,
  [834] = 834,
  [835] = 835,
  [836] = 836,
  [837] = 837,
  [838] = 838,
  [839] = 839,
  [840] = 840,
  [841] = 841,
  [842] = 842,
  [843] = 843,
  [844] = 844,
  [845] = 845,
  [846] = 846,
  [847] = 847,
  [848] = 690,
  [849] = 849,
  [850] = 850,
  [851] = 765,
  [852] = 603,
  [853] = 632,
  [854] = 648,
  [855] = 826,
  [856] = 828,
  [857] = 857,
  [858] = 841,
  [859] = 847,
  [860] = 849,
  [861] = 850,
  [862] = 862,
  [863] = 863,
  [864] = 842,
  [865] = 828,
  [866] = 857,
  [867] = 847,
  [868] = 828,
  [869] = 857,
  [870] = 828,
  [871] = 828,
  [872] = 872,
  [873] = 862,
  [874] = 874,
  [875] = 857,
  [876] = 831,
  [877] = 832,
  [878] = 840,
  [879] = 879,
  [880] = 880,
  [881] = 881,
  [882] = 874,
  [883] = 883,
  [884] = 863,
  [885] = 883,
  [886] = 886,
  [887] = 660,
  [888] = 598,
  [889] = 889,
  [890] = 656,
  [891] = 658,
  [892] = 653,
  [893] = 660,
  [894] = 661,
  [895] = 662,
  [896] = 669,
  [897] = 671,
  [898] = 672,
  [899] = 673,
  [900] = 674,
  [901] = 675,
  [902] = 676,
  [903] = 677,
  [904] = 678,
  [905] = 679,
  [906] = 680,
  [907] = 667,
  [908] = 654,
  [909] = 664,
  [910] = 666,
  [911] = 668,
  [912] = 655,
  [913] = 663,
  [914] = 914,
  [915] = 915,
  [916] = 916,
  [917] = 590,
  [918] = 663,
  [919] = 656,
  [920] = 658,
  [921] = 653,
  [922] = 590,
  [923] = 662,
  [924] = 669,
  [925] = 671,
  [926] = 672,
  [927] = 673,
  [928] = 674,
  [929] = 675,
  [930] = 676,
  [931] = 677,
  [932] = 678,
  [933] = 679,
  [934] = 680,
  [935] = 667,
  [936] = 654,
  [937] = 664,
  [938] = 666,
  [939] = 668,
  [940] = 655,
  [941] = 941,
  [942] = 656,
  [943] = 658,
  [944] = 653,
  [945] = 660,
  [946] = 661,
  [947] = 662,
  [948] = 669,
  [949] = 671,
  [950] = 672,
  [951] = 673,
  [952] = 674,
  [953] = 675,
  [954] = 676,
  [955] = 677,
  [956] = 678,
  [957] = 679,
  [958] = 680,
  [959] = 667,
  [960] = 654,
  [961] = 664,
  [962] = 666,
  [963] = 668,
  [964] = 655,
  [965] = 827,
  [966] = 663,
  [967] = 661,
  [968] = 968,
  [969] = 969,
  [970] = 968,
  [971] = 971,
  [972] = 968,
  [973] = 968,
  [974] = 968,
  [975] = 975,
  [976] = 975,
  [977] = 968,
  [978] = 826,
  [979] = 979,
  [980] = 979,
  [981] = 981,
  [982] = 982,
  [983] = 983,
  [984] = 982,
  [985] = 985,
  [986] = 986,
  [987] = 986,
  [988] = 982,
  [989] = 989,
  [990] = 983,
  [991] = 983,
  [992] = 992,
  [993] = 993,
  [994] = 994,
  [995] = 993,
  [996] = 994,
  [997] = 993,
  [998] = 994,
  [999] = 993,
  [1000] = 994,
  [1001] = 994,
  [1002] = 993,
  [1003] = 993,
  [1004] = 994,
  [1005] = 1005,
  [1006] = 1005,
  [1007] = 1005,
  [1008] = 1005,
  [1009] = 1009,
  [1010] = 1005,
  [1011] = 1005,
  [1012] = 1012,
  [1013] = 1013,
  [1014] = 1014,
  [1015] = 1015,
  [1016] = 1016,
  [1017] = 1017,
  [1018] = 1018,
  [1019] = 1019,
  [1020] = 1020,
  [1021] = 1021,
  [1022] = 1022,
  [1023] = 1023,
  [1024] = 1024,
  [1025] = 1025,
  [1026] = 1026,
  [1027] = 1027,
  [1028] = 1028,
  [1029] = 1029,
  [1030] = 1030,
  [1031] = 1031,
  [1032] = 1032,
  [1033] = 1033,
  [1034] = 1034,
  [1035] = 1035,
  [1036] = 1036,
  [1037] = 1037,
  [1038] = 1038,
  [1039] = 1039,
  [1040] = 1040,
  [1041] = 1041,
  [1042] = 88,
  [1043] = 1043,
  [1044] = 1044,
  [1045] = 1045,
  [1046] = 1046,
  [1047] = 91,
  [1048] = 1048,
  [1049] = 1049,
  [1050] = 1050,
  [1051] = 1051,
  [1052] = 1052,
  [1053] = 1053,
  [1054] = 1054,
  [1055] = 1055,
  [1056] = 1056,
  [1057] = 1057,
  [1058] = 1058,
  [1059] = 1059,
  [1060] = 1060,
  [1061] = 1061,
  [1062] = 1062,
  [1063] = 1063,
  [1064] = 1064,
  [1065] = 1065,
  [1066] = 1066,
  [1067] = 1038,
  [1068] = 1068,
  [1069] = 1069,
  [1070] = 1070,
  [1071] = 1071,
  [1072] = 1072,
  [1073] = 1073,
  [1074] = 1074,
  [1075] = 87,
  [1076] = 1018,
  [1077] = 1077,
  [1078] = 1078,
  [1079] = 1017,
  [1080] = 1080,
  [1081] = 1081,
  [1082] = 1026,
  [1083] = 1026,
  [1084] = 1072,
  [1085] = 89,
  [1086] = 1086,
  [1087] = 1086,
  [1088] = 1030,
  [1089] = 1089,
  [1090] = 1032,
  [1091] = 1033,
  [1092] = 1092,
  [1093] = 1093,
  [1094] = 1094,
  [1095] = 1095,
  [1096] = 1096,
  [1097] = 1097,
  [1098] = 1098,
  [1099] = 1029,
  [1100] = 634,
  [1101] = 1101,
  [1102] = 1031,
  [1103] = 1103,
  [1104] = 1029,
  [1105] = 1089,
  [1106] = 1106,
  [1107] = 1107,
  [1108] = 1103,
  [1109] = 1098,
  [1110] = 1101,
  [1111] = 1111,
  [1112] = 1111,
  [1113] = 1092,
  [1114] = 1093,
  [1115] = 1095,
  [1116] = 1096,
  [1117] = 1107,
  [1118] = 601,
  [1119] = 638,
  [1120] = 1120,
  [1121] = 1121,
  [1122] = 1122,
  [1123] = 1123,
  [1124] = 1120,
  [1125] = 1125,
  [1126] = 1126,
  [1127] = 1127,
  [1128] = 1128,
  [1129] = 1123,
  [1130] = 1125,
  [1131] = 1122,
  [1132] = 1132,
  [1133] = 1126,
  [1134] = 1121,
  [1135] = 1127,
  [1136] = 1128,
  [1137] = 1132,
  [1138] = 1138,
  [1139] = 1138,
  [1140] = 1138,
  [1141] = 1138,
  [1142] = 1138,
  [1143] = 1138,
  [1144] = 1144,
  [1145] = 1145,
  [1146] = 1146,
  [1147] = 1145,
  [1148] = 1146,
  [1149] = 1149,
  [1150] = 1149,
  [1151] = 1145,
  [1152] = 1145,
  [1153] = 1149,
  [1154] = 1146,
  [1155] = 1149,
  [1156] = 1146,
  [1157] = 1157,
  [1158] = 1158,
  [1159] = 1158,
  [1160] = 1160,
  [1161] = 1157,
  [1162] = 1158,
  [1163] = 1158,
  [1164] = 1157,
  [1165] = 1157,
  [1166] = 1166,
  [1167] = 1167,
  [1168] = 1168,
  [1169] = 1169,
  [1170] = 1170,
  [1171] = 1171,
  [1172] = 1172,
  [1173] = 1173,
  [1174] = 1174,
  [1175] = 1173,
  [1176] = 1174,
  [1177] = 1177,
  [1178] = 1167,
  [1179] = 1173,
  [1180] = 1166,
  [1181] = 1173,
  [1182] = 1177,
  [1183] = 1177,
  [1184] = 1167,
  [1185] = 1177,
  [1186] = 1167,
  [1187] = 1166,
  [1188] = 1166,
  [1189] = 1174,
  [1190] = 1174,
  [1191] = 1191,
  [1192] = 1192,
  [1193] = 1193,
  [1194] = 1194,
  [1195] = 1195,
  [1196] = 1196,
  [1197] = 1197,
  [1198] = 1198,
  [1199] = 1199,
  [1200] = 1200,
  [1201] = 1201,
  [1202] = 1202,
  [1203] = 1203,
  [1204] = 1204,
  [1205] = 1205,
  [1206] = 1206,
  [1207] = 1207,
  [1208] = 1208,
  [1209] = 1209,
  [1210] = 1210,
  [1211] = 1210,
  [1212] = 1210,
  [1213] = 1213,
  [1214] = 1210,
  [1215] = 1215,
  [1216] = 1216,
  [1217] = 1210,
  [1218] = 1218,
  [1219] = 1219,
  [1220] = 1210,
  [1221] = 1221,
  [1222] = 1222,
  [1223] = 1223,
  [1224] = 1224,
  [1225] = 1225,
  [1226] = 1226,
  [1227] = 1227,
  [1228] = 1228,
  [1229] = 1229,
  [1230] = 1230,
  [1231] = 1231,
  [1232] = 1232,
  [1233] = 1233,
  [1234] = 1222,
  [1235] = 640,
  [1236] = 1236,
  [1237] = 614,
  [1238] = 1206,
  [1239] = 1239,
  [1240] = 1240,
  [1241] = 1241,
  [1242] = 1229,
  [1243] = 1243,
  [1244] = 1240,
  [1245] = 1245,
  [1246] = 1246,
  [1247] = 1218,
  [1248] = 1207,
  [1249] = 1219,
  [1250] = 1250,
  [1251] = 1251,
  [1252] = 1252,
  [1253] = 1253,
  [1254] = 1254,
  [1255] = 608,
  [1256] = 593,
  [1257] = 1257,
  [1258] = 610,
  [1259] = 611,
  [1260] = 1260,
  [1261] = 1261,
  [1262] = 1262,
  [1263] = 621,
  [1264] = 622,
  [1265] = 623,
  [1266] = 1266,
  [1267] = 1267,
  [1268] = 1261,
  [1269] = 624,
  [1270] = 592,
  [1271] = 1271,
  [1272] = 1272,
  [1273] = 1273,
  [1274] = 614,
  [1275] = 593,
  [1276] = 1276,
  [1277] = 1277,
  [1278] = 1278,
  [1279] = 1279,
  [1280] = 1280,
  [1281] = 610,
  [1282] = 1282,
  [1283] = 1283,
  [1284] = 1284,
  [1285] = 1285,
  [1286] = 1284,
  [1287] = 1287,
  [1288] = 1260,
  [1289] = 1273,
  [1290] = 611,
  [1291] = 1291,
  [1292] = 1292,
  [1293] = 1293,
  [1294] = 621,
  [1295] = 1279,
  [1296] = 1262,
  [1297] = 1297,
  [1298] = 1271,
  [1299] = 622,
  [1300] = 623,
  [1301] = 1301,
  [1302] = 624,
  [1303] = 1260,
  [1304] = 1304,
  [1305] = 1305,
  [1306] = 1267,
  [1307] = 1282,
  [1308] = 592,
  [1309] = 1309,
  [1310] = 1293,
  [1311] = 1311,
  [1312] = 608,
  [1313] = 1279,
  [1314] = 1266,
  [1315] = 1279,
  [1316] = 640,
  [1317] = 1317,
  [1318] = 1260,
  [1319] = 1287,
  [1320] = 1320,
  [1321] = 1321,
  [1322] = 1322,
  [1323] = 1323,
  [1324] = 1253,
  [1325] = 1325,
  [1326] = 1326,
  [1327] = 1327,
  [1328] = 1328,
  [1329] = 1329,
  [1330] = 1330,
  [1331] = 1331,
  [1332] = 1330,
  [1333] = 1326,
  [1334] = 1331,
  [1335] = 1335,
  [1336] = 1336,
  [1337] = 1337,
  [1338] = 1338,
  [1339] = 1330,
  [1340] = 1251,
  [1341] = 1341,
  [1342] = 1342,
  [1343] = 1322,
  [1344] = 1204,
  [1345] = 1345,
  [1346] = 1227,
  [1347] = 1347,
  [1348] = 1348,
  [1349] = 1328,
  [1350] = 1328,
  [1351] = 1329,
  [1352] = 1329,
  [1353] = 1353,
  [1354] = 1354,
  [1355] = 1355,
  [1356] = 1356,
  [1357] = 1330,
  [1358] = 1358,
  [1359] = 1359,
  [1360] = 1360,
  [1361] = 1328,
  [1362] = 1362,
  [1363] = 1329,
  [1364] = 1331,
  [1365] = 1365,
  [1366] = 1241,
  [1367] = 1353,
  [1368] = 1362,
  [1369] = 1331,
  [1370] = 1370,
  [1371] = 1371,
  [1372] = 1372,
  [1373] = 1345,
  [1374] = 1336,
  [1375] = 1375,
  [1376] = 1376,
  [1377] = 1377,
  [1378] = 1378,
  [1379] = 1224,
  [1380] = 1380,
  [1381] = 1381,
  [1382] = 1347,
  [1383] = 1383,
  [1384] = 1384,
  [1385] = 1385,
  [1386] = 1378,
  [1387] = 1376,
  [1388] = 1381,
  [1389] = 1389,
  [1390] = 1390,
  [1391] = 1391,
  [1392] = 1360,
  [1393] = 1323,
  [1394] = 1394,
  [1395] = 1395,
  [1396] = 1396,
  [1397] = 1397,
  [1398] = 1398,
  [1399] = 1399,
  [1400] = 1400,
  [1401] = 1401,
  [1402] = 1402,
  [1403] = 411,
  [1404] = 1404,
  [1405] = 1405,
  [1406] = 1402,
  [1407] = 1407,
  [1408] = 1408,
  [1409] = 1409,
  [1410] = 1410,
  [1411] = 1411,
  [1412] = 1412,
  [1413] = 1413,
  [1414] = 1414,
  [1415] = 1415,
  [1416] = 1416,
  [1417] = 1417,
  [1418] = 1418,
  [1419] = 1419,
  [1420] = 1416,
  [1421] = 1421,
  [1422] = 1422,
  [1423] = 1423,
  [1424] = 1424,
  [1425] = 1425,
  [1426] = 1426,
  [1427] = 1427,
  [1428] = 1424,
  [1429] = 1429,
  [1430] = 1430,
  [1431] = 1431,
  [1432] = 1432,
  [1433] = 1433,
  [1434] = 1434,
  [1435] = 1435,
  [1436] = 1436,
  [1437] = 1405,
  [1438] = 1432,
  [1439] = 1439,
  [1440] = 1440,
  [1441] = 1422,
  [1442] = 1442,
  [1443] = 1443,
  [1444] = 1424,
  [1445] = 1432,
  [1446] = 1446,
  [1447] = 1447,
  [1448] = 1448,
  [1449] = 1449,
  [1450] = 1450,
  [1451] = 1430,
  [1452] = 1419,
  [1453] = 1453,
  [1454] = 1454,
  [1455] = 1436,
  [1456] = 1456,
  [1457] = 1454,
  [1458] = 1402,
  [1459] = 1459,
  [1460] = 1460,
  [1461] = 1449,
  [1462] = 1462,
  [1463] = 1463,
  [1464] = 1418,
  [1465] = 1436,
  [1466] = 1466,
  [1467] = 1422,
  [1468] = 1468,
  [1469] = 1469,
  [1470] = 1470,
  [1471] = 1399,
  [1472] = 1472,
  [1473] = 1401,
  [1474] = 1474,
  [1475] = 1475,
  [1476] = 1466,
  [1477] = 1477,
  [1478] = 1404,
  [1479] = 1462,
  [1480] = 1405,
  [1481] = 1472,
  [1482] = 1482,
  [1483] = 1429,
  [1484] = 1413,
  [1485] = 1460,
  [1486] = 1486,
  [1487] = 1418,
  [1488] = 1488,
  [1489] = 1489,
  [1490] = 1450,
  [1491] = 1409,
  [1492] = 1400,
  [1493] = 1493,
  [1494] = 1411,
  [1495] = 1495,
  [1496] = 1424,
  [1497] = 1497,
  [1498] = 1440,
  [1499] = 1449,
  [1500] = 1500,
  [1501] = 1501,
  [1502] = 1502,
  [1503] = 1503,
  [1504] = 1504,
  [1505] = 1505,
  [1506] = 1506,
  [1507] = 1503,
  [1508] = 1508,
  [1509] = 1504,
  [1510] = 1510,
  [1511] = 1511,
  [1512] = 1512,
  [1513] = 1513,
  [1514] = 1514,
  [1515] = 1515,
  [1516] = 1516,
  [1517] = 1517,
  [1518] = 1518,
  [1519] = 1519,
  [1520] = 1520,
  [1521] = 616,
  [1522] = 617,
  [1523] = 1506,
  [1524] = 1524,
  [1525] = 1525,
  [1526] = 1526,
  [1527] = 1511,
  [1528] = 1528,
  [1529] = 1529,
  [1530] = 1530,
  [1531] = 1531,
  [1532] = 1532,
  [1533] = 1533,
  [1534] = 1512,
  [1535] = 1535,
  [1536] = 1536,
  [1537] = 1537,
  [1538] = 1538,
  [1539] = 1539,
  [1540] = 1540,
  [1541] = 1541,
  [1542] = 1542,
  [1543] = 1543,
  [1544] = 1544,
  [1545] = 1545,
  [1546] = 1546,
  [1547] = 1547,
  [1548] = 1389,
  [1549] = 1511,
  [1550] = 1512,
  [1551] = 1551,
  [1552] = 1552,
  [1553] = 1553,
  [1554] = 1554,
  [1555] = 1519,
  [1556] = 1556,
  [1557] = 1542,
  [1558] = 1558,
  [1559] = 1559,
  [1560] = 1560,
  [1561] = 1561,
  [1562] = 1562,
  [1563] = 1563,
  [1564] = 1564,
  [1565] = 1565,
  [1566] = 1566,
  [1567] = 1567,
  [1568] = 1568,
  [1569] = 1569,
  [1570] = 1570,
  [1571] = 1571,
  [1572] = 1572,
  [1573] = 1501,
  [1574] = 1502,
  [1575] = 1575,
  [1576] = 1576,
  [1577] = 1577,
  [1578] = 1578,
  [1579] = 1579,
  [1580] = 1580,
  [1581] = 1581,
  [1582] = 1582,
  [1583] = 1529,
  [1584] = 1539,
  [1585] = 1540,
  [1586] = 1518,
  [1587] = 1587,
  [1588] = 1588,
  [1589] = 1589,
  [1590] = 1590,
  [1591] = 1591,
  [1592] = 1592,
  [1593] = 1552,
  [1594] = 1559,
  [1595] = 1561,
  [1596] = 1562,
  [1597] = 1565,
  [1598] = 1568,
  [1599] = 1569,
  [1600] = 1570,
  [1601] = 1572,
  [1602] = 1575,
  [1603] = 1576,
  [1604] = 1577,
  [1605] = 1605,
  [1606] = 1606,
  [1607] = 1537,
  [1608] = 1551,
  [1609] = 1609,
  [1610] = 1610,
  [1611] = 1611,
  [1612] = 1612,
  [1613] = 1613,
  [1614] = 1614,
  [1615] = 1558,
  [1616] = 1564,
  [1617] = 1571,
  [1618] = 1578,
  [1619] = 1579,
  [1620] = 1620,
  [1621] = 1589,
  [1622] = 1622,
  [1623] = 1623,
  [1624] = 1624,
  [1625] = 1625,
  [1626] = 1626,
  [1627] = 1605,
  [1628] = 1622,
  [1629] = 1623,
  [1630] = 1630,
  [1631] = 1631,
  [1632] = 1624,
  [1633] = 1633,
  [1634] = 1634,
  [1635] = 1635,
  [1636] = 1636,
  [1637] = 1637,
  [1638] = 1638,
  [1639] = 1639,
  [1640] = 1640,
  [1641] = 1641,
  [1642] = 1642,
  [1643] = 1524,
  [1644] = 1515,
  [1645] = 1530,
  [1646] = 1646,
  [1647] = 1647,
  [1648] = 1648,
  [1649] = 1649,
  [1650] = 1650,
  [1651] = 1651,
  [1652] = 1652,
  [1653] = 1653,
  [1654] = 1654,
  [1655] = 1655,
  [1656] = 1554,
  [1657] = 1517,
  [1658] = 1658,
  [1659] = 1659,
  [1660] = 1660,
  [1661] = 1661,
  [1662] = 1662,
  [1663] = 1560,
  [1664] = 1664,
  [1665] = 1531,
  [1666] = 1666,
  [1667] = 1667,
  [1668] = 1520,
  [1669] = 1669,
  [1670] = 1563,
  [1671] = 1532,
  [1672] = 1533,
  [1673] = 1673,
  [1674] = 1674,
  [1675] = 1536,
  [1676] = 1676,
  [1677] = 1677,
  [1678] = 1678,
  [1679] = 1626,
  [1680] = 1680,
  [1681] = 1681,
  [1682] = 1682,
  [1683] = 1683,
  [1684] = 1684,
  [1685] = 1685,
  [1686] = 1686,
  [1687] = 1687,
  [1688] = 1688,
  [1689] = 1689,
  [1690] = 1556,
  [1691] = 1646,
  [1692] = 1692,
  [1693] = 1647,
  [1694] = 1580,
  [1695] = 1695,
  [1696] = 1566,
  [1697] = 1535,
  [1698] = 1698,
  [1699] = 1699,
  [1700] = 1700,
  [1701] = 1650,
  [1702] = 1702,
  [1703] = 1703,
  [1704] = 1630,
  [1705] = 1631,
  [1706] = 1633,
  [1707] = 1634,
  [1708] = 1708,
  [1709] = 1587,
  [1710] = 1553,
  [1711] = 1635,
  [1712] = 1609,
  [1713] = 1588,
  [1714] = 1567,
  [1715] = 1610,
  [1716] = 1716,
  [1717] = 1651,
  [1718] = 1611,
  [1719] = 1636,
  [1720] = 1637,
  [1721] = 1638,
  [1722] = 1722,
  [1723] = 1652,
  [1724] = 1724,
  [1725] = 1653,
  [1726] = 1581,
  [1727] = 1727,
  [1728] = 1728,
  [1729] = 1729,
  [1730] = 1582,
  [1731] = 1731,
  [1732] = 1732,
  [1733] = 1511,
  [1734] = 1734,
  [1735] = 1639,
  [1736] = 1512,
  [1737] = 639,
  [1738] = 1510,
  [1739] = 1612,
  [1740] = 1613,
  [1741] = 1614,
  [1742] = 1640,
  [1743] = 1641,
  [1744] = 1642,
  [1745] = 1590,
  [1746] = 1684,
  [1747] = 1747,
  [1748] = 1748,
  [1749] = 1749,
  [1750] = 1748,
  [1751] = 1751,
  [1752] = 1752,
  [1753] = 1753,
  [1754] = 1754,
  [1755] = 1755,
  [1756] = 1756,
  [1757] = 1757,
  [1758] = 1758,
  [1759] = 1759,
  [1760] = 1760,
  [1761] = 1761,
  [1762] = 1762,
  [1763] = 1763,
  [1764] = 1764,
  [1765] = 1765,
  [1766] = 1766,
  [1767] = 1342,
  [1768] = 1768,
  [1769] = 1769,
  [1770] = 1770,
  [1771] = 1756,
  [1772] = 1772,
  [1773] = 1773,
  [1774] = 1774,
  [1775] = 1775,
  [1776] = 1762,
  [1777] = 1756,
  [1778] = 1763,
  [1779] = 1763,
  [1780] = 1768,
  [1781] = 1781,
  [1782] = 1782,
  [1783] = 1783,
  [1784] = 1784,
  [1785] = 1769,
  [1786] = 1786,
  [1787] = 1770,
  [1788] = 1788,
  [1789] = 1789,
  [1790] = 1762,
  [1791] = 1791,
  [1792] = 1792,
  [1793] = 1793,
  [1794] = 1758,
  [1795] = 1764,
  [1796] = 1796,
  [1797] = 1797,
  [1798] = 1756,
  [1799] = 1799,
  [1800] = 1800,
  [1801] = 1801,
  [1802] = 1762,
  [1803] = 1763,
  [1804] = 1804,
  [1805] = 1805,
  [1806] = 1806,
  [1807] = 1807,
  [1808] = 1770,
  [1809] = 1756,
  [1810] = 1768,
  [1811] = 1804,
  [1812] = 1760,
  [1813] = 1782,
  [1814] = 1799,
  [1815] = 1762,
  [1816] = 1763,
  [1817] = 1804,
  [1818] = 1768,
  [1819] = 1768,
  [1820] = 1781,
  [1821] = 1782,
  [1822] = 1770,
  [1823] = 1781,
  [1824] = 1782,
  [1825] = 1825,
  [1826] = 1804,
  [1827] = 1781,
  [1828] = 1828,
  [1829] = 1769,
  [1830] = 1756,
  [1831] = 1831,
  [1832] = 1769,
  [1833] = 1833,
  [1834] = 1782,
  [1835] = 1835,
  [1836] = 1836,
  [1837] = 1837,
  [1838] = 1831,
  [1839] = 1839,
  [1840] = 1840,
  [1841] = 1749,
  [1842] = 1842,
  [1843] = 1789,
  [1844] = 1844,
  [1845] = 1765,
  [1846] = 1799,
  [1847] = 1847,
  [1848] = 1848,
  [1849] = 1763,
  [1850] = 1850,
  [1851] = 1768,
  [1852] = 1835,
  [1853] = 1853,
  [1854] = 1854,
  [1855] = 1855,
  [1856] = 1781,
  [1857] = 1762,
  [1858] = 1858,
  [1859] = 1782,
  [1860] = 1860,
  [1861] = 1799,
  [1862] = 1862,
  [1863] = 1863,
  [1864] = 1864,
  [1865] = 1754,
  [1866] = 1866,
  [1867] = 1858,
  [1868] = 1847,
  [1869] = 1781,
};

static const TSSymbol ts_supertype_symbols[SUPERTYPE_COUNT] = {
  sym_declaration,
  sym_expression,
  sym_pattern,
  sym_primary_expression,
  sym_statement,
};

static const TSMapSlice ts_supertype_map_slices[] = {
  [sym_declaration] = {.index = 0, .length = 6},
  [sym_expression] = {.index = 6, .length = 12},
  [sym_pattern] = {.index = 18, .length = 7},
  [sym_primary_expression] = {.index = 25, .length = 22},
  [sym_statement] = {.index = 47, .length = 20},
};

static const TSSymbol ts_supertype_map_entries[] = {
  [0] =
    sym_class_declaration,
    sym_function_declaration,
    sym_generator_function_declaration,
    sym_lexical_declaration,
    sym_using_declaration,
    sym_variable_declaration,
  [6] =
    sym_assignment_expression,
    sym_augmented_assignment_expression,
    sym_await_expression,
    sym_binary_expression,
    sym_jsx_element,
    sym_jsx_self_closing_element,
    sym_new_expression,
    sym_primary_expression,
    sym_ternary_expression,
    sym_unary_expression,
    sym_update_expression,
    sym_yield_expression,
  [18] =
    sym_array_pattern,
    sym_identifier,
    sym_member_expression,
    sym_object_pattern,
    sym_rest_pattern,
    sym_subscript_expression,
    sym_undefined,
  [25] =
    sym_array,
    sym_arrow_function,
    sym_call_expression,
    sym_class,
    sym_false,
    sym_function_expression,
    sym_generator_function,
    sym_identifier,
    sym_member_expression,
    sym_meta_property,
    sym_null,
    sym_number,
    sym_object,
    sym_parenthesized_expression,
    sym_regex,
    sym_string,
    sym_subscript_expression,
    sym_super,
    sym_template_string,
    sym_this,
    sym_true,
    sym_undefined,
  [47] =
    sym_break_statement,
    sym_continue_statement,
    sym_debugger_statement,
    sym_declaration,
    sym_do_statement,
    sym_empty_statement,
    sym_export_statement,
    sym_expression_statement,
    sym_for_in_statement,
    sym_for_statement,
    sym_if_statement,
    sym_import_statement,
    sym_labeled_statement,
    sym_return_statement,
    sym_statement_block,
    sym_switch_statement,
    sym_throw_statement,
    sym_try_statement,
    sym_while_statement,
    sym_with_statement,
};

static const TSCharacterRange extras_character_set_1[] = {
  {'\t', '\r'}, {' ', ' '}, {0xa0, 0xa0}, {0x1680, 0x1680}, {0x2000, 0x200b}, {0x2028, 0x2029}, {0x202f, 0x202f}, {0x205f, 0x2060},
  {0x3000, 0x3000}, {0xfeff, 0xfeff},
};

static const TSCharacterRange sym_identifier_character_set_1[] = {
  {'$', '$'}, {'A', 'Z'}, {'\\', '\\'}, {'_', '_'}, {'a', 'z'}, {0x7f, 0x9f}, {0xa1, 0x167f}, {0x1681, 0x1fff},
  {0x200c, 0x2027}, {0x202a, 0x202e}, {0x2030, 0x205e}, {0x2061, 0x2fff}, {0x3001, 0xfefe}, {0xff00, 0x10ffff},
};

static const TSCharacterRange sym_identifier_character_set_2[] = {
  {'$', '$'}, {'0', '9'}, {'A', 'Z'}, {'\\', '\\'}, {'_', '_'}, {'a', 'z'}, {0x7f, 0x9f}, {0xa1, 0x167f},
  {0x1681, 0x1fff}, {0x200c, 0x2027}, {0x202a, 0x202e}, {0x2030, 0x205e}, {0x2061, 0x2fff}, {0x3001, 0xfefe}, {0xff00, 0x10ffff},
};

static bool ts_lex(TSLexer *lexer, TSStateId state) {
  START_LEXER();
  eof = lexer->eof(lexer);
  switch (state) {
    case 0:
      if (eof) ADVANCE(126);
      ADVANCE_MAP(
        '!', 226,
        '"', 159,
        '#', 4,
        '$', 273,
        '%', 214,
        '&', 201,
        '\'', 160,
        '(', 134,
        ')', 136,
        '*', 129,
        '+', 208,
        ',', 132,
        '-', 210,
        '.', 155,
        '/', 253,
        '0', 258,
        ':', 137,
        ';', 135,
        '<', 147,
        '=', 140,
        '>', 151,
        '?', 28,
        '@', 277,
        '[', 142,
        '\\', 80,
        ']', 143,
        '^', 204,
        '`', 251,
        's', 271,
        '{', 131,
        '|', 205,
        '}', 133,
        '~', 227,
      );
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(259);
      if (set_contains(extras_character_set_1, 10, lookahead)) SKIP(123);
      if (lookahead > '@') ADVANCE(275);
      END_STATE();
    case 1:
      if (lookahead == '\n') ADVANCE(278);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(1);
      END_STATE();
    case 2:
      if (lookahead == '\n') SKIP(33);
      if (lookahead == '/') ADVANCE(21);
      if (lookahead == '[') ADVANCE(76);
      if (lookahead == '\\') ADVANCE(122);
      if (set_contains(extras_character_set_1, 10, lookahead)) ADVANCE(254);
      if (lookahead != 0) ADVANCE(255);
      END_STATE();
    case 3:
      ADVANCE_MAP(
        '!', 226,
        '"', 159,
        '#', 75,
        '%', 214,
        '&', 201,
        '\'', 160,
        '(', 134,
        ')', 136,
        '*', 129,
        '+', 208,
        ',', 132,
        '-', 210,
        '.', 156,
        '/', 212,
        '0', 258,
        ':', 137,
        ';', 135,
        '<', 148,
        '=', 140,
        '>', 151,
        '?', 28,
        '@', 277,
        '[', 142,
        '\\', 82,
        ']', 143,
        '^', 204,
        '`', 251,
        '{', 131,
        '|', 205,
        '}', 133,
        '~', 227,
      );
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(259);
      if (set_contains(extras_character_set_1, 10, lookahead)) SKIP(3);
      if (lookahead > '#') ADVANCE(275);
      END_STATE();
    case 4:
      if (lookahead == '!') ADVANCE(127);
      if (lookahead == '\\') ADVANCE(81);
      if (set_contains(sym_identifier_character_set_1, 14, lookahead)) ADVANCE(276);
      END_STATE();
    case 5:
      ADVANCE_MAP(
        '!', 225,
        '"', 159,
        '#', 75,
        '\'', 160,
        '(', 134,
        '*', 128,
        '+', 207,
        ',', 132,
        '-', 209,
        '.', 27,
        '/', 211,
        '0', 258,
        ';', 135,
        '<', 145,
        '@', 277,
        '[', 142,
        '\\', 82,
        '`', 251,
        's', 271,
        '{', 131,
        '}', 133,
        '~', 227,
      );
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(259);
      if (set_contains(extras_character_set_1, 10, lookahead)) SKIP(5);
      if (lookahead > '#' &&
          (lookahead < '%' || '@' < lookahead) &&
          (lookahead < '[' || '^' < lookahead) &&
          (lookahead < '{' || '~' < lookahead)) ADVANCE(275);
      END_STATE();
    case 6:
      ADVANCE_MAP(
        '!', 225,
        '"', 159,
        '#', 75,
        '\'', 160,
        '(', 134,
        '+', 207,
        '-', 209,
        '.', 156,
        '/', 211,
        '0', 258,
        ':', 137,
        '<', 145,
        '>', 150,
        '@', 277,
        '[', 142,
        '\\', 82,
        '`', 251,
        '{', 131,
        '~', 227,
      );
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(259);
      if (set_contains(extras_character_set_1, 10, lookahead)) SKIP(6);
      if (lookahead > '#' &&
          (lookahead < '%' || '@' < lookahead) &&
          (lookahead < '[' || '^' < lookahead) &&
          (lookahead < '{' || '~' < lookahead)) ADVANCE(275);
      END_STATE();
    case 7:
      ADVANCE_MAP(
        '!', 73,
        '%', 213,
        '&', 202,
        '(', 134,
        ')', 136,
        '*', 130,
        '+', 207,
        ',', 132,
        '-', 209,
        '.', 154,
        '/', 211,
        ':', 137,
        ';', 135,
        '<', 149,
        '=', 74,
        '>', 152,
        '?', 29,
        '[', 142,
        '\\', 82,
        ']', 143,
        '^', 203,
        '`', 251,
        '{', 131,
        '|', 206,
        '}', 133,
      );
      if (('a' <= lookahead && lookahead <= 'z')) ADVANCE(256);
      if (set_contains(extras_character_set_1, 10, lookahead)) SKIP(8);
      if (lookahead > '#' &&
          (lookahead < '%' || '@' < lookahead) &&
          (lookahead < '`' || '~' < lookahead)) ADVANCE(275);
      END_STATE();
    case 8:
      ADVANCE_MAP(
        '!', 73,
        '%', 213,
        '&', 202,
        '(', 134,
        ')', 136,
        '*', 130,
        '+', 207,
        ',', 132,
        '-', 209,
        '.', 154,
        '/', 211,
        ':', 137,
        ';', 135,
        '<', 149,
        '=', 74,
        '>', 152,
        '?', 29,
        '[', 142,
        '\\', 82,
        ']', 143,
        '^', 203,
        '`', 251,
        '{', 131,
        '|', 206,
        '}', 133,
      );
      if (set_contains(extras_character_set_1, 10, lookahead)) SKIP(8);
      if (lookahead > '#' &&
          (lookahead < '%' || '@' < lookahead) &&
          (lookahead < '{' || '~' < lookahead)) ADVANCE(275);
      END_STATE();
    case 9:
      ADVANCE_MAP(
        '"', 159,
        '#', 75,
        '&', 13,
        '\'', 160,
        '(', 134,
        '*', 128,
        '.', 156,
        '/', 21,
        '0', 258,
        '<', 146,
        '@', 277,
        '[', 142,
        '\\', 82,
        's', 271,
        '{', 131,
      );
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(259);
      if (set_contains(extras_character_set_1, 10, lookahead)) SKIP(9);
      if (lookahead > '#' &&
          (lookahead < '%' || '@' < lookahead) &&
          (lookahead < '[' || '^' < lookahead) &&
          lookahead != '`' &&
          (lookahead < '{' || '~' < lookahead)) ADVANCE(275);
      END_STATE();
    case 10:
      if (lookahead == '"') ADVANCE(159);
      if (lookahead == '&') ADVANCE(14);
      if (lookahead == '/') ADVANCE(162);
      if (set_contains(extras_character_set_1, 10, lookahead)) ADVANCE(161);
      if (lookahead != 0) ADVANCE(163);
      END_STATE();
    case 11:
      if (lookahead == '"') ADVANCE(159);
      if (lookahead == '/') ADVANCE(21);
      if (set_contains(extras_character_set_1, 10, lookahead)) SKIP(11);
      END_STATE();
    case 12:
      if (lookahead == '"') ADVANCE(159);
      if (lookahead == '/') ADVANCE(230);
      if (lookahead == '\\') ADVANCE(83);
      if (lookahead == '\n' ||
          lookahead == '\r') SKIP(11);
      if (set_contains(extras_character_set_1, 10, lookahead)) ADVANCE(233);
      if (lookahead != 0) ADVANCE(235);
      END_STATE();
    case 13:
      if (lookahead == '#') ADVANCE(94);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(72);
      END_STATE();
    case 14:
      if (lookahead == '#') ADVANCE(94);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(72);
      if (lookahead != 0) ADVANCE(163);
      END_STATE();
    case 15:
      if (lookahead == '#') ADVANCE(94);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(72);
      if (lookahead != 0) ADVANCE(169);
      END_STATE();
    case 16:
      if (lookahead == '$') ADVANCE(84);
      if (lookahead == '/') ADVANCE(21);
      if (lookahead == '\\') ADVANCE(83);
      if (lookahead == '`') ADVANCE(251);
      if (set_contains(extras_character_set_1, 10, lookahead)) SKIP(17);
      END_STATE();
    case 17:
      if (lookahead == '$') ADVANCE(84);
      if (lookahead == '/') ADVANCE(21);
      if (lookahead == '`') ADVANCE(251);
      if (set_contains(extras_character_set_1, 10, lookahead)) SKIP(17);
      END_STATE();
    case 18:
      if (lookahead == '&') ADVANCE(15);
      if (lookahead == '\'') ADVANCE(160);
      if (lookahead == '/') ADVANCE(168);
      if (set_contains(extras_character_set_1, 10, lookahead)) ADVANCE(167);
      if (lookahead != 0) ADVANCE(169);
      END_STATE();
    case 19:
      if (lookahead == '\'') ADVANCE(160);
      if (lookahead == '/') ADVANCE(21);
      if (set_contains(extras_character_set_1, 10, lookahead)) SKIP(19);
      END_STATE();
    case 20:
      if (lookahead == '\'') ADVANCE(160);
      if (lookahead == '/') ADVANCE(236);
      if (lookahead == '\\') ADVANCE(83);
      if (lookahead == '\n' ||
          lookahead == '\r') SKIP(19);
      if (set_contains(extras_character_set_1, 10, lookahead)) ADVANCE(239);
      if (lookahead != 0) ADVANCE(241);
      END_STATE();
    case 21:
      if (lookahead == '*') ADVANCE(24);
      if (lookahead == '/') ADVANCE(250);
      END_STATE();
    case 22:
      if (lookahead == '*') ADVANCE(24);
      if (lookahead == '/') ADVANCE(250);
      if (lookahead == '>') ADVANCE(158);
      END_STATE();
    case 23:
      if (lookahead == '*') ADVANCE(23);
      if (lookahead == '/') ADVANCE(247);
      if (lookahead != 0) ADVANCE(24);
      END_STATE();
    case 24:
      if (lookahead == '*') ADVANCE(23);
      if (lookahead != 0) ADVANCE(24);
      END_STATE();
    case 25:
      if (lookahead == '*') ADVANCE(164);
      if (lookahead == '#' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(24);
      if (lookahead != 0) ADVANCE(165);
      END_STATE();
    case 26:
      if (lookahead == '*') ADVANCE(170);
      if (lookahead == '#' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(24);
      if (lookahead != 0) ADVANCE(171);
      END_STATE();
    case 27:
      if (lookahead == '.') ADVANCE(30);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(264);
      END_STATE();
    case 28:
      if (lookahead == '.') ADVANCE(174);
      if (lookahead == '?') ADVANCE(224);
      END_STATE();
    case 29:
      if (lookahead == '.') ADVANCE(174);
      if (lookahead == '?') ADVANCE(223);
      END_STATE();
    case 30:
      if (lookahead == '.') ADVANCE(190);
      END_STATE();
    case 31:
      if (lookahead == '.') ADVANCE(154);
      if (lookahead == '/') ADVANCE(22);
      if (lookahead == ':') ADVANCE(137);
      if (lookahead == '=') ADVANCE(138);
      if (lookahead == '>') ADVANCE(150);
      if (lookahead == '\\') ADVANCE(82);
      if (lookahead == '{') ADVANCE(131);
      if (set_contains(extras_character_set_1, 10, lookahead)) SKIP(31);
      if (lookahead == '$' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(267);
      if (lookahead > '~') ADVANCE(275);
      END_STATE();
    case 32:
      if (lookahead == '/') ADVANCE(253);
      if (set_contains(extras_character_set_1, 10, lookahead)) SKIP(33);
      END_STATE();
    case 33:
      if (lookahead == '/') ADVANCE(21);
      if (set_contains(extras_character_set_1, 10, lookahead)) SKIP(33);
      END_STATE();
    case 34:
      if (lookahead == ';') ADVANCE(144);
      END_STATE();
    case 35:
      if (lookahead == ';') ADVANCE(144);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(34);
      END_STATE();
    case 36:
      if (lookahead == ';') ADVANCE(144);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(35);
      END_STATE();
    case 37:
      if (lookahead == ';') ADVANCE(144);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(36);
      END_STATE();
    case 38:
      if (lookahead == ';') ADVANCE(144);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(37);
      END_STATE();
    case 39:
      if (lookahead == ';') ADVANCE(144);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(34);
      END_STATE();
    case 40:
      if (lookahead == ';') ADVANCE(144);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(39);
      END_STATE();
    case 41:
      if (lookahead == ';') ADVANCE(144);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(40);
      END_STATE();
    case 42:
      if (lookahead == ';') ADVANCE(144);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(41);
      END_STATE();
    case 43:
      if (lookahead == ';') ADVANCE(144);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(42);
      END_STATE();
    case 44:
      if (lookahead == ';') ADVANCE(144);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(34);
      END_STATE();
    case 45:
      if (lookahead == ';') ADVANCE(144);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(44);
      END_STATE();
    case 46:
      if (lookahead == ';') ADVANCE(144);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(45);
      END_STATE();
    case 47:
      if (lookahead == ';') ADVANCE(144);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(46);
      END_STATE();
    case 48:
      if (lookahead == ';') ADVANCE(144);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(47);
      END_STATE();
    case 49:
      if (lookahead == ';') ADVANCE(144);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(48);
      END_STATE();
    case 50:
      if (lookahead == ';') ADVANCE(144);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(49);
      END_STATE();
    case 51:
      if (lookahead == ';') ADVANCE(144);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(50);
      END_STATE();
    case 52:
      if (lookahead == ';') ADVANCE(144);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(51);
      END_STATE();
    case 53:
      if (lookahead == ';') ADVANCE(144);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(52);
      END_STATE();
    case 54:
      if (lookahead == ';') ADVANCE(144);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(53);
      END_STATE();
    case 55:
      if (lookahead == ';') ADVANCE(144);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(54);
      END_STATE();
    case 56:
      if (lookahead == ';') ADVANCE(144);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(55);
      END_STATE();
    case 57:
      if (lookahead == ';') ADVANCE(144);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(56);
      END_STATE();
    case 58:
      if (lookahead == ';') ADVANCE(144);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(57);
      END_STATE();
    case 59:
      if (lookahead == ';') ADVANCE(144);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(58);
      END_STATE();
    case 60:
      if (lookahead == ';') ADVANCE(144);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(59);
      END_STATE();
    case 61:
      if (lookahead == ';') ADVANCE(144);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(60);
      END_STATE();
    case 62:
      if (lookahead == ';') ADVANCE(144);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(61);
      END_STATE();
    case 63:
      if (lookahead == ';') ADVANCE(144);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(62);
      END_STATE();
    case 64:
      if (lookahead == ';') ADVANCE(144);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(63);
      END_STATE();
    case 65:
      if (lookahead == ';') ADVANCE(144);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(64);
      END_STATE();
    case 66:
      if (lookahead == ';') ADVANCE(144);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(65);
      END_STATE();
    case 67:
      if (lookahead == ';') ADVANCE(144);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(66);
      END_STATE();
    case 68:
      if (lookahead == ';') ADVANCE(144);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(67);
      END_STATE();
    case 69:
      if (lookahead == ';') ADVANCE(144);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(68);
      END_STATE();
    case 70:
      if (lookahead == ';') ADVANCE(144);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(69);
      END_STATE();
    case 71:
      if (lookahead == ';') ADVANCE(144);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(70);
      END_STATE();
    case 72:
      if (lookahead == ';') ADVANCE(144);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(71);
      END_STATE();
    case 73:
      if (lookahead == '=') ADVANCE(220);
      END_STATE();
    case 74:
      if (lookahead == '=') ADVANCE(218);
      END_STATE();
    case 75:
      if (lookahead == '\\') ADVANCE(81);
      if (set_contains(sym_identifier_character_set_1, 14, lookahead)) ADVANCE(276);
      END_STATE();
    case 76:
      if (lookahead == '\\') ADVANCE(121);
      if (lookahead == ']') ADVANCE(255);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(76);
      END_STATE();
    case 77:
      if (lookahead == 'e') ADVANCE(79);
      END_STATE();
    case 78:
      if (lookahead == 'g') ADVANCE(77);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(78);
      END_STATE();
    case 79:
      if (lookahead == 't') ADVANCE(1);
      END_STATE();
    case 80:
      if (lookahead == 'u') ADVANCE(85);
      if (lookahead == 'x') ADVANCE(111);
      if (lookahead == '\r' ||
          lookahead == '?') ADVANCE(244);
      if (('0' <= lookahead && lookahead <= '7')) ADVANCE(246);
      if (lookahead != 0) ADVANCE(242);
      END_STATE();
    case 81:
      if (lookahead == 'u') ADVANCE(86);
      END_STATE();
    case 82:
      if (lookahead == 'u') ADVANCE(87);
      END_STATE();
    case 83:
      if (lookahead == 'u') ADVANCE(88);
      if (lookahead == 'x') ADVANCE(111);
      if (lookahead == '\r' ||
          lookahead == '?') ADVANCE(244);
      if (('0' <= lookahead && lookahead <= '7')) ADVANCE(246);
      if (lookahead != 0) ADVANCE(242);
      END_STATE();
    case 84:
      if (lookahead == '{') ADVANCE(252);
      END_STATE();
    case 85:
      if (lookahead == '{') ADVANCE(105);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(116);
      END_STATE();
    case 86:
      if (lookahead == '{') ADVANCE(109);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(117);
      END_STATE();
    case 87:
      if (lookahead == '{') ADVANCE(110);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(118);
      END_STATE();
    case 88:
      if (lookahead == '{') ADVANCE(112);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(108);
      END_STATE();
    case 89:
      if (lookahead == '}') ADVANCE(275);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(89);
      END_STATE();
    case 90:
      if (lookahead == '}') ADVANCE(276);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(90);
      END_STATE();
    case 91:
      if (lookahead == '}') ADVANCE(242);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(91);
      END_STATE();
    case 92:
      if (lookahead == '}') ADVANCE(243);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(92);
      END_STATE();
    case 93:
      if (lookahead == '+' ||
          lookahead == '-') ADVANCE(100);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(265);
      END_STATE();
    case 94:
      if (lookahead == 'X' ||
          lookahead == 'x') ADVANCE(107);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(38);
      END_STATE();
    case 95:
      if (lookahead == '0' ||
          lookahead == '1') ADVANCE(261);
      END_STATE();
    case 96:
      if (('0' <= lookahead && lookahead <= '7')) ADVANCE(262);
      END_STATE();
    case 97:
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(259);
      END_STATE();
    case 98:
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(264);
      END_STATE();
    case 99:
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(260);
      END_STATE();
    case 100:
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(265);
      END_STATE();
    case 101:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(275);
      END_STATE();
    case 102:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(276);
      END_STATE();
    case 103:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(242);
      END_STATE();
    case 104:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(263);
      END_STATE();
    case 105:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(92);
      END_STATE();
    case 106:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(243);
      END_STATE();
    case 107:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(43);
      END_STATE();
    case 108:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(111);
      END_STATE();
    case 109:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(90);
      END_STATE();
    case 110:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(89);
      END_STATE();
    case 111:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(103);
      END_STATE();
    case 112:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(91);
      END_STATE();
    case 113:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(106);
      END_STATE();
    case 114:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(102);
      END_STATE();
    case 115:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(101);
      END_STATE();
    case 116:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(113);
      END_STATE();
    case 117:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(114);
      END_STATE();
    case 118:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(115);
      END_STATE();
    case 119:
      if (lookahead != 0 &&
          lookahead != '#' &&
          (lookahead < 'A' || 'Z' < lookahead) &&
          (lookahead < 'a' || 'z' < lookahead)) ADVANCE(163);
      END_STATE();
    case 120:
      if (lookahead != 0 &&
          lookahead != '#' &&
          (lookahead < 'A' || 'Z' < lookahead) &&
          (lookahead < 'a' || 'z' < lookahead)) ADVANCE(169);
      END_STATE();
    case 121:
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(76);
      END_STATE();
    case 122:
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(255);
      END_STATE();
    case 123:
      if (eof) ADVANCE(126);
      ADVANCE_MAP(
        '!', 226,
        '"', 159,
        '#', 4,
        '$', 273,
        '%', 214,
        '&', 201,
        '\'', 160,
        '(', 134,
        ')', 136,
        '*', 129,
        '+', 208,
        ',', 132,
        '-', 210,
        '.', 155,
        '/', 212,
        '0', 258,
        ':', 137,
        ';', 135,
        '<', 147,
        '=', 140,
        '>', 151,
        '?', 28,
        '@', 277,
        '[', 142,
        '\\', 82,
        ']', 143,
        '^', 204,
        '`', 251,
        's', 271,
        '{', 131,
        '|', 205,
        '}', 133,
        '~', 227,
      );
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(259);
      if (set_contains(extras_character_set_1, 10, lookahead)) SKIP(123);
      if (lookahead > '@') ADVANCE(275);
      END_STATE();
    case 124:
      if (eof) ADVANCE(126);
      ADVANCE_MAP(
        '!', 226,
        '"', 159,
        '#', 75,
        '%', 213,
        '&', 202,
        '\'', 160,
        '(', 134,
        ')', 136,
        '*', 130,
        '+', 207,
        ',', 132,
        '-', 209,
        '.', 156,
        '/', 211,
        '0', 258,
        ':', 137,
        ';', 135,
        '<', 149,
        '=', 139,
        '>', 152,
        '?', 29,
        '@', 277,
        '[', 142,
        '\\', 82,
        ']', 143,
        '^', 203,
        '`', 251,
        '{', 131,
        '|', 206,
        '}', 133,
        '~', 227,
      );
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(259);
      if (set_contains(extras_character_set_1, 10, lookahead)) SKIP(124);
      if (lookahead > '#') ADVANCE(275);
      END_STATE();
    case 125:
      if (eof) ADVANCE(126);
      ADVANCE_MAP(
        '!', 225,
        '"', 159,
        '#', 4,
        '\'', 160,
        '(', 134,
        ')', 136,
        '*', 128,
        '+', 207,
        ',', 132,
        '-', 209,
        '.', 27,
        '/', 211,
        '0', 258,
        ':', 137,
        ';', 135,
        '<', 145,
        '=', 141,
        '>', 150,
        '@', 277,
        '[', 142,
        '\\', 82,
        ']', 143,
        '`', 251,
        '{', 131,
        '}', 133,
        '~', 227,
      );
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(259);
      if (set_contains(extras_character_set_1, 10, lookahead)) SKIP(125);
      if (lookahead > '#' &&
          (lookahead < '%' || '@' < lookahead) &&
          (lookahead < '[' || '^' < lookahead) &&
          (lookahead < '{' || '~' < lookahead)) ADVANCE(275);
      END_STATE();
    case 126:
      ACCEPT_TOKEN(ts_builtin_sym_end);
      END_STATE();
    case 127:
      ACCEPT_TOKEN(sym_hash_bang_line);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(127);
      END_STATE();
    case 128:
      ACCEPT_TOKEN(anon_sym_STAR);
      END_STATE();
    case 129:
      ACCEPT_TOKEN(anon_sym_STAR);
      if (lookahead == '*') ADVANCE(216);
      if (lookahead == '=') ADVANCE(177);
      END_STATE();
    case 130:
      ACCEPT_TOKEN(anon_sym_STAR);
      if (lookahead == '*') ADVANCE(215);
      END_STATE();
    case 131:
      ACCEPT_TOKEN(anon_sym_LBRACE);
      END_STATE();
    case 132:
      ACCEPT_TOKEN(anon_sym_COMMA);
      END_STATE();
    case 133:
      ACCEPT_TOKEN(anon_sym_RBRACE);
      END_STATE();
    case 134:
      ACCEPT_TOKEN(anon_sym_LPAREN);
      END_STATE();
    case 135:
      ACCEPT_TOKEN(anon_sym_SEMI);
      END_STATE();
    case 136:
      ACCEPT_TOKEN(anon_sym_RPAREN);
      END_STATE();
    case 137:
      ACCEPT_TOKEN(anon_sym_COLON);
      END_STATE();
    case 138:
      ACCEPT_TOKEN(anon_sym_EQ);
      END_STATE();
    case 139:
      ACCEPT_TOKEN(anon_sym_EQ);
      if (lookahead == '=') ADVANCE(218);
      END_STATE();
    case 140:
      ACCEPT_TOKEN(anon_sym_EQ);
      if (lookahead == '=') ADVANCE(218);
      if (lookahead == '>') ADVANCE(173);
      END_STATE();
    case 141:
      ACCEPT_TOKEN(anon_sym_EQ);
      if (lookahead == '>') ADVANCE(173);
      END_STATE();
    case 142:
      ACCEPT_TOKEN(anon_sym_LBRACK);
      END_STATE();
    case 143:
      ACCEPT_TOKEN(anon_sym_RBRACK);
      END_STATE();
    case 144:
      ACCEPT_TOKEN(sym_html_character_reference);
      END_STATE();
    case 145:
      ACCEPT_TOKEN(anon_sym_LT);
      END_STATE();
    case 146:
      ACCEPT_TOKEN(anon_sym_LT);
      if (lookahead == '/') ADVANCE(157);
      END_STATE();
    case 147:
      ACCEPT_TOKEN(anon_sym_LT);
      if (lookahead == '/') ADVANCE(157);
      if (lookahead == '<') ADVANCE(200);
      if (lookahead == '=') ADVANCE(217);
      END_STATE();
    case 148:
      ACCEPT_TOKEN(anon_sym_LT);
      if (lookahead == '<') ADVANCE(200);
      if (lookahead == '=') ADVANCE(217);
      END_STATE();
    case 149:
      ACCEPT_TOKEN(anon_sym_LT);
      if (lookahead == '<') ADVANCE(199);
      if (lookahead == '=') ADVANCE(217);
      END_STATE();
    case 150:
      ACCEPT_TOKEN(anon_sym_GT);
      END_STATE();
    case 151:
      ACCEPT_TOKEN(anon_sym_GT);
      if (lookahead == '=') ADVANCE(222);
      if (lookahead == '>') ADVANCE(195);
      END_STATE();
    case 152:
      ACCEPT_TOKEN(anon_sym_GT);
      if (lookahead == '=') ADVANCE(222);
      if (lookahead == '>') ADVANCE(196);
      END_STATE();
    case 153:
      ACCEPT_TOKEN(sym_jsx_identifier);
      if (lookahead == '$' ||
          lookahead == '-' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(153);
      END_STATE();
    case 154:
      ACCEPT_TOKEN(anon_sym_DOT);
      END_STATE();
    case 155:
      ACCEPT_TOKEN(anon_sym_DOT);
      if (lookahead == '.') ADVANCE(30);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(264);
      END_STATE();
    case 156:
      ACCEPT_TOKEN(anon_sym_DOT);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(264);
      END_STATE();
    case 157:
      ACCEPT_TOKEN(anon_sym_LT_SLASH);
      END_STATE();
    case 158:
      ACCEPT_TOKEN(anon_sym_SLASH_GT);
      END_STATE();
    case 159:
      ACCEPT_TOKEN(anon_sym_DQUOTE);
      END_STATE();
    case 160:
      ACCEPT_TOKEN(anon_sym_SQUOTE);
      END_STATE();
    case 161:
      ACCEPT_TOKEN(sym_unescaped_double_jsx_string_fragment);
      if (lookahead == '&') ADVANCE(14);
      if (lookahead == '/') ADVANCE(162);
      if (set_contains(extras_character_set_1, 10, lookahead)) ADVANCE(161);
      if (lookahead != 0 &&
          lookahead != '"') ADVANCE(163);
      END_STATE();
    case 162:
      ACCEPT_TOKEN(sym_unescaped_double_jsx_string_fragment);
      if (lookahead == '&') ADVANCE(119);
      if (lookahead == '*') ADVANCE(165);
      if (lookahead == '/') ADVANCE(166);
      if (lookahead != 0 &&
          lookahead != '"') ADVANCE(163);
      END_STATE();
    case 163:
      ACCEPT_TOKEN(sym_unescaped_double_jsx_string_fragment);
      if (lookahead == '&') ADVANCE(119);
      if (lookahead != 0 &&
          lookahead != '"') ADVANCE(163);
      END_STATE();
    case 164:
      ACCEPT_TOKEN(sym_unescaped_double_jsx_string_fragment);
      if (lookahead == '&') ADVANCE(25);
      if (lookahead == '*') ADVANCE(164);
      if (lookahead == '/') ADVANCE(163);
      if (lookahead != 0 &&
          lookahead != '"') ADVANCE(165);
      END_STATE();
    case 165:
      ACCEPT_TOKEN(sym_unescaped_double_jsx_string_fragment);
      if (lookahead == '&') ADVANCE(25);
      if (lookahead == '*') ADVANCE(164);
      if (lookahead != 0 &&
          lookahead != '"') ADVANCE(165);
      END_STATE();
    case 166:
      ACCEPT_TOKEN(sym_unescaped_double_jsx_string_fragment);
      if (lookahead == '&') ADVANCE(248);
      if (lookahead == '\n' ||
          lookahead == '\r' ||
          lookahead == 0x2028 ||
          lookahead == 0x2029) ADVANCE(163);
      if (lookahead != 0 &&
          lookahead != '"') ADVANCE(166);
      END_STATE();
    case 167:
      ACCEPT_TOKEN(sym_unescaped_single_jsx_string_fragment);
      if (lookahead == '&') ADVANCE(15);
      if (lookahead == '/') ADVANCE(168);
      if (set_contains(extras_character_set_1, 10, lookahead)) ADVANCE(167);
      if (lookahead != 0 &&
          lookahead != '&' &&
          lookahead != '\'') ADVANCE(169);
      END_STATE();
    case 168:
      ACCEPT_TOKEN(sym_unescaped_single_jsx_string_fragment);
      if (lookahead == '&') ADVANCE(120);
      if (lookahead == '*') ADVANCE(171);
      if (lookahead == '/') ADVANCE(172);
      if (lookahead != 0 &&
          lookahead != '&' &&
          lookahead != '\'') ADVANCE(169);
      END_STATE();
    case 169:
      ACCEPT_TOKEN(sym_unescaped_single_jsx_string_fragment);
      if (lookahead == '&') ADVANCE(120);
      if (lookahead != 0 &&
          lookahead != '&' &&
          lookahead != '\'') ADVANCE(169);
      END_STATE();
    case 170:
      ACCEPT_TOKEN(sym_unescaped_single_jsx_string_fragment);
      if (lookahead == '&') ADVANCE(26);
      if (lookahead == '*') ADVANCE(170);
      if (lookahead == '/') ADVANCE(169);
      if (lookahead != 0 &&
          lookahead != '&' &&
          lookahead != '\'') ADVANCE(171);
      END_STATE();
    case 171:
      ACCEPT_TOKEN(sym_unescaped_single_jsx_string_fragment);
      if (lookahead == '&') ADVANCE(26);
      if (lookahead == '*') ADVANCE(170);
      if (lookahead != 0 &&
          lookahead != '&' &&
          lookahead != '\'') ADVANCE(171);
      END_STATE();
    case 172:
      ACCEPT_TOKEN(sym_unescaped_single_jsx_string_fragment);
      if (lookahead == '&') ADVANCE(249);
      if (lookahead == '\n' ||
          lookahead == '\r' ||
          lookahead == 0x2028 ||
          lookahead == 0x2029) ADVANCE(169);
      if (lookahead != 0 &&
          lookahead != '&' &&
          lookahead != '\'') ADVANCE(172);
      END_STATE();
    case 173:
      ACCEPT_TOKEN(anon_sym_EQ_GT);
      END_STATE();
    case 174:
      ACCEPT_TOKEN(sym_optional_chain);
      END_STATE();
    case 175:
      ACCEPT_TOKEN(anon_sym_PLUS_EQ);
      END_STATE();
    case 176:
      ACCEPT_TOKEN(anon_sym_DASH_EQ);
      END_STATE();
    case 177:
      ACCEPT_TOKEN(anon_sym_STAR_EQ);
      END_STATE();
    case 178:
      ACCEPT_TOKEN(anon_sym_SLASH_EQ);
      END_STATE();
    case 179:
      ACCEPT_TOKEN(anon_sym_PERCENT_EQ);
      END_STATE();
    case 180:
      ACCEPT_TOKEN(anon_sym_CARET_EQ);
      END_STATE();
    case 181:
      ACCEPT_TOKEN(anon_sym_AMP_EQ);
      END_STATE();
    case 182:
      ACCEPT_TOKEN(anon_sym_PIPE_EQ);
      END_STATE();
    case 183:
      ACCEPT_TOKEN(anon_sym_GT_GT_EQ);
      END_STATE();
    case 184:
      ACCEPT_TOKEN(anon_sym_GT_GT_GT_EQ);
      END_STATE();
    case 185:
      ACCEPT_TOKEN(anon_sym_LT_LT_EQ);
      END_STATE();
    case 186:
      ACCEPT_TOKEN(anon_sym_STAR_STAR_EQ);
      END_STATE();
    case 187:
      ACCEPT_TOKEN(anon_sym_AMP_AMP_EQ);
      END_STATE();
    case 188:
      ACCEPT_TOKEN(anon_sym_PIPE_PIPE_EQ);
      END_STATE();
    case 189:
      ACCEPT_TOKEN(anon_sym_QMARK_QMARK_EQ);
      END_STATE();
    case 190:
      ACCEPT_TOKEN(anon_sym_DOT_DOT_DOT);
      END_STATE();
    case 191:
      ACCEPT_TOKEN(anon_sym_AMP_AMP);
      END_STATE();
    case 192:
      ACCEPT_TOKEN(anon_sym_AMP_AMP);
      if (lookahead == '=') ADVANCE(187);
      END_STATE();
    case 193:
      ACCEPT_TOKEN(anon_sym_PIPE_PIPE);
      END_STATE();
    case 194:
      ACCEPT_TOKEN(anon_sym_PIPE_PIPE);
      if (lookahead == '=') ADVANCE(188);
      END_STATE();
    case 195:
      ACCEPT_TOKEN(anon_sym_GT_GT);
      if (lookahead == '=') ADVANCE(183);
      if (lookahead == '>') ADVANCE(198);
      END_STATE();
    case 196:
      ACCEPT_TOKEN(anon_sym_GT_GT);
      if (lookahead == '>') ADVANCE(197);
      END_STATE();
    case 197:
      ACCEPT_TOKEN(anon_sym_GT_GT_GT);
      END_STATE();
    case 198:
      ACCEPT_TOKEN(anon_sym_GT_GT_GT);
      if (lookahead == '=') ADVANCE(184);
      END_STATE();
    case 199:
      ACCEPT_TOKEN(anon_sym_LT_LT);
      END_STATE();
    case 200:
      ACCEPT_TOKEN(anon_sym_LT_LT);
      if (lookahead == '=') ADVANCE(185);
      END_STATE();
    case 201:
      ACCEPT_TOKEN(anon_sym_AMP);
      if (lookahead == '&') ADVANCE(192);
      if (lookahead == '=') ADVANCE(181);
      END_STATE();
    case 202:
      ACCEPT_TOKEN(anon_sym_AMP);
      if (lookahead == '&') ADVANCE(191);
      END_STATE();
    case 203:
      ACCEPT_TOKEN(anon_sym_CARET);
      END_STATE();
    case 204:
      ACCEPT_TOKEN(anon_sym_CARET);
      if (lookahead == '=') ADVANCE(180);
      END_STATE();
    case 205:
      ACCEPT_TOKEN(anon_sym_PIPE);
      if (lookahead == '=') ADVANCE(182);
      if (lookahead == '|') ADVANCE(194);
      END_STATE();
    case 206:
      ACCEPT_TOKEN(anon_sym_PIPE);
      if (lookahead == '|') ADVANCE(193);
      END_STATE();
    case 207:
      ACCEPT_TOKEN(anon_sym_PLUS);
      if (lookahead == '+') ADVANCE(228);
      END_STATE();
    case 208:
      ACCEPT_TOKEN(anon_sym_PLUS);
      if (lookahead == '+') ADVANCE(228);
      if (lookahead == '=') ADVANCE(175);
      END_STATE();
    case 209:
      ACCEPT_TOKEN(anon_sym_DASH);
      if (lookahead == '-') ADVANCE(229);
      END_STATE();
    case 210:
      ACCEPT_TOKEN(anon_sym_DASH);
      if (lookahead == '-') ADVANCE(229);
      if (lookahead == '=') ADVANCE(176);
      END_STATE();
    case 211:
      ACCEPT_TOKEN(anon_sym_SLASH);
      if (lookahead == '*') ADVANCE(24);
      if (lookahead == '/') ADVANCE(250);
      END_STATE();
    case 212:
      ACCEPT_TOKEN(anon_sym_SLASH);
      if (lookahead == '*') ADVANCE(24);
      if (lookahead == '/') ADVANCE(250);
      if (lookahead == '=') ADVANCE(178);
      END_STATE();
    case 213:
      ACCEPT_TOKEN(anon_sym_PERCENT);
      END_STATE();
    case 214:
      ACCEPT_TOKEN(anon_sym_PERCENT);
      if (lookahead == '=') ADVANCE(179);
      END_STATE();
    case 215:
      ACCEPT_TOKEN(anon_sym_STAR_STAR);
      END_STATE();
    case 216:
      ACCEPT_TOKEN(anon_sym_STAR_STAR);
      if (lookahead == '=') ADVANCE(186);
      END_STATE();
    case 217:
      ACCEPT_TOKEN(anon_sym_LT_EQ);
      END_STATE();
    case 218:
      ACCEPT_TOKEN(anon_sym_EQ_EQ);
      if (lookahead == '=') ADVANCE(219);
      END_STATE();
    case 219:
      ACCEPT_TOKEN(anon_sym_EQ_EQ_EQ);
      END_STATE();
    case 220:
      ACCEPT_TOKEN(anon_sym_BANG_EQ);
      if (lookahead == '=') ADVANCE(221);
      END_STATE();
    case 221:
      ACCEPT_TOKEN(anon_sym_BANG_EQ_EQ);
      END_STATE();
    case 222:
      ACCEPT_TOKEN(anon_sym_GT_EQ);
      END_STATE();
    case 223:
      ACCEPT_TOKEN(anon_sym_QMARK_QMARK);
      END_STATE();
    case 224:
      ACCEPT_TOKEN(anon_sym_QMARK_QMARK);
      if (lookahead == '=') ADVANCE(189);
      END_STATE();
    case 225:
      ACCEPT_TOKEN(anon_sym_BANG);
      END_STATE();
    case 226:
      ACCEPT_TOKEN(anon_sym_BANG);
      if (lookahead == '=') ADVANCE(220);
      END_STATE();
    case 227:
      ACCEPT_TOKEN(anon_sym_TILDE);
      END_STATE();
    case 228:
      ACCEPT_TOKEN(anon_sym_PLUS_PLUS);
      END_STATE();
    case 229:
      ACCEPT_TOKEN(anon_sym_DASH_DASH);
      END_STATE();
    case 230:
      ACCEPT_TOKEN(sym_unescaped_double_string_fragment);
      if (lookahead == '*') ADVANCE(232);
      if (lookahead == '/') ADVANCE(234);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '\r' &&
          lookahead != '"' &&
          lookahead != '\\') ADVANCE(235);
      END_STATE();
    case 231:
      ACCEPT_TOKEN(sym_unescaped_double_string_fragment);
      if (lookahead == '*') ADVANCE(231);
      if (lookahead == '/') ADVANCE(235);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '\r' &&
          lookahead != '"' &&
          lookahead != '\\') ADVANCE(232);
      END_STATE();
    case 232:
      ACCEPT_TOKEN(sym_unescaped_double_string_fragment);
      if (lookahead == '*') ADVANCE(231);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '\r' &&
          lookahead != '"' &&
          lookahead != '\\') ADVANCE(232);
      END_STATE();
    case 233:
      ACCEPT_TOKEN(sym_unescaped_double_string_fragment);
      if (lookahead == '/') ADVANCE(230);
      if ((set_contains(extras_character_set_1, 10, lookahead)) &&
          lookahead != '\n' &&
          lookahead != '\r') ADVANCE(233);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != '"' &&
          lookahead != '\\') ADVANCE(235);
      END_STATE();
    case 234:
      ACCEPT_TOKEN(sym_unescaped_double_string_fragment);
      if (lookahead == 0x2028 ||
          lookahead == 0x2029) ADVANCE(235);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '\r' &&
          lookahead != '"' &&
          lookahead != '\\') ADVANCE(234);
      END_STATE();
    case 235:
      ACCEPT_TOKEN(sym_unescaped_double_string_fragment);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '\r' &&
          lookahead != '"' &&
          lookahead != '\\') ADVANCE(235);
      END_STATE();
    case 236:
      ACCEPT_TOKEN(sym_unescaped_single_string_fragment);
      if (lookahead == '*') ADVANCE(238);
      if (lookahead == '/') ADVANCE(240);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '\r' &&
          lookahead != '\'' &&
          lookahead != '\\') ADVANCE(241);
      END_STATE();
    case 237:
      ACCEPT_TOKEN(sym_unescaped_single_string_fragment);
      if (lookahead == '*') ADVANCE(237);
      if (lookahead == '/') ADVANCE(241);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '\r' &&
          lookahead != '\'' &&
          lookahead != '\\') ADVANCE(238);
      END_STATE();
    case 238:
      ACCEPT_TOKEN(sym_unescaped_single_string_fragment);
      if (lookahead == '*') ADVANCE(237);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '\r' &&
          lookahead != '\'' &&
          lookahead != '\\') ADVANCE(238);
      END_STATE();
    case 239:
      ACCEPT_TOKEN(sym_unescaped_single_string_fragment);
      if (lookahead == '/') ADVANCE(236);
      if ((set_contains(extras_character_set_1, 10, lookahead)) &&
          lookahead != '\n' &&
          lookahead != '\r') ADVANCE(239);
      if (lookahead != 0 &&
          (lookahead < '\t' || '\r' < lookahead) &&
          lookahead != '\'' &&
          lookahead != '\\') ADVANCE(241);
      END_STATE();
    case 240:
      ACCEPT_TOKEN(sym_unescaped_single_string_fragment);
      if (lookahead == 0x2028 ||
          lookahead == 0x2029) ADVANCE(241);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '\r' &&
          lookahead != '\'' &&
          lookahead != '\\') ADVANCE(240);
      END_STATE();
    case 241:
      ACCEPT_TOKEN(sym_unescaped_single_string_fragment);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '\r' &&
          lookahead != '\'' &&
          lookahead != '\\') ADVANCE(241);
      END_STATE();
    case 242:
      ACCEPT_TOKEN(sym_escape_sequence);
      END_STATE();
    case 243:
      ACCEPT_TOKEN(sym_escape_sequence);
      if (lookahead == '\\') ADVANCE(82);
      if (set_contains(sym_identifier_character_set_2, 15, lookahead)) ADVANCE(275);
      END_STATE();
    case 244:
      ACCEPT_TOKEN(sym_escape_sequence);
      if (lookahead == '\n' ||
          lookahead == 0x2028 ||
          lookahead == 0x2029) ADVANCE(242);
      END_STATE();
    case 245:
      ACCEPT_TOKEN(sym_escape_sequence);
      if (('0' <= lookahead && lookahead <= '7')) ADVANCE(242);
      END_STATE();
    case 246:
      ACCEPT_TOKEN(sym_escape_sequence);
      if (('0' <= lookahead && lookahead <= '7')) ADVANCE(245);
      END_STATE();
    case 247:
      ACCEPT_TOKEN(sym_comment);
      END_STATE();
    case 248:
      ACCEPT_TOKEN(sym_comment);
      if (lookahead == '\n' ||
          lookahead == '\r' ||
          lookahead == 0x2028 ||
          lookahead == 0x2029) ADVANCE(163);
      if (lookahead == '#' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(250);
      if (lookahead != 0) ADVANCE(166);
      END_STATE();
    case 249:
      ACCEPT_TOKEN(sym_comment);
      if (lookahead == '\n' ||
          lookahead == '\r' ||
          lookahead == 0x2028 ||
          lookahead == 0x2029) ADVANCE(169);
      if (lookahead == '#' ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(250);
      if (lookahead != 0) ADVANCE(172);
      END_STATE();
    case 250:
      ACCEPT_TOKEN(sym_comment);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '\r' &&
          lookahead != 0x2028 &&
          lookahead != 0x2029) ADVANCE(250);
      END_STATE();
    case 251:
      ACCEPT_TOKEN(anon_sym_BQUOTE);
      END_STATE();
    case 252:
      ACCEPT_TOKEN(anon_sym_DOLLAR_LBRACE);
      END_STATE();
    case 253:
      ACCEPT_TOKEN(anon_sym_SLASH2);
      END_STATE();
    case 254:
      ACCEPT_TOKEN(sym_regex_pattern);
      if (lookahead == '\n') SKIP(33);
      if (lookahead == '/') ADVANCE(21);
      if (lookahead == '[') ADVANCE(76);
      if (lookahead == '\\') ADVANCE(122);
      if (set_contains(extras_character_set_1, 10, lookahead)) ADVANCE(254);
      if (lookahead != 0) ADVANCE(255);
      END_STATE();
    case 255:
      ACCEPT_TOKEN(sym_regex_pattern);
      if (lookahead == '[') ADVANCE(76);
      if (lookahead == '\\') ADVANCE(122);
      if (lookahead != 0 &&
          lookahead != '\n' &&
          lookahead != '/') ADVANCE(255);
      END_STATE();
    case 256:
      ACCEPT_TOKEN(sym_regex_flags);
      if (lookahead == '\\') ADVANCE(82);
      if (('a' <= lookahead && lookahead <= 'z')) ADVANCE(256);
      if (set_contains(sym_identifier_character_set_2, 15, lookahead)) ADVANCE(275);
      END_STATE();
    case 257:
      ACCEPT_TOKEN(sym_number);
      END_STATE();
    case 258:
      ACCEPT_TOKEN(sym_number);
      ADVANCE_MAP(
        '.', 266,
        '0', 260,
        '_', 99,
        'n', 257,
        'B', 95,
        'b', 95,
        'E', 93,
        'e', 93,
        'O', 96,
        'o', 96,
        'X', 104,
        'x', 104,
      );
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(259);
      END_STATE();
    case 259:
      ACCEPT_TOKEN(sym_number);
      if (lookahead == '.') ADVANCE(266);
      if (lookahead == '_') ADVANCE(97);
      if (lookahead == 'n') ADVANCE(257);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(93);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(259);
      END_STATE();
    case 260:
      ACCEPT_TOKEN(sym_number);
      if (lookahead == '_') ADVANCE(99);
      if (lookahead == 'n') ADVANCE(257);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(260);
      END_STATE();
    case 261:
      ACCEPT_TOKEN(sym_number);
      if (lookahead == '_') ADVANCE(95);
      if (lookahead == 'n') ADVANCE(257);
      if (lookahead == '0' ||
          lookahead == '1') ADVANCE(261);
      END_STATE();
    case 262:
      ACCEPT_TOKEN(sym_number);
      if (lookahead == '_') ADVANCE(96);
      if (lookahead == 'n') ADVANCE(257);
      if (('0' <= lookahead && lookahead <= '7')) ADVANCE(262);
      END_STATE();
    case 263:
      ACCEPT_TOKEN(sym_number);
      if (lookahead == '_') ADVANCE(104);
      if (lookahead == 'n') ADVANCE(257);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(263);
      END_STATE();
    case 264:
      ACCEPT_TOKEN(sym_number);
      if (lookahead == '_') ADVANCE(98);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(93);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(264);
      END_STATE();
    case 265:
      ACCEPT_TOKEN(sym_number);
      if (lookahead == '_') ADVANCE(100);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(265);
      END_STATE();
    case 266:
      ACCEPT_TOKEN(sym_number);
      if (lookahead == 'E' ||
          lookahead == 'e') ADVANCE(93);
      if (('0' <= lookahead && lookahead <= '9')) ADVANCE(264);
      END_STATE();
    case 267:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '-') ADVANCE(153);
      if (lookahead == '\\') ADVANCE(82);
      if (lookahead == '$' ||
          ('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(267);
      if (set_contains(sym_identifier_character_set_2, 15, lookahead)) ADVANCE(275);
      END_STATE();
    case 268:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '\\') ADVANCE(82);
      if (lookahead == 'a') ADVANCE(272);
      if (set_contains(sym_identifier_character_set_2, 15, lookahead)) ADVANCE(275);
      END_STATE();
    case 269:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '\\') ADVANCE(82);
      if (lookahead == 'c') ADVANCE(274);
      if (set_contains(sym_identifier_character_set_2, 15, lookahead)) ADVANCE(275);
      END_STATE();
    case 270:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '\\') ADVANCE(82);
      if (lookahead == 'i') ADVANCE(269);
      if (set_contains(sym_identifier_character_set_2, 15, lookahead)) ADVANCE(275);
      END_STATE();
    case 271:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '\\') ADVANCE(82);
      if (lookahead == 't') ADVANCE(268);
      if (set_contains(sym_identifier_character_set_2, 15, lookahead)) ADVANCE(275);
      END_STATE();
    case 272:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '\\') ADVANCE(82);
      if (lookahead == 't') ADVANCE(270);
      if (set_contains(sym_identifier_character_set_2, 15, lookahead)) ADVANCE(275);
      END_STATE();
    case 273:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '\\') ADVANCE(82);
      if (lookahead == '{') ADVANCE(252);
      if (set_contains(sym_identifier_character_set_2, 15, lookahead)) ADVANCE(275);
      END_STATE();
    case 274:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '\\') ADVANCE(82);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(78);
      if (set_contains(sym_identifier_character_set_2, 15, lookahead)) ADVANCE(275);
      END_STATE();
    case 275:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '\\') ADVANCE(82);
      if (set_contains(sym_identifier_character_set_2, 15, lookahead)) ADVANCE(275);
      END_STATE();
    case 276:
      ACCEPT_TOKEN(sym_private_property_identifier);
      if (lookahead == '\\') ADVANCE(81);
      if (set_contains(sym_identifier_character_set_2, 15, lookahead)) ADVANCE(276);
      END_STATE();
    case 277:
      ACCEPT_TOKEN(anon_sym_AT);
      END_STATE();
    case 278:
      ACCEPT_TOKEN(aux_sym_method_definition_token1);
      if (lookahead == '\n') ADVANCE(278);
      if (('\t' <= lookahead && lookahead <= '\r') ||
          lookahead == ' ') ADVANCE(1);
      END_STATE();
    default:
      return false;
  }
}

static bool ts_lex_keywords(TSLexer *lexer, TSStateId state) {
  START_LEXER();
  eof = lexer->eof(lexer);
  switch (state) {
    case 0:
      ADVANCE_MAP(
        'a', 1,
        'b', 2,
        'c', 3,
        'd', 4,
        'e', 5,
        'f', 6,
        'g', 7,
        'i', 8,
        'l', 9,
        'm', 10,
        'n', 11,
        'o', 12,
        'r', 13,
        's', 14,
        't', 15,
        'u', 16,
        'v', 17,
        'w', 18,
        'y', 19,
      );
      if (set_contains(extras_character_set_1, 10, lookahead)) SKIP(0);
      END_STATE();
    case 1:
      if (lookahead == 's') ADVANCE(20);
      if (lookahead == 'w') ADVANCE(21);
      END_STATE();
    case 2:
      if (lookahead == 'r') ADVANCE(22);
      END_STATE();
    case 3:
      if (lookahead == 'a') ADVANCE(23);
      if (lookahead == 'l') ADVANCE(24);
      if (lookahead == 'o') ADVANCE(25);
      END_STATE();
    case 4:
      if (lookahead == 'e') ADVANCE(26);
      if (lookahead == 'o') ADVANCE(27);
      END_STATE();
    case 5:
      if (lookahead == 'l') ADVANCE(28);
      if (lookahead == 'x') ADVANCE(29);
      END_STATE();
    case 6:
      if (lookahead == 'a') ADVANCE(30);
      if (lookahead == 'i') ADVANCE(31);
      if (lookahead == 'o') ADVANCE(32);
      if (lookahead == 'r') ADVANCE(33);
      if (lookahead == 'u') ADVANCE(34);
      END_STATE();
    case 7:
      if (lookahead == 'e') ADVANCE(35);
      END_STATE();
    case 8:
      if (lookahead == 'f') ADVANCE(36);
      if (lookahead == 'm') ADVANCE(37);
      if (lookahead == 'n') ADVANCE(38);
      END_STATE();
    case 9:
      if (lookahead == 'e') ADVANCE(39);
      END_STATE();
    case 10:
      if (lookahead == 'e') ADVANCE(40);
      END_STATE();
    case 11:
      if (lookahead == 'e') ADVANCE(41);
      if (lookahead == 'u') ADVANCE(42);
      END_STATE();
    case 12:
      if (lookahead == 'f') ADVANCE(43);
      END_STATE();
    case 13:
      if (lookahead == 'e') ADVANCE(44);
      END_STATE();
    case 14:
      if (lookahead == 'e') ADVANCE(45);
      if (lookahead == 't') ADVANCE(46);
      if (lookahead == 'u') ADVANCE(47);
      if (lookahead == 'w') ADVANCE(48);
      END_STATE();
    case 15:
      if (lookahead == 'a') ADVANCE(49);
      if (lookahead == 'h') ADVANCE(50);
      if (lookahead == 'r') ADVANCE(51);
      if (lookahead == 'y') ADVANCE(52);
      END_STATE();
    case 16:
      if (lookahead == 'n') ADVANCE(53);
      if (lookahead == 's') ADVANCE(54);
      END_STATE();
    case 17:
      if (lookahead == 'a') ADVANCE(55);
      if (lookahead == 'o') ADVANCE(56);
      END_STATE();
    case 18:
      if (lookahead == 'h') ADVANCE(57);
      if (lookahead == 'i') ADVANCE(58);
      END_STATE();
    case 19:
      if (lookahead == 'i') ADVANCE(59);
      END_STATE();
    case 20:
      ACCEPT_TOKEN(anon_sym_as);
      if (lookahead == 'y') ADVANCE(60);
      END_STATE();
    case 21:
      if (lookahead == 'a') ADVANCE(61);
      END_STATE();
    case 22:
      if (lookahead == 'e') ADVANCE(62);
      END_STATE();
    case 23:
      if (lookahead == 's') ADVANCE(63);
      if (lookahead == 't') ADVANCE(64);
      END_STATE();
    case 24:
      if (lookahead == 'a') ADVANCE(65);
      END_STATE();
    case 25:
      if (lookahead == 'n') ADVANCE(66);
      END_STATE();
    case 26:
      if (lookahead == 'b') ADVANCE(67);
      if (lookahead == 'f') ADVANCE(68);
      if (lookahead == 'l') ADVANCE(69);
      END_STATE();
    case 27:
      ACCEPT_TOKEN(anon_sym_do);
      END_STATE();
    case 28:
      if (lookahead == 's') ADVANCE(70);
      END_STATE();
    case 29:
      if (lookahead == 'p') ADVANCE(71);
      if (lookahead == 't') ADVANCE(72);
      END_STATE();
    case 30:
      if (lookahead == 'l') ADVANCE(73);
      END_STATE();
    case 31:
      if (lookahead == 'n') ADVANCE(74);
      END_STATE();
    case 32:
      if (lookahead == 'r') ADVANCE(75);
      END_STATE();
    case 33:
      if (lookahead == 'o') ADVANCE(76);
      END_STATE();
    case 34:
      if (lookahead == 'n') ADVANCE(77);
      END_STATE();
    case 35:
      if (lookahead == 't') ADVANCE(78);
      END_STATE();
    case 36:
      ACCEPT_TOKEN(anon_sym_if);
      END_STATE();
    case 37:
      if (lookahead == 'p') ADVANCE(79);
      END_STATE();
    case 38:
      ACCEPT_TOKEN(anon_sym_in);
      if (lookahead == 's') ADVANCE(80);
      END_STATE();
    case 39:
      if (lookahead == 't') ADVANCE(81);
      END_STATE();
    case 40:
      if (lookahead == 't') ADVANCE(82);
      END_STATE();
    case 41:
      if (lookahead == 'w') ADVANCE(83);
      END_STATE();
    case 42:
      if (lookahead == 'l') ADVANCE(84);
      END_STATE();
    case 43:
      ACCEPT_TOKEN(anon_sym_of);
      END_STATE();
    case 44:
      if (lookahead == 't') ADVANCE(85);
      END_STATE();
    case 45:
      if (lookahead == 't') ADVANCE(86);
      END_STATE();
    case 46:
      if (lookahead == 'a') ADVANCE(87);
      END_STATE();
    case 47:
      if (lookahead == 'p') ADVANCE(88);
      END_STATE();
    case 48:
      if (lookahead == 'i') ADVANCE(89);
      END_STATE();
    case 49:
      if (lookahead == 'r') ADVANCE(90);
      END_STATE();
    case 50:
      if (lookahead == 'i') ADVANCE(91);
      if (lookahead == 'r') ADVANCE(92);
      END_STATE();
    case 51:
      if (lookahead == 'u') ADVANCE(93);
      if (lookahead == 'y') ADVANCE(94);
      END_STATE();
    case 52:
      if (lookahead == 'p') ADVANCE(95);
      END_STATE();
    case 53:
      if (lookahead == 'd') ADVANCE(96);
      END_STATE();
    case 54:
      if (lookahead == 'i') ADVANCE(97);
      END_STATE();
    case 55:
      if (lookahead == 'r') ADVANCE(98);
      END_STATE();
    case 56:
      if (lookahead == 'i') ADVANCE(99);
      END_STATE();
    case 57:
      if (lookahead == 'i') ADVANCE(100);
      END_STATE();
    case 58:
      if (lookahead == 't') ADVANCE(101);
      END_STATE();
    case 59:
      if (lookahead == 'e') ADVANCE(102);
      END_STATE();
    case 60:
      if (lookahead == 'n') ADVANCE(103);
      END_STATE();
    case 61:
      if (lookahead == 'i') ADVANCE(104);
      END_STATE();
    case 62:
      if (lookahead == 'a') ADVANCE(105);
      END_STATE();
    case 63:
      if (lookahead == 'e') ADVANCE(106);
      END_STATE();
    case 64:
      if (lookahead == 'c') ADVANCE(107);
      END_STATE();
    case 65:
      if (lookahead == 's') ADVANCE(108);
      END_STATE();
    case 66:
      if (lookahead == 's') ADVANCE(109);
      if (lookahead == 't') ADVANCE(110);
      END_STATE();
    case 67:
      if (lookahead == 'u') ADVANCE(111);
      END_STATE();
    case 68:
      if (lookahead == 'a') ADVANCE(112);
      END_STATE();
    case 69:
      if (lookahead == 'e') ADVANCE(113);
      END_STATE();
    case 70:
      if (lookahead == 'e') ADVANCE(114);
      END_STATE();
    case 71:
      if (lookahead == 'o') ADVANCE(115);
      END_STATE();
    case 72:
      if (lookahead == 'e') ADVANCE(116);
      END_STATE();
    case 73:
      if (lookahead == 's') ADVANCE(117);
      END_STATE();
    case 74:
      if (lookahead == 'a') ADVANCE(118);
      END_STATE();
    case 75:
      ACCEPT_TOKEN(anon_sym_for);
      END_STATE();
    case 76:
      if (lookahead == 'm') ADVANCE(119);
      END_STATE();
    case 77:
      if (lookahead == 'c') ADVANCE(120);
      END_STATE();
    case 78:
      ACCEPT_TOKEN(anon_sym_get);
      END_STATE();
    case 79:
      if (lookahead == 'o') ADVANCE(121);
      END_STATE();
    case 80:
      if (lookahead == 't') ADVANCE(122);
      END_STATE();
    case 81:
      ACCEPT_TOKEN(anon_sym_let);
      END_STATE();
    case 82:
      if (lookahead == 'a') ADVANCE(123);
      END_STATE();
    case 83:
      ACCEPT_TOKEN(anon_sym_new);
      END_STATE();
    case 84:
      if (lookahead == 'l') ADVANCE(124);
      END_STATE();
    case 85:
      if (lookahead == 'u') ADVANCE(125);
      END_STATE();
    case 86:
      ACCEPT_TOKEN(anon_sym_set);
      END_STATE();
    case 87:
      if (lookahead == 't') ADVANCE(126);
      END_STATE();
    case 88:
      if (lookahead == 'e') ADVANCE(127);
      END_STATE();
    case 89:
      if (lookahead == 't') ADVANCE(128);
      END_STATE();
    case 90:
      if (lookahead == 'g') ADVANCE(129);
      END_STATE();
    case 91:
      if (lookahead == 's') ADVANCE(130);
      END_STATE();
    case 92:
      if (lookahead == 'o') ADVANCE(131);
      END_STATE();
    case 93:
      if (lookahead == 'e') ADVANCE(132);
      END_STATE();
    case 94:
      ACCEPT_TOKEN(anon_sym_try);
      END_STATE();
    case 95:
      if (lookahead == 'e') ADVANCE(133);
      END_STATE();
    case 96:
      if (lookahead == 'e') ADVANCE(134);
      END_STATE();
    case 97:
      if (lookahead == 'n') ADVANCE(135);
      END_STATE();
    case 98:
      ACCEPT_TOKEN(anon_sym_var);
      END_STATE();
    case 99:
      if (lookahead == 'd') ADVANCE(136);
      END_STATE();
    case 100:
      if (lookahead == 'l') ADVANCE(137);
      END_STATE();
    case 101:
      if (lookahead == 'h') ADVANCE(138);
      END_STATE();
    case 102:
      if (lookahead == 'l') ADVANCE(139);
      END_STATE();
    case 103:
      if (lookahead == 'c') ADVANCE(140);
      END_STATE();
    case 104:
      if (lookahead == 't') ADVANCE(141);
      END_STATE();
    case 105:
      if (lookahead == 'k') ADVANCE(142);
      END_STATE();
    case 106:
      ACCEPT_TOKEN(anon_sym_case);
      END_STATE();
    case 107:
      if (lookahead == 'h') ADVANCE(143);
      END_STATE();
    case 108:
      if (lookahead == 's') ADVANCE(144);
      END_STATE();
    case 109:
      if (lookahead == 't') ADVANCE(145);
      END_STATE();
    case 110:
      if (lookahead == 'i') ADVANCE(146);
      END_STATE();
    case 111:
      if (lookahead == 'g') ADVANCE(147);
      END_STATE();
    case 112:
      if (lookahead == 'u') ADVANCE(148);
      END_STATE();
    case 113:
      if (lookahead == 't') ADVANCE(149);
      END_STATE();
    case 114:
      ACCEPT_TOKEN(anon_sym_else);
      END_STATE();
    case 115:
      if (lookahead == 'r') ADVANCE(150);
      END_STATE();
    case 116:
      if (lookahead == 'n') ADVANCE(151);
      END_STATE();
    case 117:
      if (lookahead == 'e') ADVANCE(152);
      END_STATE();
    case 118:
      if (lookahead == 'l') ADVANCE(153);
      END_STATE();
    case 119:
      ACCEPT_TOKEN(anon_sym_from);
      END_STATE();
    case 120:
      if (lookahead == 't') ADVANCE(154);
      END_STATE();
    case 121:
      if (lookahead == 'r') ADVANCE(155);
      END_STATE();
    case 122:
      if (lookahead == 'a') ADVANCE(156);
      END_STATE();
    case 123:
      ACCEPT_TOKEN(anon_sym_meta);
      END_STATE();
    case 124:
      ACCEPT_TOKEN(sym_null);
      END_STATE();
    case 125:
      if (lookahead == 'r') ADVANCE(157);
      END_STATE();
    case 126:
      if (lookahead == 'i') ADVANCE(158);
      END_STATE();
    case 127:
      if (lookahead == 'r') ADVANCE(159);
      END_STATE();
    case 128:
      if (lookahead == 'c') ADVANCE(160);
      END_STATE();
    case 129:
      if (lookahead == 'e') ADVANCE(161);
      END_STATE();
    case 130:
      ACCEPT_TOKEN(sym_this);
      END_STATE();
    case 131:
      if (lookahead == 'w') ADVANCE(162);
      END_STATE();
    case 132:
      ACCEPT_TOKEN(sym_true);
      END_STATE();
    case 133:
      if (lookahead == 'o') ADVANCE(163);
      END_STATE();
    case 134:
      if (lookahead == 'f') ADVANCE(164);
      END_STATE();
    case 135:
      if (lookahead == 'g') ADVANCE(165);
      END_STATE();
    case 136:
      ACCEPT_TOKEN(anon_sym_void);
      END_STATE();
    case 137:
      if (lookahead == 'e') ADVANCE(166);
      END_STATE();
    case 138:
      ACCEPT_TOKEN(anon_sym_with);
      END_STATE();
    case 139:
      if (lookahead == 'd') ADVANCE(167);
      END_STATE();
    case 140:
      ACCEPT_TOKEN(anon_sym_async);
      END_STATE();
    case 141:
      ACCEPT_TOKEN(anon_sym_await);
      END_STATE();
    case 142:
      ACCEPT_TOKEN(anon_sym_break);
      END_STATE();
    case 143:
      ACCEPT_TOKEN(anon_sym_catch);
      END_STATE();
    case 144:
      ACCEPT_TOKEN(anon_sym_class);
      END_STATE();
    case 145:
      ACCEPT_TOKEN(anon_sym_const);
      END_STATE();
    case 146:
      if (lookahead == 'n') ADVANCE(168);
      END_STATE();
    case 147:
      if (lookahead == 'g') ADVANCE(169);
      END_STATE();
    case 148:
      if (lookahead == 'l') ADVANCE(170);
      END_STATE();
    case 149:
      if (lookahead == 'e') ADVANCE(171);
      END_STATE();
    case 150:
      if (lookahead == 't') ADVANCE(172);
      END_STATE();
    case 151:
      if (lookahead == 'd') ADVANCE(173);
      END_STATE();
    case 152:
      ACCEPT_TOKEN(sym_false);
      END_STATE();
    case 153:
      if (lookahead == 'l') ADVANCE(174);
      END_STATE();
    case 154:
      if (lookahead == 'i') ADVANCE(175);
      END_STATE();
    case 155:
      if (lookahead == 't') ADVANCE(176);
      END_STATE();
    case 156:
      if (lookahead == 'n') ADVANCE(177);
      END_STATE();
    case 157:
      if (lookahead == 'n') ADVANCE(178);
      END_STATE();
    case 158:
      if (lookahead == 'c') ADVANCE(179);
      END_STATE();
    case 159:
      ACCEPT_TOKEN(sym_super);
      END_STATE();
    case 160:
      if (lookahead == 'h') ADVANCE(180);
      END_STATE();
    case 161:
      if (lookahead == 't') ADVANCE(181);
      END_STATE();
    case 162:
      ACCEPT_TOKEN(anon_sym_throw);
      END_STATE();
    case 163:
      if (lookahead == 'f') ADVANCE(182);
      END_STATE();
    case 164:
      if (lookahead == 'i') ADVANCE(183);
      END_STATE();
    case 165:
      ACCEPT_TOKEN(anon_sym_using);
      END_STATE();
    case 166:
      ACCEPT_TOKEN(anon_sym_while);
      END_STATE();
    case 167:
      ACCEPT_TOKEN(anon_sym_yield);
      END_STATE();
    case 168:
      if (lookahead == 'u') ADVANCE(184);
      END_STATE();
    case 169:
      if (lookahead == 'e') ADVANCE(185);
      END_STATE();
    case 170:
      if (lookahead == 't') ADVANCE(186);
      END_STATE();
    case 171:
      ACCEPT_TOKEN(anon_sym_delete);
      END_STATE();
    case 172:
      ACCEPT_TOKEN(anon_sym_export);
      END_STATE();
    case 173:
      if (lookahead == 's') ADVANCE(187);
      END_STATE();
    case 174:
      if (lookahead == 'y') ADVANCE(188);
      END_STATE();
    case 175:
      if (lookahead == 'o') ADVANCE(189);
      END_STATE();
    case 176:
      ACCEPT_TOKEN(anon_sym_import);
      END_STATE();
    case 177:
      if (lookahead == 'c') ADVANCE(190);
      END_STATE();
    case 178:
      ACCEPT_TOKEN(anon_sym_return);
      END_STATE();
    case 179:
      ACCEPT_TOKEN(anon_sym_static);
      END_STATE();
    case 180:
      ACCEPT_TOKEN(anon_sym_switch);
      END_STATE();
    case 181:
      ACCEPT_TOKEN(anon_sym_target);
      END_STATE();
    case 182:
      ACCEPT_TOKEN(anon_sym_typeof);
      END_STATE();
    case 183:
      if (lookahead == 'n') ADVANCE(191);
      END_STATE();
    case 184:
      if (lookahead == 'e') ADVANCE(192);
      END_STATE();
    case 185:
      if (lookahead == 'r') ADVANCE(193);
      END_STATE();
    case 186:
      ACCEPT_TOKEN(anon_sym_default);
      END_STATE();
    case 187:
      ACCEPT_TOKEN(anon_sym_extends);
      END_STATE();
    case 188:
      ACCEPT_TOKEN(anon_sym_finally);
      END_STATE();
    case 189:
      if (lookahead == 'n') ADVANCE(194);
      END_STATE();
    case 190:
      if (lookahead == 'e') ADVANCE(195);
      END_STATE();
    case 191:
      if (lookahead == 'e') ADVANCE(196);
      END_STATE();
    case 192:
      ACCEPT_TOKEN(anon_sym_continue);
      END_STATE();
    case 193:
      ACCEPT_TOKEN(anon_sym_debugger);
      END_STATE();
    case 194:
      ACCEPT_TOKEN(anon_sym_function);
      END_STATE();
    case 195:
      if (lookahead == 'o') ADVANCE(197);
      END_STATE();
    case 196:
      if (lookahead == 'd') ADVANCE(198);
      END_STATE();
    case 197:
      if (lookahead == 'f') ADVANCE(199);
      END_STATE();
    case 198:
      ACCEPT_TOKEN(sym_undefined);
      END_STATE();
    case 199:
      ACCEPT_TOKEN(anon_sym_instanceof);
      END_STATE();
    default:
      return false;
  }
}

static const TSLexerMode ts_lex_modes[STATE_COUNT] = {
  [0] = {.lex_state = 0, .external_lex_state = 1},
  [1] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [2] = {.lex_state = 5, .external_lex_state = 2},
  [3] = {.lex_state = 5, .external_lex_state = 2},
  [4] = {.lex_state = 5, .external_lex_state = 2},
  [5] = {.lex_state = 5, .external_lex_state = 2},
  [6] = {.lex_state = 5, .external_lex_state = 2},
  [7] = {.lex_state = 3, .external_lex_state = 3, .reserved_word_set_id = 2},
  [8] = {.lex_state = 3, .external_lex_state = 3, .reserved_word_set_id = 2},
  [9] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 3},
  [10] = {.lex_state = 3, .external_lex_state = 3, .reserved_word_set_id = 2},
  [11] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [12] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [13] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [14] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [15] = {.lex_state = 3, .external_lex_state = 4, .reserved_word_set_id = 2},
  [16] = {.lex_state = 3, .external_lex_state = 4, .reserved_word_set_id = 2},
  [17] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [18] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 4},
  [19] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 4},
  [20] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [21] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [22] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 4},
  [23] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 4},
  [24] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 4},
  [25] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 4},
  [26] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [27] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [28] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 4},
  [29] = {.lex_state = 3, .external_lex_state = 3, .reserved_word_set_id = 2},
  [30] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 4},
  [31] = {.lex_state = 3, .external_lex_state = 3, .reserved_word_set_id = 2},
  [32] = {.lex_state = 3, .external_lex_state = 4, .reserved_word_set_id = 2},
  [33] = {.lex_state = 3, .external_lex_state = 3, .reserved_word_set_id = 1},
  [34] = {.lex_state = 3, .external_lex_state = 4, .reserved_word_set_id = 1},
  [35] = {.lex_state = 3, .external_lex_state = 3, .reserved_word_set_id = 2},
  [36] = {.lex_state = 3, .external_lex_state = 4, .reserved_word_set_id = 2},
  [37] = {.lex_state = 3, .external_lex_state = 3, .reserved_word_set_id = 2},
  [38] = {.lex_state = 3, .external_lex_state = 4, .reserved_word_set_id = 2},
  [39] = {.lex_state = 3, .external_lex_state = 3, .reserved_word_set_id = 2},
  [40] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 4},
  [41] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 4},
  [42] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 4},
  [43] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 4},
  [44] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 4},
  [45] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 4},
  [46] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 4},
  [47] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 4},
  [48] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 4},
  [49] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 4},
  [50] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 4},
  [51] = {.lex_state = 3, .external_lex_state = 4, .reserved_word_set_id = 2},
  [52] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 4},
  [53] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 4},
  [54] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 4},
  [55] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 4},
  [56] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 4},
  [57] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 4},
  [58] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 4},
  [59] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 4},
  [60] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 4},
  [61] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 4},
  [62] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 4},
  [63] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 4},
  [64] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 4},
  [65] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 4},
  [66] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 4},
  [67] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 4},
  [68] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 4},
  [69] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 4},
  [70] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 4},
  [71] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 4},
  [72] = {.lex_state = 3, .external_lex_state = 3, .reserved_word_set_id = 2},
  [73] = {.lex_state = 3, .external_lex_state = 4, .reserved_word_set_id = 1},
  [74] = {.lex_state = 3, .external_lex_state = 4, .reserved_word_set_id = 1},
  [75] = {.lex_state = 3, .external_lex_state = 4, .reserved_word_set_id = 1},
  [76] = {.lex_state = 3, .external_lex_state = 4, .reserved_word_set_id = 2},
  [77] = {.lex_state = 3, .external_lex_state = 4, .reserved_word_set_id = 2},
  [78] = {.lex_state = 3, .external_lex_state = 4, .reserved_word_set_id = 2},
  [79] = {.lex_state = 3, .external_lex_state = 4, .reserved_word_set_id = 1},
  [80] = {.lex_state = 124, .external_lex_state = 4, .reserved_word_set_id = 2},
  [81] = {.lex_state = 124, .external_lex_state = 3, .reserved_word_set_id = 2},
  [82] = {.lex_state = 124, .external_lex_state = 3, .reserved_word_set_id = 2},
  [83] = {.lex_state = 124, .external_lex_state = 4, .reserved_word_set_id = 2},
  [84] = {.lex_state = 124, .external_lex_state = 4, .reserved_word_set_id = 2},
  [85] = {.lex_state = 124, .external_lex_state = 4, .reserved_word_set_id = 2},
  [86] = {.lex_state = 124, .external_lex_state = 3, .reserved_word_set_id = 5},
  [87] = {.lex_state = 124, .external_lex_state = 3, .reserved_word_set_id = 5},
  [88] = {.lex_state = 124, .external_lex_state = 3, .reserved_word_set_id = 5},
  [89] = {.lex_state = 124, .external_lex_state = 3, .reserved_word_set_id = 5},
  [90] = {.lex_state = 124, .external_lex_state = 3, .reserved_word_set_id = 5},
  [91] = {.lex_state = 124, .external_lex_state = 3, .reserved_word_set_id = 5},
  [92] = {.lex_state = 124, .external_lex_state = 3, .reserved_word_set_id = 5},
  [93] = {.lex_state = 124, .external_lex_state = 3, .reserved_word_set_id = 5},
  [94] = {.lex_state = 124, .external_lex_state = 3, .reserved_word_set_id = 5},
  [95] = {.lex_state = 124, .external_lex_state = 3, .reserved_word_set_id = 5},
  [96] = {.lex_state = 124, .external_lex_state = 3, .reserved_word_set_id = 5},
  [97] = {.lex_state = 124, .external_lex_state = 3, .reserved_word_set_id = 5},
  [98] = {.lex_state = 124, .external_lex_state = 3, .reserved_word_set_id = 5},
  [99] = {.lex_state = 124, .external_lex_state = 3, .reserved_word_set_id = 5},
  [100] = {.lex_state = 124, .external_lex_state = 3, .reserved_word_set_id = 5},
  [101] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 6},
  [102] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 6},
  [103] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 6},
  [104] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 6},
  [105] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [106] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [107] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [108] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [109] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [110] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [111] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [112] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 6},
  [113] = {.lex_state = 125, .external_lex_state = 5, .reserved_word_set_id = 1},
  [114] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 6},
  [115] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [116] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [117] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [118] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [119] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [120] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [121] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [122] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [123] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [124] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [125] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [126] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [127] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [128] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [129] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [130] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [131] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [132] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [133] = {.lex_state = 3, .external_lex_state = 3},
  [134] = {.lex_state = 3, .external_lex_state = 3},
  [135] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [136] = {.lex_state = 125, .external_lex_state = 5, .reserved_word_set_id = 1},
  [137] = {.lex_state = 3, .external_lex_state = 3},
  [138] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [139] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [140] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [141] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [142] = {.lex_state = 125, .external_lex_state = 5, .reserved_word_set_id = 7},
  [143] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [144] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [145] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [146] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [147] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [148] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [149] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [150] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [151] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [152] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [153] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [154] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [155] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [156] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [157] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 7},
  [158] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [159] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [160] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [161] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [162] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [163] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [164] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [165] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 1},
  [166] = {.lex_state = 3, .external_lex_state = 3},
  [167] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [168] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [169] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [170] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [171] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [172] = {.lex_state = 3, .external_lex_state = 3},
  [173] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [174] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [175] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [176] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [177] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [178] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [179] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [180] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [181] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [182] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [183] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [184] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [185] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [186] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [187] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [188] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [189] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [190] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [191] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [192] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 7},
  [193] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [194] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [195] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [196] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [197] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [198] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [199] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [200] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [201] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [202] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [203] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [204] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [205] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [206] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [207] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [208] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [209] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [210] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 7},
  [211] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [212] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [213] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 1},
  [214] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [215] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [216] = {.lex_state = 6, .external_lex_state = 2, .reserved_word_set_id = 1},
  [217] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [218] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [219] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [220] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [221] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [222] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [223] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [224] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [225] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [226] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [227] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [228] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [229] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [230] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [231] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [232] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [233] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [234] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [235] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [236] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [237] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [238] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [239] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [240] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [241] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [242] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [243] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [244] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [245] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [246] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [247] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [248] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [249] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [250] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [251] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [252] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [253] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [254] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [255] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [256] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [257] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [258] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [259] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [260] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [261] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [262] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [263] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [264] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [265] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [266] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [267] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [268] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [269] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [270] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [271] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [272] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [273] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [274] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [275] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [276] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [277] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [278] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [279] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [280] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [281] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [282] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [283] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [284] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [285] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [286] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [287] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [288] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [289] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [290] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [291] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [292] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [293] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [294] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [295] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [296] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [297] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [298] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [299] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [300] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [301] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [302] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [303] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [304] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [305] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [306] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [307] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [308] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [309] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [310] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [311] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [312] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [313] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [314] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [315] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [316] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [317] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [318] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [319] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [320] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [321] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [322] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [323] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [324] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [325] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [326] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [327] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [328] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [329] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [330] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [331] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [332] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [333] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [334] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [335] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [336] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [337] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [338] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [339] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [340] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [341] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [342] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [343] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [344] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [345] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [346] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [347] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [348] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [349] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [350] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [351] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [352] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [353] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [354] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [355] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [356] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 7},
  [357] = {.lex_state = 3, .external_lex_state = 3},
  [358] = {.lex_state = 3, .external_lex_state = 3},
  [359] = {.lex_state = 3, .external_lex_state = 3},
  [360] = {.lex_state = 3, .external_lex_state = 3},
  [361] = {.lex_state = 3, .external_lex_state = 3},
  [362] = {.lex_state = 3, .external_lex_state = 3},
  [363] = {.lex_state = 3, .external_lex_state = 3},
  [364] = {.lex_state = 3, .external_lex_state = 3},
  [365] = {.lex_state = 3, .external_lex_state = 3},
  [366] = {.lex_state = 3, .external_lex_state = 4, .reserved_word_set_id = 8},
  [367] = {.lex_state = 3, .external_lex_state = 4, .reserved_word_set_id = 8},
  [368] = {.lex_state = 3, .external_lex_state = 3, .reserved_word_set_id = 9},
  [369] = {.lex_state = 125, .external_lex_state = 5, .reserved_word_set_id = 10},
  [370] = {.lex_state = 3, .external_lex_state = 3, .reserved_word_set_id = 9},
  [371] = {.lex_state = 3, .external_lex_state = 3, .reserved_word_set_id = 8},
  [372] = {.lex_state = 3, .external_lex_state = 3, .reserved_word_set_id = 8},
  [373] = {.lex_state = 3, .external_lex_state = 3, .reserved_word_set_id = 1},
  [374] = {.lex_state = 125, .external_lex_state = 5, .reserved_word_set_id = 10},
  [375] = {.lex_state = 3, .external_lex_state = 3, .reserved_word_set_id = 8},
  [376] = {.lex_state = 3, .external_lex_state = 3, .reserved_word_set_id = 8},
  [377] = {.lex_state = 3, .external_lex_state = 3, .reserved_word_set_id = 8},
  [378] = {.lex_state = 3, .external_lex_state = 4, .reserved_word_set_id = 8},
  [379] = {.lex_state = 3, .external_lex_state = 4, .reserved_word_set_id = 8},
  [380] = {.lex_state = 3, .external_lex_state = 3, .reserved_word_set_id = 9},
  [381] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 10},
  [382] = {.lex_state = 3, .external_lex_state = 4, .reserved_word_set_id = 8},
  [383] = {.lex_state = 3, .external_lex_state = 4, .reserved_word_set_id = 1},
  [384] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 10},
  [385] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 10},
  [386] = {.lex_state = 3, .external_lex_state = 4, .reserved_word_set_id = 8},
  [387] = {.lex_state = 3, .external_lex_state = 4, .reserved_word_set_id = 1},
  [388] = {.lex_state = 3, .external_lex_state = 3, .reserved_word_set_id = 8},
  [389] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 11},
  [390] = {.lex_state = 3, .external_lex_state = 3, .reserved_word_set_id = 8},
  [391] = {.lex_state = 125, .external_lex_state = 5, .reserved_word_set_id = 12},
  [392] = {.lex_state = 125, .external_lex_state = 5, .reserved_word_set_id = 12},
  [393] = {.lex_state = 3, .external_lex_state = 4, .reserved_word_set_id = 1},
  [394] = {.lex_state = 125, .external_lex_state = 5, .reserved_word_set_id = 12},
  [395] = {.lex_state = 125, .external_lex_state = 5, .reserved_word_set_id = 12},
  [396] = {.lex_state = 3, .external_lex_state = 3, .reserved_word_set_id = 9},
  [397] = {.lex_state = 3, .external_lex_state = 3, .reserved_word_set_id = 9},
  [398] = {.lex_state = 125, .external_lex_state = 5, .reserved_word_set_id = 12},
  [399] = {.lex_state = 125, .external_lex_state = 5, .reserved_word_set_id = 12},
  [400] = {.lex_state = 125, .external_lex_state = 5, .reserved_word_set_id = 12},
  [401] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 11},
  [402] = {.lex_state = 125, .external_lex_state = 5, .reserved_word_set_id = 12},
  [403] = {.lex_state = 125, .external_lex_state = 5, .reserved_word_set_id = 12},
  [404] = {.lex_state = 125, .external_lex_state = 5, .reserved_word_set_id = 12},
  [405] = {.lex_state = 125, .external_lex_state = 5, .reserved_word_set_id = 12},
  [406] = {.lex_state = 3, .external_lex_state = 4, .reserved_word_set_id = 8},
  [407] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 11},
  [408] = {.lex_state = 125, .external_lex_state = 5, .reserved_word_set_id = 12},
  [409] = {.lex_state = 125, .external_lex_state = 5, .reserved_word_set_id = 12},
  [410] = {.lex_state = 3, .external_lex_state = 4, .reserved_word_set_id = 8},
  [411] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 12},
  [412] = {.lex_state = 3, .external_lex_state = 4, .reserved_word_set_id = 8},
  [413] = {.lex_state = 125, .external_lex_state = 5, .reserved_word_set_id = 12},
  [414] = {.lex_state = 3, .external_lex_state = 4, .reserved_word_set_id = 1},
  [415] = {.lex_state = 125, .external_lex_state = 5, .reserved_word_set_id = 12},
  [416] = {.lex_state = 125, .external_lex_state = 5, .reserved_word_set_id = 12},
  [417] = {.lex_state = 3, .external_lex_state = 4, .reserved_word_set_id = 1},
  [418] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 12},
  [419] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 12},
  [420] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 12},
  [421] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 12},
  [422] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 12},
  [423] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 12},
  [424] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 12},
  [425] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 12},
  [426] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 12},
  [427] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 12},
  [428] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 12},
  [429] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 12},
  [430] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 12},
  [431] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 12},
  [432] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 12},
  [433] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 12},
  [434] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 12},
  [435] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 12},
  [436] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 12},
  [437] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 12},
  [438] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 12},
  [439] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 12},
  [440] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 12},
  [441] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 12},
  [442] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 12},
  [443] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 12},
  [444] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 12},
  [445] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 12},
  [446] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 12},
  [447] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 12},
  [448] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 12},
  [449] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 12},
  [450] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 12},
  [451] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 12},
  [452] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 12},
  [453] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 12},
  [454] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 12},
  [455] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 12},
  [456] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 12},
  [457] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 12},
  [458] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 12},
  [459] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 12},
  [460] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 12},
  [461] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 12},
  [462] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 12},
  [463] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 12},
  [464] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 12},
  [465] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 12},
  [466] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 12},
  [467] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 12},
  [468] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 12},
  [469] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 12},
  [470] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 12},
  [471] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 12},
  [472] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 12},
  [473] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 12},
  [474] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 12},
  [475] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 12},
  [476] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 12},
  [477] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 12},
  [478] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 12},
  [479] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 12},
  [480] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 12},
  [481] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 12},
  [482] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 12},
  [483] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 12},
  [484] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 12},
  [485] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 12},
  [486] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 12},
  [487] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 12},
  [488] = {.lex_state = 3, .external_lex_state = 4, .reserved_word_set_id = 1},
  [489] = {.lex_state = 3, .external_lex_state = 3, .reserved_word_set_id = 9},
  [490] = {.lex_state = 3, .external_lex_state = 3},
  [491] = {.lex_state = 3, .external_lex_state = 4},
  [492] = {.lex_state = 3, .external_lex_state = 3},
  [493] = {.lex_state = 3, .external_lex_state = 4},
  [494] = {.lex_state = 3, .external_lex_state = 4},
  [495] = {.lex_state = 3, .external_lex_state = 4},
  [496] = {.lex_state = 3, .external_lex_state = 4},
  [497] = {.lex_state = 3, .external_lex_state = 4},
  [498] = {.lex_state = 3, .external_lex_state = 4},
  [499] = {.lex_state = 3, .external_lex_state = 3},
  [500] = {.lex_state = 3, .external_lex_state = 4},
  [501] = {.lex_state = 3, .external_lex_state = 4},
  [502] = {.lex_state = 3, .external_lex_state = 4},
  [503] = {.lex_state = 3, .external_lex_state = 4},
  [504] = {.lex_state = 3, .external_lex_state = 4},
  [505] = {.lex_state = 3, .external_lex_state = 4, .reserved_word_set_id = 1},
  [506] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [507] = {.lex_state = 3, .external_lex_state = 4},
  [508] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [509] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [510] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [511] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [512] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [513] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [514] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [515] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [516] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [517] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [518] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [519] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [520] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [521] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [522] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [523] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [524] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 4},
  [525] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [526] = {.lex_state = 3, .external_lex_state = 3},
  [527] = {.lex_state = 3, .external_lex_state = 3},
  [528] = {.lex_state = 3, .external_lex_state = 3},
  [529] = {.lex_state = 3, .external_lex_state = 3},
  [530] = {.lex_state = 3, .external_lex_state = 3},
  [531] = {.lex_state = 3, .external_lex_state = 3},
  [532] = {.lex_state = 3, .external_lex_state = 4},
  [533] = {.lex_state = 3, .external_lex_state = 4},
  [534] = {.lex_state = 3, .external_lex_state = 3},
  [535] = {.lex_state = 3, .external_lex_state = 3},
  [536] = {.lex_state = 3, .external_lex_state = 4},
  [537] = {.lex_state = 3, .external_lex_state = 3},
  [538] = {.lex_state = 3, .external_lex_state = 3},
  [539] = {.lex_state = 3, .external_lex_state = 4},
  [540] = {.lex_state = 3, .external_lex_state = 4},
  [541] = {.lex_state = 3, .external_lex_state = 3},
  [542] = {.lex_state = 3, .external_lex_state = 3},
  [543] = {.lex_state = 3, .external_lex_state = 3},
  [544] = {.lex_state = 3, .external_lex_state = 3},
  [545] = {.lex_state = 3, .external_lex_state = 3},
  [546] = {.lex_state = 3, .external_lex_state = 3},
  [547] = {.lex_state = 3, .external_lex_state = 3},
  [548] = {.lex_state = 3, .external_lex_state = 3},
  [549] = {.lex_state = 3, .external_lex_state = 4},
  [550] = {.lex_state = 3, .external_lex_state = 3},
  [551] = {.lex_state = 3, .external_lex_state = 3},
  [552] = {.lex_state = 3, .external_lex_state = 3},
  [553] = {.lex_state = 3, .external_lex_state = 4},
  [554] = {.lex_state = 3, .external_lex_state = 4},
  [555] = {.lex_state = 3, .external_lex_state = 4},
  [556] = {.lex_state = 3, .external_lex_state = 4},
  [557] = {.lex_state = 3, .external_lex_state = 4},
  [558] = {.lex_state = 3, .external_lex_state = 4},
  [559] = {.lex_state = 3, .external_lex_state = 4},
  [560] = {.lex_state = 3, .external_lex_state = 3},
  [561] = {.lex_state = 3, .external_lex_state = 4},
  [562] = {.lex_state = 3, .external_lex_state = 4},
  [563] = {.lex_state = 3, .external_lex_state = 4},
  [564] = {.lex_state = 3, .external_lex_state = 4},
  [565] = {.lex_state = 3, .external_lex_state = 4},
  [566] = {.lex_state = 3, .external_lex_state = 4},
  [567] = {.lex_state = 3, .external_lex_state = 4},
  [568] = {.lex_state = 3, .external_lex_state = 4},
  [569] = {.lex_state = 3, .external_lex_state = 4},
  [570] = {.lex_state = 3, .external_lex_state = 4},
  [571] = {.lex_state = 3, .external_lex_state = 4},
  [572] = {.lex_state = 3, .external_lex_state = 4},
  [573] = {.lex_state = 3, .external_lex_state = 4},
  [574] = {.lex_state = 3, .external_lex_state = 4},
  [575] = {.lex_state = 3, .external_lex_state = 4},
  [576] = {.lex_state = 3, .external_lex_state = 4},
  [577] = {.lex_state = 3, .external_lex_state = 4},
  [578] = {.lex_state = 3, .external_lex_state = 4},
  [579] = {.lex_state = 3, .external_lex_state = 4},
  [580] = {.lex_state = 3, .external_lex_state = 4},
  [581] = {.lex_state = 124, .external_lex_state = 4},
  [582] = {.lex_state = 124, .external_lex_state = 3},
  [583] = {.lex_state = 124, .external_lex_state = 4},
  [584] = {.lex_state = 124, .external_lex_state = 3},
  [585] = {.lex_state = 124, .external_lex_state = 4},
  [586] = {.lex_state = 124, .external_lex_state = 3},
  [587] = {.lex_state = 124, .external_lex_state = 4},
  [588] = {.lex_state = 124, .external_lex_state = 4},
  [589] = {.lex_state = 124, .external_lex_state = 4},
  [590] = {.lex_state = 124, .external_lex_state = 4},
  [591] = {.lex_state = 124, .external_lex_state = 4},
  [592] = {.lex_state = 124, .external_lex_state = 4},
  [593] = {.lex_state = 124, .external_lex_state = 4},
  [594] = {.lex_state = 124, .external_lex_state = 4},
  [595] = {.lex_state = 124, .external_lex_state = 4},
  [596] = {.lex_state = 124, .external_lex_state = 4},
  [597] = {.lex_state = 124, .external_lex_state = 4},
  [598] = {.lex_state = 7, .external_lex_state = 4},
  [599] = {.lex_state = 124, .external_lex_state = 4},
  [600] = {.lex_state = 124, .external_lex_state = 4},
  [601] = {.lex_state = 124, .external_lex_state = 4},
  [602] = {.lex_state = 124, .external_lex_state = 4},
  [603] = {.lex_state = 124, .external_lex_state = 4},
  [604] = {.lex_state = 124, .external_lex_state = 4},
  [605] = {.lex_state = 124, .external_lex_state = 4},
  [606] = {.lex_state = 124, .external_lex_state = 4},
  [607] = {.lex_state = 124, .external_lex_state = 4},
  [608] = {.lex_state = 124, .external_lex_state = 4},
  [609] = {.lex_state = 124, .external_lex_state = 4},
  [610] = {.lex_state = 124, .external_lex_state = 4},
  [611] = {.lex_state = 124, .external_lex_state = 4},
  [612] = {.lex_state = 124, .external_lex_state = 4},
  [613] = {.lex_state = 124, .external_lex_state = 4},
  [614] = {.lex_state = 124, .external_lex_state = 4},
  [615] = {.lex_state = 124, .external_lex_state = 4},
  [616] = {.lex_state = 124, .external_lex_state = 4},
  [617] = {.lex_state = 124, .external_lex_state = 4},
  [618] = {.lex_state = 124, .external_lex_state = 4},
  [619] = {.lex_state = 124, .external_lex_state = 4},
  [620] = {.lex_state = 124, .external_lex_state = 4},
  [621] = {.lex_state = 124, .external_lex_state = 4},
  [622] = {.lex_state = 124, .external_lex_state = 4},
  [623] = {.lex_state = 124, .external_lex_state = 4},
  [624] = {.lex_state = 124, .external_lex_state = 4},
  [625] = {.lex_state = 124, .external_lex_state = 4},
  [626] = {.lex_state = 124, .external_lex_state = 4},
  [627] = {.lex_state = 124, .external_lex_state = 4},
  [628] = {.lex_state = 124, .external_lex_state = 4},
  [629] = {.lex_state = 124, .external_lex_state = 4},
  [630] = {.lex_state = 124, .external_lex_state = 4},
  [631] = {.lex_state = 124, .external_lex_state = 4},
  [632] = {.lex_state = 124, .external_lex_state = 4},
  [633] = {.lex_state = 124, .external_lex_state = 4},
  [634] = {.lex_state = 124, .external_lex_state = 4},
  [635] = {.lex_state = 124, .external_lex_state = 4},
  [636] = {.lex_state = 124, .external_lex_state = 4},
  [637] = {.lex_state = 124, .external_lex_state = 4},
  [638] = {.lex_state = 124, .external_lex_state = 4},
  [639] = {.lex_state = 124, .external_lex_state = 4},
  [640] = {.lex_state = 124, .external_lex_state = 4},
  [641] = {.lex_state = 124, .external_lex_state = 4},
  [642] = {.lex_state = 124, .external_lex_state = 4},
  [643] = {.lex_state = 124, .external_lex_state = 4},
  [644] = {.lex_state = 124, .external_lex_state = 4},
  [645] = {.lex_state = 124, .external_lex_state = 4},
  [646] = {.lex_state = 124, .external_lex_state = 4},
  [647] = {.lex_state = 124, .external_lex_state = 4},
  [648] = {.lex_state = 124, .external_lex_state = 4},
  [649] = {.lex_state = 124, .external_lex_state = 4},
  [650] = {.lex_state = 124, .external_lex_state = 4},
  [651] = {.lex_state = 124, .external_lex_state = 4},
  [652] = {.lex_state = 124, .external_lex_state = 4},
  [653] = {.lex_state = 124, .external_lex_state = 4},
  [654] = {.lex_state = 124, .external_lex_state = 4},
  [655] = {.lex_state = 124, .external_lex_state = 4},
  [656] = {.lex_state = 124, .external_lex_state = 4},
  [657] = {.lex_state = 124, .external_lex_state = 3},
  [658] = {.lex_state = 124, .external_lex_state = 4},
  [659] = {.lex_state = 124, .external_lex_state = 4},
  [660] = {.lex_state = 124, .external_lex_state = 4},
  [661] = {.lex_state = 124, .external_lex_state = 4},
  [662] = {.lex_state = 124, .external_lex_state = 4},
  [663] = {.lex_state = 124, .external_lex_state = 4},
  [664] = {.lex_state = 124, .external_lex_state = 4},
  [665] = {.lex_state = 124, .external_lex_state = 4},
  [666] = {.lex_state = 124, .external_lex_state = 4},
  [667] = {.lex_state = 124, .external_lex_state = 4},
  [668] = {.lex_state = 124, .external_lex_state = 4},
  [669] = {.lex_state = 124, .external_lex_state = 4},
  [670] = {.lex_state = 124, .external_lex_state = 4},
  [671] = {.lex_state = 124, .external_lex_state = 4},
  [672] = {.lex_state = 124, .external_lex_state = 4},
  [673] = {.lex_state = 124, .external_lex_state = 4},
  [674] = {.lex_state = 124, .external_lex_state = 4},
  [675] = {.lex_state = 124, .external_lex_state = 4},
  [676] = {.lex_state = 124, .external_lex_state = 4},
  [677] = {.lex_state = 124, .external_lex_state = 4},
  [678] = {.lex_state = 124, .external_lex_state = 4},
  [679] = {.lex_state = 124, .external_lex_state = 4},
  [680] = {.lex_state = 124, .external_lex_state = 4},
  [681] = {.lex_state = 124, .external_lex_state = 3},
  [682] = {.lex_state = 124, .external_lex_state = 3},
  [683] = {.lex_state = 124, .external_lex_state = 3},
  [684] = {.lex_state = 124, .external_lex_state = 3},
  [685] = {.lex_state = 124, .external_lex_state = 3},
  [686] = {.lex_state = 124, .external_lex_state = 3},
  [687] = {.lex_state = 124, .external_lex_state = 3},
  [688] = {.lex_state = 124, .external_lex_state = 3},
  [689] = {.lex_state = 124, .external_lex_state = 3},
  [690] = {.lex_state = 124, .external_lex_state = 4},
  [691] = {.lex_state = 124, .external_lex_state = 3},
  [692] = {.lex_state = 124, .external_lex_state = 3},
  [693] = {.lex_state = 124, .external_lex_state = 3},
  [694] = {.lex_state = 124, .external_lex_state = 3},
  [695] = {.lex_state = 124, .external_lex_state = 3},
  [696] = {.lex_state = 124, .external_lex_state = 3},
  [697] = {.lex_state = 124, .external_lex_state = 3},
  [698] = {.lex_state = 124, .external_lex_state = 3},
  [699] = {.lex_state = 124, .external_lex_state = 3},
  [700] = {.lex_state = 124, .external_lex_state = 3},
  [701] = {.lex_state = 7, .external_lex_state = 3},
  [702] = {.lex_state = 124, .external_lex_state = 3},
  [703] = {.lex_state = 124, .external_lex_state = 3},
  [704] = {.lex_state = 124, .external_lex_state = 3},
  [705] = {.lex_state = 124, .external_lex_state = 3},
  [706] = {.lex_state = 124, .external_lex_state = 3},
  [707] = {.lex_state = 124, .external_lex_state = 3},
  [708] = {.lex_state = 124, .external_lex_state = 3},
  [709] = {.lex_state = 124, .external_lex_state = 3},
  [710] = {.lex_state = 124, .external_lex_state = 3},
  [711] = {.lex_state = 124, .external_lex_state = 3},
  [712] = {.lex_state = 124, .external_lex_state = 3},
  [713] = {.lex_state = 124, .external_lex_state = 3},
  [714] = {.lex_state = 124, .external_lex_state = 3},
  [715] = {.lex_state = 124, .external_lex_state = 3},
  [716] = {.lex_state = 124, .external_lex_state = 3},
  [717] = {.lex_state = 124, .external_lex_state = 3},
  [718] = {.lex_state = 124, .external_lex_state = 3},
  [719] = {.lex_state = 124, .external_lex_state = 3},
  [720] = {.lex_state = 124, .external_lex_state = 3},
  [721] = {.lex_state = 124, .external_lex_state = 3},
  [722] = {.lex_state = 124, .external_lex_state = 3},
  [723] = {.lex_state = 124, .external_lex_state = 3},
  [724] = {.lex_state = 124, .external_lex_state = 3},
  [725] = {.lex_state = 124, .external_lex_state = 3},
  [726] = {.lex_state = 124, .external_lex_state = 3},
  [727] = {.lex_state = 124, .external_lex_state = 3},
  [728] = {.lex_state = 124, .external_lex_state = 3},
  [729] = {.lex_state = 124, .external_lex_state = 3},
  [730] = {.lex_state = 124, .external_lex_state = 4},
  [731] = {.lex_state = 124, .external_lex_state = 4},
  [732] = {.lex_state = 124, .external_lex_state = 3},
  [733] = {.lex_state = 124, .external_lex_state = 4},
  [734] = {.lex_state = 124, .external_lex_state = 3},
  [735] = {.lex_state = 124, .external_lex_state = 3},
  [736] = {.lex_state = 124, .external_lex_state = 3},
  [737] = {.lex_state = 124, .external_lex_state = 3},
  [738] = {.lex_state = 124, .external_lex_state = 3},
  [739] = {.lex_state = 124, .external_lex_state = 3},
  [740] = {.lex_state = 124, .external_lex_state = 3},
  [741] = {.lex_state = 124, .external_lex_state = 3},
  [742] = {.lex_state = 124, .external_lex_state = 3},
  [743] = {.lex_state = 124, .external_lex_state = 3},
  [744] = {.lex_state = 124, .external_lex_state = 3},
  [745] = {.lex_state = 124, .external_lex_state = 3},
  [746] = {.lex_state = 124, .external_lex_state = 3},
  [747] = {.lex_state = 124, .external_lex_state = 3},
  [748] = {.lex_state = 124, .external_lex_state = 3},
  [749] = {.lex_state = 124, .external_lex_state = 3},
  [750] = {.lex_state = 124, .external_lex_state = 3},
  [751] = {.lex_state = 124, .external_lex_state = 3},
  [752] = {.lex_state = 124, .external_lex_state = 3},
  [753] = {.lex_state = 124, .external_lex_state = 3},
  [754] = {.lex_state = 124, .external_lex_state = 3},
  [755] = {.lex_state = 124, .external_lex_state = 3},
  [756] = {.lex_state = 124, .external_lex_state = 3},
  [757] = {.lex_state = 124, .external_lex_state = 3},
  [758] = {.lex_state = 124, .external_lex_state = 3},
  [759] = {.lex_state = 124, .external_lex_state = 3},
  [760] = {.lex_state = 124, .external_lex_state = 3},
  [761] = {.lex_state = 124, .external_lex_state = 4},
  [762] = {.lex_state = 124, .external_lex_state = 3},
  [763] = {.lex_state = 124, .external_lex_state = 3},
  [764] = {.lex_state = 124, .external_lex_state = 3},
  [765] = {.lex_state = 124, .external_lex_state = 4},
  [766] = {.lex_state = 124, .external_lex_state = 3},
  [767] = {.lex_state = 124, .external_lex_state = 3},
  [768] = {.lex_state = 124, .external_lex_state = 3},
  [769] = {.lex_state = 124, .external_lex_state = 3},
  [770] = {.lex_state = 124, .external_lex_state = 3},
  [771] = {.lex_state = 124, .external_lex_state = 3},
  [772] = {.lex_state = 124, .external_lex_state = 3},
  [773] = {.lex_state = 124, .external_lex_state = 3},
  [774] = {.lex_state = 124, .external_lex_state = 3},
  [775] = {.lex_state = 124, .external_lex_state = 3},
  [776] = {.lex_state = 124, .external_lex_state = 3},
  [777] = {.lex_state = 124, .external_lex_state = 3},
  [778] = {.lex_state = 124, .external_lex_state = 3},
  [779] = {.lex_state = 124, .external_lex_state = 3},
  [780] = {.lex_state = 124, .external_lex_state = 3},
  [781] = {.lex_state = 124, .external_lex_state = 3},
  [782] = {.lex_state = 124, .external_lex_state = 3},
  [783] = {.lex_state = 124, .external_lex_state = 3},
  [784] = {.lex_state = 124, .external_lex_state = 3},
  [785] = {.lex_state = 124, .external_lex_state = 3},
  [786] = {.lex_state = 124, .external_lex_state = 3},
  [787] = {.lex_state = 124, .external_lex_state = 3},
  [788] = {.lex_state = 124, .external_lex_state = 3},
  [789] = {.lex_state = 124, .external_lex_state = 3},
  [790] = {.lex_state = 124, .external_lex_state = 3},
  [791] = {.lex_state = 124, .external_lex_state = 3},
  [792] = {.lex_state = 124, .external_lex_state = 3},
  [793] = {.lex_state = 124, .external_lex_state = 3},
  [794] = {.lex_state = 124, .external_lex_state = 3},
  [795] = {.lex_state = 7, .external_lex_state = 3},
  [796] = {.lex_state = 124, .external_lex_state = 3},
  [797] = {.lex_state = 124, .external_lex_state = 3},
  [798] = {.lex_state = 124, .external_lex_state = 3},
  [799] = {.lex_state = 124, .external_lex_state = 3},
  [800] = {.lex_state = 124, .external_lex_state = 3},
  [801] = {.lex_state = 124, .external_lex_state = 3},
  [802] = {.lex_state = 124, .external_lex_state = 3},
  [803] = {.lex_state = 124, .external_lex_state = 3},
  [804] = {.lex_state = 124, .external_lex_state = 4},
  [805] = {.lex_state = 124, .external_lex_state = 3},
  [806] = {.lex_state = 124, .external_lex_state = 3},
  [807] = {.lex_state = 124, .external_lex_state = 3},
  [808] = {.lex_state = 124, .external_lex_state = 3},
  [809] = {.lex_state = 124, .external_lex_state = 3},
  [810] = {.lex_state = 124, .external_lex_state = 3},
  [811] = {.lex_state = 124, .external_lex_state = 3},
  [812] = {.lex_state = 124, .external_lex_state = 4},
  [813] = {.lex_state = 124, .external_lex_state = 3},
  [814] = {.lex_state = 124, .external_lex_state = 3},
  [815] = {.lex_state = 124, .external_lex_state = 3},
  [816] = {.lex_state = 124, .external_lex_state = 3},
  [817] = {.lex_state = 124, .external_lex_state = 4},
  [818] = {.lex_state = 124, .external_lex_state = 3},
  [819] = {.lex_state = 124, .external_lex_state = 4},
  [820] = {.lex_state = 124, .external_lex_state = 4},
  [821] = {.lex_state = 124, .external_lex_state = 3},
  [822] = {.lex_state = 124, .external_lex_state = 3},
  [823] = {.lex_state = 124, .external_lex_state = 3},
  [824] = {.lex_state = 124, .external_lex_state = 4},
  [825] = {.lex_state = 124, .external_lex_state = 4},
  [826] = {.lex_state = 124, .external_lex_state = 3},
  [827] = {.lex_state = 124, .external_lex_state = 4},
  [828] = {.lex_state = 5, .external_lex_state = 2},
  [829] = {.lex_state = 124, .external_lex_state = 4},
  [830] = {.lex_state = 124, .external_lex_state = 4},
  [831] = {.lex_state = 124, .external_lex_state = 4},
  [832] = {.lex_state = 124, .external_lex_state = 4},
  [833] = {.lex_state = 124, .external_lex_state = 4},
  [834] = {.lex_state = 124, .external_lex_state = 4},
  [835] = {.lex_state = 124, .external_lex_state = 4},
  [836] = {.lex_state = 124, .external_lex_state = 4},
  [837] = {.lex_state = 124, .external_lex_state = 4},
  [838] = {.lex_state = 124, .external_lex_state = 4},
  [839] = {.lex_state = 124, .external_lex_state = 4},
  [840] = {.lex_state = 124, .external_lex_state = 4},
  [841] = {.lex_state = 124, .external_lex_state = 4},
  [842] = {.lex_state = 124, .external_lex_state = 4},
  [843] = {.lex_state = 124, .external_lex_state = 4},
  [844] = {.lex_state = 124, .external_lex_state = 4},
  [845] = {.lex_state = 124, .external_lex_state = 4},
  [846] = {.lex_state = 124, .external_lex_state = 4},
  [847] = {.lex_state = 124, .external_lex_state = 4},
  [848] = {.lex_state = 124, .external_lex_state = 3},
  [849] = {.lex_state = 124, .external_lex_state = 4},
  [850] = {.lex_state = 124, .external_lex_state = 4},
  [851] = {.lex_state = 124, .external_lex_state = 4},
  [852] = {.lex_state = 124, .external_lex_state = 4},
  [853] = {.lex_state = 124, .external_lex_state = 4},
  [854] = {.lex_state = 124, .external_lex_state = 4},
  [855] = {.lex_state = 124, .external_lex_state = 3},
  [856] = {.lex_state = 5, .external_lex_state = 2},
  [857] = {.lex_state = 124, .external_lex_state = 4},
  [858] = {.lex_state = 124, .external_lex_state = 4},
  [859] = {.lex_state = 124, .external_lex_state = 4},
  [860] = {.lex_state = 124, .external_lex_state = 4},
  [861] = {.lex_state = 124, .external_lex_state = 4},
  [862] = {.lex_state = 124, .external_lex_state = 4},
  [863] = {.lex_state = 124, .external_lex_state = 4},
  [864] = {.lex_state = 124, .external_lex_state = 4},
  [865] = {.lex_state = 5, .external_lex_state = 2},
  [866] = {.lex_state = 124, .external_lex_state = 4},
  [867] = {.lex_state = 124, .external_lex_state = 4},
  [868] = {.lex_state = 5, .external_lex_state = 2},
  [869] = {.lex_state = 124, .external_lex_state = 4},
  [870] = {.lex_state = 5, .external_lex_state = 2},
  [871] = {.lex_state = 5, .external_lex_state = 2},
  [872] = {.lex_state = 124, .external_lex_state = 4},
  [873] = {.lex_state = 124, .external_lex_state = 4},
  [874] = {.lex_state = 124, .external_lex_state = 4},
  [875] = {.lex_state = 124, .external_lex_state = 4},
  [876] = {.lex_state = 124, .external_lex_state = 4},
  [877] = {.lex_state = 124, .external_lex_state = 4},
  [878] = {.lex_state = 124, .external_lex_state = 4},
  [879] = {.lex_state = 124, .external_lex_state = 4},
  [880] = {.lex_state = 124, .external_lex_state = 4},
  [881] = {.lex_state = 124, .external_lex_state = 4},
  [882] = {.lex_state = 124, .external_lex_state = 4},
  [883] = {.lex_state = 124, .external_lex_state = 4},
  [884] = {.lex_state = 124, .external_lex_state = 4},
  [885] = {.lex_state = 124, .external_lex_state = 4},
  [886] = {.lex_state = 124, .external_lex_state = 4},
  [887] = {.lex_state = 124, .external_lex_state = 4},
  [888] = {.lex_state = 7, .external_lex_state = 4},
  [889] = {.lex_state = 124, .external_lex_state = 3},
  [890] = {.lex_state = 124, .external_lex_state = 4},
  [891] = {.lex_state = 124, .external_lex_state = 4},
  [892] = {.lex_state = 124, .external_lex_state = 4},
  [893] = {.lex_state = 124, .external_lex_state = 4},
  [894] = {.lex_state = 124, .external_lex_state = 4},
  [895] = {.lex_state = 124, .external_lex_state = 4},
  [896] = {.lex_state = 124, .external_lex_state = 4},
  [897] = {.lex_state = 124, .external_lex_state = 4},
  [898] = {.lex_state = 124, .external_lex_state = 4},
  [899] = {.lex_state = 124, .external_lex_state = 4},
  [900] = {.lex_state = 124, .external_lex_state = 4},
  [901] = {.lex_state = 124, .external_lex_state = 4},
  [902] = {.lex_state = 124, .external_lex_state = 4},
  [903] = {.lex_state = 124, .external_lex_state = 4},
  [904] = {.lex_state = 124, .external_lex_state = 4},
  [905] = {.lex_state = 124, .external_lex_state = 4},
  [906] = {.lex_state = 124, .external_lex_state = 4},
  [907] = {.lex_state = 124, .external_lex_state = 4},
  [908] = {.lex_state = 124, .external_lex_state = 4},
  [909] = {.lex_state = 124, .external_lex_state = 4},
  [910] = {.lex_state = 124, .external_lex_state = 4},
  [911] = {.lex_state = 124, .external_lex_state = 4},
  [912] = {.lex_state = 124, .external_lex_state = 4},
  [913] = {.lex_state = 124, .external_lex_state = 4},
  [914] = {.lex_state = 124, .external_lex_state = 3},
  [915] = {.lex_state = 124, .external_lex_state = 4},
  [916] = {.lex_state = 124, .external_lex_state = 4},
  [917] = {.lex_state = 124, .external_lex_state = 4},
  [918] = {.lex_state = 124, .external_lex_state = 4},
  [919] = {.lex_state = 124, .external_lex_state = 4},
  [920] = {.lex_state = 124, .external_lex_state = 4},
  [921] = {.lex_state = 124, .external_lex_state = 4},
  [922] = {.lex_state = 124, .external_lex_state = 4},
  [923] = {.lex_state = 124, .external_lex_state = 4},
  [924] = {.lex_state = 124, .external_lex_state = 4},
  [925] = {.lex_state = 124, .external_lex_state = 4},
  [926] = {.lex_state = 124, .external_lex_state = 4},
  [927] = {.lex_state = 124, .external_lex_state = 4},
  [928] = {.lex_state = 124, .external_lex_state = 4},
  [929] = {.lex_state = 124, .external_lex_state = 4},
  [930] = {.lex_state = 124, .external_lex_state = 4},
  [931] = {.lex_state = 124, .external_lex_state = 4},
  [932] = {.lex_state = 124, .external_lex_state = 4},
  [933] = {.lex_state = 124, .external_lex_state = 4},
  [934] = {.lex_state = 124, .external_lex_state = 4},
  [935] = {.lex_state = 124, .external_lex_state = 4},
  [936] = {.lex_state = 124, .external_lex_state = 4},
  [937] = {.lex_state = 124, .external_lex_state = 4},
  [938] = {.lex_state = 124, .external_lex_state = 4},
  [939] = {.lex_state = 124, .external_lex_state = 4},
  [940] = {.lex_state = 124, .external_lex_state = 4},
  [941] = {.lex_state = 124, .external_lex_state = 4},
  [942] = {.lex_state = 124, .external_lex_state = 4},
  [943] = {.lex_state = 124, .external_lex_state = 4},
  [944] = {.lex_state = 124, .external_lex_state = 4},
  [945] = {.lex_state = 124, .external_lex_state = 4},
  [946] = {.lex_state = 124, .external_lex_state = 4},
  [947] = {.lex_state = 124, .external_lex_state = 4},
  [948] = {.lex_state = 124, .external_lex_state = 4},
  [949] = {.lex_state = 124, .external_lex_state = 4},
  [950] = {.lex_state = 124, .external_lex_state = 4},
  [951] = {.lex_state = 124, .external_lex_state = 4},
  [952] = {.lex_state = 124, .external_lex_state = 4},
  [953] = {.lex_state = 124, .external_lex_state = 4},
  [954] = {.lex_state = 124, .external_lex_state = 4},
  [955] = {.lex_state = 124, .external_lex_state = 4},
  [956] = {.lex_state = 124, .external_lex_state = 4},
  [957] = {.lex_state = 124, .external_lex_state = 4},
  [958] = {.lex_state = 124, .external_lex_state = 4},
  [959] = {.lex_state = 124, .external_lex_state = 4},
  [960] = {.lex_state = 124, .external_lex_state = 4},
  [961] = {.lex_state = 124, .external_lex_state = 4},
  [962] = {.lex_state = 124, .external_lex_state = 4},
  [963] = {.lex_state = 124, .external_lex_state = 4},
  [964] = {.lex_state = 124, .external_lex_state = 4},
  [965] = {.lex_state = 124, .external_lex_state = 4},
  [966] = {.lex_state = 124, .external_lex_state = 4},
  [967] = {.lex_state = 124, .external_lex_state = 4},
  [968] = {.lex_state = 124, .external_lex_state = 4},
  [969] = {.lex_state = 5, .external_lex_state = 2},
  [970] = {.lex_state = 124, .external_lex_state = 4},
  [971] = {.lex_state = 124, .external_lex_state = 4},
  [972] = {.lex_state = 124, .external_lex_state = 4},
  [973] = {.lex_state = 124, .external_lex_state = 4},
  [974] = {.lex_state = 124, .external_lex_state = 4},
  [975] = {.lex_state = 124, .external_lex_state = 4},
  [976] = {.lex_state = 124, .external_lex_state = 4},
  [977] = {.lex_state = 124, .external_lex_state = 4},
  [978] = {.lex_state = 124, .external_lex_state = 4},
  [979] = {.lex_state = 124, .external_lex_state = 4},
  [980] = {.lex_state = 124, .external_lex_state = 4},
  [981] = {.lex_state = 5, .external_lex_state = 2},
  [982] = {.lex_state = 5, .external_lex_state = 2},
  [983] = {.lex_state = 5, .external_lex_state = 2},
  [984] = {.lex_state = 5, .external_lex_state = 2},
  [985] = {.lex_state = 5, .external_lex_state = 2},
  [986] = {.lex_state = 125, .external_lex_state = 2},
  [987] = {.lex_state = 125, .external_lex_state = 2},
  [988] = {.lex_state = 5, .external_lex_state = 2},
  [989] = {.lex_state = 5, .external_lex_state = 2},
  [990] = {.lex_state = 5, .external_lex_state = 2},
  [991] = {.lex_state = 5, .external_lex_state = 2},
  [992] = {.lex_state = 125, .external_lex_state = 2},
  [993] = {.lex_state = 125, .external_lex_state = 2},
  [994] = {.lex_state = 125, .external_lex_state = 2},
  [995] = {.lex_state = 125, .external_lex_state = 2},
  [996] = {.lex_state = 125, .external_lex_state = 2},
  [997] = {.lex_state = 125, .external_lex_state = 2},
  [998] = {.lex_state = 125, .external_lex_state = 2},
  [999] = {.lex_state = 125, .external_lex_state = 2},
  [1000] = {.lex_state = 125, .external_lex_state = 2},
  [1001] = {.lex_state = 125, .external_lex_state = 2},
  [1002] = {.lex_state = 125, .external_lex_state = 2},
  [1003] = {.lex_state = 125, .external_lex_state = 2},
  [1004] = {.lex_state = 125, .external_lex_state = 2},
  [1005] = {.lex_state = 125, .external_lex_state = 2},
  [1006] = {.lex_state = 125, .external_lex_state = 2},
  [1007] = {.lex_state = 125, .external_lex_state = 2},
  [1008] = {.lex_state = 125, .external_lex_state = 2},
  [1009] = {.lex_state = 125, .external_lex_state = 5},
  [1010] = {.lex_state = 125, .external_lex_state = 2},
  [1011] = {.lex_state = 125, .external_lex_state = 2},
  [1012] = {.lex_state = 125, .external_lex_state = 2},
  [1013] = {.lex_state = 125, .external_lex_state = 2},
  [1014] = {.lex_state = 5, .external_lex_state = 2},
  [1015] = {.lex_state = 125, .external_lex_state = 2},
  [1016] = {.lex_state = 125, .external_lex_state = 2},
  [1017] = {.lex_state = 125, .external_lex_state = 5},
  [1018] = {.lex_state = 125, .external_lex_state = 5},
  [1019] = {.lex_state = 5, .external_lex_state = 2},
  [1020] = {.lex_state = 125, .external_lex_state = 2},
  [1021] = {.lex_state = 125, .external_lex_state = 2},
  [1022] = {.lex_state = 125, .external_lex_state = 5},
  [1023] = {.lex_state = 125, .external_lex_state = 5},
  [1024] = {.lex_state = 125, .external_lex_state = 2},
  [1025] = {.lex_state = 5, .external_lex_state = 2},
  [1026] = {.lex_state = 125, .external_lex_state = 5},
  [1027] = {.lex_state = 125, .external_lex_state = 2},
  [1028] = {.lex_state = 125, .external_lex_state = 5},
  [1029] = {.lex_state = 125, .external_lex_state = 5},
  [1030] = {.lex_state = 125, .external_lex_state = 5},
  [1031] = {.lex_state = 125, .external_lex_state = 5},
  [1032] = {.lex_state = 125, .external_lex_state = 5},
  [1033] = {.lex_state = 125, .external_lex_state = 5},
  [1034] = {.lex_state = 9, .external_lex_state = 2},
  [1035] = {.lex_state = 125, .external_lex_state = 2},
  [1036] = {.lex_state = 125, .external_lex_state = 2},
  [1037] = {.lex_state = 5, .external_lex_state = 2},
  [1038] = {.lex_state = 125, .external_lex_state = 2},
  [1039] = {.lex_state = 5, .external_lex_state = 2},
  [1040] = {.lex_state = 5, .external_lex_state = 2},
  [1041] = {.lex_state = 5, .external_lex_state = 2},
  [1042] = {.lex_state = 5, .external_lex_state = 5},
  [1043] = {.lex_state = 9, .external_lex_state = 2},
  [1044] = {.lex_state = 5, .external_lex_state = 2},
  [1045] = {.lex_state = 5, .external_lex_state = 2},
  [1046] = {.lex_state = 5, .external_lex_state = 2},
  [1047] = {.lex_state = 5, .external_lex_state = 5},
  [1048] = {.lex_state = 5, .external_lex_state = 2},
  [1049] = {.lex_state = 5, .external_lex_state = 2},
  [1050] = {.lex_state = 5, .external_lex_state = 2},
  [1051] = {.lex_state = 5, .external_lex_state = 2},
  [1052] = {.lex_state = 5, .external_lex_state = 2},
  [1053] = {.lex_state = 5, .external_lex_state = 2},
  [1054] = {.lex_state = 5, .external_lex_state = 2},
  [1055] = {.lex_state = 5, .external_lex_state = 2},
  [1056] = {.lex_state = 5, .external_lex_state = 2},
  [1057] = {.lex_state = 5, .external_lex_state = 2},
  [1058] = {.lex_state = 5, .external_lex_state = 2},
  [1059] = {.lex_state = 5, .external_lex_state = 2},
  [1060] = {.lex_state = 5, .external_lex_state = 2},
  [1061] = {.lex_state = 5, .external_lex_state = 2},
  [1062] = {.lex_state = 5, .external_lex_state = 2},
  [1063] = {.lex_state = 5, .external_lex_state = 2},
  [1064] = {.lex_state = 5, .external_lex_state = 2},
  [1065] = {.lex_state = 5, .external_lex_state = 2},
  [1066] = {.lex_state = 5, .external_lex_state = 2},
  [1067] = {.lex_state = 125, .external_lex_state = 2},
  [1068] = {.lex_state = 5, .external_lex_state = 2},
  [1069] = {.lex_state = 5, .external_lex_state = 2},
  [1070] = {.lex_state = 5, .external_lex_state = 2},
  [1071] = {.lex_state = 5, .external_lex_state = 2},
  [1072] = {.lex_state = 125, .external_lex_state = 2},
  [1073] = {.lex_state = 5, .external_lex_state = 2},
  [1074] = {.lex_state = 125, .external_lex_state = 2},
  [1075] = {.lex_state = 5, .external_lex_state = 2},
  [1076] = {.lex_state = 125, .external_lex_state = 2},
  [1077] = {.lex_state = 125, .external_lex_state = 2},
  [1078] = {.lex_state = 5, .external_lex_state = 2},
  [1079] = {.lex_state = 125, .external_lex_state = 2},
  [1080] = {.lex_state = 5, .external_lex_state = 2},
  [1081] = {.lex_state = 5, .external_lex_state = 2},
  [1082] = {.lex_state = 125, .external_lex_state = 2},
  [1083] = {.lex_state = 125, .external_lex_state = 2},
  [1084] = {.lex_state = 125, .external_lex_state = 2},
  [1085] = {.lex_state = 5, .external_lex_state = 2},
  [1086] = {.lex_state = 125, .external_lex_state = 2},
  [1087] = {.lex_state = 125, .external_lex_state = 2},
  [1088] = {.lex_state = 125, .external_lex_state = 2},
  [1089] = {.lex_state = 125, .external_lex_state = 2},
  [1090] = {.lex_state = 125, .external_lex_state = 2},
  [1091] = {.lex_state = 125, .external_lex_state = 2},
  [1092] = {.lex_state = 125, .external_lex_state = 2},
  [1093] = {.lex_state = 125, .external_lex_state = 2},
  [1094] = {.lex_state = 5, .external_lex_state = 2},
  [1095] = {.lex_state = 125, .external_lex_state = 2},
  [1096] = {.lex_state = 125, .external_lex_state = 2},
  [1097] = {.lex_state = 5, .external_lex_state = 2},
  [1098] = {.lex_state = 125, .external_lex_state = 2},
  [1099] = {.lex_state = 125, .external_lex_state = 2},
  [1100] = {.lex_state = 5, .external_lex_state = 2},
  [1101] = {.lex_state = 125, .external_lex_state = 2},
  [1102] = {.lex_state = 125, .external_lex_state = 2},
  [1103] = {.lex_state = 125, .external_lex_state = 2},
  [1104] = {.lex_state = 125, .external_lex_state = 2},
  [1105] = {.lex_state = 125, .external_lex_state = 2},
  [1106] = {.lex_state = 5, .external_lex_state = 2},
  [1107] = {.lex_state = 125, .external_lex_state = 2},
  [1108] = {.lex_state = 125, .external_lex_state = 2},
  [1109] = {.lex_state = 125, .external_lex_state = 2},
  [1110] = {.lex_state = 125, .external_lex_state = 2},
  [1111] = {.lex_state = 125, .external_lex_state = 2},
  [1112] = {.lex_state = 125, .external_lex_state = 2},
  [1113] = {.lex_state = 125, .external_lex_state = 2},
  [1114] = {.lex_state = 125, .external_lex_state = 2},
  [1115] = {.lex_state = 125, .external_lex_state = 2},
  [1116] = {.lex_state = 125, .external_lex_state = 2},
  [1117] = {.lex_state = 125, .external_lex_state = 2},
  [1118] = {.lex_state = 5, .external_lex_state = 2},
  [1119] = {.lex_state = 5, .external_lex_state = 2},
  [1120] = {.lex_state = 125, .external_lex_state = 2},
  [1121] = {.lex_state = 125, .external_lex_state = 2},
  [1122] = {.lex_state = 125, .external_lex_state = 2},
  [1123] = {.lex_state = 125, .external_lex_state = 2},
  [1124] = {.lex_state = 125, .external_lex_state = 2},
  [1125] = {.lex_state = 125, .external_lex_state = 2},
  [1126] = {.lex_state = 125, .external_lex_state = 2},
  [1127] = {.lex_state = 125, .external_lex_state = 2},
  [1128] = {.lex_state = 125, .external_lex_state = 2},
  [1129] = {.lex_state = 125, .external_lex_state = 2},
  [1130] = {.lex_state = 125, .external_lex_state = 2},
  [1131] = {.lex_state = 125, .external_lex_state = 2},
  [1132] = {.lex_state = 125, .external_lex_state = 2},
  [1133] = {.lex_state = 125, .external_lex_state = 2},
  [1134] = {.lex_state = 125, .external_lex_state = 2},
  [1135] = {.lex_state = 125, .external_lex_state = 2},
  [1136] = {.lex_state = 125, .external_lex_state = 2},
  [1137] = {.lex_state = 125, .external_lex_state = 2},
  [1138] = {.lex_state = 3, .external_lex_state = 2},
  [1139] = {.lex_state = 3, .external_lex_state = 2},
  [1140] = {.lex_state = 3, .external_lex_state = 2},
  [1141] = {.lex_state = 3, .external_lex_state = 2},
  [1142] = {.lex_state = 3, .external_lex_state = 2},
  [1143] = {.lex_state = 3, .external_lex_state = 2},
  [1144] = {.lex_state = 3, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1145] = {.lex_state = 31, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1146] = {.lex_state = 9, .external_lex_state = 6},
  [1147] = {.lex_state = 31, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1148] = {.lex_state = 9, .external_lex_state = 6},
  [1149] = {.lex_state = 9, .external_lex_state = 6},
  [1150] = {.lex_state = 9, .external_lex_state = 6},
  [1151] = {.lex_state = 31, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1152] = {.lex_state = 31, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1153] = {.lex_state = 9, .external_lex_state = 6},
  [1154] = {.lex_state = 9, .external_lex_state = 6},
  [1155] = {.lex_state = 9, .external_lex_state = 6},
  [1156] = {.lex_state = 9, .external_lex_state = 6},
  [1157] = {.lex_state = 31, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1158] = {.lex_state = 31, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1159] = {.lex_state = 31, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1160] = {.lex_state = 9, .external_lex_state = 6},
  [1161] = {.lex_state = 31, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1162] = {.lex_state = 31, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1163] = {.lex_state = 31, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1164] = {.lex_state = 31, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1165] = {.lex_state = 31, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1166] = {.lex_state = 31, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1167] = {.lex_state = 31, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1168] = {.lex_state = 125, .external_lex_state = 2},
  [1169] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1170] = {.lex_state = 125, .external_lex_state = 2},
  [1171] = {.lex_state = 31, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1172] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1173] = {.lex_state = 31, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1174] = {.lex_state = 31, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1175] = {.lex_state = 31, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1176] = {.lex_state = 31, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1177] = {.lex_state = 31, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1178] = {.lex_state = 31, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1179] = {.lex_state = 31, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1180] = {.lex_state = 31, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1181] = {.lex_state = 31, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1182] = {.lex_state = 31, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1183] = {.lex_state = 31, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1184] = {.lex_state = 31, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1185] = {.lex_state = 31, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1186] = {.lex_state = 31, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1187] = {.lex_state = 31, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1188] = {.lex_state = 31, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1189] = {.lex_state = 31, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1190] = {.lex_state = 31, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1191] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1192] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1193] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1194] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1195] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1196] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1197] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1198] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1199] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1200] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1201] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1202] = {.lex_state = 125, .external_lex_state = 5},
  [1203] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1204] = {.lex_state = 125, .external_lex_state = 2},
  [1205] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1206] = {.lex_state = 125, .external_lex_state = 2},
  [1207] = {.lex_state = 125, .external_lex_state = 2},
  [1208] = {.lex_state = 125, .external_lex_state = 5},
  [1209] = {.lex_state = 125, .external_lex_state = 5},
  [1210] = {.lex_state = 125, .external_lex_state = 2},
  [1211] = {.lex_state = 125, .external_lex_state = 2},
  [1212] = {.lex_state = 125, .external_lex_state = 2},
  [1213] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1214] = {.lex_state = 125, .external_lex_state = 2},
  [1215] = {.lex_state = 31, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1216] = {.lex_state = 125, .external_lex_state = 5},
  [1217] = {.lex_state = 125, .external_lex_state = 2},
  [1218] = {.lex_state = 125, .external_lex_state = 2},
  [1219] = {.lex_state = 125, .external_lex_state = 2},
  [1220] = {.lex_state = 125, .external_lex_state = 2},
  [1221] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1222] = {.lex_state = 125, .external_lex_state = 2},
  [1223] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1224] = {.lex_state = 125, .external_lex_state = 2},
  [1225] = {.lex_state = 31, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1226] = {.lex_state = 31, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1227] = {.lex_state = 125, .external_lex_state = 5},
  [1228] = {.lex_state = 125, .external_lex_state = 5},
  [1229] = {.lex_state = 16, .external_lex_state = 7},
  [1230] = {.lex_state = 31, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1231] = {.lex_state = 125, .external_lex_state = 5},
  [1232] = {.lex_state = 125, .external_lex_state = 5},
  [1233] = {.lex_state = 16, .external_lex_state = 7},
  [1234] = {.lex_state = 125, .external_lex_state = 5},
  [1235] = {.lex_state = 125, .external_lex_state = 2},
  [1236] = {.lex_state = 125, .external_lex_state = 2},
  [1237] = {.lex_state = 125, .external_lex_state = 2},
  [1238] = {.lex_state = 125, .external_lex_state = 5},
  [1239] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1240] = {.lex_state = 16, .external_lex_state = 7},
  [1241] = {.lex_state = 125, .external_lex_state = 5},
  [1242] = {.lex_state = 16, .external_lex_state = 7},
  [1243] = {.lex_state = 125, .external_lex_state = 2},
  [1244] = {.lex_state = 16, .external_lex_state = 7},
  [1245] = {.lex_state = 125, .external_lex_state = 5},
  [1246] = {.lex_state = 125, .external_lex_state = 2},
  [1247] = {.lex_state = 125, .external_lex_state = 5},
  [1248] = {.lex_state = 125, .external_lex_state = 5},
  [1249] = {.lex_state = 125, .external_lex_state = 5},
  [1250] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1251] = {.lex_state = 125, .external_lex_state = 5},
  [1252] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1253] = {.lex_state = 125, .external_lex_state = 2},
  [1254] = {.lex_state = 9, .external_lex_state = 6},
  [1255] = {.lex_state = 9, .external_lex_state = 6},
  [1256] = {.lex_state = 31, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1257] = {.lex_state = 31, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1258] = {.lex_state = 31, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1259] = {.lex_state = 31, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1260] = {.lex_state = 31, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1261] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 13},
  [1262] = {.lex_state = 125, .external_lex_state = 2},
  [1263] = {.lex_state = 31, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1264] = {.lex_state = 31, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1265] = {.lex_state = 31, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1266] = {.lex_state = 9, .external_lex_state = 6},
  [1267] = {.lex_state = 125, .external_lex_state = 2},
  [1268] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 13},
  [1269] = {.lex_state = 31, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1270] = {.lex_state = 31, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1271] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 13},
  [1272] = {.lex_state = 9, .external_lex_state = 6},
  [1273] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 13},
  [1274] = {.lex_state = 125, .external_lex_state = 5},
  [1275] = {.lex_state = 9, .external_lex_state = 6},
  [1276] = {.lex_state = 31, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1277] = {.lex_state = 125, .external_lex_state = 2},
  [1278] = {.lex_state = 9, .external_lex_state = 6},
  [1279] = {.lex_state = 31, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1280] = {.lex_state = 125, .external_lex_state = 2},
  [1281] = {.lex_state = 9, .external_lex_state = 6},
  [1282] = {.lex_state = 125, .external_lex_state = 2},
  [1283] = {.lex_state = 125, .external_lex_state = 5},
  [1284] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 13},
  [1285] = {.lex_state = 125, .external_lex_state = 5},
  [1286] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 13},
  [1287] = {.lex_state = 9, .external_lex_state = 6},
  [1288] = {.lex_state = 31, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1289] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 13},
  [1290] = {.lex_state = 9, .external_lex_state = 6},
  [1291] = {.lex_state = 31, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1292] = {.lex_state = 31, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1293] = {.lex_state = 125, .external_lex_state = 2},
  [1294] = {.lex_state = 9, .external_lex_state = 6},
  [1295] = {.lex_state = 31, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1296] = {.lex_state = 125, .external_lex_state = 2},
  [1297] = {.lex_state = 125, .external_lex_state = 2},
  [1298] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 13},
  [1299] = {.lex_state = 9, .external_lex_state = 6},
  [1300] = {.lex_state = 9, .external_lex_state = 6},
  [1301] = {.lex_state = 9, .external_lex_state = 6},
  [1302] = {.lex_state = 9, .external_lex_state = 6},
  [1303] = {.lex_state = 31, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1304] = {.lex_state = 125, .external_lex_state = 5},
  [1305] = {.lex_state = 31, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1306] = {.lex_state = 125, .external_lex_state = 2},
  [1307] = {.lex_state = 125, .external_lex_state = 2},
  [1308] = {.lex_state = 9, .external_lex_state = 6},
  [1309] = {.lex_state = 9, .external_lex_state = 6},
  [1310] = {.lex_state = 125, .external_lex_state = 2},
  [1311] = {.lex_state = 125, .external_lex_state = 5},
  [1312] = {.lex_state = 31, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1313] = {.lex_state = 31, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1314] = {.lex_state = 31, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1315] = {.lex_state = 31, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1316] = {.lex_state = 125, .external_lex_state = 5},
  [1317] = {.lex_state = 9, .external_lex_state = 6},
  [1318] = {.lex_state = 31, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1319] = {.lex_state = 31, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1320] = {.lex_state = 10, .external_lex_state = 2},
  [1321] = {.lex_state = 125, .external_lex_state = 5},
  [1322] = {.lex_state = 125, .external_lex_state = 2},
  [1323] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1324] = {.lex_state = 125, .external_lex_state = 2},
  [1325] = {.lex_state = 125, .external_lex_state = 5},
  [1326] = {.lex_state = 3, .external_lex_state = 2},
  [1327] = {.lex_state = 125, .external_lex_state = 2},
  [1328] = {.lex_state = 12, .external_lex_state = 8},
  [1329] = {.lex_state = 20, .external_lex_state = 8},
  [1330] = {.lex_state = 12, .external_lex_state = 8},
  [1331] = {.lex_state = 20, .external_lex_state = 8},
  [1332] = {.lex_state = 12, .external_lex_state = 8},
  [1333] = {.lex_state = 3, .external_lex_state = 2},
  [1334] = {.lex_state = 20, .external_lex_state = 8},
  [1335] = {.lex_state = 125, .external_lex_state = 5},
  [1336] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1337] = {.lex_state = 125, .external_lex_state = 2},
  [1338] = {.lex_state = 125, .external_lex_state = 2},
  [1339] = {.lex_state = 12, .external_lex_state = 8},
  [1340] = {.lex_state = 125, .external_lex_state = 2},
  [1341] = {.lex_state = 125, .external_lex_state = 5},
  [1342] = {.lex_state = 125, .external_lex_state = 5},
  [1343] = {.lex_state = 125, .external_lex_state = 2},
  [1344] = {.lex_state = 125, .external_lex_state = 5},
  [1345] = {.lex_state = 125, .external_lex_state = 2},
  [1346] = {.lex_state = 125, .external_lex_state = 2},
  [1347] = {.lex_state = 125, .external_lex_state = 2},
  [1348] = {.lex_state = 18, .external_lex_state = 2},
  [1349] = {.lex_state = 12, .external_lex_state = 8},
  [1350] = {.lex_state = 12, .external_lex_state = 8},
  [1351] = {.lex_state = 20, .external_lex_state = 8},
  [1352] = {.lex_state = 20, .external_lex_state = 8},
  [1353] = {.lex_state = 125, .external_lex_state = 2},
  [1354] = {.lex_state = 125, .external_lex_state = 2},
  [1355] = {.lex_state = 18, .external_lex_state = 2},
  [1356] = {.lex_state = 10, .external_lex_state = 2},
  [1357] = {.lex_state = 12, .external_lex_state = 8},
  [1358] = {.lex_state = 125, .external_lex_state = 5},
  [1359] = {.lex_state = 125, .external_lex_state = 5},
  [1360] = {.lex_state = 125, .external_lex_state = 2},
  [1361] = {.lex_state = 12, .external_lex_state = 8},
  [1362] = {.lex_state = 125, .external_lex_state = 2},
  [1363] = {.lex_state = 20, .external_lex_state = 8},
  [1364] = {.lex_state = 20, .external_lex_state = 8},
  [1365] = {.lex_state = 125, .external_lex_state = 5},
  [1366] = {.lex_state = 125, .external_lex_state = 2},
  [1367] = {.lex_state = 125, .external_lex_state = 2},
  [1368] = {.lex_state = 125, .external_lex_state = 2},
  [1369] = {.lex_state = 20, .external_lex_state = 8},
  [1370] = {.lex_state = 125, .external_lex_state = 2},
  [1371] = {.lex_state = 125, .external_lex_state = 2},
  [1372] = {.lex_state = 125, .external_lex_state = 5},
  [1373] = {.lex_state = 125, .external_lex_state = 2},
  [1374] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1375] = {.lex_state = 16, .external_lex_state = 7},
  [1376] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1377] = {.lex_state = 125, .external_lex_state = 2},
  [1378] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1379] = {.lex_state = 125, .external_lex_state = 5},
  [1380] = {.lex_state = 125, .external_lex_state = 5},
  [1381] = {.lex_state = 125, .external_lex_state = 2},
  [1382] = {.lex_state = 125, .external_lex_state = 2},
  [1383] = {.lex_state = 125, .external_lex_state = 2},
  [1384] = {.lex_state = 10, .external_lex_state = 2},
  [1385] = {.lex_state = 125, .external_lex_state = 5},
  [1386] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1387] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1388] = {.lex_state = 125, .external_lex_state = 2},
  [1389] = {.lex_state = 125, .external_lex_state = 5},
  [1390] = {.lex_state = 125, .external_lex_state = 5},
  [1391] = {.lex_state = 125, .external_lex_state = 5},
  [1392] = {.lex_state = 125, .external_lex_state = 2},
  [1393] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1394] = {.lex_state = 125, .external_lex_state = 2},
  [1395] = {.lex_state = 12, .external_lex_state = 8},
  [1396] = {.lex_state = 20, .external_lex_state = 8},
  [1397] = {.lex_state = 125, .external_lex_state = 2},
  [1398] = {.lex_state = 18, .external_lex_state = 2},
  [1399] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1400] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1401] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1402] = {.lex_state = 125, .external_lex_state = 2},
  [1403] = {.lex_state = 125, .external_lex_state = 2},
  [1404] = {.lex_state = 3, .external_lex_state = 2},
  [1405] = {.lex_state = 125, .external_lex_state = 2},
  [1406] = {.lex_state = 125, .external_lex_state = 2},
  [1407] = {.lex_state = 125, .external_lex_state = 2},
  [1408] = {.lex_state = 125, .external_lex_state = 5},
  [1409] = {.lex_state = 125, .external_lex_state = 2},
  [1410] = {.lex_state = 125, .external_lex_state = 2},
  [1411] = {.lex_state = 125, .external_lex_state = 2},
  [1412] = {.lex_state = 125, .external_lex_state = 2},
  [1413] = {.lex_state = 125, .external_lex_state = 2},
  [1414] = {.lex_state = 125, .external_lex_state = 2},
  [1415] = {.lex_state = 125, .external_lex_state = 5},
  [1416] = {.lex_state = 125, .external_lex_state = 2},
  [1417] = {.lex_state = 125, .external_lex_state = 2},
  [1418] = {.lex_state = 125, .external_lex_state = 2},
  [1419] = {.lex_state = 125, .external_lex_state = 2},
  [1420] = {.lex_state = 125, .external_lex_state = 2},
  [1421] = {.lex_state = 125, .external_lex_state = 2},
  [1422] = {.lex_state = 125, .external_lex_state = 2},
  [1423] = {.lex_state = 125, .external_lex_state = 2},
  [1424] = {.lex_state = 6, .external_lex_state = 2},
  [1425] = {.lex_state = 125, .external_lex_state = 5},
  [1426] = {.lex_state = 125, .external_lex_state = 2},
  [1427] = {.lex_state = 125, .external_lex_state = 2},
  [1428] = {.lex_state = 6, .external_lex_state = 2},
  [1429] = {.lex_state = 125, .external_lex_state = 2},
  [1430] = {.lex_state = 125, .external_lex_state = 2},
  [1431] = {.lex_state = 125, .external_lex_state = 2},
  [1432] = {.lex_state = 125, .external_lex_state = 2},
  [1433] = {.lex_state = 125, .external_lex_state = 2},
  [1434] = {.lex_state = 125, .external_lex_state = 2},
  [1435] = {.lex_state = 125, .external_lex_state = 2},
  [1436] = {.lex_state = 125, .external_lex_state = 2},
  [1437] = {.lex_state = 125, .external_lex_state = 2},
  [1438] = {.lex_state = 125, .external_lex_state = 2},
  [1439] = {.lex_state = 125, .external_lex_state = 5},
  [1440] = {.lex_state = 125, .external_lex_state = 2},
  [1441] = {.lex_state = 125, .external_lex_state = 2},
  [1442] = {.lex_state = 125, .external_lex_state = 2},
  [1443] = {.lex_state = 125, .external_lex_state = 2},
  [1444] = {.lex_state = 6, .external_lex_state = 2},
  [1445] = {.lex_state = 125, .external_lex_state = 2},
  [1446] = {.lex_state = 125, .external_lex_state = 5, .reserved_word_set_id = 1},
  [1447] = {.lex_state = 125, .external_lex_state = 5},
  [1448] = {.lex_state = 125, .external_lex_state = 2},
  [1449] = {.lex_state = 125, .external_lex_state = 2},
  [1450] = {.lex_state = 125, .external_lex_state = 2},
  [1451] = {.lex_state = 125, .external_lex_state = 2},
  [1452] = {.lex_state = 125, .external_lex_state = 2},
  [1453] = {.lex_state = 125, .external_lex_state = 5},
  [1454] = {.lex_state = 125, .external_lex_state = 2},
  [1455] = {.lex_state = 125, .external_lex_state = 2},
  [1456] = {.lex_state = 125, .external_lex_state = 5, .reserved_word_set_id = 1},
  [1457] = {.lex_state = 125, .external_lex_state = 2},
  [1458] = {.lex_state = 125, .external_lex_state = 2},
  [1459] = {.lex_state = 125, .external_lex_state = 2},
  [1460] = {.lex_state = 125, .external_lex_state = 2},
  [1461] = {.lex_state = 125, .external_lex_state = 2},
  [1462] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1463] = {.lex_state = 125, .external_lex_state = 2},
  [1464] = {.lex_state = 125, .external_lex_state = 2},
  [1465] = {.lex_state = 125, .external_lex_state = 2},
  [1466] = {.lex_state = 125, .external_lex_state = 2},
  [1467] = {.lex_state = 125, .external_lex_state = 2},
  [1468] = {.lex_state = 125, .external_lex_state = 5},
  [1469] = {.lex_state = 125, .external_lex_state = 5},
  [1470] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1471] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1472] = {.lex_state = 125, .external_lex_state = 2},
  [1473] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1474] = {.lex_state = 125, .external_lex_state = 2},
  [1475] = {.lex_state = 125, .external_lex_state = 2},
  [1476] = {.lex_state = 125, .external_lex_state = 2},
  [1477] = {.lex_state = 125, .external_lex_state = 2},
  [1478] = {.lex_state = 3, .external_lex_state = 2},
  [1479] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1480] = {.lex_state = 125, .external_lex_state = 2},
  [1481] = {.lex_state = 125, .external_lex_state = 2},
  [1482] = {.lex_state = 125, .external_lex_state = 5},
  [1483] = {.lex_state = 125, .external_lex_state = 2},
  [1484] = {.lex_state = 125, .external_lex_state = 2},
  [1485] = {.lex_state = 125, .external_lex_state = 2},
  [1486] = {.lex_state = 125, .external_lex_state = 2},
  [1487] = {.lex_state = 125, .external_lex_state = 2},
  [1488] = {.lex_state = 125, .external_lex_state = 2},
  [1489] = {.lex_state = 125, .external_lex_state = 2},
  [1490] = {.lex_state = 125, .external_lex_state = 2},
  [1491] = {.lex_state = 125, .external_lex_state = 2},
  [1492] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1493] = {.lex_state = 125, .external_lex_state = 5},
  [1494] = {.lex_state = 125, .external_lex_state = 2},
  [1495] = {.lex_state = 125, .external_lex_state = 5},
  [1496] = {.lex_state = 6, .external_lex_state = 2},
  [1497] = {.lex_state = 125, .external_lex_state = 5},
  [1498] = {.lex_state = 125, .external_lex_state = 2},
  [1499] = {.lex_state = 125, .external_lex_state = 2},
  [1500] = {.lex_state = 125, .external_lex_state = 5},
  [1501] = {.lex_state = 125, .external_lex_state = 2},
  [1502] = {.lex_state = 125, .external_lex_state = 2},
  [1503] = {.lex_state = 125, .external_lex_state = 2},
  [1504] = {.lex_state = 125, .external_lex_state = 2},
  [1505] = {.lex_state = 125, .external_lex_state = 5},
  [1506] = {.lex_state = 125, .external_lex_state = 2},
  [1507] = {.lex_state = 125, .external_lex_state = 2},
  [1508] = {.lex_state = 125, .external_lex_state = 2},
  [1509] = {.lex_state = 125, .external_lex_state = 2},
  [1510] = {.lex_state = 125, .external_lex_state = 2},
  [1511] = {.lex_state = 125, .external_lex_state = 2},
  [1512] = {.lex_state = 6, .external_lex_state = 2},
  [1513] = {.lex_state = 125, .external_lex_state = 5},
  [1514] = {.lex_state = 125, .external_lex_state = 2},
  [1515] = {.lex_state = 125, .external_lex_state = 2},
  [1516] = {.lex_state = 125, .external_lex_state = 2},
  [1517] = {.lex_state = 125, .external_lex_state = 2},
  [1518] = {.lex_state = 125, .external_lex_state = 2},
  [1519] = {.lex_state = 125, .external_lex_state = 2},
  [1520] = {.lex_state = 125, .external_lex_state = 2},
  [1521] = {.lex_state = 125, .external_lex_state = 5},
  [1522] = {.lex_state = 125, .external_lex_state = 5},
  [1523] = {.lex_state = 125, .external_lex_state = 2},
  [1524] = {.lex_state = 125, .external_lex_state = 2},
  [1525] = {.lex_state = 125, .external_lex_state = 2},
  [1526] = {.lex_state = 125, .external_lex_state = 5},
  [1527] = {.lex_state = 125, .external_lex_state = 2},
  [1528] = {.lex_state = 125, .external_lex_state = 2},
  [1529] = {.lex_state = 125, .external_lex_state = 2},
  [1530] = {.lex_state = 125, .external_lex_state = 2},
  [1531] = {.lex_state = 125, .external_lex_state = 2},
  [1532] = {.lex_state = 125, .external_lex_state = 2},
  [1533] = {.lex_state = 125, .external_lex_state = 2},
  [1534] = {.lex_state = 6, .external_lex_state = 2},
  [1535] = {.lex_state = 125, .external_lex_state = 2},
  [1536] = {.lex_state = 125, .external_lex_state = 2},
  [1537] = {.lex_state = 125, .external_lex_state = 2},
  [1538] = {.lex_state = 125, .external_lex_state = 2},
  [1539] = {.lex_state = 125, .external_lex_state = 2},
  [1540] = {.lex_state = 125, .external_lex_state = 2},
  [1541] = {.lex_state = 125, .external_lex_state = 2},
  [1542] = {.lex_state = 125, .external_lex_state = 2},
  [1543] = {.lex_state = 125, .external_lex_state = 2},
  [1544] = {.lex_state = 125, .external_lex_state = 2},
  [1545] = {.lex_state = 125, .external_lex_state = 2},
  [1546] = {.lex_state = 125, .external_lex_state = 2},
  [1547] = {.lex_state = 125, .external_lex_state = 2},
  [1548] = {.lex_state = 125, .external_lex_state = 2},
  [1549] = {.lex_state = 125, .external_lex_state = 2},
  [1550] = {.lex_state = 6, .external_lex_state = 2},
  [1551] = {.lex_state = 125, .external_lex_state = 2},
  [1552] = {.lex_state = 125, .external_lex_state = 2},
  [1553] = {.lex_state = 125, .external_lex_state = 2},
  [1554] = {.lex_state = 125, .external_lex_state = 2},
  [1555] = {.lex_state = 125, .external_lex_state = 2},
  [1556] = {.lex_state = 125, .external_lex_state = 2},
  [1557] = {.lex_state = 125, .external_lex_state = 2},
  [1558] = {.lex_state = 125, .external_lex_state = 2},
  [1559] = {.lex_state = 125, .external_lex_state = 2},
  [1560] = {.lex_state = 125, .external_lex_state = 2},
  [1561] = {.lex_state = 125, .external_lex_state = 2},
  [1562] = {.lex_state = 125, .external_lex_state = 2},
  [1563] = {.lex_state = 125, .external_lex_state = 2},
  [1564] = {.lex_state = 125, .external_lex_state = 2},
  [1565] = {.lex_state = 125, .external_lex_state = 2},
  [1566] = {.lex_state = 125, .external_lex_state = 2},
  [1567] = {.lex_state = 125, .external_lex_state = 2},
  [1568] = {.lex_state = 125, .external_lex_state = 2},
  [1569] = {.lex_state = 125, .external_lex_state = 2},
  [1570] = {.lex_state = 125, .external_lex_state = 2},
  [1571] = {.lex_state = 125, .external_lex_state = 2},
  [1572] = {.lex_state = 125, .external_lex_state = 2},
  [1573] = {.lex_state = 125, .external_lex_state = 2},
  [1574] = {.lex_state = 125, .external_lex_state = 2},
  [1575] = {.lex_state = 125, .external_lex_state = 2},
  [1576] = {.lex_state = 125, .external_lex_state = 2},
  [1577] = {.lex_state = 125, .external_lex_state = 2},
  [1578] = {.lex_state = 125, .external_lex_state = 2},
  [1579] = {.lex_state = 125, .external_lex_state = 2},
  [1580] = {.lex_state = 125, .external_lex_state = 2},
  [1581] = {.lex_state = 125, .external_lex_state = 2},
  [1582] = {.lex_state = 125, .external_lex_state = 2},
  [1583] = {.lex_state = 125, .external_lex_state = 2},
  [1584] = {.lex_state = 125, .external_lex_state = 2},
  [1585] = {.lex_state = 125, .external_lex_state = 2},
  [1586] = {.lex_state = 125, .external_lex_state = 2},
  [1587] = {.lex_state = 125, .external_lex_state = 2},
  [1588] = {.lex_state = 125, .external_lex_state = 2},
  [1589] = {.lex_state = 125, .external_lex_state = 2},
  [1590] = {.lex_state = 125, .external_lex_state = 2},
  [1591] = {.lex_state = 125, .external_lex_state = 2},
  [1592] = {.lex_state = 125, .external_lex_state = 2},
  [1593] = {.lex_state = 125, .external_lex_state = 2},
  [1594] = {.lex_state = 125, .external_lex_state = 2},
  [1595] = {.lex_state = 125, .external_lex_state = 2},
  [1596] = {.lex_state = 125, .external_lex_state = 2},
  [1597] = {.lex_state = 125, .external_lex_state = 2},
  [1598] = {.lex_state = 125, .external_lex_state = 2},
  [1599] = {.lex_state = 125, .external_lex_state = 2},
  [1600] = {.lex_state = 125, .external_lex_state = 2},
  [1601] = {.lex_state = 125, .external_lex_state = 2},
  [1602] = {.lex_state = 125, .external_lex_state = 2},
  [1603] = {.lex_state = 125, .external_lex_state = 2},
  [1604] = {.lex_state = 125, .external_lex_state = 2},
  [1605] = {.lex_state = 125, .external_lex_state = 2},
  [1606] = {.lex_state = 125, .external_lex_state = 2},
  [1607] = {.lex_state = 125, .external_lex_state = 2},
  [1608] = {.lex_state = 125, .external_lex_state = 2},
  [1609] = {.lex_state = 125, .external_lex_state = 2},
  [1610] = {.lex_state = 125, .external_lex_state = 2},
  [1611] = {.lex_state = 125, .external_lex_state = 2},
  [1612] = {.lex_state = 125, .external_lex_state = 2},
  [1613] = {.lex_state = 125, .external_lex_state = 2},
  [1614] = {.lex_state = 125, .external_lex_state = 2},
  [1615] = {.lex_state = 125, .external_lex_state = 2},
  [1616] = {.lex_state = 125, .external_lex_state = 2},
  [1617] = {.lex_state = 125, .external_lex_state = 2},
  [1618] = {.lex_state = 125, .external_lex_state = 2},
  [1619] = {.lex_state = 125, .external_lex_state = 2},
  [1620] = {.lex_state = 125, .external_lex_state = 5},
  [1621] = {.lex_state = 125, .external_lex_state = 2},
  [1622] = {.lex_state = 125, .external_lex_state = 2},
  [1623] = {.lex_state = 125, .external_lex_state = 2},
  [1624] = {.lex_state = 125, .external_lex_state = 2},
  [1625] = {.lex_state = 125, .external_lex_state = 2},
  [1626] = {.lex_state = 125, .external_lex_state = 2},
  [1627] = {.lex_state = 125, .external_lex_state = 2},
  [1628] = {.lex_state = 125, .external_lex_state = 2},
  [1629] = {.lex_state = 125, .external_lex_state = 2},
  [1630] = {.lex_state = 125, .external_lex_state = 2},
  [1631] = {.lex_state = 125, .external_lex_state = 2},
  [1632] = {.lex_state = 125, .external_lex_state = 2},
  [1633] = {.lex_state = 125, .external_lex_state = 2},
  [1634] = {.lex_state = 125, .external_lex_state = 2},
  [1635] = {.lex_state = 125, .external_lex_state = 2},
  [1636] = {.lex_state = 125, .external_lex_state = 2},
  [1637] = {.lex_state = 125, .external_lex_state = 2},
  [1638] = {.lex_state = 125, .external_lex_state = 2},
  [1639] = {.lex_state = 125, .external_lex_state = 2},
  [1640] = {.lex_state = 125, .external_lex_state = 2},
  [1641] = {.lex_state = 125, .external_lex_state = 2},
  [1642] = {.lex_state = 125, .external_lex_state = 2},
  [1643] = {.lex_state = 125, .external_lex_state = 2},
  [1644] = {.lex_state = 125, .external_lex_state = 2},
  [1645] = {.lex_state = 125, .external_lex_state = 2},
  [1646] = {.lex_state = 125, .external_lex_state = 2},
  [1647] = {.lex_state = 125, .external_lex_state = 2},
  [1648] = {.lex_state = 125, .external_lex_state = 2},
  [1649] = {.lex_state = 125, .external_lex_state = 2},
  [1650] = {.lex_state = 125, .external_lex_state = 2},
  [1651] = {.lex_state = 125, .external_lex_state = 2},
  [1652] = {.lex_state = 125, .external_lex_state = 2},
  [1653] = {.lex_state = 125, .external_lex_state = 2},
  [1654] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1655] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1656] = {.lex_state = 125, .external_lex_state = 2},
  [1657] = {.lex_state = 125, .external_lex_state = 2},
  [1658] = {.lex_state = 125, .external_lex_state = 2},
  [1659] = {.lex_state = 125, .external_lex_state = 2},
  [1660] = {.lex_state = 125, .external_lex_state = 2},
  [1661] = {.lex_state = 125, .external_lex_state = 2},
  [1662] = {.lex_state = 125, .external_lex_state = 5},
  [1663] = {.lex_state = 125, .external_lex_state = 2},
  [1664] = {.lex_state = 125, .external_lex_state = 5},
  [1665] = {.lex_state = 125, .external_lex_state = 2},
  [1666] = {.lex_state = 125, .external_lex_state = 5},
  [1667] = {.lex_state = 125, .external_lex_state = 2},
  [1668] = {.lex_state = 125, .external_lex_state = 2},
  [1669] = {.lex_state = 125, .external_lex_state = 2},
  [1670] = {.lex_state = 125, .external_lex_state = 2},
  [1671] = {.lex_state = 125, .external_lex_state = 2},
  [1672] = {.lex_state = 125, .external_lex_state = 2},
  [1673] = {.lex_state = 125, .external_lex_state = 5},
  [1674] = {.lex_state = 125, .external_lex_state = 5},
  [1675] = {.lex_state = 125, .external_lex_state = 2},
  [1676] = {.lex_state = 125, .external_lex_state = 2},
  [1677] = {.lex_state = 125, .external_lex_state = 2},
  [1678] = {.lex_state = 125, .external_lex_state = 5},
  [1679] = {.lex_state = 125, .external_lex_state = 2},
  [1680] = {.lex_state = 125, .external_lex_state = 2},
  [1681] = {.lex_state = 125, .external_lex_state = 2},
  [1682] = {.lex_state = 125, .external_lex_state = 2},
  [1683] = {.lex_state = 125, .external_lex_state = 2},
  [1684] = {.lex_state = 125, .external_lex_state = 2},
  [1685] = {.lex_state = 125, .external_lex_state = 5},
  [1686] = {.lex_state = 125, .external_lex_state = 5},
  [1687] = {.lex_state = 125, .external_lex_state = 2},
  [1688] = {.lex_state = 125, .external_lex_state = 2},
  [1689] = {.lex_state = 125, .external_lex_state = 5},
  [1690] = {.lex_state = 125, .external_lex_state = 2},
  [1691] = {.lex_state = 125, .external_lex_state = 2},
  [1692] = {.lex_state = 125, .external_lex_state = 2},
  [1693] = {.lex_state = 125, .external_lex_state = 2},
  [1694] = {.lex_state = 125, .external_lex_state = 2},
  [1695] = {.lex_state = 125, .external_lex_state = 5},
  [1696] = {.lex_state = 125, .external_lex_state = 2},
  [1697] = {.lex_state = 125, .external_lex_state = 2},
  [1698] = {.lex_state = 125, .external_lex_state = 2},
  [1699] = {.lex_state = 125, .external_lex_state = 2},
  [1700] = {.lex_state = 125, .external_lex_state = 2},
  [1701] = {.lex_state = 125, .external_lex_state = 2},
  [1702] = {.lex_state = 125, .external_lex_state = 5},
  [1703] = {.lex_state = 125, .external_lex_state = 2},
  [1704] = {.lex_state = 125, .external_lex_state = 2},
  [1705] = {.lex_state = 125, .external_lex_state = 2},
  [1706] = {.lex_state = 125, .external_lex_state = 2},
  [1707] = {.lex_state = 125, .external_lex_state = 2},
  [1708] = {.lex_state = 125, .external_lex_state = 2},
  [1709] = {.lex_state = 125, .external_lex_state = 2},
  [1710] = {.lex_state = 125, .external_lex_state = 2},
  [1711] = {.lex_state = 125, .external_lex_state = 2},
  [1712] = {.lex_state = 125, .external_lex_state = 2},
  [1713] = {.lex_state = 125, .external_lex_state = 2},
  [1714] = {.lex_state = 125, .external_lex_state = 2},
  [1715] = {.lex_state = 125, .external_lex_state = 2},
  [1716] = {.lex_state = 125, .external_lex_state = 2},
  [1717] = {.lex_state = 125, .external_lex_state = 2},
  [1718] = {.lex_state = 125, .external_lex_state = 2},
  [1719] = {.lex_state = 125, .external_lex_state = 2},
  [1720] = {.lex_state = 125, .external_lex_state = 2},
  [1721] = {.lex_state = 125, .external_lex_state = 2},
  [1722] = {.lex_state = 125, .external_lex_state = 2},
  [1723] = {.lex_state = 125, .external_lex_state = 2},
  [1724] = {.lex_state = 31, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1725] = {.lex_state = 125, .external_lex_state = 2},
  [1726] = {.lex_state = 125, .external_lex_state = 2},
  [1727] = {.lex_state = 125, .external_lex_state = 5},
  [1728] = {.lex_state = 125, .external_lex_state = 2},
  [1729] = {.lex_state = 125, .external_lex_state = 2},
  [1730] = {.lex_state = 125, .external_lex_state = 2},
  [1731] = {.lex_state = 125, .external_lex_state = 5},
  [1732] = {.lex_state = 125, .external_lex_state = 5},
  [1733] = {.lex_state = 125, .external_lex_state = 2},
  [1734] = {.lex_state = 125, .external_lex_state = 2},
  [1735] = {.lex_state = 125, .external_lex_state = 2},
  [1736] = {.lex_state = 6, .external_lex_state = 2},
  [1737] = {.lex_state = 125, .external_lex_state = 5},
  [1738] = {.lex_state = 125, .external_lex_state = 2},
  [1739] = {.lex_state = 125, .external_lex_state = 2},
  [1740] = {.lex_state = 125, .external_lex_state = 2},
  [1741] = {.lex_state = 125, .external_lex_state = 2},
  [1742] = {.lex_state = 125, .external_lex_state = 2},
  [1743] = {.lex_state = 125, .external_lex_state = 2},
  [1744] = {.lex_state = 125, .external_lex_state = 2},
  [1745] = {.lex_state = 125, .external_lex_state = 2},
  [1746] = {.lex_state = 125, .external_lex_state = 2},
  [1747] = {.lex_state = 125, .external_lex_state = 2},
  [1748] = {.lex_state = 125, .external_lex_state = 2},
  [1749] = {.lex_state = 125, .external_lex_state = 2},
  [1750] = {.lex_state = 125, .external_lex_state = 2},
  [1751] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1752] = {.lex_state = 125, .external_lex_state = 2},
  [1753] = {.lex_state = 125, .external_lex_state = 2},
  [1754] = {.lex_state = 125, .external_lex_state = 2},
  [1755] = {.lex_state = 125, .external_lex_state = 2},
  [1756] = {.lex_state = 125, .external_lex_state = 2},
  [1757] = {.lex_state = 125, .external_lex_state = 2},
  [1758] = {.lex_state = 125, .external_lex_state = 2},
  [1759] = {.lex_state = 125, .external_lex_state = 2},
  [1760] = {.lex_state = 125, .external_lex_state = 2},
  [1761] = {.lex_state = 125, .external_lex_state = 2},
  [1762] = {.lex_state = 125, .external_lex_state = 2},
  [1763] = {.lex_state = 125, .external_lex_state = 2},
  [1764] = {.lex_state = 125, .external_lex_state = 2},
  [1765] = {.lex_state = 125, .external_lex_state = 2},
  [1766] = {.lex_state = 125, .external_lex_state = 2},
  [1767] = {.lex_state = 125, .external_lex_state = 2},
  [1768] = {.lex_state = 125, .external_lex_state = 2},
  [1769] = {.lex_state = 2, .external_lex_state = 9},
  [1770] = {.lex_state = 32, .external_lex_state = 2},
  [1771] = {.lex_state = 125, .external_lex_state = 2},
  [1772] = {.lex_state = 125, .external_lex_state = 2},
  [1773] = {.lex_state = 125, .external_lex_state = 2},
  [1774] = {.lex_state = 125, .external_lex_state = 2},
  [1775] = {.lex_state = 125, .external_lex_state = 2},
  [1776] = {.lex_state = 125, .external_lex_state = 2},
  [1777] = {.lex_state = 125, .external_lex_state = 2},
  [1778] = {.lex_state = 125, .external_lex_state = 2},
  [1779] = {.lex_state = 125, .external_lex_state = 2},
  [1780] = {.lex_state = 125, .external_lex_state = 2},
  [1781] = {.lex_state = 125, .external_lex_state = 2},
  [1782] = {.lex_state = 125, .external_lex_state = 2},
  [1783] = {.lex_state = 125, .external_lex_state = 2},
  [1784] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1785] = {.lex_state = 2, .external_lex_state = 9},
  [1786] = {.lex_state = 125, .external_lex_state = 2},
  [1787] = {.lex_state = 32, .external_lex_state = 2},
  [1788] = {.lex_state = 125, .external_lex_state = 2},
  [1789] = {.lex_state = 125, .external_lex_state = 2},
  [1790] = {.lex_state = 125, .external_lex_state = 2},
  [1791] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1792] = {.lex_state = 125, .external_lex_state = 2},
  [1793] = {.lex_state = 125, .external_lex_state = 2},
  [1794] = {.lex_state = 125, .external_lex_state = 2},
  [1795] = {.lex_state = 125, .external_lex_state = 2},
  [1796] = {.lex_state = 125, .external_lex_state = 2},
  [1797] = {.lex_state = 125, .external_lex_state = 2},
  [1798] = {.lex_state = 125, .external_lex_state = 2},
  [1799] = {.lex_state = 125, .external_lex_state = 2},
  [1800] = {.lex_state = 125, .external_lex_state = 2},
  [1801] = {.lex_state = 125, .external_lex_state = 2},
  [1802] = {.lex_state = 125, .external_lex_state = 2},
  [1803] = {.lex_state = 125, .external_lex_state = 2},
  [1804] = {.lex_state = 125, .external_lex_state = 2},
  [1805] = {.lex_state = 125, .external_lex_state = 2},
  [1806] = {.lex_state = 125, .external_lex_state = 2},
  [1807] = {.lex_state = 125, .external_lex_state = 2},
  [1808] = {.lex_state = 32, .external_lex_state = 2},
  [1809] = {.lex_state = 125, .external_lex_state = 2},
  [1810] = {.lex_state = 125, .external_lex_state = 2},
  [1811] = {.lex_state = 125, .external_lex_state = 2},
  [1812] = {.lex_state = 125, .external_lex_state = 2},
  [1813] = {.lex_state = 125, .external_lex_state = 2},
  [1814] = {.lex_state = 125, .external_lex_state = 2},
  [1815] = {.lex_state = 125, .external_lex_state = 2},
  [1816] = {.lex_state = 125, .external_lex_state = 2},
  [1817] = {.lex_state = 125, .external_lex_state = 2},
  [1818] = {.lex_state = 125, .external_lex_state = 2},
  [1819] = {.lex_state = 125, .external_lex_state = 2},
  [1820] = {.lex_state = 125, .external_lex_state = 2},
  [1821] = {.lex_state = 125, .external_lex_state = 2},
  [1822] = {.lex_state = 32, .external_lex_state = 2},
  [1823] = {.lex_state = 125, .external_lex_state = 2},
  [1824] = {.lex_state = 125, .external_lex_state = 2},
  [1825] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1826] = {.lex_state = 125, .external_lex_state = 2},
  [1827] = {.lex_state = 125, .external_lex_state = 2},
  [1828] = {.lex_state = 125, .external_lex_state = 2},
  [1829] = {.lex_state = 2, .external_lex_state = 9},
  [1830] = {.lex_state = 125, .external_lex_state = 2},
  [1831] = {.lex_state = 125, .external_lex_state = 2},
  [1832] = {.lex_state = 2, .external_lex_state = 9},
  [1833] = {.lex_state = 125, .external_lex_state = 2},
  [1834] = {.lex_state = 125, .external_lex_state = 2},
  [1835] = {.lex_state = 125, .external_lex_state = 2},
  [1836] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1837] = {.lex_state = 125, .external_lex_state = 2},
  [1838] = {.lex_state = 125, .external_lex_state = 2},
  [1839] = {.lex_state = 125, .external_lex_state = 2},
  [1840] = {.lex_state = 125, .external_lex_state = 2},
  [1841] = {.lex_state = 125, .external_lex_state = 2},
  [1842] = {.lex_state = 125, .external_lex_state = 2},
  [1843] = {.lex_state = 125, .external_lex_state = 2},
  [1844] = {.lex_state = 125, .external_lex_state = 2},
  [1845] = {.lex_state = 125, .external_lex_state = 2},
  [1846] = {.lex_state = 125, .external_lex_state = 2},
  [1847] = {.lex_state = 125, .external_lex_state = 2},
  [1848] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1849] = {.lex_state = 125, .external_lex_state = 2},
  [1850] = {.lex_state = 125, .external_lex_state = 2},
  [1851] = {.lex_state = 125, .external_lex_state = 2},
  [1852] = {.lex_state = 125, .external_lex_state = 2},
  [1853] = {.lex_state = 125, .external_lex_state = 2},
  [1854] = {.lex_state = 125, .external_lex_state = 2},
  [1855] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1856] = {.lex_state = 125, .external_lex_state = 2},
  [1857] = {.lex_state = 125, .external_lex_state = 2},
  [1858] = {.lex_state = 125, .external_lex_state = 2},
  [1859] = {.lex_state = 125, .external_lex_state = 2},
  [1860] = {.lex_state = 125, .external_lex_state = 2, .reserved_word_set_id = 1},
  [1861] = {.lex_state = 125, .external_lex_state = 2},
  [1862] = {.lex_state = 125, .external_lex_state = 2},
  [1863] = {.lex_state = 125, .external_lex_state = 2},
  [1864] = {.lex_state = 125, .external_lex_state = 2},
  [1865] = {.lex_state = 125, .external_lex_state = 2},
  [1866] = {.lex_state = 125, .external_lex_state = 2},
  [1867] = {.lex_state = 125, .external_lex_state = 2},
  [1868] = {.lex_state = 125, .external_lex_state = 2},
  [1869] = {.lex_state = 125, .external_lex_state = 2},
};

static const TSSymbol ts_reserved_words[14][MAX_RESERVED_WORD_SET_SIZE] = {
  [1] = {
    anon_sym_export,
    anon_sym_default,
    anon_sym_import,
    anon_sym_with,
    anon_sym_var,
    anon_sym_const,
    anon_sym_else,
    anon_sym_if,
    anon_sym_switch,
    anon_sym_for,
    anon_sym_in,
    anon_sym_while,
    anon_sym_do,
    anon_sym_try,
    anon_sym_break,
    anon_sym_continue,
    anon_sym_debugger,
    anon_sym_return,
    anon_sym_throw,
    anon_sym_case,
    anon_sym_catch,
    anon_sym_finally,
    anon_sym_class,
    anon_sym_extends,
    anon_sym_function,
    anon_sym_new,
    anon_sym_instanceof,
    anon_sym_typeof,
    anon_sym_void,
    anon_sym_delete,
    sym_this,
    sym_super,
    sym_true,
    sym_false,
    sym_null,
  },
  [2] = {
    anon_sym_default,
    anon_sym_with,
    anon_sym_var,
    anon_sym_const,
    anon_sym_else,
    anon_sym_if,
    anon_sym_switch,
    anon_sym_for,
    anon_sym_while,
    anon_sym_do,
    anon_sym_try,
    anon_sym_break,
    anon_sym_continue,
    anon_sym_debugger,
    anon_sym_return,
    anon_sym_throw,
    anon_sym_case,
    anon_sym_catch,
    anon_sym_finally,
    anon_sym_extends,
  },
  [3] = {
    anon_sym_else,
    anon_sym_in,
    anon_sym_catch,
    anon_sym_finally,
    anon_sym_extends,
    anon_sym_instanceof,
  },
  [4] = {
    anon_sym_default,
    anon_sym_else,
    anon_sym_in,
    anon_sym_case,
    anon_sym_catch,
    anon_sym_finally,
    anon_sym_extends,
    anon_sym_instanceof,
  },
  [5] = {
    anon_sym_catch,
    anon_sym_finally,
    anon_sym_extends,
  },
  [6] = {
    anon_sym_default,
    anon_sym_with,
    anon_sym_else,
    anon_sym_if,
    anon_sym_switch,
    anon_sym_for,
    anon_sym_in,
    anon_sym_while,
    anon_sym_do,
    anon_sym_try,
    anon_sym_break,
    anon_sym_continue,
    anon_sym_debugger,
    anon_sym_return,
    anon_sym_throw,
    anon_sym_case,
    anon_sym_catch,
    anon_sym_finally,
    anon_sym_extends,
    anon_sym_instanceof,
  },
  [7] = {
    anon_sym_default,
    anon_sym_with,
    anon_sym_var,
    anon_sym_const,
    anon_sym_else,
    anon_sym_if,
    anon_sym_switch,
    anon_sym_for,
    anon_sym_in,
    anon_sym_while,
    anon_sym_do,
    anon_sym_try,
    anon_sym_break,
    anon_sym_continue,
    anon_sym_debugger,
    anon_sym_return,
    anon_sym_throw,
    anon_sym_case,
    anon_sym_catch,
    anon_sym_finally,
    anon_sym_extends,
    anon_sym_instanceof,
  },
  [8] = {
    anon_sym_default,
    anon_sym_import,
    anon_sym_with,
    anon_sym_var,
    anon_sym_const,
    anon_sym_else,
    anon_sym_if,
    anon_sym_switch,
    anon_sym_for,
    anon_sym_while,
    anon_sym_do,
    anon_sym_try,
    anon_sym_break,
    anon_sym_continue,
    anon_sym_debugger,
    anon_sym_return,
    anon_sym_throw,
    anon_sym_case,
    anon_sym_catch,
    anon_sym_finally,
    anon_sym_class,
    anon_sym_extends,
    anon_sym_new,
    anon_sym_typeof,
    anon_sym_void,
    anon_sym_delete,
    sym_this,
    sym_super,
    sym_true,
    sym_false,
    sym_null,
  },
  [9] = {
    anon_sym_export,
    anon_sym_default,
    anon_sym_import,
    anon_sym_with,
    anon_sym_var,
    anon_sym_const,
    anon_sym_else,
    anon_sym_if,
    anon_sym_switch,
    anon_sym_for,
    anon_sym_while,
    anon_sym_do,
    anon_sym_try,
    anon_sym_break,
    anon_sym_continue,
    anon_sym_debugger,
    anon_sym_return,
    anon_sym_throw,
    anon_sym_case,
    anon_sym_catch,
    anon_sym_finally,
    anon_sym_class,
    anon_sym_extends,
    anon_sym_function,
    anon_sym_new,
    anon_sym_typeof,
    anon_sym_void,
    anon_sym_delete,
    sym_this,
    sym_super,
    sym_true,
    sym_false,
    sym_null,
  },
  [10] = {
    anon_sym_in,
    anon_sym_extends,
    anon_sym_instanceof,
  },
  [11] = {
    anon_sym_in,
    anon_sym_catch,
    anon_sym_extends,
    anon_sym_instanceof,
  },
  [12] = {
    anon_sym_in,
    anon_sym_catch,
    anon_sym_finally,
    anon_sym_extends,
    anon_sym_instanceof,
  },
  [13] = {
    anon_sym_export,
    anon_sym_default,
    anon_sym_import,
    anon_sym_with,
    anon_sym_var,
    anon_sym_const,
    anon_sym_else,
    anon_sym_if,
    anon_sym_switch,
    anon_sym_for,
    anon_sym_in,
    anon_sym_while,
    anon_sym_do,
    anon_sym_try,
    anon_sym_break,
    anon_sym_continue,
    anon_sym_debugger,
    anon_sym_return,
    anon_sym_throw,
    anon_sym_case,
    anon_sym_catch,
    anon_sym_finally,
    anon_sym_class,
    anon_sym_function,
    anon_sym_new,
    anon_sym_instanceof,
    anon_sym_typeof,
    anon_sym_void,
    anon_sym_delete,
    sym_this,
    sym_super,
    sym_true,
    sym_false,
    sym_null,
  },
};

static const uint16_t ts_parse_table[LARGE_STATE_COUNT][SYMBOL_COUNT] = {
  [STATE(0)] = {
    [ts_builtin_sym_end] = ACTIONS(1),
    [sym_identifier] = ACTIONS(1),
    [sym_hash_bang_line] = ACTIONS(1),
    [anon_sym_export] = ACTIONS(1),
    [anon_sym_STAR] = ACTIONS(1),
    [anon_sym_default] = ACTIONS(1),
    [anon_sym_as] = ACTIONS(1),
    [anon_sym_LBRACE] = ACTIONS(1),
    [anon_sym_COMMA] = ACTIONS(1),
    [anon_sym_RBRACE] = ACTIONS(1),
    [anon_sym_import] = ACTIONS(1),
    [anon_sym_from] = ACTIONS(1),
    [anon_sym_with] = ACTIONS(1),
    [anon_sym_var] = ACTIONS(1),
    [anon_sym_let] = ACTIONS(1),
    [anon_sym_const] = ACTIONS(1),
    [anon_sym_using] = ACTIONS(1),
    [anon_sym_await] = ACTIONS(1),
    [anon_sym_of] = ACTIONS(1),
    [anon_sym_else] = ACTIONS(1),
    [anon_sym_if] = ACTIONS(1),
    [anon_sym_switch] = ACTIONS(1),
    [anon_sym_for] = ACTIONS(1),
    [anon_sym_LPAREN] = ACTIONS(1),
    [anon_sym_SEMI] = ACTIONS(1),
    [anon_sym_RPAREN] = ACTIONS(1),
    [anon_sym_in] = ACTIONS(1),
    [anon_sym_while] = ACTIONS(1),
    [anon_sym_do] = ACTIONS(1),
    [anon_sym_try] = ACTIONS(1),
    [anon_sym_break] = ACTIONS(1),
    [anon_sym_continue] = ACTIONS(1),
    [anon_sym_debugger] = ACTIONS(1),
    [anon_sym_return] = ACTIONS(1),
    [anon_sym_throw] = ACTIONS(1),
    [anon_sym_COLON] = ACTIONS(1),
    [anon_sym_case] = ACTIONS(1),
    [anon_sym_catch] = ACTIONS(1),
    [anon_sym_finally] = ACTIONS(1),
    [anon_sym_yield] = ACTIONS(1),
    [anon_sym_EQ] = ACTIONS(1),
    [anon_sym_LBRACK] = ACTIONS(1),
    [anon_sym_RBRACK] = ACTIONS(1),
    [anon_sym_LT] = ACTIONS(1),
    [anon_sym_GT] = ACTIONS(1),
    [anon_sym_DOT] = ACTIONS(1),
    [anon_sym_LT_SLASH] = ACTIONS(1),
    [anon_sym_DQUOTE] = ACTIONS(1),
    [anon_sym_SQUOTE] = ACTIONS(1),
    [anon_sym_class] = ACTIONS(1),
    [anon_sym_extends] = ACTIONS(1),
    [anon_sym_async] = ACTIONS(1),
    [anon_sym_function] = ACTIONS(1),
    [anon_sym_EQ_GT] = ACTIONS(1),
    [sym_optional_chain] = ACTIONS(1),
    [anon_sym_new] = ACTIONS(1),
    [anon_sym_PLUS_EQ] = ACTIONS(1),
    [anon_sym_DASH_EQ] = ACTIONS(1),
    [anon_sym_STAR_EQ] = ACTIONS(1),
    [anon_sym_SLASH_EQ] = ACTIONS(1),
    [anon_sym_PERCENT_EQ] = ACTIONS(1),
    [anon_sym_CARET_EQ] = ACTIONS(1),
    [anon_sym_AMP_EQ] = ACTIONS(1),
    [anon_sym_PIPE_EQ] = ACTIONS(1),
    [anon_sym_GT_GT_EQ] = ACTIONS(1),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(1),
    [anon_sym_LT_LT_EQ] = ACTIONS(1),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(1),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(1),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(1),
    [anon_sym_QMARK_QMARK_EQ] = ACTIONS(1),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(1),
    [anon_sym_AMP_AMP] = ACTIONS(1),
    [anon_sym_PIPE_PIPE] = ACTIONS(1),
    [anon_sym_GT_GT] = ACTIONS(1),
    [anon_sym_GT_GT_GT] = ACTIONS(1),
    [anon_sym_LT_LT] = ACTIONS(1),
    [anon_sym_AMP] = ACTIONS(1),
    [anon_sym_CARET] = ACTIONS(1),
    [anon_sym_PIPE] = ACTIONS(1),
    [anon_sym_PLUS] = ACTIONS(1),
    [anon_sym_DASH] = ACTIONS(1),
    [anon_sym_SLASH] = ACTIONS(1),
    [anon_sym_PERCENT] = ACTIONS(1),
    [anon_sym_STAR_STAR] = ACTIONS(1),
    [anon_sym_LT_EQ] = ACTIONS(1),
    [anon_sym_EQ_EQ] = ACTIONS(1),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(1),
    [anon_sym_BANG_EQ] = ACTIONS(1),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(1),
    [anon_sym_GT_EQ] = ACTIONS(1),
    [anon_sym_QMARK_QMARK] = ACTIONS(1),
    [anon_sym_instanceof] = ACTIONS(1),
    [anon_sym_BANG] = ACTIONS(1),
    [anon_sym_TILDE] = ACTIONS(1),
    [anon_sym_typeof] = ACTIONS(1),
    [anon_sym_void] = ACTIONS(1),
    [anon_sym_delete] = ACTIONS(1),
    [anon_sym_PLUS_PLUS] = ACTIONS(1),
    [anon_sym_DASH_DASH] = ACTIONS(1),
    [sym_escape_sequence] = ACTIONS(1),
    [sym_comment] = ACTIONS(3),
    [anon_sym_BQUOTE] = ACTIONS(1),
    [anon_sym_DOLLAR_LBRACE] = ACTIONS(1),
    [anon_sym_SLASH2] = ACTIONS(1),
    [sym_number] = ACTIONS(1),
    [sym_private_property_identifier] = ACTIONS(1),
    [anon_sym_target] = ACTIONS(1),
    [anon_sym_meta] = ACTIONS(1),
    [sym_this] = ACTIONS(1),
    [sym_super] = ACTIONS(1),
    [sym_true] = ACTIONS(1),
    [sym_false] = ACTIONS(1),
    [sym_null] = ACTIONS(1),
    [sym_undefined] = ACTIONS(1),
    [anon_sym_AT] = ACTIONS(1),
    [anon_sym_static] = ACTIONS(1),
    [aux_sym_method_definition_token1] = ACTIONS(1),
    [anon_sym_get] = ACTIONS(1),
    [anon_sym_set] = ACTIONS(1),
    [sym__automatic_semicolon] = ACTIONS(1),
    [sym__template_chars] = ACTIONS(1),
    [sym__ternary_qmark] = ACTIONS(1),
    [sym_html_comment] = ACTIONS(5),
    [sym_jsx_text] = ACTIONS(1),
  },
  [STATE(1)] = {
    [sym_program] = STATE(1786),
    [sym_export_statement] = STATE(471),
    [sym_declaration] = STATE(471),
    [sym_import] = STATE(1326),
    [sym_import_statement] = STATE(471),
    [sym_statement] = STATE(20),
    [sym_expression_statement] = STATE(471),
    [sym_variable_declaration] = STATE(419),
    [sym_lexical_declaration] = STATE(419),
    [sym_using_declaration] = STATE(419),
    [sym_statement_block] = STATE(471),
    [sym_if_statement] = STATE(471),
    [sym_switch_statement] = STATE(471),
    [sym_for_statement] = STATE(471),
    [sym_for_in_statement] = STATE(471),
    [sym_while_statement] = STATE(471),
    [sym_do_statement] = STATE(471),
    [sym_try_statement] = STATE(471),
    [sym_with_statement] = STATE(471),
    [sym_break_statement] = STATE(471),
    [sym_continue_statement] = STATE(471),
    [sym_debugger_statement] = STATE(471),
    [sym_return_statement] = STATE(471),
    [sym_throw_statement] = STATE(471),
    [sym_empty_statement] = STATE(471),
    [sym_labeled_statement] = STATE(471),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(823),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1790),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1790),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_class_declaration] = STATE(419),
    [sym_function_expression] = STATE(740),
    [sym_function_declaration] = STATE(419),
    [sym_generator_function] = STATE(740),
    [sym_generator_function_declaration] = STATE(419),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1790),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_sequence_expression] = STATE(1662),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [aux_sym_program_repeat1] = STATE(20),
    [aux_sym_export_statement_repeat1] = STATE(1267),
    [ts_builtin_sym_end] = ACTIONS(7),
    [sym_identifier] = ACTIONS(9),
    [sym_hash_bang_line] = ACTIONS(11),
    [anon_sym_export] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(15),
    [anon_sym_import] = ACTIONS(17),
    [anon_sym_with] = ACTIONS(19),
    [anon_sym_var] = ACTIONS(21),
    [anon_sym_let] = ACTIONS(23),
    [anon_sym_const] = ACTIONS(25),
    [anon_sym_using] = ACTIONS(27),
    [anon_sym_await] = ACTIONS(29),
    [anon_sym_if] = ACTIONS(31),
    [anon_sym_switch] = ACTIONS(33),
    [anon_sym_for] = ACTIONS(35),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_SEMI] = ACTIONS(39),
    [anon_sym_while] = ACTIONS(41),
    [anon_sym_do] = ACTIONS(43),
    [anon_sym_try] = ACTIONS(45),
    [anon_sym_break] = ACTIONS(47),
    [anon_sym_continue] = ACTIONS(49),
    [anon_sym_debugger] = ACTIONS(51),
    [anon_sym_return] = ACTIONS(53),
    [anon_sym_throw] = ACTIONS(55),
    [anon_sym_yield] = ACTIONS(57),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(67),
    [anon_sym_async] = ACTIONS(69),
    [anon_sym_function] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(75),
    [anon_sym_DASH] = ACTIONS(75),
    [anon_sym_SLASH] = ACTIONS(77),
    [anon_sym_BANG] = ACTIONS(79),
    [anon_sym_TILDE] = ACTIONS(79),
    [anon_sym_typeof] = ACTIONS(75),
    [anon_sym_void] = ACTIONS(75),
    [anon_sym_delete] = ACTIONS(75),
    [anon_sym_PLUS_PLUS] = ACTIONS(81),
    [anon_sym_DASH_DASH] = ACTIONS(81),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(87),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(95),
    [anon_sym_get] = ACTIONS(95),
    [anon_sym_set] = ACTIONS(95),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(2)] = {
    [sym_export_statement] = STATE(471),
    [sym_declaration] = STATE(471),
    [sym_import] = STATE(1326),
    [sym_import_statement] = STATE(471),
    [sym_statement] = STATE(18),
    [sym_expression_statement] = STATE(471),
    [sym_variable_declaration] = STATE(419),
    [sym_lexical_declaration] = STATE(419),
    [sym_using_declaration] = STATE(419),
    [sym_statement_block] = STATE(471),
    [sym_if_statement] = STATE(471),
    [sym_switch_statement] = STATE(471),
    [sym_for_statement] = STATE(471),
    [sym_for_in_statement] = STATE(471),
    [sym_while_statement] = STATE(471),
    [sym_do_statement] = STATE(471),
    [sym_try_statement] = STATE(471),
    [sym_with_statement] = STATE(471),
    [sym_break_statement] = STATE(471),
    [sym_continue_statement] = STATE(471),
    [sym_debugger_statement] = STATE(471),
    [sym_return_statement] = STATE(471),
    [sym_throw_statement] = STATE(471),
    [sym_empty_statement] = STATE(471),
    [sym_labeled_statement] = STATE(471),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(823),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1853),
    [sym_object_assignment_pattern] = STATE(1454),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1853),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_class_declaration] = STATE(419),
    [sym_function_expression] = STATE(740),
    [sym_function_declaration] = STATE(419),
    [sym_generator_function] = STATE(740),
    [sym_generator_function_declaration] = STATE(419),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1853),
    [sym_spread_element] = STATE(1406),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_sequence_expression] = STATE(1662),
    [sym_string] = STATE(815),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [sym_rest_pattern] = STATE(1454),
    [sym_method_definition] = STATE(1406),
    [sym_pair] = STATE(1406),
    [sym_pair_pattern] = STATE(1454),
    [sym__property_name] = STATE(1407),
    [sym_computed_property_name] = STATE(1407),
    [aux_sym_program_repeat1] = STATE(18),
    [aux_sym_export_statement_repeat1] = STATE(1014),
    [aux_sym_object_repeat1] = STATE(1499),
    [aux_sym_object_pattern_repeat1] = STATE(1476),
    [sym_identifier] = ACTIONS(97),
    [anon_sym_export] = ACTIONS(99),
    [anon_sym_STAR] = ACTIONS(101),
    [anon_sym_LBRACE] = ACTIONS(15),
    [anon_sym_COMMA] = ACTIONS(103),
    [anon_sym_RBRACE] = ACTIONS(105),
    [anon_sym_import] = ACTIONS(17),
    [anon_sym_with] = ACTIONS(19),
    [anon_sym_var] = ACTIONS(21),
    [anon_sym_let] = ACTIONS(107),
    [anon_sym_const] = ACTIONS(25),
    [anon_sym_using] = ACTIONS(27),
    [anon_sym_await] = ACTIONS(109),
    [anon_sym_if] = ACTIONS(31),
    [anon_sym_switch] = ACTIONS(33),
    [anon_sym_for] = ACTIONS(35),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_SEMI] = ACTIONS(39),
    [anon_sym_while] = ACTIONS(41),
    [anon_sym_do] = ACTIONS(43),
    [anon_sym_try] = ACTIONS(45),
    [anon_sym_break] = ACTIONS(47),
    [anon_sym_continue] = ACTIONS(49),
    [anon_sym_debugger] = ACTIONS(51),
    [anon_sym_return] = ACTIONS(53),
    [anon_sym_throw] = ACTIONS(55),
    [anon_sym_yield] = ACTIONS(57),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(67),
    [anon_sym_async] = ACTIONS(113),
    [anon_sym_function] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(115),
    [anon_sym_PLUS] = ACTIONS(75),
    [anon_sym_DASH] = ACTIONS(75),
    [anon_sym_SLASH] = ACTIONS(77),
    [anon_sym_BANG] = ACTIONS(79),
    [anon_sym_TILDE] = ACTIONS(79),
    [anon_sym_typeof] = ACTIONS(75),
    [anon_sym_void] = ACTIONS(75),
    [anon_sym_delete] = ACTIONS(75),
    [anon_sym_PLUS_PLUS] = ACTIONS(81),
    [anon_sym_DASH_DASH] = ACTIONS(81),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(117),
    [sym_private_property_identifier] = ACTIONS(119),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(121),
    [aux_sym_method_definition_token1] = ACTIONS(123),
    [anon_sym_get] = ACTIONS(125),
    [anon_sym_set] = ACTIONS(125),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(3)] = {
    [sym_export_statement] = STATE(471),
    [sym_declaration] = STATE(471),
    [sym_import] = STATE(1326),
    [sym_import_statement] = STATE(471),
    [sym_statement] = STATE(18),
    [sym_expression_statement] = STATE(471),
    [sym_variable_declaration] = STATE(419),
    [sym_lexical_declaration] = STATE(419),
    [sym_using_declaration] = STATE(419),
    [sym_statement_block] = STATE(471),
    [sym_if_statement] = STATE(471),
    [sym_switch_statement] = STATE(471),
    [sym_for_statement] = STATE(471),
    [sym_for_in_statement] = STATE(471),
    [sym_while_statement] = STATE(471),
    [sym_do_statement] = STATE(471),
    [sym_try_statement] = STATE(471),
    [sym_with_statement] = STATE(471),
    [sym_break_statement] = STATE(471),
    [sym_continue_statement] = STATE(471),
    [sym_debugger_statement] = STATE(471),
    [sym_return_statement] = STATE(471),
    [sym_throw_statement] = STATE(471),
    [sym_empty_statement] = STATE(471),
    [sym_labeled_statement] = STATE(471),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(823),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1853),
    [sym_object_assignment_pattern] = STATE(1454),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1853),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_class_declaration] = STATE(419),
    [sym_function_expression] = STATE(740),
    [sym_function_declaration] = STATE(419),
    [sym_generator_function] = STATE(740),
    [sym_generator_function_declaration] = STATE(419),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1853),
    [sym_spread_element] = STATE(1406),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_sequence_expression] = STATE(1662),
    [sym_string] = STATE(815),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [sym_rest_pattern] = STATE(1454),
    [sym_method_definition] = STATE(1406),
    [sym_pair] = STATE(1406),
    [sym_pair_pattern] = STATE(1454),
    [sym__property_name] = STATE(1407),
    [sym_computed_property_name] = STATE(1407),
    [aux_sym_program_repeat1] = STATE(18),
    [aux_sym_export_statement_repeat1] = STATE(1014),
    [aux_sym_object_repeat1] = STATE(1499),
    [aux_sym_object_pattern_repeat1] = STATE(1476),
    [sym_identifier] = ACTIONS(97),
    [anon_sym_export] = ACTIONS(99),
    [anon_sym_STAR] = ACTIONS(101),
    [anon_sym_LBRACE] = ACTIONS(15),
    [anon_sym_COMMA] = ACTIONS(103),
    [anon_sym_RBRACE] = ACTIONS(127),
    [anon_sym_import] = ACTIONS(17),
    [anon_sym_with] = ACTIONS(19),
    [anon_sym_var] = ACTIONS(21),
    [anon_sym_let] = ACTIONS(107),
    [anon_sym_const] = ACTIONS(25),
    [anon_sym_using] = ACTIONS(27),
    [anon_sym_await] = ACTIONS(109),
    [anon_sym_if] = ACTIONS(31),
    [anon_sym_switch] = ACTIONS(33),
    [anon_sym_for] = ACTIONS(35),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_SEMI] = ACTIONS(39),
    [anon_sym_while] = ACTIONS(41),
    [anon_sym_do] = ACTIONS(43),
    [anon_sym_try] = ACTIONS(45),
    [anon_sym_break] = ACTIONS(47),
    [anon_sym_continue] = ACTIONS(49),
    [anon_sym_debugger] = ACTIONS(51),
    [anon_sym_return] = ACTIONS(53),
    [anon_sym_throw] = ACTIONS(55),
    [anon_sym_yield] = ACTIONS(57),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(67),
    [anon_sym_async] = ACTIONS(113),
    [anon_sym_function] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(115),
    [anon_sym_PLUS] = ACTIONS(75),
    [anon_sym_DASH] = ACTIONS(75),
    [anon_sym_SLASH] = ACTIONS(77),
    [anon_sym_BANG] = ACTIONS(79),
    [anon_sym_TILDE] = ACTIONS(79),
    [anon_sym_typeof] = ACTIONS(75),
    [anon_sym_void] = ACTIONS(75),
    [anon_sym_delete] = ACTIONS(75),
    [anon_sym_PLUS_PLUS] = ACTIONS(81),
    [anon_sym_DASH_DASH] = ACTIONS(81),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(117),
    [sym_private_property_identifier] = ACTIONS(119),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(121),
    [aux_sym_method_definition_token1] = ACTIONS(123),
    [anon_sym_get] = ACTIONS(125),
    [anon_sym_set] = ACTIONS(125),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(4)] = {
    [sym_export_statement] = STATE(471),
    [sym_declaration] = STATE(471),
    [sym_import] = STATE(1326),
    [sym_import_statement] = STATE(471),
    [sym_statement] = STATE(22),
    [sym_expression_statement] = STATE(471),
    [sym_variable_declaration] = STATE(419),
    [sym_lexical_declaration] = STATE(419),
    [sym_using_declaration] = STATE(419),
    [sym_statement_block] = STATE(471),
    [sym_if_statement] = STATE(471),
    [sym_switch_statement] = STATE(471),
    [sym_for_statement] = STATE(471),
    [sym_for_in_statement] = STATE(471),
    [sym_while_statement] = STATE(471),
    [sym_do_statement] = STATE(471),
    [sym_try_statement] = STATE(471),
    [sym_with_statement] = STATE(471),
    [sym_break_statement] = STATE(471),
    [sym_continue_statement] = STATE(471),
    [sym_debugger_statement] = STATE(471),
    [sym_return_statement] = STATE(471),
    [sym_throw_statement] = STATE(471),
    [sym_empty_statement] = STATE(471),
    [sym_labeled_statement] = STATE(471),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(823),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1853),
    [sym_object_assignment_pattern] = STATE(1454),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1853),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_class_declaration] = STATE(419),
    [sym_function_expression] = STATE(740),
    [sym_function_declaration] = STATE(419),
    [sym_generator_function] = STATE(740),
    [sym_generator_function_declaration] = STATE(419),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1853),
    [sym_spread_element] = STATE(1406),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_sequence_expression] = STATE(1662),
    [sym_string] = STATE(815),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [sym_rest_pattern] = STATE(1454),
    [sym_method_definition] = STATE(1406),
    [sym_pair] = STATE(1406),
    [sym_pair_pattern] = STATE(1454),
    [sym__property_name] = STATE(1407),
    [sym_computed_property_name] = STATE(1407),
    [aux_sym_program_repeat1] = STATE(22),
    [aux_sym_export_statement_repeat1] = STATE(1014),
    [aux_sym_object_repeat1] = STATE(1499),
    [aux_sym_object_pattern_repeat1] = STATE(1476),
    [sym_identifier] = ACTIONS(97),
    [anon_sym_export] = ACTIONS(99),
    [anon_sym_STAR] = ACTIONS(101),
    [anon_sym_LBRACE] = ACTIONS(15),
    [anon_sym_COMMA] = ACTIONS(103),
    [anon_sym_RBRACE] = ACTIONS(129),
    [anon_sym_import] = ACTIONS(17),
    [anon_sym_with] = ACTIONS(19),
    [anon_sym_var] = ACTIONS(21),
    [anon_sym_let] = ACTIONS(107),
    [anon_sym_const] = ACTIONS(25),
    [anon_sym_using] = ACTIONS(27),
    [anon_sym_await] = ACTIONS(109),
    [anon_sym_if] = ACTIONS(31),
    [anon_sym_switch] = ACTIONS(33),
    [anon_sym_for] = ACTIONS(35),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_SEMI] = ACTIONS(39),
    [anon_sym_while] = ACTIONS(41),
    [anon_sym_do] = ACTIONS(43),
    [anon_sym_try] = ACTIONS(45),
    [anon_sym_break] = ACTIONS(47),
    [anon_sym_continue] = ACTIONS(49),
    [anon_sym_debugger] = ACTIONS(51),
    [anon_sym_return] = ACTIONS(53),
    [anon_sym_throw] = ACTIONS(55),
    [anon_sym_yield] = ACTIONS(57),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(67),
    [anon_sym_async] = ACTIONS(113),
    [anon_sym_function] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(115),
    [anon_sym_PLUS] = ACTIONS(75),
    [anon_sym_DASH] = ACTIONS(75),
    [anon_sym_SLASH] = ACTIONS(77),
    [anon_sym_BANG] = ACTIONS(79),
    [anon_sym_TILDE] = ACTIONS(79),
    [anon_sym_typeof] = ACTIONS(75),
    [anon_sym_void] = ACTIONS(75),
    [anon_sym_delete] = ACTIONS(75),
    [anon_sym_PLUS_PLUS] = ACTIONS(81),
    [anon_sym_DASH_DASH] = ACTIONS(81),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(117),
    [sym_private_property_identifier] = ACTIONS(119),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(121),
    [aux_sym_method_definition_token1] = ACTIONS(123),
    [anon_sym_get] = ACTIONS(125),
    [anon_sym_set] = ACTIONS(125),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(5)] = {
    [sym_export_statement] = STATE(471),
    [sym_declaration] = STATE(471),
    [sym_import] = STATE(1326),
    [sym_import_statement] = STATE(471),
    [sym_statement] = STATE(24),
    [sym_expression_statement] = STATE(471),
    [sym_variable_declaration] = STATE(419),
    [sym_lexical_declaration] = STATE(419),
    [sym_using_declaration] = STATE(419),
    [sym_statement_block] = STATE(471),
    [sym_if_statement] = STATE(471),
    [sym_switch_statement] = STATE(471),
    [sym_for_statement] = STATE(471),
    [sym_for_in_statement] = STATE(471),
    [sym_while_statement] = STATE(471),
    [sym_do_statement] = STATE(471),
    [sym_try_statement] = STATE(471),
    [sym_with_statement] = STATE(471),
    [sym_break_statement] = STATE(471),
    [sym_continue_statement] = STATE(471),
    [sym_debugger_statement] = STATE(471),
    [sym_return_statement] = STATE(471),
    [sym_throw_statement] = STATE(471),
    [sym_empty_statement] = STATE(471),
    [sym_labeled_statement] = STATE(471),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(823),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1853),
    [sym_object_assignment_pattern] = STATE(1454),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1853),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_class_declaration] = STATE(419),
    [sym_function_expression] = STATE(740),
    [sym_function_declaration] = STATE(419),
    [sym_generator_function] = STATE(740),
    [sym_generator_function_declaration] = STATE(419),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1853),
    [sym_spread_element] = STATE(1458),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_sequence_expression] = STATE(1662),
    [sym_string] = STATE(815),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [sym_rest_pattern] = STATE(1454),
    [sym_method_definition] = STATE(1458),
    [sym_pair] = STATE(1458),
    [sym_pair_pattern] = STATE(1454),
    [sym__property_name] = STATE(1407),
    [sym_computed_property_name] = STATE(1407),
    [aux_sym_program_repeat1] = STATE(24),
    [aux_sym_export_statement_repeat1] = STATE(1014),
    [aux_sym_object_repeat1] = STATE(1461),
    [aux_sym_object_pattern_repeat1] = STATE(1476),
    [sym_identifier] = ACTIONS(131),
    [anon_sym_export] = ACTIONS(133),
    [anon_sym_STAR] = ACTIONS(101),
    [anon_sym_LBRACE] = ACTIONS(15),
    [anon_sym_COMMA] = ACTIONS(103),
    [anon_sym_RBRACE] = ACTIONS(135),
    [anon_sym_import] = ACTIONS(17),
    [anon_sym_with] = ACTIONS(19),
    [anon_sym_var] = ACTIONS(21),
    [anon_sym_let] = ACTIONS(137),
    [anon_sym_const] = ACTIONS(25),
    [anon_sym_using] = ACTIONS(27),
    [anon_sym_await] = ACTIONS(139),
    [anon_sym_if] = ACTIONS(31),
    [anon_sym_switch] = ACTIONS(33),
    [anon_sym_for] = ACTIONS(35),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_SEMI] = ACTIONS(39),
    [anon_sym_while] = ACTIONS(41),
    [anon_sym_do] = ACTIONS(43),
    [anon_sym_try] = ACTIONS(45),
    [anon_sym_break] = ACTIONS(47),
    [anon_sym_continue] = ACTIONS(49),
    [anon_sym_debugger] = ACTIONS(51),
    [anon_sym_return] = ACTIONS(53),
    [anon_sym_throw] = ACTIONS(55),
    [anon_sym_yield] = ACTIONS(57),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(67),
    [anon_sym_async] = ACTIONS(141),
    [anon_sym_function] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(115),
    [anon_sym_PLUS] = ACTIONS(75),
    [anon_sym_DASH] = ACTIONS(75),
    [anon_sym_SLASH] = ACTIONS(77),
    [anon_sym_BANG] = ACTIONS(79),
    [anon_sym_TILDE] = ACTIONS(79),
    [anon_sym_typeof] = ACTIONS(75),
    [anon_sym_void] = ACTIONS(75),
    [anon_sym_delete] = ACTIONS(75),
    [anon_sym_PLUS_PLUS] = ACTIONS(81),
    [anon_sym_DASH_DASH] = ACTIONS(81),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(117),
    [sym_private_property_identifier] = ACTIONS(119),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(143),
    [aux_sym_method_definition_token1] = ACTIONS(123),
    [anon_sym_get] = ACTIONS(145),
    [anon_sym_set] = ACTIONS(145),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(6)] = {
    [sym_export_statement] = STATE(471),
    [sym_declaration] = STATE(471),
    [sym_import] = STATE(1326),
    [sym_import_statement] = STATE(471),
    [sym_statement] = STATE(24),
    [sym_expression_statement] = STATE(471),
    [sym_variable_declaration] = STATE(419),
    [sym_lexical_declaration] = STATE(419),
    [sym_using_declaration] = STATE(419),
    [sym_statement_block] = STATE(471),
    [sym_if_statement] = STATE(471),
    [sym_switch_statement] = STATE(471),
    [sym_for_statement] = STATE(471),
    [sym_for_in_statement] = STATE(471),
    [sym_while_statement] = STATE(471),
    [sym_do_statement] = STATE(471),
    [sym_try_statement] = STATE(471),
    [sym_with_statement] = STATE(471),
    [sym_break_statement] = STATE(471),
    [sym_continue_statement] = STATE(471),
    [sym_debugger_statement] = STATE(471),
    [sym_return_statement] = STATE(471),
    [sym_throw_statement] = STATE(471),
    [sym_empty_statement] = STATE(471),
    [sym_labeled_statement] = STATE(471),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(823),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1853),
    [sym_object_assignment_pattern] = STATE(1454),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1853),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_class_declaration] = STATE(419),
    [sym_function_expression] = STATE(740),
    [sym_function_declaration] = STATE(419),
    [sym_generator_function] = STATE(740),
    [sym_generator_function_declaration] = STATE(419),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1853),
    [sym_spread_element] = STATE(1458),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_sequence_expression] = STATE(1662),
    [sym_string] = STATE(815),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [sym_rest_pattern] = STATE(1454),
    [sym_method_definition] = STATE(1458),
    [sym_pair] = STATE(1458),
    [sym_pair_pattern] = STATE(1454),
    [sym__property_name] = STATE(1407),
    [sym_computed_property_name] = STATE(1407),
    [aux_sym_program_repeat1] = STATE(24),
    [aux_sym_export_statement_repeat1] = STATE(1014),
    [aux_sym_object_repeat1] = STATE(1461),
    [aux_sym_object_pattern_repeat1] = STATE(1476),
    [sym_identifier] = ACTIONS(147),
    [anon_sym_export] = ACTIONS(149),
    [anon_sym_STAR] = ACTIONS(101),
    [anon_sym_LBRACE] = ACTIONS(15),
    [anon_sym_COMMA] = ACTIONS(103),
    [anon_sym_RBRACE] = ACTIONS(135),
    [anon_sym_import] = ACTIONS(17),
    [anon_sym_with] = ACTIONS(19),
    [anon_sym_var] = ACTIONS(21),
    [anon_sym_let] = ACTIONS(151),
    [anon_sym_const] = ACTIONS(25),
    [anon_sym_using] = ACTIONS(27),
    [anon_sym_await] = ACTIONS(153),
    [anon_sym_if] = ACTIONS(31),
    [anon_sym_switch] = ACTIONS(33),
    [anon_sym_for] = ACTIONS(35),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_SEMI] = ACTIONS(39),
    [anon_sym_while] = ACTIONS(41),
    [anon_sym_do] = ACTIONS(43),
    [anon_sym_try] = ACTIONS(45),
    [anon_sym_break] = ACTIONS(47),
    [anon_sym_continue] = ACTIONS(49),
    [anon_sym_debugger] = ACTIONS(51),
    [anon_sym_return] = ACTIONS(53),
    [anon_sym_throw] = ACTIONS(55),
    [anon_sym_yield] = ACTIONS(57),
    [anon_sym_LBRACK] = ACTIONS(111),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(67),
    [anon_sym_async] = ACTIONS(155),
    [anon_sym_function] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(115),
    [anon_sym_PLUS] = ACTIONS(75),
    [anon_sym_DASH] = ACTIONS(75),
    [anon_sym_SLASH] = ACTIONS(77),
    [anon_sym_BANG] = ACTIONS(79),
    [anon_sym_TILDE] = ACTIONS(79),
    [anon_sym_typeof] = ACTIONS(75),
    [anon_sym_void] = ACTIONS(75),
    [anon_sym_delete] = ACTIONS(75),
    [anon_sym_PLUS_PLUS] = ACTIONS(81),
    [anon_sym_DASH_DASH] = ACTIONS(81),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(117),
    [sym_private_property_identifier] = ACTIONS(119),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(157),
    [aux_sym_method_definition_token1] = ACTIONS(123),
    [anon_sym_get] = ACTIONS(159),
    [anon_sym_set] = ACTIONS(159),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(7)] = {
    [sym_import] = STATE(1326),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(692),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1790),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1790),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_function_expression] = STATE(740),
    [sym_generator_function] = STATE(740),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1790),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [aux_sym_export_statement_repeat1] = STATE(1381),
    [aux_sym_object_repeat1] = STATE(1487),
    [aux_sym_object_pattern_repeat1] = STATE(1452),
    [sym_identifier] = ACTIONS(161),
    [anon_sym_export] = ACTIONS(163),
    [anon_sym_STAR] = ACTIONS(165),
    [anon_sym_LBRACE] = ACTIONS(167),
    [anon_sym_COMMA] = ACTIONS(169),
    [anon_sym_RBRACE] = ACTIONS(171),
    [anon_sym_import] = ACTIONS(173),
    [anon_sym_let] = ACTIONS(163),
    [anon_sym_using] = ACTIONS(175),
    [anon_sym_await] = ACTIONS(177),
    [anon_sym_LPAREN] = ACTIONS(179),
    [anon_sym_SEMI] = ACTIONS(169),
    [anon_sym_in] = ACTIONS(165),
    [anon_sym_COLON] = ACTIONS(183),
    [anon_sym_yield] = ACTIONS(57),
    [anon_sym_EQ] = ACTIONS(186),
    [anon_sym_LBRACK] = ACTIONS(188),
    [anon_sym_LT] = ACTIONS(191),
    [anon_sym_GT] = ACTIONS(165),
    [anon_sym_DOT] = ACTIONS(165),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(194),
    [anon_sym_async] = ACTIONS(196),
    [anon_sym_function] = ACTIONS(198),
    [anon_sym_EQ_GT] = ACTIONS(200),
    [sym_optional_chain] = ACTIONS(169),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_PLUS_EQ] = ACTIONS(202),
    [anon_sym_DASH_EQ] = ACTIONS(202),
    [anon_sym_STAR_EQ] = ACTIONS(202),
    [anon_sym_SLASH_EQ] = ACTIONS(202),
    [anon_sym_PERCENT_EQ] = ACTIONS(202),
    [anon_sym_CARET_EQ] = ACTIONS(202),
    [anon_sym_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_LT_LT_EQ] = ACTIONS(202),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(202),
    [anon_sym_QMARK_QMARK_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP] = ACTIONS(165),
    [anon_sym_PIPE_PIPE] = ACTIONS(165),
    [anon_sym_GT_GT] = ACTIONS(165),
    [anon_sym_GT_GT_GT] = ACTIONS(165),
    [anon_sym_LT_LT] = ACTIONS(165),
    [anon_sym_AMP] = ACTIONS(165),
    [anon_sym_CARET] = ACTIONS(165),
    [anon_sym_PIPE] = ACTIONS(165),
    [anon_sym_PLUS] = ACTIONS(204),
    [anon_sym_DASH] = ACTIONS(204),
    [anon_sym_SLASH] = ACTIONS(207),
    [anon_sym_PERCENT] = ACTIONS(165),
    [anon_sym_STAR_STAR] = ACTIONS(165),
    [anon_sym_LT_EQ] = ACTIONS(169),
    [anon_sym_EQ_EQ] = ACTIONS(165),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(169),
    [anon_sym_BANG_EQ] = ACTIONS(165),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(169),
    [anon_sym_GT_EQ] = ACTIONS(169),
    [anon_sym_QMARK_QMARK] = ACTIONS(165),
    [anon_sym_instanceof] = ACTIONS(165),
    [anon_sym_BANG] = ACTIONS(75),
    [anon_sym_TILDE] = ACTIONS(79),
    [anon_sym_typeof] = ACTIONS(75),
    [anon_sym_void] = ACTIONS(75),
    [anon_sym_delete] = ACTIONS(75),
    [anon_sym_PLUS_PLUS] = ACTIONS(210),
    [anon_sym_DASH_DASH] = ACTIONS(210),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(213),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(87),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(163),
    [anon_sym_get] = ACTIONS(163),
    [anon_sym_set] = ACTIONS(163),
    [sym__automatic_semicolon] = ACTIONS(169),
    [sym__ternary_qmark] = ACTIONS(169),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(8)] = {
    [sym_import] = STATE(1326),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(692),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1790),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1790),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_function_expression] = STATE(740),
    [sym_generator_function] = STATE(740),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1790),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [aux_sym_export_statement_repeat1] = STATE(1381),
    [aux_sym_object_repeat1] = STATE(1418),
    [aux_sym_object_pattern_repeat1] = STATE(1452),
    [sym_identifier] = ACTIONS(161),
    [anon_sym_export] = ACTIONS(163),
    [anon_sym_STAR] = ACTIONS(165),
    [anon_sym_LBRACE] = ACTIONS(167),
    [anon_sym_COMMA] = ACTIONS(169),
    [anon_sym_RBRACE] = ACTIONS(216),
    [anon_sym_import] = ACTIONS(173),
    [anon_sym_let] = ACTIONS(163),
    [anon_sym_using] = ACTIONS(175),
    [anon_sym_await] = ACTIONS(177),
    [anon_sym_LPAREN] = ACTIONS(179),
    [anon_sym_SEMI] = ACTIONS(169),
    [anon_sym_in] = ACTIONS(165),
    [anon_sym_COLON] = ACTIONS(183),
    [anon_sym_yield] = ACTIONS(57),
    [anon_sym_EQ] = ACTIONS(186),
    [anon_sym_LBRACK] = ACTIONS(188),
    [anon_sym_LT] = ACTIONS(191),
    [anon_sym_GT] = ACTIONS(165),
    [anon_sym_DOT] = ACTIONS(165),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(194),
    [anon_sym_async] = ACTIONS(196),
    [anon_sym_function] = ACTIONS(198),
    [anon_sym_EQ_GT] = ACTIONS(200),
    [sym_optional_chain] = ACTIONS(169),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_PLUS_EQ] = ACTIONS(202),
    [anon_sym_DASH_EQ] = ACTIONS(202),
    [anon_sym_STAR_EQ] = ACTIONS(202),
    [anon_sym_SLASH_EQ] = ACTIONS(202),
    [anon_sym_PERCENT_EQ] = ACTIONS(202),
    [anon_sym_CARET_EQ] = ACTIONS(202),
    [anon_sym_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_LT_LT_EQ] = ACTIONS(202),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(202),
    [anon_sym_QMARK_QMARK_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP] = ACTIONS(165),
    [anon_sym_PIPE_PIPE] = ACTIONS(165),
    [anon_sym_GT_GT] = ACTIONS(165),
    [anon_sym_GT_GT_GT] = ACTIONS(165),
    [anon_sym_LT_LT] = ACTIONS(165),
    [anon_sym_AMP] = ACTIONS(165),
    [anon_sym_CARET] = ACTIONS(165),
    [anon_sym_PIPE] = ACTIONS(165),
    [anon_sym_PLUS] = ACTIONS(204),
    [anon_sym_DASH] = ACTIONS(204),
    [anon_sym_SLASH] = ACTIONS(207),
    [anon_sym_PERCENT] = ACTIONS(165),
    [anon_sym_STAR_STAR] = ACTIONS(165),
    [anon_sym_LT_EQ] = ACTIONS(169),
    [anon_sym_EQ_EQ] = ACTIONS(165),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(169),
    [anon_sym_BANG_EQ] = ACTIONS(165),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(169),
    [anon_sym_GT_EQ] = ACTIONS(169),
    [anon_sym_QMARK_QMARK] = ACTIONS(165),
    [anon_sym_instanceof] = ACTIONS(165),
    [anon_sym_BANG] = ACTIONS(75),
    [anon_sym_TILDE] = ACTIONS(79),
    [anon_sym_typeof] = ACTIONS(75),
    [anon_sym_void] = ACTIONS(75),
    [anon_sym_delete] = ACTIONS(75),
    [anon_sym_PLUS_PLUS] = ACTIONS(210),
    [anon_sym_DASH_DASH] = ACTIONS(210),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(213),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(87),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(163),
    [anon_sym_get] = ACTIONS(163),
    [anon_sym_set] = ACTIONS(163),
    [sym__automatic_semicolon] = ACTIONS(169),
    [sym__ternary_qmark] = ACTIONS(169),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(9)] = {
    [sym_export_statement] = STATE(471),
    [sym_declaration] = STATE(471),
    [sym_import] = STATE(1326),
    [sym_import_statement] = STATE(471),
    [sym_statement] = STATE(9),
    [sym_expression_statement] = STATE(471),
    [sym_variable_declaration] = STATE(419),
    [sym_lexical_declaration] = STATE(419),
    [sym_using_declaration] = STATE(419),
    [sym_statement_block] = STATE(471),
    [sym_if_statement] = STATE(471),
    [sym_switch_statement] = STATE(471),
    [sym_for_statement] = STATE(471),
    [sym_for_in_statement] = STATE(471),
    [sym_while_statement] = STATE(471),
    [sym_do_statement] = STATE(471),
    [sym_try_statement] = STATE(471),
    [sym_with_statement] = STATE(471),
    [sym_break_statement] = STATE(471),
    [sym_continue_statement] = STATE(471),
    [sym_debugger_statement] = STATE(471),
    [sym_return_statement] = STATE(471),
    [sym_throw_statement] = STATE(471),
    [sym_empty_statement] = STATE(471),
    [sym_labeled_statement] = STATE(471),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(823),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1790),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1790),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_class_declaration] = STATE(419),
    [sym_function_expression] = STATE(740),
    [sym_function_declaration] = STATE(419),
    [sym_generator_function] = STATE(740),
    [sym_generator_function_declaration] = STATE(419),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1790),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_sequence_expression] = STATE(1662),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [aux_sym_program_repeat1] = STATE(9),
    [aux_sym_export_statement_repeat1] = STATE(1267),
    [ts_builtin_sym_end] = ACTIONS(218),
    [sym_identifier] = ACTIONS(220),
    [anon_sym_export] = ACTIONS(223),
    [anon_sym_default] = ACTIONS(226),
    [anon_sym_LBRACE] = ACTIONS(228),
    [anon_sym_RBRACE] = ACTIONS(218),
    [anon_sym_import] = ACTIONS(231),
    [anon_sym_with] = ACTIONS(234),
    [anon_sym_var] = ACTIONS(237),
    [anon_sym_let] = ACTIONS(240),
    [anon_sym_const] = ACTIONS(243),
    [anon_sym_using] = ACTIONS(246),
    [anon_sym_await] = ACTIONS(249),
    [anon_sym_if] = ACTIONS(252),
    [anon_sym_switch] = ACTIONS(255),
    [anon_sym_for] = ACTIONS(258),
    [anon_sym_LPAREN] = ACTIONS(261),
    [anon_sym_SEMI] = ACTIONS(264),
    [anon_sym_while] = ACTIONS(267),
    [anon_sym_do] = ACTIONS(270),
    [anon_sym_try] = ACTIONS(273),
    [anon_sym_break] = ACTIONS(276),
    [anon_sym_continue] = ACTIONS(279),
    [anon_sym_debugger] = ACTIONS(282),
    [anon_sym_return] = ACTIONS(285),
    [anon_sym_throw] = ACTIONS(288),
    [anon_sym_case] = ACTIONS(226),
    [anon_sym_yield] = ACTIONS(291),
    [anon_sym_LBRACK] = ACTIONS(294),
    [anon_sym_LT] = ACTIONS(297),
    [anon_sym_DQUOTE] = ACTIONS(300),
    [anon_sym_SQUOTE] = ACTIONS(303),
    [anon_sym_class] = ACTIONS(306),
    [anon_sym_async] = ACTIONS(309),
    [anon_sym_function] = ACTIONS(312),
    [anon_sym_new] = ACTIONS(315),
    [anon_sym_PLUS] = ACTIONS(318),
    [anon_sym_DASH] = ACTIONS(318),
    [anon_sym_SLASH] = ACTIONS(321),
    [anon_sym_BANG] = ACTIONS(324),
    [anon_sym_TILDE] = ACTIONS(324),
    [anon_sym_typeof] = ACTIONS(318),
    [anon_sym_void] = ACTIONS(318),
    [anon_sym_delete] = ACTIONS(318),
    [anon_sym_PLUS_PLUS] = ACTIONS(327),
    [anon_sym_DASH_DASH] = ACTIONS(327),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(330),
    [sym_number] = ACTIONS(333),
    [sym_private_property_identifier] = ACTIONS(336),
    [sym_this] = ACTIONS(339),
    [sym_super] = ACTIONS(339),
    [sym_true] = ACTIONS(339),
    [sym_false] = ACTIONS(339),
    [sym_null] = ACTIONS(339),
    [sym_undefined] = ACTIONS(342),
    [anon_sym_AT] = ACTIONS(345),
    [anon_sym_static] = ACTIONS(348),
    [anon_sym_get] = ACTIONS(348),
    [anon_sym_set] = ACTIONS(348),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(10)] = {
    [sym_import] = STATE(1326),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(692),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1790),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1790),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_function_expression] = STATE(740),
    [sym_generator_function] = STATE(740),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1790),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [aux_sym_export_statement_repeat1] = STATE(1381),
    [aux_sym_object_repeat1] = STATE(1418),
    [aux_sym_object_pattern_repeat1] = STATE(1452),
    [sym_identifier] = ACTIONS(161),
    [anon_sym_export] = ACTIONS(163),
    [anon_sym_STAR] = ACTIONS(165),
    [anon_sym_LBRACE] = ACTIONS(167),
    [anon_sym_COMMA] = ACTIONS(169),
    [anon_sym_RBRACE] = ACTIONS(351),
    [anon_sym_import] = ACTIONS(173),
    [anon_sym_let] = ACTIONS(163),
    [anon_sym_using] = ACTIONS(175),
    [anon_sym_await] = ACTIONS(177),
    [anon_sym_LPAREN] = ACTIONS(179),
    [anon_sym_SEMI] = ACTIONS(169),
    [anon_sym_in] = ACTIONS(165),
    [anon_sym_COLON] = ACTIONS(183),
    [anon_sym_yield] = ACTIONS(57),
    [anon_sym_EQ] = ACTIONS(186),
    [anon_sym_LBRACK] = ACTIONS(188),
    [anon_sym_LT] = ACTIONS(191),
    [anon_sym_GT] = ACTIONS(165),
    [anon_sym_DOT] = ACTIONS(165),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(194),
    [anon_sym_async] = ACTIONS(196),
    [anon_sym_function] = ACTIONS(198),
    [anon_sym_EQ_GT] = ACTIONS(200),
    [sym_optional_chain] = ACTIONS(169),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_PLUS_EQ] = ACTIONS(202),
    [anon_sym_DASH_EQ] = ACTIONS(202),
    [anon_sym_STAR_EQ] = ACTIONS(202),
    [anon_sym_SLASH_EQ] = ACTIONS(202),
    [anon_sym_PERCENT_EQ] = ACTIONS(202),
    [anon_sym_CARET_EQ] = ACTIONS(202),
    [anon_sym_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_LT_LT_EQ] = ACTIONS(202),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(202),
    [anon_sym_QMARK_QMARK_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP] = ACTIONS(165),
    [anon_sym_PIPE_PIPE] = ACTIONS(165),
    [anon_sym_GT_GT] = ACTIONS(165),
    [anon_sym_GT_GT_GT] = ACTIONS(165),
    [anon_sym_LT_LT] = ACTIONS(165),
    [anon_sym_AMP] = ACTIONS(165),
    [anon_sym_CARET] = ACTIONS(165),
    [anon_sym_PIPE] = ACTIONS(165),
    [anon_sym_PLUS] = ACTIONS(204),
    [anon_sym_DASH] = ACTIONS(204),
    [anon_sym_SLASH] = ACTIONS(207),
    [anon_sym_PERCENT] = ACTIONS(165),
    [anon_sym_STAR_STAR] = ACTIONS(165),
    [anon_sym_LT_EQ] = ACTIONS(169),
    [anon_sym_EQ_EQ] = ACTIONS(165),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(169),
    [anon_sym_BANG_EQ] = ACTIONS(165),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(169),
    [anon_sym_GT_EQ] = ACTIONS(169),
    [anon_sym_QMARK_QMARK] = ACTIONS(165),
    [anon_sym_instanceof] = ACTIONS(165),
    [anon_sym_BANG] = ACTIONS(75),
    [anon_sym_TILDE] = ACTIONS(79),
    [anon_sym_typeof] = ACTIONS(75),
    [anon_sym_void] = ACTIONS(75),
    [anon_sym_delete] = ACTIONS(75),
    [anon_sym_PLUS_PLUS] = ACTIONS(210),
    [anon_sym_DASH_DASH] = ACTIONS(210),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(213),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(87),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(163),
    [anon_sym_get] = ACTIONS(163),
    [anon_sym_set] = ACTIONS(163),
    [sym__automatic_semicolon] = ACTIONS(169),
    [sym__ternary_qmark] = ACTIONS(169),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(11)] = {
    [sym_export_statement] = STATE(471),
    [sym_declaration] = STATE(471),
    [sym_import] = STATE(1326),
    [sym_import_statement] = STATE(471),
    [sym_statement] = STATE(12),
    [sym_expression_statement] = STATE(471),
    [sym_variable_declaration] = STATE(419),
    [sym_lexical_declaration] = STATE(419),
    [sym_using_declaration] = STATE(419),
    [sym_statement_block] = STATE(471),
    [sym_if_statement] = STATE(471),
    [sym_switch_statement] = STATE(471),
    [sym_for_statement] = STATE(471),
    [sym_for_in_statement] = STATE(471),
    [sym_while_statement] = STATE(471),
    [sym_do_statement] = STATE(471),
    [sym_try_statement] = STATE(471),
    [sym_with_statement] = STATE(471),
    [sym_break_statement] = STATE(471),
    [sym_continue_statement] = STATE(471),
    [sym_debugger_statement] = STATE(471),
    [sym_return_statement] = STATE(471),
    [sym_throw_statement] = STATE(471),
    [sym_empty_statement] = STATE(471),
    [sym_labeled_statement] = STATE(471),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(823),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1790),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1790),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_class_declaration] = STATE(419),
    [sym_function_expression] = STATE(740),
    [sym_function_declaration] = STATE(419),
    [sym_generator_function] = STATE(740),
    [sym_generator_function_declaration] = STATE(419),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1790),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_sequence_expression] = STATE(1662),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [aux_sym_program_repeat1] = STATE(12),
    [aux_sym_export_statement_repeat1] = STATE(1267),
    [sym_identifier] = ACTIONS(9),
    [anon_sym_export] = ACTIONS(13),
    [anon_sym_default] = ACTIONS(353),
    [anon_sym_LBRACE] = ACTIONS(15),
    [anon_sym_RBRACE] = ACTIONS(355),
    [anon_sym_import] = ACTIONS(17),
    [anon_sym_with] = ACTIONS(19),
    [anon_sym_var] = ACTIONS(21),
    [anon_sym_let] = ACTIONS(23),
    [anon_sym_const] = ACTIONS(25),
    [anon_sym_using] = ACTIONS(27),
    [anon_sym_await] = ACTIONS(29),
    [anon_sym_if] = ACTIONS(31),
    [anon_sym_switch] = ACTIONS(33),
    [anon_sym_for] = ACTIONS(35),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_SEMI] = ACTIONS(39),
    [anon_sym_while] = ACTIONS(41),
    [anon_sym_do] = ACTIONS(43),
    [anon_sym_try] = ACTIONS(45),
    [anon_sym_break] = ACTIONS(47),
    [anon_sym_continue] = ACTIONS(49),
    [anon_sym_debugger] = ACTIONS(51),
    [anon_sym_return] = ACTIONS(53),
    [anon_sym_throw] = ACTIONS(55),
    [anon_sym_case] = ACTIONS(353),
    [anon_sym_yield] = ACTIONS(57),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(67),
    [anon_sym_async] = ACTIONS(69),
    [anon_sym_function] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(75),
    [anon_sym_DASH] = ACTIONS(75),
    [anon_sym_SLASH] = ACTIONS(77),
    [anon_sym_BANG] = ACTIONS(79),
    [anon_sym_TILDE] = ACTIONS(79),
    [anon_sym_typeof] = ACTIONS(75),
    [anon_sym_void] = ACTIONS(75),
    [anon_sym_delete] = ACTIONS(75),
    [anon_sym_PLUS_PLUS] = ACTIONS(81),
    [anon_sym_DASH_DASH] = ACTIONS(81),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(87),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(95),
    [anon_sym_get] = ACTIONS(95),
    [anon_sym_set] = ACTIONS(95),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(12)] = {
    [sym_export_statement] = STATE(471),
    [sym_declaration] = STATE(471),
    [sym_import] = STATE(1326),
    [sym_import_statement] = STATE(471),
    [sym_statement] = STATE(9),
    [sym_expression_statement] = STATE(471),
    [sym_variable_declaration] = STATE(419),
    [sym_lexical_declaration] = STATE(419),
    [sym_using_declaration] = STATE(419),
    [sym_statement_block] = STATE(471),
    [sym_if_statement] = STATE(471),
    [sym_switch_statement] = STATE(471),
    [sym_for_statement] = STATE(471),
    [sym_for_in_statement] = STATE(471),
    [sym_while_statement] = STATE(471),
    [sym_do_statement] = STATE(471),
    [sym_try_statement] = STATE(471),
    [sym_with_statement] = STATE(471),
    [sym_break_statement] = STATE(471),
    [sym_continue_statement] = STATE(471),
    [sym_debugger_statement] = STATE(471),
    [sym_return_statement] = STATE(471),
    [sym_throw_statement] = STATE(471),
    [sym_empty_statement] = STATE(471),
    [sym_labeled_statement] = STATE(471),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(823),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1790),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1790),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_class_declaration] = STATE(419),
    [sym_function_expression] = STATE(740),
    [sym_function_declaration] = STATE(419),
    [sym_generator_function] = STATE(740),
    [sym_generator_function_declaration] = STATE(419),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1790),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_sequence_expression] = STATE(1662),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [aux_sym_program_repeat1] = STATE(9),
    [aux_sym_export_statement_repeat1] = STATE(1267),
    [sym_identifier] = ACTIONS(9),
    [anon_sym_export] = ACTIONS(13),
    [anon_sym_default] = ACTIONS(357),
    [anon_sym_LBRACE] = ACTIONS(15),
    [anon_sym_RBRACE] = ACTIONS(359),
    [anon_sym_import] = ACTIONS(17),
    [anon_sym_with] = ACTIONS(19),
    [anon_sym_var] = ACTIONS(21),
    [anon_sym_let] = ACTIONS(23),
    [anon_sym_const] = ACTIONS(25),
    [anon_sym_using] = ACTIONS(27),
    [anon_sym_await] = ACTIONS(29),
    [anon_sym_if] = ACTIONS(31),
    [anon_sym_switch] = ACTIONS(33),
    [anon_sym_for] = ACTIONS(35),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_SEMI] = ACTIONS(39),
    [anon_sym_while] = ACTIONS(41),
    [anon_sym_do] = ACTIONS(43),
    [anon_sym_try] = ACTIONS(45),
    [anon_sym_break] = ACTIONS(47),
    [anon_sym_continue] = ACTIONS(49),
    [anon_sym_debugger] = ACTIONS(51),
    [anon_sym_return] = ACTIONS(53),
    [anon_sym_throw] = ACTIONS(55),
    [anon_sym_case] = ACTIONS(357),
    [anon_sym_yield] = ACTIONS(57),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(67),
    [anon_sym_async] = ACTIONS(69),
    [anon_sym_function] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(75),
    [anon_sym_DASH] = ACTIONS(75),
    [anon_sym_SLASH] = ACTIONS(77),
    [anon_sym_BANG] = ACTIONS(79),
    [anon_sym_TILDE] = ACTIONS(79),
    [anon_sym_typeof] = ACTIONS(75),
    [anon_sym_void] = ACTIONS(75),
    [anon_sym_delete] = ACTIONS(75),
    [anon_sym_PLUS_PLUS] = ACTIONS(81),
    [anon_sym_DASH_DASH] = ACTIONS(81),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(87),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(95),
    [anon_sym_get] = ACTIONS(95),
    [anon_sym_set] = ACTIONS(95),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(13)] = {
    [sym_export_statement] = STATE(471),
    [sym_declaration] = STATE(471),
    [sym_import] = STATE(1326),
    [sym_import_statement] = STATE(471),
    [sym_statement] = STATE(9),
    [sym_expression_statement] = STATE(471),
    [sym_variable_declaration] = STATE(419),
    [sym_lexical_declaration] = STATE(419),
    [sym_using_declaration] = STATE(419),
    [sym_statement_block] = STATE(471),
    [sym_if_statement] = STATE(471),
    [sym_switch_statement] = STATE(471),
    [sym_for_statement] = STATE(471),
    [sym_for_in_statement] = STATE(471),
    [sym_while_statement] = STATE(471),
    [sym_do_statement] = STATE(471),
    [sym_try_statement] = STATE(471),
    [sym_with_statement] = STATE(471),
    [sym_break_statement] = STATE(471),
    [sym_continue_statement] = STATE(471),
    [sym_debugger_statement] = STATE(471),
    [sym_return_statement] = STATE(471),
    [sym_throw_statement] = STATE(471),
    [sym_empty_statement] = STATE(471),
    [sym_labeled_statement] = STATE(471),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(823),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1790),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1790),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_class_declaration] = STATE(419),
    [sym_function_expression] = STATE(740),
    [sym_function_declaration] = STATE(419),
    [sym_generator_function] = STATE(740),
    [sym_generator_function_declaration] = STATE(419),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1790),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_sequence_expression] = STATE(1662),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [aux_sym_program_repeat1] = STATE(9),
    [aux_sym_export_statement_repeat1] = STATE(1267),
    [sym_identifier] = ACTIONS(9),
    [anon_sym_export] = ACTIONS(13),
    [anon_sym_default] = ACTIONS(361),
    [anon_sym_LBRACE] = ACTIONS(15),
    [anon_sym_RBRACE] = ACTIONS(363),
    [anon_sym_import] = ACTIONS(17),
    [anon_sym_with] = ACTIONS(19),
    [anon_sym_var] = ACTIONS(21),
    [anon_sym_let] = ACTIONS(23),
    [anon_sym_const] = ACTIONS(25),
    [anon_sym_using] = ACTIONS(27),
    [anon_sym_await] = ACTIONS(29),
    [anon_sym_if] = ACTIONS(31),
    [anon_sym_switch] = ACTIONS(33),
    [anon_sym_for] = ACTIONS(35),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_SEMI] = ACTIONS(39),
    [anon_sym_while] = ACTIONS(41),
    [anon_sym_do] = ACTIONS(43),
    [anon_sym_try] = ACTIONS(45),
    [anon_sym_break] = ACTIONS(47),
    [anon_sym_continue] = ACTIONS(49),
    [anon_sym_debugger] = ACTIONS(51),
    [anon_sym_return] = ACTIONS(53),
    [anon_sym_throw] = ACTIONS(55),
    [anon_sym_case] = ACTIONS(361),
    [anon_sym_yield] = ACTIONS(57),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(67),
    [anon_sym_async] = ACTIONS(69),
    [anon_sym_function] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(75),
    [anon_sym_DASH] = ACTIONS(75),
    [anon_sym_SLASH] = ACTIONS(77),
    [anon_sym_BANG] = ACTIONS(79),
    [anon_sym_TILDE] = ACTIONS(79),
    [anon_sym_typeof] = ACTIONS(75),
    [anon_sym_void] = ACTIONS(75),
    [anon_sym_delete] = ACTIONS(75),
    [anon_sym_PLUS_PLUS] = ACTIONS(81),
    [anon_sym_DASH_DASH] = ACTIONS(81),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(87),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(95),
    [anon_sym_get] = ACTIONS(95),
    [anon_sym_set] = ACTIONS(95),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(14)] = {
    [sym_export_statement] = STATE(471),
    [sym_declaration] = STATE(471),
    [sym_import] = STATE(1326),
    [sym_import_statement] = STATE(471),
    [sym_statement] = STATE(13),
    [sym_expression_statement] = STATE(471),
    [sym_variable_declaration] = STATE(419),
    [sym_lexical_declaration] = STATE(419),
    [sym_using_declaration] = STATE(419),
    [sym_statement_block] = STATE(471),
    [sym_if_statement] = STATE(471),
    [sym_switch_statement] = STATE(471),
    [sym_for_statement] = STATE(471),
    [sym_for_in_statement] = STATE(471),
    [sym_while_statement] = STATE(471),
    [sym_do_statement] = STATE(471),
    [sym_try_statement] = STATE(471),
    [sym_with_statement] = STATE(471),
    [sym_break_statement] = STATE(471),
    [sym_continue_statement] = STATE(471),
    [sym_debugger_statement] = STATE(471),
    [sym_return_statement] = STATE(471),
    [sym_throw_statement] = STATE(471),
    [sym_empty_statement] = STATE(471),
    [sym_labeled_statement] = STATE(471),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(823),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1790),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1790),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_class_declaration] = STATE(419),
    [sym_function_expression] = STATE(740),
    [sym_function_declaration] = STATE(419),
    [sym_generator_function] = STATE(740),
    [sym_generator_function_declaration] = STATE(419),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1790),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_sequence_expression] = STATE(1662),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [aux_sym_program_repeat1] = STATE(13),
    [aux_sym_export_statement_repeat1] = STATE(1267),
    [sym_identifier] = ACTIONS(9),
    [anon_sym_export] = ACTIONS(13),
    [anon_sym_default] = ACTIONS(365),
    [anon_sym_LBRACE] = ACTIONS(15),
    [anon_sym_RBRACE] = ACTIONS(367),
    [anon_sym_import] = ACTIONS(17),
    [anon_sym_with] = ACTIONS(19),
    [anon_sym_var] = ACTIONS(21),
    [anon_sym_let] = ACTIONS(23),
    [anon_sym_const] = ACTIONS(25),
    [anon_sym_using] = ACTIONS(27),
    [anon_sym_await] = ACTIONS(29),
    [anon_sym_if] = ACTIONS(31),
    [anon_sym_switch] = ACTIONS(33),
    [anon_sym_for] = ACTIONS(35),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_SEMI] = ACTIONS(39),
    [anon_sym_while] = ACTIONS(41),
    [anon_sym_do] = ACTIONS(43),
    [anon_sym_try] = ACTIONS(45),
    [anon_sym_break] = ACTIONS(47),
    [anon_sym_continue] = ACTIONS(49),
    [anon_sym_debugger] = ACTIONS(51),
    [anon_sym_return] = ACTIONS(53),
    [anon_sym_throw] = ACTIONS(55),
    [anon_sym_case] = ACTIONS(365),
    [anon_sym_yield] = ACTIONS(57),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(67),
    [anon_sym_async] = ACTIONS(69),
    [anon_sym_function] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(75),
    [anon_sym_DASH] = ACTIONS(75),
    [anon_sym_SLASH] = ACTIONS(77),
    [anon_sym_BANG] = ACTIONS(79),
    [anon_sym_TILDE] = ACTIONS(79),
    [anon_sym_typeof] = ACTIONS(75),
    [anon_sym_void] = ACTIONS(75),
    [anon_sym_delete] = ACTIONS(75),
    [anon_sym_PLUS_PLUS] = ACTIONS(81),
    [anon_sym_DASH_DASH] = ACTIONS(81),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(87),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(95),
    [anon_sym_get] = ACTIONS(95),
    [anon_sym_set] = ACTIONS(95),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(15)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(507),
    [sym_expression] = STATE(587),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1815),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1815),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(507),
    [sym_subscript_expression] = STATE(507),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1139),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1815),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1763),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(369),
    [anon_sym_export] = ACTIONS(371),
    [anon_sym_STAR] = ACTIONS(165),
    [anon_sym_LBRACE] = ACTIONS(373),
    [anon_sym_COMMA] = ACTIONS(169),
    [anon_sym_RBRACE] = ACTIONS(169),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(371),
    [anon_sym_await] = ACTIONS(377),
    [anon_sym_LPAREN] = ACTIONS(379),
    [anon_sym_SEMI] = ACTIONS(169),
    [anon_sym_RPAREN] = ACTIONS(169),
    [anon_sym_in] = ACTIONS(165),
    [anon_sym_COLON] = ACTIONS(169),
    [anon_sym_yield] = ACTIONS(382),
    [anon_sym_EQ] = ACTIONS(384),
    [anon_sym_LBRACK] = ACTIONS(386),
    [anon_sym_RBRACK] = ACTIONS(169),
    [anon_sym_LT] = ACTIONS(389),
    [anon_sym_GT] = ACTIONS(165),
    [anon_sym_DOT] = ACTIONS(165),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(398),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_EQ_GT] = ACTIONS(402),
    [sym_optional_chain] = ACTIONS(169),
    [anon_sym_new] = ACTIONS(404),
    [anon_sym_PLUS_EQ] = ACTIONS(202),
    [anon_sym_DASH_EQ] = ACTIONS(202),
    [anon_sym_STAR_EQ] = ACTIONS(202),
    [anon_sym_SLASH_EQ] = ACTIONS(202),
    [anon_sym_PERCENT_EQ] = ACTIONS(202),
    [anon_sym_CARET_EQ] = ACTIONS(202),
    [anon_sym_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_LT_LT_EQ] = ACTIONS(202),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(202),
    [anon_sym_QMARK_QMARK_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP] = ACTIONS(165),
    [anon_sym_PIPE_PIPE] = ACTIONS(165),
    [anon_sym_GT_GT] = ACTIONS(165),
    [anon_sym_GT_GT_GT] = ACTIONS(165),
    [anon_sym_LT_LT] = ACTIONS(165),
    [anon_sym_AMP] = ACTIONS(165),
    [anon_sym_CARET] = ACTIONS(165),
    [anon_sym_PIPE] = ACTIONS(165),
    [anon_sym_PLUS] = ACTIONS(406),
    [anon_sym_DASH] = ACTIONS(406),
    [anon_sym_SLASH] = ACTIONS(409),
    [anon_sym_PERCENT] = ACTIONS(165),
    [anon_sym_STAR_STAR] = ACTIONS(165),
    [anon_sym_LT_EQ] = ACTIONS(169),
    [anon_sym_EQ_EQ] = ACTIONS(165),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(169),
    [anon_sym_BANG_EQ] = ACTIONS(165),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(169),
    [anon_sym_GT_EQ] = ACTIONS(169),
    [anon_sym_QMARK_QMARK] = ACTIONS(165),
    [anon_sym_instanceof] = ACTIONS(165),
    [anon_sym_BANG] = ACTIONS(412),
    [anon_sym_TILDE] = ACTIONS(414),
    [anon_sym_typeof] = ACTIONS(412),
    [anon_sym_void] = ACTIONS(412),
    [anon_sym_delete] = ACTIONS(412),
    [anon_sym_PLUS_PLUS] = ACTIONS(416),
    [anon_sym_DASH_DASH] = ACTIONS(416),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(419),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(424),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(428),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(371),
    [anon_sym_get] = ACTIONS(371),
    [anon_sym_set] = ACTIONS(371),
    [sym__ternary_qmark] = ACTIONS(169),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(16)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(498),
    [sym_expression] = STATE(587),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1762),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1762),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(498),
    [sym_subscript_expression] = STATE(498),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1138),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1762),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1816),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(430),
    [anon_sym_export] = ACTIONS(432),
    [anon_sym_STAR] = ACTIONS(165),
    [anon_sym_LBRACE] = ACTIONS(434),
    [anon_sym_COMMA] = ACTIONS(169),
    [anon_sym_RBRACE] = ACTIONS(169),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(432),
    [anon_sym_await] = ACTIONS(436),
    [anon_sym_LPAREN] = ACTIONS(379),
    [anon_sym_SEMI] = ACTIONS(169),
    [anon_sym_RPAREN] = ACTIONS(169),
    [anon_sym_in] = ACTIONS(165),
    [anon_sym_COLON] = ACTIONS(169),
    [anon_sym_yield] = ACTIONS(438),
    [anon_sym_EQ] = ACTIONS(440),
    [anon_sym_LBRACK] = ACTIONS(442),
    [anon_sym_RBRACK] = ACTIONS(169),
    [anon_sym_LT] = ACTIONS(389),
    [anon_sym_GT] = ACTIONS(165),
    [anon_sym_DOT] = ACTIONS(165),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(445),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_EQ_GT] = ACTIONS(402),
    [sym_optional_chain] = ACTIONS(169),
    [anon_sym_new] = ACTIONS(447),
    [anon_sym_PLUS_EQ] = ACTIONS(202),
    [anon_sym_DASH_EQ] = ACTIONS(202),
    [anon_sym_STAR_EQ] = ACTIONS(202),
    [anon_sym_SLASH_EQ] = ACTIONS(202),
    [anon_sym_PERCENT_EQ] = ACTIONS(202),
    [anon_sym_CARET_EQ] = ACTIONS(202),
    [anon_sym_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_LT_LT_EQ] = ACTIONS(202),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(202),
    [anon_sym_QMARK_QMARK_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP] = ACTIONS(165),
    [anon_sym_PIPE_PIPE] = ACTIONS(165),
    [anon_sym_GT_GT] = ACTIONS(165),
    [anon_sym_GT_GT_GT] = ACTIONS(165),
    [anon_sym_LT_LT] = ACTIONS(165),
    [anon_sym_AMP] = ACTIONS(165),
    [anon_sym_CARET] = ACTIONS(165),
    [anon_sym_PIPE] = ACTIONS(165),
    [anon_sym_PLUS] = ACTIONS(449),
    [anon_sym_DASH] = ACTIONS(449),
    [anon_sym_SLASH] = ACTIONS(409),
    [anon_sym_PERCENT] = ACTIONS(165),
    [anon_sym_STAR_STAR] = ACTIONS(165),
    [anon_sym_LT_EQ] = ACTIONS(169),
    [anon_sym_EQ_EQ] = ACTIONS(165),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(169),
    [anon_sym_BANG_EQ] = ACTIONS(165),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(169),
    [anon_sym_GT_EQ] = ACTIONS(169),
    [anon_sym_QMARK_QMARK] = ACTIONS(165),
    [anon_sym_instanceof] = ACTIONS(165),
    [anon_sym_BANG] = ACTIONS(452),
    [anon_sym_TILDE] = ACTIONS(454),
    [anon_sym_typeof] = ACTIONS(452),
    [anon_sym_void] = ACTIONS(452),
    [anon_sym_delete] = ACTIONS(452),
    [anon_sym_PLUS_PLUS] = ACTIONS(456),
    [anon_sym_DASH_DASH] = ACTIONS(456),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(419),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(459),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(461),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(432),
    [anon_sym_get] = ACTIONS(432),
    [anon_sym_set] = ACTIONS(432),
    [sym__ternary_qmark] = ACTIONS(169),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(17)] = {
    [sym_export_statement] = STATE(471),
    [sym_declaration] = STATE(471),
    [sym_import] = STATE(1326),
    [sym_import_statement] = STATE(471),
    [sym_statement] = STATE(21),
    [sym_expression_statement] = STATE(471),
    [sym_variable_declaration] = STATE(419),
    [sym_lexical_declaration] = STATE(419),
    [sym_using_declaration] = STATE(419),
    [sym_statement_block] = STATE(471),
    [sym_if_statement] = STATE(471),
    [sym_switch_statement] = STATE(471),
    [sym_for_statement] = STATE(471),
    [sym_for_in_statement] = STATE(471),
    [sym_while_statement] = STATE(471),
    [sym_do_statement] = STATE(471),
    [sym_try_statement] = STATE(471),
    [sym_with_statement] = STATE(471),
    [sym_break_statement] = STATE(471),
    [sym_continue_statement] = STATE(471),
    [sym_debugger_statement] = STATE(471),
    [sym_return_statement] = STATE(471),
    [sym_throw_statement] = STATE(471),
    [sym_empty_statement] = STATE(471),
    [sym_labeled_statement] = STATE(471),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(823),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1790),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1790),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_class_declaration] = STATE(419),
    [sym_function_expression] = STATE(740),
    [sym_function_declaration] = STATE(419),
    [sym_generator_function] = STATE(740),
    [sym_generator_function_declaration] = STATE(419),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1790),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_sequence_expression] = STATE(1662),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [aux_sym_program_repeat1] = STATE(21),
    [aux_sym_export_statement_repeat1] = STATE(1267),
    [ts_builtin_sym_end] = ACTIONS(463),
    [sym_identifier] = ACTIONS(9),
    [anon_sym_export] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(15),
    [anon_sym_import] = ACTIONS(17),
    [anon_sym_with] = ACTIONS(19),
    [anon_sym_var] = ACTIONS(21),
    [anon_sym_let] = ACTIONS(23),
    [anon_sym_const] = ACTIONS(25),
    [anon_sym_using] = ACTIONS(27),
    [anon_sym_await] = ACTIONS(29),
    [anon_sym_if] = ACTIONS(31),
    [anon_sym_switch] = ACTIONS(33),
    [anon_sym_for] = ACTIONS(35),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_SEMI] = ACTIONS(39),
    [anon_sym_while] = ACTIONS(41),
    [anon_sym_do] = ACTIONS(43),
    [anon_sym_try] = ACTIONS(45),
    [anon_sym_break] = ACTIONS(47),
    [anon_sym_continue] = ACTIONS(49),
    [anon_sym_debugger] = ACTIONS(51),
    [anon_sym_return] = ACTIONS(53),
    [anon_sym_throw] = ACTIONS(55),
    [anon_sym_yield] = ACTIONS(57),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(67),
    [anon_sym_async] = ACTIONS(69),
    [anon_sym_function] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(75),
    [anon_sym_DASH] = ACTIONS(75),
    [anon_sym_SLASH] = ACTIONS(77),
    [anon_sym_BANG] = ACTIONS(79),
    [anon_sym_TILDE] = ACTIONS(79),
    [anon_sym_typeof] = ACTIONS(75),
    [anon_sym_void] = ACTIONS(75),
    [anon_sym_delete] = ACTIONS(75),
    [anon_sym_PLUS_PLUS] = ACTIONS(81),
    [anon_sym_DASH_DASH] = ACTIONS(81),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(87),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(95),
    [anon_sym_get] = ACTIONS(95),
    [anon_sym_set] = ACTIONS(95),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(18)] = {
    [sym_export_statement] = STATE(471),
    [sym_declaration] = STATE(471),
    [sym_import] = STATE(1326),
    [sym_import_statement] = STATE(471),
    [sym_statement] = STATE(9),
    [sym_expression_statement] = STATE(471),
    [sym_variable_declaration] = STATE(419),
    [sym_lexical_declaration] = STATE(419),
    [sym_using_declaration] = STATE(419),
    [sym_statement_block] = STATE(471),
    [sym_if_statement] = STATE(471),
    [sym_switch_statement] = STATE(471),
    [sym_for_statement] = STATE(471),
    [sym_for_in_statement] = STATE(471),
    [sym_while_statement] = STATE(471),
    [sym_do_statement] = STATE(471),
    [sym_try_statement] = STATE(471),
    [sym_with_statement] = STATE(471),
    [sym_break_statement] = STATE(471),
    [sym_continue_statement] = STATE(471),
    [sym_debugger_statement] = STATE(471),
    [sym_return_statement] = STATE(471),
    [sym_throw_statement] = STATE(471),
    [sym_empty_statement] = STATE(471),
    [sym_labeled_statement] = STATE(471),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(823),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1790),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1790),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_class_declaration] = STATE(419),
    [sym_function_expression] = STATE(740),
    [sym_function_declaration] = STATE(419),
    [sym_generator_function] = STATE(740),
    [sym_generator_function_declaration] = STATE(419),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1790),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_sequence_expression] = STATE(1662),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [aux_sym_program_repeat1] = STATE(9),
    [aux_sym_export_statement_repeat1] = STATE(1267),
    [sym_identifier] = ACTIONS(9),
    [anon_sym_export] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(15),
    [anon_sym_RBRACE] = ACTIONS(465),
    [anon_sym_import] = ACTIONS(17),
    [anon_sym_with] = ACTIONS(19),
    [anon_sym_var] = ACTIONS(21),
    [anon_sym_let] = ACTIONS(23),
    [anon_sym_const] = ACTIONS(25),
    [anon_sym_using] = ACTIONS(27),
    [anon_sym_await] = ACTIONS(29),
    [anon_sym_if] = ACTIONS(31),
    [anon_sym_switch] = ACTIONS(33),
    [anon_sym_for] = ACTIONS(35),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_SEMI] = ACTIONS(39),
    [anon_sym_while] = ACTIONS(41),
    [anon_sym_do] = ACTIONS(43),
    [anon_sym_try] = ACTIONS(45),
    [anon_sym_break] = ACTIONS(47),
    [anon_sym_continue] = ACTIONS(49),
    [anon_sym_debugger] = ACTIONS(51),
    [anon_sym_return] = ACTIONS(53),
    [anon_sym_throw] = ACTIONS(55),
    [anon_sym_yield] = ACTIONS(57),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(67),
    [anon_sym_async] = ACTIONS(69),
    [anon_sym_function] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(75),
    [anon_sym_DASH] = ACTIONS(75),
    [anon_sym_SLASH] = ACTIONS(77),
    [anon_sym_BANG] = ACTIONS(79),
    [anon_sym_TILDE] = ACTIONS(79),
    [anon_sym_typeof] = ACTIONS(75),
    [anon_sym_void] = ACTIONS(75),
    [anon_sym_delete] = ACTIONS(75),
    [anon_sym_PLUS_PLUS] = ACTIONS(81),
    [anon_sym_DASH_DASH] = ACTIONS(81),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(87),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(95),
    [anon_sym_get] = ACTIONS(95),
    [anon_sym_set] = ACTIONS(95),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(19)] = {
    [sym_export_statement] = STATE(471),
    [sym_declaration] = STATE(471),
    [sym_import] = STATE(1326),
    [sym_import_statement] = STATE(471),
    [sym_statement] = STATE(18),
    [sym_expression_statement] = STATE(471),
    [sym_variable_declaration] = STATE(419),
    [sym_lexical_declaration] = STATE(419),
    [sym_using_declaration] = STATE(419),
    [sym_statement_block] = STATE(471),
    [sym_if_statement] = STATE(471),
    [sym_switch_statement] = STATE(471),
    [sym_for_statement] = STATE(471),
    [sym_for_in_statement] = STATE(471),
    [sym_while_statement] = STATE(471),
    [sym_do_statement] = STATE(471),
    [sym_try_statement] = STATE(471),
    [sym_with_statement] = STATE(471),
    [sym_break_statement] = STATE(471),
    [sym_continue_statement] = STATE(471),
    [sym_debugger_statement] = STATE(471),
    [sym_return_statement] = STATE(471),
    [sym_throw_statement] = STATE(471),
    [sym_empty_statement] = STATE(471),
    [sym_labeled_statement] = STATE(471),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(823),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1790),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1790),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_class_declaration] = STATE(419),
    [sym_function_expression] = STATE(740),
    [sym_function_declaration] = STATE(419),
    [sym_generator_function] = STATE(740),
    [sym_generator_function_declaration] = STATE(419),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1790),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_sequence_expression] = STATE(1662),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [aux_sym_program_repeat1] = STATE(18),
    [aux_sym_export_statement_repeat1] = STATE(1267),
    [sym_identifier] = ACTIONS(9),
    [anon_sym_export] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(15),
    [anon_sym_RBRACE] = ACTIONS(467),
    [anon_sym_import] = ACTIONS(17),
    [anon_sym_with] = ACTIONS(19),
    [anon_sym_var] = ACTIONS(21),
    [anon_sym_let] = ACTIONS(23),
    [anon_sym_const] = ACTIONS(25),
    [anon_sym_using] = ACTIONS(27),
    [anon_sym_await] = ACTIONS(29),
    [anon_sym_if] = ACTIONS(31),
    [anon_sym_switch] = ACTIONS(33),
    [anon_sym_for] = ACTIONS(35),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_SEMI] = ACTIONS(39),
    [anon_sym_while] = ACTIONS(41),
    [anon_sym_do] = ACTIONS(43),
    [anon_sym_try] = ACTIONS(45),
    [anon_sym_break] = ACTIONS(47),
    [anon_sym_continue] = ACTIONS(49),
    [anon_sym_debugger] = ACTIONS(51),
    [anon_sym_return] = ACTIONS(53),
    [anon_sym_throw] = ACTIONS(55),
    [anon_sym_yield] = ACTIONS(57),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(67),
    [anon_sym_async] = ACTIONS(69),
    [anon_sym_function] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(75),
    [anon_sym_DASH] = ACTIONS(75),
    [anon_sym_SLASH] = ACTIONS(77),
    [anon_sym_BANG] = ACTIONS(79),
    [anon_sym_TILDE] = ACTIONS(79),
    [anon_sym_typeof] = ACTIONS(75),
    [anon_sym_void] = ACTIONS(75),
    [anon_sym_delete] = ACTIONS(75),
    [anon_sym_PLUS_PLUS] = ACTIONS(81),
    [anon_sym_DASH_DASH] = ACTIONS(81),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(87),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(95),
    [anon_sym_get] = ACTIONS(95),
    [anon_sym_set] = ACTIONS(95),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(20)] = {
    [sym_export_statement] = STATE(471),
    [sym_declaration] = STATE(471),
    [sym_import] = STATE(1326),
    [sym_import_statement] = STATE(471),
    [sym_statement] = STATE(9),
    [sym_expression_statement] = STATE(471),
    [sym_variable_declaration] = STATE(419),
    [sym_lexical_declaration] = STATE(419),
    [sym_using_declaration] = STATE(419),
    [sym_statement_block] = STATE(471),
    [sym_if_statement] = STATE(471),
    [sym_switch_statement] = STATE(471),
    [sym_for_statement] = STATE(471),
    [sym_for_in_statement] = STATE(471),
    [sym_while_statement] = STATE(471),
    [sym_do_statement] = STATE(471),
    [sym_try_statement] = STATE(471),
    [sym_with_statement] = STATE(471),
    [sym_break_statement] = STATE(471),
    [sym_continue_statement] = STATE(471),
    [sym_debugger_statement] = STATE(471),
    [sym_return_statement] = STATE(471),
    [sym_throw_statement] = STATE(471),
    [sym_empty_statement] = STATE(471),
    [sym_labeled_statement] = STATE(471),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(823),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1790),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1790),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_class_declaration] = STATE(419),
    [sym_function_expression] = STATE(740),
    [sym_function_declaration] = STATE(419),
    [sym_generator_function] = STATE(740),
    [sym_generator_function_declaration] = STATE(419),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1790),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_sequence_expression] = STATE(1662),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [aux_sym_program_repeat1] = STATE(9),
    [aux_sym_export_statement_repeat1] = STATE(1267),
    [ts_builtin_sym_end] = ACTIONS(463),
    [sym_identifier] = ACTIONS(9),
    [anon_sym_export] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(15),
    [anon_sym_import] = ACTIONS(17),
    [anon_sym_with] = ACTIONS(19),
    [anon_sym_var] = ACTIONS(21),
    [anon_sym_let] = ACTIONS(23),
    [anon_sym_const] = ACTIONS(25),
    [anon_sym_using] = ACTIONS(27),
    [anon_sym_await] = ACTIONS(29),
    [anon_sym_if] = ACTIONS(31),
    [anon_sym_switch] = ACTIONS(33),
    [anon_sym_for] = ACTIONS(35),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_SEMI] = ACTIONS(39),
    [anon_sym_while] = ACTIONS(41),
    [anon_sym_do] = ACTIONS(43),
    [anon_sym_try] = ACTIONS(45),
    [anon_sym_break] = ACTIONS(47),
    [anon_sym_continue] = ACTIONS(49),
    [anon_sym_debugger] = ACTIONS(51),
    [anon_sym_return] = ACTIONS(53),
    [anon_sym_throw] = ACTIONS(55),
    [anon_sym_yield] = ACTIONS(57),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(67),
    [anon_sym_async] = ACTIONS(69),
    [anon_sym_function] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(75),
    [anon_sym_DASH] = ACTIONS(75),
    [anon_sym_SLASH] = ACTIONS(77),
    [anon_sym_BANG] = ACTIONS(79),
    [anon_sym_TILDE] = ACTIONS(79),
    [anon_sym_typeof] = ACTIONS(75),
    [anon_sym_void] = ACTIONS(75),
    [anon_sym_delete] = ACTIONS(75),
    [anon_sym_PLUS_PLUS] = ACTIONS(81),
    [anon_sym_DASH_DASH] = ACTIONS(81),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(87),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(95),
    [anon_sym_get] = ACTIONS(95),
    [anon_sym_set] = ACTIONS(95),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(21)] = {
    [sym_export_statement] = STATE(471),
    [sym_declaration] = STATE(471),
    [sym_import] = STATE(1326),
    [sym_import_statement] = STATE(471),
    [sym_statement] = STATE(9),
    [sym_expression_statement] = STATE(471),
    [sym_variable_declaration] = STATE(419),
    [sym_lexical_declaration] = STATE(419),
    [sym_using_declaration] = STATE(419),
    [sym_statement_block] = STATE(471),
    [sym_if_statement] = STATE(471),
    [sym_switch_statement] = STATE(471),
    [sym_for_statement] = STATE(471),
    [sym_for_in_statement] = STATE(471),
    [sym_while_statement] = STATE(471),
    [sym_do_statement] = STATE(471),
    [sym_try_statement] = STATE(471),
    [sym_with_statement] = STATE(471),
    [sym_break_statement] = STATE(471),
    [sym_continue_statement] = STATE(471),
    [sym_debugger_statement] = STATE(471),
    [sym_return_statement] = STATE(471),
    [sym_throw_statement] = STATE(471),
    [sym_empty_statement] = STATE(471),
    [sym_labeled_statement] = STATE(471),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(823),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1790),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1790),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_class_declaration] = STATE(419),
    [sym_function_expression] = STATE(740),
    [sym_function_declaration] = STATE(419),
    [sym_generator_function] = STATE(740),
    [sym_generator_function_declaration] = STATE(419),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1790),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_sequence_expression] = STATE(1662),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [aux_sym_program_repeat1] = STATE(9),
    [aux_sym_export_statement_repeat1] = STATE(1267),
    [ts_builtin_sym_end] = ACTIONS(469),
    [sym_identifier] = ACTIONS(9),
    [anon_sym_export] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(15),
    [anon_sym_import] = ACTIONS(17),
    [anon_sym_with] = ACTIONS(19),
    [anon_sym_var] = ACTIONS(21),
    [anon_sym_let] = ACTIONS(23),
    [anon_sym_const] = ACTIONS(25),
    [anon_sym_using] = ACTIONS(27),
    [anon_sym_await] = ACTIONS(29),
    [anon_sym_if] = ACTIONS(31),
    [anon_sym_switch] = ACTIONS(33),
    [anon_sym_for] = ACTIONS(35),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_SEMI] = ACTIONS(39),
    [anon_sym_while] = ACTIONS(41),
    [anon_sym_do] = ACTIONS(43),
    [anon_sym_try] = ACTIONS(45),
    [anon_sym_break] = ACTIONS(47),
    [anon_sym_continue] = ACTIONS(49),
    [anon_sym_debugger] = ACTIONS(51),
    [anon_sym_return] = ACTIONS(53),
    [anon_sym_throw] = ACTIONS(55),
    [anon_sym_yield] = ACTIONS(57),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(67),
    [anon_sym_async] = ACTIONS(69),
    [anon_sym_function] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(75),
    [anon_sym_DASH] = ACTIONS(75),
    [anon_sym_SLASH] = ACTIONS(77),
    [anon_sym_BANG] = ACTIONS(79),
    [anon_sym_TILDE] = ACTIONS(79),
    [anon_sym_typeof] = ACTIONS(75),
    [anon_sym_void] = ACTIONS(75),
    [anon_sym_delete] = ACTIONS(75),
    [anon_sym_PLUS_PLUS] = ACTIONS(81),
    [anon_sym_DASH_DASH] = ACTIONS(81),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(87),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(95),
    [anon_sym_get] = ACTIONS(95),
    [anon_sym_set] = ACTIONS(95),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(22)] = {
    [sym_export_statement] = STATE(471),
    [sym_declaration] = STATE(471),
    [sym_import] = STATE(1326),
    [sym_import_statement] = STATE(471),
    [sym_statement] = STATE(9),
    [sym_expression_statement] = STATE(471),
    [sym_variable_declaration] = STATE(419),
    [sym_lexical_declaration] = STATE(419),
    [sym_using_declaration] = STATE(419),
    [sym_statement_block] = STATE(471),
    [sym_if_statement] = STATE(471),
    [sym_switch_statement] = STATE(471),
    [sym_for_statement] = STATE(471),
    [sym_for_in_statement] = STATE(471),
    [sym_while_statement] = STATE(471),
    [sym_do_statement] = STATE(471),
    [sym_try_statement] = STATE(471),
    [sym_with_statement] = STATE(471),
    [sym_break_statement] = STATE(471),
    [sym_continue_statement] = STATE(471),
    [sym_debugger_statement] = STATE(471),
    [sym_return_statement] = STATE(471),
    [sym_throw_statement] = STATE(471),
    [sym_empty_statement] = STATE(471),
    [sym_labeled_statement] = STATE(471),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(823),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1790),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1790),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_class_declaration] = STATE(419),
    [sym_function_expression] = STATE(740),
    [sym_function_declaration] = STATE(419),
    [sym_generator_function] = STATE(740),
    [sym_generator_function_declaration] = STATE(419),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1790),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_sequence_expression] = STATE(1662),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [aux_sym_program_repeat1] = STATE(9),
    [aux_sym_export_statement_repeat1] = STATE(1267),
    [sym_identifier] = ACTIONS(9),
    [anon_sym_export] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(15),
    [anon_sym_RBRACE] = ACTIONS(471),
    [anon_sym_import] = ACTIONS(17),
    [anon_sym_with] = ACTIONS(19),
    [anon_sym_var] = ACTIONS(21),
    [anon_sym_let] = ACTIONS(23),
    [anon_sym_const] = ACTIONS(25),
    [anon_sym_using] = ACTIONS(27),
    [anon_sym_await] = ACTIONS(29),
    [anon_sym_if] = ACTIONS(31),
    [anon_sym_switch] = ACTIONS(33),
    [anon_sym_for] = ACTIONS(35),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_SEMI] = ACTIONS(39),
    [anon_sym_while] = ACTIONS(41),
    [anon_sym_do] = ACTIONS(43),
    [anon_sym_try] = ACTIONS(45),
    [anon_sym_break] = ACTIONS(47),
    [anon_sym_continue] = ACTIONS(49),
    [anon_sym_debugger] = ACTIONS(51),
    [anon_sym_return] = ACTIONS(53),
    [anon_sym_throw] = ACTIONS(55),
    [anon_sym_yield] = ACTIONS(57),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(67),
    [anon_sym_async] = ACTIONS(69),
    [anon_sym_function] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(75),
    [anon_sym_DASH] = ACTIONS(75),
    [anon_sym_SLASH] = ACTIONS(77),
    [anon_sym_BANG] = ACTIONS(79),
    [anon_sym_TILDE] = ACTIONS(79),
    [anon_sym_typeof] = ACTIONS(75),
    [anon_sym_void] = ACTIONS(75),
    [anon_sym_delete] = ACTIONS(75),
    [anon_sym_PLUS_PLUS] = ACTIONS(81),
    [anon_sym_DASH_DASH] = ACTIONS(81),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(87),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(95),
    [anon_sym_get] = ACTIONS(95),
    [anon_sym_set] = ACTIONS(95),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(23)] = {
    [sym_export_statement] = STATE(471),
    [sym_declaration] = STATE(471),
    [sym_import] = STATE(1326),
    [sym_import_statement] = STATE(471),
    [sym_statement] = STATE(22),
    [sym_expression_statement] = STATE(471),
    [sym_variable_declaration] = STATE(419),
    [sym_lexical_declaration] = STATE(419),
    [sym_using_declaration] = STATE(419),
    [sym_statement_block] = STATE(471),
    [sym_if_statement] = STATE(471),
    [sym_switch_statement] = STATE(471),
    [sym_for_statement] = STATE(471),
    [sym_for_in_statement] = STATE(471),
    [sym_while_statement] = STATE(471),
    [sym_do_statement] = STATE(471),
    [sym_try_statement] = STATE(471),
    [sym_with_statement] = STATE(471),
    [sym_break_statement] = STATE(471),
    [sym_continue_statement] = STATE(471),
    [sym_debugger_statement] = STATE(471),
    [sym_return_statement] = STATE(471),
    [sym_throw_statement] = STATE(471),
    [sym_empty_statement] = STATE(471),
    [sym_labeled_statement] = STATE(471),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(823),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1790),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1790),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_class_declaration] = STATE(419),
    [sym_function_expression] = STATE(740),
    [sym_function_declaration] = STATE(419),
    [sym_generator_function] = STATE(740),
    [sym_generator_function_declaration] = STATE(419),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1790),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_sequence_expression] = STATE(1662),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [aux_sym_program_repeat1] = STATE(22),
    [aux_sym_export_statement_repeat1] = STATE(1267),
    [sym_identifier] = ACTIONS(9),
    [anon_sym_export] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(15),
    [anon_sym_RBRACE] = ACTIONS(473),
    [anon_sym_import] = ACTIONS(17),
    [anon_sym_with] = ACTIONS(19),
    [anon_sym_var] = ACTIONS(21),
    [anon_sym_let] = ACTIONS(23),
    [anon_sym_const] = ACTIONS(25),
    [anon_sym_using] = ACTIONS(27),
    [anon_sym_await] = ACTIONS(29),
    [anon_sym_if] = ACTIONS(31),
    [anon_sym_switch] = ACTIONS(33),
    [anon_sym_for] = ACTIONS(35),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_SEMI] = ACTIONS(39),
    [anon_sym_while] = ACTIONS(41),
    [anon_sym_do] = ACTIONS(43),
    [anon_sym_try] = ACTIONS(45),
    [anon_sym_break] = ACTIONS(47),
    [anon_sym_continue] = ACTIONS(49),
    [anon_sym_debugger] = ACTIONS(51),
    [anon_sym_return] = ACTIONS(53),
    [anon_sym_throw] = ACTIONS(55),
    [anon_sym_yield] = ACTIONS(57),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(67),
    [anon_sym_async] = ACTIONS(69),
    [anon_sym_function] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(75),
    [anon_sym_DASH] = ACTIONS(75),
    [anon_sym_SLASH] = ACTIONS(77),
    [anon_sym_BANG] = ACTIONS(79),
    [anon_sym_TILDE] = ACTIONS(79),
    [anon_sym_typeof] = ACTIONS(75),
    [anon_sym_void] = ACTIONS(75),
    [anon_sym_delete] = ACTIONS(75),
    [anon_sym_PLUS_PLUS] = ACTIONS(81),
    [anon_sym_DASH_DASH] = ACTIONS(81),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(87),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(95),
    [anon_sym_get] = ACTIONS(95),
    [anon_sym_set] = ACTIONS(95),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(24)] = {
    [sym_export_statement] = STATE(471),
    [sym_declaration] = STATE(471),
    [sym_import] = STATE(1326),
    [sym_import_statement] = STATE(471),
    [sym_statement] = STATE(9),
    [sym_expression_statement] = STATE(471),
    [sym_variable_declaration] = STATE(419),
    [sym_lexical_declaration] = STATE(419),
    [sym_using_declaration] = STATE(419),
    [sym_statement_block] = STATE(471),
    [sym_if_statement] = STATE(471),
    [sym_switch_statement] = STATE(471),
    [sym_for_statement] = STATE(471),
    [sym_for_in_statement] = STATE(471),
    [sym_while_statement] = STATE(471),
    [sym_do_statement] = STATE(471),
    [sym_try_statement] = STATE(471),
    [sym_with_statement] = STATE(471),
    [sym_break_statement] = STATE(471),
    [sym_continue_statement] = STATE(471),
    [sym_debugger_statement] = STATE(471),
    [sym_return_statement] = STATE(471),
    [sym_throw_statement] = STATE(471),
    [sym_empty_statement] = STATE(471),
    [sym_labeled_statement] = STATE(471),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(823),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1790),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1790),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_class_declaration] = STATE(419),
    [sym_function_expression] = STATE(740),
    [sym_function_declaration] = STATE(419),
    [sym_generator_function] = STATE(740),
    [sym_generator_function_declaration] = STATE(419),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1790),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_sequence_expression] = STATE(1662),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [aux_sym_program_repeat1] = STATE(9),
    [aux_sym_export_statement_repeat1] = STATE(1267),
    [sym_identifier] = ACTIONS(9),
    [anon_sym_export] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(15),
    [anon_sym_RBRACE] = ACTIONS(475),
    [anon_sym_import] = ACTIONS(17),
    [anon_sym_with] = ACTIONS(19),
    [anon_sym_var] = ACTIONS(21),
    [anon_sym_let] = ACTIONS(23),
    [anon_sym_const] = ACTIONS(25),
    [anon_sym_using] = ACTIONS(27),
    [anon_sym_await] = ACTIONS(29),
    [anon_sym_if] = ACTIONS(31),
    [anon_sym_switch] = ACTIONS(33),
    [anon_sym_for] = ACTIONS(35),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_SEMI] = ACTIONS(39),
    [anon_sym_while] = ACTIONS(41),
    [anon_sym_do] = ACTIONS(43),
    [anon_sym_try] = ACTIONS(45),
    [anon_sym_break] = ACTIONS(47),
    [anon_sym_continue] = ACTIONS(49),
    [anon_sym_debugger] = ACTIONS(51),
    [anon_sym_return] = ACTIONS(53),
    [anon_sym_throw] = ACTIONS(55),
    [anon_sym_yield] = ACTIONS(57),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(67),
    [anon_sym_async] = ACTIONS(69),
    [anon_sym_function] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(75),
    [anon_sym_DASH] = ACTIONS(75),
    [anon_sym_SLASH] = ACTIONS(77),
    [anon_sym_BANG] = ACTIONS(79),
    [anon_sym_TILDE] = ACTIONS(79),
    [anon_sym_typeof] = ACTIONS(75),
    [anon_sym_void] = ACTIONS(75),
    [anon_sym_delete] = ACTIONS(75),
    [anon_sym_PLUS_PLUS] = ACTIONS(81),
    [anon_sym_DASH_DASH] = ACTIONS(81),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(87),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(95),
    [anon_sym_get] = ACTIONS(95),
    [anon_sym_set] = ACTIONS(95),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(25)] = {
    [sym_export_statement] = STATE(471),
    [sym_declaration] = STATE(471),
    [sym_import] = STATE(1326),
    [sym_import_statement] = STATE(471),
    [sym_statement] = STATE(24),
    [sym_expression_statement] = STATE(471),
    [sym_variable_declaration] = STATE(419),
    [sym_lexical_declaration] = STATE(419),
    [sym_using_declaration] = STATE(419),
    [sym_statement_block] = STATE(471),
    [sym_if_statement] = STATE(471),
    [sym_switch_statement] = STATE(471),
    [sym_for_statement] = STATE(471),
    [sym_for_in_statement] = STATE(471),
    [sym_while_statement] = STATE(471),
    [sym_do_statement] = STATE(471),
    [sym_try_statement] = STATE(471),
    [sym_with_statement] = STATE(471),
    [sym_break_statement] = STATE(471),
    [sym_continue_statement] = STATE(471),
    [sym_debugger_statement] = STATE(471),
    [sym_return_statement] = STATE(471),
    [sym_throw_statement] = STATE(471),
    [sym_empty_statement] = STATE(471),
    [sym_labeled_statement] = STATE(471),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(823),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1790),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1790),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_class_declaration] = STATE(419),
    [sym_function_expression] = STATE(740),
    [sym_function_declaration] = STATE(419),
    [sym_generator_function] = STATE(740),
    [sym_generator_function_declaration] = STATE(419),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1790),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_sequence_expression] = STATE(1662),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [aux_sym_program_repeat1] = STATE(24),
    [aux_sym_export_statement_repeat1] = STATE(1267),
    [sym_identifier] = ACTIONS(9),
    [anon_sym_export] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(15),
    [anon_sym_RBRACE] = ACTIONS(477),
    [anon_sym_import] = ACTIONS(17),
    [anon_sym_with] = ACTIONS(19),
    [anon_sym_var] = ACTIONS(21),
    [anon_sym_let] = ACTIONS(23),
    [anon_sym_const] = ACTIONS(25),
    [anon_sym_using] = ACTIONS(27),
    [anon_sym_await] = ACTIONS(29),
    [anon_sym_if] = ACTIONS(31),
    [anon_sym_switch] = ACTIONS(33),
    [anon_sym_for] = ACTIONS(35),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_SEMI] = ACTIONS(39),
    [anon_sym_while] = ACTIONS(41),
    [anon_sym_do] = ACTIONS(43),
    [anon_sym_try] = ACTIONS(45),
    [anon_sym_break] = ACTIONS(47),
    [anon_sym_continue] = ACTIONS(49),
    [anon_sym_debugger] = ACTIONS(51),
    [anon_sym_return] = ACTIONS(53),
    [anon_sym_throw] = ACTIONS(55),
    [anon_sym_yield] = ACTIONS(57),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(67),
    [anon_sym_async] = ACTIONS(69),
    [anon_sym_function] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(75),
    [anon_sym_DASH] = ACTIONS(75),
    [anon_sym_SLASH] = ACTIONS(77),
    [anon_sym_BANG] = ACTIONS(79),
    [anon_sym_TILDE] = ACTIONS(79),
    [anon_sym_typeof] = ACTIONS(75),
    [anon_sym_void] = ACTIONS(75),
    [anon_sym_delete] = ACTIONS(75),
    [anon_sym_PLUS_PLUS] = ACTIONS(81),
    [anon_sym_DASH_DASH] = ACTIONS(81),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(87),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(95),
    [anon_sym_get] = ACTIONS(95),
    [anon_sym_set] = ACTIONS(95),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(26)] = {
    [sym_export_statement] = STATE(471),
    [sym_declaration] = STATE(471),
    [sym_import] = STATE(1326),
    [sym_import_statement] = STATE(471),
    [sym_statement] = STATE(9),
    [sym_expression_statement] = STATE(471),
    [sym_variable_declaration] = STATE(419),
    [sym_lexical_declaration] = STATE(419),
    [sym_using_declaration] = STATE(419),
    [sym_statement_block] = STATE(471),
    [sym_if_statement] = STATE(471),
    [sym_switch_statement] = STATE(471),
    [sym_for_statement] = STATE(471),
    [sym_for_in_statement] = STATE(471),
    [sym_while_statement] = STATE(471),
    [sym_do_statement] = STATE(471),
    [sym_try_statement] = STATE(471),
    [sym_with_statement] = STATE(471),
    [sym_break_statement] = STATE(471),
    [sym_continue_statement] = STATE(471),
    [sym_debugger_statement] = STATE(471),
    [sym_return_statement] = STATE(471),
    [sym_throw_statement] = STATE(471),
    [sym_empty_statement] = STATE(471),
    [sym_labeled_statement] = STATE(471),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(823),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1790),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1790),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_class_declaration] = STATE(419),
    [sym_function_expression] = STATE(740),
    [sym_function_declaration] = STATE(419),
    [sym_generator_function] = STATE(740),
    [sym_generator_function_declaration] = STATE(419),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1790),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_sequence_expression] = STATE(1662),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [aux_sym_program_repeat1] = STATE(9),
    [aux_sym_export_statement_repeat1] = STATE(1267),
    [sym_identifier] = ACTIONS(9),
    [anon_sym_export] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(15),
    [anon_sym_RBRACE] = ACTIONS(479),
    [anon_sym_import] = ACTIONS(17),
    [anon_sym_with] = ACTIONS(19),
    [anon_sym_var] = ACTIONS(21),
    [anon_sym_let] = ACTIONS(23),
    [anon_sym_const] = ACTIONS(25),
    [anon_sym_using] = ACTIONS(27),
    [anon_sym_await] = ACTIONS(29),
    [anon_sym_if] = ACTIONS(31),
    [anon_sym_switch] = ACTIONS(33),
    [anon_sym_for] = ACTIONS(35),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_SEMI] = ACTIONS(39),
    [anon_sym_while] = ACTIONS(41),
    [anon_sym_do] = ACTIONS(43),
    [anon_sym_try] = ACTIONS(45),
    [anon_sym_break] = ACTIONS(47),
    [anon_sym_continue] = ACTIONS(49),
    [anon_sym_debugger] = ACTIONS(51),
    [anon_sym_return] = ACTIONS(53),
    [anon_sym_throw] = ACTIONS(55),
    [anon_sym_yield] = ACTIONS(57),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(67),
    [anon_sym_async] = ACTIONS(69),
    [anon_sym_function] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(75),
    [anon_sym_DASH] = ACTIONS(75),
    [anon_sym_SLASH] = ACTIONS(77),
    [anon_sym_BANG] = ACTIONS(79),
    [anon_sym_TILDE] = ACTIONS(79),
    [anon_sym_typeof] = ACTIONS(75),
    [anon_sym_void] = ACTIONS(75),
    [anon_sym_delete] = ACTIONS(75),
    [anon_sym_PLUS_PLUS] = ACTIONS(81),
    [anon_sym_DASH_DASH] = ACTIONS(81),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(87),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(95),
    [anon_sym_get] = ACTIONS(95),
    [anon_sym_set] = ACTIONS(95),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(27)] = {
    [sym_export_statement] = STATE(471),
    [sym_declaration] = STATE(471),
    [sym_import] = STATE(1326),
    [sym_import_statement] = STATE(471),
    [sym_statement] = STATE(26),
    [sym_expression_statement] = STATE(471),
    [sym_variable_declaration] = STATE(419),
    [sym_lexical_declaration] = STATE(419),
    [sym_using_declaration] = STATE(419),
    [sym_statement_block] = STATE(471),
    [sym_if_statement] = STATE(471),
    [sym_switch_statement] = STATE(471),
    [sym_for_statement] = STATE(471),
    [sym_for_in_statement] = STATE(471),
    [sym_while_statement] = STATE(471),
    [sym_do_statement] = STATE(471),
    [sym_try_statement] = STATE(471),
    [sym_with_statement] = STATE(471),
    [sym_break_statement] = STATE(471),
    [sym_continue_statement] = STATE(471),
    [sym_debugger_statement] = STATE(471),
    [sym_return_statement] = STATE(471),
    [sym_throw_statement] = STATE(471),
    [sym_empty_statement] = STATE(471),
    [sym_labeled_statement] = STATE(471),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(823),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1790),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1790),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_class_declaration] = STATE(419),
    [sym_function_expression] = STATE(740),
    [sym_function_declaration] = STATE(419),
    [sym_generator_function] = STATE(740),
    [sym_generator_function_declaration] = STATE(419),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1790),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_sequence_expression] = STATE(1662),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [aux_sym_program_repeat1] = STATE(26),
    [aux_sym_export_statement_repeat1] = STATE(1267),
    [sym_identifier] = ACTIONS(9),
    [anon_sym_export] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(15),
    [anon_sym_RBRACE] = ACTIONS(481),
    [anon_sym_import] = ACTIONS(17),
    [anon_sym_with] = ACTIONS(19),
    [anon_sym_var] = ACTIONS(21),
    [anon_sym_let] = ACTIONS(23),
    [anon_sym_const] = ACTIONS(25),
    [anon_sym_using] = ACTIONS(27),
    [anon_sym_await] = ACTIONS(29),
    [anon_sym_if] = ACTIONS(31),
    [anon_sym_switch] = ACTIONS(33),
    [anon_sym_for] = ACTIONS(35),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_SEMI] = ACTIONS(39),
    [anon_sym_while] = ACTIONS(41),
    [anon_sym_do] = ACTIONS(43),
    [anon_sym_try] = ACTIONS(45),
    [anon_sym_break] = ACTIONS(47),
    [anon_sym_continue] = ACTIONS(49),
    [anon_sym_debugger] = ACTIONS(51),
    [anon_sym_return] = ACTIONS(53),
    [anon_sym_throw] = ACTIONS(55),
    [anon_sym_yield] = ACTIONS(57),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(67),
    [anon_sym_async] = ACTIONS(69),
    [anon_sym_function] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(75),
    [anon_sym_DASH] = ACTIONS(75),
    [anon_sym_SLASH] = ACTIONS(77),
    [anon_sym_BANG] = ACTIONS(79),
    [anon_sym_TILDE] = ACTIONS(79),
    [anon_sym_typeof] = ACTIONS(75),
    [anon_sym_void] = ACTIONS(75),
    [anon_sym_delete] = ACTIONS(75),
    [anon_sym_PLUS_PLUS] = ACTIONS(81),
    [anon_sym_DASH_DASH] = ACTIONS(81),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(87),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(95),
    [anon_sym_get] = ACTIONS(95),
    [anon_sym_set] = ACTIONS(95),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(28)] = {
    [sym_export_statement] = STATE(471),
    [sym_declaration] = STATE(471),
    [sym_import] = STATE(1326),
    [sym_import_statement] = STATE(471),
    [sym_statement] = STATE(30),
    [sym_expression_statement] = STATE(471),
    [sym_variable_declaration] = STATE(419),
    [sym_lexical_declaration] = STATE(419),
    [sym_using_declaration] = STATE(419),
    [sym_statement_block] = STATE(471),
    [sym_if_statement] = STATE(471),
    [sym_switch_statement] = STATE(471),
    [sym_for_statement] = STATE(471),
    [sym_for_in_statement] = STATE(471),
    [sym_while_statement] = STATE(471),
    [sym_do_statement] = STATE(471),
    [sym_try_statement] = STATE(471),
    [sym_with_statement] = STATE(471),
    [sym_break_statement] = STATE(471),
    [sym_continue_statement] = STATE(471),
    [sym_debugger_statement] = STATE(471),
    [sym_return_statement] = STATE(471),
    [sym_throw_statement] = STATE(471),
    [sym_empty_statement] = STATE(471),
    [sym_labeled_statement] = STATE(471),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(823),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1790),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1790),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_class_declaration] = STATE(419),
    [sym_function_expression] = STATE(740),
    [sym_function_declaration] = STATE(419),
    [sym_generator_function] = STATE(740),
    [sym_generator_function_declaration] = STATE(419),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1790),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_sequence_expression] = STATE(1662),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [aux_sym_program_repeat1] = STATE(30),
    [aux_sym_export_statement_repeat1] = STATE(1267),
    [sym_identifier] = ACTIONS(9),
    [anon_sym_export] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(15),
    [anon_sym_RBRACE] = ACTIONS(483),
    [anon_sym_import] = ACTIONS(17),
    [anon_sym_with] = ACTIONS(19),
    [anon_sym_var] = ACTIONS(21),
    [anon_sym_let] = ACTIONS(23),
    [anon_sym_const] = ACTIONS(25),
    [anon_sym_using] = ACTIONS(27),
    [anon_sym_await] = ACTIONS(29),
    [anon_sym_if] = ACTIONS(31),
    [anon_sym_switch] = ACTIONS(33),
    [anon_sym_for] = ACTIONS(35),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_SEMI] = ACTIONS(39),
    [anon_sym_while] = ACTIONS(41),
    [anon_sym_do] = ACTIONS(43),
    [anon_sym_try] = ACTIONS(45),
    [anon_sym_break] = ACTIONS(47),
    [anon_sym_continue] = ACTIONS(49),
    [anon_sym_debugger] = ACTIONS(51),
    [anon_sym_return] = ACTIONS(53),
    [anon_sym_throw] = ACTIONS(55),
    [anon_sym_yield] = ACTIONS(57),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(67),
    [anon_sym_async] = ACTIONS(69),
    [anon_sym_function] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(75),
    [anon_sym_DASH] = ACTIONS(75),
    [anon_sym_SLASH] = ACTIONS(77),
    [anon_sym_BANG] = ACTIONS(79),
    [anon_sym_TILDE] = ACTIONS(79),
    [anon_sym_typeof] = ACTIONS(75),
    [anon_sym_void] = ACTIONS(75),
    [anon_sym_delete] = ACTIONS(75),
    [anon_sym_PLUS_PLUS] = ACTIONS(81),
    [anon_sym_DASH_DASH] = ACTIONS(81),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(87),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(95),
    [anon_sym_get] = ACTIONS(95),
    [anon_sym_set] = ACTIONS(95),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(29)] = {
    [sym_import] = STATE(1326),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(692),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1790),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1790),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_function_expression] = STATE(740),
    [sym_generator_function] = STATE(740),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1790),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [aux_sym_export_statement_repeat1] = STATE(1381),
    [sym_identifier] = ACTIONS(161),
    [anon_sym_export] = ACTIONS(163),
    [anon_sym_STAR] = ACTIONS(165),
    [anon_sym_LBRACE] = ACTIONS(167),
    [anon_sym_COMMA] = ACTIONS(169),
    [anon_sym_import] = ACTIONS(173),
    [anon_sym_let] = ACTIONS(163),
    [anon_sym_using] = ACTIONS(175),
    [anon_sym_await] = ACTIONS(177),
    [anon_sym_LPAREN] = ACTIONS(485),
    [anon_sym_SEMI] = ACTIONS(169),
    [anon_sym_in] = ACTIONS(165),
    [anon_sym_COLON] = ACTIONS(488),
    [anon_sym_yield] = ACTIONS(57),
    [anon_sym_EQ] = ACTIONS(490),
    [anon_sym_LBRACK] = ACTIONS(188),
    [anon_sym_LT] = ACTIONS(191),
    [anon_sym_GT] = ACTIONS(165),
    [anon_sym_DOT] = ACTIONS(165),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(194),
    [anon_sym_async] = ACTIONS(196),
    [anon_sym_function] = ACTIONS(198),
    [anon_sym_EQ_GT] = ACTIONS(200),
    [sym_optional_chain] = ACTIONS(169),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_PLUS_EQ] = ACTIONS(202),
    [anon_sym_DASH_EQ] = ACTIONS(202),
    [anon_sym_STAR_EQ] = ACTIONS(202),
    [anon_sym_SLASH_EQ] = ACTIONS(202),
    [anon_sym_PERCENT_EQ] = ACTIONS(202),
    [anon_sym_CARET_EQ] = ACTIONS(202),
    [anon_sym_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_LT_LT_EQ] = ACTIONS(202),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(202),
    [anon_sym_QMARK_QMARK_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP] = ACTIONS(165),
    [anon_sym_PIPE_PIPE] = ACTIONS(165),
    [anon_sym_GT_GT] = ACTIONS(165),
    [anon_sym_GT_GT_GT] = ACTIONS(165),
    [anon_sym_LT_LT] = ACTIONS(165),
    [anon_sym_AMP] = ACTIONS(165),
    [anon_sym_CARET] = ACTIONS(165),
    [anon_sym_PIPE] = ACTIONS(165),
    [anon_sym_PLUS] = ACTIONS(204),
    [anon_sym_DASH] = ACTIONS(204),
    [anon_sym_SLASH] = ACTIONS(207),
    [anon_sym_PERCENT] = ACTIONS(165),
    [anon_sym_STAR_STAR] = ACTIONS(165),
    [anon_sym_LT_EQ] = ACTIONS(169),
    [anon_sym_EQ_EQ] = ACTIONS(165),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(169),
    [anon_sym_BANG_EQ] = ACTIONS(165),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(169),
    [anon_sym_GT_EQ] = ACTIONS(169),
    [anon_sym_QMARK_QMARK] = ACTIONS(165),
    [anon_sym_instanceof] = ACTIONS(165),
    [anon_sym_BANG] = ACTIONS(75),
    [anon_sym_TILDE] = ACTIONS(79),
    [anon_sym_typeof] = ACTIONS(75),
    [anon_sym_void] = ACTIONS(75),
    [anon_sym_delete] = ACTIONS(75),
    [anon_sym_PLUS_PLUS] = ACTIONS(210),
    [anon_sym_DASH_DASH] = ACTIONS(210),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(213),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(87),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(163),
    [anon_sym_get] = ACTIONS(163),
    [anon_sym_set] = ACTIONS(163),
    [sym__automatic_semicolon] = ACTIONS(169),
    [sym__ternary_qmark] = ACTIONS(169),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(30)] = {
    [sym_export_statement] = STATE(471),
    [sym_declaration] = STATE(471),
    [sym_import] = STATE(1326),
    [sym_import_statement] = STATE(471),
    [sym_statement] = STATE(9),
    [sym_expression_statement] = STATE(471),
    [sym_variable_declaration] = STATE(419),
    [sym_lexical_declaration] = STATE(419),
    [sym_using_declaration] = STATE(419),
    [sym_statement_block] = STATE(471),
    [sym_if_statement] = STATE(471),
    [sym_switch_statement] = STATE(471),
    [sym_for_statement] = STATE(471),
    [sym_for_in_statement] = STATE(471),
    [sym_while_statement] = STATE(471),
    [sym_do_statement] = STATE(471),
    [sym_try_statement] = STATE(471),
    [sym_with_statement] = STATE(471),
    [sym_break_statement] = STATE(471),
    [sym_continue_statement] = STATE(471),
    [sym_debugger_statement] = STATE(471),
    [sym_return_statement] = STATE(471),
    [sym_throw_statement] = STATE(471),
    [sym_empty_statement] = STATE(471),
    [sym_labeled_statement] = STATE(471),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(823),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1790),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1790),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_class_declaration] = STATE(419),
    [sym_function_expression] = STATE(740),
    [sym_function_declaration] = STATE(419),
    [sym_generator_function] = STATE(740),
    [sym_generator_function_declaration] = STATE(419),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1790),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_sequence_expression] = STATE(1662),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [aux_sym_program_repeat1] = STATE(9),
    [aux_sym_export_statement_repeat1] = STATE(1267),
    [sym_identifier] = ACTIONS(9),
    [anon_sym_export] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(15),
    [anon_sym_RBRACE] = ACTIONS(492),
    [anon_sym_import] = ACTIONS(17),
    [anon_sym_with] = ACTIONS(19),
    [anon_sym_var] = ACTIONS(21),
    [anon_sym_let] = ACTIONS(23),
    [anon_sym_const] = ACTIONS(25),
    [anon_sym_using] = ACTIONS(27),
    [anon_sym_await] = ACTIONS(29),
    [anon_sym_if] = ACTIONS(31),
    [anon_sym_switch] = ACTIONS(33),
    [anon_sym_for] = ACTIONS(35),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_SEMI] = ACTIONS(39),
    [anon_sym_while] = ACTIONS(41),
    [anon_sym_do] = ACTIONS(43),
    [anon_sym_try] = ACTIONS(45),
    [anon_sym_break] = ACTIONS(47),
    [anon_sym_continue] = ACTIONS(49),
    [anon_sym_debugger] = ACTIONS(51),
    [anon_sym_return] = ACTIONS(53),
    [anon_sym_throw] = ACTIONS(55),
    [anon_sym_yield] = ACTIONS(57),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(67),
    [anon_sym_async] = ACTIONS(69),
    [anon_sym_function] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(75),
    [anon_sym_DASH] = ACTIONS(75),
    [anon_sym_SLASH] = ACTIONS(77),
    [anon_sym_BANG] = ACTIONS(79),
    [anon_sym_TILDE] = ACTIONS(79),
    [anon_sym_typeof] = ACTIONS(75),
    [anon_sym_void] = ACTIONS(75),
    [anon_sym_delete] = ACTIONS(75),
    [anon_sym_PLUS_PLUS] = ACTIONS(81),
    [anon_sym_DASH_DASH] = ACTIONS(81),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(87),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(95),
    [anon_sym_get] = ACTIONS(95),
    [anon_sym_set] = ACTIONS(95),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(31)] = {
    [sym_import] = STATE(1326),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(692),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1790),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1790),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_function_expression] = STATE(740),
    [sym_generator_function] = STATE(740),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1790),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [aux_sym_export_statement_repeat1] = STATE(1381),
    [sym_identifier] = ACTIONS(161),
    [anon_sym_export] = ACTIONS(163),
    [anon_sym_STAR] = ACTIONS(165),
    [anon_sym_LBRACE] = ACTIONS(167),
    [anon_sym_COMMA] = ACTIONS(169),
    [anon_sym_import] = ACTIONS(173),
    [anon_sym_let] = ACTIONS(163),
    [anon_sym_using] = ACTIONS(175),
    [anon_sym_await] = ACTIONS(177),
    [anon_sym_LPAREN] = ACTIONS(485),
    [anon_sym_SEMI] = ACTIONS(169),
    [anon_sym_in] = ACTIONS(165),
    [anon_sym_COLON] = ACTIONS(494),
    [anon_sym_yield] = ACTIONS(57),
    [anon_sym_EQ] = ACTIONS(490),
    [anon_sym_LBRACK] = ACTIONS(188),
    [anon_sym_LT] = ACTIONS(191),
    [anon_sym_GT] = ACTIONS(165),
    [anon_sym_DOT] = ACTIONS(165),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(194),
    [anon_sym_async] = ACTIONS(196),
    [anon_sym_function] = ACTIONS(198),
    [anon_sym_EQ_GT] = ACTIONS(200),
    [sym_optional_chain] = ACTIONS(169),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_PLUS_EQ] = ACTIONS(202),
    [anon_sym_DASH_EQ] = ACTIONS(202),
    [anon_sym_STAR_EQ] = ACTIONS(202),
    [anon_sym_SLASH_EQ] = ACTIONS(202),
    [anon_sym_PERCENT_EQ] = ACTIONS(202),
    [anon_sym_CARET_EQ] = ACTIONS(202),
    [anon_sym_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_LT_LT_EQ] = ACTIONS(202),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(202),
    [anon_sym_QMARK_QMARK_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP] = ACTIONS(165),
    [anon_sym_PIPE_PIPE] = ACTIONS(165),
    [anon_sym_GT_GT] = ACTIONS(165),
    [anon_sym_GT_GT_GT] = ACTIONS(165),
    [anon_sym_LT_LT] = ACTIONS(165),
    [anon_sym_AMP] = ACTIONS(165),
    [anon_sym_CARET] = ACTIONS(165),
    [anon_sym_PIPE] = ACTIONS(165),
    [anon_sym_PLUS] = ACTIONS(204),
    [anon_sym_DASH] = ACTIONS(204),
    [anon_sym_SLASH] = ACTIONS(207),
    [anon_sym_PERCENT] = ACTIONS(165),
    [anon_sym_STAR_STAR] = ACTIONS(165),
    [anon_sym_LT_EQ] = ACTIONS(169),
    [anon_sym_EQ_EQ] = ACTIONS(165),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(169),
    [anon_sym_BANG_EQ] = ACTIONS(165),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(169),
    [anon_sym_GT_EQ] = ACTIONS(169),
    [anon_sym_QMARK_QMARK] = ACTIONS(165),
    [anon_sym_instanceof] = ACTIONS(165),
    [anon_sym_BANG] = ACTIONS(75),
    [anon_sym_TILDE] = ACTIONS(79),
    [anon_sym_typeof] = ACTIONS(75),
    [anon_sym_void] = ACTIONS(75),
    [anon_sym_delete] = ACTIONS(75),
    [anon_sym_PLUS_PLUS] = ACTIONS(210),
    [anon_sym_DASH_DASH] = ACTIONS(210),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(213),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(87),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(163),
    [anon_sym_get] = ACTIONS(163),
    [anon_sym_set] = ACTIONS(163),
    [sym__automatic_semicolon] = ACTIONS(169),
    [sym__ternary_qmark] = ACTIONS(169),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(32)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(498),
    [sym_expression] = STATE(587),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1762),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1762),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(498),
    [sym_subscript_expression] = STATE(498),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1138),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1762),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1816),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(430),
    [anon_sym_export] = ACTIONS(432),
    [anon_sym_STAR] = ACTIONS(165),
    [anon_sym_LBRACE] = ACTIONS(434),
    [anon_sym_COMMA] = ACTIONS(496),
    [anon_sym_RBRACE] = ACTIONS(496),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(432),
    [anon_sym_await] = ACTIONS(436),
    [anon_sym_LPAREN] = ACTIONS(379),
    [anon_sym_RPAREN] = ACTIONS(496),
    [anon_sym_in] = ACTIONS(165),
    [anon_sym_yield] = ACTIONS(438),
    [anon_sym_EQ] = ACTIONS(440),
    [anon_sym_LBRACK] = ACTIONS(442),
    [anon_sym_RBRACK] = ACTIONS(496),
    [anon_sym_LT] = ACTIONS(389),
    [anon_sym_GT] = ACTIONS(165),
    [anon_sym_DOT] = ACTIONS(165),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(445),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_EQ_GT] = ACTIONS(498),
    [sym_optional_chain] = ACTIONS(169),
    [anon_sym_new] = ACTIONS(447),
    [anon_sym_PLUS_EQ] = ACTIONS(202),
    [anon_sym_DASH_EQ] = ACTIONS(202),
    [anon_sym_STAR_EQ] = ACTIONS(202),
    [anon_sym_SLASH_EQ] = ACTIONS(202),
    [anon_sym_PERCENT_EQ] = ACTIONS(202),
    [anon_sym_CARET_EQ] = ACTIONS(202),
    [anon_sym_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_LT_LT_EQ] = ACTIONS(202),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(202),
    [anon_sym_QMARK_QMARK_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP] = ACTIONS(165),
    [anon_sym_PIPE_PIPE] = ACTIONS(165),
    [anon_sym_GT_GT] = ACTIONS(165),
    [anon_sym_GT_GT_GT] = ACTIONS(165),
    [anon_sym_LT_LT] = ACTIONS(165),
    [anon_sym_AMP] = ACTIONS(165),
    [anon_sym_CARET] = ACTIONS(165),
    [anon_sym_PIPE] = ACTIONS(165),
    [anon_sym_PLUS] = ACTIONS(449),
    [anon_sym_DASH] = ACTIONS(449),
    [anon_sym_SLASH] = ACTIONS(409),
    [anon_sym_PERCENT] = ACTIONS(165),
    [anon_sym_STAR_STAR] = ACTIONS(165),
    [anon_sym_LT_EQ] = ACTIONS(169),
    [anon_sym_EQ_EQ] = ACTIONS(165),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(169),
    [anon_sym_BANG_EQ] = ACTIONS(165),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(169),
    [anon_sym_GT_EQ] = ACTIONS(169),
    [anon_sym_QMARK_QMARK] = ACTIONS(165),
    [anon_sym_instanceof] = ACTIONS(165),
    [anon_sym_BANG] = ACTIONS(452),
    [anon_sym_TILDE] = ACTIONS(454),
    [anon_sym_typeof] = ACTIONS(452),
    [anon_sym_void] = ACTIONS(452),
    [anon_sym_delete] = ACTIONS(452),
    [anon_sym_PLUS_PLUS] = ACTIONS(456),
    [anon_sym_DASH_DASH] = ACTIONS(456),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(419),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(459),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(461),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(432),
    [anon_sym_get] = ACTIONS(432),
    [anon_sym_set] = ACTIONS(432),
    [sym__ternary_qmark] = ACTIONS(169),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(33)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(498),
    [sym_expression] = STATE(587),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1762),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1762),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(498),
    [sym_subscript_expression] = STATE(498),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1138),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1762),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1816),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(430),
    [anon_sym_export] = ACTIONS(432),
    [anon_sym_STAR] = ACTIONS(165),
    [anon_sym_LBRACE] = ACTIONS(434),
    [anon_sym_COMMA] = ACTIONS(169),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(432),
    [anon_sym_await] = ACTIONS(436),
    [anon_sym_of] = ACTIONS(165),
    [anon_sym_LPAREN] = ACTIONS(379),
    [anon_sym_SEMI] = ACTIONS(169),
    [anon_sym_in] = ACTIONS(165),
    [anon_sym_yield] = ACTIONS(438),
    [anon_sym_EQ] = ACTIONS(440),
    [anon_sym_LBRACK] = ACTIONS(442),
    [anon_sym_LT] = ACTIONS(389),
    [anon_sym_GT] = ACTIONS(165),
    [anon_sym_DOT] = ACTIONS(165),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(445),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_EQ_GT] = ACTIONS(500),
    [sym_optional_chain] = ACTIONS(169),
    [anon_sym_new] = ACTIONS(447),
    [anon_sym_PLUS_EQ] = ACTIONS(202),
    [anon_sym_DASH_EQ] = ACTIONS(202),
    [anon_sym_STAR_EQ] = ACTIONS(202),
    [anon_sym_SLASH_EQ] = ACTIONS(202),
    [anon_sym_PERCENT_EQ] = ACTIONS(202),
    [anon_sym_CARET_EQ] = ACTIONS(202),
    [anon_sym_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_LT_LT_EQ] = ACTIONS(202),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(202),
    [anon_sym_QMARK_QMARK_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP] = ACTIONS(165),
    [anon_sym_PIPE_PIPE] = ACTIONS(165),
    [anon_sym_GT_GT] = ACTIONS(165),
    [anon_sym_GT_GT_GT] = ACTIONS(165),
    [anon_sym_LT_LT] = ACTIONS(165),
    [anon_sym_AMP] = ACTIONS(165),
    [anon_sym_CARET] = ACTIONS(165),
    [anon_sym_PIPE] = ACTIONS(165),
    [anon_sym_PLUS] = ACTIONS(449),
    [anon_sym_DASH] = ACTIONS(449),
    [anon_sym_SLASH] = ACTIONS(409),
    [anon_sym_PERCENT] = ACTIONS(165),
    [anon_sym_STAR_STAR] = ACTIONS(165),
    [anon_sym_LT_EQ] = ACTIONS(169),
    [anon_sym_EQ_EQ] = ACTIONS(165),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(169),
    [anon_sym_BANG_EQ] = ACTIONS(165),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(169),
    [anon_sym_GT_EQ] = ACTIONS(169),
    [anon_sym_QMARK_QMARK] = ACTIONS(165),
    [anon_sym_instanceof] = ACTIONS(165),
    [anon_sym_BANG] = ACTIONS(452),
    [anon_sym_TILDE] = ACTIONS(454),
    [anon_sym_typeof] = ACTIONS(452),
    [anon_sym_void] = ACTIONS(452),
    [anon_sym_delete] = ACTIONS(452),
    [anon_sym_PLUS_PLUS] = ACTIONS(456),
    [anon_sym_DASH_DASH] = ACTIONS(456),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(419),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(459),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(461),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(432),
    [anon_sym_get] = ACTIONS(432),
    [anon_sym_set] = ACTIONS(432),
    [sym__automatic_semicolon] = ACTIONS(169),
    [sym__ternary_qmark] = ACTIONS(169),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(34)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(507),
    [sym_expression] = STATE(587),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1815),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1815),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(507),
    [sym_subscript_expression] = STATE(507),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1139),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1815),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1763),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(369),
    [anon_sym_export] = ACTIONS(371),
    [anon_sym_STAR] = ACTIONS(165),
    [anon_sym_LBRACE] = ACTIONS(373),
    [anon_sym_COMMA] = ACTIONS(169),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(371),
    [anon_sym_using] = ACTIONS(502),
    [anon_sym_await] = ACTIONS(377),
    [anon_sym_of] = ACTIONS(504),
    [anon_sym_LPAREN] = ACTIONS(379),
    [anon_sym_SEMI] = ACTIONS(169),
    [anon_sym_in] = ACTIONS(506),
    [anon_sym_yield] = ACTIONS(382),
    [anon_sym_EQ] = ACTIONS(384),
    [anon_sym_LBRACK] = ACTIONS(386),
    [anon_sym_LT] = ACTIONS(389),
    [anon_sym_GT] = ACTIONS(165),
    [anon_sym_DOT] = ACTIONS(165),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(398),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_EQ_GT] = ACTIONS(402),
    [sym_optional_chain] = ACTIONS(169),
    [anon_sym_new] = ACTIONS(404),
    [anon_sym_PLUS_EQ] = ACTIONS(202),
    [anon_sym_DASH_EQ] = ACTIONS(202),
    [anon_sym_STAR_EQ] = ACTIONS(202),
    [anon_sym_SLASH_EQ] = ACTIONS(202),
    [anon_sym_PERCENT_EQ] = ACTIONS(202),
    [anon_sym_CARET_EQ] = ACTIONS(202),
    [anon_sym_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_LT_LT_EQ] = ACTIONS(202),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(202),
    [anon_sym_QMARK_QMARK_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP] = ACTIONS(165),
    [anon_sym_PIPE_PIPE] = ACTIONS(165),
    [anon_sym_GT_GT] = ACTIONS(165),
    [anon_sym_GT_GT_GT] = ACTIONS(165),
    [anon_sym_LT_LT] = ACTIONS(165),
    [anon_sym_AMP] = ACTIONS(165),
    [anon_sym_CARET] = ACTIONS(165),
    [anon_sym_PIPE] = ACTIONS(165),
    [anon_sym_PLUS] = ACTIONS(406),
    [anon_sym_DASH] = ACTIONS(406),
    [anon_sym_SLASH] = ACTIONS(409),
    [anon_sym_PERCENT] = ACTIONS(165),
    [anon_sym_STAR_STAR] = ACTIONS(165),
    [anon_sym_LT_EQ] = ACTIONS(169),
    [anon_sym_EQ_EQ] = ACTIONS(165),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(169),
    [anon_sym_BANG_EQ] = ACTIONS(165),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(169),
    [anon_sym_GT_EQ] = ACTIONS(169),
    [anon_sym_QMARK_QMARK] = ACTIONS(165),
    [anon_sym_instanceof] = ACTIONS(165),
    [anon_sym_BANG] = ACTIONS(412),
    [anon_sym_TILDE] = ACTIONS(414),
    [anon_sym_typeof] = ACTIONS(412),
    [anon_sym_void] = ACTIONS(412),
    [anon_sym_delete] = ACTIONS(412),
    [anon_sym_PLUS_PLUS] = ACTIONS(416),
    [anon_sym_DASH_DASH] = ACTIONS(416),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(419),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(424),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(428),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(371),
    [anon_sym_get] = ACTIONS(371),
    [anon_sym_set] = ACTIONS(371),
    [sym__ternary_qmark] = ACTIONS(169),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(35)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(498),
    [sym_expression] = STATE(587),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1762),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1762),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(498),
    [sym_subscript_expression] = STATE(498),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1138),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1762),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1816),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(430),
    [anon_sym_export] = ACTIONS(432),
    [anon_sym_STAR] = ACTIONS(165),
    [anon_sym_LBRACE] = ACTIONS(434),
    [anon_sym_COMMA] = ACTIONS(169),
    [anon_sym_RBRACE] = ACTIONS(169),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(432),
    [anon_sym_await] = ACTIONS(436),
    [anon_sym_LPAREN] = ACTIONS(379),
    [anon_sym_SEMI] = ACTIONS(169),
    [anon_sym_in] = ACTIONS(165),
    [anon_sym_yield] = ACTIONS(438),
    [anon_sym_EQ] = ACTIONS(440),
    [anon_sym_LBRACK] = ACTIONS(442),
    [anon_sym_LT] = ACTIONS(389),
    [anon_sym_GT] = ACTIONS(165),
    [anon_sym_DOT] = ACTIONS(165),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(445),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_EQ_GT] = ACTIONS(200),
    [sym_optional_chain] = ACTIONS(169),
    [anon_sym_new] = ACTIONS(447),
    [anon_sym_PLUS_EQ] = ACTIONS(202),
    [anon_sym_DASH_EQ] = ACTIONS(202),
    [anon_sym_STAR_EQ] = ACTIONS(202),
    [anon_sym_SLASH_EQ] = ACTIONS(202),
    [anon_sym_PERCENT_EQ] = ACTIONS(202),
    [anon_sym_CARET_EQ] = ACTIONS(202),
    [anon_sym_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_LT_LT_EQ] = ACTIONS(202),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(202),
    [anon_sym_QMARK_QMARK_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP] = ACTIONS(165),
    [anon_sym_PIPE_PIPE] = ACTIONS(165),
    [anon_sym_GT_GT] = ACTIONS(165),
    [anon_sym_GT_GT_GT] = ACTIONS(165),
    [anon_sym_LT_LT] = ACTIONS(165),
    [anon_sym_AMP] = ACTIONS(165),
    [anon_sym_CARET] = ACTIONS(165),
    [anon_sym_PIPE] = ACTIONS(165),
    [anon_sym_PLUS] = ACTIONS(449),
    [anon_sym_DASH] = ACTIONS(449),
    [anon_sym_SLASH] = ACTIONS(409),
    [anon_sym_PERCENT] = ACTIONS(165),
    [anon_sym_STAR_STAR] = ACTIONS(165),
    [anon_sym_LT_EQ] = ACTIONS(169),
    [anon_sym_EQ_EQ] = ACTIONS(165),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(169),
    [anon_sym_BANG_EQ] = ACTIONS(165),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(169),
    [anon_sym_GT_EQ] = ACTIONS(169),
    [anon_sym_QMARK_QMARK] = ACTIONS(165),
    [anon_sym_instanceof] = ACTIONS(165),
    [anon_sym_BANG] = ACTIONS(452),
    [anon_sym_TILDE] = ACTIONS(454),
    [anon_sym_typeof] = ACTIONS(452),
    [anon_sym_void] = ACTIONS(452),
    [anon_sym_delete] = ACTIONS(452),
    [anon_sym_PLUS_PLUS] = ACTIONS(456),
    [anon_sym_DASH_DASH] = ACTIONS(456),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(419),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(459),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(461),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(432),
    [anon_sym_get] = ACTIONS(432),
    [anon_sym_set] = ACTIONS(432),
    [sym__automatic_semicolon] = ACTIONS(169),
    [sym__ternary_qmark] = ACTIONS(169),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(36)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(498),
    [sym_expression] = STATE(587),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1762),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1762),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(498),
    [sym_subscript_expression] = STATE(498),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1138),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1762),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1816),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(430),
    [anon_sym_export] = ACTIONS(432),
    [anon_sym_STAR] = ACTIONS(165),
    [anon_sym_LBRACE] = ACTIONS(434),
    [anon_sym_COMMA] = ACTIONS(509),
    [anon_sym_RBRACE] = ACTIONS(509),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(432),
    [anon_sym_await] = ACTIONS(436),
    [anon_sym_LPAREN] = ACTIONS(379),
    [anon_sym_RPAREN] = ACTIONS(509),
    [anon_sym_in] = ACTIONS(165),
    [anon_sym_yield] = ACTIONS(438),
    [anon_sym_EQ] = ACTIONS(511),
    [anon_sym_LBRACK] = ACTIONS(442),
    [anon_sym_RBRACK] = ACTIONS(509),
    [anon_sym_LT] = ACTIONS(389),
    [anon_sym_GT] = ACTIONS(165),
    [anon_sym_DOT] = ACTIONS(165),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(445),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_EQ_GT] = ACTIONS(498),
    [sym_optional_chain] = ACTIONS(169),
    [anon_sym_new] = ACTIONS(447),
    [anon_sym_PLUS_EQ] = ACTIONS(202),
    [anon_sym_DASH_EQ] = ACTIONS(202),
    [anon_sym_STAR_EQ] = ACTIONS(202),
    [anon_sym_SLASH_EQ] = ACTIONS(202),
    [anon_sym_PERCENT_EQ] = ACTIONS(202),
    [anon_sym_CARET_EQ] = ACTIONS(202),
    [anon_sym_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_LT_LT_EQ] = ACTIONS(202),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(202),
    [anon_sym_QMARK_QMARK_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP] = ACTIONS(165),
    [anon_sym_PIPE_PIPE] = ACTIONS(165),
    [anon_sym_GT_GT] = ACTIONS(165),
    [anon_sym_GT_GT_GT] = ACTIONS(165),
    [anon_sym_LT_LT] = ACTIONS(165),
    [anon_sym_AMP] = ACTIONS(165),
    [anon_sym_CARET] = ACTIONS(165),
    [anon_sym_PIPE] = ACTIONS(165),
    [anon_sym_PLUS] = ACTIONS(449),
    [anon_sym_DASH] = ACTIONS(449),
    [anon_sym_SLASH] = ACTIONS(409),
    [anon_sym_PERCENT] = ACTIONS(165),
    [anon_sym_STAR_STAR] = ACTIONS(165),
    [anon_sym_LT_EQ] = ACTIONS(169),
    [anon_sym_EQ_EQ] = ACTIONS(165),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(169),
    [anon_sym_BANG_EQ] = ACTIONS(165),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(169),
    [anon_sym_GT_EQ] = ACTIONS(169),
    [anon_sym_QMARK_QMARK] = ACTIONS(165),
    [anon_sym_instanceof] = ACTIONS(165),
    [anon_sym_BANG] = ACTIONS(452),
    [anon_sym_TILDE] = ACTIONS(454),
    [anon_sym_typeof] = ACTIONS(452),
    [anon_sym_void] = ACTIONS(452),
    [anon_sym_delete] = ACTIONS(452),
    [anon_sym_PLUS_PLUS] = ACTIONS(456),
    [anon_sym_DASH_DASH] = ACTIONS(456),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(419),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(459),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(461),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(432),
    [anon_sym_get] = ACTIONS(432),
    [anon_sym_set] = ACTIONS(432),
    [sym__ternary_qmark] = ACTIONS(169),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(37)] = {
    [sym_import] = STATE(1326),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(692),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1790),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1790),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_function_expression] = STATE(740),
    [sym_generator_function] = STATE(740),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1790),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [aux_sym_export_statement_repeat1] = STATE(1381),
    [sym_identifier] = ACTIONS(161),
    [anon_sym_export] = ACTIONS(163),
    [anon_sym_STAR] = ACTIONS(165),
    [anon_sym_LBRACE] = ACTIONS(167),
    [anon_sym_COMMA] = ACTIONS(169),
    [anon_sym_RBRACE] = ACTIONS(169),
    [anon_sym_import] = ACTIONS(173),
    [anon_sym_let] = ACTIONS(163),
    [anon_sym_await] = ACTIONS(177),
    [anon_sym_LPAREN] = ACTIONS(485),
    [anon_sym_SEMI] = ACTIONS(169),
    [anon_sym_in] = ACTIONS(165),
    [anon_sym_yield] = ACTIONS(57),
    [anon_sym_EQ] = ACTIONS(490),
    [anon_sym_LBRACK] = ACTIONS(188),
    [anon_sym_LT] = ACTIONS(191),
    [anon_sym_GT] = ACTIONS(165),
    [anon_sym_DOT] = ACTIONS(165),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(194),
    [anon_sym_async] = ACTIONS(196),
    [anon_sym_function] = ACTIONS(198),
    [anon_sym_EQ_GT] = ACTIONS(200),
    [sym_optional_chain] = ACTIONS(169),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_PLUS_EQ] = ACTIONS(202),
    [anon_sym_DASH_EQ] = ACTIONS(202),
    [anon_sym_STAR_EQ] = ACTIONS(202),
    [anon_sym_SLASH_EQ] = ACTIONS(202),
    [anon_sym_PERCENT_EQ] = ACTIONS(202),
    [anon_sym_CARET_EQ] = ACTIONS(202),
    [anon_sym_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_LT_LT_EQ] = ACTIONS(202),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(202),
    [anon_sym_QMARK_QMARK_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP] = ACTIONS(165),
    [anon_sym_PIPE_PIPE] = ACTIONS(165),
    [anon_sym_GT_GT] = ACTIONS(165),
    [anon_sym_GT_GT_GT] = ACTIONS(165),
    [anon_sym_LT_LT] = ACTIONS(165),
    [anon_sym_AMP] = ACTIONS(165),
    [anon_sym_CARET] = ACTIONS(165),
    [anon_sym_PIPE] = ACTIONS(165),
    [anon_sym_PLUS] = ACTIONS(204),
    [anon_sym_DASH] = ACTIONS(204),
    [anon_sym_SLASH] = ACTIONS(207),
    [anon_sym_PERCENT] = ACTIONS(165),
    [anon_sym_STAR_STAR] = ACTIONS(165),
    [anon_sym_LT_EQ] = ACTIONS(169),
    [anon_sym_EQ_EQ] = ACTIONS(165),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(169),
    [anon_sym_BANG_EQ] = ACTIONS(165),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(169),
    [anon_sym_GT_EQ] = ACTIONS(169),
    [anon_sym_QMARK_QMARK] = ACTIONS(165),
    [anon_sym_instanceof] = ACTIONS(165),
    [anon_sym_BANG] = ACTIONS(75),
    [anon_sym_TILDE] = ACTIONS(79),
    [anon_sym_typeof] = ACTIONS(75),
    [anon_sym_void] = ACTIONS(75),
    [anon_sym_delete] = ACTIONS(75),
    [anon_sym_PLUS_PLUS] = ACTIONS(210),
    [anon_sym_DASH_DASH] = ACTIONS(210),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(213),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(87),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(163),
    [anon_sym_get] = ACTIONS(163),
    [anon_sym_set] = ACTIONS(163),
    [sym__automatic_semicolon] = ACTIONS(169),
    [sym__ternary_qmark] = ACTIONS(169),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(38)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(507),
    [sym_expression] = STATE(587),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1815),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1815),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(507),
    [sym_subscript_expression] = STATE(507),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1139),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1815),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1763),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(369),
    [anon_sym_export] = ACTIONS(371),
    [anon_sym_STAR] = ACTIONS(165),
    [anon_sym_LBRACE] = ACTIONS(373),
    [anon_sym_COMMA] = ACTIONS(514),
    [anon_sym_RBRACE] = ACTIONS(514),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(371),
    [anon_sym_await] = ACTIONS(377),
    [anon_sym_LPAREN] = ACTIONS(379),
    [anon_sym_RPAREN] = ACTIONS(514),
    [anon_sym_in] = ACTIONS(165),
    [anon_sym_yield] = ACTIONS(382),
    [anon_sym_EQ] = ACTIONS(517),
    [anon_sym_LBRACK] = ACTIONS(386),
    [anon_sym_RBRACK] = ACTIONS(514),
    [anon_sym_LT] = ACTIONS(389),
    [anon_sym_GT] = ACTIONS(165),
    [anon_sym_DOT] = ACTIONS(165),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(398),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_EQ_GT] = ACTIONS(402),
    [sym_optional_chain] = ACTIONS(169),
    [anon_sym_new] = ACTIONS(404),
    [anon_sym_PLUS_EQ] = ACTIONS(202),
    [anon_sym_DASH_EQ] = ACTIONS(202),
    [anon_sym_STAR_EQ] = ACTIONS(202),
    [anon_sym_SLASH_EQ] = ACTIONS(202),
    [anon_sym_PERCENT_EQ] = ACTIONS(202),
    [anon_sym_CARET_EQ] = ACTIONS(202),
    [anon_sym_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_LT_LT_EQ] = ACTIONS(202),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(202),
    [anon_sym_QMARK_QMARK_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP] = ACTIONS(165),
    [anon_sym_PIPE_PIPE] = ACTIONS(165),
    [anon_sym_GT_GT] = ACTIONS(165),
    [anon_sym_GT_GT_GT] = ACTIONS(165),
    [anon_sym_LT_LT] = ACTIONS(165),
    [anon_sym_AMP] = ACTIONS(165),
    [anon_sym_CARET] = ACTIONS(165),
    [anon_sym_PIPE] = ACTIONS(165),
    [anon_sym_PLUS] = ACTIONS(406),
    [anon_sym_DASH] = ACTIONS(406),
    [anon_sym_SLASH] = ACTIONS(409),
    [anon_sym_PERCENT] = ACTIONS(165),
    [anon_sym_STAR_STAR] = ACTIONS(165),
    [anon_sym_LT_EQ] = ACTIONS(169),
    [anon_sym_EQ_EQ] = ACTIONS(165),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(169),
    [anon_sym_BANG_EQ] = ACTIONS(165),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(169),
    [anon_sym_GT_EQ] = ACTIONS(169),
    [anon_sym_QMARK_QMARK] = ACTIONS(165),
    [anon_sym_instanceof] = ACTIONS(165),
    [anon_sym_BANG] = ACTIONS(412),
    [anon_sym_TILDE] = ACTIONS(414),
    [anon_sym_typeof] = ACTIONS(412),
    [anon_sym_void] = ACTIONS(412),
    [anon_sym_delete] = ACTIONS(412),
    [anon_sym_PLUS_PLUS] = ACTIONS(416),
    [anon_sym_DASH_DASH] = ACTIONS(416),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(419),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(424),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(428),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(371),
    [anon_sym_get] = ACTIONS(371),
    [anon_sym_set] = ACTIONS(371),
    [sym__ternary_qmark] = ACTIONS(169),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(39)] = {
    [sym_import] = STATE(1326),
    [sym_parenthesized_expression] = STATE(560),
    [sym_expression] = STATE(692),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1802),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1802),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_function_expression] = STATE(740),
    [sym_generator_function] = STATE(740),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(560),
    [sym_subscript_expression] = STATE(560),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1143),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1802),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1803),
    [aux_sym_export_statement_repeat1] = STATE(1381),
    [sym_identifier] = ACTIONS(520),
    [anon_sym_export] = ACTIONS(522),
    [anon_sym_STAR] = ACTIONS(165),
    [anon_sym_LBRACE] = ACTIONS(167),
    [anon_sym_COMMA] = ACTIONS(169),
    [anon_sym_import] = ACTIONS(173),
    [anon_sym_let] = ACTIONS(522),
    [anon_sym_await] = ACTIONS(524),
    [anon_sym_of] = ACTIONS(165),
    [anon_sym_LPAREN] = ACTIONS(485),
    [anon_sym_SEMI] = ACTIONS(169),
    [anon_sym_in] = ACTIONS(165),
    [anon_sym_yield] = ACTIONS(526),
    [anon_sym_EQ] = ACTIONS(528),
    [anon_sym_LBRACK] = ACTIONS(188),
    [anon_sym_LT] = ACTIONS(191),
    [anon_sym_GT] = ACTIONS(165),
    [anon_sym_DOT] = ACTIONS(165),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(194),
    [anon_sym_async] = ACTIONS(530),
    [anon_sym_function] = ACTIONS(198),
    [anon_sym_EQ_GT] = ACTIONS(500),
    [sym_optional_chain] = ACTIONS(169),
    [anon_sym_new] = ACTIONS(532),
    [anon_sym_PLUS_EQ] = ACTIONS(202),
    [anon_sym_DASH_EQ] = ACTIONS(202),
    [anon_sym_STAR_EQ] = ACTIONS(202),
    [anon_sym_SLASH_EQ] = ACTIONS(202),
    [anon_sym_PERCENT_EQ] = ACTIONS(202),
    [anon_sym_CARET_EQ] = ACTIONS(202),
    [anon_sym_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_LT_LT_EQ] = ACTIONS(202),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(202),
    [anon_sym_QMARK_QMARK_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP] = ACTIONS(165),
    [anon_sym_PIPE_PIPE] = ACTIONS(165),
    [anon_sym_GT_GT] = ACTIONS(165),
    [anon_sym_GT_GT_GT] = ACTIONS(165),
    [anon_sym_LT_LT] = ACTIONS(165),
    [anon_sym_AMP] = ACTIONS(165),
    [anon_sym_CARET] = ACTIONS(165),
    [anon_sym_PIPE] = ACTIONS(165),
    [anon_sym_PLUS] = ACTIONS(534),
    [anon_sym_DASH] = ACTIONS(534),
    [anon_sym_SLASH] = ACTIONS(537),
    [anon_sym_PERCENT] = ACTIONS(165),
    [anon_sym_STAR_STAR] = ACTIONS(165),
    [anon_sym_LT_EQ] = ACTIONS(169),
    [anon_sym_EQ_EQ] = ACTIONS(165),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(169),
    [anon_sym_BANG_EQ] = ACTIONS(165),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(169),
    [anon_sym_GT_EQ] = ACTIONS(169),
    [anon_sym_QMARK_QMARK] = ACTIONS(165),
    [anon_sym_instanceof] = ACTIONS(165),
    [anon_sym_BANG] = ACTIONS(540),
    [anon_sym_TILDE] = ACTIONS(542),
    [anon_sym_typeof] = ACTIONS(540),
    [anon_sym_void] = ACTIONS(540),
    [anon_sym_delete] = ACTIONS(540),
    [anon_sym_PLUS_PLUS] = ACTIONS(544),
    [anon_sym_DASH_DASH] = ACTIONS(544),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(213),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(547),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(549),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(522),
    [anon_sym_get] = ACTIONS(522),
    [anon_sym_set] = ACTIONS(522),
    [sym__automatic_semicolon] = ACTIONS(169),
    [sym__ternary_qmark] = ACTIONS(169),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(40)] = {
    [sym_export_statement] = STATE(471),
    [sym_declaration] = STATE(471),
    [sym_import] = STATE(1326),
    [sym_import_statement] = STATE(471),
    [sym_statement] = STATE(1403),
    [sym_expression_statement] = STATE(471),
    [sym_variable_declaration] = STATE(419),
    [sym_lexical_declaration] = STATE(419),
    [sym_using_declaration] = STATE(419),
    [sym_statement_block] = STATE(471),
    [sym_if_statement] = STATE(471),
    [sym_switch_statement] = STATE(471),
    [sym_for_statement] = STATE(471),
    [sym_for_in_statement] = STATE(471),
    [sym_while_statement] = STATE(471),
    [sym_do_statement] = STATE(471),
    [sym_try_statement] = STATE(471),
    [sym_with_statement] = STATE(471),
    [sym_break_statement] = STATE(471),
    [sym_continue_statement] = STATE(471),
    [sym_debugger_statement] = STATE(471),
    [sym_return_statement] = STATE(471),
    [sym_throw_statement] = STATE(471),
    [sym_empty_statement] = STATE(471),
    [sym_labeled_statement] = STATE(471),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(823),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1790),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1790),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_class_declaration] = STATE(419),
    [sym_function_expression] = STATE(740),
    [sym_function_declaration] = STATE(419),
    [sym_generator_function] = STATE(740),
    [sym_generator_function_declaration] = STATE(419),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1790),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_sequence_expression] = STATE(1662),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [aux_sym_export_statement_repeat1] = STATE(1306),
    [sym_identifier] = ACTIONS(551),
    [anon_sym_export] = ACTIONS(553),
    [anon_sym_LBRACE] = ACTIONS(555),
    [anon_sym_import] = ACTIONS(17),
    [anon_sym_with] = ACTIONS(557),
    [anon_sym_var] = ACTIONS(21),
    [anon_sym_let] = ACTIONS(559),
    [anon_sym_const] = ACTIONS(25),
    [anon_sym_using] = ACTIONS(27),
    [anon_sym_await] = ACTIONS(561),
    [anon_sym_if] = ACTIONS(563),
    [anon_sym_switch] = ACTIONS(33),
    [anon_sym_for] = ACTIONS(565),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_SEMI] = ACTIONS(39),
    [anon_sym_while] = ACTIONS(567),
    [anon_sym_do] = ACTIONS(43),
    [anon_sym_try] = ACTIONS(45),
    [anon_sym_break] = ACTIONS(47),
    [anon_sym_continue] = ACTIONS(49),
    [anon_sym_debugger] = ACTIONS(51),
    [anon_sym_return] = ACTIONS(53),
    [anon_sym_throw] = ACTIONS(55),
    [anon_sym_yield] = ACTIONS(57),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(569),
    [anon_sym_async] = ACTIONS(571),
    [anon_sym_function] = ACTIONS(573),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(75),
    [anon_sym_DASH] = ACTIONS(75),
    [anon_sym_SLASH] = ACTIONS(77),
    [anon_sym_BANG] = ACTIONS(79),
    [anon_sym_TILDE] = ACTIONS(79),
    [anon_sym_typeof] = ACTIONS(75),
    [anon_sym_void] = ACTIONS(75),
    [anon_sym_delete] = ACTIONS(75),
    [anon_sym_PLUS_PLUS] = ACTIONS(81),
    [anon_sym_DASH_DASH] = ACTIONS(81),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(87),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(575),
    [anon_sym_get] = ACTIONS(575),
    [anon_sym_set] = ACTIONS(575),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(41)] = {
    [sym_export_statement] = STATE(471),
    [sym_declaration] = STATE(471),
    [sym_import] = STATE(1326),
    [sym_import_statement] = STATE(471),
    [sym_statement] = STATE(1755),
    [sym_expression_statement] = STATE(471),
    [sym_variable_declaration] = STATE(419),
    [sym_lexical_declaration] = STATE(419),
    [sym_using_declaration] = STATE(419),
    [sym_statement_block] = STATE(471),
    [sym_if_statement] = STATE(471),
    [sym_switch_statement] = STATE(471),
    [sym_for_statement] = STATE(471),
    [sym_for_in_statement] = STATE(471),
    [sym_while_statement] = STATE(471),
    [sym_do_statement] = STATE(471),
    [sym_try_statement] = STATE(471),
    [sym_with_statement] = STATE(471),
    [sym_break_statement] = STATE(471),
    [sym_continue_statement] = STATE(471),
    [sym_debugger_statement] = STATE(471),
    [sym_return_statement] = STATE(471),
    [sym_throw_statement] = STATE(471),
    [sym_empty_statement] = STATE(471),
    [sym_labeled_statement] = STATE(471),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(823),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1790),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1790),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_class_declaration] = STATE(419),
    [sym_function_expression] = STATE(740),
    [sym_function_declaration] = STATE(419),
    [sym_generator_function] = STATE(740),
    [sym_generator_function_declaration] = STATE(419),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1790),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_sequence_expression] = STATE(1662),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [aux_sym_export_statement_repeat1] = STATE(1306),
    [sym_identifier] = ACTIONS(551),
    [anon_sym_export] = ACTIONS(553),
    [anon_sym_LBRACE] = ACTIONS(555),
    [anon_sym_import] = ACTIONS(17),
    [anon_sym_with] = ACTIONS(557),
    [anon_sym_var] = ACTIONS(21),
    [anon_sym_let] = ACTIONS(559),
    [anon_sym_const] = ACTIONS(25),
    [anon_sym_using] = ACTIONS(27),
    [anon_sym_await] = ACTIONS(561),
    [anon_sym_if] = ACTIONS(563),
    [anon_sym_switch] = ACTIONS(33),
    [anon_sym_for] = ACTIONS(565),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_SEMI] = ACTIONS(39),
    [anon_sym_while] = ACTIONS(567),
    [anon_sym_do] = ACTIONS(43),
    [anon_sym_try] = ACTIONS(45),
    [anon_sym_break] = ACTIONS(47),
    [anon_sym_continue] = ACTIONS(49),
    [anon_sym_debugger] = ACTIONS(51),
    [anon_sym_return] = ACTIONS(53),
    [anon_sym_throw] = ACTIONS(55),
    [anon_sym_yield] = ACTIONS(57),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(569),
    [anon_sym_async] = ACTIONS(571),
    [anon_sym_function] = ACTIONS(573),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(75),
    [anon_sym_DASH] = ACTIONS(75),
    [anon_sym_SLASH] = ACTIONS(77),
    [anon_sym_BANG] = ACTIONS(79),
    [anon_sym_TILDE] = ACTIONS(79),
    [anon_sym_typeof] = ACTIONS(75),
    [anon_sym_void] = ACTIONS(75),
    [anon_sym_delete] = ACTIONS(75),
    [anon_sym_PLUS_PLUS] = ACTIONS(81),
    [anon_sym_DASH_DASH] = ACTIONS(81),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(87),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(575),
    [anon_sym_get] = ACTIONS(575),
    [anon_sym_set] = ACTIONS(575),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(42)] = {
    [sym_export_statement] = STATE(471),
    [sym_declaration] = STATE(471),
    [sym_import] = STATE(1326),
    [sym_import_statement] = STATE(471),
    [sym_statement] = STATE(424),
    [sym_expression_statement] = STATE(471),
    [sym_variable_declaration] = STATE(419),
    [sym_lexical_declaration] = STATE(419),
    [sym_using_declaration] = STATE(419),
    [sym_statement_block] = STATE(471),
    [sym_if_statement] = STATE(471),
    [sym_switch_statement] = STATE(471),
    [sym_for_statement] = STATE(471),
    [sym_for_in_statement] = STATE(471),
    [sym_while_statement] = STATE(471),
    [sym_do_statement] = STATE(471),
    [sym_try_statement] = STATE(471),
    [sym_with_statement] = STATE(471),
    [sym_break_statement] = STATE(471),
    [sym_continue_statement] = STATE(471),
    [sym_debugger_statement] = STATE(471),
    [sym_return_statement] = STATE(471),
    [sym_throw_statement] = STATE(471),
    [sym_empty_statement] = STATE(471),
    [sym_labeled_statement] = STATE(471),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(823),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1790),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1790),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_class_declaration] = STATE(419),
    [sym_function_expression] = STATE(740),
    [sym_function_declaration] = STATE(419),
    [sym_generator_function] = STATE(740),
    [sym_generator_function_declaration] = STATE(419),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1790),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_sequence_expression] = STATE(1662),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [aux_sym_export_statement_repeat1] = STATE(1267),
    [sym_identifier] = ACTIONS(9),
    [anon_sym_export] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(15),
    [anon_sym_import] = ACTIONS(17),
    [anon_sym_with] = ACTIONS(19),
    [anon_sym_var] = ACTIONS(21),
    [anon_sym_let] = ACTIONS(23),
    [anon_sym_const] = ACTIONS(25),
    [anon_sym_using] = ACTIONS(27),
    [anon_sym_await] = ACTIONS(29),
    [anon_sym_if] = ACTIONS(31),
    [anon_sym_switch] = ACTIONS(33),
    [anon_sym_for] = ACTIONS(35),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_SEMI] = ACTIONS(39),
    [anon_sym_while] = ACTIONS(41),
    [anon_sym_do] = ACTIONS(43),
    [anon_sym_try] = ACTIONS(45),
    [anon_sym_break] = ACTIONS(47),
    [anon_sym_continue] = ACTIONS(49),
    [anon_sym_debugger] = ACTIONS(51),
    [anon_sym_return] = ACTIONS(53),
    [anon_sym_throw] = ACTIONS(55),
    [anon_sym_yield] = ACTIONS(57),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(67),
    [anon_sym_async] = ACTIONS(69),
    [anon_sym_function] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(75),
    [anon_sym_DASH] = ACTIONS(75),
    [anon_sym_SLASH] = ACTIONS(77),
    [anon_sym_BANG] = ACTIONS(79),
    [anon_sym_TILDE] = ACTIONS(79),
    [anon_sym_typeof] = ACTIONS(75),
    [anon_sym_void] = ACTIONS(75),
    [anon_sym_delete] = ACTIONS(75),
    [anon_sym_PLUS_PLUS] = ACTIONS(81),
    [anon_sym_DASH_DASH] = ACTIONS(81),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(87),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(95),
    [anon_sym_get] = ACTIONS(95),
    [anon_sym_set] = ACTIONS(95),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(43)] = {
    [sym_export_statement] = STATE(471),
    [sym_declaration] = STATE(471),
    [sym_import] = STATE(1326),
    [sym_import_statement] = STATE(471),
    [sym_statement] = STATE(448),
    [sym_expression_statement] = STATE(471),
    [sym_variable_declaration] = STATE(419),
    [sym_lexical_declaration] = STATE(419),
    [sym_using_declaration] = STATE(419),
    [sym_statement_block] = STATE(471),
    [sym_if_statement] = STATE(471),
    [sym_switch_statement] = STATE(471),
    [sym_for_statement] = STATE(471),
    [sym_for_in_statement] = STATE(471),
    [sym_while_statement] = STATE(471),
    [sym_do_statement] = STATE(471),
    [sym_try_statement] = STATE(471),
    [sym_with_statement] = STATE(471),
    [sym_break_statement] = STATE(471),
    [sym_continue_statement] = STATE(471),
    [sym_debugger_statement] = STATE(471),
    [sym_return_statement] = STATE(471),
    [sym_throw_statement] = STATE(471),
    [sym_empty_statement] = STATE(471),
    [sym_labeled_statement] = STATE(471),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(823),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1790),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1790),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_class_declaration] = STATE(419),
    [sym_function_expression] = STATE(740),
    [sym_function_declaration] = STATE(419),
    [sym_generator_function] = STATE(740),
    [sym_generator_function_declaration] = STATE(419),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1790),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_sequence_expression] = STATE(1662),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [aux_sym_export_statement_repeat1] = STATE(1267),
    [sym_identifier] = ACTIONS(9),
    [anon_sym_export] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(15),
    [anon_sym_import] = ACTIONS(17),
    [anon_sym_with] = ACTIONS(19),
    [anon_sym_var] = ACTIONS(21),
    [anon_sym_let] = ACTIONS(23),
    [anon_sym_const] = ACTIONS(25),
    [anon_sym_using] = ACTIONS(27),
    [anon_sym_await] = ACTIONS(29),
    [anon_sym_if] = ACTIONS(31),
    [anon_sym_switch] = ACTIONS(33),
    [anon_sym_for] = ACTIONS(35),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_SEMI] = ACTIONS(39),
    [anon_sym_while] = ACTIONS(41),
    [anon_sym_do] = ACTIONS(43),
    [anon_sym_try] = ACTIONS(45),
    [anon_sym_break] = ACTIONS(47),
    [anon_sym_continue] = ACTIONS(49),
    [anon_sym_debugger] = ACTIONS(51),
    [anon_sym_return] = ACTIONS(53),
    [anon_sym_throw] = ACTIONS(55),
    [anon_sym_yield] = ACTIONS(57),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(67),
    [anon_sym_async] = ACTIONS(69),
    [anon_sym_function] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(75),
    [anon_sym_DASH] = ACTIONS(75),
    [anon_sym_SLASH] = ACTIONS(77),
    [anon_sym_BANG] = ACTIONS(79),
    [anon_sym_TILDE] = ACTIONS(79),
    [anon_sym_typeof] = ACTIONS(75),
    [anon_sym_void] = ACTIONS(75),
    [anon_sym_delete] = ACTIONS(75),
    [anon_sym_PLUS_PLUS] = ACTIONS(81),
    [anon_sym_DASH_DASH] = ACTIONS(81),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(87),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(95),
    [anon_sym_get] = ACTIONS(95),
    [anon_sym_set] = ACTIONS(95),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(44)] = {
    [sym_export_statement] = STATE(471),
    [sym_declaration] = STATE(471),
    [sym_import] = STATE(1326),
    [sym_import_statement] = STATE(471),
    [sym_statement] = STATE(463),
    [sym_expression_statement] = STATE(471),
    [sym_variable_declaration] = STATE(419),
    [sym_lexical_declaration] = STATE(419),
    [sym_using_declaration] = STATE(419),
    [sym_statement_block] = STATE(471),
    [sym_if_statement] = STATE(471),
    [sym_switch_statement] = STATE(471),
    [sym_for_statement] = STATE(471),
    [sym_for_in_statement] = STATE(471),
    [sym_while_statement] = STATE(471),
    [sym_do_statement] = STATE(471),
    [sym_try_statement] = STATE(471),
    [sym_with_statement] = STATE(471),
    [sym_break_statement] = STATE(471),
    [sym_continue_statement] = STATE(471),
    [sym_debugger_statement] = STATE(471),
    [sym_return_statement] = STATE(471),
    [sym_throw_statement] = STATE(471),
    [sym_empty_statement] = STATE(471),
    [sym_labeled_statement] = STATE(471),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(823),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1790),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1790),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_class_declaration] = STATE(419),
    [sym_function_expression] = STATE(740),
    [sym_function_declaration] = STATE(419),
    [sym_generator_function] = STATE(740),
    [sym_generator_function_declaration] = STATE(419),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1790),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_sequence_expression] = STATE(1662),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [aux_sym_export_statement_repeat1] = STATE(1267),
    [sym_identifier] = ACTIONS(9),
    [anon_sym_export] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(15),
    [anon_sym_import] = ACTIONS(17),
    [anon_sym_with] = ACTIONS(19),
    [anon_sym_var] = ACTIONS(21),
    [anon_sym_let] = ACTIONS(23),
    [anon_sym_const] = ACTIONS(25),
    [anon_sym_using] = ACTIONS(27),
    [anon_sym_await] = ACTIONS(29),
    [anon_sym_if] = ACTIONS(31),
    [anon_sym_switch] = ACTIONS(33),
    [anon_sym_for] = ACTIONS(35),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_SEMI] = ACTIONS(39),
    [anon_sym_while] = ACTIONS(41),
    [anon_sym_do] = ACTIONS(43),
    [anon_sym_try] = ACTIONS(45),
    [anon_sym_break] = ACTIONS(47),
    [anon_sym_continue] = ACTIONS(49),
    [anon_sym_debugger] = ACTIONS(51),
    [anon_sym_return] = ACTIONS(53),
    [anon_sym_throw] = ACTIONS(55),
    [anon_sym_yield] = ACTIONS(57),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(67),
    [anon_sym_async] = ACTIONS(69),
    [anon_sym_function] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(75),
    [anon_sym_DASH] = ACTIONS(75),
    [anon_sym_SLASH] = ACTIONS(77),
    [anon_sym_BANG] = ACTIONS(79),
    [anon_sym_TILDE] = ACTIONS(79),
    [anon_sym_typeof] = ACTIONS(75),
    [anon_sym_void] = ACTIONS(75),
    [anon_sym_delete] = ACTIONS(75),
    [anon_sym_PLUS_PLUS] = ACTIONS(81),
    [anon_sym_DASH_DASH] = ACTIONS(81),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(87),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(95),
    [anon_sym_get] = ACTIONS(95),
    [anon_sym_set] = ACTIONS(95),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(45)] = {
    [sym_export_statement] = STATE(471),
    [sym_declaration] = STATE(471),
    [sym_import] = STATE(1326),
    [sym_import_statement] = STATE(471),
    [sym_statement] = STATE(464),
    [sym_expression_statement] = STATE(471),
    [sym_variable_declaration] = STATE(419),
    [sym_lexical_declaration] = STATE(419),
    [sym_using_declaration] = STATE(419),
    [sym_statement_block] = STATE(471),
    [sym_if_statement] = STATE(471),
    [sym_switch_statement] = STATE(471),
    [sym_for_statement] = STATE(471),
    [sym_for_in_statement] = STATE(471),
    [sym_while_statement] = STATE(471),
    [sym_do_statement] = STATE(471),
    [sym_try_statement] = STATE(471),
    [sym_with_statement] = STATE(471),
    [sym_break_statement] = STATE(471),
    [sym_continue_statement] = STATE(471),
    [sym_debugger_statement] = STATE(471),
    [sym_return_statement] = STATE(471),
    [sym_throw_statement] = STATE(471),
    [sym_empty_statement] = STATE(471),
    [sym_labeled_statement] = STATE(471),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(823),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1790),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1790),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_class_declaration] = STATE(419),
    [sym_function_expression] = STATE(740),
    [sym_function_declaration] = STATE(419),
    [sym_generator_function] = STATE(740),
    [sym_generator_function_declaration] = STATE(419),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1790),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_sequence_expression] = STATE(1662),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [aux_sym_export_statement_repeat1] = STATE(1267),
    [sym_identifier] = ACTIONS(9),
    [anon_sym_export] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(15),
    [anon_sym_import] = ACTIONS(17),
    [anon_sym_with] = ACTIONS(19),
    [anon_sym_var] = ACTIONS(21),
    [anon_sym_let] = ACTIONS(23),
    [anon_sym_const] = ACTIONS(25),
    [anon_sym_using] = ACTIONS(27),
    [anon_sym_await] = ACTIONS(29),
    [anon_sym_if] = ACTIONS(31),
    [anon_sym_switch] = ACTIONS(33),
    [anon_sym_for] = ACTIONS(35),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_SEMI] = ACTIONS(39),
    [anon_sym_while] = ACTIONS(41),
    [anon_sym_do] = ACTIONS(43),
    [anon_sym_try] = ACTIONS(45),
    [anon_sym_break] = ACTIONS(47),
    [anon_sym_continue] = ACTIONS(49),
    [anon_sym_debugger] = ACTIONS(51),
    [anon_sym_return] = ACTIONS(53),
    [anon_sym_throw] = ACTIONS(55),
    [anon_sym_yield] = ACTIONS(57),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(67),
    [anon_sym_async] = ACTIONS(69),
    [anon_sym_function] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(75),
    [anon_sym_DASH] = ACTIONS(75),
    [anon_sym_SLASH] = ACTIONS(77),
    [anon_sym_BANG] = ACTIONS(79),
    [anon_sym_TILDE] = ACTIONS(79),
    [anon_sym_typeof] = ACTIONS(75),
    [anon_sym_void] = ACTIONS(75),
    [anon_sym_delete] = ACTIONS(75),
    [anon_sym_PLUS_PLUS] = ACTIONS(81),
    [anon_sym_DASH_DASH] = ACTIONS(81),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(87),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(95),
    [anon_sym_get] = ACTIONS(95),
    [anon_sym_set] = ACTIONS(95),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(46)] = {
    [sym_export_statement] = STATE(471),
    [sym_declaration] = STATE(471),
    [sym_import] = STATE(1326),
    [sym_import_statement] = STATE(471),
    [sym_statement] = STATE(466),
    [sym_expression_statement] = STATE(471),
    [sym_variable_declaration] = STATE(419),
    [sym_lexical_declaration] = STATE(419),
    [sym_using_declaration] = STATE(419),
    [sym_statement_block] = STATE(471),
    [sym_if_statement] = STATE(471),
    [sym_switch_statement] = STATE(471),
    [sym_for_statement] = STATE(471),
    [sym_for_in_statement] = STATE(471),
    [sym_while_statement] = STATE(471),
    [sym_do_statement] = STATE(471),
    [sym_try_statement] = STATE(471),
    [sym_with_statement] = STATE(471),
    [sym_break_statement] = STATE(471),
    [sym_continue_statement] = STATE(471),
    [sym_debugger_statement] = STATE(471),
    [sym_return_statement] = STATE(471),
    [sym_throw_statement] = STATE(471),
    [sym_empty_statement] = STATE(471),
    [sym_labeled_statement] = STATE(471),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(823),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1790),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1790),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_class_declaration] = STATE(419),
    [sym_function_expression] = STATE(740),
    [sym_function_declaration] = STATE(419),
    [sym_generator_function] = STATE(740),
    [sym_generator_function_declaration] = STATE(419),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1790),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_sequence_expression] = STATE(1662),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [aux_sym_export_statement_repeat1] = STATE(1267),
    [sym_identifier] = ACTIONS(9),
    [anon_sym_export] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(15),
    [anon_sym_import] = ACTIONS(17),
    [anon_sym_with] = ACTIONS(19),
    [anon_sym_var] = ACTIONS(21),
    [anon_sym_let] = ACTIONS(23),
    [anon_sym_const] = ACTIONS(25),
    [anon_sym_using] = ACTIONS(27),
    [anon_sym_await] = ACTIONS(29),
    [anon_sym_if] = ACTIONS(31),
    [anon_sym_switch] = ACTIONS(33),
    [anon_sym_for] = ACTIONS(35),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_SEMI] = ACTIONS(39),
    [anon_sym_while] = ACTIONS(41),
    [anon_sym_do] = ACTIONS(43),
    [anon_sym_try] = ACTIONS(45),
    [anon_sym_break] = ACTIONS(47),
    [anon_sym_continue] = ACTIONS(49),
    [anon_sym_debugger] = ACTIONS(51),
    [anon_sym_return] = ACTIONS(53),
    [anon_sym_throw] = ACTIONS(55),
    [anon_sym_yield] = ACTIONS(57),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(67),
    [anon_sym_async] = ACTIONS(69),
    [anon_sym_function] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(75),
    [anon_sym_DASH] = ACTIONS(75),
    [anon_sym_SLASH] = ACTIONS(77),
    [anon_sym_BANG] = ACTIONS(79),
    [anon_sym_TILDE] = ACTIONS(79),
    [anon_sym_typeof] = ACTIONS(75),
    [anon_sym_void] = ACTIONS(75),
    [anon_sym_delete] = ACTIONS(75),
    [anon_sym_PLUS_PLUS] = ACTIONS(81),
    [anon_sym_DASH_DASH] = ACTIONS(81),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(87),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(95),
    [anon_sym_get] = ACTIONS(95),
    [anon_sym_set] = ACTIONS(95),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(47)] = {
    [sym_export_statement] = STATE(471),
    [sym_declaration] = STATE(471),
    [sym_import] = STATE(1326),
    [sym_import_statement] = STATE(471),
    [sym_statement] = STATE(479),
    [sym_expression_statement] = STATE(471),
    [sym_variable_declaration] = STATE(419),
    [sym_lexical_declaration] = STATE(419),
    [sym_using_declaration] = STATE(419),
    [sym_statement_block] = STATE(471),
    [sym_if_statement] = STATE(471),
    [sym_switch_statement] = STATE(471),
    [sym_for_statement] = STATE(471),
    [sym_for_in_statement] = STATE(471),
    [sym_while_statement] = STATE(471),
    [sym_do_statement] = STATE(471),
    [sym_try_statement] = STATE(471),
    [sym_with_statement] = STATE(471),
    [sym_break_statement] = STATE(471),
    [sym_continue_statement] = STATE(471),
    [sym_debugger_statement] = STATE(471),
    [sym_return_statement] = STATE(471),
    [sym_throw_statement] = STATE(471),
    [sym_empty_statement] = STATE(471),
    [sym_labeled_statement] = STATE(471),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(823),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1790),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1790),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_class_declaration] = STATE(419),
    [sym_function_expression] = STATE(740),
    [sym_function_declaration] = STATE(419),
    [sym_generator_function] = STATE(740),
    [sym_generator_function_declaration] = STATE(419),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1790),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_sequence_expression] = STATE(1662),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [aux_sym_export_statement_repeat1] = STATE(1267),
    [sym_identifier] = ACTIONS(9),
    [anon_sym_export] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(15),
    [anon_sym_import] = ACTIONS(17),
    [anon_sym_with] = ACTIONS(19),
    [anon_sym_var] = ACTIONS(21),
    [anon_sym_let] = ACTIONS(23),
    [anon_sym_const] = ACTIONS(25),
    [anon_sym_using] = ACTIONS(27),
    [anon_sym_await] = ACTIONS(29),
    [anon_sym_if] = ACTIONS(31),
    [anon_sym_switch] = ACTIONS(33),
    [anon_sym_for] = ACTIONS(35),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_SEMI] = ACTIONS(39),
    [anon_sym_while] = ACTIONS(41),
    [anon_sym_do] = ACTIONS(43),
    [anon_sym_try] = ACTIONS(45),
    [anon_sym_break] = ACTIONS(47),
    [anon_sym_continue] = ACTIONS(49),
    [anon_sym_debugger] = ACTIONS(51),
    [anon_sym_return] = ACTIONS(53),
    [anon_sym_throw] = ACTIONS(55),
    [anon_sym_yield] = ACTIONS(57),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(67),
    [anon_sym_async] = ACTIONS(69),
    [anon_sym_function] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(75),
    [anon_sym_DASH] = ACTIONS(75),
    [anon_sym_SLASH] = ACTIONS(77),
    [anon_sym_BANG] = ACTIONS(79),
    [anon_sym_TILDE] = ACTIONS(79),
    [anon_sym_typeof] = ACTIONS(75),
    [anon_sym_void] = ACTIONS(75),
    [anon_sym_delete] = ACTIONS(75),
    [anon_sym_PLUS_PLUS] = ACTIONS(81),
    [anon_sym_DASH_DASH] = ACTIONS(81),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(87),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(95),
    [anon_sym_get] = ACTIONS(95),
    [anon_sym_set] = ACTIONS(95),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(48)] = {
    [sym_export_statement] = STATE(471),
    [sym_declaration] = STATE(471),
    [sym_import] = STATE(1326),
    [sym_import_statement] = STATE(471),
    [sym_statement] = STATE(480),
    [sym_expression_statement] = STATE(471),
    [sym_variable_declaration] = STATE(419),
    [sym_lexical_declaration] = STATE(419),
    [sym_using_declaration] = STATE(419),
    [sym_statement_block] = STATE(471),
    [sym_if_statement] = STATE(471),
    [sym_switch_statement] = STATE(471),
    [sym_for_statement] = STATE(471),
    [sym_for_in_statement] = STATE(471),
    [sym_while_statement] = STATE(471),
    [sym_do_statement] = STATE(471),
    [sym_try_statement] = STATE(471),
    [sym_with_statement] = STATE(471),
    [sym_break_statement] = STATE(471),
    [sym_continue_statement] = STATE(471),
    [sym_debugger_statement] = STATE(471),
    [sym_return_statement] = STATE(471),
    [sym_throw_statement] = STATE(471),
    [sym_empty_statement] = STATE(471),
    [sym_labeled_statement] = STATE(471),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(823),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1790),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1790),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_class_declaration] = STATE(419),
    [sym_function_expression] = STATE(740),
    [sym_function_declaration] = STATE(419),
    [sym_generator_function] = STATE(740),
    [sym_generator_function_declaration] = STATE(419),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1790),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_sequence_expression] = STATE(1662),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [aux_sym_export_statement_repeat1] = STATE(1267),
    [sym_identifier] = ACTIONS(9),
    [anon_sym_export] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(15),
    [anon_sym_import] = ACTIONS(17),
    [anon_sym_with] = ACTIONS(19),
    [anon_sym_var] = ACTIONS(21),
    [anon_sym_let] = ACTIONS(23),
    [anon_sym_const] = ACTIONS(25),
    [anon_sym_using] = ACTIONS(27),
    [anon_sym_await] = ACTIONS(29),
    [anon_sym_if] = ACTIONS(31),
    [anon_sym_switch] = ACTIONS(33),
    [anon_sym_for] = ACTIONS(35),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_SEMI] = ACTIONS(39),
    [anon_sym_while] = ACTIONS(41),
    [anon_sym_do] = ACTIONS(43),
    [anon_sym_try] = ACTIONS(45),
    [anon_sym_break] = ACTIONS(47),
    [anon_sym_continue] = ACTIONS(49),
    [anon_sym_debugger] = ACTIONS(51),
    [anon_sym_return] = ACTIONS(53),
    [anon_sym_throw] = ACTIONS(55),
    [anon_sym_yield] = ACTIONS(57),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(67),
    [anon_sym_async] = ACTIONS(69),
    [anon_sym_function] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(75),
    [anon_sym_DASH] = ACTIONS(75),
    [anon_sym_SLASH] = ACTIONS(77),
    [anon_sym_BANG] = ACTIONS(79),
    [anon_sym_TILDE] = ACTIONS(79),
    [anon_sym_typeof] = ACTIONS(75),
    [anon_sym_void] = ACTIONS(75),
    [anon_sym_delete] = ACTIONS(75),
    [anon_sym_PLUS_PLUS] = ACTIONS(81),
    [anon_sym_DASH_DASH] = ACTIONS(81),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(87),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(95),
    [anon_sym_get] = ACTIONS(95),
    [anon_sym_set] = ACTIONS(95),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(49)] = {
    [sym_export_statement] = STATE(471),
    [sym_declaration] = STATE(471),
    [sym_import] = STATE(1326),
    [sym_import_statement] = STATE(471),
    [sym_statement] = STATE(481),
    [sym_expression_statement] = STATE(471),
    [sym_variable_declaration] = STATE(419),
    [sym_lexical_declaration] = STATE(419),
    [sym_using_declaration] = STATE(419),
    [sym_statement_block] = STATE(471),
    [sym_if_statement] = STATE(471),
    [sym_switch_statement] = STATE(471),
    [sym_for_statement] = STATE(471),
    [sym_for_in_statement] = STATE(471),
    [sym_while_statement] = STATE(471),
    [sym_do_statement] = STATE(471),
    [sym_try_statement] = STATE(471),
    [sym_with_statement] = STATE(471),
    [sym_break_statement] = STATE(471),
    [sym_continue_statement] = STATE(471),
    [sym_debugger_statement] = STATE(471),
    [sym_return_statement] = STATE(471),
    [sym_throw_statement] = STATE(471),
    [sym_empty_statement] = STATE(471),
    [sym_labeled_statement] = STATE(471),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(823),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1790),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1790),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_class_declaration] = STATE(419),
    [sym_function_expression] = STATE(740),
    [sym_function_declaration] = STATE(419),
    [sym_generator_function] = STATE(740),
    [sym_generator_function_declaration] = STATE(419),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1790),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_sequence_expression] = STATE(1662),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [aux_sym_export_statement_repeat1] = STATE(1267),
    [sym_identifier] = ACTIONS(9),
    [anon_sym_export] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(15),
    [anon_sym_import] = ACTIONS(17),
    [anon_sym_with] = ACTIONS(19),
    [anon_sym_var] = ACTIONS(21),
    [anon_sym_let] = ACTIONS(23),
    [anon_sym_const] = ACTIONS(25),
    [anon_sym_using] = ACTIONS(27),
    [anon_sym_await] = ACTIONS(29),
    [anon_sym_if] = ACTIONS(31),
    [anon_sym_switch] = ACTIONS(33),
    [anon_sym_for] = ACTIONS(35),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_SEMI] = ACTIONS(39),
    [anon_sym_while] = ACTIONS(41),
    [anon_sym_do] = ACTIONS(43),
    [anon_sym_try] = ACTIONS(45),
    [anon_sym_break] = ACTIONS(47),
    [anon_sym_continue] = ACTIONS(49),
    [anon_sym_debugger] = ACTIONS(51),
    [anon_sym_return] = ACTIONS(53),
    [anon_sym_throw] = ACTIONS(55),
    [anon_sym_yield] = ACTIONS(57),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(67),
    [anon_sym_async] = ACTIONS(69),
    [anon_sym_function] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(75),
    [anon_sym_DASH] = ACTIONS(75),
    [anon_sym_SLASH] = ACTIONS(77),
    [anon_sym_BANG] = ACTIONS(79),
    [anon_sym_TILDE] = ACTIONS(79),
    [anon_sym_typeof] = ACTIONS(75),
    [anon_sym_void] = ACTIONS(75),
    [anon_sym_delete] = ACTIONS(75),
    [anon_sym_PLUS_PLUS] = ACTIONS(81),
    [anon_sym_DASH_DASH] = ACTIONS(81),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(87),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(95),
    [anon_sym_get] = ACTIONS(95),
    [anon_sym_set] = ACTIONS(95),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(50)] = {
    [sym_export_statement] = STATE(471),
    [sym_declaration] = STATE(471),
    [sym_import] = STATE(1326),
    [sym_import_statement] = STATE(471),
    [sym_statement] = STATE(483),
    [sym_expression_statement] = STATE(471),
    [sym_variable_declaration] = STATE(419),
    [sym_lexical_declaration] = STATE(419),
    [sym_using_declaration] = STATE(419),
    [sym_statement_block] = STATE(471),
    [sym_if_statement] = STATE(471),
    [sym_switch_statement] = STATE(471),
    [sym_for_statement] = STATE(471),
    [sym_for_in_statement] = STATE(471),
    [sym_while_statement] = STATE(471),
    [sym_do_statement] = STATE(471),
    [sym_try_statement] = STATE(471),
    [sym_with_statement] = STATE(471),
    [sym_break_statement] = STATE(471),
    [sym_continue_statement] = STATE(471),
    [sym_debugger_statement] = STATE(471),
    [sym_return_statement] = STATE(471),
    [sym_throw_statement] = STATE(471),
    [sym_empty_statement] = STATE(471),
    [sym_labeled_statement] = STATE(471),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(823),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1790),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1790),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_class_declaration] = STATE(419),
    [sym_function_expression] = STATE(740),
    [sym_function_declaration] = STATE(419),
    [sym_generator_function] = STATE(740),
    [sym_generator_function_declaration] = STATE(419),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1790),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_sequence_expression] = STATE(1662),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [aux_sym_export_statement_repeat1] = STATE(1267),
    [sym_identifier] = ACTIONS(9),
    [anon_sym_export] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(15),
    [anon_sym_import] = ACTIONS(17),
    [anon_sym_with] = ACTIONS(19),
    [anon_sym_var] = ACTIONS(21),
    [anon_sym_let] = ACTIONS(23),
    [anon_sym_const] = ACTIONS(25),
    [anon_sym_using] = ACTIONS(27),
    [anon_sym_await] = ACTIONS(29),
    [anon_sym_if] = ACTIONS(31),
    [anon_sym_switch] = ACTIONS(33),
    [anon_sym_for] = ACTIONS(35),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_SEMI] = ACTIONS(39),
    [anon_sym_while] = ACTIONS(41),
    [anon_sym_do] = ACTIONS(43),
    [anon_sym_try] = ACTIONS(45),
    [anon_sym_break] = ACTIONS(47),
    [anon_sym_continue] = ACTIONS(49),
    [anon_sym_debugger] = ACTIONS(51),
    [anon_sym_return] = ACTIONS(53),
    [anon_sym_throw] = ACTIONS(55),
    [anon_sym_yield] = ACTIONS(57),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(67),
    [anon_sym_async] = ACTIONS(69),
    [anon_sym_function] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(75),
    [anon_sym_DASH] = ACTIONS(75),
    [anon_sym_SLASH] = ACTIONS(77),
    [anon_sym_BANG] = ACTIONS(79),
    [anon_sym_TILDE] = ACTIONS(79),
    [anon_sym_typeof] = ACTIONS(75),
    [anon_sym_void] = ACTIONS(75),
    [anon_sym_delete] = ACTIONS(75),
    [anon_sym_PLUS_PLUS] = ACTIONS(81),
    [anon_sym_DASH_DASH] = ACTIONS(81),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(87),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(95),
    [anon_sym_get] = ACTIONS(95),
    [anon_sym_set] = ACTIONS(95),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(51)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(507),
    [sym_expression] = STATE(587),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1815),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1815),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(507),
    [sym_subscript_expression] = STATE(507),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1139),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1815),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1763),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(369),
    [anon_sym_export] = ACTIONS(371),
    [anon_sym_STAR] = ACTIONS(165),
    [anon_sym_LBRACE] = ACTIONS(373),
    [anon_sym_COMMA] = ACTIONS(577),
    [anon_sym_RBRACE] = ACTIONS(577),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(371),
    [anon_sym_await] = ACTIONS(377),
    [anon_sym_LPAREN] = ACTIONS(379),
    [anon_sym_in] = ACTIONS(165),
    [anon_sym_yield] = ACTIONS(382),
    [anon_sym_EQ] = ACTIONS(384),
    [anon_sym_LBRACK] = ACTIONS(386),
    [anon_sym_RBRACK] = ACTIONS(577),
    [anon_sym_LT] = ACTIONS(389),
    [anon_sym_GT] = ACTIONS(165),
    [anon_sym_DOT] = ACTIONS(165),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(398),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_EQ_GT] = ACTIONS(402),
    [sym_optional_chain] = ACTIONS(169),
    [anon_sym_new] = ACTIONS(404),
    [anon_sym_PLUS_EQ] = ACTIONS(202),
    [anon_sym_DASH_EQ] = ACTIONS(202),
    [anon_sym_STAR_EQ] = ACTIONS(202),
    [anon_sym_SLASH_EQ] = ACTIONS(202),
    [anon_sym_PERCENT_EQ] = ACTIONS(202),
    [anon_sym_CARET_EQ] = ACTIONS(202),
    [anon_sym_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_LT_LT_EQ] = ACTIONS(202),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(202),
    [anon_sym_QMARK_QMARK_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP] = ACTIONS(165),
    [anon_sym_PIPE_PIPE] = ACTIONS(165),
    [anon_sym_GT_GT] = ACTIONS(165),
    [anon_sym_GT_GT_GT] = ACTIONS(165),
    [anon_sym_LT_LT] = ACTIONS(165),
    [anon_sym_AMP] = ACTIONS(165),
    [anon_sym_CARET] = ACTIONS(165),
    [anon_sym_PIPE] = ACTIONS(165),
    [anon_sym_PLUS] = ACTIONS(406),
    [anon_sym_DASH] = ACTIONS(406),
    [anon_sym_SLASH] = ACTIONS(409),
    [anon_sym_PERCENT] = ACTIONS(165),
    [anon_sym_STAR_STAR] = ACTIONS(165),
    [anon_sym_LT_EQ] = ACTIONS(169),
    [anon_sym_EQ_EQ] = ACTIONS(165),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(169),
    [anon_sym_BANG_EQ] = ACTIONS(165),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(169),
    [anon_sym_GT_EQ] = ACTIONS(169),
    [anon_sym_QMARK_QMARK] = ACTIONS(165),
    [anon_sym_instanceof] = ACTIONS(165),
    [anon_sym_BANG] = ACTIONS(412),
    [anon_sym_TILDE] = ACTIONS(414),
    [anon_sym_typeof] = ACTIONS(412),
    [anon_sym_void] = ACTIONS(412),
    [anon_sym_delete] = ACTIONS(412),
    [anon_sym_PLUS_PLUS] = ACTIONS(416),
    [anon_sym_DASH_DASH] = ACTIONS(416),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(419),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(424),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(428),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(371),
    [anon_sym_get] = ACTIONS(371),
    [anon_sym_set] = ACTIONS(371),
    [sym__ternary_qmark] = ACTIONS(169),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(52)] = {
    [sym_export_statement] = STATE(471),
    [sym_declaration] = STATE(471),
    [sym_import] = STATE(1326),
    [sym_import_statement] = STATE(471),
    [sym_statement] = STATE(452),
    [sym_expression_statement] = STATE(471),
    [sym_variable_declaration] = STATE(419),
    [sym_lexical_declaration] = STATE(419),
    [sym_using_declaration] = STATE(419),
    [sym_statement_block] = STATE(471),
    [sym_if_statement] = STATE(471),
    [sym_switch_statement] = STATE(471),
    [sym_for_statement] = STATE(471),
    [sym_for_in_statement] = STATE(471),
    [sym_while_statement] = STATE(471),
    [sym_do_statement] = STATE(471),
    [sym_try_statement] = STATE(471),
    [sym_with_statement] = STATE(471),
    [sym_break_statement] = STATE(471),
    [sym_continue_statement] = STATE(471),
    [sym_debugger_statement] = STATE(471),
    [sym_return_statement] = STATE(471),
    [sym_throw_statement] = STATE(471),
    [sym_empty_statement] = STATE(471),
    [sym_labeled_statement] = STATE(471),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(823),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1790),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1790),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_class_declaration] = STATE(419),
    [sym_function_expression] = STATE(740),
    [sym_function_declaration] = STATE(419),
    [sym_generator_function] = STATE(740),
    [sym_generator_function_declaration] = STATE(419),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1790),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_sequence_expression] = STATE(1662),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [aux_sym_export_statement_repeat1] = STATE(1306),
    [sym_identifier] = ACTIONS(551),
    [anon_sym_export] = ACTIONS(553),
    [anon_sym_LBRACE] = ACTIONS(555),
    [anon_sym_import] = ACTIONS(17),
    [anon_sym_with] = ACTIONS(557),
    [anon_sym_var] = ACTIONS(21),
    [anon_sym_let] = ACTIONS(559),
    [anon_sym_const] = ACTIONS(25),
    [anon_sym_using] = ACTIONS(27),
    [anon_sym_await] = ACTIONS(561),
    [anon_sym_if] = ACTIONS(563),
    [anon_sym_switch] = ACTIONS(33),
    [anon_sym_for] = ACTIONS(565),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_SEMI] = ACTIONS(39),
    [anon_sym_while] = ACTIONS(567),
    [anon_sym_do] = ACTIONS(43),
    [anon_sym_try] = ACTIONS(45),
    [anon_sym_break] = ACTIONS(47),
    [anon_sym_continue] = ACTIONS(49),
    [anon_sym_debugger] = ACTIONS(51),
    [anon_sym_return] = ACTIONS(53),
    [anon_sym_throw] = ACTIONS(55),
    [anon_sym_yield] = ACTIONS(57),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(569),
    [anon_sym_async] = ACTIONS(571),
    [anon_sym_function] = ACTIONS(573),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(75),
    [anon_sym_DASH] = ACTIONS(75),
    [anon_sym_SLASH] = ACTIONS(77),
    [anon_sym_BANG] = ACTIONS(79),
    [anon_sym_TILDE] = ACTIONS(79),
    [anon_sym_typeof] = ACTIONS(75),
    [anon_sym_void] = ACTIONS(75),
    [anon_sym_delete] = ACTIONS(75),
    [anon_sym_PLUS_PLUS] = ACTIONS(81),
    [anon_sym_DASH_DASH] = ACTIONS(81),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(87),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(575),
    [anon_sym_get] = ACTIONS(575),
    [anon_sym_set] = ACTIONS(575),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(53)] = {
    [sym_export_statement] = STATE(471),
    [sym_declaration] = STATE(471),
    [sym_import] = STATE(1326),
    [sym_import_statement] = STATE(471),
    [sym_statement] = STATE(444),
    [sym_expression_statement] = STATE(471),
    [sym_variable_declaration] = STATE(419),
    [sym_lexical_declaration] = STATE(419),
    [sym_using_declaration] = STATE(419),
    [sym_statement_block] = STATE(471),
    [sym_if_statement] = STATE(471),
    [sym_switch_statement] = STATE(471),
    [sym_for_statement] = STATE(471),
    [sym_for_in_statement] = STATE(471),
    [sym_while_statement] = STATE(471),
    [sym_do_statement] = STATE(471),
    [sym_try_statement] = STATE(471),
    [sym_with_statement] = STATE(471),
    [sym_break_statement] = STATE(471),
    [sym_continue_statement] = STATE(471),
    [sym_debugger_statement] = STATE(471),
    [sym_return_statement] = STATE(471),
    [sym_throw_statement] = STATE(471),
    [sym_empty_statement] = STATE(471),
    [sym_labeled_statement] = STATE(471),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(823),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1790),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1790),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_class_declaration] = STATE(419),
    [sym_function_expression] = STATE(740),
    [sym_function_declaration] = STATE(419),
    [sym_generator_function] = STATE(740),
    [sym_generator_function_declaration] = STATE(419),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1790),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_sequence_expression] = STATE(1662),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [aux_sym_export_statement_repeat1] = STATE(1267),
    [sym_identifier] = ACTIONS(9),
    [anon_sym_export] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(15),
    [anon_sym_import] = ACTIONS(17),
    [anon_sym_with] = ACTIONS(19),
    [anon_sym_var] = ACTIONS(21),
    [anon_sym_let] = ACTIONS(23),
    [anon_sym_const] = ACTIONS(25),
    [anon_sym_using] = ACTIONS(27),
    [anon_sym_await] = ACTIONS(29),
    [anon_sym_if] = ACTIONS(31),
    [anon_sym_switch] = ACTIONS(33),
    [anon_sym_for] = ACTIONS(35),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_SEMI] = ACTIONS(39),
    [anon_sym_while] = ACTIONS(41),
    [anon_sym_do] = ACTIONS(43),
    [anon_sym_try] = ACTIONS(45),
    [anon_sym_break] = ACTIONS(47),
    [anon_sym_continue] = ACTIONS(49),
    [anon_sym_debugger] = ACTIONS(51),
    [anon_sym_return] = ACTIONS(53),
    [anon_sym_throw] = ACTIONS(55),
    [anon_sym_yield] = ACTIONS(57),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(67),
    [anon_sym_async] = ACTIONS(69),
    [anon_sym_function] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(75),
    [anon_sym_DASH] = ACTIONS(75),
    [anon_sym_SLASH] = ACTIONS(77),
    [anon_sym_BANG] = ACTIONS(79),
    [anon_sym_TILDE] = ACTIONS(79),
    [anon_sym_typeof] = ACTIONS(75),
    [anon_sym_void] = ACTIONS(75),
    [anon_sym_delete] = ACTIONS(75),
    [anon_sym_PLUS_PLUS] = ACTIONS(81),
    [anon_sym_DASH_DASH] = ACTIONS(81),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(87),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(95),
    [anon_sym_get] = ACTIONS(95),
    [anon_sym_set] = ACTIONS(95),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(54)] = {
    [sym_export_statement] = STATE(471),
    [sym_declaration] = STATE(471),
    [sym_import] = STATE(1326),
    [sym_import_statement] = STATE(471),
    [sym_statement] = STATE(473),
    [sym_expression_statement] = STATE(471),
    [sym_variable_declaration] = STATE(419),
    [sym_lexical_declaration] = STATE(419),
    [sym_using_declaration] = STATE(419),
    [sym_statement_block] = STATE(471),
    [sym_if_statement] = STATE(471),
    [sym_switch_statement] = STATE(471),
    [sym_for_statement] = STATE(471),
    [sym_for_in_statement] = STATE(471),
    [sym_while_statement] = STATE(471),
    [sym_do_statement] = STATE(471),
    [sym_try_statement] = STATE(471),
    [sym_with_statement] = STATE(471),
    [sym_break_statement] = STATE(471),
    [sym_continue_statement] = STATE(471),
    [sym_debugger_statement] = STATE(471),
    [sym_return_statement] = STATE(471),
    [sym_throw_statement] = STATE(471),
    [sym_empty_statement] = STATE(471),
    [sym_labeled_statement] = STATE(471),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(823),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1790),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1790),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_class_declaration] = STATE(419),
    [sym_function_expression] = STATE(740),
    [sym_function_declaration] = STATE(419),
    [sym_generator_function] = STATE(740),
    [sym_generator_function_declaration] = STATE(419),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1790),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_sequence_expression] = STATE(1662),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [aux_sym_export_statement_repeat1] = STATE(1306),
    [sym_identifier] = ACTIONS(551),
    [anon_sym_export] = ACTIONS(553),
    [anon_sym_LBRACE] = ACTIONS(555),
    [anon_sym_import] = ACTIONS(17),
    [anon_sym_with] = ACTIONS(557),
    [anon_sym_var] = ACTIONS(21),
    [anon_sym_let] = ACTIONS(559),
    [anon_sym_const] = ACTIONS(25),
    [anon_sym_using] = ACTIONS(27),
    [anon_sym_await] = ACTIONS(561),
    [anon_sym_if] = ACTIONS(563),
    [anon_sym_switch] = ACTIONS(33),
    [anon_sym_for] = ACTIONS(565),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_SEMI] = ACTIONS(39),
    [anon_sym_while] = ACTIONS(567),
    [anon_sym_do] = ACTIONS(43),
    [anon_sym_try] = ACTIONS(45),
    [anon_sym_break] = ACTIONS(47),
    [anon_sym_continue] = ACTIONS(49),
    [anon_sym_debugger] = ACTIONS(51),
    [anon_sym_return] = ACTIONS(53),
    [anon_sym_throw] = ACTIONS(55),
    [anon_sym_yield] = ACTIONS(57),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(569),
    [anon_sym_async] = ACTIONS(571),
    [anon_sym_function] = ACTIONS(573),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(75),
    [anon_sym_DASH] = ACTIONS(75),
    [anon_sym_SLASH] = ACTIONS(77),
    [anon_sym_BANG] = ACTIONS(79),
    [anon_sym_TILDE] = ACTIONS(79),
    [anon_sym_typeof] = ACTIONS(75),
    [anon_sym_void] = ACTIONS(75),
    [anon_sym_delete] = ACTIONS(75),
    [anon_sym_PLUS_PLUS] = ACTIONS(81),
    [anon_sym_DASH_DASH] = ACTIONS(81),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(87),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(575),
    [anon_sym_get] = ACTIONS(575),
    [anon_sym_set] = ACTIONS(575),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(55)] = {
    [sym_export_statement] = STATE(471),
    [sym_declaration] = STATE(471),
    [sym_import] = STATE(1326),
    [sym_import_statement] = STATE(471),
    [sym_statement] = STATE(429),
    [sym_expression_statement] = STATE(471),
    [sym_variable_declaration] = STATE(419),
    [sym_lexical_declaration] = STATE(419),
    [sym_using_declaration] = STATE(419),
    [sym_statement_block] = STATE(471),
    [sym_if_statement] = STATE(471),
    [sym_switch_statement] = STATE(471),
    [sym_for_statement] = STATE(471),
    [sym_for_in_statement] = STATE(471),
    [sym_while_statement] = STATE(471),
    [sym_do_statement] = STATE(471),
    [sym_try_statement] = STATE(471),
    [sym_with_statement] = STATE(471),
    [sym_break_statement] = STATE(471),
    [sym_continue_statement] = STATE(471),
    [sym_debugger_statement] = STATE(471),
    [sym_return_statement] = STATE(471),
    [sym_throw_statement] = STATE(471),
    [sym_empty_statement] = STATE(471),
    [sym_labeled_statement] = STATE(471),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(823),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1790),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1790),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_class_declaration] = STATE(419),
    [sym_function_expression] = STATE(740),
    [sym_function_declaration] = STATE(419),
    [sym_generator_function] = STATE(740),
    [sym_generator_function_declaration] = STATE(419),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1790),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_sequence_expression] = STATE(1662),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [aux_sym_export_statement_repeat1] = STATE(1306),
    [sym_identifier] = ACTIONS(551),
    [anon_sym_export] = ACTIONS(553),
    [anon_sym_LBRACE] = ACTIONS(555),
    [anon_sym_import] = ACTIONS(17),
    [anon_sym_with] = ACTIONS(557),
    [anon_sym_var] = ACTIONS(21),
    [anon_sym_let] = ACTIONS(559),
    [anon_sym_const] = ACTIONS(25),
    [anon_sym_using] = ACTIONS(27),
    [anon_sym_await] = ACTIONS(561),
    [anon_sym_if] = ACTIONS(563),
    [anon_sym_switch] = ACTIONS(33),
    [anon_sym_for] = ACTIONS(565),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_SEMI] = ACTIONS(39),
    [anon_sym_while] = ACTIONS(567),
    [anon_sym_do] = ACTIONS(43),
    [anon_sym_try] = ACTIONS(45),
    [anon_sym_break] = ACTIONS(47),
    [anon_sym_continue] = ACTIONS(49),
    [anon_sym_debugger] = ACTIONS(51),
    [anon_sym_return] = ACTIONS(53),
    [anon_sym_throw] = ACTIONS(55),
    [anon_sym_yield] = ACTIONS(57),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(569),
    [anon_sym_async] = ACTIONS(571),
    [anon_sym_function] = ACTIONS(573),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(75),
    [anon_sym_DASH] = ACTIONS(75),
    [anon_sym_SLASH] = ACTIONS(77),
    [anon_sym_BANG] = ACTIONS(79),
    [anon_sym_TILDE] = ACTIONS(79),
    [anon_sym_typeof] = ACTIONS(75),
    [anon_sym_void] = ACTIONS(75),
    [anon_sym_delete] = ACTIONS(75),
    [anon_sym_PLUS_PLUS] = ACTIONS(81),
    [anon_sym_DASH_DASH] = ACTIONS(81),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(87),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(575),
    [anon_sym_get] = ACTIONS(575),
    [anon_sym_set] = ACTIONS(575),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(56)] = {
    [sym_export_statement] = STATE(471),
    [sym_declaration] = STATE(471),
    [sym_import] = STATE(1326),
    [sym_import_statement] = STATE(471),
    [sym_statement] = STATE(432),
    [sym_expression_statement] = STATE(471),
    [sym_variable_declaration] = STATE(419),
    [sym_lexical_declaration] = STATE(419),
    [sym_using_declaration] = STATE(419),
    [sym_statement_block] = STATE(471),
    [sym_if_statement] = STATE(471),
    [sym_switch_statement] = STATE(471),
    [sym_for_statement] = STATE(471),
    [sym_for_in_statement] = STATE(471),
    [sym_while_statement] = STATE(471),
    [sym_do_statement] = STATE(471),
    [sym_try_statement] = STATE(471),
    [sym_with_statement] = STATE(471),
    [sym_break_statement] = STATE(471),
    [sym_continue_statement] = STATE(471),
    [sym_debugger_statement] = STATE(471),
    [sym_return_statement] = STATE(471),
    [sym_throw_statement] = STATE(471),
    [sym_empty_statement] = STATE(471),
    [sym_labeled_statement] = STATE(471),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(823),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1790),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1790),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_class_declaration] = STATE(419),
    [sym_function_expression] = STATE(740),
    [sym_function_declaration] = STATE(419),
    [sym_generator_function] = STATE(740),
    [sym_generator_function_declaration] = STATE(419),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1790),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_sequence_expression] = STATE(1662),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [aux_sym_export_statement_repeat1] = STATE(1306),
    [sym_identifier] = ACTIONS(551),
    [anon_sym_export] = ACTIONS(553),
    [anon_sym_LBRACE] = ACTIONS(555),
    [anon_sym_import] = ACTIONS(17),
    [anon_sym_with] = ACTIONS(557),
    [anon_sym_var] = ACTIONS(21),
    [anon_sym_let] = ACTIONS(559),
    [anon_sym_const] = ACTIONS(25),
    [anon_sym_using] = ACTIONS(27),
    [anon_sym_await] = ACTIONS(561),
    [anon_sym_if] = ACTIONS(563),
    [anon_sym_switch] = ACTIONS(33),
    [anon_sym_for] = ACTIONS(565),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_SEMI] = ACTIONS(39),
    [anon_sym_while] = ACTIONS(567),
    [anon_sym_do] = ACTIONS(43),
    [anon_sym_try] = ACTIONS(45),
    [anon_sym_break] = ACTIONS(47),
    [anon_sym_continue] = ACTIONS(49),
    [anon_sym_debugger] = ACTIONS(51),
    [anon_sym_return] = ACTIONS(53),
    [anon_sym_throw] = ACTIONS(55),
    [anon_sym_yield] = ACTIONS(57),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(569),
    [anon_sym_async] = ACTIONS(571),
    [anon_sym_function] = ACTIONS(573),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(75),
    [anon_sym_DASH] = ACTIONS(75),
    [anon_sym_SLASH] = ACTIONS(77),
    [anon_sym_BANG] = ACTIONS(79),
    [anon_sym_TILDE] = ACTIONS(79),
    [anon_sym_typeof] = ACTIONS(75),
    [anon_sym_void] = ACTIONS(75),
    [anon_sym_delete] = ACTIONS(75),
    [anon_sym_PLUS_PLUS] = ACTIONS(81),
    [anon_sym_DASH_DASH] = ACTIONS(81),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(87),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(575),
    [anon_sym_get] = ACTIONS(575),
    [anon_sym_set] = ACTIONS(575),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(57)] = {
    [sym_export_statement] = STATE(471),
    [sym_declaration] = STATE(471),
    [sym_import] = STATE(1326),
    [sym_import_statement] = STATE(471),
    [sym_statement] = STATE(411),
    [sym_expression_statement] = STATE(471),
    [sym_variable_declaration] = STATE(419),
    [sym_lexical_declaration] = STATE(419),
    [sym_using_declaration] = STATE(419),
    [sym_statement_block] = STATE(471),
    [sym_if_statement] = STATE(471),
    [sym_switch_statement] = STATE(471),
    [sym_for_statement] = STATE(471),
    [sym_for_in_statement] = STATE(471),
    [sym_while_statement] = STATE(471),
    [sym_do_statement] = STATE(471),
    [sym_try_statement] = STATE(471),
    [sym_with_statement] = STATE(471),
    [sym_break_statement] = STATE(471),
    [sym_continue_statement] = STATE(471),
    [sym_debugger_statement] = STATE(471),
    [sym_return_statement] = STATE(471),
    [sym_throw_statement] = STATE(471),
    [sym_empty_statement] = STATE(471),
    [sym_labeled_statement] = STATE(471),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(823),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1790),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1790),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_class_declaration] = STATE(419),
    [sym_function_expression] = STATE(740),
    [sym_function_declaration] = STATE(419),
    [sym_generator_function] = STATE(740),
    [sym_generator_function_declaration] = STATE(419),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1790),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_sequence_expression] = STATE(1662),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [aux_sym_export_statement_repeat1] = STATE(1267),
    [sym_identifier] = ACTIONS(9),
    [anon_sym_export] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(15),
    [anon_sym_import] = ACTIONS(17),
    [anon_sym_with] = ACTIONS(19),
    [anon_sym_var] = ACTIONS(21),
    [anon_sym_let] = ACTIONS(23),
    [anon_sym_const] = ACTIONS(25),
    [anon_sym_using] = ACTIONS(27),
    [anon_sym_await] = ACTIONS(29),
    [anon_sym_if] = ACTIONS(31),
    [anon_sym_switch] = ACTIONS(33),
    [anon_sym_for] = ACTIONS(35),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_SEMI] = ACTIONS(39),
    [anon_sym_while] = ACTIONS(41),
    [anon_sym_do] = ACTIONS(43),
    [anon_sym_try] = ACTIONS(45),
    [anon_sym_break] = ACTIONS(47),
    [anon_sym_continue] = ACTIONS(49),
    [anon_sym_debugger] = ACTIONS(51),
    [anon_sym_return] = ACTIONS(53),
    [anon_sym_throw] = ACTIONS(55),
    [anon_sym_yield] = ACTIONS(57),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(67),
    [anon_sym_async] = ACTIONS(69),
    [anon_sym_function] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(75),
    [anon_sym_DASH] = ACTIONS(75),
    [anon_sym_SLASH] = ACTIONS(77),
    [anon_sym_BANG] = ACTIONS(79),
    [anon_sym_TILDE] = ACTIONS(79),
    [anon_sym_typeof] = ACTIONS(75),
    [anon_sym_void] = ACTIONS(75),
    [anon_sym_delete] = ACTIONS(75),
    [anon_sym_PLUS_PLUS] = ACTIONS(81),
    [anon_sym_DASH_DASH] = ACTIONS(81),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(87),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(95),
    [anon_sym_get] = ACTIONS(95),
    [anon_sym_set] = ACTIONS(95),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(58)] = {
    [sym_export_statement] = STATE(471),
    [sym_declaration] = STATE(471),
    [sym_import] = STATE(1326),
    [sym_import_statement] = STATE(471),
    [sym_statement] = STATE(444),
    [sym_expression_statement] = STATE(471),
    [sym_variable_declaration] = STATE(419),
    [sym_lexical_declaration] = STATE(419),
    [sym_using_declaration] = STATE(419),
    [sym_statement_block] = STATE(471),
    [sym_if_statement] = STATE(471),
    [sym_switch_statement] = STATE(471),
    [sym_for_statement] = STATE(471),
    [sym_for_in_statement] = STATE(471),
    [sym_while_statement] = STATE(471),
    [sym_do_statement] = STATE(471),
    [sym_try_statement] = STATE(471),
    [sym_with_statement] = STATE(471),
    [sym_break_statement] = STATE(471),
    [sym_continue_statement] = STATE(471),
    [sym_debugger_statement] = STATE(471),
    [sym_return_statement] = STATE(471),
    [sym_throw_statement] = STATE(471),
    [sym_empty_statement] = STATE(471),
    [sym_labeled_statement] = STATE(471),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(823),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1790),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1790),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_class_declaration] = STATE(419),
    [sym_function_expression] = STATE(740),
    [sym_function_declaration] = STATE(419),
    [sym_generator_function] = STATE(740),
    [sym_generator_function_declaration] = STATE(419),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1790),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_sequence_expression] = STATE(1662),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [aux_sym_export_statement_repeat1] = STATE(1306),
    [sym_identifier] = ACTIONS(551),
    [anon_sym_export] = ACTIONS(553),
    [anon_sym_LBRACE] = ACTIONS(555),
    [anon_sym_import] = ACTIONS(17),
    [anon_sym_with] = ACTIONS(557),
    [anon_sym_var] = ACTIONS(21),
    [anon_sym_let] = ACTIONS(559),
    [anon_sym_const] = ACTIONS(25),
    [anon_sym_using] = ACTIONS(27),
    [anon_sym_await] = ACTIONS(561),
    [anon_sym_if] = ACTIONS(563),
    [anon_sym_switch] = ACTIONS(33),
    [anon_sym_for] = ACTIONS(565),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_SEMI] = ACTIONS(39),
    [anon_sym_while] = ACTIONS(567),
    [anon_sym_do] = ACTIONS(43),
    [anon_sym_try] = ACTIONS(45),
    [anon_sym_break] = ACTIONS(47),
    [anon_sym_continue] = ACTIONS(49),
    [anon_sym_debugger] = ACTIONS(51),
    [anon_sym_return] = ACTIONS(53),
    [anon_sym_throw] = ACTIONS(55),
    [anon_sym_yield] = ACTIONS(57),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(569),
    [anon_sym_async] = ACTIONS(571),
    [anon_sym_function] = ACTIONS(573),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(75),
    [anon_sym_DASH] = ACTIONS(75),
    [anon_sym_SLASH] = ACTIONS(77),
    [anon_sym_BANG] = ACTIONS(79),
    [anon_sym_TILDE] = ACTIONS(79),
    [anon_sym_typeof] = ACTIONS(75),
    [anon_sym_void] = ACTIONS(75),
    [anon_sym_delete] = ACTIONS(75),
    [anon_sym_PLUS_PLUS] = ACTIONS(81),
    [anon_sym_DASH_DASH] = ACTIONS(81),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(87),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(575),
    [anon_sym_get] = ACTIONS(575),
    [anon_sym_set] = ACTIONS(575),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(59)] = {
    [sym_export_statement] = STATE(471),
    [sym_declaration] = STATE(471),
    [sym_import] = STATE(1326),
    [sym_import_statement] = STATE(471),
    [sym_statement] = STATE(424),
    [sym_expression_statement] = STATE(471),
    [sym_variable_declaration] = STATE(419),
    [sym_lexical_declaration] = STATE(419),
    [sym_using_declaration] = STATE(419),
    [sym_statement_block] = STATE(471),
    [sym_if_statement] = STATE(471),
    [sym_switch_statement] = STATE(471),
    [sym_for_statement] = STATE(471),
    [sym_for_in_statement] = STATE(471),
    [sym_while_statement] = STATE(471),
    [sym_do_statement] = STATE(471),
    [sym_try_statement] = STATE(471),
    [sym_with_statement] = STATE(471),
    [sym_break_statement] = STATE(471),
    [sym_continue_statement] = STATE(471),
    [sym_debugger_statement] = STATE(471),
    [sym_return_statement] = STATE(471),
    [sym_throw_statement] = STATE(471),
    [sym_empty_statement] = STATE(471),
    [sym_labeled_statement] = STATE(471),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(823),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1790),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1790),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_class_declaration] = STATE(419),
    [sym_function_expression] = STATE(740),
    [sym_function_declaration] = STATE(419),
    [sym_generator_function] = STATE(740),
    [sym_generator_function_declaration] = STATE(419),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1790),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_sequence_expression] = STATE(1662),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [aux_sym_export_statement_repeat1] = STATE(1306),
    [sym_identifier] = ACTIONS(551),
    [anon_sym_export] = ACTIONS(553),
    [anon_sym_LBRACE] = ACTIONS(555),
    [anon_sym_import] = ACTIONS(17),
    [anon_sym_with] = ACTIONS(557),
    [anon_sym_var] = ACTIONS(21),
    [anon_sym_let] = ACTIONS(559),
    [anon_sym_const] = ACTIONS(25),
    [anon_sym_using] = ACTIONS(27),
    [anon_sym_await] = ACTIONS(561),
    [anon_sym_if] = ACTIONS(563),
    [anon_sym_switch] = ACTIONS(33),
    [anon_sym_for] = ACTIONS(565),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_SEMI] = ACTIONS(39),
    [anon_sym_while] = ACTIONS(567),
    [anon_sym_do] = ACTIONS(43),
    [anon_sym_try] = ACTIONS(45),
    [anon_sym_break] = ACTIONS(47),
    [anon_sym_continue] = ACTIONS(49),
    [anon_sym_debugger] = ACTIONS(51),
    [anon_sym_return] = ACTIONS(53),
    [anon_sym_throw] = ACTIONS(55),
    [anon_sym_yield] = ACTIONS(57),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(569),
    [anon_sym_async] = ACTIONS(571),
    [anon_sym_function] = ACTIONS(573),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(75),
    [anon_sym_DASH] = ACTIONS(75),
    [anon_sym_SLASH] = ACTIONS(77),
    [anon_sym_BANG] = ACTIONS(79),
    [anon_sym_TILDE] = ACTIONS(79),
    [anon_sym_typeof] = ACTIONS(75),
    [anon_sym_void] = ACTIONS(75),
    [anon_sym_delete] = ACTIONS(75),
    [anon_sym_PLUS_PLUS] = ACTIONS(81),
    [anon_sym_DASH_DASH] = ACTIONS(81),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(87),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(575),
    [anon_sym_get] = ACTIONS(575),
    [anon_sym_set] = ACTIONS(575),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(60)] = {
    [sym_export_statement] = STATE(471),
    [sym_declaration] = STATE(471),
    [sym_import] = STATE(1326),
    [sym_import_statement] = STATE(471),
    [sym_statement] = STATE(448),
    [sym_expression_statement] = STATE(471),
    [sym_variable_declaration] = STATE(419),
    [sym_lexical_declaration] = STATE(419),
    [sym_using_declaration] = STATE(419),
    [sym_statement_block] = STATE(471),
    [sym_if_statement] = STATE(471),
    [sym_switch_statement] = STATE(471),
    [sym_for_statement] = STATE(471),
    [sym_for_in_statement] = STATE(471),
    [sym_while_statement] = STATE(471),
    [sym_do_statement] = STATE(471),
    [sym_try_statement] = STATE(471),
    [sym_with_statement] = STATE(471),
    [sym_break_statement] = STATE(471),
    [sym_continue_statement] = STATE(471),
    [sym_debugger_statement] = STATE(471),
    [sym_return_statement] = STATE(471),
    [sym_throw_statement] = STATE(471),
    [sym_empty_statement] = STATE(471),
    [sym_labeled_statement] = STATE(471),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(823),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1790),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1790),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_class_declaration] = STATE(419),
    [sym_function_expression] = STATE(740),
    [sym_function_declaration] = STATE(419),
    [sym_generator_function] = STATE(740),
    [sym_generator_function_declaration] = STATE(419),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1790),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_sequence_expression] = STATE(1662),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [aux_sym_export_statement_repeat1] = STATE(1306),
    [sym_identifier] = ACTIONS(551),
    [anon_sym_export] = ACTIONS(553),
    [anon_sym_LBRACE] = ACTIONS(555),
    [anon_sym_import] = ACTIONS(17),
    [anon_sym_with] = ACTIONS(557),
    [anon_sym_var] = ACTIONS(21),
    [anon_sym_let] = ACTIONS(559),
    [anon_sym_const] = ACTIONS(25),
    [anon_sym_using] = ACTIONS(27),
    [anon_sym_await] = ACTIONS(561),
    [anon_sym_if] = ACTIONS(563),
    [anon_sym_switch] = ACTIONS(33),
    [anon_sym_for] = ACTIONS(565),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_SEMI] = ACTIONS(39),
    [anon_sym_while] = ACTIONS(567),
    [anon_sym_do] = ACTIONS(43),
    [anon_sym_try] = ACTIONS(45),
    [anon_sym_break] = ACTIONS(47),
    [anon_sym_continue] = ACTIONS(49),
    [anon_sym_debugger] = ACTIONS(51),
    [anon_sym_return] = ACTIONS(53),
    [anon_sym_throw] = ACTIONS(55),
    [anon_sym_yield] = ACTIONS(57),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(569),
    [anon_sym_async] = ACTIONS(571),
    [anon_sym_function] = ACTIONS(573),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(75),
    [anon_sym_DASH] = ACTIONS(75),
    [anon_sym_SLASH] = ACTIONS(77),
    [anon_sym_BANG] = ACTIONS(79),
    [anon_sym_TILDE] = ACTIONS(79),
    [anon_sym_typeof] = ACTIONS(75),
    [anon_sym_void] = ACTIONS(75),
    [anon_sym_delete] = ACTIONS(75),
    [anon_sym_PLUS_PLUS] = ACTIONS(81),
    [anon_sym_DASH_DASH] = ACTIONS(81),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(87),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(575),
    [anon_sym_get] = ACTIONS(575),
    [anon_sym_set] = ACTIONS(575),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(61)] = {
    [sym_export_statement] = STATE(471),
    [sym_declaration] = STATE(471),
    [sym_import] = STATE(1326),
    [sym_import_statement] = STATE(471),
    [sym_statement] = STATE(463),
    [sym_expression_statement] = STATE(471),
    [sym_variable_declaration] = STATE(419),
    [sym_lexical_declaration] = STATE(419),
    [sym_using_declaration] = STATE(419),
    [sym_statement_block] = STATE(471),
    [sym_if_statement] = STATE(471),
    [sym_switch_statement] = STATE(471),
    [sym_for_statement] = STATE(471),
    [sym_for_in_statement] = STATE(471),
    [sym_while_statement] = STATE(471),
    [sym_do_statement] = STATE(471),
    [sym_try_statement] = STATE(471),
    [sym_with_statement] = STATE(471),
    [sym_break_statement] = STATE(471),
    [sym_continue_statement] = STATE(471),
    [sym_debugger_statement] = STATE(471),
    [sym_return_statement] = STATE(471),
    [sym_throw_statement] = STATE(471),
    [sym_empty_statement] = STATE(471),
    [sym_labeled_statement] = STATE(471),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(823),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1790),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1790),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_class_declaration] = STATE(419),
    [sym_function_expression] = STATE(740),
    [sym_function_declaration] = STATE(419),
    [sym_generator_function] = STATE(740),
    [sym_generator_function_declaration] = STATE(419),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1790),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_sequence_expression] = STATE(1662),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [aux_sym_export_statement_repeat1] = STATE(1306),
    [sym_identifier] = ACTIONS(551),
    [anon_sym_export] = ACTIONS(553),
    [anon_sym_LBRACE] = ACTIONS(555),
    [anon_sym_import] = ACTIONS(17),
    [anon_sym_with] = ACTIONS(557),
    [anon_sym_var] = ACTIONS(21),
    [anon_sym_let] = ACTIONS(559),
    [anon_sym_const] = ACTIONS(25),
    [anon_sym_using] = ACTIONS(27),
    [anon_sym_await] = ACTIONS(561),
    [anon_sym_if] = ACTIONS(563),
    [anon_sym_switch] = ACTIONS(33),
    [anon_sym_for] = ACTIONS(565),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_SEMI] = ACTIONS(39),
    [anon_sym_while] = ACTIONS(567),
    [anon_sym_do] = ACTIONS(43),
    [anon_sym_try] = ACTIONS(45),
    [anon_sym_break] = ACTIONS(47),
    [anon_sym_continue] = ACTIONS(49),
    [anon_sym_debugger] = ACTIONS(51),
    [anon_sym_return] = ACTIONS(53),
    [anon_sym_throw] = ACTIONS(55),
    [anon_sym_yield] = ACTIONS(57),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(569),
    [anon_sym_async] = ACTIONS(571),
    [anon_sym_function] = ACTIONS(573),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(75),
    [anon_sym_DASH] = ACTIONS(75),
    [anon_sym_SLASH] = ACTIONS(77),
    [anon_sym_BANG] = ACTIONS(79),
    [anon_sym_TILDE] = ACTIONS(79),
    [anon_sym_typeof] = ACTIONS(75),
    [anon_sym_void] = ACTIONS(75),
    [anon_sym_delete] = ACTIONS(75),
    [anon_sym_PLUS_PLUS] = ACTIONS(81),
    [anon_sym_DASH_DASH] = ACTIONS(81),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(87),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(575),
    [anon_sym_get] = ACTIONS(575),
    [anon_sym_set] = ACTIONS(575),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(62)] = {
    [sym_export_statement] = STATE(471),
    [sym_declaration] = STATE(471),
    [sym_import] = STATE(1326),
    [sym_import_statement] = STATE(471),
    [sym_statement] = STATE(464),
    [sym_expression_statement] = STATE(471),
    [sym_variable_declaration] = STATE(419),
    [sym_lexical_declaration] = STATE(419),
    [sym_using_declaration] = STATE(419),
    [sym_statement_block] = STATE(471),
    [sym_if_statement] = STATE(471),
    [sym_switch_statement] = STATE(471),
    [sym_for_statement] = STATE(471),
    [sym_for_in_statement] = STATE(471),
    [sym_while_statement] = STATE(471),
    [sym_do_statement] = STATE(471),
    [sym_try_statement] = STATE(471),
    [sym_with_statement] = STATE(471),
    [sym_break_statement] = STATE(471),
    [sym_continue_statement] = STATE(471),
    [sym_debugger_statement] = STATE(471),
    [sym_return_statement] = STATE(471),
    [sym_throw_statement] = STATE(471),
    [sym_empty_statement] = STATE(471),
    [sym_labeled_statement] = STATE(471),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(823),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1790),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1790),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_class_declaration] = STATE(419),
    [sym_function_expression] = STATE(740),
    [sym_function_declaration] = STATE(419),
    [sym_generator_function] = STATE(740),
    [sym_generator_function_declaration] = STATE(419),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1790),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_sequence_expression] = STATE(1662),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [aux_sym_export_statement_repeat1] = STATE(1306),
    [sym_identifier] = ACTIONS(551),
    [anon_sym_export] = ACTIONS(553),
    [anon_sym_LBRACE] = ACTIONS(555),
    [anon_sym_import] = ACTIONS(17),
    [anon_sym_with] = ACTIONS(557),
    [anon_sym_var] = ACTIONS(21),
    [anon_sym_let] = ACTIONS(559),
    [anon_sym_const] = ACTIONS(25),
    [anon_sym_using] = ACTIONS(27),
    [anon_sym_await] = ACTIONS(561),
    [anon_sym_if] = ACTIONS(563),
    [anon_sym_switch] = ACTIONS(33),
    [anon_sym_for] = ACTIONS(565),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_SEMI] = ACTIONS(39),
    [anon_sym_while] = ACTIONS(567),
    [anon_sym_do] = ACTIONS(43),
    [anon_sym_try] = ACTIONS(45),
    [anon_sym_break] = ACTIONS(47),
    [anon_sym_continue] = ACTIONS(49),
    [anon_sym_debugger] = ACTIONS(51),
    [anon_sym_return] = ACTIONS(53),
    [anon_sym_throw] = ACTIONS(55),
    [anon_sym_yield] = ACTIONS(57),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(569),
    [anon_sym_async] = ACTIONS(571),
    [anon_sym_function] = ACTIONS(573),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(75),
    [anon_sym_DASH] = ACTIONS(75),
    [anon_sym_SLASH] = ACTIONS(77),
    [anon_sym_BANG] = ACTIONS(79),
    [anon_sym_TILDE] = ACTIONS(79),
    [anon_sym_typeof] = ACTIONS(75),
    [anon_sym_void] = ACTIONS(75),
    [anon_sym_delete] = ACTIONS(75),
    [anon_sym_PLUS_PLUS] = ACTIONS(81),
    [anon_sym_DASH_DASH] = ACTIONS(81),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(87),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(575),
    [anon_sym_get] = ACTIONS(575),
    [anon_sym_set] = ACTIONS(575),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(63)] = {
    [sym_export_statement] = STATE(471),
    [sym_declaration] = STATE(471),
    [sym_import] = STATE(1326),
    [sym_import_statement] = STATE(471),
    [sym_statement] = STATE(466),
    [sym_expression_statement] = STATE(471),
    [sym_variable_declaration] = STATE(419),
    [sym_lexical_declaration] = STATE(419),
    [sym_using_declaration] = STATE(419),
    [sym_statement_block] = STATE(471),
    [sym_if_statement] = STATE(471),
    [sym_switch_statement] = STATE(471),
    [sym_for_statement] = STATE(471),
    [sym_for_in_statement] = STATE(471),
    [sym_while_statement] = STATE(471),
    [sym_do_statement] = STATE(471),
    [sym_try_statement] = STATE(471),
    [sym_with_statement] = STATE(471),
    [sym_break_statement] = STATE(471),
    [sym_continue_statement] = STATE(471),
    [sym_debugger_statement] = STATE(471),
    [sym_return_statement] = STATE(471),
    [sym_throw_statement] = STATE(471),
    [sym_empty_statement] = STATE(471),
    [sym_labeled_statement] = STATE(471),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(823),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1790),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1790),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_class_declaration] = STATE(419),
    [sym_function_expression] = STATE(740),
    [sym_function_declaration] = STATE(419),
    [sym_generator_function] = STATE(740),
    [sym_generator_function_declaration] = STATE(419),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1790),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_sequence_expression] = STATE(1662),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [aux_sym_export_statement_repeat1] = STATE(1306),
    [sym_identifier] = ACTIONS(551),
    [anon_sym_export] = ACTIONS(553),
    [anon_sym_LBRACE] = ACTIONS(555),
    [anon_sym_import] = ACTIONS(17),
    [anon_sym_with] = ACTIONS(557),
    [anon_sym_var] = ACTIONS(21),
    [anon_sym_let] = ACTIONS(559),
    [anon_sym_const] = ACTIONS(25),
    [anon_sym_using] = ACTIONS(27),
    [anon_sym_await] = ACTIONS(561),
    [anon_sym_if] = ACTIONS(563),
    [anon_sym_switch] = ACTIONS(33),
    [anon_sym_for] = ACTIONS(565),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_SEMI] = ACTIONS(39),
    [anon_sym_while] = ACTIONS(567),
    [anon_sym_do] = ACTIONS(43),
    [anon_sym_try] = ACTIONS(45),
    [anon_sym_break] = ACTIONS(47),
    [anon_sym_continue] = ACTIONS(49),
    [anon_sym_debugger] = ACTIONS(51),
    [anon_sym_return] = ACTIONS(53),
    [anon_sym_throw] = ACTIONS(55),
    [anon_sym_yield] = ACTIONS(57),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(569),
    [anon_sym_async] = ACTIONS(571),
    [anon_sym_function] = ACTIONS(573),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(75),
    [anon_sym_DASH] = ACTIONS(75),
    [anon_sym_SLASH] = ACTIONS(77),
    [anon_sym_BANG] = ACTIONS(79),
    [anon_sym_TILDE] = ACTIONS(79),
    [anon_sym_typeof] = ACTIONS(75),
    [anon_sym_void] = ACTIONS(75),
    [anon_sym_delete] = ACTIONS(75),
    [anon_sym_PLUS_PLUS] = ACTIONS(81),
    [anon_sym_DASH_DASH] = ACTIONS(81),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(87),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(575),
    [anon_sym_get] = ACTIONS(575),
    [anon_sym_set] = ACTIONS(575),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(64)] = {
    [sym_export_statement] = STATE(471),
    [sym_declaration] = STATE(471),
    [sym_import] = STATE(1326),
    [sym_import_statement] = STATE(471),
    [sym_statement] = STATE(479),
    [sym_expression_statement] = STATE(471),
    [sym_variable_declaration] = STATE(419),
    [sym_lexical_declaration] = STATE(419),
    [sym_using_declaration] = STATE(419),
    [sym_statement_block] = STATE(471),
    [sym_if_statement] = STATE(471),
    [sym_switch_statement] = STATE(471),
    [sym_for_statement] = STATE(471),
    [sym_for_in_statement] = STATE(471),
    [sym_while_statement] = STATE(471),
    [sym_do_statement] = STATE(471),
    [sym_try_statement] = STATE(471),
    [sym_with_statement] = STATE(471),
    [sym_break_statement] = STATE(471),
    [sym_continue_statement] = STATE(471),
    [sym_debugger_statement] = STATE(471),
    [sym_return_statement] = STATE(471),
    [sym_throw_statement] = STATE(471),
    [sym_empty_statement] = STATE(471),
    [sym_labeled_statement] = STATE(471),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(823),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1790),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1790),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_class_declaration] = STATE(419),
    [sym_function_expression] = STATE(740),
    [sym_function_declaration] = STATE(419),
    [sym_generator_function] = STATE(740),
    [sym_generator_function_declaration] = STATE(419),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1790),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_sequence_expression] = STATE(1662),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [aux_sym_export_statement_repeat1] = STATE(1306),
    [sym_identifier] = ACTIONS(551),
    [anon_sym_export] = ACTIONS(553),
    [anon_sym_LBRACE] = ACTIONS(555),
    [anon_sym_import] = ACTIONS(17),
    [anon_sym_with] = ACTIONS(557),
    [anon_sym_var] = ACTIONS(21),
    [anon_sym_let] = ACTIONS(559),
    [anon_sym_const] = ACTIONS(25),
    [anon_sym_using] = ACTIONS(27),
    [anon_sym_await] = ACTIONS(561),
    [anon_sym_if] = ACTIONS(563),
    [anon_sym_switch] = ACTIONS(33),
    [anon_sym_for] = ACTIONS(565),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_SEMI] = ACTIONS(39),
    [anon_sym_while] = ACTIONS(567),
    [anon_sym_do] = ACTIONS(43),
    [anon_sym_try] = ACTIONS(45),
    [anon_sym_break] = ACTIONS(47),
    [anon_sym_continue] = ACTIONS(49),
    [anon_sym_debugger] = ACTIONS(51),
    [anon_sym_return] = ACTIONS(53),
    [anon_sym_throw] = ACTIONS(55),
    [anon_sym_yield] = ACTIONS(57),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(569),
    [anon_sym_async] = ACTIONS(571),
    [anon_sym_function] = ACTIONS(573),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(75),
    [anon_sym_DASH] = ACTIONS(75),
    [anon_sym_SLASH] = ACTIONS(77),
    [anon_sym_BANG] = ACTIONS(79),
    [anon_sym_TILDE] = ACTIONS(79),
    [anon_sym_typeof] = ACTIONS(75),
    [anon_sym_void] = ACTIONS(75),
    [anon_sym_delete] = ACTIONS(75),
    [anon_sym_PLUS_PLUS] = ACTIONS(81),
    [anon_sym_DASH_DASH] = ACTIONS(81),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(87),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(575),
    [anon_sym_get] = ACTIONS(575),
    [anon_sym_set] = ACTIONS(575),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(65)] = {
    [sym_export_statement] = STATE(471),
    [sym_declaration] = STATE(471),
    [sym_import] = STATE(1326),
    [sym_import_statement] = STATE(471),
    [sym_statement] = STATE(480),
    [sym_expression_statement] = STATE(471),
    [sym_variable_declaration] = STATE(419),
    [sym_lexical_declaration] = STATE(419),
    [sym_using_declaration] = STATE(419),
    [sym_statement_block] = STATE(471),
    [sym_if_statement] = STATE(471),
    [sym_switch_statement] = STATE(471),
    [sym_for_statement] = STATE(471),
    [sym_for_in_statement] = STATE(471),
    [sym_while_statement] = STATE(471),
    [sym_do_statement] = STATE(471),
    [sym_try_statement] = STATE(471),
    [sym_with_statement] = STATE(471),
    [sym_break_statement] = STATE(471),
    [sym_continue_statement] = STATE(471),
    [sym_debugger_statement] = STATE(471),
    [sym_return_statement] = STATE(471),
    [sym_throw_statement] = STATE(471),
    [sym_empty_statement] = STATE(471),
    [sym_labeled_statement] = STATE(471),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(823),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1790),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1790),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_class_declaration] = STATE(419),
    [sym_function_expression] = STATE(740),
    [sym_function_declaration] = STATE(419),
    [sym_generator_function] = STATE(740),
    [sym_generator_function_declaration] = STATE(419),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1790),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_sequence_expression] = STATE(1662),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [aux_sym_export_statement_repeat1] = STATE(1306),
    [sym_identifier] = ACTIONS(551),
    [anon_sym_export] = ACTIONS(553),
    [anon_sym_LBRACE] = ACTIONS(555),
    [anon_sym_import] = ACTIONS(17),
    [anon_sym_with] = ACTIONS(557),
    [anon_sym_var] = ACTIONS(21),
    [anon_sym_let] = ACTIONS(559),
    [anon_sym_const] = ACTIONS(25),
    [anon_sym_using] = ACTIONS(27),
    [anon_sym_await] = ACTIONS(561),
    [anon_sym_if] = ACTIONS(563),
    [anon_sym_switch] = ACTIONS(33),
    [anon_sym_for] = ACTIONS(565),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_SEMI] = ACTIONS(39),
    [anon_sym_while] = ACTIONS(567),
    [anon_sym_do] = ACTIONS(43),
    [anon_sym_try] = ACTIONS(45),
    [anon_sym_break] = ACTIONS(47),
    [anon_sym_continue] = ACTIONS(49),
    [anon_sym_debugger] = ACTIONS(51),
    [anon_sym_return] = ACTIONS(53),
    [anon_sym_throw] = ACTIONS(55),
    [anon_sym_yield] = ACTIONS(57),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(569),
    [anon_sym_async] = ACTIONS(571),
    [anon_sym_function] = ACTIONS(573),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(75),
    [anon_sym_DASH] = ACTIONS(75),
    [anon_sym_SLASH] = ACTIONS(77),
    [anon_sym_BANG] = ACTIONS(79),
    [anon_sym_TILDE] = ACTIONS(79),
    [anon_sym_typeof] = ACTIONS(75),
    [anon_sym_void] = ACTIONS(75),
    [anon_sym_delete] = ACTIONS(75),
    [anon_sym_PLUS_PLUS] = ACTIONS(81),
    [anon_sym_DASH_DASH] = ACTIONS(81),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(87),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(575),
    [anon_sym_get] = ACTIONS(575),
    [anon_sym_set] = ACTIONS(575),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(66)] = {
    [sym_export_statement] = STATE(471),
    [sym_declaration] = STATE(471),
    [sym_import] = STATE(1326),
    [sym_import_statement] = STATE(471),
    [sym_statement] = STATE(481),
    [sym_expression_statement] = STATE(471),
    [sym_variable_declaration] = STATE(419),
    [sym_lexical_declaration] = STATE(419),
    [sym_using_declaration] = STATE(419),
    [sym_statement_block] = STATE(471),
    [sym_if_statement] = STATE(471),
    [sym_switch_statement] = STATE(471),
    [sym_for_statement] = STATE(471),
    [sym_for_in_statement] = STATE(471),
    [sym_while_statement] = STATE(471),
    [sym_do_statement] = STATE(471),
    [sym_try_statement] = STATE(471),
    [sym_with_statement] = STATE(471),
    [sym_break_statement] = STATE(471),
    [sym_continue_statement] = STATE(471),
    [sym_debugger_statement] = STATE(471),
    [sym_return_statement] = STATE(471),
    [sym_throw_statement] = STATE(471),
    [sym_empty_statement] = STATE(471),
    [sym_labeled_statement] = STATE(471),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(823),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1790),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1790),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_class_declaration] = STATE(419),
    [sym_function_expression] = STATE(740),
    [sym_function_declaration] = STATE(419),
    [sym_generator_function] = STATE(740),
    [sym_generator_function_declaration] = STATE(419),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1790),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_sequence_expression] = STATE(1662),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [aux_sym_export_statement_repeat1] = STATE(1306),
    [sym_identifier] = ACTIONS(551),
    [anon_sym_export] = ACTIONS(553),
    [anon_sym_LBRACE] = ACTIONS(555),
    [anon_sym_import] = ACTIONS(17),
    [anon_sym_with] = ACTIONS(557),
    [anon_sym_var] = ACTIONS(21),
    [anon_sym_let] = ACTIONS(559),
    [anon_sym_const] = ACTIONS(25),
    [anon_sym_using] = ACTIONS(27),
    [anon_sym_await] = ACTIONS(561),
    [anon_sym_if] = ACTIONS(563),
    [anon_sym_switch] = ACTIONS(33),
    [anon_sym_for] = ACTIONS(565),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_SEMI] = ACTIONS(39),
    [anon_sym_while] = ACTIONS(567),
    [anon_sym_do] = ACTIONS(43),
    [anon_sym_try] = ACTIONS(45),
    [anon_sym_break] = ACTIONS(47),
    [anon_sym_continue] = ACTIONS(49),
    [anon_sym_debugger] = ACTIONS(51),
    [anon_sym_return] = ACTIONS(53),
    [anon_sym_throw] = ACTIONS(55),
    [anon_sym_yield] = ACTIONS(57),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(569),
    [anon_sym_async] = ACTIONS(571),
    [anon_sym_function] = ACTIONS(573),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(75),
    [anon_sym_DASH] = ACTIONS(75),
    [anon_sym_SLASH] = ACTIONS(77),
    [anon_sym_BANG] = ACTIONS(79),
    [anon_sym_TILDE] = ACTIONS(79),
    [anon_sym_typeof] = ACTIONS(75),
    [anon_sym_void] = ACTIONS(75),
    [anon_sym_delete] = ACTIONS(75),
    [anon_sym_PLUS_PLUS] = ACTIONS(81),
    [anon_sym_DASH_DASH] = ACTIONS(81),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(87),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(575),
    [anon_sym_get] = ACTIONS(575),
    [anon_sym_set] = ACTIONS(575),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(67)] = {
    [sym_export_statement] = STATE(471),
    [sym_declaration] = STATE(471),
    [sym_import] = STATE(1326),
    [sym_import_statement] = STATE(471),
    [sym_statement] = STATE(483),
    [sym_expression_statement] = STATE(471),
    [sym_variable_declaration] = STATE(419),
    [sym_lexical_declaration] = STATE(419),
    [sym_using_declaration] = STATE(419),
    [sym_statement_block] = STATE(471),
    [sym_if_statement] = STATE(471),
    [sym_switch_statement] = STATE(471),
    [sym_for_statement] = STATE(471),
    [sym_for_in_statement] = STATE(471),
    [sym_while_statement] = STATE(471),
    [sym_do_statement] = STATE(471),
    [sym_try_statement] = STATE(471),
    [sym_with_statement] = STATE(471),
    [sym_break_statement] = STATE(471),
    [sym_continue_statement] = STATE(471),
    [sym_debugger_statement] = STATE(471),
    [sym_return_statement] = STATE(471),
    [sym_throw_statement] = STATE(471),
    [sym_empty_statement] = STATE(471),
    [sym_labeled_statement] = STATE(471),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(823),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1790),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1790),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_class_declaration] = STATE(419),
    [sym_function_expression] = STATE(740),
    [sym_function_declaration] = STATE(419),
    [sym_generator_function] = STATE(740),
    [sym_generator_function_declaration] = STATE(419),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1790),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_sequence_expression] = STATE(1662),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [aux_sym_export_statement_repeat1] = STATE(1306),
    [sym_identifier] = ACTIONS(551),
    [anon_sym_export] = ACTIONS(553),
    [anon_sym_LBRACE] = ACTIONS(555),
    [anon_sym_import] = ACTIONS(17),
    [anon_sym_with] = ACTIONS(557),
    [anon_sym_var] = ACTIONS(21),
    [anon_sym_let] = ACTIONS(559),
    [anon_sym_const] = ACTIONS(25),
    [anon_sym_using] = ACTIONS(27),
    [anon_sym_await] = ACTIONS(561),
    [anon_sym_if] = ACTIONS(563),
    [anon_sym_switch] = ACTIONS(33),
    [anon_sym_for] = ACTIONS(565),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_SEMI] = ACTIONS(39),
    [anon_sym_while] = ACTIONS(567),
    [anon_sym_do] = ACTIONS(43),
    [anon_sym_try] = ACTIONS(45),
    [anon_sym_break] = ACTIONS(47),
    [anon_sym_continue] = ACTIONS(49),
    [anon_sym_debugger] = ACTIONS(51),
    [anon_sym_return] = ACTIONS(53),
    [anon_sym_throw] = ACTIONS(55),
    [anon_sym_yield] = ACTIONS(57),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(569),
    [anon_sym_async] = ACTIONS(571),
    [anon_sym_function] = ACTIONS(573),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(75),
    [anon_sym_DASH] = ACTIONS(75),
    [anon_sym_SLASH] = ACTIONS(77),
    [anon_sym_BANG] = ACTIONS(79),
    [anon_sym_TILDE] = ACTIONS(79),
    [anon_sym_typeof] = ACTIONS(75),
    [anon_sym_void] = ACTIONS(75),
    [anon_sym_delete] = ACTIONS(75),
    [anon_sym_PLUS_PLUS] = ACTIONS(81),
    [anon_sym_DASH_DASH] = ACTIONS(81),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(87),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(575),
    [anon_sym_get] = ACTIONS(575),
    [anon_sym_set] = ACTIONS(575),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(68)] = {
    [sym_export_statement] = STATE(471),
    [sym_declaration] = STATE(471),
    [sym_import] = STATE(1326),
    [sym_import_statement] = STATE(471),
    [sym_statement] = STATE(452),
    [sym_expression_statement] = STATE(471),
    [sym_variable_declaration] = STATE(419),
    [sym_lexical_declaration] = STATE(419),
    [sym_using_declaration] = STATE(419),
    [sym_statement_block] = STATE(471),
    [sym_if_statement] = STATE(471),
    [sym_switch_statement] = STATE(471),
    [sym_for_statement] = STATE(471),
    [sym_for_in_statement] = STATE(471),
    [sym_while_statement] = STATE(471),
    [sym_do_statement] = STATE(471),
    [sym_try_statement] = STATE(471),
    [sym_with_statement] = STATE(471),
    [sym_break_statement] = STATE(471),
    [sym_continue_statement] = STATE(471),
    [sym_debugger_statement] = STATE(471),
    [sym_return_statement] = STATE(471),
    [sym_throw_statement] = STATE(471),
    [sym_empty_statement] = STATE(471),
    [sym_labeled_statement] = STATE(471),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(823),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1790),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1790),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_class_declaration] = STATE(419),
    [sym_function_expression] = STATE(740),
    [sym_function_declaration] = STATE(419),
    [sym_generator_function] = STATE(740),
    [sym_generator_function_declaration] = STATE(419),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1790),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_sequence_expression] = STATE(1662),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [aux_sym_export_statement_repeat1] = STATE(1267),
    [sym_identifier] = ACTIONS(9),
    [anon_sym_export] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(15),
    [anon_sym_import] = ACTIONS(17),
    [anon_sym_with] = ACTIONS(19),
    [anon_sym_var] = ACTIONS(21),
    [anon_sym_let] = ACTIONS(23),
    [anon_sym_const] = ACTIONS(25),
    [anon_sym_using] = ACTIONS(27),
    [anon_sym_await] = ACTIONS(29),
    [anon_sym_if] = ACTIONS(31),
    [anon_sym_switch] = ACTIONS(33),
    [anon_sym_for] = ACTIONS(35),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_SEMI] = ACTIONS(39),
    [anon_sym_while] = ACTIONS(41),
    [anon_sym_do] = ACTIONS(43),
    [anon_sym_try] = ACTIONS(45),
    [anon_sym_break] = ACTIONS(47),
    [anon_sym_continue] = ACTIONS(49),
    [anon_sym_debugger] = ACTIONS(51),
    [anon_sym_return] = ACTIONS(53),
    [anon_sym_throw] = ACTIONS(55),
    [anon_sym_yield] = ACTIONS(57),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(67),
    [anon_sym_async] = ACTIONS(69),
    [anon_sym_function] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(75),
    [anon_sym_DASH] = ACTIONS(75),
    [anon_sym_SLASH] = ACTIONS(77),
    [anon_sym_BANG] = ACTIONS(79),
    [anon_sym_TILDE] = ACTIONS(79),
    [anon_sym_typeof] = ACTIONS(75),
    [anon_sym_void] = ACTIONS(75),
    [anon_sym_delete] = ACTIONS(75),
    [anon_sym_PLUS_PLUS] = ACTIONS(81),
    [anon_sym_DASH_DASH] = ACTIONS(81),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(87),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(95),
    [anon_sym_get] = ACTIONS(95),
    [anon_sym_set] = ACTIONS(95),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(69)] = {
    [sym_export_statement] = STATE(471),
    [sym_declaration] = STATE(471),
    [sym_import] = STATE(1326),
    [sym_import_statement] = STATE(471),
    [sym_statement] = STATE(429),
    [sym_expression_statement] = STATE(471),
    [sym_variable_declaration] = STATE(419),
    [sym_lexical_declaration] = STATE(419),
    [sym_using_declaration] = STATE(419),
    [sym_statement_block] = STATE(471),
    [sym_if_statement] = STATE(471),
    [sym_switch_statement] = STATE(471),
    [sym_for_statement] = STATE(471),
    [sym_for_in_statement] = STATE(471),
    [sym_while_statement] = STATE(471),
    [sym_do_statement] = STATE(471),
    [sym_try_statement] = STATE(471),
    [sym_with_statement] = STATE(471),
    [sym_break_statement] = STATE(471),
    [sym_continue_statement] = STATE(471),
    [sym_debugger_statement] = STATE(471),
    [sym_return_statement] = STATE(471),
    [sym_throw_statement] = STATE(471),
    [sym_empty_statement] = STATE(471),
    [sym_labeled_statement] = STATE(471),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(823),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1790),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1790),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_class_declaration] = STATE(419),
    [sym_function_expression] = STATE(740),
    [sym_function_declaration] = STATE(419),
    [sym_generator_function] = STATE(740),
    [sym_generator_function_declaration] = STATE(419),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1790),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_sequence_expression] = STATE(1662),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [aux_sym_export_statement_repeat1] = STATE(1267),
    [sym_identifier] = ACTIONS(9),
    [anon_sym_export] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(15),
    [anon_sym_import] = ACTIONS(17),
    [anon_sym_with] = ACTIONS(19),
    [anon_sym_var] = ACTIONS(21),
    [anon_sym_let] = ACTIONS(23),
    [anon_sym_const] = ACTIONS(25),
    [anon_sym_using] = ACTIONS(27),
    [anon_sym_await] = ACTIONS(29),
    [anon_sym_if] = ACTIONS(31),
    [anon_sym_switch] = ACTIONS(33),
    [anon_sym_for] = ACTIONS(35),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_SEMI] = ACTIONS(39),
    [anon_sym_while] = ACTIONS(41),
    [anon_sym_do] = ACTIONS(43),
    [anon_sym_try] = ACTIONS(45),
    [anon_sym_break] = ACTIONS(47),
    [anon_sym_continue] = ACTIONS(49),
    [anon_sym_debugger] = ACTIONS(51),
    [anon_sym_return] = ACTIONS(53),
    [anon_sym_throw] = ACTIONS(55),
    [anon_sym_yield] = ACTIONS(57),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(67),
    [anon_sym_async] = ACTIONS(69),
    [anon_sym_function] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(75),
    [anon_sym_DASH] = ACTIONS(75),
    [anon_sym_SLASH] = ACTIONS(77),
    [anon_sym_BANG] = ACTIONS(79),
    [anon_sym_TILDE] = ACTIONS(79),
    [anon_sym_typeof] = ACTIONS(75),
    [anon_sym_void] = ACTIONS(75),
    [anon_sym_delete] = ACTIONS(75),
    [anon_sym_PLUS_PLUS] = ACTIONS(81),
    [anon_sym_DASH_DASH] = ACTIONS(81),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(87),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(95),
    [anon_sym_get] = ACTIONS(95),
    [anon_sym_set] = ACTIONS(95),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(70)] = {
    [sym_export_statement] = STATE(471),
    [sym_declaration] = STATE(471),
    [sym_import] = STATE(1326),
    [sym_import_statement] = STATE(471),
    [sym_statement] = STATE(432),
    [sym_expression_statement] = STATE(471),
    [sym_variable_declaration] = STATE(419),
    [sym_lexical_declaration] = STATE(419),
    [sym_using_declaration] = STATE(419),
    [sym_statement_block] = STATE(471),
    [sym_if_statement] = STATE(471),
    [sym_switch_statement] = STATE(471),
    [sym_for_statement] = STATE(471),
    [sym_for_in_statement] = STATE(471),
    [sym_while_statement] = STATE(471),
    [sym_do_statement] = STATE(471),
    [sym_try_statement] = STATE(471),
    [sym_with_statement] = STATE(471),
    [sym_break_statement] = STATE(471),
    [sym_continue_statement] = STATE(471),
    [sym_debugger_statement] = STATE(471),
    [sym_return_statement] = STATE(471),
    [sym_throw_statement] = STATE(471),
    [sym_empty_statement] = STATE(471),
    [sym_labeled_statement] = STATE(471),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(823),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1790),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1790),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_class_declaration] = STATE(419),
    [sym_function_expression] = STATE(740),
    [sym_function_declaration] = STATE(419),
    [sym_generator_function] = STATE(740),
    [sym_generator_function_declaration] = STATE(419),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1790),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_sequence_expression] = STATE(1662),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [aux_sym_export_statement_repeat1] = STATE(1267),
    [sym_identifier] = ACTIONS(9),
    [anon_sym_export] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(15),
    [anon_sym_import] = ACTIONS(17),
    [anon_sym_with] = ACTIONS(19),
    [anon_sym_var] = ACTIONS(21),
    [anon_sym_let] = ACTIONS(23),
    [anon_sym_const] = ACTIONS(25),
    [anon_sym_using] = ACTIONS(27),
    [anon_sym_await] = ACTIONS(29),
    [anon_sym_if] = ACTIONS(31),
    [anon_sym_switch] = ACTIONS(33),
    [anon_sym_for] = ACTIONS(35),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_SEMI] = ACTIONS(39),
    [anon_sym_while] = ACTIONS(41),
    [anon_sym_do] = ACTIONS(43),
    [anon_sym_try] = ACTIONS(45),
    [anon_sym_break] = ACTIONS(47),
    [anon_sym_continue] = ACTIONS(49),
    [anon_sym_debugger] = ACTIONS(51),
    [anon_sym_return] = ACTIONS(53),
    [anon_sym_throw] = ACTIONS(55),
    [anon_sym_yield] = ACTIONS(57),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(67),
    [anon_sym_async] = ACTIONS(69),
    [anon_sym_function] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(75),
    [anon_sym_DASH] = ACTIONS(75),
    [anon_sym_SLASH] = ACTIONS(77),
    [anon_sym_BANG] = ACTIONS(79),
    [anon_sym_TILDE] = ACTIONS(79),
    [anon_sym_typeof] = ACTIONS(75),
    [anon_sym_void] = ACTIONS(75),
    [anon_sym_delete] = ACTIONS(75),
    [anon_sym_PLUS_PLUS] = ACTIONS(81),
    [anon_sym_DASH_DASH] = ACTIONS(81),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(87),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(95),
    [anon_sym_get] = ACTIONS(95),
    [anon_sym_set] = ACTIONS(95),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(71)] = {
    [sym_export_statement] = STATE(471),
    [sym_declaration] = STATE(471),
    [sym_import] = STATE(1326),
    [sym_import_statement] = STATE(471),
    [sym_statement] = STATE(473),
    [sym_expression_statement] = STATE(471),
    [sym_variable_declaration] = STATE(419),
    [sym_lexical_declaration] = STATE(419),
    [sym_using_declaration] = STATE(419),
    [sym_statement_block] = STATE(471),
    [sym_if_statement] = STATE(471),
    [sym_switch_statement] = STATE(471),
    [sym_for_statement] = STATE(471),
    [sym_for_in_statement] = STATE(471),
    [sym_while_statement] = STATE(471),
    [sym_do_statement] = STATE(471),
    [sym_try_statement] = STATE(471),
    [sym_with_statement] = STATE(471),
    [sym_break_statement] = STATE(471),
    [sym_continue_statement] = STATE(471),
    [sym_debugger_statement] = STATE(471),
    [sym_return_statement] = STATE(471),
    [sym_throw_statement] = STATE(471),
    [sym_empty_statement] = STATE(471),
    [sym_labeled_statement] = STATE(471),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(823),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1790),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1790),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_class_declaration] = STATE(419),
    [sym_function_expression] = STATE(740),
    [sym_function_declaration] = STATE(419),
    [sym_generator_function] = STATE(740),
    [sym_generator_function_declaration] = STATE(419),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1790),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_sequence_expression] = STATE(1662),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [aux_sym_export_statement_repeat1] = STATE(1267),
    [sym_identifier] = ACTIONS(9),
    [anon_sym_export] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(15),
    [anon_sym_import] = ACTIONS(17),
    [anon_sym_with] = ACTIONS(19),
    [anon_sym_var] = ACTIONS(21),
    [anon_sym_let] = ACTIONS(23),
    [anon_sym_const] = ACTIONS(25),
    [anon_sym_using] = ACTIONS(27),
    [anon_sym_await] = ACTIONS(29),
    [anon_sym_if] = ACTIONS(31),
    [anon_sym_switch] = ACTIONS(33),
    [anon_sym_for] = ACTIONS(35),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_SEMI] = ACTIONS(39),
    [anon_sym_while] = ACTIONS(41),
    [anon_sym_do] = ACTIONS(43),
    [anon_sym_try] = ACTIONS(45),
    [anon_sym_break] = ACTIONS(47),
    [anon_sym_continue] = ACTIONS(49),
    [anon_sym_debugger] = ACTIONS(51),
    [anon_sym_return] = ACTIONS(53),
    [anon_sym_throw] = ACTIONS(55),
    [anon_sym_yield] = ACTIONS(57),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(67),
    [anon_sym_async] = ACTIONS(69),
    [anon_sym_function] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(75),
    [anon_sym_DASH] = ACTIONS(75),
    [anon_sym_SLASH] = ACTIONS(77),
    [anon_sym_BANG] = ACTIONS(79),
    [anon_sym_TILDE] = ACTIONS(79),
    [anon_sym_typeof] = ACTIONS(75),
    [anon_sym_void] = ACTIONS(75),
    [anon_sym_delete] = ACTIONS(75),
    [anon_sym_PLUS_PLUS] = ACTIONS(81),
    [anon_sym_DASH_DASH] = ACTIONS(81),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(87),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(95),
    [anon_sym_get] = ACTIONS(95),
    [anon_sym_set] = ACTIONS(95),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(72)] = {
    [sym_import] = STATE(1326),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(692),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1790),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1790),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_function_expression] = STATE(740),
    [sym_generator_function] = STATE(740),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1790),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [aux_sym_export_statement_repeat1] = STATE(1381),
    [sym_identifier] = ACTIONS(161),
    [anon_sym_export] = ACTIONS(163),
    [anon_sym_STAR] = ACTIONS(165),
    [anon_sym_LBRACE] = ACTIONS(167),
    [anon_sym_import] = ACTIONS(173),
    [anon_sym_let] = ACTIONS(163),
    [anon_sym_using] = ACTIONS(175),
    [anon_sym_await] = ACTIONS(177),
    [anon_sym_LPAREN] = ACTIONS(485),
    [anon_sym_SEMI] = ACTIONS(169),
    [anon_sym_in] = ACTIONS(165),
    [anon_sym_yield] = ACTIONS(57),
    [anon_sym_EQ] = ACTIONS(490),
    [anon_sym_LBRACK] = ACTIONS(188),
    [anon_sym_LT] = ACTIONS(191),
    [anon_sym_GT] = ACTIONS(165),
    [anon_sym_DOT] = ACTIONS(165),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(194),
    [anon_sym_async] = ACTIONS(196),
    [anon_sym_function] = ACTIONS(198),
    [anon_sym_EQ_GT] = ACTIONS(200),
    [sym_optional_chain] = ACTIONS(169),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_PLUS_EQ] = ACTIONS(202),
    [anon_sym_DASH_EQ] = ACTIONS(202),
    [anon_sym_STAR_EQ] = ACTIONS(202),
    [anon_sym_SLASH_EQ] = ACTIONS(202),
    [anon_sym_PERCENT_EQ] = ACTIONS(202),
    [anon_sym_CARET_EQ] = ACTIONS(202),
    [anon_sym_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_LT_LT_EQ] = ACTIONS(202),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(202),
    [anon_sym_QMARK_QMARK_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP] = ACTIONS(165),
    [anon_sym_PIPE_PIPE] = ACTIONS(165),
    [anon_sym_GT_GT] = ACTIONS(165),
    [anon_sym_GT_GT_GT] = ACTIONS(165),
    [anon_sym_LT_LT] = ACTIONS(165),
    [anon_sym_AMP] = ACTIONS(165),
    [anon_sym_CARET] = ACTIONS(165),
    [anon_sym_PIPE] = ACTIONS(165),
    [anon_sym_PLUS] = ACTIONS(204),
    [anon_sym_DASH] = ACTIONS(204),
    [anon_sym_SLASH] = ACTIONS(207),
    [anon_sym_PERCENT] = ACTIONS(165),
    [anon_sym_STAR_STAR] = ACTIONS(165),
    [anon_sym_LT_EQ] = ACTIONS(169),
    [anon_sym_EQ_EQ] = ACTIONS(165),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(169),
    [anon_sym_BANG_EQ] = ACTIONS(165),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(169),
    [anon_sym_GT_EQ] = ACTIONS(169),
    [anon_sym_QMARK_QMARK] = ACTIONS(165),
    [anon_sym_instanceof] = ACTIONS(165),
    [anon_sym_BANG] = ACTIONS(75),
    [anon_sym_TILDE] = ACTIONS(79),
    [anon_sym_typeof] = ACTIONS(75),
    [anon_sym_void] = ACTIONS(75),
    [anon_sym_delete] = ACTIONS(75),
    [anon_sym_PLUS_PLUS] = ACTIONS(210),
    [anon_sym_DASH_DASH] = ACTIONS(210),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(213),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(87),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(163),
    [anon_sym_get] = ACTIONS(163),
    [anon_sym_set] = ACTIONS(163),
    [sym__automatic_semicolon] = ACTIONS(169),
    [sym__ternary_qmark] = ACTIONS(169),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(73)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(498),
    [sym_expression] = STATE(587),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1762),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1762),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(498),
    [sym_subscript_expression] = STATE(498),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1138),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1762),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1816),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(430),
    [anon_sym_export] = ACTIONS(432),
    [anon_sym_STAR] = ACTIONS(165),
    [anon_sym_LBRACE] = ACTIONS(434),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(432),
    [anon_sym_using] = ACTIONS(502),
    [anon_sym_await] = ACTIONS(436),
    [anon_sym_of] = ACTIONS(504),
    [anon_sym_LPAREN] = ACTIONS(379),
    [anon_sym_in] = ACTIONS(506),
    [anon_sym_yield] = ACTIONS(438),
    [anon_sym_EQ] = ACTIONS(440),
    [anon_sym_LBRACK] = ACTIONS(442),
    [anon_sym_LT] = ACTIONS(389),
    [anon_sym_GT] = ACTIONS(165),
    [anon_sym_DOT] = ACTIONS(165),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(445),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_EQ_GT] = ACTIONS(498),
    [sym_optional_chain] = ACTIONS(169),
    [anon_sym_new] = ACTIONS(447),
    [anon_sym_PLUS_EQ] = ACTIONS(202),
    [anon_sym_DASH_EQ] = ACTIONS(202),
    [anon_sym_STAR_EQ] = ACTIONS(202),
    [anon_sym_SLASH_EQ] = ACTIONS(202),
    [anon_sym_PERCENT_EQ] = ACTIONS(202),
    [anon_sym_CARET_EQ] = ACTIONS(202),
    [anon_sym_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_LT_LT_EQ] = ACTIONS(202),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(202),
    [anon_sym_QMARK_QMARK_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP] = ACTIONS(165),
    [anon_sym_PIPE_PIPE] = ACTIONS(165),
    [anon_sym_GT_GT] = ACTIONS(165),
    [anon_sym_GT_GT_GT] = ACTIONS(165),
    [anon_sym_LT_LT] = ACTIONS(165),
    [anon_sym_AMP] = ACTIONS(165),
    [anon_sym_CARET] = ACTIONS(165),
    [anon_sym_PIPE] = ACTIONS(165),
    [anon_sym_PLUS] = ACTIONS(449),
    [anon_sym_DASH] = ACTIONS(449),
    [anon_sym_SLASH] = ACTIONS(409),
    [anon_sym_PERCENT] = ACTIONS(165),
    [anon_sym_STAR_STAR] = ACTIONS(165),
    [anon_sym_LT_EQ] = ACTIONS(169),
    [anon_sym_EQ_EQ] = ACTIONS(165),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(169),
    [anon_sym_BANG_EQ] = ACTIONS(165),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(169),
    [anon_sym_GT_EQ] = ACTIONS(169),
    [anon_sym_QMARK_QMARK] = ACTIONS(165),
    [anon_sym_instanceof] = ACTIONS(165),
    [anon_sym_BANG] = ACTIONS(452),
    [anon_sym_TILDE] = ACTIONS(454),
    [anon_sym_typeof] = ACTIONS(452),
    [anon_sym_void] = ACTIONS(452),
    [anon_sym_delete] = ACTIONS(452),
    [anon_sym_PLUS_PLUS] = ACTIONS(456),
    [anon_sym_DASH_DASH] = ACTIONS(456),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(419),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(459),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(461),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(432),
    [anon_sym_get] = ACTIONS(432),
    [anon_sym_set] = ACTIONS(432),
    [sym__ternary_qmark] = ACTIONS(169),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(74)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(507),
    [sym_expression] = STATE(587),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1815),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1815),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(507),
    [sym_subscript_expression] = STATE(507),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1139),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1815),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1763),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(369),
    [anon_sym_export] = ACTIONS(371),
    [anon_sym_STAR] = ACTIONS(165),
    [anon_sym_LBRACE] = ACTIONS(373),
    [anon_sym_COMMA] = ACTIONS(509),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(371),
    [anon_sym_await] = ACTIONS(377),
    [anon_sym_LPAREN] = ACTIONS(379),
    [anon_sym_in] = ACTIONS(165),
    [anon_sym_yield] = ACTIONS(382),
    [anon_sym_EQ] = ACTIONS(517),
    [anon_sym_LBRACK] = ACTIONS(386),
    [anon_sym_RBRACK] = ACTIONS(514),
    [anon_sym_LT] = ACTIONS(389),
    [anon_sym_GT] = ACTIONS(165),
    [anon_sym_DOT] = ACTIONS(165),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(398),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_EQ_GT] = ACTIONS(402),
    [sym_optional_chain] = ACTIONS(169),
    [anon_sym_new] = ACTIONS(404),
    [anon_sym_PLUS_EQ] = ACTIONS(202),
    [anon_sym_DASH_EQ] = ACTIONS(202),
    [anon_sym_STAR_EQ] = ACTIONS(202),
    [anon_sym_SLASH_EQ] = ACTIONS(202),
    [anon_sym_PERCENT_EQ] = ACTIONS(202),
    [anon_sym_CARET_EQ] = ACTIONS(202),
    [anon_sym_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_LT_LT_EQ] = ACTIONS(202),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(202),
    [anon_sym_QMARK_QMARK_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP] = ACTIONS(165),
    [anon_sym_PIPE_PIPE] = ACTIONS(165),
    [anon_sym_GT_GT] = ACTIONS(165),
    [anon_sym_GT_GT_GT] = ACTIONS(165),
    [anon_sym_LT_LT] = ACTIONS(165),
    [anon_sym_AMP] = ACTIONS(165),
    [anon_sym_CARET] = ACTIONS(165),
    [anon_sym_PIPE] = ACTIONS(165),
    [anon_sym_PLUS] = ACTIONS(406),
    [anon_sym_DASH] = ACTIONS(406),
    [anon_sym_SLASH] = ACTIONS(409),
    [anon_sym_PERCENT] = ACTIONS(165),
    [anon_sym_STAR_STAR] = ACTIONS(165),
    [anon_sym_LT_EQ] = ACTIONS(169),
    [anon_sym_EQ_EQ] = ACTIONS(165),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(169),
    [anon_sym_BANG_EQ] = ACTIONS(165),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(169),
    [anon_sym_GT_EQ] = ACTIONS(169),
    [anon_sym_QMARK_QMARK] = ACTIONS(165),
    [anon_sym_instanceof] = ACTIONS(165),
    [anon_sym_BANG] = ACTIONS(412),
    [anon_sym_TILDE] = ACTIONS(414),
    [anon_sym_typeof] = ACTIONS(412),
    [anon_sym_void] = ACTIONS(412),
    [anon_sym_delete] = ACTIONS(412),
    [anon_sym_PLUS_PLUS] = ACTIONS(416),
    [anon_sym_DASH_DASH] = ACTIONS(416),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(419),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(424),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(428),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(371),
    [anon_sym_get] = ACTIONS(371),
    [anon_sym_set] = ACTIONS(371),
    [sym__ternary_qmark] = ACTIONS(169),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(75)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(498),
    [sym_expression] = STATE(587),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1762),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1762),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(498),
    [sym_subscript_expression] = STATE(498),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1138),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1762),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1816),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(430),
    [anon_sym_export] = ACTIONS(432),
    [anon_sym_STAR] = ACTIONS(165),
    [anon_sym_LBRACE] = ACTIONS(434),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(432),
    [anon_sym_await] = ACTIONS(436),
    [anon_sym_of] = ACTIONS(165),
    [anon_sym_LPAREN] = ACTIONS(379),
    [anon_sym_in] = ACTIONS(165),
    [anon_sym_yield] = ACTIONS(438),
    [anon_sym_EQ] = ACTIONS(440),
    [anon_sym_LBRACK] = ACTIONS(442),
    [anon_sym_LT] = ACTIONS(389),
    [anon_sym_GT] = ACTIONS(165),
    [anon_sym_DOT] = ACTIONS(165),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(445),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_EQ_GT] = ACTIONS(580),
    [sym_optional_chain] = ACTIONS(169),
    [anon_sym_new] = ACTIONS(447),
    [anon_sym_PLUS_EQ] = ACTIONS(202),
    [anon_sym_DASH_EQ] = ACTIONS(202),
    [anon_sym_STAR_EQ] = ACTIONS(202),
    [anon_sym_SLASH_EQ] = ACTIONS(202),
    [anon_sym_PERCENT_EQ] = ACTIONS(202),
    [anon_sym_CARET_EQ] = ACTIONS(202),
    [anon_sym_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_LT_LT_EQ] = ACTIONS(202),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(202),
    [anon_sym_QMARK_QMARK_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP] = ACTIONS(165),
    [anon_sym_PIPE_PIPE] = ACTIONS(165),
    [anon_sym_GT_GT] = ACTIONS(165),
    [anon_sym_GT_GT_GT] = ACTIONS(165),
    [anon_sym_LT_LT] = ACTIONS(165),
    [anon_sym_AMP] = ACTIONS(165),
    [anon_sym_CARET] = ACTIONS(165),
    [anon_sym_PIPE] = ACTIONS(165),
    [anon_sym_PLUS] = ACTIONS(449),
    [anon_sym_DASH] = ACTIONS(449),
    [anon_sym_SLASH] = ACTIONS(409),
    [anon_sym_PERCENT] = ACTIONS(165),
    [anon_sym_STAR_STAR] = ACTIONS(165),
    [anon_sym_LT_EQ] = ACTIONS(169),
    [anon_sym_EQ_EQ] = ACTIONS(165),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(169),
    [anon_sym_BANG_EQ] = ACTIONS(165),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(169),
    [anon_sym_GT_EQ] = ACTIONS(169),
    [anon_sym_QMARK_QMARK] = ACTIONS(165),
    [anon_sym_instanceof] = ACTIONS(165),
    [anon_sym_BANG] = ACTIONS(452),
    [anon_sym_TILDE] = ACTIONS(454),
    [anon_sym_typeof] = ACTIONS(452),
    [anon_sym_void] = ACTIONS(452),
    [anon_sym_delete] = ACTIONS(452),
    [anon_sym_PLUS_PLUS] = ACTIONS(456),
    [anon_sym_DASH_DASH] = ACTIONS(456),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(419),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(459),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(461),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(432),
    [anon_sym_get] = ACTIONS(432),
    [anon_sym_set] = ACTIONS(432),
    [sym__ternary_qmark] = ACTIONS(169),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(76)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(498),
    [sym_expression] = STATE(587),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1762),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1762),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(498),
    [sym_subscript_expression] = STATE(498),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1138),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1762),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1816),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(430),
    [anon_sym_export] = ACTIONS(432),
    [anon_sym_STAR] = ACTIONS(165),
    [anon_sym_LBRACE] = ACTIONS(434),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(432),
    [anon_sym_await] = ACTIONS(436),
    [anon_sym_LPAREN] = ACTIONS(379),
    [anon_sym_in] = ACTIONS(165),
    [anon_sym_COLON] = ACTIONS(169),
    [anon_sym_yield] = ACTIONS(438),
    [anon_sym_EQ] = ACTIONS(440),
    [anon_sym_LBRACK] = ACTIONS(442),
    [anon_sym_LT] = ACTIONS(389),
    [anon_sym_GT] = ACTIONS(165),
    [anon_sym_DOT] = ACTIONS(165),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(445),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_EQ_GT] = ACTIONS(498),
    [sym_optional_chain] = ACTIONS(169),
    [anon_sym_new] = ACTIONS(447),
    [anon_sym_PLUS_EQ] = ACTIONS(202),
    [anon_sym_DASH_EQ] = ACTIONS(202),
    [anon_sym_STAR_EQ] = ACTIONS(202),
    [anon_sym_SLASH_EQ] = ACTIONS(202),
    [anon_sym_PERCENT_EQ] = ACTIONS(202),
    [anon_sym_CARET_EQ] = ACTIONS(202),
    [anon_sym_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_LT_LT_EQ] = ACTIONS(202),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(202),
    [anon_sym_QMARK_QMARK_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP] = ACTIONS(165),
    [anon_sym_PIPE_PIPE] = ACTIONS(165),
    [anon_sym_GT_GT] = ACTIONS(165),
    [anon_sym_GT_GT_GT] = ACTIONS(165),
    [anon_sym_LT_LT] = ACTIONS(165),
    [anon_sym_AMP] = ACTIONS(165),
    [anon_sym_CARET] = ACTIONS(165),
    [anon_sym_PIPE] = ACTIONS(165),
    [anon_sym_PLUS] = ACTIONS(449),
    [anon_sym_DASH] = ACTIONS(449),
    [anon_sym_SLASH] = ACTIONS(409),
    [anon_sym_PERCENT] = ACTIONS(165),
    [anon_sym_STAR_STAR] = ACTIONS(165),
    [anon_sym_LT_EQ] = ACTIONS(169),
    [anon_sym_EQ_EQ] = ACTIONS(165),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(169),
    [anon_sym_BANG_EQ] = ACTIONS(165),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(169),
    [anon_sym_GT_EQ] = ACTIONS(169),
    [anon_sym_QMARK_QMARK] = ACTIONS(165),
    [anon_sym_instanceof] = ACTIONS(165),
    [anon_sym_BANG] = ACTIONS(452),
    [anon_sym_TILDE] = ACTIONS(454),
    [anon_sym_typeof] = ACTIONS(452),
    [anon_sym_void] = ACTIONS(452),
    [anon_sym_delete] = ACTIONS(452),
    [anon_sym_PLUS_PLUS] = ACTIONS(456),
    [anon_sym_DASH_DASH] = ACTIONS(456),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(419),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(459),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(461),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(432),
    [anon_sym_get] = ACTIONS(432),
    [anon_sym_set] = ACTIONS(432),
    [sym__ternary_qmark] = ACTIONS(169),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(77)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(579),
    [sym_expression] = STATE(587),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1857),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1857),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(579),
    [sym_subscript_expression] = STATE(579),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1141),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1857),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1779),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(582),
    [anon_sym_export] = ACTIONS(584),
    [anon_sym_STAR] = ACTIONS(165),
    [anon_sym_LBRACE] = ACTIONS(434),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(584),
    [anon_sym_await] = ACTIONS(586),
    [anon_sym_of] = ACTIONS(165),
    [anon_sym_LPAREN] = ACTIONS(379),
    [anon_sym_in] = ACTIONS(165),
    [anon_sym_yield] = ACTIONS(588),
    [anon_sym_EQ] = ACTIONS(590),
    [anon_sym_LBRACK] = ACTIONS(442),
    [anon_sym_LT] = ACTIONS(389),
    [anon_sym_GT] = ACTIONS(165),
    [anon_sym_DOT] = ACTIONS(165),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(592),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_EQ_GT] = ACTIONS(580),
    [sym_optional_chain] = ACTIONS(169),
    [anon_sym_new] = ACTIONS(594),
    [anon_sym_PLUS_EQ] = ACTIONS(202),
    [anon_sym_DASH_EQ] = ACTIONS(202),
    [anon_sym_STAR_EQ] = ACTIONS(202),
    [anon_sym_SLASH_EQ] = ACTIONS(202),
    [anon_sym_PERCENT_EQ] = ACTIONS(202),
    [anon_sym_CARET_EQ] = ACTIONS(202),
    [anon_sym_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_LT_LT_EQ] = ACTIONS(202),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(202),
    [anon_sym_QMARK_QMARK_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP] = ACTIONS(165),
    [anon_sym_PIPE_PIPE] = ACTIONS(165),
    [anon_sym_GT_GT] = ACTIONS(165),
    [anon_sym_GT_GT_GT] = ACTIONS(165),
    [anon_sym_LT_LT] = ACTIONS(165),
    [anon_sym_AMP] = ACTIONS(165),
    [anon_sym_CARET] = ACTIONS(165),
    [anon_sym_PIPE] = ACTIONS(165),
    [anon_sym_PLUS] = ACTIONS(596),
    [anon_sym_DASH] = ACTIONS(596),
    [anon_sym_SLASH] = ACTIONS(599),
    [anon_sym_PERCENT] = ACTIONS(165),
    [anon_sym_STAR_STAR] = ACTIONS(165),
    [anon_sym_LT_EQ] = ACTIONS(169),
    [anon_sym_EQ_EQ] = ACTIONS(165),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(169),
    [anon_sym_BANG_EQ] = ACTIONS(165),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(169),
    [anon_sym_GT_EQ] = ACTIONS(169),
    [anon_sym_QMARK_QMARK] = ACTIONS(165),
    [anon_sym_instanceof] = ACTIONS(165),
    [anon_sym_BANG] = ACTIONS(602),
    [anon_sym_TILDE] = ACTIONS(604),
    [anon_sym_typeof] = ACTIONS(602),
    [anon_sym_void] = ACTIONS(602),
    [anon_sym_delete] = ACTIONS(602),
    [anon_sym_PLUS_PLUS] = ACTIONS(606),
    [anon_sym_DASH_DASH] = ACTIONS(606),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(419),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(609),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(611),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(584),
    [anon_sym_get] = ACTIONS(584),
    [anon_sym_set] = ACTIONS(584),
    [sym__ternary_qmark] = ACTIONS(169),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(78)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(578),
    [sym_expression] = STATE(587),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1776),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1776),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(578),
    [sym_subscript_expression] = STATE(578),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1142),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1776),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1778),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(613),
    [anon_sym_export] = ACTIONS(615),
    [anon_sym_STAR] = ACTIONS(165),
    [anon_sym_LBRACE] = ACTIONS(617),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(615),
    [anon_sym_await] = ACTIONS(620),
    [anon_sym_LPAREN] = ACTIONS(379),
    [anon_sym_in] = ACTIONS(165),
    [anon_sym_yield] = ACTIONS(622),
    [anon_sym_EQ] = ACTIONS(624),
    [anon_sym_LBRACK] = ACTIONS(442),
    [anon_sym_LT] = ACTIONS(389),
    [anon_sym_GT] = ACTIONS(165),
    [anon_sym_DOT] = ACTIONS(165),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(626),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_EQ_GT] = ACTIONS(628),
    [sym_optional_chain] = ACTIONS(169),
    [anon_sym_new] = ACTIONS(630),
    [anon_sym_PLUS_EQ] = ACTIONS(202),
    [anon_sym_DASH_EQ] = ACTIONS(202),
    [anon_sym_STAR_EQ] = ACTIONS(202),
    [anon_sym_SLASH_EQ] = ACTIONS(202),
    [anon_sym_PERCENT_EQ] = ACTIONS(202),
    [anon_sym_CARET_EQ] = ACTIONS(202),
    [anon_sym_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_LT_LT_EQ] = ACTIONS(202),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(202),
    [anon_sym_QMARK_QMARK_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP] = ACTIONS(165),
    [anon_sym_PIPE_PIPE] = ACTIONS(165),
    [anon_sym_GT_GT] = ACTIONS(165),
    [anon_sym_GT_GT_GT] = ACTIONS(165),
    [anon_sym_LT_LT] = ACTIONS(165),
    [anon_sym_AMP] = ACTIONS(165),
    [anon_sym_CARET] = ACTIONS(165),
    [anon_sym_PIPE] = ACTIONS(165),
    [anon_sym_PLUS] = ACTIONS(632),
    [anon_sym_DASH] = ACTIONS(632),
    [anon_sym_SLASH] = ACTIONS(409),
    [anon_sym_PERCENT] = ACTIONS(165),
    [anon_sym_STAR_STAR] = ACTIONS(165),
    [anon_sym_LT_EQ] = ACTIONS(169),
    [anon_sym_EQ_EQ] = ACTIONS(165),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(169),
    [anon_sym_BANG_EQ] = ACTIONS(165),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(169),
    [anon_sym_GT_EQ] = ACTIONS(169),
    [anon_sym_QMARK_QMARK] = ACTIONS(165),
    [anon_sym_instanceof] = ACTIONS(165),
    [anon_sym_BANG] = ACTIONS(635),
    [anon_sym_TILDE] = ACTIONS(637),
    [anon_sym_typeof] = ACTIONS(635),
    [anon_sym_void] = ACTIONS(635),
    [anon_sym_delete] = ACTIONS(635),
    [anon_sym_PLUS_PLUS] = ACTIONS(639),
    [anon_sym_DASH_DASH] = ACTIONS(639),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(419),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(642),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(644),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(615),
    [anon_sym_get] = ACTIONS(615),
    [anon_sym_set] = ACTIONS(615),
    [sym__ternary_qmark] = ACTIONS(169),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(79)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(498),
    [sym_expression] = STATE(587),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1762),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1762),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(498),
    [sym_subscript_expression] = STATE(498),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1138),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1762),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1816),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(430),
    [anon_sym_export] = ACTIONS(432),
    [anon_sym_STAR] = ACTIONS(165),
    [anon_sym_LBRACE] = ACTIONS(617),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(432),
    [anon_sym_await] = ACTIONS(436),
    [anon_sym_LPAREN] = ACTIONS(379),
    [anon_sym_in] = ACTIONS(165),
    [anon_sym_yield] = ACTIONS(438),
    [anon_sym_EQ] = ACTIONS(440),
    [anon_sym_LBRACK] = ACTIONS(442),
    [anon_sym_LT] = ACTIONS(389),
    [anon_sym_GT] = ACTIONS(165),
    [anon_sym_DOT] = ACTIONS(165),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(445),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_EQ_GT] = ACTIONS(628),
    [sym_optional_chain] = ACTIONS(169),
    [anon_sym_new] = ACTIONS(447),
    [anon_sym_PLUS_EQ] = ACTIONS(202),
    [anon_sym_DASH_EQ] = ACTIONS(202),
    [anon_sym_STAR_EQ] = ACTIONS(202),
    [anon_sym_SLASH_EQ] = ACTIONS(202),
    [anon_sym_PERCENT_EQ] = ACTIONS(202),
    [anon_sym_CARET_EQ] = ACTIONS(202),
    [anon_sym_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_LT_LT_EQ] = ACTIONS(202),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(202),
    [anon_sym_QMARK_QMARK_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP] = ACTIONS(165),
    [anon_sym_PIPE_PIPE] = ACTIONS(165),
    [anon_sym_GT_GT] = ACTIONS(165),
    [anon_sym_GT_GT_GT] = ACTIONS(165),
    [anon_sym_LT_LT] = ACTIONS(165),
    [anon_sym_AMP] = ACTIONS(165),
    [anon_sym_CARET] = ACTIONS(165),
    [anon_sym_PIPE] = ACTIONS(165),
    [anon_sym_PLUS] = ACTIONS(449),
    [anon_sym_DASH] = ACTIONS(449),
    [anon_sym_SLASH] = ACTIONS(409),
    [anon_sym_PERCENT] = ACTIONS(165),
    [anon_sym_STAR_STAR] = ACTIONS(165),
    [anon_sym_LT_EQ] = ACTIONS(169),
    [anon_sym_EQ_EQ] = ACTIONS(165),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(169),
    [anon_sym_BANG_EQ] = ACTIONS(165),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(169),
    [anon_sym_GT_EQ] = ACTIONS(169),
    [anon_sym_QMARK_QMARK] = ACTIONS(165),
    [anon_sym_instanceof] = ACTIONS(165),
    [anon_sym_BANG] = ACTIONS(452),
    [anon_sym_TILDE] = ACTIONS(454),
    [anon_sym_typeof] = ACTIONS(452),
    [anon_sym_void] = ACTIONS(452),
    [anon_sym_delete] = ACTIONS(452),
    [anon_sym_PLUS_PLUS] = ACTIONS(456),
    [anon_sym_DASH_DASH] = ACTIONS(456),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(419),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(459),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(461),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(432),
    [anon_sym_get] = ACTIONS(432),
    [anon_sym_set] = ACTIONS(432),
    [sym__ternary_qmark] = ACTIONS(169),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(80)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(507),
    [sym_expression] = STATE(663),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1815),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1815),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(507),
    [sym_subscript_expression] = STATE(507),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1139),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1815),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1763),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(369),
    [anon_sym_export] = ACTIONS(371),
    [anon_sym_STAR] = ACTIONS(646),
    [anon_sym_LBRACE] = ACTIONS(373),
    [anon_sym_COMMA] = ACTIONS(648),
    [anon_sym_RBRACE] = ACTIONS(648),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(371),
    [anon_sym_await] = ACTIONS(377),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_SEMI] = ACTIONS(648),
    [anon_sym_RPAREN] = ACTIONS(648),
    [anon_sym_in] = ACTIONS(652),
    [anon_sym_COLON] = ACTIONS(648),
    [anon_sym_yield] = ACTIONS(382),
    [anon_sym_LBRACK] = ACTIONS(654),
    [anon_sym_RBRACK] = ACTIONS(648),
    [anon_sym_LT] = ACTIONS(656),
    [anon_sym_GT] = ACTIONS(652),
    [anon_sym_DOT] = ACTIONS(652),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(398),
    [anon_sym_function] = ACTIONS(400),
    [sym_optional_chain] = ACTIONS(648),
    [anon_sym_new] = ACTIONS(404),
    [anon_sym_AMP_AMP] = ACTIONS(648),
    [anon_sym_PIPE_PIPE] = ACTIONS(648),
    [anon_sym_GT_GT] = ACTIONS(652),
    [anon_sym_GT_GT_GT] = ACTIONS(648),
    [anon_sym_LT_LT] = ACTIONS(648),
    [anon_sym_AMP] = ACTIONS(652),
    [anon_sym_CARET] = ACTIONS(648),
    [anon_sym_PIPE] = ACTIONS(652),
    [anon_sym_PLUS] = ACTIONS(412),
    [anon_sym_DASH] = ACTIONS(412),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_PERCENT] = ACTIONS(648),
    [anon_sym_STAR_STAR] = ACTIONS(648),
    [anon_sym_LT_EQ] = ACTIONS(648),
    [anon_sym_EQ_EQ] = ACTIONS(652),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(648),
    [anon_sym_BANG_EQ] = ACTIONS(652),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(648),
    [anon_sym_GT_EQ] = ACTIONS(648),
    [anon_sym_QMARK_QMARK] = ACTIONS(648),
    [anon_sym_instanceof] = ACTIONS(652),
    [anon_sym_BANG] = ACTIONS(412),
    [anon_sym_TILDE] = ACTIONS(414),
    [anon_sym_typeof] = ACTIONS(412),
    [anon_sym_void] = ACTIONS(412),
    [anon_sym_delete] = ACTIONS(412),
    [anon_sym_PLUS_PLUS] = ACTIONS(660),
    [anon_sym_DASH_DASH] = ACTIONS(660),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(424),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(428),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(371),
    [anon_sym_get] = ACTIONS(371),
    [anon_sym_set] = ACTIONS(371),
    [sym__ternary_qmark] = ACTIONS(648),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(81)] = {
    [sym_import] = STATE(1326),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(802),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1790),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1790),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_function_expression] = STATE(740),
    [sym_generator_function] = STATE(740),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1790),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [aux_sym_export_statement_repeat1] = STATE(1381),
    [sym_identifier] = ACTIONS(161),
    [anon_sym_export] = ACTIONS(163),
    [anon_sym_STAR] = ACTIONS(664),
    [anon_sym_LBRACE] = ACTIONS(167),
    [anon_sym_COMMA] = ACTIONS(648),
    [anon_sym_RBRACE] = ACTIONS(648),
    [anon_sym_import] = ACTIONS(173),
    [anon_sym_let] = ACTIONS(163),
    [anon_sym_await] = ACTIONS(177),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_SEMI] = ACTIONS(648),
    [anon_sym_in] = ACTIONS(652),
    [anon_sym_yield] = ACTIONS(57),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(666),
    [anon_sym_GT] = ACTIONS(652),
    [anon_sym_DOT] = ACTIONS(652),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(194),
    [anon_sym_async] = ACTIONS(196),
    [anon_sym_function] = ACTIONS(198),
    [sym_optional_chain] = ACTIONS(648),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_AMP_AMP] = ACTIONS(648),
    [anon_sym_PIPE_PIPE] = ACTIONS(648),
    [anon_sym_GT_GT] = ACTIONS(652),
    [anon_sym_GT_GT_GT] = ACTIONS(648),
    [anon_sym_LT_LT] = ACTIONS(648),
    [anon_sym_AMP] = ACTIONS(652),
    [anon_sym_CARET] = ACTIONS(648),
    [anon_sym_PIPE] = ACTIONS(652),
    [anon_sym_PLUS] = ACTIONS(75),
    [anon_sym_DASH] = ACTIONS(75),
    [anon_sym_SLASH] = ACTIONS(77),
    [anon_sym_PERCENT] = ACTIONS(648),
    [anon_sym_STAR_STAR] = ACTIONS(648),
    [anon_sym_LT_EQ] = ACTIONS(648),
    [anon_sym_EQ_EQ] = ACTIONS(652),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(648),
    [anon_sym_BANG_EQ] = ACTIONS(652),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(648),
    [anon_sym_GT_EQ] = ACTIONS(648),
    [anon_sym_QMARK_QMARK] = ACTIONS(648),
    [anon_sym_instanceof] = ACTIONS(652),
    [anon_sym_BANG] = ACTIONS(75),
    [anon_sym_TILDE] = ACTIONS(79),
    [anon_sym_typeof] = ACTIONS(75),
    [anon_sym_void] = ACTIONS(75),
    [anon_sym_delete] = ACTIONS(75),
    [anon_sym_PLUS_PLUS] = ACTIONS(81),
    [anon_sym_DASH_DASH] = ACTIONS(81),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(87),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(163),
    [anon_sym_get] = ACTIONS(163),
    [anon_sym_set] = ACTIONS(163),
    [sym__automatic_semicolon] = ACTIONS(648),
    [sym__ternary_qmark] = ACTIONS(648),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(82)] = {
    [sym_import] = STATE(1326),
    [sym_parenthesized_expression] = STATE(560),
    [sym_expression] = STATE(796),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1802),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1802),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_function_expression] = STATE(740),
    [sym_generator_function] = STATE(740),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(560),
    [sym_subscript_expression] = STATE(560),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1143),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1802),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1803),
    [aux_sym_export_statement_repeat1] = STATE(1381),
    [sym_identifier] = ACTIONS(520),
    [anon_sym_export] = ACTIONS(522),
    [anon_sym_STAR] = ACTIONS(668),
    [anon_sym_LBRACE] = ACTIONS(167),
    [anon_sym_COMMA] = ACTIONS(648),
    [anon_sym_import] = ACTIONS(173),
    [anon_sym_let] = ACTIONS(522),
    [anon_sym_await] = ACTIONS(524),
    [anon_sym_of] = ACTIONS(652),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_SEMI] = ACTIONS(648),
    [anon_sym_in] = ACTIONS(652),
    [anon_sym_yield] = ACTIONS(526),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(666),
    [anon_sym_GT] = ACTIONS(652),
    [anon_sym_DOT] = ACTIONS(652),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(194),
    [anon_sym_async] = ACTIONS(530),
    [anon_sym_function] = ACTIONS(198),
    [sym_optional_chain] = ACTIONS(648),
    [anon_sym_new] = ACTIONS(532),
    [anon_sym_AMP_AMP] = ACTIONS(648),
    [anon_sym_PIPE_PIPE] = ACTIONS(648),
    [anon_sym_GT_GT] = ACTIONS(652),
    [anon_sym_GT_GT_GT] = ACTIONS(648),
    [anon_sym_LT_LT] = ACTIONS(648),
    [anon_sym_AMP] = ACTIONS(652),
    [anon_sym_CARET] = ACTIONS(648),
    [anon_sym_PIPE] = ACTIONS(652),
    [anon_sym_PLUS] = ACTIONS(540),
    [anon_sym_DASH] = ACTIONS(540),
    [anon_sym_SLASH] = ACTIONS(670),
    [anon_sym_PERCENT] = ACTIONS(648),
    [anon_sym_STAR_STAR] = ACTIONS(648),
    [anon_sym_LT_EQ] = ACTIONS(648),
    [anon_sym_EQ_EQ] = ACTIONS(652),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(648),
    [anon_sym_BANG_EQ] = ACTIONS(652),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(648),
    [anon_sym_GT_EQ] = ACTIONS(648),
    [anon_sym_QMARK_QMARK] = ACTIONS(648),
    [anon_sym_instanceof] = ACTIONS(652),
    [anon_sym_BANG] = ACTIONS(540),
    [anon_sym_TILDE] = ACTIONS(542),
    [anon_sym_typeof] = ACTIONS(540),
    [anon_sym_void] = ACTIONS(540),
    [anon_sym_delete] = ACTIONS(540),
    [anon_sym_PLUS_PLUS] = ACTIONS(672),
    [anon_sym_DASH_DASH] = ACTIONS(672),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(547),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(549),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(522),
    [anon_sym_get] = ACTIONS(522),
    [anon_sym_set] = ACTIONS(522),
    [sym__automatic_semicolon] = ACTIONS(648),
    [sym__ternary_qmark] = ACTIONS(648),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(83)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(579),
    [sym_expression] = STATE(913),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1857),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1857),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(579),
    [sym_subscript_expression] = STATE(579),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1141),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1857),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1779),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(582),
    [anon_sym_export] = ACTIONS(584),
    [anon_sym_STAR] = ACTIONS(674),
    [anon_sym_LBRACE] = ACTIONS(434),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(584),
    [anon_sym_await] = ACTIONS(586),
    [anon_sym_of] = ACTIONS(652),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_in] = ACTIONS(652),
    [anon_sym_yield] = ACTIONS(588),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_LT] = ACTIONS(656),
    [anon_sym_GT] = ACTIONS(652),
    [anon_sym_DOT] = ACTIONS(652),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(592),
    [anon_sym_function] = ACTIONS(400),
    [sym_optional_chain] = ACTIONS(648),
    [anon_sym_new] = ACTIONS(594),
    [anon_sym_AMP_AMP] = ACTIONS(648),
    [anon_sym_PIPE_PIPE] = ACTIONS(648),
    [anon_sym_GT_GT] = ACTIONS(652),
    [anon_sym_GT_GT_GT] = ACTIONS(648),
    [anon_sym_LT_LT] = ACTIONS(648),
    [anon_sym_AMP] = ACTIONS(652),
    [anon_sym_CARET] = ACTIONS(648),
    [anon_sym_PIPE] = ACTIONS(652),
    [anon_sym_PLUS] = ACTIONS(602),
    [anon_sym_DASH] = ACTIONS(602),
    [anon_sym_SLASH] = ACTIONS(678),
    [anon_sym_PERCENT] = ACTIONS(648),
    [anon_sym_STAR_STAR] = ACTIONS(648),
    [anon_sym_LT_EQ] = ACTIONS(648),
    [anon_sym_EQ_EQ] = ACTIONS(652),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(648),
    [anon_sym_BANG_EQ] = ACTIONS(652),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(648),
    [anon_sym_GT_EQ] = ACTIONS(648),
    [anon_sym_QMARK_QMARK] = ACTIONS(648),
    [anon_sym_instanceof] = ACTIONS(652),
    [anon_sym_BANG] = ACTIONS(602),
    [anon_sym_TILDE] = ACTIONS(604),
    [anon_sym_typeof] = ACTIONS(602),
    [anon_sym_void] = ACTIONS(602),
    [anon_sym_delete] = ACTIONS(602),
    [anon_sym_PLUS_PLUS] = ACTIONS(680),
    [anon_sym_DASH_DASH] = ACTIONS(680),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(609),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(611),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(584),
    [anon_sym_get] = ACTIONS(584),
    [anon_sym_set] = ACTIONS(584),
    [sym__ternary_qmark] = ACTIONS(648),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(84)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(498),
    [sym_expression] = STATE(966),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1762),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1762),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(498),
    [sym_subscript_expression] = STATE(498),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1138),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1762),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1816),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(430),
    [anon_sym_export] = ACTIONS(432),
    [anon_sym_STAR] = ACTIONS(682),
    [anon_sym_LBRACE] = ACTIONS(434),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(432),
    [anon_sym_await] = ACTIONS(436),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_in] = ACTIONS(652),
    [anon_sym_COLON] = ACTIONS(648),
    [anon_sym_yield] = ACTIONS(438),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_LT] = ACTIONS(656),
    [anon_sym_GT] = ACTIONS(652),
    [anon_sym_DOT] = ACTIONS(652),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(445),
    [anon_sym_function] = ACTIONS(400),
    [sym_optional_chain] = ACTIONS(648),
    [anon_sym_new] = ACTIONS(447),
    [anon_sym_AMP_AMP] = ACTIONS(648),
    [anon_sym_PIPE_PIPE] = ACTIONS(648),
    [anon_sym_GT_GT] = ACTIONS(652),
    [anon_sym_GT_GT_GT] = ACTIONS(648),
    [anon_sym_LT_LT] = ACTIONS(648),
    [anon_sym_AMP] = ACTIONS(652),
    [anon_sym_CARET] = ACTIONS(648),
    [anon_sym_PIPE] = ACTIONS(652),
    [anon_sym_PLUS] = ACTIONS(452),
    [anon_sym_DASH] = ACTIONS(452),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_PERCENT] = ACTIONS(648),
    [anon_sym_STAR_STAR] = ACTIONS(648),
    [anon_sym_LT_EQ] = ACTIONS(648),
    [anon_sym_EQ_EQ] = ACTIONS(652),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(648),
    [anon_sym_BANG_EQ] = ACTIONS(652),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(648),
    [anon_sym_GT_EQ] = ACTIONS(648),
    [anon_sym_QMARK_QMARK] = ACTIONS(648),
    [anon_sym_instanceof] = ACTIONS(652),
    [anon_sym_BANG] = ACTIONS(452),
    [anon_sym_TILDE] = ACTIONS(454),
    [anon_sym_typeof] = ACTIONS(452),
    [anon_sym_void] = ACTIONS(452),
    [anon_sym_delete] = ACTIONS(452),
    [anon_sym_PLUS_PLUS] = ACTIONS(684),
    [anon_sym_DASH_DASH] = ACTIONS(684),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(459),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(461),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(432),
    [anon_sym_get] = ACTIONS(432),
    [anon_sym_set] = ACTIONS(432),
    [sym__ternary_qmark] = ACTIONS(648),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(85)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(578),
    [sym_expression] = STATE(918),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1776),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1776),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(578),
    [sym_subscript_expression] = STATE(578),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1142),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1776),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1778),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(613),
    [anon_sym_export] = ACTIONS(615),
    [anon_sym_STAR] = ACTIONS(686),
    [anon_sym_LBRACE] = ACTIONS(434),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(615),
    [anon_sym_await] = ACTIONS(620),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_in] = ACTIONS(652),
    [anon_sym_yield] = ACTIONS(622),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_LT] = ACTIONS(656),
    [anon_sym_GT] = ACTIONS(652),
    [anon_sym_DOT] = ACTIONS(652),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(626),
    [anon_sym_function] = ACTIONS(400),
    [sym_optional_chain] = ACTIONS(648),
    [anon_sym_new] = ACTIONS(630),
    [anon_sym_AMP_AMP] = ACTIONS(648),
    [anon_sym_PIPE_PIPE] = ACTIONS(648),
    [anon_sym_GT_GT] = ACTIONS(652),
    [anon_sym_GT_GT_GT] = ACTIONS(648),
    [anon_sym_LT_LT] = ACTIONS(648),
    [anon_sym_AMP] = ACTIONS(652),
    [anon_sym_CARET] = ACTIONS(648),
    [anon_sym_PIPE] = ACTIONS(652),
    [anon_sym_PLUS] = ACTIONS(635),
    [anon_sym_DASH] = ACTIONS(635),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_PERCENT] = ACTIONS(648),
    [anon_sym_STAR_STAR] = ACTIONS(648),
    [anon_sym_LT_EQ] = ACTIONS(648),
    [anon_sym_EQ_EQ] = ACTIONS(652),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(648),
    [anon_sym_BANG_EQ] = ACTIONS(652),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(648),
    [anon_sym_GT_EQ] = ACTIONS(648),
    [anon_sym_QMARK_QMARK] = ACTIONS(648),
    [anon_sym_instanceof] = ACTIONS(652),
    [anon_sym_BANG] = ACTIONS(635),
    [anon_sym_TILDE] = ACTIONS(637),
    [anon_sym_typeof] = ACTIONS(635),
    [anon_sym_void] = ACTIONS(635),
    [anon_sym_delete] = ACTIONS(635),
    [anon_sym_PLUS_PLUS] = ACTIONS(688),
    [anon_sym_DASH_DASH] = ACTIONS(688),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(642),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(644),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(615),
    [anon_sym_get] = ACTIONS(615),
    [anon_sym_set] = ACTIONS(615),
    [sym__ternary_qmark] = ACTIONS(648),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(86)] = {
    [ts_builtin_sym_end] = ACTIONS(690),
    [sym_identifier] = ACTIONS(692),
    [anon_sym_export] = ACTIONS(692),
    [anon_sym_STAR] = ACTIONS(692),
    [anon_sym_default] = ACTIONS(692),
    [anon_sym_LBRACE] = ACTIONS(690),
    [anon_sym_COMMA] = ACTIONS(690),
    [anon_sym_RBRACE] = ACTIONS(690),
    [anon_sym_import] = ACTIONS(692),
    [anon_sym_with] = ACTIONS(692),
    [anon_sym_var] = ACTIONS(692),
    [anon_sym_let] = ACTIONS(692),
    [anon_sym_const] = ACTIONS(692),
    [anon_sym_using] = ACTIONS(692),
    [anon_sym_await] = ACTIONS(692),
    [anon_sym_of] = ACTIONS(692),
    [anon_sym_else] = ACTIONS(692),
    [anon_sym_if] = ACTIONS(692),
    [anon_sym_switch] = ACTIONS(692),
    [anon_sym_for] = ACTIONS(692),
    [anon_sym_LPAREN] = ACTIONS(690),
    [anon_sym_SEMI] = ACTIONS(690),
    [anon_sym_in] = ACTIONS(692),
    [anon_sym_while] = ACTIONS(692),
    [anon_sym_do] = ACTIONS(692),
    [anon_sym_try] = ACTIONS(692),
    [anon_sym_break] = ACTIONS(692),
    [anon_sym_continue] = ACTIONS(692),
    [anon_sym_debugger] = ACTIONS(692),
    [anon_sym_return] = ACTIONS(692),
    [anon_sym_throw] = ACTIONS(692),
    [anon_sym_case] = ACTIONS(692),
    [anon_sym_yield] = ACTIONS(692),
    [anon_sym_LBRACK] = ACTIONS(690),
    [anon_sym_LT] = ACTIONS(692),
    [anon_sym_GT] = ACTIONS(692),
    [anon_sym_DOT] = ACTIONS(692),
    [anon_sym_DQUOTE] = ACTIONS(690),
    [anon_sym_SQUOTE] = ACTIONS(690),
    [anon_sym_class] = ACTIONS(692),
    [anon_sym_async] = ACTIONS(692),
    [anon_sym_function] = ACTIONS(692),
    [sym_optional_chain] = ACTIONS(690),
    [anon_sym_new] = ACTIONS(692),
    [anon_sym_AMP_AMP] = ACTIONS(690),
    [anon_sym_PIPE_PIPE] = ACTIONS(690),
    [anon_sym_GT_GT] = ACTIONS(692),
    [anon_sym_GT_GT_GT] = ACTIONS(690),
    [anon_sym_LT_LT] = ACTIONS(690),
    [anon_sym_AMP] = ACTIONS(692),
    [anon_sym_CARET] = ACTIONS(690),
    [anon_sym_PIPE] = ACTIONS(692),
    [anon_sym_PLUS] = ACTIONS(692),
    [anon_sym_DASH] = ACTIONS(692),
    [anon_sym_SLASH] = ACTIONS(692),
    [anon_sym_PERCENT] = ACTIONS(690),
    [anon_sym_STAR_STAR] = ACTIONS(690),
    [anon_sym_LT_EQ] = ACTIONS(690),
    [anon_sym_EQ_EQ] = ACTIONS(692),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(690),
    [anon_sym_BANG_EQ] = ACTIONS(692),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(690),
    [anon_sym_GT_EQ] = ACTIONS(690),
    [anon_sym_QMARK_QMARK] = ACTIONS(690),
    [anon_sym_instanceof] = ACTIONS(692),
    [anon_sym_BANG] = ACTIONS(692),
    [anon_sym_TILDE] = ACTIONS(690),
    [anon_sym_typeof] = ACTIONS(692),
    [anon_sym_void] = ACTIONS(692),
    [anon_sym_delete] = ACTIONS(692),
    [anon_sym_PLUS_PLUS] = ACTIONS(690),
    [anon_sym_DASH_DASH] = ACTIONS(690),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(690),
    [sym_number] = ACTIONS(690),
    [sym_private_property_identifier] = ACTIONS(690),
    [sym_this] = ACTIONS(692),
    [sym_super] = ACTIONS(692),
    [sym_true] = ACTIONS(692),
    [sym_false] = ACTIONS(692),
    [sym_null] = ACTIONS(692),
    [sym_undefined] = ACTIONS(692),
    [anon_sym_AT] = ACTIONS(690),
    [anon_sym_static] = ACTIONS(692),
    [anon_sym_get] = ACTIONS(692),
    [anon_sym_set] = ACTIONS(692),
    [sym__automatic_semicolon] = ACTIONS(690),
    [sym__ternary_qmark] = ACTIONS(690),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(87)] = {
    [ts_builtin_sym_end] = ACTIONS(694),
    [sym_identifier] = ACTIONS(696),
    [anon_sym_export] = ACTIONS(696),
    [anon_sym_STAR] = ACTIONS(696),
    [anon_sym_default] = ACTIONS(696),
    [anon_sym_LBRACE] = ACTIONS(694),
    [anon_sym_COMMA] = ACTIONS(694),
    [anon_sym_RBRACE] = ACTIONS(694),
    [anon_sym_import] = ACTIONS(696),
    [anon_sym_with] = ACTIONS(696),
    [anon_sym_var] = ACTIONS(696),
    [anon_sym_let] = ACTIONS(696),
    [anon_sym_const] = ACTIONS(696),
    [anon_sym_using] = ACTIONS(696),
    [anon_sym_await] = ACTIONS(696),
    [anon_sym_of] = ACTIONS(696),
    [anon_sym_else] = ACTIONS(696),
    [anon_sym_if] = ACTIONS(696),
    [anon_sym_switch] = ACTIONS(696),
    [anon_sym_for] = ACTIONS(696),
    [anon_sym_LPAREN] = ACTIONS(694),
    [anon_sym_SEMI] = ACTIONS(694),
    [anon_sym_in] = ACTIONS(696),
    [anon_sym_while] = ACTIONS(696),
    [anon_sym_do] = ACTIONS(696),
    [anon_sym_try] = ACTIONS(696),
    [anon_sym_break] = ACTIONS(696),
    [anon_sym_continue] = ACTIONS(696),
    [anon_sym_debugger] = ACTIONS(696),
    [anon_sym_return] = ACTIONS(696),
    [anon_sym_throw] = ACTIONS(696),
    [anon_sym_case] = ACTIONS(696),
    [anon_sym_yield] = ACTIONS(696),
    [anon_sym_LBRACK] = ACTIONS(694),
    [anon_sym_LT] = ACTIONS(696),
    [anon_sym_GT] = ACTIONS(696),
    [anon_sym_DOT] = ACTIONS(696),
    [anon_sym_DQUOTE] = ACTIONS(694),
    [anon_sym_SQUOTE] = ACTIONS(694),
    [anon_sym_class] = ACTIONS(696),
    [anon_sym_async] = ACTIONS(696),
    [anon_sym_function] = ACTIONS(696),
    [sym_optional_chain] = ACTIONS(694),
    [anon_sym_new] = ACTIONS(696),
    [anon_sym_AMP_AMP] = ACTIONS(694),
    [anon_sym_PIPE_PIPE] = ACTIONS(694),
    [anon_sym_GT_GT] = ACTIONS(696),
    [anon_sym_GT_GT_GT] = ACTIONS(694),
    [anon_sym_LT_LT] = ACTIONS(694),
    [anon_sym_AMP] = ACTIONS(696),
    [anon_sym_CARET] = ACTIONS(694),
    [anon_sym_PIPE] = ACTIONS(696),
    [anon_sym_PLUS] = ACTIONS(696),
    [anon_sym_DASH] = ACTIONS(696),
    [anon_sym_SLASH] = ACTIONS(696),
    [anon_sym_PERCENT] = ACTIONS(694),
    [anon_sym_STAR_STAR] = ACTIONS(694),
    [anon_sym_LT_EQ] = ACTIONS(694),
    [anon_sym_EQ_EQ] = ACTIONS(696),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(694),
    [anon_sym_BANG_EQ] = ACTIONS(696),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(694),
    [anon_sym_GT_EQ] = ACTIONS(694),
    [anon_sym_QMARK_QMARK] = ACTIONS(694),
    [anon_sym_instanceof] = ACTIONS(696),
    [anon_sym_BANG] = ACTIONS(696),
    [anon_sym_TILDE] = ACTIONS(694),
    [anon_sym_typeof] = ACTIONS(696),
    [anon_sym_void] = ACTIONS(696),
    [anon_sym_delete] = ACTIONS(696),
    [anon_sym_PLUS_PLUS] = ACTIONS(694),
    [anon_sym_DASH_DASH] = ACTIONS(694),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(694),
    [sym_number] = ACTIONS(694),
    [sym_private_property_identifier] = ACTIONS(694),
    [sym_this] = ACTIONS(696),
    [sym_super] = ACTIONS(696),
    [sym_true] = ACTIONS(696),
    [sym_false] = ACTIONS(696),
    [sym_null] = ACTIONS(696),
    [sym_undefined] = ACTIONS(696),
    [anon_sym_AT] = ACTIONS(694),
    [anon_sym_static] = ACTIONS(696),
    [anon_sym_get] = ACTIONS(696),
    [anon_sym_set] = ACTIONS(696),
    [sym__automatic_semicolon] = ACTIONS(694),
    [sym__ternary_qmark] = ACTIONS(694),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(88)] = {
    [ts_builtin_sym_end] = ACTIONS(694),
    [sym_identifier] = ACTIONS(696),
    [anon_sym_export] = ACTIONS(696),
    [anon_sym_STAR] = ACTIONS(696),
    [anon_sym_default] = ACTIONS(696),
    [anon_sym_LBRACE] = ACTIONS(694),
    [anon_sym_COMMA] = ACTIONS(694),
    [anon_sym_RBRACE] = ACTIONS(694),
    [anon_sym_import] = ACTIONS(696),
    [anon_sym_with] = ACTIONS(696),
    [anon_sym_var] = ACTIONS(696),
    [anon_sym_let] = ACTIONS(696),
    [anon_sym_const] = ACTIONS(696),
    [anon_sym_using] = ACTIONS(696),
    [anon_sym_await] = ACTIONS(696),
    [anon_sym_of] = ACTIONS(696),
    [anon_sym_else] = ACTIONS(696),
    [anon_sym_if] = ACTIONS(696),
    [anon_sym_switch] = ACTIONS(696),
    [anon_sym_for] = ACTIONS(696),
    [anon_sym_LPAREN] = ACTIONS(694),
    [anon_sym_SEMI] = ACTIONS(694),
    [anon_sym_in] = ACTIONS(696),
    [anon_sym_while] = ACTIONS(696),
    [anon_sym_do] = ACTIONS(696),
    [anon_sym_try] = ACTIONS(696),
    [anon_sym_break] = ACTIONS(696),
    [anon_sym_continue] = ACTIONS(696),
    [anon_sym_debugger] = ACTIONS(696),
    [anon_sym_return] = ACTIONS(696),
    [anon_sym_throw] = ACTIONS(696),
    [anon_sym_case] = ACTIONS(696),
    [anon_sym_yield] = ACTIONS(696),
    [anon_sym_LBRACK] = ACTIONS(694),
    [anon_sym_LT] = ACTIONS(696),
    [anon_sym_GT] = ACTIONS(696),
    [anon_sym_DOT] = ACTIONS(696),
    [anon_sym_DQUOTE] = ACTIONS(694),
    [anon_sym_SQUOTE] = ACTIONS(694),
    [anon_sym_class] = ACTIONS(696),
    [anon_sym_async] = ACTIONS(696),
    [anon_sym_function] = ACTIONS(696),
    [sym_optional_chain] = ACTIONS(694),
    [anon_sym_new] = ACTIONS(696),
    [anon_sym_AMP_AMP] = ACTIONS(694),
    [anon_sym_PIPE_PIPE] = ACTIONS(694),
    [anon_sym_GT_GT] = ACTIONS(696),
    [anon_sym_GT_GT_GT] = ACTIONS(694),
    [anon_sym_LT_LT] = ACTIONS(694),
    [anon_sym_AMP] = ACTIONS(696),
    [anon_sym_CARET] = ACTIONS(694),
    [anon_sym_PIPE] = ACTIONS(696),
    [anon_sym_PLUS] = ACTIONS(696),
    [anon_sym_DASH] = ACTIONS(696),
    [anon_sym_SLASH] = ACTIONS(696),
    [anon_sym_PERCENT] = ACTIONS(694),
    [anon_sym_STAR_STAR] = ACTIONS(694),
    [anon_sym_LT_EQ] = ACTIONS(694),
    [anon_sym_EQ_EQ] = ACTIONS(696),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(694),
    [anon_sym_BANG_EQ] = ACTIONS(696),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(694),
    [anon_sym_GT_EQ] = ACTIONS(694),
    [anon_sym_QMARK_QMARK] = ACTIONS(694),
    [anon_sym_instanceof] = ACTIONS(696),
    [anon_sym_BANG] = ACTIONS(696),
    [anon_sym_TILDE] = ACTIONS(694),
    [anon_sym_typeof] = ACTIONS(696),
    [anon_sym_void] = ACTIONS(696),
    [anon_sym_delete] = ACTIONS(696),
    [anon_sym_PLUS_PLUS] = ACTIONS(694),
    [anon_sym_DASH_DASH] = ACTIONS(694),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(694),
    [sym_number] = ACTIONS(694),
    [sym_private_property_identifier] = ACTIONS(694),
    [sym_this] = ACTIONS(696),
    [sym_super] = ACTIONS(696),
    [sym_true] = ACTIONS(696),
    [sym_false] = ACTIONS(696),
    [sym_null] = ACTIONS(696),
    [sym_undefined] = ACTIONS(696),
    [anon_sym_AT] = ACTIONS(694),
    [anon_sym_static] = ACTIONS(696),
    [anon_sym_get] = ACTIONS(696),
    [anon_sym_set] = ACTIONS(696),
    [sym__automatic_semicolon] = ACTIONS(698),
    [sym__ternary_qmark] = ACTIONS(694),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(89)] = {
    [ts_builtin_sym_end] = ACTIONS(700),
    [sym_identifier] = ACTIONS(702),
    [anon_sym_export] = ACTIONS(702),
    [anon_sym_STAR] = ACTIONS(702),
    [anon_sym_default] = ACTIONS(702),
    [anon_sym_LBRACE] = ACTIONS(700),
    [anon_sym_COMMA] = ACTIONS(700),
    [anon_sym_RBRACE] = ACTIONS(700),
    [anon_sym_import] = ACTIONS(702),
    [anon_sym_with] = ACTIONS(702),
    [anon_sym_var] = ACTIONS(702),
    [anon_sym_let] = ACTIONS(702),
    [anon_sym_const] = ACTIONS(702),
    [anon_sym_using] = ACTIONS(702),
    [anon_sym_await] = ACTIONS(702),
    [anon_sym_of] = ACTIONS(702),
    [anon_sym_else] = ACTIONS(702),
    [anon_sym_if] = ACTIONS(702),
    [anon_sym_switch] = ACTIONS(702),
    [anon_sym_for] = ACTIONS(702),
    [anon_sym_LPAREN] = ACTIONS(700),
    [anon_sym_SEMI] = ACTIONS(700),
    [anon_sym_in] = ACTIONS(702),
    [anon_sym_while] = ACTIONS(702),
    [anon_sym_do] = ACTIONS(702),
    [anon_sym_try] = ACTIONS(702),
    [anon_sym_break] = ACTIONS(702),
    [anon_sym_continue] = ACTIONS(702),
    [anon_sym_debugger] = ACTIONS(702),
    [anon_sym_return] = ACTIONS(702),
    [anon_sym_throw] = ACTIONS(702),
    [anon_sym_case] = ACTIONS(702),
    [anon_sym_yield] = ACTIONS(702),
    [anon_sym_LBRACK] = ACTIONS(700),
    [anon_sym_LT] = ACTIONS(702),
    [anon_sym_GT] = ACTIONS(702),
    [anon_sym_DOT] = ACTIONS(702),
    [anon_sym_DQUOTE] = ACTIONS(700),
    [anon_sym_SQUOTE] = ACTIONS(700),
    [anon_sym_class] = ACTIONS(702),
    [anon_sym_async] = ACTIONS(702),
    [anon_sym_function] = ACTIONS(702),
    [sym_optional_chain] = ACTIONS(700),
    [anon_sym_new] = ACTIONS(702),
    [anon_sym_AMP_AMP] = ACTIONS(700),
    [anon_sym_PIPE_PIPE] = ACTIONS(700),
    [anon_sym_GT_GT] = ACTIONS(702),
    [anon_sym_GT_GT_GT] = ACTIONS(700),
    [anon_sym_LT_LT] = ACTIONS(700),
    [anon_sym_AMP] = ACTIONS(702),
    [anon_sym_CARET] = ACTIONS(700),
    [anon_sym_PIPE] = ACTIONS(702),
    [anon_sym_PLUS] = ACTIONS(702),
    [anon_sym_DASH] = ACTIONS(702),
    [anon_sym_SLASH] = ACTIONS(702),
    [anon_sym_PERCENT] = ACTIONS(700),
    [anon_sym_STAR_STAR] = ACTIONS(700),
    [anon_sym_LT_EQ] = ACTIONS(700),
    [anon_sym_EQ_EQ] = ACTIONS(702),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(700),
    [anon_sym_BANG_EQ] = ACTIONS(702),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(700),
    [anon_sym_GT_EQ] = ACTIONS(700),
    [anon_sym_QMARK_QMARK] = ACTIONS(700),
    [anon_sym_instanceof] = ACTIONS(702),
    [anon_sym_BANG] = ACTIONS(702),
    [anon_sym_TILDE] = ACTIONS(700),
    [anon_sym_typeof] = ACTIONS(702),
    [anon_sym_void] = ACTIONS(702),
    [anon_sym_delete] = ACTIONS(702),
    [anon_sym_PLUS_PLUS] = ACTIONS(700),
    [anon_sym_DASH_DASH] = ACTIONS(700),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(700),
    [sym_number] = ACTIONS(700),
    [sym_private_property_identifier] = ACTIONS(700),
    [sym_this] = ACTIONS(702),
    [sym_super] = ACTIONS(702),
    [sym_true] = ACTIONS(702),
    [sym_false] = ACTIONS(702),
    [sym_null] = ACTIONS(702),
    [sym_undefined] = ACTIONS(702),
    [anon_sym_AT] = ACTIONS(700),
    [anon_sym_static] = ACTIONS(702),
    [anon_sym_get] = ACTIONS(702),
    [anon_sym_set] = ACTIONS(702),
    [sym__automatic_semicolon] = ACTIONS(700),
    [sym__ternary_qmark] = ACTIONS(700),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(90)] = {
    [ts_builtin_sym_end] = ACTIONS(704),
    [sym_identifier] = ACTIONS(706),
    [anon_sym_export] = ACTIONS(706),
    [anon_sym_STAR] = ACTIONS(706),
    [anon_sym_default] = ACTIONS(706),
    [anon_sym_LBRACE] = ACTIONS(704),
    [anon_sym_COMMA] = ACTIONS(704),
    [anon_sym_RBRACE] = ACTIONS(704),
    [anon_sym_import] = ACTIONS(706),
    [anon_sym_with] = ACTIONS(706),
    [anon_sym_var] = ACTIONS(706),
    [anon_sym_let] = ACTIONS(706),
    [anon_sym_const] = ACTIONS(706),
    [anon_sym_using] = ACTIONS(706),
    [anon_sym_await] = ACTIONS(706),
    [anon_sym_of] = ACTIONS(706),
    [anon_sym_else] = ACTIONS(706),
    [anon_sym_if] = ACTIONS(706),
    [anon_sym_switch] = ACTIONS(706),
    [anon_sym_for] = ACTIONS(706),
    [anon_sym_LPAREN] = ACTIONS(704),
    [anon_sym_SEMI] = ACTIONS(704),
    [anon_sym_in] = ACTIONS(706),
    [anon_sym_while] = ACTIONS(706),
    [anon_sym_do] = ACTIONS(706),
    [anon_sym_try] = ACTIONS(706),
    [anon_sym_break] = ACTIONS(706),
    [anon_sym_continue] = ACTIONS(706),
    [anon_sym_debugger] = ACTIONS(706),
    [anon_sym_return] = ACTIONS(706),
    [anon_sym_throw] = ACTIONS(706),
    [anon_sym_case] = ACTIONS(706),
    [anon_sym_yield] = ACTIONS(706),
    [anon_sym_LBRACK] = ACTIONS(704),
    [anon_sym_LT] = ACTIONS(706),
    [anon_sym_GT] = ACTIONS(706),
    [anon_sym_DOT] = ACTIONS(706),
    [anon_sym_DQUOTE] = ACTIONS(704),
    [anon_sym_SQUOTE] = ACTIONS(704),
    [anon_sym_class] = ACTIONS(706),
    [anon_sym_async] = ACTIONS(706),
    [anon_sym_function] = ACTIONS(706),
    [sym_optional_chain] = ACTIONS(704),
    [anon_sym_new] = ACTIONS(706),
    [anon_sym_AMP_AMP] = ACTIONS(704),
    [anon_sym_PIPE_PIPE] = ACTIONS(704),
    [anon_sym_GT_GT] = ACTIONS(706),
    [anon_sym_GT_GT_GT] = ACTIONS(704),
    [anon_sym_LT_LT] = ACTIONS(704),
    [anon_sym_AMP] = ACTIONS(706),
    [anon_sym_CARET] = ACTIONS(704),
    [anon_sym_PIPE] = ACTIONS(706),
    [anon_sym_PLUS] = ACTIONS(706),
    [anon_sym_DASH] = ACTIONS(706),
    [anon_sym_SLASH] = ACTIONS(706),
    [anon_sym_PERCENT] = ACTIONS(704),
    [anon_sym_STAR_STAR] = ACTIONS(704),
    [anon_sym_LT_EQ] = ACTIONS(704),
    [anon_sym_EQ_EQ] = ACTIONS(706),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(704),
    [anon_sym_BANG_EQ] = ACTIONS(706),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(704),
    [anon_sym_GT_EQ] = ACTIONS(704),
    [anon_sym_QMARK_QMARK] = ACTIONS(704),
    [anon_sym_instanceof] = ACTIONS(706),
    [anon_sym_BANG] = ACTIONS(706),
    [anon_sym_TILDE] = ACTIONS(704),
    [anon_sym_typeof] = ACTIONS(706),
    [anon_sym_void] = ACTIONS(706),
    [anon_sym_delete] = ACTIONS(706),
    [anon_sym_PLUS_PLUS] = ACTIONS(704),
    [anon_sym_DASH_DASH] = ACTIONS(704),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(704),
    [sym_number] = ACTIONS(704),
    [sym_private_property_identifier] = ACTIONS(704),
    [sym_this] = ACTIONS(706),
    [sym_super] = ACTIONS(706),
    [sym_true] = ACTIONS(706),
    [sym_false] = ACTIONS(706),
    [sym_null] = ACTIONS(706),
    [sym_undefined] = ACTIONS(706),
    [anon_sym_AT] = ACTIONS(704),
    [anon_sym_static] = ACTIONS(706),
    [anon_sym_get] = ACTIONS(706),
    [anon_sym_set] = ACTIONS(706),
    [sym__automatic_semicolon] = ACTIONS(704),
    [sym__ternary_qmark] = ACTIONS(704),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(91)] = {
    [ts_builtin_sym_end] = ACTIONS(708),
    [sym_identifier] = ACTIONS(710),
    [anon_sym_export] = ACTIONS(710),
    [anon_sym_STAR] = ACTIONS(710),
    [anon_sym_default] = ACTIONS(710),
    [anon_sym_LBRACE] = ACTIONS(708),
    [anon_sym_COMMA] = ACTIONS(708),
    [anon_sym_RBRACE] = ACTIONS(708),
    [anon_sym_import] = ACTIONS(710),
    [anon_sym_with] = ACTIONS(710),
    [anon_sym_var] = ACTIONS(710),
    [anon_sym_let] = ACTIONS(710),
    [anon_sym_const] = ACTIONS(710),
    [anon_sym_using] = ACTIONS(710),
    [anon_sym_await] = ACTIONS(710),
    [anon_sym_of] = ACTIONS(710),
    [anon_sym_else] = ACTIONS(710),
    [anon_sym_if] = ACTIONS(710),
    [anon_sym_switch] = ACTIONS(710),
    [anon_sym_for] = ACTIONS(710),
    [anon_sym_LPAREN] = ACTIONS(708),
    [anon_sym_SEMI] = ACTIONS(708),
    [anon_sym_in] = ACTIONS(710),
    [anon_sym_while] = ACTIONS(710),
    [anon_sym_do] = ACTIONS(710),
    [anon_sym_try] = ACTIONS(710),
    [anon_sym_break] = ACTIONS(710),
    [anon_sym_continue] = ACTIONS(710),
    [anon_sym_debugger] = ACTIONS(710),
    [anon_sym_return] = ACTIONS(710),
    [anon_sym_throw] = ACTIONS(710),
    [anon_sym_case] = ACTIONS(710),
    [anon_sym_yield] = ACTIONS(710),
    [anon_sym_LBRACK] = ACTIONS(708),
    [anon_sym_LT] = ACTIONS(710),
    [anon_sym_GT] = ACTIONS(710),
    [anon_sym_DOT] = ACTIONS(710),
    [anon_sym_DQUOTE] = ACTIONS(708),
    [anon_sym_SQUOTE] = ACTIONS(708),
    [anon_sym_class] = ACTIONS(710),
    [anon_sym_async] = ACTIONS(710),
    [anon_sym_function] = ACTIONS(710),
    [sym_optional_chain] = ACTIONS(708),
    [anon_sym_new] = ACTIONS(710),
    [anon_sym_AMP_AMP] = ACTIONS(708),
    [anon_sym_PIPE_PIPE] = ACTIONS(708),
    [anon_sym_GT_GT] = ACTIONS(710),
    [anon_sym_GT_GT_GT] = ACTIONS(708),
    [anon_sym_LT_LT] = ACTIONS(708),
    [anon_sym_AMP] = ACTIONS(710),
    [anon_sym_CARET] = ACTIONS(708),
    [anon_sym_PIPE] = ACTIONS(710),
    [anon_sym_PLUS] = ACTIONS(710),
    [anon_sym_DASH] = ACTIONS(710),
    [anon_sym_SLASH] = ACTIONS(710),
    [anon_sym_PERCENT] = ACTIONS(708),
    [anon_sym_STAR_STAR] = ACTIONS(708),
    [anon_sym_LT_EQ] = ACTIONS(708),
    [anon_sym_EQ_EQ] = ACTIONS(710),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(708),
    [anon_sym_BANG_EQ] = ACTIONS(710),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(708),
    [anon_sym_GT_EQ] = ACTIONS(708),
    [anon_sym_QMARK_QMARK] = ACTIONS(708),
    [anon_sym_instanceof] = ACTIONS(710),
    [anon_sym_BANG] = ACTIONS(710),
    [anon_sym_TILDE] = ACTIONS(708),
    [anon_sym_typeof] = ACTIONS(710),
    [anon_sym_void] = ACTIONS(710),
    [anon_sym_delete] = ACTIONS(710),
    [anon_sym_PLUS_PLUS] = ACTIONS(708),
    [anon_sym_DASH_DASH] = ACTIONS(708),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(708),
    [sym_number] = ACTIONS(708),
    [sym_private_property_identifier] = ACTIONS(708),
    [sym_this] = ACTIONS(710),
    [sym_super] = ACTIONS(710),
    [sym_true] = ACTIONS(710),
    [sym_false] = ACTIONS(710),
    [sym_null] = ACTIONS(710),
    [sym_undefined] = ACTIONS(710),
    [anon_sym_AT] = ACTIONS(708),
    [anon_sym_static] = ACTIONS(710),
    [anon_sym_get] = ACTIONS(710),
    [anon_sym_set] = ACTIONS(710),
    [sym__automatic_semicolon] = ACTIONS(712),
    [sym__ternary_qmark] = ACTIONS(708),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(92)] = {
    [ts_builtin_sym_end] = ACTIONS(708),
    [sym_identifier] = ACTIONS(710),
    [anon_sym_export] = ACTIONS(710),
    [anon_sym_STAR] = ACTIONS(714),
    [anon_sym_default] = ACTIONS(710),
    [anon_sym_LBRACE] = ACTIONS(708),
    [anon_sym_COMMA] = ACTIONS(716),
    [anon_sym_RBRACE] = ACTIONS(708),
    [anon_sym_import] = ACTIONS(710),
    [anon_sym_with] = ACTIONS(710),
    [anon_sym_var] = ACTIONS(710),
    [anon_sym_let] = ACTIONS(710),
    [anon_sym_const] = ACTIONS(710),
    [anon_sym_using] = ACTIONS(710),
    [anon_sym_await] = ACTIONS(710),
    [anon_sym_else] = ACTIONS(710),
    [anon_sym_if] = ACTIONS(710),
    [anon_sym_switch] = ACTIONS(710),
    [anon_sym_for] = ACTIONS(710),
    [anon_sym_LPAREN] = ACTIONS(708),
    [anon_sym_SEMI] = ACTIONS(708),
    [anon_sym_in] = ACTIONS(714),
    [anon_sym_while] = ACTIONS(710),
    [anon_sym_do] = ACTIONS(710),
    [anon_sym_try] = ACTIONS(710),
    [anon_sym_break] = ACTIONS(710),
    [anon_sym_continue] = ACTIONS(710),
    [anon_sym_debugger] = ACTIONS(710),
    [anon_sym_return] = ACTIONS(710),
    [anon_sym_throw] = ACTIONS(710),
    [anon_sym_case] = ACTIONS(710),
    [anon_sym_yield] = ACTIONS(710),
    [anon_sym_EQ] = ACTIONS(718),
    [anon_sym_LBRACK] = ACTIONS(708),
    [anon_sym_LT] = ACTIONS(710),
    [anon_sym_GT] = ACTIONS(714),
    [anon_sym_DOT] = ACTIONS(714),
    [anon_sym_DQUOTE] = ACTIONS(708),
    [anon_sym_SQUOTE] = ACTIONS(708),
    [anon_sym_class] = ACTIONS(710),
    [anon_sym_async] = ACTIONS(710),
    [anon_sym_function] = ACTIONS(710),
    [sym_optional_chain] = ACTIONS(716),
    [anon_sym_new] = ACTIONS(710),
    [anon_sym_AMP_AMP] = ACTIONS(716),
    [anon_sym_PIPE_PIPE] = ACTIONS(716),
    [anon_sym_GT_GT] = ACTIONS(714),
    [anon_sym_GT_GT_GT] = ACTIONS(716),
    [anon_sym_LT_LT] = ACTIONS(716),
    [anon_sym_AMP] = ACTIONS(714),
    [anon_sym_CARET] = ACTIONS(716),
    [anon_sym_PIPE] = ACTIONS(714),
    [anon_sym_PLUS] = ACTIONS(710),
    [anon_sym_DASH] = ACTIONS(710),
    [anon_sym_SLASH] = ACTIONS(710),
    [anon_sym_PERCENT] = ACTIONS(716),
    [anon_sym_STAR_STAR] = ACTIONS(716),
    [anon_sym_LT_EQ] = ACTIONS(716),
    [anon_sym_EQ_EQ] = ACTIONS(714),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(716),
    [anon_sym_BANG_EQ] = ACTIONS(714),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(716),
    [anon_sym_GT_EQ] = ACTIONS(716),
    [anon_sym_QMARK_QMARK] = ACTIONS(716),
    [anon_sym_instanceof] = ACTIONS(714),
    [anon_sym_BANG] = ACTIONS(710),
    [anon_sym_TILDE] = ACTIONS(708),
    [anon_sym_typeof] = ACTIONS(710),
    [anon_sym_void] = ACTIONS(710),
    [anon_sym_delete] = ACTIONS(710),
    [anon_sym_PLUS_PLUS] = ACTIONS(708),
    [anon_sym_DASH_DASH] = ACTIONS(708),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(708),
    [sym_number] = ACTIONS(708),
    [sym_private_property_identifier] = ACTIONS(708),
    [sym_this] = ACTIONS(710),
    [sym_super] = ACTIONS(710),
    [sym_true] = ACTIONS(710),
    [sym_false] = ACTIONS(710),
    [sym_null] = ACTIONS(710),
    [sym_undefined] = ACTIONS(710),
    [anon_sym_AT] = ACTIONS(708),
    [anon_sym_static] = ACTIONS(710),
    [anon_sym_get] = ACTIONS(710),
    [anon_sym_set] = ACTIONS(710),
    [sym__automatic_semicolon] = ACTIONS(720),
    [sym__ternary_qmark] = ACTIONS(716),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(93)] = {
    [ts_builtin_sym_end] = ACTIONS(722),
    [sym_identifier] = ACTIONS(724),
    [anon_sym_export] = ACTIONS(724),
    [anon_sym_STAR] = ACTIONS(726),
    [anon_sym_default] = ACTIONS(724),
    [anon_sym_LBRACE] = ACTIONS(722),
    [anon_sym_COMMA] = ACTIONS(728),
    [anon_sym_RBRACE] = ACTIONS(722),
    [anon_sym_import] = ACTIONS(724),
    [anon_sym_with] = ACTIONS(724),
    [anon_sym_var] = ACTIONS(724),
    [anon_sym_let] = ACTIONS(724),
    [anon_sym_const] = ACTIONS(724),
    [anon_sym_using] = ACTIONS(724),
    [anon_sym_await] = ACTIONS(724),
    [anon_sym_else] = ACTIONS(724),
    [anon_sym_if] = ACTIONS(724),
    [anon_sym_switch] = ACTIONS(724),
    [anon_sym_for] = ACTIONS(724),
    [anon_sym_LPAREN] = ACTIONS(722),
    [anon_sym_SEMI] = ACTIONS(722),
    [anon_sym_in] = ACTIONS(726),
    [anon_sym_while] = ACTIONS(724),
    [anon_sym_do] = ACTIONS(724),
    [anon_sym_try] = ACTIONS(724),
    [anon_sym_break] = ACTIONS(724),
    [anon_sym_continue] = ACTIONS(724),
    [anon_sym_debugger] = ACTIONS(724),
    [anon_sym_return] = ACTIONS(724),
    [anon_sym_throw] = ACTIONS(724),
    [anon_sym_case] = ACTIONS(724),
    [anon_sym_yield] = ACTIONS(724),
    [anon_sym_LBRACK] = ACTIONS(722),
    [anon_sym_LT] = ACTIONS(724),
    [anon_sym_GT] = ACTIONS(726),
    [anon_sym_DOT] = ACTIONS(726),
    [anon_sym_DQUOTE] = ACTIONS(722),
    [anon_sym_SQUOTE] = ACTIONS(722),
    [anon_sym_class] = ACTIONS(724),
    [anon_sym_async] = ACTIONS(724),
    [anon_sym_function] = ACTIONS(724),
    [sym_optional_chain] = ACTIONS(728),
    [anon_sym_new] = ACTIONS(724),
    [anon_sym_AMP_AMP] = ACTIONS(728),
    [anon_sym_PIPE_PIPE] = ACTIONS(728),
    [anon_sym_GT_GT] = ACTIONS(726),
    [anon_sym_GT_GT_GT] = ACTIONS(728),
    [anon_sym_LT_LT] = ACTIONS(728),
    [anon_sym_AMP] = ACTIONS(726),
    [anon_sym_CARET] = ACTIONS(728),
    [anon_sym_PIPE] = ACTIONS(726),
    [anon_sym_PLUS] = ACTIONS(724),
    [anon_sym_DASH] = ACTIONS(724),
    [anon_sym_SLASH] = ACTIONS(724),
    [anon_sym_PERCENT] = ACTIONS(728),
    [anon_sym_STAR_STAR] = ACTIONS(728),
    [anon_sym_LT_EQ] = ACTIONS(728),
    [anon_sym_EQ_EQ] = ACTIONS(726),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(728),
    [anon_sym_BANG_EQ] = ACTIONS(726),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(728),
    [anon_sym_GT_EQ] = ACTIONS(728),
    [anon_sym_QMARK_QMARK] = ACTIONS(728),
    [anon_sym_instanceof] = ACTIONS(726),
    [anon_sym_BANG] = ACTIONS(724),
    [anon_sym_TILDE] = ACTIONS(722),
    [anon_sym_typeof] = ACTIONS(724),
    [anon_sym_void] = ACTIONS(724),
    [anon_sym_delete] = ACTIONS(724),
    [anon_sym_PLUS_PLUS] = ACTIONS(722),
    [anon_sym_DASH_DASH] = ACTIONS(722),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(722),
    [sym_number] = ACTIONS(722),
    [sym_private_property_identifier] = ACTIONS(722),
    [sym_this] = ACTIONS(724),
    [sym_super] = ACTIONS(724),
    [sym_true] = ACTIONS(724),
    [sym_false] = ACTIONS(724),
    [sym_null] = ACTIONS(724),
    [sym_undefined] = ACTIONS(724),
    [anon_sym_AT] = ACTIONS(722),
    [anon_sym_static] = ACTIONS(724),
    [anon_sym_get] = ACTIONS(724),
    [anon_sym_set] = ACTIONS(724),
    [sym__automatic_semicolon] = ACTIONS(730),
    [sym__ternary_qmark] = ACTIONS(728),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(94)] = {
    [ts_builtin_sym_end] = ACTIONS(732),
    [sym_identifier] = ACTIONS(734),
    [anon_sym_export] = ACTIONS(734),
    [anon_sym_STAR] = ACTIONS(736),
    [anon_sym_default] = ACTIONS(734),
    [anon_sym_LBRACE] = ACTIONS(732),
    [anon_sym_COMMA] = ACTIONS(738),
    [anon_sym_RBRACE] = ACTIONS(732),
    [anon_sym_import] = ACTIONS(734),
    [anon_sym_with] = ACTIONS(734),
    [anon_sym_var] = ACTIONS(734),
    [anon_sym_let] = ACTIONS(734),
    [anon_sym_const] = ACTIONS(734),
    [anon_sym_using] = ACTIONS(734),
    [anon_sym_await] = ACTIONS(734),
    [anon_sym_else] = ACTIONS(734),
    [anon_sym_if] = ACTIONS(734),
    [anon_sym_switch] = ACTIONS(734),
    [anon_sym_for] = ACTIONS(734),
    [anon_sym_LPAREN] = ACTIONS(732),
    [anon_sym_SEMI] = ACTIONS(732),
    [anon_sym_in] = ACTIONS(736),
    [anon_sym_while] = ACTIONS(734),
    [anon_sym_do] = ACTIONS(734),
    [anon_sym_try] = ACTIONS(734),
    [anon_sym_break] = ACTIONS(734),
    [anon_sym_continue] = ACTIONS(734),
    [anon_sym_debugger] = ACTIONS(734),
    [anon_sym_return] = ACTIONS(734),
    [anon_sym_throw] = ACTIONS(734),
    [anon_sym_case] = ACTIONS(734),
    [anon_sym_yield] = ACTIONS(734),
    [anon_sym_LBRACK] = ACTIONS(732),
    [anon_sym_LT] = ACTIONS(734),
    [anon_sym_GT] = ACTIONS(736),
    [anon_sym_DOT] = ACTIONS(736),
    [anon_sym_DQUOTE] = ACTIONS(732),
    [anon_sym_SQUOTE] = ACTIONS(732),
    [anon_sym_class] = ACTIONS(734),
    [anon_sym_async] = ACTIONS(734),
    [anon_sym_function] = ACTIONS(734),
    [sym_optional_chain] = ACTIONS(738),
    [anon_sym_new] = ACTIONS(734),
    [anon_sym_AMP_AMP] = ACTIONS(738),
    [anon_sym_PIPE_PIPE] = ACTIONS(738),
    [anon_sym_GT_GT] = ACTIONS(736),
    [anon_sym_GT_GT_GT] = ACTIONS(738),
    [anon_sym_LT_LT] = ACTIONS(738),
    [anon_sym_AMP] = ACTIONS(736),
    [anon_sym_CARET] = ACTIONS(738),
    [anon_sym_PIPE] = ACTIONS(736),
    [anon_sym_PLUS] = ACTIONS(734),
    [anon_sym_DASH] = ACTIONS(734),
    [anon_sym_SLASH] = ACTIONS(734),
    [anon_sym_PERCENT] = ACTIONS(738),
    [anon_sym_STAR_STAR] = ACTIONS(738),
    [anon_sym_LT_EQ] = ACTIONS(738),
    [anon_sym_EQ_EQ] = ACTIONS(736),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(738),
    [anon_sym_BANG_EQ] = ACTIONS(736),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(738),
    [anon_sym_GT_EQ] = ACTIONS(738),
    [anon_sym_QMARK_QMARK] = ACTIONS(738),
    [anon_sym_instanceof] = ACTIONS(736),
    [anon_sym_BANG] = ACTIONS(734),
    [anon_sym_TILDE] = ACTIONS(732),
    [anon_sym_typeof] = ACTIONS(734),
    [anon_sym_void] = ACTIONS(734),
    [anon_sym_delete] = ACTIONS(734),
    [anon_sym_PLUS_PLUS] = ACTIONS(732),
    [anon_sym_DASH_DASH] = ACTIONS(732),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(732),
    [sym_number] = ACTIONS(732),
    [sym_private_property_identifier] = ACTIONS(732),
    [sym_this] = ACTIONS(734),
    [sym_super] = ACTIONS(734),
    [sym_true] = ACTIONS(734),
    [sym_false] = ACTIONS(734),
    [sym_null] = ACTIONS(734),
    [sym_undefined] = ACTIONS(734),
    [anon_sym_AT] = ACTIONS(732),
    [anon_sym_static] = ACTIONS(734),
    [anon_sym_get] = ACTIONS(734),
    [anon_sym_set] = ACTIONS(734),
    [sym__automatic_semicolon] = ACTIONS(740),
    [sym__ternary_qmark] = ACTIONS(738),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(95)] = {
    [ts_builtin_sym_end] = ACTIONS(742),
    [sym_identifier] = ACTIONS(744),
    [anon_sym_export] = ACTIONS(744),
    [anon_sym_STAR] = ACTIONS(746),
    [anon_sym_default] = ACTIONS(744),
    [anon_sym_LBRACE] = ACTIONS(742),
    [anon_sym_COMMA] = ACTIONS(748),
    [anon_sym_RBRACE] = ACTIONS(742),
    [anon_sym_import] = ACTIONS(744),
    [anon_sym_with] = ACTIONS(744),
    [anon_sym_var] = ACTIONS(744),
    [anon_sym_let] = ACTIONS(744),
    [anon_sym_const] = ACTIONS(744),
    [anon_sym_using] = ACTIONS(744),
    [anon_sym_await] = ACTIONS(744),
    [anon_sym_else] = ACTIONS(744),
    [anon_sym_if] = ACTIONS(744),
    [anon_sym_switch] = ACTIONS(744),
    [anon_sym_for] = ACTIONS(744),
    [anon_sym_LPAREN] = ACTIONS(742),
    [anon_sym_SEMI] = ACTIONS(742),
    [anon_sym_in] = ACTIONS(746),
    [anon_sym_while] = ACTIONS(744),
    [anon_sym_do] = ACTIONS(744),
    [anon_sym_try] = ACTIONS(744),
    [anon_sym_break] = ACTIONS(744),
    [anon_sym_continue] = ACTIONS(744),
    [anon_sym_debugger] = ACTIONS(744),
    [anon_sym_return] = ACTIONS(744),
    [anon_sym_throw] = ACTIONS(744),
    [anon_sym_case] = ACTIONS(744),
    [anon_sym_yield] = ACTIONS(744),
    [anon_sym_LBRACK] = ACTIONS(742),
    [anon_sym_LT] = ACTIONS(744),
    [anon_sym_GT] = ACTIONS(746),
    [anon_sym_DOT] = ACTIONS(746),
    [anon_sym_DQUOTE] = ACTIONS(742),
    [anon_sym_SQUOTE] = ACTIONS(742),
    [anon_sym_class] = ACTIONS(744),
    [anon_sym_async] = ACTIONS(744),
    [anon_sym_function] = ACTIONS(744),
    [sym_optional_chain] = ACTIONS(748),
    [anon_sym_new] = ACTIONS(744),
    [anon_sym_AMP_AMP] = ACTIONS(748),
    [anon_sym_PIPE_PIPE] = ACTIONS(748),
    [anon_sym_GT_GT] = ACTIONS(746),
    [anon_sym_GT_GT_GT] = ACTIONS(748),
    [anon_sym_LT_LT] = ACTIONS(748),
    [anon_sym_AMP] = ACTIONS(746),
    [anon_sym_CARET] = ACTIONS(748),
    [anon_sym_PIPE] = ACTIONS(746),
    [anon_sym_PLUS] = ACTIONS(744),
    [anon_sym_DASH] = ACTIONS(744),
    [anon_sym_SLASH] = ACTIONS(744),
    [anon_sym_PERCENT] = ACTIONS(748),
    [anon_sym_STAR_STAR] = ACTIONS(748),
    [anon_sym_LT_EQ] = ACTIONS(748),
    [anon_sym_EQ_EQ] = ACTIONS(746),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(748),
    [anon_sym_BANG_EQ] = ACTIONS(746),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(748),
    [anon_sym_GT_EQ] = ACTIONS(748),
    [anon_sym_QMARK_QMARK] = ACTIONS(748),
    [anon_sym_instanceof] = ACTIONS(746),
    [anon_sym_BANG] = ACTIONS(744),
    [anon_sym_TILDE] = ACTIONS(742),
    [anon_sym_typeof] = ACTIONS(744),
    [anon_sym_void] = ACTIONS(744),
    [anon_sym_delete] = ACTIONS(744),
    [anon_sym_PLUS_PLUS] = ACTIONS(742),
    [anon_sym_DASH_DASH] = ACTIONS(742),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(742),
    [sym_number] = ACTIONS(742),
    [sym_private_property_identifier] = ACTIONS(742),
    [sym_this] = ACTIONS(744),
    [sym_super] = ACTIONS(744),
    [sym_true] = ACTIONS(744),
    [sym_false] = ACTIONS(744),
    [sym_null] = ACTIONS(744),
    [sym_undefined] = ACTIONS(744),
    [anon_sym_AT] = ACTIONS(742),
    [anon_sym_static] = ACTIONS(744),
    [anon_sym_get] = ACTIONS(744),
    [anon_sym_set] = ACTIONS(744),
    [sym__automatic_semicolon] = ACTIONS(750),
    [sym__ternary_qmark] = ACTIONS(748),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(96)] = {
    [ts_builtin_sym_end] = ACTIONS(752),
    [sym_identifier] = ACTIONS(754),
    [anon_sym_export] = ACTIONS(754),
    [anon_sym_STAR] = ACTIONS(756),
    [anon_sym_default] = ACTIONS(754),
    [anon_sym_LBRACE] = ACTIONS(752),
    [anon_sym_COMMA] = ACTIONS(758),
    [anon_sym_RBRACE] = ACTIONS(752),
    [anon_sym_import] = ACTIONS(754),
    [anon_sym_with] = ACTIONS(754),
    [anon_sym_var] = ACTIONS(754),
    [anon_sym_let] = ACTIONS(754),
    [anon_sym_const] = ACTIONS(754),
    [anon_sym_using] = ACTIONS(754),
    [anon_sym_await] = ACTIONS(754),
    [anon_sym_else] = ACTIONS(754),
    [anon_sym_if] = ACTIONS(754),
    [anon_sym_switch] = ACTIONS(754),
    [anon_sym_for] = ACTIONS(754),
    [anon_sym_LPAREN] = ACTIONS(752),
    [anon_sym_SEMI] = ACTIONS(752),
    [anon_sym_in] = ACTIONS(756),
    [anon_sym_while] = ACTIONS(754),
    [anon_sym_do] = ACTIONS(754),
    [anon_sym_try] = ACTIONS(754),
    [anon_sym_break] = ACTIONS(754),
    [anon_sym_continue] = ACTIONS(754),
    [anon_sym_debugger] = ACTIONS(754),
    [anon_sym_return] = ACTIONS(754),
    [anon_sym_throw] = ACTIONS(754),
    [anon_sym_case] = ACTIONS(754),
    [anon_sym_yield] = ACTIONS(754),
    [anon_sym_LBRACK] = ACTIONS(752),
    [anon_sym_LT] = ACTIONS(754),
    [anon_sym_GT] = ACTIONS(756),
    [anon_sym_DOT] = ACTIONS(756),
    [anon_sym_DQUOTE] = ACTIONS(752),
    [anon_sym_SQUOTE] = ACTIONS(752),
    [anon_sym_class] = ACTIONS(754),
    [anon_sym_async] = ACTIONS(754),
    [anon_sym_function] = ACTIONS(754),
    [sym_optional_chain] = ACTIONS(758),
    [anon_sym_new] = ACTIONS(754),
    [anon_sym_AMP_AMP] = ACTIONS(758),
    [anon_sym_PIPE_PIPE] = ACTIONS(758),
    [anon_sym_GT_GT] = ACTIONS(756),
    [anon_sym_GT_GT_GT] = ACTIONS(758),
    [anon_sym_LT_LT] = ACTIONS(758),
    [anon_sym_AMP] = ACTIONS(756),
    [anon_sym_CARET] = ACTIONS(758),
    [anon_sym_PIPE] = ACTIONS(756),
    [anon_sym_PLUS] = ACTIONS(754),
    [anon_sym_DASH] = ACTIONS(754),
    [anon_sym_SLASH] = ACTIONS(754),
    [anon_sym_PERCENT] = ACTIONS(758),
    [anon_sym_STAR_STAR] = ACTIONS(758),
    [anon_sym_LT_EQ] = ACTIONS(758),
    [anon_sym_EQ_EQ] = ACTIONS(756),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(758),
    [anon_sym_BANG_EQ] = ACTIONS(756),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(758),
    [anon_sym_GT_EQ] = ACTIONS(758),
    [anon_sym_QMARK_QMARK] = ACTIONS(758),
    [anon_sym_instanceof] = ACTIONS(756),
    [anon_sym_BANG] = ACTIONS(754),
    [anon_sym_TILDE] = ACTIONS(752),
    [anon_sym_typeof] = ACTIONS(754),
    [anon_sym_void] = ACTIONS(754),
    [anon_sym_delete] = ACTIONS(754),
    [anon_sym_PLUS_PLUS] = ACTIONS(752),
    [anon_sym_DASH_DASH] = ACTIONS(752),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(752),
    [sym_number] = ACTIONS(752),
    [sym_private_property_identifier] = ACTIONS(752),
    [sym_this] = ACTIONS(754),
    [sym_super] = ACTIONS(754),
    [sym_true] = ACTIONS(754),
    [sym_false] = ACTIONS(754),
    [sym_null] = ACTIONS(754),
    [sym_undefined] = ACTIONS(754),
    [anon_sym_AT] = ACTIONS(752),
    [anon_sym_static] = ACTIONS(754),
    [anon_sym_get] = ACTIONS(754),
    [anon_sym_set] = ACTIONS(754),
    [sym__automatic_semicolon] = ACTIONS(760),
    [sym__ternary_qmark] = ACTIONS(758),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(97)] = {
    [ts_builtin_sym_end] = ACTIONS(762),
    [sym_identifier] = ACTIONS(764),
    [anon_sym_export] = ACTIONS(764),
    [anon_sym_STAR] = ACTIONS(766),
    [anon_sym_default] = ACTIONS(764),
    [anon_sym_LBRACE] = ACTIONS(762),
    [anon_sym_COMMA] = ACTIONS(768),
    [anon_sym_RBRACE] = ACTIONS(762),
    [anon_sym_import] = ACTIONS(764),
    [anon_sym_with] = ACTIONS(764),
    [anon_sym_var] = ACTIONS(764),
    [anon_sym_let] = ACTIONS(764),
    [anon_sym_const] = ACTIONS(764),
    [anon_sym_using] = ACTIONS(764),
    [anon_sym_await] = ACTIONS(764),
    [anon_sym_else] = ACTIONS(764),
    [anon_sym_if] = ACTIONS(764),
    [anon_sym_switch] = ACTIONS(764),
    [anon_sym_for] = ACTIONS(764),
    [anon_sym_LPAREN] = ACTIONS(762),
    [anon_sym_SEMI] = ACTIONS(762),
    [anon_sym_in] = ACTIONS(766),
    [anon_sym_while] = ACTIONS(764),
    [anon_sym_do] = ACTIONS(764),
    [anon_sym_try] = ACTIONS(764),
    [anon_sym_break] = ACTIONS(764),
    [anon_sym_continue] = ACTIONS(764),
    [anon_sym_debugger] = ACTIONS(764),
    [anon_sym_return] = ACTIONS(764),
    [anon_sym_throw] = ACTIONS(764),
    [anon_sym_case] = ACTIONS(764),
    [anon_sym_yield] = ACTIONS(764),
    [anon_sym_LBRACK] = ACTIONS(762),
    [anon_sym_LT] = ACTIONS(764),
    [anon_sym_GT] = ACTIONS(766),
    [anon_sym_DOT] = ACTIONS(766),
    [anon_sym_DQUOTE] = ACTIONS(762),
    [anon_sym_SQUOTE] = ACTIONS(762),
    [anon_sym_class] = ACTIONS(764),
    [anon_sym_async] = ACTIONS(764),
    [anon_sym_function] = ACTIONS(764),
    [sym_optional_chain] = ACTIONS(768),
    [anon_sym_new] = ACTIONS(764),
    [anon_sym_AMP_AMP] = ACTIONS(768),
    [anon_sym_PIPE_PIPE] = ACTIONS(768),
    [anon_sym_GT_GT] = ACTIONS(766),
    [anon_sym_GT_GT_GT] = ACTIONS(768),
    [anon_sym_LT_LT] = ACTIONS(768),
    [anon_sym_AMP] = ACTIONS(766),
    [anon_sym_CARET] = ACTIONS(768),
    [anon_sym_PIPE] = ACTIONS(766),
    [anon_sym_PLUS] = ACTIONS(764),
    [anon_sym_DASH] = ACTIONS(764),
    [anon_sym_SLASH] = ACTIONS(764),
    [anon_sym_PERCENT] = ACTIONS(768),
    [anon_sym_STAR_STAR] = ACTIONS(768),
    [anon_sym_LT_EQ] = ACTIONS(768),
    [anon_sym_EQ_EQ] = ACTIONS(766),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(768),
    [anon_sym_BANG_EQ] = ACTIONS(766),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(768),
    [anon_sym_GT_EQ] = ACTIONS(768),
    [anon_sym_QMARK_QMARK] = ACTIONS(768),
    [anon_sym_instanceof] = ACTIONS(766),
    [anon_sym_BANG] = ACTIONS(764),
    [anon_sym_TILDE] = ACTIONS(762),
    [anon_sym_typeof] = ACTIONS(764),
    [anon_sym_void] = ACTIONS(764),
    [anon_sym_delete] = ACTIONS(764),
    [anon_sym_PLUS_PLUS] = ACTIONS(762),
    [anon_sym_DASH_DASH] = ACTIONS(762),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(762),
    [sym_number] = ACTIONS(762),
    [sym_private_property_identifier] = ACTIONS(762),
    [sym_this] = ACTIONS(764),
    [sym_super] = ACTIONS(764),
    [sym_true] = ACTIONS(764),
    [sym_false] = ACTIONS(764),
    [sym_null] = ACTIONS(764),
    [sym_undefined] = ACTIONS(764),
    [anon_sym_AT] = ACTIONS(762),
    [anon_sym_static] = ACTIONS(764),
    [anon_sym_get] = ACTIONS(764),
    [anon_sym_set] = ACTIONS(764),
    [sym__automatic_semicolon] = ACTIONS(770),
    [sym__ternary_qmark] = ACTIONS(768),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(98)] = {
    [ts_builtin_sym_end] = ACTIONS(772),
    [sym_identifier] = ACTIONS(774),
    [anon_sym_export] = ACTIONS(774),
    [anon_sym_STAR] = ACTIONS(776),
    [anon_sym_default] = ACTIONS(774),
    [anon_sym_LBRACE] = ACTIONS(772),
    [anon_sym_COMMA] = ACTIONS(778),
    [anon_sym_RBRACE] = ACTIONS(772),
    [anon_sym_import] = ACTIONS(774),
    [anon_sym_with] = ACTIONS(774),
    [anon_sym_var] = ACTIONS(774),
    [anon_sym_let] = ACTIONS(774),
    [anon_sym_const] = ACTIONS(774),
    [anon_sym_using] = ACTIONS(774),
    [anon_sym_await] = ACTIONS(774),
    [anon_sym_else] = ACTIONS(774),
    [anon_sym_if] = ACTIONS(774),
    [anon_sym_switch] = ACTIONS(774),
    [anon_sym_for] = ACTIONS(774),
    [anon_sym_LPAREN] = ACTIONS(772),
    [anon_sym_SEMI] = ACTIONS(772),
    [anon_sym_in] = ACTIONS(776),
    [anon_sym_while] = ACTIONS(774),
    [anon_sym_do] = ACTIONS(774),
    [anon_sym_try] = ACTIONS(774),
    [anon_sym_break] = ACTIONS(774),
    [anon_sym_continue] = ACTIONS(774),
    [anon_sym_debugger] = ACTIONS(774),
    [anon_sym_return] = ACTIONS(774),
    [anon_sym_throw] = ACTIONS(774),
    [anon_sym_case] = ACTIONS(774),
    [anon_sym_yield] = ACTIONS(774),
    [anon_sym_LBRACK] = ACTIONS(772),
    [anon_sym_LT] = ACTIONS(774),
    [anon_sym_GT] = ACTIONS(776),
    [anon_sym_DOT] = ACTIONS(776),
    [anon_sym_DQUOTE] = ACTIONS(772),
    [anon_sym_SQUOTE] = ACTIONS(772),
    [anon_sym_class] = ACTIONS(774),
    [anon_sym_async] = ACTIONS(774),
    [anon_sym_function] = ACTIONS(774),
    [sym_optional_chain] = ACTIONS(778),
    [anon_sym_new] = ACTIONS(774),
    [anon_sym_AMP_AMP] = ACTIONS(778),
    [anon_sym_PIPE_PIPE] = ACTIONS(778),
    [anon_sym_GT_GT] = ACTIONS(776),
    [anon_sym_GT_GT_GT] = ACTIONS(778),
    [anon_sym_LT_LT] = ACTIONS(778),
    [anon_sym_AMP] = ACTIONS(776),
    [anon_sym_CARET] = ACTIONS(778),
    [anon_sym_PIPE] = ACTIONS(776),
    [anon_sym_PLUS] = ACTIONS(774),
    [anon_sym_DASH] = ACTIONS(774),
    [anon_sym_SLASH] = ACTIONS(774),
    [anon_sym_PERCENT] = ACTIONS(778),
    [anon_sym_STAR_STAR] = ACTIONS(778),
    [anon_sym_LT_EQ] = ACTIONS(778),
    [anon_sym_EQ_EQ] = ACTIONS(776),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(778),
    [anon_sym_BANG_EQ] = ACTIONS(776),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(778),
    [anon_sym_GT_EQ] = ACTIONS(778),
    [anon_sym_QMARK_QMARK] = ACTIONS(778),
    [anon_sym_instanceof] = ACTIONS(776),
    [anon_sym_BANG] = ACTIONS(774),
    [anon_sym_TILDE] = ACTIONS(772),
    [anon_sym_typeof] = ACTIONS(774),
    [anon_sym_void] = ACTIONS(774),
    [anon_sym_delete] = ACTIONS(774),
    [anon_sym_PLUS_PLUS] = ACTIONS(772),
    [anon_sym_DASH_DASH] = ACTIONS(772),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(772),
    [sym_number] = ACTIONS(772),
    [sym_private_property_identifier] = ACTIONS(772),
    [sym_this] = ACTIONS(774),
    [sym_super] = ACTIONS(774),
    [sym_true] = ACTIONS(774),
    [sym_false] = ACTIONS(774),
    [sym_null] = ACTIONS(774),
    [sym_undefined] = ACTIONS(774),
    [anon_sym_AT] = ACTIONS(772),
    [anon_sym_static] = ACTIONS(774),
    [anon_sym_get] = ACTIONS(774),
    [anon_sym_set] = ACTIONS(774),
    [sym__automatic_semicolon] = ACTIONS(780),
    [sym__ternary_qmark] = ACTIONS(778),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(99)] = {
    [ts_builtin_sym_end] = ACTIONS(782),
    [sym_identifier] = ACTIONS(784),
    [anon_sym_export] = ACTIONS(784),
    [anon_sym_STAR] = ACTIONS(786),
    [anon_sym_default] = ACTIONS(784),
    [anon_sym_LBRACE] = ACTIONS(782),
    [anon_sym_COMMA] = ACTIONS(788),
    [anon_sym_RBRACE] = ACTIONS(782),
    [anon_sym_import] = ACTIONS(784),
    [anon_sym_with] = ACTIONS(784),
    [anon_sym_var] = ACTIONS(784),
    [anon_sym_let] = ACTIONS(784),
    [anon_sym_const] = ACTIONS(784),
    [anon_sym_using] = ACTIONS(784),
    [anon_sym_await] = ACTIONS(784),
    [anon_sym_else] = ACTIONS(784),
    [anon_sym_if] = ACTIONS(784),
    [anon_sym_switch] = ACTIONS(784),
    [anon_sym_for] = ACTIONS(784),
    [anon_sym_LPAREN] = ACTIONS(782),
    [anon_sym_SEMI] = ACTIONS(782),
    [anon_sym_in] = ACTIONS(786),
    [anon_sym_while] = ACTIONS(784),
    [anon_sym_do] = ACTIONS(784),
    [anon_sym_try] = ACTIONS(784),
    [anon_sym_break] = ACTIONS(784),
    [anon_sym_continue] = ACTIONS(784),
    [anon_sym_debugger] = ACTIONS(784),
    [anon_sym_return] = ACTIONS(784),
    [anon_sym_throw] = ACTIONS(784),
    [anon_sym_case] = ACTIONS(784),
    [anon_sym_yield] = ACTIONS(784),
    [anon_sym_LBRACK] = ACTIONS(782),
    [anon_sym_LT] = ACTIONS(784),
    [anon_sym_GT] = ACTIONS(786),
    [anon_sym_DOT] = ACTIONS(786),
    [anon_sym_DQUOTE] = ACTIONS(782),
    [anon_sym_SQUOTE] = ACTIONS(782),
    [anon_sym_class] = ACTIONS(784),
    [anon_sym_async] = ACTIONS(784),
    [anon_sym_function] = ACTIONS(784),
    [sym_optional_chain] = ACTIONS(788),
    [anon_sym_new] = ACTIONS(784),
    [anon_sym_AMP_AMP] = ACTIONS(788),
    [anon_sym_PIPE_PIPE] = ACTIONS(788),
    [anon_sym_GT_GT] = ACTIONS(786),
    [anon_sym_GT_GT_GT] = ACTIONS(788),
    [anon_sym_LT_LT] = ACTIONS(788),
    [anon_sym_AMP] = ACTIONS(786),
    [anon_sym_CARET] = ACTIONS(788),
    [anon_sym_PIPE] = ACTIONS(786),
    [anon_sym_PLUS] = ACTIONS(784),
    [anon_sym_DASH] = ACTIONS(784),
    [anon_sym_SLASH] = ACTIONS(784),
    [anon_sym_PERCENT] = ACTIONS(788),
    [anon_sym_STAR_STAR] = ACTIONS(788),
    [anon_sym_LT_EQ] = ACTIONS(788),
    [anon_sym_EQ_EQ] = ACTIONS(786),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(788),
    [anon_sym_BANG_EQ] = ACTIONS(786),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(788),
    [anon_sym_GT_EQ] = ACTIONS(788),
    [anon_sym_QMARK_QMARK] = ACTIONS(788),
    [anon_sym_instanceof] = ACTIONS(786),
    [anon_sym_BANG] = ACTIONS(784),
    [anon_sym_TILDE] = ACTIONS(782),
    [anon_sym_typeof] = ACTIONS(784),
    [anon_sym_void] = ACTIONS(784),
    [anon_sym_delete] = ACTIONS(784),
    [anon_sym_PLUS_PLUS] = ACTIONS(782),
    [anon_sym_DASH_DASH] = ACTIONS(782),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(782),
    [sym_number] = ACTIONS(782),
    [sym_private_property_identifier] = ACTIONS(782),
    [sym_this] = ACTIONS(784),
    [sym_super] = ACTIONS(784),
    [sym_true] = ACTIONS(784),
    [sym_false] = ACTIONS(784),
    [sym_null] = ACTIONS(784),
    [sym_undefined] = ACTIONS(784),
    [anon_sym_AT] = ACTIONS(782),
    [anon_sym_static] = ACTIONS(784),
    [anon_sym_get] = ACTIONS(784),
    [anon_sym_set] = ACTIONS(784),
    [sym__automatic_semicolon] = ACTIONS(790),
    [sym__ternary_qmark] = ACTIONS(788),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(100)] = {
    [ts_builtin_sym_end] = ACTIONS(792),
    [sym_identifier] = ACTIONS(794),
    [anon_sym_export] = ACTIONS(794),
    [anon_sym_STAR] = ACTIONS(796),
    [anon_sym_default] = ACTIONS(794),
    [anon_sym_LBRACE] = ACTIONS(792),
    [anon_sym_COMMA] = ACTIONS(798),
    [anon_sym_RBRACE] = ACTIONS(792),
    [anon_sym_import] = ACTIONS(794),
    [anon_sym_with] = ACTIONS(794),
    [anon_sym_var] = ACTIONS(794),
    [anon_sym_let] = ACTIONS(794),
    [anon_sym_const] = ACTIONS(794),
    [anon_sym_using] = ACTIONS(794),
    [anon_sym_await] = ACTIONS(794),
    [anon_sym_else] = ACTIONS(794),
    [anon_sym_if] = ACTIONS(794),
    [anon_sym_switch] = ACTIONS(794),
    [anon_sym_for] = ACTIONS(794),
    [anon_sym_LPAREN] = ACTIONS(792),
    [anon_sym_SEMI] = ACTIONS(792),
    [anon_sym_in] = ACTIONS(796),
    [anon_sym_while] = ACTIONS(794),
    [anon_sym_do] = ACTIONS(794),
    [anon_sym_try] = ACTIONS(794),
    [anon_sym_break] = ACTIONS(794),
    [anon_sym_continue] = ACTIONS(794),
    [anon_sym_debugger] = ACTIONS(794),
    [anon_sym_return] = ACTIONS(794),
    [anon_sym_throw] = ACTIONS(794),
    [anon_sym_case] = ACTIONS(794),
    [anon_sym_yield] = ACTIONS(794),
    [anon_sym_LBRACK] = ACTIONS(792),
    [anon_sym_LT] = ACTIONS(794),
    [anon_sym_GT] = ACTIONS(796),
    [anon_sym_DOT] = ACTIONS(796),
    [anon_sym_DQUOTE] = ACTIONS(792),
    [anon_sym_SQUOTE] = ACTIONS(792),
    [anon_sym_class] = ACTIONS(794),
    [anon_sym_async] = ACTIONS(794),
    [anon_sym_function] = ACTIONS(794),
    [sym_optional_chain] = ACTIONS(798),
    [anon_sym_new] = ACTIONS(794),
    [anon_sym_AMP_AMP] = ACTIONS(798),
    [anon_sym_PIPE_PIPE] = ACTIONS(798),
    [anon_sym_GT_GT] = ACTIONS(796),
    [anon_sym_GT_GT_GT] = ACTIONS(798),
    [anon_sym_LT_LT] = ACTIONS(798),
    [anon_sym_AMP] = ACTIONS(796),
    [anon_sym_CARET] = ACTIONS(798),
    [anon_sym_PIPE] = ACTIONS(796),
    [anon_sym_PLUS] = ACTIONS(794),
    [anon_sym_DASH] = ACTIONS(794),
    [anon_sym_SLASH] = ACTIONS(794),
    [anon_sym_PERCENT] = ACTIONS(798),
    [anon_sym_STAR_STAR] = ACTIONS(798),
    [anon_sym_LT_EQ] = ACTIONS(798),
    [anon_sym_EQ_EQ] = ACTIONS(796),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(798),
    [anon_sym_BANG_EQ] = ACTIONS(796),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(798),
    [anon_sym_GT_EQ] = ACTIONS(798),
    [anon_sym_QMARK_QMARK] = ACTIONS(798),
    [anon_sym_instanceof] = ACTIONS(796),
    [anon_sym_BANG] = ACTIONS(794),
    [anon_sym_TILDE] = ACTIONS(792),
    [anon_sym_typeof] = ACTIONS(794),
    [anon_sym_void] = ACTIONS(794),
    [anon_sym_delete] = ACTIONS(794),
    [anon_sym_PLUS_PLUS] = ACTIONS(792),
    [anon_sym_DASH_DASH] = ACTIONS(792),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(792),
    [sym_number] = ACTIONS(792),
    [sym_private_property_identifier] = ACTIONS(792),
    [sym_this] = ACTIONS(794),
    [sym_super] = ACTIONS(794),
    [sym_true] = ACTIONS(794),
    [sym_false] = ACTIONS(794),
    [sym_null] = ACTIONS(794),
    [sym_undefined] = ACTIONS(794),
    [anon_sym_AT] = ACTIONS(792),
    [anon_sym_static] = ACTIONS(794),
    [anon_sym_get] = ACTIONS(794),
    [anon_sym_set] = ACTIONS(794),
    [sym__automatic_semicolon] = ACTIONS(800),
    [sym__ternary_qmark] = ACTIONS(798),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(101)] = {
    [sym_declaration] = STATE(469),
    [sym_import] = STATE(1326),
    [sym_variable_declaration] = STATE(419),
    [sym_lexical_declaration] = STATE(419),
    [sym_using_declaration] = STATE(419),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(914),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1790),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1790),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_class_declaration] = STATE(419),
    [sym_function_expression] = STATE(740),
    [sym_function_declaration] = STATE(419),
    [sym_generator_function] = STATE(740),
    [sym_generator_function_declaration] = STATE(419),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1790),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [aux_sym_export_statement_repeat1] = STATE(1343),
    [sym_identifier] = ACTIONS(161),
    [anon_sym_export] = ACTIONS(163),
    [anon_sym_LBRACE] = ACTIONS(167),
    [anon_sym_import] = ACTIONS(173),
    [anon_sym_var] = ACTIONS(21),
    [anon_sym_let] = ACTIONS(802),
    [anon_sym_const] = ACTIONS(25),
    [anon_sym_using] = ACTIONS(27),
    [anon_sym_await] = ACTIONS(804),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_yield] = ACTIONS(57),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(67),
    [anon_sym_async] = ACTIONS(806),
    [anon_sym_function] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(75),
    [anon_sym_DASH] = ACTIONS(75),
    [anon_sym_SLASH] = ACTIONS(77),
    [anon_sym_BANG] = ACTIONS(79),
    [anon_sym_TILDE] = ACTIONS(79),
    [anon_sym_typeof] = ACTIONS(75),
    [anon_sym_void] = ACTIONS(75),
    [anon_sym_delete] = ACTIONS(75),
    [anon_sym_PLUS_PLUS] = ACTIONS(81),
    [anon_sym_DASH_DASH] = ACTIONS(81),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(87),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(163),
    [anon_sym_get] = ACTIONS(163),
    [anon_sym_set] = ACTIONS(163),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(102)] = {
    [sym_declaration] = STATE(451),
    [sym_import] = STATE(1326),
    [sym_variable_declaration] = STATE(419),
    [sym_lexical_declaration] = STATE(419),
    [sym_using_declaration] = STATE(419),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(889),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1790),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1790),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_class_declaration] = STATE(419),
    [sym_function_expression] = STATE(740),
    [sym_function_declaration] = STATE(419),
    [sym_generator_function] = STATE(740),
    [sym_generator_function_declaration] = STATE(419),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1790),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [aux_sym_export_statement_repeat1] = STATE(1322),
    [sym_identifier] = ACTIONS(161),
    [anon_sym_export] = ACTIONS(163),
    [anon_sym_LBRACE] = ACTIONS(167),
    [anon_sym_import] = ACTIONS(173),
    [anon_sym_var] = ACTIONS(21),
    [anon_sym_let] = ACTIONS(802),
    [anon_sym_const] = ACTIONS(25),
    [anon_sym_using] = ACTIONS(27),
    [anon_sym_await] = ACTIONS(804),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_yield] = ACTIONS(57),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(569),
    [anon_sym_async] = ACTIONS(808),
    [anon_sym_function] = ACTIONS(573),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(75),
    [anon_sym_DASH] = ACTIONS(75),
    [anon_sym_SLASH] = ACTIONS(77),
    [anon_sym_BANG] = ACTIONS(79),
    [anon_sym_TILDE] = ACTIONS(79),
    [anon_sym_typeof] = ACTIONS(75),
    [anon_sym_void] = ACTIONS(75),
    [anon_sym_delete] = ACTIONS(75),
    [anon_sym_PLUS_PLUS] = ACTIONS(81),
    [anon_sym_DASH_DASH] = ACTIONS(81),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(87),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(163),
    [anon_sym_get] = ACTIONS(163),
    [anon_sym_set] = ACTIONS(163),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(103)] = {
    [sym_declaration] = STATE(451),
    [sym_import] = STATE(1326),
    [sym_variable_declaration] = STATE(419),
    [sym_lexical_declaration] = STATE(419),
    [sym_using_declaration] = STATE(419),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(889),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1790),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1790),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_class_declaration] = STATE(419),
    [sym_function_expression] = STATE(740),
    [sym_function_declaration] = STATE(419),
    [sym_generator_function] = STATE(740),
    [sym_generator_function_declaration] = STATE(419),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1790),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [aux_sym_export_statement_repeat1] = STATE(1343),
    [sym_identifier] = ACTIONS(161),
    [anon_sym_export] = ACTIONS(163),
    [anon_sym_LBRACE] = ACTIONS(167),
    [anon_sym_import] = ACTIONS(173),
    [anon_sym_var] = ACTIONS(21),
    [anon_sym_let] = ACTIONS(802),
    [anon_sym_const] = ACTIONS(25),
    [anon_sym_using] = ACTIONS(27),
    [anon_sym_await] = ACTIONS(804),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_yield] = ACTIONS(57),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(67),
    [anon_sym_async] = ACTIONS(806),
    [anon_sym_function] = ACTIONS(71),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(75),
    [anon_sym_DASH] = ACTIONS(75),
    [anon_sym_SLASH] = ACTIONS(77),
    [anon_sym_BANG] = ACTIONS(79),
    [anon_sym_TILDE] = ACTIONS(79),
    [anon_sym_typeof] = ACTIONS(75),
    [anon_sym_void] = ACTIONS(75),
    [anon_sym_delete] = ACTIONS(75),
    [anon_sym_PLUS_PLUS] = ACTIONS(81),
    [anon_sym_DASH_DASH] = ACTIONS(81),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(87),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(163),
    [anon_sym_get] = ACTIONS(163),
    [anon_sym_set] = ACTIONS(163),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(104)] = {
    [sym_declaration] = STATE(469),
    [sym_import] = STATE(1326),
    [sym_variable_declaration] = STATE(419),
    [sym_lexical_declaration] = STATE(419),
    [sym_using_declaration] = STATE(419),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(914),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1790),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1790),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_class_declaration] = STATE(419),
    [sym_function_expression] = STATE(740),
    [sym_function_declaration] = STATE(419),
    [sym_generator_function] = STATE(740),
    [sym_generator_function_declaration] = STATE(419),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1790),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [aux_sym_export_statement_repeat1] = STATE(1322),
    [sym_identifier] = ACTIONS(161),
    [anon_sym_export] = ACTIONS(163),
    [anon_sym_LBRACE] = ACTIONS(167),
    [anon_sym_import] = ACTIONS(173),
    [anon_sym_var] = ACTIONS(21),
    [anon_sym_let] = ACTIONS(802),
    [anon_sym_const] = ACTIONS(25),
    [anon_sym_using] = ACTIONS(27),
    [anon_sym_await] = ACTIONS(804),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_yield] = ACTIONS(57),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(569),
    [anon_sym_async] = ACTIONS(808),
    [anon_sym_function] = ACTIONS(573),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(75),
    [anon_sym_DASH] = ACTIONS(75),
    [anon_sym_SLASH] = ACTIONS(77),
    [anon_sym_BANG] = ACTIONS(79),
    [anon_sym_TILDE] = ACTIONS(79),
    [anon_sym_typeof] = ACTIONS(75),
    [anon_sym_void] = ACTIONS(75),
    [anon_sym_delete] = ACTIONS(75),
    [anon_sym_PLUS_PLUS] = ACTIONS(81),
    [anon_sym_DASH_DASH] = ACTIONS(81),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(87),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(163),
    [anon_sym_get] = ACTIONS(163),
    [anon_sym_set] = ACTIONS(163),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(105)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(507),
    [sym_expression] = STATE(858),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1282),
    [sym_assignment_pattern] = STATE(1413),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1282),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(557),
    [sym_subscript_expression] = STATE(557),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1139),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1282),
    [sym_spread_element] = STATE(1485),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1763),
    [sym_pattern] = STATE(1382),
    [sym_rest_pattern] = STATE(1280),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [aux_sym_array_repeat1] = STATE(1490),
    [aux_sym_array_pattern_repeat1] = STATE(1409),
    [sym_identifier] = ACTIONS(810),
    [anon_sym_export] = ACTIONS(812),
    [anon_sym_LBRACE] = ACTIONS(434),
    [anon_sym_COMMA] = ACTIONS(814),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(812),
    [anon_sym_await] = ACTIONS(816),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(382),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_RBRACK] = ACTIONS(818),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(822),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(404),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(115),
    [anon_sym_PLUS] = ACTIONS(412),
    [anon_sym_DASH] = ACTIONS(412),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(414),
    [anon_sym_TILDE] = ACTIONS(414),
    [anon_sym_typeof] = ACTIONS(412),
    [anon_sym_void] = ACTIONS(412),
    [anon_sym_delete] = ACTIONS(412),
    [anon_sym_PLUS_PLUS] = ACTIONS(660),
    [anon_sym_DASH_DASH] = ACTIONS(660),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(424),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(824),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(812),
    [anon_sym_get] = ACTIONS(812),
    [anon_sym_set] = ACTIONS(812),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(106)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(507),
    [sym_expression] = STATE(858),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1282),
    [sym_assignment_pattern] = STATE(1413),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1282),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(557),
    [sym_subscript_expression] = STATE(557),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1139),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1282),
    [sym_spread_element] = STATE(1485),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1763),
    [sym_pattern] = STATE(1382),
    [sym_rest_pattern] = STATE(1280),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [aux_sym_array_repeat1] = STATE(1490),
    [aux_sym_array_pattern_repeat1] = STATE(1409),
    [sym_identifier] = ACTIONS(810),
    [anon_sym_export] = ACTIONS(812),
    [anon_sym_LBRACE] = ACTIONS(434),
    [anon_sym_COMMA] = ACTIONS(814),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(812),
    [anon_sym_await] = ACTIONS(816),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(382),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_RBRACK] = ACTIONS(826),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(822),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(404),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(115),
    [anon_sym_PLUS] = ACTIONS(412),
    [anon_sym_DASH] = ACTIONS(412),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(414),
    [anon_sym_TILDE] = ACTIONS(414),
    [anon_sym_typeof] = ACTIONS(412),
    [anon_sym_void] = ACTIONS(412),
    [anon_sym_delete] = ACTIONS(412),
    [anon_sym_PLUS_PLUS] = ACTIONS(660),
    [anon_sym_DASH_DASH] = ACTIONS(660),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(424),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(824),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(812),
    [anon_sym_get] = ACTIONS(812),
    [anon_sym_set] = ACTIONS(812),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(107)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(507),
    [sym_expression] = STATE(872),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1282),
    [sym_assignment_pattern] = STATE(1413),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1282),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(557),
    [sym_subscript_expression] = STATE(557),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1139),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1282),
    [sym_spread_element] = STATE(1460),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1763),
    [sym_pattern] = STATE(1382),
    [sym_rest_pattern] = STATE(1280),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [aux_sym_array_repeat1] = STATE(1450),
    [aux_sym_array_pattern_repeat1] = STATE(1409),
    [sym_identifier] = ACTIONS(810),
    [anon_sym_export] = ACTIONS(812),
    [anon_sym_LBRACE] = ACTIONS(434),
    [anon_sym_COMMA] = ACTIONS(814),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(812),
    [anon_sym_await] = ACTIONS(816),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(382),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_RBRACK] = ACTIONS(828),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(822),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(404),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(115),
    [anon_sym_PLUS] = ACTIONS(412),
    [anon_sym_DASH] = ACTIONS(412),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(414),
    [anon_sym_TILDE] = ACTIONS(414),
    [anon_sym_typeof] = ACTIONS(412),
    [anon_sym_void] = ACTIONS(412),
    [anon_sym_delete] = ACTIONS(412),
    [anon_sym_PLUS_PLUS] = ACTIONS(660),
    [anon_sym_DASH_DASH] = ACTIONS(660),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(424),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(824),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(812),
    [anon_sym_get] = ACTIONS(812),
    [anon_sym_set] = ACTIONS(812),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(108)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(507),
    [sym_expression] = STATE(858),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1282),
    [sym_assignment_pattern] = STATE(1413),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1282),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(557),
    [sym_subscript_expression] = STATE(557),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1139),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1282),
    [sym_spread_element] = STATE(1485),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1763),
    [sym_pattern] = STATE(1382),
    [sym_rest_pattern] = STATE(1280),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [aux_sym_array_repeat1] = STATE(1490),
    [aux_sym_array_pattern_repeat1] = STATE(1409),
    [sym_identifier] = ACTIONS(810),
    [anon_sym_export] = ACTIONS(812),
    [anon_sym_LBRACE] = ACTIONS(434),
    [anon_sym_COMMA] = ACTIONS(814),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(812),
    [anon_sym_await] = ACTIONS(816),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(382),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_RBRACK] = ACTIONS(830),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(822),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(404),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(115),
    [anon_sym_PLUS] = ACTIONS(412),
    [anon_sym_DASH] = ACTIONS(412),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(414),
    [anon_sym_TILDE] = ACTIONS(414),
    [anon_sym_typeof] = ACTIONS(412),
    [anon_sym_void] = ACTIONS(412),
    [anon_sym_delete] = ACTIONS(412),
    [anon_sym_PLUS_PLUS] = ACTIONS(660),
    [anon_sym_DASH_DASH] = ACTIONS(660),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(424),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(824),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(812),
    [anon_sym_get] = ACTIONS(812),
    [anon_sym_set] = ACTIONS(812),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(109)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(507),
    [sym_expression] = STATE(858),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1282),
    [sym_assignment_pattern] = STATE(1413),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1282),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(557),
    [sym_subscript_expression] = STATE(557),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1139),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1282),
    [sym_spread_element] = STATE(1485),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1763),
    [sym_pattern] = STATE(1382),
    [sym_rest_pattern] = STATE(1280),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [aux_sym_array_repeat1] = STATE(1490),
    [aux_sym_array_pattern_repeat1] = STATE(1409),
    [sym_identifier] = ACTIONS(810),
    [anon_sym_export] = ACTIONS(812),
    [anon_sym_LBRACE] = ACTIONS(434),
    [anon_sym_COMMA] = ACTIONS(814),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(812),
    [anon_sym_await] = ACTIONS(816),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(382),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_RBRACK] = ACTIONS(832),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(822),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(404),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(115),
    [anon_sym_PLUS] = ACTIONS(412),
    [anon_sym_DASH] = ACTIONS(412),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(414),
    [anon_sym_TILDE] = ACTIONS(414),
    [anon_sym_typeof] = ACTIONS(412),
    [anon_sym_void] = ACTIONS(412),
    [anon_sym_delete] = ACTIONS(412),
    [anon_sym_PLUS_PLUS] = ACTIONS(660),
    [anon_sym_DASH_DASH] = ACTIONS(660),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(424),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(824),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(812),
    [anon_sym_get] = ACTIONS(812),
    [anon_sym_set] = ACTIONS(812),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(110)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(507),
    [sym_expression] = STATE(858),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1282),
    [sym_assignment_pattern] = STATE(1413),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1282),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(557),
    [sym_subscript_expression] = STATE(557),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1139),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1282),
    [sym_spread_element] = STATE(1485),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1763),
    [sym_pattern] = STATE(1382),
    [sym_rest_pattern] = STATE(1280),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [aux_sym_array_repeat1] = STATE(1490),
    [aux_sym_array_pattern_repeat1] = STATE(1409),
    [sym_identifier] = ACTIONS(810),
    [anon_sym_export] = ACTIONS(812),
    [anon_sym_LBRACE] = ACTIONS(434),
    [anon_sym_COMMA] = ACTIONS(814),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(812),
    [anon_sym_await] = ACTIONS(816),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(382),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_RBRACK] = ACTIONS(834),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(822),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(404),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(115),
    [anon_sym_PLUS] = ACTIONS(412),
    [anon_sym_DASH] = ACTIONS(412),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(414),
    [anon_sym_TILDE] = ACTIONS(414),
    [anon_sym_typeof] = ACTIONS(412),
    [anon_sym_void] = ACTIONS(412),
    [anon_sym_delete] = ACTIONS(412),
    [anon_sym_PLUS_PLUS] = ACTIONS(660),
    [anon_sym_DASH_DASH] = ACTIONS(660),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(424),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(824),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(812),
    [anon_sym_get] = ACTIONS(812),
    [anon_sym_set] = ACTIONS(812),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(111)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(507),
    [sym_expression] = STATE(841),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1282),
    [sym_assignment_pattern] = STATE(1413),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1282),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(557),
    [sym_subscript_expression] = STATE(557),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1139),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1282),
    [sym_spread_element] = STATE(1460),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1763),
    [sym_pattern] = STATE(1382),
    [sym_rest_pattern] = STATE(1280),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [aux_sym_array_repeat1] = STATE(1450),
    [aux_sym_array_pattern_repeat1] = STATE(1409),
    [sym_identifier] = ACTIONS(810),
    [anon_sym_export] = ACTIONS(812),
    [anon_sym_LBRACE] = ACTIONS(434),
    [anon_sym_COMMA] = ACTIONS(814),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(812),
    [anon_sym_await] = ACTIONS(816),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(382),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_RBRACK] = ACTIONS(828),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(822),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(404),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(115),
    [anon_sym_PLUS] = ACTIONS(412),
    [anon_sym_DASH] = ACTIONS(412),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(414),
    [anon_sym_TILDE] = ACTIONS(414),
    [anon_sym_typeof] = ACTIONS(412),
    [anon_sym_void] = ACTIONS(412),
    [anon_sym_delete] = ACTIONS(412),
    [anon_sym_PLUS_PLUS] = ACTIONS(660),
    [anon_sym_DASH_DASH] = ACTIONS(660),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(424),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(824),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(812),
    [anon_sym_get] = ACTIONS(812),
    [anon_sym_set] = ACTIONS(812),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(112)] = {
    [sym_import] = STATE(1333),
    [sym_variable_declaration] = STATE(141),
    [sym_lexical_declaration] = STATE(141),
    [sym_empty_statement] = STATE(141),
    [sym_parenthesized_expression] = STATE(564),
    [sym_expression] = STATE(842),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1481),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1481),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(564),
    [sym_subscript_expression] = STATE(564),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1139),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1481),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_sequence_expression] = STATE(1847),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1763),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(836),
    [anon_sym_export] = ACTIONS(838),
    [anon_sym_LBRACE] = ACTIONS(840),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_var] = ACTIONS(842),
    [anon_sym_let] = ACTIONS(844),
    [anon_sym_const] = ACTIONS(846),
    [anon_sym_using] = ACTIONS(848),
    [anon_sym_await] = ACTIONS(850),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_SEMI] = ACTIONS(39),
    [anon_sym_yield] = ACTIONS(382),
    [anon_sym_LBRACK] = ACTIONS(852),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(854),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(404),
    [anon_sym_PLUS] = ACTIONS(412),
    [anon_sym_DASH] = ACTIONS(412),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(414),
    [anon_sym_TILDE] = ACTIONS(414),
    [anon_sym_typeof] = ACTIONS(412),
    [anon_sym_void] = ACTIONS(412),
    [anon_sym_delete] = ACTIONS(412),
    [anon_sym_PLUS_PLUS] = ACTIONS(660),
    [anon_sym_DASH_DASH] = ACTIONS(660),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(424),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(856),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(838),
    [anon_sym_get] = ACTIONS(838),
    [anon_sym_set] = ACTIONS(838),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(113)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(507),
    [sym_expression] = STATE(880),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1815),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1815),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(507),
    [sym_subscript_expression] = STATE(507),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1139),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__initializer] = STATE(1408),
    [sym__destructuring_pattern] = STATE(1815),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_sequence_expression] = STATE(1854),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1763),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(369),
    [anon_sym_export] = ACTIONS(371),
    [anon_sym_LBRACE] = ACTIONS(373),
    [anon_sym_COMMA] = ACTIONS(858),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(371),
    [anon_sym_await] = ACTIONS(377),
    [anon_sym_of] = ACTIONS(860),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_SEMI] = ACTIONS(858),
    [anon_sym_in] = ACTIONS(860),
    [anon_sym_yield] = ACTIONS(382),
    [anon_sym_EQ] = ACTIONS(862),
    [anon_sym_LBRACK] = ACTIONS(654),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(398),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(404),
    [anon_sym_PLUS] = ACTIONS(412),
    [anon_sym_DASH] = ACTIONS(412),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(414),
    [anon_sym_TILDE] = ACTIONS(414),
    [anon_sym_typeof] = ACTIONS(412),
    [anon_sym_void] = ACTIONS(412),
    [anon_sym_delete] = ACTIONS(412),
    [anon_sym_PLUS_PLUS] = ACTIONS(660),
    [anon_sym_DASH_DASH] = ACTIONS(660),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(424),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(428),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(371),
    [anon_sym_get] = ACTIONS(371),
    [anon_sym_set] = ACTIONS(371),
    [sym__automatic_semicolon] = ACTIONS(864),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(114)] = {
    [sym_import] = STATE(1333),
    [sym_variable_declaration] = STATE(143),
    [sym_lexical_declaration] = STATE(143),
    [sym_empty_statement] = STATE(143),
    [sym_parenthesized_expression] = STATE(564),
    [sym_expression] = STATE(864),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1481),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1481),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(564),
    [sym_subscript_expression] = STATE(564),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1139),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1481),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_sequence_expression] = STATE(1868),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1763),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(836),
    [anon_sym_export] = ACTIONS(838),
    [anon_sym_LBRACE] = ACTIONS(840),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_var] = ACTIONS(842),
    [anon_sym_let] = ACTIONS(844),
    [anon_sym_const] = ACTIONS(846),
    [anon_sym_using] = ACTIONS(848),
    [anon_sym_await] = ACTIONS(850),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_SEMI] = ACTIONS(39),
    [anon_sym_yield] = ACTIONS(382),
    [anon_sym_LBRACK] = ACTIONS(852),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(854),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(404),
    [anon_sym_PLUS] = ACTIONS(412),
    [anon_sym_DASH] = ACTIONS(412),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(414),
    [anon_sym_TILDE] = ACTIONS(414),
    [anon_sym_typeof] = ACTIONS(412),
    [anon_sym_void] = ACTIONS(412),
    [anon_sym_delete] = ACTIONS(412),
    [anon_sym_PLUS_PLUS] = ACTIONS(660),
    [anon_sym_DASH_DASH] = ACTIONS(660),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(424),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(856),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(838),
    [anon_sym_get] = ACTIONS(838),
    [anon_sym_set] = ACTIONS(838),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(115)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(498),
    [sym_expression] = STATE(979),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1307),
    [sym_assignment_pattern] = STATE(1413),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1307),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(558),
    [sym_subscript_expression] = STATE(558),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1138),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1307),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1816),
    [sym_pattern] = STATE(1382),
    [sym_rest_pattern] = STATE(1280),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [aux_sym_array_pattern_repeat1] = STATE(1409),
    [sym_identifier] = ACTIONS(867),
    [anon_sym_export] = ACTIONS(869),
    [anon_sym_LBRACE] = ACTIONS(871),
    [anon_sym_COMMA] = ACTIONS(873),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(869),
    [anon_sym_await] = ACTIONS(875),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(438),
    [anon_sym_LBRACK] = ACTIONS(877),
    [anon_sym_RBRACK] = ACTIONS(879),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(881),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(447),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(883),
    [anon_sym_PLUS] = ACTIONS(452),
    [anon_sym_DASH] = ACTIONS(452),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(454),
    [anon_sym_TILDE] = ACTIONS(454),
    [anon_sym_typeof] = ACTIONS(452),
    [anon_sym_void] = ACTIONS(452),
    [anon_sym_delete] = ACTIONS(452),
    [anon_sym_PLUS_PLUS] = ACTIONS(684),
    [anon_sym_DASH_DASH] = ACTIONS(684),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(459),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(885),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(869),
    [anon_sym_get] = ACTIONS(869),
    [anon_sym_set] = ACTIONS(869),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(116)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(507),
    [sym_expression] = STATE(975),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1282),
    [sym_assignment_pattern] = STATE(1413),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1282),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(572),
    [sym_subscript_expression] = STATE(572),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1139),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1282),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1763),
    [sym_pattern] = STATE(1382),
    [sym_rest_pattern] = STATE(1280),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [aux_sym_array_pattern_repeat1] = STATE(1409),
    [sym_identifier] = ACTIONS(887),
    [anon_sym_export] = ACTIONS(889),
    [anon_sym_LBRACE] = ACTIONS(891),
    [anon_sym_COMMA] = ACTIONS(873),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(889),
    [anon_sym_await] = ACTIONS(893),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(382),
    [anon_sym_LBRACK] = ACTIONS(895),
    [anon_sym_RBRACK] = ACTIONS(879),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(897),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(404),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(883),
    [anon_sym_PLUS] = ACTIONS(412),
    [anon_sym_DASH] = ACTIONS(412),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(414),
    [anon_sym_TILDE] = ACTIONS(414),
    [anon_sym_typeof] = ACTIONS(412),
    [anon_sym_void] = ACTIONS(412),
    [anon_sym_delete] = ACTIONS(412),
    [anon_sym_PLUS_PLUS] = ACTIONS(660),
    [anon_sym_DASH_DASH] = ACTIONS(660),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(424),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(899),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(889),
    [anon_sym_get] = ACTIONS(889),
    [anon_sym_set] = ACTIONS(889),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(117)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(507),
    [sym_expression] = STATE(845),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1282),
    [sym_assignment_pattern] = STATE(1699),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1282),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(557),
    [sym_subscript_expression] = STATE(557),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1139),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1282),
    [sym_spread_element] = STATE(1488),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1763),
    [sym_pattern] = STATE(1410),
    [sym_rest_pattern] = STATE(1280),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(810),
    [anon_sym_export] = ACTIONS(812),
    [anon_sym_LBRACE] = ACTIONS(434),
    [anon_sym_COMMA] = ACTIONS(901),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(812),
    [anon_sym_await] = ACTIONS(816),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(382),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_RBRACK] = ACTIONS(901),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(822),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(404),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(115),
    [anon_sym_PLUS] = ACTIONS(412),
    [anon_sym_DASH] = ACTIONS(412),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(414),
    [anon_sym_TILDE] = ACTIONS(414),
    [anon_sym_typeof] = ACTIONS(412),
    [anon_sym_void] = ACTIONS(412),
    [anon_sym_delete] = ACTIONS(412),
    [anon_sym_PLUS_PLUS] = ACTIONS(660),
    [anon_sym_DASH_DASH] = ACTIONS(660),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(424),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(824),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(812),
    [anon_sym_get] = ACTIONS(812),
    [anon_sym_set] = ACTIONS(812),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(118)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(498),
    [sym_expression] = STATE(979),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1307),
    [sym_assignment_pattern] = STATE(1484),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1307),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(558),
    [sym_subscript_expression] = STATE(558),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1138),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1307),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1816),
    [sym_pattern] = STATE(1347),
    [sym_rest_pattern] = STATE(1280),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [aux_sym_array_pattern_repeat1] = STATE(1491),
    [sym_identifier] = ACTIONS(867),
    [anon_sym_export] = ACTIONS(869),
    [anon_sym_LBRACE] = ACTIONS(871),
    [anon_sym_COMMA] = ACTIONS(873),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(869),
    [anon_sym_await] = ACTIONS(875),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(438),
    [anon_sym_LBRACK] = ACTIONS(877),
    [anon_sym_RBRACK] = ACTIONS(904),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(881),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(447),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(883),
    [anon_sym_PLUS] = ACTIONS(452),
    [anon_sym_DASH] = ACTIONS(452),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(454),
    [anon_sym_TILDE] = ACTIONS(454),
    [anon_sym_typeof] = ACTIONS(452),
    [anon_sym_void] = ACTIONS(452),
    [anon_sym_delete] = ACTIONS(452),
    [anon_sym_PLUS_PLUS] = ACTIONS(684),
    [anon_sym_DASH_DASH] = ACTIONS(684),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(459),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(885),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(869),
    [anon_sym_get] = ACTIONS(869),
    [anon_sym_set] = ACTIONS(869),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(119)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(507),
    [sym_expression] = STATE(875),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1282),
    [sym_assignment_pattern] = STATE(1426),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1282),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(557),
    [sym_subscript_expression] = STATE(557),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1139),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1282),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_sequence_expression] = STATE(1817),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1763),
    [sym_pattern] = STATE(1370),
    [sym_rest_pattern] = STATE(1280),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(810),
    [anon_sym_export] = ACTIONS(812),
    [anon_sym_LBRACE] = ACTIONS(434),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(812),
    [anon_sym_await] = ACTIONS(816),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_RPAREN] = ACTIONS(906),
    [anon_sym_yield] = ACTIONS(382),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(822),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(404),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(883),
    [anon_sym_PLUS] = ACTIONS(412),
    [anon_sym_DASH] = ACTIONS(412),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(414),
    [anon_sym_TILDE] = ACTIONS(414),
    [anon_sym_typeof] = ACTIONS(412),
    [anon_sym_void] = ACTIONS(412),
    [anon_sym_delete] = ACTIONS(412),
    [anon_sym_PLUS_PLUS] = ACTIONS(660),
    [anon_sym_DASH_DASH] = ACTIONS(660),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(424),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(824),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(812),
    [anon_sym_get] = ACTIONS(812),
    [anon_sym_set] = ACTIONS(812),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(120)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(507),
    [sym_expression] = STATE(866),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1282),
    [sym_assignment_pattern] = STATE(1426),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1282),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(557),
    [sym_subscript_expression] = STATE(557),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1139),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1282),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_sequence_expression] = STATE(1804),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1763),
    [sym_pattern] = STATE(1370),
    [sym_rest_pattern] = STATE(1280),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(810),
    [anon_sym_export] = ACTIONS(812),
    [anon_sym_LBRACE] = ACTIONS(434),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(812),
    [anon_sym_await] = ACTIONS(816),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_RPAREN] = ACTIONS(906),
    [anon_sym_yield] = ACTIONS(382),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(822),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(404),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(883),
    [anon_sym_PLUS] = ACTIONS(412),
    [anon_sym_DASH] = ACTIONS(412),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(414),
    [anon_sym_TILDE] = ACTIONS(414),
    [anon_sym_typeof] = ACTIONS(412),
    [anon_sym_void] = ACTIONS(412),
    [anon_sym_delete] = ACTIONS(412),
    [anon_sym_PLUS_PLUS] = ACTIONS(660),
    [anon_sym_DASH_DASH] = ACTIONS(660),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(424),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(824),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(812),
    [anon_sym_get] = ACTIONS(812),
    [anon_sym_set] = ACTIONS(812),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(121)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(498),
    [sym_expression] = STATE(979),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1307),
    [sym_assignment_pattern] = STATE(1699),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1307),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(558),
    [sym_subscript_expression] = STATE(558),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1138),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1307),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1816),
    [sym_pattern] = STATE(1410),
    [sym_rest_pattern] = STATE(1280),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(867),
    [anon_sym_export] = ACTIONS(869),
    [anon_sym_LBRACE] = ACTIONS(871),
    [anon_sym_COMMA] = ACTIONS(908),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(869),
    [anon_sym_await] = ACTIONS(875),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(438),
    [anon_sym_LBRACK] = ACTIONS(877),
    [anon_sym_RBRACK] = ACTIONS(908),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(881),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(447),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(883),
    [anon_sym_PLUS] = ACTIONS(452),
    [anon_sym_DASH] = ACTIONS(452),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(454),
    [anon_sym_TILDE] = ACTIONS(454),
    [anon_sym_typeof] = ACTIONS(452),
    [anon_sym_void] = ACTIONS(452),
    [anon_sym_delete] = ACTIONS(452),
    [anon_sym_PLUS_PLUS] = ACTIONS(684),
    [anon_sym_DASH_DASH] = ACTIONS(684),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(459),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(885),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(869),
    [anon_sym_get] = ACTIONS(869),
    [anon_sym_set] = ACTIONS(869),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(122)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(507),
    [sym_expression] = STATE(867),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1815),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1815),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(507),
    [sym_subscript_expression] = STATE(507),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1139),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1815),
    [sym_spread_element] = STATE(1405),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1763),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [aux_sym_array_repeat1] = STATE(1445),
    [sym_identifier] = ACTIONS(369),
    [anon_sym_export] = ACTIONS(371),
    [anon_sym_LBRACE] = ACTIONS(373),
    [anon_sym_COMMA] = ACTIONS(910),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(371),
    [anon_sym_await] = ACTIONS(377),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_RPAREN] = ACTIONS(912),
    [anon_sym_yield] = ACTIONS(382),
    [anon_sym_LBRACK] = ACTIONS(654),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(398),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(404),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(914),
    [anon_sym_PLUS] = ACTIONS(412),
    [anon_sym_DASH] = ACTIONS(412),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(414),
    [anon_sym_TILDE] = ACTIONS(414),
    [anon_sym_typeof] = ACTIONS(412),
    [anon_sym_void] = ACTIONS(412),
    [anon_sym_delete] = ACTIONS(412),
    [anon_sym_PLUS_PLUS] = ACTIONS(660),
    [anon_sym_DASH_DASH] = ACTIONS(660),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(424),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(428),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(371),
    [anon_sym_get] = ACTIONS(371),
    [anon_sym_set] = ACTIONS(371),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(123)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(507),
    [sym_expression] = STATE(847),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1815),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1815),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(507),
    [sym_subscript_expression] = STATE(507),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1139),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1815),
    [sym_spread_element] = STATE(1480),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1763),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [aux_sym_array_repeat1] = STATE(1432),
    [sym_identifier] = ACTIONS(369),
    [anon_sym_export] = ACTIONS(371),
    [anon_sym_LBRACE] = ACTIONS(373),
    [anon_sym_COMMA] = ACTIONS(910),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(371),
    [anon_sym_await] = ACTIONS(377),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_RPAREN] = ACTIONS(916),
    [anon_sym_yield] = ACTIONS(382),
    [anon_sym_LBRACK] = ACTIONS(654),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(398),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(404),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(914),
    [anon_sym_PLUS] = ACTIONS(412),
    [anon_sym_DASH] = ACTIONS(412),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(414),
    [anon_sym_TILDE] = ACTIONS(414),
    [anon_sym_typeof] = ACTIONS(412),
    [anon_sym_void] = ACTIONS(412),
    [anon_sym_delete] = ACTIONS(412),
    [anon_sym_PLUS_PLUS] = ACTIONS(660),
    [anon_sym_DASH_DASH] = ACTIONS(660),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(424),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(428),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(371),
    [anon_sym_get] = ACTIONS(371),
    [anon_sym_set] = ACTIONS(371),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(124)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(498),
    [sym_expression] = STATE(979),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1307),
    [sym_assignment_pattern] = STATE(1661),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1307),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(558),
    [sym_subscript_expression] = STATE(558),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1138),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1307),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1816),
    [sym_pattern] = STATE(1474),
    [sym_rest_pattern] = STATE(1280),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(867),
    [anon_sym_export] = ACTIONS(869),
    [anon_sym_LBRACE] = ACTIONS(871),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(869),
    [anon_sym_await] = ACTIONS(875),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_RPAREN] = ACTIONS(918),
    [anon_sym_yield] = ACTIONS(438),
    [anon_sym_LBRACK] = ACTIONS(877),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(881),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(447),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(883),
    [anon_sym_PLUS] = ACTIONS(452),
    [anon_sym_DASH] = ACTIONS(452),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(454),
    [anon_sym_TILDE] = ACTIONS(454),
    [anon_sym_typeof] = ACTIONS(452),
    [anon_sym_void] = ACTIONS(452),
    [anon_sym_delete] = ACTIONS(452),
    [anon_sym_PLUS_PLUS] = ACTIONS(684),
    [anon_sym_DASH_DASH] = ACTIONS(684),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(459),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(885),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(869),
    [anon_sym_get] = ACTIONS(869),
    [anon_sym_set] = ACTIONS(869),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(125)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(507),
    [sym_expression] = STATE(859),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1815),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1815),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(507),
    [sym_subscript_expression] = STATE(507),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1139),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1815),
    [sym_spread_element] = STATE(1437),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1763),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [aux_sym_array_repeat1] = STATE(1438),
    [sym_identifier] = ACTIONS(369),
    [anon_sym_export] = ACTIONS(371),
    [anon_sym_LBRACE] = ACTIONS(373),
    [anon_sym_COMMA] = ACTIONS(910),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(371),
    [anon_sym_await] = ACTIONS(377),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_RPAREN] = ACTIONS(920),
    [anon_sym_yield] = ACTIONS(382),
    [anon_sym_LBRACK] = ACTIONS(654),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(398),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(404),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(914),
    [anon_sym_PLUS] = ACTIONS(412),
    [anon_sym_DASH] = ACTIONS(412),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(414),
    [anon_sym_TILDE] = ACTIONS(414),
    [anon_sym_typeof] = ACTIONS(412),
    [anon_sym_void] = ACTIONS(412),
    [anon_sym_delete] = ACTIONS(412),
    [anon_sym_PLUS_PLUS] = ACTIONS(660),
    [anon_sym_DASH_DASH] = ACTIONS(660),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(424),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(428),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(371),
    [anon_sym_get] = ACTIONS(371),
    [anon_sym_set] = ACTIONS(371),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(126)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(498),
    [sym_expression] = STATE(979),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1307),
    [sym_assignment_pattern] = STATE(1426),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1307),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(558),
    [sym_subscript_expression] = STATE(558),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1138),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1307),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1816),
    [sym_pattern] = STATE(1370),
    [sym_rest_pattern] = STATE(1280),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(867),
    [anon_sym_export] = ACTIONS(869),
    [anon_sym_LBRACE] = ACTIONS(871),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(869),
    [anon_sym_await] = ACTIONS(875),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_RPAREN] = ACTIONS(906),
    [anon_sym_yield] = ACTIONS(438),
    [anon_sym_LBRACK] = ACTIONS(877),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(881),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(447),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(883),
    [anon_sym_PLUS] = ACTIONS(452),
    [anon_sym_DASH] = ACTIONS(452),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(454),
    [anon_sym_TILDE] = ACTIONS(454),
    [anon_sym_typeof] = ACTIONS(452),
    [anon_sym_void] = ACTIONS(452),
    [anon_sym_delete] = ACTIONS(452),
    [anon_sym_PLUS_PLUS] = ACTIONS(684),
    [anon_sym_DASH_DASH] = ACTIONS(684),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(459),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(885),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(869),
    [anon_sym_get] = ACTIONS(869),
    [anon_sym_set] = ACTIONS(869),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(127)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(507),
    [sym_expression] = STATE(845),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1815),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1815),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(507),
    [sym_subscript_expression] = STATE(507),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1139),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1815),
    [sym_spread_element] = STATE(1488),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1763),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(369),
    [anon_sym_export] = ACTIONS(371),
    [anon_sym_LBRACE] = ACTIONS(373),
    [anon_sym_COMMA] = ACTIONS(922),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(371),
    [anon_sym_await] = ACTIONS(377),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_RPAREN] = ACTIONS(922),
    [anon_sym_yield] = ACTIONS(382),
    [anon_sym_LBRACK] = ACTIONS(654),
    [anon_sym_RBRACK] = ACTIONS(922),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(398),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(404),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(914),
    [anon_sym_PLUS] = ACTIONS(412),
    [anon_sym_DASH] = ACTIONS(412),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(414),
    [anon_sym_TILDE] = ACTIONS(414),
    [anon_sym_typeof] = ACTIONS(412),
    [anon_sym_void] = ACTIONS(412),
    [anon_sym_delete] = ACTIONS(412),
    [anon_sym_PLUS_PLUS] = ACTIONS(660),
    [anon_sym_DASH_DASH] = ACTIONS(660),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(424),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(428),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(371),
    [anon_sym_get] = ACTIONS(371),
    [anon_sym_set] = ACTIONS(371),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(128)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(498),
    [sym_expression] = STATE(979),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1307),
    [sym_assignment_pattern] = STATE(1661),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1307),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(558),
    [sym_subscript_expression] = STATE(558),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1138),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1307),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1816),
    [sym_pattern] = STATE(1474),
    [sym_rest_pattern] = STATE(1280),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(867),
    [anon_sym_export] = ACTIONS(869),
    [anon_sym_LBRACE] = ACTIONS(871),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(869),
    [anon_sym_await] = ACTIONS(875),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_RPAREN] = ACTIONS(924),
    [anon_sym_yield] = ACTIONS(438),
    [anon_sym_LBRACK] = ACTIONS(877),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(881),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(447),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(883),
    [anon_sym_PLUS] = ACTIONS(452),
    [anon_sym_DASH] = ACTIONS(452),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(454),
    [anon_sym_TILDE] = ACTIONS(454),
    [anon_sym_typeof] = ACTIONS(452),
    [anon_sym_void] = ACTIONS(452),
    [anon_sym_delete] = ACTIONS(452),
    [anon_sym_PLUS_PLUS] = ACTIONS(684),
    [anon_sym_DASH_DASH] = ACTIONS(684),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(459),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(885),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(869),
    [anon_sym_get] = ACTIONS(869),
    [anon_sym_set] = ACTIONS(869),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(129)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(507),
    [sym_expression] = STATE(850),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1815),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1815),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(507),
    [sym_subscript_expression] = STATE(507),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1139),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1815),
    [sym_spread_element] = STATE(1845),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_sequence_expression] = STATE(1845),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1763),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(369),
    [anon_sym_export] = ACTIONS(371),
    [anon_sym_LBRACE] = ACTIONS(373),
    [anon_sym_RBRACE] = ACTIONS(926),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(371),
    [anon_sym_await] = ACTIONS(377),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(382),
    [anon_sym_LBRACK] = ACTIONS(654),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(398),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(404),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(914),
    [anon_sym_PLUS] = ACTIONS(412),
    [anon_sym_DASH] = ACTIONS(412),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(414),
    [anon_sym_TILDE] = ACTIONS(414),
    [anon_sym_typeof] = ACTIONS(412),
    [anon_sym_void] = ACTIONS(412),
    [anon_sym_delete] = ACTIONS(412),
    [anon_sym_PLUS_PLUS] = ACTIONS(660),
    [anon_sym_DASH_DASH] = ACTIONS(660),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(424),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(428),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(371),
    [anon_sym_get] = ACTIONS(371),
    [anon_sym_set] = ACTIONS(371),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(130)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(498),
    [sym_expression] = STATE(979),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1307),
    [sym_assignment_pattern] = STATE(1722),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1307),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(558),
    [sym_subscript_expression] = STATE(558),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1138),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1307),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1816),
    [sym_pattern] = STATE(1434),
    [sym_rest_pattern] = STATE(1280),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(867),
    [anon_sym_export] = ACTIONS(869),
    [anon_sym_LBRACE] = ACTIONS(871),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(869),
    [anon_sym_await] = ACTIONS(875),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(438),
    [anon_sym_LBRACK] = ACTIONS(877),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(881),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(447),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(883),
    [anon_sym_PLUS] = ACTIONS(452),
    [anon_sym_DASH] = ACTIONS(452),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(454),
    [anon_sym_TILDE] = ACTIONS(454),
    [anon_sym_typeof] = ACTIONS(452),
    [anon_sym_void] = ACTIONS(452),
    [anon_sym_delete] = ACTIONS(452),
    [anon_sym_PLUS_PLUS] = ACTIONS(684),
    [anon_sym_DASH_DASH] = ACTIONS(684),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(459),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(885),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(869),
    [anon_sym_get] = ACTIONS(869),
    [anon_sym_set] = ACTIONS(869),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(131)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(507),
    [sym_expression] = STATE(861),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1815),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1815),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(507),
    [sym_subscript_expression] = STATE(507),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1139),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1815),
    [sym_spread_element] = STATE(1765),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_sequence_expression] = STATE(1765),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1763),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(369),
    [anon_sym_export] = ACTIONS(371),
    [anon_sym_LBRACE] = ACTIONS(373),
    [anon_sym_RBRACE] = ACTIONS(928),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(371),
    [anon_sym_await] = ACTIONS(377),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(382),
    [anon_sym_LBRACK] = ACTIONS(654),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(398),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(404),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(914),
    [anon_sym_PLUS] = ACTIONS(412),
    [anon_sym_DASH] = ACTIONS(412),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(414),
    [anon_sym_TILDE] = ACTIONS(414),
    [anon_sym_typeof] = ACTIONS(412),
    [anon_sym_void] = ACTIONS(412),
    [anon_sym_delete] = ACTIONS(412),
    [anon_sym_PLUS_PLUS] = ACTIONS(660),
    [anon_sym_DASH_DASH] = ACTIONS(660),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(424),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(428),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(371),
    [anon_sym_get] = ACTIONS(371),
    [anon_sym_set] = ACTIONS(371),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(132)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(507),
    [sym_expression] = STATE(941),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1282),
    [sym_assignment_pattern] = STATE(1722),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1282),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(557),
    [sym_subscript_expression] = STATE(557),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1139),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1282),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1763),
    [sym_pattern] = STATE(1434),
    [sym_rest_pattern] = STATE(1280),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(810),
    [anon_sym_export] = ACTIONS(812),
    [anon_sym_LBRACE] = ACTIONS(434),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(812),
    [anon_sym_await] = ACTIONS(816),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(382),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(822),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(404),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(883),
    [anon_sym_PLUS] = ACTIONS(412),
    [anon_sym_DASH] = ACTIONS(412),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(414),
    [anon_sym_TILDE] = ACTIONS(414),
    [anon_sym_typeof] = ACTIONS(412),
    [anon_sym_void] = ACTIONS(412),
    [anon_sym_delete] = ACTIONS(412),
    [anon_sym_PLUS_PLUS] = ACTIONS(660),
    [anon_sym_DASH_DASH] = ACTIONS(660),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(424),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(824),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(812),
    [anon_sym_get] = ACTIONS(812),
    [anon_sym_set] = ACTIONS(812),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(133)] = {
    [sym_namespace_export] = STATE(1692),
    [sym_export_clause] = STATE(1358),
    [sym_declaration] = STATE(433),
    [sym_variable_declaration] = STATE(419),
    [sym_lexical_declaration] = STATE(419),
    [sym_using_declaration] = STATE(419),
    [sym_class_declaration] = STATE(419),
    [sym_function_declaration] = STATE(419),
    [sym_generator_function_declaration] = STATE(419),
    [sym_decorator] = STATE(1097),
    [aux_sym_export_statement_repeat1] = STATE(1327),
    [aux_sym_object_repeat1] = STATE(1418),
    [aux_sym_object_pattern_repeat1] = STATE(1452),
    [anon_sym_STAR] = ACTIONS(930),
    [anon_sym_default] = ACTIONS(932),
    [anon_sym_LBRACE] = ACTIONS(934),
    [anon_sym_COMMA] = ACTIONS(169),
    [anon_sym_RBRACE] = ACTIONS(351),
    [anon_sym_var] = ACTIONS(936),
    [anon_sym_let] = ACTIONS(938),
    [anon_sym_const] = ACTIONS(938),
    [anon_sym_using] = ACTIONS(940),
    [anon_sym_await] = ACTIONS(942),
    [anon_sym_LPAREN] = ACTIONS(944),
    [anon_sym_SEMI] = ACTIONS(169),
    [anon_sym_in] = ACTIONS(165),
    [anon_sym_COLON] = ACTIONS(183),
    [anon_sym_EQ] = ACTIONS(186),
    [anon_sym_LBRACK] = ACTIONS(169),
    [anon_sym_LT] = ACTIONS(165),
    [anon_sym_GT] = ACTIONS(165),
    [anon_sym_DOT] = ACTIONS(169),
    [anon_sym_class] = ACTIONS(947),
    [anon_sym_async] = ACTIONS(949),
    [anon_sym_function] = ACTIONS(951),
    [anon_sym_EQ_GT] = ACTIONS(200),
    [sym_optional_chain] = ACTIONS(169),
    [anon_sym_PLUS_EQ] = ACTIONS(202),
    [anon_sym_DASH_EQ] = ACTIONS(202),
    [anon_sym_STAR_EQ] = ACTIONS(202),
    [anon_sym_SLASH_EQ] = ACTIONS(202),
    [anon_sym_PERCENT_EQ] = ACTIONS(202),
    [anon_sym_CARET_EQ] = ACTIONS(202),
    [anon_sym_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_LT_LT_EQ] = ACTIONS(202),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(202),
    [anon_sym_QMARK_QMARK_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP] = ACTIONS(165),
    [anon_sym_PIPE_PIPE] = ACTIONS(165),
    [anon_sym_GT_GT] = ACTIONS(165),
    [anon_sym_GT_GT_GT] = ACTIONS(165),
    [anon_sym_LT_LT] = ACTIONS(165),
    [anon_sym_AMP] = ACTIONS(165),
    [anon_sym_CARET] = ACTIONS(165),
    [anon_sym_PIPE] = ACTIONS(165),
    [anon_sym_PLUS] = ACTIONS(165),
    [anon_sym_DASH] = ACTIONS(165),
    [anon_sym_SLASH] = ACTIONS(165),
    [anon_sym_PERCENT] = ACTIONS(165),
    [anon_sym_STAR_STAR] = ACTIONS(165),
    [anon_sym_LT_EQ] = ACTIONS(169),
    [anon_sym_EQ_EQ] = ACTIONS(165),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(169),
    [anon_sym_BANG_EQ] = ACTIONS(165),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(169),
    [anon_sym_GT_EQ] = ACTIONS(169),
    [anon_sym_QMARK_QMARK] = ACTIONS(165),
    [anon_sym_instanceof] = ACTIONS(169),
    [anon_sym_PLUS_PLUS] = ACTIONS(169),
    [anon_sym_DASH_DASH] = ACTIONS(169),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(169),
    [anon_sym_AT] = ACTIONS(93),
    [sym__automatic_semicolon] = ACTIONS(169),
    [sym__ternary_qmark] = ACTIONS(169),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(134)] = {
    [sym_namespace_export] = STATE(1692),
    [sym_export_clause] = STATE(1358),
    [sym_declaration] = STATE(433),
    [sym_variable_declaration] = STATE(419),
    [sym_lexical_declaration] = STATE(419),
    [sym_using_declaration] = STATE(419),
    [sym_class_declaration] = STATE(419),
    [sym_function_declaration] = STATE(419),
    [sym_generator_function_declaration] = STATE(419),
    [sym_decorator] = STATE(1097),
    [aux_sym_export_statement_repeat1] = STATE(1327),
    [aux_sym_object_repeat1] = STATE(1487),
    [aux_sym_object_pattern_repeat1] = STATE(1452),
    [anon_sym_STAR] = ACTIONS(930),
    [anon_sym_default] = ACTIONS(932),
    [anon_sym_LBRACE] = ACTIONS(934),
    [anon_sym_COMMA] = ACTIONS(169),
    [anon_sym_RBRACE] = ACTIONS(171),
    [anon_sym_var] = ACTIONS(936),
    [anon_sym_let] = ACTIONS(938),
    [anon_sym_const] = ACTIONS(938),
    [anon_sym_using] = ACTIONS(940),
    [anon_sym_await] = ACTIONS(942),
    [anon_sym_LPAREN] = ACTIONS(944),
    [anon_sym_SEMI] = ACTIONS(169),
    [anon_sym_in] = ACTIONS(165),
    [anon_sym_COLON] = ACTIONS(183),
    [anon_sym_EQ] = ACTIONS(186),
    [anon_sym_LBRACK] = ACTIONS(169),
    [anon_sym_LT] = ACTIONS(165),
    [anon_sym_GT] = ACTIONS(165),
    [anon_sym_DOT] = ACTIONS(169),
    [anon_sym_class] = ACTIONS(947),
    [anon_sym_async] = ACTIONS(949),
    [anon_sym_function] = ACTIONS(951),
    [anon_sym_EQ_GT] = ACTIONS(200),
    [sym_optional_chain] = ACTIONS(169),
    [anon_sym_PLUS_EQ] = ACTIONS(202),
    [anon_sym_DASH_EQ] = ACTIONS(202),
    [anon_sym_STAR_EQ] = ACTIONS(202),
    [anon_sym_SLASH_EQ] = ACTIONS(202),
    [anon_sym_PERCENT_EQ] = ACTIONS(202),
    [anon_sym_CARET_EQ] = ACTIONS(202),
    [anon_sym_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_LT_LT_EQ] = ACTIONS(202),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(202),
    [anon_sym_QMARK_QMARK_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP] = ACTIONS(165),
    [anon_sym_PIPE_PIPE] = ACTIONS(165),
    [anon_sym_GT_GT] = ACTIONS(165),
    [anon_sym_GT_GT_GT] = ACTIONS(165),
    [anon_sym_LT_LT] = ACTIONS(165),
    [anon_sym_AMP] = ACTIONS(165),
    [anon_sym_CARET] = ACTIONS(165),
    [anon_sym_PIPE] = ACTIONS(165),
    [anon_sym_PLUS] = ACTIONS(165),
    [anon_sym_DASH] = ACTIONS(165),
    [anon_sym_SLASH] = ACTIONS(165),
    [anon_sym_PERCENT] = ACTIONS(165),
    [anon_sym_STAR_STAR] = ACTIONS(165),
    [anon_sym_LT_EQ] = ACTIONS(169),
    [anon_sym_EQ_EQ] = ACTIONS(165),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(169),
    [anon_sym_BANG_EQ] = ACTIONS(165),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(169),
    [anon_sym_GT_EQ] = ACTIONS(169),
    [anon_sym_QMARK_QMARK] = ACTIONS(165),
    [anon_sym_instanceof] = ACTIONS(169),
    [anon_sym_PLUS_PLUS] = ACTIONS(169),
    [anon_sym_DASH_DASH] = ACTIONS(169),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(169),
    [anon_sym_AT] = ACTIONS(93),
    [sym__automatic_semicolon] = ACTIONS(169),
    [sym__ternary_qmark] = ACTIONS(169),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(135)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(498),
    [sym_expression] = STATE(979),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1307),
    [sym_assignment_pattern] = STATE(1661),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1307),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(558),
    [sym_subscript_expression] = STATE(558),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1138),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1307),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1816),
    [sym_pattern] = STATE(1474),
    [sym_rest_pattern] = STATE(1280),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(867),
    [anon_sym_export] = ACTIONS(869),
    [anon_sym_LBRACE] = ACTIONS(871),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(869),
    [anon_sym_await] = ACTIONS(875),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(438),
    [anon_sym_LBRACK] = ACTIONS(877),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(881),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(447),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(883),
    [anon_sym_PLUS] = ACTIONS(452),
    [anon_sym_DASH] = ACTIONS(452),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(454),
    [anon_sym_TILDE] = ACTIONS(454),
    [anon_sym_typeof] = ACTIONS(452),
    [anon_sym_void] = ACTIONS(452),
    [anon_sym_delete] = ACTIONS(452),
    [anon_sym_PLUS_PLUS] = ACTIONS(684),
    [anon_sym_DASH_DASH] = ACTIONS(684),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(459),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(885),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(869),
    [anon_sym_get] = ACTIONS(869),
    [anon_sym_set] = ACTIONS(869),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(136)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(507),
    [sym_expression] = STATE(880),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1815),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1815),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(507),
    [sym_subscript_expression] = STATE(507),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1139),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1815),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_sequence_expression] = STATE(1854),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1763),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(369),
    [anon_sym_export] = ACTIONS(371),
    [anon_sym_LBRACE] = ACTIONS(373),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(371),
    [anon_sym_await] = ACTIONS(377),
    [anon_sym_of] = ACTIONS(860),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_in] = ACTIONS(860),
    [anon_sym_yield] = ACTIONS(382),
    [anon_sym_LBRACK] = ACTIONS(654),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(398),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(404),
    [anon_sym_PLUS] = ACTIONS(412),
    [anon_sym_DASH] = ACTIONS(412),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(414),
    [anon_sym_TILDE] = ACTIONS(414),
    [anon_sym_typeof] = ACTIONS(412),
    [anon_sym_void] = ACTIONS(412),
    [anon_sym_delete] = ACTIONS(412),
    [anon_sym_PLUS_PLUS] = ACTIONS(660),
    [anon_sym_DASH_DASH] = ACTIONS(660),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(424),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(428),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(371),
    [anon_sym_get] = ACTIONS(371),
    [anon_sym_set] = ACTIONS(371),
    [sym__automatic_semicolon] = ACTIONS(953),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(137)] = {
    [sym_namespace_export] = STATE(1692),
    [sym_export_clause] = STATE(1358),
    [sym_declaration] = STATE(433),
    [sym_variable_declaration] = STATE(419),
    [sym_lexical_declaration] = STATE(419),
    [sym_using_declaration] = STATE(419),
    [sym_class_declaration] = STATE(419),
    [sym_function_declaration] = STATE(419),
    [sym_generator_function_declaration] = STATE(419),
    [sym_decorator] = STATE(1097),
    [aux_sym_export_statement_repeat1] = STATE(1327),
    [aux_sym_object_repeat1] = STATE(1418),
    [aux_sym_object_pattern_repeat1] = STATE(1452),
    [anon_sym_STAR] = ACTIONS(930),
    [anon_sym_default] = ACTIONS(932),
    [anon_sym_LBRACE] = ACTIONS(934),
    [anon_sym_COMMA] = ACTIONS(169),
    [anon_sym_RBRACE] = ACTIONS(216),
    [anon_sym_var] = ACTIONS(936),
    [anon_sym_let] = ACTIONS(938),
    [anon_sym_const] = ACTIONS(938),
    [anon_sym_using] = ACTIONS(940),
    [anon_sym_await] = ACTIONS(942),
    [anon_sym_LPAREN] = ACTIONS(944),
    [anon_sym_SEMI] = ACTIONS(169),
    [anon_sym_in] = ACTIONS(165),
    [anon_sym_COLON] = ACTIONS(183),
    [anon_sym_EQ] = ACTIONS(186),
    [anon_sym_LBRACK] = ACTIONS(169),
    [anon_sym_LT] = ACTIONS(165),
    [anon_sym_GT] = ACTIONS(165),
    [anon_sym_DOT] = ACTIONS(169),
    [anon_sym_class] = ACTIONS(947),
    [anon_sym_async] = ACTIONS(949),
    [anon_sym_function] = ACTIONS(951),
    [anon_sym_EQ_GT] = ACTIONS(200),
    [sym_optional_chain] = ACTIONS(169),
    [anon_sym_PLUS_EQ] = ACTIONS(202),
    [anon_sym_DASH_EQ] = ACTIONS(202),
    [anon_sym_STAR_EQ] = ACTIONS(202),
    [anon_sym_SLASH_EQ] = ACTIONS(202),
    [anon_sym_PERCENT_EQ] = ACTIONS(202),
    [anon_sym_CARET_EQ] = ACTIONS(202),
    [anon_sym_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_LT_LT_EQ] = ACTIONS(202),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(202),
    [anon_sym_QMARK_QMARK_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP] = ACTIONS(165),
    [anon_sym_PIPE_PIPE] = ACTIONS(165),
    [anon_sym_GT_GT] = ACTIONS(165),
    [anon_sym_GT_GT_GT] = ACTIONS(165),
    [anon_sym_LT_LT] = ACTIONS(165),
    [anon_sym_AMP] = ACTIONS(165),
    [anon_sym_CARET] = ACTIONS(165),
    [anon_sym_PIPE] = ACTIONS(165),
    [anon_sym_PLUS] = ACTIONS(165),
    [anon_sym_DASH] = ACTIONS(165),
    [anon_sym_SLASH] = ACTIONS(165),
    [anon_sym_PERCENT] = ACTIONS(165),
    [anon_sym_STAR_STAR] = ACTIONS(165),
    [anon_sym_LT_EQ] = ACTIONS(169),
    [anon_sym_EQ_EQ] = ACTIONS(165),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(169),
    [anon_sym_BANG_EQ] = ACTIONS(165),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(169),
    [anon_sym_GT_EQ] = ACTIONS(169),
    [anon_sym_QMARK_QMARK] = ACTIONS(165),
    [anon_sym_instanceof] = ACTIONS(169),
    [anon_sym_PLUS_PLUS] = ACTIONS(169),
    [anon_sym_DASH_DASH] = ACTIONS(169),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(169),
    [anon_sym_AT] = ACTIONS(93),
    [sym__automatic_semicolon] = ACTIONS(169),
    [sym__ternary_qmark] = ACTIONS(169),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(138)] = {
    [sym_import] = STATE(1333),
    [sym_empty_statement] = STATE(151),
    [sym_parenthesized_expression] = STATE(507),
    [sym_expression] = STATE(885),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1815),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1815),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(507),
    [sym_subscript_expression] = STATE(507),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1139),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1815),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_sequence_expression] = STATE(1852),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1763),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(369),
    [anon_sym_export] = ACTIONS(371),
    [anon_sym_LBRACE] = ACTIONS(373),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(371),
    [anon_sym_await] = ACTIONS(377),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_SEMI] = ACTIONS(39),
    [anon_sym_yield] = ACTIONS(382),
    [anon_sym_LBRACK] = ACTIONS(654),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(398),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(404),
    [anon_sym_PLUS] = ACTIONS(412),
    [anon_sym_DASH] = ACTIONS(412),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(414),
    [anon_sym_TILDE] = ACTIONS(414),
    [anon_sym_typeof] = ACTIONS(412),
    [anon_sym_void] = ACTIONS(412),
    [anon_sym_delete] = ACTIONS(412),
    [anon_sym_PLUS_PLUS] = ACTIONS(660),
    [anon_sym_DASH_DASH] = ACTIONS(660),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(424),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(428),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(371),
    [anon_sym_get] = ACTIONS(371),
    [anon_sym_set] = ACTIONS(371),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(139)] = {
    [sym_import] = STATE(1333),
    [sym_empty_statement] = STATE(150),
    [sym_parenthesized_expression] = STATE(507),
    [sym_expression] = STATE(883),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1815),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1815),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(507),
    [sym_subscript_expression] = STATE(507),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1139),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1815),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_sequence_expression] = STATE(1835),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1763),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(369),
    [anon_sym_export] = ACTIONS(371),
    [anon_sym_LBRACE] = ACTIONS(373),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(371),
    [anon_sym_await] = ACTIONS(377),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_SEMI] = ACTIONS(39),
    [anon_sym_yield] = ACTIONS(382),
    [anon_sym_LBRACK] = ACTIONS(654),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(398),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(404),
    [anon_sym_PLUS] = ACTIONS(412),
    [anon_sym_DASH] = ACTIONS(412),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(414),
    [anon_sym_TILDE] = ACTIONS(414),
    [anon_sym_typeof] = ACTIONS(412),
    [anon_sym_void] = ACTIONS(412),
    [anon_sym_delete] = ACTIONS(412),
    [anon_sym_PLUS_PLUS] = ACTIONS(660),
    [anon_sym_DASH_DASH] = ACTIONS(660),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(424),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(428),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(371),
    [anon_sym_get] = ACTIONS(371),
    [anon_sym_set] = ACTIONS(371),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(140)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(580),
    [sym_expression] = STATE(979),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1472),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1472),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(580),
    [sym_subscript_expression] = STATE(580),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1138),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1472),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1816),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(955),
    [anon_sym_export] = ACTIONS(957),
    [anon_sym_LBRACE] = ACTIONS(840),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_var] = ACTIONS(959),
    [anon_sym_let] = ACTIONS(961),
    [anon_sym_const] = ACTIONS(848),
    [anon_sym_using] = ACTIONS(848),
    [anon_sym_await] = ACTIONS(963),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(438),
    [anon_sym_LBRACK] = ACTIONS(852),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(965),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(447),
    [anon_sym_PLUS] = ACTIONS(452),
    [anon_sym_DASH] = ACTIONS(452),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(454),
    [anon_sym_TILDE] = ACTIONS(454),
    [anon_sym_typeof] = ACTIONS(452),
    [anon_sym_void] = ACTIONS(452),
    [anon_sym_delete] = ACTIONS(452),
    [anon_sym_PLUS_PLUS] = ACTIONS(684),
    [anon_sym_DASH_DASH] = ACTIONS(684),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(459),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(967),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(957),
    [anon_sym_get] = ACTIONS(957),
    [anon_sym_set] = ACTIONS(957),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(141)] = {
    [sym_import] = STATE(1333),
    [sym_empty_statement] = STATE(144),
    [sym_parenthesized_expression] = STATE(507),
    [sym_expression] = STATE(863),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1815),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1815),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(507),
    [sym_subscript_expression] = STATE(507),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1139),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1815),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_sequence_expression] = STATE(1789),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1763),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(369),
    [anon_sym_export] = ACTIONS(371),
    [anon_sym_LBRACE] = ACTIONS(373),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(371),
    [anon_sym_await] = ACTIONS(377),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_SEMI] = ACTIONS(39),
    [anon_sym_yield] = ACTIONS(382),
    [anon_sym_LBRACK] = ACTIONS(654),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(398),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(404),
    [anon_sym_PLUS] = ACTIONS(412),
    [anon_sym_DASH] = ACTIONS(412),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(414),
    [anon_sym_TILDE] = ACTIONS(414),
    [anon_sym_typeof] = ACTIONS(412),
    [anon_sym_void] = ACTIONS(412),
    [anon_sym_delete] = ACTIONS(412),
    [anon_sym_PLUS_PLUS] = ACTIONS(660),
    [anon_sym_DASH_DASH] = ACTIONS(660),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(424),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(428),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(371),
    [anon_sym_get] = ACTIONS(371),
    [anon_sym_set] = ACTIONS(371),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(142)] = {
    [sym_import] = STATE(1326),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(822),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1790),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1790),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_function_expression] = STATE(740),
    [sym_generator_function] = STATE(740),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1790),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_sequence_expression] = STATE(1695),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [aux_sym_export_statement_repeat1] = STATE(1381),
    [sym_identifier] = ACTIONS(161),
    [anon_sym_export] = ACTIONS(163),
    [anon_sym_LBRACE] = ACTIONS(167),
    [anon_sym_import] = ACTIONS(173),
    [anon_sym_let] = ACTIONS(163),
    [anon_sym_await] = ACTIONS(177),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_SEMI] = ACTIONS(969),
    [anon_sym_yield] = ACTIONS(57),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(194),
    [anon_sym_async] = ACTIONS(196),
    [anon_sym_function] = ACTIONS(198),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(75),
    [anon_sym_DASH] = ACTIONS(75),
    [anon_sym_SLASH] = ACTIONS(77),
    [anon_sym_BANG] = ACTIONS(79),
    [anon_sym_TILDE] = ACTIONS(79),
    [anon_sym_typeof] = ACTIONS(75),
    [anon_sym_void] = ACTIONS(75),
    [anon_sym_delete] = ACTIONS(75),
    [anon_sym_PLUS_PLUS] = ACTIONS(81),
    [anon_sym_DASH_DASH] = ACTIONS(81),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(87),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(163),
    [anon_sym_get] = ACTIONS(163),
    [anon_sym_set] = ACTIONS(163),
    [sym__automatic_semicolon] = ACTIONS(969),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(143)] = {
    [sym_import] = STATE(1333),
    [sym_empty_statement] = STATE(147),
    [sym_parenthesized_expression] = STATE(507),
    [sym_expression] = STATE(884),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1815),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1815),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(507),
    [sym_subscript_expression] = STATE(507),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1139),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1815),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_sequence_expression] = STATE(1843),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1763),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(369),
    [anon_sym_export] = ACTIONS(371),
    [anon_sym_LBRACE] = ACTIONS(373),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(371),
    [anon_sym_await] = ACTIONS(377),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_SEMI] = ACTIONS(39),
    [anon_sym_yield] = ACTIONS(382),
    [anon_sym_LBRACK] = ACTIONS(654),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(398),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(404),
    [anon_sym_PLUS] = ACTIONS(412),
    [anon_sym_DASH] = ACTIONS(412),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(414),
    [anon_sym_TILDE] = ACTIONS(414),
    [anon_sym_typeof] = ACTIONS(412),
    [anon_sym_void] = ACTIONS(412),
    [anon_sym_delete] = ACTIONS(412),
    [anon_sym_PLUS_PLUS] = ACTIONS(660),
    [anon_sym_DASH_DASH] = ACTIONS(660),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(424),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(428),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(371),
    [anon_sym_get] = ACTIONS(371),
    [anon_sym_set] = ACTIONS(371),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(144)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(507),
    [sym_expression] = STATE(882),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1815),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1815),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(507),
    [sym_subscript_expression] = STATE(507),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1139),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1815),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_sequence_expression] = STATE(1831),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1763),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(369),
    [anon_sym_export] = ACTIONS(371),
    [anon_sym_LBRACE] = ACTIONS(373),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(371),
    [anon_sym_await] = ACTIONS(377),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_RPAREN] = ACTIONS(971),
    [anon_sym_yield] = ACTIONS(382),
    [anon_sym_LBRACK] = ACTIONS(654),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(398),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(404),
    [anon_sym_PLUS] = ACTIONS(412),
    [anon_sym_DASH] = ACTIONS(412),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(414),
    [anon_sym_TILDE] = ACTIONS(414),
    [anon_sym_typeof] = ACTIONS(412),
    [anon_sym_void] = ACTIONS(412),
    [anon_sym_delete] = ACTIONS(412),
    [anon_sym_PLUS_PLUS] = ACTIONS(660),
    [anon_sym_DASH_DASH] = ACTIONS(660),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(424),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(428),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(371),
    [anon_sym_get] = ACTIONS(371),
    [anon_sym_set] = ACTIONS(371),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(145)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(507),
    [sym_expression] = STATE(840),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1815),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1815),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(507),
    [sym_subscript_expression] = STATE(507),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1139),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1815),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_sequence_expression] = STATE(1795),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1763),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(369),
    [anon_sym_export] = ACTIONS(371),
    [anon_sym_LBRACE] = ACTIONS(373),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(371),
    [anon_sym_await] = ACTIONS(377),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_RPAREN] = ACTIONS(973),
    [anon_sym_yield] = ACTIONS(382),
    [anon_sym_LBRACK] = ACTIONS(654),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(398),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(404),
    [anon_sym_PLUS] = ACTIONS(412),
    [anon_sym_DASH] = ACTIONS(412),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(414),
    [anon_sym_TILDE] = ACTIONS(414),
    [anon_sym_typeof] = ACTIONS(412),
    [anon_sym_void] = ACTIONS(412),
    [anon_sym_delete] = ACTIONS(412),
    [anon_sym_PLUS_PLUS] = ACTIONS(660),
    [anon_sym_DASH_DASH] = ACTIONS(660),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(424),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(428),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(371),
    [anon_sym_get] = ACTIONS(371),
    [anon_sym_set] = ACTIONS(371),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(146)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(507),
    [sym_expression] = STATE(876),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1815),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1815),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(507),
    [sym_subscript_expression] = STATE(507),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1139),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1815),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_sequence_expression] = STATE(1812),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1763),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(369),
    [anon_sym_export] = ACTIONS(371),
    [anon_sym_LBRACE] = ACTIONS(373),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(371),
    [anon_sym_await] = ACTIONS(377),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_RPAREN] = ACTIONS(975),
    [anon_sym_yield] = ACTIONS(382),
    [anon_sym_LBRACK] = ACTIONS(654),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(398),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(404),
    [anon_sym_PLUS] = ACTIONS(412),
    [anon_sym_DASH] = ACTIONS(412),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(414),
    [anon_sym_TILDE] = ACTIONS(414),
    [anon_sym_typeof] = ACTIONS(412),
    [anon_sym_void] = ACTIONS(412),
    [anon_sym_delete] = ACTIONS(412),
    [anon_sym_PLUS_PLUS] = ACTIONS(660),
    [anon_sym_DASH_DASH] = ACTIONS(660),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(424),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(428),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(371),
    [anon_sym_get] = ACTIONS(371),
    [anon_sym_set] = ACTIONS(371),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(147)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(507),
    [sym_expression] = STATE(874),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1815),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1815),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(507),
    [sym_subscript_expression] = STATE(507),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1139),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1815),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_sequence_expression] = STATE(1838),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1763),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(369),
    [anon_sym_export] = ACTIONS(371),
    [anon_sym_LBRACE] = ACTIONS(373),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(371),
    [anon_sym_await] = ACTIONS(377),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_RPAREN] = ACTIONS(977),
    [anon_sym_yield] = ACTIONS(382),
    [anon_sym_LBRACK] = ACTIONS(654),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(398),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(404),
    [anon_sym_PLUS] = ACTIONS(412),
    [anon_sym_DASH] = ACTIONS(412),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(414),
    [anon_sym_TILDE] = ACTIONS(414),
    [anon_sym_typeof] = ACTIONS(412),
    [anon_sym_void] = ACTIONS(412),
    [anon_sym_delete] = ACTIONS(412),
    [anon_sym_PLUS_PLUS] = ACTIONS(660),
    [anon_sym_DASH_DASH] = ACTIONS(660),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(424),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(428),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(371),
    [anon_sym_get] = ACTIONS(371),
    [anon_sym_set] = ACTIONS(371),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(148)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(507),
    [sym_expression] = STATE(878),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1815),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1815),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(507),
    [sym_subscript_expression] = STATE(507),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1139),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1815),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_sequence_expression] = STATE(1764),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1763),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(369),
    [anon_sym_export] = ACTIONS(371),
    [anon_sym_LBRACE] = ACTIONS(373),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(371),
    [anon_sym_await] = ACTIONS(377),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_RPAREN] = ACTIONS(979),
    [anon_sym_yield] = ACTIONS(382),
    [anon_sym_LBRACK] = ACTIONS(654),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(398),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(404),
    [anon_sym_PLUS] = ACTIONS(412),
    [anon_sym_DASH] = ACTIONS(412),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(414),
    [anon_sym_TILDE] = ACTIONS(414),
    [anon_sym_typeof] = ACTIONS(412),
    [anon_sym_void] = ACTIONS(412),
    [anon_sym_delete] = ACTIONS(412),
    [anon_sym_PLUS_PLUS] = ACTIONS(660),
    [anon_sym_DASH_DASH] = ACTIONS(660),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(424),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(428),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(371),
    [anon_sym_get] = ACTIONS(371),
    [anon_sym_set] = ACTIONS(371),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(149)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(507),
    [sym_expression] = STATE(831),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1815),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1815),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(507),
    [sym_subscript_expression] = STATE(507),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1139),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1815),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_sequence_expression] = STATE(1760),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1763),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(369),
    [anon_sym_export] = ACTIONS(371),
    [anon_sym_LBRACE] = ACTIONS(373),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(371),
    [anon_sym_await] = ACTIONS(377),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_RPAREN] = ACTIONS(981),
    [anon_sym_yield] = ACTIONS(382),
    [anon_sym_LBRACK] = ACTIONS(654),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(398),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(404),
    [anon_sym_PLUS] = ACTIONS(412),
    [anon_sym_DASH] = ACTIONS(412),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(414),
    [anon_sym_TILDE] = ACTIONS(414),
    [anon_sym_typeof] = ACTIONS(412),
    [anon_sym_void] = ACTIONS(412),
    [anon_sym_delete] = ACTIONS(412),
    [anon_sym_PLUS_PLUS] = ACTIONS(660),
    [anon_sym_DASH_DASH] = ACTIONS(660),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(424),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(428),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(371),
    [anon_sym_get] = ACTIONS(371),
    [anon_sym_set] = ACTIONS(371),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(150)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(507),
    [sym_expression] = STATE(832),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1815),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1815),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(507),
    [sym_subscript_expression] = STATE(507),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1139),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1815),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_sequence_expression] = STATE(1748),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1763),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(369),
    [anon_sym_export] = ACTIONS(371),
    [anon_sym_LBRACE] = ACTIONS(373),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(371),
    [anon_sym_await] = ACTIONS(377),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_RPAREN] = ACTIONS(983),
    [anon_sym_yield] = ACTIONS(382),
    [anon_sym_LBRACK] = ACTIONS(654),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(398),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(404),
    [anon_sym_PLUS] = ACTIONS(412),
    [anon_sym_DASH] = ACTIONS(412),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(414),
    [anon_sym_TILDE] = ACTIONS(414),
    [anon_sym_typeof] = ACTIONS(412),
    [anon_sym_void] = ACTIONS(412),
    [anon_sym_delete] = ACTIONS(412),
    [anon_sym_PLUS_PLUS] = ACTIONS(660),
    [anon_sym_DASH_DASH] = ACTIONS(660),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(424),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(428),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(371),
    [anon_sym_get] = ACTIONS(371),
    [anon_sym_set] = ACTIONS(371),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(151)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(507),
    [sym_expression] = STATE(877),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1815),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1815),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(507),
    [sym_subscript_expression] = STATE(507),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1139),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1815),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_sequence_expression] = STATE(1750),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1763),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(369),
    [anon_sym_export] = ACTIONS(371),
    [anon_sym_LBRACE] = ACTIONS(373),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(371),
    [anon_sym_await] = ACTIONS(377),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_RPAREN] = ACTIONS(985),
    [anon_sym_yield] = ACTIONS(382),
    [anon_sym_LBRACK] = ACTIONS(654),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(398),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(404),
    [anon_sym_PLUS] = ACTIONS(412),
    [anon_sym_DASH] = ACTIONS(412),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(414),
    [anon_sym_TILDE] = ACTIONS(414),
    [anon_sym_typeof] = ACTIONS(412),
    [anon_sym_void] = ACTIONS(412),
    [anon_sym_delete] = ACTIONS(412),
    [anon_sym_PLUS_PLUS] = ACTIONS(660),
    [anon_sym_DASH_DASH] = ACTIONS(660),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(424),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(428),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(371),
    [anon_sym_get] = ACTIONS(371),
    [anon_sym_set] = ACTIONS(371),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(152)] = {
    [sym_import] = STATE(1333),
    [sym_statement_block] = STATE(628),
    [sym_parenthesized_expression] = STATE(507),
    [sym_expression] = STATE(666),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1815),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1815),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(507),
    [sym_subscript_expression] = STATE(507),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1139),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1815),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1763),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(369),
    [anon_sym_export] = ACTIONS(371),
    [anon_sym_LBRACE] = ACTIONS(987),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(371),
    [anon_sym_await] = ACTIONS(377),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(382),
    [anon_sym_LBRACK] = ACTIONS(654),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(398),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(404),
    [anon_sym_PLUS] = ACTIONS(412),
    [anon_sym_DASH] = ACTIONS(412),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(414),
    [anon_sym_TILDE] = ACTIONS(414),
    [anon_sym_typeof] = ACTIONS(412),
    [anon_sym_void] = ACTIONS(412),
    [anon_sym_delete] = ACTIONS(412),
    [anon_sym_PLUS_PLUS] = ACTIONS(660),
    [anon_sym_DASH_DASH] = ACTIONS(660),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(424),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(428),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(371),
    [anon_sym_get] = ACTIONS(371),
    [anon_sym_set] = ACTIONS(371),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(153)] = {
    [sym_import] = STATE(1333),
    [sym_statement_block] = STATE(628),
    [sym_parenthesized_expression] = STATE(578),
    [sym_expression] = STATE(910),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1776),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1776),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(578),
    [sym_subscript_expression] = STATE(578),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1142),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1776),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1778),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(613),
    [anon_sym_export] = ACTIONS(615),
    [anon_sym_LBRACE] = ACTIONS(989),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(615),
    [anon_sym_await] = ACTIONS(620),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(622),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(626),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(630),
    [anon_sym_PLUS] = ACTIONS(635),
    [anon_sym_DASH] = ACTIONS(635),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(637),
    [anon_sym_TILDE] = ACTIONS(637),
    [anon_sym_typeof] = ACTIONS(635),
    [anon_sym_void] = ACTIONS(635),
    [anon_sym_delete] = ACTIONS(635),
    [anon_sym_PLUS_PLUS] = ACTIONS(688),
    [anon_sym_DASH_DASH] = ACTIONS(688),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(642),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(644),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(615),
    [anon_sym_get] = ACTIONS(615),
    [anon_sym_set] = ACTIONS(615),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(154)] = {
    [sym_import] = STATE(1333),
    [sym_statement_block] = STATE(629),
    [sym_parenthesized_expression] = STATE(578),
    [sym_expression] = STATE(911),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1776),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1776),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(578),
    [sym_subscript_expression] = STATE(578),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1142),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1776),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1778),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(613),
    [anon_sym_export] = ACTIONS(615),
    [anon_sym_LBRACE] = ACTIONS(989),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(615),
    [anon_sym_await] = ACTIONS(620),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(622),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(626),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(630),
    [anon_sym_PLUS] = ACTIONS(635),
    [anon_sym_DASH] = ACTIONS(635),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(637),
    [anon_sym_TILDE] = ACTIONS(637),
    [anon_sym_typeof] = ACTIONS(635),
    [anon_sym_void] = ACTIONS(635),
    [anon_sym_delete] = ACTIONS(635),
    [anon_sym_PLUS_PLUS] = ACTIONS(688),
    [anon_sym_DASH_DASH] = ACTIONS(688),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(642),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(644),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(615),
    [anon_sym_get] = ACTIONS(615),
    [anon_sym_set] = ACTIONS(615),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(155)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(507),
    [sym_expression] = STATE(880),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1815),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1815),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(507),
    [sym_subscript_expression] = STATE(507),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1139),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1815),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_sequence_expression] = STATE(1854),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1763),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(369),
    [anon_sym_export] = ACTIONS(371),
    [anon_sym_LBRACE] = ACTIONS(373),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(371),
    [anon_sym_await] = ACTIONS(377),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(382),
    [anon_sym_LBRACK] = ACTIONS(654),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(398),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(404),
    [anon_sym_PLUS] = ACTIONS(412),
    [anon_sym_DASH] = ACTIONS(412),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(414),
    [anon_sym_TILDE] = ACTIONS(414),
    [anon_sym_typeof] = ACTIONS(412),
    [anon_sym_void] = ACTIONS(412),
    [anon_sym_delete] = ACTIONS(412),
    [anon_sym_PLUS_PLUS] = ACTIONS(660),
    [anon_sym_DASH_DASH] = ACTIONS(660),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(424),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(428),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(371),
    [anon_sym_get] = ACTIONS(371),
    [anon_sym_set] = ACTIONS(371),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(156)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(507),
    [sym_expression] = STATE(836),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1815),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1815),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(507),
    [sym_subscript_expression] = STATE(507),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1139),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1815),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_sequence_expression] = STATE(1837),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1763),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(369),
    [anon_sym_export] = ACTIONS(371),
    [anon_sym_LBRACE] = ACTIONS(373),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(371),
    [anon_sym_await] = ACTIONS(377),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(382),
    [anon_sym_LBRACK] = ACTIONS(654),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(398),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(404),
    [anon_sym_PLUS] = ACTIONS(412),
    [anon_sym_DASH] = ACTIONS(412),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(414),
    [anon_sym_TILDE] = ACTIONS(414),
    [anon_sym_typeof] = ACTIONS(412),
    [anon_sym_void] = ACTIONS(412),
    [anon_sym_delete] = ACTIONS(412),
    [anon_sym_PLUS_PLUS] = ACTIONS(660),
    [anon_sym_DASH_DASH] = ACTIONS(660),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(424),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(428),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(371),
    [anon_sym_get] = ACTIONS(371),
    [anon_sym_set] = ACTIONS(371),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(157)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(498),
    [sym_expression] = STATE(979),
    [sym_primary_expression] = STATE(581),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1762),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1762),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(583),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(498),
    [sym_subscript_expression] = STATE(498),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1138),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1762),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1816),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(430),
    [anon_sym_export] = ACTIONS(432),
    [anon_sym_LBRACE] = ACTIONS(434),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(432),
    [anon_sym_await] = ACTIONS(436),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(438),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DOT] = ACTIONS(991),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(445),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(447),
    [anon_sym_PLUS] = ACTIONS(452),
    [anon_sym_DASH] = ACTIONS(452),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(454),
    [anon_sym_TILDE] = ACTIONS(454),
    [anon_sym_typeof] = ACTIONS(452),
    [anon_sym_void] = ACTIONS(452),
    [anon_sym_delete] = ACTIONS(452),
    [anon_sym_PLUS_PLUS] = ACTIONS(684),
    [anon_sym_DASH_DASH] = ACTIONS(684),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(459),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(461),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(432),
    [anon_sym_get] = ACTIONS(432),
    [anon_sym_set] = ACTIONS(432),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(158)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(507),
    [sym_expression] = STATE(886),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1815),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1815),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(507),
    [sym_subscript_expression] = STATE(507),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1139),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1815),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_sequence_expression] = STATE(1773),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1763),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(369),
    [anon_sym_export] = ACTIONS(371),
    [anon_sym_LBRACE] = ACTIONS(373),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(371),
    [anon_sym_await] = ACTIONS(377),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(382),
    [anon_sym_LBRACK] = ACTIONS(654),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(398),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(404),
    [anon_sym_PLUS] = ACTIONS(412),
    [anon_sym_DASH] = ACTIONS(412),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(414),
    [anon_sym_TILDE] = ACTIONS(414),
    [anon_sym_typeof] = ACTIONS(412),
    [anon_sym_void] = ACTIONS(412),
    [anon_sym_delete] = ACTIONS(412),
    [anon_sym_PLUS_PLUS] = ACTIONS(660),
    [anon_sym_DASH_DASH] = ACTIONS(660),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(424),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(428),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(371),
    [anon_sym_get] = ACTIONS(371),
    [anon_sym_set] = ACTIONS(371),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(159)] = {
    [sym_import] = STATE(1326),
    [sym_statement_block] = STATE(737),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(729),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1790),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1790),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_function_expression] = STATE(740),
    [sym_generator_function] = STATE(740),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1790),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [aux_sym_export_statement_repeat1] = STATE(1381),
    [sym_identifier] = ACTIONS(161),
    [anon_sym_export] = ACTIONS(163),
    [anon_sym_LBRACE] = ACTIONS(993),
    [anon_sym_import] = ACTIONS(173),
    [anon_sym_let] = ACTIONS(163),
    [anon_sym_await] = ACTIONS(177),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_yield] = ACTIONS(57),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(194),
    [anon_sym_async] = ACTIONS(196),
    [anon_sym_function] = ACTIONS(198),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(75),
    [anon_sym_DASH] = ACTIONS(75),
    [anon_sym_SLASH] = ACTIONS(77),
    [anon_sym_BANG] = ACTIONS(79),
    [anon_sym_TILDE] = ACTIONS(79),
    [anon_sym_typeof] = ACTIONS(75),
    [anon_sym_void] = ACTIONS(75),
    [anon_sym_delete] = ACTIONS(75),
    [anon_sym_PLUS_PLUS] = ACTIONS(81),
    [anon_sym_DASH_DASH] = ACTIONS(81),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(87),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(163),
    [anon_sym_get] = ACTIONS(163),
    [anon_sym_set] = ACTIONS(163),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(160)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(507),
    [sym_expression] = STATE(881),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1815),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1815),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(507),
    [sym_subscript_expression] = STATE(507),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1139),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1815),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_sequence_expression] = STATE(1807),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1763),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(369),
    [anon_sym_export] = ACTIONS(371),
    [anon_sym_LBRACE] = ACTIONS(373),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(371),
    [anon_sym_await] = ACTIONS(377),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(382),
    [anon_sym_LBRACK] = ACTIONS(654),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(398),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(404),
    [anon_sym_PLUS] = ACTIONS(412),
    [anon_sym_DASH] = ACTIONS(412),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(414),
    [anon_sym_TILDE] = ACTIONS(414),
    [anon_sym_typeof] = ACTIONS(412),
    [anon_sym_void] = ACTIONS(412),
    [anon_sym_delete] = ACTIONS(412),
    [anon_sym_PLUS_PLUS] = ACTIONS(660),
    [anon_sym_DASH_DASH] = ACTIONS(660),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(424),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(428),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(371),
    [anon_sym_get] = ACTIONS(371),
    [anon_sym_set] = ACTIONS(371),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(161)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(507),
    [sym_expression] = STATE(873),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1815),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1815),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(507),
    [sym_subscript_expression] = STATE(507),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1139),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1815),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_sequence_expression] = STATE(1749),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1763),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(369),
    [anon_sym_export] = ACTIONS(371),
    [anon_sym_LBRACE] = ACTIONS(373),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(371),
    [anon_sym_await] = ACTIONS(377),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(382),
    [anon_sym_LBRACK] = ACTIONS(654),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(398),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(404),
    [anon_sym_PLUS] = ACTIONS(412),
    [anon_sym_DASH] = ACTIONS(412),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(414),
    [anon_sym_TILDE] = ACTIONS(414),
    [anon_sym_typeof] = ACTIONS(412),
    [anon_sym_void] = ACTIONS(412),
    [anon_sym_delete] = ACTIONS(412),
    [anon_sym_PLUS_PLUS] = ACTIONS(660),
    [anon_sym_DASH_DASH] = ACTIONS(660),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(424),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(428),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(371),
    [anon_sym_get] = ACTIONS(371),
    [anon_sym_set] = ACTIONS(371),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(162)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(507),
    [sym_expression] = STATE(837),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1815),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1815),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(507),
    [sym_subscript_expression] = STATE(507),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1139),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1815),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_sequence_expression] = STATE(1842),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1763),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(369),
    [anon_sym_export] = ACTIONS(371),
    [anon_sym_LBRACE] = ACTIONS(373),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(371),
    [anon_sym_await] = ACTIONS(377),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(382),
    [anon_sym_LBRACK] = ACTIONS(654),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(398),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(404),
    [anon_sym_PLUS] = ACTIONS(412),
    [anon_sym_DASH] = ACTIONS(412),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(414),
    [anon_sym_TILDE] = ACTIONS(414),
    [anon_sym_typeof] = ACTIONS(412),
    [anon_sym_void] = ACTIONS(412),
    [anon_sym_delete] = ACTIONS(412),
    [anon_sym_PLUS_PLUS] = ACTIONS(660),
    [anon_sym_DASH_DASH] = ACTIONS(660),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(424),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(428),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(371),
    [anon_sym_get] = ACTIONS(371),
    [anon_sym_set] = ACTIONS(371),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(163)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(507),
    [sym_expression] = STATE(838),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1815),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1815),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(507),
    [sym_subscript_expression] = STATE(507),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1139),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1815),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_sequence_expression] = STATE(1757),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1763),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(369),
    [anon_sym_export] = ACTIONS(371),
    [anon_sym_LBRACE] = ACTIONS(373),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(371),
    [anon_sym_await] = ACTIONS(377),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(382),
    [anon_sym_LBRACK] = ACTIONS(654),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(398),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(404),
    [anon_sym_PLUS] = ACTIONS(412),
    [anon_sym_DASH] = ACTIONS(412),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(414),
    [anon_sym_TILDE] = ACTIONS(414),
    [anon_sym_typeof] = ACTIONS(412),
    [anon_sym_void] = ACTIONS(412),
    [anon_sym_delete] = ACTIONS(412),
    [anon_sym_PLUS_PLUS] = ACTIONS(660),
    [anon_sym_DASH_DASH] = ACTIONS(660),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(424),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(428),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(371),
    [anon_sym_get] = ACTIONS(371),
    [anon_sym_set] = ACTIONS(371),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(164)] = {
    [sym_import] = STATE(1333),
    [sym_statement_block] = STATE(630),
    [sym_parenthesized_expression] = STATE(507),
    [sym_expression] = STATE(658),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1815),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1815),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(507),
    [sym_subscript_expression] = STATE(507),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1139),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1815),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1763),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(369),
    [anon_sym_export] = ACTIONS(371),
    [anon_sym_LBRACE] = ACTIONS(987),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(371),
    [anon_sym_await] = ACTIONS(377),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(382),
    [anon_sym_LBRACK] = ACTIONS(654),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(398),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(404),
    [anon_sym_PLUS] = ACTIONS(412),
    [anon_sym_DASH] = ACTIONS(412),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(414),
    [anon_sym_TILDE] = ACTIONS(414),
    [anon_sym_typeof] = ACTIONS(412),
    [anon_sym_void] = ACTIONS(412),
    [anon_sym_delete] = ACTIONS(412),
    [anon_sym_PLUS_PLUS] = ACTIONS(660),
    [anon_sym_DASH_DASH] = ACTIONS(660),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(424),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(428),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(371),
    [anon_sym_get] = ACTIONS(371),
    [anon_sym_set] = ACTIONS(371),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(165)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(498),
    [sym_expression] = STATE(979),
    [sym_primary_expression] = STATE(581),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1762),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1762),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(583),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(498),
    [sym_subscript_expression] = STATE(498),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1138),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1762),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1778),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(995),
    [anon_sym_export] = ACTIONS(997),
    [anon_sym_LBRACE] = ACTIONS(434),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(997),
    [anon_sym_await] = ACTIONS(999),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(438),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DOT] = ACTIONS(991),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(1001),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(630),
    [anon_sym_PLUS] = ACTIONS(452),
    [anon_sym_DASH] = ACTIONS(452),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(454),
    [anon_sym_TILDE] = ACTIONS(454),
    [anon_sym_typeof] = ACTIONS(452),
    [anon_sym_void] = ACTIONS(452),
    [anon_sym_delete] = ACTIONS(452),
    [anon_sym_PLUS_PLUS] = ACTIONS(684),
    [anon_sym_DASH_DASH] = ACTIONS(684),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(459),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(461),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(997),
    [anon_sym_get] = ACTIONS(997),
    [anon_sym_set] = ACTIONS(997),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(166)] = {
    [sym_namespace_export] = STATE(1692),
    [sym_export_clause] = STATE(1358),
    [sym_declaration] = STATE(433),
    [sym_variable_declaration] = STATE(419),
    [sym_lexical_declaration] = STATE(419),
    [sym_using_declaration] = STATE(419),
    [sym_class_declaration] = STATE(419),
    [sym_function_declaration] = STATE(419),
    [sym_generator_function_declaration] = STATE(419),
    [sym_decorator] = STATE(1097),
    [aux_sym_export_statement_repeat1] = STATE(1327),
    [anon_sym_STAR] = ACTIONS(930),
    [anon_sym_default] = ACTIONS(932),
    [anon_sym_LBRACE] = ACTIONS(934),
    [anon_sym_COMMA] = ACTIONS(169),
    [anon_sym_var] = ACTIONS(936),
    [anon_sym_let] = ACTIONS(938),
    [anon_sym_const] = ACTIONS(938),
    [anon_sym_using] = ACTIONS(940),
    [anon_sym_await] = ACTIONS(942),
    [anon_sym_LPAREN] = ACTIONS(169),
    [anon_sym_SEMI] = ACTIONS(169),
    [anon_sym_in] = ACTIONS(165),
    [anon_sym_COLON] = ACTIONS(494),
    [anon_sym_EQ] = ACTIONS(490),
    [anon_sym_LBRACK] = ACTIONS(169),
    [anon_sym_LT] = ACTIONS(165),
    [anon_sym_GT] = ACTIONS(165),
    [anon_sym_DOT] = ACTIONS(169),
    [anon_sym_class] = ACTIONS(947),
    [anon_sym_async] = ACTIONS(949),
    [anon_sym_function] = ACTIONS(951),
    [anon_sym_EQ_GT] = ACTIONS(200),
    [sym_optional_chain] = ACTIONS(169),
    [anon_sym_PLUS_EQ] = ACTIONS(202),
    [anon_sym_DASH_EQ] = ACTIONS(202),
    [anon_sym_STAR_EQ] = ACTIONS(202),
    [anon_sym_SLASH_EQ] = ACTIONS(202),
    [anon_sym_PERCENT_EQ] = ACTIONS(202),
    [anon_sym_CARET_EQ] = ACTIONS(202),
    [anon_sym_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_LT_LT_EQ] = ACTIONS(202),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(202),
    [anon_sym_QMARK_QMARK_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP] = ACTIONS(165),
    [anon_sym_PIPE_PIPE] = ACTIONS(165),
    [anon_sym_GT_GT] = ACTIONS(165),
    [anon_sym_GT_GT_GT] = ACTIONS(165),
    [anon_sym_LT_LT] = ACTIONS(165),
    [anon_sym_AMP] = ACTIONS(165),
    [anon_sym_CARET] = ACTIONS(165),
    [anon_sym_PIPE] = ACTIONS(165),
    [anon_sym_PLUS] = ACTIONS(165),
    [anon_sym_DASH] = ACTIONS(165),
    [anon_sym_SLASH] = ACTIONS(165),
    [anon_sym_PERCENT] = ACTIONS(165),
    [anon_sym_STAR_STAR] = ACTIONS(165),
    [anon_sym_LT_EQ] = ACTIONS(169),
    [anon_sym_EQ_EQ] = ACTIONS(165),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(169),
    [anon_sym_BANG_EQ] = ACTIONS(165),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(169),
    [anon_sym_GT_EQ] = ACTIONS(169),
    [anon_sym_QMARK_QMARK] = ACTIONS(165),
    [anon_sym_instanceof] = ACTIONS(169),
    [anon_sym_PLUS_PLUS] = ACTIONS(169),
    [anon_sym_DASH_DASH] = ACTIONS(169),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(169),
    [anon_sym_AT] = ACTIONS(93),
    [sym__automatic_semicolon] = ACTIONS(169),
    [sym__ternary_qmark] = ACTIONS(169),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(167)] = {
    [sym_import] = STATE(1326),
    [sym_statement_block] = STATE(801),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(803),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1790),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1790),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_function_expression] = STATE(740),
    [sym_generator_function] = STATE(740),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1790),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [aux_sym_export_statement_repeat1] = STATE(1381),
    [sym_identifier] = ACTIONS(161),
    [anon_sym_export] = ACTIONS(163),
    [anon_sym_LBRACE] = ACTIONS(993),
    [anon_sym_import] = ACTIONS(173),
    [anon_sym_let] = ACTIONS(163),
    [anon_sym_await] = ACTIONS(177),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_yield] = ACTIONS(57),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(194),
    [anon_sym_async] = ACTIONS(196),
    [anon_sym_function] = ACTIONS(198),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(75),
    [anon_sym_DASH] = ACTIONS(75),
    [anon_sym_SLASH] = ACTIONS(77),
    [anon_sym_BANG] = ACTIONS(79),
    [anon_sym_TILDE] = ACTIONS(79),
    [anon_sym_typeof] = ACTIONS(75),
    [anon_sym_void] = ACTIONS(75),
    [anon_sym_delete] = ACTIONS(75),
    [anon_sym_PLUS_PLUS] = ACTIONS(81),
    [anon_sym_DASH_DASH] = ACTIONS(81),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(87),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(163),
    [anon_sym_get] = ACTIONS(163),
    [anon_sym_set] = ACTIONS(163),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(168)] = {
    [sym_import] = STATE(1333),
    [sym_statement_block] = STATE(600),
    [sym_parenthesized_expression] = STATE(507),
    [sym_expression] = STATE(661),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1815),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1815),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(507),
    [sym_subscript_expression] = STATE(507),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1139),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1815),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1763),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(369),
    [anon_sym_export] = ACTIONS(371),
    [anon_sym_LBRACE] = ACTIONS(987),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(371),
    [anon_sym_await] = ACTIONS(377),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(382),
    [anon_sym_LBRACK] = ACTIONS(654),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(398),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(404),
    [anon_sym_PLUS] = ACTIONS(412),
    [anon_sym_DASH] = ACTIONS(412),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(414),
    [anon_sym_TILDE] = ACTIONS(414),
    [anon_sym_typeof] = ACTIONS(412),
    [anon_sym_void] = ACTIONS(412),
    [anon_sym_delete] = ACTIONS(412),
    [anon_sym_PLUS_PLUS] = ACTIONS(660),
    [anon_sym_DASH_DASH] = ACTIONS(660),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(424),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(428),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(371),
    [anon_sym_get] = ACTIONS(371),
    [anon_sym_set] = ACTIONS(371),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(169)] = {
    [sym_import] = STATE(1333),
    [sym_statement_block] = STATE(613),
    [sym_parenthesized_expression] = STATE(507),
    [sym_expression] = STATE(654),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1815),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1815),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(507),
    [sym_subscript_expression] = STATE(507),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1139),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1815),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1763),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(369),
    [anon_sym_export] = ACTIONS(371),
    [anon_sym_LBRACE] = ACTIONS(987),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(371),
    [anon_sym_await] = ACTIONS(377),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(382),
    [anon_sym_LBRACK] = ACTIONS(654),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(398),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(404),
    [anon_sym_PLUS] = ACTIONS(412),
    [anon_sym_DASH] = ACTIONS(412),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(414),
    [anon_sym_TILDE] = ACTIONS(414),
    [anon_sym_typeof] = ACTIONS(412),
    [anon_sym_void] = ACTIONS(412),
    [anon_sym_delete] = ACTIONS(412),
    [anon_sym_PLUS_PLUS] = ACTIONS(660),
    [anon_sym_DASH_DASH] = ACTIONS(660),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(424),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(428),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(371),
    [anon_sym_get] = ACTIONS(371),
    [anon_sym_set] = ACTIONS(371),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(170)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(507),
    [sym_expression] = STATE(839),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1815),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1815),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(507),
    [sym_subscript_expression] = STATE(507),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1139),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1815),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_sequence_expression] = STATE(1761),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1763),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(369),
    [anon_sym_export] = ACTIONS(371),
    [anon_sym_LBRACE] = ACTIONS(373),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(371),
    [anon_sym_await] = ACTIONS(377),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(382),
    [anon_sym_LBRACK] = ACTIONS(654),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(398),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(404),
    [anon_sym_PLUS] = ACTIONS(412),
    [anon_sym_DASH] = ACTIONS(412),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(414),
    [anon_sym_TILDE] = ACTIONS(414),
    [anon_sym_typeof] = ACTIONS(412),
    [anon_sym_void] = ACTIONS(412),
    [anon_sym_delete] = ACTIONS(412),
    [anon_sym_PLUS_PLUS] = ACTIONS(660),
    [anon_sym_DASH_DASH] = ACTIONS(660),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(424),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(428),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(371),
    [anon_sym_get] = ACTIONS(371),
    [anon_sym_set] = ACTIONS(371),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(171)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(507),
    [sym_expression] = STATE(833),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1815),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1815),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(507),
    [sym_subscript_expression] = STATE(507),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1139),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1815),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_sequence_expression] = STATE(1850),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1763),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(369),
    [anon_sym_export] = ACTIONS(371),
    [anon_sym_LBRACE] = ACTIONS(373),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(371),
    [anon_sym_await] = ACTIONS(377),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(382),
    [anon_sym_LBRACK] = ACTIONS(654),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(398),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(404),
    [anon_sym_PLUS] = ACTIONS(412),
    [anon_sym_DASH] = ACTIONS(412),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(414),
    [anon_sym_TILDE] = ACTIONS(414),
    [anon_sym_typeof] = ACTIONS(412),
    [anon_sym_void] = ACTIONS(412),
    [anon_sym_delete] = ACTIONS(412),
    [anon_sym_PLUS_PLUS] = ACTIONS(660),
    [anon_sym_DASH_DASH] = ACTIONS(660),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(424),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(428),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(371),
    [anon_sym_get] = ACTIONS(371),
    [anon_sym_set] = ACTIONS(371),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(172)] = {
    [sym_namespace_export] = STATE(1692),
    [sym_export_clause] = STATE(1358),
    [sym_declaration] = STATE(433),
    [sym_variable_declaration] = STATE(419),
    [sym_lexical_declaration] = STATE(419),
    [sym_using_declaration] = STATE(419),
    [sym_class_declaration] = STATE(419),
    [sym_function_declaration] = STATE(419),
    [sym_generator_function_declaration] = STATE(419),
    [sym_decorator] = STATE(1097),
    [aux_sym_export_statement_repeat1] = STATE(1327),
    [anon_sym_STAR] = ACTIONS(930),
    [anon_sym_default] = ACTIONS(1003),
    [anon_sym_LBRACE] = ACTIONS(934),
    [anon_sym_COMMA] = ACTIONS(169),
    [anon_sym_var] = ACTIONS(936),
    [anon_sym_let] = ACTIONS(938),
    [anon_sym_const] = ACTIONS(938),
    [anon_sym_using] = ACTIONS(940),
    [anon_sym_await] = ACTIONS(942),
    [anon_sym_LPAREN] = ACTIONS(169),
    [anon_sym_SEMI] = ACTIONS(169),
    [anon_sym_in] = ACTIONS(165),
    [anon_sym_COLON] = ACTIONS(488),
    [anon_sym_EQ] = ACTIONS(490),
    [anon_sym_LBRACK] = ACTIONS(169),
    [anon_sym_LT] = ACTIONS(165),
    [anon_sym_GT] = ACTIONS(165),
    [anon_sym_DOT] = ACTIONS(169),
    [anon_sym_class] = ACTIONS(947),
    [anon_sym_async] = ACTIONS(949),
    [anon_sym_function] = ACTIONS(951),
    [anon_sym_EQ_GT] = ACTIONS(200),
    [sym_optional_chain] = ACTIONS(169),
    [anon_sym_PLUS_EQ] = ACTIONS(202),
    [anon_sym_DASH_EQ] = ACTIONS(202),
    [anon_sym_STAR_EQ] = ACTIONS(202),
    [anon_sym_SLASH_EQ] = ACTIONS(202),
    [anon_sym_PERCENT_EQ] = ACTIONS(202),
    [anon_sym_CARET_EQ] = ACTIONS(202),
    [anon_sym_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_LT_LT_EQ] = ACTIONS(202),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(202),
    [anon_sym_QMARK_QMARK_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP] = ACTIONS(165),
    [anon_sym_PIPE_PIPE] = ACTIONS(165),
    [anon_sym_GT_GT] = ACTIONS(165),
    [anon_sym_GT_GT_GT] = ACTIONS(165),
    [anon_sym_LT_LT] = ACTIONS(165),
    [anon_sym_AMP] = ACTIONS(165),
    [anon_sym_CARET] = ACTIONS(165),
    [anon_sym_PIPE] = ACTIONS(165),
    [anon_sym_PLUS] = ACTIONS(165),
    [anon_sym_DASH] = ACTIONS(165),
    [anon_sym_SLASH] = ACTIONS(165),
    [anon_sym_PERCENT] = ACTIONS(165),
    [anon_sym_STAR_STAR] = ACTIONS(165),
    [anon_sym_LT_EQ] = ACTIONS(169),
    [anon_sym_EQ_EQ] = ACTIONS(165),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(169),
    [anon_sym_BANG_EQ] = ACTIONS(165),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(169),
    [anon_sym_GT_EQ] = ACTIONS(169),
    [anon_sym_QMARK_QMARK] = ACTIONS(165),
    [anon_sym_instanceof] = ACTIONS(169),
    [anon_sym_PLUS_PLUS] = ACTIONS(169),
    [anon_sym_DASH_DASH] = ACTIONS(169),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(169),
    [anon_sym_AT] = ACTIONS(93),
    [sym__automatic_semicolon] = ACTIONS(169),
    [sym__ternary_qmark] = ACTIONS(169),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(173)] = {
    [sym_import] = STATE(1326),
    [sym_statement_block] = STATE(720),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(725),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1790),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1790),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_function_expression] = STATE(740),
    [sym_generator_function] = STATE(740),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1790),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [aux_sym_export_statement_repeat1] = STATE(1381),
    [sym_identifier] = ACTIONS(161),
    [anon_sym_export] = ACTIONS(163),
    [anon_sym_LBRACE] = ACTIONS(993),
    [anon_sym_import] = ACTIONS(173),
    [anon_sym_let] = ACTIONS(163),
    [anon_sym_await] = ACTIONS(177),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_yield] = ACTIONS(57),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(194),
    [anon_sym_async] = ACTIONS(196),
    [anon_sym_function] = ACTIONS(198),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(75),
    [anon_sym_DASH] = ACTIONS(75),
    [anon_sym_SLASH] = ACTIONS(77),
    [anon_sym_BANG] = ACTIONS(79),
    [anon_sym_TILDE] = ACTIONS(79),
    [anon_sym_typeof] = ACTIONS(75),
    [anon_sym_void] = ACTIONS(75),
    [anon_sym_delete] = ACTIONS(75),
    [anon_sym_PLUS_PLUS] = ACTIONS(81),
    [anon_sym_DASH_DASH] = ACTIONS(81),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(87),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(163),
    [anon_sym_get] = ACTIONS(163),
    [anon_sym_set] = ACTIONS(163),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(174)] = {
    [sym_import] = STATE(1326),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(818),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1790),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1790),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_function_expression] = STATE(740),
    [sym_generator_function] = STATE(740),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1790),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_sequence_expression] = STATE(1505),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [aux_sym_export_statement_repeat1] = STATE(1381),
    [sym_identifier] = ACTIONS(161),
    [anon_sym_export] = ACTIONS(163),
    [anon_sym_LBRACE] = ACTIONS(167),
    [anon_sym_import] = ACTIONS(173),
    [anon_sym_let] = ACTIONS(163),
    [anon_sym_await] = ACTIONS(177),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_yield] = ACTIONS(57),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(194),
    [anon_sym_async] = ACTIONS(196),
    [anon_sym_function] = ACTIONS(198),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(75),
    [anon_sym_DASH] = ACTIONS(75),
    [anon_sym_SLASH] = ACTIONS(77),
    [anon_sym_BANG] = ACTIONS(79),
    [anon_sym_TILDE] = ACTIONS(79),
    [anon_sym_typeof] = ACTIONS(75),
    [anon_sym_void] = ACTIONS(75),
    [anon_sym_delete] = ACTIONS(75),
    [anon_sym_PLUS_PLUS] = ACTIONS(81),
    [anon_sym_DASH_DASH] = ACTIONS(81),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(87),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(163),
    [anon_sym_get] = ACTIONS(163),
    [anon_sym_set] = ACTIONS(163),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(175)] = {
    [sym_import] = STATE(1333),
    [sym_statement_block] = STATE(626),
    [sym_parenthesized_expression] = STATE(507),
    [sym_expression] = STATE(664),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1815),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1815),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(507),
    [sym_subscript_expression] = STATE(507),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1139),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1815),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1763),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(369),
    [anon_sym_export] = ACTIONS(371),
    [anon_sym_LBRACE] = ACTIONS(987),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(371),
    [anon_sym_await] = ACTIONS(377),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(382),
    [anon_sym_LBRACK] = ACTIONS(654),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(398),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(404),
    [anon_sym_PLUS] = ACTIONS(412),
    [anon_sym_DASH] = ACTIONS(412),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(414),
    [anon_sym_TILDE] = ACTIONS(414),
    [anon_sym_typeof] = ACTIONS(412),
    [anon_sym_void] = ACTIONS(412),
    [anon_sym_delete] = ACTIONS(412),
    [anon_sym_PLUS_PLUS] = ACTIONS(660),
    [anon_sym_DASH_DASH] = ACTIONS(660),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(424),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(428),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(371),
    [anon_sym_get] = ACTIONS(371),
    [anon_sym_set] = ACTIONS(371),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(176)] = {
    [sym_import] = STATE(1326),
    [sym_statement_block] = STATE(737),
    [sym_parenthesized_expression] = STATE(560),
    [sym_expression] = STATE(809),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1802),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1802),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_function_expression] = STATE(740),
    [sym_generator_function] = STATE(740),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(560),
    [sym_subscript_expression] = STATE(560),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1143),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1802),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1803),
    [aux_sym_export_statement_repeat1] = STATE(1381),
    [sym_identifier] = ACTIONS(520),
    [anon_sym_export] = ACTIONS(522),
    [anon_sym_LBRACE] = ACTIONS(993),
    [anon_sym_import] = ACTIONS(173),
    [anon_sym_let] = ACTIONS(522),
    [anon_sym_await] = ACTIONS(524),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_yield] = ACTIONS(526),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(194),
    [anon_sym_async] = ACTIONS(530),
    [anon_sym_function] = ACTIONS(198),
    [anon_sym_new] = ACTIONS(532),
    [anon_sym_PLUS] = ACTIONS(540),
    [anon_sym_DASH] = ACTIONS(540),
    [anon_sym_SLASH] = ACTIONS(670),
    [anon_sym_BANG] = ACTIONS(542),
    [anon_sym_TILDE] = ACTIONS(542),
    [anon_sym_typeof] = ACTIONS(540),
    [anon_sym_void] = ACTIONS(540),
    [anon_sym_delete] = ACTIONS(540),
    [anon_sym_PLUS_PLUS] = ACTIONS(672),
    [anon_sym_DASH_DASH] = ACTIONS(672),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(547),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(549),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(522),
    [anon_sym_get] = ACTIONS(522),
    [anon_sym_set] = ACTIONS(522),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(177)] = {
    [sym_import] = STATE(1326),
    [sym_statement_block] = STATE(794),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(806),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1790),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1790),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_function_expression] = STATE(740),
    [sym_generator_function] = STATE(740),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1790),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [aux_sym_export_statement_repeat1] = STATE(1381),
    [sym_identifier] = ACTIONS(161),
    [anon_sym_export] = ACTIONS(163),
    [anon_sym_LBRACE] = ACTIONS(993),
    [anon_sym_import] = ACTIONS(173),
    [anon_sym_let] = ACTIONS(163),
    [anon_sym_await] = ACTIONS(177),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_yield] = ACTIONS(57),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(194),
    [anon_sym_async] = ACTIONS(196),
    [anon_sym_function] = ACTIONS(198),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(75),
    [anon_sym_DASH] = ACTIONS(75),
    [anon_sym_SLASH] = ACTIONS(77),
    [anon_sym_BANG] = ACTIONS(79),
    [anon_sym_TILDE] = ACTIONS(79),
    [anon_sym_typeof] = ACTIONS(75),
    [anon_sym_void] = ACTIONS(75),
    [anon_sym_delete] = ACTIONS(75),
    [anon_sym_PLUS_PLUS] = ACTIONS(81),
    [anon_sym_DASH_DASH] = ACTIONS(81),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(87),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(163),
    [anon_sym_get] = ACTIONS(163),
    [anon_sym_set] = ACTIONS(163),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(178)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(507),
    [sym_expression] = STATE(869),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1815),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1815),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(507),
    [sym_subscript_expression] = STATE(507),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1139),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1815),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_sequence_expression] = STATE(1811),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1763),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(369),
    [anon_sym_export] = ACTIONS(371),
    [anon_sym_LBRACE] = ACTIONS(373),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(371),
    [anon_sym_await] = ACTIONS(377),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(382),
    [anon_sym_LBRACK] = ACTIONS(654),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(398),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(404),
    [anon_sym_PLUS] = ACTIONS(412),
    [anon_sym_DASH] = ACTIONS(412),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(414),
    [anon_sym_TILDE] = ACTIONS(414),
    [anon_sym_typeof] = ACTIONS(412),
    [anon_sym_void] = ACTIONS(412),
    [anon_sym_delete] = ACTIONS(412),
    [anon_sym_PLUS_PLUS] = ACTIONS(660),
    [anon_sym_DASH_DASH] = ACTIONS(660),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(424),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(428),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(371),
    [anon_sym_get] = ACTIONS(371),
    [anon_sym_set] = ACTIONS(371),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(179)] = {
    [sym_import] = STATE(1333),
    [sym_statement_block] = STATE(629),
    [sym_parenthesized_expression] = STATE(507),
    [sym_expression] = STATE(668),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1815),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1815),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(507),
    [sym_subscript_expression] = STATE(507),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1139),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1815),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1763),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(369),
    [anon_sym_export] = ACTIONS(371),
    [anon_sym_LBRACE] = ACTIONS(987),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(371),
    [anon_sym_await] = ACTIONS(377),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(382),
    [anon_sym_LBRACK] = ACTIONS(654),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(398),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(404),
    [anon_sym_PLUS] = ACTIONS(412),
    [anon_sym_DASH] = ACTIONS(412),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(414),
    [anon_sym_TILDE] = ACTIONS(414),
    [anon_sym_typeof] = ACTIONS(412),
    [anon_sym_void] = ACTIONS(412),
    [anon_sym_delete] = ACTIONS(412),
    [anon_sym_PLUS_PLUS] = ACTIONS(660),
    [anon_sym_DASH_DASH] = ACTIONS(660),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(424),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(428),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(371),
    [anon_sym_get] = ACTIONS(371),
    [anon_sym_set] = ACTIONS(371),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(180)] = {
    [sym_import] = STATE(1326),
    [sym_statement_block] = STATE(720),
    [sym_parenthesized_expression] = STATE(560),
    [sym_expression] = STATE(703),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1802),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1802),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_function_expression] = STATE(740),
    [sym_generator_function] = STATE(740),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(560),
    [sym_subscript_expression] = STATE(560),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1143),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1802),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1803),
    [aux_sym_export_statement_repeat1] = STATE(1381),
    [sym_identifier] = ACTIONS(520),
    [anon_sym_export] = ACTIONS(522),
    [anon_sym_LBRACE] = ACTIONS(993),
    [anon_sym_import] = ACTIONS(173),
    [anon_sym_let] = ACTIONS(522),
    [anon_sym_await] = ACTIONS(524),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_yield] = ACTIONS(526),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(194),
    [anon_sym_async] = ACTIONS(530),
    [anon_sym_function] = ACTIONS(198),
    [anon_sym_new] = ACTIONS(532),
    [anon_sym_PLUS] = ACTIONS(540),
    [anon_sym_DASH] = ACTIONS(540),
    [anon_sym_SLASH] = ACTIONS(670),
    [anon_sym_BANG] = ACTIONS(542),
    [anon_sym_TILDE] = ACTIONS(542),
    [anon_sym_typeof] = ACTIONS(540),
    [anon_sym_void] = ACTIONS(540),
    [anon_sym_delete] = ACTIONS(540),
    [anon_sym_PLUS_PLUS] = ACTIONS(672),
    [anon_sym_DASH_DASH] = ACTIONS(672),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(547),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(549),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(522),
    [anon_sym_get] = ACTIONS(522),
    [anon_sym_set] = ACTIONS(522),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(181)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(507),
    [sym_expression] = STATE(860),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1815),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1815),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(507),
    [sym_subscript_expression] = STATE(507),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1139),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1815),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_sequence_expression] = STATE(1858),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1763),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(369),
    [anon_sym_export] = ACTIONS(371),
    [anon_sym_LBRACE] = ACTIONS(373),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(371),
    [anon_sym_await] = ACTIONS(377),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(382),
    [anon_sym_LBRACK] = ACTIONS(654),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(398),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(404),
    [anon_sym_PLUS] = ACTIONS(412),
    [anon_sym_DASH] = ACTIONS(412),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(414),
    [anon_sym_TILDE] = ACTIONS(414),
    [anon_sym_typeof] = ACTIONS(412),
    [anon_sym_void] = ACTIONS(412),
    [anon_sym_delete] = ACTIONS(412),
    [anon_sym_PLUS_PLUS] = ACTIONS(660),
    [anon_sym_DASH_DASH] = ACTIONS(660),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(424),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(428),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(371),
    [anon_sym_get] = ACTIONS(371),
    [anon_sym_set] = ACTIONS(371),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(182)] = {
    [sym_import] = STATE(1326),
    [sym_statement_block] = STATE(801),
    [sym_parenthesized_expression] = STATE(560),
    [sym_expression] = STATE(718),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1802),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1802),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_function_expression] = STATE(740),
    [sym_generator_function] = STATE(740),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(560),
    [sym_subscript_expression] = STATE(560),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1143),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1802),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1803),
    [aux_sym_export_statement_repeat1] = STATE(1381),
    [sym_identifier] = ACTIONS(520),
    [anon_sym_export] = ACTIONS(522),
    [anon_sym_LBRACE] = ACTIONS(993),
    [anon_sym_import] = ACTIONS(173),
    [anon_sym_let] = ACTIONS(522),
    [anon_sym_await] = ACTIONS(524),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_yield] = ACTIONS(526),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(194),
    [anon_sym_async] = ACTIONS(530),
    [anon_sym_function] = ACTIONS(198),
    [anon_sym_new] = ACTIONS(532),
    [anon_sym_PLUS] = ACTIONS(540),
    [anon_sym_DASH] = ACTIONS(540),
    [anon_sym_SLASH] = ACTIONS(670),
    [anon_sym_BANG] = ACTIONS(542),
    [anon_sym_TILDE] = ACTIONS(542),
    [anon_sym_typeof] = ACTIONS(540),
    [anon_sym_void] = ACTIONS(540),
    [anon_sym_delete] = ACTIONS(540),
    [anon_sym_PLUS_PLUS] = ACTIONS(672),
    [anon_sym_DASH_DASH] = ACTIONS(672),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(547),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(549),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(522),
    [anon_sym_get] = ACTIONS(522),
    [anon_sym_set] = ACTIONS(522),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(183)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(507),
    [sym_expression] = STATE(834),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1815),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1815),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(507),
    [sym_subscript_expression] = STATE(507),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1139),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1815),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_sequence_expression] = STATE(1844),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1763),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(369),
    [anon_sym_export] = ACTIONS(371),
    [anon_sym_LBRACE] = ACTIONS(373),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(371),
    [anon_sym_await] = ACTIONS(377),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(382),
    [anon_sym_LBRACK] = ACTIONS(654),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(398),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(404),
    [anon_sym_PLUS] = ACTIONS(412),
    [anon_sym_DASH] = ACTIONS(412),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(414),
    [anon_sym_TILDE] = ACTIONS(414),
    [anon_sym_typeof] = ACTIONS(412),
    [anon_sym_void] = ACTIONS(412),
    [anon_sym_delete] = ACTIONS(412),
    [anon_sym_PLUS_PLUS] = ACTIONS(660),
    [anon_sym_DASH_DASH] = ACTIONS(660),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(424),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(428),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(371),
    [anon_sym_get] = ACTIONS(371),
    [anon_sym_set] = ACTIONS(371),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(184)] = {
    [sym_import] = STATE(1326),
    [sym_statement_block] = STATE(759),
    [sym_parenthesized_expression] = STATE(560),
    [sym_expression] = STATE(721),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1802),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1802),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_function_expression] = STATE(740),
    [sym_generator_function] = STATE(740),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(560),
    [sym_subscript_expression] = STATE(560),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1143),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1802),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1803),
    [aux_sym_export_statement_repeat1] = STATE(1381),
    [sym_identifier] = ACTIONS(520),
    [anon_sym_export] = ACTIONS(522),
    [anon_sym_LBRACE] = ACTIONS(993),
    [anon_sym_import] = ACTIONS(173),
    [anon_sym_let] = ACTIONS(522),
    [anon_sym_await] = ACTIONS(524),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_yield] = ACTIONS(526),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(194),
    [anon_sym_async] = ACTIONS(530),
    [anon_sym_function] = ACTIONS(198),
    [anon_sym_new] = ACTIONS(532),
    [anon_sym_PLUS] = ACTIONS(540),
    [anon_sym_DASH] = ACTIONS(540),
    [anon_sym_SLASH] = ACTIONS(670),
    [anon_sym_BANG] = ACTIONS(542),
    [anon_sym_TILDE] = ACTIONS(542),
    [anon_sym_typeof] = ACTIONS(540),
    [anon_sym_void] = ACTIONS(540),
    [anon_sym_delete] = ACTIONS(540),
    [anon_sym_PLUS_PLUS] = ACTIONS(672),
    [anon_sym_DASH_DASH] = ACTIONS(672),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(547),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(549),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(522),
    [anon_sym_get] = ACTIONS(522),
    [anon_sym_set] = ACTIONS(522),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(185)] = {
    [sym_import] = STATE(1326),
    [sym_statement_block] = STATE(794),
    [sym_parenthesized_expression] = STATE(560),
    [sym_expression] = STATE(722),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1802),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1802),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_function_expression] = STATE(740),
    [sym_generator_function] = STATE(740),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(560),
    [sym_subscript_expression] = STATE(560),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1143),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1802),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1803),
    [aux_sym_export_statement_repeat1] = STATE(1381),
    [sym_identifier] = ACTIONS(520),
    [anon_sym_export] = ACTIONS(522),
    [anon_sym_LBRACE] = ACTIONS(993),
    [anon_sym_import] = ACTIONS(173),
    [anon_sym_let] = ACTIONS(522),
    [anon_sym_await] = ACTIONS(524),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_yield] = ACTIONS(526),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(194),
    [anon_sym_async] = ACTIONS(530),
    [anon_sym_function] = ACTIONS(198),
    [anon_sym_new] = ACTIONS(532),
    [anon_sym_PLUS] = ACTIONS(540),
    [anon_sym_DASH] = ACTIONS(540),
    [anon_sym_SLASH] = ACTIONS(670),
    [anon_sym_BANG] = ACTIONS(542),
    [anon_sym_TILDE] = ACTIONS(542),
    [anon_sym_typeof] = ACTIONS(540),
    [anon_sym_void] = ACTIONS(540),
    [anon_sym_delete] = ACTIONS(540),
    [anon_sym_PLUS_PLUS] = ACTIONS(672),
    [anon_sym_DASH_DASH] = ACTIONS(672),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(547),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(549),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(522),
    [anon_sym_get] = ACTIONS(522),
    [anon_sym_set] = ACTIONS(522),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(186)] = {
    [sym_import] = STATE(1326),
    [sym_statement_block] = STATE(742),
    [sym_parenthesized_expression] = STATE(560),
    [sym_expression] = STATE(723),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1802),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1802),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_function_expression] = STATE(740),
    [sym_generator_function] = STATE(740),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(560),
    [sym_subscript_expression] = STATE(560),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1143),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1802),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1803),
    [aux_sym_export_statement_repeat1] = STATE(1381),
    [sym_identifier] = ACTIONS(520),
    [anon_sym_export] = ACTIONS(522),
    [anon_sym_LBRACE] = ACTIONS(993),
    [anon_sym_import] = ACTIONS(173),
    [anon_sym_let] = ACTIONS(522),
    [anon_sym_await] = ACTIONS(524),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_yield] = ACTIONS(526),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(194),
    [anon_sym_async] = ACTIONS(530),
    [anon_sym_function] = ACTIONS(198),
    [anon_sym_new] = ACTIONS(532),
    [anon_sym_PLUS] = ACTIONS(540),
    [anon_sym_DASH] = ACTIONS(540),
    [anon_sym_SLASH] = ACTIONS(670),
    [anon_sym_BANG] = ACTIONS(542),
    [anon_sym_TILDE] = ACTIONS(542),
    [anon_sym_typeof] = ACTIONS(540),
    [anon_sym_void] = ACTIONS(540),
    [anon_sym_delete] = ACTIONS(540),
    [anon_sym_PLUS_PLUS] = ACTIONS(672),
    [anon_sym_DASH_DASH] = ACTIONS(672),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(547),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(549),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(522),
    [anon_sym_get] = ACTIONS(522),
    [anon_sym_set] = ACTIONS(522),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(187)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(507),
    [sym_expression] = STATE(862),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1815),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1815),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(507),
    [sym_subscript_expression] = STATE(507),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1139),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1815),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_sequence_expression] = STATE(1841),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1763),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(369),
    [anon_sym_export] = ACTIONS(371),
    [anon_sym_LBRACE] = ACTIONS(373),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(371),
    [anon_sym_await] = ACTIONS(377),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(382),
    [anon_sym_LBRACK] = ACTIONS(654),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(398),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(404),
    [anon_sym_PLUS] = ACTIONS(412),
    [anon_sym_DASH] = ACTIONS(412),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(414),
    [anon_sym_TILDE] = ACTIONS(414),
    [anon_sym_typeof] = ACTIONS(412),
    [anon_sym_void] = ACTIONS(412),
    [anon_sym_delete] = ACTIONS(412),
    [anon_sym_PLUS_PLUS] = ACTIONS(660),
    [anon_sym_DASH_DASH] = ACTIONS(660),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(424),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(428),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(371),
    [anon_sym_get] = ACTIONS(371),
    [anon_sym_set] = ACTIONS(371),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(188)] = {
    [sym_import] = STATE(1326),
    [sym_statement_block] = STATE(742),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(811),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1790),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1790),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_function_expression] = STATE(740),
    [sym_generator_function] = STATE(740),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1790),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [aux_sym_export_statement_repeat1] = STATE(1381),
    [sym_identifier] = ACTIONS(161),
    [anon_sym_export] = ACTIONS(163),
    [anon_sym_LBRACE] = ACTIONS(993),
    [anon_sym_import] = ACTIONS(173),
    [anon_sym_let] = ACTIONS(163),
    [anon_sym_await] = ACTIONS(177),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_yield] = ACTIONS(57),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(194),
    [anon_sym_async] = ACTIONS(196),
    [anon_sym_function] = ACTIONS(198),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(75),
    [anon_sym_DASH] = ACTIONS(75),
    [anon_sym_SLASH] = ACTIONS(77),
    [anon_sym_BANG] = ACTIONS(79),
    [anon_sym_TILDE] = ACTIONS(79),
    [anon_sym_typeof] = ACTIONS(75),
    [anon_sym_void] = ACTIONS(75),
    [anon_sym_delete] = ACTIONS(75),
    [anon_sym_PLUS_PLUS] = ACTIONS(81),
    [anon_sym_DASH_DASH] = ACTIONS(81),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(87),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(163),
    [anon_sym_get] = ACTIONS(163),
    [anon_sym_set] = ACTIONS(163),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(189)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(507),
    [sym_expression] = STATE(849),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1815),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1815),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(507),
    [sym_subscript_expression] = STATE(507),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1139),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1815),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_sequence_expression] = STATE(1867),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1763),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(369),
    [anon_sym_export] = ACTIONS(371),
    [anon_sym_LBRACE] = ACTIONS(373),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(371),
    [anon_sym_await] = ACTIONS(377),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(382),
    [anon_sym_LBRACK] = ACTIONS(654),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(398),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(404),
    [anon_sym_PLUS] = ACTIONS(412),
    [anon_sym_DASH] = ACTIONS(412),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(414),
    [anon_sym_TILDE] = ACTIONS(414),
    [anon_sym_typeof] = ACTIONS(412),
    [anon_sym_void] = ACTIONS(412),
    [anon_sym_delete] = ACTIONS(412),
    [anon_sym_PLUS_PLUS] = ACTIONS(660),
    [anon_sym_DASH_DASH] = ACTIONS(660),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(424),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(428),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(371),
    [anon_sym_get] = ACTIONS(371),
    [anon_sym_set] = ACTIONS(371),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(190)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(507),
    [sym_expression] = STATE(835),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1815),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1815),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(507),
    [sym_subscript_expression] = STATE(507),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1139),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1815),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_sequence_expression] = STATE(1864),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1763),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(369),
    [anon_sym_export] = ACTIONS(371),
    [anon_sym_LBRACE] = ACTIONS(373),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(371),
    [anon_sym_await] = ACTIONS(377),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(382),
    [anon_sym_LBRACK] = ACTIONS(654),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(398),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(404),
    [anon_sym_PLUS] = ACTIONS(412),
    [anon_sym_DASH] = ACTIONS(412),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(414),
    [anon_sym_TILDE] = ACTIONS(414),
    [anon_sym_typeof] = ACTIONS(412),
    [anon_sym_void] = ACTIONS(412),
    [anon_sym_delete] = ACTIONS(412),
    [anon_sym_PLUS_PLUS] = ACTIONS(660),
    [anon_sym_DASH_DASH] = ACTIONS(660),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(424),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(428),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(371),
    [anon_sym_get] = ACTIONS(371),
    [anon_sym_set] = ACTIONS(371),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(191)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(507),
    [sym_expression] = STATE(879),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1815),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1815),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(507),
    [sym_subscript_expression] = STATE(507),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1139),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1815),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_sequence_expression] = STATE(1839),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1763),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(369),
    [anon_sym_export] = ACTIONS(371),
    [anon_sym_LBRACE] = ACTIONS(373),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(371),
    [anon_sym_await] = ACTIONS(377),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(382),
    [anon_sym_LBRACK] = ACTIONS(654),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(398),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(404),
    [anon_sym_PLUS] = ACTIONS(412),
    [anon_sym_DASH] = ACTIONS(412),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(414),
    [anon_sym_TILDE] = ACTIONS(414),
    [anon_sym_typeof] = ACTIONS(412),
    [anon_sym_void] = ACTIONS(412),
    [anon_sym_delete] = ACTIONS(412),
    [anon_sym_PLUS_PLUS] = ACTIONS(660),
    [anon_sym_DASH_DASH] = ACTIONS(660),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(424),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(428),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(371),
    [anon_sym_get] = ACTIONS(371),
    [anon_sym_set] = ACTIONS(371),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(192)] = {
    [sym_import] = STATE(1326),
    [sym_parenthesized_expression] = STATE(528),
    [sym_expression] = STATE(980),
    [sym_primary_expression] = STATE(657),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1762),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1762),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(740),
    [sym_function_expression] = STATE(740),
    [sym_generator_function] = STATE(740),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(681),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(528),
    [sym_subscript_expression] = STATE(528),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1138),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1762),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [aux_sym_export_statement_repeat1] = STATE(1381),
    [sym_identifier] = ACTIONS(1005),
    [anon_sym_export] = ACTIONS(1007),
    [anon_sym_LBRACE] = ACTIONS(167),
    [anon_sym_import] = ACTIONS(173),
    [anon_sym_let] = ACTIONS(1007),
    [anon_sym_await] = ACTIONS(1009),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_yield] = ACTIONS(438),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DOT] = ACTIONS(1011),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(194),
    [anon_sym_async] = ACTIONS(1013),
    [anon_sym_function] = ACTIONS(198),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(452),
    [anon_sym_DASH] = ACTIONS(452),
    [anon_sym_SLASH] = ACTIONS(77),
    [anon_sym_BANG] = ACTIONS(454),
    [anon_sym_TILDE] = ACTIONS(454),
    [anon_sym_typeof] = ACTIONS(452),
    [anon_sym_void] = ACTIONS(452),
    [anon_sym_delete] = ACTIONS(452),
    [anon_sym_PLUS_PLUS] = ACTIONS(684),
    [anon_sym_DASH_DASH] = ACTIONS(684),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(459),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(1015),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(1007),
    [anon_sym_get] = ACTIONS(1007),
    [anon_sym_set] = ACTIONS(1007),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(193)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(507),
    [sym_expression] = STATE(830),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1815),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1815),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(507),
    [sym_subscript_expression] = STATE(507),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1139),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1815),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_sequence_expression] = STATE(1806),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1763),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(369),
    [anon_sym_export] = ACTIONS(371),
    [anon_sym_LBRACE] = ACTIONS(373),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(371),
    [anon_sym_await] = ACTIONS(377),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(382),
    [anon_sym_LBRACK] = ACTIONS(654),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(398),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(404),
    [anon_sym_PLUS] = ACTIONS(412),
    [anon_sym_DASH] = ACTIONS(412),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(414),
    [anon_sym_TILDE] = ACTIONS(414),
    [anon_sym_typeof] = ACTIONS(412),
    [anon_sym_void] = ACTIONS(412),
    [anon_sym_delete] = ACTIONS(412),
    [anon_sym_PLUS_PLUS] = ACTIONS(660),
    [anon_sym_DASH_DASH] = ACTIONS(660),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(424),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(428),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(371),
    [anon_sym_get] = ACTIONS(371),
    [anon_sym_set] = ACTIONS(371),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(194)] = {
    [sym_import] = STATE(1333),
    [sym_statement_block] = STATE(630),
    [sym_parenthesized_expression] = STATE(498),
    [sym_expression] = STATE(920),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1762),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1762),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(498),
    [sym_subscript_expression] = STATE(498),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1138),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1762),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1816),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(430),
    [anon_sym_export] = ACTIONS(432),
    [anon_sym_LBRACE] = ACTIONS(989),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(432),
    [anon_sym_await] = ACTIONS(436),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(438),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(445),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(447),
    [anon_sym_PLUS] = ACTIONS(452),
    [anon_sym_DASH] = ACTIONS(452),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(454),
    [anon_sym_TILDE] = ACTIONS(454),
    [anon_sym_typeof] = ACTIONS(452),
    [anon_sym_void] = ACTIONS(452),
    [anon_sym_delete] = ACTIONS(452),
    [anon_sym_PLUS_PLUS] = ACTIONS(684),
    [anon_sym_DASH_DASH] = ACTIONS(684),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(459),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(461),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(432),
    [anon_sym_get] = ACTIONS(432),
    [anon_sym_set] = ACTIONS(432),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(195)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(507),
    [sym_expression] = STATE(857),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1815),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1815),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(507),
    [sym_subscript_expression] = STATE(507),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1139),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1815),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_sequence_expression] = STATE(1826),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1763),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(369),
    [anon_sym_export] = ACTIONS(371),
    [anon_sym_LBRACE] = ACTIONS(373),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(371),
    [anon_sym_await] = ACTIONS(377),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(382),
    [anon_sym_LBRACK] = ACTIONS(654),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(398),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(404),
    [anon_sym_PLUS] = ACTIONS(412),
    [anon_sym_DASH] = ACTIONS(412),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(414),
    [anon_sym_TILDE] = ACTIONS(414),
    [anon_sym_typeof] = ACTIONS(412),
    [anon_sym_void] = ACTIONS(412),
    [anon_sym_delete] = ACTIONS(412),
    [anon_sym_PLUS_PLUS] = ACTIONS(660),
    [anon_sym_DASH_DASH] = ACTIONS(660),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(424),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(428),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(371),
    [anon_sym_get] = ACTIONS(371),
    [anon_sym_set] = ACTIONS(371),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(196)] = {
    [sym_import] = STATE(1333),
    [sym_statement_block] = STATE(600),
    [sym_parenthesized_expression] = STATE(498),
    [sym_expression] = STATE(967),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1762),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1762),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(498),
    [sym_subscript_expression] = STATE(498),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1138),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1762),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1816),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(430),
    [anon_sym_export] = ACTIONS(432),
    [anon_sym_LBRACE] = ACTIONS(989),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(432),
    [anon_sym_await] = ACTIONS(436),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(438),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(445),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(447),
    [anon_sym_PLUS] = ACTIONS(452),
    [anon_sym_DASH] = ACTIONS(452),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(454),
    [anon_sym_TILDE] = ACTIONS(454),
    [anon_sym_typeof] = ACTIONS(452),
    [anon_sym_void] = ACTIONS(452),
    [anon_sym_delete] = ACTIONS(452),
    [anon_sym_PLUS_PLUS] = ACTIONS(684),
    [anon_sym_DASH_DASH] = ACTIONS(684),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(459),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(461),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(432),
    [anon_sym_get] = ACTIONS(432),
    [anon_sym_set] = ACTIONS(432),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(197)] = {
    [sym_import] = STATE(1333),
    [sym_statement_block] = STATE(613),
    [sym_parenthesized_expression] = STATE(498),
    [sym_expression] = STATE(936),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1762),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1762),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(498),
    [sym_subscript_expression] = STATE(498),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1138),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1762),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1816),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(430),
    [anon_sym_export] = ACTIONS(432),
    [anon_sym_LBRACE] = ACTIONS(989),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(432),
    [anon_sym_await] = ACTIONS(436),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(438),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(445),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(447),
    [anon_sym_PLUS] = ACTIONS(452),
    [anon_sym_DASH] = ACTIONS(452),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(454),
    [anon_sym_TILDE] = ACTIONS(454),
    [anon_sym_typeof] = ACTIONS(452),
    [anon_sym_void] = ACTIONS(452),
    [anon_sym_delete] = ACTIONS(452),
    [anon_sym_PLUS_PLUS] = ACTIONS(684),
    [anon_sym_DASH_DASH] = ACTIONS(684),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(459),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(461),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(432),
    [anon_sym_get] = ACTIONS(432),
    [anon_sym_set] = ACTIONS(432),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(198)] = {
    [sym_import] = STATE(1333),
    [sym_statement_block] = STATE(626),
    [sym_parenthesized_expression] = STATE(498),
    [sym_expression] = STATE(937),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1762),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1762),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(498),
    [sym_subscript_expression] = STATE(498),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1138),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1762),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1816),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(430),
    [anon_sym_export] = ACTIONS(432),
    [anon_sym_LBRACE] = ACTIONS(989),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(432),
    [anon_sym_await] = ACTIONS(436),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(438),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(445),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(447),
    [anon_sym_PLUS] = ACTIONS(452),
    [anon_sym_DASH] = ACTIONS(452),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(454),
    [anon_sym_TILDE] = ACTIONS(454),
    [anon_sym_typeof] = ACTIONS(452),
    [anon_sym_void] = ACTIONS(452),
    [anon_sym_delete] = ACTIONS(452),
    [anon_sym_PLUS_PLUS] = ACTIONS(684),
    [anon_sym_DASH_DASH] = ACTIONS(684),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(459),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(461),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(432),
    [anon_sym_get] = ACTIONS(432),
    [anon_sym_set] = ACTIONS(432),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(199)] = {
    [sym_import] = STATE(1333),
    [sym_statement_block] = STATE(628),
    [sym_parenthesized_expression] = STATE(498),
    [sym_expression] = STATE(938),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1762),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1762),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(498),
    [sym_subscript_expression] = STATE(498),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1138),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1762),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1816),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(430),
    [anon_sym_export] = ACTIONS(432),
    [anon_sym_LBRACE] = ACTIONS(989),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(432),
    [anon_sym_await] = ACTIONS(436),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(438),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(445),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(447),
    [anon_sym_PLUS] = ACTIONS(452),
    [anon_sym_DASH] = ACTIONS(452),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(454),
    [anon_sym_TILDE] = ACTIONS(454),
    [anon_sym_typeof] = ACTIONS(452),
    [anon_sym_void] = ACTIONS(452),
    [anon_sym_delete] = ACTIONS(452),
    [anon_sym_PLUS_PLUS] = ACTIONS(684),
    [anon_sym_DASH_DASH] = ACTIONS(684),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(459),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(461),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(432),
    [anon_sym_get] = ACTIONS(432),
    [anon_sym_set] = ACTIONS(432),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(200)] = {
    [sym_import] = STATE(1333),
    [sym_statement_block] = STATE(629),
    [sym_parenthesized_expression] = STATE(498),
    [sym_expression] = STATE(939),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1762),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1762),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(498),
    [sym_subscript_expression] = STATE(498),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1138),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1762),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1816),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(430),
    [anon_sym_export] = ACTIONS(432),
    [anon_sym_LBRACE] = ACTIONS(989),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(432),
    [anon_sym_await] = ACTIONS(436),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(438),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(445),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(447),
    [anon_sym_PLUS] = ACTIONS(452),
    [anon_sym_DASH] = ACTIONS(452),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(454),
    [anon_sym_TILDE] = ACTIONS(454),
    [anon_sym_typeof] = ACTIONS(452),
    [anon_sym_void] = ACTIONS(452),
    [anon_sym_delete] = ACTIONS(452),
    [anon_sym_PLUS_PLUS] = ACTIONS(684),
    [anon_sym_DASH_DASH] = ACTIONS(684),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(459),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(461),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(432),
    [anon_sym_get] = ACTIONS(432),
    [anon_sym_set] = ACTIONS(432),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(201)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(507),
    [sym_expression] = STATE(843),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1815),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1815),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(507),
    [sym_subscript_expression] = STATE(507),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1139),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1815),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_sequence_expression] = STATE(1863),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1763),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(369),
    [anon_sym_export] = ACTIONS(371),
    [anon_sym_LBRACE] = ACTIONS(373),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(371),
    [anon_sym_await] = ACTIONS(377),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(382),
    [anon_sym_LBRACK] = ACTIONS(654),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(398),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(404),
    [anon_sym_PLUS] = ACTIONS(412),
    [anon_sym_DASH] = ACTIONS(412),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(414),
    [anon_sym_TILDE] = ACTIONS(414),
    [anon_sym_typeof] = ACTIONS(412),
    [anon_sym_void] = ACTIONS(412),
    [anon_sym_delete] = ACTIONS(412),
    [anon_sym_PLUS_PLUS] = ACTIONS(660),
    [anon_sym_DASH_DASH] = ACTIONS(660),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(424),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(428),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(371),
    [anon_sym_get] = ACTIONS(371),
    [anon_sym_set] = ACTIONS(371),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(202)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(507),
    [sym_expression] = STATE(844),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1815),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1815),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(507),
    [sym_subscript_expression] = STATE(507),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1139),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1815),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_sequence_expression] = STATE(1774),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1763),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(369),
    [anon_sym_export] = ACTIONS(371),
    [anon_sym_LBRACE] = ACTIONS(373),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(371),
    [anon_sym_await] = ACTIONS(377),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(382),
    [anon_sym_LBRACK] = ACTIONS(654),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(398),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(404),
    [anon_sym_PLUS] = ACTIONS(412),
    [anon_sym_DASH] = ACTIONS(412),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(414),
    [anon_sym_TILDE] = ACTIONS(414),
    [anon_sym_typeof] = ACTIONS(412),
    [anon_sym_void] = ACTIONS(412),
    [anon_sym_delete] = ACTIONS(412),
    [anon_sym_PLUS_PLUS] = ACTIONS(660),
    [anon_sym_DASH_DASH] = ACTIONS(660),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(424),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(428),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(371),
    [anon_sym_get] = ACTIONS(371),
    [anon_sym_set] = ACTIONS(371),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(203)] = {
    [sym_import] = STATE(1333),
    [sym_statement_block] = STATE(630),
    [sym_parenthesized_expression] = STATE(579),
    [sym_expression] = STATE(943),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1857),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1857),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(579),
    [sym_subscript_expression] = STATE(579),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1141),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1857),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1779),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(582),
    [anon_sym_export] = ACTIONS(584),
    [anon_sym_LBRACE] = ACTIONS(989),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(584),
    [anon_sym_await] = ACTIONS(586),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(588),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(592),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(594),
    [anon_sym_PLUS] = ACTIONS(602),
    [anon_sym_DASH] = ACTIONS(602),
    [anon_sym_SLASH] = ACTIONS(678),
    [anon_sym_BANG] = ACTIONS(604),
    [anon_sym_TILDE] = ACTIONS(604),
    [anon_sym_typeof] = ACTIONS(602),
    [anon_sym_void] = ACTIONS(602),
    [anon_sym_delete] = ACTIONS(602),
    [anon_sym_PLUS_PLUS] = ACTIONS(680),
    [anon_sym_DASH_DASH] = ACTIONS(680),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(609),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(611),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(584),
    [anon_sym_get] = ACTIONS(584),
    [anon_sym_set] = ACTIONS(584),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(204)] = {
    [sym_import] = STATE(1333),
    [sym_statement_block] = STATE(600),
    [sym_parenthesized_expression] = STATE(579),
    [sym_expression] = STATE(946),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1857),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1857),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(579),
    [sym_subscript_expression] = STATE(579),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1141),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1857),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1779),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(582),
    [anon_sym_export] = ACTIONS(584),
    [anon_sym_LBRACE] = ACTIONS(989),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(584),
    [anon_sym_await] = ACTIONS(586),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(588),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(592),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(594),
    [anon_sym_PLUS] = ACTIONS(602),
    [anon_sym_DASH] = ACTIONS(602),
    [anon_sym_SLASH] = ACTIONS(678),
    [anon_sym_BANG] = ACTIONS(604),
    [anon_sym_TILDE] = ACTIONS(604),
    [anon_sym_typeof] = ACTIONS(602),
    [anon_sym_void] = ACTIONS(602),
    [anon_sym_delete] = ACTIONS(602),
    [anon_sym_PLUS_PLUS] = ACTIONS(680),
    [anon_sym_DASH_DASH] = ACTIONS(680),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(609),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(611),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(584),
    [anon_sym_get] = ACTIONS(584),
    [anon_sym_set] = ACTIONS(584),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(205)] = {
    [sym_import] = STATE(1333),
    [sym_statement_block] = STATE(613),
    [sym_parenthesized_expression] = STATE(579),
    [sym_expression] = STATE(960),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1857),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1857),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(579),
    [sym_subscript_expression] = STATE(579),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1141),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1857),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1779),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(582),
    [anon_sym_export] = ACTIONS(584),
    [anon_sym_LBRACE] = ACTIONS(989),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(584),
    [anon_sym_await] = ACTIONS(586),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(588),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(592),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(594),
    [anon_sym_PLUS] = ACTIONS(602),
    [anon_sym_DASH] = ACTIONS(602),
    [anon_sym_SLASH] = ACTIONS(678),
    [anon_sym_BANG] = ACTIONS(604),
    [anon_sym_TILDE] = ACTIONS(604),
    [anon_sym_typeof] = ACTIONS(602),
    [anon_sym_void] = ACTIONS(602),
    [anon_sym_delete] = ACTIONS(602),
    [anon_sym_PLUS_PLUS] = ACTIONS(680),
    [anon_sym_DASH_DASH] = ACTIONS(680),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(609),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(611),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(584),
    [anon_sym_get] = ACTIONS(584),
    [anon_sym_set] = ACTIONS(584),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(206)] = {
    [sym_import] = STATE(1333),
    [sym_statement_block] = STATE(626),
    [sym_parenthesized_expression] = STATE(579),
    [sym_expression] = STATE(961),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1857),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1857),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(579),
    [sym_subscript_expression] = STATE(579),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1141),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1857),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1779),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(582),
    [anon_sym_export] = ACTIONS(584),
    [anon_sym_LBRACE] = ACTIONS(989),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(584),
    [anon_sym_await] = ACTIONS(586),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(588),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(592),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(594),
    [anon_sym_PLUS] = ACTIONS(602),
    [anon_sym_DASH] = ACTIONS(602),
    [anon_sym_SLASH] = ACTIONS(678),
    [anon_sym_BANG] = ACTIONS(604),
    [anon_sym_TILDE] = ACTIONS(604),
    [anon_sym_typeof] = ACTIONS(602),
    [anon_sym_void] = ACTIONS(602),
    [anon_sym_delete] = ACTIONS(602),
    [anon_sym_PLUS_PLUS] = ACTIONS(680),
    [anon_sym_DASH_DASH] = ACTIONS(680),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(609),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(611),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(584),
    [anon_sym_get] = ACTIONS(584),
    [anon_sym_set] = ACTIONS(584),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(207)] = {
    [sym_import] = STATE(1333),
    [sym_statement_block] = STATE(628),
    [sym_parenthesized_expression] = STATE(579),
    [sym_expression] = STATE(962),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1857),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1857),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(579),
    [sym_subscript_expression] = STATE(579),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1141),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1857),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1779),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(582),
    [anon_sym_export] = ACTIONS(584),
    [anon_sym_LBRACE] = ACTIONS(989),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(584),
    [anon_sym_await] = ACTIONS(586),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(588),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(592),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(594),
    [anon_sym_PLUS] = ACTIONS(602),
    [anon_sym_DASH] = ACTIONS(602),
    [anon_sym_SLASH] = ACTIONS(678),
    [anon_sym_BANG] = ACTIONS(604),
    [anon_sym_TILDE] = ACTIONS(604),
    [anon_sym_typeof] = ACTIONS(602),
    [anon_sym_void] = ACTIONS(602),
    [anon_sym_delete] = ACTIONS(602),
    [anon_sym_PLUS_PLUS] = ACTIONS(680),
    [anon_sym_DASH_DASH] = ACTIONS(680),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(609),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(611),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(584),
    [anon_sym_get] = ACTIONS(584),
    [anon_sym_set] = ACTIONS(584),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(208)] = {
    [sym_import] = STATE(1333),
    [sym_statement_block] = STATE(629),
    [sym_parenthesized_expression] = STATE(579),
    [sym_expression] = STATE(963),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1857),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1857),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(579),
    [sym_subscript_expression] = STATE(579),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1141),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1857),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1779),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(582),
    [anon_sym_export] = ACTIONS(584),
    [anon_sym_LBRACE] = ACTIONS(989),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(584),
    [anon_sym_await] = ACTIONS(586),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(588),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(592),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(594),
    [anon_sym_PLUS] = ACTIONS(602),
    [anon_sym_DASH] = ACTIONS(602),
    [anon_sym_SLASH] = ACTIONS(678),
    [anon_sym_BANG] = ACTIONS(604),
    [anon_sym_TILDE] = ACTIONS(604),
    [anon_sym_typeof] = ACTIONS(602),
    [anon_sym_void] = ACTIONS(602),
    [anon_sym_delete] = ACTIONS(602),
    [anon_sym_PLUS_PLUS] = ACTIONS(680),
    [anon_sym_DASH_DASH] = ACTIONS(680),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(609),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(611),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(584),
    [anon_sym_get] = ACTIONS(584),
    [anon_sym_set] = ACTIONS(584),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(209)] = {
    [sym_import] = STATE(1333),
    [sym_statement_block] = STATE(630),
    [sym_parenthesized_expression] = STATE(578),
    [sym_expression] = STATE(891),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1776),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1776),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(578),
    [sym_subscript_expression] = STATE(578),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1142),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1776),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1778),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(613),
    [anon_sym_export] = ACTIONS(615),
    [anon_sym_LBRACE] = ACTIONS(989),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(615),
    [anon_sym_await] = ACTIONS(620),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(622),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(626),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(630),
    [anon_sym_PLUS] = ACTIONS(635),
    [anon_sym_DASH] = ACTIONS(635),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(637),
    [anon_sym_TILDE] = ACTIONS(637),
    [anon_sym_typeof] = ACTIONS(635),
    [anon_sym_void] = ACTIONS(635),
    [anon_sym_delete] = ACTIONS(635),
    [anon_sym_PLUS_PLUS] = ACTIONS(688),
    [anon_sym_DASH_DASH] = ACTIONS(688),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(642),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(644),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(615),
    [anon_sym_get] = ACTIONS(615),
    [anon_sym_set] = ACTIONS(615),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(210)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(498),
    [sym_expression] = STATE(979),
    [sym_primary_expression] = STATE(581),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1762),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1762),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(583),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(498),
    [sym_subscript_expression] = STATE(498),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1138),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1762),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1763),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(1017),
    [anon_sym_export] = ACTIONS(1019),
    [anon_sym_LBRACE] = ACTIONS(373),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(1019),
    [anon_sym_await] = ACTIONS(1021),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(438),
    [anon_sym_LBRACK] = ACTIONS(654),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DOT] = ACTIONS(991),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(1023),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(404),
    [anon_sym_PLUS] = ACTIONS(452),
    [anon_sym_DASH] = ACTIONS(452),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(454),
    [anon_sym_TILDE] = ACTIONS(454),
    [anon_sym_typeof] = ACTIONS(452),
    [anon_sym_void] = ACTIONS(452),
    [anon_sym_delete] = ACTIONS(452),
    [anon_sym_PLUS_PLUS] = ACTIONS(684),
    [anon_sym_DASH_DASH] = ACTIONS(684),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(459),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(461),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(1019),
    [anon_sym_get] = ACTIONS(1019),
    [anon_sym_set] = ACTIONS(1019),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(211)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(507),
    [sym_expression] = STATE(846),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1815),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1815),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(507),
    [sym_subscript_expression] = STATE(507),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1139),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1815),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_sequence_expression] = STATE(1753),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1763),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(369),
    [anon_sym_export] = ACTIONS(371),
    [anon_sym_LBRACE] = ACTIONS(373),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(371),
    [anon_sym_await] = ACTIONS(377),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(382),
    [anon_sym_LBRACK] = ACTIONS(654),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(398),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(404),
    [anon_sym_PLUS] = ACTIONS(412),
    [anon_sym_DASH] = ACTIONS(412),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(414),
    [anon_sym_TILDE] = ACTIONS(414),
    [anon_sym_typeof] = ACTIONS(412),
    [anon_sym_void] = ACTIONS(412),
    [anon_sym_delete] = ACTIONS(412),
    [anon_sym_PLUS_PLUS] = ACTIONS(660),
    [anon_sym_DASH_DASH] = ACTIONS(660),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(424),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(428),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(371),
    [anon_sym_get] = ACTIONS(371),
    [anon_sym_set] = ACTIONS(371),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(212)] = {
    [sym_import] = STATE(1333),
    [sym_statement_block] = STATE(600),
    [sym_parenthesized_expression] = STATE(578),
    [sym_expression] = STATE(894),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1776),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1776),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(578),
    [sym_subscript_expression] = STATE(578),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1142),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1776),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1778),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(613),
    [anon_sym_export] = ACTIONS(615),
    [anon_sym_LBRACE] = ACTIONS(989),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(615),
    [anon_sym_await] = ACTIONS(620),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(622),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(626),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(630),
    [anon_sym_PLUS] = ACTIONS(635),
    [anon_sym_DASH] = ACTIONS(635),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(637),
    [anon_sym_TILDE] = ACTIONS(637),
    [anon_sym_typeof] = ACTIONS(635),
    [anon_sym_void] = ACTIONS(635),
    [anon_sym_delete] = ACTIONS(635),
    [anon_sym_PLUS_PLUS] = ACTIONS(688),
    [anon_sym_DASH_DASH] = ACTIONS(688),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(642),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(644),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(615),
    [anon_sym_get] = ACTIONS(615),
    [anon_sym_set] = ACTIONS(615),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(213)] = {
    [sym_import] = STATE(1326),
    [sym_parenthesized_expression] = STATE(528),
    [sym_expression] = STATE(980),
    [sym_primary_expression] = STATE(657),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1762),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1762),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(740),
    [sym_function_expression] = STATE(740),
    [sym_generator_function] = STATE(740),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(681),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(528),
    [sym_subscript_expression] = STATE(528),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1138),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1762),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1803),
    [aux_sym_export_statement_repeat1] = STATE(1381),
    [sym_identifier] = ACTIONS(1025),
    [anon_sym_export] = ACTIONS(1027),
    [anon_sym_LBRACE] = ACTIONS(167),
    [anon_sym_import] = ACTIONS(173),
    [anon_sym_let] = ACTIONS(1027),
    [anon_sym_await] = ACTIONS(1029),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_yield] = ACTIONS(438),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DOT] = ACTIONS(1011),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(194),
    [anon_sym_async] = ACTIONS(1031),
    [anon_sym_function] = ACTIONS(198),
    [anon_sym_new] = ACTIONS(532),
    [anon_sym_PLUS] = ACTIONS(452),
    [anon_sym_DASH] = ACTIONS(452),
    [anon_sym_SLASH] = ACTIONS(670),
    [anon_sym_BANG] = ACTIONS(454),
    [anon_sym_TILDE] = ACTIONS(454),
    [anon_sym_typeof] = ACTIONS(452),
    [anon_sym_void] = ACTIONS(452),
    [anon_sym_delete] = ACTIONS(452),
    [anon_sym_PLUS_PLUS] = ACTIONS(684),
    [anon_sym_DASH_DASH] = ACTIONS(684),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(459),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(1015),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(1027),
    [anon_sym_get] = ACTIONS(1027),
    [anon_sym_set] = ACTIONS(1027),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(214)] = {
    [sym_import] = STATE(1333),
    [sym_statement_block] = STATE(613),
    [sym_parenthesized_expression] = STATE(578),
    [sym_expression] = STATE(908),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1776),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1776),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(578),
    [sym_subscript_expression] = STATE(578),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1142),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1776),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1778),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(613),
    [anon_sym_export] = ACTIONS(615),
    [anon_sym_LBRACE] = ACTIONS(989),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(615),
    [anon_sym_await] = ACTIONS(620),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(622),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(626),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(630),
    [anon_sym_PLUS] = ACTIONS(635),
    [anon_sym_DASH] = ACTIONS(635),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(637),
    [anon_sym_TILDE] = ACTIONS(637),
    [anon_sym_typeof] = ACTIONS(635),
    [anon_sym_void] = ACTIONS(635),
    [anon_sym_delete] = ACTIONS(635),
    [anon_sym_PLUS_PLUS] = ACTIONS(688),
    [anon_sym_DASH_DASH] = ACTIONS(688),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(642),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(644),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(615),
    [anon_sym_get] = ACTIONS(615),
    [anon_sym_set] = ACTIONS(615),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(215)] = {
    [sym_import] = STATE(1333),
    [sym_statement_block] = STATE(626),
    [sym_parenthesized_expression] = STATE(578),
    [sym_expression] = STATE(909),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1776),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1776),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(578),
    [sym_subscript_expression] = STATE(578),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1142),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1776),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1778),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(613),
    [anon_sym_export] = ACTIONS(615),
    [anon_sym_LBRACE] = ACTIONS(989),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(615),
    [anon_sym_await] = ACTIONS(620),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(622),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(626),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(630),
    [anon_sym_PLUS] = ACTIONS(635),
    [anon_sym_DASH] = ACTIONS(635),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(637),
    [anon_sym_TILDE] = ACTIONS(637),
    [anon_sym_typeof] = ACTIONS(635),
    [anon_sym_void] = ACTIONS(635),
    [anon_sym_delete] = ACTIONS(635),
    [anon_sym_PLUS_PLUS] = ACTIONS(688),
    [anon_sym_DASH_DASH] = ACTIONS(688),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(642),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(644),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(615),
    [anon_sym_get] = ACTIONS(615),
    [anon_sym_set] = ACTIONS(615),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(216)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(498),
    [sym_expression] = STATE(979),
    [sym_primary_expression] = STATE(581),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1762),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1762),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(583),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(498),
    [sym_subscript_expression] = STATE(498),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1138),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1762),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1779),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(1033),
    [anon_sym_export] = ACTIONS(1035),
    [anon_sym_LBRACE] = ACTIONS(434),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(1035),
    [anon_sym_await] = ACTIONS(1037),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(438),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DOT] = ACTIONS(991),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(1039),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(594),
    [anon_sym_PLUS] = ACTIONS(452),
    [anon_sym_DASH] = ACTIONS(452),
    [anon_sym_SLASH] = ACTIONS(678),
    [anon_sym_BANG] = ACTIONS(454),
    [anon_sym_TILDE] = ACTIONS(454),
    [anon_sym_typeof] = ACTIONS(452),
    [anon_sym_void] = ACTIONS(452),
    [anon_sym_delete] = ACTIONS(452),
    [anon_sym_PLUS_PLUS] = ACTIONS(684),
    [anon_sym_DASH_DASH] = ACTIONS(684),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(459),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(461),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(1035),
    [anon_sym_get] = ACTIONS(1035),
    [anon_sym_set] = ACTIONS(1035),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(217)] = {
    [sym_import] = STATE(1326),
    [sym_statement_block] = STATE(759),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(792),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1790),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1790),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_function_expression] = STATE(740),
    [sym_generator_function] = STATE(740),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1790),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [aux_sym_export_statement_repeat1] = STATE(1381),
    [sym_identifier] = ACTIONS(161),
    [anon_sym_export] = ACTIONS(163),
    [anon_sym_LBRACE] = ACTIONS(993),
    [anon_sym_import] = ACTIONS(173),
    [anon_sym_let] = ACTIONS(163),
    [anon_sym_await] = ACTIONS(177),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_yield] = ACTIONS(57),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(194),
    [anon_sym_async] = ACTIONS(196),
    [anon_sym_function] = ACTIONS(198),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(75),
    [anon_sym_DASH] = ACTIONS(75),
    [anon_sym_SLASH] = ACTIONS(77),
    [anon_sym_BANG] = ACTIONS(79),
    [anon_sym_TILDE] = ACTIONS(79),
    [anon_sym_typeof] = ACTIONS(75),
    [anon_sym_void] = ACTIONS(75),
    [anon_sym_delete] = ACTIONS(75),
    [anon_sym_PLUS_PLUS] = ACTIONS(81),
    [anon_sym_DASH_DASH] = ACTIONS(81),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(87),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(163),
    [anon_sym_get] = ACTIONS(163),
    [anon_sym_set] = ACTIONS(163),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(218)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(507),
    [sym_expression] = STATE(829),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1815),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1815),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(507),
    [sym_subscript_expression] = STATE(507),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1139),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1815),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_sequence_expression] = STATE(1792),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1763),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(369),
    [anon_sym_export] = ACTIONS(371),
    [anon_sym_LBRACE] = ACTIONS(373),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(371),
    [anon_sym_await] = ACTIONS(377),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(382),
    [anon_sym_LBRACK] = ACTIONS(654),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(398),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(404),
    [anon_sym_PLUS] = ACTIONS(412),
    [anon_sym_DASH] = ACTIONS(412),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(414),
    [anon_sym_TILDE] = ACTIONS(414),
    [anon_sym_typeof] = ACTIONS(412),
    [anon_sym_void] = ACTIONS(412),
    [anon_sym_delete] = ACTIONS(412),
    [anon_sym_PLUS_PLUS] = ACTIONS(660),
    [anon_sym_DASH_DASH] = ACTIONS(660),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(424),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(428),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(371),
    [anon_sym_get] = ACTIONS(371),
    [anon_sym_set] = ACTIONS(371),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(219)] = {
    [sym_import] = STATE(1326),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(726),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1790),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1790),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_function_expression] = STATE(740),
    [sym_generator_function] = STATE(740),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1790),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [aux_sym_export_statement_repeat1] = STATE(1381),
    [sym_identifier] = ACTIONS(161),
    [anon_sym_export] = ACTIONS(163),
    [anon_sym_LBRACE] = ACTIONS(167),
    [anon_sym_import] = ACTIONS(173),
    [anon_sym_let] = ACTIONS(163),
    [anon_sym_await] = ACTIONS(177),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_yield] = ACTIONS(57),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(194),
    [anon_sym_async] = ACTIONS(196),
    [anon_sym_function] = ACTIONS(198),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(75),
    [anon_sym_DASH] = ACTIONS(75),
    [anon_sym_SLASH] = ACTIONS(77),
    [anon_sym_BANG] = ACTIONS(79),
    [anon_sym_TILDE] = ACTIONS(79),
    [anon_sym_typeof] = ACTIONS(75),
    [anon_sym_void] = ACTIONS(75),
    [anon_sym_delete] = ACTIONS(75),
    [anon_sym_PLUS_PLUS] = ACTIONS(81),
    [anon_sym_DASH_DASH] = ACTIONS(81),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(87),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(163),
    [anon_sym_get] = ACTIONS(163),
    [anon_sym_set] = ACTIONS(163),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(220)] = {
    [sym_import] = STATE(1326),
    [sym_parenthesized_expression] = STATE(560),
    [sym_expression] = STATE(695),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1802),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1802),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_function_expression] = STATE(740),
    [sym_generator_function] = STATE(740),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(560),
    [sym_subscript_expression] = STATE(560),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1143),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1802),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1803),
    [aux_sym_export_statement_repeat1] = STATE(1381),
    [sym_identifier] = ACTIONS(520),
    [anon_sym_export] = ACTIONS(522),
    [anon_sym_LBRACE] = ACTIONS(167),
    [anon_sym_import] = ACTIONS(173),
    [anon_sym_let] = ACTIONS(522),
    [anon_sym_await] = ACTIONS(524),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_yield] = ACTIONS(526),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(194),
    [anon_sym_async] = ACTIONS(530),
    [anon_sym_function] = ACTIONS(198),
    [anon_sym_new] = ACTIONS(532),
    [anon_sym_PLUS] = ACTIONS(540),
    [anon_sym_DASH] = ACTIONS(540),
    [anon_sym_SLASH] = ACTIONS(670),
    [anon_sym_BANG] = ACTIONS(542),
    [anon_sym_TILDE] = ACTIONS(542),
    [anon_sym_typeof] = ACTIONS(540),
    [anon_sym_void] = ACTIONS(540),
    [anon_sym_delete] = ACTIONS(540),
    [anon_sym_PLUS_PLUS] = ACTIONS(672),
    [anon_sym_DASH_DASH] = ACTIONS(672),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(547),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(549),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(522),
    [anon_sym_get] = ACTIONS(522),
    [anon_sym_set] = ACTIONS(522),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(221)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(507),
    [sym_expression] = STATE(915),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1815),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1815),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(507),
    [sym_subscript_expression] = STATE(507),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1139),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1815),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1763),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(369),
    [anon_sym_export] = ACTIONS(371),
    [anon_sym_LBRACE] = ACTIONS(373),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(371),
    [anon_sym_await] = ACTIONS(377),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(382),
    [anon_sym_LBRACK] = ACTIONS(654),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(398),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(404),
    [anon_sym_PLUS] = ACTIONS(412),
    [anon_sym_DASH] = ACTIONS(412),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(414),
    [anon_sym_TILDE] = ACTIONS(414),
    [anon_sym_typeof] = ACTIONS(412),
    [anon_sym_void] = ACTIONS(412),
    [anon_sym_delete] = ACTIONS(412),
    [anon_sym_PLUS_PLUS] = ACTIONS(660),
    [anon_sym_DASH_DASH] = ACTIONS(660),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(424),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(428),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(371),
    [anon_sym_get] = ACTIONS(371),
    [anon_sym_set] = ACTIONS(371),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(222)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(507),
    [sym_expression] = STATE(916),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1815),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1815),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(507),
    [sym_subscript_expression] = STATE(507),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1139),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1815),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1763),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(369),
    [anon_sym_export] = ACTIONS(371),
    [anon_sym_LBRACE] = ACTIONS(373),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(371),
    [anon_sym_await] = ACTIONS(377),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(382),
    [anon_sym_LBRACK] = ACTIONS(654),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(398),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(404),
    [anon_sym_PLUS] = ACTIONS(412),
    [anon_sym_DASH] = ACTIONS(412),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(414),
    [anon_sym_TILDE] = ACTIONS(414),
    [anon_sym_typeof] = ACTIONS(412),
    [anon_sym_void] = ACTIONS(412),
    [anon_sym_delete] = ACTIONS(412),
    [anon_sym_PLUS_PLUS] = ACTIONS(660),
    [anon_sym_DASH_DASH] = ACTIONS(660),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(424),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(428),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(371),
    [anon_sym_get] = ACTIONS(371),
    [anon_sym_set] = ACTIONS(371),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(223)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(498),
    [sym_expression] = STATE(979),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1253),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1253),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(556),
    [sym_subscript_expression] = STATE(556),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1138),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1253),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1816),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(1041),
    [anon_sym_export] = ACTIONS(1043),
    [anon_sym_LBRACE] = ACTIONS(871),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(1043),
    [anon_sym_await] = ACTIONS(1045),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(438),
    [anon_sym_LBRACK] = ACTIONS(877),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(1047),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(447),
    [anon_sym_PLUS] = ACTIONS(452),
    [anon_sym_DASH] = ACTIONS(452),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(454),
    [anon_sym_TILDE] = ACTIONS(454),
    [anon_sym_typeof] = ACTIONS(452),
    [anon_sym_void] = ACTIONS(452),
    [anon_sym_delete] = ACTIONS(452),
    [anon_sym_PLUS_PLUS] = ACTIONS(684),
    [anon_sym_DASH_DASH] = ACTIONS(684),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(459),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(1049),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(1043),
    [anon_sym_get] = ACTIONS(1043),
    [anon_sym_set] = ACTIONS(1043),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(224)] = {
    [sym_import] = STATE(1326),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(695),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1790),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1790),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_function_expression] = STATE(740),
    [sym_generator_function] = STATE(740),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1790),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [aux_sym_export_statement_repeat1] = STATE(1381),
    [sym_identifier] = ACTIONS(161),
    [anon_sym_export] = ACTIONS(163),
    [anon_sym_LBRACE] = ACTIONS(167),
    [anon_sym_import] = ACTIONS(173),
    [anon_sym_let] = ACTIONS(163),
    [anon_sym_await] = ACTIONS(177),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_yield] = ACTIONS(57),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(194),
    [anon_sym_async] = ACTIONS(196),
    [anon_sym_function] = ACTIONS(198),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(75),
    [anon_sym_DASH] = ACTIONS(75),
    [anon_sym_SLASH] = ACTIONS(77),
    [anon_sym_BANG] = ACTIONS(79),
    [anon_sym_TILDE] = ACTIONS(79),
    [anon_sym_typeof] = ACTIONS(75),
    [anon_sym_void] = ACTIONS(75),
    [anon_sym_delete] = ACTIONS(75),
    [anon_sym_PLUS_PLUS] = ACTIONS(81),
    [anon_sym_DASH_DASH] = ACTIONS(81),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(87),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(163),
    [anon_sym_get] = ACTIONS(163),
    [anon_sym_set] = ACTIONS(163),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(225)] = {
    [sym_import] = STATE(1326),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(787),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1790),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1790),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_function_expression] = STATE(740),
    [sym_generator_function] = STATE(740),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1790),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [aux_sym_export_statement_repeat1] = STATE(1381),
    [sym_identifier] = ACTIONS(161),
    [anon_sym_export] = ACTIONS(163),
    [anon_sym_LBRACE] = ACTIONS(167),
    [anon_sym_import] = ACTIONS(173),
    [anon_sym_let] = ACTIONS(163),
    [anon_sym_await] = ACTIONS(177),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_yield] = ACTIONS(57),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(194),
    [anon_sym_async] = ACTIONS(196),
    [anon_sym_function] = ACTIONS(198),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(75),
    [anon_sym_DASH] = ACTIONS(75),
    [anon_sym_SLASH] = ACTIONS(77),
    [anon_sym_BANG] = ACTIONS(79),
    [anon_sym_TILDE] = ACTIONS(79),
    [anon_sym_typeof] = ACTIONS(75),
    [anon_sym_void] = ACTIONS(75),
    [anon_sym_delete] = ACTIONS(75),
    [anon_sym_PLUS_PLUS] = ACTIONS(81),
    [anon_sym_DASH_DASH] = ACTIONS(81),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(87),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(163),
    [anon_sym_get] = ACTIONS(163),
    [anon_sym_set] = ACTIONS(163),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(226)] = {
    [sym_import] = STATE(1326),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(769),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1790),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1790),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_function_expression] = STATE(740),
    [sym_generator_function] = STATE(740),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1790),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [aux_sym_export_statement_repeat1] = STATE(1381),
    [sym_identifier] = ACTIONS(161),
    [anon_sym_export] = ACTIONS(163),
    [anon_sym_LBRACE] = ACTIONS(167),
    [anon_sym_import] = ACTIONS(173),
    [anon_sym_let] = ACTIONS(163),
    [anon_sym_await] = ACTIONS(177),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_yield] = ACTIONS(57),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(194),
    [anon_sym_async] = ACTIONS(196),
    [anon_sym_function] = ACTIONS(198),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(75),
    [anon_sym_DASH] = ACTIONS(75),
    [anon_sym_SLASH] = ACTIONS(77),
    [anon_sym_BANG] = ACTIONS(79),
    [anon_sym_TILDE] = ACTIONS(79),
    [anon_sym_typeof] = ACTIONS(75),
    [anon_sym_void] = ACTIONS(75),
    [anon_sym_delete] = ACTIONS(75),
    [anon_sym_PLUS_PLUS] = ACTIONS(81),
    [anon_sym_DASH_DASH] = ACTIONS(81),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(87),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(163),
    [anon_sym_get] = ACTIONS(163),
    [anon_sym_set] = ACTIONS(163),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(227)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(507),
    [sym_expression] = STATE(941),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1815),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1815),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(507),
    [sym_subscript_expression] = STATE(507),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1139),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1815),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1763),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(369),
    [anon_sym_export] = ACTIONS(371),
    [anon_sym_LBRACE] = ACTIONS(373),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(371),
    [anon_sym_await] = ACTIONS(377),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(382),
    [anon_sym_LBRACK] = ACTIONS(654),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(398),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(404),
    [anon_sym_PLUS] = ACTIONS(412),
    [anon_sym_DASH] = ACTIONS(412),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(414),
    [anon_sym_TILDE] = ACTIONS(414),
    [anon_sym_typeof] = ACTIONS(412),
    [anon_sym_void] = ACTIONS(412),
    [anon_sym_delete] = ACTIONS(412),
    [anon_sym_PLUS_PLUS] = ACTIONS(660),
    [anon_sym_DASH_DASH] = ACTIONS(660),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(424),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(428),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(371),
    [anon_sym_get] = ACTIONS(371),
    [anon_sym_set] = ACTIONS(371),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(228)] = {
    [sym_import] = STATE(1326),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(762),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1790),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1790),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_function_expression] = STATE(740),
    [sym_generator_function] = STATE(740),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1790),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [aux_sym_export_statement_repeat1] = STATE(1381),
    [sym_identifier] = ACTIONS(161),
    [anon_sym_export] = ACTIONS(163),
    [anon_sym_LBRACE] = ACTIONS(167),
    [anon_sym_import] = ACTIONS(173),
    [anon_sym_let] = ACTIONS(163),
    [anon_sym_await] = ACTIONS(177),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_yield] = ACTIONS(57),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(194),
    [anon_sym_async] = ACTIONS(196),
    [anon_sym_function] = ACTIONS(198),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(75),
    [anon_sym_DASH] = ACTIONS(75),
    [anon_sym_SLASH] = ACTIONS(77),
    [anon_sym_BANG] = ACTIONS(79),
    [anon_sym_TILDE] = ACTIONS(79),
    [anon_sym_typeof] = ACTIONS(75),
    [anon_sym_void] = ACTIONS(75),
    [anon_sym_delete] = ACTIONS(75),
    [anon_sym_PLUS_PLUS] = ACTIONS(81),
    [anon_sym_DASH_DASH] = ACTIONS(81),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(87),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(163),
    [anon_sym_get] = ACTIONS(163),
    [anon_sym_set] = ACTIONS(163),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(229)] = {
    [sym_import] = STATE(1326),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(691),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1790),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1790),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_function_expression] = STATE(740),
    [sym_generator_function] = STATE(740),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1790),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [aux_sym_export_statement_repeat1] = STATE(1381),
    [sym_identifier] = ACTIONS(161),
    [anon_sym_export] = ACTIONS(163),
    [anon_sym_LBRACE] = ACTIONS(167),
    [anon_sym_import] = ACTIONS(173),
    [anon_sym_let] = ACTIONS(163),
    [anon_sym_await] = ACTIONS(177),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_yield] = ACTIONS(57),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(194),
    [anon_sym_async] = ACTIONS(196),
    [anon_sym_function] = ACTIONS(198),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(75),
    [anon_sym_DASH] = ACTIONS(75),
    [anon_sym_SLASH] = ACTIONS(77),
    [anon_sym_BANG] = ACTIONS(79),
    [anon_sym_TILDE] = ACTIONS(79),
    [anon_sym_typeof] = ACTIONS(75),
    [anon_sym_void] = ACTIONS(75),
    [anon_sym_delete] = ACTIONS(75),
    [anon_sym_PLUS_PLUS] = ACTIONS(81),
    [anon_sym_DASH_DASH] = ACTIONS(81),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(87),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(163),
    [anon_sym_get] = ACTIONS(163),
    [anon_sym_set] = ACTIONS(163),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(230)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(578),
    [sym_expression] = STATE(971),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1776),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1776),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(578),
    [sym_subscript_expression] = STATE(578),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1142),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1776),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1778),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(613),
    [anon_sym_export] = ACTIONS(615),
    [anon_sym_LBRACE] = ACTIONS(434),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(615),
    [anon_sym_await] = ACTIONS(620),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(622),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(626),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(630),
    [anon_sym_PLUS] = ACTIONS(635),
    [anon_sym_DASH] = ACTIONS(635),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(637),
    [anon_sym_TILDE] = ACTIONS(637),
    [anon_sym_typeof] = ACTIONS(635),
    [anon_sym_void] = ACTIONS(635),
    [anon_sym_delete] = ACTIONS(635),
    [anon_sym_PLUS_PLUS] = ACTIONS(688),
    [anon_sym_DASH_DASH] = ACTIONS(688),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(642),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(644),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(615),
    [anon_sym_get] = ACTIONS(615),
    [anon_sym_set] = ACTIONS(615),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(231)] = {
    [sym_import] = STATE(1326),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(719),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1790),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1790),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_function_expression] = STATE(740),
    [sym_generator_function] = STATE(740),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1790),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [aux_sym_export_statement_repeat1] = STATE(1381),
    [sym_identifier] = ACTIONS(161),
    [anon_sym_export] = ACTIONS(163),
    [anon_sym_LBRACE] = ACTIONS(167),
    [anon_sym_import] = ACTIONS(173),
    [anon_sym_let] = ACTIONS(163),
    [anon_sym_await] = ACTIONS(177),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_yield] = ACTIONS(57),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(194),
    [anon_sym_async] = ACTIONS(196),
    [anon_sym_function] = ACTIONS(198),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(75),
    [anon_sym_DASH] = ACTIONS(75),
    [anon_sym_SLASH] = ACTIONS(77),
    [anon_sym_BANG] = ACTIONS(79),
    [anon_sym_TILDE] = ACTIONS(79),
    [anon_sym_typeof] = ACTIONS(75),
    [anon_sym_void] = ACTIONS(75),
    [anon_sym_delete] = ACTIONS(75),
    [anon_sym_PLUS_PLUS] = ACTIONS(81),
    [anon_sym_DASH_DASH] = ACTIONS(81),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(87),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(163),
    [anon_sym_get] = ACTIONS(163),
    [anon_sym_set] = ACTIONS(163),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(232)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(498),
    [sym_expression] = STATE(968),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1762),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1762),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(498),
    [sym_subscript_expression] = STATE(498),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1138),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1762),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1816),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(430),
    [anon_sym_export] = ACTIONS(432),
    [anon_sym_LBRACE] = ACTIONS(434),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(432),
    [anon_sym_await] = ACTIONS(436),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(438),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(445),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(447),
    [anon_sym_PLUS] = ACTIONS(452),
    [anon_sym_DASH] = ACTIONS(452),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(454),
    [anon_sym_TILDE] = ACTIONS(454),
    [anon_sym_typeof] = ACTIONS(452),
    [anon_sym_void] = ACTIONS(452),
    [anon_sym_delete] = ACTIONS(452),
    [anon_sym_PLUS_PLUS] = ACTIONS(684),
    [anon_sym_DASH_DASH] = ACTIONS(684),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(459),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(461),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(432),
    [anon_sym_get] = ACTIONS(432),
    [anon_sym_set] = ACTIONS(432),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(233)] = {
    [sym_import] = STATE(1326),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(749),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1790),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1790),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_function_expression] = STATE(740),
    [sym_generator_function] = STATE(740),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1790),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [aux_sym_export_statement_repeat1] = STATE(1381),
    [sym_identifier] = ACTIONS(161),
    [anon_sym_export] = ACTIONS(163),
    [anon_sym_LBRACE] = ACTIONS(167),
    [anon_sym_import] = ACTIONS(173),
    [anon_sym_let] = ACTIONS(163),
    [anon_sym_await] = ACTIONS(177),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_yield] = ACTIONS(57),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(194),
    [anon_sym_async] = ACTIONS(196),
    [anon_sym_function] = ACTIONS(198),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(75),
    [anon_sym_DASH] = ACTIONS(75),
    [anon_sym_SLASH] = ACTIONS(77),
    [anon_sym_BANG] = ACTIONS(79),
    [anon_sym_TILDE] = ACTIONS(79),
    [anon_sym_typeof] = ACTIONS(75),
    [anon_sym_void] = ACTIONS(75),
    [anon_sym_delete] = ACTIONS(75),
    [anon_sym_PLUS_PLUS] = ACTIONS(81),
    [anon_sym_DASH_DASH] = ACTIONS(81),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(87),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(163),
    [anon_sym_get] = ACTIONS(163),
    [anon_sym_set] = ACTIONS(163),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(234)] = {
    [sym_import] = STATE(1326),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(848),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1790),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1790),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_function_expression] = STATE(740),
    [sym_generator_function] = STATE(740),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1790),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [aux_sym_export_statement_repeat1] = STATE(1381),
    [sym_identifier] = ACTIONS(161),
    [anon_sym_export] = ACTIONS(163),
    [anon_sym_LBRACE] = ACTIONS(167),
    [anon_sym_import] = ACTIONS(173),
    [anon_sym_let] = ACTIONS(163),
    [anon_sym_await] = ACTIONS(177),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_yield] = ACTIONS(57),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(194),
    [anon_sym_async] = ACTIONS(196),
    [anon_sym_function] = ACTIONS(198),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(75),
    [anon_sym_DASH] = ACTIONS(75),
    [anon_sym_SLASH] = ACTIONS(77),
    [anon_sym_BANG] = ACTIONS(79),
    [anon_sym_TILDE] = ACTIONS(79),
    [anon_sym_typeof] = ACTIONS(75),
    [anon_sym_void] = ACTIONS(75),
    [anon_sym_delete] = ACTIONS(75),
    [anon_sym_PLUS_PLUS] = ACTIONS(81),
    [anon_sym_DASH_DASH] = ACTIONS(81),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(87),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(163),
    [anon_sym_get] = ACTIONS(163),
    [anon_sym_set] = ACTIONS(163),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(235)] = {
    [sym_import] = STATE(1326),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(751),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1790),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1790),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_function_expression] = STATE(740),
    [sym_generator_function] = STATE(740),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1790),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [aux_sym_export_statement_repeat1] = STATE(1381),
    [sym_identifier] = ACTIONS(161),
    [anon_sym_export] = ACTIONS(163),
    [anon_sym_LBRACE] = ACTIONS(167),
    [anon_sym_import] = ACTIONS(173),
    [anon_sym_let] = ACTIONS(163),
    [anon_sym_await] = ACTIONS(177),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_yield] = ACTIONS(57),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(194),
    [anon_sym_async] = ACTIONS(196),
    [anon_sym_function] = ACTIONS(198),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(75),
    [anon_sym_DASH] = ACTIONS(75),
    [anon_sym_SLASH] = ACTIONS(77),
    [anon_sym_BANG] = ACTIONS(79),
    [anon_sym_TILDE] = ACTIONS(79),
    [anon_sym_typeof] = ACTIONS(75),
    [anon_sym_void] = ACTIONS(75),
    [anon_sym_delete] = ACTIONS(75),
    [anon_sym_PLUS_PLUS] = ACTIONS(81),
    [anon_sym_DASH_DASH] = ACTIONS(81),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(87),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(163),
    [anon_sym_get] = ACTIONS(163),
    [anon_sym_set] = ACTIONS(163),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(236)] = {
    [sym_import] = STATE(1326),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(753),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1790),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1790),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_function_expression] = STATE(740),
    [sym_generator_function] = STATE(740),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1790),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [aux_sym_export_statement_repeat1] = STATE(1381),
    [sym_identifier] = ACTIONS(161),
    [anon_sym_export] = ACTIONS(163),
    [anon_sym_LBRACE] = ACTIONS(167),
    [anon_sym_import] = ACTIONS(173),
    [anon_sym_let] = ACTIONS(163),
    [anon_sym_await] = ACTIONS(177),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_yield] = ACTIONS(57),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(194),
    [anon_sym_async] = ACTIONS(196),
    [anon_sym_function] = ACTIONS(198),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(75),
    [anon_sym_DASH] = ACTIONS(75),
    [anon_sym_SLASH] = ACTIONS(77),
    [anon_sym_BANG] = ACTIONS(79),
    [anon_sym_TILDE] = ACTIONS(79),
    [anon_sym_typeof] = ACTIONS(75),
    [anon_sym_void] = ACTIONS(75),
    [anon_sym_delete] = ACTIONS(75),
    [anon_sym_PLUS_PLUS] = ACTIONS(81),
    [anon_sym_DASH_DASH] = ACTIONS(81),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(87),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(163),
    [anon_sym_get] = ACTIONS(163),
    [anon_sym_set] = ACTIONS(163),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(237)] = {
    [sym_import] = STATE(1326),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(756),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1790),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1790),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_function_expression] = STATE(740),
    [sym_generator_function] = STATE(740),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1790),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [aux_sym_export_statement_repeat1] = STATE(1381),
    [sym_identifier] = ACTIONS(161),
    [anon_sym_export] = ACTIONS(163),
    [anon_sym_LBRACE] = ACTIONS(167),
    [anon_sym_import] = ACTIONS(173),
    [anon_sym_let] = ACTIONS(163),
    [anon_sym_await] = ACTIONS(177),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_yield] = ACTIONS(57),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(194),
    [anon_sym_async] = ACTIONS(196),
    [anon_sym_function] = ACTIONS(198),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(75),
    [anon_sym_DASH] = ACTIONS(75),
    [anon_sym_SLASH] = ACTIONS(77),
    [anon_sym_BANG] = ACTIONS(79),
    [anon_sym_TILDE] = ACTIONS(79),
    [anon_sym_typeof] = ACTIONS(75),
    [anon_sym_void] = ACTIONS(75),
    [anon_sym_delete] = ACTIONS(75),
    [anon_sym_PLUS_PLUS] = ACTIONS(81),
    [anon_sym_DASH_DASH] = ACTIONS(81),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(87),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(163),
    [anon_sym_get] = ACTIONS(163),
    [anon_sym_set] = ACTIONS(163),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(238)] = {
    [sym_import] = STATE(1326),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(760),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1790),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1790),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_function_expression] = STATE(740),
    [sym_generator_function] = STATE(740),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1790),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [aux_sym_export_statement_repeat1] = STATE(1381),
    [sym_identifier] = ACTIONS(161),
    [anon_sym_export] = ACTIONS(163),
    [anon_sym_LBRACE] = ACTIONS(167),
    [anon_sym_import] = ACTIONS(173),
    [anon_sym_let] = ACTIONS(163),
    [anon_sym_await] = ACTIONS(177),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_yield] = ACTIONS(57),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(194),
    [anon_sym_async] = ACTIONS(196),
    [anon_sym_function] = ACTIONS(198),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(75),
    [anon_sym_DASH] = ACTIONS(75),
    [anon_sym_SLASH] = ACTIONS(77),
    [anon_sym_BANG] = ACTIONS(79),
    [anon_sym_TILDE] = ACTIONS(79),
    [anon_sym_typeof] = ACTIONS(75),
    [anon_sym_void] = ACTIONS(75),
    [anon_sym_delete] = ACTIONS(75),
    [anon_sym_PLUS_PLUS] = ACTIONS(81),
    [anon_sym_DASH_DASH] = ACTIONS(81),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(87),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(163),
    [anon_sym_get] = ACTIONS(163),
    [anon_sym_set] = ACTIONS(163),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(239)] = {
    [sym_import] = STATE(1326),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(763),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1790),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1790),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_function_expression] = STATE(740),
    [sym_generator_function] = STATE(740),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1790),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [aux_sym_export_statement_repeat1] = STATE(1381),
    [sym_identifier] = ACTIONS(161),
    [anon_sym_export] = ACTIONS(163),
    [anon_sym_LBRACE] = ACTIONS(167),
    [anon_sym_import] = ACTIONS(173),
    [anon_sym_let] = ACTIONS(163),
    [anon_sym_await] = ACTIONS(177),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_yield] = ACTIONS(57),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(194),
    [anon_sym_async] = ACTIONS(196),
    [anon_sym_function] = ACTIONS(198),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(75),
    [anon_sym_DASH] = ACTIONS(75),
    [anon_sym_SLASH] = ACTIONS(77),
    [anon_sym_BANG] = ACTIONS(79),
    [anon_sym_TILDE] = ACTIONS(79),
    [anon_sym_typeof] = ACTIONS(75),
    [anon_sym_void] = ACTIONS(75),
    [anon_sym_delete] = ACTIONS(75),
    [anon_sym_PLUS_PLUS] = ACTIONS(81),
    [anon_sym_DASH_DASH] = ACTIONS(81),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(87),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(163),
    [anon_sym_get] = ACTIONS(163),
    [anon_sym_set] = ACTIONS(163),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(240)] = {
    [sym_import] = STATE(1326),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(775),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1790),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1790),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_function_expression] = STATE(740),
    [sym_generator_function] = STATE(740),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1790),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [aux_sym_export_statement_repeat1] = STATE(1381),
    [sym_identifier] = ACTIONS(161),
    [anon_sym_export] = ACTIONS(163),
    [anon_sym_LBRACE] = ACTIONS(167),
    [anon_sym_import] = ACTIONS(173),
    [anon_sym_let] = ACTIONS(163),
    [anon_sym_await] = ACTIONS(177),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_yield] = ACTIONS(57),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(194),
    [anon_sym_async] = ACTIONS(196),
    [anon_sym_function] = ACTIONS(198),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(75),
    [anon_sym_DASH] = ACTIONS(75),
    [anon_sym_SLASH] = ACTIONS(77),
    [anon_sym_BANG] = ACTIONS(79),
    [anon_sym_TILDE] = ACTIONS(79),
    [anon_sym_typeof] = ACTIONS(75),
    [anon_sym_void] = ACTIONS(75),
    [anon_sym_delete] = ACTIONS(75),
    [anon_sym_PLUS_PLUS] = ACTIONS(81),
    [anon_sym_DASH_DASH] = ACTIONS(81),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(87),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(163),
    [anon_sym_get] = ACTIONS(163),
    [anon_sym_set] = ACTIONS(163),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(241)] = {
    [sym_import] = STATE(1326),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(776),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1790),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1790),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_function_expression] = STATE(740),
    [sym_generator_function] = STATE(740),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1790),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [aux_sym_export_statement_repeat1] = STATE(1381),
    [sym_identifier] = ACTIONS(161),
    [anon_sym_export] = ACTIONS(163),
    [anon_sym_LBRACE] = ACTIONS(167),
    [anon_sym_import] = ACTIONS(173),
    [anon_sym_let] = ACTIONS(163),
    [anon_sym_await] = ACTIONS(177),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_yield] = ACTIONS(57),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(194),
    [anon_sym_async] = ACTIONS(196),
    [anon_sym_function] = ACTIONS(198),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(75),
    [anon_sym_DASH] = ACTIONS(75),
    [anon_sym_SLASH] = ACTIONS(77),
    [anon_sym_BANG] = ACTIONS(79),
    [anon_sym_TILDE] = ACTIONS(79),
    [anon_sym_typeof] = ACTIONS(75),
    [anon_sym_void] = ACTIONS(75),
    [anon_sym_delete] = ACTIONS(75),
    [anon_sym_PLUS_PLUS] = ACTIONS(81),
    [anon_sym_DASH_DASH] = ACTIONS(81),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(87),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(163),
    [anon_sym_get] = ACTIONS(163),
    [anon_sym_set] = ACTIONS(163),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(242)] = {
    [sym_import] = STATE(1326),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(777),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1790),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1790),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_function_expression] = STATE(740),
    [sym_generator_function] = STATE(740),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1790),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [aux_sym_export_statement_repeat1] = STATE(1381),
    [sym_identifier] = ACTIONS(161),
    [anon_sym_export] = ACTIONS(163),
    [anon_sym_LBRACE] = ACTIONS(167),
    [anon_sym_import] = ACTIONS(173),
    [anon_sym_let] = ACTIONS(163),
    [anon_sym_await] = ACTIONS(177),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_yield] = ACTIONS(57),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(194),
    [anon_sym_async] = ACTIONS(196),
    [anon_sym_function] = ACTIONS(198),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(75),
    [anon_sym_DASH] = ACTIONS(75),
    [anon_sym_SLASH] = ACTIONS(77),
    [anon_sym_BANG] = ACTIONS(79),
    [anon_sym_TILDE] = ACTIONS(79),
    [anon_sym_typeof] = ACTIONS(75),
    [anon_sym_void] = ACTIONS(75),
    [anon_sym_delete] = ACTIONS(75),
    [anon_sym_PLUS_PLUS] = ACTIONS(81),
    [anon_sym_DASH_DASH] = ACTIONS(81),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(87),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(163),
    [anon_sym_get] = ACTIONS(163),
    [anon_sym_set] = ACTIONS(163),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(243)] = {
    [sym_import] = STATE(1326),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(778),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1790),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1790),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_function_expression] = STATE(740),
    [sym_generator_function] = STATE(740),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1790),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [aux_sym_export_statement_repeat1] = STATE(1381),
    [sym_identifier] = ACTIONS(161),
    [anon_sym_export] = ACTIONS(163),
    [anon_sym_LBRACE] = ACTIONS(167),
    [anon_sym_import] = ACTIONS(173),
    [anon_sym_let] = ACTIONS(163),
    [anon_sym_await] = ACTIONS(177),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_yield] = ACTIONS(57),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(194),
    [anon_sym_async] = ACTIONS(196),
    [anon_sym_function] = ACTIONS(198),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(75),
    [anon_sym_DASH] = ACTIONS(75),
    [anon_sym_SLASH] = ACTIONS(77),
    [anon_sym_BANG] = ACTIONS(79),
    [anon_sym_TILDE] = ACTIONS(79),
    [anon_sym_typeof] = ACTIONS(75),
    [anon_sym_void] = ACTIONS(75),
    [anon_sym_delete] = ACTIONS(75),
    [anon_sym_PLUS_PLUS] = ACTIONS(81),
    [anon_sym_DASH_DASH] = ACTIONS(81),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(87),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(163),
    [anon_sym_get] = ACTIONS(163),
    [anon_sym_set] = ACTIONS(163),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(244)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(507),
    [sym_expression] = STATE(588),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1815),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1815),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(507),
    [sym_subscript_expression] = STATE(507),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1139),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1815),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1763),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(369),
    [anon_sym_export] = ACTIONS(371),
    [anon_sym_LBRACE] = ACTIONS(373),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(371),
    [anon_sym_await] = ACTIONS(377),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(382),
    [anon_sym_LBRACK] = ACTIONS(654),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(398),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(404),
    [anon_sym_PLUS] = ACTIONS(412),
    [anon_sym_DASH] = ACTIONS(412),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(414),
    [anon_sym_TILDE] = ACTIONS(414),
    [anon_sym_typeof] = ACTIONS(412),
    [anon_sym_void] = ACTIONS(412),
    [anon_sym_delete] = ACTIONS(412),
    [anon_sym_PLUS_PLUS] = ACTIONS(660),
    [anon_sym_DASH_DASH] = ACTIONS(660),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(424),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(428),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(371),
    [anon_sym_get] = ACTIONS(371),
    [anon_sym_set] = ACTIONS(371),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(245)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(507),
    [sym_expression] = STATE(591),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1815),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1815),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(507),
    [sym_subscript_expression] = STATE(507),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1139),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1815),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1763),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(369),
    [anon_sym_export] = ACTIONS(371),
    [anon_sym_LBRACE] = ACTIONS(373),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(371),
    [anon_sym_await] = ACTIONS(377),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(382),
    [anon_sym_LBRACK] = ACTIONS(654),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(398),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(404),
    [anon_sym_PLUS] = ACTIONS(412),
    [anon_sym_DASH] = ACTIONS(412),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(414),
    [anon_sym_TILDE] = ACTIONS(414),
    [anon_sym_typeof] = ACTIONS(412),
    [anon_sym_void] = ACTIONS(412),
    [anon_sym_delete] = ACTIONS(412),
    [anon_sym_PLUS_PLUS] = ACTIONS(660),
    [anon_sym_DASH_DASH] = ACTIONS(660),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(424),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(428),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(371),
    [anon_sym_get] = ACTIONS(371),
    [anon_sym_set] = ACTIONS(371),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(246)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(507),
    [sym_expression] = STATE(656),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1815),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1815),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(507),
    [sym_subscript_expression] = STATE(507),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1139),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1815),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1763),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(369),
    [anon_sym_export] = ACTIONS(371),
    [anon_sym_LBRACE] = ACTIONS(373),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(371),
    [anon_sym_await] = ACTIONS(377),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(382),
    [anon_sym_LBRACK] = ACTIONS(654),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(398),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(404),
    [anon_sym_PLUS] = ACTIONS(412),
    [anon_sym_DASH] = ACTIONS(412),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(414),
    [anon_sym_TILDE] = ACTIONS(414),
    [anon_sym_typeof] = ACTIONS(412),
    [anon_sym_void] = ACTIONS(412),
    [anon_sym_delete] = ACTIONS(412),
    [anon_sym_PLUS_PLUS] = ACTIONS(660),
    [anon_sym_DASH_DASH] = ACTIONS(660),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(424),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(428),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(371),
    [anon_sym_get] = ACTIONS(371),
    [anon_sym_set] = ACTIONS(371),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(247)] = {
    [sym_import] = STATE(1326),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(780),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1790),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1790),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_function_expression] = STATE(740),
    [sym_generator_function] = STATE(740),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1790),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [aux_sym_export_statement_repeat1] = STATE(1381),
    [sym_identifier] = ACTIONS(161),
    [anon_sym_export] = ACTIONS(163),
    [anon_sym_LBRACE] = ACTIONS(167),
    [anon_sym_import] = ACTIONS(173),
    [anon_sym_let] = ACTIONS(163),
    [anon_sym_await] = ACTIONS(177),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_yield] = ACTIONS(57),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(194),
    [anon_sym_async] = ACTIONS(196),
    [anon_sym_function] = ACTIONS(198),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(75),
    [anon_sym_DASH] = ACTIONS(75),
    [anon_sym_SLASH] = ACTIONS(77),
    [anon_sym_BANG] = ACTIONS(79),
    [anon_sym_TILDE] = ACTIONS(79),
    [anon_sym_typeof] = ACTIONS(75),
    [anon_sym_void] = ACTIONS(75),
    [anon_sym_delete] = ACTIONS(75),
    [anon_sym_PLUS_PLUS] = ACTIONS(81),
    [anon_sym_DASH_DASH] = ACTIONS(81),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(87),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(163),
    [anon_sym_get] = ACTIONS(163),
    [anon_sym_set] = ACTIONS(163),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(248)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(507),
    [sym_expression] = STATE(653),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1815),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1815),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(507),
    [sym_subscript_expression] = STATE(507),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1139),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1815),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1763),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(369),
    [anon_sym_export] = ACTIONS(371),
    [anon_sym_LBRACE] = ACTIONS(373),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(371),
    [anon_sym_await] = ACTIONS(377),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(382),
    [anon_sym_LBRACK] = ACTIONS(654),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(398),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(404),
    [anon_sym_PLUS] = ACTIONS(412),
    [anon_sym_DASH] = ACTIONS(412),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(414),
    [anon_sym_TILDE] = ACTIONS(414),
    [anon_sym_typeof] = ACTIONS(412),
    [anon_sym_void] = ACTIONS(412),
    [anon_sym_delete] = ACTIONS(412),
    [anon_sym_PLUS_PLUS] = ACTIONS(660),
    [anon_sym_DASH_DASH] = ACTIONS(660),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(424),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(428),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(371),
    [anon_sym_get] = ACTIONS(371),
    [anon_sym_set] = ACTIONS(371),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(249)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(507),
    [sym_expression] = STATE(660),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1815),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1815),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(507),
    [sym_subscript_expression] = STATE(507),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1139),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1815),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1763),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(369),
    [anon_sym_export] = ACTIONS(371),
    [anon_sym_LBRACE] = ACTIONS(373),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(371),
    [anon_sym_await] = ACTIONS(377),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(382),
    [anon_sym_LBRACK] = ACTIONS(654),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(398),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(404),
    [anon_sym_PLUS] = ACTIONS(412),
    [anon_sym_DASH] = ACTIONS(412),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(414),
    [anon_sym_TILDE] = ACTIONS(414),
    [anon_sym_typeof] = ACTIONS(412),
    [anon_sym_void] = ACTIONS(412),
    [anon_sym_delete] = ACTIONS(412),
    [anon_sym_PLUS_PLUS] = ACTIONS(660),
    [anon_sym_DASH_DASH] = ACTIONS(660),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(424),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(428),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(371),
    [anon_sym_get] = ACTIONS(371),
    [anon_sym_set] = ACTIONS(371),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(250)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(507),
    [sym_expression] = STATE(662),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1815),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1815),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(507),
    [sym_subscript_expression] = STATE(507),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1139),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1815),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1763),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(369),
    [anon_sym_export] = ACTIONS(371),
    [anon_sym_LBRACE] = ACTIONS(373),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(371),
    [anon_sym_await] = ACTIONS(377),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(382),
    [anon_sym_LBRACK] = ACTIONS(654),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(398),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(404),
    [anon_sym_PLUS] = ACTIONS(412),
    [anon_sym_DASH] = ACTIONS(412),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(414),
    [anon_sym_TILDE] = ACTIONS(414),
    [anon_sym_typeof] = ACTIONS(412),
    [anon_sym_void] = ACTIONS(412),
    [anon_sym_delete] = ACTIONS(412),
    [anon_sym_PLUS_PLUS] = ACTIONS(660),
    [anon_sym_DASH_DASH] = ACTIONS(660),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(424),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(428),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(371),
    [anon_sym_get] = ACTIONS(371),
    [anon_sym_set] = ACTIONS(371),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(251)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(507),
    [sym_expression] = STATE(669),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1815),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1815),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(507),
    [sym_subscript_expression] = STATE(507),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1139),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1815),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1763),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(369),
    [anon_sym_export] = ACTIONS(371),
    [anon_sym_LBRACE] = ACTIONS(373),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(371),
    [anon_sym_await] = ACTIONS(377),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(382),
    [anon_sym_LBRACK] = ACTIONS(654),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(398),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(404),
    [anon_sym_PLUS] = ACTIONS(412),
    [anon_sym_DASH] = ACTIONS(412),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(414),
    [anon_sym_TILDE] = ACTIONS(414),
    [anon_sym_typeof] = ACTIONS(412),
    [anon_sym_void] = ACTIONS(412),
    [anon_sym_delete] = ACTIONS(412),
    [anon_sym_PLUS_PLUS] = ACTIONS(660),
    [anon_sym_DASH_DASH] = ACTIONS(660),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(424),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(428),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(371),
    [anon_sym_get] = ACTIONS(371),
    [anon_sym_set] = ACTIONS(371),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(252)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(507),
    [sym_expression] = STATE(690),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1815),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1815),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(507),
    [sym_subscript_expression] = STATE(507),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1139),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1815),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1763),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(369),
    [anon_sym_export] = ACTIONS(371),
    [anon_sym_LBRACE] = ACTIONS(373),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(371),
    [anon_sym_await] = ACTIONS(377),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(382),
    [anon_sym_LBRACK] = ACTIONS(654),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(398),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(404),
    [anon_sym_PLUS] = ACTIONS(412),
    [anon_sym_DASH] = ACTIONS(412),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(414),
    [anon_sym_TILDE] = ACTIONS(414),
    [anon_sym_typeof] = ACTIONS(412),
    [anon_sym_void] = ACTIONS(412),
    [anon_sym_delete] = ACTIONS(412),
    [anon_sym_PLUS_PLUS] = ACTIONS(660),
    [anon_sym_DASH_DASH] = ACTIONS(660),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(424),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(428),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(371),
    [anon_sym_get] = ACTIONS(371),
    [anon_sym_set] = ACTIONS(371),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(253)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(507),
    [sym_expression] = STATE(671),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1815),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1815),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(507),
    [sym_subscript_expression] = STATE(507),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1139),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1815),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1763),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(369),
    [anon_sym_export] = ACTIONS(371),
    [anon_sym_LBRACE] = ACTIONS(373),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(371),
    [anon_sym_await] = ACTIONS(377),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(382),
    [anon_sym_LBRACK] = ACTIONS(654),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(398),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(404),
    [anon_sym_PLUS] = ACTIONS(412),
    [anon_sym_DASH] = ACTIONS(412),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(414),
    [anon_sym_TILDE] = ACTIONS(414),
    [anon_sym_typeof] = ACTIONS(412),
    [anon_sym_void] = ACTIONS(412),
    [anon_sym_delete] = ACTIONS(412),
    [anon_sym_PLUS_PLUS] = ACTIONS(660),
    [anon_sym_DASH_DASH] = ACTIONS(660),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(424),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(428),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(371),
    [anon_sym_get] = ACTIONS(371),
    [anon_sym_set] = ACTIONS(371),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(254)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(507),
    [sym_expression] = STATE(672),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1815),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1815),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(507),
    [sym_subscript_expression] = STATE(507),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1139),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1815),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1763),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(369),
    [anon_sym_export] = ACTIONS(371),
    [anon_sym_LBRACE] = ACTIONS(373),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(371),
    [anon_sym_await] = ACTIONS(377),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(382),
    [anon_sym_LBRACK] = ACTIONS(654),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(398),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(404),
    [anon_sym_PLUS] = ACTIONS(412),
    [anon_sym_DASH] = ACTIONS(412),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(414),
    [anon_sym_TILDE] = ACTIONS(414),
    [anon_sym_typeof] = ACTIONS(412),
    [anon_sym_void] = ACTIONS(412),
    [anon_sym_delete] = ACTIONS(412),
    [anon_sym_PLUS_PLUS] = ACTIONS(660),
    [anon_sym_DASH_DASH] = ACTIONS(660),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(424),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(428),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(371),
    [anon_sym_get] = ACTIONS(371),
    [anon_sym_set] = ACTIONS(371),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(255)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(507),
    [sym_expression] = STATE(673),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1815),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1815),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(507),
    [sym_subscript_expression] = STATE(507),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1139),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1815),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1763),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(369),
    [anon_sym_export] = ACTIONS(371),
    [anon_sym_LBRACE] = ACTIONS(373),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(371),
    [anon_sym_await] = ACTIONS(377),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(382),
    [anon_sym_LBRACK] = ACTIONS(654),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(398),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(404),
    [anon_sym_PLUS] = ACTIONS(412),
    [anon_sym_DASH] = ACTIONS(412),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(414),
    [anon_sym_TILDE] = ACTIONS(414),
    [anon_sym_typeof] = ACTIONS(412),
    [anon_sym_void] = ACTIONS(412),
    [anon_sym_delete] = ACTIONS(412),
    [anon_sym_PLUS_PLUS] = ACTIONS(660),
    [anon_sym_DASH_DASH] = ACTIONS(660),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(424),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(428),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(371),
    [anon_sym_get] = ACTIONS(371),
    [anon_sym_set] = ACTIONS(371),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(256)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(507),
    [sym_expression] = STATE(674),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1815),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1815),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(507),
    [sym_subscript_expression] = STATE(507),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1139),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1815),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1763),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(369),
    [anon_sym_export] = ACTIONS(371),
    [anon_sym_LBRACE] = ACTIONS(373),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(371),
    [anon_sym_await] = ACTIONS(377),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(382),
    [anon_sym_LBRACK] = ACTIONS(654),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(398),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(404),
    [anon_sym_PLUS] = ACTIONS(412),
    [anon_sym_DASH] = ACTIONS(412),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(414),
    [anon_sym_TILDE] = ACTIONS(414),
    [anon_sym_typeof] = ACTIONS(412),
    [anon_sym_void] = ACTIONS(412),
    [anon_sym_delete] = ACTIONS(412),
    [anon_sym_PLUS_PLUS] = ACTIONS(660),
    [anon_sym_DASH_DASH] = ACTIONS(660),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(424),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(428),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(371),
    [anon_sym_get] = ACTIONS(371),
    [anon_sym_set] = ACTIONS(371),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(257)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(507),
    [sym_expression] = STATE(675),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1815),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1815),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(507),
    [sym_subscript_expression] = STATE(507),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1139),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1815),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1763),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(369),
    [anon_sym_export] = ACTIONS(371),
    [anon_sym_LBRACE] = ACTIONS(373),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(371),
    [anon_sym_await] = ACTIONS(377),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(382),
    [anon_sym_LBRACK] = ACTIONS(654),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(398),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(404),
    [anon_sym_PLUS] = ACTIONS(412),
    [anon_sym_DASH] = ACTIONS(412),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(414),
    [anon_sym_TILDE] = ACTIONS(414),
    [anon_sym_typeof] = ACTIONS(412),
    [anon_sym_void] = ACTIONS(412),
    [anon_sym_delete] = ACTIONS(412),
    [anon_sym_PLUS_PLUS] = ACTIONS(660),
    [anon_sym_DASH_DASH] = ACTIONS(660),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(424),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(428),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(371),
    [anon_sym_get] = ACTIONS(371),
    [anon_sym_set] = ACTIONS(371),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(258)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(507),
    [sym_expression] = STATE(676),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1815),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1815),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(507),
    [sym_subscript_expression] = STATE(507),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1139),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1815),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1763),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(369),
    [anon_sym_export] = ACTIONS(371),
    [anon_sym_LBRACE] = ACTIONS(373),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(371),
    [anon_sym_await] = ACTIONS(377),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(382),
    [anon_sym_LBRACK] = ACTIONS(654),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(398),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(404),
    [anon_sym_PLUS] = ACTIONS(412),
    [anon_sym_DASH] = ACTIONS(412),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(414),
    [anon_sym_TILDE] = ACTIONS(414),
    [anon_sym_typeof] = ACTIONS(412),
    [anon_sym_void] = ACTIONS(412),
    [anon_sym_delete] = ACTIONS(412),
    [anon_sym_PLUS_PLUS] = ACTIONS(660),
    [anon_sym_DASH_DASH] = ACTIONS(660),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(424),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(428),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(371),
    [anon_sym_get] = ACTIONS(371),
    [anon_sym_set] = ACTIONS(371),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(259)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(507),
    [sym_expression] = STATE(677),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1815),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1815),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(507),
    [sym_subscript_expression] = STATE(507),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1139),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1815),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1763),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(369),
    [anon_sym_export] = ACTIONS(371),
    [anon_sym_LBRACE] = ACTIONS(373),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(371),
    [anon_sym_await] = ACTIONS(377),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(382),
    [anon_sym_LBRACK] = ACTIONS(654),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(398),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(404),
    [anon_sym_PLUS] = ACTIONS(412),
    [anon_sym_DASH] = ACTIONS(412),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(414),
    [anon_sym_TILDE] = ACTIONS(414),
    [anon_sym_typeof] = ACTIONS(412),
    [anon_sym_void] = ACTIONS(412),
    [anon_sym_delete] = ACTIONS(412),
    [anon_sym_PLUS_PLUS] = ACTIONS(660),
    [anon_sym_DASH_DASH] = ACTIONS(660),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(424),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(428),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(371),
    [anon_sym_get] = ACTIONS(371),
    [anon_sym_set] = ACTIONS(371),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(260)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(507),
    [sym_expression] = STATE(678),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1815),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1815),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(507),
    [sym_subscript_expression] = STATE(507),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1139),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1815),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1763),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(369),
    [anon_sym_export] = ACTIONS(371),
    [anon_sym_LBRACE] = ACTIONS(373),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(371),
    [anon_sym_await] = ACTIONS(377),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(382),
    [anon_sym_LBRACK] = ACTIONS(654),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(398),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(404),
    [anon_sym_PLUS] = ACTIONS(412),
    [anon_sym_DASH] = ACTIONS(412),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(414),
    [anon_sym_TILDE] = ACTIONS(414),
    [anon_sym_typeof] = ACTIONS(412),
    [anon_sym_void] = ACTIONS(412),
    [anon_sym_delete] = ACTIONS(412),
    [anon_sym_PLUS_PLUS] = ACTIONS(660),
    [anon_sym_DASH_DASH] = ACTIONS(660),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(424),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(428),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(371),
    [anon_sym_get] = ACTIONS(371),
    [anon_sym_set] = ACTIONS(371),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(261)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(507),
    [sym_expression] = STATE(679),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1815),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1815),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(507),
    [sym_subscript_expression] = STATE(507),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1139),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1815),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1763),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(369),
    [anon_sym_export] = ACTIONS(371),
    [anon_sym_LBRACE] = ACTIONS(373),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(371),
    [anon_sym_await] = ACTIONS(377),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(382),
    [anon_sym_LBRACK] = ACTIONS(654),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(398),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(404),
    [anon_sym_PLUS] = ACTIONS(412),
    [anon_sym_DASH] = ACTIONS(412),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(414),
    [anon_sym_TILDE] = ACTIONS(414),
    [anon_sym_typeof] = ACTIONS(412),
    [anon_sym_void] = ACTIONS(412),
    [anon_sym_delete] = ACTIONS(412),
    [anon_sym_PLUS_PLUS] = ACTIONS(660),
    [anon_sym_DASH_DASH] = ACTIONS(660),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(424),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(428),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(371),
    [anon_sym_get] = ACTIONS(371),
    [anon_sym_set] = ACTIONS(371),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(262)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(507),
    [sym_expression] = STATE(680),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1815),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1815),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(507),
    [sym_subscript_expression] = STATE(507),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1139),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1815),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1763),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(369),
    [anon_sym_export] = ACTIONS(371),
    [anon_sym_LBRACE] = ACTIONS(373),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(371),
    [anon_sym_await] = ACTIONS(377),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(382),
    [anon_sym_LBRACK] = ACTIONS(654),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(398),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(404),
    [anon_sym_PLUS] = ACTIONS(412),
    [anon_sym_DASH] = ACTIONS(412),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(414),
    [anon_sym_TILDE] = ACTIONS(414),
    [anon_sym_typeof] = ACTIONS(412),
    [anon_sym_void] = ACTIONS(412),
    [anon_sym_delete] = ACTIONS(412),
    [anon_sym_PLUS_PLUS] = ACTIONS(660),
    [anon_sym_DASH_DASH] = ACTIONS(660),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(424),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(428),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(371),
    [anon_sym_get] = ACTIONS(371),
    [anon_sym_set] = ACTIONS(371),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(263)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(507),
    [sym_expression] = STATE(667),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1815),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1815),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(507),
    [sym_subscript_expression] = STATE(507),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1139),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1815),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1763),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(369),
    [anon_sym_export] = ACTIONS(371),
    [anon_sym_LBRACE] = ACTIONS(373),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(371),
    [anon_sym_await] = ACTIONS(377),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(382),
    [anon_sym_LBRACK] = ACTIONS(654),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(398),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(404),
    [anon_sym_PLUS] = ACTIONS(412),
    [anon_sym_DASH] = ACTIONS(412),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(414),
    [anon_sym_TILDE] = ACTIONS(414),
    [anon_sym_typeof] = ACTIONS(412),
    [anon_sym_void] = ACTIONS(412),
    [anon_sym_delete] = ACTIONS(412),
    [anon_sym_PLUS_PLUS] = ACTIONS(660),
    [anon_sym_DASH_DASH] = ACTIONS(660),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(424),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(428),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(371),
    [anon_sym_get] = ACTIONS(371),
    [anon_sym_set] = ACTIONS(371),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(264)] = {
    [sym_import] = STATE(1326),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(791),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1790),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1790),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_function_expression] = STATE(740),
    [sym_generator_function] = STATE(740),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1790),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [aux_sym_export_statement_repeat1] = STATE(1381),
    [sym_identifier] = ACTIONS(161),
    [anon_sym_export] = ACTIONS(163),
    [anon_sym_LBRACE] = ACTIONS(167),
    [anon_sym_import] = ACTIONS(173),
    [anon_sym_let] = ACTIONS(163),
    [anon_sym_await] = ACTIONS(177),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_yield] = ACTIONS(57),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(194),
    [anon_sym_async] = ACTIONS(196),
    [anon_sym_function] = ACTIONS(198),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(75),
    [anon_sym_DASH] = ACTIONS(75),
    [anon_sym_SLASH] = ACTIONS(77),
    [anon_sym_BANG] = ACTIONS(79),
    [anon_sym_TILDE] = ACTIONS(79),
    [anon_sym_typeof] = ACTIONS(75),
    [anon_sym_void] = ACTIONS(75),
    [anon_sym_delete] = ACTIONS(75),
    [anon_sym_PLUS_PLUS] = ACTIONS(81),
    [anon_sym_DASH_DASH] = ACTIONS(81),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(87),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(163),
    [anon_sym_get] = ACTIONS(163),
    [anon_sym_set] = ACTIONS(163),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(265)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(507),
    [sym_expression] = STATE(825),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1324),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1324),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(562),
    [sym_subscript_expression] = STATE(562),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1139),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1324),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1763),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(1051),
    [anon_sym_export] = ACTIONS(1053),
    [anon_sym_LBRACE] = ACTIONS(434),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(1053),
    [anon_sym_await] = ACTIONS(1055),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(382),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(1057),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(404),
    [anon_sym_PLUS] = ACTIONS(412),
    [anon_sym_DASH] = ACTIONS(412),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(414),
    [anon_sym_TILDE] = ACTIONS(414),
    [anon_sym_typeof] = ACTIONS(412),
    [anon_sym_void] = ACTIONS(412),
    [anon_sym_delete] = ACTIONS(412),
    [anon_sym_PLUS_PLUS] = ACTIONS(660),
    [anon_sym_DASH_DASH] = ACTIONS(660),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(424),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(1059),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(1053),
    [anon_sym_get] = ACTIONS(1053),
    [anon_sym_set] = ACTIONS(1053),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(266)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(579),
    [sym_expression] = STATE(978),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1857),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1857),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(579),
    [sym_subscript_expression] = STATE(579),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1141),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1857),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1779),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(582),
    [anon_sym_export] = ACTIONS(584),
    [anon_sym_LBRACE] = ACTIONS(434),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(584),
    [anon_sym_await] = ACTIONS(586),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(588),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(592),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(594),
    [anon_sym_PLUS] = ACTIONS(602),
    [anon_sym_DASH] = ACTIONS(602),
    [anon_sym_SLASH] = ACTIONS(678),
    [anon_sym_BANG] = ACTIONS(604),
    [anon_sym_TILDE] = ACTIONS(604),
    [anon_sym_typeof] = ACTIONS(602),
    [anon_sym_void] = ACTIONS(602),
    [anon_sym_delete] = ACTIONS(602),
    [anon_sym_PLUS_PLUS] = ACTIONS(680),
    [anon_sym_DASH_DASH] = ACTIONS(680),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(609),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(611),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(584),
    [anon_sym_get] = ACTIONS(584),
    [anon_sym_set] = ACTIONS(584),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(267)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(507),
    [sym_expression] = STATE(655),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1815),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1815),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(507),
    [sym_subscript_expression] = STATE(507),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1139),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1815),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1763),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(369),
    [anon_sym_export] = ACTIONS(371),
    [anon_sym_LBRACE] = ACTIONS(373),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(371),
    [anon_sym_await] = ACTIONS(377),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(382),
    [anon_sym_LBRACK] = ACTIONS(654),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(398),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(404),
    [anon_sym_PLUS] = ACTIONS(412),
    [anon_sym_DASH] = ACTIONS(412),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(414),
    [anon_sym_TILDE] = ACTIONS(414),
    [anon_sym_typeof] = ACTIONS(412),
    [anon_sym_void] = ACTIONS(412),
    [anon_sym_delete] = ACTIONS(412),
    [anon_sym_PLUS_PLUS] = ACTIONS(660),
    [anon_sym_DASH_DASH] = ACTIONS(660),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(424),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(428),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(371),
    [anon_sym_get] = ACTIONS(371),
    [anon_sym_set] = ACTIONS(371),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(268)] = {
    [sym_import] = STATE(1326),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(814),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1790),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1790),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_function_expression] = STATE(740),
    [sym_generator_function] = STATE(740),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1790),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [aux_sym_export_statement_repeat1] = STATE(1381),
    [sym_identifier] = ACTIONS(161),
    [anon_sym_export] = ACTIONS(163),
    [anon_sym_LBRACE] = ACTIONS(167),
    [anon_sym_import] = ACTIONS(173),
    [anon_sym_let] = ACTIONS(163),
    [anon_sym_await] = ACTIONS(177),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_yield] = ACTIONS(57),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(194),
    [anon_sym_async] = ACTIONS(196),
    [anon_sym_function] = ACTIONS(198),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(75),
    [anon_sym_DASH] = ACTIONS(75),
    [anon_sym_SLASH] = ACTIONS(77),
    [anon_sym_BANG] = ACTIONS(79),
    [anon_sym_TILDE] = ACTIONS(79),
    [anon_sym_typeof] = ACTIONS(75),
    [anon_sym_void] = ACTIONS(75),
    [anon_sym_delete] = ACTIONS(75),
    [anon_sym_PLUS_PLUS] = ACTIONS(81),
    [anon_sym_DASH_DASH] = ACTIONS(81),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(87),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(163),
    [anon_sym_get] = ACTIONS(163),
    [anon_sym_set] = ACTIONS(163),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(269)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(507),
    [sym_expression] = STATE(975),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1815),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1815),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(507),
    [sym_subscript_expression] = STATE(507),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1139),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1815),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1763),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(369),
    [anon_sym_export] = ACTIONS(371),
    [anon_sym_LBRACE] = ACTIONS(373),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(371),
    [anon_sym_await] = ACTIONS(377),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(382),
    [anon_sym_LBRACK] = ACTIONS(654),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(398),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(404),
    [anon_sym_PLUS] = ACTIONS(412),
    [anon_sym_DASH] = ACTIONS(412),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(414),
    [anon_sym_TILDE] = ACTIONS(414),
    [anon_sym_typeof] = ACTIONS(412),
    [anon_sym_void] = ACTIONS(412),
    [anon_sym_delete] = ACTIONS(412),
    [anon_sym_PLUS_PLUS] = ACTIONS(660),
    [anon_sym_DASH_DASH] = ACTIONS(660),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(424),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(428),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(371),
    [anon_sym_get] = ACTIONS(371),
    [anon_sym_set] = ACTIONS(371),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(270)] = {
    [sym_import] = STATE(1326),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(813),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1790),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1790),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_function_expression] = STATE(740),
    [sym_generator_function] = STATE(740),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1790),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [aux_sym_export_statement_repeat1] = STATE(1381),
    [sym_identifier] = ACTIONS(161),
    [anon_sym_export] = ACTIONS(163),
    [anon_sym_LBRACE] = ACTIONS(167),
    [anon_sym_import] = ACTIONS(173),
    [anon_sym_let] = ACTIONS(163),
    [anon_sym_await] = ACTIONS(177),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_yield] = ACTIONS(57),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(194),
    [anon_sym_async] = ACTIONS(196),
    [anon_sym_function] = ACTIONS(198),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(75),
    [anon_sym_DASH] = ACTIONS(75),
    [anon_sym_SLASH] = ACTIONS(77),
    [anon_sym_BANG] = ACTIONS(79),
    [anon_sym_TILDE] = ACTIONS(79),
    [anon_sym_typeof] = ACTIONS(75),
    [anon_sym_void] = ACTIONS(75),
    [anon_sym_delete] = ACTIONS(75),
    [anon_sym_PLUS_PLUS] = ACTIONS(81),
    [anon_sym_DASH_DASH] = ACTIONS(81),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(87),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(163),
    [anon_sym_get] = ACTIONS(163),
    [anon_sym_set] = ACTIONS(163),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(271)] = {
    [sym_import] = STATE(1326),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(821),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1790),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1790),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_function_expression] = STATE(740),
    [sym_generator_function] = STATE(740),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1790),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [aux_sym_export_statement_repeat1] = STATE(1381),
    [sym_identifier] = ACTIONS(161),
    [anon_sym_export] = ACTIONS(163),
    [anon_sym_LBRACE] = ACTIONS(167),
    [anon_sym_import] = ACTIONS(173),
    [anon_sym_let] = ACTIONS(163),
    [anon_sym_await] = ACTIONS(177),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_yield] = ACTIONS(57),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(194),
    [anon_sym_async] = ACTIONS(196),
    [anon_sym_function] = ACTIONS(198),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(75),
    [anon_sym_DASH] = ACTIONS(75),
    [anon_sym_SLASH] = ACTIONS(77),
    [anon_sym_BANG] = ACTIONS(79),
    [anon_sym_TILDE] = ACTIONS(79),
    [anon_sym_typeof] = ACTIONS(75),
    [anon_sym_void] = ACTIONS(75),
    [anon_sym_delete] = ACTIONS(75),
    [anon_sym_PLUS_PLUS] = ACTIONS(81),
    [anon_sym_DASH_DASH] = ACTIONS(81),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(87),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(163),
    [anon_sym_get] = ACTIONS(163),
    [anon_sym_set] = ACTIONS(163),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(272)] = {
    [sym_import] = STATE(1326),
    [sym_parenthesized_expression] = STATE(560),
    [sym_expression] = STATE(826),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1802),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1802),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_function_expression] = STATE(740),
    [sym_generator_function] = STATE(740),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(560),
    [sym_subscript_expression] = STATE(560),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1143),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1802),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1803),
    [aux_sym_export_statement_repeat1] = STATE(1381),
    [sym_identifier] = ACTIONS(520),
    [anon_sym_export] = ACTIONS(522),
    [anon_sym_LBRACE] = ACTIONS(167),
    [anon_sym_import] = ACTIONS(173),
    [anon_sym_let] = ACTIONS(522),
    [anon_sym_await] = ACTIONS(524),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_yield] = ACTIONS(526),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(194),
    [anon_sym_async] = ACTIONS(530),
    [anon_sym_function] = ACTIONS(198),
    [anon_sym_new] = ACTIONS(532),
    [anon_sym_PLUS] = ACTIONS(540),
    [anon_sym_DASH] = ACTIONS(540),
    [anon_sym_SLASH] = ACTIONS(670),
    [anon_sym_BANG] = ACTIONS(542),
    [anon_sym_TILDE] = ACTIONS(542),
    [anon_sym_typeof] = ACTIONS(540),
    [anon_sym_void] = ACTIONS(540),
    [anon_sym_delete] = ACTIONS(540),
    [anon_sym_PLUS_PLUS] = ACTIONS(672),
    [anon_sym_DASH_DASH] = ACTIONS(672),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(547),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(549),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(522),
    [anon_sym_get] = ACTIONS(522),
    [anon_sym_set] = ACTIONS(522),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(273)] = {
    [sym_import] = STATE(1326),
    [sym_parenthesized_expression] = STATE(552),
    [sym_expression] = STATE(855),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1790),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1790),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_function_expression] = STATE(740),
    [sym_generator_function] = STATE(740),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(552),
    [sym_subscript_expression] = STATE(552),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1140),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1790),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1849),
    [aux_sym_export_statement_repeat1] = STATE(1381),
    [sym_identifier] = ACTIONS(161),
    [anon_sym_export] = ACTIONS(163),
    [anon_sym_LBRACE] = ACTIONS(167),
    [anon_sym_import] = ACTIONS(173),
    [anon_sym_let] = ACTIONS(163),
    [anon_sym_await] = ACTIONS(177),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_yield] = ACTIONS(57),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(194),
    [anon_sym_async] = ACTIONS(196),
    [anon_sym_function] = ACTIONS(198),
    [anon_sym_new] = ACTIONS(73),
    [anon_sym_PLUS] = ACTIONS(75),
    [anon_sym_DASH] = ACTIONS(75),
    [anon_sym_SLASH] = ACTIONS(77),
    [anon_sym_BANG] = ACTIONS(79),
    [anon_sym_TILDE] = ACTIONS(79),
    [anon_sym_typeof] = ACTIONS(75),
    [anon_sym_void] = ACTIONS(75),
    [anon_sym_delete] = ACTIONS(75),
    [anon_sym_PLUS_PLUS] = ACTIONS(81),
    [anon_sym_DASH_DASH] = ACTIONS(81),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(87),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(91),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(163),
    [anon_sym_get] = ACTIONS(163),
    [anon_sym_set] = ACTIONS(163),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(274)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(578),
    [sym_expression] = STATE(588),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1776),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1776),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(578),
    [sym_subscript_expression] = STATE(578),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1142),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1776),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1778),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(613),
    [anon_sym_export] = ACTIONS(615),
    [anon_sym_LBRACE] = ACTIONS(434),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(615),
    [anon_sym_await] = ACTIONS(620),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(622),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(626),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(630),
    [anon_sym_PLUS] = ACTIONS(635),
    [anon_sym_DASH] = ACTIONS(635),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(637),
    [anon_sym_TILDE] = ACTIONS(637),
    [anon_sym_typeof] = ACTIONS(635),
    [anon_sym_void] = ACTIONS(635),
    [anon_sym_delete] = ACTIONS(635),
    [anon_sym_PLUS_PLUS] = ACTIONS(688),
    [anon_sym_DASH_DASH] = ACTIONS(688),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(642),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(644),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(615),
    [anon_sym_get] = ACTIONS(615),
    [anon_sym_set] = ACTIONS(615),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(275)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(578),
    [sym_expression] = STATE(591),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1776),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1776),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(578),
    [sym_subscript_expression] = STATE(578),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1142),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1776),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1778),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(613),
    [anon_sym_export] = ACTIONS(615),
    [anon_sym_LBRACE] = ACTIONS(434),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(615),
    [anon_sym_await] = ACTIONS(620),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(622),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(626),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(630),
    [anon_sym_PLUS] = ACTIONS(635),
    [anon_sym_DASH] = ACTIONS(635),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(637),
    [anon_sym_TILDE] = ACTIONS(637),
    [anon_sym_typeof] = ACTIONS(635),
    [anon_sym_void] = ACTIONS(635),
    [anon_sym_delete] = ACTIONS(635),
    [anon_sym_PLUS_PLUS] = ACTIONS(688),
    [anon_sym_DASH_DASH] = ACTIONS(688),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(642),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(644),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(615),
    [anon_sym_get] = ACTIONS(615),
    [anon_sym_set] = ACTIONS(615),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(276)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(578),
    [sym_expression] = STATE(890),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1776),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1776),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(578),
    [sym_subscript_expression] = STATE(578),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1142),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1776),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1778),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(613),
    [anon_sym_export] = ACTIONS(615),
    [anon_sym_LBRACE] = ACTIONS(434),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(615),
    [anon_sym_await] = ACTIONS(620),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(622),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(626),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(630),
    [anon_sym_PLUS] = ACTIONS(635),
    [anon_sym_DASH] = ACTIONS(635),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(637),
    [anon_sym_TILDE] = ACTIONS(637),
    [anon_sym_typeof] = ACTIONS(635),
    [anon_sym_void] = ACTIONS(635),
    [anon_sym_delete] = ACTIONS(635),
    [anon_sym_PLUS_PLUS] = ACTIONS(688),
    [anon_sym_DASH_DASH] = ACTIONS(688),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(642),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(644),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(615),
    [anon_sym_get] = ACTIONS(615),
    [anon_sym_set] = ACTIONS(615),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(277)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(498),
    [sym_expression] = STATE(974),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1762),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1762),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(498),
    [sym_subscript_expression] = STATE(498),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1138),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1762),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1816),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(430),
    [anon_sym_export] = ACTIONS(432),
    [anon_sym_LBRACE] = ACTIONS(434),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(432),
    [anon_sym_await] = ACTIONS(436),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(438),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(445),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(447),
    [anon_sym_PLUS] = ACTIONS(452),
    [anon_sym_DASH] = ACTIONS(452),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(454),
    [anon_sym_TILDE] = ACTIONS(454),
    [anon_sym_typeof] = ACTIONS(452),
    [anon_sym_void] = ACTIONS(452),
    [anon_sym_delete] = ACTIONS(452),
    [anon_sym_PLUS_PLUS] = ACTIONS(684),
    [anon_sym_DASH_DASH] = ACTIONS(684),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(459),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(461),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(432),
    [anon_sym_get] = ACTIONS(432),
    [anon_sym_set] = ACTIONS(432),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(278)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(578),
    [sym_expression] = STATE(893),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1776),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1776),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(578),
    [sym_subscript_expression] = STATE(578),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1142),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1776),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1778),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(613),
    [anon_sym_export] = ACTIONS(615),
    [anon_sym_LBRACE] = ACTIONS(434),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(615),
    [anon_sym_await] = ACTIONS(620),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(622),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(626),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(630),
    [anon_sym_PLUS] = ACTIONS(635),
    [anon_sym_DASH] = ACTIONS(635),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(637),
    [anon_sym_TILDE] = ACTIONS(637),
    [anon_sym_typeof] = ACTIONS(635),
    [anon_sym_void] = ACTIONS(635),
    [anon_sym_delete] = ACTIONS(635),
    [anon_sym_PLUS_PLUS] = ACTIONS(688),
    [anon_sym_DASH_DASH] = ACTIONS(688),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(642),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(644),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(615),
    [anon_sym_get] = ACTIONS(615),
    [anon_sym_set] = ACTIONS(615),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(279)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(578),
    [sym_expression] = STATE(895),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1776),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1776),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(578),
    [sym_subscript_expression] = STATE(578),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1142),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1776),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1778),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(613),
    [anon_sym_export] = ACTIONS(615),
    [anon_sym_LBRACE] = ACTIONS(434),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(615),
    [anon_sym_await] = ACTIONS(620),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(622),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(626),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(630),
    [anon_sym_PLUS] = ACTIONS(635),
    [anon_sym_DASH] = ACTIONS(635),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(637),
    [anon_sym_TILDE] = ACTIONS(637),
    [anon_sym_typeof] = ACTIONS(635),
    [anon_sym_void] = ACTIONS(635),
    [anon_sym_delete] = ACTIONS(635),
    [anon_sym_PLUS_PLUS] = ACTIONS(688),
    [anon_sym_DASH_DASH] = ACTIONS(688),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(642),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(644),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(615),
    [anon_sym_get] = ACTIONS(615),
    [anon_sym_set] = ACTIONS(615),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(280)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(578),
    [sym_expression] = STATE(896),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1776),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1776),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(578),
    [sym_subscript_expression] = STATE(578),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1142),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1776),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1778),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(613),
    [anon_sym_export] = ACTIONS(615),
    [anon_sym_LBRACE] = ACTIONS(434),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(615),
    [anon_sym_await] = ACTIONS(620),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(622),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(626),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(630),
    [anon_sym_PLUS] = ACTIONS(635),
    [anon_sym_DASH] = ACTIONS(635),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(637),
    [anon_sym_TILDE] = ACTIONS(637),
    [anon_sym_typeof] = ACTIONS(635),
    [anon_sym_void] = ACTIONS(635),
    [anon_sym_delete] = ACTIONS(635),
    [anon_sym_PLUS_PLUS] = ACTIONS(688),
    [anon_sym_DASH_DASH] = ACTIONS(688),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(642),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(644),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(615),
    [anon_sym_get] = ACTIONS(615),
    [anon_sym_set] = ACTIONS(615),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(281)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(578),
    [sym_expression] = STATE(897),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1776),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1776),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(578),
    [sym_subscript_expression] = STATE(578),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1142),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1776),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1778),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(613),
    [anon_sym_export] = ACTIONS(615),
    [anon_sym_LBRACE] = ACTIONS(434),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(615),
    [anon_sym_await] = ACTIONS(620),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(622),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(626),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(630),
    [anon_sym_PLUS] = ACTIONS(635),
    [anon_sym_DASH] = ACTIONS(635),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(637),
    [anon_sym_TILDE] = ACTIONS(637),
    [anon_sym_typeof] = ACTIONS(635),
    [anon_sym_void] = ACTIONS(635),
    [anon_sym_delete] = ACTIONS(635),
    [anon_sym_PLUS_PLUS] = ACTIONS(688),
    [anon_sym_DASH_DASH] = ACTIONS(688),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(642),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(644),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(615),
    [anon_sym_get] = ACTIONS(615),
    [anon_sym_set] = ACTIONS(615),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(282)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(578),
    [sym_expression] = STATE(898),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1776),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1776),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(578),
    [sym_subscript_expression] = STATE(578),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1142),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1776),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1778),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(613),
    [anon_sym_export] = ACTIONS(615),
    [anon_sym_LBRACE] = ACTIONS(434),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(615),
    [anon_sym_await] = ACTIONS(620),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(622),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(626),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(630),
    [anon_sym_PLUS] = ACTIONS(635),
    [anon_sym_DASH] = ACTIONS(635),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(637),
    [anon_sym_TILDE] = ACTIONS(637),
    [anon_sym_typeof] = ACTIONS(635),
    [anon_sym_void] = ACTIONS(635),
    [anon_sym_delete] = ACTIONS(635),
    [anon_sym_PLUS_PLUS] = ACTIONS(688),
    [anon_sym_DASH_DASH] = ACTIONS(688),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(642),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(644),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(615),
    [anon_sym_get] = ACTIONS(615),
    [anon_sym_set] = ACTIONS(615),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(283)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(578),
    [sym_expression] = STATE(899),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1776),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1776),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(578),
    [sym_subscript_expression] = STATE(578),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1142),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1776),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1778),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(613),
    [anon_sym_export] = ACTIONS(615),
    [anon_sym_LBRACE] = ACTIONS(434),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(615),
    [anon_sym_await] = ACTIONS(620),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(622),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(626),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(630),
    [anon_sym_PLUS] = ACTIONS(635),
    [anon_sym_DASH] = ACTIONS(635),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(637),
    [anon_sym_TILDE] = ACTIONS(637),
    [anon_sym_typeof] = ACTIONS(635),
    [anon_sym_void] = ACTIONS(635),
    [anon_sym_delete] = ACTIONS(635),
    [anon_sym_PLUS_PLUS] = ACTIONS(688),
    [anon_sym_DASH_DASH] = ACTIONS(688),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(642),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(644),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(615),
    [anon_sym_get] = ACTIONS(615),
    [anon_sym_set] = ACTIONS(615),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(284)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(578),
    [sym_expression] = STATE(900),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1776),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1776),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(578),
    [sym_subscript_expression] = STATE(578),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1142),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1776),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1778),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(613),
    [anon_sym_export] = ACTIONS(615),
    [anon_sym_LBRACE] = ACTIONS(434),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(615),
    [anon_sym_await] = ACTIONS(620),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(622),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(626),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(630),
    [anon_sym_PLUS] = ACTIONS(635),
    [anon_sym_DASH] = ACTIONS(635),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(637),
    [anon_sym_TILDE] = ACTIONS(637),
    [anon_sym_typeof] = ACTIONS(635),
    [anon_sym_void] = ACTIONS(635),
    [anon_sym_delete] = ACTIONS(635),
    [anon_sym_PLUS_PLUS] = ACTIONS(688),
    [anon_sym_DASH_DASH] = ACTIONS(688),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(642),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(644),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(615),
    [anon_sym_get] = ACTIONS(615),
    [anon_sym_set] = ACTIONS(615),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(285)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(578),
    [sym_expression] = STATE(901),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1776),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1776),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(578),
    [sym_subscript_expression] = STATE(578),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1142),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1776),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1778),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(613),
    [anon_sym_export] = ACTIONS(615),
    [anon_sym_LBRACE] = ACTIONS(434),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(615),
    [anon_sym_await] = ACTIONS(620),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(622),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(626),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(630),
    [anon_sym_PLUS] = ACTIONS(635),
    [anon_sym_DASH] = ACTIONS(635),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(637),
    [anon_sym_TILDE] = ACTIONS(637),
    [anon_sym_typeof] = ACTIONS(635),
    [anon_sym_void] = ACTIONS(635),
    [anon_sym_delete] = ACTIONS(635),
    [anon_sym_PLUS_PLUS] = ACTIONS(688),
    [anon_sym_DASH_DASH] = ACTIONS(688),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(642),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(644),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(615),
    [anon_sym_get] = ACTIONS(615),
    [anon_sym_set] = ACTIONS(615),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(286)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(578),
    [sym_expression] = STATE(902),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1776),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1776),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(578),
    [sym_subscript_expression] = STATE(578),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1142),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1776),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1778),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(613),
    [anon_sym_export] = ACTIONS(615),
    [anon_sym_LBRACE] = ACTIONS(434),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(615),
    [anon_sym_await] = ACTIONS(620),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(622),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(626),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(630),
    [anon_sym_PLUS] = ACTIONS(635),
    [anon_sym_DASH] = ACTIONS(635),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(637),
    [anon_sym_TILDE] = ACTIONS(637),
    [anon_sym_typeof] = ACTIONS(635),
    [anon_sym_void] = ACTIONS(635),
    [anon_sym_delete] = ACTIONS(635),
    [anon_sym_PLUS_PLUS] = ACTIONS(688),
    [anon_sym_DASH_DASH] = ACTIONS(688),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(642),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(644),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(615),
    [anon_sym_get] = ACTIONS(615),
    [anon_sym_set] = ACTIONS(615),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(287)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(578),
    [sym_expression] = STATE(903),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1776),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1776),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(578),
    [sym_subscript_expression] = STATE(578),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1142),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1776),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1778),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(613),
    [anon_sym_export] = ACTIONS(615),
    [anon_sym_LBRACE] = ACTIONS(434),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(615),
    [anon_sym_await] = ACTIONS(620),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(622),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(626),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(630),
    [anon_sym_PLUS] = ACTIONS(635),
    [anon_sym_DASH] = ACTIONS(635),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(637),
    [anon_sym_TILDE] = ACTIONS(637),
    [anon_sym_typeof] = ACTIONS(635),
    [anon_sym_void] = ACTIONS(635),
    [anon_sym_delete] = ACTIONS(635),
    [anon_sym_PLUS_PLUS] = ACTIONS(688),
    [anon_sym_DASH_DASH] = ACTIONS(688),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(642),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(644),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(615),
    [anon_sym_get] = ACTIONS(615),
    [anon_sym_set] = ACTIONS(615),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(288)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(578),
    [sym_expression] = STATE(904),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1776),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1776),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(578),
    [sym_subscript_expression] = STATE(578),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1142),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1776),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1778),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(613),
    [anon_sym_export] = ACTIONS(615),
    [anon_sym_LBRACE] = ACTIONS(434),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(615),
    [anon_sym_await] = ACTIONS(620),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(622),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(626),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(630),
    [anon_sym_PLUS] = ACTIONS(635),
    [anon_sym_DASH] = ACTIONS(635),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(637),
    [anon_sym_TILDE] = ACTIONS(637),
    [anon_sym_typeof] = ACTIONS(635),
    [anon_sym_void] = ACTIONS(635),
    [anon_sym_delete] = ACTIONS(635),
    [anon_sym_PLUS_PLUS] = ACTIONS(688),
    [anon_sym_DASH_DASH] = ACTIONS(688),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(642),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(644),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(615),
    [anon_sym_get] = ACTIONS(615),
    [anon_sym_set] = ACTIONS(615),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(289)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(578),
    [sym_expression] = STATE(905),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1776),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1776),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(578),
    [sym_subscript_expression] = STATE(578),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1142),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1776),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1778),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(613),
    [anon_sym_export] = ACTIONS(615),
    [anon_sym_LBRACE] = ACTIONS(434),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(615),
    [anon_sym_await] = ACTIONS(620),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(622),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(626),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(630),
    [anon_sym_PLUS] = ACTIONS(635),
    [anon_sym_DASH] = ACTIONS(635),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(637),
    [anon_sym_TILDE] = ACTIONS(637),
    [anon_sym_typeof] = ACTIONS(635),
    [anon_sym_void] = ACTIONS(635),
    [anon_sym_delete] = ACTIONS(635),
    [anon_sym_PLUS_PLUS] = ACTIONS(688),
    [anon_sym_DASH_DASH] = ACTIONS(688),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(642),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(644),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(615),
    [anon_sym_get] = ACTIONS(615),
    [anon_sym_set] = ACTIONS(615),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(290)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(578),
    [sym_expression] = STATE(906),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1776),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1776),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(578),
    [sym_subscript_expression] = STATE(578),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1142),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1776),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1778),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(613),
    [anon_sym_export] = ACTIONS(615),
    [anon_sym_LBRACE] = ACTIONS(434),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(615),
    [anon_sym_await] = ACTIONS(620),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(622),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(626),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(630),
    [anon_sym_PLUS] = ACTIONS(635),
    [anon_sym_DASH] = ACTIONS(635),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(637),
    [anon_sym_TILDE] = ACTIONS(637),
    [anon_sym_typeof] = ACTIONS(635),
    [anon_sym_void] = ACTIONS(635),
    [anon_sym_delete] = ACTIONS(635),
    [anon_sym_PLUS_PLUS] = ACTIONS(688),
    [anon_sym_DASH_DASH] = ACTIONS(688),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(642),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(644),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(615),
    [anon_sym_get] = ACTIONS(615),
    [anon_sym_set] = ACTIONS(615),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(291)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(578),
    [sym_expression] = STATE(907),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1776),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1776),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(578),
    [sym_subscript_expression] = STATE(578),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1142),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1776),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1778),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(613),
    [anon_sym_export] = ACTIONS(615),
    [anon_sym_LBRACE] = ACTIONS(434),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(615),
    [anon_sym_await] = ACTIONS(620),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(622),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(626),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(630),
    [anon_sym_PLUS] = ACTIONS(635),
    [anon_sym_DASH] = ACTIONS(635),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(637),
    [anon_sym_TILDE] = ACTIONS(637),
    [anon_sym_typeof] = ACTIONS(635),
    [anon_sym_void] = ACTIONS(635),
    [anon_sym_delete] = ACTIONS(635),
    [anon_sym_PLUS_PLUS] = ACTIONS(688),
    [anon_sym_DASH_DASH] = ACTIONS(688),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(642),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(644),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(615),
    [anon_sym_get] = ACTIONS(615),
    [anon_sym_set] = ACTIONS(615),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(292)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(578),
    [sym_expression] = STATE(912),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1776),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1776),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(578),
    [sym_subscript_expression] = STATE(578),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1142),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1776),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1778),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(613),
    [anon_sym_export] = ACTIONS(615),
    [anon_sym_LBRACE] = ACTIONS(434),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(615),
    [anon_sym_await] = ACTIONS(620),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(622),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(626),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(630),
    [anon_sym_PLUS] = ACTIONS(635),
    [anon_sym_DASH] = ACTIONS(635),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(637),
    [anon_sym_TILDE] = ACTIONS(637),
    [anon_sym_typeof] = ACTIONS(635),
    [anon_sym_void] = ACTIONS(635),
    [anon_sym_delete] = ACTIONS(635),
    [anon_sym_PLUS_PLUS] = ACTIONS(688),
    [anon_sym_DASH_DASH] = ACTIONS(688),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(642),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(644),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(615),
    [anon_sym_get] = ACTIONS(615),
    [anon_sym_set] = ACTIONS(615),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(293)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(507),
    [sym_expression] = STATE(824),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1815),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1815),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(507),
    [sym_subscript_expression] = STATE(507),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1139),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1815),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1763),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(369),
    [anon_sym_export] = ACTIONS(371),
    [anon_sym_LBRACE] = ACTIONS(373),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(371),
    [anon_sym_await] = ACTIONS(377),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(382),
    [anon_sym_LBRACK] = ACTIONS(654),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(398),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(404),
    [anon_sym_PLUS] = ACTIONS(412),
    [anon_sym_DASH] = ACTIONS(412),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(414),
    [anon_sym_TILDE] = ACTIONS(414),
    [anon_sym_typeof] = ACTIONS(412),
    [anon_sym_void] = ACTIONS(412),
    [anon_sym_delete] = ACTIONS(412),
    [anon_sym_PLUS_PLUS] = ACTIONS(660),
    [anon_sym_DASH_DASH] = ACTIONS(660),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(424),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(428),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(371),
    [anon_sym_get] = ACTIONS(371),
    [anon_sym_set] = ACTIONS(371),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(294)] = {
    [sym_import] = STATE(1326),
    [sym_parenthesized_expression] = STATE(560),
    [sym_expression] = STATE(691),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1802),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1802),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_function_expression] = STATE(740),
    [sym_generator_function] = STATE(740),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(560),
    [sym_subscript_expression] = STATE(560),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1143),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1802),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1803),
    [aux_sym_export_statement_repeat1] = STATE(1381),
    [sym_identifier] = ACTIONS(520),
    [anon_sym_export] = ACTIONS(522),
    [anon_sym_LBRACE] = ACTIONS(167),
    [anon_sym_import] = ACTIONS(173),
    [anon_sym_let] = ACTIONS(522),
    [anon_sym_await] = ACTIONS(524),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_yield] = ACTIONS(526),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(194),
    [anon_sym_async] = ACTIONS(530),
    [anon_sym_function] = ACTIONS(198),
    [anon_sym_new] = ACTIONS(532),
    [anon_sym_PLUS] = ACTIONS(540),
    [anon_sym_DASH] = ACTIONS(540),
    [anon_sym_SLASH] = ACTIONS(670),
    [anon_sym_BANG] = ACTIONS(542),
    [anon_sym_TILDE] = ACTIONS(542),
    [anon_sym_typeof] = ACTIONS(540),
    [anon_sym_void] = ACTIONS(540),
    [anon_sym_delete] = ACTIONS(540),
    [anon_sym_PLUS_PLUS] = ACTIONS(672),
    [anon_sym_DASH_DASH] = ACTIONS(672),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(547),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(549),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(522),
    [anon_sym_get] = ACTIONS(522),
    [anon_sym_set] = ACTIONS(522),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(295)] = {
    [sym_import] = STATE(1326),
    [sym_parenthesized_expression] = STATE(560),
    [sym_expression] = STATE(808),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1802),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1802),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_function_expression] = STATE(740),
    [sym_generator_function] = STATE(740),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(560),
    [sym_subscript_expression] = STATE(560),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1143),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1802),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1803),
    [aux_sym_export_statement_repeat1] = STATE(1381),
    [sym_identifier] = ACTIONS(520),
    [anon_sym_export] = ACTIONS(522),
    [anon_sym_LBRACE] = ACTIONS(167),
    [anon_sym_import] = ACTIONS(173),
    [anon_sym_let] = ACTIONS(522),
    [anon_sym_await] = ACTIONS(524),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_yield] = ACTIONS(526),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(194),
    [anon_sym_async] = ACTIONS(530),
    [anon_sym_function] = ACTIONS(198),
    [anon_sym_new] = ACTIONS(532),
    [anon_sym_PLUS] = ACTIONS(540),
    [anon_sym_DASH] = ACTIONS(540),
    [anon_sym_SLASH] = ACTIONS(670),
    [anon_sym_BANG] = ACTIONS(542),
    [anon_sym_TILDE] = ACTIONS(542),
    [anon_sym_typeof] = ACTIONS(540),
    [anon_sym_void] = ACTIONS(540),
    [anon_sym_delete] = ACTIONS(540),
    [anon_sym_PLUS_PLUS] = ACTIONS(672),
    [anon_sym_DASH_DASH] = ACTIONS(672),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(547),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(549),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(522),
    [anon_sym_get] = ACTIONS(522),
    [anon_sym_set] = ACTIONS(522),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(296)] = {
    [sym_import] = STATE(1326),
    [sym_parenthesized_expression] = STATE(560),
    [sym_expression] = STATE(698),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1802),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1802),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_function_expression] = STATE(740),
    [sym_generator_function] = STATE(740),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(560),
    [sym_subscript_expression] = STATE(560),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1143),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1802),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1803),
    [aux_sym_export_statement_repeat1] = STATE(1381),
    [sym_identifier] = ACTIONS(520),
    [anon_sym_export] = ACTIONS(522),
    [anon_sym_LBRACE] = ACTIONS(167),
    [anon_sym_import] = ACTIONS(173),
    [anon_sym_let] = ACTIONS(522),
    [anon_sym_await] = ACTIONS(524),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_yield] = ACTIONS(526),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(194),
    [anon_sym_async] = ACTIONS(530),
    [anon_sym_function] = ACTIONS(198),
    [anon_sym_new] = ACTIONS(532),
    [anon_sym_PLUS] = ACTIONS(540),
    [anon_sym_DASH] = ACTIONS(540),
    [anon_sym_SLASH] = ACTIONS(670),
    [anon_sym_BANG] = ACTIONS(542),
    [anon_sym_TILDE] = ACTIONS(542),
    [anon_sym_typeof] = ACTIONS(540),
    [anon_sym_void] = ACTIONS(540),
    [anon_sym_delete] = ACTIONS(540),
    [anon_sym_PLUS_PLUS] = ACTIONS(672),
    [anon_sym_DASH_DASH] = ACTIONS(672),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(547),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(549),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(522),
    [anon_sym_get] = ACTIONS(522),
    [anon_sym_set] = ACTIONS(522),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(297)] = {
    [sym_import] = STATE(1326),
    [sym_parenthesized_expression] = STATE(560),
    [sym_expression] = STATE(696),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1802),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1802),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_function_expression] = STATE(740),
    [sym_generator_function] = STATE(740),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(560),
    [sym_subscript_expression] = STATE(560),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1143),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1802),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1803),
    [aux_sym_export_statement_repeat1] = STATE(1381),
    [sym_identifier] = ACTIONS(520),
    [anon_sym_export] = ACTIONS(522),
    [anon_sym_LBRACE] = ACTIONS(167),
    [anon_sym_import] = ACTIONS(173),
    [anon_sym_let] = ACTIONS(522),
    [anon_sym_await] = ACTIONS(524),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_yield] = ACTIONS(526),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(194),
    [anon_sym_async] = ACTIONS(530),
    [anon_sym_function] = ACTIONS(198),
    [anon_sym_new] = ACTIONS(532),
    [anon_sym_PLUS] = ACTIONS(540),
    [anon_sym_DASH] = ACTIONS(540),
    [anon_sym_SLASH] = ACTIONS(670),
    [anon_sym_BANG] = ACTIONS(542),
    [anon_sym_TILDE] = ACTIONS(542),
    [anon_sym_typeof] = ACTIONS(540),
    [anon_sym_void] = ACTIONS(540),
    [anon_sym_delete] = ACTIONS(540),
    [anon_sym_PLUS_PLUS] = ACTIONS(672),
    [anon_sym_DASH_DASH] = ACTIONS(672),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(547),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(549),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(522),
    [anon_sym_get] = ACTIONS(522),
    [anon_sym_set] = ACTIONS(522),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(298)] = {
    [sym_import] = STATE(1326),
    [sym_parenthesized_expression] = STATE(560),
    [sym_expression] = STATE(704),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1802),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1802),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_function_expression] = STATE(740),
    [sym_generator_function] = STATE(740),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(560),
    [sym_subscript_expression] = STATE(560),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1143),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1802),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1803),
    [aux_sym_export_statement_repeat1] = STATE(1381),
    [sym_identifier] = ACTIONS(520),
    [anon_sym_export] = ACTIONS(522),
    [anon_sym_LBRACE] = ACTIONS(167),
    [anon_sym_import] = ACTIONS(173),
    [anon_sym_let] = ACTIONS(522),
    [anon_sym_await] = ACTIONS(524),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_yield] = ACTIONS(526),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(194),
    [anon_sym_async] = ACTIONS(530),
    [anon_sym_function] = ACTIONS(198),
    [anon_sym_new] = ACTIONS(532),
    [anon_sym_PLUS] = ACTIONS(540),
    [anon_sym_DASH] = ACTIONS(540),
    [anon_sym_SLASH] = ACTIONS(670),
    [anon_sym_BANG] = ACTIONS(542),
    [anon_sym_TILDE] = ACTIONS(542),
    [anon_sym_typeof] = ACTIONS(540),
    [anon_sym_void] = ACTIONS(540),
    [anon_sym_delete] = ACTIONS(540),
    [anon_sym_PLUS_PLUS] = ACTIONS(672),
    [anon_sym_DASH_DASH] = ACTIONS(672),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(547),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(549),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(522),
    [anon_sym_get] = ACTIONS(522),
    [anon_sym_set] = ACTIONS(522),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(299)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(498),
    [sym_expression] = STATE(972),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1762),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1762),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(498),
    [sym_subscript_expression] = STATE(498),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1138),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1762),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1816),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(430),
    [anon_sym_export] = ACTIONS(432),
    [anon_sym_LBRACE] = ACTIONS(434),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(432),
    [anon_sym_await] = ACTIONS(436),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(438),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(445),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(447),
    [anon_sym_PLUS] = ACTIONS(452),
    [anon_sym_DASH] = ACTIONS(452),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(454),
    [anon_sym_TILDE] = ACTIONS(454),
    [anon_sym_typeof] = ACTIONS(452),
    [anon_sym_void] = ACTIONS(452),
    [anon_sym_delete] = ACTIONS(452),
    [anon_sym_PLUS_PLUS] = ACTIONS(684),
    [anon_sym_DASH_DASH] = ACTIONS(684),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(459),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(461),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(432),
    [anon_sym_get] = ACTIONS(432),
    [anon_sym_set] = ACTIONS(432),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(300)] = {
    [sym_import] = STATE(1326),
    [sym_parenthesized_expression] = STATE(560),
    [sym_expression] = STATE(705),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1802),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1802),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_function_expression] = STATE(740),
    [sym_generator_function] = STATE(740),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(560),
    [sym_subscript_expression] = STATE(560),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1143),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1802),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1803),
    [aux_sym_export_statement_repeat1] = STATE(1381),
    [sym_identifier] = ACTIONS(520),
    [anon_sym_export] = ACTIONS(522),
    [anon_sym_LBRACE] = ACTIONS(167),
    [anon_sym_import] = ACTIONS(173),
    [anon_sym_let] = ACTIONS(522),
    [anon_sym_await] = ACTIONS(524),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_yield] = ACTIONS(526),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(194),
    [anon_sym_async] = ACTIONS(530),
    [anon_sym_function] = ACTIONS(198),
    [anon_sym_new] = ACTIONS(532),
    [anon_sym_PLUS] = ACTIONS(540),
    [anon_sym_DASH] = ACTIONS(540),
    [anon_sym_SLASH] = ACTIONS(670),
    [anon_sym_BANG] = ACTIONS(542),
    [anon_sym_TILDE] = ACTIONS(542),
    [anon_sym_typeof] = ACTIONS(540),
    [anon_sym_void] = ACTIONS(540),
    [anon_sym_delete] = ACTIONS(540),
    [anon_sym_PLUS_PLUS] = ACTIONS(672),
    [anon_sym_DASH_DASH] = ACTIONS(672),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(547),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(549),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(522),
    [anon_sym_get] = ACTIONS(522),
    [anon_sym_set] = ACTIONS(522),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(301)] = {
    [sym_import] = STATE(1326),
    [sym_parenthesized_expression] = STATE(560),
    [sym_expression] = STATE(707),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1802),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1802),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_function_expression] = STATE(740),
    [sym_generator_function] = STATE(740),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(560),
    [sym_subscript_expression] = STATE(560),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1143),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1802),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1803),
    [aux_sym_export_statement_repeat1] = STATE(1381),
    [sym_identifier] = ACTIONS(520),
    [anon_sym_export] = ACTIONS(522),
    [anon_sym_LBRACE] = ACTIONS(167),
    [anon_sym_import] = ACTIONS(173),
    [anon_sym_let] = ACTIONS(522),
    [anon_sym_await] = ACTIONS(524),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_yield] = ACTIONS(526),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(194),
    [anon_sym_async] = ACTIONS(530),
    [anon_sym_function] = ACTIONS(198),
    [anon_sym_new] = ACTIONS(532),
    [anon_sym_PLUS] = ACTIONS(540),
    [anon_sym_DASH] = ACTIONS(540),
    [anon_sym_SLASH] = ACTIONS(670),
    [anon_sym_BANG] = ACTIONS(542),
    [anon_sym_TILDE] = ACTIONS(542),
    [anon_sym_typeof] = ACTIONS(540),
    [anon_sym_void] = ACTIONS(540),
    [anon_sym_delete] = ACTIONS(540),
    [anon_sym_PLUS_PLUS] = ACTIONS(672),
    [anon_sym_DASH_DASH] = ACTIONS(672),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(547),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(549),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(522),
    [anon_sym_get] = ACTIONS(522),
    [anon_sym_set] = ACTIONS(522),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(302)] = {
    [sym_import] = STATE(1326),
    [sym_parenthesized_expression] = STATE(560),
    [sym_expression] = STATE(708),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1802),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1802),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_function_expression] = STATE(740),
    [sym_generator_function] = STATE(740),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(560),
    [sym_subscript_expression] = STATE(560),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1143),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1802),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1803),
    [aux_sym_export_statement_repeat1] = STATE(1381),
    [sym_identifier] = ACTIONS(520),
    [anon_sym_export] = ACTIONS(522),
    [anon_sym_LBRACE] = ACTIONS(167),
    [anon_sym_import] = ACTIONS(173),
    [anon_sym_let] = ACTIONS(522),
    [anon_sym_await] = ACTIONS(524),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_yield] = ACTIONS(526),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(194),
    [anon_sym_async] = ACTIONS(530),
    [anon_sym_function] = ACTIONS(198),
    [anon_sym_new] = ACTIONS(532),
    [anon_sym_PLUS] = ACTIONS(540),
    [anon_sym_DASH] = ACTIONS(540),
    [anon_sym_SLASH] = ACTIONS(670),
    [anon_sym_BANG] = ACTIONS(542),
    [anon_sym_TILDE] = ACTIONS(542),
    [anon_sym_typeof] = ACTIONS(540),
    [anon_sym_void] = ACTIONS(540),
    [anon_sym_delete] = ACTIONS(540),
    [anon_sym_PLUS_PLUS] = ACTIONS(672),
    [anon_sym_DASH_DASH] = ACTIONS(672),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(547),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(549),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(522),
    [anon_sym_get] = ACTIONS(522),
    [anon_sym_set] = ACTIONS(522),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(303)] = {
    [sym_import] = STATE(1326),
    [sym_parenthesized_expression] = STATE(560),
    [sym_expression] = STATE(709),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1802),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1802),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_function_expression] = STATE(740),
    [sym_generator_function] = STATE(740),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(560),
    [sym_subscript_expression] = STATE(560),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1143),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1802),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1803),
    [aux_sym_export_statement_repeat1] = STATE(1381),
    [sym_identifier] = ACTIONS(520),
    [anon_sym_export] = ACTIONS(522),
    [anon_sym_LBRACE] = ACTIONS(167),
    [anon_sym_import] = ACTIONS(173),
    [anon_sym_let] = ACTIONS(522),
    [anon_sym_await] = ACTIONS(524),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_yield] = ACTIONS(526),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(194),
    [anon_sym_async] = ACTIONS(530),
    [anon_sym_function] = ACTIONS(198),
    [anon_sym_new] = ACTIONS(532),
    [anon_sym_PLUS] = ACTIONS(540),
    [anon_sym_DASH] = ACTIONS(540),
    [anon_sym_SLASH] = ACTIONS(670),
    [anon_sym_BANG] = ACTIONS(542),
    [anon_sym_TILDE] = ACTIONS(542),
    [anon_sym_typeof] = ACTIONS(540),
    [anon_sym_void] = ACTIONS(540),
    [anon_sym_delete] = ACTIONS(540),
    [anon_sym_PLUS_PLUS] = ACTIONS(672),
    [anon_sym_DASH_DASH] = ACTIONS(672),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(547),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(549),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(522),
    [anon_sym_get] = ACTIONS(522),
    [anon_sym_set] = ACTIONS(522),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(304)] = {
    [sym_import] = STATE(1326),
    [sym_parenthesized_expression] = STATE(560),
    [sym_expression] = STATE(710),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1802),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1802),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_function_expression] = STATE(740),
    [sym_generator_function] = STATE(740),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(560),
    [sym_subscript_expression] = STATE(560),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1143),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1802),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1803),
    [aux_sym_export_statement_repeat1] = STATE(1381),
    [sym_identifier] = ACTIONS(520),
    [anon_sym_export] = ACTIONS(522),
    [anon_sym_LBRACE] = ACTIONS(167),
    [anon_sym_import] = ACTIONS(173),
    [anon_sym_let] = ACTIONS(522),
    [anon_sym_await] = ACTIONS(524),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_yield] = ACTIONS(526),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(194),
    [anon_sym_async] = ACTIONS(530),
    [anon_sym_function] = ACTIONS(198),
    [anon_sym_new] = ACTIONS(532),
    [anon_sym_PLUS] = ACTIONS(540),
    [anon_sym_DASH] = ACTIONS(540),
    [anon_sym_SLASH] = ACTIONS(670),
    [anon_sym_BANG] = ACTIONS(542),
    [anon_sym_TILDE] = ACTIONS(542),
    [anon_sym_typeof] = ACTIONS(540),
    [anon_sym_void] = ACTIONS(540),
    [anon_sym_delete] = ACTIONS(540),
    [anon_sym_PLUS_PLUS] = ACTIONS(672),
    [anon_sym_DASH_DASH] = ACTIONS(672),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(547),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(549),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(522),
    [anon_sym_get] = ACTIONS(522),
    [anon_sym_set] = ACTIONS(522),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(305)] = {
    [sym_import] = STATE(1326),
    [sym_parenthesized_expression] = STATE(560),
    [sym_expression] = STATE(711),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1802),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1802),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_function_expression] = STATE(740),
    [sym_generator_function] = STATE(740),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(560),
    [sym_subscript_expression] = STATE(560),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1143),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1802),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1803),
    [aux_sym_export_statement_repeat1] = STATE(1381),
    [sym_identifier] = ACTIONS(520),
    [anon_sym_export] = ACTIONS(522),
    [anon_sym_LBRACE] = ACTIONS(167),
    [anon_sym_import] = ACTIONS(173),
    [anon_sym_let] = ACTIONS(522),
    [anon_sym_await] = ACTIONS(524),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_yield] = ACTIONS(526),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(194),
    [anon_sym_async] = ACTIONS(530),
    [anon_sym_function] = ACTIONS(198),
    [anon_sym_new] = ACTIONS(532),
    [anon_sym_PLUS] = ACTIONS(540),
    [anon_sym_DASH] = ACTIONS(540),
    [anon_sym_SLASH] = ACTIONS(670),
    [anon_sym_BANG] = ACTIONS(542),
    [anon_sym_TILDE] = ACTIONS(542),
    [anon_sym_typeof] = ACTIONS(540),
    [anon_sym_void] = ACTIONS(540),
    [anon_sym_delete] = ACTIONS(540),
    [anon_sym_PLUS_PLUS] = ACTIONS(672),
    [anon_sym_DASH_DASH] = ACTIONS(672),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(547),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(549),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(522),
    [anon_sym_get] = ACTIONS(522),
    [anon_sym_set] = ACTIONS(522),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(306)] = {
    [sym_import] = STATE(1326),
    [sym_parenthesized_expression] = STATE(560),
    [sym_expression] = STATE(712),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1802),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1802),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_function_expression] = STATE(740),
    [sym_generator_function] = STATE(740),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(560),
    [sym_subscript_expression] = STATE(560),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1143),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1802),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1803),
    [aux_sym_export_statement_repeat1] = STATE(1381),
    [sym_identifier] = ACTIONS(520),
    [anon_sym_export] = ACTIONS(522),
    [anon_sym_LBRACE] = ACTIONS(167),
    [anon_sym_import] = ACTIONS(173),
    [anon_sym_let] = ACTIONS(522),
    [anon_sym_await] = ACTIONS(524),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_yield] = ACTIONS(526),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(194),
    [anon_sym_async] = ACTIONS(530),
    [anon_sym_function] = ACTIONS(198),
    [anon_sym_new] = ACTIONS(532),
    [anon_sym_PLUS] = ACTIONS(540),
    [anon_sym_DASH] = ACTIONS(540),
    [anon_sym_SLASH] = ACTIONS(670),
    [anon_sym_BANG] = ACTIONS(542),
    [anon_sym_TILDE] = ACTIONS(542),
    [anon_sym_typeof] = ACTIONS(540),
    [anon_sym_void] = ACTIONS(540),
    [anon_sym_delete] = ACTIONS(540),
    [anon_sym_PLUS_PLUS] = ACTIONS(672),
    [anon_sym_DASH_DASH] = ACTIONS(672),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(547),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(549),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(522),
    [anon_sym_get] = ACTIONS(522),
    [anon_sym_set] = ACTIONS(522),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(307)] = {
    [sym_import] = STATE(1326),
    [sym_parenthesized_expression] = STATE(560),
    [sym_expression] = STATE(713),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1802),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1802),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_function_expression] = STATE(740),
    [sym_generator_function] = STATE(740),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(560),
    [sym_subscript_expression] = STATE(560),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1143),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1802),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1803),
    [aux_sym_export_statement_repeat1] = STATE(1381),
    [sym_identifier] = ACTIONS(520),
    [anon_sym_export] = ACTIONS(522),
    [anon_sym_LBRACE] = ACTIONS(167),
    [anon_sym_import] = ACTIONS(173),
    [anon_sym_let] = ACTIONS(522),
    [anon_sym_await] = ACTIONS(524),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_yield] = ACTIONS(526),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(194),
    [anon_sym_async] = ACTIONS(530),
    [anon_sym_function] = ACTIONS(198),
    [anon_sym_new] = ACTIONS(532),
    [anon_sym_PLUS] = ACTIONS(540),
    [anon_sym_DASH] = ACTIONS(540),
    [anon_sym_SLASH] = ACTIONS(670),
    [anon_sym_BANG] = ACTIONS(542),
    [anon_sym_TILDE] = ACTIONS(542),
    [anon_sym_typeof] = ACTIONS(540),
    [anon_sym_void] = ACTIONS(540),
    [anon_sym_delete] = ACTIONS(540),
    [anon_sym_PLUS_PLUS] = ACTIONS(672),
    [anon_sym_DASH_DASH] = ACTIONS(672),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(547),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(549),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(522),
    [anon_sym_get] = ACTIONS(522),
    [anon_sym_set] = ACTIONS(522),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(308)] = {
    [sym_import] = STATE(1326),
    [sym_parenthesized_expression] = STATE(560),
    [sym_expression] = STATE(714),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1802),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1802),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_function_expression] = STATE(740),
    [sym_generator_function] = STATE(740),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(560),
    [sym_subscript_expression] = STATE(560),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1143),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1802),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1803),
    [aux_sym_export_statement_repeat1] = STATE(1381),
    [sym_identifier] = ACTIONS(520),
    [anon_sym_export] = ACTIONS(522),
    [anon_sym_LBRACE] = ACTIONS(167),
    [anon_sym_import] = ACTIONS(173),
    [anon_sym_let] = ACTIONS(522),
    [anon_sym_await] = ACTIONS(524),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_yield] = ACTIONS(526),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(194),
    [anon_sym_async] = ACTIONS(530),
    [anon_sym_function] = ACTIONS(198),
    [anon_sym_new] = ACTIONS(532),
    [anon_sym_PLUS] = ACTIONS(540),
    [anon_sym_DASH] = ACTIONS(540),
    [anon_sym_SLASH] = ACTIONS(670),
    [anon_sym_BANG] = ACTIONS(542),
    [anon_sym_TILDE] = ACTIONS(542),
    [anon_sym_typeof] = ACTIONS(540),
    [anon_sym_void] = ACTIONS(540),
    [anon_sym_delete] = ACTIONS(540),
    [anon_sym_PLUS_PLUS] = ACTIONS(672),
    [anon_sym_DASH_DASH] = ACTIONS(672),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(547),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(549),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(522),
    [anon_sym_get] = ACTIONS(522),
    [anon_sym_set] = ACTIONS(522),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(309)] = {
    [sym_import] = STATE(1326),
    [sym_parenthesized_expression] = STATE(560),
    [sym_expression] = STATE(715),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1802),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1802),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_function_expression] = STATE(740),
    [sym_generator_function] = STATE(740),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(560),
    [sym_subscript_expression] = STATE(560),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1143),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1802),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1803),
    [aux_sym_export_statement_repeat1] = STATE(1381),
    [sym_identifier] = ACTIONS(520),
    [anon_sym_export] = ACTIONS(522),
    [anon_sym_LBRACE] = ACTIONS(167),
    [anon_sym_import] = ACTIONS(173),
    [anon_sym_let] = ACTIONS(522),
    [anon_sym_await] = ACTIONS(524),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_yield] = ACTIONS(526),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(194),
    [anon_sym_async] = ACTIONS(530),
    [anon_sym_function] = ACTIONS(198),
    [anon_sym_new] = ACTIONS(532),
    [anon_sym_PLUS] = ACTIONS(540),
    [anon_sym_DASH] = ACTIONS(540),
    [anon_sym_SLASH] = ACTIONS(670),
    [anon_sym_BANG] = ACTIONS(542),
    [anon_sym_TILDE] = ACTIONS(542),
    [anon_sym_typeof] = ACTIONS(540),
    [anon_sym_void] = ACTIONS(540),
    [anon_sym_delete] = ACTIONS(540),
    [anon_sym_PLUS_PLUS] = ACTIONS(672),
    [anon_sym_DASH_DASH] = ACTIONS(672),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(547),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(549),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(522),
    [anon_sym_get] = ACTIONS(522),
    [anon_sym_set] = ACTIONS(522),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(310)] = {
    [sym_import] = STATE(1326),
    [sym_parenthesized_expression] = STATE(560),
    [sym_expression] = STATE(716),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1802),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1802),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_function_expression] = STATE(740),
    [sym_generator_function] = STATE(740),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(560),
    [sym_subscript_expression] = STATE(560),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1143),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1802),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1803),
    [aux_sym_export_statement_repeat1] = STATE(1381),
    [sym_identifier] = ACTIONS(520),
    [anon_sym_export] = ACTIONS(522),
    [anon_sym_LBRACE] = ACTIONS(167),
    [anon_sym_import] = ACTIONS(173),
    [anon_sym_let] = ACTIONS(522),
    [anon_sym_await] = ACTIONS(524),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_yield] = ACTIONS(526),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(194),
    [anon_sym_async] = ACTIONS(530),
    [anon_sym_function] = ACTIONS(198),
    [anon_sym_new] = ACTIONS(532),
    [anon_sym_PLUS] = ACTIONS(540),
    [anon_sym_DASH] = ACTIONS(540),
    [anon_sym_SLASH] = ACTIONS(670),
    [anon_sym_BANG] = ACTIONS(542),
    [anon_sym_TILDE] = ACTIONS(542),
    [anon_sym_typeof] = ACTIONS(540),
    [anon_sym_void] = ACTIONS(540),
    [anon_sym_delete] = ACTIONS(540),
    [anon_sym_PLUS_PLUS] = ACTIONS(672),
    [anon_sym_DASH_DASH] = ACTIONS(672),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(547),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(549),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(522),
    [anon_sym_get] = ACTIONS(522),
    [anon_sym_set] = ACTIONS(522),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(311)] = {
    [sym_import] = STATE(1326),
    [sym_parenthesized_expression] = STATE(560),
    [sym_expression] = STATE(717),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1802),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1802),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_function_expression] = STATE(740),
    [sym_generator_function] = STATE(740),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(560),
    [sym_subscript_expression] = STATE(560),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1143),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1802),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1803),
    [aux_sym_export_statement_repeat1] = STATE(1381),
    [sym_identifier] = ACTIONS(520),
    [anon_sym_export] = ACTIONS(522),
    [anon_sym_LBRACE] = ACTIONS(167),
    [anon_sym_import] = ACTIONS(173),
    [anon_sym_let] = ACTIONS(522),
    [anon_sym_await] = ACTIONS(524),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_yield] = ACTIONS(526),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(194),
    [anon_sym_async] = ACTIONS(530),
    [anon_sym_function] = ACTIONS(198),
    [anon_sym_new] = ACTIONS(532),
    [anon_sym_PLUS] = ACTIONS(540),
    [anon_sym_DASH] = ACTIONS(540),
    [anon_sym_SLASH] = ACTIONS(670),
    [anon_sym_BANG] = ACTIONS(542),
    [anon_sym_TILDE] = ACTIONS(542),
    [anon_sym_typeof] = ACTIONS(540),
    [anon_sym_void] = ACTIONS(540),
    [anon_sym_delete] = ACTIONS(540),
    [anon_sym_PLUS_PLUS] = ACTIONS(672),
    [anon_sym_DASH_DASH] = ACTIONS(672),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(547),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(549),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(522),
    [anon_sym_get] = ACTIONS(522),
    [anon_sym_set] = ACTIONS(522),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(312)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(507),
    [sym_expression] = STATE(976),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1815),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1815),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(507),
    [sym_subscript_expression] = STATE(507),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1139),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1815),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1763),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(369),
    [anon_sym_export] = ACTIONS(371),
    [anon_sym_LBRACE] = ACTIONS(373),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(371),
    [anon_sym_await] = ACTIONS(377),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(382),
    [anon_sym_LBRACK] = ACTIONS(654),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(398),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(404),
    [anon_sym_PLUS] = ACTIONS(412),
    [anon_sym_DASH] = ACTIONS(412),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(414),
    [anon_sym_TILDE] = ACTIONS(414),
    [anon_sym_typeof] = ACTIONS(412),
    [anon_sym_void] = ACTIONS(412),
    [anon_sym_delete] = ACTIONS(412),
    [anon_sym_PLUS_PLUS] = ACTIONS(660),
    [anon_sym_DASH_DASH] = ACTIONS(660),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(424),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(428),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(371),
    [anon_sym_get] = ACTIONS(371),
    [anon_sym_set] = ACTIONS(371),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(313)] = {
    [sym_import] = STATE(1326),
    [sym_parenthesized_expression] = STATE(560),
    [sym_expression] = STATE(727),
    [sym_primary_expression] = STATE(686),
    [sym_yield_expression] = STATE(746),
    [sym_object] = STATE(740),
    [sym_object_pattern] = STATE(1802),
    [sym_array] = STATE(740),
    [sym_array_pattern] = STATE(1802),
    [sym_jsx_element] = STATE(746),
    [sym_jsx_opening_element] = STATE(1155),
    [sym_jsx_self_closing_element] = STATE(746),
    [sym_class] = STATE(740),
    [sym_function_expression] = STATE(740),
    [sym_generator_function] = STATE(740),
    [sym_arrow_function] = STATE(740),
    [sym_call_expression] = STATE(740),
    [sym_new_expression] = STATE(682),
    [sym_await_expression] = STATE(746),
    [sym_member_expression] = STATE(560),
    [sym_subscript_expression] = STATE(560),
    [sym_assignment_expression] = STATE(746),
    [sym__augmented_assignment_lhs] = STATE(1143),
    [sym_augmented_assignment_expression] = STATE(746),
    [sym__destructuring_pattern] = STATE(1802),
    [sym_ternary_expression] = STATE(746),
    [sym_binary_expression] = STATE(746),
    [sym_unary_expression] = STATE(746),
    [sym_update_expression] = STATE(746),
    [sym_string] = STATE(740),
    [sym_template_string] = STATE(740),
    [sym_regex] = STATE(740),
    [sym_meta_property] = STATE(740),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1803),
    [aux_sym_export_statement_repeat1] = STATE(1381),
    [sym_identifier] = ACTIONS(520),
    [anon_sym_export] = ACTIONS(522),
    [anon_sym_LBRACE] = ACTIONS(167),
    [anon_sym_import] = ACTIONS(173),
    [anon_sym_let] = ACTIONS(522),
    [anon_sym_await] = ACTIONS(524),
    [anon_sym_LPAREN] = ACTIONS(37),
    [anon_sym_yield] = ACTIONS(526),
    [anon_sym_LBRACK] = ACTIONS(59),
    [anon_sym_LT] = ACTIONS(61),
    [anon_sym_DQUOTE] = ACTIONS(63),
    [anon_sym_SQUOTE] = ACTIONS(65),
    [anon_sym_class] = ACTIONS(194),
    [anon_sym_async] = ACTIONS(530),
    [anon_sym_function] = ACTIONS(198),
    [anon_sym_new] = ACTIONS(532),
    [anon_sym_PLUS] = ACTIONS(540),
    [anon_sym_DASH] = ACTIONS(540),
    [anon_sym_SLASH] = ACTIONS(670),
    [anon_sym_BANG] = ACTIONS(542),
    [anon_sym_TILDE] = ACTIONS(542),
    [anon_sym_typeof] = ACTIONS(540),
    [anon_sym_void] = ACTIONS(540),
    [anon_sym_delete] = ACTIONS(540),
    [anon_sym_PLUS_PLUS] = ACTIONS(672),
    [anon_sym_DASH_DASH] = ACTIONS(672),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(83),
    [sym_number] = ACTIONS(85),
    [sym_private_property_identifier] = ACTIONS(547),
    [sym_this] = ACTIONS(89),
    [sym_super] = ACTIONS(89),
    [sym_true] = ACTIONS(89),
    [sym_false] = ACTIONS(89),
    [sym_null] = ACTIONS(89),
    [sym_undefined] = ACTIONS(549),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(522),
    [anon_sym_get] = ACTIONS(522),
    [anon_sym_set] = ACTIONS(522),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(314)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(507),
    [sym_expression] = STATE(825),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1815),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1815),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(507),
    [sym_subscript_expression] = STATE(507),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1139),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1815),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1763),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(369),
    [anon_sym_export] = ACTIONS(371),
    [anon_sym_LBRACE] = ACTIONS(373),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(371),
    [anon_sym_await] = ACTIONS(377),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(382),
    [anon_sym_LBRACK] = ACTIONS(654),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(398),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(404),
    [anon_sym_PLUS] = ACTIONS(412),
    [anon_sym_DASH] = ACTIONS(412),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(414),
    [anon_sym_TILDE] = ACTIONS(414),
    [anon_sym_typeof] = ACTIONS(412),
    [anon_sym_void] = ACTIONS(412),
    [anon_sym_delete] = ACTIONS(412),
    [anon_sym_PLUS_PLUS] = ACTIONS(660),
    [anon_sym_DASH_DASH] = ACTIONS(660),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(424),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(428),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(371),
    [anon_sym_get] = ACTIONS(371),
    [anon_sym_set] = ACTIONS(371),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(315)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(498),
    [sym_expression] = STATE(588),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1762),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1762),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(498),
    [sym_subscript_expression] = STATE(498),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1138),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1762),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1816),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(430),
    [anon_sym_export] = ACTIONS(432),
    [anon_sym_LBRACE] = ACTIONS(434),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(432),
    [anon_sym_await] = ACTIONS(436),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(438),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(445),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(447),
    [anon_sym_PLUS] = ACTIONS(452),
    [anon_sym_DASH] = ACTIONS(452),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(454),
    [anon_sym_TILDE] = ACTIONS(454),
    [anon_sym_typeof] = ACTIONS(452),
    [anon_sym_void] = ACTIONS(452),
    [anon_sym_delete] = ACTIONS(452),
    [anon_sym_PLUS_PLUS] = ACTIONS(684),
    [anon_sym_DASH_DASH] = ACTIONS(684),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(459),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(461),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(432),
    [anon_sym_get] = ACTIONS(432),
    [anon_sym_set] = ACTIONS(432),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(316)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(498),
    [sym_expression] = STATE(591),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1762),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1762),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(498),
    [sym_subscript_expression] = STATE(498),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1138),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1762),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1816),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(430),
    [anon_sym_export] = ACTIONS(432),
    [anon_sym_LBRACE] = ACTIONS(434),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(432),
    [anon_sym_await] = ACTIONS(436),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(438),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(445),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(447),
    [anon_sym_PLUS] = ACTIONS(452),
    [anon_sym_DASH] = ACTIONS(452),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(454),
    [anon_sym_TILDE] = ACTIONS(454),
    [anon_sym_typeof] = ACTIONS(452),
    [anon_sym_void] = ACTIONS(452),
    [anon_sym_delete] = ACTIONS(452),
    [anon_sym_PLUS_PLUS] = ACTIONS(684),
    [anon_sym_DASH_DASH] = ACTIONS(684),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(459),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(461),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(432),
    [anon_sym_get] = ACTIONS(432),
    [anon_sym_set] = ACTIONS(432),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(317)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(498),
    [sym_expression] = STATE(919),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1762),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1762),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(498),
    [sym_subscript_expression] = STATE(498),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1138),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1762),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1816),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(430),
    [anon_sym_export] = ACTIONS(432),
    [anon_sym_LBRACE] = ACTIONS(434),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(432),
    [anon_sym_await] = ACTIONS(436),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(438),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(445),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(447),
    [anon_sym_PLUS] = ACTIONS(452),
    [anon_sym_DASH] = ACTIONS(452),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(454),
    [anon_sym_TILDE] = ACTIONS(454),
    [anon_sym_typeof] = ACTIONS(452),
    [anon_sym_void] = ACTIONS(452),
    [anon_sym_delete] = ACTIONS(452),
    [anon_sym_PLUS_PLUS] = ACTIONS(684),
    [anon_sym_DASH_DASH] = ACTIONS(684),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(459),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(461),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(432),
    [anon_sym_get] = ACTIONS(432),
    [anon_sym_set] = ACTIONS(432),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(318)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(498),
    [sym_expression] = STATE(921),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1762),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1762),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(498),
    [sym_subscript_expression] = STATE(498),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1138),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1762),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1816),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(430),
    [anon_sym_export] = ACTIONS(432),
    [anon_sym_LBRACE] = ACTIONS(434),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(432),
    [anon_sym_await] = ACTIONS(436),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(438),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(445),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(447),
    [anon_sym_PLUS] = ACTIONS(452),
    [anon_sym_DASH] = ACTIONS(452),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(454),
    [anon_sym_TILDE] = ACTIONS(454),
    [anon_sym_typeof] = ACTIONS(452),
    [anon_sym_void] = ACTIONS(452),
    [anon_sym_delete] = ACTIONS(452),
    [anon_sym_PLUS_PLUS] = ACTIONS(684),
    [anon_sym_DASH_DASH] = ACTIONS(684),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(459),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(461),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(432),
    [anon_sym_get] = ACTIONS(432),
    [anon_sym_set] = ACTIONS(432),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(319)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(498),
    [sym_expression] = STATE(887),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1762),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1762),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(498),
    [sym_subscript_expression] = STATE(498),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1138),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1762),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1816),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(430),
    [anon_sym_export] = ACTIONS(432),
    [anon_sym_LBRACE] = ACTIONS(434),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(432),
    [anon_sym_await] = ACTIONS(436),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(438),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(445),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(447),
    [anon_sym_PLUS] = ACTIONS(452),
    [anon_sym_DASH] = ACTIONS(452),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(454),
    [anon_sym_TILDE] = ACTIONS(454),
    [anon_sym_typeof] = ACTIONS(452),
    [anon_sym_void] = ACTIONS(452),
    [anon_sym_delete] = ACTIONS(452),
    [anon_sym_PLUS_PLUS] = ACTIONS(684),
    [anon_sym_DASH_DASH] = ACTIONS(684),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(459),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(461),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(432),
    [anon_sym_get] = ACTIONS(432),
    [anon_sym_set] = ACTIONS(432),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(320)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(498),
    [sym_expression] = STATE(923),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1762),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1762),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(498),
    [sym_subscript_expression] = STATE(498),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1138),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1762),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1816),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(430),
    [anon_sym_export] = ACTIONS(432),
    [anon_sym_LBRACE] = ACTIONS(434),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(432),
    [anon_sym_await] = ACTIONS(436),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(438),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(445),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(447),
    [anon_sym_PLUS] = ACTIONS(452),
    [anon_sym_DASH] = ACTIONS(452),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(454),
    [anon_sym_TILDE] = ACTIONS(454),
    [anon_sym_typeof] = ACTIONS(452),
    [anon_sym_void] = ACTIONS(452),
    [anon_sym_delete] = ACTIONS(452),
    [anon_sym_PLUS_PLUS] = ACTIONS(684),
    [anon_sym_DASH_DASH] = ACTIONS(684),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(459),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(461),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(432),
    [anon_sym_get] = ACTIONS(432),
    [anon_sym_set] = ACTIONS(432),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(321)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(498),
    [sym_expression] = STATE(924),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1762),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1762),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(498),
    [sym_subscript_expression] = STATE(498),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1138),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1762),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1816),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(430),
    [anon_sym_export] = ACTIONS(432),
    [anon_sym_LBRACE] = ACTIONS(434),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(432),
    [anon_sym_await] = ACTIONS(436),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(438),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(445),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(447),
    [anon_sym_PLUS] = ACTIONS(452),
    [anon_sym_DASH] = ACTIONS(452),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(454),
    [anon_sym_TILDE] = ACTIONS(454),
    [anon_sym_typeof] = ACTIONS(452),
    [anon_sym_void] = ACTIONS(452),
    [anon_sym_delete] = ACTIONS(452),
    [anon_sym_PLUS_PLUS] = ACTIONS(684),
    [anon_sym_DASH_DASH] = ACTIONS(684),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(459),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(461),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(432),
    [anon_sym_get] = ACTIONS(432),
    [anon_sym_set] = ACTIONS(432),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(322)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(498),
    [sym_expression] = STATE(925),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1762),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1762),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(498),
    [sym_subscript_expression] = STATE(498),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1138),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1762),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1816),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(430),
    [anon_sym_export] = ACTIONS(432),
    [anon_sym_LBRACE] = ACTIONS(434),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(432),
    [anon_sym_await] = ACTIONS(436),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(438),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(445),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(447),
    [anon_sym_PLUS] = ACTIONS(452),
    [anon_sym_DASH] = ACTIONS(452),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(454),
    [anon_sym_TILDE] = ACTIONS(454),
    [anon_sym_typeof] = ACTIONS(452),
    [anon_sym_void] = ACTIONS(452),
    [anon_sym_delete] = ACTIONS(452),
    [anon_sym_PLUS_PLUS] = ACTIONS(684),
    [anon_sym_DASH_DASH] = ACTIONS(684),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(459),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(461),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(432),
    [anon_sym_get] = ACTIONS(432),
    [anon_sym_set] = ACTIONS(432),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(323)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(498),
    [sym_expression] = STATE(926),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1762),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1762),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(498),
    [sym_subscript_expression] = STATE(498),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1138),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1762),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1816),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(430),
    [anon_sym_export] = ACTIONS(432),
    [anon_sym_LBRACE] = ACTIONS(434),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(432),
    [anon_sym_await] = ACTIONS(436),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(438),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(445),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(447),
    [anon_sym_PLUS] = ACTIONS(452),
    [anon_sym_DASH] = ACTIONS(452),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(454),
    [anon_sym_TILDE] = ACTIONS(454),
    [anon_sym_typeof] = ACTIONS(452),
    [anon_sym_void] = ACTIONS(452),
    [anon_sym_delete] = ACTIONS(452),
    [anon_sym_PLUS_PLUS] = ACTIONS(684),
    [anon_sym_DASH_DASH] = ACTIONS(684),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(459),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(461),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(432),
    [anon_sym_get] = ACTIONS(432),
    [anon_sym_set] = ACTIONS(432),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(324)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(498),
    [sym_expression] = STATE(927),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1762),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1762),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(498),
    [sym_subscript_expression] = STATE(498),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1138),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1762),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1816),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(430),
    [anon_sym_export] = ACTIONS(432),
    [anon_sym_LBRACE] = ACTIONS(434),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(432),
    [anon_sym_await] = ACTIONS(436),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(438),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(445),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(447),
    [anon_sym_PLUS] = ACTIONS(452),
    [anon_sym_DASH] = ACTIONS(452),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(454),
    [anon_sym_TILDE] = ACTIONS(454),
    [anon_sym_typeof] = ACTIONS(452),
    [anon_sym_void] = ACTIONS(452),
    [anon_sym_delete] = ACTIONS(452),
    [anon_sym_PLUS_PLUS] = ACTIONS(684),
    [anon_sym_DASH_DASH] = ACTIONS(684),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(459),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(461),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(432),
    [anon_sym_get] = ACTIONS(432),
    [anon_sym_set] = ACTIONS(432),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(325)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(498),
    [sym_expression] = STATE(928),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1762),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1762),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(498),
    [sym_subscript_expression] = STATE(498),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1138),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1762),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1816),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(430),
    [anon_sym_export] = ACTIONS(432),
    [anon_sym_LBRACE] = ACTIONS(434),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(432),
    [anon_sym_await] = ACTIONS(436),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(438),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(445),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(447),
    [anon_sym_PLUS] = ACTIONS(452),
    [anon_sym_DASH] = ACTIONS(452),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(454),
    [anon_sym_TILDE] = ACTIONS(454),
    [anon_sym_typeof] = ACTIONS(452),
    [anon_sym_void] = ACTIONS(452),
    [anon_sym_delete] = ACTIONS(452),
    [anon_sym_PLUS_PLUS] = ACTIONS(684),
    [anon_sym_DASH_DASH] = ACTIONS(684),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(459),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(461),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(432),
    [anon_sym_get] = ACTIONS(432),
    [anon_sym_set] = ACTIONS(432),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(326)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(498),
    [sym_expression] = STATE(929),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1762),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1762),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(498),
    [sym_subscript_expression] = STATE(498),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1138),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1762),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1816),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(430),
    [anon_sym_export] = ACTIONS(432),
    [anon_sym_LBRACE] = ACTIONS(434),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(432),
    [anon_sym_await] = ACTIONS(436),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(438),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(445),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(447),
    [anon_sym_PLUS] = ACTIONS(452),
    [anon_sym_DASH] = ACTIONS(452),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(454),
    [anon_sym_TILDE] = ACTIONS(454),
    [anon_sym_typeof] = ACTIONS(452),
    [anon_sym_void] = ACTIONS(452),
    [anon_sym_delete] = ACTIONS(452),
    [anon_sym_PLUS_PLUS] = ACTIONS(684),
    [anon_sym_DASH_DASH] = ACTIONS(684),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(459),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(461),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(432),
    [anon_sym_get] = ACTIONS(432),
    [anon_sym_set] = ACTIONS(432),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(327)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(498),
    [sym_expression] = STATE(930),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1762),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1762),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(498),
    [sym_subscript_expression] = STATE(498),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1138),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1762),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1816),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(430),
    [anon_sym_export] = ACTIONS(432),
    [anon_sym_LBRACE] = ACTIONS(434),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(432),
    [anon_sym_await] = ACTIONS(436),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(438),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(445),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(447),
    [anon_sym_PLUS] = ACTIONS(452),
    [anon_sym_DASH] = ACTIONS(452),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(454),
    [anon_sym_TILDE] = ACTIONS(454),
    [anon_sym_typeof] = ACTIONS(452),
    [anon_sym_void] = ACTIONS(452),
    [anon_sym_delete] = ACTIONS(452),
    [anon_sym_PLUS_PLUS] = ACTIONS(684),
    [anon_sym_DASH_DASH] = ACTIONS(684),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(459),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(461),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(432),
    [anon_sym_get] = ACTIONS(432),
    [anon_sym_set] = ACTIONS(432),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(328)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(498),
    [sym_expression] = STATE(931),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1762),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1762),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(498),
    [sym_subscript_expression] = STATE(498),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1138),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1762),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1816),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(430),
    [anon_sym_export] = ACTIONS(432),
    [anon_sym_LBRACE] = ACTIONS(434),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(432),
    [anon_sym_await] = ACTIONS(436),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(438),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(445),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(447),
    [anon_sym_PLUS] = ACTIONS(452),
    [anon_sym_DASH] = ACTIONS(452),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(454),
    [anon_sym_TILDE] = ACTIONS(454),
    [anon_sym_typeof] = ACTIONS(452),
    [anon_sym_void] = ACTIONS(452),
    [anon_sym_delete] = ACTIONS(452),
    [anon_sym_PLUS_PLUS] = ACTIONS(684),
    [anon_sym_DASH_DASH] = ACTIONS(684),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(459),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(461),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(432),
    [anon_sym_get] = ACTIONS(432),
    [anon_sym_set] = ACTIONS(432),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(329)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(498),
    [sym_expression] = STATE(932),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1762),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1762),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(498),
    [sym_subscript_expression] = STATE(498),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1138),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1762),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1816),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(430),
    [anon_sym_export] = ACTIONS(432),
    [anon_sym_LBRACE] = ACTIONS(434),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(432),
    [anon_sym_await] = ACTIONS(436),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(438),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(445),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(447),
    [anon_sym_PLUS] = ACTIONS(452),
    [anon_sym_DASH] = ACTIONS(452),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(454),
    [anon_sym_TILDE] = ACTIONS(454),
    [anon_sym_typeof] = ACTIONS(452),
    [anon_sym_void] = ACTIONS(452),
    [anon_sym_delete] = ACTIONS(452),
    [anon_sym_PLUS_PLUS] = ACTIONS(684),
    [anon_sym_DASH_DASH] = ACTIONS(684),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(459),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(461),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(432),
    [anon_sym_get] = ACTIONS(432),
    [anon_sym_set] = ACTIONS(432),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(330)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(498),
    [sym_expression] = STATE(933),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1762),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1762),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(498),
    [sym_subscript_expression] = STATE(498),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1138),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1762),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1816),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(430),
    [anon_sym_export] = ACTIONS(432),
    [anon_sym_LBRACE] = ACTIONS(434),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(432),
    [anon_sym_await] = ACTIONS(436),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(438),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(445),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(447),
    [anon_sym_PLUS] = ACTIONS(452),
    [anon_sym_DASH] = ACTIONS(452),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(454),
    [anon_sym_TILDE] = ACTIONS(454),
    [anon_sym_typeof] = ACTIONS(452),
    [anon_sym_void] = ACTIONS(452),
    [anon_sym_delete] = ACTIONS(452),
    [anon_sym_PLUS_PLUS] = ACTIONS(684),
    [anon_sym_DASH_DASH] = ACTIONS(684),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(459),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(461),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(432),
    [anon_sym_get] = ACTIONS(432),
    [anon_sym_set] = ACTIONS(432),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(331)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(498),
    [sym_expression] = STATE(934),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1762),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1762),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(498),
    [sym_subscript_expression] = STATE(498),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1138),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1762),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1816),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(430),
    [anon_sym_export] = ACTIONS(432),
    [anon_sym_LBRACE] = ACTIONS(434),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(432),
    [anon_sym_await] = ACTIONS(436),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(438),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(445),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(447),
    [anon_sym_PLUS] = ACTIONS(452),
    [anon_sym_DASH] = ACTIONS(452),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(454),
    [anon_sym_TILDE] = ACTIONS(454),
    [anon_sym_typeof] = ACTIONS(452),
    [anon_sym_void] = ACTIONS(452),
    [anon_sym_delete] = ACTIONS(452),
    [anon_sym_PLUS_PLUS] = ACTIONS(684),
    [anon_sym_DASH_DASH] = ACTIONS(684),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(459),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(461),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(432),
    [anon_sym_get] = ACTIONS(432),
    [anon_sym_set] = ACTIONS(432),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(332)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(498),
    [sym_expression] = STATE(935),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1762),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1762),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(498),
    [sym_subscript_expression] = STATE(498),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1138),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1762),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1816),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(430),
    [anon_sym_export] = ACTIONS(432),
    [anon_sym_LBRACE] = ACTIONS(434),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(432),
    [anon_sym_await] = ACTIONS(436),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(438),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(445),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(447),
    [anon_sym_PLUS] = ACTIONS(452),
    [anon_sym_DASH] = ACTIONS(452),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(454),
    [anon_sym_TILDE] = ACTIONS(454),
    [anon_sym_typeof] = ACTIONS(452),
    [anon_sym_void] = ACTIONS(452),
    [anon_sym_delete] = ACTIONS(452),
    [anon_sym_PLUS_PLUS] = ACTIONS(684),
    [anon_sym_DASH_DASH] = ACTIONS(684),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(459),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(461),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(432),
    [anon_sym_get] = ACTIONS(432),
    [anon_sym_set] = ACTIONS(432),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(333)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(498),
    [sym_expression] = STATE(940),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1762),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1762),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(498),
    [sym_subscript_expression] = STATE(498),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1138),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1762),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1816),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(430),
    [anon_sym_export] = ACTIONS(432),
    [anon_sym_LBRACE] = ACTIONS(434),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(432),
    [anon_sym_await] = ACTIONS(436),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(438),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(445),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(447),
    [anon_sym_PLUS] = ACTIONS(452),
    [anon_sym_DASH] = ACTIONS(452),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(454),
    [anon_sym_TILDE] = ACTIONS(454),
    [anon_sym_typeof] = ACTIONS(452),
    [anon_sym_void] = ACTIONS(452),
    [anon_sym_delete] = ACTIONS(452),
    [anon_sym_PLUS_PLUS] = ACTIONS(684),
    [anon_sym_DASH_DASH] = ACTIONS(684),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(459),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(461),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(432),
    [anon_sym_get] = ACTIONS(432),
    [anon_sym_set] = ACTIONS(432),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(334)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(498),
    [sym_expression] = STATE(973),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1762),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1762),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(498),
    [sym_subscript_expression] = STATE(498),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1138),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1762),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1816),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(430),
    [anon_sym_export] = ACTIONS(432),
    [anon_sym_LBRACE] = ACTIONS(434),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(432),
    [anon_sym_await] = ACTIONS(436),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(438),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(445),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(447),
    [anon_sym_PLUS] = ACTIONS(452),
    [anon_sym_DASH] = ACTIONS(452),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(454),
    [anon_sym_TILDE] = ACTIONS(454),
    [anon_sym_typeof] = ACTIONS(452),
    [anon_sym_void] = ACTIONS(452),
    [anon_sym_delete] = ACTIONS(452),
    [anon_sym_PLUS_PLUS] = ACTIONS(684),
    [anon_sym_DASH_DASH] = ACTIONS(684),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(459),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(461),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(432),
    [anon_sym_get] = ACTIONS(432),
    [anon_sym_set] = ACTIONS(432),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(335)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(579),
    [sym_expression] = STATE(588),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1857),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1857),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(579),
    [sym_subscript_expression] = STATE(579),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1141),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1857),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1779),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(582),
    [anon_sym_export] = ACTIONS(584),
    [anon_sym_LBRACE] = ACTIONS(434),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(584),
    [anon_sym_await] = ACTIONS(586),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(588),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(592),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(594),
    [anon_sym_PLUS] = ACTIONS(602),
    [anon_sym_DASH] = ACTIONS(602),
    [anon_sym_SLASH] = ACTIONS(678),
    [anon_sym_BANG] = ACTIONS(604),
    [anon_sym_TILDE] = ACTIONS(604),
    [anon_sym_typeof] = ACTIONS(602),
    [anon_sym_void] = ACTIONS(602),
    [anon_sym_delete] = ACTIONS(602),
    [anon_sym_PLUS_PLUS] = ACTIONS(680),
    [anon_sym_DASH_DASH] = ACTIONS(680),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(609),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(611),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(584),
    [anon_sym_get] = ACTIONS(584),
    [anon_sym_set] = ACTIONS(584),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(336)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(579),
    [sym_expression] = STATE(591),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1857),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1857),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(579),
    [sym_subscript_expression] = STATE(579),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1141),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1857),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1779),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(582),
    [anon_sym_export] = ACTIONS(584),
    [anon_sym_LBRACE] = ACTIONS(434),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(584),
    [anon_sym_await] = ACTIONS(586),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(588),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(592),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(594),
    [anon_sym_PLUS] = ACTIONS(602),
    [anon_sym_DASH] = ACTIONS(602),
    [anon_sym_SLASH] = ACTIONS(678),
    [anon_sym_BANG] = ACTIONS(604),
    [anon_sym_TILDE] = ACTIONS(604),
    [anon_sym_typeof] = ACTIONS(602),
    [anon_sym_void] = ACTIONS(602),
    [anon_sym_delete] = ACTIONS(602),
    [anon_sym_PLUS_PLUS] = ACTIONS(680),
    [anon_sym_DASH_DASH] = ACTIONS(680),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(609),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(611),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(584),
    [anon_sym_get] = ACTIONS(584),
    [anon_sym_set] = ACTIONS(584),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(337)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(579),
    [sym_expression] = STATE(942),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1857),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1857),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(579),
    [sym_subscript_expression] = STATE(579),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1141),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1857),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1779),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(582),
    [anon_sym_export] = ACTIONS(584),
    [anon_sym_LBRACE] = ACTIONS(434),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(584),
    [anon_sym_await] = ACTIONS(586),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(588),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(592),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(594),
    [anon_sym_PLUS] = ACTIONS(602),
    [anon_sym_DASH] = ACTIONS(602),
    [anon_sym_SLASH] = ACTIONS(678),
    [anon_sym_BANG] = ACTIONS(604),
    [anon_sym_TILDE] = ACTIONS(604),
    [anon_sym_typeof] = ACTIONS(602),
    [anon_sym_void] = ACTIONS(602),
    [anon_sym_delete] = ACTIONS(602),
    [anon_sym_PLUS_PLUS] = ACTIONS(680),
    [anon_sym_DASH_DASH] = ACTIONS(680),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(609),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(611),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(584),
    [anon_sym_get] = ACTIONS(584),
    [anon_sym_set] = ACTIONS(584),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(338)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(579),
    [sym_expression] = STATE(944),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1857),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1857),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(579),
    [sym_subscript_expression] = STATE(579),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1141),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1857),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1779),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(582),
    [anon_sym_export] = ACTIONS(584),
    [anon_sym_LBRACE] = ACTIONS(434),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(584),
    [anon_sym_await] = ACTIONS(586),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(588),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(592),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(594),
    [anon_sym_PLUS] = ACTIONS(602),
    [anon_sym_DASH] = ACTIONS(602),
    [anon_sym_SLASH] = ACTIONS(678),
    [anon_sym_BANG] = ACTIONS(604),
    [anon_sym_TILDE] = ACTIONS(604),
    [anon_sym_typeof] = ACTIONS(602),
    [anon_sym_void] = ACTIONS(602),
    [anon_sym_delete] = ACTIONS(602),
    [anon_sym_PLUS_PLUS] = ACTIONS(680),
    [anon_sym_DASH_DASH] = ACTIONS(680),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(609),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(611),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(584),
    [anon_sym_get] = ACTIONS(584),
    [anon_sym_set] = ACTIONS(584),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(339)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(579),
    [sym_expression] = STATE(945),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1857),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1857),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(579),
    [sym_subscript_expression] = STATE(579),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1141),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1857),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1779),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(582),
    [anon_sym_export] = ACTIONS(584),
    [anon_sym_LBRACE] = ACTIONS(434),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(584),
    [anon_sym_await] = ACTIONS(586),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(588),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(592),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(594),
    [anon_sym_PLUS] = ACTIONS(602),
    [anon_sym_DASH] = ACTIONS(602),
    [anon_sym_SLASH] = ACTIONS(678),
    [anon_sym_BANG] = ACTIONS(604),
    [anon_sym_TILDE] = ACTIONS(604),
    [anon_sym_typeof] = ACTIONS(602),
    [anon_sym_void] = ACTIONS(602),
    [anon_sym_delete] = ACTIONS(602),
    [anon_sym_PLUS_PLUS] = ACTIONS(680),
    [anon_sym_DASH_DASH] = ACTIONS(680),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(609),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(611),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(584),
    [anon_sym_get] = ACTIONS(584),
    [anon_sym_set] = ACTIONS(584),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(340)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(579),
    [sym_expression] = STATE(947),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1857),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1857),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(579),
    [sym_subscript_expression] = STATE(579),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1141),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1857),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1779),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(582),
    [anon_sym_export] = ACTIONS(584),
    [anon_sym_LBRACE] = ACTIONS(434),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(584),
    [anon_sym_await] = ACTIONS(586),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(588),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(592),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(594),
    [anon_sym_PLUS] = ACTIONS(602),
    [anon_sym_DASH] = ACTIONS(602),
    [anon_sym_SLASH] = ACTIONS(678),
    [anon_sym_BANG] = ACTIONS(604),
    [anon_sym_TILDE] = ACTIONS(604),
    [anon_sym_typeof] = ACTIONS(602),
    [anon_sym_void] = ACTIONS(602),
    [anon_sym_delete] = ACTIONS(602),
    [anon_sym_PLUS_PLUS] = ACTIONS(680),
    [anon_sym_DASH_DASH] = ACTIONS(680),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(609),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(611),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(584),
    [anon_sym_get] = ACTIONS(584),
    [anon_sym_set] = ACTIONS(584),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(341)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(579),
    [sym_expression] = STATE(948),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1857),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1857),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(579),
    [sym_subscript_expression] = STATE(579),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1141),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1857),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1779),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(582),
    [anon_sym_export] = ACTIONS(584),
    [anon_sym_LBRACE] = ACTIONS(434),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(584),
    [anon_sym_await] = ACTIONS(586),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(588),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(592),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(594),
    [anon_sym_PLUS] = ACTIONS(602),
    [anon_sym_DASH] = ACTIONS(602),
    [anon_sym_SLASH] = ACTIONS(678),
    [anon_sym_BANG] = ACTIONS(604),
    [anon_sym_TILDE] = ACTIONS(604),
    [anon_sym_typeof] = ACTIONS(602),
    [anon_sym_void] = ACTIONS(602),
    [anon_sym_delete] = ACTIONS(602),
    [anon_sym_PLUS_PLUS] = ACTIONS(680),
    [anon_sym_DASH_DASH] = ACTIONS(680),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(609),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(611),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(584),
    [anon_sym_get] = ACTIONS(584),
    [anon_sym_set] = ACTIONS(584),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(342)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(579),
    [sym_expression] = STATE(949),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1857),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1857),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(579),
    [sym_subscript_expression] = STATE(579),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1141),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1857),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1779),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(582),
    [anon_sym_export] = ACTIONS(584),
    [anon_sym_LBRACE] = ACTIONS(434),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(584),
    [anon_sym_await] = ACTIONS(586),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(588),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(592),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(594),
    [anon_sym_PLUS] = ACTIONS(602),
    [anon_sym_DASH] = ACTIONS(602),
    [anon_sym_SLASH] = ACTIONS(678),
    [anon_sym_BANG] = ACTIONS(604),
    [anon_sym_TILDE] = ACTIONS(604),
    [anon_sym_typeof] = ACTIONS(602),
    [anon_sym_void] = ACTIONS(602),
    [anon_sym_delete] = ACTIONS(602),
    [anon_sym_PLUS_PLUS] = ACTIONS(680),
    [anon_sym_DASH_DASH] = ACTIONS(680),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(609),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(611),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(584),
    [anon_sym_get] = ACTIONS(584),
    [anon_sym_set] = ACTIONS(584),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(343)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(579),
    [sym_expression] = STATE(950),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1857),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1857),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(579),
    [sym_subscript_expression] = STATE(579),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1141),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1857),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1779),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(582),
    [anon_sym_export] = ACTIONS(584),
    [anon_sym_LBRACE] = ACTIONS(434),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(584),
    [anon_sym_await] = ACTIONS(586),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(588),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(592),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(594),
    [anon_sym_PLUS] = ACTIONS(602),
    [anon_sym_DASH] = ACTIONS(602),
    [anon_sym_SLASH] = ACTIONS(678),
    [anon_sym_BANG] = ACTIONS(604),
    [anon_sym_TILDE] = ACTIONS(604),
    [anon_sym_typeof] = ACTIONS(602),
    [anon_sym_void] = ACTIONS(602),
    [anon_sym_delete] = ACTIONS(602),
    [anon_sym_PLUS_PLUS] = ACTIONS(680),
    [anon_sym_DASH_DASH] = ACTIONS(680),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(609),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(611),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(584),
    [anon_sym_get] = ACTIONS(584),
    [anon_sym_set] = ACTIONS(584),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(344)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(579),
    [sym_expression] = STATE(951),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1857),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1857),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(579),
    [sym_subscript_expression] = STATE(579),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1141),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1857),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1779),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(582),
    [anon_sym_export] = ACTIONS(584),
    [anon_sym_LBRACE] = ACTIONS(434),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(584),
    [anon_sym_await] = ACTIONS(586),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(588),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(592),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(594),
    [anon_sym_PLUS] = ACTIONS(602),
    [anon_sym_DASH] = ACTIONS(602),
    [anon_sym_SLASH] = ACTIONS(678),
    [anon_sym_BANG] = ACTIONS(604),
    [anon_sym_TILDE] = ACTIONS(604),
    [anon_sym_typeof] = ACTIONS(602),
    [anon_sym_void] = ACTIONS(602),
    [anon_sym_delete] = ACTIONS(602),
    [anon_sym_PLUS_PLUS] = ACTIONS(680),
    [anon_sym_DASH_DASH] = ACTIONS(680),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(609),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(611),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(584),
    [anon_sym_get] = ACTIONS(584),
    [anon_sym_set] = ACTIONS(584),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(345)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(579),
    [sym_expression] = STATE(952),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1857),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1857),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(579),
    [sym_subscript_expression] = STATE(579),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1141),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1857),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1779),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(582),
    [anon_sym_export] = ACTIONS(584),
    [anon_sym_LBRACE] = ACTIONS(434),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(584),
    [anon_sym_await] = ACTIONS(586),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(588),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(592),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(594),
    [anon_sym_PLUS] = ACTIONS(602),
    [anon_sym_DASH] = ACTIONS(602),
    [anon_sym_SLASH] = ACTIONS(678),
    [anon_sym_BANG] = ACTIONS(604),
    [anon_sym_TILDE] = ACTIONS(604),
    [anon_sym_typeof] = ACTIONS(602),
    [anon_sym_void] = ACTIONS(602),
    [anon_sym_delete] = ACTIONS(602),
    [anon_sym_PLUS_PLUS] = ACTIONS(680),
    [anon_sym_DASH_DASH] = ACTIONS(680),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(609),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(611),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(584),
    [anon_sym_get] = ACTIONS(584),
    [anon_sym_set] = ACTIONS(584),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(346)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(579),
    [sym_expression] = STATE(953),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1857),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1857),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(579),
    [sym_subscript_expression] = STATE(579),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1141),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1857),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1779),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(582),
    [anon_sym_export] = ACTIONS(584),
    [anon_sym_LBRACE] = ACTIONS(434),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(584),
    [anon_sym_await] = ACTIONS(586),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(588),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(592),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(594),
    [anon_sym_PLUS] = ACTIONS(602),
    [anon_sym_DASH] = ACTIONS(602),
    [anon_sym_SLASH] = ACTIONS(678),
    [anon_sym_BANG] = ACTIONS(604),
    [anon_sym_TILDE] = ACTIONS(604),
    [anon_sym_typeof] = ACTIONS(602),
    [anon_sym_void] = ACTIONS(602),
    [anon_sym_delete] = ACTIONS(602),
    [anon_sym_PLUS_PLUS] = ACTIONS(680),
    [anon_sym_DASH_DASH] = ACTIONS(680),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(609),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(611),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(584),
    [anon_sym_get] = ACTIONS(584),
    [anon_sym_set] = ACTIONS(584),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(347)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(579),
    [sym_expression] = STATE(954),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1857),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1857),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(579),
    [sym_subscript_expression] = STATE(579),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1141),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1857),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1779),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(582),
    [anon_sym_export] = ACTIONS(584),
    [anon_sym_LBRACE] = ACTIONS(434),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(584),
    [anon_sym_await] = ACTIONS(586),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(588),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(592),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(594),
    [anon_sym_PLUS] = ACTIONS(602),
    [anon_sym_DASH] = ACTIONS(602),
    [anon_sym_SLASH] = ACTIONS(678),
    [anon_sym_BANG] = ACTIONS(604),
    [anon_sym_TILDE] = ACTIONS(604),
    [anon_sym_typeof] = ACTIONS(602),
    [anon_sym_void] = ACTIONS(602),
    [anon_sym_delete] = ACTIONS(602),
    [anon_sym_PLUS_PLUS] = ACTIONS(680),
    [anon_sym_DASH_DASH] = ACTIONS(680),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(609),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(611),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(584),
    [anon_sym_get] = ACTIONS(584),
    [anon_sym_set] = ACTIONS(584),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(348)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(579),
    [sym_expression] = STATE(955),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1857),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1857),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(579),
    [sym_subscript_expression] = STATE(579),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1141),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1857),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1779),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(582),
    [anon_sym_export] = ACTIONS(584),
    [anon_sym_LBRACE] = ACTIONS(434),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(584),
    [anon_sym_await] = ACTIONS(586),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(588),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(592),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(594),
    [anon_sym_PLUS] = ACTIONS(602),
    [anon_sym_DASH] = ACTIONS(602),
    [anon_sym_SLASH] = ACTIONS(678),
    [anon_sym_BANG] = ACTIONS(604),
    [anon_sym_TILDE] = ACTIONS(604),
    [anon_sym_typeof] = ACTIONS(602),
    [anon_sym_void] = ACTIONS(602),
    [anon_sym_delete] = ACTIONS(602),
    [anon_sym_PLUS_PLUS] = ACTIONS(680),
    [anon_sym_DASH_DASH] = ACTIONS(680),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(609),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(611),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(584),
    [anon_sym_get] = ACTIONS(584),
    [anon_sym_set] = ACTIONS(584),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(349)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(579),
    [sym_expression] = STATE(956),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1857),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1857),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(579),
    [sym_subscript_expression] = STATE(579),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1141),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1857),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1779),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(582),
    [anon_sym_export] = ACTIONS(584),
    [anon_sym_LBRACE] = ACTIONS(434),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(584),
    [anon_sym_await] = ACTIONS(586),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(588),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(592),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(594),
    [anon_sym_PLUS] = ACTIONS(602),
    [anon_sym_DASH] = ACTIONS(602),
    [anon_sym_SLASH] = ACTIONS(678),
    [anon_sym_BANG] = ACTIONS(604),
    [anon_sym_TILDE] = ACTIONS(604),
    [anon_sym_typeof] = ACTIONS(602),
    [anon_sym_void] = ACTIONS(602),
    [anon_sym_delete] = ACTIONS(602),
    [anon_sym_PLUS_PLUS] = ACTIONS(680),
    [anon_sym_DASH_DASH] = ACTIONS(680),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(609),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(611),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(584),
    [anon_sym_get] = ACTIONS(584),
    [anon_sym_set] = ACTIONS(584),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(350)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(579),
    [sym_expression] = STATE(957),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1857),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1857),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(579),
    [sym_subscript_expression] = STATE(579),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1141),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1857),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1779),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(582),
    [anon_sym_export] = ACTIONS(584),
    [anon_sym_LBRACE] = ACTIONS(434),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(584),
    [anon_sym_await] = ACTIONS(586),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(588),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(592),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(594),
    [anon_sym_PLUS] = ACTIONS(602),
    [anon_sym_DASH] = ACTIONS(602),
    [anon_sym_SLASH] = ACTIONS(678),
    [anon_sym_BANG] = ACTIONS(604),
    [anon_sym_TILDE] = ACTIONS(604),
    [anon_sym_typeof] = ACTIONS(602),
    [anon_sym_void] = ACTIONS(602),
    [anon_sym_delete] = ACTIONS(602),
    [anon_sym_PLUS_PLUS] = ACTIONS(680),
    [anon_sym_DASH_DASH] = ACTIONS(680),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(609),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(611),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(584),
    [anon_sym_get] = ACTIONS(584),
    [anon_sym_set] = ACTIONS(584),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(351)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(579),
    [sym_expression] = STATE(958),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1857),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1857),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(579),
    [sym_subscript_expression] = STATE(579),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1141),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1857),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1779),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(582),
    [anon_sym_export] = ACTIONS(584),
    [anon_sym_LBRACE] = ACTIONS(434),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(584),
    [anon_sym_await] = ACTIONS(586),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(588),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(592),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(594),
    [anon_sym_PLUS] = ACTIONS(602),
    [anon_sym_DASH] = ACTIONS(602),
    [anon_sym_SLASH] = ACTIONS(678),
    [anon_sym_BANG] = ACTIONS(604),
    [anon_sym_TILDE] = ACTIONS(604),
    [anon_sym_typeof] = ACTIONS(602),
    [anon_sym_void] = ACTIONS(602),
    [anon_sym_delete] = ACTIONS(602),
    [anon_sym_PLUS_PLUS] = ACTIONS(680),
    [anon_sym_DASH_DASH] = ACTIONS(680),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(609),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(611),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(584),
    [anon_sym_get] = ACTIONS(584),
    [anon_sym_set] = ACTIONS(584),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(352)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(579),
    [sym_expression] = STATE(959),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1857),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1857),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(579),
    [sym_subscript_expression] = STATE(579),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1141),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1857),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1779),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(582),
    [anon_sym_export] = ACTIONS(584),
    [anon_sym_LBRACE] = ACTIONS(434),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(584),
    [anon_sym_await] = ACTIONS(586),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(588),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(592),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(594),
    [anon_sym_PLUS] = ACTIONS(602),
    [anon_sym_DASH] = ACTIONS(602),
    [anon_sym_SLASH] = ACTIONS(678),
    [anon_sym_BANG] = ACTIONS(604),
    [anon_sym_TILDE] = ACTIONS(604),
    [anon_sym_typeof] = ACTIONS(602),
    [anon_sym_void] = ACTIONS(602),
    [anon_sym_delete] = ACTIONS(602),
    [anon_sym_PLUS_PLUS] = ACTIONS(680),
    [anon_sym_DASH_DASH] = ACTIONS(680),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(609),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(611),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(584),
    [anon_sym_get] = ACTIONS(584),
    [anon_sym_set] = ACTIONS(584),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(353)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(579),
    [sym_expression] = STATE(964),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1857),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1857),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(579),
    [sym_subscript_expression] = STATE(579),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1141),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1857),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1779),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(582),
    [anon_sym_export] = ACTIONS(584),
    [anon_sym_LBRACE] = ACTIONS(434),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(584),
    [anon_sym_await] = ACTIONS(586),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(588),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(592),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(594),
    [anon_sym_PLUS] = ACTIONS(602),
    [anon_sym_DASH] = ACTIONS(602),
    [anon_sym_SLASH] = ACTIONS(678),
    [anon_sym_BANG] = ACTIONS(604),
    [anon_sym_TILDE] = ACTIONS(604),
    [anon_sym_typeof] = ACTIONS(602),
    [anon_sym_void] = ACTIONS(602),
    [anon_sym_delete] = ACTIONS(602),
    [anon_sym_PLUS_PLUS] = ACTIONS(680),
    [anon_sym_DASH_DASH] = ACTIONS(680),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(609),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(611),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(584),
    [anon_sym_get] = ACTIONS(584),
    [anon_sym_set] = ACTIONS(584),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(354)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(498),
    [sym_expression] = STATE(970),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1762),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1762),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(498),
    [sym_subscript_expression] = STATE(498),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1138),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1762),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1816),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(430),
    [anon_sym_export] = ACTIONS(432),
    [anon_sym_LBRACE] = ACTIONS(434),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(432),
    [anon_sym_await] = ACTIONS(436),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(438),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(445),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(447),
    [anon_sym_PLUS] = ACTIONS(452),
    [anon_sym_DASH] = ACTIONS(452),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(454),
    [anon_sym_TILDE] = ACTIONS(454),
    [anon_sym_typeof] = ACTIONS(452),
    [anon_sym_void] = ACTIONS(452),
    [anon_sym_delete] = ACTIONS(452),
    [anon_sym_PLUS_PLUS] = ACTIONS(684),
    [anon_sym_DASH_DASH] = ACTIONS(684),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(459),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(461),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(432),
    [anon_sym_get] = ACTIONS(432),
    [anon_sym_set] = ACTIONS(432),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(355)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(498),
    [sym_expression] = STATE(977),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1762),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1762),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(498),
    [sym_subscript_expression] = STATE(498),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1138),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1762),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1816),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(430),
    [anon_sym_export] = ACTIONS(432),
    [anon_sym_LBRACE] = ACTIONS(434),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(432),
    [anon_sym_await] = ACTIONS(436),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(438),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(445),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(447),
    [anon_sym_PLUS] = ACTIONS(452),
    [anon_sym_DASH] = ACTIONS(452),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(454),
    [anon_sym_TILDE] = ACTIONS(454),
    [anon_sym_typeof] = ACTIONS(452),
    [anon_sym_void] = ACTIONS(452),
    [anon_sym_delete] = ACTIONS(452),
    [anon_sym_PLUS_PLUS] = ACTIONS(684),
    [anon_sym_DASH_DASH] = ACTIONS(684),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(459),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(461),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(432),
    [anon_sym_get] = ACTIONS(432),
    [anon_sym_set] = ACTIONS(432),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(356)] = {
    [sym_import] = STATE(1333),
    [sym_parenthesized_expression] = STATE(578),
    [sym_expression] = STATE(892),
    [sym_primary_expression] = STATE(589),
    [sym_yield_expression] = STATE(602),
    [sym_object] = STATE(646),
    [sym_object_pattern] = STATE(1776),
    [sym_array] = STATE(646),
    [sym_array_pattern] = STATE(1776),
    [sym_jsx_element] = STATE(602),
    [sym_jsx_opening_element] = STATE(1153),
    [sym_jsx_self_closing_element] = STATE(602),
    [sym_class] = STATE(646),
    [sym_function_expression] = STATE(646),
    [sym_generator_function] = STATE(646),
    [sym_arrow_function] = STATE(646),
    [sym_call_expression] = STATE(646),
    [sym_new_expression] = STATE(585),
    [sym_await_expression] = STATE(602),
    [sym_member_expression] = STATE(578),
    [sym_subscript_expression] = STATE(578),
    [sym_assignment_expression] = STATE(602),
    [sym__augmented_assignment_lhs] = STATE(1142),
    [sym_augmented_assignment_expression] = STATE(602),
    [sym__destructuring_pattern] = STATE(1776),
    [sym_ternary_expression] = STATE(602),
    [sym_binary_expression] = STATE(602),
    [sym_unary_expression] = STATE(602),
    [sym_update_expression] = STATE(602),
    [sym_string] = STATE(646),
    [sym_template_string] = STATE(646),
    [sym_regex] = STATE(646),
    [sym_meta_property] = STATE(646),
    [sym_decorator] = STATE(1097),
    [sym_formal_parameters] = STATE(1778),
    [aux_sym_export_statement_repeat1] = STATE(1388),
    [sym_identifier] = ACTIONS(613),
    [anon_sym_export] = ACTIONS(615),
    [anon_sym_LBRACE] = ACTIONS(434),
    [anon_sym_import] = ACTIONS(375),
    [anon_sym_let] = ACTIONS(615),
    [anon_sym_await] = ACTIONS(620),
    [anon_sym_LPAREN] = ACTIONS(650),
    [anon_sym_yield] = ACTIONS(622),
    [anon_sym_LBRACK] = ACTIONS(676),
    [anon_sym_LT] = ACTIONS(820),
    [anon_sym_DQUOTE] = ACTIONS(392),
    [anon_sym_SQUOTE] = ACTIONS(394),
    [anon_sym_class] = ACTIONS(396),
    [anon_sym_async] = ACTIONS(626),
    [anon_sym_function] = ACTIONS(400),
    [anon_sym_new] = ACTIONS(630),
    [anon_sym_PLUS] = ACTIONS(635),
    [anon_sym_DASH] = ACTIONS(635),
    [anon_sym_SLASH] = ACTIONS(658),
    [anon_sym_BANG] = ACTIONS(637),
    [anon_sym_TILDE] = ACTIONS(637),
    [anon_sym_typeof] = ACTIONS(635),
    [anon_sym_void] = ACTIONS(635),
    [anon_sym_delete] = ACTIONS(635),
    [anon_sym_PLUS_PLUS] = ACTIONS(688),
    [anon_sym_DASH_DASH] = ACTIONS(688),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(662),
    [sym_number] = ACTIONS(422),
    [sym_private_property_identifier] = ACTIONS(642),
    [sym_this] = ACTIONS(426),
    [sym_super] = ACTIONS(426),
    [sym_true] = ACTIONS(426),
    [sym_false] = ACTIONS(426),
    [sym_null] = ACTIONS(426),
    [sym_undefined] = ACTIONS(644),
    [anon_sym_AT] = ACTIONS(93),
    [anon_sym_static] = ACTIONS(615),
    [anon_sym_get] = ACTIONS(615),
    [anon_sym_set] = ACTIONS(615),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(357)] = {
    [sym_string] = STATE(1710),
    [sym_formal_parameters] = STATE(1834),
    [sym__property_name] = STATE(1710),
    [sym_computed_property_name] = STATE(1710),
    [aux_sym_object_repeat1] = STATE(1418),
    [aux_sym_object_pattern_repeat1] = STATE(1452),
    [sym_identifier] = ACTIONS(1061),
    [anon_sym_export] = ACTIONS(1063),
    [anon_sym_STAR] = ACTIONS(1065),
    [anon_sym_COMMA] = ACTIONS(169),
    [anon_sym_RBRACE] = ACTIONS(351),
    [anon_sym_let] = ACTIONS(1063),
    [anon_sym_await] = ACTIONS(1063),
    [anon_sym_LPAREN] = ACTIONS(1068),
    [anon_sym_SEMI] = ACTIONS(169),
    [anon_sym_in] = ACTIONS(165),
    [anon_sym_COLON] = ACTIONS(183),
    [anon_sym_EQ] = ACTIONS(186),
    [anon_sym_LBRACK] = ACTIONS(1072),
    [anon_sym_LT] = ACTIONS(165),
    [anon_sym_GT] = ACTIONS(165),
    [anon_sym_DOT] = ACTIONS(165),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [anon_sym_SQUOTE] = ACTIONS(1077),
    [anon_sym_async] = ACTIONS(1063),
    [anon_sym_function] = ACTIONS(1079),
    [anon_sym_EQ_GT] = ACTIONS(200),
    [sym_optional_chain] = ACTIONS(169),
    [anon_sym_PLUS_EQ] = ACTIONS(202),
    [anon_sym_DASH_EQ] = ACTIONS(202),
    [anon_sym_STAR_EQ] = ACTIONS(202),
    [anon_sym_SLASH_EQ] = ACTIONS(202),
    [anon_sym_PERCENT_EQ] = ACTIONS(202),
    [anon_sym_CARET_EQ] = ACTIONS(202),
    [anon_sym_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_LT_LT_EQ] = ACTIONS(202),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(202),
    [anon_sym_QMARK_QMARK_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP] = ACTIONS(165),
    [anon_sym_PIPE_PIPE] = ACTIONS(165),
    [anon_sym_GT_GT] = ACTIONS(165),
    [anon_sym_GT_GT_GT] = ACTIONS(165),
    [anon_sym_LT_LT] = ACTIONS(165),
    [anon_sym_AMP] = ACTIONS(165),
    [anon_sym_CARET] = ACTIONS(165),
    [anon_sym_PIPE] = ACTIONS(165),
    [anon_sym_PLUS] = ACTIONS(165),
    [anon_sym_DASH] = ACTIONS(165),
    [anon_sym_SLASH] = ACTIONS(165),
    [anon_sym_PERCENT] = ACTIONS(165),
    [anon_sym_STAR_STAR] = ACTIONS(165),
    [anon_sym_LT_EQ] = ACTIONS(169),
    [anon_sym_EQ_EQ] = ACTIONS(165),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(169),
    [anon_sym_BANG_EQ] = ACTIONS(165),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(169),
    [anon_sym_GT_EQ] = ACTIONS(169),
    [anon_sym_QMARK_QMARK] = ACTIONS(165),
    [anon_sym_instanceof] = ACTIONS(165),
    [anon_sym_PLUS_PLUS] = ACTIONS(169),
    [anon_sym_DASH_DASH] = ACTIONS(169),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(169),
    [sym_number] = ACTIONS(1081),
    [sym_private_property_identifier] = ACTIONS(1081),
    [anon_sym_static] = ACTIONS(1063),
    [anon_sym_get] = ACTIONS(1083),
    [anon_sym_set] = ACTIONS(1083),
    [sym__automatic_semicolon] = ACTIONS(169),
    [sym__ternary_qmark] = ACTIONS(169),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(358)] = {
    [sym_string] = STATE(1710),
    [sym_formal_parameters] = STATE(1834),
    [sym__property_name] = STATE(1710),
    [sym_computed_property_name] = STATE(1710),
    [aux_sym_object_repeat1] = STATE(1487),
    [aux_sym_object_pattern_repeat1] = STATE(1452),
    [sym_identifier] = ACTIONS(1061),
    [anon_sym_export] = ACTIONS(1063),
    [anon_sym_STAR] = ACTIONS(1065),
    [anon_sym_COMMA] = ACTIONS(169),
    [anon_sym_RBRACE] = ACTIONS(171),
    [anon_sym_let] = ACTIONS(1063),
    [anon_sym_await] = ACTIONS(1063),
    [anon_sym_LPAREN] = ACTIONS(1068),
    [anon_sym_SEMI] = ACTIONS(169),
    [anon_sym_in] = ACTIONS(165),
    [anon_sym_COLON] = ACTIONS(183),
    [anon_sym_EQ] = ACTIONS(186),
    [anon_sym_LBRACK] = ACTIONS(1072),
    [anon_sym_LT] = ACTIONS(165),
    [anon_sym_GT] = ACTIONS(165),
    [anon_sym_DOT] = ACTIONS(165),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [anon_sym_SQUOTE] = ACTIONS(1077),
    [anon_sym_async] = ACTIONS(1063),
    [anon_sym_function] = ACTIONS(1079),
    [anon_sym_EQ_GT] = ACTIONS(200),
    [sym_optional_chain] = ACTIONS(169),
    [anon_sym_PLUS_EQ] = ACTIONS(202),
    [anon_sym_DASH_EQ] = ACTIONS(202),
    [anon_sym_STAR_EQ] = ACTIONS(202),
    [anon_sym_SLASH_EQ] = ACTIONS(202),
    [anon_sym_PERCENT_EQ] = ACTIONS(202),
    [anon_sym_CARET_EQ] = ACTIONS(202),
    [anon_sym_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_LT_LT_EQ] = ACTIONS(202),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(202),
    [anon_sym_QMARK_QMARK_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP] = ACTIONS(165),
    [anon_sym_PIPE_PIPE] = ACTIONS(165),
    [anon_sym_GT_GT] = ACTIONS(165),
    [anon_sym_GT_GT_GT] = ACTIONS(165),
    [anon_sym_LT_LT] = ACTIONS(165),
    [anon_sym_AMP] = ACTIONS(165),
    [anon_sym_CARET] = ACTIONS(165),
    [anon_sym_PIPE] = ACTIONS(165),
    [anon_sym_PLUS] = ACTIONS(165),
    [anon_sym_DASH] = ACTIONS(165),
    [anon_sym_SLASH] = ACTIONS(165),
    [anon_sym_PERCENT] = ACTIONS(165),
    [anon_sym_STAR_STAR] = ACTIONS(165),
    [anon_sym_LT_EQ] = ACTIONS(169),
    [anon_sym_EQ_EQ] = ACTIONS(165),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(169),
    [anon_sym_BANG_EQ] = ACTIONS(165),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(169),
    [anon_sym_GT_EQ] = ACTIONS(169),
    [anon_sym_QMARK_QMARK] = ACTIONS(165),
    [anon_sym_instanceof] = ACTIONS(165),
    [anon_sym_PLUS_PLUS] = ACTIONS(169),
    [anon_sym_DASH_DASH] = ACTIONS(169),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(169),
    [sym_number] = ACTIONS(1081),
    [sym_private_property_identifier] = ACTIONS(1081),
    [anon_sym_static] = ACTIONS(1063),
    [anon_sym_get] = ACTIONS(1083),
    [anon_sym_set] = ACTIONS(1083),
    [sym__automatic_semicolon] = ACTIONS(169),
    [sym__ternary_qmark] = ACTIONS(169),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(359)] = {
    [sym_string] = STATE(1710),
    [sym_formal_parameters] = STATE(1834),
    [sym__property_name] = STATE(1710),
    [sym_computed_property_name] = STATE(1710),
    [aux_sym_object_repeat1] = STATE(1418),
    [aux_sym_object_pattern_repeat1] = STATE(1452),
    [sym_identifier] = ACTIONS(1061),
    [anon_sym_export] = ACTIONS(1063),
    [anon_sym_STAR] = ACTIONS(1065),
    [anon_sym_COMMA] = ACTIONS(169),
    [anon_sym_RBRACE] = ACTIONS(216),
    [anon_sym_let] = ACTIONS(1063),
    [anon_sym_await] = ACTIONS(1063),
    [anon_sym_LPAREN] = ACTIONS(1068),
    [anon_sym_SEMI] = ACTIONS(169),
    [anon_sym_in] = ACTIONS(165),
    [anon_sym_COLON] = ACTIONS(183),
    [anon_sym_EQ] = ACTIONS(186),
    [anon_sym_LBRACK] = ACTIONS(1072),
    [anon_sym_LT] = ACTIONS(165),
    [anon_sym_GT] = ACTIONS(165),
    [anon_sym_DOT] = ACTIONS(165),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [anon_sym_SQUOTE] = ACTIONS(1077),
    [anon_sym_async] = ACTIONS(1063),
    [anon_sym_function] = ACTIONS(1079),
    [anon_sym_EQ_GT] = ACTIONS(200),
    [sym_optional_chain] = ACTIONS(169),
    [anon_sym_PLUS_EQ] = ACTIONS(202),
    [anon_sym_DASH_EQ] = ACTIONS(202),
    [anon_sym_STAR_EQ] = ACTIONS(202),
    [anon_sym_SLASH_EQ] = ACTIONS(202),
    [anon_sym_PERCENT_EQ] = ACTIONS(202),
    [anon_sym_CARET_EQ] = ACTIONS(202),
    [anon_sym_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_LT_LT_EQ] = ACTIONS(202),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(202),
    [anon_sym_QMARK_QMARK_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP] = ACTIONS(165),
    [anon_sym_PIPE_PIPE] = ACTIONS(165),
    [anon_sym_GT_GT] = ACTIONS(165),
    [anon_sym_GT_GT_GT] = ACTIONS(165),
    [anon_sym_LT_LT] = ACTIONS(165),
    [anon_sym_AMP] = ACTIONS(165),
    [anon_sym_CARET] = ACTIONS(165),
    [anon_sym_PIPE] = ACTIONS(165),
    [anon_sym_PLUS] = ACTIONS(165),
    [anon_sym_DASH] = ACTIONS(165),
    [anon_sym_SLASH] = ACTIONS(165),
    [anon_sym_PERCENT] = ACTIONS(165),
    [anon_sym_STAR_STAR] = ACTIONS(165),
    [anon_sym_LT_EQ] = ACTIONS(169),
    [anon_sym_EQ_EQ] = ACTIONS(165),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(169),
    [anon_sym_BANG_EQ] = ACTIONS(165),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(169),
    [anon_sym_GT_EQ] = ACTIONS(169),
    [anon_sym_QMARK_QMARK] = ACTIONS(165),
    [anon_sym_instanceof] = ACTIONS(165),
    [anon_sym_PLUS_PLUS] = ACTIONS(169),
    [anon_sym_DASH_DASH] = ACTIONS(169),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(169),
    [sym_number] = ACTIONS(1081),
    [sym_private_property_identifier] = ACTIONS(1081),
    [anon_sym_static] = ACTIONS(1063),
    [anon_sym_get] = ACTIONS(1083),
    [anon_sym_set] = ACTIONS(1083),
    [sym__automatic_semicolon] = ACTIONS(169),
    [sym__ternary_qmark] = ACTIONS(169),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(360)] = {
    [sym_string] = STATE(1710),
    [sym__property_name] = STATE(1710),
    [sym_computed_property_name] = STATE(1710),
    [aux_sym_object_repeat1] = STATE(1487),
    [aux_sym_object_pattern_repeat1] = STATE(1452),
    [sym_identifier] = ACTIONS(1085),
    [anon_sym_export] = ACTIONS(1085),
    [anon_sym_STAR] = ACTIONS(165),
    [anon_sym_COMMA] = ACTIONS(169),
    [anon_sym_RBRACE] = ACTIONS(171),
    [anon_sym_let] = ACTIONS(1085),
    [anon_sym_await] = ACTIONS(1085),
    [anon_sym_LPAREN] = ACTIONS(944),
    [anon_sym_SEMI] = ACTIONS(169),
    [anon_sym_in] = ACTIONS(165),
    [anon_sym_COLON] = ACTIONS(183),
    [anon_sym_EQ] = ACTIONS(186),
    [anon_sym_LBRACK] = ACTIONS(1072),
    [anon_sym_LT] = ACTIONS(165),
    [anon_sym_GT] = ACTIONS(165),
    [anon_sym_DOT] = ACTIONS(165),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [anon_sym_SQUOTE] = ACTIONS(1077),
    [anon_sym_async] = ACTIONS(1085),
    [anon_sym_EQ_GT] = ACTIONS(200),
    [sym_optional_chain] = ACTIONS(169),
    [anon_sym_PLUS_EQ] = ACTIONS(202),
    [anon_sym_DASH_EQ] = ACTIONS(202),
    [anon_sym_STAR_EQ] = ACTIONS(202),
    [anon_sym_SLASH_EQ] = ACTIONS(202),
    [anon_sym_PERCENT_EQ] = ACTIONS(202),
    [anon_sym_CARET_EQ] = ACTIONS(202),
    [anon_sym_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_LT_LT_EQ] = ACTIONS(202),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(202),
    [anon_sym_QMARK_QMARK_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP] = ACTIONS(165),
    [anon_sym_PIPE_PIPE] = ACTIONS(165),
    [anon_sym_GT_GT] = ACTIONS(165),
    [anon_sym_GT_GT_GT] = ACTIONS(165),
    [anon_sym_LT_LT] = ACTIONS(165),
    [anon_sym_AMP] = ACTIONS(165),
    [anon_sym_CARET] = ACTIONS(165),
    [anon_sym_PIPE] = ACTIONS(165),
    [anon_sym_PLUS] = ACTIONS(165),
    [anon_sym_DASH] = ACTIONS(165),
    [anon_sym_SLASH] = ACTIONS(165),
    [anon_sym_PERCENT] = ACTIONS(165),
    [anon_sym_STAR_STAR] = ACTIONS(165),
    [anon_sym_LT_EQ] = ACTIONS(169),
    [anon_sym_EQ_EQ] = ACTIONS(165),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(169),
    [anon_sym_BANG_EQ] = ACTIONS(165),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(169),
    [anon_sym_GT_EQ] = ACTIONS(169),
    [anon_sym_QMARK_QMARK] = ACTIONS(165),
    [anon_sym_instanceof] = ACTIONS(165),
    [anon_sym_PLUS_PLUS] = ACTIONS(169),
    [anon_sym_DASH_DASH] = ACTIONS(169),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(169),
    [sym_number] = ACTIONS(1081),
    [sym_private_property_identifier] = ACTIONS(1081),
    [anon_sym_static] = ACTIONS(1085),
    [anon_sym_get] = ACTIONS(1085),
    [anon_sym_set] = ACTIONS(1085),
    [sym__automatic_semicolon] = ACTIONS(169),
    [sym__ternary_qmark] = ACTIONS(169),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(361)] = {
    [sym_string] = STATE(1710),
    [sym__property_name] = STATE(1710),
    [sym_computed_property_name] = STATE(1710),
    [aux_sym_object_repeat1] = STATE(1418),
    [aux_sym_object_pattern_repeat1] = STATE(1452),
    [sym_identifier] = ACTIONS(1085),
    [anon_sym_export] = ACTIONS(1085),
    [anon_sym_STAR] = ACTIONS(1065),
    [anon_sym_COMMA] = ACTIONS(169),
    [anon_sym_RBRACE] = ACTIONS(351),
    [anon_sym_let] = ACTIONS(1085),
    [anon_sym_await] = ACTIONS(1085),
    [anon_sym_LPAREN] = ACTIONS(944),
    [anon_sym_SEMI] = ACTIONS(169),
    [anon_sym_in] = ACTIONS(165),
    [anon_sym_COLON] = ACTIONS(183),
    [anon_sym_EQ] = ACTIONS(186),
    [anon_sym_LBRACK] = ACTIONS(1072),
    [anon_sym_LT] = ACTIONS(165),
    [anon_sym_GT] = ACTIONS(165),
    [anon_sym_DOT] = ACTIONS(165),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [anon_sym_SQUOTE] = ACTIONS(1077),
    [anon_sym_async] = ACTIONS(1087),
    [anon_sym_EQ_GT] = ACTIONS(200),
    [sym_optional_chain] = ACTIONS(169),
    [anon_sym_PLUS_EQ] = ACTIONS(202),
    [anon_sym_DASH_EQ] = ACTIONS(202),
    [anon_sym_STAR_EQ] = ACTIONS(202),
    [anon_sym_SLASH_EQ] = ACTIONS(202),
    [anon_sym_PERCENT_EQ] = ACTIONS(202),
    [anon_sym_CARET_EQ] = ACTIONS(202),
    [anon_sym_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_LT_LT_EQ] = ACTIONS(202),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(202),
    [anon_sym_QMARK_QMARK_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP] = ACTIONS(165),
    [anon_sym_PIPE_PIPE] = ACTIONS(165),
    [anon_sym_GT_GT] = ACTIONS(165),
    [anon_sym_GT_GT_GT] = ACTIONS(165),
    [anon_sym_LT_LT] = ACTIONS(165),
    [anon_sym_AMP] = ACTIONS(165),
    [anon_sym_CARET] = ACTIONS(165),
    [anon_sym_PIPE] = ACTIONS(165),
    [anon_sym_PLUS] = ACTIONS(165),
    [anon_sym_DASH] = ACTIONS(165),
    [anon_sym_SLASH] = ACTIONS(165),
    [anon_sym_PERCENT] = ACTIONS(165),
    [anon_sym_STAR_STAR] = ACTIONS(165),
    [anon_sym_LT_EQ] = ACTIONS(169),
    [anon_sym_EQ_EQ] = ACTIONS(165),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(169),
    [anon_sym_BANG_EQ] = ACTIONS(165),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(169),
    [anon_sym_GT_EQ] = ACTIONS(169),
    [anon_sym_QMARK_QMARK] = ACTIONS(165),
    [anon_sym_instanceof] = ACTIONS(165),
    [anon_sym_PLUS_PLUS] = ACTIONS(169),
    [anon_sym_DASH_DASH] = ACTIONS(169),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(169),
    [sym_number] = ACTIONS(1081),
    [sym_private_property_identifier] = ACTIONS(1081),
    [anon_sym_static] = ACTIONS(1085),
    [anon_sym_get] = ACTIONS(1089),
    [anon_sym_set] = ACTIONS(1089),
    [sym__automatic_semicolon] = ACTIONS(169),
    [sym__ternary_qmark] = ACTIONS(169),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(362)] = {
    [sym_string] = STATE(1710),
    [sym__property_name] = STATE(1710),
    [sym_computed_property_name] = STATE(1710),
    [aux_sym_object_repeat1] = STATE(1418),
    [aux_sym_object_pattern_repeat1] = STATE(1452),
    [sym_identifier] = ACTIONS(1085),
    [anon_sym_export] = ACTIONS(1085),
    [anon_sym_STAR] = ACTIONS(165),
    [anon_sym_COMMA] = ACTIONS(169),
    [anon_sym_RBRACE] = ACTIONS(351),
    [anon_sym_let] = ACTIONS(1085),
    [anon_sym_await] = ACTIONS(1085),
    [anon_sym_LPAREN] = ACTIONS(944),
    [anon_sym_SEMI] = ACTIONS(169),
    [anon_sym_in] = ACTIONS(165),
    [anon_sym_COLON] = ACTIONS(183),
    [anon_sym_EQ] = ACTIONS(186),
    [anon_sym_LBRACK] = ACTIONS(1072),
    [anon_sym_LT] = ACTIONS(165),
    [anon_sym_GT] = ACTIONS(165),
    [anon_sym_DOT] = ACTIONS(165),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [anon_sym_SQUOTE] = ACTIONS(1077),
    [anon_sym_async] = ACTIONS(1085),
    [anon_sym_EQ_GT] = ACTIONS(200),
    [sym_optional_chain] = ACTIONS(169),
    [anon_sym_PLUS_EQ] = ACTIONS(202),
    [anon_sym_DASH_EQ] = ACTIONS(202),
    [anon_sym_STAR_EQ] = ACTIONS(202),
    [anon_sym_SLASH_EQ] = ACTIONS(202),
    [anon_sym_PERCENT_EQ] = ACTIONS(202),
    [anon_sym_CARET_EQ] = ACTIONS(202),
    [anon_sym_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_LT_LT_EQ] = ACTIONS(202),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(202),
    [anon_sym_QMARK_QMARK_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP] = ACTIONS(165),
    [anon_sym_PIPE_PIPE] = ACTIONS(165),
    [anon_sym_GT_GT] = ACTIONS(165),
    [anon_sym_GT_GT_GT] = ACTIONS(165),
    [anon_sym_LT_LT] = ACTIONS(165),
    [anon_sym_AMP] = ACTIONS(165),
    [anon_sym_CARET] = ACTIONS(165),
    [anon_sym_PIPE] = ACTIONS(165),
    [anon_sym_PLUS] = ACTIONS(165),
    [anon_sym_DASH] = ACTIONS(165),
    [anon_sym_SLASH] = ACTIONS(165),
    [anon_sym_PERCENT] = ACTIONS(165),
    [anon_sym_STAR_STAR] = ACTIONS(165),
    [anon_sym_LT_EQ] = ACTIONS(169),
    [anon_sym_EQ_EQ] = ACTIONS(165),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(169),
    [anon_sym_BANG_EQ] = ACTIONS(165),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(169),
    [anon_sym_GT_EQ] = ACTIONS(169),
    [anon_sym_QMARK_QMARK] = ACTIONS(165),
    [anon_sym_instanceof] = ACTIONS(165),
    [anon_sym_PLUS_PLUS] = ACTIONS(169),
    [anon_sym_DASH_DASH] = ACTIONS(169),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(169),
    [sym_number] = ACTIONS(1081),
    [sym_private_property_identifier] = ACTIONS(1081),
    [anon_sym_static] = ACTIONS(1085),
    [anon_sym_get] = ACTIONS(1085),
    [anon_sym_set] = ACTIONS(1085),
    [sym__automatic_semicolon] = ACTIONS(169),
    [sym__ternary_qmark] = ACTIONS(169),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(363)] = {
    [sym_string] = STATE(1710),
    [sym__property_name] = STATE(1710),
    [sym_computed_property_name] = STATE(1710),
    [aux_sym_object_repeat1] = STATE(1487),
    [aux_sym_object_pattern_repeat1] = STATE(1452),
    [sym_identifier] = ACTIONS(1085),
    [anon_sym_export] = ACTIONS(1085),
    [anon_sym_STAR] = ACTIONS(1065),
    [anon_sym_COMMA] = ACTIONS(169),
    [anon_sym_RBRACE] = ACTIONS(171),
    [anon_sym_let] = ACTIONS(1085),
    [anon_sym_await] = ACTIONS(1085),
    [anon_sym_LPAREN] = ACTIONS(944),
    [anon_sym_SEMI] = ACTIONS(169),
    [anon_sym_in] = ACTIONS(165),
    [anon_sym_COLON] = ACTIONS(183),
    [anon_sym_EQ] = ACTIONS(186),
    [anon_sym_LBRACK] = ACTIONS(1072),
    [anon_sym_LT] = ACTIONS(165),
    [anon_sym_GT] = ACTIONS(165),
    [anon_sym_DOT] = ACTIONS(165),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [anon_sym_SQUOTE] = ACTIONS(1077),
    [anon_sym_async] = ACTIONS(1087),
    [anon_sym_EQ_GT] = ACTIONS(200),
    [sym_optional_chain] = ACTIONS(169),
    [anon_sym_PLUS_EQ] = ACTIONS(202),
    [anon_sym_DASH_EQ] = ACTIONS(202),
    [anon_sym_STAR_EQ] = ACTIONS(202),
    [anon_sym_SLASH_EQ] = ACTIONS(202),
    [anon_sym_PERCENT_EQ] = ACTIONS(202),
    [anon_sym_CARET_EQ] = ACTIONS(202),
    [anon_sym_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_LT_LT_EQ] = ACTIONS(202),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(202),
    [anon_sym_QMARK_QMARK_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP] = ACTIONS(165),
    [anon_sym_PIPE_PIPE] = ACTIONS(165),
    [anon_sym_GT_GT] = ACTIONS(165),
    [anon_sym_GT_GT_GT] = ACTIONS(165),
    [anon_sym_LT_LT] = ACTIONS(165),
    [anon_sym_AMP] = ACTIONS(165),
    [anon_sym_CARET] = ACTIONS(165),
    [anon_sym_PIPE] = ACTIONS(165),
    [anon_sym_PLUS] = ACTIONS(165),
    [anon_sym_DASH] = ACTIONS(165),
    [anon_sym_SLASH] = ACTIONS(165),
    [anon_sym_PERCENT] = ACTIONS(165),
    [anon_sym_STAR_STAR] = ACTIONS(165),
    [anon_sym_LT_EQ] = ACTIONS(169),
    [anon_sym_EQ_EQ] = ACTIONS(165),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(169),
    [anon_sym_BANG_EQ] = ACTIONS(165),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(169),
    [anon_sym_GT_EQ] = ACTIONS(169),
    [anon_sym_QMARK_QMARK] = ACTIONS(165),
    [anon_sym_instanceof] = ACTIONS(165),
    [anon_sym_PLUS_PLUS] = ACTIONS(169),
    [anon_sym_DASH_DASH] = ACTIONS(169),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(169),
    [sym_number] = ACTIONS(1081),
    [sym_private_property_identifier] = ACTIONS(1081),
    [anon_sym_static] = ACTIONS(1085),
    [anon_sym_get] = ACTIONS(1089),
    [anon_sym_set] = ACTIONS(1089),
    [sym__automatic_semicolon] = ACTIONS(169),
    [sym__ternary_qmark] = ACTIONS(169),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(364)] = {
    [sym_string] = STATE(1710),
    [sym__property_name] = STATE(1710),
    [sym_computed_property_name] = STATE(1710),
    [aux_sym_object_repeat1] = STATE(1418),
    [aux_sym_object_pattern_repeat1] = STATE(1452),
    [sym_identifier] = ACTIONS(1085),
    [anon_sym_export] = ACTIONS(1085),
    [anon_sym_STAR] = ACTIONS(1065),
    [anon_sym_COMMA] = ACTIONS(169),
    [anon_sym_RBRACE] = ACTIONS(216),
    [anon_sym_let] = ACTIONS(1085),
    [anon_sym_await] = ACTIONS(1085),
    [anon_sym_LPAREN] = ACTIONS(944),
    [anon_sym_SEMI] = ACTIONS(169),
    [anon_sym_in] = ACTIONS(165),
    [anon_sym_COLON] = ACTIONS(183),
    [anon_sym_EQ] = ACTIONS(186),
    [anon_sym_LBRACK] = ACTIONS(1072),
    [anon_sym_LT] = ACTIONS(165),
    [anon_sym_GT] = ACTIONS(165),
    [anon_sym_DOT] = ACTIONS(165),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [anon_sym_SQUOTE] = ACTIONS(1077),
    [anon_sym_async] = ACTIONS(1087),
    [anon_sym_EQ_GT] = ACTIONS(200),
    [sym_optional_chain] = ACTIONS(169),
    [anon_sym_PLUS_EQ] = ACTIONS(202),
    [anon_sym_DASH_EQ] = ACTIONS(202),
    [anon_sym_STAR_EQ] = ACTIONS(202),
    [anon_sym_SLASH_EQ] = ACTIONS(202),
    [anon_sym_PERCENT_EQ] = ACTIONS(202),
    [anon_sym_CARET_EQ] = ACTIONS(202),
    [anon_sym_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_LT_LT_EQ] = ACTIONS(202),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(202),
    [anon_sym_QMARK_QMARK_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP] = ACTIONS(165),
    [anon_sym_PIPE_PIPE] = ACTIONS(165),
    [anon_sym_GT_GT] = ACTIONS(165),
    [anon_sym_GT_GT_GT] = ACTIONS(165),
    [anon_sym_LT_LT] = ACTIONS(165),
    [anon_sym_AMP] = ACTIONS(165),
    [anon_sym_CARET] = ACTIONS(165),
    [anon_sym_PIPE] = ACTIONS(165),
    [anon_sym_PLUS] = ACTIONS(165),
    [anon_sym_DASH] = ACTIONS(165),
    [anon_sym_SLASH] = ACTIONS(165),
    [anon_sym_PERCENT] = ACTIONS(165),
    [anon_sym_STAR_STAR] = ACTIONS(165),
    [anon_sym_LT_EQ] = ACTIONS(169),
    [anon_sym_EQ_EQ] = ACTIONS(165),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(169),
    [anon_sym_BANG_EQ] = ACTIONS(165),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(169),
    [anon_sym_GT_EQ] = ACTIONS(169),
    [anon_sym_QMARK_QMARK] = ACTIONS(165),
    [anon_sym_instanceof] = ACTIONS(165),
    [anon_sym_PLUS_PLUS] = ACTIONS(169),
    [anon_sym_DASH_DASH] = ACTIONS(169),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(169),
    [sym_number] = ACTIONS(1081),
    [sym_private_property_identifier] = ACTIONS(1081),
    [anon_sym_static] = ACTIONS(1085),
    [anon_sym_get] = ACTIONS(1089),
    [anon_sym_set] = ACTIONS(1089),
    [sym__automatic_semicolon] = ACTIONS(169),
    [sym__ternary_qmark] = ACTIONS(169),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(365)] = {
    [sym_string] = STATE(1710),
    [sym__property_name] = STATE(1710),
    [sym_computed_property_name] = STATE(1710),
    [aux_sym_object_repeat1] = STATE(1418),
    [aux_sym_object_pattern_repeat1] = STATE(1452),
    [sym_identifier] = ACTIONS(1085),
    [anon_sym_export] = ACTIONS(1085),
    [anon_sym_STAR] = ACTIONS(165),
    [anon_sym_COMMA] = ACTIONS(169),
    [anon_sym_RBRACE] = ACTIONS(216),
    [anon_sym_let] = ACTIONS(1085),
    [anon_sym_await] = ACTIONS(1085),
    [anon_sym_LPAREN] = ACTIONS(944),
    [anon_sym_SEMI] = ACTIONS(169),
    [anon_sym_in] = ACTIONS(165),
    [anon_sym_COLON] = ACTIONS(183),
    [anon_sym_EQ] = ACTIONS(186),
    [anon_sym_LBRACK] = ACTIONS(1072),
    [anon_sym_LT] = ACTIONS(165),
    [anon_sym_GT] = ACTIONS(165),
    [anon_sym_DOT] = ACTIONS(165),
    [anon_sym_DQUOTE] = ACTIONS(1075),
    [anon_sym_SQUOTE] = ACTIONS(1077),
    [anon_sym_async] = ACTIONS(1085),
    [anon_sym_EQ_GT] = ACTIONS(200),
    [sym_optional_chain] = ACTIONS(169),
    [anon_sym_PLUS_EQ] = ACTIONS(202),
    [anon_sym_DASH_EQ] = ACTIONS(202),
    [anon_sym_STAR_EQ] = ACTIONS(202),
    [anon_sym_SLASH_EQ] = ACTIONS(202),
    [anon_sym_PERCENT_EQ] = ACTIONS(202),
    [anon_sym_CARET_EQ] = ACTIONS(202),
    [anon_sym_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_LT_LT_EQ] = ACTIONS(202),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(202),
    [anon_sym_QMARK_QMARK_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP] = ACTIONS(165),
    [anon_sym_PIPE_PIPE] = ACTIONS(165),
    [anon_sym_GT_GT] = ACTIONS(165),
    [anon_sym_GT_GT_GT] = ACTIONS(165),
    [anon_sym_LT_LT] = ACTIONS(165),
    [anon_sym_AMP] = ACTIONS(165),
    [anon_sym_CARET] = ACTIONS(165),
    [anon_sym_PIPE] = ACTIONS(165),
    [anon_sym_PLUS] = ACTIONS(165),
    [anon_sym_DASH] = ACTIONS(165),
    [anon_sym_SLASH] = ACTIONS(165),
    [anon_sym_PERCENT] = ACTIONS(165),
    [anon_sym_STAR_STAR] = ACTIONS(165),
    [anon_sym_LT_EQ] = ACTIONS(169),
    [anon_sym_EQ_EQ] = ACTIONS(165),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(169),
    [anon_sym_BANG_EQ] = ACTIONS(165),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(169),
    [anon_sym_GT_EQ] = ACTIONS(169),
    [anon_sym_QMARK_QMARK] = ACTIONS(165),
    [anon_sym_instanceof] = ACTIONS(165),
    [anon_sym_PLUS_PLUS] = ACTIONS(169),
    [anon_sym_DASH_DASH] = ACTIONS(169),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(169),
    [sym_number] = ACTIONS(1081),
    [sym_private_property_identifier] = ACTIONS(1081),
    [anon_sym_static] = ACTIONS(1085),
    [anon_sym_get] = ACTIONS(1085),
    [anon_sym_set] = ACTIONS(1085),
    [sym__automatic_semicolon] = ACTIONS(169),
    [sym__ternary_qmark] = ACTIONS(169),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(366)] = {
    [sym_formal_parameters] = STATE(1859),
    [sym_identifier] = ACTIONS(1091),
    [anon_sym_export] = ACTIONS(1093),
    [anon_sym_STAR] = ACTIONS(165),
    [anon_sym_COMMA] = ACTIONS(169),
    [anon_sym_RBRACE] = ACTIONS(169),
    [anon_sym_let] = ACTIONS(1093),
    [anon_sym_await] = ACTIONS(1093),
    [anon_sym_LPAREN] = ACTIONS(1095),
    [anon_sym_SEMI] = ACTIONS(169),
    [anon_sym_RPAREN] = ACTIONS(169),
    [anon_sym_in] = ACTIONS(165),
    [anon_sym_COLON] = ACTIONS(169),
    [anon_sym_EQ] = ACTIONS(384),
    [anon_sym_LBRACK] = ACTIONS(169),
    [anon_sym_RBRACK] = ACTIONS(169),
    [anon_sym_LT] = ACTIONS(165),
    [anon_sym_GT] = ACTIONS(165),
    [anon_sym_DOT] = ACTIONS(169),
    [anon_sym_async] = ACTIONS(1093),
    [anon_sym_function] = ACTIONS(1098),
    [anon_sym_EQ_GT] = ACTIONS(402),
    [sym_optional_chain] = ACTIONS(169),
    [anon_sym_PLUS_EQ] = ACTIONS(202),
    [anon_sym_DASH_EQ] = ACTIONS(202),
    [anon_sym_STAR_EQ] = ACTIONS(202),
    [anon_sym_SLASH_EQ] = ACTIONS(202),
    [anon_sym_PERCENT_EQ] = ACTIONS(202),
    [anon_sym_CARET_EQ] = ACTIONS(202),
    [anon_sym_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_LT_LT_EQ] = ACTIONS(202),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(202),
    [anon_sym_QMARK_QMARK_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP] = ACTIONS(165),
    [anon_sym_PIPE_PIPE] = ACTIONS(165),
    [anon_sym_GT_GT] = ACTIONS(165),
    [anon_sym_GT_GT_GT] = ACTIONS(165),
    [anon_sym_LT_LT] = ACTIONS(165),
    [anon_sym_AMP] = ACTIONS(165),
    [anon_sym_CARET] = ACTIONS(165),
    [anon_sym_PIPE] = ACTIONS(165),
    [anon_sym_PLUS] = ACTIONS(165),
    [anon_sym_DASH] = ACTIONS(165),
    [anon_sym_SLASH] = ACTIONS(165),
    [anon_sym_PERCENT] = ACTIONS(165),
    [anon_sym_STAR_STAR] = ACTIONS(165),
    [anon_sym_LT_EQ] = ACTIONS(169),
    [anon_sym_EQ_EQ] = ACTIONS(165),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(169),
    [anon_sym_BANG_EQ] = ACTIONS(165),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(169),
    [anon_sym_GT_EQ] = ACTIONS(169),
    [anon_sym_QMARK_QMARK] = ACTIONS(165),
    [anon_sym_instanceof] = ACTIONS(165),
    [anon_sym_PLUS_PLUS] = ACTIONS(169),
    [anon_sym_DASH_DASH] = ACTIONS(169),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(169),
    [anon_sym_static] = ACTIONS(1093),
    [anon_sym_get] = ACTIONS(1093),
    [anon_sym_set] = ACTIONS(1093),
    [sym__ternary_qmark] = ACTIONS(169),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(367)] = {
    [sym_formal_parameters] = STATE(1859),
    [sym_identifier] = ACTIONS(1091),
    [anon_sym_export] = ACTIONS(1093),
    [anon_sym_STAR] = ACTIONS(165),
    [anon_sym_COMMA] = ACTIONS(169),
    [anon_sym_RBRACE] = ACTIONS(169),
    [anon_sym_let] = ACTIONS(1093),
    [anon_sym_await] = ACTIONS(1093),
    [anon_sym_LPAREN] = ACTIONS(1095),
    [anon_sym_SEMI] = ACTIONS(169),
    [anon_sym_RPAREN] = ACTIONS(169),
    [anon_sym_in] = ACTIONS(165),
    [anon_sym_COLON] = ACTIONS(169),
    [anon_sym_EQ] = ACTIONS(440),
    [anon_sym_LBRACK] = ACTIONS(169),
    [anon_sym_RBRACK] = ACTIONS(169),
    [anon_sym_LT] = ACTIONS(165),
    [anon_sym_GT] = ACTIONS(165),
    [anon_sym_DOT] = ACTIONS(169),
    [anon_sym_async] = ACTIONS(1093),
    [anon_sym_function] = ACTIONS(1098),
    [anon_sym_EQ_GT] = ACTIONS(402),
    [sym_optional_chain] = ACTIONS(169),
    [anon_sym_PLUS_EQ] = ACTIONS(202),
    [anon_sym_DASH_EQ] = ACTIONS(202),
    [anon_sym_STAR_EQ] = ACTIONS(202),
    [anon_sym_SLASH_EQ] = ACTIONS(202),
    [anon_sym_PERCENT_EQ] = ACTIONS(202),
    [anon_sym_CARET_EQ] = ACTIONS(202),
    [anon_sym_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_LT_LT_EQ] = ACTIONS(202),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(202),
    [anon_sym_QMARK_QMARK_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP] = ACTIONS(165),
    [anon_sym_PIPE_PIPE] = ACTIONS(165),
    [anon_sym_GT_GT] = ACTIONS(165),
    [anon_sym_GT_GT_GT] = ACTIONS(165),
    [anon_sym_LT_LT] = ACTIONS(165),
    [anon_sym_AMP] = ACTIONS(165),
    [anon_sym_CARET] = ACTIONS(165),
    [anon_sym_PIPE] = ACTIONS(165),
    [anon_sym_PLUS] = ACTIONS(165),
    [anon_sym_DASH] = ACTIONS(165),
    [anon_sym_SLASH] = ACTIONS(165),
    [anon_sym_PERCENT] = ACTIONS(165),
    [anon_sym_STAR_STAR] = ACTIONS(165),
    [anon_sym_LT_EQ] = ACTIONS(169),
    [anon_sym_EQ_EQ] = ACTIONS(165),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(169),
    [anon_sym_BANG_EQ] = ACTIONS(165),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(169),
    [anon_sym_GT_EQ] = ACTIONS(169),
    [anon_sym_QMARK_QMARK] = ACTIONS(165),
    [anon_sym_instanceof] = ACTIONS(165),
    [anon_sym_PLUS_PLUS] = ACTIONS(169),
    [anon_sym_DASH_DASH] = ACTIONS(169),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(169),
    [anon_sym_static] = ACTIONS(1093),
    [anon_sym_get] = ACTIONS(1093),
    [anon_sym_set] = ACTIONS(1093),
    [sym__ternary_qmark] = ACTIONS(169),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(368)] = {
    [sym_variable_declarator] = STATE(1385),
    [sym_object_pattern] = STATE(1311),
    [sym_array_pattern] = STATE(1311),
    [sym__destructuring_pattern] = STATE(1311),
    [aux_sym_object_repeat1] = STATE(1418),
    [aux_sym_object_pattern_repeat1] = STATE(1452),
    [sym_identifier] = ACTIONS(1100),
    [anon_sym_STAR] = ACTIONS(165),
    [anon_sym_LBRACE] = ACTIONS(1102),
    [anon_sym_COMMA] = ACTIONS(169),
    [anon_sym_RBRACE] = ACTIONS(351),
    [anon_sym_of] = ACTIONS(1104),
    [anon_sym_LPAREN] = ACTIONS(944),
    [anon_sym_SEMI] = ACTIONS(169),
    [anon_sym_in] = ACTIONS(165),
    [anon_sym_COLON] = ACTIONS(183),
    [anon_sym_EQ] = ACTIONS(186),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_LT] = ACTIONS(165),
    [anon_sym_GT] = ACTIONS(165),
    [anon_sym_DOT] = ACTIONS(169),
    [anon_sym_EQ_GT] = ACTIONS(200),
    [sym_optional_chain] = ACTIONS(169),
    [anon_sym_PLUS_EQ] = ACTIONS(202),
    [anon_sym_DASH_EQ] = ACTIONS(202),
    [anon_sym_STAR_EQ] = ACTIONS(202),
    [anon_sym_SLASH_EQ] = ACTIONS(202),
    [anon_sym_PERCENT_EQ] = ACTIONS(202),
    [anon_sym_CARET_EQ] = ACTIONS(202),
    [anon_sym_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_LT_LT_EQ] = ACTIONS(202),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(202),
    [anon_sym_QMARK_QMARK_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP] = ACTIONS(165),
    [anon_sym_PIPE_PIPE] = ACTIONS(165),
    [anon_sym_GT_GT] = ACTIONS(165),
    [anon_sym_GT_GT_GT] = ACTIONS(165),
    [anon_sym_LT_LT] = ACTIONS(165),
    [anon_sym_AMP] = ACTIONS(165),
    [anon_sym_CARET] = ACTIONS(165),
    [anon_sym_PIPE] = ACTIONS(165),
    [anon_sym_PLUS] = ACTIONS(165),
    [anon_sym_DASH] = ACTIONS(165),
    [anon_sym_SLASH] = ACTIONS(165),
    [anon_sym_PERCENT] = ACTIONS(165),
    [anon_sym_STAR_STAR] = ACTIONS(165),
    [anon_sym_LT_EQ] = ACTIONS(169),
    [anon_sym_EQ_EQ] = ACTIONS(165),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(169),
    [anon_sym_BANG_EQ] = ACTIONS(165),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(169),
    [anon_sym_GT_EQ] = ACTIONS(169),
    [anon_sym_QMARK_QMARK] = ACTIONS(165),
    [anon_sym_instanceof] = ACTIONS(165),
    [anon_sym_PLUS_PLUS] = ACTIONS(169),
    [anon_sym_DASH_DASH] = ACTIONS(169),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(169),
    [sym__automatic_semicolon] = ACTIONS(169),
    [sym__ternary_qmark] = ACTIONS(169),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(369)] = {
    [ts_builtin_sym_end] = ACTIONS(694),
    [sym_identifier] = ACTIONS(696),
    [anon_sym_export] = ACTIONS(696),
    [anon_sym_default] = ACTIONS(696),
    [anon_sym_LBRACE] = ACTIONS(694),
    [anon_sym_COMMA] = ACTIONS(694),
    [anon_sym_RBRACE] = ACTIONS(694),
    [anon_sym_import] = ACTIONS(696),
    [anon_sym_with] = ACTIONS(696),
    [anon_sym_var] = ACTIONS(696),
    [anon_sym_let] = ACTIONS(696),
    [anon_sym_const] = ACTIONS(696),
    [anon_sym_using] = ACTIONS(696),
    [anon_sym_await] = ACTIONS(696),
    [anon_sym_else] = ACTIONS(696),
    [anon_sym_if] = ACTIONS(696),
    [anon_sym_switch] = ACTIONS(696),
    [anon_sym_for] = ACTIONS(696),
    [anon_sym_LPAREN] = ACTIONS(694),
    [anon_sym_SEMI] = ACTIONS(694),
    [anon_sym_while] = ACTIONS(696),
    [anon_sym_do] = ACTIONS(696),
    [anon_sym_try] = ACTIONS(696),
    [anon_sym_break] = ACTIONS(696),
    [anon_sym_continue] = ACTIONS(696),
    [anon_sym_debugger] = ACTIONS(696),
    [anon_sym_return] = ACTIONS(696),
    [anon_sym_throw] = ACTIONS(696),
    [anon_sym_case] = ACTIONS(696),
    [anon_sym_catch] = ACTIONS(696),
    [anon_sym_finally] = ACTIONS(696),
    [anon_sym_yield] = ACTIONS(696),
    [anon_sym_LBRACK] = ACTIONS(694),
    [anon_sym_LT] = ACTIONS(694),
    [anon_sym_DQUOTE] = ACTIONS(694),
    [anon_sym_SQUOTE] = ACTIONS(694),
    [anon_sym_class] = ACTIONS(696),
    [anon_sym_async] = ACTIONS(696),
    [anon_sym_function] = ACTIONS(696),
    [anon_sym_new] = ACTIONS(696),
    [anon_sym_PLUS] = ACTIONS(696),
    [anon_sym_DASH] = ACTIONS(696),
    [anon_sym_SLASH] = ACTIONS(696),
    [anon_sym_BANG] = ACTIONS(694),
    [anon_sym_TILDE] = ACTIONS(694),
    [anon_sym_typeof] = ACTIONS(696),
    [anon_sym_void] = ACTIONS(696),
    [anon_sym_delete] = ACTIONS(696),
    [anon_sym_PLUS_PLUS] = ACTIONS(694),
    [anon_sym_DASH_DASH] = ACTIONS(694),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(694),
    [sym_number] = ACTIONS(694),
    [sym_private_property_identifier] = ACTIONS(694),
    [sym_this] = ACTIONS(696),
    [sym_super] = ACTIONS(696),
    [sym_true] = ACTIONS(696),
    [sym_false] = ACTIONS(696),
    [sym_null] = ACTIONS(696),
    [sym_undefined] = ACTIONS(696),
    [anon_sym_AT] = ACTIONS(694),
    [anon_sym_static] = ACTIONS(696),
    [anon_sym_get] = ACTIONS(696),
    [anon_sym_set] = ACTIONS(696),
    [sym__automatic_semicolon] = ACTIONS(1108),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(370)] = {
    [sym_variable_declarator] = STATE(1385),
    [sym_object_pattern] = STATE(1311),
    [sym_array_pattern] = STATE(1311),
    [sym__destructuring_pattern] = STATE(1311),
    [aux_sym_object_repeat1] = STATE(1487),
    [aux_sym_object_pattern_repeat1] = STATE(1452),
    [sym_identifier] = ACTIONS(1100),
    [anon_sym_STAR] = ACTIONS(165),
    [anon_sym_LBRACE] = ACTIONS(1102),
    [anon_sym_COMMA] = ACTIONS(169),
    [anon_sym_RBRACE] = ACTIONS(171),
    [anon_sym_of] = ACTIONS(1104),
    [anon_sym_LPAREN] = ACTIONS(944),
    [anon_sym_SEMI] = ACTIONS(169),
    [anon_sym_in] = ACTIONS(165),
    [anon_sym_COLON] = ACTIONS(183),
    [anon_sym_EQ] = ACTIONS(186),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_LT] = ACTIONS(165),
    [anon_sym_GT] = ACTIONS(165),
    [anon_sym_DOT] = ACTIONS(169),
    [anon_sym_EQ_GT] = ACTIONS(200),
    [sym_optional_chain] = ACTIONS(169),
    [anon_sym_PLUS_EQ] = ACTIONS(202),
    [anon_sym_DASH_EQ] = ACTIONS(202),
    [anon_sym_STAR_EQ] = ACTIONS(202),
    [anon_sym_SLASH_EQ] = ACTIONS(202),
    [anon_sym_PERCENT_EQ] = ACTIONS(202),
    [anon_sym_CARET_EQ] = ACTIONS(202),
    [anon_sym_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_LT_LT_EQ] = ACTIONS(202),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(202),
    [anon_sym_QMARK_QMARK_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP] = ACTIONS(165),
    [anon_sym_PIPE_PIPE] = ACTIONS(165),
    [anon_sym_GT_GT] = ACTIONS(165),
    [anon_sym_GT_GT_GT] = ACTIONS(165),
    [anon_sym_LT_LT] = ACTIONS(165),
    [anon_sym_AMP] = ACTIONS(165),
    [anon_sym_CARET] = ACTIONS(165),
    [anon_sym_PIPE] = ACTIONS(165),
    [anon_sym_PLUS] = ACTIONS(165),
    [anon_sym_DASH] = ACTIONS(165),
    [anon_sym_SLASH] = ACTIONS(165),
    [anon_sym_PERCENT] = ACTIONS(165),
    [anon_sym_STAR_STAR] = ACTIONS(165),
    [anon_sym_LT_EQ] = ACTIONS(169),
    [anon_sym_EQ_EQ] = ACTIONS(165),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(169),
    [anon_sym_BANG_EQ] = ACTIONS(165),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(169),
    [anon_sym_GT_EQ] = ACTIONS(169),
    [anon_sym_QMARK_QMARK] = ACTIONS(165),
    [anon_sym_instanceof] = ACTIONS(165),
    [anon_sym_PLUS_PLUS] = ACTIONS(169),
    [anon_sym_DASH_DASH] = ACTIONS(169),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(169),
    [sym__automatic_semicolon] = ACTIONS(169),
    [sym__ternary_qmark] = ACTIONS(169),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(371)] = {
    [sym_formal_parameters] = STATE(1834),
    [sym_identifier] = ACTIONS(1110),
    [anon_sym_export] = ACTIONS(1112),
    [anon_sym_STAR] = ACTIONS(165),
    [anon_sym_COMMA] = ACTIONS(169),
    [anon_sym_let] = ACTIONS(1112),
    [anon_sym_await] = ACTIONS(1112),
    [anon_sym_LPAREN] = ACTIONS(1095),
    [anon_sym_SEMI] = ACTIONS(169),
    [anon_sym_in] = ACTIONS(165),
    [anon_sym_COLON] = ACTIONS(494),
    [anon_sym_EQ] = ACTIONS(490),
    [anon_sym_LBRACK] = ACTIONS(169),
    [anon_sym_LT] = ACTIONS(165),
    [anon_sym_GT] = ACTIONS(165),
    [anon_sym_DOT] = ACTIONS(169),
    [anon_sym_async] = ACTIONS(1112),
    [anon_sym_function] = ACTIONS(1079),
    [anon_sym_EQ_GT] = ACTIONS(200),
    [sym_optional_chain] = ACTIONS(169),
    [anon_sym_PLUS_EQ] = ACTIONS(202),
    [anon_sym_DASH_EQ] = ACTIONS(202),
    [anon_sym_STAR_EQ] = ACTIONS(202),
    [anon_sym_SLASH_EQ] = ACTIONS(202),
    [anon_sym_PERCENT_EQ] = ACTIONS(202),
    [anon_sym_CARET_EQ] = ACTIONS(202),
    [anon_sym_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_LT_LT_EQ] = ACTIONS(202),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(202),
    [anon_sym_QMARK_QMARK_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP] = ACTIONS(165),
    [anon_sym_PIPE_PIPE] = ACTIONS(165),
    [anon_sym_GT_GT] = ACTIONS(165),
    [anon_sym_GT_GT_GT] = ACTIONS(165),
    [anon_sym_LT_LT] = ACTIONS(165),
    [anon_sym_AMP] = ACTIONS(165),
    [anon_sym_CARET] = ACTIONS(165),
    [anon_sym_PIPE] = ACTIONS(165),
    [anon_sym_PLUS] = ACTIONS(165),
    [anon_sym_DASH] = ACTIONS(165),
    [anon_sym_SLASH] = ACTIONS(165),
    [anon_sym_PERCENT] = ACTIONS(165),
    [anon_sym_STAR_STAR] = ACTIONS(165),
    [anon_sym_LT_EQ] = ACTIONS(169),
    [anon_sym_EQ_EQ] = ACTIONS(165),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(169),
    [anon_sym_BANG_EQ] = ACTIONS(165),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(169),
    [anon_sym_GT_EQ] = ACTIONS(169),
    [anon_sym_QMARK_QMARK] = ACTIONS(165),
    [anon_sym_instanceof] = ACTIONS(165),
    [anon_sym_PLUS_PLUS] = ACTIONS(169),
    [anon_sym_DASH_DASH] = ACTIONS(169),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(169),
    [anon_sym_static] = ACTIONS(1112),
    [anon_sym_get] = ACTIONS(1112),
    [anon_sym_set] = ACTIONS(1112),
    [sym__automatic_semicolon] = ACTIONS(169),
    [sym__ternary_qmark] = ACTIONS(169),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(372)] = {
    [sym_formal_parameters] = STATE(1834),
    [sym_identifier] = ACTIONS(1110),
    [anon_sym_export] = ACTIONS(1112),
    [anon_sym_STAR] = ACTIONS(165),
    [anon_sym_COMMA] = ACTIONS(169),
    [anon_sym_RBRACE] = ACTIONS(169),
    [anon_sym_let] = ACTIONS(1112),
    [anon_sym_await] = ACTIONS(1112),
    [anon_sym_LPAREN] = ACTIONS(1095),
    [anon_sym_SEMI] = ACTIONS(169),
    [anon_sym_in] = ACTIONS(165),
    [anon_sym_EQ] = ACTIONS(440),
    [anon_sym_LBRACK] = ACTIONS(169),
    [anon_sym_LT] = ACTIONS(165),
    [anon_sym_GT] = ACTIONS(165),
    [anon_sym_DOT] = ACTIONS(169),
    [anon_sym_async] = ACTIONS(1112),
    [anon_sym_function] = ACTIONS(1114),
    [anon_sym_EQ_GT] = ACTIONS(200),
    [sym_optional_chain] = ACTIONS(169),
    [anon_sym_PLUS_EQ] = ACTIONS(202),
    [anon_sym_DASH_EQ] = ACTIONS(202),
    [anon_sym_STAR_EQ] = ACTIONS(202),
    [anon_sym_SLASH_EQ] = ACTIONS(202),
    [anon_sym_PERCENT_EQ] = ACTIONS(202),
    [anon_sym_CARET_EQ] = ACTIONS(202),
    [anon_sym_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_LT_LT_EQ] = ACTIONS(202),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(202),
    [anon_sym_QMARK_QMARK_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP] = ACTIONS(165),
    [anon_sym_PIPE_PIPE] = ACTIONS(165),
    [anon_sym_GT_GT] = ACTIONS(165),
    [anon_sym_GT_GT_GT] = ACTIONS(165),
    [anon_sym_LT_LT] = ACTIONS(165),
    [anon_sym_AMP] = ACTIONS(165),
    [anon_sym_CARET] = ACTIONS(165),
    [anon_sym_PIPE] = ACTIONS(165),
    [anon_sym_PLUS] = ACTIONS(165),
    [anon_sym_DASH] = ACTIONS(165),
    [anon_sym_SLASH] = ACTIONS(165),
    [anon_sym_PERCENT] = ACTIONS(165),
    [anon_sym_STAR_STAR] = ACTIONS(165),
    [anon_sym_LT_EQ] = ACTIONS(169),
    [anon_sym_EQ_EQ] = ACTIONS(165),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(169),
    [anon_sym_BANG_EQ] = ACTIONS(165),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(169),
    [anon_sym_GT_EQ] = ACTIONS(169),
    [anon_sym_QMARK_QMARK] = ACTIONS(165),
    [anon_sym_instanceof] = ACTIONS(165),
    [anon_sym_PLUS_PLUS] = ACTIONS(169),
    [anon_sym_DASH_DASH] = ACTIONS(169),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(169),
    [anon_sym_static] = ACTIONS(1112),
    [anon_sym_get] = ACTIONS(1112),
    [anon_sym_set] = ACTIONS(1112),
    [sym__automatic_semicolon] = ACTIONS(169),
    [sym__ternary_qmark] = ACTIONS(169),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(373)] = {
    [sym_formal_parameters] = STATE(1813),
    [sym_identifier] = ACTIONS(1116),
    [anon_sym_export] = ACTIONS(1118),
    [anon_sym_STAR] = ACTIONS(165),
    [anon_sym_COMMA] = ACTIONS(169),
    [anon_sym_let] = ACTIONS(1118),
    [anon_sym_await] = ACTIONS(1118),
    [anon_sym_of] = ACTIONS(165),
    [anon_sym_LPAREN] = ACTIONS(1095),
    [anon_sym_SEMI] = ACTIONS(169),
    [anon_sym_in] = ACTIONS(165),
    [anon_sym_EQ] = ACTIONS(440),
    [anon_sym_LBRACK] = ACTIONS(169),
    [anon_sym_LT] = ACTIONS(165),
    [anon_sym_GT] = ACTIONS(165),
    [anon_sym_DOT] = ACTIONS(169),
    [anon_sym_async] = ACTIONS(1118),
    [anon_sym_function] = ACTIONS(1114),
    [anon_sym_EQ_GT] = ACTIONS(500),
    [sym_optional_chain] = ACTIONS(169),
    [anon_sym_PLUS_EQ] = ACTIONS(202),
    [anon_sym_DASH_EQ] = ACTIONS(202),
    [anon_sym_STAR_EQ] = ACTIONS(202),
    [anon_sym_SLASH_EQ] = ACTIONS(202),
    [anon_sym_PERCENT_EQ] = ACTIONS(202),
    [anon_sym_CARET_EQ] = ACTIONS(202),
    [anon_sym_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_LT_LT_EQ] = ACTIONS(202),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(202),
    [anon_sym_QMARK_QMARK_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP] = ACTIONS(165),
    [anon_sym_PIPE_PIPE] = ACTIONS(165),
    [anon_sym_GT_GT] = ACTIONS(165),
    [anon_sym_GT_GT_GT] = ACTIONS(165),
    [anon_sym_LT_LT] = ACTIONS(165),
    [anon_sym_AMP] = ACTIONS(165),
    [anon_sym_CARET] = ACTIONS(165),
    [anon_sym_PIPE] = ACTIONS(165),
    [anon_sym_PLUS] = ACTIONS(165),
    [anon_sym_DASH] = ACTIONS(165),
    [anon_sym_SLASH] = ACTIONS(165),
    [anon_sym_PERCENT] = ACTIONS(165),
    [anon_sym_STAR_STAR] = ACTIONS(165),
    [anon_sym_LT_EQ] = ACTIONS(169),
    [anon_sym_EQ_EQ] = ACTIONS(165),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(169),
    [anon_sym_BANG_EQ] = ACTIONS(165),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(169),
    [anon_sym_GT_EQ] = ACTIONS(169),
    [anon_sym_QMARK_QMARK] = ACTIONS(165),
    [anon_sym_instanceof] = ACTIONS(165),
    [anon_sym_PLUS_PLUS] = ACTIONS(169),
    [anon_sym_DASH_DASH] = ACTIONS(169),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(169),
    [anon_sym_static] = ACTIONS(1118),
    [anon_sym_get] = ACTIONS(1118),
    [anon_sym_set] = ACTIONS(1118),
    [sym__automatic_semicolon] = ACTIONS(169),
    [sym__ternary_qmark] = ACTIONS(169),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(374)] = {
    [ts_builtin_sym_end] = ACTIONS(708),
    [sym_identifier] = ACTIONS(710),
    [anon_sym_export] = ACTIONS(710),
    [anon_sym_default] = ACTIONS(710),
    [anon_sym_LBRACE] = ACTIONS(708),
    [anon_sym_COMMA] = ACTIONS(708),
    [anon_sym_RBRACE] = ACTIONS(708),
    [anon_sym_import] = ACTIONS(710),
    [anon_sym_with] = ACTIONS(710),
    [anon_sym_var] = ACTIONS(710),
    [anon_sym_let] = ACTIONS(710),
    [anon_sym_const] = ACTIONS(710),
    [anon_sym_using] = ACTIONS(710),
    [anon_sym_await] = ACTIONS(710),
    [anon_sym_else] = ACTIONS(710),
    [anon_sym_if] = ACTIONS(710),
    [anon_sym_switch] = ACTIONS(710),
    [anon_sym_for] = ACTIONS(710),
    [anon_sym_LPAREN] = ACTIONS(708),
    [anon_sym_SEMI] = ACTIONS(708),
    [anon_sym_while] = ACTIONS(710),
    [anon_sym_do] = ACTIONS(710),
    [anon_sym_try] = ACTIONS(710),
    [anon_sym_break] = ACTIONS(710),
    [anon_sym_continue] = ACTIONS(710),
    [anon_sym_debugger] = ACTIONS(710),
    [anon_sym_return] = ACTIONS(710),
    [anon_sym_throw] = ACTIONS(710),
    [anon_sym_case] = ACTIONS(710),
    [anon_sym_catch] = ACTIONS(710),
    [anon_sym_finally] = ACTIONS(710),
    [anon_sym_yield] = ACTIONS(710),
    [anon_sym_LBRACK] = ACTIONS(708),
    [anon_sym_LT] = ACTIONS(708),
    [anon_sym_DQUOTE] = ACTIONS(708),
    [anon_sym_SQUOTE] = ACTIONS(708),
    [anon_sym_class] = ACTIONS(710),
    [anon_sym_async] = ACTIONS(710),
    [anon_sym_function] = ACTIONS(710),
    [anon_sym_new] = ACTIONS(710),
    [anon_sym_PLUS] = ACTIONS(710),
    [anon_sym_DASH] = ACTIONS(710),
    [anon_sym_SLASH] = ACTIONS(710),
    [anon_sym_BANG] = ACTIONS(708),
    [anon_sym_TILDE] = ACTIONS(708),
    [anon_sym_typeof] = ACTIONS(710),
    [anon_sym_void] = ACTIONS(710),
    [anon_sym_delete] = ACTIONS(710),
    [anon_sym_PLUS_PLUS] = ACTIONS(708),
    [anon_sym_DASH_DASH] = ACTIONS(708),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(708),
    [sym_number] = ACTIONS(708),
    [sym_private_property_identifier] = ACTIONS(708),
    [sym_this] = ACTIONS(710),
    [sym_super] = ACTIONS(710),
    [sym_true] = ACTIONS(710),
    [sym_false] = ACTIONS(710),
    [sym_null] = ACTIONS(710),
    [sym_undefined] = ACTIONS(710),
    [anon_sym_AT] = ACTIONS(708),
    [anon_sym_static] = ACTIONS(710),
    [anon_sym_get] = ACTIONS(710),
    [anon_sym_set] = ACTIONS(710),
    [sym__automatic_semicolon] = ACTIONS(720),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(375)] = {
    [sym_formal_parameters] = STATE(1834),
    [sym_identifier] = ACTIONS(1110),
    [anon_sym_export] = ACTIONS(1112),
    [anon_sym_STAR] = ACTIONS(165),
    [anon_sym_COMMA] = ACTIONS(169),
    [anon_sym_RBRACE] = ACTIONS(169),
    [anon_sym_let] = ACTIONS(1112),
    [anon_sym_await] = ACTIONS(1112),
    [anon_sym_LPAREN] = ACTIONS(1095),
    [anon_sym_SEMI] = ACTIONS(169),
    [anon_sym_in] = ACTIONS(165),
    [anon_sym_EQ] = ACTIONS(490),
    [anon_sym_LBRACK] = ACTIONS(169),
    [anon_sym_LT] = ACTIONS(165),
    [anon_sym_GT] = ACTIONS(165),
    [anon_sym_DOT] = ACTIONS(169),
    [anon_sym_async] = ACTIONS(1112),
    [anon_sym_function] = ACTIONS(1114),
    [anon_sym_EQ_GT] = ACTIONS(200),
    [sym_optional_chain] = ACTIONS(169),
    [anon_sym_PLUS_EQ] = ACTIONS(202),
    [anon_sym_DASH_EQ] = ACTIONS(202),
    [anon_sym_STAR_EQ] = ACTIONS(202),
    [anon_sym_SLASH_EQ] = ACTIONS(202),
    [anon_sym_PERCENT_EQ] = ACTIONS(202),
    [anon_sym_CARET_EQ] = ACTIONS(202),
    [anon_sym_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_LT_LT_EQ] = ACTIONS(202),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(202),
    [anon_sym_QMARK_QMARK_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP] = ACTIONS(165),
    [anon_sym_PIPE_PIPE] = ACTIONS(165),
    [anon_sym_GT_GT] = ACTIONS(165),
    [anon_sym_GT_GT_GT] = ACTIONS(165),
    [anon_sym_LT_LT] = ACTIONS(165),
    [anon_sym_AMP] = ACTIONS(165),
    [anon_sym_CARET] = ACTIONS(165),
    [anon_sym_PIPE] = ACTIONS(165),
    [anon_sym_PLUS] = ACTIONS(165),
    [anon_sym_DASH] = ACTIONS(165),
    [anon_sym_SLASH] = ACTIONS(165),
    [anon_sym_PERCENT] = ACTIONS(165),
    [anon_sym_STAR_STAR] = ACTIONS(165),
    [anon_sym_LT_EQ] = ACTIONS(169),
    [anon_sym_EQ_EQ] = ACTIONS(165),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(169),
    [anon_sym_BANG_EQ] = ACTIONS(165),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(169),
    [anon_sym_GT_EQ] = ACTIONS(169),
    [anon_sym_QMARK_QMARK] = ACTIONS(165),
    [anon_sym_instanceof] = ACTIONS(165),
    [anon_sym_PLUS_PLUS] = ACTIONS(169),
    [anon_sym_DASH_DASH] = ACTIONS(169),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(169),
    [anon_sym_static] = ACTIONS(1112),
    [anon_sym_get] = ACTIONS(1112),
    [anon_sym_set] = ACTIONS(1112),
    [sym__automatic_semicolon] = ACTIONS(169),
    [sym__ternary_qmark] = ACTIONS(169),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(376)] = {
    [sym_formal_parameters] = STATE(1834),
    [sym_identifier] = ACTIONS(1110),
    [anon_sym_export] = ACTIONS(1112),
    [anon_sym_STAR] = ACTIONS(165),
    [anon_sym_COMMA] = ACTIONS(169),
    [anon_sym_let] = ACTIONS(1112),
    [anon_sym_await] = ACTIONS(1112),
    [anon_sym_LPAREN] = ACTIONS(1095),
    [anon_sym_SEMI] = ACTIONS(169),
    [anon_sym_in] = ACTIONS(165),
    [anon_sym_COLON] = ACTIONS(488),
    [anon_sym_EQ] = ACTIONS(490),
    [anon_sym_LBRACK] = ACTIONS(169),
    [anon_sym_LT] = ACTIONS(165),
    [anon_sym_GT] = ACTIONS(165),
    [anon_sym_DOT] = ACTIONS(169),
    [anon_sym_async] = ACTIONS(1112),
    [anon_sym_function] = ACTIONS(1120),
    [anon_sym_EQ_GT] = ACTIONS(200),
    [sym_optional_chain] = ACTIONS(169),
    [anon_sym_PLUS_EQ] = ACTIONS(202),
    [anon_sym_DASH_EQ] = ACTIONS(202),
    [anon_sym_STAR_EQ] = ACTIONS(202),
    [anon_sym_SLASH_EQ] = ACTIONS(202),
    [anon_sym_PERCENT_EQ] = ACTIONS(202),
    [anon_sym_CARET_EQ] = ACTIONS(202),
    [anon_sym_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_LT_LT_EQ] = ACTIONS(202),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(202),
    [anon_sym_QMARK_QMARK_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP] = ACTIONS(165),
    [anon_sym_PIPE_PIPE] = ACTIONS(165),
    [anon_sym_GT_GT] = ACTIONS(165),
    [anon_sym_GT_GT_GT] = ACTIONS(165),
    [anon_sym_LT_LT] = ACTIONS(165),
    [anon_sym_AMP] = ACTIONS(165),
    [anon_sym_CARET] = ACTIONS(165),
    [anon_sym_PIPE] = ACTIONS(165),
    [anon_sym_PLUS] = ACTIONS(165),
    [anon_sym_DASH] = ACTIONS(165),
    [anon_sym_SLASH] = ACTIONS(165),
    [anon_sym_PERCENT] = ACTIONS(165),
    [anon_sym_STAR_STAR] = ACTIONS(165),
    [anon_sym_LT_EQ] = ACTIONS(169),
    [anon_sym_EQ_EQ] = ACTIONS(165),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(169),
    [anon_sym_BANG_EQ] = ACTIONS(165),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(169),
    [anon_sym_GT_EQ] = ACTIONS(169),
    [anon_sym_QMARK_QMARK] = ACTIONS(165),
    [anon_sym_instanceof] = ACTIONS(165),
    [anon_sym_PLUS_PLUS] = ACTIONS(169),
    [anon_sym_DASH_DASH] = ACTIONS(169),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(169),
    [anon_sym_static] = ACTIONS(1112),
    [anon_sym_get] = ACTIONS(1112),
    [anon_sym_set] = ACTIONS(1112),
    [sym__automatic_semicolon] = ACTIONS(169),
    [sym__ternary_qmark] = ACTIONS(169),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(377)] = {
    [sym_formal_parameters] = STATE(1813),
    [sym_identifier] = ACTIONS(1116),
    [anon_sym_export] = ACTIONS(1118),
    [anon_sym_STAR] = ACTIONS(165),
    [anon_sym_COMMA] = ACTIONS(169),
    [anon_sym_let] = ACTIONS(1118),
    [anon_sym_await] = ACTIONS(1118),
    [anon_sym_of] = ACTIONS(165),
    [anon_sym_LPAREN] = ACTIONS(1095),
    [anon_sym_SEMI] = ACTIONS(169),
    [anon_sym_in] = ACTIONS(165),
    [anon_sym_EQ] = ACTIONS(528),
    [anon_sym_LBRACK] = ACTIONS(169),
    [anon_sym_LT] = ACTIONS(165),
    [anon_sym_GT] = ACTIONS(165),
    [anon_sym_DOT] = ACTIONS(169),
    [anon_sym_async] = ACTIONS(1118),
    [anon_sym_function] = ACTIONS(1114),
    [anon_sym_EQ_GT] = ACTIONS(500),
    [sym_optional_chain] = ACTIONS(169),
    [anon_sym_PLUS_EQ] = ACTIONS(202),
    [anon_sym_DASH_EQ] = ACTIONS(202),
    [anon_sym_STAR_EQ] = ACTIONS(202),
    [anon_sym_SLASH_EQ] = ACTIONS(202),
    [anon_sym_PERCENT_EQ] = ACTIONS(202),
    [anon_sym_CARET_EQ] = ACTIONS(202),
    [anon_sym_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_LT_LT_EQ] = ACTIONS(202),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(202),
    [anon_sym_QMARK_QMARK_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP] = ACTIONS(165),
    [anon_sym_PIPE_PIPE] = ACTIONS(165),
    [anon_sym_GT_GT] = ACTIONS(165),
    [anon_sym_GT_GT_GT] = ACTIONS(165),
    [anon_sym_LT_LT] = ACTIONS(165),
    [anon_sym_AMP] = ACTIONS(165),
    [anon_sym_CARET] = ACTIONS(165),
    [anon_sym_PIPE] = ACTIONS(165),
    [anon_sym_PLUS] = ACTIONS(165),
    [anon_sym_DASH] = ACTIONS(165),
    [anon_sym_SLASH] = ACTIONS(165),
    [anon_sym_PERCENT] = ACTIONS(165),
    [anon_sym_STAR_STAR] = ACTIONS(165),
    [anon_sym_LT_EQ] = ACTIONS(169),
    [anon_sym_EQ_EQ] = ACTIONS(165),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(169),
    [anon_sym_BANG_EQ] = ACTIONS(165),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(169),
    [anon_sym_GT_EQ] = ACTIONS(169),
    [anon_sym_QMARK_QMARK] = ACTIONS(165),
    [anon_sym_instanceof] = ACTIONS(165),
    [anon_sym_PLUS_PLUS] = ACTIONS(169),
    [anon_sym_DASH_DASH] = ACTIONS(169),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(169),
    [anon_sym_static] = ACTIONS(1118),
    [anon_sym_get] = ACTIONS(1118),
    [anon_sym_set] = ACTIONS(1118),
    [sym__automatic_semicolon] = ACTIONS(169),
    [sym__ternary_qmark] = ACTIONS(169),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(378)] = {
    [sym_formal_parameters] = STATE(1824),
    [sym_identifier] = ACTIONS(1122),
    [anon_sym_export] = ACTIONS(1124),
    [anon_sym_STAR] = ACTIONS(165),
    [anon_sym_COMMA] = ACTIONS(509),
    [anon_sym_RBRACE] = ACTIONS(509),
    [anon_sym_let] = ACTIONS(1124),
    [anon_sym_await] = ACTIONS(1124),
    [anon_sym_LPAREN] = ACTIONS(1095),
    [anon_sym_RPAREN] = ACTIONS(509),
    [anon_sym_in] = ACTIONS(165),
    [anon_sym_EQ] = ACTIONS(511),
    [anon_sym_LBRACK] = ACTIONS(169),
    [anon_sym_RBRACK] = ACTIONS(509),
    [anon_sym_LT] = ACTIONS(165),
    [anon_sym_GT] = ACTIONS(165),
    [anon_sym_DOT] = ACTIONS(169),
    [anon_sym_async] = ACTIONS(1124),
    [anon_sym_function] = ACTIONS(1098),
    [anon_sym_EQ_GT] = ACTIONS(498),
    [sym_optional_chain] = ACTIONS(169),
    [anon_sym_PLUS_EQ] = ACTIONS(202),
    [anon_sym_DASH_EQ] = ACTIONS(202),
    [anon_sym_STAR_EQ] = ACTIONS(202),
    [anon_sym_SLASH_EQ] = ACTIONS(202),
    [anon_sym_PERCENT_EQ] = ACTIONS(202),
    [anon_sym_CARET_EQ] = ACTIONS(202),
    [anon_sym_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_LT_LT_EQ] = ACTIONS(202),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(202),
    [anon_sym_QMARK_QMARK_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP] = ACTIONS(165),
    [anon_sym_PIPE_PIPE] = ACTIONS(165),
    [anon_sym_GT_GT] = ACTIONS(165),
    [anon_sym_GT_GT_GT] = ACTIONS(165),
    [anon_sym_LT_LT] = ACTIONS(165),
    [anon_sym_AMP] = ACTIONS(165),
    [anon_sym_CARET] = ACTIONS(165),
    [anon_sym_PIPE] = ACTIONS(165),
    [anon_sym_PLUS] = ACTIONS(165),
    [anon_sym_DASH] = ACTIONS(165),
    [anon_sym_SLASH] = ACTIONS(165),
    [anon_sym_PERCENT] = ACTIONS(165),
    [anon_sym_STAR_STAR] = ACTIONS(165),
    [anon_sym_LT_EQ] = ACTIONS(169),
    [anon_sym_EQ_EQ] = ACTIONS(165),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(169),
    [anon_sym_BANG_EQ] = ACTIONS(165),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(169),
    [anon_sym_GT_EQ] = ACTIONS(169),
    [anon_sym_QMARK_QMARK] = ACTIONS(165),
    [anon_sym_instanceof] = ACTIONS(165),
    [anon_sym_PLUS_PLUS] = ACTIONS(169),
    [anon_sym_DASH_DASH] = ACTIONS(169),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(169),
    [anon_sym_static] = ACTIONS(1124),
    [anon_sym_get] = ACTIONS(1124),
    [anon_sym_set] = ACTIONS(1124),
    [sym__ternary_qmark] = ACTIONS(169),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(379)] = {
    [sym_formal_parameters] = STATE(1859),
    [sym_identifier] = ACTIONS(1091),
    [anon_sym_export] = ACTIONS(1093),
    [anon_sym_STAR] = ACTIONS(165),
    [anon_sym_COMMA] = ACTIONS(514),
    [anon_sym_RBRACE] = ACTIONS(514),
    [anon_sym_let] = ACTIONS(1093),
    [anon_sym_await] = ACTIONS(1093),
    [anon_sym_LPAREN] = ACTIONS(1095),
    [anon_sym_RPAREN] = ACTIONS(514),
    [anon_sym_in] = ACTIONS(165),
    [anon_sym_EQ] = ACTIONS(517),
    [anon_sym_LBRACK] = ACTIONS(169),
    [anon_sym_RBRACK] = ACTIONS(514),
    [anon_sym_LT] = ACTIONS(165),
    [anon_sym_GT] = ACTIONS(165),
    [anon_sym_DOT] = ACTIONS(169),
    [anon_sym_async] = ACTIONS(1093),
    [anon_sym_function] = ACTIONS(1098),
    [anon_sym_EQ_GT] = ACTIONS(402),
    [sym_optional_chain] = ACTIONS(169),
    [anon_sym_PLUS_EQ] = ACTIONS(202),
    [anon_sym_DASH_EQ] = ACTIONS(202),
    [anon_sym_STAR_EQ] = ACTIONS(202),
    [anon_sym_SLASH_EQ] = ACTIONS(202),
    [anon_sym_PERCENT_EQ] = ACTIONS(202),
    [anon_sym_CARET_EQ] = ACTIONS(202),
    [anon_sym_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_LT_LT_EQ] = ACTIONS(202),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(202),
    [anon_sym_QMARK_QMARK_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP] = ACTIONS(165),
    [anon_sym_PIPE_PIPE] = ACTIONS(165),
    [anon_sym_GT_GT] = ACTIONS(165),
    [anon_sym_GT_GT_GT] = ACTIONS(165),
    [anon_sym_LT_LT] = ACTIONS(165),
    [anon_sym_AMP] = ACTIONS(165),
    [anon_sym_CARET] = ACTIONS(165),
    [anon_sym_PIPE] = ACTIONS(165),
    [anon_sym_PLUS] = ACTIONS(165),
    [anon_sym_DASH] = ACTIONS(165),
    [anon_sym_SLASH] = ACTIONS(165),
    [anon_sym_PERCENT] = ACTIONS(165),
    [anon_sym_STAR_STAR] = ACTIONS(165),
    [anon_sym_LT_EQ] = ACTIONS(169),
    [anon_sym_EQ_EQ] = ACTIONS(165),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(169),
    [anon_sym_BANG_EQ] = ACTIONS(165),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(169),
    [anon_sym_GT_EQ] = ACTIONS(169),
    [anon_sym_QMARK_QMARK] = ACTIONS(165),
    [anon_sym_instanceof] = ACTIONS(165),
    [anon_sym_PLUS_PLUS] = ACTIONS(169),
    [anon_sym_DASH_DASH] = ACTIONS(169),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(169),
    [anon_sym_static] = ACTIONS(1093),
    [anon_sym_get] = ACTIONS(1093),
    [anon_sym_set] = ACTIONS(1093),
    [sym__ternary_qmark] = ACTIONS(169),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(380)] = {
    [sym_variable_declarator] = STATE(1385),
    [sym_object_pattern] = STATE(1311),
    [sym_array_pattern] = STATE(1311),
    [sym__destructuring_pattern] = STATE(1311),
    [aux_sym_object_repeat1] = STATE(1418),
    [aux_sym_object_pattern_repeat1] = STATE(1452),
    [sym_identifier] = ACTIONS(1100),
    [anon_sym_STAR] = ACTIONS(165),
    [anon_sym_LBRACE] = ACTIONS(1102),
    [anon_sym_COMMA] = ACTIONS(169),
    [anon_sym_RBRACE] = ACTIONS(216),
    [anon_sym_of] = ACTIONS(1104),
    [anon_sym_LPAREN] = ACTIONS(944),
    [anon_sym_SEMI] = ACTIONS(169),
    [anon_sym_in] = ACTIONS(165),
    [anon_sym_COLON] = ACTIONS(183),
    [anon_sym_EQ] = ACTIONS(186),
    [anon_sym_LBRACK] = ACTIONS(1106),
    [anon_sym_LT] = ACTIONS(165),
    [anon_sym_GT] = ACTIONS(165),
    [anon_sym_DOT] = ACTIONS(169),
    [anon_sym_EQ_GT] = ACTIONS(200),
    [sym_optional_chain] = ACTIONS(169),
    [anon_sym_PLUS_EQ] = ACTIONS(202),
    [anon_sym_DASH_EQ] = ACTIONS(202),
    [anon_sym_STAR_EQ] = ACTIONS(202),
    [anon_sym_SLASH_EQ] = ACTIONS(202),
    [anon_sym_PERCENT_EQ] = ACTIONS(202),
    [anon_sym_CARET_EQ] = ACTIONS(202),
    [anon_sym_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_LT_LT_EQ] = ACTIONS(202),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(202),
    [anon_sym_QMARK_QMARK_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP] = ACTIONS(165),
    [anon_sym_PIPE_PIPE] = ACTIONS(165),
    [anon_sym_GT_GT] = ACTIONS(165),
    [anon_sym_GT_GT_GT] = ACTIONS(165),
    [anon_sym_LT_LT] = ACTIONS(165),
    [anon_sym_AMP] = ACTIONS(165),
    [anon_sym_CARET] = ACTIONS(165),
    [anon_sym_PIPE] = ACTIONS(165),
    [anon_sym_PLUS] = ACTIONS(165),
    [anon_sym_DASH] = ACTIONS(165),
    [anon_sym_SLASH] = ACTIONS(165),
    [anon_sym_PERCENT] = ACTIONS(165),
    [anon_sym_STAR_STAR] = ACTIONS(165),
    [anon_sym_LT_EQ] = ACTIONS(169),
    [anon_sym_EQ_EQ] = ACTIONS(165),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(169),
    [anon_sym_BANG_EQ] = ACTIONS(165),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(169),
    [anon_sym_GT_EQ] = ACTIONS(169),
    [anon_sym_QMARK_QMARK] = ACTIONS(165),
    [anon_sym_instanceof] = ACTIONS(165),
    [anon_sym_PLUS_PLUS] = ACTIONS(169),
    [anon_sym_DASH_DASH] = ACTIONS(169),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(169),
    [sym__automatic_semicolon] = ACTIONS(169),
    [sym__ternary_qmark] = ACTIONS(169),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(381)] = {
    [sym_catch_clause] = STATE(389),
    [sym_finally_clause] = STATE(435),
    [ts_builtin_sym_end] = ACTIONS(1126),
    [sym_identifier] = ACTIONS(1128),
    [anon_sym_export] = ACTIONS(1128),
    [anon_sym_default] = ACTIONS(1128),
    [anon_sym_LBRACE] = ACTIONS(1126),
    [anon_sym_RBRACE] = ACTIONS(1126),
    [anon_sym_import] = ACTIONS(1128),
    [anon_sym_with] = ACTIONS(1128),
    [anon_sym_var] = ACTIONS(1128),
    [anon_sym_let] = ACTIONS(1128),
    [anon_sym_const] = ACTIONS(1128),
    [anon_sym_using] = ACTIONS(1128),
    [anon_sym_await] = ACTIONS(1128),
    [anon_sym_else] = ACTIONS(1128),
    [anon_sym_if] = ACTIONS(1128),
    [anon_sym_switch] = ACTIONS(1128),
    [anon_sym_for] = ACTIONS(1128),
    [anon_sym_LPAREN] = ACTIONS(1126),
    [anon_sym_SEMI] = ACTIONS(1126),
    [anon_sym_while] = ACTIONS(1128),
    [anon_sym_do] = ACTIONS(1128),
    [anon_sym_try] = ACTIONS(1128),
    [anon_sym_break] = ACTIONS(1128),
    [anon_sym_continue] = ACTIONS(1128),
    [anon_sym_debugger] = ACTIONS(1128),
    [anon_sym_return] = ACTIONS(1128),
    [anon_sym_throw] = ACTIONS(1128),
    [anon_sym_case] = ACTIONS(1128),
    [anon_sym_catch] = ACTIONS(1130),
    [anon_sym_finally] = ACTIONS(1132),
    [anon_sym_yield] = ACTIONS(1128),
    [anon_sym_LBRACK] = ACTIONS(1126),
    [anon_sym_LT] = ACTIONS(1126),
    [anon_sym_DQUOTE] = ACTIONS(1126),
    [anon_sym_SQUOTE] = ACTIONS(1126),
    [anon_sym_class] = ACTIONS(1128),
    [anon_sym_async] = ACTIONS(1128),
    [anon_sym_function] = ACTIONS(1128),
    [anon_sym_new] = ACTIONS(1128),
    [anon_sym_PLUS] = ACTIONS(1128),
    [anon_sym_DASH] = ACTIONS(1128),
    [anon_sym_SLASH] = ACTIONS(1128),
    [anon_sym_BANG] = ACTIONS(1126),
    [anon_sym_TILDE] = ACTIONS(1126),
    [anon_sym_typeof] = ACTIONS(1128),
    [anon_sym_void] = ACTIONS(1128),
    [anon_sym_delete] = ACTIONS(1128),
    [anon_sym_PLUS_PLUS] = ACTIONS(1126),
    [anon_sym_DASH_DASH] = ACTIONS(1126),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(1126),
    [sym_number] = ACTIONS(1126),
    [sym_private_property_identifier] = ACTIONS(1126),
    [sym_this] = ACTIONS(1128),
    [sym_super] = ACTIONS(1128),
    [sym_true] = ACTIONS(1128),
    [sym_false] = ACTIONS(1128),
    [sym_null] = ACTIONS(1128),
    [sym_undefined] = ACTIONS(1128),
    [anon_sym_AT] = ACTIONS(1126),
    [anon_sym_static] = ACTIONS(1128),
    [anon_sym_get] = ACTIONS(1128),
    [anon_sym_set] = ACTIONS(1128),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(382)] = {
    [sym_formal_parameters] = STATE(1824),
    [sym_identifier] = ACTIONS(1122),
    [anon_sym_export] = ACTIONS(1124),
    [anon_sym_STAR] = ACTIONS(165),
    [anon_sym_COMMA] = ACTIONS(496),
    [anon_sym_RBRACE] = ACTIONS(496),
    [anon_sym_let] = ACTIONS(1124),
    [anon_sym_await] = ACTIONS(1124),
    [anon_sym_LPAREN] = ACTIONS(1095),
    [anon_sym_RPAREN] = ACTIONS(496),
    [anon_sym_in] = ACTIONS(165),
    [anon_sym_EQ] = ACTIONS(440),
    [anon_sym_LBRACK] = ACTIONS(169),
    [anon_sym_RBRACK] = ACTIONS(496),
    [anon_sym_LT] = ACTIONS(165),
    [anon_sym_GT] = ACTIONS(165),
    [anon_sym_DOT] = ACTIONS(169),
    [anon_sym_async] = ACTIONS(1124),
    [anon_sym_function] = ACTIONS(1098),
    [anon_sym_EQ_GT] = ACTIONS(498),
    [sym_optional_chain] = ACTIONS(169),
    [anon_sym_PLUS_EQ] = ACTIONS(202),
    [anon_sym_DASH_EQ] = ACTIONS(202),
    [anon_sym_STAR_EQ] = ACTIONS(202),
    [anon_sym_SLASH_EQ] = ACTIONS(202),
    [anon_sym_PERCENT_EQ] = ACTIONS(202),
    [anon_sym_CARET_EQ] = ACTIONS(202),
    [anon_sym_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_LT_LT_EQ] = ACTIONS(202),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(202),
    [anon_sym_QMARK_QMARK_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP] = ACTIONS(165),
    [anon_sym_PIPE_PIPE] = ACTIONS(165),
    [anon_sym_GT_GT] = ACTIONS(165),
    [anon_sym_GT_GT_GT] = ACTIONS(165),
    [anon_sym_LT_LT] = ACTIONS(165),
    [anon_sym_AMP] = ACTIONS(165),
    [anon_sym_CARET] = ACTIONS(165),
    [anon_sym_PIPE] = ACTIONS(165),
    [anon_sym_PLUS] = ACTIONS(165),
    [anon_sym_DASH] = ACTIONS(165),
    [anon_sym_SLASH] = ACTIONS(165),
    [anon_sym_PERCENT] = ACTIONS(165),
    [anon_sym_STAR_STAR] = ACTIONS(165),
    [anon_sym_LT_EQ] = ACTIONS(169),
    [anon_sym_EQ_EQ] = ACTIONS(165),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(169),
    [anon_sym_BANG_EQ] = ACTIONS(165),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(169),
    [anon_sym_GT_EQ] = ACTIONS(169),
    [anon_sym_QMARK_QMARK] = ACTIONS(165),
    [anon_sym_instanceof] = ACTIONS(165),
    [anon_sym_PLUS_PLUS] = ACTIONS(169),
    [anon_sym_DASH_DASH] = ACTIONS(169),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(169),
    [anon_sym_static] = ACTIONS(1124),
    [anon_sym_get] = ACTIONS(1124),
    [anon_sym_set] = ACTIONS(1124),
    [sym__ternary_qmark] = ACTIONS(169),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(383)] = {
    [sym_formal_parameters] = STATE(1859),
    [sym_identifier] = ACTIONS(1091),
    [anon_sym_export] = ACTIONS(1093),
    [anon_sym_STAR] = ACTIONS(165),
    [anon_sym_COMMA] = ACTIONS(169),
    [anon_sym_let] = ACTIONS(1093),
    [anon_sym_await] = ACTIONS(1093),
    [anon_sym_of] = ACTIONS(504),
    [anon_sym_LPAREN] = ACTIONS(1095),
    [anon_sym_SEMI] = ACTIONS(169),
    [anon_sym_in] = ACTIONS(506),
    [anon_sym_EQ] = ACTIONS(384),
    [anon_sym_LBRACK] = ACTIONS(169),
    [anon_sym_LT] = ACTIONS(165),
    [anon_sym_GT] = ACTIONS(165),
    [anon_sym_DOT] = ACTIONS(169),
    [anon_sym_async] = ACTIONS(1093),
    [anon_sym_function] = ACTIONS(1098),
    [anon_sym_EQ_GT] = ACTIONS(402),
    [sym_optional_chain] = ACTIONS(169),
    [anon_sym_PLUS_EQ] = ACTIONS(202),
    [anon_sym_DASH_EQ] = ACTIONS(202),
    [anon_sym_STAR_EQ] = ACTIONS(202),
    [anon_sym_SLASH_EQ] = ACTIONS(202),
    [anon_sym_PERCENT_EQ] = ACTIONS(202),
    [anon_sym_CARET_EQ] = ACTIONS(202),
    [anon_sym_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_LT_LT_EQ] = ACTIONS(202),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(202),
    [anon_sym_QMARK_QMARK_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP] = ACTIONS(165),
    [anon_sym_PIPE_PIPE] = ACTIONS(165),
    [anon_sym_GT_GT] = ACTIONS(165),
    [anon_sym_GT_GT_GT] = ACTIONS(165),
    [anon_sym_LT_LT] = ACTIONS(165),
    [anon_sym_AMP] = ACTIONS(165),
    [anon_sym_CARET] = ACTIONS(165),
    [anon_sym_PIPE] = ACTIONS(165),
    [anon_sym_PLUS] = ACTIONS(165),
    [anon_sym_DASH] = ACTIONS(165),
    [anon_sym_SLASH] = ACTIONS(165),
    [anon_sym_PERCENT] = ACTIONS(165),
    [anon_sym_STAR_STAR] = ACTIONS(165),
    [anon_sym_LT_EQ] = ACTIONS(169),
    [anon_sym_EQ_EQ] = ACTIONS(165),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(169),
    [anon_sym_BANG_EQ] = ACTIONS(165),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(169),
    [anon_sym_GT_EQ] = ACTIONS(169),
    [anon_sym_QMARK_QMARK] = ACTIONS(165),
    [anon_sym_instanceof] = ACTIONS(165),
    [anon_sym_PLUS_PLUS] = ACTIONS(169),
    [anon_sym_DASH_DASH] = ACTIONS(169),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(169),
    [anon_sym_static] = ACTIONS(1093),
    [anon_sym_get] = ACTIONS(1093),
    [anon_sym_set] = ACTIONS(1093),
    [sym__ternary_qmark] = ACTIONS(169),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(384)] = {
    [ts_builtin_sym_end] = ACTIONS(700),
    [sym_identifier] = ACTIONS(702),
    [anon_sym_export] = ACTIONS(702),
    [anon_sym_default] = ACTIONS(702),
    [anon_sym_LBRACE] = ACTIONS(700),
    [anon_sym_COMMA] = ACTIONS(700),
    [anon_sym_RBRACE] = ACTIONS(700),
    [anon_sym_import] = ACTIONS(702),
    [anon_sym_with] = ACTIONS(702),
    [anon_sym_var] = ACTIONS(702),
    [anon_sym_let] = ACTIONS(702),
    [anon_sym_const] = ACTIONS(702),
    [anon_sym_using] = ACTIONS(702),
    [anon_sym_await] = ACTIONS(702),
    [anon_sym_else] = ACTIONS(702),
    [anon_sym_if] = ACTIONS(702),
    [anon_sym_switch] = ACTIONS(702),
    [anon_sym_for] = ACTIONS(702),
    [anon_sym_LPAREN] = ACTIONS(700),
    [anon_sym_SEMI] = ACTIONS(700),
    [anon_sym_while] = ACTIONS(702),
    [anon_sym_do] = ACTIONS(702),
    [anon_sym_try] = ACTIONS(702),
    [anon_sym_break] = ACTIONS(702),
    [anon_sym_continue] = ACTIONS(702),
    [anon_sym_debugger] = ACTIONS(702),
    [anon_sym_return] = ACTIONS(702),
    [anon_sym_throw] = ACTIONS(702),
    [anon_sym_case] = ACTIONS(702),
    [anon_sym_catch] = ACTIONS(702),
    [anon_sym_finally] = ACTIONS(702),
    [anon_sym_yield] = ACTIONS(702),
    [anon_sym_LBRACK] = ACTIONS(700),
    [anon_sym_LT] = ACTIONS(700),
    [anon_sym_DQUOTE] = ACTIONS(700),
    [anon_sym_SQUOTE] = ACTIONS(700),
    [anon_sym_class] = ACTIONS(702),
    [anon_sym_async] = ACTIONS(702),
    [anon_sym_function] = ACTIONS(702),
    [anon_sym_new] = ACTIONS(702),
    [anon_sym_PLUS] = ACTIONS(702),
    [anon_sym_DASH] = ACTIONS(702),
    [anon_sym_SLASH] = ACTIONS(702),
    [anon_sym_BANG] = ACTIONS(700),
    [anon_sym_TILDE] = ACTIONS(700),
    [anon_sym_typeof] = ACTIONS(702),
    [anon_sym_void] = ACTIONS(702),
    [anon_sym_delete] = ACTIONS(702),
    [anon_sym_PLUS_PLUS] = ACTIONS(700),
    [anon_sym_DASH_DASH] = ACTIONS(700),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(700),
    [sym_number] = ACTIONS(700),
    [sym_private_property_identifier] = ACTIONS(700),
    [sym_this] = ACTIONS(702),
    [sym_super] = ACTIONS(702),
    [sym_true] = ACTIONS(702),
    [sym_false] = ACTIONS(702),
    [sym_null] = ACTIONS(702),
    [sym_undefined] = ACTIONS(702),
    [anon_sym_AT] = ACTIONS(700),
    [anon_sym_static] = ACTIONS(702),
    [anon_sym_get] = ACTIONS(702),
    [anon_sym_set] = ACTIONS(702),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(385)] = {
    [ts_builtin_sym_end] = ACTIONS(694),
    [sym_identifier] = ACTIONS(696),
    [anon_sym_export] = ACTIONS(696),
    [anon_sym_default] = ACTIONS(696),
    [anon_sym_LBRACE] = ACTIONS(694),
    [anon_sym_COMMA] = ACTIONS(694),
    [anon_sym_RBRACE] = ACTIONS(694),
    [anon_sym_import] = ACTIONS(696),
    [anon_sym_with] = ACTIONS(696),
    [anon_sym_var] = ACTIONS(696),
    [anon_sym_let] = ACTIONS(696),
    [anon_sym_const] = ACTIONS(696),
    [anon_sym_using] = ACTIONS(696),
    [anon_sym_await] = ACTIONS(696),
    [anon_sym_else] = ACTIONS(696),
    [anon_sym_if] = ACTIONS(696),
    [anon_sym_switch] = ACTIONS(696),
    [anon_sym_for] = ACTIONS(696),
    [anon_sym_LPAREN] = ACTIONS(694),
    [anon_sym_SEMI] = ACTIONS(694),
    [anon_sym_while] = ACTIONS(696),
    [anon_sym_do] = ACTIONS(696),
    [anon_sym_try] = ACTIONS(696),
    [anon_sym_break] = ACTIONS(696),
    [anon_sym_continue] = ACTIONS(696),
    [anon_sym_debugger] = ACTIONS(696),
    [anon_sym_return] = ACTIONS(696),
    [anon_sym_throw] = ACTIONS(696),
    [anon_sym_case] = ACTIONS(696),
    [anon_sym_catch] = ACTIONS(696),
    [anon_sym_finally] = ACTIONS(696),
    [anon_sym_yield] = ACTIONS(696),
    [anon_sym_LBRACK] = ACTIONS(694),
    [anon_sym_LT] = ACTIONS(694),
    [anon_sym_DQUOTE] = ACTIONS(694),
    [anon_sym_SQUOTE] = ACTIONS(694),
    [anon_sym_class] = ACTIONS(696),
    [anon_sym_async] = ACTIONS(696),
    [anon_sym_function] = ACTIONS(696),
    [anon_sym_new] = ACTIONS(696),
    [anon_sym_PLUS] = ACTIONS(696),
    [anon_sym_DASH] = ACTIONS(696),
    [anon_sym_SLASH] = ACTIONS(696),
    [anon_sym_BANG] = ACTIONS(694),
    [anon_sym_TILDE] = ACTIONS(694),
    [anon_sym_typeof] = ACTIONS(696),
    [anon_sym_void] = ACTIONS(696),
    [anon_sym_delete] = ACTIONS(696),
    [anon_sym_PLUS_PLUS] = ACTIONS(694),
    [anon_sym_DASH_DASH] = ACTIONS(694),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(694),
    [sym_number] = ACTIONS(694),
    [sym_private_property_identifier] = ACTIONS(694),
    [sym_this] = ACTIONS(696),
    [sym_super] = ACTIONS(696),
    [sym_true] = ACTIONS(696),
    [sym_false] = ACTIONS(696),
    [sym_null] = ACTIONS(696),
    [sym_undefined] = ACTIONS(696),
    [anon_sym_AT] = ACTIONS(694),
    [anon_sym_static] = ACTIONS(696),
    [anon_sym_get] = ACTIONS(696),
    [anon_sym_set] = ACTIONS(696),
    [sym_html_comment] = ACTIONS(5),
  },
  [STATE(386)] = {
    [sym_formal_parameters] = STATE(1859),
    [sym_identifier] = ACTIONS(1091),
    [anon_sym_export] = ACTIONS(1093),
    [anon_sym_STAR] = ACTIONS(165),
    [anon_sym_COMMA] = ACTIONS(577),
    [anon_sym_RBRACE] = ACTIONS(577),
    [anon_sym_let] = ACTIONS(1093),
    [anon_sym_await] = ACTIONS(1093),
    [anon_sym_LPAREN] = ACTIONS(1095),
    [anon_sym_in] = ACTIONS(165),
    [anon_sym_EQ] = ACTIONS(384),
    [anon_sym_LBRACK] = ACTIONS(169),
    [anon_sym_RBRACK] = ACTIONS(577),
    [anon_sym_LT] = ACTIONS(165),
    [anon_sym_GT] = ACTIONS(165),
    [anon_sym_DOT] = ACTIONS(169),
    [anon_sym_async] = ACTIONS(1093),
    [anon_sym_function] = ACTIONS(1098),
    [anon_sym_EQ_GT] = ACTIONS(402),
    [sym_optional_chain] = ACTIONS(169),
    [anon_sym_PLUS_EQ] = ACTIONS(202),
    [anon_sym_DASH_EQ] = ACTIONS(202),
    [anon_sym_STAR_EQ] = ACTIONS(202),
    [anon_sym_SLASH_EQ] = ACTIONS(202),
    [anon_sym_PERCENT_EQ] = ACTIONS(202),
    [anon_sym_CARET_EQ] = ACTIONS(202),
    [anon_sym_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_GT_GT_GT_EQ] = ACTIONS(202),
    [anon_sym_LT_LT_EQ] = ACTIONS(202),
    [anon_sym_STAR_STAR_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP_EQ] = ACTIONS(202),
    [anon_sym_PIPE_PIPE_EQ] = ACTIONS(202),
    [anon_sym_QMARK_QMARK_EQ] = ACTIONS(202),
    [anon_sym_AMP_AMP] = ACTIONS(165),
    [anon_sym_PIPE_PIPE] = ACTIONS(165),
    [anon_sym_GT_GT] = ACTIONS(165),
    [anon_sym_GT_GT_GT] = ACTIONS(165),
    [anon_sym_LT_LT] = ACTIONS(165),
    [anon_sym_AMP] = ACTIONS(165),
    [anon_sym_CARET] = ACTIONS(165),
    [anon_sym_PIPE] = ACTIONS(165),
    [anon_sym_PLUS] = ACTIONS(165),
    [anon_sym_DASH] = ACTIONS(165),
    [anon_sym_SLASH] = ACTIONS(165),
    [anon_sym_PERCENT] = ACTIONS(165),
    [anon_sym_STAR_STAR] = ACTIONS(165),
    [anon_sym_LT_EQ] = ACTIONS(169),
    [anon_sym_EQ_EQ] = ACTIONS(165),
    [anon_sym_EQ_EQ_EQ] = ACTIONS(169),
    [anon_sym_BANG_EQ] = ACTIONS(165),
    [anon_sym_BANG_EQ_EQ] = ACTIONS(169),
    [anon_sym_GT_EQ] = ACTIONS(169),
    [anon_sym_QMARK_QMARK] = ACTIONS(165),
    [anon_sym_instanceof] = ACTIONS(165),
    [anon_sym_PLUS_PLUS] = ACTIONS(169),
    [anon_sym_DASH_DASH] = ACTIONS(169),
    [sym_comment] = ACTIONS(5),
    [anon_sym_BQUOTE] = ACTIONS(169),
    [anon_sym_static] = ACTIONS(1093),
    [anon_sym_get] = ACTIONS(1093),
    [anon_sym_set] = ACTIONS(1093),
    [sym__ternary_qmark] = ACTIONS(169),
    [sym_html_comment] = ACTIONS(5),
  },
};

static const uint16_t ts_small_parse_table[] = {
  [0] = 13,
    ACTIONS(402), 1,
      anon_sym_EQ_GT,
    ACTIONS(509), 1,
      anon_sym_COMMA,
    ACTIONS(514), 1,
      anon_sym_RBRACK,
    ACTIONS(517), 1,
      anon_sym_EQ,
    ACTIONS(1091), 1,
      sym_identifier,
    ACTIONS(1095), 1,
      anon_sym_LPAREN,
    ACTIONS(1098), 1,
      anon_sym_function,
    STATE(1859), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1093), 7,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
    ACTIONS(169), 11,
      sym__ternary_qmark,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
    ACTIONS(202), 15,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_CARET_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_AMP_AMP_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(165), 21,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_AMP,
      anon_sym_CARET,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
  [91] = 11,
    ACTIONS(200), 1,
      anon_sym_EQ_GT,
    ACTIONS(490), 1,
      anon_sym_EQ,
    ACTIONS(1095), 1,
      anon_sym_LPAREN,
    ACTIONS(1110), 1,
      sym_identifier,
    ACTIONS(1120), 1,
      anon_sym_function,
    STATE(1834), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1112), 7,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
    ACTIONS(169), 13,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
    ACTIONS(202), 15,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_CARET_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_AMP_AMP_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(165), 21,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_AMP,
      anon_sym_CARET,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
  [178] = 5,
    ACTIONS(1132), 1,
      anon_sym_finally,
    STATE(460), 1,
      sym_finally_clause,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1134), 17,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(1136), 43,
      anon_sym_export,
      anon_sym_default,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_else,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_case,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [253] = 11,
    ACTIONS(200), 1,
      anon_sym_EQ_GT,
    ACTIONS(490), 1,
      anon_sym_EQ,
    ACTIONS(1079), 1,
      anon_sym_function,
    ACTIONS(1095), 1,
      anon_sym_LPAREN,
    ACTIONS(1110), 1,
      sym_identifier,
    STATE(1834), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1112), 7,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
    ACTIONS(169), 13,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
    ACTIONS(202), 15,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_CARET_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_AMP_AMP_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(165), 21,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_AMP,
      anon_sym_CARET,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
  [340] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(694), 18,
      sym__automatic_semicolon,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(696), 43,
      anon_sym_export,
      anon_sym_default,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_else,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_case,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [410] = 4,
    ACTIONS(730), 1,
      sym__automatic_semicolon,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(722), 17,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(724), 43,
      anon_sym_export,
      anon_sym_default,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_else,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_case,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [482] = 11,
    ACTIONS(440), 1,
      anon_sym_EQ,
    ACTIONS(628), 1,
      anon_sym_EQ_GT,
    ACTIONS(1095), 1,
      anon_sym_LPAREN,
    ACTIONS(1098), 1,
      anon_sym_function,
    ACTIONS(1138), 1,
      sym_identifier,
    STATE(1821), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1140), 7,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
    ACTIONS(169), 12,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
    ACTIONS(202), 15,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_CARET_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_AMP_AMP_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(165), 21,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_AMP,
      anon_sym_CARET,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
  [568] = 4,
    ACTIONS(1142), 1,
      sym__automatic_semicolon,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(708), 17,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(710), 43,
      anon_sym_export,
      anon_sym_default,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_else,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_case,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [640] = 4,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1148), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
    ACTIONS(1144), 16,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(1146), 43,
      anon_sym_export,
      anon_sym_default,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_else,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_case,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [712] = 13,
    ACTIONS(200), 1,
      anon_sym_EQ_GT,
    ACTIONS(490), 1,
      anon_sym_EQ,
    ACTIONS(494), 1,
      anon_sym_COLON,
    ACTIONS(1100), 1,
      sym_identifier,
    ACTIONS(1102), 1,
      anon_sym_LBRACE,
    ACTIONS(1104), 1,
      anon_sym_of,
    ACTIONS(1106), 1,
      anon_sym_LBRACK,
    STATE(1385), 1,
      sym_variable_declarator,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    STATE(1311), 3,
      sym_object_pattern,
      sym_array_pattern,
      sym__destructuring_pattern,
    ACTIONS(169), 14,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
    ACTIONS(202), 15,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_CARET_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_AMP_AMP_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(165), 21,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_AMP,
      anon_sym_CARET,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
  [802] = 13,
    ACTIONS(200), 1,
      anon_sym_EQ_GT,
    ACTIONS(488), 1,
      anon_sym_COLON,
    ACTIONS(490), 1,
      anon_sym_EQ,
    ACTIONS(1100), 1,
      sym_identifier,
    ACTIONS(1102), 1,
      anon_sym_LBRACE,
    ACTIONS(1104), 1,
      anon_sym_of,
    ACTIONS(1106), 1,
      anon_sym_LBRACK,
    STATE(1385), 1,
      sym_variable_declarator,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    STATE(1311), 3,
      sym_object_pattern,
      sym_array_pattern,
      sym__destructuring_pattern,
    ACTIONS(169), 14,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
    ACTIONS(202), 15,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_CARET_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_AMP_AMP_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(165), 21,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_AMP,
      anon_sym_CARET,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
  [892] = 4,
    ACTIONS(780), 1,
      sym__automatic_semicolon,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(772), 17,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(774), 43,
      anon_sym_export,
      anon_sym_default,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_else,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_case,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [964] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(704), 18,
      sym__automatic_semicolon,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(706), 43,
      anon_sym_export,
      anon_sym_default,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_else,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_case,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [1034] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(700), 18,
      sym__automatic_semicolon,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(702), 43,
      anon_sym_export,
      anon_sym_default,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_else,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_case,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [1104] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1150), 17,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(1152), 44,
      anon_sym_export,
      anon_sym_default,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_else,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_case,
      anon_sym_finally,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [1174] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(690), 18,
      sym__automatic_semicolon,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(692), 43,
      anon_sym_export,
      anon_sym_default,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_else,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_case,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [1244] = 4,
    ACTIONS(740), 1,
      sym__automatic_semicolon,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(732), 17,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(734), 43,
      anon_sym_export,
      anon_sym_default,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_else,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_case,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [1316] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1154), 18,
      sym__automatic_semicolon,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(1156), 43,
      anon_sym_export,
      anon_sym_default,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_else,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_case,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [1386] = 4,
    ACTIONS(770), 1,
      sym__automatic_semicolon,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(762), 17,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(764), 43,
      anon_sym_export,
      anon_sym_default,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_else,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_case,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [1458] = 11,
    ACTIONS(624), 1,
      anon_sym_EQ,
    ACTIONS(628), 1,
      anon_sym_EQ_GT,
    ACTIONS(1095), 1,
      anon_sym_LPAREN,
    ACTIONS(1098), 1,
      anon_sym_function,
    ACTIONS(1138), 1,
      sym_identifier,
    STATE(1821), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1140), 7,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
    ACTIONS(169), 12,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
    ACTIONS(202), 15,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_CARET_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_AMP_AMP_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(165), 21,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_AMP,
      anon_sym_CARET,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
  [1544] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1158), 17,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(1160), 44,
      anon_sym_export,
      anon_sym_default,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_else,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_case,
      anon_sym_finally,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [1614] = 4,
    ACTIONS(760), 1,
      sym__automatic_semicolon,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(752), 17,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(754), 43,
      anon_sym_export,
      anon_sym_default,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_else,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_case,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [1686] = 4,
    ACTIONS(790), 1,
      sym__automatic_semicolon,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(782), 17,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(784), 43,
      anon_sym_export,
      anon_sym_default,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_else,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_case,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [1758] = 11,
    ACTIONS(440), 1,
      anon_sym_EQ,
    ACTIONS(498), 1,
      anon_sym_EQ_GT,
    ACTIONS(1095), 1,
      anon_sym_LPAREN,
    ACTIONS(1098), 1,
      anon_sym_function,
    ACTIONS(1122), 1,
      sym_identifier,
    STATE(1824), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1124), 7,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
    ACTIONS(169), 12,
      sym__ternary_qmark,
      anon_sym_COLON,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
    ACTIONS(202), 15,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_CARET_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_AMP_AMP_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(165), 21,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_AMP,
      anon_sym_CARET,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
  [1844] = 5,
    ACTIONS(1166), 1,
      anon_sym_else,
    STATE(428), 1,
      sym_else_clause,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1162), 17,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(1164), 42,
      anon_sym_export,
      anon_sym_default,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_case,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [1918] = 11,
    ACTIONS(580), 1,
      anon_sym_EQ_GT,
    ACTIONS(590), 1,
      anon_sym_EQ,
    ACTIONS(1095), 1,
      anon_sym_LPAREN,
    ACTIONS(1098), 1,
      anon_sym_function,
    ACTIONS(1168), 1,
      sym_identifier,
    STATE(1782), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1170), 7,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
    ACTIONS(169), 11,
      sym__ternary_qmark,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
    ACTIONS(202), 15,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_CARET_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_AMP_AMP_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(165), 22,
      anon_sym_STAR,
      anon_sym_of,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_AMP,
      anon_sym_CARET,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
  [2004] = 4,
    ACTIONS(800), 1,
      sym__automatic_semicolon,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(792), 17,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(794), 43,
      anon_sym_export,
      anon_sym_default,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_else,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_case,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [2076] = 11,
    ACTIONS(440), 1,
      anon_sym_EQ,
    ACTIONS(580), 1,
      anon_sym_EQ_GT,
    ACTIONS(1095), 1,
      anon_sym_LPAREN,
    ACTIONS(1098), 1,
      anon_sym_function,
    ACTIONS(1168), 1,
      sym_identifier,
    STATE(1782), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1170), 7,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
    ACTIONS(169), 11,
      sym__ternary_qmark,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
    ACTIONS(202), 15,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_CARET_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_AMP_AMP_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(165), 22,
      anon_sym_STAR,
      anon_sym_of,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_AMP,
      anon_sym_CARET,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
  [2162] = 4,
    ACTIONS(750), 1,
      sym__automatic_semicolon,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(742), 17,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(744), 43,
      anon_sym_export,
      anon_sym_default,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_else,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_case,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [2234] = 4,
    ACTIONS(1172), 1,
      sym__automatic_semicolon,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(694), 17,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(696), 43,
      anon_sym_export,
      anon_sym_default,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_else,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_case,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [2306] = 13,
    ACTIONS(440), 1,
      anon_sym_EQ,
    ACTIONS(498), 1,
      anon_sym_EQ_GT,
    ACTIONS(504), 1,
      anon_sym_of,
    ACTIONS(506), 1,
      anon_sym_in,
    ACTIONS(1095), 1,
      anon_sym_LPAREN,
    ACTIONS(1098), 1,
      anon_sym_function,
    ACTIONS(1122), 1,
      sym_identifier,
    STATE(1824), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1124), 7,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
    ACTIONS(169), 11,
      sym__ternary_qmark,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
    ACTIONS(202), 15,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_CARET_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_AMP_AMP_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(165), 20,
      anon_sym_STAR,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_AMP,
      anon_sym_CARET,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
  [2396] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1174), 18,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(1176), 43,
      anon_sym_export,
      anon_sym_default,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_else,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_case,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [2466] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1178), 17,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(1180), 43,
      anon_sym_export,
      anon_sym_default,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_else,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_case,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [2535] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1182), 17,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(1184), 43,
      anon_sym_export,
      anon_sym_default,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_else,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_case,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [2604] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1186), 17,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(1188), 43,
      anon_sym_export,
      anon_sym_default,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_else,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_case,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [2673] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1186), 17,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(1188), 43,
      anon_sym_export,
      anon_sym_default,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_else,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_case,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [2742] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1190), 17,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(1192), 43,
      anon_sym_export,
      anon_sym_default,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_else,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_case,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [2811] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1194), 17,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(1196), 43,
      anon_sym_export,
      anon_sym_default,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_else,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_case,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [2880] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1198), 17,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(1200), 43,
      anon_sym_export,
      anon_sym_default,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_else,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_case,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [2949] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1202), 17,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(1204), 43,
      anon_sym_export,
      anon_sym_default,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_else,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_case,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [3018] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1206), 17,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(1208), 43,
      anon_sym_export,
      anon_sym_default,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_else,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_case,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [3087] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1210), 17,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(1212), 43,
      anon_sym_export,
      anon_sym_default,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_else,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_case,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [3156] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1214), 17,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(1216), 43,
      anon_sym_export,
      anon_sym_default,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_else,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_case,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [3225] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1218), 17,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(1220), 43,
      anon_sym_export,
      anon_sym_default,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_else,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_case,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [3294] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1222), 17,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(1224), 43,
      anon_sym_export,
      anon_sym_default,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_else,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_case,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [3363] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1226), 17,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(1228), 43,
      anon_sym_export,
      anon_sym_default,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_else,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_case,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [3432] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1230), 17,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(1232), 43,
      anon_sym_export,
      anon_sym_default,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_else,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_case,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [3501] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1234), 17,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(1236), 43,
      anon_sym_export,
      anon_sym_default,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_else,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_case,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [3570] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1238), 17,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(1240), 43,
      anon_sym_export,
      anon_sym_default,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_else,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_case,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [3639] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1242), 17,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(1244), 43,
      anon_sym_export,
      anon_sym_default,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_else,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_case,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [3708] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1246), 17,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(1248), 43,
      anon_sym_export,
      anon_sym_default,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_else,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_case,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [3777] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1250), 17,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(1252), 43,
      anon_sym_export,
      anon_sym_default,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_else,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_case,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [3846] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1254), 17,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(1256), 43,
      anon_sym_export,
      anon_sym_default,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_else,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_case,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [3915] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1258), 17,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(1260), 43,
      anon_sym_export,
      anon_sym_default,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_else,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_case,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [3984] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1258), 17,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(1260), 43,
      anon_sym_export,
      anon_sym_default,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_else,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_case,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [4053] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1262), 17,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(1264), 43,
      anon_sym_export,
      anon_sym_default,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_else,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_case,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [4122] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1266), 17,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(1268), 43,
      anon_sym_export,
      anon_sym_default,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_else,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_case,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [4191] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1270), 17,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(1272), 43,
      anon_sym_export,
      anon_sym_default,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_else,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_case,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [4260] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1274), 17,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(1276), 43,
      anon_sym_export,
      anon_sym_default,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_else,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_case,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [4329] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1206), 17,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(1208), 43,
      anon_sym_export,
      anon_sym_default,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_else,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_case,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [4398] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1278), 17,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(1280), 43,
      anon_sym_export,
      anon_sym_default,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_else,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_case,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [4467] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1282), 17,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(1284), 43,
      anon_sym_export,
      anon_sym_default,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_else,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_case,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [4536] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1206), 17,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(1208), 43,
      anon_sym_export,
      anon_sym_default,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_else,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_case,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [4605] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1286), 17,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(1288), 43,
      anon_sym_export,
      anon_sym_default,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_else,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_case,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [4674] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1290), 17,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(1292), 43,
      anon_sym_export,
      anon_sym_default,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_else,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_case,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [4743] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1294), 17,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(1296), 43,
      anon_sym_export,
      anon_sym_default,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_else,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_case,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [4812] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1298), 17,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(1300), 43,
      anon_sym_export,
      anon_sym_default,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_else,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_case,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [4881] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1298), 17,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(1300), 43,
      anon_sym_export,
      anon_sym_default,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_else,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_case,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [4950] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1302), 17,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(1304), 43,
      anon_sym_export,
      anon_sym_default,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_else,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_case,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [5019] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1306), 17,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(1308), 43,
      anon_sym_export,
      anon_sym_default,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_else,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_case,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [5088] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1310), 17,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(1312), 43,
      anon_sym_export,
      anon_sym_default,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_else,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_case,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [5157] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1314), 17,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(1316), 43,
      anon_sym_export,
      anon_sym_default,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_else,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_case,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [5226] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1318), 17,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(1320), 43,
      anon_sym_export,
      anon_sym_default,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_else,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_case,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [5295] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1322), 17,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(1324), 43,
      anon_sym_export,
      anon_sym_default,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_else,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_case,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [5364] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1326), 17,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(1328), 43,
      anon_sym_export,
      anon_sym_default,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_else,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_case,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [5433] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1330), 17,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(1332), 43,
      anon_sym_export,
      anon_sym_default,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_else,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_case,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [5502] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1334), 17,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(1336), 43,
      anon_sym_export,
      anon_sym_default,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_else,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_case,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [5571] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1338), 17,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(1340), 43,
      anon_sym_export,
      anon_sym_default,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_else,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_case,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [5640] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1342), 17,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(1344), 43,
      anon_sym_export,
      anon_sym_default,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_else,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_case,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [5709] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1346), 17,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(1348), 43,
      anon_sym_export,
      anon_sym_default,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_else,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_case,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [5778] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1350), 17,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(1352), 43,
      anon_sym_export,
      anon_sym_default,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_else,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_case,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [5847] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1354), 17,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(1356), 43,
      anon_sym_export,
      anon_sym_default,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_else,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_case,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [5916] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1358), 17,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(1360), 43,
      anon_sym_export,
      anon_sym_default,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_else,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_case,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [5985] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1362), 17,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(1364), 43,
      anon_sym_export,
      anon_sym_default,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_else,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_case,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [6054] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1366), 17,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(1368), 43,
      anon_sym_export,
      anon_sym_default,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_else,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_case,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [6123] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1370), 17,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(1372), 43,
      anon_sym_export,
      anon_sym_default,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_else,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_case,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [6192] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1374), 17,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(1376), 43,
      anon_sym_export,
      anon_sym_default,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_else,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_case,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [6261] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1206), 17,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(1208), 43,
      anon_sym_export,
      anon_sym_default,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_else,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_case,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [6330] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1378), 17,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(1380), 43,
      anon_sym_export,
      anon_sym_default,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_else,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_case,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [6399] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1382), 17,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(1384), 43,
      anon_sym_export,
      anon_sym_default,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_else,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_case,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [6468] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1386), 17,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(1388), 43,
      anon_sym_export,
      anon_sym_default,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_else,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_case,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [6537] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1390), 17,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(1392), 43,
      anon_sym_export,
      anon_sym_default,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_else,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_case,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [6606] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1394), 17,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(1396), 43,
      anon_sym_export,
      anon_sym_default,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_else,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_case,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [6675] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1398), 17,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(1400), 43,
      anon_sym_export,
      anon_sym_default,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_else,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_case,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [6744] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1402), 17,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(1404), 43,
      anon_sym_export,
      anon_sym_default,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_else,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_case,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [6813] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1206), 17,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(1208), 43,
      anon_sym_export,
      anon_sym_default,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_else,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_case,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [6882] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1406), 17,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(1408), 43,
      anon_sym_export,
      anon_sym_default,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_else,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_case,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [6951] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1410), 17,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(1412), 43,
      anon_sym_export,
      anon_sym_default,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_else,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_case,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [7020] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1414), 17,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(1416), 43,
      anon_sym_export,
      anon_sym_default,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_else,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_case,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [7089] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1418), 17,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(1420), 43,
      anon_sym_export,
      anon_sym_default,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_else,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_case,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [7158] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1206), 17,
      ts_builtin_sym_end,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(1208), 43,
      anon_sym_export,
      anon_sym_default,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_else,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_case,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [7227] = 13,
    ACTIONS(384), 1,
      anon_sym_EQ,
    ACTIONS(402), 1,
      anon_sym_EQ_GT,
    ACTIONS(506), 1,
      anon_sym_in,
    ACTIONS(1102), 1,
      anon_sym_LBRACE,
    ACTIONS(1106), 1,
      anon_sym_LBRACK,
    ACTIONS(1422), 1,
      sym_identifier,
    ACTIONS(1424), 1,
      anon_sym_of,
    STATE(1385), 1,
      sym_variable_declarator,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    STATE(1202), 3,
      sym_object_pattern,
      sym_array_pattern,
      sym__destructuring_pattern,
    ACTIONS(169), 13,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
    ACTIONS(202), 15,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_CARET_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_AMP_AMP_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(165), 20,
      anon_sym_STAR,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_AMP,
      anon_sym_CARET,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
  [7315] = 12,
    ACTIONS(200), 1,
      anon_sym_EQ_GT,
    ACTIONS(490), 1,
      anon_sym_EQ,
    ACTIONS(1100), 1,
      sym_identifier,
    ACTIONS(1102), 1,
      anon_sym_LBRACE,
    ACTIONS(1104), 1,
      anon_sym_of,
    ACTIONS(1106), 1,
      anon_sym_LBRACK,
    STATE(1385), 1,
      sym_variable_declarator,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    STATE(1311), 3,
      sym_object_pattern,
      sym_array_pattern,
      sym__destructuring_pattern,
    ACTIONS(169), 13,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
    ACTIONS(202), 15,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_CARET_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_AMP_AMP_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(165), 21,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_AMP,
      anon_sym_CARET,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
  [7401] = 11,
    ACTIONS(183), 1,
      anon_sym_COLON,
    ACTIONS(351), 1,
      anon_sym_RBRACE,
    ACTIONS(1430), 1,
      anon_sym_LPAREN,
    ACTIONS(1433), 1,
      anon_sym_EQ,
    ACTIONS(1435), 1,
      anon_sym_EQ_GT,
    STATE(1418), 1,
      aux_sym_object_repeat1,
    STATE(1452), 1,
      aux_sym_object_pattern_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1428), 15,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
    ACTIONS(1437), 15,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_CARET_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_AMP_AMP_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1426), 20,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_AMP,
      anon_sym_CARET,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_QMARK_QMARK,
  [7483] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1156), 21,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_AMP,
      anon_sym_CARET,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_QMARK_QMARK,
    ACTIONS(1154), 36,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_CARET_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_AMP_AMP_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [7549] = 11,
    ACTIONS(171), 1,
      anon_sym_RBRACE,
    ACTIONS(183), 1,
      anon_sym_COLON,
    ACTIONS(1430), 1,
      anon_sym_LPAREN,
    ACTIONS(1433), 1,
      anon_sym_EQ,
    ACTIONS(1435), 1,
      anon_sym_EQ_GT,
    STATE(1452), 1,
      aux_sym_object_pattern_repeat1,
    STATE(1487), 1,
      aux_sym_object_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1428), 15,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
    ACTIONS(1437), 15,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_CARET_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_AMP_AMP_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1426), 20,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_AMP,
      anon_sym_CARET,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_QMARK_QMARK,
  [7631] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1439), 21,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_AMP,
      anon_sym_CARET,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_QMARK_QMARK,
    ACTIONS(1441), 36,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_CARET_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_AMP_AMP_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [7697] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1443), 21,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_AMP,
      anon_sym_CARET,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_QMARK_QMARK,
    ACTIONS(1445), 36,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_CARET_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_AMP_AMP_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [7763] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1447), 21,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_AMP,
      anon_sym_CARET,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_QMARK_QMARK,
    ACTIONS(1449), 36,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_CARET_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_AMP_AMP_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [7829] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1451), 21,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_AMP,
      anon_sym_CARET,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_QMARK_QMARK,
    ACTIONS(1453), 36,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_CARET_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_AMP_AMP_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [7895] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1455), 21,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_AMP,
      anon_sym_CARET,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_QMARK_QMARK,
    ACTIONS(1457), 36,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_CARET_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_AMP_AMP_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [7961] = 5,
    ACTIONS(1459), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1437), 15,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_CARET_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_AMP_AMP_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1426), 20,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_AMP,
      anon_sym_CARET,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_QMARK_QMARK,
    ACTIONS(1428), 21,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [8031] = 11,
    ACTIONS(183), 1,
      anon_sym_COLON,
    ACTIONS(216), 1,
      anon_sym_RBRACE,
    ACTIONS(1430), 1,
      anon_sym_LPAREN,
    ACTIONS(1433), 1,
      anon_sym_EQ,
    ACTIONS(1435), 1,
      anon_sym_EQ_GT,
    STATE(1418), 1,
      aux_sym_object_repeat1,
    STATE(1452), 1,
      aux_sym_object_pattern_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1428), 15,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
    ACTIONS(1437), 15,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_CARET_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_AMP_AMP_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1426), 20,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_AMP,
      anon_sym_CARET,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_QMARK_QMARK,
  [8113] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1461), 21,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_AMP,
      anon_sym_CARET,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_QMARK_QMARK,
    ACTIONS(1463), 36,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_CARET_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_AMP_AMP_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [8179] = 6,
    ACTIONS(1465), 1,
      anon_sym_EQ,
    ACTIONS(1467), 1,
      anon_sym_EQ_GT,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1437), 15,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_CARET_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_AMP_AMP_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1428), 19,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
    ACTIONS(1426), 20,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_AMP,
      anon_sym_CARET,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_QMARK_QMARK,
  [8250] = 6,
    ACTIONS(384), 1,
      anon_sym_EQ,
    ACTIONS(402), 1,
      anon_sym_EQ_GT,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(202), 15,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_CARET_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_AMP_AMP_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(169), 19,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
    ACTIONS(165), 20,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_AMP,
      anon_sym_CARET,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_QMARK_QMARK,
  [8321] = 6,
    ACTIONS(1459), 1,
      anon_sym_EQ,
    ACTIONS(1467), 1,
      anon_sym_EQ_GT,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1437), 15,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_CARET_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_AMP_AMP_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1428), 19,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
    ACTIONS(1426), 20,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_AMP,
      anon_sym_CARET,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_QMARK_QMARK,
  [8392] = 6,
    ACTIONS(402), 1,
      anon_sym_EQ_GT,
    ACTIONS(440), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(202), 15,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_CARET_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_AMP_AMP_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(169), 19,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
    ACTIONS(165), 20,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_AMP,
      anon_sym_CARET,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_QMARK_QMARK,
  [8463] = 12,
    ACTIONS(440), 1,
      anon_sym_EQ,
    ACTIONS(498), 1,
      anon_sym_EQ_GT,
    ACTIONS(506), 1,
      anon_sym_in,
    ACTIONS(1102), 1,
      anon_sym_LBRACE,
    ACTIONS(1469), 1,
      sym_identifier,
    ACTIONS(1471), 1,
      anon_sym_of,
    ACTIONS(1473), 1,
      anon_sym_LBRACK,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    STATE(1469), 3,
      sym_object_pattern,
      sym_array_pattern,
      sym__destructuring_pattern,
    ACTIONS(169), 11,
      sym__ternary_qmark,
      anon_sym_LPAREN,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
    ACTIONS(202), 15,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_CARET_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_AMP_AMP_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(165), 20,
      anon_sym_STAR,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_AMP,
      anon_sym_CARET,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
  [8546] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1478), 15,
      anon_sym_LBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(1476), 40,
      anon_sym_export,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [8610] = 5,
    ACTIONS(1465), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1437), 15,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_CARET_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_AMP_AMP_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1428), 19,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
    ACTIONS(1426), 20,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_AMP,
      anon_sym_CARET,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_QMARK_QMARK,
  [8678] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1478), 15,
      anon_sym_LBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(1476), 40,
      anon_sym_export,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [8742] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1482), 15,
      anon_sym_LBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(1480), 40,
      anon_sym_export,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [8806] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1482), 15,
      anon_sym_LBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(1480), 40,
      anon_sym_export,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [8870] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1486), 15,
      anon_sym_LBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(1484), 40,
      anon_sym_export,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [8934] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1490), 15,
      anon_sym_LBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(1488), 40,
      anon_sym_export,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [8998] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1494), 15,
      anon_sym_LBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(1492), 40,
      anon_sym_export,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [9062] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1498), 15,
      anon_sym_LBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(1496), 40,
      anon_sym_export,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [9126] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1502), 15,
      anon_sym_LBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(1500), 40,
      anon_sym_export,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [9190] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1506), 15,
      anon_sym_LBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(1504), 40,
      anon_sym_export,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [9254] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1510), 15,
      anon_sym_LBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(1508), 40,
      anon_sym_export,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [9318] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1478), 15,
      anon_sym_LBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(1476), 40,
      anon_sym_export,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [9382] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1514), 15,
      anon_sym_LBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(1512), 40,
      anon_sym_export,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [9446] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1482), 15,
      anon_sym_LBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(1480), 40,
      anon_sym_export,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [9510] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1478), 15,
      anon_sym_LBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(1476), 40,
      anon_sym_export,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [9574] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1518), 15,
      anon_sym_LBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(1516), 40,
      anon_sym_export,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [9638] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1522), 15,
      anon_sym_LBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(1520), 40,
      anon_sym_export,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [9702] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1154), 15,
      anon_sym_LBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(1156), 40,
      anon_sym_export,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [9766] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1482), 15,
      anon_sym_LBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_LT,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_TILDE,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
    ACTIONS(1480), 40,
      anon_sym_export,
      anon_sym_import,
      anon_sym_with,
      anon_sym_var,
      anon_sym_let,
      anon_sym_const,
      anon_sym_using,
      anon_sym_await,
      anon_sym_if,
      anon_sym_switch,
      anon_sym_for,
      anon_sym_while,
      anon_sym_do,
      anon_sym_try,
      anon_sym_break,
      anon_sym_continue,
      anon_sym_debugger,
      anon_sym_return,
      anon_sym_throw,
      anon_sym_yield,
      anon_sym_class,
      anon_sym_async,
      anon_sym_function,
      anon_sym_new,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_typeof,
      anon_sym_void,
      anon_sym_delete,
      sym_identifier,
      sym_this,
      sym_super,
      sym_true,
      sym_false,
      sym_null,
      sym_undefined,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [9830] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1455), 21,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_AMP,
      anon_sym_CARET,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_QMARK_QMARK,
    ACTIONS(1457), 33,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_CARET_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_AMP_AMP_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [9893] = 6,
    ACTIONS(1524), 1,
      anon_sym_EQ,
    ACTIONS(1526), 1,
      anon_sym_EQ_GT,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1437), 15,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_CARET_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_AMP_AMP_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1428), 17,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
    ACTIONS(1426), 20,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_AMP,
      anon_sym_CARET,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_QMARK_QMARK,
  [9962] = 5,
    ACTIONS(1459), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1437), 15,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_CARET_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_AMP_AMP_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1428), 18,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
    ACTIONS(1426), 20,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_AMP,
      anon_sym_CARET,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_QMARK_QMARK,
  [10029] = 6,
    ACTIONS(1435), 1,
      anon_sym_EQ_GT,
    ACTIONS(1459), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1437), 15,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_CARET_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_AMP_AMP_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1428), 17,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
    ACTIONS(1426), 20,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_AMP,
      anon_sym_CARET,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_QMARK_QMARK,
  [10098] = 6,
    ACTIONS(200), 1,
      anon_sym_EQ_GT,
    ACTIONS(440), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(202), 15,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_CARET_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_AMP_AMP_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(169), 17,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
    ACTIONS(165), 20,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_AMP,
      anon_sym_CARET,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_QMARK_QMARK,
  [10167] = 7,
    ACTIONS(494), 1,
      anon_sym_COLON,
    ACTIONS(1435), 1,
      anon_sym_EQ_GT,
    ACTIONS(1528), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1437), 15,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_CARET_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_AMP_AMP_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1428), 16,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
    ACTIONS(1426), 20,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_AMP,
      anon_sym_CARET,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_QMARK_QMARK,
  [10238] = 7,
    ACTIONS(498), 1,
      anon_sym_EQ_GT,
    ACTIONS(511), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(509), 4,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
    ACTIONS(169), 13,
      sym__ternary_qmark,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
    ACTIONS(202), 15,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_CARET_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_AMP_AMP_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(165), 20,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_AMP,
      anon_sym_CARET,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_QMARK_QMARK,
  [10309] = 7,
    ACTIONS(1532), 1,
      anon_sym_EQ,
    ACTIONS(1535), 1,
      anon_sym_EQ_GT,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1530), 4,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
    ACTIONS(1428), 13,
      sym__ternary_qmark,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
    ACTIONS(1437), 15,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_CARET_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_AMP_AMP_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1426), 20,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_AMP,
      anon_sym_CARET,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_QMARK_QMARK,
  [10380] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1461), 21,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_AMP,
      anon_sym_CARET,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_QMARK_QMARK,
    ACTIONS(1463), 33,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_CARET_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_AMP_AMP_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [10443] = 7,
    ACTIONS(200), 1,
      anon_sym_EQ_GT,
    ACTIONS(490), 1,
      anon_sym_EQ,
    ACTIONS(494), 1,
      anon_sym_COLON,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(202), 15,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_CARET_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_AMP_AMP_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(169), 16,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
    ACTIONS(165), 20,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_AMP,
      anon_sym_CARET,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_QMARK_QMARK,
  [10514] = 7,
    ACTIONS(1467), 1,
      anon_sym_EQ_GT,
    ACTIONS(1540), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1537), 4,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
    ACTIONS(1428), 13,
      sym__ternary_qmark,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
    ACTIONS(1437), 15,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_CARET_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_AMP_AMP_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1426), 20,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_AMP,
      anon_sym_CARET,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_QMARK_QMARK,
  [10585] = 6,
    ACTIONS(200), 1,
      anon_sym_EQ_GT,
    ACTIONS(490), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(202), 15,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_CARET_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_AMP_AMP_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(169), 17,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
    ACTIONS(165), 20,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_AMP,
      anon_sym_CARET,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_QMARK_QMARK,
  [10654] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1156), 21,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_AMP,
      anon_sym_CARET,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_QMARK_QMARK,
    ACTIONS(1154), 33,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_CARET_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_AMP_AMP_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [10717] = 7,
    ACTIONS(440), 1,
      anon_sym_EQ,
    ACTIONS(498), 1,
      anon_sym_EQ_GT,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(496), 4,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
    ACTIONS(169), 13,
      sym__ternary_qmark,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
    ACTIONS(202), 15,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_CARET_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_AMP_AMP_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(165), 20,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_AMP,
      anon_sym_CARET,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_QMARK_QMARK,
  [10788] = 7,
    ACTIONS(1459), 1,
      anon_sym_EQ,
    ACTIONS(1535), 1,
      anon_sym_EQ_GT,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1543), 4,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
    ACTIONS(1428), 13,
      sym__ternary_qmark,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
    ACTIONS(1437), 15,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_CARET_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_AMP_AMP_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1426), 20,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_AMP,
      anon_sym_CARET,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_QMARK_QMARK,
  [10859] = 7,
    ACTIONS(488), 1,
      anon_sym_COLON,
    ACTIONS(1435), 1,
      anon_sym_EQ_GT,
    ACTIONS(1528), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1437), 15,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_CARET_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_AMP_AMP_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1428), 16,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
    ACTIONS(1426), 20,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_AMP,
      anon_sym_CARET,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_QMARK_QMARK,
  [10930] = 7,
    ACTIONS(200), 1,
      anon_sym_EQ_GT,
    ACTIONS(488), 1,
      anon_sym_COLON,
    ACTIONS(490), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(202), 15,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_CARET_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_AMP_AMP_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(169), 16,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
    ACTIONS(165), 20,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_AMP,
      anon_sym_CARET,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_QMARK_QMARK,
  [11001] = 6,
    ACTIONS(1435), 1,
      anon_sym_EQ_GT,
    ACTIONS(1528), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1437), 15,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_CARET_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_AMP_AMP_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1428), 17,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
    ACTIONS(1426), 20,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_AMP,
      anon_sym_CARET,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_QMARK_QMARK,
  [11070] = 6,
    ACTIONS(500), 1,
      anon_sym_EQ_GT,
    ACTIONS(528), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(202), 15,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_CARET_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_AMP_AMP_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(169), 17,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
    ACTIONS(165), 20,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_AMP,
      anon_sym_CARET,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_QMARK_QMARK,
  [11139] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1443), 21,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_AMP,
      anon_sym_CARET,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_QMARK_QMARK,
    ACTIONS(1445), 33,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_CARET_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_AMP_AMP_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [11202] = 6,
    ACTIONS(440), 1,
      anon_sym_EQ,
    ACTIONS(500), 1,
      anon_sym_EQ_GT,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(202), 15,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_CARET_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_AMP_AMP_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(169), 17,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
    ACTIONS(165), 20,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_AMP,
      anon_sym_CARET,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_QMARK_QMARK,
  [11271] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1447), 21,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_AMP,
      anon_sym_CARET,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_QMARK_QMARK,
    ACTIONS(1449), 33,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_CARET_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_AMP_AMP_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [11334] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1451), 21,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_AMP,
      anon_sym_CARET,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_QMARK_QMARK,
    ACTIONS(1453), 33,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_CARET_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_AMP_AMP_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [11397] = 7,
    ACTIONS(402), 1,
      anon_sym_EQ_GT,
    ACTIONS(517), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(514), 4,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
    ACTIONS(169), 13,
      sym__ternary_qmark,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
    ACTIONS(202), 15,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_CARET_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_AMP_AMP_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(165), 20,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_AMP,
      anon_sym_CARET,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_QMARK_QMARK,
  [11468] = 6,
    ACTIONS(1459), 1,
      anon_sym_EQ,
    ACTIONS(1526), 1,
      anon_sym_EQ_GT,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1437), 15,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_CARET_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_AMP_AMP_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1428), 17,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
    ACTIONS(1426), 20,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_AMP,
      anon_sym_CARET,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_QMARK_QMARK,
  [11537] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1439), 21,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_AMP,
      anon_sym_CARET,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_QMARK_QMARK,
    ACTIONS(1441), 33,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_CARET_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_AMP_AMP_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [11600] = 5,
    ACTIONS(1528), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1437), 15,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_CARET_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_AMP_AMP_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1428), 17,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
    ACTIONS(1426), 20,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_AMP,
      anon_sym_CARET,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_QMARK_QMARK,
  [11666] = 8,
    ACTIONS(384), 1,
      anon_sym_EQ,
    ACTIONS(402), 1,
      anon_sym_EQ_GT,
    ACTIONS(506), 1,
      anon_sym_in,
    ACTIONS(1545), 1,
      anon_sym_of,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(169), 15,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
    ACTIONS(202), 15,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_CARET_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_AMP_AMP_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(165), 19,
      anon_sym_STAR,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_AMP,
      anon_sym_CARET,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_QMARK_QMARK,
  [11738] = 7,
    ACTIONS(1465), 1,
      anon_sym_EQ,
    ACTIONS(1467), 1,
      anon_sym_EQ_GT,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1547), 3,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
    ACTIONS(1428), 13,
      sym__ternary_qmark,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
    ACTIONS(1437), 15,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_CARET_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_AMP_AMP_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1426), 20,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_AMP,
      anon_sym_CARET,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_QMARK_QMARK,
  [11808] = 8,
    ACTIONS(1465), 1,
      anon_sym_EQ,
    ACTIONS(1467), 1,
      anon_sym_EQ_GT,
    ACTIONS(1550), 1,
      anon_sym_of,
    ACTIONS(1552), 1,
      anon_sym_in,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1428), 15,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
    ACTIONS(1437), 15,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_CARET_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_AMP_AMP_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1426), 19,
      anon_sym_STAR,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_AMP,
      anon_sym_CARET,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_QMARK_QMARK,
  [11880] = 6,
    ACTIONS(1459), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1543), 4,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
    ACTIONS(1428), 13,
      sym__ternary_qmark,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
    ACTIONS(1437), 15,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_CARET_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_AMP_AMP_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1426), 20,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_AMP,
      anon_sym_CARET,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_QMARK_QMARK,
  [11948] = 6,
    ACTIONS(1540), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1537), 4,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
    ACTIONS(1428), 13,
      sym__ternary_qmark,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
    ACTIONS(1437), 15,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_CARET_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_AMP_AMP_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1426), 20,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_AMP,
      anon_sym_CARET,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_QMARK_QMARK,
  [12016] = 6,
    ACTIONS(1532), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1530), 4,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
    ACTIONS(1428), 13,
      sym__ternary_qmark,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
    ACTIONS(1437), 15,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_CARET_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_AMP_AMP_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1426), 20,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_AMP,
      anon_sym_CARET,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_QMARK_QMARK,
  [12084] = 7,
    ACTIONS(384), 1,
      anon_sym_EQ,
    ACTIONS(402), 1,
      anon_sym_EQ_GT,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(577), 3,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
    ACTIONS(169), 13,
      sym__ternary_qmark,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
    ACTIONS(202), 15,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_CARET_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_AMP_AMP_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(165), 20,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_AMP,
      anon_sym_CARET,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_QMARK_QMARK,
  [12154] = 5,
    ACTIONS(1524), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1437), 15,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_CARET_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_AMP_AMP_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1428), 17,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
    ACTIONS(1426), 20,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_AMP,
      anon_sym_CARET,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_QMARK_QMARK,
  [12220] = 8,
    ACTIONS(402), 1,
      anon_sym_EQ_GT,
    ACTIONS(509), 1,
      anon_sym_COMMA,
    ACTIONS(514), 1,
      anon_sym_RBRACK,
    ACTIONS(517), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(169), 13,
      sym__ternary_qmark,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
    ACTIONS(202), 15,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_CARET_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_AMP_AMP_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(165), 20,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_AMP,
      anon_sym_CARET,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_QMARK_QMARK,
  [12291] = 6,
    ACTIONS(1465), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1547), 3,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
    ACTIONS(1428), 13,
      sym__ternary_qmark,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
    ACTIONS(1437), 15,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_CARET_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_AMP_AMP_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1426), 20,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_AMP,
      anon_sym_CARET,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_QMARK_QMARK,
  [12358] = 8,
    ACTIONS(1467), 1,
      anon_sym_EQ_GT,
    ACTIONS(1530), 1,
      anon_sym_COMMA,
    ACTIONS(1537), 1,
      anon_sym_RBRACK,
    ACTIONS(1540), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1428), 13,
      sym__ternary_qmark,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
    ACTIONS(1437), 15,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_CARET_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_AMP_AMP_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1426), 20,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_AMP,
      anon_sym_CARET,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_QMARK_QMARK,
  [12429] = 7,
    ACTIONS(1465), 1,
      anon_sym_EQ,
    ACTIONS(1550), 1,
      anon_sym_of,
    ACTIONS(1552), 1,
      anon_sym_in,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1428), 15,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
    ACTIONS(1437), 15,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_CARET_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_AMP_AMP_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1426), 19,
      anon_sym_STAR,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_AMP,
      anon_sym_CARET,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_QMARK_QMARK,
  [12498] = 6,
    ACTIONS(440), 1,
      anon_sym_EQ,
    ACTIONS(628), 1,
      anon_sym_EQ_GT,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(169), 14,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
    ACTIONS(202), 15,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_CARET_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_AMP_AMP_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(165), 20,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_AMP,
      anon_sym_CARET,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_QMARK_QMARK,
  [12564] = 6,
    ACTIONS(1459), 1,
      anon_sym_EQ,
    ACTIONS(1555), 1,
      anon_sym_EQ_GT,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1428), 14,
      sym__ternary_qmark,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
    ACTIONS(1437), 15,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_CARET_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_AMP_AMP_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1426), 20,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_AMP,
      anon_sym_CARET,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_QMARK_QMARK,
  [12630] = 6,
    ACTIONS(624), 1,
      anon_sym_EQ,
    ACTIONS(628), 1,
      anon_sym_EQ_GT,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(169), 14,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
    ACTIONS(202), 15,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_CARET_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_AMP_AMP_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(165), 20,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_AMP,
      anon_sym_CARET,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_QMARK_QMARK,
  [12696] = 6,
    ACTIONS(1557), 1,
      anon_sym_EQ,
    ACTIONS(1559), 1,
      anon_sym_EQ_GT,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1428), 14,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
    ACTIONS(1437), 15,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_CARET_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_AMP_AMP_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1426), 20,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_AMP,
      anon_sym_CARET,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_QMARK_QMARK,
  [12762] = 6,
    ACTIONS(580), 1,
      anon_sym_EQ_GT,
    ACTIONS(590), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(169), 14,
      sym__ternary_qmark,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
    ACTIONS(202), 15,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_CARET_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_AMP_AMP_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(165), 20,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_AMP,
      anon_sym_CARET,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_QMARK_QMARK,
  [12828] = 6,
    ACTIONS(1459), 1,
      anon_sym_EQ,
    ACTIONS(1559), 1,
      anon_sym_EQ_GT,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1428), 14,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
    ACTIONS(1437), 15,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_CARET_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_AMP_AMP_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1426), 20,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_AMP,
      anon_sym_CARET,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_QMARK_QMARK,
  [12894] = 6,
    ACTIONS(440), 1,
      anon_sym_EQ,
    ACTIONS(498), 1,
      anon_sym_EQ_GT,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(169), 14,
      sym__ternary_qmark,
      anon_sym_LPAREN,
      anon_sym_COLON,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
    ACTIONS(202), 15,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_CARET_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_AMP_AMP_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(165), 20,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_AMP,
      anon_sym_CARET,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_QMARK_QMARK,
  [12960] = 7,
    ACTIONS(1530), 1,
      anon_sym_COMMA,
    ACTIONS(1537), 1,
      anon_sym_RBRACK,
    ACTIONS(1540), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1428), 13,
      sym__ternary_qmark,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
    ACTIONS(1437), 15,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_CARET_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_AMP_AMP_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1426), 20,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_AMP,
      anon_sym_CARET,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_QMARK_QMARK,
  [13028] = 6,
    ACTIONS(440), 1,
      anon_sym_EQ,
    ACTIONS(580), 1,
      anon_sym_EQ_GT,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(169), 14,
      sym__ternary_qmark,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
    ACTIONS(202), 15,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_CARET_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_AMP_AMP_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(165), 20,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_AMP,
      anon_sym_CARET,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_QMARK_QMARK,
  [13094] = 8,
    ACTIONS(440), 1,
      anon_sym_EQ,
    ACTIONS(498), 1,
      anon_sym_EQ_GT,
    ACTIONS(506), 1,
      anon_sym_in,
    ACTIONS(1545), 1,
      anon_sym_of,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(169), 13,
      sym__ternary_qmark,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
    ACTIONS(202), 15,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_CARET_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_AMP_AMP_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(165), 19,
      anon_sym_STAR,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_AMP,
      anon_sym_CARET,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_QMARK_QMARK,
  [13164] = 8,
    ACTIONS(1459), 1,
      anon_sym_EQ,
    ACTIONS(1535), 1,
      anon_sym_EQ_GT,
    ACTIONS(1550), 1,
      anon_sym_of,
    ACTIONS(1552), 1,
      anon_sym_in,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1428), 13,
      sym__ternary_qmark,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
    ACTIONS(1437), 15,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_CARET_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_AMP_AMP_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1426), 19,
      anon_sym_STAR,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_AMP,
      anon_sym_CARET,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_QMARK_QMARK,
  [13234] = 6,
    ACTIONS(1459), 1,
      anon_sym_EQ,
    ACTIONS(1535), 1,
      anon_sym_EQ_GT,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1428), 14,
      sym__ternary_qmark,
      anon_sym_LPAREN,
      anon_sym_COLON,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
    ACTIONS(1437), 15,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_CARET_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_AMP_AMP_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1426), 20,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_AMP,
      anon_sym_CARET,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_QMARK_QMARK,
  [13300] = 6,
    ACTIONS(1555), 1,
      anon_sym_EQ_GT,
    ACTIONS(1561), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1428), 14,
      sym__ternary_qmark,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
    ACTIONS(1437), 15,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_CARET_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_AMP_AMP_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1426), 20,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_AMP,
      anon_sym_CARET,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_QMARK_QMARK,
  [13366] = 5,
    ACTIONS(1557), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1428), 14,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
    ACTIONS(1437), 15,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_CARET_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_AMP_AMP_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1426), 20,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_AMP,
      anon_sym_CARET,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_QMARK_QMARK,
  [13429] = 5,
    ACTIONS(1561), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1428), 14,
      sym__ternary_qmark,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
    ACTIONS(1437), 15,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_CARET_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_AMP_AMP_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1426), 20,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_AMP,
      anon_sym_CARET,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_QMARK_QMARK,
  [13492] = 7,
    ACTIONS(1459), 1,
      anon_sym_EQ,
    ACTIONS(1550), 1,
      anon_sym_of,
    ACTIONS(1552), 1,
      anon_sym_in,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1428), 13,
      sym__ternary_qmark,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
    ACTIONS(1437), 15,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_CARET_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_AMP_AMP_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
    ACTIONS(1426), 19,
      anon_sym_STAR,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_AMP,
      anon_sym_CARET,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_QMARK_QMARK,
  [13559] = 10,
    ACTIONS(662), 1,
      anon_sym_BQUOTE,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1573), 1,
      sym_optional_chain,
    STATE(597), 1,
      sym_arguments,
    STATE(607), 1,
      sym_template_string,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1563), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1565), 24,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_RBRACK,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
  [13625] = 5,
    ACTIONS(718), 1,
      anon_sym_EQ,
    ACTIONS(1575), 1,
      sym__automatic_semicolon,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(710), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(708), 29,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [13681] = 7,
    ACTIONS(662), 1,
      anon_sym_BQUOTE,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    STATE(597), 1,
      sym_arguments,
    STATE(607), 1,
      sym_template_string,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1563), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1565), 27,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
  [13741] = 4,
    ACTIONS(1575), 1,
      sym__automatic_semicolon,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(710), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(708), 29,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [13794] = 5,
    ACTIONS(662), 1,
      anon_sym_BQUOTE,
    STATE(607), 1,
      sym_template_string,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1577), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1579), 28,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
  [13849] = 4,
    ACTIONS(1581), 1,
      sym__automatic_semicolon,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(696), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(694), 29,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [13902] = 9,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1583), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1585), 23,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_RBRACK,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_BQUOTE,
  [13965] = 9,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1591), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1593), 23,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_RBRACK,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_BQUOTE,
  [14028] = 8,
    ACTIONS(662), 1,
      anon_sym_BQUOTE,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1573), 1,
      sym_optional_chain,
    STATE(607), 1,
      sym_template_string,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1577), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1579), 25,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_RBRACK,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
  [14089] = 4,
    ACTIONS(1459), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1426), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1428), 29,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [14142] = 8,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1595), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1597), 25,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_RBRACK,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [14203] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1599), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1601), 29,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [14253] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1603), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1605), 29,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [14303] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(706), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(704), 29,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [14353] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1607), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1609), 29,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [14403] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1611), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1613), 29,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [14453] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1615), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1617), 29,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [14503] = 4,
    ACTIONS(1623), 1,
      sym_regex_flags,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1619), 13,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_instanceof,
    ACTIONS(1621), 27,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [14555] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1625), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1627), 29,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [14605] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1629), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1631), 29,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [14655] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1633), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1635), 29,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [14705] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1577), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1579), 29,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [14755] = 5,
    ACTIONS(1644), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1641), 4,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
    ACTIONS(1637), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1639), 24,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_COLON,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [14809] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1646), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1648), 29,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [14859] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(746), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(748), 29,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [14909] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1650), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1652), 29,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [14959] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1654), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1656), 29,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [15009] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1658), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1660), 29,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [15059] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1662), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1664), 29,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [15109] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1666), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1668), 29,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [15159] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1670), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1672), 29,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [15209] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(726), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(728), 29,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [15259] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1674), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1676), 29,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [15309] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1678), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1680), 29,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [15359] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1682), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1684), 29,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [15409] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1686), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1688), 29,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [15459] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1690), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1692), 29,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [15509] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(702), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(700), 29,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [15559] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(766), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(768), 29,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [15609] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(736), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(738), 29,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [15659] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1694), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1696), 29,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [15709] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1694), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1696), 29,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [15759] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1694), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1696), 29,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [15809] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1694), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1696), 29,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [15859] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(692), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(690), 29,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [15909] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1698), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1700), 29,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [15959] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1702), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1704), 29,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [16009] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1706), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1708), 29,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [16059] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1710), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1712), 29,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [16109] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1714), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1716), 29,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [16159] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1718), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1720), 29,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [16209] = 5,
    ACTIONS(1729), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1726), 4,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
    ACTIONS(1722), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1724), 24,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_COLON,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [16263] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1731), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1733), 29,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [16313] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1735), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1737), 29,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [16363] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1739), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1741), 29,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [16413] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(756), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(758), 29,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [16463] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1654), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1656), 29,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [16513] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1743), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1745), 29,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [16563] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1747), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1749), 29,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [16613] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1751), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1753), 29,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [16663] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1755), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1757), 29,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [16713] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1759), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1761), 29,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [16763] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1763), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1765), 29,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [16813] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(776), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(778), 29,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [16863] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(786), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(788), 29,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [16913] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1426), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1428), 29,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [16963] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(696), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(694), 29,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [17013] = 5,
    ACTIONS(718), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1767), 4,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
    ACTIONS(714), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(716), 24,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_COLON,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [17067] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1770), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1772), 29,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [17117] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(796), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(798), 29,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [17167] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1774), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1776), 29,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [17217] = 4,
    ACTIONS(1729), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1722), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1724), 27,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [17268] = 25,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(1784), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1786), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1788), 1,
      anon_sym_GT_GT,
    ACTIONS(1792), 1,
      anon_sym_AMP,
    ACTIONS(1794), 1,
      anon_sym_CARET,
    ACTIONS(1796), 1,
      anon_sym_PIPE,
    ACTIONS(1800), 1,
      anon_sym_PERCENT,
    ACTIONS(1802), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1810), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1812), 1,
      sym__ternary_qmark,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1778), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1790), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1798), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1806), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1808), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1782), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1804), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1780), 7,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_RBRACK,
      anon_sym_BQUOTE,
  [17361] = 25,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(1784), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1786), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1788), 1,
      anon_sym_GT_GT,
    ACTIONS(1792), 1,
      anon_sym_AMP,
    ACTIONS(1794), 1,
      anon_sym_CARET,
    ACTIONS(1796), 1,
      anon_sym_PIPE,
    ACTIONS(1800), 1,
      anon_sym_PERCENT,
    ACTIONS(1802), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1810), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1812), 1,
      sym__ternary_qmark,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1778), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1790), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1798), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1806), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1808), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1782), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1804), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1676), 7,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_RBRACK,
      anon_sym_BQUOTE,
  [17454] = 25,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(1784), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1786), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1788), 1,
      anon_sym_GT_GT,
    ACTIONS(1792), 1,
      anon_sym_AMP,
    ACTIONS(1794), 1,
      anon_sym_CARET,
    ACTIONS(1796), 1,
      anon_sym_PIPE,
    ACTIONS(1800), 1,
      anon_sym_PERCENT,
    ACTIONS(1802), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1810), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1812), 1,
      sym__ternary_qmark,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1778), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1790), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1798), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1806), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1808), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1782), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1804), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1814), 7,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_RBRACK,
      anon_sym_BQUOTE,
  [17547] = 25,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(1784), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1786), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1788), 1,
      anon_sym_GT_GT,
    ACTIONS(1792), 1,
      anon_sym_AMP,
    ACTIONS(1794), 1,
      anon_sym_CARET,
    ACTIONS(1796), 1,
      anon_sym_PIPE,
    ACTIONS(1800), 1,
      anon_sym_PERCENT,
    ACTIONS(1802), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1810), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1812), 1,
      sym__ternary_qmark,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1778), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1790), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1798), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1806), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1808), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1782), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1804), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1816), 7,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_RBRACK,
      anon_sym_BQUOTE,
  [17640] = 10,
    ACTIONS(83), 1,
      anon_sym_BQUOTE,
    ACTIONS(1818), 1,
      anon_sym_LPAREN,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_DOT,
    ACTIONS(1824), 1,
      sym_optional_chain,
    STATE(700), 1,
      sym_arguments,
    STATE(758), 1,
      sym_template_string,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1563), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1565), 21,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_SEMI,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
  [17703] = 25,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(1784), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1786), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1788), 1,
      anon_sym_GT_GT,
    ACTIONS(1792), 1,
      anon_sym_AMP,
    ACTIONS(1794), 1,
      anon_sym_CARET,
    ACTIONS(1796), 1,
      anon_sym_PIPE,
    ACTIONS(1800), 1,
      anon_sym_PERCENT,
    ACTIONS(1802), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1810), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1812), 1,
      sym__ternary_qmark,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1778), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1790), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1798), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1806), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1808), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1782), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1804), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1716), 7,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_RBRACK,
      anon_sym_BQUOTE,
  [17796] = 4,
    ACTIONS(1465), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1426), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1428), 27,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [17847] = 25,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(1784), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1786), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1788), 1,
      anon_sym_GT_GT,
    ACTIONS(1792), 1,
      anon_sym_AMP,
    ACTIONS(1794), 1,
      anon_sym_CARET,
    ACTIONS(1796), 1,
      anon_sym_PIPE,
    ACTIONS(1800), 1,
      anon_sym_PERCENT,
    ACTIONS(1802), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1810), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1812), 1,
      sym__ternary_qmark,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1778), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1790), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1798), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1806), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1808), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1782), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1804), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1826), 7,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_RBRACK,
      anon_sym_BQUOTE,
  [17940] = 25,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(1784), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1786), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1788), 1,
      anon_sym_GT_GT,
    ACTIONS(1792), 1,
      anon_sym_AMP,
    ACTIONS(1794), 1,
      anon_sym_CARET,
    ACTIONS(1796), 1,
      anon_sym_PIPE,
    ACTIONS(1800), 1,
      anon_sym_PERCENT,
    ACTIONS(1802), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1810), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1812), 1,
      sym__ternary_qmark,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1778), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1790), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1798), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1806), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1808), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1782), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1804), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1631), 7,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_RBRACK,
      anon_sym_BQUOTE,
  [18033] = 15,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(1788), 1,
      anon_sym_GT_GT,
    ACTIONS(1800), 1,
      anon_sym_PERCENT,
    ACTIONS(1802), 1,
      anon_sym_STAR_STAR,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1778), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1790), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1798), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1830), 7,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1828), 17,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_RBRACK,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_CARET,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_BQUOTE,
  [18106] = 25,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(1784), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1786), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1788), 1,
      anon_sym_GT_GT,
    ACTIONS(1792), 1,
      anon_sym_AMP,
    ACTIONS(1794), 1,
      anon_sym_CARET,
    ACTIONS(1796), 1,
      anon_sym_PIPE,
    ACTIONS(1800), 1,
      anon_sym_PERCENT,
    ACTIONS(1802), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1810), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1812), 1,
      sym__ternary_qmark,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1778), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1790), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1798), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1806), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1808), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1782), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1804), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1832), 7,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_RBRACK,
      anon_sym_BQUOTE,
  [18199] = 25,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(1784), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1786), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1788), 1,
      anon_sym_GT_GT,
    ACTIONS(1792), 1,
      anon_sym_AMP,
    ACTIONS(1794), 1,
      anon_sym_CARET,
    ACTIONS(1796), 1,
      anon_sym_PIPE,
    ACTIONS(1800), 1,
      anon_sym_PERCENT,
    ACTIONS(1802), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1810), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1812), 1,
      sym__ternary_qmark,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1778), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1790), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1798), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1806), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1808), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1782), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1804), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1700), 7,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_RBRACK,
      anon_sym_BQUOTE,
  [18292] = 4,
    ACTIONS(718), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(714), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(716), 27,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [18343] = 25,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(1784), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1786), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1788), 1,
      anon_sym_GT_GT,
    ACTIONS(1792), 1,
      anon_sym_AMP,
    ACTIONS(1794), 1,
      anon_sym_CARET,
    ACTIONS(1796), 1,
      anon_sym_PIPE,
    ACTIONS(1800), 1,
      anon_sym_PERCENT,
    ACTIONS(1802), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1810), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1812), 1,
      sym__ternary_qmark,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1778), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1790), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1798), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1806), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1808), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1782), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1804), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1708), 7,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_RBRACK,
      anon_sym_BQUOTE,
  [18436] = 25,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(1784), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1786), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1788), 1,
      anon_sym_GT_GT,
    ACTIONS(1792), 1,
      anon_sym_AMP,
    ACTIONS(1794), 1,
      anon_sym_CARET,
    ACTIONS(1796), 1,
      anon_sym_PIPE,
    ACTIONS(1800), 1,
      anon_sym_PERCENT,
    ACTIONS(1802), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1810), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1812), 1,
      sym__ternary_qmark,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1778), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1790), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1798), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1806), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1808), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1782), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1804), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1834), 7,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_RBRACK,
      anon_sym_BQUOTE,
  [18529] = 25,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(1784), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1786), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1788), 1,
      anon_sym_GT_GT,
    ACTIONS(1792), 1,
      anon_sym_AMP,
    ACTIONS(1794), 1,
      anon_sym_CARET,
    ACTIONS(1796), 1,
      anon_sym_PIPE,
    ACTIONS(1800), 1,
      anon_sym_PERCENT,
    ACTIONS(1802), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1810), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1812), 1,
      sym__ternary_qmark,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1778), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1790), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1798), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1806), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1808), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1782), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1804), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1712), 7,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_RBRACK,
      anon_sym_BQUOTE,
  [18622] = 10,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(1802), 1,
      anon_sym_STAR_STAR,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1830), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1828), 20,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_RBRACK,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_BQUOTE,
  [18685] = 4,
    ACTIONS(1644), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1637), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1639), 27,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [18736] = 21,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(1788), 1,
      anon_sym_GT_GT,
    ACTIONS(1792), 1,
      anon_sym_AMP,
    ACTIONS(1794), 1,
      anon_sym_CARET,
    ACTIONS(1796), 1,
      anon_sym_PIPE,
    ACTIONS(1800), 1,
      anon_sym_PERCENT,
    ACTIONS(1802), 1,
      anon_sym_STAR_STAR,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1778), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1790), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1798), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1806), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1808), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1782), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1804), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1828), 11,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_RBRACK,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_QMARK,
      anon_sym_BQUOTE,
  [18821] = 22,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(1784), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1788), 1,
      anon_sym_GT_GT,
    ACTIONS(1792), 1,
      anon_sym_AMP,
    ACTIONS(1794), 1,
      anon_sym_CARET,
    ACTIONS(1796), 1,
      anon_sym_PIPE,
    ACTIONS(1800), 1,
      anon_sym_PERCENT,
    ACTIONS(1802), 1,
      anon_sym_STAR_STAR,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1778), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1790), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1798), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1806), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1808), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1782), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1804), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1828), 10,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_RBRACK,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_QMARK,
      anon_sym_BQUOTE,
  [18908] = 13,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(1800), 1,
      anon_sym_PERCENT,
    ACTIONS(1802), 1,
      anon_sym_STAR_STAR,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1778), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1798), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1830), 8,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1828), 19,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_RBRACK,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_BQUOTE,
  [18977] = 19,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(1788), 1,
      anon_sym_GT_GT,
    ACTIONS(1800), 1,
      anon_sym_PERCENT,
    ACTIONS(1802), 1,
      anon_sym_STAR_STAR,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1778), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1790), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1798), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1806), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1808), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1830), 2,
      anon_sym_AMP,
      anon_sym_PIPE,
    ACTIONS(1782), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1804), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1828), 12,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_RBRACK,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_CARET,
      anon_sym_QMARK_QMARK,
      anon_sym_BQUOTE,
  [19058] = 20,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(1788), 1,
      anon_sym_GT_GT,
    ACTIONS(1792), 1,
      anon_sym_AMP,
    ACTIONS(1800), 1,
      anon_sym_PERCENT,
    ACTIONS(1802), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1830), 1,
      anon_sym_PIPE,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1778), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1790), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1798), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1806), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1808), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1782), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1804), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1828), 12,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_RBRACK,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_CARET,
      anon_sym_QMARK_QMARK,
      anon_sym_BQUOTE,
  [19141] = 21,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(1788), 1,
      anon_sym_GT_GT,
    ACTIONS(1792), 1,
      anon_sym_AMP,
    ACTIONS(1794), 1,
      anon_sym_CARET,
    ACTIONS(1800), 1,
      anon_sym_PERCENT,
    ACTIONS(1802), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1830), 1,
      anon_sym_PIPE,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1778), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1790), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1798), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1806), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1808), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1782), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1804), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1828), 11,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_RBRACK,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_QMARK,
      anon_sym_BQUOTE,
  [19226] = 12,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(1800), 1,
      anon_sym_PERCENT,
    ACTIONS(1802), 1,
      anon_sym_STAR_STAR,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1778), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1830), 10,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1828), 19,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_RBRACK,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_BQUOTE,
  [19293] = 10,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(1802), 1,
      anon_sym_STAR_STAR,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1830), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1828), 20,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_RBRACK,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_BQUOTE,
  [19356] = 17,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(1788), 1,
      anon_sym_GT_GT,
    ACTIONS(1800), 1,
      anon_sym_PERCENT,
    ACTIONS(1802), 1,
      anon_sym_STAR_STAR,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1778), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1790), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1798), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1782), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1804), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1830), 4,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1828), 14,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_RBRACK,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_CARET,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_BQUOTE,
  [19433] = 23,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(1784), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1786), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1788), 1,
      anon_sym_GT_GT,
    ACTIONS(1792), 1,
      anon_sym_AMP,
    ACTIONS(1794), 1,
      anon_sym_CARET,
    ACTIONS(1796), 1,
      anon_sym_PIPE,
    ACTIONS(1800), 1,
      anon_sym_PERCENT,
    ACTIONS(1802), 1,
      anon_sym_STAR_STAR,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1778), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1790), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1798), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1806), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1808), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1782), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1804), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1828), 9,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_RBRACK,
      anon_sym_QMARK_QMARK,
      anon_sym_BQUOTE,
  [19522] = 7,
    ACTIONS(83), 1,
      anon_sym_BQUOTE,
    ACTIONS(1818), 1,
      anon_sym_LPAREN,
    STATE(700), 1,
      sym_arguments,
    STATE(758), 1,
      sym_template_string,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1563), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1565), 24,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
  [19579] = 5,
    ACTIONS(83), 1,
      anon_sym_BQUOTE,
    STATE(758), 1,
      sym_template_string,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1577), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1579), 25,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
  [19631] = 4,
    ACTIONS(1644), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1637), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1639), 26,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [19681] = 5,
    ACTIONS(712), 1,
      sym__automatic_semicolon,
    ACTIONS(718), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(710), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(708), 25,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [19733] = 4,
    ACTIONS(1729), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1722), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1724), 26,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [19783] = 8,
    ACTIONS(83), 1,
      anon_sym_BQUOTE,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_DOT,
    ACTIONS(1824), 1,
      sym_optional_chain,
    STATE(758), 1,
      sym_template_string,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1577), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1579), 22,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
  [19841] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1678), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1680), 27,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [19889] = 4,
    ACTIONS(1459), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1426), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1428), 26,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [19939] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1751), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1753), 27,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_COLON,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [19987] = 25,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(1784), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1786), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1788), 1,
      anon_sym_GT_GT,
    ACTIONS(1792), 1,
      anon_sym_AMP,
    ACTIONS(1794), 1,
      anon_sym_CARET,
    ACTIONS(1796), 1,
      anon_sym_PIPE,
    ACTIONS(1800), 1,
      anon_sym_PERCENT,
    ACTIONS(1802), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1810), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1812), 1,
      sym__ternary_qmark,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1778), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1790), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1798), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1806), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1808), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1782), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1804), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1836), 6,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_RBRACK,
  [20079] = 8,
    ACTIONS(1818), 1,
      anon_sym_LPAREN,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_DOT,
    ACTIONS(1838), 1,
      sym_optional_chain,
    STATE(797), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1595), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1597), 22,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_SEMI,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [20137] = 9,
    ACTIONS(1818), 1,
      anon_sym_LPAREN,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_DOT,
    ACTIONS(1838), 1,
      sym_optional_chain,
    STATE(797), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1840), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1583), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1585), 20,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_SEMI,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_BQUOTE,
  [20197] = 6,
    ACTIONS(718), 1,
      anon_sym_EQ,
    ACTIONS(720), 1,
      sym__automatic_semicolon,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(708), 2,
      anon_sym_else,
      anon_sym_while,
    ACTIONS(714), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(716), 23,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [20251] = 4,
    ACTIONS(718), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(714), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(716), 26,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [20301] = 9,
    ACTIONS(1818), 1,
      anon_sym_LPAREN,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_DOT,
    ACTIONS(1838), 1,
      sym_optional_chain,
    STATE(797), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1840), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1591), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1593), 20,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_SEMI,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_BQUOTE,
  [20361] = 25,
    ACTIONS(1818), 1,
      anon_sym_LPAREN,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_DOT,
    ACTIONS(1838), 1,
      sym_optional_chain,
    ACTIONS(1846), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1850), 1,
      anon_sym_GT_GT,
    ACTIONS(1854), 1,
      anon_sym_AMP,
    ACTIONS(1856), 1,
      anon_sym_CARET,
    ACTIONS(1858), 1,
      anon_sym_PIPE,
    ACTIONS(1862), 1,
      anon_sym_PERCENT,
    ACTIONS(1864), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1872), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1874), 1,
      sym__ternary_qmark,
    STATE(797), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1840), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1842), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1852), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1860), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1868), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1870), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1844), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1866), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1826), 5,
      sym__automatic_semicolon,
      anon_sym_COMMA,
      anon_sym_of,
      anon_sym_SEMI,
      anon_sym_BQUOTE,
  [20452] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1690), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1692), 26,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [20499] = 25,
    ACTIONS(1818), 1,
      anon_sym_LPAREN,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_DOT,
    ACTIONS(1838), 1,
      sym_optional_chain,
    ACTIONS(1846), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1850), 1,
      anon_sym_GT_GT,
    ACTIONS(1854), 1,
      anon_sym_AMP,
    ACTIONS(1856), 1,
      anon_sym_CARET,
    ACTIONS(1858), 1,
      anon_sym_PIPE,
    ACTIONS(1862), 1,
      anon_sym_PERCENT,
    ACTIONS(1864), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1872), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1874), 1,
      sym__ternary_qmark,
    STATE(797), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1840), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1842), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1852), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1860), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1868), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1870), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1844), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1866), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1780), 5,
      sym__automatic_semicolon,
      anon_sym_COMMA,
      anon_sym_of,
      anon_sym_SEMI,
      anon_sym_BQUOTE,
  [20590] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1611), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1613), 26,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [20637] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1615), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1617), 26,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [20684] = 4,
    ACTIONS(1876), 1,
      sym_regex_flags,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1619), 13,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_instanceof,
    ACTIONS(1621), 24,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [20733] = 5,
    ACTIONS(740), 1,
      sym__automatic_semicolon,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(732), 2,
      anon_sym_else,
      anon_sym_while,
    ACTIONS(736), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(738), 23,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [20784] = 25,
    ACTIONS(1818), 1,
      anon_sym_LPAREN,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_DOT,
    ACTIONS(1838), 1,
      sym_optional_chain,
    ACTIONS(1846), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1850), 1,
      anon_sym_GT_GT,
    ACTIONS(1854), 1,
      anon_sym_AMP,
    ACTIONS(1856), 1,
      anon_sym_CARET,
    ACTIONS(1858), 1,
      anon_sym_PIPE,
    ACTIONS(1862), 1,
      anon_sym_PERCENT,
    ACTIONS(1864), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1872), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1874), 1,
      sym__ternary_qmark,
    STATE(797), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1840), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1842), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1852), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1860), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1868), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1870), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1844), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1866), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1631), 5,
      sym__automatic_semicolon,
      anon_sym_COMMA,
      anon_sym_of,
      anon_sym_SEMI,
      anon_sym_BQUOTE,
  [20875] = 15,
    ACTIONS(1818), 1,
      anon_sym_LPAREN,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_DOT,
    ACTIONS(1838), 1,
      sym_optional_chain,
    ACTIONS(1850), 1,
      anon_sym_GT_GT,
    ACTIONS(1862), 1,
      anon_sym_PERCENT,
    ACTIONS(1864), 1,
      anon_sym_STAR_STAR,
    STATE(797), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1840), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1842), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1852), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1860), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1830), 7,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1828), 15,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_of,
      anon_sym_SEMI,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_CARET,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_BQUOTE,
  [20946] = 10,
    ACTIONS(1818), 1,
      anon_sym_LPAREN,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_DOT,
    ACTIONS(1838), 1,
      sym_optional_chain,
    ACTIONS(1864), 1,
      anon_sym_STAR_STAR,
    STATE(797), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1840), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1830), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1828), 18,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_of,
      anon_sym_SEMI,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_BQUOTE,
  [21007] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1625), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1627), 26,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [21054] = 21,
    ACTIONS(1818), 1,
      anon_sym_LPAREN,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_DOT,
    ACTIONS(1838), 1,
      sym_optional_chain,
    ACTIONS(1850), 1,
      anon_sym_GT_GT,
    ACTIONS(1854), 1,
      anon_sym_AMP,
    ACTIONS(1856), 1,
      anon_sym_CARET,
    ACTIONS(1858), 1,
      anon_sym_PIPE,
    ACTIONS(1862), 1,
      anon_sym_PERCENT,
    ACTIONS(1864), 1,
      anon_sym_STAR_STAR,
    STATE(797), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1840), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1842), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1852), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1860), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1868), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1870), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1844), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1866), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1828), 9,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_of,
      anon_sym_SEMI,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_QMARK,
      anon_sym_BQUOTE,
  [21137] = 22,
    ACTIONS(1818), 1,
      anon_sym_LPAREN,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_DOT,
    ACTIONS(1838), 1,
      sym_optional_chain,
    ACTIONS(1846), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1850), 1,
      anon_sym_GT_GT,
    ACTIONS(1854), 1,
      anon_sym_AMP,
    ACTIONS(1856), 1,
      anon_sym_CARET,
    ACTIONS(1858), 1,
      anon_sym_PIPE,
    ACTIONS(1862), 1,
      anon_sym_PERCENT,
    ACTIONS(1864), 1,
      anon_sym_STAR_STAR,
    STATE(797), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1840), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1842), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1852), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1860), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1868), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1870), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1844), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1866), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1828), 8,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_of,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_QMARK,
      anon_sym_BQUOTE,
  [21222] = 13,
    ACTIONS(1818), 1,
      anon_sym_LPAREN,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_DOT,
    ACTIONS(1838), 1,
      sym_optional_chain,
    ACTIONS(1862), 1,
      anon_sym_PERCENT,
    ACTIONS(1864), 1,
      anon_sym_STAR_STAR,
    STATE(797), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1840), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1842), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1860), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1830), 8,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1828), 17,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_of,
      anon_sym_SEMI,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_BQUOTE,
  [21289] = 19,
    ACTIONS(1818), 1,
      anon_sym_LPAREN,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_DOT,
    ACTIONS(1838), 1,
      sym_optional_chain,
    ACTIONS(1850), 1,
      anon_sym_GT_GT,
    ACTIONS(1862), 1,
      anon_sym_PERCENT,
    ACTIONS(1864), 1,
      anon_sym_STAR_STAR,
    STATE(797), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1830), 2,
      anon_sym_AMP,
      anon_sym_PIPE,
    ACTIONS(1840), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1842), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1852), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1860), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1868), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1870), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1844), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1866), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1828), 10,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_of,
      anon_sym_SEMI,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_CARET,
      anon_sym_QMARK_QMARK,
      anon_sym_BQUOTE,
  [21368] = 20,
    ACTIONS(1818), 1,
      anon_sym_LPAREN,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_DOT,
    ACTIONS(1830), 1,
      anon_sym_PIPE,
    ACTIONS(1838), 1,
      sym_optional_chain,
    ACTIONS(1850), 1,
      anon_sym_GT_GT,
    ACTIONS(1854), 1,
      anon_sym_AMP,
    ACTIONS(1862), 1,
      anon_sym_PERCENT,
    ACTIONS(1864), 1,
      anon_sym_STAR_STAR,
    STATE(797), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1840), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1842), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1852), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1860), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1868), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1870), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1844), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1866), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1828), 10,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_of,
      anon_sym_SEMI,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_CARET,
      anon_sym_QMARK_QMARK,
      anon_sym_BQUOTE,
  [21449] = 21,
    ACTIONS(1818), 1,
      anon_sym_LPAREN,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_DOT,
    ACTIONS(1830), 1,
      anon_sym_PIPE,
    ACTIONS(1838), 1,
      sym_optional_chain,
    ACTIONS(1850), 1,
      anon_sym_GT_GT,
    ACTIONS(1854), 1,
      anon_sym_AMP,
    ACTIONS(1856), 1,
      anon_sym_CARET,
    ACTIONS(1862), 1,
      anon_sym_PERCENT,
    ACTIONS(1864), 1,
      anon_sym_STAR_STAR,
    STATE(797), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1840), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1842), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1852), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1860), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1868), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1870), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1844), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1866), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1828), 9,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_of,
      anon_sym_SEMI,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_QMARK,
      anon_sym_BQUOTE,
  [21532] = 12,
    ACTIONS(1818), 1,
      anon_sym_LPAREN,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_DOT,
    ACTIONS(1838), 1,
      sym_optional_chain,
    ACTIONS(1862), 1,
      anon_sym_PERCENT,
    ACTIONS(1864), 1,
      anon_sym_STAR_STAR,
    STATE(797), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1840), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1842), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1830), 10,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1828), 17,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_of,
      anon_sym_SEMI,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_BQUOTE,
  [21597] = 10,
    ACTIONS(1818), 1,
      anon_sym_LPAREN,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_DOT,
    ACTIONS(1838), 1,
      sym_optional_chain,
    ACTIONS(1864), 1,
      anon_sym_STAR_STAR,
    STATE(797), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1840), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1830), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1828), 18,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_of,
      anon_sym_SEMI,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_BQUOTE,
  [21658] = 17,
    ACTIONS(1818), 1,
      anon_sym_LPAREN,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_DOT,
    ACTIONS(1838), 1,
      sym_optional_chain,
    ACTIONS(1850), 1,
      anon_sym_GT_GT,
    ACTIONS(1862), 1,
      anon_sym_PERCENT,
    ACTIONS(1864), 1,
      anon_sym_STAR_STAR,
    STATE(797), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1840), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1842), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1852), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1860), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1844), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1866), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1830), 4,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1828), 12,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_of,
      anon_sym_SEMI,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_CARET,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_BQUOTE,
  [21733] = 23,
    ACTIONS(1818), 1,
      anon_sym_LPAREN,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_DOT,
    ACTIONS(1838), 1,
      sym_optional_chain,
    ACTIONS(1846), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1850), 1,
      anon_sym_GT_GT,
    ACTIONS(1854), 1,
      anon_sym_AMP,
    ACTIONS(1856), 1,
      anon_sym_CARET,
    ACTIONS(1858), 1,
      anon_sym_PIPE,
    ACTIONS(1862), 1,
      anon_sym_PERCENT,
    ACTIONS(1864), 1,
      anon_sym_STAR_STAR,
    STATE(797), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1840), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1842), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1852), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1860), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1868), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1870), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1844), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1866), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1828), 7,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_of,
      anon_sym_SEMI,
      anon_sym_QMARK_QMARK,
      anon_sym_BQUOTE,
  [21820] = 25,
    ACTIONS(1818), 1,
      anon_sym_LPAREN,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_DOT,
    ACTIONS(1838), 1,
      sym_optional_chain,
    ACTIONS(1846), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1850), 1,
      anon_sym_GT_GT,
    ACTIONS(1854), 1,
      anon_sym_AMP,
    ACTIONS(1856), 1,
      anon_sym_CARET,
    ACTIONS(1858), 1,
      anon_sym_PIPE,
    ACTIONS(1862), 1,
      anon_sym_PERCENT,
    ACTIONS(1864), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1872), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1874), 1,
      sym__ternary_qmark,
    STATE(797), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1840), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1842), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1852), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1860), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1868), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1870), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1844), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1866), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1834), 5,
      sym__automatic_semicolon,
      anon_sym_COMMA,
      anon_sym_of,
      anon_sym_SEMI,
      anon_sym_BQUOTE,
  [21911] = 25,
    ACTIONS(1818), 1,
      anon_sym_LPAREN,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_DOT,
    ACTIONS(1838), 1,
      sym_optional_chain,
    ACTIONS(1846), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1850), 1,
      anon_sym_GT_GT,
    ACTIONS(1854), 1,
      anon_sym_AMP,
    ACTIONS(1856), 1,
      anon_sym_CARET,
    ACTIONS(1858), 1,
      anon_sym_PIPE,
    ACTIONS(1862), 1,
      anon_sym_PERCENT,
    ACTIONS(1864), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1872), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1874), 1,
      sym__ternary_qmark,
    STATE(797), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1840), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1842), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1852), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1860), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1868), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1870), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1844), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1866), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1676), 5,
      sym__automatic_semicolon,
      anon_sym_COMMA,
      anon_sym_of,
      anon_sym_SEMI,
      anon_sym_BQUOTE,
  [22002] = 25,
    ACTIONS(1818), 1,
      anon_sym_LPAREN,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_DOT,
    ACTIONS(1838), 1,
      sym_optional_chain,
    ACTIONS(1882), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1884), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1886), 1,
      anon_sym_GT_GT,
    ACTIONS(1890), 1,
      anon_sym_AMP,
    ACTIONS(1892), 1,
      anon_sym_CARET,
    ACTIONS(1894), 1,
      anon_sym_PIPE,
    ACTIONS(1898), 1,
      anon_sym_PERCENT,
    ACTIONS(1900), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1908), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1910), 1,
      sym__ternary_qmark,
    STATE(797), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1840), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1878), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1888), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1896), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1904), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1906), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1880), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1902), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1826), 5,
      sym__automatic_semicolon,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_SEMI,
      anon_sym_BQUOTE,
  [22093] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1629), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1631), 26,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [22140] = 25,
    ACTIONS(1818), 1,
      anon_sym_LPAREN,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_DOT,
    ACTIONS(1838), 1,
      sym_optional_chain,
    ACTIONS(1846), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1850), 1,
      anon_sym_GT_GT,
    ACTIONS(1854), 1,
      anon_sym_AMP,
    ACTIONS(1856), 1,
      anon_sym_CARET,
    ACTIONS(1858), 1,
      anon_sym_PIPE,
    ACTIONS(1862), 1,
      anon_sym_PERCENT,
    ACTIONS(1864), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1872), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1874), 1,
      sym__ternary_qmark,
    STATE(797), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1840), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1842), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1852), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1860), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1868), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1870), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1844), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1866), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1700), 5,
      sym__automatic_semicolon,
      anon_sym_COMMA,
      anon_sym_of,
      anon_sym_SEMI,
      anon_sym_BQUOTE,
  [22231] = 25,
    ACTIONS(1818), 1,
      anon_sym_LPAREN,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_DOT,
    ACTIONS(1838), 1,
      sym_optional_chain,
    ACTIONS(1846), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1850), 1,
      anon_sym_GT_GT,
    ACTIONS(1854), 1,
      anon_sym_AMP,
    ACTIONS(1856), 1,
      anon_sym_CARET,
    ACTIONS(1858), 1,
      anon_sym_PIPE,
    ACTIONS(1862), 1,
      anon_sym_PERCENT,
    ACTIONS(1864), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1872), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1874), 1,
      sym__ternary_qmark,
    STATE(797), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1840), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1842), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1852), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1860), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1868), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1870), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1844), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1866), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1708), 5,
      sym__automatic_semicolon,
      anon_sym_COMMA,
      anon_sym_of,
      anon_sym_SEMI,
      anon_sym_BQUOTE,
  [22322] = 25,
    ACTIONS(1818), 1,
      anon_sym_LPAREN,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_DOT,
    ACTIONS(1838), 1,
      sym_optional_chain,
    ACTIONS(1846), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1850), 1,
      anon_sym_GT_GT,
    ACTIONS(1854), 1,
      anon_sym_AMP,
    ACTIONS(1856), 1,
      anon_sym_CARET,
    ACTIONS(1858), 1,
      anon_sym_PIPE,
    ACTIONS(1862), 1,
      anon_sym_PERCENT,
    ACTIONS(1864), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1872), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1874), 1,
      sym__ternary_qmark,
    STATE(797), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1840), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1842), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1852), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1860), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1868), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1870), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1844), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1866), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1712), 5,
      sym__automatic_semicolon,
      anon_sym_COMMA,
      anon_sym_of,
      anon_sym_SEMI,
      anon_sym_BQUOTE,
  [22413] = 5,
    ACTIONS(730), 1,
      sym__automatic_semicolon,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(722), 2,
      anon_sym_else,
      anon_sym_while,
    ACTIONS(726), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(728), 23,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [22464] = 25,
    ACTIONS(1818), 1,
      anon_sym_LPAREN,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_DOT,
    ACTIONS(1838), 1,
      sym_optional_chain,
    ACTIONS(1882), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1884), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1886), 1,
      anon_sym_GT_GT,
    ACTIONS(1890), 1,
      anon_sym_AMP,
    ACTIONS(1892), 1,
      anon_sym_CARET,
    ACTIONS(1894), 1,
      anon_sym_PIPE,
    ACTIONS(1898), 1,
      anon_sym_PERCENT,
    ACTIONS(1900), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1908), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1910), 1,
      sym__ternary_qmark,
    STATE(797), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1840), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1878), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1888), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1896), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1904), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1906), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1880), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1902), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1631), 5,
      sym__automatic_semicolon,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_SEMI,
      anon_sym_BQUOTE,
  [22555] = 15,
    ACTIONS(1818), 1,
      anon_sym_LPAREN,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_DOT,
    ACTIONS(1838), 1,
      sym_optional_chain,
    ACTIONS(1886), 1,
      anon_sym_GT_GT,
    ACTIONS(1898), 1,
      anon_sym_PERCENT,
    ACTIONS(1900), 1,
      anon_sym_STAR_STAR,
    STATE(797), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1840), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1878), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1888), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1896), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1830), 7,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1828), 15,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_SEMI,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_CARET,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_BQUOTE,
  [22626] = 25,
    ACTIONS(1818), 1,
      anon_sym_LPAREN,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_DOT,
    ACTIONS(1838), 1,
      sym_optional_chain,
    ACTIONS(1846), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1850), 1,
      anon_sym_GT_GT,
    ACTIONS(1854), 1,
      anon_sym_AMP,
    ACTIONS(1856), 1,
      anon_sym_CARET,
    ACTIONS(1858), 1,
      anon_sym_PIPE,
    ACTIONS(1862), 1,
      anon_sym_PERCENT,
    ACTIONS(1864), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1872), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1874), 1,
      sym__ternary_qmark,
    STATE(797), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1840), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1842), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1852), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1860), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1868), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1870), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1844), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1866), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1814), 5,
      sym__automatic_semicolon,
      anon_sym_COMMA,
      anon_sym_of,
      anon_sym_SEMI,
      anon_sym_BQUOTE,
  [22717] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1633), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1635), 26,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [22764] = 25,
    ACTIONS(1818), 1,
      anon_sym_LPAREN,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_DOT,
    ACTIONS(1838), 1,
      sym_optional_chain,
    ACTIONS(1882), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1884), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1886), 1,
      anon_sym_GT_GT,
    ACTIONS(1890), 1,
      anon_sym_AMP,
    ACTIONS(1892), 1,
      anon_sym_CARET,
    ACTIONS(1894), 1,
      anon_sym_PIPE,
    ACTIONS(1898), 1,
      anon_sym_PERCENT,
    ACTIONS(1900), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1908), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1910), 1,
      sym__ternary_qmark,
    STATE(797), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1840), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1878), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1888), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1896), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1904), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1906), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1880), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1902), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1716), 5,
      sym__automatic_semicolon,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_SEMI,
      anon_sym_BQUOTE,
  [22855] = 5,
    ACTIONS(1644), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1912), 4,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
    ACTIONS(1637), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1639), 21,
      sym__ternary_qmark,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [22906] = 5,
    ACTIONS(1729), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1914), 4,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
    ACTIONS(1722), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1724), 21,
      sym__ternary_qmark,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [22957] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1646), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1648), 26,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [23004] = 5,
    ACTIONS(718), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1916), 4,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
    ACTIONS(714), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(716), 21,
      sym__ternary_qmark,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [23055] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1686), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1688), 26,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [23102] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1774), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1776), 26,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [23149] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(756), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(758), 26,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [23196] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1714), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1716), 26,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [23243] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1694), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1696), 26,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [23290] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1739), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1741), 26,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [23337] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1426), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1428), 26,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [23384] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1747), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1749), 26,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [23431] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1710), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1712), 26,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [23478] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(776), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(778), 26,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [23525] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1650), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1652), 26,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [23572] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1658), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1660), 26,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [23619] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1577), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1579), 26,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [23666] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1735), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1737), 26,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [23713] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1682), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1684), 26,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [23760] = 10,
    ACTIONS(1818), 1,
      anon_sym_LPAREN,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_DOT,
    ACTIONS(1838), 1,
      sym_optional_chain,
    ACTIONS(1900), 1,
      anon_sym_STAR_STAR,
    STATE(797), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1840), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1830), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1828), 18,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_SEMI,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_BQUOTE,
  [23821] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1759), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1761), 26,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [23868] = 21,
    ACTIONS(1818), 1,
      anon_sym_LPAREN,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_DOT,
    ACTIONS(1838), 1,
      sym_optional_chain,
    ACTIONS(1886), 1,
      anon_sym_GT_GT,
    ACTIONS(1890), 1,
      anon_sym_AMP,
    ACTIONS(1892), 1,
      anon_sym_CARET,
    ACTIONS(1894), 1,
      anon_sym_PIPE,
    ACTIONS(1898), 1,
      anon_sym_PERCENT,
    ACTIONS(1900), 1,
      anon_sym_STAR_STAR,
    STATE(797), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1840), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1878), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1888), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1896), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1904), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1906), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1880), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1902), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1828), 9,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_SEMI,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_QMARK,
      anon_sym_BQUOTE,
  [23951] = 5,
    ACTIONS(750), 1,
      sym__automatic_semicolon,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(742), 2,
      anon_sym_else,
      anon_sym_while,
    ACTIONS(746), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(748), 23,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [24002] = 22,
    ACTIONS(1818), 1,
      anon_sym_LPAREN,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_DOT,
    ACTIONS(1838), 1,
      sym_optional_chain,
    ACTIONS(1882), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1886), 1,
      anon_sym_GT_GT,
    ACTIONS(1890), 1,
      anon_sym_AMP,
    ACTIONS(1892), 1,
      anon_sym_CARET,
    ACTIONS(1894), 1,
      anon_sym_PIPE,
    ACTIONS(1898), 1,
      anon_sym_PERCENT,
    ACTIONS(1900), 1,
      anon_sym_STAR_STAR,
    STATE(797), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1840), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1878), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1888), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1896), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1904), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1906), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1880), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1902), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1828), 8,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_SEMI,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_QMARK,
      anon_sym_BQUOTE,
  [24087] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(786), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(788), 26,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [24134] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(796), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(798), 26,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [24181] = 13,
    ACTIONS(1818), 1,
      anon_sym_LPAREN,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_DOT,
    ACTIONS(1838), 1,
      sym_optional_chain,
    ACTIONS(1898), 1,
      anon_sym_PERCENT,
    ACTIONS(1900), 1,
      anon_sym_STAR_STAR,
    STATE(797), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1840), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1878), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1896), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1830), 8,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1828), 17,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_SEMI,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_BQUOTE,
  [24248] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1694), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1696), 26,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [24295] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1654), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1656), 26,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [24342] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1698), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1700), 26,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [24389] = 19,
    ACTIONS(1818), 1,
      anon_sym_LPAREN,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_DOT,
    ACTIONS(1838), 1,
      sym_optional_chain,
    ACTIONS(1886), 1,
      anon_sym_GT_GT,
    ACTIONS(1898), 1,
      anon_sym_PERCENT,
    ACTIONS(1900), 1,
      anon_sym_STAR_STAR,
    STATE(797), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1830), 2,
      anon_sym_AMP,
      anon_sym_PIPE,
    ACTIONS(1840), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1878), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1888), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1896), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1904), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1906), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1880), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1902), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1828), 10,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_SEMI,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_CARET,
      anon_sym_QMARK_QMARK,
      anon_sym_BQUOTE,
  [24468] = 5,
    ACTIONS(1459), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1543), 4,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
    ACTIONS(1426), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1428), 21,
      sym__ternary_qmark,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [24519] = 25,
    ACTIONS(1818), 1,
      anon_sym_LPAREN,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_DOT,
    ACTIONS(1838), 1,
      sym_optional_chain,
    ACTIONS(1882), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1884), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1886), 1,
      anon_sym_GT_GT,
    ACTIONS(1890), 1,
      anon_sym_AMP,
    ACTIONS(1892), 1,
      anon_sym_CARET,
    ACTIONS(1894), 1,
      anon_sym_PIPE,
    ACTIONS(1898), 1,
      anon_sym_PERCENT,
    ACTIONS(1900), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1908), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1910), 1,
      sym__ternary_qmark,
    STATE(797), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1840), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1878), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1888), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1896), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1904), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1906), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1880), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1902), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1780), 5,
      sym__automatic_semicolon,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_SEMI,
      anon_sym_BQUOTE,
  [24610] = 20,
    ACTIONS(1818), 1,
      anon_sym_LPAREN,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_DOT,
    ACTIONS(1830), 1,
      anon_sym_PIPE,
    ACTIONS(1838), 1,
      sym_optional_chain,
    ACTIONS(1886), 1,
      anon_sym_GT_GT,
    ACTIONS(1890), 1,
      anon_sym_AMP,
    ACTIONS(1898), 1,
      anon_sym_PERCENT,
    ACTIONS(1900), 1,
      anon_sym_STAR_STAR,
    STATE(797), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1840), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1878), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1888), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1896), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1904), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1906), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1880), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1902), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1828), 10,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_SEMI,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_CARET,
      anon_sym_QMARK_QMARK,
      anon_sym_BQUOTE,
  [24691] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1718), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1720), 26,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [24738] = 5,
    ACTIONS(1532), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1530), 4,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
    ACTIONS(1426), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1428), 21,
      sym__ternary_qmark,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [24789] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(766), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(768), 26,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [24836] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1770), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1772), 26,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [24883] = 5,
    ACTIONS(780), 1,
      sym__automatic_semicolon,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(772), 2,
      anon_sym_else,
      anon_sym_while,
    ACTIONS(776), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(778), 23,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [24934] = 25,
    ACTIONS(1818), 1,
      anon_sym_LPAREN,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_DOT,
    ACTIONS(1838), 1,
      sym_optional_chain,
    ACTIONS(1882), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1884), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1886), 1,
      anon_sym_GT_GT,
    ACTIONS(1890), 1,
      anon_sym_AMP,
    ACTIONS(1892), 1,
      anon_sym_CARET,
    ACTIONS(1894), 1,
      anon_sym_PIPE,
    ACTIONS(1898), 1,
      anon_sym_PERCENT,
    ACTIONS(1900), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1908), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1910), 1,
      sym__ternary_qmark,
    STATE(797), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1840), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1878), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1888), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1896), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1904), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1906), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1880), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1902), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1814), 5,
      sym__automatic_semicolon,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_SEMI,
      anon_sym_BQUOTE,
  [25025] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1603), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1605), 26,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [25072] = 4,
    ACTIONS(1528), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1426), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1428), 25,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [25121] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(736), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(738), 26,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [25168] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(746), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(748), 26,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [25215] = 5,
    ACTIONS(760), 1,
      sym__automatic_semicolon,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(752), 2,
      anon_sym_else,
      anon_sym_while,
    ACTIONS(756), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(758), 23,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [25266] = 21,
    ACTIONS(1818), 1,
      anon_sym_LPAREN,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_DOT,
    ACTIONS(1830), 1,
      anon_sym_PIPE,
    ACTIONS(1838), 1,
      sym_optional_chain,
    ACTIONS(1886), 1,
      anon_sym_GT_GT,
    ACTIONS(1890), 1,
      anon_sym_AMP,
    ACTIONS(1892), 1,
      anon_sym_CARET,
    ACTIONS(1898), 1,
      anon_sym_PERCENT,
    ACTIONS(1900), 1,
      anon_sym_STAR_STAR,
    STATE(797), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1840), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1878), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1888), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1896), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1904), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1906), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1880), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1902), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1828), 9,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_SEMI,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_QMARK,
      anon_sym_BQUOTE,
  [25349] = 12,
    ACTIONS(1818), 1,
      anon_sym_LPAREN,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_DOT,
    ACTIONS(1838), 1,
      sym_optional_chain,
    ACTIONS(1898), 1,
      anon_sym_PERCENT,
    ACTIONS(1900), 1,
      anon_sym_STAR_STAR,
    STATE(797), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1840), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1878), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1830), 10,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1828), 17,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_SEMI,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_BQUOTE,
  [25414] = 10,
    ACTIONS(1818), 1,
      anon_sym_LPAREN,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_DOT,
    ACTIONS(1838), 1,
      sym_optional_chain,
    ACTIONS(1900), 1,
      anon_sym_STAR_STAR,
    STATE(797), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1840), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1830), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1828), 18,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_SEMI,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_BQUOTE,
  [25475] = 17,
    ACTIONS(1818), 1,
      anon_sym_LPAREN,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_DOT,
    ACTIONS(1838), 1,
      sym_optional_chain,
    ACTIONS(1886), 1,
      anon_sym_GT_GT,
    ACTIONS(1898), 1,
      anon_sym_PERCENT,
    ACTIONS(1900), 1,
      anon_sym_STAR_STAR,
    STATE(797), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1840), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1878), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1888), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1896), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1880), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1902), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1830), 4,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1828), 12,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_SEMI,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_CARET,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_BQUOTE,
  [25550] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1702), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1704), 26,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [25597] = 23,
    ACTIONS(1818), 1,
      anon_sym_LPAREN,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_DOT,
    ACTIONS(1838), 1,
      sym_optional_chain,
    ACTIONS(1882), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1884), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1886), 1,
      anon_sym_GT_GT,
    ACTIONS(1890), 1,
      anon_sym_AMP,
    ACTIONS(1892), 1,
      anon_sym_CARET,
    ACTIONS(1894), 1,
      anon_sym_PIPE,
    ACTIONS(1898), 1,
      anon_sym_PERCENT,
    ACTIONS(1900), 1,
      anon_sym_STAR_STAR,
    STATE(797), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1840), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1878), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1888), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1896), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1904), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1906), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1880), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1902), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1828), 7,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_SEMI,
      anon_sym_QMARK_QMARK,
      anon_sym_BQUOTE,
  [25684] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1694), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1696), 26,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [25731] = 5,
    ACTIONS(790), 1,
      sym__automatic_semicolon,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(782), 2,
      anon_sym_else,
      anon_sym_while,
    ACTIONS(786), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(788), 23,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [25782] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1607), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1609), 26,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [25829] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1662), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1664), 26,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [25876] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1694), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1696), 26,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [25923] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(726), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(728), 26,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [25970] = 25,
    ACTIONS(1818), 1,
      anon_sym_LPAREN,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_DOT,
    ACTIONS(1838), 1,
      sym_optional_chain,
    ACTIONS(1882), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1884), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1886), 1,
      anon_sym_GT_GT,
    ACTIONS(1890), 1,
      anon_sym_AMP,
    ACTIONS(1892), 1,
      anon_sym_CARET,
    ACTIONS(1894), 1,
      anon_sym_PIPE,
    ACTIONS(1898), 1,
      anon_sym_PERCENT,
    ACTIONS(1900), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1908), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1910), 1,
      sym__ternary_qmark,
    STATE(797), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1840), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1878), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1888), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1896), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1904), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1906), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1880), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1902), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1816), 5,
      sym__automatic_semicolon,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_SEMI,
      anon_sym_BQUOTE,
  [26061] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1666), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1668), 26,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [26108] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1755), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1757), 26,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [26155] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1670), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1672), 26,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [26202] = 25,
    ACTIONS(1818), 1,
      anon_sym_LPAREN,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_DOT,
    ACTIONS(1838), 1,
      sym_optional_chain,
    ACTIONS(1882), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1884), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1886), 1,
      anon_sym_GT_GT,
    ACTIONS(1890), 1,
      anon_sym_AMP,
    ACTIONS(1892), 1,
      anon_sym_CARET,
    ACTIONS(1894), 1,
      anon_sym_PIPE,
    ACTIONS(1898), 1,
      anon_sym_PERCENT,
    ACTIONS(1900), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1908), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1910), 1,
      sym__ternary_qmark,
    STATE(797), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1840), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1878), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1888), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1896), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1904), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1906), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1880), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1902), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1834), 5,
      sym__automatic_semicolon,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_SEMI,
      anon_sym_BQUOTE,
  [26293] = 25,
    ACTIONS(1818), 1,
      anon_sym_LPAREN,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_DOT,
    ACTIONS(1838), 1,
      sym_optional_chain,
    ACTIONS(1882), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1884), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1886), 1,
      anon_sym_GT_GT,
    ACTIONS(1890), 1,
      anon_sym_AMP,
    ACTIONS(1892), 1,
      anon_sym_CARET,
    ACTIONS(1894), 1,
      anon_sym_PIPE,
    ACTIONS(1898), 1,
      anon_sym_PERCENT,
    ACTIONS(1900), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1908), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1910), 1,
      sym__ternary_qmark,
    STATE(797), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1840), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1878), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1888), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1896), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1904), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1906), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1880), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1902), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1700), 5,
      sym__automatic_semicolon,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_SEMI,
      anon_sym_BQUOTE,
  [26384] = 4,
    ACTIONS(1524), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1426), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1428), 25,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [26433] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1706), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1708), 26,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [26480] = 4,
    ACTIONS(1876), 1,
      sym_regex_flags,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1619), 14,
      anon_sym_STAR,
      anon_sym_of,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_instanceof,
    ACTIONS(1621), 23,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [26529] = 25,
    ACTIONS(1818), 1,
      anon_sym_LPAREN,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_DOT,
    ACTIONS(1838), 1,
      sym_optional_chain,
    ACTIONS(1846), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1850), 1,
      anon_sym_GT_GT,
    ACTIONS(1854), 1,
      anon_sym_AMP,
    ACTIONS(1856), 1,
      anon_sym_CARET,
    ACTIONS(1858), 1,
      anon_sym_PIPE,
    ACTIONS(1862), 1,
      anon_sym_PERCENT,
    ACTIONS(1864), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1872), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1874), 1,
      sym__ternary_qmark,
    STATE(797), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1840), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1842), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1852), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1860), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1868), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1870), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1844), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1866), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1832), 5,
      sym__automatic_semicolon,
      anon_sym_COMMA,
      anon_sym_of,
      anon_sym_SEMI,
      anon_sym_BQUOTE,
  [26620] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1654), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1656), 26,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [26667] = 5,
    ACTIONS(770), 1,
      sym__automatic_semicolon,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(762), 2,
      anon_sym_else,
      anon_sym_while,
    ACTIONS(766), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(768), 23,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [26718] = 5,
    ACTIONS(800), 1,
      sym__automatic_semicolon,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(792), 2,
      anon_sym_else,
      anon_sym_while,
    ACTIONS(796), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(798), 23,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [26769] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1743), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1745), 26,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [26816] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1674), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1676), 26,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [26863] = 25,
    ACTIONS(1818), 1,
      anon_sym_LPAREN,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_DOT,
    ACTIONS(1838), 1,
      sym_optional_chain,
    ACTIONS(1882), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1884), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1886), 1,
      anon_sym_GT_GT,
    ACTIONS(1890), 1,
      anon_sym_AMP,
    ACTIONS(1892), 1,
      anon_sym_CARET,
    ACTIONS(1894), 1,
      anon_sym_PIPE,
    ACTIONS(1898), 1,
      anon_sym_PERCENT,
    ACTIONS(1900), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1908), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1910), 1,
      sym__ternary_qmark,
    STATE(797), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1840), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1878), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1888), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1896), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1904), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1906), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1880), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1902), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1832), 5,
      sym__automatic_semicolon,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_SEMI,
      anon_sym_BQUOTE,
  [26954] = 25,
    ACTIONS(1818), 1,
      anon_sym_LPAREN,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_DOT,
    ACTIONS(1838), 1,
      sym_optional_chain,
    ACTIONS(1882), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1884), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1886), 1,
      anon_sym_GT_GT,
    ACTIONS(1890), 1,
      anon_sym_AMP,
    ACTIONS(1892), 1,
      anon_sym_CARET,
    ACTIONS(1894), 1,
      anon_sym_PIPE,
    ACTIONS(1898), 1,
      anon_sym_PERCENT,
    ACTIONS(1900), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1908), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1910), 1,
      sym__ternary_qmark,
    STATE(797), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1840), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1878), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1888), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1896), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1904), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1906), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1880), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1902), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1676), 5,
      sym__automatic_semicolon,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_SEMI,
      anon_sym_BQUOTE,
  [27045] = 5,
    ACTIONS(1540), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1537), 4,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
    ACTIONS(1426), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1428), 21,
      sym__ternary_qmark,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [27096] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1599), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1601), 26,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [27143] = 25,
    ACTIONS(1818), 1,
      anon_sym_LPAREN,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_DOT,
    ACTIONS(1838), 1,
      sym_optional_chain,
    ACTIONS(1882), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1884), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1886), 1,
      anon_sym_GT_GT,
    ACTIONS(1890), 1,
      anon_sym_AMP,
    ACTIONS(1892), 1,
      anon_sym_CARET,
    ACTIONS(1894), 1,
      anon_sym_PIPE,
    ACTIONS(1898), 1,
      anon_sym_PERCENT,
    ACTIONS(1900), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1908), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1910), 1,
      sym__ternary_qmark,
    STATE(797), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1840), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1878), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1888), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1896), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1904), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1906), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1880), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1902), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1708), 5,
      sym__automatic_semicolon,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_SEMI,
      anon_sym_BQUOTE,
  [27234] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1763), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1765), 26,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [27281] = 25,
    ACTIONS(1818), 1,
      anon_sym_LPAREN,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_DOT,
    ACTIONS(1838), 1,
      sym_optional_chain,
    ACTIONS(1846), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1850), 1,
      anon_sym_GT_GT,
    ACTIONS(1854), 1,
      anon_sym_AMP,
    ACTIONS(1856), 1,
      anon_sym_CARET,
    ACTIONS(1858), 1,
      anon_sym_PIPE,
    ACTIONS(1862), 1,
      anon_sym_PERCENT,
    ACTIONS(1864), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1872), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1874), 1,
      sym__ternary_qmark,
    STATE(797), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1840), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1842), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1852), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1860), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1868), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1870), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1844), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1866), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1816), 5,
      sym__automatic_semicolon,
      anon_sym_COMMA,
      anon_sym_of,
      anon_sym_SEMI,
      anon_sym_BQUOTE,
  [27372] = 25,
    ACTIONS(1818), 1,
      anon_sym_LPAREN,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_DOT,
    ACTIONS(1838), 1,
      sym_optional_chain,
    ACTIONS(1846), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1850), 1,
      anon_sym_GT_GT,
    ACTIONS(1854), 1,
      anon_sym_AMP,
    ACTIONS(1856), 1,
      anon_sym_CARET,
    ACTIONS(1858), 1,
      anon_sym_PIPE,
    ACTIONS(1862), 1,
      anon_sym_PERCENT,
    ACTIONS(1864), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1872), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1874), 1,
      sym__ternary_qmark,
    STATE(797), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1840), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1842), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1852), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1860), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1868), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1870), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1844), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1866), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1716), 5,
      sym__automatic_semicolon,
      anon_sym_COMMA,
      anon_sym_of,
      anon_sym_SEMI,
      anon_sym_BQUOTE,
  [27463] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1731), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1733), 26,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [27510] = 25,
    ACTIONS(1818), 1,
      anon_sym_LPAREN,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_DOT,
    ACTIONS(1838), 1,
      sym_optional_chain,
    ACTIONS(1882), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1884), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1886), 1,
      anon_sym_GT_GT,
    ACTIONS(1890), 1,
      anon_sym_AMP,
    ACTIONS(1892), 1,
      anon_sym_CARET,
    ACTIONS(1894), 1,
      anon_sym_PIPE,
    ACTIONS(1898), 1,
      anon_sym_PERCENT,
    ACTIONS(1900), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1908), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1910), 1,
      sym__ternary_qmark,
    STATE(797), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1840), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1878), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1888), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1896), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1904), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1906), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1880), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1902), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1712), 5,
      sym__automatic_semicolon,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_SEMI,
      anon_sym_BQUOTE,
  [27601] = 6,
    ACTIONS(718), 1,
      anon_sym_EQ,
    ACTIONS(1916), 1,
      anon_sym_of,
    ACTIONS(1918), 1,
      anon_sym_in,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(714), 11,
      anon_sym_STAR,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(716), 23,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [27653] = 27,
    ACTIONS(1818), 1,
      anon_sym_LPAREN,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_DOT,
    ACTIONS(1838), 1,
      sym_optional_chain,
    ACTIONS(1882), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1884), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1886), 1,
      anon_sym_GT_GT,
    ACTIONS(1890), 1,
      anon_sym_AMP,
    ACTIONS(1892), 1,
      anon_sym_CARET,
    ACTIONS(1894), 1,
      anon_sym_PIPE,
    ACTIONS(1898), 1,
      anon_sym_PERCENT,
    ACTIONS(1900), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1908), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1910), 1,
      sym__ternary_qmark,
    ACTIONS(1921), 1,
      anon_sym_COMMA,
    ACTIONS(1924), 1,
      anon_sym_RBRACE,
    STATE(797), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1826), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
    ACTIONS(1840), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1878), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1888), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1896), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1904), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1906), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1880), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1902), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [27747] = 27,
    ACTIONS(1818), 1,
      anon_sym_LPAREN,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_DOT,
    ACTIONS(1838), 1,
      sym_optional_chain,
    ACTIONS(1882), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1884), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1886), 1,
      anon_sym_GT_GT,
    ACTIONS(1890), 1,
      anon_sym_AMP,
    ACTIONS(1892), 1,
      anon_sym_CARET,
    ACTIONS(1894), 1,
      anon_sym_PIPE,
    ACTIONS(1898), 1,
      anon_sym_PERCENT,
    ACTIONS(1900), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1908), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1910), 1,
      sym__ternary_qmark,
    ACTIONS(1924), 1,
      anon_sym_RBRACE,
    ACTIONS(1926), 1,
      anon_sym_COMMA,
    STATE(797), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1816), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
    ACTIONS(1840), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1878), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1888), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1896), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1904), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1906), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1880), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1902), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [27841] = 5,
    ACTIONS(1929), 1,
      anon_sym_LPAREN,
    ACTIONS(1932), 1,
      anon_sym_COLON,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1426), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1428), 23,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [27891] = 5,
    ACTIONS(1934), 1,
      anon_sym_LPAREN,
    ACTIONS(1937), 1,
      anon_sym_COLON,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1770), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1772), 23,
      sym__automatic_semicolon,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [27941] = 6,
    ACTIONS(1644), 1,
      anon_sym_EQ,
    ACTIONS(1912), 1,
      anon_sym_of,
    ACTIONS(1939), 1,
      anon_sym_in,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1637), 11,
      anon_sym_STAR,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1639), 23,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [27993] = 27,
    ACTIONS(1818), 1,
      anon_sym_LPAREN,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_DOT,
    ACTIONS(1838), 1,
      sym_optional_chain,
    ACTIONS(1882), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1884), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1886), 1,
      anon_sym_GT_GT,
    ACTIONS(1890), 1,
      anon_sym_AMP,
    ACTIONS(1892), 1,
      anon_sym_CARET,
    ACTIONS(1894), 1,
      anon_sym_PIPE,
    ACTIONS(1898), 1,
      anon_sym_PERCENT,
    ACTIONS(1900), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1908), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1910), 1,
      sym__ternary_qmark,
    ACTIONS(1942), 1,
      anon_sym_COMMA,
    STATE(797), 1,
      sym_arguments,
    STATE(1379), 1,
      aux_sym_sequence_expression_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1840), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1878), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1888), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1896), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1904), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1906), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1944), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
    ACTIONS(1880), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1902), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [28087] = 5,
    ACTIONS(1465), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1547), 3,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
    ACTIONS(1426), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1428), 21,
      sym__ternary_qmark,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [28137] = 6,
    ACTIONS(1729), 1,
      anon_sym_EQ,
    ACTIONS(1914), 1,
      anon_sym_of,
    ACTIONS(1946), 1,
      anon_sym_in,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1722), 11,
      anon_sym_STAR,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1724), 23,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [28189] = 27,
    ACTIONS(1818), 1,
      anon_sym_LPAREN,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_DOT,
    ACTIONS(1838), 1,
      sym_optional_chain,
    ACTIONS(1882), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1884), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1886), 1,
      anon_sym_GT_GT,
    ACTIONS(1890), 1,
      anon_sym_AMP,
    ACTIONS(1892), 1,
      anon_sym_CARET,
    ACTIONS(1894), 1,
      anon_sym_PIPE,
    ACTIONS(1898), 1,
      anon_sym_PERCENT,
    ACTIONS(1900), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1908), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1910), 1,
      sym__ternary_qmark,
    ACTIONS(1949), 1,
      anon_sym_COMMA,
    ACTIONS(1952), 1,
      anon_sym_RBRACE,
    STATE(797), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1826), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
    ACTIONS(1840), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1878), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1888), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1896), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1904), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1906), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1880), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1902), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [28283] = 27,
    ACTIONS(1818), 1,
      anon_sym_LPAREN,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_DOT,
    ACTIONS(1838), 1,
      sym_optional_chain,
    ACTIONS(1882), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1884), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1886), 1,
      anon_sym_GT_GT,
    ACTIONS(1890), 1,
      anon_sym_AMP,
    ACTIONS(1892), 1,
      anon_sym_CARET,
    ACTIONS(1894), 1,
      anon_sym_PIPE,
    ACTIONS(1898), 1,
      anon_sym_PERCENT,
    ACTIONS(1900), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1908), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1910), 1,
      sym__ternary_qmark,
    ACTIONS(1942), 1,
      anon_sym_COMMA,
    STATE(797), 1,
      sym_arguments,
    STATE(1379), 1,
      aux_sym_sequence_expression_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1840), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1878), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1888), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1896), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1904), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1906), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1954), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
    ACTIONS(1880), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1902), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [28377] = 27,
    ACTIONS(1818), 1,
      anon_sym_LPAREN,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_DOT,
    ACTIONS(1838), 1,
      sym_optional_chain,
    ACTIONS(1882), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1884), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1886), 1,
      anon_sym_GT_GT,
    ACTIONS(1890), 1,
      anon_sym_AMP,
    ACTIONS(1892), 1,
      anon_sym_CARET,
    ACTIONS(1894), 1,
      anon_sym_PIPE,
    ACTIONS(1898), 1,
      anon_sym_PERCENT,
    ACTIONS(1900), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1908), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1910), 1,
      sym__ternary_qmark,
    ACTIONS(1942), 1,
      anon_sym_COMMA,
    STATE(797), 1,
      sym_arguments,
    STATE(1379), 1,
      aux_sym_sequence_expression_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1840), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1878), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1888), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1896), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1904), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1906), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1956), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
    ACTIONS(1880), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1902), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [28471] = 25,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(1784), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1786), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1788), 1,
      anon_sym_GT_GT,
    ACTIONS(1792), 1,
      anon_sym_AMP,
    ACTIONS(1794), 1,
      anon_sym_CARET,
    ACTIONS(1796), 1,
      anon_sym_PIPE,
    ACTIONS(1800), 1,
      anon_sym_PERCENT,
    ACTIONS(1802), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1810), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1812), 1,
      sym__ternary_qmark,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1778), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1790), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1798), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1806), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1808), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1782), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1804), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1958), 4,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
  [28561] = 25,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(1784), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1786), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1788), 1,
      anon_sym_GT_GT,
    ACTIONS(1792), 1,
      anon_sym_AMP,
    ACTIONS(1794), 1,
      anon_sym_CARET,
    ACTIONS(1796), 1,
      anon_sym_PIPE,
    ACTIONS(1800), 1,
      anon_sym_PERCENT,
    ACTIONS(1802), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1810), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1812), 1,
      sym__ternary_qmark,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1778), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1790), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1798), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1806), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1808), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1782), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1804), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1960), 4,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
  [28651] = 26,
    ACTIONS(1818), 1,
      anon_sym_LPAREN,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_DOT,
    ACTIONS(1838), 1,
      sym_optional_chain,
    ACTIONS(1846), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1848), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1850), 1,
      anon_sym_GT_GT,
    ACTIONS(1854), 1,
      anon_sym_AMP,
    ACTIONS(1856), 1,
      anon_sym_CARET,
    ACTIONS(1858), 1,
      anon_sym_PIPE,
    ACTIONS(1862), 1,
      anon_sym_PERCENT,
    ACTIONS(1864), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1872), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1874), 1,
      sym__ternary_qmark,
    ACTIONS(1964), 1,
      anon_sym_in,
    STATE(797), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1840), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1842), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1844), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1852), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1860), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1868), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1870), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1866), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1962), 4,
      sym__automatic_semicolon,
      anon_sym_COMMA,
      anon_sym_of,
      anon_sym_SEMI,
  [28743] = 6,
    ACTIONS(1465), 1,
      anon_sym_EQ,
    ACTIONS(1550), 1,
      anon_sym_of,
    ACTIONS(1552), 1,
      anon_sym_in,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1426), 11,
      anon_sym_STAR,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1428), 23,
      sym__ternary_qmark,
      anon_sym_COMMA,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [28795] = 24,
    ACTIONS(93), 1,
      anon_sym_AT,
    ACTIONS(101), 1,
      anon_sym_STAR,
    ACTIONS(103), 1,
      anon_sym_COMMA,
    ACTIONS(115), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(123), 1,
      aux_sym_method_definition_token1,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(1969), 1,
      anon_sym_LBRACE,
    ACTIONS(1971), 1,
      anon_sym_RBRACE,
    ACTIONS(1973), 1,
      anon_sym_LBRACK,
    ACTIONS(1975), 1,
      anon_sym_async,
    ACTIONS(1979), 1,
      anon_sym_static,
    STATE(1019), 1,
      aux_sym_export_statement_repeat1,
    STATE(1097), 1,
      sym_decorator,
    STATE(1476), 1,
      aux_sym_object_pattern_repeat1,
    STATE(1499), 1,
      aux_sym_object_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1977), 2,
      sym_number,
      sym_private_property_identifier,
    ACTIONS(1981), 2,
      anon_sym_get,
      anon_sym_set,
    STATE(1406), 3,
      sym_spread_element,
      sym_method_definition,
      sym_pair,
    STATE(1407), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    STATE(1454), 3,
      sym_object_assignment_pattern,
      sym_rest_pattern,
      sym_pair_pattern,
    STATE(1788), 3,
      sym_object_pattern,
      sym_array_pattern,
      sym__destructuring_pattern,
    ACTIONS(1967), 4,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      sym_identifier,
  [28882] = 27,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(1784), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1786), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1788), 1,
      anon_sym_GT_GT,
    ACTIONS(1792), 1,
      anon_sym_AMP,
    ACTIONS(1794), 1,
      anon_sym_CARET,
    ACTIONS(1796), 1,
      anon_sym_PIPE,
    ACTIONS(1800), 1,
      anon_sym_PERCENT,
    ACTIONS(1802), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1810), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1812), 1,
      sym__ternary_qmark,
    ACTIONS(1983), 1,
      anon_sym_COMMA,
    ACTIONS(1985), 1,
      anon_sym_RPAREN,
    STATE(637), 1,
      sym_arguments,
    STATE(1224), 1,
      aux_sym_sequence_expression_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1778), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1790), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1798), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1806), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1808), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1782), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1804), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [28975] = 27,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(1784), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1786), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1788), 1,
      anon_sym_GT_GT,
    ACTIONS(1792), 1,
      anon_sym_AMP,
    ACTIONS(1794), 1,
      anon_sym_CARET,
    ACTIONS(1796), 1,
      anon_sym_PIPE,
    ACTIONS(1800), 1,
      anon_sym_PERCENT,
    ACTIONS(1802), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1810), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1812), 1,
      sym__ternary_qmark,
    ACTIONS(1983), 1,
      anon_sym_COMMA,
    ACTIONS(1987), 1,
      anon_sym_RPAREN,
    STATE(637), 1,
      sym_arguments,
    STATE(1224), 1,
      aux_sym_sequence_expression_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1778), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1790), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1798), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1806), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1808), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1782), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1804), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [29068] = 27,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(1784), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1786), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1788), 1,
      anon_sym_GT_GT,
    ACTIONS(1792), 1,
      anon_sym_AMP,
    ACTIONS(1794), 1,
      anon_sym_CARET,
    ACTIONS(1796), 1,
      anon_sym_PIPE,
    ACTIONS(1800), 1,
      anon_sym_PERCENT,
    ACTIONS(1802), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1810), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1812), 1,
      sym__ternary_qmark,
    ACTIONS(1983), 1,
      anon_sym_COMMA,
    ACTIONS(1989), 1,
      anon_sym_RPAREN,
    STATE(637), 1,
      sym_arguments,
    STATE(1224), 1,
      aux_sym_sequence_expression_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1778), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1790), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1798), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1806), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1808), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1782), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1804), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [29161] = 27,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(1784), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1786), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1788), 1,
      anon_sym_GT_GT,
    ACTIONS(1792), 1,
      anon_sym_AMP,
    ACTIONS(1794), 1,
      anon_sym_CARET,
    ACTIONS(1796), 1,
      anon_sym_PIPE,
    ACTIONS(1800), 1,
      anon_sym_PERCENT,
    ACTIONS(1802), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1810), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1812), 1,
      sym__ternary_qmark,
    ACTIONS(1983), 1,
      anon_sym_COMMA,
    ACTIONS(1991), 1,
      anon_sym_RPAREN,
    STATE(637), 1,
      sym_arguments,
    STATE(1224), 1,
      aux_sym_sequence_expression_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1778), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1790), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1798), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1806), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1808), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1782), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1804), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [29254] = 27,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(1784), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1786), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1788), 1,
      anon_sym_GT_GT,
    ACTIONS(1792), 1,
      anon_sym_AMP,
    ACTIONS(1794), 1,
      anon_sym_CARET,
    ACTIONS(1796), 1,
      anon_sym_PIPE,
    ACTIONS(1800), 1,
      anon_sym_PERCENT,
    ACTIONS(1802), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1810), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1812), 1,
      sym__ternary_qmark,
    ACTIONS(1983), 1,
      anon_sym_COMMA,
    ACTIONS(1993), 1,
      anon_sym_RPAREN,
    STATE(637), 1,
      sym_arguments,
    STATE(1224), 1,
      aux_sym_sequence_expression_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1778), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1790), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1798), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1806), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1808), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1782), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1804), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [29347] = 27,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(1784), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1786), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1788), 1,
      anon_sym_GT_GT,
    ACTIONS(1792), 1,
      anon_sym_AMP,
    ACTIONS(1794), 1,
      anon_sym_CARET,
    ACTIONS(1796), 1,
      anon_sym_PIPE,
    ACTIONS(1800), 1,
      anon_sym_PERCENT,
    ACTIONS(1802), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1810), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1812), 1,
      sym__ternary_qmark,
    ACTIONS(1983), 1,
      anon_sym_COMMA,
    ACTIONS(1995), 1,
      anon_sym_RPAREN,
    STATE(637), 1,
      sym_arguments,
    STATE(1224), 1,
      aux_sym_sequence_expression_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1778), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1790), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1798), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1806), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1808), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1782), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1804), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [29440] = 27,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(1784), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1786), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1788), 1,
      anon_sym_GT_GT,
    ACTIONS(1792), 1,
      anon_sym_AMP,
    ACTIONS(1794), 1,
      anon_sym_CARET,
    ACTIONS(1796), 1,
      anon_sym_PIPE,
    ACTIONS(1800), 1,
      anon_sym_PERCENT,
    ACTIONS(1802), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1810), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1812), 1,
      sym__ternary_qmark,
    ACTIONS(1983), 1,
      anon_sym_COMMA,
    ACTIONS(1997), 1,
      anon_sym_RPAREN,
    STATE(637), 1,
      sym_arguments,
    STATE(1224), 1,
      aux_sym_sequence_expression_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1778), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1790), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1798), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1806), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1808), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1782), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1804), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [29533] = 27,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(1784), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1786), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1788), 1,
      anon_sym_GT_GT,
    ACTIONS(1792), 1,
      anon_sym_AMP,
    ACTIONS(1794), 1,
      anon_sym_CARET,
    ACTIONS(1796), 1,
      anon_sym_PIPE,
    ACTIONS(1800), 1,
      anon_sym_PERCENT,
    ACTIONS(1802), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1810), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1812), 1,
      sym__ternary_qmark,
    ACTIONS(1983), 1,
      anon_sym_COMMA,
    ACTIONS(1999), 1,
      anon_sym_RPAREN,
    STATE(637), 1,
      sym_arguments,
    STATE(1224), 1,
      aux_sym_sequence_expression_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1778), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1790), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1798), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1806), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1808), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1782), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1804), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [29626] = 27,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(1784), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1786), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1788), 1,
      anon_sym_GT_GT,
    ACTIONS(1792), 1,
      anon_sym_AMP,
    ACTIONS(1794), 1,
      anon_sym_CARET,
    ACTIONS(1796), 1,
      anon_sym_PIPE,
    ACTIONS(1800), 1,
      anon_sym_PERCENT,
    ACTIONS(1802), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1810), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1812), 1,
      sym__ternary_qmark,
    ACTIONS(1983), 1,
      anon_sym_COMMA,
    ACTIONS(2001), 1,
      anon_sym_RPAREN,
    STATE(637), 1,
      sym_arguments,
    STATE(1224), 1,
      aux_sym_sequence_expression_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1778), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1790), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1798), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1806), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1808), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1782), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1804), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [29719] = 27,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(1784), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1786), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1788), 1,
      anon_sym_GT_GT,
    ACTIONS(1792), 1,
      anon_sym_AMP,
    ACTIONS(1794), 1,
      anon_sym_CARET,
    ACTIONS(1796), 1,
      anon_sym_PIPE,
    ACTIONS(1800), 1,
      anon_sym_PERCENT,
    ACTIONS(1802), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1810), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1812), 1,
      sym__ternary_qmark,
    ACTIONS(1983), 1,
      anon_sym_COMMA,
    ACTIONS(2003), 1,
      anon_sym_RPAREN,
    STATE(637), 1,
      sym_arguments,
    STATE(1224), 1,
      aux_sym_sequence_expression_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1778), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1790), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1798), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1806), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1808), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1782), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1804), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [29812] = 27,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(1784), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1786), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1788), 1,
      anon_sym_GT_GT,
    ACTIONS(1792), 1,
      anon_sym_AMP,
    ACTIONS(1794), 1,
      anon_sym_CARET,
    ACTIONS(1796), 1,
      anon_sym_PIPE,
    ACTIONS(1800), 1,
      anon_sym_PERCENT,
    ACTIONS(1802), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1810), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1812), 1,
      sym__ternary_qmark,
    ACTIONS(1983), 1,
      anon_sym_COMMA,
    ACTIONS(2005), 1,
      anon_sym_RPAREN,
    STATE(637), 1,
      sym_arguments,
    STATE(1224), 1,
      aux_sym_sequence_expression_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1778), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1790), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1798), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1806), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1808), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1782), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1804), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [29905] = 27,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(1784), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1786), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1788), 1,
      anon_sym_GT_GT,
    ACTIONS(1792), 1,
      anon_sym_AMP,
    ACTIONS(1794), 1,
      anon_sym_CARET,
    ACTIONS(1796), 1,
      anon_sym_PIPE,
    ACTIONS(1800), 1,
      anon_sym_PERCENT,
    ACTIONS(1802), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1810), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1812), 1,
      sym__ternary_qmark,
    ACTIONS(1983), 1,
      anon_sym_COMMA,
    ACTIONS(2007), 1,
      anon_sym_RPAREN,
    STATE(637), 1,
      sym_arguments,
    STATE(1224), 1,
      aux_sym_sequence_expression_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1778), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1790), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1798), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1806), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1808), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1782), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1804), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [29998] = 27,
    ACTIONS(910), 1,
      anon_sym_COMMA,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(1784), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1786), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1788), 1,
      anon_sym_GT_GT,
    ACTIONS(1792), 1,
      anon_sym_AMP,
    ACTIONS(1794), 1,
      anon_sym_CARET,
    ACTIONS(1796), 1,
      anon_sym_PIPE,
    ACTIONS(1800), 1,
      anon_sym_PERCENT,
    ACTIONS(1802), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1810), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1812), 1,
      sym__ternary_qmark,
    ACTIONS(2009), 1,
      anon_sym_RBRACK,
    STATE(637), 1,
      sym_arguments,
    STATE(1483), 1,
      aux_sym_array_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1778), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1790), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1798), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1806), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1808), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1782), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1804), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [30091] = 27,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(1784), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1786), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1788), 1,
      anon_sym_GT_GT,
    ACTIONS(1792), 1,
      anon_sym_AMP,
    ACTIONS(1794), 1,
      anon_sym_CARET,
    ACTIONS(1796), 1,
      anon_sym_PIPE,
    ACTIONS(1800), 1,
      anon_sym_PERCENT,
    ACTIONS(1802), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1810), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1812), 1,
      sym__ternary_qmark,
    ACTIONS(1983), 1,
      anon_sym_COMMA,
    ACTIONS(2011), 1,
      anon_sym_SEMI,
    STATE(637), 1,
      sym_arguments,
    STATE(1224), 1,
      aux_sym_sequence_expression_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1778), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1790), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1798), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1806), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1808), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1782), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1804), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [30184] = 27,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(1784), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1786), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1788), 1,
      anon_sym_GT_GT,
    ACTIONS(1792), 1,
      anon_sym_AMP,
    ACTIONS(1794), 1,
      anon_sym_CARET,
    ACTIONS(1796), 1,
      anon_sym_PIPE,
    ACTIONS(1800), 1,
      anon_sym_PERCENT,
    ACTIONS(1802), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1810), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1812), 1,
      sym__ternary_qmark,
    ACTIONS(1983), 1,
      anon_sym_COMMA,
    ACTIONS(2013), 1,
      anon_sym_RPAREN,
    STATE(637), 1,
      sym_arguments,
    STATE(1224), 1,
      aux_sym_sequence_expression_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1778), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1790), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1798), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1806), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1808), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1782), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1804), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [30277] = 27,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(1784), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1786), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1788), 1,
      anon_sym_GT_GT,
    ACTIONS(1792), 1,
      anon_sym_AMP,
    ACTIONS(1794), 1,
      anon_sym_CARET,
    ACTIONS(1796), 1,
      anon_sym_PIPE,
    ACTIONS(1800), 1,
      anon_sym_PERCENT,
    ACTIONS(1802), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1810), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1812), 1,
      sym__ternary_qmark,
    ACTIONS(1983), 1,
      anon_sym_COMMA,
    ACTIONS(2015), 1,
      anon_sym_RPAREN,
    STATE(637), 1,
      sym_arguments,
    STATE(1224), 1,
      aux_sym_sequence_expression_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1778), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1790), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1798), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1806), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1808), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1782), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1804), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [30370] = 25,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(1784), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1786), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1788), 1,
      anon_sym_GT_GT,
    ACTIONS(1792), 1,
      anon_sym_AMP,
    ACTIONS(1794), 1,
      anon_sym_CARET,
    ACTIONS(1796), 1,
      anon_sym_PIPE,
    ACTIONS(1800), 1,
      anon_sym_PERCENT,
    ACTIONS(1802), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1810), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1812), 1,
      sym__ternary_qmark,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1778), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1790), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1798), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1806), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1808), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1782), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1804), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(2017), 3,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
  [30459] = 27,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(1784), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1786), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1788), 1,
      anon_sym_GT_GT,
    ACTIONS(1792), 1,
      anon_sym_AMP,
    ACTIONS(1794), 1,
      anon_sym_CARET,
    ACTIONS(1796), 1,
      anon_sym_PIPE,
    ACTIONS(1800), 1,
      anon_sym_PERCENT,
    ACTIONS(1802), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1810), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1812), 1,
      sym__ternary_qmark,
    ACTIONS(1983), 1,
      anon_sym_COMMA,
    ACTIONS(2019), 1,
      anon_sym_RBRACE,
    STATE(637), 1,
      sym_arguments,
    STATE(1224), 1,
      aux_sym_sequence_expression_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1778), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1790), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1798), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1806), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1808), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1782), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1804), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [30552] = 27,
    ACTIONS(910), 1,
      anon_sym_COMMA,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(1784), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1786), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1788), 1,
      anon_sym_GT_GT,
    ACTIONS(1792), 1,
      anon_sym_AMP,
    ACTIONS(1794), 1,
      anon_sym_CARET,
    ACTIONS(1796), 1,
      anon_sym_PIPE,
    ACTIONS(1800), 1,
      anon_sym_PERCENT,
    ACTIONS(1802), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1810), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1812), 1,
      sym__ternary_qmark,
    ACTIONS(2021), 1,
      anon_sym_RPAREN,
    STATE(637), 1,
      sym_arguments,
    STATE(1465), 1,
      aux_sym_array_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1778), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1790), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1798), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1806), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1808), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1782), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1804), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [30645] = 25,
    ACTIONS(1818), 1,
      anon_sym_LPAREN,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_DOT,
    ACTIONS(1838), 1,
      sym_optional_chain,
    ACTIONS(1882), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1884), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1886), 1,
      anon_sym_GT_GT,
    ACTIONS(1890), 1,
      anon_sym_AMP,
    ACTIONS(1892), 1,
      anon_sym_CARET,
    ACTIONS(1894), 1,
      anon_sym_PIPE,
    ACTIONS(1898), 1,
      anon_sym_PERCENT,
    ACTIONS(1900), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1908), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1910), 1,
      sym__ternary_qmark,
    STATE(797), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1840), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1878), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1888), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1896), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1904), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1906), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1836), 3,
      sym__automatic_semicolon,
      anon_sym_COMMA,
      anon_sym_SEMI,
    ACTIONS(1880), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1902), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [30734] = 27,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(1784), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1786), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1788), 1,
      anon_sym_GT_GT,
    ACTIONS(1792), 1,
      anon_sym_AMP,
    ACTIONS(1794), 1,
      anon_sym_CARET,
    ACTIONS(1796), 1,
      anon_sym_PIPE,
    ACTIONS(1800), 1,
      anon_sym_PERCENT,
    ACTIONS(1802), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1810), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1812), 1,
      sym__ternary_qmark,
    ACTIONS(1983), 1,
      anon_sym_COMMA,
    ACTIONS(2023), 1,
      anon_sym_RBRACK,
    STATE(637), 1,
      sym_arguments,
    STATE(1224), 1,
      aux_sym_sequence_expression_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1778), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1790), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1798), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1806), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1808), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1782), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1804), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [30827] = 27,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(1784), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1786), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1788), 1,
      anon_sym_GT_GT,
    ACTIONS(1792), 1,
      anon_sym_AMP,
    ACTIONS(1794), 1,
      anon_sym_CARET,
    ACTIONS(1796), 1,
      anon_sym_PIPE,
    ACTIONS(1800), 1,
      anon_sym_PERCENT,
    ACTIONS(1802), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1810), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1812), 1,
      sym__ternary_qmark,
    ACTIONS(1983), 1,
      anon_sym_COMMA,
    ACTIONS(2025), 1,
      anon_sym_RBRACE,
    STATE(637), 1,
      sym_arguments,
    STATE(1224), 1,
      aux_sym_sequence_expression_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1778), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1790), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1798), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1806), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1808), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1782), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1804), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [30920] = 6,
    ACTIONS(1530), 1,
      anon_sym_COMMA,
    ACTIONS(1537), 1,
      anon_sym_RBRACK,
    ACTIONS(1540), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1426), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1428), 21,
      sym__ternary_qmark,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [30971] = 6,
    ACTIONS(1641), 1,
      anon_sym_RBRACK,
    ACTIONS(1644), 1,
      anon_sym_EQ,
    ACTIONS(1912), 1,
      anon_sym_COMMA,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1637), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1639), 21,
      sym__ternary_qmark,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [31022] = 6,
    ACTIONS(1726), 1,
      anon_sym_RBRACK,
    ACTIONS(1729), 1,
      anon_sym_EQ,
    ACTIONS(1914), 1,
      anon_sym_COMMA,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1722), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1724), 21,
      sym__ternary_qmark,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [31073] = 6,
    ACTIONS(718), 1,
      anon_sym_EQ,
    ACTIONS(1767), 1,
      anon_sym_RBRACK,
    ACTIONS(1916), 1,
      anon_sym_COMMA,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(714), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(716), 21,
      sym__ternary_qmark,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [31124] = 25,
    ACTIONS(1818), 1,
      anon_sym_LPAREN,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_DOT,
    ACTIONS(1838), 1,
      sym_optional_chain,
    ACTIONS(1882), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1884), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1886), 1,
      anon_sym_GT_GT,
    ACTIONS(1890), 1,
      anon_sym_AMP,
    ACTIONS(1892), 1,
      anon_sym_CARET,
    ACTIONS(1894), 1,
      anon_sym_PIPE,
    ACTIONS(1898), 1,
      anon_sym_PERCENT,
    ACTIONS(1900), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1908), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1910), 1,
      sym__ternary_qmark,
    STATE(797), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1840), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1878), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1888), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1896), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1904), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1906), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1880), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1902), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1962), 3,
      sym__automatic_semicolon,
      anon_sym_COMMA,
      anon_sym_SEMI,
  [31213] = 24,
    ACTIONS(93), 1,
      anon_sym_AT,
    ACTIONS(101), 1,
      anon_sym_STAR,
    ACTIONS(103), 1,
      anon_sym_COMMA,
    ACTIONS(115), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(123), 1,
      aux_sym_method_definition_token1,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(1969), 1,
      anon_sym_LBRACE,
    ACTIONS(1973), 1,
      anon_sym_LBRACK,
    ACTIONS(2029), 1,
      anon_sym_RBRACE,
    ACTIONS(2031), 1,
      anon_sym_async,
    ACTIONS(2033), 1,
      anon_sym_static,
    STATE(1019), 1,
      aux_sym_export_statement_repeat1,
    STATE(1097), 1,
      sym_decorator,
    STATE(1461), 1,
      aux_sym_object_repeat1,
    STATE(1476), 1,
      aux_sym_object_pattern_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1977), 2,
      sym_number,
      sym_private_property_identifier,
    ACTIONS(2035), 2,
      anon_sym_get,
      anon_sym_set,
    STATE(1407), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    STATE(1454), 3,
      sym_object_assignment_pattern,
      sym_rest_pattern,
      sym_pair_pattern,
    STATE(1458), 3,
      sym_spread_element,
      sym_method_definition,
      sym_pair,
    STATE(1788), 3,
      sym_object_pattern,
      sym_array_pattern,
      sym__destructuring_pattern,
    ACTIONS(2027), 4,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      sym_identifier,
  [31300] = 27,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(1784), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1786), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1788), 1,
      anon_sym_GT_GT,
    ACTIONS(1792), 1,
      anon_sym_AMP,
    ACTIONS(1794), 1,
      anon_sym_CARET,
    ACTIONS(1796), 1,
      anon_sym_PIPE,
    ACTIONS(1800), 1,
      anon_sym_PERCENT,
    ACTIONS(1802), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1810), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1812), 1,
      sym__ternary_qmark,
    ACTIONS(1983), 1,
      anon_sym_COMMA,
    ACTIONS(2037), 1,
      anon_sym_RPAREN,
    STATE(637), 1,
      sym_arguments,
    STATE(1224), 1,
      aux_sym_sequence_expression_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1778), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1790), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1798), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1806), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1808), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1782), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1804), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [31393] = 27,
    ACTIONS(910), 1,
      anon_sym_COMMA,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(1784), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1786), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1788), 1,
      anon_sym_GT_GT,
    ACTIONS(1792), 1,
      anon_sym_AMP,
    ACTIONS(1794), 1,
      anon_sym_CARET,
    ACTIONS(1796), 1,
      anon_sym_PIPE,
    ACTIONS(1800), 1,
      anon_sym_PERCENT,
    ACTIONS(1802), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1810), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1812), 1,
      sym__ternary_qmark,
    ACTIONS(2039), 1,
      anon_sym_RBRACK,
    STATE(637), 1,
      sym_arguments,
    STATE(1429), 1,
      aux_sym_array_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1778), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1790), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1798), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1806), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1808), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1782), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1804), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [31486] = 27,
    ACTIONS(910), 1,
      anon_sym_COMMA,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(1784), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1786), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1788), 1,
      anon_sym_GT_GT,
    ACTIONS(1792), 1,
      anon_sym_AMP,
    ACTIONS(1794), 1,
      anon_sym_CARET,
    ACTIONS(1796), 1,
      anon_sym_PIPE,
    ACTIONS(1800), 1,
      anon_sym_PERCENT,
    ACTIONS(1802), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1810), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1812), 1,
      sym__ternary_qmark,
    ACTIONS(2041), 1,
      anon_sym_RPAREN,
    STATE(637), 1,
      sym_arguments,
    STATE(1455), 1,
      aux_sym_array_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1778), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1790), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1798), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1806), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1808), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1782), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1804), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [31579] = 27,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(1784), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1786), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1788), 1,
      anon_sym_GT_GT,
    ACTIONS(1792), 1,
      anon_sym_AMP,
    ACTIONS(1794), 1,
      anon_sym_CARET,
    ACTIONS(1796), 1,
      anon_sym_PIPE,
    ACTIONS(1800), 1,
      anon_sym_PERCENT,
    ACTIONS(1802), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1810), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1812), 1,
      sym__ternary_qmark,
    ACTIONS(1983), 1,
      anon_sym_COMMA,
    ACTIONS(2043), 1,
      anon_sym_RBRACK,
    STATE(637), 1,
      sym_arguments,
    STATE(1224), 1,
      aux_sym_sequence_expression_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1778), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1790), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1798), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1806), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1808), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1782), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1804), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [31672] = 27,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(1784), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1786), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1788), 1,
      anon_sym_GT_GT,
    ACTIONS(1792), 1,
      anon_sym_AMP,
    ACTIONS(1794), 1,
      anon_sym_CARET,
    ACTIONS(1796), 1,
      anon_sym_PIPE,
    ACTIONS(1800), 1,
      anon_sym_PERCENT,
    ACTIONS(1802), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1810), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1812), 1,
      sym__ternary_qmark,
    ACTIONS(1983), 1,
      anon_sym_COMMA,
    ACTIONS(2045), 1,
      anon_sym_RBRACE,
    STATE(637), 1,
      sym_arguments,
    STATE(1224), 1,
      aux_sym_sequence_expression_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1778), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1790), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1798), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1806), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1808), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1782), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1804), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [31765] = 27,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(1784), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1786), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1788), 1,
      anon_sym_GT_GT,
    ACTIONS(1792), 1,
      anon_sym_AMP,
    ACTIONS(1794), 1,
      anon_sym_CARET,
    ACTIONS(1796), 1,
      anon_sym_PIPE,
    ACTIONS(1800), 1,
      anon_sym_PERCENT,
    ACTIONS(1802), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1810), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1812), 1,
      sym__ternary_qmark,
    ACTIONS(1983), 1,
      anon_sym_COMMA,
    ACTIONS(2047), 1,
      anon_sym_RBRACK,
    STATE(637), 1,
      sym_arguments,
    STATE(1224), 1,
      aux_sym_sequence_expression_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1778), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1790), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1798), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1806), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1808), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1782), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1804), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [31858] = 27,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(1784), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1786), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1788), 1,
      anon_sym_GT_GT,
    ACTIONS(1792), 1,
      anon_sym_AMP,
    ACTIONS(1794), 1,
      anon_sym_CARET,
    ACTIONS(1796), 1,
      anon_sym_PIPE,
    ACTIONS(1800), 1,
      anon_sym_PERCENT,
    ACTIONS(1802), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1810), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1812), 1,
      sym__ternary_qmark,
    ACTIONS(1983), 1,
      anon_sym_COMMA,
    ACTIONS(2049), 1,
      anon_sym_SEMI,
    STATE(637), 1,
      sym_arguments,
    STATE(1224), 1,
      aux_sym_sequence_expression_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1778), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1790), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1798), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1806), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1808), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1782), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1804), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [31951] = 27,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(1784), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1786), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1788), 1,
      anon_sym_GT_GT,
    ACTIONS(1792), 1,
      anon_sym_AMP,
    ACTIONS(1794), 1,
      anon_sym_CARET,
    ACTIONS(1796), 1,
      anon_sym_PIPE,
    ACTIONS(1800), 1,
      anon_sym_PERCENT,
    ACTIONS(1802), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1810), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1812), 1,
      sym__ternary_qmark,
    ACTIONS(1983), 1,
      anon_sym_COMMA,
    ACTIONS(2051), 1,
      anon_sym_SEMI,
    STATE(637), 1,
      sym_arguments,
    STATE(1224), 1,
      aux_sym_sequence_expression_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1778), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1790), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1798), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1806), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1808), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1782), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1804), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [32044] = 24,
    ACTIONS(93), 1,
      anon_sym_AT,
    ACTIONS(101), 1,
      anon_sym_STAR,
    ACTIONS(103), 1,
      anon_sym_COMMA,
    ACTIONS(115), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(123), 1,
      aux_sym_method_definition_token1,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(1969), 1,
      anon_sym_LBRACE,
    ACTIONS(1973), 1,
      anon_sym_LBRACK,
    ACTIONS(2055), 1,
      anon_sym_RBRACE,
    ACTIONS(2057), 1,
      anon_sym_async,
    ACTIONS(2059), 1,
      anon_sym_static,
    STATE(1019), 1,
      aux_sym_export_statement_repeat1,
    STATE(1097), 1,
      sym_decorator,
    STATE(1461), 1,
      aux_sym_object_repeat1,
    STATE(1476), 1,
      aux_sym_object_pattern_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1977), 2,
      sym_number,
      sym_private_property_identifier,
    ACTIONS(2061), 2,
      anon_sym_get,
      anon_sym_set,
    STATE(1407), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    STATE(1454), 3,
      sym_object_assignment_pattern,
      sym_rest_pattern,
      sym_pair_pattern,
    STATE(1458), 3,
      sym_spread_element,
      sym_method_definition,
      sym_pair,
    STATE(1788), 3,
      sym_object_pattern,
      sym_array_pattern,
      sym__destructuring_pattern,
    ACTIONS(2053), 4,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      sym_identifier,
  [32131] = 27,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(1784), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1786), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1788), 1,
      anon_sym_GT_GT,
    ACTIONS(1792), 1,
      anon_sym_AMP,
    ACTIONS(1794), 1,
      anon_sym_CARET,
    ACTIONS(1796), 1,
      anon_sym_PIPE,
    ACTIONS(1800), 1,
      anon_sym_PERCENT,
    ACTIONS(1802), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1810), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1812), 1,
      sym__ternary_qmark,
    ACTIONS(1983), 1,
      anon_sym_COMMA,
    ACTIONS(2063), 1,
      anon_sym_RPAREN,
    STATE(637), 1,
      sym_arguments,
    STATE(1224), 1,
      aux_sym_sequence_expression_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1778), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1790), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1798), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1806), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1808), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1782), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1804), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [32224] = 27,
    ACTIONS(910), 1,
      anon_sym_COMMA,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(1784), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1786), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1788), 1,
      anon_sym_GT_GT,
    ACTIONS(1792), 1,
      anon_sym_AMP,
    ACTIONS(1794), 1,
      anon_sym_CARET,
    ACTIONS(1796), 1,
      anon_sym_PIPE,
    ACTIONS(1800), 1,
      anon_sym_PERCENT,
    ACTIONS(1802), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1810), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1812), 1,
      sym__ternary_qmark,
    ACTIONS(2065), 1,
      anon_sym_RPAREN,
    STATE(637), 1,
      sym_arguments,
    STATE(1436), 1,
      aux_sym_array_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1778), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1790), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1798), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1806), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1808), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1782), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1804), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [32317] = 24,
    ACTIONS(93), 1,
      anon_sym_AT,
    ACTIONS(101), 1,
      anon_sym_STAR,
    ACTIONS(103), 1,
      anon_sym_COMMA,
    ACTIONS(115), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(123), 1,
      aux_sym_method_definition_token1,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(1969), 1,
      anon_sym_LBRACE,
    ACTIONS(1973), 1,
      anon_sym_LBRACK,
    ACTIONS(2069), 1,
      anon_sym_RBRACE,
    ACTIONS(2071), 1,
      anon_sym_async,
    ACTIONS(2073), 1,
      anon_sym_static,
    STATE(1019), 1,
      aux_sym_export_statement_repeat1,
    STATE(1097), 1,
      sym_decorator,
    STATE(1461), 1,
      aux_sym_object_repeat1,
    STATE(1476), 1,
      aux_sym_object_pattern_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1977), 2,
      sym_number,
      sym_private_property_identifier,
    ACTIONS(2075), 2,
      anon_sym_get,
      anon_sym_set,
    STATE(1407), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    STATE(1454), 3,
      sym_object_assignment_pattern,
      sym_rest_pattern,
      sym_pair_pattern,
    STATE(1458), 3,
      sym_spread_element,
      sym_method_definition,
      sym_pair,
    STATE(1788), 3,
      sym_object_pattern,
      sym_array_pattern,
      sym__destructuring_pattern,
    ACTIONS(2067), 4,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      sym_identifier,
  [32404] = 27,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(1784), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1786), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1788), 1,
      anon_sym_GT_GT,
    ACTIONS(1792), 1,
      anon_sym_AMP,
    ACTIONS(1794), 1,
      anon_sym_CARET,
    ACTIONS(1796), 1,
      anon_sym_PIPE,
    ACTIONS(1800), 1,
      anon_sym_PERCENT,
    ACTIONS(1802), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1810), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1812), 1,
      sym__ternary_qmark,
    ACTIONS(1983), 1,
      anon_sym_COMMA,
    ACTIONS(2077), 1,
      anon_sym_RPAREN,
    STATE(637), 1,
      sym_arguments,
    STATE(1224), 1,
      aux_sym_sequence_expression_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1778), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1790), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1798), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1806), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1808), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1782), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1804), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [32497] = 24,
    ACTIONS(93), 1,
      anon_sym_AT,
    ACTIONS(101), 1,
      anon_sym_STAR,
    ACTIONS(103), 1,
      anon_sym_COMMA,
    ACTIONS(115), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(123), 1,
      aux_sym_method_definition_token1,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(1969), 1,
      anon_sym_LBRACE,
    ACTIONS(1973), 1,
      anon_sym_LBRACK,
    ACTIONS(2081), 1,
      anon_sym_RBRACE,
    ACTIONS(2083), 1,
      anon_sym_async,
    ACTIONS(2085), 1,
      anon_sym_static,
    STATE(1019), 1,
      aux_sym_export_statement_repeat1,
    STATE(1097), 1,
      sym_decorator,
    STATE(1461), 1,
      aux_sym_object_repeat1,
    STATE(1476), 1,
      aux_sym_object_pattern_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1977), 2,
      sym_number,
      sym_private_property_identifier,
    ACTIONS(2087), 2,
      anon_sym_get,
      anon_sym_set,
    STATE(1407), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    STATE(1454), 3,
      sym_object_assignment_pattern,
      sym_rest_pattern,
      sym_pair_pattern,
    STATE(1458), 3,
      sym_spread_element,
      sym_method_definition,
      sym_pair,
    STATE(1788), 3,
      sym_object_pattern,
      sym_array_pattern,
      sym__destructuring_pattern,
    ACTIONS(2079), 4,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      sym_identifier,
  [32584] = 24,
    ACTIONS(93), 1,
      anon_sym_AT,
    ACTIONS(101), 1,
      anon_sym_STAR,
    ACTIONS(103), 1,
      anon_sym_COMMA,
    ACTIONS(115), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(123), 1,
      aux_sym_method_definition_token1,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(1969), 1,
      anon_sym_LBRACE,
    ACTIONS(1973), 1,
      anon_sym_LBRACK,
    ACTIONS(2091), 1,
      anon_sym_RBRACE,
    ACTIONS(2093), 1,
      anon_sym_async,
    ACTIONS(2095), 1,
      anon_sym_static,
    STATE(1019), 1,
      aux_sym_export_statement_repeat1,
    STATE(1097), 1,
      sym_decorator,
    STATE(1461), 1,
      aux_sym_object_repeat1,
    STATE(1476), 1,
      aux_sym_object_pattern_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1977), 2,
      sym_number,
      sym_private_property_identifier,
    ACTIONS(2097), 2,
      anon_sym_get,
      anon_sym_set,
    STATE(1407), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    STATE(1454), 3,
      sym_object_assignment_pattern,
      sym_rest_pattern,
      sym_pair_pattern,
    STATE(1458), 3,
      sym_spread_element,
      sym_method_definition,
      sym_pair,
    STATE(1788), 3,
      sym_object_pattern,
      sym_array_pattern,
      sym__destructuring_pattern,
    ACTIONS(2089), 4,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      sym_identifier,
  [32671] = 27,
    ACTIONS(910), 1,
      anon_sym_COMMA,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(1784), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1786), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1788), 1,
      anon_sym_GT_GT,
    ACTIONS(1792), 1,
      anon_sym_AMP,
    ACTIONS(1794), 1,
      anon_sym_CARET,
    ACTIONS(1796), 1,
      anon_sym_PIPE,
    ACTIONS(1800), 1,
      anon_sym_PERCENT,
    ACTIONS(1802), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1810), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1812), 1,
      sym__ternary_qmark,
    ACTIONS(2099), 1,
      anon_sym_RBRACK,
    STATE(637), 1,
      sym_arguments,
    STATE(1483), 1,
      aux_sym_array_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1778), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1790), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1798), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1806), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1808), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1782), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1804), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [32764] = 27,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(1784), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1786), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1788), 1,
      anon_sym_GT_GT,
    ACTIONS(1792), 1,
      anon_sym_AMP,
    ACTIONS(1794), 1,
      anon_sym_CARET,
    ACTIONS(1796), 1,
      anon_sym_PIPE,
    ACTIONS(1800), 1,
      anon_sym_PERCENT,
    ACTIONS(1802), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1810), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1812), 1,
      sym__ternary_qmark,
    ACTIONS(1983), 1,
      anon_sym_COMMA,
    ACTIONS(2101), 1,
      anon_sym_RBRACK,
    STATE(637), 1,
      sym_arguments,
    STATE(1224), 1,
      aux_sym_sequence_expression_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1778), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1790), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1798), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1806), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1808), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1782), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1804), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [32857] = 27,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(1784), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1786), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1788), 1,
      anon_sym_GT_GT,
    ACTIONS(1792), 1,
      anon_sym_AMP,
    ACTIONS(1794), 1,
      anon_sym_CARET,
    ACTIONS(1796), 1,
      anon_sym_PIPE,
    ACTIONS(1800), 1,
      anon_sym_PERCENT,
    ACTIONS(1802), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1810), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1812), 1,
      sym__ternary_qmark,
    ACTIONS(1983), 1,
      anon_sym_COMMA,
    ACTIONS(2103), 1,
      anon_sym_RPAREN,
    STATE(637), 1,
      sym_arguments,
    STATE(1224), 1,
      aux_sym_sequence_expression_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1778), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1790), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1798), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1806), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1808), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1782), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1804), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [32950] = 27,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(1784), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1786), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1788), 1,
      anon_sym_GT_GT,
    ACTIONS(1792), 1,
      anon_sym_AMP,
    ACTIONS(1794), 1,
      anon_sym_CARET,
    ACTIONS(1796), 1,
      anon_sym_PIPE,
    ACTIONS(1800), 1,
      anon_sym_PERCENT,
    ACTIONS(1802), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1810), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1812), 1,
      sym__ternary_qmark,
    ACTIONS(1983), 1,
      anon_sym_COMMA,
    ACTIONS(2105), 1,
      anon_sym_RPAREN,
    STATE(637), 1,
      sym_arguments,
    STATE(1224), 1,
      aux_sym_sequence_expression_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1778), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1790), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1798), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1806), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1808), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1782), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1804), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [33043] = 27,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(1784), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1786), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1788), 1,
      anon_sym_GT_GT,
    ACTIONS(1792), 1,
      anon_sym_AMP,
    ACTIONS(1794), 1,
      anon_sym_CARET,
    ACTIONS(1796), 1,
      anon_sym_PIPE,
    ACTIONS(1800), 1,
      anon_sym_PERCENT,
    ACTIONS(1802), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1810), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1812), 1,
      sym__ternary_qmark,
    ACTIONS(1983), 1,
      anon_sym_COMMA,
    ACTIONS(2107), 1,
      anon_sym_RPAREN,
    STATE(637), 1,
      sym_arguments,
    STATE(1224), 1,
      aux_sym_sequence_expression_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1778), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1790), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1798), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1806), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1808), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1782), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1804), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [33136] = 27,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(1784), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1786), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1788), 1,
      anon_sym_GT_GT,
    ACTIONS(1792), 1,
      anon_sym_AMP,
    ACTIONS(1794), 1,
      anon_sym_CARET,
    ACTIONS(1796), 1,
      anon_sym_PIPE,
    ACTIONS(1800), 1,
      anon_sym_PERCENT,
    ACTIONS(1802), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1810), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1812), 1,
      sym__ternary_qmark,
    ACTIONS(1983), 1,
      anon_sym_COMMA,
    ACTIONS(2109), 1,
      anon_sym_RPAREN,
    STATE(637), 1,
      sym_arguments,
    STATE(1224), 1,
      aux_sym_sequence_expression_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1778), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1790), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1798), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1806), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1808), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1782), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1804), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [33229] = 27,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(1784), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1786), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1788), 1,
      anon_sym_GT_GT,
    ACTIONS(1792), 1,
      anon_sym_AMP,
    ACTIONS(1794), 1,
      anon_sym_CARET,
    ACTIONS(1796), 1,
      anon_sym_PIPE,
    ACTIONS(1800), 1,
      anon_sym_PERCENT,
    ACTIONS(1802), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1810), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1812), 1,
      sym__ternary_qmark,
    ACTIONS(1983), 1,
      anon_sym_COMMA,
    ACTIONS(2111), 1,
      anon_sym_RPAREN,
    STATE(637), 1,
      sym_arguments,
    STATE(1224), 1,
      aux_sym_sequence_expression_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1778), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1790), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1798), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1806), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1808), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1782), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1804), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [33322] = 27,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(1784), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1786), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1788), 1,
      anon_sym_GT_GT,
    ACTIONS(1792), 1,
      anon_sym_AMP,
    ACTIONS(1794), 1,
      anon_sym_CARET,
    ACTIONS(1796), 1,
      anon_sym_PIPE,
    ACTIONS(1800), 1,
      anon_sym_PERCENT,
    ACTIONS(1802), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1810), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1812), 1,
      sym__ternary_qmark,
    ACTIONS(1983), 1,
      anon_sym_COMMA,
    ACTIONS(2113), 1,
      anon_sym_COLON,
    STATE(637), 1,
      sym_arguments,
    STATE(1224), 1,
      aux_sym_sequence_expression_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1778), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1790), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1798), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1806), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1808), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1782), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1804), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [33415] = 27,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(1784), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1786), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1788), 1,
      anon_sym_GT_GT,
    ACTIONS(1792), 1,
      anon_sym_AMP,
    ACTIONS(1794), 1,
      anon_sym_CARET,
    ACTIONS(1796), 1,
      anon_sym_PIPE,
    ACTIONS(1800), 1,
      anon_sym_PERCENT,
    ACTIONS(1802), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1810), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1812), 1,
      sym__ternary_qmark,
    ACTIONS(1983), 1,
      anon_sym_COMMA,
    ACTIONS(2115), 1,
      anon_sym_RPAREN,
    STATE(637), 1,
      sym_arguments,
    STATE(1224), 1,
      aux_sym_sequence_expression_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1778), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1790), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1798), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1806), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1808), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1782), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1804), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [33508] = 27,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(1784), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1786), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1788), 1,
      anon_sym_GT_GT,
    ACTIONS(1792), 1,
      anon_sym_AMP,
    ACTIONS(1794), 1,
      anon_sym_CARET,
    ACTIONS(1796), 1,
      anon_sym_PIPE,
    ACTIONS(1800), 1,
      anon_sym_PERCENT,
    ACTIONS(1802), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1810), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1812), 1,
      sym__ternary_qmark,
    ACTIONS(1983), 1,
      anon_sym_COMMA,
    ACTIONS(2117), 1,
      anon_sym_RPAREN,
    STATE(637), 1,
      sym_arguments,
    STATE(1224), 1,
      aux_sym_sequence_expression_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1778), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1790), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1798), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1806), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1808), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1782), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1804), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [33601] = 27,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(1784), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1786), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1788), 1,
      anon_sym_GT_GT,
    ACTIONS(1792), 1,
      anon_sym_AMP,
    ACTIONS(1794), 1,
      anon_sym_CARET,
    ACTIONS(1796), 1,
      anon_sym_PIPE,
    ACTIONS(1800), 1,
      anon_sym_PERCENT,
    ACTIONS(1802), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1810), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1812), 1,
      sym__ternary_qmark,
    ACTIONS(1983), 1,
      anon_sym_COMMA,
    ACTIONS(2119), 1,
      anon_sym_RPAREN,
    STATE(637), 1,
      sym_arguments,
    STATE(1224), 1,
      aux_sym_sequence_expression_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1778), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1790), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1798), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1806), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1808), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1782), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1804), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [33694] = 27,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(1784), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1786), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1788), 1,
      anon_sym_GT_GT,
    ACTIONS(1792), 1,
      anon_sym_AMP,
    ACTIONS(1794), 1,
      anon_sym_CARET,
    ACTIONS(1796), 1,
      anon_sym_PIPE,
    ACTIONS(1800), 1,
      anon_sym_PERCENT,
    ACTIONS(1802), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1810), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1812), 1,
      sym__ternary_qmark,
    ACTIONS(1983), 1,
      anon_sym_COMMA,
    ACTIONS(2121), 1,
      anon_sym_SEMI,
    STATE(637), 1,
      sym_arguments,
    STATE(1224), 1,
      aux_sym_sequence_expression_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1778), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1790), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1798), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1806), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1808), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1782), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1804), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [33787] = 27,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(1784), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1786), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1788), 1,
      anon_sym_GT_GT,
    ACTIONS(1792), 1,
      anon_sym_AMP,
    ACTIONS(1794), 1,
      anon_sym_CARET,
    ACTIONS(1796), 1,
      anon_sym_PIPE,
    ACTIONS(1800), 1,
      anon_sym_PERCENT,
    ACTIONS(1802), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1810), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1812), 1,
      sym__ternary_qmark,
    ACTIONS(1983), 1,
      anon_sym_COMMA,
    ACTIONS(2123), 1,
      anon_sym_SEMI,
    STATE(637), 1,
      sym_arguments,
    STATE(1224), 1,
      aux_sym_sequence_expression_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1778), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1790), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1798), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1806), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1808), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1782), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1804), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [33880] = 27,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(1784), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1786), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1788), 1,
      anon_sym_GT_GT,
    ACTIONS(1792), 1,
      anon_sym_AMP,
    ACTIONS(1794), 1,
      anon_sym_CARET,
    ACTIONS(1796), 1,
      anon_sym_PIPE,
    ACTIONS(1800), 1,
      anon_sym_PERCENT,
    ACTIONS(1802), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1810), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1812), 1,
      sym__ternary_qmark,
    ACTIONS(1983), 1,
      anon_sym_COMMA,
    ACTIONS(2125), 1,
      anon_sym_SEMI,
    STATE(637), 1,
      sym_arguments,
    STATE(1224), 1,
      aux_sym_sequence_expression_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1778), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1790), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1798), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1806), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1808), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1782), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1804), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [33973] = 27,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(1784), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1786), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1788), 1,
      anon_sym_GT_GT,
    ACTIONS(1792), 1,
      anon_sym_AMP,
    ACTIONS(1794), 1,
      anon_sym_CARET,
    ACTIONS(1796), 1,
      anon_sym_PIPE,
    ACTIONS(1800), 1,
      anon_sym_PERCENT,
    ACTIONS(1802), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1810), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1812), 1,
      sym__ternary_qmark,
    ACTIONS(1983), 1,
      anon_sym_COMMA,
    ACTIONS(2127), 1,
      anon_sym_RPAREN,
    STATE(637), 1,
      sym_arguments,
    STATE(1224), 1,
      aux_sym_sequence_expression_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1778), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1790), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1798), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1806), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1808), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1782), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1804), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [34066] = 25,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(2133), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2135), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2137), 1,
      anon_sym_GT_GT,
    ACTIONS(2141), 1,
      anon_sym_AMP,
    ACTIONS(2143), 1,
      anon_sym_CARET,
    ACTIONS(2145), 1,
      anon_sym_PIPE,
    ACTIONS(2149), 1,
      anon_sym_PERCENT,
    ACTIONS(2151), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2159), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2161), 1,
      sym__ternary_qmark,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1826), 2,
      anon_sym_COLON,
      anon_sym_BQUOTE,
    ACTIONS(2129), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2139), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(2147), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2155), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2157), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(2131), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(2153), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [34154] = 4,
    ACTIONS(1623), 1,
      sym_regex_flags,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1619), 14,
      anon_sym_STAR,
      anon_sym_of,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_instanceof,
    ACTIONS(1621), 20,
      sym__ternary_qmark,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [34200] = 25,
    ACTIONS(1818), 1,
      anon_sym_LPAREN,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_DOT,
    ACTIONS(1838), 1,
      sym_optional_chain,
    ACTIONS(1882), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1884), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1886), 1,
      anon_sym_GT_GT,
    ACTIONS(1890), 1,
      anon_sym_AMP,
    ACTIONS(1892), 1,
      anon_sym_CARET,
    ACTIONS(1894), 1,
      anon_sym_PIPE,
    ACTIONS(1898), 1,
      anon_sym_PERCENT,
    ACTIONS(1900), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1908), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1910), 1,
      sym__ternary_qmark,
    STATE(797), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1840), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1878), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1888), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1896), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1904), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1906), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(2163), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
    ACTIONS(1880), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1902), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [34288] = 25,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(2169), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2171), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2173), 1,
      anon_sym_GT_GT,
    ACTIONS(2177), 1,
      anon_sym_AMP,
    ACTIONS(2179), 1,
      anon_sym_CARET,
    ACTIONS(2181), 1,
      anon_sym_PIPE,
    ACTIONS(2185), 1,
      anon_sym_PERCENT,
    ACTIONS(2187), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2195), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2197), 1,
      sym__ternary_qmark,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1816), 2,
      anon_sym_LBRACE,
      anon_sym_BQUOTE,
    ACTIONS(2165), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2175), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(2183), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2191), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2193), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(2167), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(2189), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [34376] = 25,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(2169), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2171), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2173), 1,
      anon_sym_GT_GT,
    ACTIONS(2177), 1,
      anon_sym_AMP,
    ACTIONS(2179), 1,
      anon_sym_CARET,
    ACTIONS(2181), 1,
      anon_sym_PIPE,
    ACTIONS(2185), 1,
      anon_sym_PERCENT,
    ACTIONS(2187), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2195), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2197), 1,
      sym__ternary_qmark,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1716), 2,
      anon_sym_LBRACE,
      anon_sym_BQUOTE,
    ACTIONS(2165), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2175), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(2183), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2191), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2193), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(2167), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(2189), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [34464] = 25,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(2169), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2171), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2173), 1,
      anon_sym_GT_GT,
    ACTIONS(2177), 1,
      anon_sym_AMP,
    ACTIONS(2179), 1,
      anon_sym_CARET,
    ACTIONS(2181), 1,
      anon_sym_PIPE,
    ACTIONS(2185), 1,
      anon_sym_PERCENT,
    ACTIONS(2187), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2195), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2197), 1,
      sym__ternary_qmark,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1780), 2,
      anon_sym_LBRACE,
      anon_sym_BQUOTE,
    ACTIONS(2165), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2175), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(2183), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2191), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2193), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(2167), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(2189), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [34552] = 25,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(2169), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2171), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2173), 1,
      anon_sym_GT_GT,
    ACTIONS(2177), 1,
      anon_sym_AMP,
    ACTIONS(2179), 1,
      anon_sym_CARET,
    ACTIONS(2181), 1,
      anon_sym_PIPE,
    ACTIONS(2185), 1,
      anon_sym_PERCENT,
    ACTIONS(2187), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2195), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2197), 1,
      sym__ternary_qmark,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1826), 2,
      anon_sym_LBRACE,
      anon_sym_BQUOTE,
    ACTIONS(2165), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2175), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(2183), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2191), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2193), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(2167), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(2189), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [34640] = 25,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(2169), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2171), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2173), 1,
      anon_sym_GT_GT,
    ACTIONS(2177), 1,
      anon_sym_AMP,
    ACTIONS(2179), 1,
      anon_sym_CARET,
    ACTIONS(2181), 1,
      anon_sym_PIPE,
    ACTIONS(2185), 1,
      anon_sym_PERCENT,
    ACTIONS(2187), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2195), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2197), 1,
      sym__ternary_qmark,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1631), 2,
      anon_sym_LBRACE,
      anon_sym_BQUOTE,
    ACTIONS(2165), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2175), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(2183), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2191), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2193), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(2167), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(2189), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [34728] = 15,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(2173), 1,
      anon_sym_GT_GT,
    ACTIONS(2185), 1,
      anon_sym_PERCENT,
    ACTIONS(2187), 1,
      anon_sym_STAR_STAR,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2165), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2175), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(2183), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1830), 7,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1828), 12,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_CARET,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_BQUOTE,
  [34796] = 10,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(2187), 1,
      anon_sym_STAR_STAR,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1830), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1828), 15,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_BQUOTE,
  [34854] = 21,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(2173), 1,
      anon_sym_GT_GT,
    ACTIONS(2177), 1,
      anon_sym_AMP,
    ACTIONS(2179), 1,
      anon_sym_CARET,
    ACTIONS(2181), 1,
      anon_sym_PIPE,
    ACTIONS(2185), 1,
      anon_sym_PERCENT,
    ACTIONS(2187), 1,
      anon_sym_STAR_STAR,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2165), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2175), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(2183), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2191), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2193), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(2167), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(2189), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1828), 6,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_QMARK,
      anon_sym_BQUOTE,
  [34934] = 22,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(2169), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2173), 1,
      anon_sym_GT_GT,
    ACTIONS(2177), 1,
      anon_sym_AMP,
    ACTIONS(2179), 1,
      anon_sym_CARET,
    ACTIONS(2181), 1,
      anon_sym_PIPE,
    ACTIONS(2185), 1,
      anon_sym_PERCENT,
    ACTIONS(2187), 1,
      anon_sym_STAR_STAR,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2165), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2175), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(2183), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2191), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2193), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(2167), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(2189), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1828), 5,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_QMARK,
      anon_sym_BQUOTE,
  [35016] = 13,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(2185), 1,
      anon_sym_PERCENT,
    ACTIONS(2187), 1,
      anon_sym_STAR_STAR,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2165), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2183), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1830), 8,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1828), 14,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_BQUOTE,
  [35080] = 19,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(2173), 1,
      anon_sym_GT_GT,
    ACTIONS(2185), 1,
      anon_sym_PERCENT,
    ACTIONS(2187), 1,
      anon_sym_STAR_STAR,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1830), 2,
      anon_sym_AMP,
      anon_sym_PIPE,
    ACTIONS(2165), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2175), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(2183), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2191), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2193), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(2167), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(2189), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1828), 7,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_CARET,
      anon_sym_QMARK_QMARK,
      anon_sym_BQUOTE,
  [35156] = 20,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(1830), 1,
      anon_sym_PIPE,
    ACTIONS(2173), 1,
      anon_sym_GT_GT,
    ACTIONS(2177), 1,
      anon_sym_AMP,
    ACTIONS(2185), 1,
      anon_sym_PERCENT,
    ACTIONS(2187), 1,
      anon_sym_STAR_STAR,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2165), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2175), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(2183), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2191), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2193), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(2167), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(2189), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1828), 7,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_CARET,
      anon_sym_QMARK_QMARK,
      anon_sym_BQUOTE,
  [35234] = 21,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(1830), 1,
      anon_sym_PIPE,
    ACTIONS(2173), 1,
      anon_sym_GT_GT,
    ACTIONS(2177), 1,
      anon_sym_AMP,
    ACTIONS(2179), 1,
      anon_sym_CARET,
    ACTIONS(2185), 1,
      anon_sym_PERCENT,
    ACTIONS(2187), 1,
      anon_sym_STAR_STAR,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2165), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2175), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(2183), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2191), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2193), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(2167), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(2189), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1828), 6,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_QMARK,
      anon_sym_BQUOTE,
  [35314] = 12,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(2185), 1,
      anon_sym_PERCENT,
    ACTIONS(2187), 1,
      anon_sym_STAR_STAR,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2165), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1830), 10,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1828), 14,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_BQUOTE,
  [35376] = 10,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(2187), 1,
      anon_sym_STAR_STAR,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1830), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1828), 15,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_BQUOTE,
  [35434] = 17,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(2173), 1,
      anon_sym_GT_GT,
    ACTIONS(2185), 1,
      anon_sym_PERCENT,
    ACTIONS(2187), 1,
      anon_sym_STAR_STAR,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2165), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2175), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(2183), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2167), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(2189), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1830), 4,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1828), 9,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_CARET,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_BQUOTE,
  [35506] = 23,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(2169), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2171), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2173), 1,
      anon_sym_GT_GT,
    ACTIONS(2177), 1,
      anon_sym_AMP,
    ACTIONS(2179), 1,
      anon_sym_CARET,
    ACTIONS(2181), 1,
      anon_sym_PIPE,
    ACTIONS(2185), 1,
      anon_sym_PERCENT,
    ACTIONS(2187), 1,
      anon_sym_STAR_STAR,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2165), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2175), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(2183), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2191), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2193), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(2167), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(2189), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1828), 4,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_QMARK_QMARK,
      anon_sym_BQUOTE,
  [35590] = 25,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(2169), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2171), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2173), 1,
      anon_sym_GT_GT,
    ACTIONS(2177), 1,
      anon_sym_AMP,
    ACTIONS(2179), 1,
      anon_sym_CARET,
    ACTIONS(2181), 1,
      anon_sym_PIPE,
    ACTIONS(2185), 1,
      anon_sym_PERCENT,
    ACTIONS(2187), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2195), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2197), 1,
      sym__ternary_qmark,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1834), 2,
      anon_sym_LBRACE,
      anon_sym_BQUOTE,
    ACTIONS(2165), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2175), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(2183), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2191), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2193), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(2167), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(2189), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [35678] = 25,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(2169), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2171), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2173), 1,
      anon_sym_GT_GT,
    ACTIONS(2177), 1,
      anon_sym_AMP,
    ACTIONS(2179), 1,
      anon_sym_CARET,
    ACTIONS(2181), 1,
      anon_sym_PIPE,
    ACTIONS(2185), 1,
      anon_sym_PERCENT,
    ACTIONS(2187), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2195), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2197), 1,
      sym__ternary_qmark,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1676), 2,
      anon_sym_LBRACE,
      anon_sym_BQUOTE,
    ACTIONS(2165), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2175), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(2183), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2191), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2193), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(2167), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(2189), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [35766] = 25,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(2169), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2171), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2173), 1,
      anon_sym_GT_GT,
    ACTIONS(2177), 1,
      anon_sym_AMP,
    ACTIONS(2179), 1,
      anon_sym_CARET,
    ACTIONS(2181), 1,
      anon_sym_PIPE,
    ACTIONS(2185), 1,
      anon_sym_PERCENT,
    ACTIONS(2187), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2195), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2197), 1,
      sym__ternary_qmark,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1700), 2,
      anon_sym_LBRACE,
      anon_sym_BQUOTE,
    ACTIONS(2165), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2175), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(2183), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2191), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2193), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(2167), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(2189), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [35854] = 25,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(2169), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2171), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2173), 1,
      anon_sym_GT_GT,
    ACTIONS(2177), 1,
      anon_sym_AMP,
    ACTIONS(2179), 1,
      anon_sym_CARET,
    ACTIONS(2181), 1,
      anon_sym_PIPE,
    ACTIONS(2185), 1,
      anon_sym_PERCENT,
    ACTIONS(2187), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2195), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2197), 1,
      sym__ternary_qmark,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1708), 2,
      anon_sym_LBRACE,
      anon_sym_BQUOTE,
    ACTIONS(2165), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2175), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(2183), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2191), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2193), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(2167), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(2189), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [35942] = 25,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(2169), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2171), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2173), 1,
      anon_sym_GT_GT,
    ACTIONS(2177), 1,
      anon_sym_AMP,
    ACTIONS(2179), 1,
      anon_sym_CARET,
    ACTIONS(2181), 1,
      anon_sym_PIPE,
    ACTIONS(2185), 1,
      anon_sym_PERCENT,
    ACTIONS(2187), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2195), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2197), 1,
      sym__ternary_qmark,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1712), 2,
      anon_sym_LBRACE,
      anon_sym_BQUOTE,
    ACTIONS(2165), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2175), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(2183), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2191), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2193), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(2167), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(2189), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [36030] = 25,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(2169), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2171), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2173), 1,
      anon_sym_GT_GT,
    ACTIONS(2177), 1,
      anon_sym_AMP,
    ACTIONS(2179), 1,
      anon_sym_CARET,
    ACTIONS(2181), 1,
      anon_sym_PIPE,
    ACTIONS(2185), 1,
      anon_sym_PERCENT,
    ACTIONS(2187), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2195), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2197), 1,
      sym__ternary_qmark,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1814), 2,
      anon_sym_LBRACE,
      anon_sym_BQUOTE,
    ACTIONS(2165), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2175), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(2183), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2191), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2193), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(2167), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(2189), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [36118] = 25,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(2203), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2205), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2207), 1,
      anon_sym_GT_GT,
    ACTIONS(2211), 1,
      anon_sym_AMP,
    ACTIONS(2213), 1,
      anon_sym_CARET,
    ACTIONS(2215), 1,
      anon_sym_PIPE,
    ACTIONS(2219), 1,
      anon_sym_PERCENT,
    ACTIONS(2221), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2229), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2231), 1,
      sym__ternary_qmark,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1832), 2,
      anon_sym_of,
      anon_sym_BQUOTE,
    ACTIONS(2199), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2209), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(2217), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2225), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2227), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(2201), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(2223), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [36206] = 26,
    ACTIONS(1818), 1,
      anon_sym_LPAREN,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_DOT,
    ACTIONS(1838), 1,
      sym_optional_chain,
    ACTIONS(1882), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1884), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1886), 1,
      anon_sym_GT_GT,
    ACTIONS(1890), 1,
      anon_sym_AMP,
    ACTIONS(1892), 1,
      anon_sym_CARET,
    ACTIONS(1894), 1,
      anon_sym_PIPE,
    ACTIONS(1898), 1,
      anon_sym_PERCENT,
    ACTIONS(1900), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1908), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1910), 1,
      sym__ternary_qmark,
    ACTIONS(2233), 1,
      anon_sym_SEMI,
    ACTIONS(2235), 1,
      sym__automatic_semicolon,
    STATE(797), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1840), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1878), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1888), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1896), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1904), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1906), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1880), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1902), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [36296] = 25,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(1784), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1786), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1788), 1,
      anon_sym_GT_GT,
    ACTIONS(1792), 1,
      anon_sym_AMP,
    ACTIONS(1794), 1,
      anon_sym_CARET,
    ACTIONS(1796), 1,
      anon_sym_PIPE,
    ACTIONS(1800), 1,
      anon_sym_PERCENT,
    ACTIONS(1802), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1810), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1812), 1,
      sym__ternary_qmark,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1778), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1790), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1798), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1806), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1808), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1924), 2,
      anon_sym_COMMA,
      anon_sym_RBRACE,
    ACTIONS(1782), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1804), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [36384] = 25,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(1784), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1786), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1788), 1,
      anon_sym_GT_GT,
    ACTIONS(1792), 1,
      anon_sym_AMP,
    ACTIONS(1794), 1,
      anon_sym_CARET,
    ACTIONS(1796), 1,
      anon_sym_PIPE,
    ACTIONS(1800), 1,
      anon_sym_PERCENT,
    ACTIONS(1802), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1810), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1812), 1,
      sym__ternary_qmark,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1778), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1790), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1798), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1806), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1808), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1952), 2,
      anon_sym_COMMA,
      anon_sym_RBRACE,
    ACTIONS(1782), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1804), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [36472] = 4,
    ACTIONS(1557), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1426), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1428), 22,
      sym__ternary_qmark,
      anon_sym_LBRACE,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [36518] = 25,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(2169), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2171), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2173), 1,
      anon_sym_GT_GT,
    ACTIONS(2177), 1,
      anon_sym_AMP,
    ACTIONS(2179), 1,
      anon_sym_CARET,
    ACTIONS(2181), 1,
      anon_sym_PIPE,
    ACTIONS(2185), 1,
      anon_sym_PERCENT,
    ACTIONS(2187), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2195), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2197), 1,
      sym__ternary_qmark,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1832), 2,
      anon_sym_LBRACE,
      anon_sym_BQUOTE,
    ACTIONS(2165), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2175), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(2183), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2191), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2193), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(2167), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(2189), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [36606] = 25,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(2133), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2135), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2137), 1,
      anon_sym_GT_GT,
    ACTIONS(2141), 1,
      anon_sym_AMP,
    ACTIONS(2143), 1,
      anon_sym_CARET,
    ACTIONS(2145), 1,
      anon_sym_PIPE,
    ACTIONS(2149), 1,
      anon_sym_PERCENT,
    ACTIONS(2151), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2159), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2161), 1,
      sym__ternary_qmark,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1816), 2,
      anon_sym_COLON,
      anon_sym_BQUOTE,
    ACTIONS(2129), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2139), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(2147), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2155), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2157), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(2131), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(2153), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [36694] = 25,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(2133), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2135), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2137), 1,
      anon_sym_GT_GT,
    ACTIONS(2141), 1,
      anon_sym_AMP,
    ACTIONS(2143), 1,
      anon_sym_CARET,
    ACTIONS(2145), 1,
      anon_sym_PIPE,
    ACTIONS(2149), 1,
      anon_sym_PERCENT,
    ACTIONS(2151), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2159), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2161), 1,
      sym__ternary_qmark,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1716), 2,
      anon_sym_COLON,
      anon_sym_BQUOTE,
    ACTIONS(2129), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2139), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(2147), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2155), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2157), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(2131), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(2153), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [36782] = 25,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(2133), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2135), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2137), 1,
      anon_sym_GT_GT,
    ACTIONS(2141), 1,
      anon_sym_AMP,
    ACTIONS(2143), 1,
      anon_sym_CARET,
    ACTIONS(2145), 1,
      anon_sym_PIPE,
    ACTIONS(2149), 1,
      anon_sym_PERCENT,
    ACTIONS(2151), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2159), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2161), 1,
      sym__ternary_qmark,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1780), 2,
      anon_sym_COLON,
      anon_sym_BQUOTE,
    ACTIONS(2129), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2139), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(2147), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2155), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2157), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(2131), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(2153), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [36870] = 4,
    ACTIONS(1561), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1426), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1428), 22,
      sym__ternary_qmark,
      anon_sym_of,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [36916] = 15,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(2137), 1,
      anon_sym_GT_GT,
    ACTIONS(2149), 1,
      anon_sym_PERCENT,
    ACTIONS(2151), 1,
      anon_sym_STAR_STAR,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2129), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2139), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(2147), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1830), 7,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1828), 12,
      sym__ternary_qmark,
      anon_sym_COLON,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_CARET,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_BQUOTE,
  [36984] = 10,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(2151), 1,
      anon_sym_STAR_STAR,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1830), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1828), 15,
      sym__ternary_qmark,
      anon_sym_COLON,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_BQUOTE,
  [37042] = 21,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(2137), 1,
      anon_sym_GT_GT,
    ACTIONS(2141), 1,
      anon_sym_AMP,
    ACTIONS(2143), 1,
      anon_sym_CARET,
    ACTIONS(2145), 1,
      anon_sym_PIPE,
    ACTIONS(2149), 1,
      anon_sym_PERCENT,
    ACTIONS(2151), 1,
      anon_sym_STAR_STAR,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2129), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2139), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(2147), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2155), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2157), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(2131), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(2153), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1828), 6,
      sym__ternary_qmark,
      anon_sym_COLON,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_QMARK,
      anon_sym_BQUOTE,
  [37122] = 22,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(2133), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2137), 1,
      anon_sym_GT_GT,
    ACTIONS(2141), 1,
      anon_sym_AMP,
    ACTIONS(2143), 1,
      anon_sym_CARET,
    ACTIONS(2145), 1,
      anon_sym_PIPE,
    ACTIONS(2149), 1,
      anon_sym_PERCENT,
    ACTIONS(2151), 1,
      anon_sym_STAR_STAR,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2129), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2139), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(2147), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2155), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2157), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(2131), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(2153), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1828), 5,
      sym__ternary_qmark,
      anon_sym_COLON,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_QMARK,
      anon_sym_BQUOTE,
  [37204] = 13,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(2149), 1,
      anon_sym_PERCENT,
    ACTIONS(2151), 1,
      anon_sym_STAR_STAR,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2129), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2147), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1830), 8,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1828), 14,
      sym__ternary_qmark,
      anon_sym_COLON,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_BQUOTE,
  [37268] = 19,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(2137), 1,
      anon_sym_GT_GT,
    ACTIONS(2149), 1,
      anon_sym_PERCENT,
    ACTIONS(2151), 1,
      anon_sym_STAR_STAR,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1830), 2,
      anon_sym_AMP,
      anon_sym_PIPE,
    ACTIONS(2129), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2139), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(2147), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2155), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2157), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(2131), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(2153), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1828), 7,
      sym__ternary_qmark,
      anon_sym_COLON,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_CARET,
      anon_sym_QMARK_QMARK,
      anon_sym_BQUOTE,
  [37344] = 20,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(1830), 1,
      anon_sym_PIPE,
    ACTIONS(2137), 1,
      anon_sym_GT_GT,
    ACTIONS(2141), 1,
      anon_sym_AMP,
    ACTIONS(2149), 1,
      anon_sym_PERCENT,
    ACTIONS(2151), 1,
      anon_sym_STAR_STAR,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2129), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2139), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(2147), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2155), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2157), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(2131), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(2153), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1828), 7,
      sym__ternary_qmark,
      anon_sym_COLON,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_CARET,
      anon_sym_QMARK_QMARK,
      anon_sym_BQUOTE,
  [37422] = 21,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(1830), 1,
      anon_sym_PIPE,
    ACTIONS(2137), 1,
      anon_sym_GT_GT,
    ACTIONS(2141), 1,
      anon_sym_AMP,
    ACTIONS(2143), 1,
      anon_sym_CARET,
    ACTIONS(2149), 1,
      anon_sym_PERCENT,
    ACTIONS(2151), 1,
      anon_sym_STAR_STAR,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2129), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2139), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(2147), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2155), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2157), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(2131), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(2153), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1828), 6,
      sym__ternary_qmark,
      anon_sym_COLON,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_QMARK,
      anon_sym_BQUOTE,
  [37502] = 12,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(2149), 1,
      anon_sym_PERCENT,
    ACTIONS(2151), 1,
      anon_sym_STAR_STAR,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2129), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1830), 10,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1828), 14,
      sym__ternary_qmark,
      anon_sym_COLON,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_BQUOTE,
  [37564] = 10,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(2151), 1,
      anon_sym_STAR_STAR,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1830), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1828), 15,
      sym__ternary_qmark,
      anon_sym_COLON,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_BQUOTE,
  [37622] = 17,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(2137), 1,
      anon_sym_GT_GT,
    ACTIONS(2149), 1,
      anon_sym_PERCENT,
    ACTIONS(2151), 1,
      anon_sym_STAR_STAR,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2129), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2139), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(2147), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2131), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(2153), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1830), 4,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1828), 9,
      sym__ternary_qmark,
      anon_sym_COLON,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_CARET,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_BQUOTE,
  [37694] = 23,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(2133), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2135), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2137), 1,
      anon_sym_GT_GT,
    ACTIONS(2141), 1,
      anon_sym_AMP,
    ACTIONS(2143), 1,
      anon_sym_CARET,
    ACTIONS(2145), 1,
      anon_sym_PIPE,
    ACTIONS(2149), 1,
      anon_sym_PERCENT,
    ACTIONS(2151), 1,
      anon_sym_STAR_STAR,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2129), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2139), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(2147), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2155), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2157), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(2131), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(2153), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1828), 4,
      sym__ternary_qmark,
      anon_sym_COLON,
      anon_sym_QMARK_QMARK,
      anon_sym_BQUOTE,
  [37778] = 25,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(2133), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2135), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2137), 1,
      anon_sym_GT_GT,
    ACTIONS(2141), 1,
      anon_sym_AMP,
    ACTIONS(2143), 1,
      anon_sym_CARET,
    ACTIONS(2145), 1,
      anon_sym_PIPE,
    ACTIONS(2149), 1,
      anon_sym_PERCENT,
    ACTIONS(2151), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2159), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2161), 1,
      sym__ternary_qmark,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1834), 2,
      anon_sym_COLON,
      anon_sym_BQUOTE,
    ACTIONS(2129), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2139), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(2147), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2155), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2157), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(2131), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(2153), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [37866] = 25,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(2133), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2135), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2137), 1,
      anon_sym_GT_GT,
    ACTIONS(2141), 1,
      anon_sym_AMP,
    ACTIONS(2143), 1,
      anon_sym_CARET,
    ACTIONS(2145), 1,
      anon_sym_PIPE,
    ACTIONS(2149), 1,
      anon_sym_PERCENT,
    ACTIONS(2151), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2159), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2161), 1,
      sym__ternary_qmark,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1676), 2,
      anon_sym_COLON,
      anon_sym_BQUOTE,
    ACTIONS(2129), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2139), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(2147), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2155), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2157), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(2131), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(2153), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [37954] = 25,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(2133), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2135), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2137), 1,
      anon_sym_GT_GT,
    ACTIONS(2141), 1,
      anon_sym_AMP,
    ACTIONS(2143), 1,
      anon_sym_CARET,
    ACTIONS(2145), 1,
      anon_sym_PIPE,
    ACTIONS(2149), 1,
      anon_sym_PERCENT,
    ACTIONS(2151), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2159), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2161), 1,
      sym__ternary_qmark,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1700), 2,
      anon_sym_COLON,
      anon_sym_BQUOTE,
    ACTIONS(2129), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2139), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(2147), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2155), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2157), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(2131), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(2153), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [38042] = 25,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(2133), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2135), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2137), 1,
      anon_sym_GT_GT,
    ACTIONS(2141), 1,
      anon_sym_AMP,
    ACTIONS(2143), 1,
      anon_sym_CARET,
    ACTIONS(2145), 1,
      anon_sym_PIPE,
    ACTIONS(2149), 1,
      anon_sym_PERCENT,
    ACTIONS(2151), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2159), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2161), 1,
      sym__ternary_qmark,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1708), 2,
      anon_sym_COLON,
      anon_sym_BQUOTE,
    ACTIONS(2129), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2139), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(2147), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2155), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2157), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(2131), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(2153), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [38130] = 25,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(2133), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2135), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2137), 1,
      anon_sym_GT_GT,
    ACTIONS(2141), 1,
      anon_sym_AMP,
    ACTIONS(2143), 1,
      anon_sym_CARET,
    ACTIONS(2145), 1,
      anon_sym_PIPE,
    ACTIONS(2149), 1,
      anon_sym_PERCENT,
    ACTIONS(2151), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2159), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2161), 1,
      sym__ternary_qmark,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1712), 2,
      anon_sym_COLON,
      anon_sym_BQUOTE,
    ACTIONS(2129), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2139), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(2147), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2155), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2157), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(2131), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(2153), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [38218] = 25,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(2133), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2135), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2137), 1,
      anon_sym_GT_GT,
    ACTIONS(2141), 1,
      anon_sym_AMP,
    ACTIONS(2143), 1,
      anon_sym_CARET,
    ACTIONS(2145), 1,
      anon_sym_PIPE,
    ACTIONS(2149), 1,
      anon_sym_PERCENT,
    ACTIONS(2151), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2159), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2161), 1,
      sym__ternary_qmark,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1814), 2,
      anon_sym_COLON,
      anon_sym_BQUOTE,
    ACTIONS(2129), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2139), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(2147), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2155), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2157), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(2131), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(2153), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [38306] = 25,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(1784), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1786), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1788), 1,
      anon_sym_GT_GT,
    ACTIONS(1792), 1,
      anon_sym_AMP,
    ACTIONS(1794), 1,
      anon_sym_CARET,
    ACTIONS(1796), 1,
      anon_sym_PIPE,
    ACTIONS(1800), 1,
      anon_sym_PERCENT,
    ACTIONS(1802), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1810), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1812), 1,
      sym__ternary_qmark,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1778), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1790), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1798), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1806), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1808), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(2237), 2,
      anon_sym_COMMA,
      anon_sym_RBRACE,
    ACTIONS(1782), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1804), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [38394] = 25,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(2203), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2205), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2207), 1,
      anon_sym_GT_GT,
    ACTIONS(2211), 1,
      anon_sym_AMP,
    ACTIONS(2213), 1,
      anon_sym_CARET,
    ACTIONS(2215), 1,
      anon_sym_PIPE,
    ACTIONS(2219), 1,
      anon_sym_PERCENT,
    ACTIONS(2221), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2229), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2231), 1,
      sym__ternary_qmark,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1816), 2,
      anon_sym_of,
      anon_sym_BQUOTE,
    ACTIONS(2199), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2209), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(2217), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2225), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2227), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(2201), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(2223), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [38482] = 25,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(2203), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2205), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2207), 1,
      anon_sym_GT_GT,
    ACTIONS(2211), 1,
      anon_sym_AMP,
    ACTIONS(2213), 1,
      anon_sym_CARET,
    ACTIONS(2215), 1,
      anon_sym_PIPE,
    ACTIONS(2219), 1,
      anon_sym_PERCENT,
    ACTIONS(2221), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2229), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2231), 1,
      sym__ternary_qmark,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1716), 2,
      anon_sym_of,
      anon_sym_BQUOTE,
    ACTIONS(2199), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2209), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(2217), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2225), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2227), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(2201), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(2223), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [38570] = 25,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(2203), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2205), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2207), 1,
      anon_sym_GT_GT,
    ACTIONS(2211), 1,
      anon_sym_AMP,
    ACTIONS(2213), 1,
      anon_sym_CARET,
    ACTIONS(2215), 1,
      anon_sym_PIPE,
    ACTIONS(2219), 1,
      anon_sym_PERCENT,
    ACTIONS(2221), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2229), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2231), 1,
      sym__ternary_qmark,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1780), 2,
      anon_sym_of,
      anon_sym_BQUOTE,
    ACTIONS(2199), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2209), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(2217), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2225), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2227), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(2201), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(2223), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [38658] = 25,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(2203), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2205), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2207), 1,
      anon_sym_GT_GT,
    ACTIONS(2211), 1,
      anon_sym_AMP,
    ACTIONS(2213), 1,
      anon_sym_CARET,
    ACTIONS(2215), 1,
      anon_sym_PIPE,
    ACTIONS(2219), 1,
      anon_sym_PERCENT,
    ACTIONS(2221), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2229), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2231), 1,
      sym__ternary_qmark,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1826), 2,
      anon_sym_of,
      anon_sym_BQUOTE,
    ACTIONS(2199), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2209), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(2217), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2225), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2227), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(2201), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(2223), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [38746] = 25,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(2203), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2205), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2207), 1,
      anon_sym_GT_GT,
    ACTIONS(2211), 1,
      anon_sym_AMP,
    ACTIONS(2213), 1,
      anon_sym_CARET,
    ACTIONS(2215), 1,
      anon_sym_PIPE,
    ACTIONS(2219), 1,
      anon_sym_PERCENT,
    ACTIONS(2221), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2229), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2231), 1,
      sym__ternary_qmark,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1631), 2,
      anon_sym_of,
      anon_sym_BQUOTE,
    ACTIONS(2199), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2209), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(2217), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2225), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2227), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(2201), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(2223), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [38834] = 15,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(2207), 1,
      anon_sym_GT_GT,
    ACTIONS(2219), 1,
      anon_sym_PERCENT,
    ACTIONS(2221), 1,
      anon_sym_STAR_STAR,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2199), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2209), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(2217), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1830), 7,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1828), 12,
      sym__ternary_qmark,
      anon_sym_of,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_CARET,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_BQUOTE,
  [38902] = 10,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(2221), 1,
      anon_sym_STAR_STAR,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1830), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1828), 15,
      sym__ternary_qmark,
      anon_sym_of,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_BQUOTE,
  [38960] = 21,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(2207), 1,
      anon_sym_GT_GT,
    ACTIONS(2211), 1,
      anon_sym_AMP,
    ACTIONS(2213), 1,
      anon_sym_CARET,
    ACTIONS(2215), 1,
      anon_sym_PIPE,
    ACTIONS(2219), 1,
      anon_sym_PERCENT,
    ACTIONS(2221), 1,
      anon_sym_STAR_STAR,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2199), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2209), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(2217), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2225), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2227), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(2201), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(2223), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1828), 6,
      sym__ternary_qmark,
      anon_sym_of,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_QMARK,
      anon_sym_BQUOTE,
  [39040] = 22,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(2203), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2207), 1,
      anon_sym_GT_GT,
    ACTIONS(2211), 1,
      anon_sym_AMP,
    ACTIONS(2213), 1,
      anon_sym_CARET,
    ACTIONS(2215), 1,
      anon_sym_PIPE,
    ACTIONS(2219), 1,
      anon_sym_PERCENT,
    ACTIONS(2221), 1,
      anon_sym_STAR_STAR,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2199), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2209), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(2217), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2225), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2227), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(2201), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(2223), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1828), 5,
      sym__ternary_qmark,
      anon_sym_of,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_QMARK,
      anon_sym_BQUOTE,
  [39122] = 13,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(2219), 1,
      anon_sym_PERCENT,
    ACTIONS(2221), 1,
      anon_sym_STAR_STAR,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2199), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2217), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1830), 8,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1828), 14,
      sym__ternary_qmark,
      anon_sym_of,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_BQUOTE,
  [39186] = 19,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(2207), 1,
      anon_sym_GT_GT,
    ACTIONS(2219), 1,
      anon_sym_PERCENT,
    ACTIONS(2221), 1,
      anon_sym_STAR_STAR,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1830), 2,
      anon_sym_AMP,
      anon_sym_PIPE,
    ACTIONS(2199), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2209), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(2217), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2225), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2227), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(2201), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(2223), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1828), 7,
      sym__ternary_qmark,
      anon_sym_of,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_CARET,
      anon_sym_QMARK_QMARK,
      anon_sym_BQUOTE,
  [39262] = 20,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(1830), 1,
      anon_sym_PIPE,
    ACTIONS(2207), 1,
      anon_sym_GT_GT,
    ACTIONS(2211), 1,
      anon_sym_AMP,
    ACTIONS(2219), 1,
      anon_sym_PERCENT,
    ACTIONS(2221), 1,
      anon_sym_STAR_STAR,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2199), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2209), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(2217), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2225), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2227), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(2201), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(2223), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1828), 7,
      sym__ternary_qmark,
      anon_sym_of,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_CARET,
      anon_sym_QMARK_QMARK,
      anon_sym_BQUOTE,
  [39340] = 21,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(1830), 1,
      anon_sym_PIPE,
    ACTIONS(2207), 1,
      anon_sym_GT_GT,
    ACTIONS(2211), 1,
      anon_sym_AMP,
    ACTIONS(2213), 1,
      anon_sym_CARET,
    ACTIONS(2219), 1,
      anon_sym_PERCENT,
    ACTIONS(2221), 1,
      anon_sym_STAR_STAR,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2199), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2209), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(2217), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2225), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2227), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(2201), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(2223), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1828), 6,
      sym__ternary_qmark,
      anon_sym_of,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_QMARK_QMARK,
      anon_sym_BQUOTE,
  [39420] = 12,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(2219), 1,
      anon_sym_PERCENT,
    ACTIONS(2221), 1,
      anon_sym_STAR_STAR,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2199), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1830), 10,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1828), 14,
      sym__ternary_qmark,
      anon_sym_of,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_BQUOTE,
  [39482] = 10,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(2221), 1,
      anon_sym_STAR_STAR,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1830), 12,
      anon_sym_STAR,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1828), 15,
      sym__ternary_qmark,
      anon_sym_of,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_BQUOTE,
  [39540] = 17,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(2207), 1,
      anon_sym_GT_GT,
    ACTIONS(2219), 1,
      anon_sym_PERCENT,
    ACTIONS(2221), 1,
      anon_sym_STAR_STAR,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2199), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2209), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(2217), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2201), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(2223), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1830), 4,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1828), 9,
      sym__ternary_qmark,
      anon_sym_of,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_CARET,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_BQUOTE,
  [39612] = 23,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(2203), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2205), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2207), 1,
      anon_sym_GT_GT,
    ACTIONS(2211), 1,
      anon_sym_AMP,
    ACTIONS(2213), 1,
      anon_sym_CARET,
    ACTIONS(2215), 1,
      anon_sym_PIPE,
    ACTIONS(2219), 1,
      anon_sym_PERCENT,
    ACTIONS(2221), 1,
      anon_sym_STAR_STAR,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2199), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2209), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(2217), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2225), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2227), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(2201), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(2223), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
    ACTIONS(1828), 4,
      sym__ternary_qmark,
      anon_sym_of,
      anon_sym_QMARK_QMARK,
      anon_sym_BQUOTE,
  [39696] = 25,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(2203), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2205), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2207), 1,
      anon_sym_GT_GT,
    ACTIONS(2211), 1,
      anon_sym_AMP,
    ACTIONS(2213), 1,
      anon_sym_CARET,
    ACTIONS(2215), 1,
      anon_sym_PIPE,
    ACTIONS(2219), 1,
      anon_sym_PERCENT,
    ACTIONS(2221), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2229), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2231), 1,
      sym__ternary_qmark,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1834), 2,
      anon_sym_of,
      anon_sym_BQUOTE,
    ACTIONS(2199), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2209), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(2217), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2225), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2227), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(2201), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(2223), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [39784] = 25,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(2203), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2205), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2207), 1,
      anon_sym_GT_GT,
    ACTIONS(2211), 1,
      anon_sym_AMP,
    ACTIONS(2213), 1,
      anon_sym_CARET,
    ACTIONS(2215), 1,
      anon_sym_PIPE,
    ACTIONS(2219), 1,
      anon_sym_PERCENT,
    ACTIONS(2221), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2229), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2231), 1,
      sym__ternary_qmark,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1676), 2,
      anon_sym_of,
      anon_sym_BQUOTE,
    ACTIONS(2199), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2209), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(2217), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2225), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2227), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(2201), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(2223), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [39872] = 25,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(2203), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2205), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2207), 1,
      anon_sym_GT_GT,
    ACTIONS(2211), 1,
      anon_sym_AMP,
    ACTIONS(2213), 1,
      anon_sym_CARET,
    ACTIONS(2215), 1,
      anon_sym_PIPE,
    ACTIONS(2219), 1,
      anon_sym_PERCENT,
    ACTIONS(2221), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2229), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2231), 1,
      sym__ternary_qmark,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1700), 2,
      anon_sym_of,
      anon_sym_BQUOTE,
    ACTIONS(2199), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2209), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(2217), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2225), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2227), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(2201), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(2223), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [39960] = 25,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(2203), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2205), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2207), 1,
      anon_sym_GT_GT,
    ACTIONS(2211), 1,
      anon_sym_AMP,
    ACTIONS(2213), 1,
      anon_sym_CARET,
    ACTIONS(2215), 1,
      anon_sym_PIPE,
    ACTIONS(2219), 1,
      anon_sym_PERCENT,
    ACTIONS(2221), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2229), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2231), 1,
      sym__ternary_qmark,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1708), 2,
      anon_sym_of,
      anon_sym_BQUOTE,
    ACTIONS(2199), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2209), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(2217), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2225), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2227), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(2201), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(2223), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [40048] = 25,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(2203), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2205), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2207), 1,
      anon_sym_GT_GT,
    ACTIONS(2211), 1,
      anon_sym_AMP,
    ACTIONS(2213), 1,
      anon_sym_CARET,
    ACTIONS(2215), 1,
      anon_sym_PIPE,
    ACTIONS(2219), 1,
      anon_sym_PERCENT,
    ACTIONS(2221), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2229), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2231), 1,
      sym__ternary_qmark,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1712), 2,
      anon_sym_of,
      anon_sym_BQUOTE,
    ACTIONS(2199), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2209), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(2217), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2225), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2227), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(2201), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(2223), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [40136] = 25,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(2203), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2205), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2207), 1,
      anon_sym_GT_GT,
    ACTIONS(2211), 1,
      anon_sym_AMP,
    ACTIONS(2213), 1,
      anon_sym_CARET,
    ACTIONS(2215), 1,
      anon_sym_PIPE,
    ACTIONS(2219), 1,
      anon_sym_PERCENT,
    ACTIONS(2221), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2229), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2231), 1,
      sym__ternary_qmark,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1814), 2,
      anon_sym_of,
      anon_sym_BQUOTE,
    ACTIONS(2199), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2209), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(2217), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2225), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2227), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(2201), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(2223), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [40224] = 6,
    ACTIONS(1459), 1,
      anon_sym_EQ,
    ACTIONS(1550), 1,
      anon_sym_of,
    ACTIONS(1552), 1,
      anon_sym_in,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1426), 11,
      anon_sym_STAR,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_GT_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_PLUS,
      anon_sym_DASH,
      anon_sym_SLASH,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1428), 21,
      sym__ternary_qmark,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DOT,
      sym_optional_chain,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
      anon_sym_CARET,
      anon_sym_PERCENT,
      anon_sym_STAR_STAR,
      anon_sym_LT_EQ,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
      anon_sym_GT_EQ,
      anon_sym_QMARK_QMARK,
      anon_sym_instanceof,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
      anon_sym_BQUOTE,
  [40274] = 25,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(2133), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2135), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2137), 1,
      anon_sym_GT_GT,
    ACTIONS(2141), 1,
      anon_sym_AMP,
    ACTIONS(2143), 1,
      anon_sym_CARET,
    ACTIONS(2145), 1,
      anon_sym_PIPE,
    ACTIONS(2149), 1,
      anon_sym_PERCENT,
    ACTIONS(2151), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2159), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2161), 1,
      sym__ternary_qmark,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1832), 2,
      anon_sym_COLON,
      anon_sym_BQUOTE,
    ACTIONS(2129), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2139), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(2147), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2155), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2157), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(2131), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(2153), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [40362] = 25,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(2133), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2135), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2137), 1,
      anon_sym_GT_GT,
    ACTIONS(2141), 1,
      anon_sym_AMP,
    ACTIONS(2143), 1,
      anon_sym_CARET,
    ACTIONS(2145), 1,
      anon_sym_PIPE,
    ACTIONS(2149), 1,
      anon_sym_PERCENT,
    ACTIONS(2151), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2159), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2161), 1,
      sym__ternary_qmark,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1631), 2,
      anon_sym_COLON,
      anon_sym_BQUOTE,
    ACTIONS(2129), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2139), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(2147), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2155), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2157), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(2131), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(2153), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [40450] = 25,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(2133), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2135), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2137), 1,
      anon_sym_GT_GT,
    ACTIONS(2141), 1,
      anon_sym_AMP,
    ACTIONS(2143), 1,
      anon_sym_CARET,
    ACTIONS(2145), 1,
      anon_sym_PIPE,
    ACTIONS(2149), 1,
      anon_sym_PERCENT,
    ACTIONS(2151), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2159), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2161), 1,
      sym__ternary_qmark,
    ACTIONS(2239), 1,
      anon_sym_COLON,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2129), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2139), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(2147), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2155), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2157), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(2131), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(2153), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [40537] = 21,
    ACTIONS(93), 1,
      anon_sym_AT,
    ACTIONS(101), 1,
      anon_sym_STAR,
    ACTIONS(115), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(123), 1,
      aux_sym_method_definition_token1,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(1969), 1,
      anon_sym_LBRACE,
    ACTIONS(1973), 1,
      anon_sym_LBRACK,
    ACTIONS(2246), 1,
      anon_sym_async,
    ACTIONS(2248), 1,
      anon_sym_static,
    STATE(1019), 1,
      aux_sym_export_statement_repeat1,
    STATE(1097), 1,
      sym_decorator,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1977), 2,
      sym_number,
      sym_private_property_identifier,
    ACTIONS(2243), 2,
      anon_sym_COMMA,
      anon_sym_RBRACE,
    ACTIONS(2250), 2,
      anon_sym_get,
      anon_sym_set,
    STATE(1407), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    STATE(1514), 3,
      sym_object_assignment_pattern,
      sym_rest_pattern,
      sym_pair_pattern,
    STATE(1516), 3,
      sym_spread_element,
      sym_method_definition,
      sym_pair,
    STATE(1788), 3,
      sym_object_pattern,
      sym_array_pattern,
      sym__destructuring_pattern,
    ACTIONS(2241), 4,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      sym_identifier,
  [40616] = 25,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(2133), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2135), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2137), 1,
      anon_sym_GT_GT,
    ACTIONS(2141), 1,
      anon_sym_AMP,
    ACTIONS(2143), 1,
      anon_sym_CARET,
    ACTIONS(2145), 1,
      anon_sym_PIPE,
    ACTIONS(2149), 1,
      anon_sym_PERCENT,
    ACTIONS(2151), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2159), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2161), 1,
      sym__ternary_qmark,
    ACTIONS(2252), 1,
      anon_sym_COLON,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2129), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2139), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(2147), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2155), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2157), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(2131), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(2153), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [40703] = 25,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(2169), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2171), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2173), 1,
      anon_sym_GT_GT,
    ACTIONS(2177), 1,
      anon_sym_AMP,
    ACTIONS(2179), 1,
      anon_sym_CARET,
    ACTIONS(2181), 1,
      anon_sym_PIPE,
    ACTIONS(2185), 1,
      anon_sym_PERCENT,
    ACTIONS(2187), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2195), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2197), 1,
      sym__ternary_qmark,
    ACTIONS(2254), 1,
      anon_sym_LBRACE,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2165), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2175), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(2183), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2191), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2193), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(2167), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(2189), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [40790] = 25,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(2133), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2135), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2137), 1,
      anon_sym_GT_GT,
    ACTIONS(2141), 1,
      anon_sym_AMP,
    ACTIONS(2143), 1,
      anon_sym_CARET,
    ACTIONS(2145), 1,
      anon_sym_PIPE,
    ACTIONS(2149), 1,
      anon_sym_PERCENT,
    ACTIONS(2151), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2159), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2161), 1,
      sym__ternary_qmark,
    ACTIONS(2256), 1,
      anon_sym_COLON,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2129), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2139), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(2147), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2155), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2157), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(2131), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(2153), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [40877] = 25,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(2133), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2135), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2137), 1,
      anon_sym_GT_GT,
    ACTIONS(2141), 1,
      anon_sym_AMP,
    ACTIONS(2143), 1,
      anon_sym_CARET,
    ACTIONS(2145), 1,
      anon_sym_PIPE,
    ACTIONS(2149), 1,
      anon_sym_PERCENT,
    ACTIONS(2151), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2159), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2161), 1,
      sym__ternary_qmark,
    ACTIONS(2258), 1,
      anon_sym_COLON,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2129), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2139), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(2147), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2155), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2157), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(2131), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(2153), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [40964] = 25,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(2133), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2135), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2137), 1,
      anon_sym_GT_GT,
    ACTIONS(2141), 1,
      anon_sym_AMP,
    ACTIONS(2143), 1,
      anon_sym_CARET,
    ACTIONS(2145), 1,
      anon_sym_PIPE,
    ACTIONS(2149), 1,
      anon_sym_PERCENT,
    ACTIONS(2151), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2159), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2161), 1,
      sym__ternary_qmark,
    ACTIONS(2260), 1,
      anon_sym_COLON,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2129), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2139), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(2147), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2155), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2157), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(2131), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(2153), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [41051] = 25,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(1784), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1786), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1788), 1,
      anon_sym_GT_GT,
    ACTIONS(1792), 1,
      anon_sym_AMP,
    ACTIONS(1794), 1,
      anon_sym_CARET,
    ACTIONS(1796), 1,
      anon_sym_PIPE,
    ACTIONS(1800), 1,
      anon_sym_PERCENT,
    ACTIONS(1802), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1810), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1812), 1,
      sym__ternary_qmark,
    ACTIONS(2262), 1,
      anon_sym_RBRACK,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1778), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1790), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1798), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1806), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1808), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1782), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1804), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [41138] = 25,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(1784), 1,
      anon_sym_AMP_AMP,
    ACTIONS(1786), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(1788), 1,
      anon_sym_GT_GT,
    ACTIONS(1792), 1,
      anon_sym_AMP,
    ACTIONS(1794), 1,
      anon_sym_CARET,
    ACTIONS(1796), 1,
      anon_sym_PIPE,
    ACTIONS(1800), 1,
      anon_sym_PERCENT,
    ACTIONS(1802), 1,
      anon_sym_STAR_STAR,
    ACTIONS(1810), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(1812), 1,
      sym__ternary_qmark,
    ACTIONS(2264), 1,
      anon_sym_RBRACK,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(1778), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(1790), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(1798), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(1806), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(1808), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(1782), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(1804), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [41225] = 25,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(2133), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2135), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2137), 1,
      anon_sym_GT_GT,
    ACTIONS(2141), 1,
      anon_sym_AMP,
    ACTIONS(2143), 1,
      anon_sym_CARET,
    ACTIONS(2145), 1,
      anon_sym_PIPE,
    ACTIONS(2149), 1,
      anon_sym_PERCENT,
    ACTIONS(2151), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2159), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2161), 1,
      sym__ternary_qmark,
    ACTIONS(2266), 1,
      anon_sym_COLON,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2129), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2139), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(2147), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2155), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2157), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(2131), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(2153), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [41312] = 26,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(1962), 1,
      anon_sym_of,
    ACTIONS(2203), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2205), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2207), 1,
      anon_sym_GT_GT,
    ACTIONS(2211), 1,
      anon_sym_AMP,
    ACTIONS(2213), 1,
      anon_sym_CARET,
    ACTIONS(2215), 1,
      anon_sym_PIPE,
    ACTIONS(2219), 1,
      anon_sym_PERCENT,
    ACTIONS(2221), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2229), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2231), 1,
      sym__ternary_qmark,
    ACTIONS(2268), 1,
      anon_sym_in,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2199), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2201), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(2209), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(2217), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2225), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2227), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(2223), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [41401] = 24,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1569), 1,
      anon_sym_LBRACK,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(1587), 1,
      sym_optional_chain,
    ACTIONS(2133), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2135), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2137), 1,
      anon_sym_GT_GT,
    ACTIONS(2141), 1,
      anon_sym_AMP,
    ACTIONS(2143), 1,
      anon_sym_CARET,
    ACTIONS(2145), 1,
      anon_sym_PIPE,
    ACTIONS(2149), 1,
      anon_sym_PERCENT,
    ACTIONS(2151), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2159), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2161), 1,
      sym__ternary_qmark,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2129), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2139), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(2147), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2155), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2157), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(2131), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(2153), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [41485] = 24,
    ACTIONS(1818), 1,
      anon_sym_LPAREN,
    ACTIONS(1820), 1,
      anon_sym_LBRACK,
    ACTIONS(1822), 1,
      anon_sym_DOT,
    ACTIONS(1838), 1,
      sym_optional_chain,
    ACTIONS(2133), 1,
      anon_sym_AMP_AMP,
    ACTIONS(2135), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(2137), 1,
      anon_sym_GT_GT,
    ACTIONS(2141), 1,
      anon_sym_AMP,
    ACTIONS(2143), 1,
      anon_sym_CARET,
    ACTIONS(2145), 1,
      anon_sym_PIPE,
    ACTIONS(2149), 1,
      anon_sym_PERCENT,
    ACTIONS(2151), 1,
      anon_sym_STAR_STAR,
    ACTIONS(2159), 1,
      anon_sym_QMARK_QMARK,
    ACTIONS(2161), 1,
      sym__ternary_qmark,
    STATE(797), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1589), 2,
      anon_sym_PLUS_PLUS,
      anon_sym_DASH_DASH,
    ACTIONS(2129), 2,
      anon_sym_STAR,
      anon_sym_SLASH,
    ACTIONS(2139), 2,
      anon_sym_GT_GT_GT,
      anon_sym_LT_LT,
    ACTIONS(2147), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2155), 2,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
    ACTIONS(2157), 2,
      anon_sym_EQ_EQ_EQ,
      anon_sym_BANG_EQ_EQ,
    ACTIONS(2131), 3,
      anon_sym_in,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(2153), 3,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_instanceof,
  [41569] = 20,
    ACTIONS(93), 1,
      anon_sym_AT,
    ACTIONS(101), 1,
      anon_sym_STAR,
    ACTIONS(123), 1,
      aux_sym_method_definition_token1,
    ACTIONS(914), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(2273), 1,
      anon_sym_COMMA,
    ACTIONS(2275), 1,
      anon_sym_RBRACE,
    ACTIONS(2277), 1,
      anon_sym_LBRACK,
    ACTIONS(2279), 1,
      anon_sym_async,
    ACTIONS(2283), 1,
      anon_sym_static,
    STATE(1019), 1,
      aux_sym_export_statement_repeat1,
    STATE(1097), 1,
      sym_decorator,
    STATE(1449), 1,
      aux_sym_object_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2281), 2,
      sym_number,
      sym_private_property_identifier,
    ACTIONS(2285), 2,
      anon_sym_get,
      anon_sym_set,
    STATE(1402), 3,
      sym_spread_element,
      sym_method_definition,
      sym_pair,
    STATE(1448), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(2271), 4,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      sym_identifier,
  [41640] = 21,
    ACTIONS(93), 1,
      anon_sym_AT,
    ACTIONS(2289), 1,
      anon_sym_STAR,
    ACTIONS(2291), 1,
      anon_sym_RBRACE,
    ACTIONS(2293), 1,
      anon_sym_SEMI,
    ACTIONS(2295), 1,
      anon_sym_LBRACK,
    ACTIONS(2297), 1,
      anon_sym_DQUOTE,
    ACTIONS(2299), 1,
      anon_sym_SQUOTE,
    ACTIONS(2301), 1,
      anon_sym_async,
    ACTIONS(2305), 1,
      anon_sym_static,
    ACTIONS(2307), 1,
      aux_sym_method_definition_token1,
    STATE(985), 1,
      aux_sym_class_body_repeat1,
    STATE(1025), 1,
      aux_sym_export_statement_repeat1,
    STATE(1071), 1,
      sym_method_definition,
    STATE(1080), 1,
      sym_class_static_block,
    STATE(1097), 1,
      sym_decorator,
    STATE(1526), 1,
      sym_field_definition,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2303), 2,
      sym_number,
      sym_private_property_identifier,
    ACTIONS(2309), 2,
      anon_sym_get,
      anon_sym_set,
    STATE(1228), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(2287), 4,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      sym_identifier,
  [41712] = 21,
    ACTIONS(93), 1,
      anon_sym_AT,
    ACTIONS(2289), 1,
      anon_sym_STAR,
    ACTIONS(2295), 1,
      anon_sym_LBRACK,
    ACTIONS(2297), 1,
      anon_sym_DQUOTE,
    ACTIONS(2299), 1,
      anon_sym_SQUOTE,
    ACTIONS(2301), 1,
      anon_sym_async,
    ACTIONS(2305), 1,
      anon_sym_static,
    ACTIONS(2307), 1,
      aux_sym_method_definition_token1,
    ACTIONS(2311), 1,
      anon_sym_RBRACE,
    ACTIONS(2313), 1,
      anon_sym_SEMI,
    STATE(984), 1,
      aux_sym_class_body_repeat1,
    STATE(1025), 1,
      aux_sym_export_statement_repeat1,
    STATE(1071), 1,
      sym_method_definition,
    STATE(1080), 1,
      sym_class_static_block,
    STATE(1097), 1,
      sym_decorator,
    STATE(1526), 1,
      sym_field_definition,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2303), 2,
      sym_number,
      sym_private_property_identifier,
    ACTIONS(2309), 2,
      anon_sym_get,
      anon_sym_set,
    STATE(1228), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(2287), 4,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      sym_identifier,
  [41784] = 21,
    ACTIONS(93), 1,
      anon_sym_AT,
    ACTIONS(2289), 1,
      anon_sym_STAR,
    ACTIONS(2293), 1,
      anon_sym_SEMI,
    ACTIONS(2295), 1,
      anon_sym_LBRACK,
    ACTIONS(2297), 1,
      anon_sym_DQUOTE,
    ACTIONS(2299), 1,
      anon_sym_SQUOTE,
    ACTIONS(2301), 1,
      anon_sym_async,
    ACTIONS(2305), 1,
      anon_sym_static,
    ACTIONS(2307), 1,
      aux_sym_method_definition_token1,
    ACTIONS(2315), 1,
      anon_sym_RBRACE,
    STATE(985), 1,
      aux_sym_class_body_repeat1,
    STATE(1025), 1,
      aux_sym_export_statement_repeat1,
    STATE(1071), 1,
      sym_method_definition,
    STATE(1080), 1,
      sym_class_static_block,
    STATE(1097), 1,
      sym_decorator,
    STATE(1526), 1,
      sym_field_definition,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2303), 2,
      sym_number,
      sym_private_property_identifier,
    ACTIONS(2309), 2,
      anon_sym_get,
      anon_sym_set,
    STATE(1228), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(2287), 4,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      sym_identifier,
  [41856] = 21,
    ACTIONS(2320), 1,
      anon_sym_STAR,
    ACTIONS(2323), 1,
      anon_sym_RBRACE,
    ACTIONS(2325), 1,
      anon_sym_SEMI,
    ACTIONS(2328), 1,
      anon_sym_LBRACK,
    ACTIONS(2331), 1,
      anon_sym_DQUOTE,
    ACTIONS(2334), 1,
      anon_sym_SQUOTE,
    ACTIONS(2337), 1,
      anon_sym_async,
    ACTIONS(2343), 1,
      anon_sym_AT,
    ACTIONS(2346), 1,
      anon_sym_static,
    ACTIONS(2349), 1,
      aux_sym_method_definition_token1,
    STATE(985), 1,
      aux_sym_class_body_repeat1,
    STATE(1025), 1,
      aux_sym_export_statement_repeat1,
    STATE(1071), 1,
      sym_method_definition,
    STATE(1080), 1,
      sym_class_static_block,
    STATE(1097), 1,
      sym_decorator,
    STATE(1526), 1,
      sym_field_definition,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2340), 2,
      sym_number,
      sym_private_property_identifier,
    ACTIONS(2352), 2,
      anon_sym_get,
      anon_sym_set,
    STATE(1228), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(2317), 4,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      sym_identifier,
  [41928] = 14,
    ACTIONS(883), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(1969), 1,
      anon_sym_LBRACE,
    ACTIONS(1973), 1,
      anon_sym_LBRACK,
    ACTIONS(2357), 1,
      anon_sym_COMMA,
    ACTIONS(2359), 1,
      anon_sym_RBRACE,
    STATE(1466), 1,
      aux_sym_object_pattern_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2361), 2,
      sym_number,
      sym_private_property_identifier,
    STATE(1457), 3,
      sym_object_assignment_pattern,
      sym_rest_pattern,
      sym_pair_pattern,
    STATE(1788), 3,
      sym_object_pattern,
      sym_array_pattern,
      sym__destructuring_pattern,
    STATE(1796), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(2355), 8,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [41986] = 14,
    ACTIONS(883), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(1969), 1,
      anon_sym_LBRACE,
    ACTIONS(1973), 1,
      anon_sym_LBRACK,
    ACTIONS(2357), 1,
      anon_sym_COMMA,
    ACTIONS(2365), 1,
      anon_sym_RBRACE,
    STATE(1476), 1,
      aux_sym_object_pattern_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2361), 2,
      sym_number,
      sym_private_property_identifier,
    STATE(1454), 3,
      sym_object_assignment_pattern,
      sym_rest_pattern,
      sym_pair_pattern,
    STATE(1788), 3,
      sym_object_pattern,
      sym_array_pattern,
      sym__destructuring_pattern,
    STATE(1796), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(2363), 8,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [42044] = 21,
    ACTIONS(93), 1,
      anon_sym_AT,
    ACTIONS(2289), 1,
      anon_sym_STAR,
    ACTIONS(2293), 1,
      anon_sym_SEMI,
    ACTIONS(2295), 1,
      anon_sym_LBRACK,
    ACTIONS(2297), 1,
      anon_sym_DQUOTE,
    ACTIONS(2299), 1,
      anon_sym_SQUOTE,
    ACTIONS(2301), 1,
      anon_sym_async,
    ACTIONS(2305), 1,
      anon_sym_static,
    ACTIONS(2307), 1,
      aux_sym_method_definition_token1,
    ACTIONS(2367), 1,
      anon_sym_RBRACE,
    STATE(985), 1,
      aux_sym_class_body_repeat1,
    STATE(1025), 1,
      aux_sym_export_statement_repeat1,
    STATE(1071), 1,
      sym_method_definition,
    STATE(1080), 1,
      sym_class_static_block,
    STATE(1097), 1,
      sym_decorator,
    STATE(1526), 1,
      sym_field_definition,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2303), 2,
      sym_number,
      sym_private_property_identifier,
    ACTIONS(2309), 2,
      anon_sym_get,
      anon_sym_set,
    STATE(1228), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(2287), 4,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      sym_identifier,
  [42116] = 18,
    ACTIONS(93), 1,
      anon_sym_AT,
    ACTIONS(101), 1,
      anon_sym_STAR,
    ACTIONS(123), 1,
      aux_sym_method_definition_token1,
    ACTIONS(914), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(2277), 1,
      anon_sym_LBRACK,
    ACTIONS(2373), 1,
      anon_sym_async,
    ACTIONS(2375), 1,
      anon_sym_static,
    STATE(1019), 1,
      aux_sym_export_statement_repeat1,
    STATE(1097), 1,
      sym_decorator,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2281), 2,
      sym_number,
      sym_private_property_identifier,
    ACTIONS(2371), 2,
      anon_sym_COMMA,
      anon_sym_RBRACE,
    ACTIONS(2377), 2,
      anon_sym_get,
      anon_sym_set,
    STATE(1448), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    STATE(1516), 3,
      sym_spread_element,
      sym_method_definition,
      sym_pair,
    ACTIONS(2369), 4,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      sym_identifier,
  [42182] = 21,
    ACTIONS(93), 1,
      anon_sym_AT,
    ACTIONS(2289), 1,
      anon_sym_STAR,
    ACTIONS(2295), 1,
      anon_sym_LBRACK,
    ACTIONS(2297), 1,
      anon_sym_DQUOTE,
    ACTIONS(2299), 1,
      anon_sym_SQUOTE,
    ACTIONS(2301), 1,
      anon_sym_async,
    ACTIONS(2305), 1,
      anon_sym_static,
    ACTIONS(2307), 1,
      aux_sym_method_definition_token1,
    ACTIONS(2379), 1,
      anon_sym_RBRACE,
    ACTIONS(2381), 1,
      anon_sym_SEMI,
    STATE(988), 1,
      aux_sym_class_body_repeat1,
    STATE(1025), 1,
      aux_sym_export_statement_repeat1,
    STATE(1071), 1,
      sym_method_definition,
    STATE(1080), 1,
      sym_class_static_block,
    STATE(1097), 1,
      sym_decorator,
    STATE(1526), 1,
      sym_field_definition,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2303), 2,
      sym_number,
      sym_private_property_identifier,
    ACTIONS(2309), 2,
      anon_sym_get,
      anon_sym_set,
    STATE(1228), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(2287), 4,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      sym_identifier,
  [42254] = 21,
    ACTIONS(93), 1,
      anon_sym_AT,
    ACTIONS(2289), 1,
      anon_sym_STAR,
    ACTIONS(2295), 1,
      anon_sym_LBRACK,
    ACTIONS(2297), 1,
      anon_sym_DQUOTE,
    ACTIONS(2299), 1,
      anon_sym_SQUOTE,
    ACTIONS(2301), 1,
      anon_sym_async,
    ACTIONS(2305), 1,
      anon_sym_static,
    ACTIONS(2307), 1,
      aux_sym_method_definition_token1,
    ACTIONS(2383), 1,
      anon_sym_RBRACE,
    ACTIONS(2385), 1,
      anon_sym_SEMI,
    STATE(982), 1,
      aux_sym_class_body_repeat1,
    STATE(1025), 1,
      aux_sym_export_statement_repeat1,
    STATE(1071), 1,
      sym_method_definition,
    STATE(1080), 1,
      sym_class_static_block,
    STATE(1097), 1,
      sym_decorator,
    STATE(1526), 1,
      sym_field_definition,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2303), 2,
      sym_number,
      sym_private_property_identifier,
    ACTIONS(2309), 2,
      anon_sym_get,
      anon_sym_set,
    STATE(1228), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(2287), 4,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      sym_identifier,
  [42326] = 12,
    ACTIONS(883), 1,
      anon_sym_DOT_DOT_DOT,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(1969), 1,
      anon_sym_LBRACE,
    ACTIONS(1973), 1,
      anon_sym_LBRACK,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2361), 2,
      sym_number,
      sym_private_property_identifier,
    ACTIONS(2389), 2,
      anon_sym_COMMA,
      anon_sym_RBRACE,
    STATE(1514), 3,
      sym_object_assignment_pattern,
      sym_rest_pattern,
      sym_pair_pattern,
    STATE(1788), 3,
      sym_object_pattern,
      sym_array_pattern,
      sym__destructuring_pattern,
    STATE(1796), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(2387), 8,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [42379] = 16,
    ACTIONS(103), 1,
      anon_sym_COMMA,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(1087), 1,
      anon_sym_async,
    ACTIONS(2277), 1,
      anon_sym_LBRACK,
    ACTIONS(2391), 1,
      anon_sym_STAR,
    ACTIONS(2393), 1,
      anon_sym_RBRACE,
    ACTIONS(2397), 1,
      anon_sym_EQ,
    STATE(1418), 1,
      aux_sym_object_repeat1,
    STATE(1452), 1,
      aux_sym_object_pattern_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1081), 2,
      sym_number,
      sym_private_property_identifier,
    ACTIONS(1089), 2,
      anon_sym_get,
      anon_sym_set,
    ACTIONS(2395), 2,
      anon_sym_LPAREN,
      anon_sym_COLON,
    STATE(1710), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(1085), 5,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      sym_identifier,
      anon_sym_static,
  [42438] = 15,
    ACTIONS(103), 1,
      anon_sym_COMMA,
    ACTIONS(171), 1,
      anon_sym_RBRACE,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(2277), 1,
      anon_sym_LBRACK,
    ACTIONS(2391), 1,
      anon_sym_STAR,
    ACTIONS(2397), 1,
      anon_sym_EQ,
    STATE(1452), 1,
      aux_sym_object_pattern_repeat1,
    STATE(1487), 1,
      aux_sym_object_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1081), 2,
      sym_number,
      sym_private_property_identifier,
    ACTIONS(1089), 2,
      anon_sym_get,
      anon_sym_set,
    ACTIONS(2395), 2,
      anon_sym_LPAREN,
      anon_sym_COLON,
    STATE(1710), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(1085), 6,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
  [42495] = 16,
    ACTIONS(103), 1,
      anon_sym_COMMA,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(1087), 1,
      anon_sym_async,
    ACTIONS(2277), 1,
      anon_sym_LBRACK,
    ACTIONS(2391), 1,
      anon_sym_STAR,
    ACTIONS(2397), 1,
      anon_sym_EQ,
    ACTIONS(2399), 1,
      anon_sym_RBRACE,
    STATE(1418), 1,
      aux_sym_object_repeat1,
    STATE(1452), 1,
      aux_sym_object_pattern_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1081), 2,
      sym_number,
      sym_private_property_identifier,
    ACTIONS(1089), 2,
      anon_sym_get,
      anon_sym_set,
    ACTIONS(2395), 2,
      anon_sym_LPAREN,
      anon_sym_COLON,
    STATE(1710), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(1085), 5,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      sym_identifier,
      anon_sym_static,
  [42554] = 15,
    ACTIONS(103), 1,
      anon_sym_COMMA,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(2277), 1,
      anon_sym_LBRACK,
    ACTIONS(2391), 1,
      anon_sym_STAR,
    ACTIONS(2397), 1,
      anon_sym_EQ,
    ACTIONS(2401), 1,
      anon_sym_RBRACE,
    STATE(1418), 1,
      aux_sym_object_repeat1,
    STATE(1452), 1,
      aux_sym_object_pattern_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1081), 2,
      sym_number,
      sym_private_property_identifier,
    ACTIONS(1089), 2,
      anon_sym_get,
      anon_sym_set,
    ACTIONS(2395), 2,
      anon_sym_LPAREN,
      anon_sym_COLON,
    STATE(1710), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(1085), 6,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
  [42611] = 16,
    ACTIONS(103), 1,
      anon_sym_COMMA,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(1087), 1,
      anon_sym_async,
    ACTIONS(2277), 1,
      anon_sym_LBRACK,
    ACTIONS(2391), 1,
      anon_sym_STAR,
    ACTIONS(2397), 1,
      anon_sym_EQ,
    ACTIONS(2401), 1,
      anon_sym_RBRACE,
    STATE(1418), 1,
      aux_sym_object_repeat1,
    STATE(1452), 1,
      aux_sym_object_pattern_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1081), 2,
      sym_number,
      sym_private_property_identifier,
    ACTIONS(1089), 2,
      anon_sym_get,
      anon_sym_set,
    ACTIONS(2395), 2,
      anon_sym_LPAREN,
      anon_sym_COLON,
    STATE(1710), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(1085), 5,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      sym_identifier,
      anon_sym_static,
  [42670] = 15,
    ACTIONS(103), 1,
      anon_sym_COMMA,
    ACTIONS(351), 1,
      anon_sym_RBRACE,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(2277), 1,
      anon_sym_LBRACK,
    ACTIONS(2391), 1,
      anon_sym_STAR,
    ACTIONS(2397), 1,
      anon_sym_EQ,
    STATE(1418), 1,
      aux_sym_object_repeat1,
    STATE(1452), 1,
      aux_sym_object_pattern_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1081), 2,
      sym_number,
      sym_private_property_identifier,
    ACTIONS(1089), 2,
      anon_sym_get,
      anon_sym_set,
    ACTIONS(2395), 2,
      anon_sym_LPAREN,
      anon_sym_COLON,
    STATE(1710), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(1085), 6,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
  [42727] = 16,
    ACTIONS(103), 1,
      anon_sym_COMMA,
    ACTIONS(351), 1,
      anon_sym_RBRACE,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(1087), 1,
      anon_sym_async,
    ACTIONS(2277), 1,
      anon_sym_LBRACK,
    ACTIONS(2391), 1,
      anon_sym_STAR,
    ACTIONS(2397), 1,
      anon_sym_EQ,
    STATE(1418), 1,
      aux_sym_object_repeat1,
    STATE(1452), 1,
      aux_sym_object_pattern_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1081), 2,
      sym_number,
      sym_private_property_identifier,
    ACTIONS(1089), 2,
      anon_sym_get,
      anon_sym_set,
    ACTIONS(2395), 2,
      anon_sym_LPAREN,
      anon_sym_COLON,
    STATE(1710), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(1085), 5,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      sym_identifier,
      anon_sym_static,
  [42786] = 15,
    ACTIONS(103), 1,
      anon_sym_COMMA,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(2277), 1,
      anon_sym_LBRACK,
    ACTIONS(2391), 1,
      anon_sym_STAR,
    ACTIONS(2397), 1,
      anon_sym_EQ,
    ACTIONS(2399), 1,
      anon_sym_RBRACE,
    STATE(1418), 1,
      aux_sym_object_repeat1,
    STATE(1452), 1,
      aux_sym_object_pattern_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1081), 2,
      sym_number,
      sym_private_property_identifier,
    ACTIONS(1089), 2,
      anon_sym_get,
      anon_sym_set,
    ACTIONS(2395), 2,
      anon_sym_LPAREN,
      anon_sym_COLON,
    STATE(1710), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(1085), 6,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
  [42843] = 15,
    ACTIONS(103), 1,
      anon_sym_COMMA,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(2277), 1,
      anon_sym_LBRACK,
    ACTIONS(2391), 1,
      anon_sym_STAR,
    ACTIONS(2393), 1,
      anon_sym_RBRACE,
    ACTIONS(2397), 1,
      anon_sym_EQ,
    STATE(1418), 1,
      aux_sym_object_repeat1,
    STATE(1452), 1,
      aux_sym_object_pattern_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1081), 2,
      sym_number,
      sym_private_property_identifier,
    ACTIONS(1089), 2,
      anon_sym_get,
      anon_sym_set,
    ACTIONS(2395), 2,
      anon_sym_LPAREN,
      anon_sym_COLON,
    STATE(1710), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(1085), 6,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
  [42900] = 16,
    ACTIONS(103), 1,
      anon_sym_COMMA,
    ACTIONS(216), 1,
      anon_sym_RBRACE,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(1087), 1,
      anon_sym_async,
    ACTIONS(2277), 1,
      anon_sym_LBRACK,
    ACTIONS(2391), 1,
      anon_sym_STAR,
    ACTIONS(2397), 1,
      anon_sym_EQ,
    STATE(1418), 1,
      aux_sym_object_repeat1,
    STATE(1452), 1,
      aux_sym_object_pattern_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1081), 2,
      sym_number,
      sym_private_property_identifier,
    ACTIONS(1089), 2,
      anon_sym_get,
      anon_sym_set,
    ACTIONS(2395), 2,
      anon_sym_LPAREN,
      anon_sym_COLON,
    STATE(1710), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(1085), 5,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      sym_identifier,
      anon_sym_static,
  [42959] = 16,
    ACTIONS(103), 1,
      anon_sym_COMMA,
    ACTIONS(171), 1,
      anon_sym_RBRACE,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(1087), 1,
      anon_sym_async,
    ACTIONS(2277), 1,
      anon_sym_LBRACK,
    ACTIONS(2391), 1,
      anon_sym_STAR,
    ACTIONS(2397), 1,
      anon_sym_EQ,
    STATE(1452), 1,
      aux_sym_object_pattern_repeat1,
    STATE(1487), 1,
      aux_sym_object_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1081), 2,
      sym_number,
      sym_private_property_identifier,
    ACTIONS(1089), 2,
      anon_sym_get,
      anon_sym_set,
    ACTIONS(2395), 2,
      anon_sym_LPAREN,
      anon_sym_COLON,
    STATE(1710), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(1085), 5,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      sym_identifier,
      anon_sym_static,
  [43018] = 15,
    ACTIONS(103), 1,
      anon_sym_COMMA,
    ACTIONS(216), 1,
      anon_sym_RBRACE,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(2277), 1,
      anon_sym_LBRACK,
    ACTIONS(2391), 1,
      anon_sym_STAR,
    ACTIONS(2397), 1,
      anon_sym_EQ,
    STATE(1418), 1,
      aux_sym_object_repeat1,
    STATE(1452), 1,
      aux_sym_object_pattern_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1081), 2,
      sym_number,
      sym_private_property_identifier,
    ACTIONS(1089), 2,
      anon_sym_get,
      anon_sym_set,
    ACTIONS(2395), 2,
      anon_sym_LPAREN,
      anon_sym_COLON,
    STATE(1710), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(1085), 6,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
  [43075] = 13,
    ACTIONS(103), 1,
      anon_sym_COMMA,
    ACTIONS(216), 1,
      anon_sym_RBRACE,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(2277), 1,
      anon_sym_LBRACK,
    ACTIONS(2397), 1,
      anon_sym_EQ,
    STATE(1418), 1,
      aux_sym_object_repeat1,
    STATE(1452), 1,
      aux_sym_object_pattern_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1081), 2,
      sym_number,
      sym_private_property_identifier,
    ACTIONS(2395), 2,
      anon_sym_LPAREN,
      anon_sym_COLON,
    STATE(1710), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(1085), 8,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [43127] = 13,
    ACTIONS(103), 1,
      anon_sym_COMMA,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(2277), 1,
      anon_sym_LBRACK,
    ACTIONS(2397), 1,
      anon_sym_EQ,
    ACTIONS(2401), 1,
      anon_sym_RBRACE,
    STATE(1418), 1,
      aux_sym_object_repeat1,
    STATE(1452), 1,
      aux_sym_object_pattern_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1081), 2,
      sym_number,
      sym_private_property_identifier,
    ACTIONS(2395), 2,
      anon_sym_LPAREN,
      anon_sym_COLON,
    STATE(1710), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(1085), 8,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [43179] = 13,
    ACTIONS(103), 1,
      anon_sym_COMMA,
    ACTIONS(171), 1,
      anon_sym_RBRACE,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(2277), 1,
      anon_sym_LBRACK,
    ACTIONS(2397), 1,
      anon_sym_EQ,
    STATE(1452), 1,
      aux_sym_object_pattern_repeat1,
    STATE(1487), 1,
      aux_sym_object_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1081), 2,
      sym_number,
      sym_private_property_identifier,
    ACTIONS(2395), 2,
      anon_sym_LPAREN,
      anon_sym_COLON,
    STATE(1710), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(1085), 8,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [43231] = 13,
    ACTIONS(103), 1,
      anon_sym_COMMA,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(2277), 1,
      anon_sym_LBRACK,
    ACTIONS(2397), 1,
      anon_sym_EQ,
    ACTIONS(2399), 1,
      anon_sym_RBRACE,
    STATE(1418), 1,
      aux_sym_object_repeat1,
    STATE(1452), 1,
      aux_sym_object_pattern_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1081), 2,
      sym_number,
      sym_private_property_identifier,
    ACTIONS(2395), 2,
      anon_sym_LPAREN,
      anon_sym_COLON,
    STATE(1710), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(1085), 8,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [43283] = 14,
    ACTIONS(2295), 1,
      anon_sym_LBRACK,
    ACTIONS(2297), 1,
      anon_sym_DQUOTE,
    ACTIONS(2299), 1,
      anon_sym_SQUOTE,
    ACTIONS(2403), 1,
      anon_sym_STAR,
    ACTIONS(2405), 1,
      anon_sym_LBRACE,
    ACTIONS(2407), 1,
      anon_sym_async,
    ACTIONS(2413), 1,
      sym__automatic_semicolon,
    STATE(1081), 1,
      sym_statement_block,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2409), 2,
      sym_number,
      sym_private_property_identifier,
    ACTIONS(2411), 2,
      anon_sym_get,
      anon_sym_set,
    ACTIONS(2395), 3,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_EQ,
    STATE(1231), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(2287), 5,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      sym_identifier,
      anon_sym_static,
  [43337] = 13,
    ACTIONS(103), 1,
      anon_sym_COMMA,
    ACTIONS(351), 1,
      anon_sym_RBRACE,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(2277), 1,
      anon_sym_LBRACK,
    ACTIONS(2397), 1,
      anon_sym_EQ,
    STATE(1418), 1,
      aux_sym_object_repeat1,
    STATE(1452), 1,
      aux_sym_object_pattern_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1081), 2,
      sym_number,
      sym_private_property_identifier,
    ACTIONS(2395), 2,
      anon_sym_LPAREN,
      anon_sym_COLON,
    STATE(1710), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(1085), 8,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [43389] = 13,
    ACTIONS(103), 1,
      anon_sym_COMMA,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(2277), 1,
      anon_sym_LBRACK,
    ACTIONS(2393), 1,
      anon_sym_RBRACE,
    ACTIONS(2397), 1,
      anon_sym_EQ,
    STATE(1418), 1,
      aux_sym_object_repeat1,
    STATE(1452), 1,
      aux_sym_object_pattern_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1081), 2,
      sym_number,
      sym_private_property_identifier,
    ACTIONS(2395), 2,
      anon_sym_LPAREN,
      anon_sym_COLON,
    STATE(1710), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(1085), 8,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [43441] = 13,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(2273), 1,
      anon_sym_COMMA,
    ACTIONS(2277), 1,
      anon_sym_LBRACK,
    ACTIONS(2391), 1,
      anon_sym_STAR,
    ACTIONS(2416), 1,
      anon_sym_RBRACE,
    STATE(1464), 1,
      aux_sym_object_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1081), 2,
      sym_number,
      sym_private_property_identifier,
    ACTIONS(1089), 2,
      anon_sym_get,
      anon_sym_set,
    ACTIONS(2395), 2,
      anon_sym_LPAREN,
      anon_sym_COLON,
    STATE(1710), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(1085), 6,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
  [43492] = 14,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(1087), 1,
      anon_sym_async,
    ACTIONS(2273), 1,
      anon_sym_COMMA,
    ACTIONS(2277), 1,
      anon_sym_LBRACK,
    ACTIONS(2391), 1,
      anon_sym_STAR,
    ACTIONS(2416), 1,
      anon_sym_RBRACE,
    STATE(1464), 1,
      aux_sym_object_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1081), 2,
      sym_number,
      sym_private_property_identifier,
    ACTIONS(1089), 2,
      anon_sym_get,
      anon_sym_set,
    ACTIONS(2395), 2,
      anon_sym_LPAREN,
      anon_sym_COLON,
    STATE(1710), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(1085), 5,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      sym_identifier,
      anon_sym_static,
  [43545] = 18,
    ACTIONS(93), 1,
      anon_sym_AT,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(2277), 1,
      anon_sym_LBRACK,
    ACTIONS(2418), 1,
      anon_sym_export,
    ACTIONS(2420), 1,
      anon_sym_STAR,
    ACTIONS(2422), 1,
      anon_sym_class,
    ACTIONS(2424), 1,
      anon_sym_async,
    ACTIONS(2428), 1,
      anon_sym_static,
    ACTIONS(2430), 1,
      aux_sym_method_definition_token1,
    ACTIONS(2432), 1,
      anon_sym_get,
    ACTIONS(2434), 1,
      anon_sym_set,
    STATE(1041), 1,
      aux_sym_export_statement_repeat1,
    STATE(1097), 1,
      sym_decorator,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2426), 2,
      sym_number,
      sym_private_property_identifier,
    ACTIONS(1085), 3,
      anon_sym_let,
      anon_sym_await,
      sym_identifier,
    STATE(1682), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
  [43606] = 13,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(1087), 1,
      anon_sym_async,
    ACTIONS(2277), 1,
      anon_sym_LBRACK,
    ACTIONS(2391), 1,
      anon_sym_STAR,
    ACTIONS(2397), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1081), 2,
      sym_number,
      sym_private_property_identifier,
    ACTIONS(1089), 2,
      anon_sym_get,
      anon_sym_set,
    ACTIONS(2395), 2,
      anon_sym_LPAREN,
      anon_sym_COLON,
    ACTIONS(2436), 2,
      anon_sym_COMMA,
      anon_sym_RBRACE,
    STATE(1710), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(1085), 5,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      sym_identifier,
      anon_sym_static,
  [43657] = 12,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(2277), 1,
      anon_sym_LBRACK,
    ACTIONS(2391), 1,
      anon_sym_STAR,
    ACTIONS(2397), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1081), 2,
      sym_number,
      sym_private_property_identifier,
    ACTIONS(1089), 2,
      anon_sym_get,
      anon_sym_set,
    ACTIONS(2395), 2,
      anon_sym_LPAREN,
      anon_sym_COLON,
    ACTIONS(2436), 2,
      anon_sym_COMMA,
      anon_sym_RBRACE,
    STATE(1710), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(1085), 6,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
  [43706] = 11,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(2277), 1,
      anon_sym_LBRACK,
    ACTIONS(2439), 1,
      anon_sym_STAR,
    ACTIONS(2443), 1,
      anon_sym_get,
    ACTIONS(2445), 1,
      anon_sym_set,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2441), 2,
      sym_number,
      sym_private_property_identifier,
    STATE(1598), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(2395), 4,
      sym__automatic_semicolon,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_EQ,
    ACTIONS(1085), 6,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
  [43752] = 11,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(2277), 1,
      anon_sym_LBRACK,
    ACTIONS(2447), 1,
      anon_sym_STAR,
    ACTIONS(2451), 1,
      anon_sym_get,
    ACTIONS(2453), 1,
      anon_sym_set,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2449), 2,
      sym_number,
      sym_private_property_identifier,
    STATE(1582), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(2395), 4,
      sym__automatic_semicolon,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_EQ,
    ACTIONS(1085), 6,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
  [43798] = 16,
    ACTIONS(93), 1,
      anon_sym_AT,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(2277), 1,
      anon_sym_LBRACK,
    ACTIONS(2420), 1,
      anon_sym_STAR,
    ACTIONS(2424), 1,
      anon_sym_async,
    ACTIONS(2428), 1,
      anon_sym_static,
    ACTIONS(2430), 1,
      aux_sym_method_definition_token1,
    ACTIONS(2432), 1,
      anon_sym_get,
    ACTIONS(2434), 1,
      anon_sym_set,
    STATE(1041), 1,
      aux_sym_export_statement_repeat1,
    STATE(1097), 1,
      sym_decorator,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2426), 2,
      sym_number,
      sym_private_property_identifier,
    STATE(1682), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(1085), 4,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      sym_identifier,
  [43854] = 11,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(2273), 1,
      anon_sym_COMMA,
    ACTIONS(2277), 1,
      anon_sym_LBRACK,
    ACTIONS(2416), 1,
      anon_sym_RBRACE,
    STATE(1464), 1,
      aux_sym_object_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1081), 2,
      sym_number,
      sym_private_property_identifier,
    ACTIONS(2395), 2,
      anon_sym_LPAREN,
      anon_sym_COLON,
    STATE(1710), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(1085), 8,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [43900] = 11,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(2277), 1,
      anon_sym_LBRACK,
    ACTIONS(2391), 1,
      anon_sym_STAR,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1081), 2,
      sym_number,
      sym_private_property_identifier,
    ACTIONS(1089), 2,
      anon_sym_get,
      anon_sym_set,
    ACTIONS(2395), 2,
      anon_sym_LPAREN,
      anon_sym_COLON,
    ACTIONS(2455), 2,
      anon_sym_COMMA,
      anon_sym_RBRACE,
    STATE(1710), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(1085), 6,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
  [43946] = 12,
    ACTIONS(2295), 1,
      anon_sym_LBRACK,
    ACTIONS(2297), 1,
      anon_sym_DQUOTE,
    ACTIONS(2299), 1,
      anon_sym_SQUOTE,
    ACTIONS(2457), 1,
      anon_sym_STAR,
    ACTIONS(2459), 1,
      anon_sym_async,
    ACTIONS(2463), 1,
      anon_sym_get,
    ACTIONS(2465), 1,
      anon_sym_set,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2461), 2,
      sym_number,
      sym_private_property_identifier,
    STATE(1245), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(2395), 4,
      sym__automatic_semicolon,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_EQ,
    ACTIONS(2287), 5,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      sym_identifier,
      anon_sym_static,
  [43994] = 10,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(2277), 1,
      anon_sym_LBRACK,
    ACTIONS(2403), 1,
      anon_sym_STAR,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2467), 2,
      sym_number,
      sym_private_property_identifier,
    ACTIONS(2469), 2,
      anon_sym_get,
      anon_sym_set,
    STATE(1553), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(2395), 4,
      sym__automatic_semicolon,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_EQ,
    ACTIONS(1085), 6,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
  [44038] = 12,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(1087), 1,
      anon_sym_async,
    ACTIONS(2277), 1,
      anon_sym_LBRACK,
    ACTIONS(2391), 1,
      anon_sym_STAR,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1081), 2,
      sym_number,
      sym_private_property_identifier,
    ACTIONS(1089), 2,
      anon_sym_get,
      anon_sym_set,
    ACTIONS(2395), 2,
      anon_sym_LPAREN,
      anon_sym_COLON,
    ACTIONS(2455), 2,
      anon_sym_COMMA,
      anon_sym_RBRACE,
    STATE(1710), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(1085), 5,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      sym_identifier,
      anon_sym_static,
  [44086] = 16,
    ACTIONS(93), 1,
      anon_sym_AT,
    ACTIONS(2295), 1,
      anon_sym_LBRACK,
    ACTIONS(2297), 1,
      anon_sym_DQUOTE,
    ACTIONS(2299), 1,
      anon_sym_SQUOTE,
    ACTIONS(2471), 1,
      anon_sym_STAR,
    ACTIONS(2473), 1,
      anon_sym_async,
    ACTIONS(2477), 1,
      anon_sym_static,
    ACTIONS(2479), 1,
      aux_sym_method_definition_token1,
    ACTIONS(2481), 1,
      anon_sym_get,
    ACTIONS(2483), 1,
      anon_sym_set,
    STATE(1041), 1,
      aux_sym_export_statement_repeat1,
    STATE(1097), 1,
      sym_decorator,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2475), 2,
      sym_number,
      sym_private_property_identifier,
    STATE(1232), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(2287), 4,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      sym_identifier,
  [44142] = 10,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(2277), 1,
      anon_sym_LBRACK,
    ACTIONS(2485), 1,
      anon_sym_STAR,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2487), 2,
      sym_number,
      sym_private_property_identifier,
    ACTIONS(2489), 2,
      anon_sym_get,
      anon_sym_set,
    STATE(1580), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(2395), 4,
      sym__automatic_semicolon,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_EQ,
    ACTIONS(1085), 6,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
  [44186] = 10,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(2277), 1,
      anon_sym_LBRACK,
    ACTIONS(2397), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1081), 2,
      sym_number,
      sym_private_property_identifier,
    ACTIONS(2395), 2,
      anon_sym_LPAREN,
      anon_sym_COLON,
    ACTIONS(2436), 2,
      anon_sym_COMMA,
      anon_sym_RBRACE,
    STATE(1710), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(1085), 8,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [44230] = 8,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(2277), 1,
      anon_sym_LBRACK,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2467), 2,
      sym_number,
      sym_private_property_identifier,
    STATE(1553), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(2395), 4,
      sym__automatic_semicolon,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_EQ,
    ACTIONS(1085), 8,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [44269] = 8,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(2277), 1,
      anon_sym_LBRACK,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2487), 2,
      sym_number,
      sym_private_property_identifier,
    STATE(1580), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(2395), 4,
      sym__automatic_semicolon,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_EQ,
    ACTIONS(1085), 8,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [44308] = 8,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(2277), 1,
      anon_sym_LBRACK,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2491), 2,
      sym_number,
      sym_private_property_identifier,
    STATE(1600), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(2395), 4,
      sym__automatic_semicolon,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_EQ,
    ACTIONS(1085), 8,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [44347] = 8,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(2277), 1,
      anon_sym_LBRACK,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2493), 2,
      sym_number,
      sym_private_property_identifier,
    STATE(1599), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(2395), 4,
      sym__automatic_semicolon,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_EQ,
    ACTIONS(1085), 8,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [44386] = 8,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(2277), 1,
      anon_sym_LBRACK,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2495), 2,
      sym_number,
      sym_private_property_identifier,
    STATE(1584), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(2395), 4,
      sym__automatic_semicolon,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_EQ,
    ACTIONS(1085), 8,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [44425] = 8,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(2277), 1,
      anon_sym_LBRACK,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2497), 2,
      sym_number,
      sym_private_property_identifier,
    STATE(1585), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(2395), 4,
      sym__automatic_semicolon,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_EQ,
    ACTIONS(1085), 8,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [44464] = 6,
    ACTIONS(2503), 1,
      anon_sym_LPAREN,
    ACTIONS(2505), 1,
      anon_sym_DOT,
    STATE(1106), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2501), 8,
      anon_sym_STAR,
      anon_sym_LBRACK,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
      aux_sym_method_definition_token1,
    ACTIONS(2499), 9,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_class,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [44499] = 9,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(2277), 1,
      anon_sym_LBRACK,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1081), 2,
      sym_number,
      sym_private_property_identifier,
    ACTIONS(2395), 2,
      anon_sym_LPAREN,
      anon_sym_COLON,
    ACTIONS(2455), 2,
      anon_sym_COMMA,
      anon_sym_RBRACE,
    STATE(1710), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(1085), 8,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [44540] = 15,
    ACTIONS(93), 1,
      anon_sym_AT,
    ACTIONS(936), 1,
      anon_sym_var,
    ACTIONS(940), 1,
      anon_sym_using,
    ACTIONS(942), 1,
      anon_sym_await,
    ACTIONS(947), 1,
      anon_sym_class,
    ACTIONS(949), 1,
      anon_sym_async,
    ACTIONS(951), 1,
      anon_sym_function,
    ACTIONS(2395), 1,
      anon_sym_LPAREN,
    ACTIONS(2507), 1,
      anon_sym_default,
    STATE(486), 1,
      sym_declaration,
    STATE(1097), 1,
      sym_decorator,
    STATE(1327), 1,
      aux_sym_export_statement_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(938), 2,
      anon_sym_let,
      anon_sym_const,
    STATE(419), 6,
      sym_variable_declaration,
      sym_lexical_declaration,
      sym_using_declaration,
      sym_class_declaration,
      sym_function_declaration,
      sym_generator_function_declaration,
  [44593] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2509), 8,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
    ACTIONS(2511), 11,
      anon_sym_STAR,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
      aux_sym_method_definition_token1,
  [44621] = 14,
    ACTIONS(93), 1,
      anon_sym_AT,
    ACTIONS(936), 1,
      anon_sym_var,
    ACTIONS(940), 1,
      anon_sym_using,
    ACTIONS(942), 1,
      anon_sym_await,
    ACTIONS(947), 1,
      anon_sym_class,
    ACTIONS(949), 1,
      anon_sym_async,
    ACTIONS(951), 1,
      anon_sym_function,
    ACTIONS(2507), 1,
      anon_sym_default,
    STATE(486), 1,
      sym_declaration,
    STATE(1097), 1,
      sym_decorator,
    STATE(1327), 1,
      aux_sym_export_statement_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(938), 2,
      anon_sym_let,
      anon_sym_const,
    STATE(419), 6,
      sym_variable_declaration,
      sym_lexical_declaration,
      sym_using_declaration,
      sym_class_declaration,
      sym_function_declaration,
      sym_generator_function_declaration,
  [44671] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2513), 8,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
    ACTIONS(2515), 11,
      anon_sym_STAR,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
      aux_sym_method_definition_token1,
  [44699] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2517), 8,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
    ACTIONS(2519), 11,
      anon_sym_STAR,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
      aux_sym_method_definition_token1,
  [44727] = 6,
    ACTIONS(2525), 1,
      anon_sym_AT,
    STATE(1041), 1,
      aux_sym_export_statement_repeat1,
    STATE(1097), 1,
      sym_decorator,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2523), 7,
      anon_sym_STAR,
      anon_sym_LBRACK,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      sym_number,
      sym_private_property_identifier,
      aux_sym_method_definition_token1,
    ACTIONS(2521), 9,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_class,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [44761] = 4,
    ACTIONS(2528), 1,
      sym__automatic_semicolon,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(696), 8,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
    ACTIONS(694), 10,
      anon_sym_STAR,
      anon_sym_RBRACE,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
      aux_sym_method_definition_token1,
  [44791] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2532), 9,
      anon_sym_STAR,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
      aux_sym_method_definition_token1,
    ACTIONS(2530), 10,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_DOT,
      anon_sym_class,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [44819] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2534), 8,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
    ACTIONS(2536), 11,
      anon_sym_STAR,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
      aux_sym_method_definition_token1,
  [44847] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2538), 8,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
    ACTIONS(2540), 11,
      anon_sym_STAR,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
      aux_sym_method_definition_token1,
  [44875] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2542), 8,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
    ACTIONS(2544), 11,
      anon_sym_STAR,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
      aux_sym_method_definition_token1,
  [44903] = 4,
    ACTIONS(2546), 1,
      sym__automatic_semicolon,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(710), 8,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
    ACTIONS(708), 10,
      anon_sym_STAR,
      anon_sym_RBRACE,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
      aux_sym_method_definition_token1,
  [44933] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2513), 8,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
    ACTIONS(2515), 11,
      anon_sym_STAR,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
      aux_sym_method_definition_token1,
  [44961] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2513), 8,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
    ACTIONS(2515), 11,
      anon_sym_STAR,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
      aux_sym_method_definition_token1,
  [44989] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2513), 8,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
    ACTIONS(2515), 11,
      anon_sym_STAR,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
      aux_sym_method_definition_token1,
  [45017] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2513), 8,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
    ACTIONS(2515), 11,
      anon_sym_STAR,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
      aux_sym_method_definition_token1,
  [45045] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2513), 8,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
    ACTIONS(2515), 11,
      anon_sym_STAR,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
      aux_sym_method_definition_token1,
  [45073] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2548), 8,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
    ACTIONS(2550), 11,
      anon_sym_STAR,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
      aux_sym_method_definition_token1,
  [45101] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2548), 8,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
    ACTIONS(2550), 11,
      anon_sym_STAR,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
      aux_sym_method_definition_token1,
  [45129] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2548), 8,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
    ACTIONS(2550), 11,
      anon_sym_STAR,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
      aux_sym_method_definition_token1,
  [45157] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2548), 8,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
    ACTIONS(2550), 11,
      anon_sym_STAR,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
      aux_sym_method_definition_token1,
  [45185] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2548), 8,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
    ACTIONS(2550), 11,
      anon_sym_STAR,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
      aux_sym_method_definition_token1,
  [45213] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2548), 8,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
    ACTIONS(2550), 11,
      anon_sym_STAR,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
      aux_sym_method_definition_token1,
  [45241] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2513), 8,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
    ACTIONS(2515), 11,
      anon_sym_STAR,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
      aux_sym_method_definition_token1,
  [45269] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2513), 8,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
    ACTIONS(2515), 11,
      anon_sym_STAR,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
      aux_sym_method_definition_token1,
  [45297] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2513), 8,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
    ACTIONS(2515), 11,
      anon_sym_STAR,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
      aux_sym_method_definition_token1,
  [45325] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2513), 8,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
    ACTIONS(2515), 11,
      anon_sym_STAR,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
      aux_sym_method_definition_token1,
  [45353] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2513), 8,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
    ACTIONS(2515), 11,
      anon_sym_STAR,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
      aux_sym_method_definition_token1,
  [45381] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2552), 8,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
    ACTIONS(2554), 11,
      anon_sym_STAR,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
      aux_sym_method_definition_token1,
  [45409] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2509), 8,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
    ACTIONS(2511), 11,
      anon_sym_STAR,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
      aux_sym_method_definition_token1,
  [45437] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2509), 8,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
    ACTIONS(2511), 11,
      anon_sym_STAR,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
      aux_sym_method_definition_token1,
  [45465] = 14,
    ACTIONS(93), 1,
      anon_sym_AT,
    ACTIONS(936), 1,
      anon_sym_var,
    ACTIONS(940), 1,
      anon_sym_using,
    ACTIONS(942), 1,
      anon_sym_await,
    ACTIONS(947), 1,
      anon_sym_class,
    ACTIONS(949), 1,
      anon_sym_async,
    ACTIONS(951), 1,
      anon_sym_function,
    ACTIONS(2556), 1,
      anon_sym_default,
    STATE(486), 1,
      sym_declaration,
    STATE(1097), 1,
      sym_decorator,
    STATE(1327), 1,
      aux_sym_export_statement_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(938), 2,
      anon_sym_let,
      anon_sym_const,
    STATE(419), 6,
      sym_variable_declaration,
      sym_lexical_declaration,
      sym_using_declaration,
      sym_class_declaration,
      sym_function_declaration,
      sym_generator_function_declaration,
  [45515] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2509), 8,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
    ACTIONS(2511), 11,
      anon_sym_STAR,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
      aux_sym_method_definition_token1,
  [45543] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2509), 8,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
    ACTIONS(2511), 11,
      anon_sym_STAR,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
      aux_sym_method_definition_token1,
  [45571] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2509), 8,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
    ACTIONS(2511), 11,
      anon_sym_STAR,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
      aux_sym_method_definition_token1,
  [45599] = 4,
    ACTIONS(2562), 1,
      anon_sym_SEMI,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2558), 8,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
    ACTIONS(2560), 9,
      anon_sym_STAR,
      anon_sym_RBRACE,
      anon_sym_LBRACK,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
      aux_sym_method_definition_token1,
  [45628] = 11,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(2277), 1,
      anon_sym_LBRACK,
    ACTIONS(2395), 1,
      anon_sym_LPAREN,
    ACTIONS(2565), 1,
      anon_sym_STAR,
    ACTIONS(2569), 1,
      anon_sym_get,
    ACTIONS(2571), 1,
      anon_sym_set,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2567), 2,
      sym_number,
      sym_private_property_identifier,
    STATE(1575), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(1085), 6,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
  [45671] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2573), 8,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
    ACTIONS(2575), 10,
      anon_sym_STAR,
      anon_sym_RBRACE,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
      aux_sym_method_definition_token1,
  [45698] = 9,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(2277), 1,
      anon_sym_LBRACK,
    ACTIONS(2395), 1,
      anon_sym_LPAREN,
    ACTIONS(2577), 1,
      anon_sym_EQ_GT,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2579), 2,
      sym_number,
      sym_private_property_identifier,
    STATE(1694), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(1085), 8,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [45737] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(696), 8,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
    ACTIONS(694), 10,
      anon_sym_STAR,
      anon_sym_RBRACE,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
      aux_sym_method_definition_token1,
  [45764] = 11,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(2277), 1,
      anon_sym_LBRACK,
    ACTIONS(2395), 1,
      anon_sym_LPAREN,
    ACTIONS(2581), 1,
      anon_sym_STAR,
    ACTIONS(2585), 1,
      anon_sym_get,
    ACTIONS(2587), 1,
      anon_sym_set,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2583), 2,
      sym_number,
      sym_private_property_identifier,
    STATE(1730), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(1085), 6,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
  [45807] = 12,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(2277), 1,
      anon_sym_LBRACK,
    ACTIONS(2395), 1,
      anon_sym_LPAREN,
    ACTIONS(2589), 1,
      anon_sym_STAR,
    ACTIONS(2591), 1,
      anon_sym_async,
    ACTIONS(2595), 1,
      anon_sym_get,
    ACTIONS(2597), 1,
      anon_sym_set,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2593), 2,
      sym_number,
      sym_private_property_identifier,
    STATE(1508), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(1085), 5,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      sym_identifier,
      anon_sym_static,
  [45852] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2599), 8,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
    ACTIONS(2601), 10,
      anon_sym_STAR,
      anon_sym_RBRACE,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
      aux_sym_method_definition_token1,
  [45879] = 11,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(2277), 1,
      anon_sym_LBRACK,
    ACTIONS(2395), 1,
      anon_sym_LPAREN,
    ACTIONS(2603), 1,
      anon_sym_STAR,
    ACTIONS(2607), 1,
      anon_sym_get,
    ACTIONS(2609), 1,
      anon_sym_set,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2605), 2,
      sym_number,
      sym_private_property_identifier,
    STATE(1568), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(1085), 6,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
  [45922] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2558), 8,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
    ACTIONS(2560), 10,
      anon_sym_STAR,
      anon_sym_RBRACE,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
      aux_sym_method_definition_token1,
  [45949] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2611), 8,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
    ACTIONS(2613), 10,
      anon_sym_STAR,
      anon_sym_RBRACE,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
      aux_sym_method_definition_token1,
  [45976] = 10,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(2277), 1,
      anon_sym_LBRACK,
    ACTIONS(2395), 1,
      anon_sym_LPAREN,
    ACTIONS(2485), 1,
      anon_sym_STAR,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2487), 2,
      sym_number,
      sym_private_property_identifier,
    ACTIONS(2489), 2,
      anon_sym_get,
      anon_sym_set,
    STATE(1580), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(1085), 6,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
  [46017] = 10,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(2277), 1,
      anon_sym_LBRACK,
    ACTIONS(2395), 1,
      anon_sym_LPAREN,
    ACTIONS(2615), 1,
      anon_sym_STAR,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2579), 2,
      sym_number,
      sym_private_property_identifier,
    ACTIONS(2617), 2,
      anon_sym_get,
      anon_sym_set,
    STATE(1694), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(1085), 6,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
  [46058] = 11,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(2277), 1,
      anon_sym_LBRACK,
    ACTIONS(2395), 1,
      anon_sym_LPAREN,
    ACTIONS(2619), 1,
      anon_sym_STAR,
    ACTIONS(2623), 1,
      anon_sym_get,
    ACTIONS(2625), 1,
      anon_sym_set,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2621), 2,
      sym_number,
      sym_private_property_identifier,
    STATE(1602), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(1085), 6,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
  [46101] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(702), 8,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
    ACTIONS(700), 10,
      anon_sym_STAR,
      anon_sym_RBRACE,
      anon_sym_SEMI,
      anon_sym_LBRACK,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
      aux_sym_method_definition_token1,
  [46128] = 8,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(2277), 1,
      anon_sym_LBRACK,
    ACTIONS(2395), 1,
      anon_sym_LPAREN,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2627), 2,
      sym_number,
      sym_private_property_identifier,
    STATE(1604), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(1085), 8,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [46164] = 8,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(2277), 1,
      anon_sym_LBRACK,
    ACTIONS(2395), 1,
      anon_sym_LPAREN,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2629), 2,
      sym_number,
      sym_private_property_identifier,
    STATE(1577), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(1085), 8,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [46200] = 8,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(2277), 1,
      anon_sym_LBRACK,
    ACTIONS(2395), 1,
      anon_sym_LPAREN,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2631), 2,
      sym_number,
      sym_private_property_identifier,
    STATE(1570), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(1085), 8,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [46236] = 11,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(2277), 1,
      anon_sym_LBRACK,
    ACTIONS(2633), 1,
      anon_sym_STAR,
    ACTIONS(2635), 1,
      anon_sym_async,
    ACTIONS(2639), 1,
      anon_sym_get,
    ACTIONS(2641), 1,
      anon_sym_set,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2637), 2,
      sym_number,
      sym_private_property_identifier,
    STATE(1529), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(1085), 5,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      sym_identifier,
      anon_sym_static,
  [46278] = 8,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(2277), 1,
      anon_sym_LBRACK,
    ACTIONS(2395), 1,
      anon_sym_LPAREN,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2643), 2,
      sym_number,
      sym_private_property_identifier,
    STATE(1539), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(1085), 8,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [46314] = 8,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(2277), 1,
      anon_sym_LBRACK,
    ACTIONS(2395), 1,
      anon_sym_LPAREN,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2645), 2,
      sym_number,
      sym_private_property_identifier,
    STATE(1540), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(1085), 8,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [46350] = 8,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(2277), 1,
      anon_sym_LBRACK,
    ACTIONS(2395), 1,
      anon_sym_LPAREN,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2647), 2,
      sym_number,
      sym_private_property_identifier,
    STATE(1715), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(1085), 8,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [46386] = 8,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(2277), 1,
      anon_sym_LBRACK,
    ACTIONS(2395), 1,
      anon_sym_LPAREN,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2649), 2,
      sym_number,
      sym_private_property_identifier,
    STATE(1718), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(1085), 8,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [46422] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2501), 8,
      anon_sym_STAR,
      anon_sym_LBRACK,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
      aux_sym_method_definition_token1,
    ACTIONS(2499), 9,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_class,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [46448] = 8,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(2277), 1,
      anon_sym_LBRACK,
    ACTIONS(2395), 1,
      anon_sym_LPAREN,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2651), 2,
      sym_number,
      sym_private_property_identifier,
    STATE(1740), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(1085), 8,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [46484] = 8,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(2277), 1,
      anon_sym_LBRACK,
    ACTIONS(2395), 1,
      anon_sym_LPAREN,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2653), 2,
      sym_number,
      sym_private_property_identifier,
    STATE(1741), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(1085), 8,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [46520] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2657), 8,
      anon_sym_STAR,
      anon_sym_LBRACK,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
      aux_sym_method_definition_token1,
    ACTIONS(2655), 9,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_class,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [46546] = 8,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(2277), 1,
      anon_sym_LBRACK,
    ACTIONS(2395), 1,
      anon_sym_LPAREN,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2659), 2,
      sym_number,
      sym_private_property_identifier,
    STATE(1562), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(1085), 8,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [46582] = 8,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(2277), 1,
      anon_sym_LBRACK,
    ACTIONS(2395), 1,
      anon_sym_LPAREN,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2579), 2,
      sym_number,
      sym_private_property_identifier,
    STATE(1694), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(1085), 8,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [46618] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1737), 8,
      anon_sym_STAR,
      anon_sym_LBRACK,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
      aux_sym_method_definition_token1,
    ACTIONS(1735), 9,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_class,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [46644] = 10,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(2277), 1,
      anon_sym_LBRACK,
    ACTIONS(2403), 1,
      anon_sym_STAR,
    ACTIONS(2661), 1,
      anon_sym_async,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2467), 2,
      sym_number,
      sym_private_property_identifier,
    ACTIONS(2469), 2,
      anon_sym_get,
      anon_sym_set,
    STATE(1553), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(1085), 5,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      sym_identifier,
      anon_sym_static,
  [46684] = 8,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(2277), 1,
      anon_sym_LBRACK,
    ACTIONS(2395), 1,
      anon_sym_LPAREN,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2663), 2,
      sym_number,
      sym_private_property_identifier,
    STATE(1569), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(1085), 8,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [46720] = 8,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(2277), 1,
      anon_sym_LBRACK,
    ACTIONS(2395), 1,
      anon_sym_LPAREN,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2665), 2,
      sym_number,
      sym_private_property_identifier,
    STATE(1561), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(1085), 8,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [46756] = 8,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(2277), 1,
      anon_sym_LBRACK,
    ACTIONS(2395), 1,
      anon_sym_LPAREN,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2487), 2,
      sym_number,
      sym_private_property_identifier,
    STATE(1580), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(1085), 8,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [46792] = 11,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(2277), 1,
      anon_sym_LBRACK,
    ACTIONS(2667), 1,
      anon_sym_STAR,
    ACTIONS(2669), 1,
      anon_sym_async,
    ACTIONS(2673), 1,
      anon_sym_get,
    ACTIONS(2675), 1,
      anon_sym_set,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2671), 2,
      sym_number,
      sym_private_property_identifier,
    STATE(1583), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(1085), 5,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      sym_identifier,
      anon_sym_static,
  [46834] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2679), 8,
      anon_sym_STAR,
      anon_sym_LBRACK,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
      aux_sym_method_definition_token1,
    ACTIONS(2677), 9,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_class,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [46860] = 8,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(2277), 1,
      anon_sym_LBRACK,
    ACTIONS(2395), 1,
      anon_sym_LPAREN,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2681), 2,
      sym_number,
      sym_private_property_identifier,
    STATE(1593), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(1085), 8,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [46896] = 8,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(2277), 1,
      anon_sym_LBRACK,
    ACTIONS(2395), 1,
      anon_sym_LPAREN,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2683), 2,
      sym_number,
      sym_private_property_identifier,
    STATE(1595), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(1085), 8,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [46932] = 8,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(2277), 1,
      anon_sym_LBRACK,
    ACTIONS(2395), 1,
      anon_sym_LPAREN,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2685), 2,
      sym_number,
      sym_private_property_identifier,
    STATE(1596), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(1085), 8,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [46968] = 10,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(1087), 1,
      anon_sym_async,
    ACTIONS(2277), 1,
      anon_sym_LBRACK,
    ACTIONS(2391), 1,
      anon_sym_STAR,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1081), 2,
      sym_number,
      sym_private_property_identifier,
    ACTIONS(1089), 2,
      anon_sym_get,
      anon_sym_set,
    STATE(1710), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(1085), 5,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      sym_identifier,
      anon_sym_static,
  [47008] = 8,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(2277), 1,
      anon_sym_LBRACK,
    ACTIONS(2395), 1,
      anon_sym_LPAREN,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2687), 2,
      sym_number,
      sym_private_property_identifier,
    STATE(1603), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(1085), 8,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [47044] = 8,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(2277), 1,
      anon_sym_LBRACK,
    ACTIONS(2395), 1,
      anon_sym_LPAREN,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2689), 2,
      sym_number,
      sym_private_property_identifier,
    STATE(1576), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(1085), 8,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [47080] = 8,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(2277), 1,
      anon_sym_LBRACK,
    ACTIONS(2395), 1,
      anon_sym_LPAREN,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2691), 2,
      sym_number,
      sym_private_property_identifier,
    STATE(1610), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(1085), 8,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [47116] = 8,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(2277), 1,
      anon_sym_LBRACK,
    ACTIONS(2395), 1,
      anon_sym_LPAREN,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2693), 2,
      sym_number,
      sym_private_property_identifier,
    STATE(1611), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(1085), 8,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [47152] = 8,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(2277), 1,
      anon_sym_LBRACK,
    ACTIONS(2395), 1,
      anon_sym_LPAREN,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2695), 2,
      sym_number,
      sym_private_property_identifier,
    STATE(1613), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(1085), 8,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [47188] = 8,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(2277), 1,
      anon_sym_LBRACK,
    ACTIONS(2395), 1,
      anon_sym_LPAREN,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2697), 2,
      sym_number,
      sym_private_property_identifier,
    STATE(1614), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(1085), 8,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [47224] = 8,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(2277), 1,
      anon_sym_LBRACK,
    ACTIONS(2395), 1,
      anon_sym_LPAREN,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2699), 2,
      sym_number,
      sym_private_property_identifier,
    STATE(1552), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(1085), 8,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [47260] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1635), 8,
      anon_sym_STAR,
      anon_sym_LBRACK,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
      aux_sym_method_definition_token1,
    ACTIONS(1633), 9,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_class,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [47286] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1745), 8,
      anon_sym_STAR,
      anon_sym_LBRACK,
      anon_sym_DQUOTE,
      anon_sym_SQUOTE,
      sym_number,
      sym_private_property_identifier,
      anon_sym_AT,
      aux_sym_method_definition_token1,
    ACTIONS(1743), 9,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_class,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [47312] = 7,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(2277), 1,
      anon_sym_LBRACK,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2701), 2,
      sym_number,
      sym_private_property_identifier,
    STATE(1597), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(1085), 8,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [47345] = 7,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(2277), 1,
      anon_sym_LBRACK,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2699), 2,
      sym_number,
      sym_private_property_identifier,
    STATE(1552), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(1085), 8,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [47378] = 7,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(2277), 1,
      anon_sym_LBRACK,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2579), 2,
      sym_number,
      sym_private_property_identifier,
    STATE(1694), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(1085), 8,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [47411] = 7,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(2277), 1,
      anon_sym_LBRACK,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2703), 2,
      sym_number,
      sym_private_property_identifier,
    STATE(1559), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(1085), 8,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [47444] = 7,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(2277), 1,
      anon_sym_LBRACK,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2705), 2,
      sym_number,
      sym_private_property_identifier,
    STATE(1565), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(1085), 8,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [47477] = 7,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(2277), 1,
      anon_sym_LBRACK,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1081), 2,
      sym_number,
      sym_private_property_identifier,
    STATE(1710), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(1085), 8,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [47510] = 7,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(2277), 1,
      anon_sym_LBRACK,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2707), 2,
      sym_number,
      sym_private_property_identifier,
    STATE(1726), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(1085), 8,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [47543] = 7,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(2277), 1,
      anon_sym_LBRACK,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2709), 2,
      sym_number,
      sym_private_property_identifier,
    STATE(1712), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(1085), 8,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [47576] = 7,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(2277), 1,
      anon_sym_LBRACK,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2711), 2,
      sym_number,
      sym_private_property_identifier,
    STATE(1739), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(1085), 8,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [47609] = 7,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(2277), 1,
      anon_sym_LBRACK,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2713), 2,
      sym_number,
      sym_private_property_identifier,
    STATE(1594), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(1085), 8,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [47642] = 7,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(2277), 1,
      anon_sym_LBRACK,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2467), 2,
      sym_number,
      sym_private_property_identifier,
    STATE(1553), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(1085), 8,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [47675] = 7,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(2277), 1,
      anon_sym_LBRACK,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2487), 2,
      sym_number,
      sym_private_property_identifier,
    STATE(1580), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(1085), 8,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [47708] = 7,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(2277), 1,
      anon_sym_LBRACK,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2715), 2,
      sym_number,
      sym_private_property_identifier,
    STATE(1601), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(1085), 8,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [47741] = 7,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(2277), 1,
      anon_sym_LBRACK,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2717), 2,
      sym_number,
      sym_private_property_identifier,
    STATE(1581), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(1085), 8,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [47774] = 7,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(2277), 1,
      anon_sym_LBRACK,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2681), 2,
      sym_number,
      sym_private_property_identifier,
    STATE(1593), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(1085), 8,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [47807] = 7,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(2277), 1,
      anon_sym_LBRACK,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2719), 2,
      sym_number,
      sym_private_property_identifier,
    STATE(1609), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(1085), 8,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [47840] = 7,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(2277), 1,
      anon_sym_LBRACK,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2721), 2,
      sym_number,
      sym_private_property_identifier,
    STATE(1612), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(1085), 8,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [47873] = 7,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(2277), 1,
      anon_sym_LBRACK,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2723), 2,
      sym_number,
      sym_private_property_identifier,
    STATE(1572), 3,
      sym_string,
      sym__property_name,
      sym_computed_property_name,
    ACTIONS(1085), 8,
      anon_sym_export,
      anon_sym_let,
      anon_sym_await,
      anon_sym_async,
      sym_identifier,
      anon_sym_static,
      anon_sym_get,
      anon_sym_set,
  [47906] = 2,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2725), 15,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_CARET_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_AMP_AMP_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
  [47928] = 2,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2727), 15,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_CARET_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_AMP_AMP_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
  [47950] = 2,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2729), 15,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_CARET_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_AMP_AMP_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
  [47972] = 2,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2731), 15,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_CARET_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_AMP_AMP_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
  [47994] = 2,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2733), 15,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_CARET_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_AMP_AMP_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
  [48016] = 2,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2735), 15,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_CARET_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_GT_GT_EQ,
      anon_sym_GT_GT_GT_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_STAR_STAR_EQ,
      anon_sym_AMP_AMP_EQ,
      anon_sym_PIPE_PIPE_EQ,
      anon_sym_QMARK_QMARK_EQ,
  [48038] = 11,
    ACTIONS(2297), 1,
      anon_sym_DQUOTE,
    ACTIONS(2299), 1,
      anon_sym_SQUOTE,
    ACTIONS(2737), 1,
      sym_identifier,
    ACTIONS(2739), 1,
      anon_sym_STAR,
    ACTIONS(2741), 1,
      anon_sym_LBRACE,
    ACTIONS(2745), 1,
      anon_sym_DOT,
    STATE(1359), 1,
      sym_string,
    STATE(1625), 1,
      sym_import_clause,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2743), 2,
      anon_sym_LPAREN,
      sym_optional_chain,
    STATE(1862), 2,
      sym_namespace_import,
      sym_named_imports,
  [48075] = 11,
    ACTIONS(2747), 1,
      sym_identifier,
    ACTIONS(2749), 1,
      anon_sym_LBRACE,
    ACTIONS(2751), 1,
      anon_sym_COLON,
    ACTIONS(2753), 1,
      anon_sym_GT,
    ACTIONS(2755), 1,
      sym_jsx_identifier,
    ACTIONS(2757), 1,
      anon_sym_DOT,
    ACTIONS(2759), 1,
      anon_sym_SLASH_GT,
    STATE(1185), 1,
      aux_sym_jsx_opening_element_repeat1,
    STATE(1226), 1,
      sym_jsx_namespace_name,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    STATE(1305), 2,
      sym_jsx_expression,
      sym_jsx_attribute,
  [48111] = 8,
    ACTIONS(2761), 1,
      anon_sym_LBRACE,
    ACTIONS(2765), 1,
      anon_sym_LT,
    ACTIONS(2767), 1,
      anon_sym_LT_SLASH,
    STATE(1149), 1,
      sym_jsx_opening_element,
    STATE(1259), 1,
      sym_jsx_closing_element,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2763), 2,
      sym_jsx_text,
      sym_html_character_reference,
    STATE(1160), 4,
      sym_jsx_element,
      sym_jsx_expression,
      sym_jsx_self_closing_element,
      aux_sym_jsx_element_repeat1,
  [48141] = 11,
    ACTIONS(2747), 1,
      sym_identifier,
    ACTIONS(2749), 1,
      anon_sym_LBRACE,
    ACTIONS(2751), 1,
      anon_sym_COLON,
    ACTIONS(2753), 1,
      anon_sym_GT,
    ACTIONS(2755), 1,
      sym_jsx_identifier,
    ACTIONS(2757), 1,
      anon_sym_DOT,
    ACTIONS(2769), 1,
      anon_sym_SLASH_GT,
    STATE(1177), 1,
      aux_sym_jsx_opening_element_repeat1,
    STATE(1226), 1,
      sym_jsx_namespace_name,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    STATE(1305), 2,
      sym_jsx_expression,
      sym_jsx_attribute,
  [48177] = 8,
    ACTIONS(2761), 1,
      anon_sym_LBRACE,
    ACTIONS(2765), 1,
      anon_sym_LT,
    ACTIONS(2771), 1,
      anon_sym_LT_SLASH,
    STATE(611), 1,
      sym_jsx_closing_element,
    STATE(1149), 1,
      sym_jsx_opening_element,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2763), 2,
      sym_jsx_text,
      sym_html_character_reference,
    STATE(1160), 4,
      sym_jsx_element,
      sym_jsx_expression,
      sym_jsx_self_closing_element,
      aux_sym_jsx_element_repeat1,
  [48207] = 8,
    ACTIONS(2761), 1,
      anon_sym_LBRACE,
    ACTIONS(2765), 1,
      anon_sym_LT,
    ACTIONS(2775), 1,
      anon_sym_LT_SLASH,
    STATE(1149), 1,
      sym_jsx_opening_element,
    STATE(1255), 1,
      sym_jsx_closing_element,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2773), 2,
      sym_jsx_text,
      sym_html_character_reference,
    STATE(1156), 4,
      sym_jsx_element,
      sym_jsx_expression,
      sym_jsx_self_closing_element,
      aux_sym_jsx_element_repeat1,
  [48237] = 8,
    ACTIONS(2761), 1,
      anon_sym_LBRACE,
    ACTIONS(2765), 1,
      anon_sym_LT,
    ACTIONS(2767), 1,
      anon_sym_LT_SLASH,
    STATE(1149), 1,
      sym_jsx_opening_element,
    STATE(1312), 1,
      sym_jsx_closing_element,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2777), 2,
      sym_jsx_text,
      sym_html_character_reference,
    STATE(1146), 4,
      sym_jsx_element,
      sym_jsx_expression,
      sym_jsx_self_closing_element,
      aux_sym_jsx_element_repeat1,
  [48267] = 11,
    ACTIONS(2747), 1,
      sym_identifier,
    ACTIONS(2749), 1,
      anon_sym_LBRACE,
    ACTIONS(2751), 1,
      anon_sym_COLON,
    ACTIONS(2753), 1,
      anon_sym_GT,
    ACTIONS(2755), 1,
      sym_jsx_identifier,
    ACTIONS(2757), 1,
      anon_sym_DOT,
    ACTIONS(2779), 1,
      anon_sym_SLASH_GT,
    STATE(1182), 1,
      aux_sym_jsx_opening_element_repeat1,
    STATE(1226), 1,
      sym_jsx_namespace_name,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    STATE(1305), 2,
      sym_jsx_expression,
      sym_jsx_attribute,
  [48303] = 11,
    ACTIONS(2747), 1,
      sym_identifier,
    ACTIONS(2749), 1,
      anon_sym_LBRACE,
    ACTIONS(2751), 1,
      anon_sym_COLON,
    ACTIONS(2753), 1,
      anon_sym_GT,
    ACTIONS(2755), 1,
      sym_jsx_identifier,
    ACTIONS(2757), 1,
      anon_sym_DOT,
    ACTIONS(2781), 1,
      anon_sym_SLASH_GT,
    STATE(1183), 1,
      aux_sym_jsx_opening_element_repeat1,
    STATE(1226), 1,
      sym_jsx_namespace_name,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    STATE(1305), 2,
      sym_jsx_expression,
      sym_jsx_attribute,
  [48339] = 8,
    ACTIONS(2761), 1,
      anon_sym_LBRACE,
    ACTIONS(2765), 1,
      anon_sym_LT,
    ACTIONS(2771), 1,
      anon_sym_LT_SLASH,
    STATE(608), 1,
      sym_jsx_closing_element,
    STATE(1149), 1,
      sym_jsx_opening_element,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2783), 2,
      sym_jsx_text,
      sym_html_character_reference,
    STATE(1148), 4,
      sym_jsx_element,
      sym_jsx_expression,
      sym_jsx_self_closing_element,
      aux_sym_jsx_element_repeat1,
  [48369] = 8,
    ACTIONS(2761), 1,
      anon_sym_LBRACE,
    ACTIONS(2765), 1,
      anon_sym_LT,
    ACTIONS(2785), 1,
      anon_sym_LT_SLASH,
    STATE(790), 1,
      sym_jsx_closing_element,
    STATE(1149), 1,
      sym_jsx_opening_element,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2763), 2,
      sym_jsx_text,
      sym_html_character_reference,
    STATE(1160), 4,
      sym_jsx_element,
      sym_jsx_expression,
      sym_jsx_self_closing_element,
      aux_sym_jsx_element_repeat1,
  [48399] = 8,
    ACTIONS(2761), 1,
      anon_sym_LBRACE,
    ACTIONS(2765), 1,
      anon_sym_LT,
    ACTIONS(2785), 1,
      anon_sym_LT_SLASH,
    STATE(745), 1,
      sym_jsx_closing_element,
    STATE(1149), 1,
      sym_jsx_opening_element,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2787), 2,
      sym_jsx_text,
      sym_html_character_reference,
    STATE(1154), 4,
      sym_jsx_element,
      sym_jsx_expression,
      sym_jsx_self_closing_element,
      aux_sym_jsx_element_repeat1,
  [48429] = 8,
    ACTIONS(2761), 1,
      anon_sym_LBRACE,
    ACTIONS(2765), 1,
      anon_sym_LT,
    ACTIONS(2775), 1,
      anon_sym_LT_SLASH,
    STATE(1149), 1,
      sym_jsx_opening_element,
    STATE(1290), 1,
      sym_jsx_closing_element,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2763), 2,
      sym_jsx_text,
      sym_html_character_reference,
    STATE(1160), 4,
      sym_jsx_element,
      sym_jsx_expression,
      sym_jsx_self_closing_element,
      aux_sym_jsx_element_repeat1,
  [48459] = 10,
    ACTIONS(2747), 1,
      sym_identifier,
    ACTIONS(2749), 1,
      anon_sym_LBRACE,
    ACTIONS(2751), 1,
      anon_sym_COLON,
    ACTIONS(2753), 1,
      anon_sym_GT,
    ACTIONS(2755), 1,
      sym_jsx_identifier,
    ACTIONS(2759), 1,
      anon_sym_SLASH_GT,
    STATE(1173), 1,
      aux_sym_jsx_opening_element_repeat1,
    STATE(1226), 1,
      sym_jsx_namespace_name,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    STATE(1305), 2,
      sym_jsx_expression,
      sym_jsx_attribute,
  [48492] = 10,
    ACTIONS(2747), 1,
      sym_identifier,
    ACTIONS(2749), 1,
      anon_sym_LBRACE,
    ACTIONS(2753), 1,
      anon_sym_GT,
    ACTIONS(2755), 1,
      sym_jsx_identifier,
    ACTIONS(2757), 1,
      anon_sym_DOT,
    ACTIONS(2769), 1,
      anon_sym_SLASH_GT,
    STATE(1178), 1,
      aux_sym_jsx_opening_element_repeat1,
    STATE(1226), 1,
      sym_jsx_namespace_name,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    STATE(1305), 2,
      sym_jsx_expression,
      sym_jsx_attribute,
  [48525] = 10,
    ACTIONS(2747), 1,
      sym_identifier,
    ACTIONS(2749), 1,
      anon_sym_LBRACE,
    ACTIONS(2753), 1,
      anon_sym_GT,
    ACTIONS(2755), 1,
      sym_jsx_identifier,
    ACTIONS(2757), 1,
      anon_sym_DOT,
    ACTIONS(2781), 1,
      anon_sym_SLASH_GT,
    STATE(1167), 1,
      aux_sym_jsx_opening_element_repeat1,
    STATE(1226), 1,
      sym_jsx_namespace_name,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    STATE(1305), 2,
      sym_jsx_expression,
      sym_jsx_attribute,
  [48558] = 7,
    ACTIONS(2789), 1,
      anon_sym_LBRACE,
    ACTIONS(2795), 1,
      anon_sym_LT,
    ACTIONS(2798), 1,
      anon_sym_LT_SLASH,
    STATE(1149), 1,
      sym_jsx_opening_element,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2792), 2,
      sym_jsx_text,
      sym_html_character_reference,
    STATE(1160), 4,
      sym_jsx_element,
      sym_jsx_expression,
      sym_jsx_self_closing_element,
      aux_sym_jsx_element_repeat1,
  [48585] = 10,
    ACTIONS(2747), 1,
      sym_identifier,
    ACTIONS(2749), 1,
      anon_sym_LBRACE,
    ACTIONS(2751), 1,
      anon_sym_COLON,
    ACTIONS(2753), 1,
      anon_sym_GT,
    ACTIONS(2755), 1,
      sym_jsx_identifier,
    ACTIONS(2779), 1,
      anon_sym_SLASH_GT,
    STATE(1181), 1,
      aux_sym_jsx_opening_element_repeat1,
    STATE(1226), 1,
      sym_jsx_namespace_name,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    STATE(1305), 2,
      sym_jsx_expression,
      sym_jsx_attribute,
  [48618] = 10,
    ACTIONS(2747), 1,
      sym_identifier,
    ACTIONS(2749), 1,
      anon_sym_LBRACE,
    ACTIONS(2753), 1,
      anon_sym_GT,
    ACTIONS(2755), 1,
      sym_jsx_identifier,
    ACTIONS(2757), 1,
      anon_sym_DOT,
    ACTIONS(2759), 1,
      anon_sym_SLASH_GT,
    STATE(1186), 1,
      aux_sym_jsx_opening_element_repeat1,
    STATE(1226), 1,
      sym_jsx_namespace_name,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    STATE(1305), 2,
      sym_jsx_expression,
      sym_jsx_attribute,
  [48651] = 10,
    ACTIONS(2747), 1,
      sym_identifier,
    ACTIONS(2749), 1,
      anon_sym_LBRACE,
    ACTIONS(2753), 1,
      anon_sym_GT,
    ACTIONS(2755), 1,
      sym_jsx_identifier,
    ACTIONS(2757), 1,
      anon_sym_DOT,
    ACTIONS(2779), 1,
      anon_sym_SLASH_GT,
    STATE(1184), 1,
      aux_sym_jsx_opening_element_repeat1,
    STATE(1226), 1,
      sym_jsx_namespace_name,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    STATE(1305), 2,
      sym_jsx_expression,
      sym_jsx_attribute,
  [48684] = 10,
    ACTIONS(2747), 1,
      sym_identifier,
    ACTIONS(2749), 1,
      anon_sym_LBRACE,
    ACTIONS(2751), 1,
      anon_sym_COLON,
    ACTIONS(2753), 1,
      anon_sym_GT,
    ACTIONS(2755), 1,
      sym_jsx_identifier,
    ACTIONS(2781), 1,
      anon_sym_SLASH_GT,
    STATE(1179), 1,
      aux_sym_jsx_opening_element_repeat1,
    STATE(1226), 1,
      sym_jsx_namespace_name,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    STATE(1305), 2,
      sym_jsx_expression,
      sym_jsx_attribute,
  [48717] = 10,
    ACTIONS(2747), 1,
      sym_identifier,
    ACTIONS(2749), 1,
      anon_sym_LBRACE,
    ACTIONS(2751), 1,
      anon_sym_COLON,
    ACTIONS(2753), 1,
      anon_sym_GT,
    ACTIONS(2755), 1,
      sym_jsx_identifier,
    ACTIONS(2769), 1,
      anon_sym_SLASH_GT,
    STATE(1175), 1,
      aux_sym_jsx_opening_element_repeat1,
    STATE(1226), 1,
      sym_jsx_namespace_name,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    STATE(1305), 2,
      sym_jsx_expression,
      sym_jsx_attribute,
  [48750] = 9,
    ACTIONS(2747), 1,
      sym_identifier,
    ACTIONS(2749), 1,
      anon_sym_LBRACE,
    ACTIONS(2755), 1,
      sym_jsx_identifier,
    ACTIONS(2800), 1,
      anon_sym_GT,
    ACTIONS(2802), 1,
      anon_sym_SLASH_GT,
    STATE(1171), 1,
      aux_sym_jsx_opening_element_repeat1,
    STATE(1226), 1,
      sym_jsx_namespace_name,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    STATE(1305), 2,
      sym_jsx_expression,
      sym_jsx_attribute,
  [48780] = 9,
    ACTIONS(2747), 1,
      sym_identifier,
    ACTIONS(2749), 1,
      anon_sym_LBRACE,
    ACTIONS(2755), 1,
      sym_jsx_identifier,
    ACTIONS(2804), 1,
      anon_sym_GT,
    ACTIONS(2806), 1,
      anon_sym_SLASH_GT,
    STATE(1171), 1,
      aux_sym_jsx_opening_element_repeat1,
    STATE(1226), 1,
      sym_jsx_namespace_name,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    STATE(1305), 2,
      sym_jsx_expression,
      sym_jsx_attribute,
  [48810] = 7,
    ACTIONS(2749), 1,
      anon_sym_LBRACE,
    ACTIONS(2808), 1,
      anon_sym_LT,
    ACTIONS(2810), 1,
      anon_sym_DQUOTE,
    ACTIONS(2812), 1,
      anon_sym_SQUOTE,
    STATE(1150), 1,
      sym_jsx_opening_element,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    STATE(1291), 4,
      sym_jsx_element,
      sym_jsx_expression,
      sym_jsx_self_closing_element,
      sym__jsx_string,
  [48836] = 9,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(2814), 1,
      sym_identifier,
    ACTIONS(2816), 1,
      anon_sym_default,
    ACTIONS(2818), 1,
      anon_sym_COMMA,
    ACTIONS(2820), 1,
      anon_sym_RBRACE,
    STATE(1443), 1,
      sym_import_specifier,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    STATE(1775), 2,
      sym__module_export_name,
      sym_string,
  [48866] = 7,
    ACTIONS(2749), 1,
      anon_sym_LBRACE,
    ACTIONS(2808), 1,
      anon_sym_LT,
    ACTIONS(2810), 1,
      anon_sym_DQUOTE,
    ACTIONS(2812), 1,
      anon_sym_SQUOTE,
    STATE(1150), 1,
      sym_jsx_opening_element,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    STATE(1292), 4,
      sym_jsx_element,
      sym_jsx_expression,
      sym_jsx_self_closing_element,
      sym__jsx_string,
  [48892] = 8,
    ACTIONS(2822), 1,
      sym_identifier,
    ACTIONS(2825), 1,
      anon_sym_LBRACE,
    ACTIONS(2830), 1,
      sym_jsx_identifier,
    STATE(1171), 1,
      aux_sym_jsx_opening_element_repeat1,
    STATE(1226), 1,
      sym_jsx_namespace_name,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2828), 2,
      anon_sym_GT,
      anon_sym_SLASH_GT,
    STATE(1305), 2,
      sym_jsx_expression,
      sym_jsx_attribute,
  [48920] = 8,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(2835), 1,
      anon_sym_COMMA,
    ACTIONS(2837), 1,
      anon_sym_RBRACE,
    STATE(1414), 1,
      sym_export_specifier,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2833), 2,
      anon_sym_default,
      sym_identifier,
    STATE(1423), 2,
      sym__module_export_name,
      sym_string,
  [48948] = 9,
    ACTIONS(2747), 1,
      sym_identifier,
    ACTIONS(2749), 1,
      anon_sym_LBRACE,
    ACTIONS(2755), 1,
      sym_jsx_identifier,
    ACTIONS(2839), 1,
      anon_sym_GT,
    ACTIONS(2841), 1,
      anon_sym_SLASH_GT,
    STATE(1171), 1,
      aux_sym_jsx_opening_element_repeat1,
    STATE(1226), 1,
      sym_jsx_namespace_name,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    STATE(1305), 2,
      sym_jsx_expression,
      sym_jsx_attribute,
  [48978] = 9,
    ACTIONS(2747), 1,
      sym_identifier,
    ACTIONS(2749), 1,
      anon_sym_LBRACE,
    ACTIONS(2753), 1,
      anon_sym_GT,
    ACTIONS(2755), 1,
      sym_jsx_identifier,
    ACTIONS(2759), 1,
      anon_sym_SLASH_GT,
    STATE(1188), 1,
      aux_sym_jsx_opening_element_repeat1,
    STATE(1226), 1,
      sym_jsx_namespace_name,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    STATE(1305), 2,
      sym_jsx_expression,
      sym_jsx_attribute,
  [49008] = 9,
    ACTIONS(2747), 1,
      sym_identifier,
    ACTIONS(2749), 1,
      anon_sym_LBRACE,
    ACTIONS(2755), 1,
      sym_jsx_identifier,
    ACTIONS(2839), 1,
      anon_sym_GT,
    ACTIONS(2843), 1,
      anon_sym_SLASH_GT,
    STATE(1171), 1,
      aux_sym_jsx_opening_element_repeat1,
    STATE(1226), 1,
      sym_jsx_namespace_name,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    STATE(1305), 2,
      sym_jsx_expression,
      sym_jsx_attribute,
  [49038] = 9,
    ACTIONS(2747), 1,
      sym_identifier,
    ACTIONS(2749), 1,
      anon_sym_LBRACE,
    ACTIONS(2753), 1,
      anon_sym_GT,
    ACTIONS(2755), 1,
      sym_jsx_identifier,
    ACTIONS(2779), 1,
      anon_sym_SLASH_GT,
    STATE(1187), 1,
      aux_sym_jsx_opening_element_repeat1,
    STATE(1226), 1,
      sym_jsx_namespace_name,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    STATE(1305), 2,
      sym_jsx_expression,
      sym_jsx_attribute,
  [49068] = 9,
    ACTIONS(2747), 1,
      sym_identifier,
    ACTIONS(2749), 1,
      anon_sym_LBRACE,
    ACTIONS(2755), 1,
      sym_jsx_identifier,
    ACTIONS(2845), 1,
      anon_sym_GT,
    ACTIONS(2847), 1,
      anon_sym_SLASH_GT,
    STATE(1171), 1,
      aux_sym_jsx_opening_element_repeat1,
    STATE(1226), 1,
      sym_jsx_namespace_name,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    STATE(1305), 2,
      sym_jsx_expression,
      sym_jsx_attribute,
  [49098] = 9,
    ACTIONS(2747), 1,
      sym_identifier,
    ACTIONS(2749), 1,
      anon_sym_LBRACE,
    ACTIONS(2755), 1,
      sym_jsx_identifier,
    ACTIONS(2804), 1,
      anon_sym_GT,
    ACTIONS(2849), 1,
      anon_sym_SLASH_GT,
    STATE(1171), 1,
      aux_sym_jsx_opening_element_repeat1,
    STATE(1226), 1,
      sym_jsx_namespace_name,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    STATE(1305), 2,
      sym_jsx_expression,
      sym_jsx_attribute,
  [49128] = 9,
    ACTIONS(2747), 1,
      sym_identifier,
    ACTIONS(2749), 1,
      anon_sym_LBRACE,
    ACTIONS(2755), 1,
      sym_jsx_identifier,
    ACTIONS(2839), 1,
      anon_sym_GT,
    ACTIONS(2851), 1,
      anon_sym_SLASH_GT,
    STATE(1171), 1,
      aux_sym_jsx_opening_element_repeat1,
    STATE(1226), 1,
      sym_jsx_namespace_name,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    STATE(1305), 2,
      sym_jsx_expression,
      sym_jsx_attribute,
  [49158] = 9,
    ACTIONS(2747), 1,
      sym_identifier,
    ACTIONS(2749), 1,
      anon_sym_LBRACE,
    ACTIONS(2755), 1,
      sym_jsx_identifier,
    ACTIONS(2800), 1,
      anon_sym_GT,
    ACTIONS(2853), 1,
      anon_sym_SLASH_GT,
    STATE(1171), 1,
      aux_sym_jsx_opening_element_repeat1,
    STATE(1226), 1,
      sym_jsx_namespace_name,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    STATE(1305), 2,
      sym_jsx_expression,
      sym_jsx_attribute,
  [49188] = 9,
    ACTIONS(2747), 1,
      sym_identifier,
    ACTIONS(2749), 1,
      anon_sym_LBRACE,
    ACTIONS(2755), 1,
      sym_jsx_identifier,
    ACTIONS(2839), 1,
      anon_sym_GT,
    ACTIONS(2855), 1,
      anon_sym_SLASH_GT,
    STATE(1171), 1,
      aux_sym_jsx_opening_element_repeat1,
    STATE(1226), 1,
      sym_jsx_namespace_name,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    STATE(1305), 2,
      sym_jsx_expression,
      sym_jsx_attribute,
  [49218] = 9,
    ACTIONS(2747), 1,
      sym_identifier,
    ACTIONS(2749), 1,
      anon_sym_LBRACE,
    ACTIONS(2755), 1,
      sym_jsx_identifier,
    ACTIONS(2845), 1,
      anon_sym_GT,
    ACTIONS(2857), 1,
      anon_sym_SLASH_GT,
    STATE(1171), 1,
      aux_sym_jsx_opening_element_repeat1,
    STATE(1226), 1,
      sym_jsx_namespace_name,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    STATE(1305), 2,
      sym_jsx_expression,
      sym_jsx_attribute,
  [49248] = 9,
    ACTIONS(2747), 1,
      sym_identifier,
    ACTIONS(2749), 1,
      anon_sym_LBRACE,
    ACTIONS(2755), 1,
      sym_jsx_identifier,
    ACTIONS(2845), 1,
      anon_sym_GT,
    ACTIONS(2859), 1,
      anon_sym_SLASH_GT,
    STATE(1171), 1,
      aux_sym_jsx_opening_element_repeat1,
    STATE(1226), 1,
      sym_jsx_namespace_name,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    STATE(1305), 2,
      sym_jsx_expression,
      sym_jsx_attribute,
  [49278] = 9,
    ACTIONS(2747), 1,
      sym_identifier,
    ACTIONS(2749), 1,
      anon_sym_LBRACE,
    ACTIONS(2755), 1,
      sym_jsx_identifier,
    ACTIONS(2804), 1,
      anon_sym_GT,
    ACTIONS(2861), 1,
      anon_sym_SLASH_GT,
    STATE(1171), 1,
      aux_sym_jsx_opening_element_repeat1,
    STATE(1226), 1,
      sym_jsx_namespace_name,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    STATE(1305), 2,
      sym_jsx_expression,
      sym_jsx_attribute,
  [49308] = 9,
    ACTIONS(2747), 1,
      sym_identifier,
    ACTIONS(2749), 1,
      anon_sym_LBRACE,
    ACTIONS(2755), 1,
      sym_jsx_identifier,
    ACTIONS(2845), 1,
      anon_sym_GT,
    ACTIONS(2863), 1,
      anon_sym_SLASH_GT,
    STATE(1171), 1,
      aux_sym_jsx_opening_element_repeat1,
    STATE(1226), 1,
      sym_jsx_namespace_name,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    STATE(1305), 2,
      sym_jsx_expression,
      sym_jsx_attribute,
  [49338] = 9,
    ACTIONS(2747), 1,
      sym_identifier,
    ACTIONS(2749), 1,
      anon_sym_LBRACE,
    ACTIONS(2755), 1,
      sym_jsx_identifier,
    ACTIONS(2804), 1,
      anon_sym_GT,
    ACTIONS(2865), 1,
      anon_sym_SLASH_GT,
    STATE(1171), 1,
      aux_sym_jsx_opening_element_repeat1,
    STATE(1226), 1,
      sym_jsx_namespace_name,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    STATE(1305), 2,
      sym_jsx_expression,
      sym_jsx_attribute,
  [49368] = 9,
    ACTIONS(2747), 1,
      sym_identifier,
    ACTIONS(2749), 1,
      anon_sym_LBRACE,
    ACTIONS(2755), 1,
      sym_jsx_identifier,
    ACTIONS(2800), 1,
      anon_sym_GT,
    ACTIONS(2867), 1,
      anon_sym_SLASH_GT,
    STATE(1171), 1,
      aux_sym_jsx_opening_element_repeat1,
    STATE(1226), 1,
      sym_jsx_namespace_name,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    STATE(1305), 2,
      sym_jsx_expression,
      sym_jsx_attribute,
  [49398] = 9,
    ACTIONS(2747), 1,
      sym_identifier,
    ACTIONS(2749), 1,
      anon_sym_LBRACE,
    ACTIONS(2755), 1,
      sym_jsx_identifier,
    ACTIONS(2800), 1,
      anon_sym_GT,
    ACTIONS(2869), 1,
      anon_sym_SLASH_GT,
    STATE(1171), 1,
      aux_sym_jsx_opening_element_repeat1,
    STATE(1226), 1,
      sym_jsx_namespace_name,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    STATE(1305), 2,
      sym_jsx_expression,
      sym_jsx_attribute,
  [49428] = 9,
    ACTIONS(2747), 1,
      sym_identifier,
    ACTIONS(2749), 1,
      anon_sym_LBRACE,
    ACTIONS(2753), 1,
      anon_sym_GT,
    ACTIONS(2755), 1,
      sym_jsx_identifier,
    ACTIONS(2781), 1,
      anon_sym_SLASH_GT,
    STATE(1166), 1,
      aux_sym_jsx_opening_element_repeat1,
    STATE(1226), 1,
      sym_jsx_namespace_name,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    STATE(1305), 2,
      sym_jsx_expression,
      sym_jsx_attribute,
  [49458] = 9,
    ACTIONS(2747), 1,
      sym_identifier,
    ACTIONS(2749), 1,
      anon_sym_LBRACE,
    ACTIONS(2753), 1,
      anon_sym_GT,
    ACTIONS(2755), 1,
      sym_jsx_identifier,
    ACTIONS(2769), 1,
      anon_sym_SLASH_GT,
    STATE(1180), 1,
      aux_sym_jsx_opening_element_repeat1,
    STATE(1226), 1,
      sym_jsx_namespace_name,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    STATE(1305), 2,
      sym_jsx_expression,
      sym_jsx_attribute,
  [49488] = 7,
    ACTIONS(1100), 1,
      sym_identifier,
    ACTIONS(1102), 1,
      anon_sym_LBRACE,
    ACTIONS(1104), 1,
      anon_sym_of,
    ACTIONS(1106), 1,
      anon_sym_LBRACK,
    STATE(1390), 1,
      sym_variable_declarator,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    STATE(1311), 3,
      sym_object_pattern,
      sym_array_pattern,
      sym__destructuring_pattern,
  [49513] = 7,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(2871), 1,
      anon_sym_RBRACE,
    STATE(1541), 1,
      sym_export_specifier,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2833), 2,
      anon_sym_default,
      sym_identifier,
    STATE(1423), 2,
      sym__module_export_name,
      sym_string,
  [49538] = 8,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(2814), 1,
      sym_identifier,
    ACTIONS(2816), 1,
      anon_sym_default,
    ACTIONS(2873), 1,
      anon_sym_RBRACE,
    STATE(1592), 1,
      sym_import_specifier,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    STATE(1775), 2,
      sym__module_export_name,
      sym_string,
  [49565] = 7,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(2875), 1,
      anon_sym_RBRACE,
    STATE(1541), 1,
      sym_export_specifier,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2833), 2,
      anon_sym_default,
      sym_identifier,
    STATE(1423), 2,
      sym__module_export_name,
      sym_string,
  [49590] = 7,
    ACTIONS(1102), 1,
      anon_sym_LBRACE,
    ACTIONS(1106), 1,
      anon_sym_LBRACK,
    ACTIONS(1422), 1,
      sym_identifier,
    ACTIONS(2877), 1,
      anon_sym_of,
    STATE(1385), 1,
      sym_variable_declarator,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    STATE(1202), 3,
      sym_object_pattern,
      sym_array_pattern,
      sym__destructuring_pattern,
  [49615] = 7,
    ACTIONS(1100), 1,
      sym_identifier,
    ACTIONS(1102), 1,
      anon_sym_LBRACE,
    ACTIONS(1104), 1,
      anon_sym_of,
    ACTIONS(1106), 1,
      anon_sym_LBRACK,
    STATE(1391), 1,
      sym_variable_declarator,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    STATE(1311), 3,
      sym_object_pattern,
      sym_array_pattern,
      sym__destructuring_pattern,
  [49640] = 7,
    ACTIONS(1100), 1,
      sym_identifier,
    ACTIONS(1102), 1,
      anon_sym_LBRACE,
    ACTIONS(1104), 1,
      anon_sym_of,
    ACTIONS(1106), 1,
      anon_sym_LBRACK,
    STATE(1380), 1,
      sym_variable_declarator,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    STATE(1311), 3,
      sym_object_pattern,
      sym_array_pattern,
      sym__destructuring_pattern,
  [49665] = 8,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(2814), 1,
      sym_identifier,
    ACTIONS(2816), 1,
      anon_sym_default,
    ACTIONS(2879), 1,
      anon_sym_RBRACE,
    STATE(1592), 1,
      sym_import_specifier,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    STATE(1775), 2,
      sym__module_export_name,
      sym_string,
  [49692] = 7,
    ACTIONS(1100), 1,
      sym_identifier,
    ACTIONS(1102), 1,
      anon_sym_LBRACE,
    ACTIONS(1104), 1,
      anon_sym_of,
    ACTIONS(1106), 1,
      anon_sym_LBRACK,
    STATE(1385), 1,
      sym_variable_declarator,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    STATE(1311), 3,
      sym_object_pattern,
      sym_array_pattern,
      sym__destructuring_pattern,
  [49717] = 7,
    ACTIONS(1102), 1,
      anon_sym_LBRACE,
    ACTIONS(1106), 1,
      anon_sym_LBRACK,
    ACTIONS(2881), 1,
      sym_identifier,
    ACTIONS(2883), 1,
      anon_sym_of,
    STATE(1380), 1,
      sym_variable_declarator,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    STATE(1208), 3,
      sym_object_pattern,
      sym_array_pattern,
      sym__destructuring_pattern,
  [49742] = 7,
    ACTIONS(1100), 1,
      sym_identifier,
    ACTIONS(1102), 1,
      anon_sym_LBRACE,
    ACTIONS(1104), 1,
      anon_sym_of,
    ACTIONS(1106), 1,
      anon_sym_LBRACK,
    STATE(1415), 1,
      sym_variable_declarator,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    STATE(1311), 3,
      sym_object_pattern,
      sym_array_pattern,
      sym__destructuring_pattern,
  [49767] = 6,
    ACTIONS(862), 1,
      anon_sym_EQ,
    ACTIONS(2889), 1,
      sym__automatic_semicolon,
    STATE(1425), 1,
      sym__initializer,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2885), 2,
      anon_sym_COMMA,
      anon_sym_SEMI,
    ACTIONS(2887), 2,
      anon_sym_of,
      anon_sym_in,
  [49789] = 6,
    ACTIONS(1102), 1,
      anon_sym_LBRACE,
    ACTIONS(1106), 1,
      anon_sym_LBRACK,
    ACTIONS(2892), 1,
      sym_identifier,
    ACTIONS(2894), 1,
      anon_sym_of,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    STATE(1495), 3,
      sym_object_pattern,
      sym_array_pattern,
      sym__destructuring_pattern,
  [49811] = 4,
    ACTIONS(2896), 1,
      anon_sym_COMMA,
    STATE(1204), 1,
      aux_sym_sequence_expression_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1836), 5,
      anon_sym_RBRACE,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_RBRACK,
  [49829] = 6,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    STATE(1541), 1,
      sym_export_specifier,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2833), 2,
      anon_sym_default,
      sym_identifier,
    STATE(1423), 2,
      sym__module_export_name,
      sym_string,
  [49851] = 2,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2899), 7,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_RPAREN,
      anon_sym_in,
      anon_sym_EQ,
      anon_sym_RBRACK,
  [49865] = 2,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2901), 7,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_RPAREN,
      anon_sym_in,
      anon_sym_EQ,
      anon_sym_RBRACK,
  [49879] = 5,
    ACTIONS(2903), 1,
      anon_sym_EQ,
    STATE(1285), 1,
      sym__initializer,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2887), 2,
      anon_sym_of,
      anon_sym_in,
    ACTIONS(2885), 3,
      sym__automatic_semicolon,
      anon_sym_COMMA,
      anon_sym_SEMI,
  [49899] = 6,
    ACTIONS(862), 1,
      anon_sym_EQ,
    ACTIONS(864), 1,
      sym__automatic_semicolon,
    STATE(1408), 1,
      sym__initializer,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(858), 2,
      anon_sym_COMMA,
      anon_sym_SEMI,
    ACTIONS(2905), 2,
      anon_sym_of,
      anon_sym_in,
  [49921] = 7,
    ACTIONS(103), 1,
      anon_sym_COMMA,
    ACTIONS(2397), 1,
      anon_sym_EQ,
    ACTIONS(2401), 1,
      anon_sym_RBRACE,
    STATE(1418), 1,
      aux_sym_object_repeat1,
    STATE(1452), 1,
      aux_sym_object_pattern_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2395), 2,
      anon_sym_LPAREN,
      anon_sym_COLON,
  [49945] = 7,
    ACTIONS(103), 1,
      anon_sym_COMMA,
    ACTIONS(2397), 1,
      anon_sym_EQ,
    ACTIONS(2399), 1,
      anon_sym_RBRACE,
    STATE(1418), 1,
      aux_sym_object_repeat1,
    STATE(1452), 1,
      aux_sym_object_pattern_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2395), 2,
      anon_sym_LPAREN,
      anon_sym_COLON,
  [49969] = 7,
    ACTIONS(103), 1,
      anon_sym_COMMA,
    ACTIONS(216), 1,
      anon_sym_RBRACE,
    ACTIONS(2397), 1,
      anon_sym_EQ,
    STATE(1418), 1,
      aux_sym_object_repeat1,
    STATE(1452), 1,
      aux_sym_object_pattern_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2395), 2,
      anon_sym_LPAREN,
      anon_sym_COLON,
  [49993] = 6,
    ACTIONS(1102), 1,
      anon_sym_LBRACE,
    ACTIONS(1106), 1,
      anon_sym_LBRACK,
    ACTIONS(1469), 1,
      sym_identifier,
    ACTIONS(2907), 1,
      anon_sym_of,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    STATE(1469), 3,
      sym_object_pattern,
      sym_array_pattern,
      sym__destructuring_pattern,
  [50015] = 7,
    ACTIONS(103), 1,
      anon_sym_COMMA,
    ACTIONS(2393), 1,
      anon_sym_RBRACE,
    ACTIONS(2397), 1,
      anon_sym_EQ,
    STATE(1418), 1,
      aux_sym_object_repeat1,
    STATE(1452), 1,
      aux_sym_object_pattern_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2395), 2,
      anon_sym_LPAREN,
      anon_sym_COLON,
  [50039] = 5,
    ACTIONS(2751), 1,
      anon_sym_COLON,
    ACTIONS(2909), 1,
      sym_identifier,
    ACTIONS(2913), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2911), 4,
      anon_sym_LBRACE,
      anon_sym_GT,
      sym_jsx_identifier,
      anon_sym_SLASH_GT,
  [50059] = 5,
    ACTIONS(2903), 1,
      anon_sym_EQ,
    STATE(1283), 1,
      sym__initializer,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2905), 2,
      anon_sym_of,
      anon_sym_in,
    ACTIONS(858), 3,
      sym__automatic_semicolon,
      anon_sym_COMMA,
      anon_sym_SEMI,
  [50079] = 7,
    ACTIONS(103), 1,
      anon_sym_COMMA,
    ACTIONS(171), 1,
      anon_sym_RBRACE,
    ACTIONS(2397), 1,
      anon_sym_EQ,
    STATE(1452), 1,
      aux_sym_object_pattern_repeat1,
    STATE(1487), 1,
      aux_sym_object_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2395), 2,
      anon_sym_LPAREN,
      anon_sym_COLON,
  [50103] = 2,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2915), 7,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_RPAREN,
      anon_sym_in,
      anon_sym_EQ,
      anon_sym_RBRACK,
  [50117] = 2,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2917), 7,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_RPAREN,
      anon_sym_in,
      anon_sym_EQ,
      anon_sym_RBRACK,
  [50131] = 7,
    ACTIONS(103), 1,
      anon_sym_COMMA,
    ACTIONS(351), 1,
      anon_sym_RBRACE,
    ACTIONS(2397), 1,
      anon_sym_EQ,
    STATE(1418), 1,
      aux_sym_object_repeat1,
    STATE(1452), 1,
      aux_sym_object_pattern_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2395), 2,
      anon_sym_LPAREN,
      anon_sym_COLON,
  [50155] = 7,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(2814), 1,
      sym_identifier,
    ACTIONS(2816), 1,
      anon_sym_default,
    STATE(1592), 1,
      sym_import_specifier,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    STATE(1775), 2,
      sym__module_export_name,
      sym_string,
  [50179] = 2,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2919), 7,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_of,
      anon_sym_RPAREN,
      anon_sym_in,
      anon_sym_EQ,
      anon_sym_RBRACK,
  [50193] = 6,
    ACTIONS(1969), 1,
      anon_sym_LBRACE,
    ACTIONS(2921), 1,
      sym_identifier,
    ACTIONS(2923), 1,
      anon_sym_of,
    ACTIONS(2925), 1,
      anon_sym_LBRACK,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    STATE(1338), 3,
      sym_object_pattern,
      sym_array_pattern,
      sym__destructuring_pattern,
  [50215] = 4,
    ACTIONS(1983), 1,
      anon_sym_COMMA,
    STATE(1204), 1,
      aux_sym_sequence_expression_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2927), 5,
      anon_sym_RBRACE,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_COLON,
      anon_sym_RBRACK,
  [50233] = 3,
    ACTIONS(2929), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2931), 5,
      anon_sym_LBRACE,
      anon_sym_EQ,
      anon_sym_GT,
      sym_jsx_identifier,
      anon_sym_SLASH_GT,
  [50248] = 4,
    ACTIONS(2933), 1,
      sym_identifier,
    ACTIONS(2937), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2935), 4,
      anon_sym_LBRACE,
      anon_sym_GT,
      sym_jsx_identifier,
      anon_sym_SLASH_GT,
  [50265] = 2,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1912), 6,
      sym__automatic_semicolon,
      anon_sym_COMMA,
      anon_sym_of,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_EQ,
  [50278] = 6,
    ACTIONS(862), 1,
      anon_sym_EQ,
    ACTIONS(2939), 1,
      anon_sym_LPAREN,
    STATE(1515), 1,
      sym_formal_parameters,
    STATE(1702), 1,
      sym__initializer,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2941), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [50299] = 5,
    ACTIONS(2945), 1,
      anon_sym_BQUOTE,
    ACTIONS(2947), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2943), 2,
      sym__template_chars,
      sym_escape_sequence,
    STATE(1240), 2,
      sym_template_substitution,
      aux_sym_template_string_repeat1,
  [50318] = 3,
    ACTIONS(2949), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2951), 5,
      anon_sym_LBRACE,
      anon_sym_GT,
      sym_jsx_identifier,
      anon_sym_DOT,
      anon_sym_SLASH_GT,
  [50333] = 6,
    ACTIONS(862), 1,
      anon_sym_EQ,
    ACTIONS(2939), 1,
      anon_sym_LPAREN,
    STATE(1500), 1,
      sym__initializer,
    STATE(1556), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2953), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [50354] = 6,
    ACTIONS(862), 1,
      anon_sym_EQ,
    ACTIONS(2939), 1,
      anon_sym_LPAREN,
    STATE(1557), 1,
      sym_formal_parameters,
    STATE(1678), 1,
      sym__initializer,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2955), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [50375] = 5,
    ACTIONS(2960), 1,
      anon_sym_BQUOTE,
    ACTIONS(2962), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2957), 2,
      sym__template_chars,
      sym_escape_sequence,
    STATE(1233), 2,
      sym_template_substitution,
      aux_sym_template_string_repeat1,
  [50394] = 2,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2919), 6,
      sym__automatic_semicolon,
      anon_sym_COMMA,
      anon_sym_of,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_EQ,
  [50407] = 2,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1753), 6,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_from,
      anon_sym_LPAREN,
      anon_sym_COLON,
  [50420] = 5,
    ACTIONS(2965), 1,
      anon_sym_default,
    ACTIONS(2967), 1,
      anon_sym_RBRACE,
    ACTIONS(2969), 1,
      anon_sym_case,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    STATE(1246), 3,
      sym_switch_case,
      sym_switch_default,
      aux_sym_switch_body_repeat1,
  [50439] = 2,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1680), 6,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_from,
      anon_sym_LPAREN,
      anon_sym_COLON,
  [50452] = 2,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2899), 6,
      sym__automatic_semicolon,
      anon_sym_COMMA,
      anon_sym_of,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_EQ,
  [50465] = 5,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2971), 2,
      anon_sym_default,
      sym_identifier,
    STATE(1805), 2,
      sym__module_export_name,
      sym_string,
  [50484] = 5,
    ACTIONS(2947), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(2975), 1,
      anon_sym_BQUOTE,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2973), 2,
      sym__template_chars,
      sym_escape_sequence,
    STATE(1233), 2,
      sym_template_substitution,
      aux_sym_template_string_repeat1,
  [50503] = 2,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1914), 6,
      sym__automatic_semicolon,
      anon_sym_COMMA,
      anon_sym_of,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_EQ,
  [50516] = 5,
    ACTIONS(2947), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(2979), 1,
      anon_sym_BQUOTE,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2977), 2,
      sym__template_chars,
      sym_escape_sequence,
    STATE(1244), 2,
      sym_template_substitution,
      aux_sym_template_string_repeat1,
  [50535] = 5,
    ACTIONS(2981), 1,
      anon_sym_default,
    ACTIONS(2984), 1,
      anon_sym_RBRACE,
    ACTIONS(2986), 1,
      anon_sym_case,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    STATE(1243), 3,
      sym_switch_case,
      sym_switch_default,
      aux_sym_switch_body_repeat1,
  [50554] = 5,
    ACTIONS(2947), 1,
      anon_sym_DOLLAR_LBRACE,
    ACTIONS(2989), 1,
      anon_sym_BQUOTE,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2973), 2,
      sym__template_chars,
      sym_escape_sequence,
    STATE(1233), 2,
      sym_template_substitution,
      aux_sym_template_string_repeat1,
  [50573] = 6,
    ACTIONS(862), 1,
      anon_sym_EQ,
    ACTIONS(2939), 1,
      anon_sym_LPAREN,
    STATE(1617), 1,
      sym_formal_parameters,
    STATE(1732), 1,
      sym__initializer,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2991), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [50594] = 5,
    ACTIONS(2965), 1,
      anon_sym_default,
    ACTIONS(2969), 1,
      anon_sym_case,
    ACTIONS(2993), 1,
      anon_sym_RBRACE,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    STATE(1243), 3,
      sym_switch_case,
      sym_switch_default,
      aux_sym_switch_body_repeat1,
  [50613] = 2,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2915), 6,
      sym__automatic_semicolon,
      anon_sym_COMMA,
      anon_sym_of,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_EQ,
  [50626] = 2,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2901), 6,
      sym__automatic_semicolon,
      anon_sym_COMMA,
      anon_sym_of,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_EQ,
  [50639] = 2,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2917), 6,
      sym__automatic_semicolon,
      anon_sym_COMMA,
      anon_sym_of,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_EQ,
  [50652] = 5,
    ACTIONS(1075), 1,
      anon_sym_DQUOTE,
    ACTIONS(1077), 1,
      anon_sym_SQUOTE,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2995), 2,
      anon_sym_default,
      sym_identifier,
    STATE(1543), 2,
      sym__module_export_name,
      sym_string,
  [50671] = 2,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1916), 6,
      sym__automatic_semicolon,
      anon_sym_COMMA,
      anon_sym_of,
      anon_sym_SEMI,
      anon_sym_in,
      anon_sym_EQ,
  [50684] = 5,
    ACTIONS(1969), 1,
      anon_sym_LBRACE,
    ACTIONS(2925), 1,
      anon_sym_LBRACK,
    ACTIONS(2997), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    STATE(1840), 3,
      sym_object_pattern,
      sym_array_pattern,
      sym__destructuring_pattern,
  [50703] = 3,
    ACTIONS(2999), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1543), 4,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
  [50717] = 3,
    ACTIONS(3003), 1,
      anon_sym_LT,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(3001), 4,
      sym_jsx_text,
      anon_sym_LBRACE,
      sym_html_character_reference,
      anon_sym_LT_SLASH,
  [50731] = 3,
    ACTIONS(1658), 1,
      anon_sym_LT,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1660), 4,
      sym_jsx_text,
      anon_sym_LBRACE,
      sym_html_character_reference,
      anon_sym_LT_SLASH,
  [50745] = 3,
    ACTIONS(1603), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1605), 4,
      anon_sym_LBRACE,
      anon_sym_GT,
      sym_jsx_identifier,
      anon_sym_SLASH_GT,
  [50759] = 3,
    ACTIONS(3005), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(3007), 4,
      anon_sym_LBRACE,
      anon_sym_GT,
      sym_jsx_identifier,
      anon_sym_SLASH_GT,
  [50773] = 3,
    ACTIONS(1666), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1668), 4,
      anon_sym_LBRACE,
      anon_sym_GT,
      sym_jsx_identifier,
      anon_sym_SLASH_GT,
  [50787] = 3,
    ACTIONS(1670), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1672), 4,
      anon_sym_LBRACE,
      anon_sym_GT,
      sym_jsx_identifier,
      anon_sym_SLASH_GT,
  [50801] = 6,
    ACTIONS(3009), 1,
      sym_identifier,
    ACTIONS(3011), 1,
      anon_sym_GT,
    ACTIONS(3013), 1,
      sym_jsx_identifier,
    STATE(1163), 1,
      sym_nested_identifier,
    STATE(1176), 1,
      sym_jsx_namespace_name,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [50821] = 6,
    ACTIONS(3015), 1,
      sym_identifier,
    ACTIONS(3017), 1,
      anon_sym_LBRACE,
    ACTIONS(3019), 1,
      anon_sym_extends,
    STATE(732), 1,
      sym_class_body,
    STATE(1510), 1,
      sym_class_heritage,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [50841] = 6,
    ACTIONS(2357), 1,
      anon_sym_COMMA,
    ACTIONS(2395), 1,
      anon_sym_COLON,
    ACTIONS(2397), 1,
      anon_sym_EQ,
    ACTIONS(3021), 1,
      anon_sym_RBRACE,
    STATE(1419), 1,
      aux_sym_object_pattern_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [50861] = 3,
    ACTIONS(1694), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1696), 4,
      anon_sym_LBRACE,
      anon_sym_GT,
      sym_jsx_identifier,
      anon_sym_SLASH_GT,
  [50875] = 3,
    ACTIONS(1694), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1696), 4,
      anon_sym_LBRACE,
      anon_sym_GT,
      sym_jsx_identifier,
      anon_sym_SLASH_GT,
  [50889] = 3,
    ACTIONS(1694), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1696), 4,
      anon_sym_LBRACE,
      anon_sym_GT,
      sym_jsx_identifier,
      anon_sym_SLASH_GT,
  [50903] = 3,
    ACTIONS(3025), 1,
      anon_sym_LT,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(3023), 4,
      sym_jsx_text,
      anon_sym_LBRACE,
      sym_html_character_reference,
      anon_sym_LT_SLASH,
  [50917] = 6,
    ACTIONS(93), 1,
      anon_sym_AT,
    ACTIONS(3027), 1,
      anon_sym_export,
    ACTIONS(3029), 1,
      anon_sym_class,
    STATE(1041), 1,
      aux_sym_export_statement_repeat1,
    STATE(1097), 1,
      sym_decorator,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [50937] = 6,
    ACTIONS(3017), 1,
      anon_sym_LBRACE,
    ACTIONS(3019), 1,
      anon_sym_extends,
    ACTIONS(3031), 1,
      sym_identifier,
    STATE(732), 1,
      sym_class_body,
    STATE(1510), 1,
      sym_class_heritage,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [50957] = 3,
    ACTIONS(1694), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1696), 4,
      anon_sym_LBRACE,
      anon_sym_GT,
      sym_jsx_identifier,
      anon_sym_SLASH_GT,
  [50971] = 3,
    ACTIONS(1599), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1601), 4,
      anon_sym_LBRACE,
      anon_sym_GT,
      sym_jsx_identifier,
      anon_sym_SLASH_GT,
  [50985] = 6,
    ACTIONS(3017), 1,
      anon_sym_LBRACE,
    ACTIONS(3019), 1,
      anon_sym_extends,
    ACTIONS(3033), 1,
      sym_identifier,
    STATE(748), 1,
      sym_class_body,
    STATE(1656), 1,
      sym_class_heritage,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [51005] = 3,
    ACTIONS(3037), 1,
      anon_sym_LT,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(3035), 4,
      sym_jsx_text,
      anon_sym_LBRACE,
      sym_html_character_reference,
      anon_sym_LT_SLASH,
  [51019] = 6,
    ACTIONS(3017), 1,
      anon_sym_LBRACE,
    ACTIONS(3019), 1,
      anon_sym_extends,
    ACTIONS(3039), 1,
      sym_identifier,
    STATE(748), 1,
      sym_class_body,
    STATE(1656), 1,
      sym_class_heritage,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [51039] = 2,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1680), 5,
      sym__automatic_semicolon,
      anon_sym_with,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_EQ,
  [51051] = 3,
    ACTIONS(1603), 1,
      anon_sym_LT,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1605), 4,
      sym_jsx_text,
      anon_sym_LBRACE,
      sym_html_character_reference,
      anon_sym_LT_SLASH,
  [51065] = 3,
    ACTIONS(3041), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(3043), 4,
      anon_sym_LBRACE,
      anon_sym_GT,
      sym_jsx_identifier,
      anon_sym_SLASH_GT,
  [51079] = 5,
    ACTIONS(2273), 1,
      anon_sym_COMMA,
    ACTIONS(2416), 1,
      anon_sym_RBRACE,
    STATE(1464), 1,
      aux_sym_object_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2395), 2,
      anon_sym_LPAREN,
      anon_sym_COLON,
  [51097] = 3,
    ACTIONS(3037), 1,
      anon_sym_LT,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(3035), 4,
      sym_jsx_text,
      anon_sym_LBRACE,
      sym_html_character_reference,
      anon_sym_LT_SLASH,
  [51111] = 6,
    ACTIONS(3045), 1,
      sym_identifier,
    ACTIONS(3047), 1,
      anon_sym_GT,
    ACTIONS(3049), 1,
      sym_jsx_identifier,
    STATE(1736), 1,
      sym_nested_identifier,
    STATE(1846), 1,
      sym_jsx_namespace_name,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [51131] = 2,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1530), 5,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_RPAREN,
      anon_sym_EQ,
      anon_sym_RBRACK,
  [51143] = 3,
    ACTIONS(1666), 1,
      anon_sym_LT,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1668), 4,
      sym_jsx_text,
      anon_sym_LBRACE,
      sym_html_character_reference,
      anon_sym_LT_SLASH,
  [51157] = 3,
    ACTIONS(3051), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1530), 4,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
  [51171] = 4,
    ACTIONS(3056), 1,
      anon_sym_of,
    ACTIONS(3058), 1,
      anon_sym_in,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(3054), 3,
      sym__automatic_semicolon,
      anon_sym_COMMA,
      anon_sym_SEMI,
  [51187] = 6,
    ACTIONS(3017), 1,
      anon_sym_LBRACE,
    ACTIONS(3019), 1,
      anon_sym_extends,
    ACTIONS(3060), 1,
      sym_identifier,
    STATE(732), 1,
      sym_class_body,
    STATE(1510), 1,
      sym_class_heritage,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [51207] = 4,
    ACTIONS(3064), 1,
      anon_sym_of,
    ACTIONS(3066), 1,
      anon_sym_in,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(3062), 3,
      sym__automatic_semicolon,
      anon_sym_COMMA,
      anon_sym_SEMI,
  [51223] = 6,
    ACTIONS(3019), 1,
      anon_sym_extends,
    ACTIONS(3068), 1,
      sym_identifier,
    ACTIONS(3070), 1,
      anon_sym_LBRACE,
    STATE(604), 1,
      sym_class_body,
    STATE(1738), 1,
      sym_class_heritage,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [51243] = 3,
    ACTIONS(3074), 1,
      anon_sym_LT,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(3072), 4,
      sym_jsx_text,
      anon_sym_LBRACE,
      sym_html_character_reference,
      anon_sym_LT_SLASH,
  [51257] = 6,
    ACTIONS(3011), 1,
      anon_sym_GT,
    ACTIONS(3076), 1,
      sym_identifier,
    ACTIONS(3078), 1,
      sym_jsx_identifier,
    STATE(1158), 1,
      sym_nested_identifier,
    STATE(1190), 1,
      sym_jsx_namespace_name,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [51277] = 6,
    ACTIONS(3017), 1,
      anon_sym_LBRACE,
    ACTIONS(3019), 1,
      anon_sym_extends,
    ACTIONS(3080), 1,
      sym_identifier,
    STATE(748), 1,
      sym_class_body,
    STATE(1656), 1,
      sym_class_heritage,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [51297] = 3,
    ACTIONS(1670), 1,
      anon_sym_LT,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1672), 4,
      sym_jsx_text,
      anon_sym_LBRACE,
      sym_html_character_reference,
      anon_sym_LT_SLASH,
  [51311] = 3,
    ACTIONS(3082), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(3084), 4,
      anon_sym_LBRACE,
      anon_sym_GT,
      sym_jsx_identifier,
      anon_sym_SLASH_GT,
  [51325] = 3,
    ACTIONS(3086), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(3088), 4,
      anon_sym_LBRACE,
      anon_sym_GT,
      sym_jsx_identifier,
      anon_sym_SLASH_GT,
  [51339] = 6,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(3090), 1,
      sym_identifier,
    ACTIONS(3092), 1,
      anon_sym_LBRACK,
    ACTIONS(3094), 1,
      sym_private_property_identifier,
    STATE(609), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [51359] = 3,
    ACTIONS(1694), 1,
      anon_sym_LT,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1696), 4,
      sym_jsx_text,
      anon_sym_LBRACE,
      sym_html_character_reference,
      anon_sym_LT_SLASH,
  [51373] = 6,
    ACTIONS(3096), 1,
      sym_identifier,
    ACTIONS(3098), 1,
      anon_sym_GT,
    ACTIONS(3100), 1,
      sym_jsx_identifier,
    STATE(1550), 1,
      sym_nested_identifier,
    STATE(1799), 1,
      sym_jsx_namespace_name,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [51393] = 6,
    ACTIONS(2357), 1,
      anon_sym_COMMA,
    ACTIONS(2395), 1,
      anon_sym_COLON,
    ACTIONS(2397), 1,
      anon_sym_EQ,
    ACTIONS(3102), 1,
      anon_sym_RBRACE,
    STATE(1452), 1,
      aux_sym_object_pattern_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [51413] = 4,
    ACTIONS(2397), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2395), 2,
      anon_sym_LPAREN,
      anon_sym_COLON,
    ACTIONS(2436), 2,
      anon_sym_COMMA,
      anon_sym_RBRACE,
  [51429] = 6,
    ACTIONS(3019), 1,
      anon_sym_extends,
    ACTIONS(3070), 1,
      anon_sym_LBRACE,
    ACTIONS(3104), 1,
      sym_identifier,
    STATE(615), 1,
      sym_class_body,
    STATE(1554), 1,
      sym_class_heritage,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [51449] = 3,
    ACTIONS(1694), 1,
      anon_sym_LT,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1696), 4,
      sym_jsx_text,
      anon_sym_LBRACE,
      sym_html_character_reference,
      anon_sym_LT_SLASH,
  [51463] = 3,
    ACTIONS(1694), 1,
      anon_sym_LT,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1696), 4,
      sym_jsx_text,
      anon_sym_LBRACE,
      sym_html_character_reference,
      anon_sym_LT_SLASH,
  [51477] = 3,
    ACTIONS(3108), 1,
      anon_sym_LT,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(3106), 4,
      sym_jsx_text,
      anon_sym_LBRACE,
      sym_html_character_reference,
      anon_sym_LT_SLASH,
  [51491] = 3,
    ACTIONS(1694), 1,
      anon_sym_LT,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1696), 4,
      sym_jsx_text,
      anon_sym_LBRACE,
      sym_html_character_reference,
      anon_sym_LT_SLASH,
  [51505] = 6,
    ACTIONS(3011), 1,
      anon_sym_GT,
    ACTIONS(3110), 1,
      sym_identifier,
    ACTIONS(3112), 1,
      sym_jsx_identifier,
    STATE(1159), 1,
      sym_nested_identifier,
    STATE(1189), 1,
      sym_jsx_namespace_name,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [51525] = 4,
    ACTIONS(862), 1,
      anon_sym_EQ,
    STATE(1408), 1,
      sym__initializer,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(858), 3,
      sym__automatic_semicolon,
      anon_sym_COMMA,
      anon_sym_SEMI,
  [51541] = 3,
    ACTIONS(3114), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(3116), 4,
      anon_sym_LBRACE,
      anon_sym_GT,
      sym_jsx_identifier,
      anon_sym_SLASH_GT,
  [51555] = 6,
    ACTIONS(93), 1,
      anon_sym_AT,
    ACTIONS(3118), 1,
      anon_sym_export,
    ACTIONS(3120), 1,
      anon_sym_class,
    STATE(1041), 1,
      aux_sym_export_statement_repeat1,
    STATE(1097), 1,
      sym_decorator,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [51575] = 3,
    ACTIONS(3122), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1530), 4,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
  [51589] = 3,
    ACTIONS(1599), 1,
      anon_sym_LT,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1601), 4,
      sym_jsx_text,
      anon_sym_LBRACE,
      sym_html_character_reference,
      anon_sym_LT_SLASH,
  [51603] = 3,
    ACTIONS(3037), 1,
      anon_sym_LT,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(3035), 4,
      sym_jsx_text,
      anon_sym_LBRACE,
      sym_html_character_reference,
      anon_sym_LT_SLASH,
  [51617] = 6,
    ACTIONS(1818), 1,
      anon_sym_LPAREN,
    ACTIONS(3125), 1,
      sym_identifier,
    ACTIONS(3127), 1,
      anon_sym_LBRACK,
    ACTIONS(3129), 1,
      sym_private_property_identifier,
    STATE(784), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [51637] = 4,
    ACTIONS(862), 1,
      anon_sym_EQ,
    STATE(1425), 1,
      sym__initializer,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2885), 3,
      sym__automatic_semicolon,
      anon_sym_COMMA,
      anon_sym_SEMI,
  [51653] = 3,
    ACTIONS(1658), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1660), 4,
      anon_sym_LBRACE,
      anon_sym_GT,
      sym_jsx_identifier,
      anon_sym_SLASH_GT,
  [51667] = 6,
    ACTIONS(3131), 1,
      sym_identifier,
    ACTIONS(3133), 1,
      anon_sym_GT,
    ACTIONS(3135), 1,
      sym_jsx_identifier,
    STATE(1534), 1,
      sym_nested_identifier,
    STATE(1814), 1,
      sym_jsx_namespace_name,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [51687] = 3,
    ACTIONS(3025), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(3023), 4,
      anon_sym_LBRACE,
      anon_sym_GT,
      sym_jsx_identifier,
      anon_sym_SLASH_GT,
  [51701] = 6,
    ACTIONS(3137), 1,
      sym_identifier,
    ACTIONS(3139), 1,
      anon_sym_GT,
    ACTIONS(3141), 1,
      sym_jsx_identifier,
    STATE(1512), 1,
      sym_nested_identifier,
    STATE(1861), 1,
      sym_jsx_namespace_name,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [51721] = 2,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1753), 5,
      sym__automatic_semicolon,
      anon_sym_with,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_EQ,
  [51733] = 3,
    ACTIONS(3037), 1,
      anon_sym_LT,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(3035), 4,
      sym_jsx_text,
      anon_sym_LBRACE,
      sym_html_character_reference,
      anon_sym_LT_SLASH,
  [51747] = 6,
    ACTIONS(3011), 1,
      anon_sym_GT,
    ACTIONS(3143), 1,
      sym_identifier,
    ACTIONS(3145), 1,
      sym_jsx_identifier,
    STATE(1162), 1,
      sym_nested_identifier,
    STATE(1174), 1,
      sym_jsx_namespace_name,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [51767] = 3,
    ACTIONS(3074), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(3072), 4,
      anon_sym_LBRACE,
      anon_sym_GT,
      sym_jsx_identifier,
      anon_sym_SLASH_GT,
  [51781] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_html_comment,
    ACTIONS(3147), 1,
      sym_html_character_reference,
    ACTIONS(3150), 1,
      anon_sym_DQUOTE,
    ACTIONS(3152), 1,
      sym_unescaped_double_jsx_string_fragment,
    STATE(1320), 1,
      aux_sym__jsx_string_repeat1,
  [51800] = 4,
    ACTIONS(3155), 1,
      anon_sym_COMMA,
    STATE(1335), 1,
      aux_sym_variable_declaration_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(3157), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [51815] = 5,
    ACTIONS(93), 1,
      anon_sym_AT,
    ACTIONS(3120), 1,
      anon_sym_class,
    STATE(1041), 1,
      aux_sym_export_statement_repeat1,
    STATE(1097), 1,
      sym_decorator,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [51832] = 5,
    ACTIONS(2939), 1,
      anon_sym_LPAREN,
    ACTIONS(3159), 1,
      sym_identifier,
    ACTIONS(3161), 1,
      anon_sym_STAR,
    STATE(1531), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [51849] = 3,
    ACTIONS(3163), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1543), 3,
      anon_sym_COMMA,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
  [51862] = 4,
    ACTIONS(3155), 1,
      anon_sym_COMMA,
    STATE(1335), 1,
      aux_sym_variable_declaration_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(3165), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [51877] = 5,
    ACTIONS(1818), 1,
      anon_sym_LPAREN,
    ACTIONS(1822), 1,
      anon_sym_DOT,
    ACTIONS(3167), 1,
      sym_optional_chain,
    STATE(797), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [51894] = 5,
    ACTIONS(93), 1,
      anon_sym_AT,
    ACTIONS(3169), 1,
      anon_sym_class,
    STATE(1041), 1,
      aux_sym_export_statement_repeat1,
    STATE(1097), 1,
      sym_decorator,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [51911] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_html_comment,
    ACTIONS(3171), 1,
      anon_sym_DQUOTE,
    STATE(1395), 1,
      aux_sym_string_repeat1,
    ACTIONS(3173), 2,
      sym_unescaped_double_string_fragment,
      sym_escape_sequence,
  [51928] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_html_comment,
    ACTIONS(3171), 1,
      anon_sym_SQUOTE,
    STATE(1396), 1,
      aux_sym_string_repeat2,
    ACTIONS(3175), 2,
      sym_unescaped_single_string_fragment,
      sym_escape_sequence,
  [51945] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_html_comment,
    ACTIONS(3177), 1,
      anon_sym_DQUOTE,
    STATE(1350), 1,
      aux_sym_string_repeat1,
    ACTIONS(3179), 2,
      sym_unescaped_double_string_fragment,
      sym_escape_sequence,
  [51962] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_html_comment,
    ACTIONS(3177), 1,
      anon_sym_SQUOTE,
    STATE(1351), 1,
      aux_sym_string_repeat2,
    ACTIONS(3181), 2,
      sym_unescaped_single_string_fragment,
      sym_escape_sequence,
  [51979] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_html_comment,
    ACTIONS(3183), 1,
      anon_sym_DQUOTE,
    STATE(1349), 1,
      aux_sym_string_repeat1,
    ACTIONS(3185), 2,
      sym_unescaped_double_string_fragment,
      sym_escape_sequence,
  [51996] = 5,
    ACTIONS(1567), 1,
      anon_sym_LPAREN,
    ACTIONS(1571), 1,
      anon_sym_DOT,
    ACTIONS(3187), 1,
      sym_optional_chain,
    STATE(637), 1,
      sym_arguments,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [52013] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_html_comment,
    ACTIONS(3183), 1,
      anon_sym_SQUOTE,
    STATE(1352), 1,
      aux_sym_string_repeat2,
    ACTIONS(3189), 2,
      sym_unescaped_single_string_fragment,
      sym_escape_sequence,
  [52030] = 4,
    ACTIONS(3191), 1,
      anon_sym_COMMA,
    STATE(1335), 1,
      aux_sym_variable_declaration_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(3194), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [52045] = 5,
    ACTIONS(2939), 1,
      anon_sym_LPAREN,
    ACTIONS(3196), 1,
      sym_identifier,
    ACTIONS(3198), 1,
      anon_sym_STAR,
    STATE(1684), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [52062] = 4,
    ACTIONS(3200), 1,
      anon_sym_EQ,
    STATE(1716), 1,
      sym__initializer,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2905), 2,
      anon_sym_of,
      anon_sym_in,
  [52077] = 4,
    ACTIONS(3200), 1,
      anon_sym_EQ,
    STATE(1729), 1,
      sym__initializer,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2887), 2,
      anon_sym_of,
      anon_sym_in,
  [52092] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_html_comment,
    ACTIONS(3202), 1,
      anon_sym_DQUOTE,
    STATE(1328), 1,
      aux_sym_string_repeat1,
    ACTIONS(3204), 2,
      sym_unescaped_double_string_fragment,
      sym_escape_sequence,
  [52109] = 2,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1916), 4,
      anon_sym_of,
      anon_sym_RPAREN,
      anon_sym_in,
      anon_sym_EQ,
  [52120] = 4,
    ACTIONS(3155), 1,
      anon_sym_COMMA,
    STATE(1335), 1,
      aux_sym_variable_declaration_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(3206), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [52135] = 2,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2395), 4,
      sym__automatic_semicolon,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_EQ,
  [52146] = 5,
    ACTIONS(93), 1,
      anon_sym_AT,
    ACTIONS(3029), 1,
      anon_sym_class,
    STATE(1041), 1,
      aux_sym_export_statement_repeat1,
    STATE(1097), 1,
      sym_decorator,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [52163] = 4,
    ACTIONS(3208), 1,
      anon_sym_COMMA,
    STATE(1344), 1,
      aux_sym_sequence_expression_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1836), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [52178] = 5,
    ACTIONS(3017), 1,
      anon_sym_LBRACE,
    ACTIONS(3211), 1,
      anon_sym_extends,
    STATE(755), 1,
      sym_class_body,
    STATE(1632), 1,
      sym_class_heritage,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [52195] = 2,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1912), 4,
      anon_sym_of,
      anon_sym_RPAREN,
      anon_sym_in,
      anon_sym_EQ,
  [52206] = 5,
    ACTIONS(873), 1,
      anon_sym_COMMA,
    ACTIONS(3213), 1,
      anon_sym_EQ,
    ACTIONS(3215), 1,
      anon_sym_RBRACK,
    STATE(1430), 1,
      aux_sym_array_pattern_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [52223] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_html_comment,
    ACTIONS(3217), 1,
      sym_html_character_reference,
    ACTIONS(3220), 1,
      anon_sym_SQUOTE,
    ACTIONS(3222), 1,
      sym_unescaped_single_jsx_string_fragment,
    STATE(1348), 1,
      aux_sym__jsx_string_repeat2,
  [52242] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_html_comment,
    ACTIONS(3225), 1,
      anon_sym_DQUOTE,
    STATE(1395), 1,
      aux_sym_string_repeat1,
    ACTIONS(3173), 2,
      sym_unescaped_double_string_fragment,
      sym_escape_sequence,
  [52259] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_html_comment,
    ACTIONS(3227), 1,
      anon_sym_DQUOTE,
    STATE(1395), 1,
      aux_sym_string_repeat1,
    ACTIONS(3173), 2,
      sym_unescaped_double_string_fragment,
      sym_escape_sequence,
  [52276] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_html_comment,
    ACTIONS(3227), 1,
      anon_sym_SQUOTE,
    STATE(1396), 1,
      aux_sym_string_repeat2,
    ACTIONS(3175), 2,
      sym_unescaped_single_string_fragment,
      sym_escape_sequence,
  [52293] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_html_comment,
    ACTIONS(3225), 1,
      anon_sym_SQUOTE,
    STATE(1396), 1,
      aux_sym_string_repeat2,
    ACTIONS(3175), 2,
      sym_unescaped_single_string_fragment,
      sym_escape_sequence,
  [52310] = 5,
    ACTIONS(3017), 1,
      anon_sym_LBRACE,
    ACTIONS(3211), 1,
      anon_sym_extends,
    STATE(798), 1,
      sym_class_body,
    STATE(1530), 1,
      sym_class_heritage,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [52327] = 4,
    ACTIONS(3229), 1,
      anon_sym_COMMA,
    STATE(1354), 1,
      aux_sym_array_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2017), 2,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
  [52342] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_html_comment,
    ACTIONS(3232), 1,
      sym_html_character_reference,
    ACTIONS(3234), 1,
      anon_sym_SQUOTE,
    ACTIONS(3236), 1,
      sym_unescaped_single_jsx_string_fragment,
    STATE(1348), 1,
      aux_sym__jsx_string_repeat2,
  [52361] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_html_comment,
    ACTIONS(3238), 1,
      sym_html_character_reference,
    ACTIONS(3240), 1,
      anon_sym_DQUOTE,
    ACTIONS(3242), 1,
      sym_unescaped_double_jsx_string_fragment,
    STATE(1384), 1,
      aux_sym__jsx_string_repeat1,
  [52380] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_html_comment,
    ACTIONS(3244), 1,
      anon_sym_DQUOTE,
    STATE(1361), 1,
      aux_sym_string_repeat1,
    ACTIONS(3246), 2,
      sym_unescaped_double_string_fragment,
      sym_escape_sequence,
  [52397] = 4,
    ACTIONS(3248), 1,
      anon_sym_from,
    STATE(1666), 1,
      sym__from_clause,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(3250), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [52412] = 4,
    ACTIONS(3252), 1,
      anon_sym_with,
    STATE(1731), 1,
      sym_import_attribute,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(3254), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [52427] = 5,
    ACTIONS(3070), 1,
      anon_sym_LBRACE,
    ACTIONS(3211), 1,
      anon_sym_extends,
    STATE(619), 1,
      sym_class_body,
    STATE(1566), 1,
      sym_class_heritage,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [52444] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_html_comment,
    ACTIONS(3256), 1,
      anon_sym_DQUOTE,
    STATE(1395), 1,
      aux_sym_string_repeat1,
    ACTIONS(3173), 2,
      sym_unescaped_double_string_fragment,
      sym_escape_sequence,
  [52461] = 5,
    ACTIONS(3017), 1,
      anon_sym_LBRACE,
    ACTIONS(3211), 1,
      anon_sym_extends,
    STATE(100), 1,
      sym_class_body,
    STATE(1537), 1,
      sym_class_heritage,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [52478] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_html_comment,
    ACTIONS(3256), 1,
      anon_sym_SQUOTE,
    STATE(1396), 1,
      aux_sym_string_repeat2,
    ACTIONS(3175), 2,
      sym_unescaped_single_string_fragment,
      sym_escape_sequence,
  [52495] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_html_comment,
    ACTIONS(3202), 1,
      anon_sym_SQUOTE,
    STATE(1329), 1,
      aux_sym_string_repeat2,
    ACTIONS(3258), 2,
      sym_unescaped_single_string_fragment,
      sym_escape_sequence,
  [52512] = 4,
    ACTIONS(3155), 1,
      anon_sym_COMMA,
    STATE(1335), 1,
      aux_sym_variable_declaration_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(3260), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [52527] = 2,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1914), 4,
      anon_sym_of,
      anon_sym_RPAREN,
      anon_sym_in,
      anon_sym_EQ,
  [52538] = 5,
    ACTIONS(3017), 1,
      anon_sym_LBRACE,
    ACTIONS(3211), 1,
      anon_sym_extends,
    STATE(97), 1,
      sym_class_body,
    STATE(1645), 1,
      sym_class_heritage,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [52555] = 5,
    ACTIONS(3017), 1,
      anon_sym_LBRACE,
    ACTIONS(3211), 1,
      anon_sym_extends,
    STATE(799), 1,
      sym_class_body,
    STATE(1607), 1,
      sym_class_heritage,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [52572] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_html_comment,
    ACTIONS(3244), 1,
      anon_sym_SQUOTE,
    STATE(1363), 1,
      aux_sym_string_repeat2,
    ACTIONS(3262), 2,
      sym_unescaped_single_string_fragment,
      sym_escape_sequence,
  [52589] = 5,
    ACTIONS(3213), 1,
      anon_sym_EQ,
    ACTIONS(3264), 1,
      anon_sym_COMMA,
    ACTIONS(3266), 1,
      anon_sym_RPAREN,
    STATE(1486), 1,
      aux_sym_formal_parameters_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [52606] = 4,
    ACTIONS(2739), 1,
      anon_sym_STAR,
    ACTIONS(2741), 1,
      anon_sym_LBRACE,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    STATE(1833), 2,
      sym_namespace_import,
      sym_named_imports,
  [52621] = 5,
    ACTIONS(3252), 1,
      anon_sym_with,
    ACTIONS(3268), 1,
      anon_sym_SEMI,
    ACTIONS(3270), 1,
      sym__automatic_semicolon,
    STATE(1674), 1,
      sym_import_attribute,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [52638] = 5,
    ACTIONS(3070), 1,
      anon_sym_LBRACE,
    ACTIONS(3211), 1,
      anon_sym_extends,
    STATE(650), 1,
      sym_class_body,
    STATE(1624), 1,
      sym_class_heritage,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [52655] = 5,
    ACTIONS(2939), 1,
      anon_sym_LPAREN,
    ACTIONS(3272), 1,
      sym_identifier,
    ACTIONS(3274), 1,
      anon_sym_STAR,
    STATE(1684), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [52672] = 2,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(3276), 4,
      sym__template_chars,
      sym_escape_sequence,
      anon_sym_BQUOTE,
      anon_sym_DOLLAR_LBRACE,
  [52683] = 5,
    ACTIONS(2939), 1,
      anon_sym_LPAREN,
    ACTIONS(3278), 1,
      sym_identifier,
    ACTIONS(3280), 1,
      anon_sym_STAR,
    STATE(1684), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [52700] = 5,
    ACTIONS(3211), 1,
      anon_sym_extends,
    ACTIONS(3282), 1,
      anon_sym_LBRACE,
    STATE(405), 1,
      sym_class_body,
    STATE(1676), 1,
      sym_class_heritage,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [52717] = 5,
    ACTIONS(2939), 1,
      anon_sym_LPAREN,
    ACTIONS(3284), 1,
      sym_identifier,
    ACTIONS(3286), 1,
      anon_sym_STAR,
    STATE(1665), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [52734] = 4,
    ACTIONS(1942), 1,
      anon_sym_COMMA,
    STATE(1344), 1,
      aux_sym_sequence_expression_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2927), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [52749] = 4,
    ACTIONS(3155), 1,
      anon_sym_COMMA,
    STATE(1325), 1,
      aux_sym_variable_declaration_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(3288), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [52764] = 5,
    ACTIONS(93), 1,
      anon_sym_AT,
    ACTIONS(3290), 1,
      anon_sym_class,
    STATE(1041), 1,
      aux_sym_export_statement_repeat1,
    STATE(1097), 1,
      sym_decorator,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [52781] = 5,
    ACTIONS(873), 1,
      anon_sym_COMMA,
    ACTIONS(3213), 1,
      anon_sym_EQ,
    ACTIONS(3292), 1,
      anon_sym_RBRACK,
    STATE(1451), 1,
      aux_sym_array_pattern_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [52798] = 5,
    ACTIONS(3211), 1,
      anon_sym_extends,
    ACTIONS(3282), 1,
      anon_sym_LBRACE,
    STATE(413), 1,
      sym_class_body,
    STATE(1547), 1,
      sym_class_heritage,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [52815] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_html_comment,
    ACTIONS(3234), 1,
      anon_sym_DQUOTE,
    ACTIONS(3294), 1,
      sym_html_character_reference,
    ACTIONS(3296), 1,
      sym_unescaped_double_jsx_string_fragment,
    STATE(1320), 1,
      aux_sym__jsx_string_repeat1,
  [52834] = 4,
    ACTIONS(3155), 1,
      anon_sym_COMMA,
    STATE(1365), 1,
      aux_sym_variable_declaration_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(3298), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [52849] = 5,
    ACTIONS(2939), 1,
      anon_sym_LPAREN,
    ACTIONS(3300), 1,
      sym_identifier,
    ACTIONS(3302), 1,
      anon_sym_STAR,
    STATE(1665), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [52866] = 5,
    ACTIONS(2939), 1,
      anon_sym_LPAREN,
    ACTIONS(3304), 1,
      sym_identifier,
    ACTIONS(3306), 1,
      anon_sym_STAR,
    STATE(1746), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [52883] = 5,
    ACTIONS(93), 1,
      anon_sym_AT,
    ACTIONS(3308), 1,
      anon_sym_class,
    STATE(1041), 1,
      aux_sym_export_statement_repeat1,
    STATE(1097), 1,
      sym_decorator,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [52900] = 2,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1937), 4,
      sym__automatic_semicolon,
      anon_sym_LPAREN,
      anon_sym_SEMI,
      anon_sym_EQ,
  [52911] = 4,
    ACTIONS(3155), 1,
      anon_sym_COMMA,
    STATE(1341), 1,
      aux_sym_variable_declaration_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(3310), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [52926] = 4,
    ACTIONS(3155), 1,
      anon_sym_COMMA,
    STATE(1321), 1,
      aux_sym_variable_declaration_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(3312), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [52941] = 5,
    ACTIONS(3017), 1,
      anon_sym_LBRACE,
    ACTIONS(3211), 1,
      anon_sym_extends,
    STATE(766), 1,
      sym_class_body,
    STATE(1696), 1,
      sym_class_heritage,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [52958] = 5,
    ACTIONS(2939), 1,
      anon_sym_LPAREN,
    ACTIONS(3314), 1,
      sym_identifier,
    ACTIONS(3316), 1,
      anon_sym_STAR,
    STATE(1665), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [52975] = 4,
    ACTIONS(2395), 1,
      anon_sym_COLON,
    ACTIONS(2397), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(3318), 2,
      anon_sym_COMMA,
      anon_sym_RBRACE,
  [52990] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_html_comment,
    ACTIONS(3320), 1,
      anon_sym_DQUOTE,
    STATE(1395), 1,
      aux_sym_string_repeat1,
    ACTIONS(3322), 2,
      sym_unescaped_double_string_fragment,
      sym_escape_sequence,
  [53007] = 5,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_html_comment,
    ACTIONS(3325), 1,
      anon_sym_SQUOTE,
    STATE(1396), 1,
      aux_sym_string_repeat2,
    ACTIONS(3327), 2,
      sym_unescaped_single_string_fragment,
      sym_escape_sequence,
  [53024] = 3,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2395), 2,
      anon_sym_LPAREN,
      anon_sym_COLON,
    ACTIONS(2455), 2,
      anon_sym_COMMA,
      anon_sym_RBRACE,
  [53037] = 6,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_html_comment,
    ACTIONS(3240), 1,
      anon_sym_SQUOTE,
    ACTIONS(3330), 1,
      sym_html_character_reference,
    ACTIONS(3332), 1,
      sym_unescaped_single_jsx_string_fragment,
    STATE(1355), 1,
      aux_sym__jsx_string_repeat2,
  [53056] = 4,
    ACTIONS(2939), 1,
      anon_sym_LPAREN,
    ACTIONS(3334), 1,
      sym_identifier,
    STATE(1573), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [53070] = 4,
    ACTIONS(2939), 1,
      anon_sym_LPAREN,
    ACTIONS(3336), 1,
      sym_identifier,
    STATE(1532), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [53084] = 4,
    ACTIONS(2939), 1,
      anon_sym_LPAREN,
    ACTIONS(3338), 1,
      sym_identifier,
    STATE(1501), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [53098] = 4,
    ACTIONS(2273), 1,
      anon_sym_COMMA,
    ACTIONS(3340), 1,
      anon_sym_RBRACE,
    STATE(1467), 1,
      aux_sym_object_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [53112] = 4,
    ACTIONS(1162), 1,
      anon_sym_while,
    ACTIONS(3342), 1,
      anon_sym_else,
    STATE(428), 1,
      sym_else_clause,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [53126] = 3,
    ACTIONS(2745), 1,
      anon_sym_DOT,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2743), 2,
      anon_sym_LPAREN,
      sym_optional_chain,
  [53138] = 4,
    ACTIONS(910), 1,
      anon_sym_COMMA,
    ACTIONS(2065), 1,
      anon_sym_RPAREN,
    STATE(1436), 1,
      aux_sym_array_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [53152] = 4,
    ACTIONS(2273), 1,
      anon_sym_COMMA,
    ACTIONS(3344), 1,
      anon_sym_RBRACE,
    STATE(1441), 1,
      aux_sym_object_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [53166] = 4,
    ACTIONS(2939), 1,
      anon_sym_LPAREN,
    ACTIONS(3346), 1,
      anon_sym_COLON,
    STATE(1644), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [53180] = 2,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(3054), 3,
      sym__automatic_semicolon,
      anon_sym_COMMA,
      anon_sym_SEMI,
  [53190] = 4,
    ACTIONS(873), 1,
      anon_sym_COMMA,
    ACTIONS(3292), 1,
      anon_sym_RBRACK,
    STATE(1433), 1,
      aux_sym_array_pattern_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [53204] = 3,
    ACTIONS(3213), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(3348), 2,
      anon_sym_COMMA,
      anon_sym_RBRACK,
  [53216] = 4,
    ACTIONS(3090), 1,
      sym_identifier,
    ACTIONS(3092), 1,
      anon_sym_LBRACK,
    ACTIONS(3094), 1,
      sym_private_property_identifier,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [53230] = 4,
    ACTIONS(3350), 1,
      anon_sym_LBRACE,
    ACTIONS(3352), 1,
      anon_sym_LPAREN,
    STATE(401), 1,
      sym_statement_block,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [53244] = 4,
    ACTIONS(873), 1,
      anon_sym_COMMA,
    ACTIONS(3292), 1,
      anon_sym_RBRACK,
    STATE(1451), 1,
      aux_sym_array_pattern_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [53258] = 4,
    ACTIONS(3354), 1,
      anon_sym_COMMA,
    ACTIONS(3356), 1,
      anon_sym_RBRACE,
    STATE(1463), 1,
      aux_sym_export_clause_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [53272] = 2,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(3194), 3,
      sym__automatic_semicolon,
      anon_sym_COMMA,
      anon_sym_SEMI,
  [53282] = 4,
    ACTIONS(2357), 1,
      anon_sym_COMMA,
    ACTIONS(3358), 1,
      anon_sym_RBRACE,
    STATE(1421), 1,
      aux_sym_object_pattern_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [53296] = 4,
    ACTIONS(3360), 1,
      anon_sym_COMMA,
    ACTIONS(3363), 1,
      anon_sym_RBRACE,
    STATE(1417), 1,
      aux_sym_object_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [53310] = 4,
    ACTIONS(2273), 1,
      anon_sym_COMMA,
    ACTIONS(3365), 1,
      anon_sym_RBRACE,
    STATE(1417), 1,
      aux_sym_object_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [53324] = 4,
    ACTIONS(2357), 1,
      anon_sym_COMMA,
    ACTIONS(3367), 1,
      anon_sym_RBRACE,
    STATE(1421), 1,
      aux_sym_object_pattern_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [53338] = 4,
    ACTIONS(2357), 1,
      anon_sym_COMMA,
    ACTIONS(3369), 1,
      anon_sym_RBRACE,
    STATE(1421), 1,
      aux_sym_object_pattern_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [53352] = 4,
    ACTIONS(3371), 1,
      anon_sym_COMMA,
    ACTIONS(3374), 1,
      anon_sym_RBRACE,
    STATE(1421), 1,
      aux_sym_object_pattern_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [53366] = 4,
    ACTIONS(2273), 1,
      anon_sym_COMMA,
    ACTIONS(3376), 1,
      anon_sym_RBRACE,
    STATE(1417), 1,
      aux_sym_object_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [53380] = 3,
    ACTIONS(3378), 1,
      anon_sym_as,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(3380), 2,
      anon_sym_COMMA,
      anon_sym_RBRACE,
  [53392] = 4,
    ACTIONS(2751), 1,
      anon_sym_COLON,
    ACTIONS(2757), 1,
      anon_sym_DOT,
    ACTIONS(3382), 1,
      anon_sym_GT,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [53406] = 2,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(3062), 3,
      sym__automatic_semicolon,
      anon_sym_COMMA,
      anon_sym_SEMI,
  [53416] = 4,
    ACTIONS(3264), 1,
      anon_sym_COMMA,
    ACTIONS(3266), 1,
      anon_sym_RPAREN,
    STATE(1486), 1,
      aux_sym_formal_parameters_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [53430] = 4,
    ACTIONS(2297), 1,
      anon_sym_DQUOTE,
    ACTIONS(2299), 1,
      anon_sym_SQUOTE,
    STATE(1439), 1,
      sym_string,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [53444] = 4,
    ACTIONS(2751), 1,
      anon_sym_COLON,
    ACTIONS(2757), 1,
      anon_sym_DOT,
    ACTIONS(3384), 1,
      anon_sym_GT,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [53458] = 4,
    ACTIONS(910), 1,
      anon_sym_COMMA,
    ACTIONS(3386), 1,
      anon_sym_RBRACK,
    STATE(1354), 1,
      aux_sym_array_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [53472] = 4,
    ACTIONS(873), 1,
      anon_sym_COMMA,
    ACTIONS(3388), 1,
      anon_sym_RBRACK,
    STATE(1433), 1,
      aux_sym_array_pattern_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [53486] = 3,
    ACTIONS(1902), 1,
      anon_sym_in,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1932), 2,
      anon_sym_LPAREN,
      anon_sym_COLON,
  [53498] = 4,
    ACTIONS(910), 1,
      anon_sym_COMMA,
    ACTIONS(2021), 1,
      anon_sym_RPAREN,
    STATE(1354), 1,
      aux_sym_array_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [53512] = 4,
    ACTIONS(3348), 1,
      anon_sym_RBRACK,
    ACTIONS(3390), 1,
      anon_sym_COMMA,
    STATE(1433), 1,
      aux_sym_array_pattern_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [53526] = 3,
    ACTIONS(3213), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(3393), 2,
      anon_sym_COMMA,
      anon_sym_RBRACE,
  [53538] = 4,
    ACTIONS(2879), 1,
      anon_sym_RBRACE,
    ACTIONS(3395), 1,
      anon_sym_COMMA,
    STATE(1477), 1,
      aux_sym_named_imports_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [53552] = 4,
    ACTIONS(910), 1,
      anon_sym_COMMA,
    ACTIONS(3397), 1,
      anon_sym_RPAREN,
    STATE(1354), 1,
      aux_sym_array_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [53566] = 4,
    ACTIONS(910), 1,
      anon_sym_COMMA,
    ACTIONS(2041), 1,
      anon_sym_RPAREN,
    STATE(1455), 1,
      aux_sym_array_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [53580] = 4,
    ACTIONS(910), 1,
      anon_sym_COMMA,
    ACTIONS(2041), 1,
      anon_sym_RPAREN,
    STATE(1354), 1,
      aux_sym_array_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [53594] = 2,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(3399), 3,
      sym__automatic_semicolon,
      anon_sym_with,
      anon_sym_SEMI,
  [53604] = 4,
    ACTIONS(3401), 1,
      anon_sym_await,
    ACTIONS(3403), 1,
      anon_sym_LPAREN,
    STATE(55), 1,
      sym__for_header,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [53618] = 4,
    ACTIONS(2273), 1,
      anon_sym_COMMA,
    ACTIONS(3405), 1,
      anon_sym_RBRACE,
    STATE(1417), 1,
      aux_sym_object_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [53632] = 3,
    ACTIONS(3407), 1,
      anon_sym_as,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(3409), 2,
      anon_sym_COMMA,
      anon_sym_RBRACE,
  [53644] = 4,
    ACTIONS(3411), 1,
      anon_sym_COMMA,
    ACTIONS(3413), 1,
      anon_sym_RBRACE,
    STATE(1435), 1,
      aux_sym_named_imports_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [53658] = 4,
    ACTIONS(2751), 1,
      anon_sym_COLON,
    ACTIONS(2757), 1,
      anon_sym_DOT,
    ACTIONS(3415), 1,
      anon_sym_GT,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [53672] = 4,
    ACTIONS(910), 1,
      anon_sym_COMMA,
    ACTIONS(2065), 1,
      anon_sym_RPAREN,
    STATE(1354), 1,
      aux_sym_array_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [53686] = 3,
    ACTIONS(3417), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(3419), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [53698] = 2,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(3421), 3,
      sym__automatic_semicolon,
      anon_sym_from,
      anon_sym_SEMI,
  [53708] = 4,
    ACTIONS(2939), 1,
      anon_sym_LPAREN,
    ACTIONS(3423), 1,
      anon_sym_COLON,
    STATE(1644), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [53722] = 4,
    ACTIONS(2273), 1,
      anon_sym_COMMA,
    ACTIONS(3340), 1,
      anon_sym_RBRACE,
    STATE(1417), 1,
      aux_sym_object_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [53736] = 4,
    ACTIONS(910), 1,
      anon_sym_COMMA,
    ACTIONS(2009), 1,
      anon_sym_RBRACK,
    STATE(1354), 1,
      aux_sym_array_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [53750] = 4,
    ACTIONS(873), 1,
      anon_sym_COMMA,
    ACTIONS(3425), 1,
      anon_sym_RBRACK,
    STATE(1433), 1,
      aux_sym_array_pattern_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [53764] = 4,
    ACTIONS(2357), 1,
      anon_sym_COMMA,
    ACTIONS(3427), 1,
      anon_sym_RBRACE,
    STATE(1421), 1,
      aux_sym_object_pattern_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [53778] = 2,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(3429), 3,
      sym__automatic_semicolon,
      anon_sym_from,
      anon_sym_SEMI,
  [53788] = 4,
    ACTIONS(2357), 1,
      anon_sym_COMMA,
    ACTIONS(3431), 1,
      anon_sym_RBRACE,
    STATE(1416), 1,
      aux_sym_object_pattern_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [53802] = 4,
    ACTIONS(910), 1,
      anon_sym_COMMA,
    ACTIONS(3433), 1,
      anon_sym_RPAREN,
    STATE(1354), 1,
      aux_sym_array_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [53816] = 3,
    ACTIONS(3435), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(3437), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [53828] = 4,
    ACTIONS(2357), 1,
      anon_sym_COMMA,
    ACTIONS(3439), 1,
      anon_sym_RBRACE,
    STATE(1420), 1,
      aux_sym_object_pattern_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [53842] = 4,
    ACTIONS(2273), 1,
      anon_sym_COMMA,
    ACTIONS(3441), 1,
      anon_sym_RBRACE,
    STATE(1422), 1,
      aux_sym_object_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [53856] = 4,
    ACTIONS(3443), 1,
      anon_sym_COMMA,
    ACTIONS(3446), 1,
      anon_sym_RBRACE,
    STATE(1459), 1,
      aux_sym_export_clause_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [53870] = 4,
    ACTIONS(910), 1,
      anon_sym_COMMA,
    ACTIONS(2009), 1,
      anon_sym_RBRACK,
    STATE(1483), 1,
      aux_sym_array_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [53884] = 4,
    ACTIONS(2273), 1,
      anon_sym_COMMA,
    ACTIONS(3441), 1,
      anon_sym_RBRACE,
    STATE(1417), 1,
      aux_sym_object_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [53898] = 4,
    ACTIONS(2939), 1,
      anon_sym_LPAREN,
    ACTIONS(3448), 1,
      sym_identifier,
    STATE(1671), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [53912] = 4,
    ACTIONS(2875), 1,
      anon_sym_RBRACE,
    ACTIONS(3450), 1,
      anon_sym_COMMA,
    STATE(1459), 1,
      aux_sym_export_clause_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [53926] = 4,
    ACTIONS(2273), 1,
      anon_sym_COMMA,
    ACTIONS(3452), 1,
      anon_sym_RBRACE,
    STATE(1417), 1,
      aux_sym_object_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [53940] = 4,
    ACTIONS(910), 1,
      anon_sym_COMMA,
    ACTIONS(3454), 1,
      anon_sym_RPAREN,
    STATE(1354), 1,
      aux_sym_array_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [53954] = 4,
    ACTIONS(2357), 1,
      anon_sym_COMMA,
    ACTIONS(3439), 1,
      anon_sym_RBRACE,
    STATE(1421), 1,
      aux_sym_object_pattern_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [53968] = 4,
    ACTIONS(2273), 1,
      anon_sym_COMMA,
    ACTIONS(3456), 1,
      anon_sym_RBRACE,
    STATE(1417), 1,
      aux_sym_object_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [53982] = 3,
    ACTIONS(953), 1,
      sym__automatic_semicolon,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2905), 2,
      anon_sym_of,
      anon_sym_in,
  [53994] = 3,
    ACTIONS(3458), 1,
      sym__automatic_semicolon,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2887), 2,
      anon_sym_of,
      anon_sym_in,
  [54006] = 4,
    ACTIONS(3460), 1,
      sym_identifier,
    STATE(1034), 1,
      sym_decorator_member_expression,
    STATE(1094), 1,
      sym_decorator_call_expression,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [54020] = 4,
    ACTIONS(2939), 1,
      anon_sym_LPAREN,
    ACTIONS(3462), 1,
      sym_identifier,
    STATE(1501), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [54034] = 3,
    ACTIONS(2999), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1550), 2,
      anon_sym_of,
      anon_sym_in,
  [54046] = 4,
    ACTIONS(2939), 1,
      anon_sym_LPAREN,
    ACTIONS(3464), 1,
      sym_identifier,
    STATE(1501), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [54060] = 3,
    ACTIONS(3213), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(3466), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [54072] = 4,
    ACTIONS(3466), 1,
      anon_sym_RPAREN,
    ACTIONS(3468), 1,
      anon_sym_COMMA,
    STATE(1475), 1,
      aux_sym_formal_parameters_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [54086] = 4,
    ACTIONS(2357), 1,
      anon_sym_COMMA,
    ACTIONS(3431), 1,
      anon_sym_RBRACE,
    STATE(1421), 1,
      aux_sym_object_pattern_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [54100] = 4,
    ACTIONS(3471), 1,
      anon_sym_COMMA,
    ACTIONS(3474), 1,
      anon_sym_RBRACE,
    STATE(1477), 1,
      aux_sym_named_imports_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [54114] = 3,
    ACTIONS(3476), 1,
      anon_sym_DOT,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2743), 2,
      anon_sym_LPAREN,
      sym_optional_chain,
  [54126] = 4,
    ACTIONS(2939), 1,
      anon_sym_LPAREN,
    ACTIONS(3478), 1,
      sym_identifier,
    STATE(1671), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [54140] = 4,
    ACTIONS(910), 1,
      anon_sym_COMMA,
    ACTIONS(2021), 1,
      anon_sym_RPAREN,
    STATE(1465), 1,
      aux_sym_array_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [54154] = 3,
    ACTIONS(3163), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1550), 2,
      anon_sym_of,
      anon_sym_in,
  [54166] = 2,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(3480), 3,
      sym__automatic_semicolon,
      anon_sym_from,
      anon_sym_SEMI,
  [54176] = 4,
    ACTIONS(910), 1,
      anon_sym_COMMA,
    ACTIONS(3482), 1,
      anon_sym_RBRACK,
    STATE(1354), 1,
      aux_sym_array_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [54190] = 4,
    ACTIONS(873), 1,
      anon_sym_COMMA,
    ACTIONS(3215), 1,
      anon_sym_RBRACK,
    STATE(1430), 1,
      aux_sym_array_pattern_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [54204] = 4,
    ACTIONS(910), 1,
      anon_sym_COMMA,
    ACTIONS(2039), 1,
      anon_sym_RBRACK,
    STATE(1429), 1,
      aux_sym_array_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [54218] = 4,
    ACTIONS(918), 1,
      anon_sym_RPAREN,
    ACTIONS(3484), 1,
      anon_sym_COMMA,
    STATE(1475), 1,
      aux_sym_formal_parameters_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [54232] = 4,
    ACTIONS(2273), 1,
      anon_sym_COMMA,
    ACTIONS(3486), 1,
      anon_sym_RBRACE,
    STATE(1417), 1,
      aux_sym_object_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [54246] = 2,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(2017), 3,
      anon_sym_COMMA,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
  [54256] = 4,
    ACTIONS(3248), 1,
      anon_sym_from,
    ACTIONS(3488), 1,
      anon_sym_as,
    STATE(1686), 1,
      sym__from_clause,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [54270] = 4,
    ACTIONS(910), 1,
      anon_sym_COMMA,
    ACTIONS(2039), 1,
      anon_sym_RBRACK,
    STATE(1354), 1,
      aux_sym_array_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [54284] = 4,
    ACTIONS(873), 1,
      anon_sym_COMMA,
    ACTIONS(3215), 1,
      anon_sym_RBRACK,
    STATE(1433), 1,
      aux_sym_array_pattern_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [54298] = 4,
    ACTIONS(2939), 1,
      anon_sym_LPAREN,
    ACTIONS(3490), 1,
      sym_identifier,
    STATE(1671), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [54312] = 3,
    ACTIONS(3494), 1,
      sym__automatic_semicolon,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(3492), 2,
      anon_sym_of,
      anon_sym_in,
  [54324] = 4,
    ACTIONS(3125), 1,
      sym_identifier,
    ACTIONS(3127), 1,
      anon_sym_LBRACK,
    ACTIONS(3129), 1,
      sym_private_property_identifier,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [54338] = 3,
    ACTIONS(3498), 1,
      sym__automatic_semicolon,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(3496), 2,
      anon_sym_of,
      anon_sym_in,
  [54350] = 4,
    ACTIONS(2751), 1,
      anon_sym_COLON,
    ACTIONS(2757), 1,
      anon_sym_DOT,
    ACTIONS(3500), 1,
      anon_sym_GT,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [54364] = 2,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(3502), 3,
      sym__automatic_semicolon,
      anon_sym_from,
      anon_sym_SEMI,
  [54374] = 4,
    ACTIONS(3504), 1,
      anon_sym_await,
    ACTIONS(3506), 1,
      anon_sym_LPAREN,
    STATE(69), 1,
      sym__for_header,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [54388] = 4,
    ACTIONS(2273), 1,
      anon_sym_COMMA,
    ACTIONS(3344), 1,
      anon_sym_RBRACE,
    STATE(1417), 1,
      aux_sym_object_repeat1,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [54402] = 2,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(3508), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [54411] = 3,
    ACTIONS(3510), 1,
      anon_sym_LBRACE,
    STATE(789), 1,
      sym_statement_block,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [54422] = 3,
    ACTIONS(3510), 1,
      anon_sym_LBRACE,
    STATE(94), 1,
      sym_statement_block,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [54433] = 3,
    ACTIONS(3512), 1,
      sym_identifier,
    ACTIONS(3514), 1,
      sym_private_property_identifier,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [54444] = 3,
    ACTIONS(3090), 1,
      sym_identifier,
    ACTIONS(3094), 1,
      sym_private_property_identifier,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [54455] = 2,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1944), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [54464] = 3,
    ACTIONS(3516), 1,
      anon_sym_LPAREN,
    STATE(70), 1,
      sym_parenthesized_expression,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [54475] = 3,
    ACTIONS(3518), 1,
      sym_identifier,
    ACTIONS(3520), 1,
      sym_private_property_identifier,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [54486] = 3,
    ACTIONS(2939), 1,
      anon_sym_LPAREN,
    STATE(1571), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [54497] = 3,
    ACTIONS(3125), 1,
      sym_identifier,
    ACTIONS(3129), 1,
      sym_private_property_identifier,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [54508] = 3,
    ACTIONS(3017), 1,
      anon_sym_LBRACE,
    STATE(783), 1,
      sym_class_body,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [54519] = 3,
    ACTIONS(2751), 1,
      anon_sym_COLON,
    ACTIONS(3382), 1,
      anon_sym_GT,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [54530] = 3,
    ACTIONS(2757), 1,
      anon_sym_DOT,
    ACTIONS(3382), 1,
      anon_sym_GT,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [54541] = 2,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1724), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [54550] = 2,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(3374), 2,
      anon_sym_COMMA,
      anon_sym_RBRACE,
  [54559] = 3,
    ACTIONS(2405), 1,
      anon_sym_LBRACE,
    STATE(1040), 1,
      sym_statement_block,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [54570] = 2,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(3363), 2,
      anon_sym_COMMA,
      anon_sym_RBRACE,
  [54579] = 3,
    ACTIONS(3516), 1,
      anon_sym_LPAREN,
    STATE(54), 1,
      sym_parenthesized_expression,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [54590] = 3,
    ACTIONS(3510), 1,
      anon_sym_LBRACE,
    STATE(95), 1,
      sym_statement_block,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [54601] = 3,
    ACTIONS(2939), 1,
      anon_sym_LPAREN,
    STATE(1714), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [54612] = 3,
    ACTIONS(3516), 1,
      anon_sym_LPAREN,
    STATE(40), 1,
      sym_parenthesized_expression,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [54623] = 2,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1688), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [54632] = 2,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1692), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [54641] = 3,
    ACTIONS(3516), 1,
      anon_sym_LPAREN,
    STATE(56), 1,
      sym_parenthesized_expression,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [54652] = 3,
    ACTIONS(3510), 1,
      anon_sym_LBRACE,
    STATE(736), 1,
      sym_statement_block,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [54663] = 3,
    ACTIONS(2395), 1,
      anon_sym_LPAREN,
    ACTIONS(2577), 1,
      anon_sym_EQ_GT,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [54674] = 2,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(3522), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [54683] = 3,
    ACTIONS(2751), 1,
      anon_sym_COLON,
    ACTIONS(3500), 1,
      anon_sym_GT,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [54694] = 3,
    ACTIONS(2395), 1,
      anon_sym_LPAREN,
    ACTIONS(3524), 1,
      anon_sym_EQ_GT,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [54705] = 3,
    ACTIONS(2939), 1,
      anon_sym_LPAREN,
    STATE(1578), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [54716] = 3,
    ACTIONS(3017), 1,
      anon_sym_LBRACE,
    STATE(768), 1,
      sym_class_body,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [54727] = 3,
    ACTIONS(3526), 1,
      anon_sym_LBRACE,
    STATE(627), 1,
      sym_statement_block,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [54738] = 3,
    ACTIONS(3526), 1,
      anon_sym_LBRACE,
    STATE(631), 1,
      sym_statement_block,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [54749] = 3,
    ACTIONS(3510), 1,
      anon_sym_LBRACE,
    STATE(782), 1,
      sym_statement_block,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [54760] = 3,
    ACTIONS(2757), 1,
      anon_sym_DOT,
    ACTIONS(3500), 1,
      anon_sym_GT,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [54771] = 3,
    ACTIONS(3528), 1,
      anon_sym_LPAREN,
    STATE(58), 1,
      sym__for_header,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [54782] = 3,
    ACTIONS(2939), 1,
      anon_sym_LPAREN,
    STATE(1533), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [54793] = 3,
    ACTIONS(3017), 1,
      anon_sym_LBRACE,
    STATE(93), 1,
      sym_class_body,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [54804] = 2,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(3530), 2,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
  [54813] = 3,
    ACTIONS(2939), 1,
      anon_sym_LPAREN,
    STATE(1579), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [54824] = 3,
    ACTIONS(2939), 1,
      anon_sym_LPAREN,
    STATE(1589), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [54835] = 2,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(3446), 2,
      anon_sym_COMMA,
      anon_sym_RBRACE,
  [54844] = 3,
    ACTIONS(3350), 1,
      anon_sym_LBRACE,
    STATE(1045), 1,
      sym_statement_block,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [54855] = 2,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(3532), 2,
      anon_sym_COMMA,
      anon_sym_RBRACE,
  [54864] = 3,
    ACTIONS(2939), 1,
      anon_sym_LPAREN,
    STATE(1700), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [54875] = 3,
    ACTIONS(3534), 1,
      anon_sym_LBRACE,
    STATE(403), 1,
      sym_statement_block,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [54886] = 3,
    ACTIONS(3534), 1,
      anon_sym_LBRACE,
    STATE(415), 1,
      sym_statement_block,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [54897] = 3,
    ACTIONS(3282), 1,
      anon_sym_LBRACE,
    STATE(392), 1,
      sym_class_body,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [54908] = 2,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1937), 2,
      anon_sym_LPAREN,
      anon_sym_COLON,
  [54917] = 3,
    ACTIONS(2751), 1,
      anon_sym_COLON,
    ACTIONS(3415), 1,
      anon_sym_GT,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [54928] = 3,
    ACTIONS(2757), 1,
      anon_sym_DOT,
    ACTIONS(3415), 1,
      anon_sym_GT,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [54939] = 3,
    ACTIONS(3350), 1,
      anon_sym_LBRACE,
    STATE(1064), 1,
      sym_statement_block,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [54950] = 3,
    ACTIONS(2939), 1,
      anon_sym_LPAREN,
    STATE(1704), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [54961] = 3,
    ACTIONS(2939), 1,
      anon_sym_LPAREN,
    STATE(1556), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [54972] = 3,
    ACTIONS(3070), 1,
      anon_sym_LBRACE,
    STATE(643), 1,
      sym_class_body,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [54983] = 3,
    ACTIONS(2939), 1,
      anon_sym_LPAREN,
    STATE(1567), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [54994] = 3,
    ACTIONS(2405), 1,
      anon_sym_LBRACE,
    STATE(1044), 1,
      sym_statement_block,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [55005] = 3,
    ACTIONS(2405), 1,
      anon_sym_LBRACE,
    STATE(1045), 1,
      sym_statement_block,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [55016] = 3,
    ACTIONS(3350), 1,
      anon_sym_LBRACE,
    STATE(1065), 1,
      sym_statement_block,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [55027] = 3,
    ACTIONS(2939), 1,
      anon_sym_LPAREN,
    STATE(1705), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [55038] = 3,
    ACTIONS(2939), 1,
      anon_sym_LPAREN,
    STATE(1574), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [55049] = 3,
    ACTIONS(2939), 1,
      anon_sym_LPAREN,
    STATE(1706), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [55060] = 3,
    ACTIONS(2939), 1,
      anon_sym_LPAREN,
    STATE(1707), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [55071] = 3,
    ACTIONS(2939), 1,
      anon_sym_LPAREN,
    STATE(1586), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [55082] = 3,
    ACTIONS(3350), 1,
      anon_sym_LBRACE,
    STATE(1066), 1,
      sym_statement_block,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [55093] = 3,
    ACTIONS(2939), 1,
      anon_sym_LPAREN,
    STATE(1711), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [55104] = 3,
    ACTIONS(3070), 1,
      anon_sym_LBRACE,
    STATE(644), 1,
      sym_class_body,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [55115] = 3,
    ACTIONS(3526), 1,
      anon_sym_LBRACE,
    STATE(645), 1,
      sym_statement_block,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [55126] = 3,
    ACTIONS(2939), 1,
      anon_sym_LPAREN,
    STATE(1719), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [55137] = 3,
    ACTIONS(2939), 1,
      anon_sym_LPAREN,
    STATE(1720), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [55148] = 3,
    ACTIONS(2939), 1,
      anon_sym_LPAREN,
    STATE(1721), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [55159] = 3,
    ACTIONS(3350), 1,
      anon_sym_LBRACE,
    STATE(1068), 1,
      sym_statement_block,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [55170] = 3,
    ACTIONS(2939), 1,
      anon_sym_LPAREN,
    STATE(1735), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [55181] = 3,
    ACTIONS(3526), 1,
      anon_sym_LBRACE,
    STATE(641), 1,
      sym_statement_block,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [55192] = 3,
    ACTIONS(3510), 1,
      anon_sym_LBRACE,
    STATE(702), 1,
      sym_statement_block,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [55203] = 3,
    ACTIONS(2939), 1,
      anon_sym_LPAREN,
    STATE(1742), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [55214] = 3,
    ACTIONS(2939), 1,
      anon_sym_LPAREN,
    STATE(1743), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [55225] = 3,
    ACTIONS(2939), 1,
      anon_sym_LPAREN,
    STATE(1744), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [55236] = 3,
    ACTIONS(3350), 1,
      anon_sym_LBRACE,
    STATE(1069), 1,
      sym_statement_block,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [55247] = 3,
    ACTIONS(3350), 1,
      anon_sym_LBRACE,
    STATE(1070), 1,
      sym_statement_block,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [55258] = 3,
    ACTIONS(2939), 1,
      anon_sym_LPAREN,
    STATE(1608), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [55269] = 3,
    ACTIONS(2939), 1,
      anon_sym_LPAREN,
    STATE(1615), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [55280] = 3,
    ACTIONS(2939), 1,
      anon_sym_LPAREN,
    STATE(1616), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [55291] = 3,
    ACTIONS(2939), 1,
      anon_sym_LPAREN,
    STATE(1618), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [55302] = 3,
    ACTIONS(2939), 1,
      anon_sym_LPAREN,
    STATE(1619), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [55313] = 3,
    ACTIONS(2939), 1,
      anon_sym_LPAREN,
    STATE(1621), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [55324] = 3,
    ACTIONS(3510), 1,
      anon_sym_LBRACE,
    STATE(752), 1,
      sym_statement_block,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [55335] = 3,
    ACTIONS(2939), 1,
      anon_sym_LPAREN,
    STATE(1622), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [55346] = 3,
    ACTIONS(2939), 1,
      anon_sym_LPAREN,
    STATE(1623), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [55357] = 3,
    ACTIONS(3350), 1,
      anon_sym_LBRACE,
    STATE(1037), 1,
      sym_statement_block,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [55368] = 3,
    ACTIONS(2939), 1,
      anon_sym_LPAREN,
    STATE(1626), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [55379] = 2,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(3536), 2,
      anon_sym_COMMA,
      anon_sym_RBRACE,
  [55388] = 2,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(3474), 2,
      anon_sym_COMMA,
      anon_sym_RBRACE,
  [55397] = 3,
    ACTIONS(2939), 1,
      anon_sym_LPAREN,
    STATE(1630), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [55408] = 3,
    ACTIONS(2939), 1,
      anon_sym_LPAREN,
    STATE(1631), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [55419] = 3,
    ACTIONS(2939), 1,
      anon_sym_LPAREN,
    STATE(1633), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [55430] = 3,
    ACTIONS(2939), 1,
      anon_sym_LPAREN,
    STATE(1634), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [55441] = 3,
    ACTIONS(2939), 1,
      anon_sym_LPAREN,
    STATE(1635), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [55452] = 3,
    ACTIONS(2939), 1,
      anon_sym_LPAREN,
    STATE(1636), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [55463] = 3,
    ACTIONS(2939), 1,
      anon_sym_LPAREN,
    STATE(1637), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [55474] = 3,
    ACTIONS(2939), 1,
      anon_sym_LPAREN,
    STATE(1638), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [55485] = 3,
    ACTIONS(2939), 1,
      anon_sym_LPAREN,
    STATE(1639), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [55496] = 3,
    ACTIONS(2939), 1,
      anon_sym_LPAREN,
    STATE(1640), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [55507] = 3,
    ACTIONS(2939), 1,
      anon_sym_LPAREN,
    STATE(1641), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [55518] = 3,
    ACTIONS(2939), 1,
      anon_sym_LPAREN,
    STATE(1642), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [55529] = 3,
    ACTIONS(2939), 1,
      anon_sym_LPAREN,
    STATE(1643), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [55540] = 3,
    ACTIONS(3538), 1,
      anon_sym_COMMA,
    ACTIONS(3540), 1,
      anon_sym_from,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [55551] = 3,
    ACTIONS(3017), 1,
      anon_sym_LBRACE,
    STATE(724), 1,
      sym_class_body,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [55562] = 3,
    ACTIONS(2405), 1,
      anon_sym_LBRACE,
    STATE(1064), 1,
      sym_statement_block,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [55573] = 3,
    ACTIONS(2939), 1,
      anon_sym_LPAREN,
    STATE(1646), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [55584] = 3,
    ACTIONS(2939), 1,
      anon_sym_LPAREN,
    STATE(1647), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [55595] = 3,
    ACTIONS(2939), 1,
      anon_sym_LPAREN,
    STATE(1650), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [55606] = 3,
    ACTIONS(2939), 1,
      anon_sym_LPAREN,
    STATE(1651), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [55617] = 3,
    ACTIONS(2939), 1,
      anon_sym_LPAREN,
    STATE(1652), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [55628] = 3,
    ACTIONS(2939), 1,
      anon_sym_LPAREN,
    STATE(1653), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [55639] = 3,
    ACTIONS(2405), 1,
      anon_sym_LBRACE,
    STATE(1065), 1,
      sym_statement_block,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [55650] = 3,
    ACTIONS(2405), 1,
      anon_sym_LBRACE,
    STATE(1066), 1,
      sym_statement_block,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [55661] = 3,
    ACTIONS(2405), 1,
      anon_sym_LBRACE,
    STATE(1068), 1,
      sym_statement_block,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [55672] = 3,
    ACTIONS(2405), 1,
      anon_sym_LBRACE,
    STATE(1069), 1,
      sym_statement_block,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [55683] = 3,
    ACTIONS(2405), 1,
      anon_sym_LBRACE,
    STATE(1070), 1,
      sym_statement_block,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [55694] = 2,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(716), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [55703] = 3,
    ACTIONS(2405), 1,
      anon_sym_LBRACE,
    STATE(1037), 1,
      sym_statement_block,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [55714] = 3,
    ACTIONS(3526), 1,
      anon_sym_LBRACE,
    STATE(620), 1,
      sym_statement_block,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [55725] = 3,
    ACTIONS(3526), 1,
      anon_sym_LBRACE,
    STATE(605), 1,
      sym_statement_block,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [55736] = 3,
    ACTIONS(3070), 1,
      anon_sym_LBRACE,
    STATE(612), 1,
      sym_class_body,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [55747] = 3,
    ACTIONS(3248), 1,
      anon_sym_from,
    STATE(1372), 1,
      sym__from_clause,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [55758] = 3,
    ACTIONS(3510), 1,
      anon_sym_LBRACE,
    STATE(774), 1,
      sym_statement_block,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [55769] = 3,
    ACTIONS(2939), 1,
      anon_sym_LPAREN,
    STATE(1524), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [55780] = 3,
    ACTIONS(3510), 1,
      anon_sym_LBRACE,
    STATE(772), 1,
      sym_statement_block,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [55791] = 3,
    ACTIONS(3510), 1,
      anon_sym_LBRACE,
    STATE(773), 1,
      sym_statement_block,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [55802] = 3,
    ACTIONS(2405), 1,
      anon_sym_LBRACE,
    STATE(1046), 1,
      sym_statement_block,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [55813] = 3,
    ACTIONS(2405), 1,
      anon_sym_LBRACE,
    STATE(1048), 1,
      sym_statement_block,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [55824] = 3,
    ACTIONS(3017), 1,
      anon_sym_LBRACE,
    STATE(786), 1,
      sym_class_body,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [55835] = 3,
    ACTIONS(2405), 1,
      anon_sym_LBRACE,
    STATE(1049), 1,
      sym_statement_block,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [55846] = 3,
    ACTIONS(2405), 1,
      anon_sym_LBRACE,
    STATE(1050), 1,
      sym_statement_block,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [55857] = 3,
    ACTIONS(2405), 1,
      anon_sym_LBRACE,
    STATE(1051), 1,
      sym_statement_block,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [55868] = 3,
    ACTIONS(2405), 1,
      anon_sym_LBRACE,
    STATE(1052), 1,
      sym_statement_block,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [55879] = 3,
    ACTIONS(2405), 1,
      anon_sym_LBRACE,
    STATE(1059), 1,
      sym_statement_block,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [55890] = 3,
    ACTIONS(2405), 1,
      anon_sym_LBRACE,
    STATE(1060), 1,
      sym_statement_block,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [55901] = 3,
    ACTIONS(2405), 1,
      anon_sym_LBRACE,
    STATE(1061), 1,
      sym_statement_block,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [55912] = 3,
    ACTIONS(2405), 1,
      anon_sym_LBRACE,
    STATE(1062), 1,
      sym_statement_block,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [55923] = 3,
    ACTIONS(2405), 1,
      anon_sym_LBRACE,
    STATE(1063), 1,
      sym_statement_block,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [55934] = 3,
    ACTIONS(2405), 1,
      anon_sym_LBRACE,
    STATE(1039), 1,
      sym_statement_block,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [55945] = 3,
    ACTIONS(3526), 1,
      anon_sym_LBRACE,
    STATE(636), 1,
      sym_statement_block,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [55956] = 3,
    ACTIONS(3350), 1,
      anon_sym_LBRACE,
    STATE(1040), 1,
      sym_statement_block,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [55967] = 3,
    ACTIONS(3017), 1,
      anon_sym_LBRACE,
    STATE(98), 1,
      sym_class_body,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [55978] = 3,
    ACTIONS(2405), 1,
      anon_sym_LBRACE,
    STATE(1053), 1,
      sym_statement_block,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [55989] = 3,
    ACTIONS(2405), 1,
      anon_sym_LBRACE,
    STATE(1054), 1,
      sym_statement_block,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [56000] = 2,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(3542), 2,
      anon_sym_of,
      anon_sym_in,
  [56009] = 2,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(3544), 2,
      anon_sym_of,
      anon_sym_in,
  [56018] = 3,
    ACTIONS(2405), 1,
      anon_sym_LBRACE,
    STATE(1055), 1,
      sym_statement_block,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [56029] = 3,
    ACTIONS(2405), 1,
      anon_sym_LBRACE,
    STATE(1056), 1,
      sym_statement_block,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [56040] = 3,
    ACTIONS(2405), 1,
      anon_sym_LBRACE,
    STATE(1057), 1,
      sym_statement_block,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [56051] = 3,
    ACTIONS(2405), 1,
      anon_sym_LBRACE,
    STATE(1058), 1,
      sym_statement_block,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [56062] = 3,
    ACTIONS(3546), 1,
      sym_identifier,
    ACTIONS(3548), 1,
      anon_sym_STAR,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [56073] = 3,
    ACTIONS(3550), 1,
      sym_identifier,
    ACTIONS(3552), 1,
      anon_sym_STAR,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [56084] = 3,
    ACTIONS(3017), 1,
      anon_sym_LBRACE,
    STATE(807), 1,
      sym_class_body,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [56095] = 3,
    ACTIONS(3516), 1,
      anon_sym_LPAREN,
    STATE(71), 1,
      sym_parenthesized_expression,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [56106] = 3,
    ACTIONS(2939), 1,
      anon_sym_LPAREN,
    STATE(1683), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [56117] = 2,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(3554), 2,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
  [56126] = 2,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(3556), 2,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
  [56135] = 2,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(3466), 2,
      anon_sym_COMMA,
      anon_sym_RPAREN,
  [56144] = 2,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1956), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [56153] = 3,
    ACTIONS(2939), 1,
      anon_sym_LPAREN,
    STATE(1502), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [56164] = 3,
    ACTIONS(3558), 1,
      anon_sym_SEMI,
    ACTIONS(3560), 1,
      sym__automatic_semicolon,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [56175] = 3,
    ACTIONS(3510), 1,
      anon_sym_LBRACE,
    STATE(779), 1,
      sym_statement_block,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [56186] = 3,
    ACTIONS(3562), 1,
      anon_sym_SEMI,
    ACTIONS(3564), 1,
      sym__automatic_semicolon,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [56197] = 3,
    ACTIONS(3566), 1,
      anon_sym_LBRACE,
    STATE(478), 1,
      sym_switch_body,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [56208] = 3,
    ACTIONS(3516), 1,
      anon_sym_LPAREN,
    STATE(57), 1,
      sym_parenthesized_expression,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [56219] = 2,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(3568), 2,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
  [56228] = 3,
    ACTIONS(2939), 1,
      anon_sym_LPAREN,
    STATE(1518), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [56239] = 3,
    ACTIONS(3510), 1,
      anon_sym_LBRACE,
    STATE(764), 1,
      sym_statement_block,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [56250] = 3,
    ACTIONS(3510), 1,
      anon_sym_LBRACE,
    STATE(99), 1,
      sym_statement_block,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [56261] = 2,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(3570), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [56270] = 3,
    ACTIONS(3572), 1,
      anon_sym_SEMI,
    ACTIONS(3574), 1,
      sym__automatic_semicolon,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [56281] = 3,
    ACTIONS(2939), 1,
      anon_sym_LPAREN,
    STATE(1672), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [56292] = 3,
    ACTIONS(3282), 1,
      anon_sym_LBRACE,
    STATE(398), 1,
      sym_class_body,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [56303] = 3,
    ACTIONS(3516), 1,
      anon_sym_LPAREN,
    STATE(1667), 1,
      sym_parenthesized_expression,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [56314] = 2,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(3576), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [56323] = 3,
    ACTIONS(3510), 1,
      anon_sym_LBRACE,
    STATE(96), 1,
      sym_statement_block,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [56334] = 3,
    ACTIONS(2939), 1,
      anon_sym_LPAREN,
    STATE(1545), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [56345] = 3,
    ACTIONS(2939), 1,
      anon_sym_LPAREN,
    STATE(1546), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [56356] = 3,
    ACTIONS(2939), 1,
      anon_sym_LPAREN,
    STATE(1542), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [56367] = 3,
    ACTIONS(3534), 1,
      anon_sym_LBRACE,
    STATE(409), 1,
      sym_statement_block,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [56378] = 3,
    ACTIONS(3510), 1,
      anon_sym_LBRACE,
    STATE(699), 1,
      sym_statement_block,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [56389] = 2,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(3578), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [56398] = 3,
    ACTIONS(3580), 1,
      anon_sym_SEMI,
    ACTIONS(3582), 1,
      sym__automatic_semicolon,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [56409] = 3,
    ACTIONS(3584), 1,
      anon_sym_LPAREN,
    STATE(395), 1,
      sym_parenthesized_expression,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [56420] = 3,
    ACTIONS(3350), 1,
      anon_sym_LBRACE,
    STATE(458), 1,
      sym_statement_block,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [56431] = 2,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(3586), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [56440] = 3,
    ACTIONS(3350), 1,
      anon_sym_LBRACE,
    STATE(1044), 1,
      sym_statement_block,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [56451] = 3,
    ACTIONS(3350), 1,
      anon_sym_LBRACE,
    STATE(1053), 1,
      sym_statement_block,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [56462] = 3,
    ACTIONS(3248), 1,
      anon_sym_from,
    STATE(1664), 1,
      sym__from_clause,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [56473] = 3,
    ACTIONS(3350), 1,
      anon_sym_LBRACE,
    STATE(1054), 1,
      sym_statement_block,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [56484] = 3,
    ACTIONS(2939), 1,
      anon_sym_LPAREN,
    STATE(1551), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [56495] = 2,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1954), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [56504] = 3,
    ACTIONS(3017), 1,
      anon_sym_LBRACE,
    STATE(743), 1,
      sym_class_body,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [56515] = 3,
    ACTIONS(3528), 1,
      anon_sym_LPAREN,
    STATE(53), 1,
      sym__for_header,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [56526] = 3,
    ACTIONS(2405), 1,
      anon_sym_LBRACE,
    STATE(1078), 1,
      sym_statement_block,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [56537] = 2,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(3348), 2,
      anon_sym_COMMA,
      anon_sym_RBRACK,
  [56546] = 3,
    ACTIONS(3534), 1,
      anon_sym_LBRACE,
    STATE(408), 1,
      sym_statement_block,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [56557] = 3,
    ACTIONS(3350), 1,
      anon_sym_LBRACE,
    STATE(1055), 1,
      sym_statement_block,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [56568] = 2,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(3588), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [56577] = 3,
    ACTIONS(3590), 1,
      anon_sym_LBRACE,
    STATE(1685), 1,
      sym_object,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [56588] = 3,
    ACTIONS(3350), 1,
      anon_sym_LBRACE,
    STATE(1046), 1,
      sym_statement_block,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [56599] = 3,
    ACTIONS(3350), 1,
      anon_sym_LBRACE,
    STATE(1048), 1,
      sym_statement_block,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [56610] = 3,
    ACTIONS(3350), 1,
      anon_sym_LBRACE,
    STATE(1049), 1,
      sym_statement_block,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [56621] = 3,
    ACTIONS(3350), 1,
      anon_sym_LBRACE,
    STATE(1050), 1,
      sym_statement_block,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [56632] = 3,
    ACTIONS(3350), 1,
      anon_sym_LBRACE,
    STATE(381), 1,
      sym_statement_block,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [56643] = 3,
    ACTIONS(2939), 1,
      anon_sym_LPAREN,
    STATE(1628), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [56654] = 3,
    ACTIONS(2939), 1,
      anon_sym_LPAREN,
    STATE(1690), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [56665] = 3,
    ACTIONS(3350), 1,
      anon_sym_LBRACE,
    STATE(1051), 1,
      sym_statement_block,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [56676] = 3,
    ACTIONS(2939), 1,
      anon_sym_LPAREN,
    STATE(1691), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [56687] = 3,
    ACTIONS(2939), 1,
      anon_sym_LPAREN,
    STATE(1629), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [56698] = 3,
    ACTIONS(3510), 1,
      anon_sym_LBRACE,
    STATE(754), 1,
      sym_statement_block,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [56709] = 3,
    ACTIONS(2939), 1,
      anon_sym_LPAREN,
    STATE(1693), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [56720] = 3,
    ACTIONS(3056), 1,
      anon_sym_of,
    ACTIONS(3058), 1,
      anon_sym_in,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [56731] = 3,
    ACTIONS(3350), 1,
      anon_sym_LBRACE,
    STATE(1056), 1,
      sym_statement_block,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [56742] = 3,
    ACTIONS(2939), 1,
      anon_sym_LPAREN,
    STATE(1701), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [56753] = 3,
    ACTIONS(3350), 1,
      anon_sym_LBRACE,
    STATE(1052), 1,
      sym_statement_block,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [56764] = 3,
    ACTIONS(3350), 1,
      anon_sym_LBRACE,
    STATE(1059), 1,
      sym_statement_block,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [56775] = 3,
    ACTIONS(3350), 1,
      anon_sym_LBRACE,
    STATE(1060), 1,
      sym_statement_block,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [56786] = 2,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(3393), 2,
      anon_sym_COMMA,
      anon_sym_RBRACE,
  [56795] = 3,
    ACTIONS(3350), 1,
      anon_sym_LBRACE,
    STATE(1057), 1,
      sym_statement_block,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [56806] = 3,
    ACTIONS(3592), 1,
      sym_identifier,
    ACTIONS(3594), 1,
      sym_jsx_identifier,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [56817] = 3,
    ACTIONS(3350), 1,
      anon_sym_LBRACE,
    STATE(1058), 1,
      sym_statement_block,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [56828] = 3,
    ACTIONS(2939), 1,
      anon_sym_LPAREN,
    STATE(1558), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [56839] = 2,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(3596), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [56848] = 3,
    ACTIONS(3350), 1,
      anon_sym_LBRACE,
    STATE(407), 1,
      sym_statement_block,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [56859] = 3,
    ACTIONS(3064), 1,
      anon_sym_of,
    ACTIONS(3066), 1,
      anon_sym_in,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [56870] = 3,
    ACTIONS(2939), 1,
      anon_sym_LPAREN,
    STATE(1564), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [56881] = 2,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(3598), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [56890] = 2,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(3600), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [56899] = 3,
    ACTIONS(2751), 1,
      anon_sym_COLON,
    ACTIONS(3384), 1,
      anon_sym_GT,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [56910] = 2,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(3602), 2,
      anon_sym_of,
      anon_sym_in,
  [56919] = 3,
    ACTIONS(3350), 1,
      anon_sym_LBRACE,
    STATE(1061), 1,
      sym_statement_block,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [56930] = 3,
    ACTIONS(2757), 1,
      anon_sym_DOT,
    ACTIONS(3384), 1,
      anon_sym_GT,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [56941] = 2,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(1749), 2,
      sym__automatic_semicolon,
      anon_sym_SEMI,
  [56950] = 3,
    ACTIONS(3070), 1,
      anon_sym_LBRACE,
    STATE(595), 1,
      sym_class_body,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [56961] = 3,
    ACTIONS(2939), 1,
      anon_sym_LPAREN,
    STATE(1717), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [56972] = 3,
    ACTIONS(2939), 1,
      anon_sym_LPAREN,
    STATE(1723), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [56983] = 3,
    ACTIONS(2939), 1,
      anon_sym_LPAREN,
    STATE(1725), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [56994] = 3,
    ACTIONS(3350), 1,
      anon_sym_LBRACE,
    STATE(1062), 1,
      sym_statement_block,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57005] = 3,
    ACTIONS(3350), 1,
      anon_sym_LBRACE,
    STATE(1063), 1,
      sym_statement_block,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57016] = 3,
    ACTIONS(3350), 1,
      anon_sym_LBRACE,
    STATE(1039), 1,
      sym_statement_block,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57027] = 3,
    ACTIONS(2939), 1,
      anon_sym_LPAREN,
    STATE(1679), 1,
      sym_formal_parameters,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57038] = 3,
    ACTIONS(3526), 1,
      anon_sym_LBRACE,
    STATE(596), 1,
      sym_statement_block,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57049] = 2,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
    ACTIONS(3604), 2,
      anon_sym_of,
      anon_sym_in,
  [57058] = 2,
    ACTIONS(1991), 1,
      anon_sym_RPAREN,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57066] = 2,
    ACTIONS(2101), 1,
      anon_sym_RBRACK,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57074] = 2,
    ACTIONS(2109), 1,
      anon_sym_RPAREN,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57082] = 2,
    ACTIONS(3606), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57090] = 2,
    ACTIONS(3413), 1,
      anon_sym_RBRACE,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57098] = 2,
    ACTIONS(2019), 1,
      anon_sym_RBRACE,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57106] = 2,
    ACTIONS(3608), 1,
      anon_sym_target,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57114] = 2,
    ACTIONS(3610), 1,
      anon_sym_while,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57122] = 2,
    ACTIONS(1804), 1,
      anon_sym_in,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57130] = 2,
    ACTIONS(2003), 1,
      anon_sym_RPAREN,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57138] = 2,
    ACTIONS(3608), 1,
      anon_sym_meta,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57146] = 2,
    ACTIONS(3612), 1,
      anon_sym_COLON,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57154] = 2,
    ACTIONS(1989), 1,
      anon_sym_RPAREN,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57162] = 2,
    ACTIONS(2005), 1,
      anon_sym_RPAREN,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57170] = 2,
    ACTIONS(2999), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57178] = 2,
    ACTIONS(3614), 1,
      anon_sym_EQ_GT,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57186] = 2,
    ACTIONS(2111), 1,
      anon_sym_RPAREN,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57194] = 2,
    ACTIONS(2045), 1,
      anon_sym_RBRACE,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57202] = 2,
    ACTIONS(3616), 1,
      anon_sym_from,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57210] = 2,
    ACTIONS(2395), 1,
      anon_sym_LPAREN,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57218] = 2,
    ACTIONS(2577), 1,
      anon_sym_EQ_GT,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57226] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_html_comment,
    ACTIONS(3618), 1,
      sym_regex_pattern,
  [57236] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_html_comment,
    ACTIONS(3620), 1,
      anon_sym_SLASH2,
  [57246] = 2,
    ACTIONS(2189), 1,
      anon_sym_in,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57254] = 2,
    ACTIONS(3622), 1,
      anon_sym_from,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57262] = 2,
    ACTIONS(2127), 1,
      anon_sym_RPAREN,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57270] = 2,
    ACTIONS(2015), 1,
      anon_sym_RPAREN,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57278] = 2,
    ACTIONS(3624), 1,
      anon_sym_as,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57286] = 2,
    ACTIONS(3626), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57294] = 2,
    ACTIONS(2223), 1,
      anon_sym_in,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57302] = 2,
    ACTIONS(3628), 1,
      anon_sym_EQ_GT,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57310] = 2,
    ACTIONS(3630), 1,
      anon_sym_EQ_GT,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57318] = 2,
    ACTIONS(3632), 1,
      anon_sym_EQ_GT,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57326] = 2,
    ACTIONS(3634), 1,
      anon_sym_EQ_GT,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57334] = 2,
    ACTIONS(3636), 1,
      anon_sym_EQ_GT,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57342] = 2,
    ACTIONS(3638), 1,
      anon_sym_using,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57350] = 2,
    ACTIONS(3640), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57358] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_html_comment,
    ACTIONS(3642), 1,
      sym_regex_pattern,
  [57368] = 2,
    ACTIONS(3644), 1,
      ts_builtin_sym_end,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57376] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_html_comment,
    ACTIONS(3646), 1,
      anon_sym_SLASH2,
  [57386] = 2,
    ACTIONS(3648), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57394] = 2,
    ACTIONS(2049), 1,
      anon_sym_SEMI,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57402] = 2,
    ACTIONS(3650), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57410] = 2,
    ACTIONS(3652), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57418] = 2,
    ACTIONS(1985), 1,
      anon_sym_RPAREN,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57426] = 2,
    ACTIONS(3654), 1,
      anon_sym_as,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57434] = 2,
    ACTIONS(3656), 1,
      anon_sym_meta,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57442] = 2,
    ACTIONS(2007), 1,
      anon_sym_RPAREN,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57450] = 2,
    ACTIONS(3658), 1,
      anon_sym_COLON,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57458] = 2,
    ACTIONS(3660), 1,
      anon_sym_from,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57466] = 2,
    ACTIONS(1866), 1,
      anon_sym_in,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57474] = 2,
    ACTIONS(3415), 1,
      anon_sym_GT,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57482] = 2,
    ACTIONS(3662), 1,
      anon_sym_from,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57490] = 2,
    ACTIONS(3664), 1,
      anon_sym_function,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57498] = 2,
    ACTIONS(3666), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57506] = 2,
    ACTIONS(3668), 1,
      anon_sym_EQ_GT,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57514] = 2,
    ACTIONS(2063), 1,
      anon_sym_RPAREN,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57522] = 2,
    ACTIONS(3670), 1,
      anon_sym_from,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57530] = 2,
    ACTIONS(1987), 1,
      anon_sym_RPAREN,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57538] = 2,
    ACTIONS(2117), 1,
      anon_sym_RPAREN,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57546] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_html_comment,
    ACTIONS(3672), 1,
      anon_sym_SLASH2,
  [57556] = 2,
    ACTIONS(2153), 1,
      anon_sym_in,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57564] = 2,
    ACTIONS(3674), 1,
      anon_sym_EQ_GT,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57572] = 2,
    ACTIONS(2077), 1,
      anon_sym_RPAREN,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57580] = 2,
    ACTIONS(2107), 1,
      anon_sym_RPAREN,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57588] = 2,
    ACTIONS(3676), 1,
      anon_sym_EQ_GT,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57596] = 2,
    ACTIONS(3500), 1,
      anon_sym_GT,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57604] = 2,
    ACTIONS(3163), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57612] = 2,
    ACTIONS(3678), 1,
      anon_sym_EQ_GT,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57620] = 2,
    ACTIONS(2105), 1,
      anon_sym_RPAREN,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57628] = 2,
    ACTIONS(3680), 1,
      anon_sym_EQ_GT,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57636] = 2,
    ACTIONS(3682), 1,
      anon_sym_EQ_GT,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57644] = 2,
    ACTIONS(3684), 1,
      anon_sym_EQ_GT,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57652] = 2,
    ACTIONS(3686), 1,
      anon_sym_EQ_GT,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57660] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_html_comment,
    ACTIONS(3688), 1,
      anon_sym_SLASH2,
  [57670] = 2,
    ACTIONS(3690), 1,
      anon_sym_EQ_GT,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57678] = 2,
    ACTIONS(3692), 1,
      anon_sym_EQ_GT,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57686] = 2,
    ACTIONS(3694), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57694] = 2,
    ACTIONS(2037), 1,
      anon_sym_RPAREN,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57702] = 2,
    ACTIONS(3524), 1,
      anon_sym_EQ_GT,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57710] = 2,
    ACTIONS(3356), 1,
      anon_sym_RBRACE,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57718] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_html_comment,
    ACTIONS(3696), 1,
      sym_regex_pattern,
  [57728] = 2,
    ACTIONS(1902), 1,
      anon_sym_in,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57736] = 2,
    ACTIONS(2119), 1,
      anon_sym_RPAREN,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57744] = 3,
    ACTIONS(3), 1,
      sym_comment,
    ACTIONS(5), 1,
      sym_html_comment,
    ACTIONS(3698), 1,
      sym_regex_pattern,
  [57754] = 2,
    ACTIONS(3700), 1,
      anon_sym_from,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57762] = 2,
    ACTIONS(3702), 1,
      anon_sym_EQ_GT,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57770] = 2,
    ACTIONS(2121), 1,
      anon_sym_SEMI,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57778] = 2,
    ACTIONS(3704), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57786] = 2,
    ACTIONS(3706), 1,
      anon_sym_RPAREN,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57794] = 2,
    ACTIONS(2103), 1,
      anon_sym_RPAREN,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57802] = 2,
    ACTIONS(2113), 1,
      anon_sym_COLON,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57810] = 2,
    ACTIONS(3708), 1,
      anon_sym_RPAREN,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57818] = 2,
    ACTIONS(2047), 1,
      anon_sym_RBRACK,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57826] = 2,
    ACTIONS(2001), 1,
      anon_sym_RPAREN,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57834] = 2,
    ACTIONS(2123), 1,
      anon_sym_SEMI,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57842] = 2,
    ACTIONS(3710), 1,
      anon_sym_RPAREN,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57850] = 2,
    ACTIONS(2025), 1,
      anon_sym_RBRACE,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57858] = 2,
    ACTIONS(3384), 1,
      anon_sym_GT,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57866] = 2,
    ACTIONS(2011), 1,
      anon_sym_SEMI,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57874] = 2,
    ACTIONS(3712), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57882] = 2,
    ACTIONS(3714), 1,
      anon_sym_EQ_GT,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57890] = 2,
    ACTIONS(3716), 1,
      anon_sym_RPAREN,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57898] = 2,
    ACTIONS(3718), 1,
      anon_sym_EQ_GT,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57906] = 2,
    ACTIONS(2125), 1,
      anon_sym_SEMI,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57914] = 2,
    ACTIONS(3720), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57922] = 2,
    ACTIONS(2115), 1,
      anon_sym_RPAREN,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57930] = 2,
    ACTIONS(3722), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57938] = 2,
    ACTIONS(3724), 1,
      anon_sym_EQ_GT,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57946] = 2,
    ACTIONS(3726), 1,
      anon_sym_EQ,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57954] = 2,
    ACTIONS(2043), 1,
      anon_sym_RBRACK,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57962] = 2,
    ACTIONS(3728), 1,
      anon_sym_EQ_GT,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57970] = 2,
    ACTIONS(3730), 1,
      sym_identifier,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57978] = 2,
    ACTIONS(3382), 1,
      anon_sym_GT,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57986] = 2,
    ACTIONS(3540), 1,
      anon_sym_from,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [57994] = 2,
    ACTIONS(2013), 1,
      anon_sym_RPAREN,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [58002] = 2,
    ACTIONS(3732), 1,
      anon_sym_RPAREN,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [58010] = 2,
    ACTIONS(3656), 1,
      anon_sym_target,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [58018] = 2,
    ACTIONS(3734), 1,
      anon_sym_from,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [58026] = 2,
    ACTIONS(2023), 1,
      anon_sym_RBRACK,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [58034] = 2,
    ACTIONS(2051), 1,
      anon_sym_SEMI,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
  [58042] = 2,
    ACTIONS(3736), 1,
      anon_sym_EQ_GT,
    ACTIONS(5), 2,
      sym_html_comment,
      sym_comment,
};

static const uint32_t ts_small_parse_table_map[] = {
  [SMALL_STATE(387)] = 0,
  [SMALL_STATE(388)] = 91,
  [SMALL_STATE(389)] = 178,
  [SMALL_STATE(390)] = 253,
  [SMALL_STATE(391)] = 340,
  [SMALL_STATE(392)] = 410,
  [SMALL_STATE(393)] = 482,
  [SMALL_STATE(394)] = 568,
  [SMALL_STATE(395)] = 640,
  [SMALL_STATE(396)] = 712,
  [SMALL_STATE(397)] = 802,
  [SMALL_STATE(398)] = 892,
  [SMALL_STATE(399)] = 964,
  [SMALL_STATE(400)] = 1034,
  [SMALL_STATE(401)] = 1104,
  [SMALL_STATE(402)] = 1174,
  [SMALL_STATE(403)] = 1244,
  [SMALL_STATE(404)] = 1316,
  [SMALL_STATE(405)] = 1386,
  [SMALL_STATE(406)] = 1458,
  [SMALL_STATE(407)] = 1544,
  [SMALL_STATE(408)] = 1614,
  [SMALL_STATE(409)] = 1686,
  [SMALL_STATE(410)] = 1758,
  [SMALL_STATE(411)] = 1844,
  [SMALL_STATE(412)] = 1918,
  [SMALL_STATE(413)] = 2004,
  [SMALL_STATE(414)] = 2076,
  [SMALL_STATE(415)] = 2162,
  [SMALL_STATE(416)] = 2234,
  [SMALL_STATE(417)] = 2306,
  [SMALL_STATE(418)] = 2396,
  [SMALL_STATE(419)] = 2466,
  [SMALL_STATE(420)] = 2535,
  [SMALL_STATE(421)] = 2604,
  [SMALL_STATE(422)] = 2673,
  [SMALL_STATE(423)] = 2742,
  [SMALL_STATE(424)] = 2811,
  [SMALL_STATE(425)] = 2880,
  [SMALL_STATE(426)] = 2949,
  [SMALL_STATE(427)] = 3018,
  [SMALL_STATE(428)] = 3087,
  [SMALL_STATE(429)] = 3156,
  [SMALL_STATE(430)] = 3225,
  [SMALL_STATE(431)] = 3294,
  [SMALL_STATE(432)] = 3363,
  [SMALL_STATE(433)] = 3432,
  [SMALL_STATE(434)] = 3501,
  [SMALL_STATE(435)] = 3570,
  [SMALL_STATE(436)] = 3639,
  [SMALL_STATE(437)] = 3708,
  [SMALL_STATE(438)] = 3777,
  [SMALL_STATE(439)] = 3846,
  [SMALL_STATE(440)] = 3915,
  [SMALL_STATE(441)] = 3984,
  [SMALL_STATE(442)] = 4053,
  [SMALL_STATE(443)] = 4122,
  [SMALL_STATE(444)] = 4191,
  [SMALL_STATE(445)] = 4260,
  [SMALL_STATE(446)] = 4329,
  [SMALL_STATE(447)] = 4398,
  [SMALL_STATE(448)] = 4467,
  [SMALL_STATE(449)] = 4536,
  [SMALL_STATE(450)] = 4605,
  [SMALL_STATE(451)] = 4674,
  [SMALL_STATE(452)] = 4743,
  [SMALL_STATE(453)] = 4812,
  [SMALL_STATE(454)] = 4881,
  [SMALL_STATE(455)] = 4950,
  [SMALL_STATE(456)] = 5019,
  [SMALL_STATE(457)] = 5088,
  [SMALL_STATE(458)] = 5157,
  [SMALL_STATE(459)] = 5226,
  [SMALL_STATE(460)] = 5295,
  [SMALL_STATE(461)] = 5364,
  [SMALL_STATE(462)] = 5433,
  [SMALL_STATE(463)] = 5502,
  [SMALL_STATE(464)] = 5571,
  [SMALL_STATE(465)] = 5640,
  [SMALL_STATE(466)] = 5709,
  [SMALL_STATE(467)] = 5778,
  [SMALL_STATE(468)] = 5847,
  [SMALL_STATE(469)] = 5916,
  [SMALL_STATE(470)] = 5985,
  [SMALL_STATE(471)] = 6054,
  [SMALL_STATE(472)] = 6123,
  [SMALL_STATE(473)] = 6192,
  [SMALL_STATE(474)] = 6261,
  [SMALL_STATE(475)] = 6330,
  [SMALL_STATE(476)] = 6399,
  [SMALL_STATE(477)] = 6468,
  [SMALL_STATE(478)] = 6537,
  [SMALL_STATE(479)] = 6606,
  [SMALL_STATE(480)] = 6675,
  [SMALL_STATE(481)] = 6744,
  [SMALL_STATE(482)] = 6813,
  [SMALL_STATE(483)] = 6882,
  [SMALL_STATE(484)] = 6951,
  [SMALL_STATE(485)] = 7020,
  [SMALL_STATE(486)] = 7089,
  [SMALL_STATE(487)] = 7158,
  [SMALL_STATE(488)] = 7227,
  [SMALL_STATE(489)] = 7315,
  [SMALL_STATE(490)] = 7401,
  [SMALL_STATE(491)] = 7483,
  [SMALL_STATE(492)] = 7549,
  [SMALL_STATE(493)] = 7631,
  [SMALL_STATE(494)] = 7697,
  [SMALL_STATE(495)] = 7763,
  [SMALL_STATE(496)] = 7829,
  [SMALL_STATE(497)] = 7895,
  [SMALL_STATE(498)] = 7961,
  [SMALL_STATE(499)] = 8031,
  [SMALL_STATE(500)] = 8113,
  [SMALL_STATE(501)] = 8179,
  [SMALL_STATE(502)] = 8250,
  [SMALL_STATE(503)] = 8321,
  [SMALL_STATE(504)] = 8392,
  [SMALL_STATE(505)] = 8463,
  [SMALL_STATE(506)] = 8546,
  [SMALL_STATE(507)] = 8610,
  [SMALL_STATE(508)] = 8678,
  [SMALL_STATE(509)] = 8742,
  [SMALL_STATE(510)] = 8806,
  [SMALL_STATE(511)] = 8870,
  [SMALL_STATE(512)] = 8934,
  [SMALL_STATE(513)] = 8998,
  [SMALL_STATE(514)] = 9062,
  [SMALL_STATE(515)] = 9126,
  [SMALL_STATE(516)] = 9190,
  [SMALL_STATE(517)] = 9254,
  [SMALL_STATE(518)] = 9318,
  [SMALL_STATE(519)] = 9382,
  [SMALL_STATE(520)] = 9446,
  [SMALL_STATE(521)] = 9510,
  [SMALL_STATE(522)] = 9574,
  [SMALL_STATE(523)] = 9638,
  [SMALL_STATE(524)] = 9702,
  [SMALL_STATE(525)] = 9766,
  [SMALL_STATE(526)] = 9830,
  [SMALL_STATE(527)] = 9893,
  [SMALL_STATE(528)] = 9962,
  [SMALL_STATE(529)] = 10029,
  [SMALL_STATE(530)] = 10098,
  [SMALL_STATE(531)] = 10167,
  [SMALL_STATE(532)] = 10238,
  [SMALL_STATE(533)] = 10309,
  [SMALL_STATE(534)] = 10380,
  [SMALL_STATE(535)] = 10443,
  [SMALL_STATE(536)] = 10514,
  [SMALL_STATE(537)] = 10585,
  [SMALL_STATE(538)] = 10654,
  [SMALL_STATE(539)] = 10717,
  [SMALL_STATE(540)] = 10788,
  [SMALL_STATE(541)] = 10859,
  [SMALL_STATE(542)] = 10930,
  [SMALL_STATE(543)] = 11001,
  [SMALL_STATE(544)] = 11070,
  [SMALL_STATE(545)] = 11139,
  [SMALL_STATE(546)] = 11202,
  [SMALL_STATE(547)] = 11271,
  [SMALL_STATE(548)] = 11334,
  [SMALL_STATE(549)] = 11397,
  [SMALL_STATE(550)] = 11468,
  [SMALL_STATE(551)] = 11537,
  [SMALL_STATE(552)] = 11600,
  [SMALL_STATE(553)] = 11666,
  [SMALL_STATE(554)] = 11738,
  [SMALL_STATE(555)] = 11808,
  [SMALL_STATE(556)] = 11880,
  [SMALL_STATE(557)] = 11948,
  [SMALL_STATE(558)] = 12016,
  [SMALL_STATE(559)] = 12084,
  [SMALL_STATE(560)] = 12154,
  [SMALL_STATE(561)] = 12220,
  [SMALL_STATE(562)] = 12291,
  [SMALL_STATE(563)] = 12358,
  [SMALL_STATE(564)] = 12429,
  [SMALL_STATE(565)] = 12498,
  [SMALL_STATE(566)] = 12564,
  [SMALL_STATE(567)] = 12630,
  [SMALL_STATE(568)] = 12696,
  [SMALL_STATE(569)] = 12762,
  [SMALL_STATE(570)] = 12828,
  [SMALL_STATE(571)] = 12894,
  [SMALL_STATE(572)] = 12960,
  [SMALL_STATE(573)] = 13028,
  [SMALL_STATE(574)] = 13094,
  [SMALL_STATE(575)] = 13164,
  [SMALL_STATE(576)] = 13234,
  [SMALL_STATE(577)] = 13300,
  [SMALL_STATE(578)] = 13366,
  [SMALL_STATE(579)] = 13429,
  [SMALL_STATE(580)] = 13492,
  [SMALL_STATE(581)] = 13559,
  [SMALL_STATE(582)] = 13625,
  [SMALL_STATE(583)] = 13681,
  [SMALL_STATE(584)] = 13741,
  [SMALL_STATE(585)] = 13794,
  [SMALL_STATE(586)] = 13849,
  [SMALL_STATE(587)] = 13902,
  [SMALL_STATE(588)] = 13965,
  [SMALL_STATE(589)] = 14028,
  [SMALL_STATE(590)] = 14089,
  [SMALL_STATE(591)] = 14142,
  [SMALL_STATE(592)] = 14203,
  [SMALL_STATE(593)] = 14253,
  [SMALL_STATE(594)] = 14303,
  [SMALL_STATE(595)] = 14353,
  [SMALL_STATE(596)] = 14403,
  [SMALL_STATE(597)] = 14453,
  [SMALL_STATE(598)] = 14503,
  [SMALL_STATE(599)] = 14555,
  [SMALL_STATE(600)] = 14605,
  [SMALL_STATE(601)] = 14655,
  [SMALL_STATE(602)] = 14705,
  [SMALL_STATE(603)] = 14755,
  [SMALL_STATE(604)] = 14809,
  [SMALL_STATE(605)] = 14859,
  [SMALL_STATE(606)] = 14909,
  [SMALL_STATE(607)] = 14959,
  [SMALL_STATE(608)] = 15009,
  [SMALL_STATE(609)] = 15059,
  [SMALL_STATE(610)] = 15109,
  [SMALL_STATE(611)] = 15159,
  [SMALL_STATE(612)] = 15209,
  [SMALL_STATE(613)] = 15259,
  [SMALL_STATE(614)] = 15309,
  [SMALL_STATE(615)] = 15359,
  [SMALL_STATE(616)] = 15409,
  [SMALL_STATE(617)] = 15459,
  [SMALL_STATE(618)] = 15509,
  [SMALL_STATE(619)] = 15559,
  [SMALL_STATE(620)] = 15609,
  [SMALL_STATE(621)] = 15659,
  [SMALL_STATE(622)] = 15709,
  [SMALL_STATE(623)] = 15759,
  [SMALL_STATE(624)] = 15809,
  [SMALL_STATE(625)] = 15859,
  [SMALL_STATE(626)] = 15909,
  [SMALL_STATE(627)] = 15959,
  [SMALL_STATE(628)] = 16009,
  [SMALL_STATE(629)] = 16059,
  [SMALL_STATE(630)] = 16109,
  [SMALL_STATE(631)] = 16159,
  [SMALL_STATE(632)] = 16209,
  [SMALL_STATE(633)] = 16263,
  [SMALL_STATE(634)] = 16313,
  [SMALL_STATE(635)] = 16363,
  [SMALL_STATE(636)] = 16413,
  [SMALL_STATE(637)] = 16463,
  [SMALL_STATE(638)] = 16513,
  [SMALL_STATE(639)] = 16563,
  [SMALL_STATE(640)] = 16613,
  [SMALL_STATE(641)] = 16663,
  [SMALL_STATE(642)] = 16713,
  [SMALL_STATE(643)] = 16763,
  [SMALL_STATE(644)] = 16813,
  [SMALL_STATE(645)] = 16863,
  [SMALL_STATE(646)] = 16913,
  [SMALL_STATE(647)] = 16963,
  [SMALL_STATE(648)] = 17013,
  [SMALL_STATE(649)] = 17067,
  [SMALL_STATE(650)] = 17117,
  [SMALL_STATE(651)] = 17167,
  [SMALL_STATE(652)] = 17217,
  [SMALL_STATE(653)] = 17268,
  [SMALL_STATE(654)] = 17361,
  [SMALL_STATE(655)] = 17454,
  [SMALL_STATE(656)] = 17547,
  [SMALL_STATE(657)] = 17640,
  [SMALL_STATE(658)] = 17703,
  [SMALL_STATE(659)] = 17796,
  [SMALL_STATE(660)] = 17847,
  [SMALL_STATE(661)] = 17940,
  [SMALL_STATE(662)] = 18033,
  [SMALL_STATE(663)] = 18106,
  [SMALL_STATE(664)] = 18199,
  [SMALL_STATE(665)] = 18292,
  [SMALL_STATE(666)] = 18343,
  [SMALL_STATE(667)] = 18436,
  [SMALL_STATE(668)] = 18529,
  [SMALL_STATE(669)] = 18622,
  [SMALL_STATE(670)] = 18685,
  [SMALL_STATE(671)] = 18736,
  [SMALL_STATE(672)] = 18821,
  [SMALL_STATE(673)] = 18908,
  [SMALL_STATE(674)] = 18977,
  [SMALL_STATE(675)] = 19058,
  [SMALL_STATE(676)] = 19141,
  [SMALL_STATE(677)] = 19226,
  [SMALL_STATE(678)] = 19293,
  [SMALL_STATE(679)] = 19356,
  [SMALL_STATE(680)] = 19433,
  [SMALL_STATE(681)] = 19522,
  [SMALL_STATE(682)] = 19579,
  [SMALL_STATE(683)] = 19631,
  [SMALL_STATE(684)] = 19681,
  [SMALL_STATE(685)] = 19733,
  [SMALL_STATE(686)] = 19783,
  [SMALL_STATE(687)] = 19841,
  [SMALL_STATE(688)] = 19889,
  [SMALL_STATE(689)] = 19939,
  [SMALL_STATE(690)] = 19987,
  [SMALL_STATE(691)] = 20079,
  [SMALL_STATE(692)] = 20137,
  [SMALL_STATE(693)] = 20197,
  [SMALL_STATE(694)] = 20251,
  [SMALL_STATE(695)] = 20301,
  [SMALL_STATE(696)] = 20361,
  [SMALL_STATE(697)] = 20452,
  [SMALL_STATE(698)] = 20499,
  [SMALL_STATE(699)] = 20590,
  [SMALL_STATE(700)] = 20637,
  [SMALL_STATE(701)] = 20684,
  [SMALL_STATE(702)] = 20733,
  [SMALL_STATE(703)] = 20784,
  [SMALL_STATE(704)] = 20875,
  [SMALL_STATE(705)] = 20946,
  [SMALL_STATE(706)] = 21007,
  [SMALL_STATE(707)] = 21054,
  [SMALL_STATE(708)] = 21137,
  [SMALL_STATE(709)] = 21222,
  [SMALL_STATE(710)] = 21289,
  [SMALL_STATE(711)] = 21368,
  [SMALL_STATE(712)] = 21449,
  [SMALL_STATE(713)] = 21532,
  [SMALL_STATE(714)] = 21597,
  [SMALL_STATE(715)] = 21658,
  [SMALL_STATE(716)] = 21733,
  [SMALL_STATE(717)] = 21820,
  [SMALL_STATE(718)] = 21911,
  [SMALL_STATE(719)] = 22002,
  [SMALL_STATE(720)] = 22093,
  [SMALL_STATE(721)] = 22140,
  [SMALL_STATE(722)] = 22231,
  [SMALL_STATE(723)] = 22322,
  [SMALL_STATE(724)] = 22413,
  [SMALL_STATE(725)] = 22464,
  [SMALL_STATE(726)] = 22555,
  [SMALL_STATE(727)] = 22626,
  [SMALL_STATE(728)] = 22717,
  [SMALL_STATE(729)] = 22764,
  [SMALL_STATE(730)] = 22855,
  [SMALL_STATE(731)] = 22906,
  [SMALL_STATE(732)] = 22957,
  [SMALL_STATE(733)] = 23004,
  [SMALL_STATE(734)] = 23055,
  [SMALL_STATE(735)] = 23102,
  [SMALL_STATE(736)] = 23149,
  [SMALL_STATE(737)] = 23196,
  [SMALL_STATE(738)] = 23243,
  [SMALL_STATE(739)] = 23290,
  [SMALL_STATE(740)] = 23337,
  [SMALL_STATE(741)] = 23384,
  [SMALL_STATE(742)] = 23431,
  [SMALL_STATE(743)] = 23478,
  [SMALL_STATE(744)] = 23525,
  [SMALL_STATE(745)] = 23572,
  [SMALL_STATE(746)] = 23619,
  [SMALL_STATE(747)] = 23666,
  [SMALL_STATE(748)] = 23713,
  [SMALL_STATE(749)] = 23760,
  [SMALL_STATE(750)] = 23821,
  [SMALL_STATE(751)] = 23868,
  [SMALL_STATE(752)] = 23951,
  [SMALL_STATE(753)] = 24002,
  [SMALL_STATE(754)] = 24087,
  [SMALL_STATE(755)] = 24134,
  [SMALL_STATE(756)] = 24181,
  [SMALL_STATE(757)] = 24248,
  [SMALL_STATE(758)] = 24295,
  [SMALL_STATE(759)] = 24342,
  [SMALL_STATE(760)] = 24389,
  [SMALL_STATE(761)] = 24468,
  [SMALL_STATE(762)] = 24519,
  [SMALL_STATE(763)] = 24610,
  [SMALL_STATE(764)] = 24691,
  [SMALL_STATE(765)] = 24738,
  [SMALL_STATE(766)] = 24789,
  [SMALL_STATE(767)] = 24836,
  [SMALL_STATE(768)] = 24883,
  [SMALL_STATE(769)] = 24934,
  [SMALL_STATE(770)] = 25025,
  [SMALL_STATE(771)] = 25072,
  [SMALL_STATE(772)] = 25121,
  [SMALL_STATE(773)] = 25168,
  [SMALL_STATE(774)] = 25215,
  [SMALL_STATE(775)] = 25266,
  [SMALL_STATE(776)] = 25349,
  [SMALL_STATE(777)] = 25414,
  [SMALL_STATE(778)] = 25475,
  [SMALL_STATE(779)] = 25550,
  [SMALL_STATE(780)] = 25597,
  [SMALL_STATE(781)] = 25684,
  [SMALL_STATE(782)] = 25731,
  [SMALL_STATE(783)] = 25782,
  [SMALL_STATE(784)] = 25829,
  [SMALL_STATE(785)] = 25876,
  [SMALL_STATE(786)] = 25923,
  [SMALL_STATE(787)] = 25970,
  [SMALL_STATE(788)] = 26061,
  [SMALL_STATE(789)] = 26108,
  [SMALL_STATE(790)] = 26155,
  [SMALL_STATE(791)] = 26202,
  [SMALL_STATE(792)] = 26293,
  [SMALL_STATE(793)] = 26384,
  [SMALL_STATE(794)] = 26433,
  [SMALL_STATE(795)] = 26480,
  [SMALL_STATE(796)] = 26529,
  [SMALL_STATE(797)] = 26620,
  [SMALL_STATE(798)] = 26667,
  [SMALL_STATE(799)] = 26718,
  [SMALL_STATE(800)] = 26769,
  [SMALL_STATE(801)] = 26816,
  [SMALL_STATE(802)] = 26863,
  [SMALL_STATE(803)] = 26954,
  [SMALL_STATE(804)] = 27045,
  [SMALL_STATE(805)] = 27096,
  [SMALL_STATE(806)] = 27143,
  [SMALL_STATE(807)] = 27234,
  [SMALL_STATE(808)] = 27281,
  [SMALL_STATE(809)] = 27372,
  [SMALL_STATE(810)] = 27463,
  [SMALL_STATE(811)] = 27510,
  [SMALL_STATE(812)] = 27601,
  [SMALL_STATE(813)] = 27653,
  [SMALL_STATE(814)] = 27747,
  [SMALL_STATE(815)] = 27841,
  [SMALL_STATE(816)] = 27891,
  [SMALL_STATE(817)] = 27941,
  [SMALL_STATE(818)] = 27993,
  [SMALL_STATE(819)] = 28087,
  [SMALL_STATE(820)] = 28137,
  [SMALL_STATE(821)] = 28189,
  [SMALL_STATE(822)] = 28283,
  [SMALL_STATE(823)] = 28377,
  [SMALL_STATE(824)] = 28471,
  [SMALL_STATE(825)] = 28561,
  [SMALL_STATE(826)] = 28651,
  [SMALL_STATE(827)] = 28743,
  [SMALL_STATE(828)] = 28795,
  [SMALL_STATE(829)] = 28882,
  [SMALL_STATE(830)] = 28975,
  [SMALL_STATE(831)] = 29068,
  [SMALL_STATE(832)] = 29161,
  [SMALL_STATE(833)] = 29254,
  [SMALL_STATE(834)] = 29347,
  [SMALL_STATE(835)] = 29440,
  [SMALL_STATE(836)] = 29533,
  [SMALL_STATE(837)] = 29626,
  [SMALL_STATE(838)] = 29719,
  [SMALL_STATE(839)] = 29812,
  [SMALL_STATE(840)] = 29905,
  [SMALL_STATE(841)] = 29998,
  [SMALL_STATE(842)] = 30091,
  [SMALL_STATE(843)] = 30184,
  [SMALL_STATE(844)] = 30277,
  [SMALL_STATE(845)] = 30370,
  [SMALL_STATE(846)] = 30459,
  [SMALL_STATE(847)] = 30552,
  [SMALL_STATE(848)] = 30645,
  [SMALL_STATE(849)] = 30734,
  [SMALL_STATE(850)] = 30827,
  [SMALL_STATE(851)] = 30920,
  [SMALL_STATE(852)] = 30971,
  [SMALL_STATE(853)] = 31022,
  [SMALL_STATE(854)] = 31073,
  [SMALL_STATE(855)] = 31124,
  [SMALL_STATE(856)] = 31213,
  [SMALL_STATE(857)] = 31300,
  [SMALL_STATE(858)] = 31393,
  [SMALL_STATE(859)] = 31486,
  [SMALL_STATE(860)] = 31579,
  [SMALL_STATE(861)] = 31672,
  [SMALL_STATE(862)] = 31765,
  [SMALL_STATE(863)] = 31858,
  [SMALL_STATE(864)] = 31951,
  [SMALL_STATE(865)] = 32044,
  [SMALL_STATE(866)] = 32131,
  [SMALL_STATE(867)] = 32224,
  [SMALL_STATE(868)] = 32317,
  [SMALL_STATE(869)] = 32404,
  [SMALL_STATE(870)] = 32497,
  [SMALL_STATE(871)] = 32584,
  [SMALL_STATE(872)] = 32671,
  [SMALL_STATE(873)] = 32764,
  [SMALL_STATE(874)] = 32857,
  [SMALL_STATE(875)] = 32950,
  [SMALL_STATE(876)] = 33043,
  [SMALL_STATE(877)] = 33136,
  [SMALL_STATE(878)] = 33229,
  [SMALL_STATE(879)] = 33322,
  [SMALL_STATE(880)] = 33415,
  [SMALL_STATE(881)] = 33508,
  [SMALL_STATE(882)] = 33601,
  [SMALL_STATE(883)] = 33694,
  [SMALL_STATE(884)] = 33787,
  [SMALL_STATE(885)] = 33880,
  [SMALL_STATE(886)] = 33973,
  [SMALL_STATE(887)] = 34066,
  [SMALL_STATE(888)] = 34154,
  [SMALL_STATE(889)] = 34200,
  [SMALL_STATE(890)] = 34288,
  [SMALL_STATE(891)] = 34376,
  [SMALL_STATE(892)] = 34464,
  [SMALL_STATE(893)] = 34552,
  [SMALL_STATE(894)] = 34640,
  [SMALL_STATE(895)] = 34728,
  [SMALL_STATE(896)] = 34796,
  [SMALL_STATE(897)] = 34854,
  [SMALL_STATE(898)] = 34934,
  [SMALL_STATE(899)] = 35016,
  [SMALL_STATE(900)] = 35080,
  [SMALL_STATE(901)] = 35156,
  [SMALL_STATE(902)] = 35234,
  [SMALL_STATE(903)] = 35314,
  [SMALL_STATE(904)] = 35376,
  [SMALL_STATE(905)] = 35434,
  [SMALL_STATE(906)] = 35506,
  [SMALL_STATE(907)] = 35590,
  [SMALL_STATE(908)] = 35678,
  [SMALL_STATE(909)] = 35766,
  [SMALL_STATE(910)] = 35854,
  [SMALL_STATE(911)] = 35942,
  [SMALL_STATE(912)] = 36030,
  [SMALL_STATE(913)] = 36118,
  [SMALL_STATE(914)] = 36206,
  [SMALL_STATE(915)] = 36296,
  [SMALL_STATE(916)] = 36384,
  [SMALL_STATE(917)] = 36472,
  [SMALL_STATE(918)] = 36518,
  [SMALL_STATE(919)] = 36606,
  [SMALL_STATE(920)] = 36694,
  [SMALL_STATE(921)] = 36782,
  [SMALL_STATE(922)] = 36870,
  [SMALL_STATE(923)] = 36916,
  [SMALL_STATE(924)] = 36984,
  [SMALL_STATE(925)] = 37042,
  [SMALL_STATE(926)] = 37122,
  [SMALL_STATE(927)] = 37204,
  [SMALL_STATE(928)] = 37268,
  [SMALL_STATE(929)] = 37344,
  [SMALL_STATE(930)] = 37422,
  [SMALL_STATE(931)] = 37502,
  [SMALL_STATE(932)] = 37564,
  [SMALL_STATE(933)] = 37622,
  [SMALL_STATE(934)] = 37694,
  [SMALL_STATE(935)] = 37778,
  [SMALL_STATE(936)] = 37866,
  [SMALL_STATE(937)] = 37954,
  [SMALL_STATE(938)] = 38042,
  [SMALL_STATE(939)] = 38130,
  [SMALL_STATE(940)] = 38218,
  [SMALL_STATE(941)] = 38306,
  [SMALL_STATE(942)] = 38394,
  [SMALL_STATE(943)] = 38482,
  [SMALL_STATE(944)] = 38570,
  [SMALL_STATE(945)] = 38658,
  [SMALL_STATE(946)] = 38746,
  [SMALL_STATE(947)] = 38834,
  [SMALL_STATE(948)] = 38902,
  [SMALL_STATE(949)] = 38960,
  [SMALL_STATE(950)] = 39040,
  [SMALL_STATE(951)] = 39122,
  [SMALL_STATE(952)] = 39186,
  [SMALL_STATE(953)] = 39262,
  [SMALL_STATE(954)] = 39340,
  [SMALL_STATE(955)] = 39420,
  [SMALL_STATE(956)] = 39482,
  [SMALL_STATE(957)] = 39540,
  [SMALL_STATE(958)] = 39612,
  [SMALL_STATE(959)] = 39696,
  [SMALL_STATE(960)] = 39784,
  [SMALL_STATE(961)] = 39872,
  [SMALL_STATE(962)] = 39960,
  [SMALL_STATE(963)] = 40048,
  [SMALL_STATE(964)] = 40136,
  [SMALL_STATE(965)] = 40224,
  [SMALL_STATE(966)] = 40274,
  [SMALL_STATE(967)] = 40362,
  [SMALL_STATE(968)] = 40450,
  [SMALL_STATE(969)] = 40537,
  [SMALL_STATE(970)] = 40616,
  [SMALL_STATE(971)] = 40703,
  [SMALL_STATE(972)] = 40790,
  [SMALL_STATE(973)] = 40877,
  [SMALL_STATE(974)] = 40964,
  [SMALL_STATE(975)] = 41051,
  [SMALL_STATE(976)] = 41138,
  [SMALL_STATE(977)] = 41225,
  [SMALL_STATE(978)] = 41312,
  [SMALL_STATE(979)] = 41401,
  [SMALL_STATE(980)] = 41485,
  [SMALL_STATE(981)] = 41569,
  [SMALL_STATE(982)] = 41640,
  [SMALL_STATE(983)] = 41712,
  [SMALL_STATE(984)] = 41784,
  [SMALL_STATE(985)] = 41856,
  [SMALL_STATE(986)] = 41928,
  [SMALL_STATE(987)] = 41986,
  [SMALL_STATE(988)] = 42044,
  [SMALL_STATE(989)] = 42116,
  [SMALL_STATE(990)] = 42182,
  [SMALL_STATE(991)] = 42254,
  [SMALL_STATE(992)] = 42326,
  [SMALL_STATE(993)] = 42379,
  [SMALL_STATE(994)] = 42438,
  [SMALL_STATE(995)] = 42495,
  [SMALL_STATE(996)] = 42554,
  [SMALL_STATE(997)] = 42611,
  [SMALL_STATE(998)] = 42670,
  [SMALL_STATE(999)] = 42727,
  [SMALL_STATE(1000)] = 42786,
  [SMALL_STATE(1001)] = 42843,
  [SMALL_STATE(1002)] = 42900,
  [SMALL_STATE(1003)] = 42959,
  [SMALL_STATE(1004)] = 43018,
  [SMALL_STATE(1005)] = 43075,
  [SMALL_STATE(1006)] = 43127,
  [SMALL_STATE(1007)] = 43179,
  [SMALL_STATE(1008)] = 43231,
  [SMALL_STATE(1009)] = 43283,
  [SMALL_STATE(1010)] = 43337,
  [SMALL_STATE(1011)] = 43389,
  [SMALL_STATE(1012)] = 43441,
  [SMALL_STATE(1013)] = 43492,
  [SMALL_STATE(1014)] = 43545,
  [SMALL_STATE(1015)] = 43606,
  [SMALL_STATE(1016)] = 43657,
  [SMALL_STATE(1017)] = 43706,
  [SMALL_STATE(1018)] = 43752,
  [SMALL_STATE(1019)] = 43798,
  [SMALL_STATE(1020)] = 43854,
  [SMALL_STATE(1021)] = 43900,
  [SMALL_STATE(1022)] = 43946,
  [SMALL_STATE(1023)] = 43994,
  [SMALL_STATE(1024)] = 44038,
  [SMALL_STATE(1025)] = 44086,
  [SMALL_STATE(1026)] = 44142,
  [SMALL_STATE(1027)] = 44186,
  [SMALL_STATE(1028)] = 44230,
  [SMALL_STATE(1029)] = 44269,
  [SMALL_STATE(1030)] = 44308,
  [SMALL_STATE(1031)] = 44347,
  [SMALL_STATE(1032)] = 44386,
  [SMALL_STATE(1033)] = 44425,
  [SMALL_STATE(1034)] = 44464,
  [SMALL_STATE(1035)] = 44499,
  [SMALL_STATE(1036)] = 44540,
  [SMALL_STATE(1037)] = 44593,
  [SMALL_STATE(1038)] = 44621,
  [SMALL_STATE(1039)] = 44671,
  [SMALL_STATE(1040)] = 44699,
  [SMALL_STATE(1041)] = 44727,
  [SMALL_STATE(1042)] = 44761,
  [SMALL_STATE(1043)] = 44791,
  [SMALL_STATE(1044)] = 44819,
  [SMALL_STATE(1045)] = 44847,
  [SMALL_STATE(1046)] = 44875,
  [SMALL_STATE(1047)] = 44903,
  [SMALL_STATE(1048)] = 44933,
  [SMALL_STATE(1049)] = 44961,
  [SMALL_STATE(1050)] = 44989,
  [SMALL_STATE(1051)] = 45017,
  [SMALL_STATE(1052)] = 45045,
  [SMALL_STATE(1053)] = 45073,
  [SMALL_STATE(1054)] = 45101,
  [SMALL_STATE(1055)] = 45129,
  [SMALL_STATE(1056)] = 45157,
  [SMALL_STATE(1057)] = 45185,
  [SMALL_STATE(1058)] = 45213,
  [SMALL_STATE(1059)] = 45241,
  [SMALL_STATE(1060)] = 45269,
  [SMALL_STATE(1061)] = 45297,
  [SMALL_STATE(1062)] = 45325,
  [SMALL_STATE(1063)] = 45353,
  [SMALL_STATE(1064)] = 45381,
  [SMALL_STATE(1065)] = 45409,
  [SMALL_STATE(1066)] = 45437,
  [SMALL_STATE(1067)] = 45465,
  [SMALL_STATE(1068)] = 45515,
  [SMALL_STATE(1069)] = 45543,
  [SMALL_STATE(1070)] = 45571,
  [SMALL_STATE(1071)] = 45599,
  [SMALL_STATE(1072)] = 45628,
  [SMALL_STATE(1073)] = 45671,
  [SMALL_STATE(1074)] = 45698,
  [SMALL_STATE(1075)] = 45737,
  [SMALL_STATE(1076)] = 45764,
  [SMALL_STATE(1077)] = 45807,
  [SMALL_STATE(1078)] = 45852,
  [SMALL_STATE(1079)] = 45879,
  [SMALL_STATE(1080)] = 45922,
  [SMALL_STATE(1081)] = 45949,
  [SMALL_STATE(1082)] = 45976,
  [SMALL_STATE(1083)] = 46017,
  [SMALL_STATE(1084)] = 46058,
  [SMALL_STATE(1085)] = 46101,
  [SMALL_STATE(1086)] = 46128,
  [SMALL_STATE(1087)] = 46164,
  [SMALL_STATE(1088)] = 46200,
  [SMALL_STATE(1089)] = 46236,
  [SMALL_STATE(1090)] = 46278,
  [SMALL_STATE(1091)] = 46314,
  [SMALL_STATE(1092)] = 46350,
  [SMALL_STATE(1093)] = 46386,
  [SMALL_STATE(1094)] = 46422,
  [SMALL_STATE(1095)] = 46448,
  [SMALL_STATE(1096)] = 46484,
  [SMALL_STATE(1097)] = 46520,
  [SMALL_STATE(1098)] = 46546,
  [SMALL_STATE(1099)] = 46582,
  [SMALL_STATE(1100)] = 46618,
  [SMALL_STATE(1101)] = 46644,
  [SMALL_STATE(1102)] = 46684,
  [SMALL_STATE(1103)] = 46720,
  [SMALL_STATE(1104)] = 46756,
  [SMALL_STATE(1105)] = 46792,
  [SMALL_STATE(1106)] = 46834,
  [SMALL_STATE(1107)] = 46860,
  [SMALL_STATE(1108)] = 46896,
  [SMALL_STATE(1109)] = 46932,
  [SMALL_STATE(1110)] = 46968,
  [SMALL_STATE(1111)] = 47008,
  [SMALL_STATE(1112)] = 47044,
  [SMALL_STATE(1113)] = 47080,
  [SMALL_STATE(1114)] = 47116,
  [SMALL_STATE(1115)] = 47152,
  [SMALL_STATE(1116)] = 47188,
  [SMALL_STATE(1117)] = 47224,
  [SMALL_STATE(1118)] = 47260,
  [SMALL_STATE(1119)] = 47286,
  [SMALL_STATE(1120)] = 47312,
  [SMALL_STATE(1121)] = 47345,
  [SMALL_STATE(1122)] = 47378,
  [SMALL_STATE(1123)] = 47411,
  [SMALL_STATE(1124)] = 47444,
  [SMALL_STATE(1125)] = 47477,
  [SMALL_STATE(1126)] = 47510,
  [SMALL_STATE(1127)] = 47543,
  [SMALL_STATE(1128)] = 47576,
  [SMALL_STATE(1129)] = 47609,
  [SMALL_STATE(1130)] = 47642,
  [SMALL_STATE(1131)] = 47675,
  [SMALL_STATE(1132)] = 47708,
  [SMALL_STATE(1133)] = 47741,
  [SMALL_STATE(1134)] = 47774,
  [SMALL_STATE(1135)] = 47807,
  [SMALL_STATE(1136)] = 47840,
  [SMALL_STATE(1137)] = 47873,
  [SMALL_STATE(1138)] = 47906,
  [SMALL_STATE(1139)] = 47928,
  [SMALL_STATE(1140)] = 47950,
  [SMALL_STATE(1141)] = 47972,
  [SMALL_STATE(1142)] = 47994,
  [SMALL_STATE(1143)] = 48016,
  [SMALL_STATE(1144)] = 48038,
  [SMALL_STATE(1145)] = 48075,
  [SMALL_STATE(1146)] = 48111,
  [SMALL_STATE(1147)] = 48141,
  [SMALL_STATE(1148)] = 48177,
  [SMALL_STATE(1149)] = 48207,
  [SMALL_STATE(1150)] = 48237,
  [SMALL_STATE(1151)] = 48267,
  [SMALL_STATE(1152)] = 48303,
  [SMALL_STATE(1153)] = 48339,
  [SMALL_STATE(1154)] = 48369,
  [SMALL_STATE(1155)] = 48399,
  [SMALL_STATE(1156)] = 48429,
  [SMALL_STATE(1157)] = 48459,
  [SMALL_STATE(1158)] = 48492,
  [SMALL_STATE(1159)] = 48525,
  [SMALL_STATE(1160)] = 48558,
  [SMALL_STATE(1161)] = 48585,
  [SMALL_STATE(1162)] = 48618,
  [SMALL_STATE(1163)] = 48651,
  [SMALL_STATE(1164)] = 48684,
  [SMALL_STATE(1165)] = 48717,
  [SMALL_STATE(1166)] = 48750,
  [SMALL_STATE(1167)] = 48780,
  [SMALL_STATE(1168)] = 48810,
  [SMALL_STATE(1169)] = 48836,
  [SMALL_STATE(1170)] = 48866,
  [SMALL_STATE(1171)] = 48892,
  [SMALL_STATE(1172)] = 48920,
  [SMALL_STATE(1173)] = 48948,
  [SMALL_STATE(1174)] = 48978,
  [SMALL_STATE(1175)] = 49008,
  [SMALL_STATE(1176)] = 49038,
  [SMALL_STATE(1177)] = 49068,
  [SMALL_STATE(1178)] = 49098,
  [SMALL_STATE(1179)] = 49128,
  [SMALL_STATE(1180)] = 49158,
  [SMALL_STATE(1181)] = 49188,
  [SMALL_STATE(1182)] = 49218,
  [SMALL_STATE(1183)] = 49248,
  [SMALL_STATE(1184)] = 49278,
  [SMALL_STATE(1185)] = 49308,
  [SMALL_STATE(1186)] = 49338,
  [SMALL_STATE(1187)] = 49368,
  [SMALL_STATE(1188)] = 49398,
  [SMALL_STATE(1189)] = 49428,
  [SMALL_STATE(1190)] = 49458,
  [SMALL_STATE(1191)] = 49488,
  [SMALL_STATE(1192)] = 49513,
  [SMALL_STATE(1193)] = 49538,
  [SMALL_STATE(1194)] = 49565,
  [SMALL_STATE(1195)] = 49590,
  [SMALL_STATE(1196)] = 49615,
  [SMALL_STATE(1197)] = 49640,
  [SMALL_STATE(1198)] = 49665,
  [SMALL_STATE(1199)] = 49692,
  [SMALL_STATE(1200)] = 49717,
  [SMALL_STATE(1201)] = 49742,
  [SMALL_STATE(1202)] = 49767,
  [SMALL_STATE(1203)] = 49789,
  [SMALL_STATE(1204)] = 49811,
  [SMALL_STATE(1205)] = 49829,
  [SMALL_STATE(1206)] = 49851,
  [SMALL_STATE(1207)] = 49865,
  [SMALL_STATE(1208)] = 49879,
  [SMALL_STATE(1209)] = 49899,
  [SMALL_STATE(1210)] = 49921,
  [SMALL_STATE(1211)] = 49945,
  [SMALL_STATE(1212)] = 49969,
  [SMALL_STATE(1213)] = 49993,
  [SMALL_STATE(1214)] = 50015,
  [SMALL_STATE(1215)] = 50039,
  [SMALL_STATE(1216)] = 50059,
  [SMALL_STATE(1217)] = 50079,
  [SMALL_STATE(1218)] = 50103,
  [SMALL_STATE(1219)] = 50117,
  [SMALL_STATE(1220)] = 50131,
  [SMALL_STATE(1221)] = 50155,
  [SMALL_STATE(1222)] = 50179,
  [SMALL_STATE(1223)] = 50193,
  [SMALL_STATE(1224)] = 50215,
  [SMALL_STATE(1225)] = 50233,
  [SMALL_STATE(1226)] = 50248,
  [SMALL_STATE(1227)] = 50265,
  [SMALL_STATE(1228)] = 50278,
  [SMALL_STATE(1229)] = 50299,
  [SMALL_STATE(1230)] = 50318,
  [SMALL_STATE(1231)] = 50333,
  [SMALL_STATE(1232)] = 50354,
  [SMALL_STATE(1233)] = 50375,
  [SMALL_STATE(1234)] = 50394,
  [SMALL_STATE(1235)] = 50407,
  [SMALL_STATE(1236)] = 50420,
  [SMALL_STATE(1237)] = 50439,
  [SMALL_STATE(1238)] = 50452,
  [SMALL_STATE(1239)] = 50465,
  [SMALL_STATE(1240)] = 50484,
  [SMALL_STATE(1241)] = 50503,
  [SMALL_STATE(1242)] = 50516,
  [SMALL_STATE(1243)] = 50535,
  [SMALL_STATE(1244)] = 50554,
  [SMALL_STATE(1245)] = 50573,
  [SMALL_STATE(1246)] = 50594,
  [SMALL_STATE(1247)] = 50613,
  [SMALL_STATE(1248)] = 50626,
  [SMALL_STATE(1249)] = 50639,
  [SMALL_STATE(1250)] = 50652,
  [SMALL_STATE(1251)] = 50671,
  [SMALL_STATE(1252)] = 50684,
  [SMALL_STATE(1253)] = 50703,
  [SMALL_STATE(1254)] = 50717,
  [SMALL_STATE(1255)] = 50731,
  [SMALL_STATE(1256)] = 50745,
  [SMALL_STATE(1257)] = 50759,
  [SMALL_STATE(1258)] = 50773,
  [SMALL_STATE(1259)] = 50787,
  [SMALL_STATE(1260)] = 50801,
  [SMALL_STATE(1261)] = 50821,
  [SMALL_STATE(1262)] = 50841,
  [SMALL_STATE(1263)] = 50861,
  [SMALL_STATE(1264)] = 50875,
  [SMALL_STATE(1265)] = 50889,
  [SMALL_STATE(1266)] = 50903,
  [SMALL_STATE(1267)] = 50917,
  [SMALL_STATE(1268)] = 50937,
  [SMALL_STATE(1269)] = 50957,
  [SMALL_STATE(1270)] = 50971,
  [SMALL_STATE(1271)] = 50985,
  [SMALL_STATE(1272)] = 51005,
  [SMALL_STATE(1273)] = 51019,
  [SMALL_STATE(1274)] = 51039,
  [SMALL_STATE(1275)] = 51051,
  [SMALL_STATE(1276)] = 51065,
  [SMALL_STATE(1277)] = 51079,
  [SMALL_STATE(1278)] = 51097,
  [SMALL_STATE(1279)] = 51111,
  [SMALL_STATE(1280)] = 51131,
  [SMALL_STATE(1281)] = 51143,
  [SMALL_STATE(1282)] = 51157,
  [SMALL_STATE(1283)] = 51171,
  [SMALL_STATE(1284)] = 51187,
  [SMALL_STATE(1285)] = 51207,
  [SMALL_STATE(1286)] = 51223,
  [SMALL_STATE(1287)] = 51243,
  [SMALL_STATE(1288)] = 51257,
  [SMALL_STATE(1289)] = 51277,
  [SMALL_STATE(1290)] = 51297,
  [SMALL_STATE(1291)] = 51311,
  [SMALL_STATE(1292)] = 51325,
  [SMALL_STATE(1293)] = 51339,
  [SMALL_STATE(1294)] = 51359,
  [SMALL_STATE(1295)] = 51373,
  [SMALL_STATE(1296)] = 51393,
  [SMALL_STATE(1297)] = 51413,
  [SMALL_STATE(1298)] = 51429,
  [SMALL_STATE(1299)] = 51449,
  [SMALL_STATE(1300)] = 51463,
  [SMALL_STATE(1301)] = 51477,
  [SMALL_STATE(1302)] = 51491,
  [SMALL_STATE(1303)] = 51505,
  [SMALL_STATE(1304)] = 51525,
  [SMALL_STATE(1305)] = 51541,
  [SMALL_STATE(1306)] = 51555,
  [SMALL_STATE(1307)] = 51575,
  [SMALL_STATE(1308)] = 51589,
  [SMALL_STATE(1309)] = 51603,
  [SMALL_STATE(1310)] = 51617,
  [SMALL_STATE(1311)] = 51637,
  [SMALL_STATE(1312)] = 51653,
  [SMALL_STATE(1313)] = 51667,
  [SMALL_STATE(1314)] = 51687,
  [SMALL_STATE(1315)] = 51701,
  [SMALL_STATE(1316)] = 51721,
  [SMALL_STATE(1317)] = 51733,
  [SMALL_STATE(1318)] = 51747,
  [SMALL_STATE(1319)] = 51767,
  [SMALL_STATE(1320)] = 51781,
  [SMALL_STATE(1321)] = 51800,
  [SMALL_STATE(1322)] = 51815,
  [SMALL_STATE(1323)] = 51832,
  [SMALL_STATE(1324)] = 51849,
  [SMALL_STATE(1325)] = 51862,
  [SMALL_STATE(1326)] = 51877,
  [SMALL_STATE(1327)] = 51894,
  [SMALL_STATE(1328)] = 51911,
  [SMALL_STATE(1329)] = 51928,
  [SMALL_STATE(1330)] = 51945,
  [SMALL_STATE(1331)] = 51962,
  [SMALL_STATE(1332)] = 51979,
  [SMALL_STATE(1333)] = 51996,
  [SMALL_STATE(1334)] = 52013,
  [SMALL_STATE(1335)] = 52030,
  [SMALL_STATE(1336)] = 52045,
  [SMALL_STATE(1337)] = 52062,
  [SMALL_STATE(1338)] = 52077,
  [SMALL_STATE(1339)] = 52092,
  [SMALL_STATE(1340)] = 52109,
  [SMALL_STATE(1341)] = 52120,
  [SMALL_STATE(1342)] = 52135,
  [SMALL_STATE(1343)] = 52146,
  [SMALL_STATE(1344)] = 52163,
  [SMALL_STATE(1345)] = 52178,
  [SMALL_STATE(1346)] = 52195,
  [SMALL_STATE(1347)] = 52206,
  [SMALL_STATE(1348)] = 52223,
  [SMALL_STATE(1349)] = 52242,
  [SMALL_STATE(1350)] = 52259,
  [SMALL_STATE(1351)] = 52276,
  [SMALL_STATE(1352)] = 52293,
  [SMALL_STATE(1353)] = 52310,
  [SMALL_STATE(1354)] = 52327,
  [SMALL_STATE(1355)] = 52342,
  [SMALL_STATE(1356)] = 52361,
  [SMALL_STATE(1357)] = 52380,
  [SMALL_STATE(1358)] = 52397,
  [SMALL_STATE(1359)] = 52412,
  [SMALL_STATE(1360)] = 52427,
  [SMALL_STATE(1361)] = 52444,
  [SMALL_STATE(1362)] = 52461,
  [SMALL_STATE(1363)] = 52478,
  [SMALL_STATE(1364)] = 52495,
  [SMALL_STATE(1365)] = 52512,
  [SMALL_STATE(1366)] = 52527,
  [SMALL_STATE(1367)] = 52538,
  [SMALL_STATE(1368)] = 52555,
  [SMALL_STATE(1369)] = 52572,
  [SMALL_STATE(1370)] = 52589,
  [SMALL_STATE(1371)] = 52606,
  [SMALL_STATE(1372)] = 52621,
  [SMALL_STATE(1373)] = 52638,
  [SMALL_STATE(1374)] = 52655,
  [SMALL_STATE(1375)] = 52672,
  [SMALL_STATE(1376)] = 52683,
  [SMALL_STATE(1377)] = 52700,
  [SMALL_STATE(1378)] = 52717,
  [SMALL_STATE(1379)] = 52734,
  [SMALL_STATE(1380)] = 52749,
  [SMALL_STATE(1381)] = 52764,
  [SMALL_STATE(1382)] = 52781,
  [SMALL_STATE(1383)] = 52798,
  [SMALL_STATE(1384)] = 52815,
  [SMALL_STATE(1385)] = 52834,
  [SMALL_STATE(1386)] = 52849,
  [SMALL_STATE(1387)] = 52866,
  [SMALL_STATE(1388)] = 52883,
  [SMALL_STATE(1389)] = 52900,
  [SMALL_STATE(1390)] = 52911,
  [SMALL_STATE(1391)] = 52926,
  [SMALL_STATE(1392)] = 52941,
  [SMALL_STATE(1393)] = 52958,
  [SMALL_STATE(1394)] = 52975,
  [SMALL_STATE(1395)] = 52990,
  [SMALL_STATE(1396)] = 53007,
  [SMALL_STATE(1397)] = 53024,
  [SMALL_STATE(1398)] = 53037,
  [SMALL_STATE(1399)] = 53056,
  [SMALL_STATE(1400)] = 53070,
  [SMALL_STATE(1401)] = 53084,
  [SMALL_STATE(1402)] = 53098,
  [SMALL_STATE(1403)] = 53112,
  [SMALL_STATE(1404)] = 53126,
  [SMALL_STATE(1405)] = 53138,
  [SMALL_STATE(1406)] = 53152,
  [SMALL_STATE(1407)] = 53166,
  [SMALL_STATE(1408)] = 53180,
  [SMALL_STATE(1409)] = 53190,
  [SMALL_STATE(1410)] = 53204,
  [SMALL_STATE(1411)] = 53216,
  [SMALL_STATE(1412)] = 53230,
  [SMALL_STATE(1413)] = 53244,
  [SMALL_STATE(1414)] = 53258,
  [SMALL_STATE(1415)] = 53272,
  [SMALL_STATE(1416)] = 53282,
  [SMALL_STATE(1417)] = 53296,
  [SMALL_STATE(1418)] = 53310,
  [SMALL_STATE(1419)] = 53324,
  [SMALL_STATE(1420)] = 53338,
  [SMALL_STATE(1421)] = 53352,
  [SMALL_STATE(1422)] = 53366,
  [SMALL_STATE(1423)] = 53380,
  [SMALL_STATE(1424)] = 53392,
  [SMALL_STATE(1425)] = 53406,
  [SMALL_STATE(1426)] = 53416,
  [SMALL_STATE(1427)] = 53430,
  [SMALL_STATE(1428)] = 53444,
  [SMALL_STATE(1429)] = 53458,
  [SMALL_STATE(1430)] = 53472,
  [SMALL_STATE(1431)] = 53486,
  [SMALL_STATE(1432)] = 53498,
  [SMALL_STATE(1433)] = 53512,
  [SMALL_STATE(1434)] = 53526,
  [SMALL_STATE(1435)] = 53538,
  [SMALL_STATE(1436)] = 53552,
  [SMALL_STATE(1437)] = 53566,
  [SMALL_STATE(1438)] = 53580,
  [SMALL_STATE(1439)] = 53594,
  [SMALL_STATE(1440)] = 53604,
  [SMALL_STATE(1441)] = 53618,
  [SMALL_STATE(1442)] = 53632,
  [SMALL_STATE(1443)] = 53644,
  [SMALL_STATE(1444)] = 53658,
  [SMALL_STATE(1445)] = 53672,
  [SMALL_STATE(1446)] = 53686,
  [SMALL_STATE(1447)] = 53698,
  [SMALL_STATE(1448)] = 53708,
  [SMALL_STATE(1449)] = 53722,
  [SMALL_STATE(1450)] = 53736,
  [SMALL_STATE(1451)] = 53750,
  [SMALL_STATE(1452)] = 53764,
  [SMALL_STATE(1453)] = 53778,
  [SMALL_STATE(1454)] = 53788,
  [SMALL_STATE(1455)] = 53802,
  [SMALL_STATE(1456)] = 53816,
  [SMALL_STATE(1457)] = 53828,
  [SMALL_STATE(1458)] = 53842,
  [SMALL_STATE(1459)] = 53856,
  [SMALL_STATE(1460)] = 53870,
  [SMALL_STATE(1461)] = 53884,
  [SMALL_STATE(1462)] = 53898,
  [SMALL_STATE(1463)] = 53912,
  [SMALL_STATE(1464)] = 53926,
  [SMALL_STATE(1465)] = 53940,
  [SMALL_STATE(1466)] = 53954,
  [SMALL_STATE(1467)] = 53968,
  [SMALL_STATE(1468)] = 53982,
  [SMALL_STATE(1469)] = 53994,
  [SMALL_STATE(1470)] = 54006,
  [SMALL_STATE(1471)] = 54020,
  [SMALL_STATE(1472)] = 54034,
  [SMALL_STATE(1473)] = 54046,
  [SMALL_STATE(1474)] = 54060,
  [SMALL_STATE(1475)] = 54072,
  [SMALL_STATE(1476)] = 54086,
  [SMALL_STATE(1477)] = 54100,
  [SMALL_STATE(1478)] = 54114,
  [SMALL_STATE(1479)] = 54126,
  [SMALL_STATE(1480)] = 54140,
  [SMALL_STATE(1481)] = 54154,
  [SMALL_STATE(1482)] = 54166,
  [SMALL_STATE(1483)] = 54176,
  [SMALL_STATE(1484)] = 54190,
  [SMALL_STATE(1485)] = 54204,
  [SMALL_STATE(1486)] = 54218,
  [SMALL_STATE(1487)] = 54232,
  [SMALL_STATE(1488)] = 54246,
  [SMALL_STATE(1489)] = 54256,
  [SMALL_STATE(1490)] = 54270,
  [SMALL_STATE(1491)] = 54284,
  [SMALL_STATE(1492)] = 54298,
  [SMALL_STATE(1493)] = 54312,
  [SMALL_STATE(1494)] = 54324,
  [SMALL_STATE(1495)] = 54338,
  [SMALL_STATE(1496)] = 54350,
  [SMALL_STATE(1497)] = 54364,
  [SMALL_STATE(1498)] = 54374,
  [SMALL_STATE(1499)] = 54388,
  [SMALL_STATE(1500)] = 54402,
  [SMALL_STATE(1501)] = 54411,
  [SMALL_STATE(1502)] = 54422,
  [SMALL_STATE(1503)] = 54433,
  [SMALL_STATE(1504)] = 54444,
  [SMALL_STATE(1505)] = 54455,
  [SMALL_STATE(1506)] = 54464,
  [SMALL_STATE(1507)] = 54475,
  [SMALL_STATE(1508)] = 54486,
  [SMALL_STATE(1509)] = 54497,
  [SMALL_STATE(1510)] = 54508,
  [SMALL_STATE(1511)] = 54519,
  [SMALL_STATE(1512)] = 54530,
  [SMALL_STATE(1513)] = 54541,
  [SMALL_STATE(1514)] = 54550,
  [SMALL_STATE(1515)] = 54559,
  [SMALL_STATE(1516)] = 54570,
  [SMALL_STATE(1517)] = 54579,
  [SMALL_STATE(1518)] = 54590,
  [SMALL_STATE(1519)] = 54601,
  [SMALL_STATE(1520)] = 54612,
  [SMALL_STATE(1521)] = 54623,
  [SMALL_STATE(1522)] = 54632,
  [SMALL_STATE(1523)] = 54641,
  [SMALL_STATE(1524)] = 54652,
  [SMALL_STATE(1525)] = 54663,
  [SMALL_STATE(1526)] = 54674,
  [SMALL_STATE(1527)] = 54683,
  [SMALL_STATE(1528)] = 54694,
  [SMALL_STATE(1529)] = 54705,
  [SMALL_STATE(1530)] = 54716,
  [SMALL_STATE(1531)] = 54727,
  [SMALL_STATE(1532)] = 54738,
  [SMALL_STATE(1533)] = 54749,
  [SMALL_STATE(1534)] = 54760,
  [SMALL_STATE(1535)] = 54771,
  [SMALL_STATE(1536)] = 54782,
  [SMALL_STATE(1537)] = 54793,
  [SMALL_STATE(1538)] = 54804,
  [SMALL_STATE(1539)] = 54813,
  [SMALL_STATE(1540)] = 54824,
  [SMALL_STATE(1541)] = 54835,
  [SMALL_STATE(1542)] = 54844,
  [SMALL_STATE(1543)] = 54855,
  [SMALL_STATE(1544)] = 54864,
  [SMALL_STATE(1545)] = 54875,
  [SMALL_STATE(1546)] = 54886,
  [SMALL_STATE(1547)] = 54897,
  [SMALL_STATE(1548)] = 54908,
  [SMALL_STATE(1549)] = 54917,
  [SMALL_STATE(1550)] = 54928,
  [SMALL_STATE(1551)] = 54939,
  [SMALL_STATE(1552)] = 54950,
  [SMALL_STATE(1553)] = 54961,
  [SMALL_STATE(1554)] = 54972,
  [SMALL_STATE(1555)] = 54983,
  [SMALL_STATE(1556)] = 54994,
  [SMALL_STATE(1557)] = 55005,
  [SMALL_STATE(1558)] = 55016,
  [SMALL_STATE(1559)] = 55027,
  [SMALL_STATE(1560)] = 55038,
  [SMALL_STATE(1561)] = 55049,
  [SMALL_STATE(1562)] = 55060,
  [SMALL_STATE(1563)] = 55071,
  [SMALL_STATE(1564)] = 55082,
  [SMALL_STATE(1565)] = 55093,
  [SMALL_STATE(1566)] = 55104,
  [SMALL_STATE(1567)] = 55115,
  [SMALL_STATE(1568)] = 55126,
  [SMALL_STATE(1569)] = 55137,
  [SMALL_STATE(1570)] = 55148,
  [SMALL_STATE(1571)] = 55159,
  [SMALL_STATE(1572)] = 55170,
  [SMALL_STATE(1573)] = 55181,
  [SMALL_STATE(1574)] = 55192,
  [SMALL_STATE(1575)] = 55203,
  [SMALL_STATE(1576)] = 55214,
  [SMALL_STATE(1577)] = 55225,
  [SMALL_STATE(1578)] = 55236,
  [SMALL_STATE(1579)] = 55247,
  [SMALL_STATE(1580)] = 55258,
  [SMALL_STATE(1581)] = 55269,
  [SMALL_STATE(1582)] = 55280,
  [SMALL_STATE(1583)] = 55291,
  [SMALL_STATE(1584)] = 55302,
  [SMALL_STATE(1585)] = 55313,
  [SMALL_STATE(1586)] = 55324,
  [SMALL_STATE(1587)] = 55335,
  [SMALL_STATE(1588)] = 55346,
  [SMALL_STATE(1589)] = 55357,
  [SMALL_STATE(1590)] = 55368,
  [SMALL_STATE(1591)] = 55379,
  [SMALL_STATE(1592)] = 55388,
  [SMALL_STATE(1593)] = 55397,
  [SMALL_STATE(1594)] = 55408,
  [SMALL_STATE(1595)] = 55419,
  [SMALL_STATE(1596)] = 55430,
  [SMALL_STATE(1597)] = 55441,
  [SMALL_STATE(1598)] = 55452,
  [SMALL_STATE(1599)] = 55463,
  [SMALL_STATE(1600)] = 55474,
  [SMALL_STATE(1601)] = 55485,
  [SMALL_STATE(1602)] = 55496,
  [SMALL_STATE(1603)] = 55507,
  [SMALL_STATE(1604)] = 55518,
  [SMALL_STATE(1605)] = 55529,
  [SMALL_STATE(1606)] = 55540,
  [SMALL_STATE(1607)] = 55551,
  [SMALL_STATE(1608)] = 55562,
  [SMALL_STATE(1609)] = 55573,
  [SMALL_STATE(1610)] = 55584,
  [SMALL_STATE(1611)] = 55595,
  [SMALL_STATE(1612)] = 55606,
  [SMALL_STATE(1613)] = 55617,
  [SMALL_STATE(1614)] = 55628,
  [SMALL_STATE(1615)] = 55639,
  [SMALL_STATE(1616)] = 55650,
  [SMALL_STATE(1617)] = 55661,
  [SMALL_STATE(1618)] = 55672,
  [SMALL_STATE(1619)] = 55683,
  [SMALL_STATE(1620)] = 55694,
  [SMALL_STATE(1621)] = 55703,
  [SMALL_STATE(1622)] = 55714,
  [SMALL_STATE(1623)] = 55725,
  [SMALL_STATE(1624)] = 55736,
  [SMALL_STATE(1625)] = 55747,
  [SMALL_STATE(1626)] = 55758,
  [SMALL_STATE(1627)] = 55769,
  [SMALL_STATE(1628)] = 55780,
  [SMALL_STATE(1629)] = 55791,
  [SMALL_STATE(1630)] = 55802,
  [SMALL_STATE(1631)] = 55813,
  [SMALL_STATE(1632)] = 55824,
  [SMALL_STATE(1633)] = 55835,
  [SMALL_STATE(1634)] = 55846,
  [SMALL_STATE(1635)] = 55857,
  [SMALL_STATE(1636)] = 55868,
  [SMALL_STATE(1637)] = 55879,
  [SMALL_STATE(1638)] = 55890,
  [SMALL_STATE(1639)] = 55901,
  [SMALL_STATE(1640)] = 55912,
  [SMALL_STATE(1641)] = 55923,
  [SMALL_STATE(1642)] = 55934,
  [SMALL_STATE(1643)] = 55945,
  [SMALL_STATE(1644)] = 55956,
  [SMALL_STATE(1645)] = 55967,
  [SMALL_STATE(1646)] = 55978,
  [SMALL_STATE(1647)] = 55989,
  [SMALL_STATE(1648)] = 56000,
  [SMALL_STATE(1649)] = 56009,
  [SMALL_STATE(1650)] = 56018,
  [SMALL_STATE(1651)] = 56029,
  [SMALL_STATE(1652)] = 56040,
  [SMALL_STATE(1653)] = 56051,
  [SMALL_STATE(1654)] = 56062,
  [SMALL_STATE(1655)] = 56073,
  [SMALL_STATE(1656)] = 56084,
  [SMALL_STATE(1657)] = 56095,
  [SMALL_STATE(1658)] = 56106,
  [SMALL_STATE(1659)] = 56117,
  [SMALL_STATE(1660)] = 56126,
  [SMALL_STATE(1661)] = 56135,
  [SMALL_STATE(1662)] = 56144,
  [SMALL_STATE(1663)] = 56153,
  [SMALL_STATE(1664)] = 56164,
  [SMALL_STATE(1665)] = 56175,
  [SMALL_STATE(1666)] = 56186,
  [SMALL_STATE(1667)] = 56197,
  [SMALL_STATE(1668)] = 56208,
  [SMALL_STATE(1669)] = 56219,
  [SMALL_STATE(1670)] = 56228,
  [SMALL_STATE(1671)] = 56239,
  [SMALL_STATE(1672)] = 56250,
  [SMALL_STATE(1673)] = 56261,
  [SMALL_STATE(1674)] = 56270,
  [SMALL_STATE(1675)] = 56281,
  [SMALL_STATE(1676)] = 56292,
  [SMALL_STATE(1677)] = 56303,
  [SMALL_STATE(1678)] = 56314,
  [SMALL_STATE(1679)] = 56323,
  [SMALL_STATE(1680)] = 56334,
  [SMALL_STATE(1681)] = 56345,
  [SMALL_STATE(1682)] = 56356,
  [SMALL_STATE(1683)] = 56367,
  [SMALL_STATE(1684)] = 56378,
  [SMALL_STATE(1685)] = 56389,
  [SMALL_STATE(1686)] = 56398,
  [SMALL_STATE(1687)] = 56409,
  [SMALL_STATE(1688)] = 56420,
  [SMALL_STATE(1689)] = 56431,
  [SMALL_STATE(1690)] = 56440,
  [SMALL_STATE(1691)] = 56451,
  [SMALL_STATE(1692)] = 56462,
  [SMALL_STATE(1693)] = 56473,
  [SMALL_STATE(1694)] = 56484,
  [SMALL_STATE(1695)] = 56495,
  [SMALL_STATE(1696)] = 56504,
  [SMALL_STATE(1697)] = 56515,
  [SMALL_STATE(1698)] = 56526,
  [SMALL_STATE(1699)] = 56537,
  [SMALL_STATE(1700)] = 56546,
  [SMALL_STATE(1701)] = 56557,
  [SMALL_STATE(1702)] = 56568,
  [SMALL_STATE(1703)] = 56577,
  [SMALL_STATE(1704)] = 56588,
  [SMALL_STATE(1705)] = 56599,
  [SMALL_STATE(1706)] = 56610,
  [SMALL_STATE(1707)] = 56621,
  [SMALL_STATE(1708)] = 56632,
  [SMALL_STATE(1709)] = 56643,
  [SMALL_STATE(1710)] = 56654,
  [SMALL_STATE(1711)] = 56665,
  [SMALL_STATE(1712)] = 56676,
  [SMALL_STATE(1713)] = 56687,
  [SMALL_STATE(1714)] = 56698,
  [SMALL_STATE(1715)] = 56709,
  [SMALL_STATE(1716)] = 56720,
  [SMALL_STATE(1717)] = 56731,
  [SMALL_STATE(1718)] = 56742,
  [SMALL_STATE(1719)] = 56753,
  [SMALL_STATE(1720)] = 56764,
  [SMALL_STATE(1721)] = 56775,
  [SMALL_STATE(1722)] = 56786,
  [SMALL_STATE(1723)] = 56795,
  [SMALL_STATE(1724)] = 56806,
  [SMALL_STATE(1725)] = 56817,
  [SMALL_STATE(1726)] = 56828,
  [SMALL_STATE(1727)] = 56839,
  [SMALL_STATE(1728)] = 56848,
  [SMALL_STATE(1729)] = 56859,
  [SMALL_STATE(1730)] = 56870,
  [SMALL_STATE(1731)] = 56881,
  [SMALL_STATE(1732)] = 56890,
  [SMALL_STATE(1733)] = 56899,
  [SMALL_STATE(1734)] = 56910,
  [SMALL_STATE(1735)] = 56919,
  [SMALL_STATE(1736)] = 56930,
  [SMALL_STATE(1737)] = 56941,
  [SMALL_STATE(1738)] = 56950,
  [SMALL_STATE(1739)] = 56961,
  [SMALL_STATE(1740)] = 56972,
  [SMALL_STATE(1741)] = 56983,
  [SMALL_STATE(1742)] = 56994,
  [SMALL_STATE(1743)] = 57005,
  [SMALL_STATE(1744)] = 57016,
  [SMALL_STATE(1745)] = 57027,
  [SMALL_STATE(1746)] = 57038,
  [SMALL_STATE(1747)] = 57049,
  [SMALL_STATE(1748)] = 57058,
  [SMALL_STATE(1749)] = 57066,
  [SMALL_STATE(1750)] = 57074,
  [SMALL_STATE(1751)] = 57082,
  [SMALL_STATE(1752)] = 57090,
  [SMALL_STATE(1753)] = 57098,
  [SMALL_STATE(1754)] = 57106,
  [SMALL_STATE(1755)] = 57114,
  [SMALL_STATE(1756)] = 57122,
  [SMALL_STATE(1757)] = 57130,
  [SMALL_STATE(1758)] = 57138,
  [SMALL_STATE(1759)] = 57146,
  [SMALL_STATE(1760)] = 57154,
  [SMALL_STATE(1761)] = 57162,
  [SMALL_STATE(1762)] = 57170,
  [SMALL_STATE(1763)] = 57178,
  [SMALL_STATE(1764)] = 57186,
  [SMALL_STATE(1765)] = 57194,
  [SMALL_STATE(1766)] = 57202,
  [SMALL_STATE(1767)] = 57210,
  [SMALL_STATE(1768)] = 57218,
  [SMALL_STATE(1769)] = 57226,
  [SMALL_STATE(1770)] = 57236,
  [SMALL_STATE(1771)] = 57246,
  [SMALL_STATE(1772)] = 57254,
  [SMALL_STATE(1773)] = 57262,
  [SMALL_STATE(1774)] = 57270,
  [SMALL_STATE(1775)] = 57278,
  [SMALL_STATE(1776)] = 57286,
  [SMALL_STATE(1777)] = 57294,
  [SMALL_STATE(1778)] = 57302,
  [SMALL_STATE(1779)] = 57310,
  [SMALL_STATE(1780)] = 57318,
  [SMALL_STATE(1781)] = 57326,
  [SMALL_STATE(1782)] = 57334,
  [SMALL_STATE(1783)] = 57342,
  [SMALL_STATE(1784)] = 57350,
  [SMALL_STATE(1785)] = 57358,
  [SMALL_STATE(1786)] = 57368,
  [SMALL_STATE(1787)] = 57376,
  [SMALL_STATE(1788)] = 57386,
  [SMALL_STATE(1789)] = 57394,
  [SMALL_STATE(1790)] = 57402,
  [SMALL_STATE(1791)] = 57410,
  [SMALL_STATE(1792)] = 57418,
  [SMALL_STATE(1793)] = 57426,
  [SMALL_STATE(1794)] = 57434,
  [SMALL_STATE(1795)] = 57442,
  [SMALL_STATE(1796)] = 57450,
  [SMALL_STATE(1797)] = 57458,
  [SMALL_STATE(1798)] = 57466,
  [SMALL_STATE(1799)] = 57474,
  [SMALL_STATE(1800)] = 57482,
  [SMALL_STATE(1801)] = 57490,
  [SMALL_STATE(1802)] = 57498,
  [SMALL_STATE(1803)] = 57506,
  [SMALL_STATE(1804)] = 57514,
  [SMALL_STATE(1805)] = 57522,
  [SMALL_STATE(1806)] = 57530,
  [SMALL_STATE(1807)] = 57538,
  [SMALL_STATE(1808)] = 57546,
  [SMALL_STATE(1809)] = 57556,
  [SMALL_STATE(1810)] = 57564,
  [SMALL_STATE(1811)] = 57572,
  [SMALL_STATE(1812)] = 57580,
  [SMALL_STATE(1813)] = 57588,
  [SMALL_STATE(1814)] = 57596,
  [SMALL_STATE(1815)] = 57604,
  [SMALL_STATE(1816)] = 57612,
  [SMALL_STATE(1817)] = 57620,
  [SMALL_STATE(1818)] = 57628,
  [SMALL_STATE(1819)] = 57636,
  [SMALL_STATE(1820)] = 57644,
  [SMALL_STATE(1821)] = 57652,
  [SMALL_STATE(1822)] = 57660,
  [SMALL_STATE(1823)] = 57670,
  [SMALL_STATE(1824)] = 57678,
  [SMALL_STATE(1825)] = 57686,
  [SMALL_STATE(1826)] = 57694,
  [SMALL_STATE(1827)] = 57702,
  [SMALL_STATE(1828)] = 57710,
  [SMALL_STATE(1829)] = 57718,
  [SMALL_STATE(1830)] = 57728,
  [SMALL_STATE(1831)] = 57736,
  [SMALL_STATE(1832)] = 57744,
  [SMALL_STATE(1833)] = 57754,
  [SMALL_STATE(1834)] = 57762,
  [SMALL_STATE(1835)] = 57770,
  [SMALL_STATE(1836)] = 57778,
  [SMALL_STATE(1837)] = 57786,
  [SMALL_STATE(1838)] = 57794,
  [SMALL_STATE(1839)] = 57802,
  [SMALL_STATE(1840)] = 57810,
  [SMALL_STATE(1841)] = 57818,
  [SMALL_STATE(1842)] = 57826,
  [SMALL_STATE(1843)] = 57834,
  [SMALL_STATE(1844)] = 57842,
  [SMALL_STATE(1845)] = 57850,
  [SMALL_STATE(1846)] = 57858,
  [SMALL_STATE(1847)] = 57866,
  [SMALL_STATE(1848)] = 57874,
  [SMALL_STATE(1849)] = 57882,
  [SMALL_STATE(1850)] = 57890,
  [SMALL_STATE(1851)] = 57898,
  [SMALL_STATE(1852)] = 57906,
  [SMALL_STATE(1853)] = 57914,
  [SMALL_STATE(1854)] = 57922,
  [SMALL_STATE(1855)] = 57930,
  [SMALL_STATE(1856)] = 57938,
  [SMALL_STATE(1857)] = 57946,
  [SMALL_STATE(1858)] = 57954,
  [SMALL_STATE(1859)] = 57962,
  [SMALL_STATE(1860)] = 57970,
  [SMALL_STATE(1861)] = 57978,
  [SMALL_STATE(1862)] = 57986,
  [SMALL_STATE(1863)] = 57994,
  [SMALL_STATE(1864)] = 58002,
  [SMALL_STATE(1865)] = 58010,
  [SMALL_STATE(1866)] = 58018,
  [SMALL_STATE(1867)] = 58026,
  [SMALL_STATE(1868)] = 58034,
  [SMALL_STATE(1869)] = 58042,
};

static const TSParseActionEntry ts_parse_actions[] = {
  [0] = {.entry = {.count = 0, .reusable = false}},
  [1] = {.entry = {.count = 1, .reusable = false}}, RECOVER(),
  [3] = {.entry = {.count = 1, .reusable = false}}, SHIFT_EXTRA(),
  [5] = {.entry = {.count = 1, .reusable = true}}, SHIFT_EXTRA(),
  [7] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_program, 0, 0, 0),
  [9] = {.entry = {.count = 1, .reusable = false}}, SHIFT(531),
  [11] = {.entry = {.count = 1, .reusable = true}}, SHIFT(17),
  [13] = {.entry = {.count = 1, .reusable = false}}, SHIFT(166),
  [15] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2),
  [17] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1144),
  [19] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1657),
  [21] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1197),
  [23] = {.entry = {.count = 1, .reusable = false}}, SHIFT(396),
  [25] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1199),
  [27] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1196),
  [29] = {.entry = {.count = 1, .reusable = false}}, SHIFT(31),
  [31] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1668),
  [33] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1677),
  [35] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1498),
  [37] = {.entry = {.count = 1, .reusable = true}}, SHIFT(119),
  [39] = {.entry = {.count = 1, .reusable = true}}, SHIFT(418),
  [41] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1506),
  [43] = {.entry = {.count = 1, .reusable = false}}, SHIFT(41),
  [45] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1708),
  [47] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1456),
  [49] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1446),
  [51] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1727),
  [53] = {.entry = {.count = 1, .reusable = false}}, SHIFT(142),
  [55] = {.entry = {.count = 1, .reusable = false}}, SHIFT(174),
  [57] = {.entry = {.count = 1, .reusable = false}}, SHIFT(81),
  [59] = {.entry = {.count = 1, .reusable = true}}, SHIFT(111),
  [61] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1318),
  [63] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1339),
  [65] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1364),
  [67] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1268),
  [69] = {.entry = {.count = 1, .reusable = false}}, SHIFT(371),
  [71] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1336),
  [73] = {.entry = {.count = 1, .reusable = false}}, SHIFT(192),
  [75] = {.entry = {.count = 1, .reusable = false}}, SHIFT(224),
  [77] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1769),
  [79] = {.entry = {.count = 1, .reusable = true}}, SHIFT(224),
  [81] = {.entry = {.count = 1, .reusable = true}}, SHIFT(229),
  [83] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1229),
  [85] = {.entry = {.count = 1, .reusable = true}}, SHIFT(740),
  [87] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1830),
  [89] = {.entry = {.count = 1, .reusable = false}}, SHIFT(740),
  [91] = {.entry = {.count = 1, .reusable = false}}, SHIFT(771),
  [93] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1470),
  [95] = {.entry = {.count = 1, .reusable = false}}, SHIFT(535),
  [97] = {.entry = {.count = 1, .reusable = false}}, SHIFT(492),
  [99] = {.entry = {.count = 1, .reusable = false}}, SHIFT(134),
  [101] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1125),
  [103] = {.entry = {.count = 1, .reusable = true}}, SHIFT(969),
  [105] = {.entry = {.count = 1, .reusable = true}}, SHIFT(92),
  [107] = {.entry = {.count = 1, .reusable = false}}, SHIFT(370),
  [109] = {.entry = {.count = 1, .reusable = false}}, SHIFT(7),
  [111] = {.entry = {.count = 1, .reusable = true}}, SHIFT(107),
  [113] = {.entry = {.count = 1, .reusable = false}}, SHIFT(358),
  [115] = {.entry = {.count = 1, .reusable = true}}, SHIFT(265),
  [117] = {.entry = {.count = 1, .reusable = true}}, SHIFT(815),
  [119] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1431),
  [121] = {.entry = {.count = 1, .reusable = false}}, SHIFT(363),
  [123] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1110),
  [125] = {.entry = {.count = 1, .reusable = false}}, SHIFT(360),
  [127] = {.entry = {.count = 1, .reusable = true}}, SHIFT(693),
  [129] = {.entry = {.count = 1, .reusable = true}}, SHIFT(684),
  [131] = {.entry = {.count = 1, .reusable = false}}, SHIFT(499),
  [133] = {.entry = {.count = 1, .reusable = false}}, SHIFT(137),
  [135] = {.entry = {.count = 1, .reusable = true}}, SHIFT(582),
  [137] = {.entry = {.count = 1, .reusable = false}}, SHIFT(380),
  [139] = {.entry = {.count = 1, .reusable = false}}, SHIFT(8),
  [141] = {.entry = {.count = 1, .reusable = false}}, SHIFT(359),
  [143] = {.entry = {.count = 1, .reusable = false}}, SHIFT(364),
  [145] = {.entry = {.count = 1, .reusable = false}}, SHIFT(365),
  [147] = {.entry = {.count = 1, .reusable = false}}, SHIFT(490),
  [149] = {.entry = {.count = 1, .reusable = false}}, SHIFT(133),
  [151] = {.entry = {.count = 1, .reusable = false}}, SHIFT(368),
  [153] = {.entry = {.count = 1, .reusable = false}}, SHIFT(10),
  [155] = {.entry = {.count = 1, .reusable = false}}, SHIFT(357),
  [157] = {.entry = {.count = 1, .reusable = false}}, SHIFT(361),
  [159] = {.entry = {.count = 1, .reusable = false}}, SHIFT(362),
  [161] = {.entry = {.count = 1, .reusable = false}}, SHIFT(543),
  [163] = {.entry = {.count = 1, .reusable = false}}, SHIFT(537),
  [165] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_primary_expression, 1, 0, 1),
  [167] = {.entry = {.count = 1, .reusable = true}}, SHIFT(828),
  [169] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_primary_expression, 1, 0, 1),
  [171] = {.entry = {.count = 1, .reusable = true}}, SHIFT(685),
  [173] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1404),
  [175] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1191),
  [177] = {.entry = {.count = 1, .reusable = false}}, SHIFT(37),
  [179] = {.entry = {.count = 3, .reusable = true}}, REDUCE(sym_primary_expression, 1, 0, 1), REDUCE(sym__property_name, 1, 0, 4), SHIFT(119),
  [183] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__property_name, 1, 0, 4), SHIFT(68),
  [186] = {.entry = {.count = 1, .reusable = false}}, SHIFT(268),
  [188] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_primary_expression, 1, 0, 1), SHIFT(111),
  [191] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_primary_expression, 1, 0, 1), SHIFT(1318),
  [194] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1284),
  [196] = {.entry = {.count = 1, .reusable = false}}, SHIFT(375),
  [198] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1376),
  [200] = {.entry = {.count = 1, .reusable = true}}, SHIFT(159),
  [202] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__augmented_assignment_lhs, 1, 0, 1),
  [204] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_primary_expression, 1, 0, 1), SHIFT(224),
  [207] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_primary_expression, 1, 0, 1), SHIFT(1769),
  [210] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_primary_expression, 1, 0, 1), SHIFT(229),
  [213] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_primary_expression, 1, 0, 1), SHIFT(1229),
  [216] = {.entry = {.count = 1, .reusable = true}}, SHIFT(652),
  [218] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0),
  [220] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(531),
  [223] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(166),
  [226] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0),
  [228] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(2),
  [231] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(1144),
  [234] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(1657),
  [237] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(1197),
  [240] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(396),
  [243] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(1199),
  [246] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(1196),
  [249] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(31),
  [252] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(1668),
  [255] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(1677),
  [258] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(1498),
  [261] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(119),
  [264] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(418),
  [267] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(1506),
  [270] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(41),
  [273] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(1708),
  [276] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(1456),
  [279] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(1446),
  [282] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(1727),
  [285] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(142),
  [288] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(174),
  [291] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(81),
  [294] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(111),
  [297] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(1318),
  [300] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(1339),
  [303] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(1364),
  [306] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(1268),
  [309] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(371),
  [312] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(1336),
  [315] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(192),
  [318] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(224),
  [321] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(1769),
  [324] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(224),
  [327] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(229),
  [330] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(1229),
  [333] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(740),
  [336] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(1830),
  [339] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(740),
  [342] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(771),
  [345] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(1470),
  [348] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_program_repeat1, 2, 0, 0), SHIFT_REPEAT(535),
  [351] = {.entry = {.count = 1, .reusable = true}}, SHIFT(632),
  [353] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_switch_default, 2, 0, 0),
  [355] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_switch_default, 2, 0, 0),
  [357] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_switch_default, 3, 0, 38),
  [359] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_switch_default, 3, 0, 38),
  [361] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_switch_case, 4, 0, 101),
  [363] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_switch_case, 4, 0, 101),
  [365] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_switch_case, 3, 0, 59),
  [367] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_switch_case, 3, 0, 59),
  [369] = {.entry = {.count = 1, .reusable = false}}, SHIFT(501),
  [371] = {.entry = {.count = 1, .reusable = false}}, SHIFT(502),
  [373] = {.entry = {.count = 1, .reusable = true}}, SHIFT(865),
  [375] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1478),
  [377] = {.entry = {.count = 1, .reusable = false}}, SHIFT(15),
  [379] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_primary_expression, 1, 0, 1), SHIFT(120),
  [382] = {.entry = {.count = 1, .reusable = false}}, SHIFT(80),
  [384] = {.entry = {.count = 1, .reusable = false}}, SHIFT(246),
  [386] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_primary_expression, 1, 0, 1), SHIFT(105),
  [389] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_primary_expression, 1, 0, 1), SHIFT(1260),
  [392] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1357),
  [394] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1369),
  [396] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1286),
  [398] = {.entry = {.count = 1, .reusable = false}}, SHIFT(366),
  [400] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1387),
  [402] = {.entry = {.count = 1, .reusable = true}}, SHIFT(164),
  [404] = {.entry = {.count = 1, .reusable = false}}, SHIFT(210),
  [406] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_primary_expression, 1, 0, 1), SHIFT(244),
  [409] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_primary_expression, 1, 0, 1), SHIFT(1832),
  [412] = {.entry = {.count = 1, .reusable = false}}, SHIFT(244),
  [414] = {.entry = {.count = 1, .reusable = true}}, SHIFT(244),
  [416] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_primary_expression, 1, 0, 1), SHIFT(245),
  [419] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_primary_expression, 1, 0, 1), SHIFT(1242),
  [422] = {.entry = {.count = 1, .reusable = true}}, SHIFT(646),
  [424] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1756),
  [426] = {.entry = {.count = 1, .reusable = false}}, SHIFT(646),
  [428] = {.entry = {.count = 1, .reusable = false}}, SHIFT(659),
  [430] = {.entry = {.count = 1, .reusable = false}}, SHIFT(576),
  [432] = {.entry = {.count = 1, .reusable = false}}, SHIFT(571),
  [434] = {.entry = {.count = 1, .reusable = true}}, SHIFT(856),
  [436] = {.entry = {.count = 1, .reusable = false}}, SHIFT(76),
  [438] = {.entry = {.count = 1, .reusable = false}}, SHIFT(84),
  [440] = {.entry = {.count = 1, .reusable = false}}, SHIFT(317),
  [442] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_primary_expression, 1, 0, 1), SHIFT(110),
  [445] = {.entry = {.count = 1, .reusable = false}}, SHIFT(410),
  [447] = {.entry = {.count = 1, .reusable = false}}, SHIFT(157),
  [449] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_primary_expression, 1, 0, 1), SHIFT(315),
  [452] = {.entry = {.count = 1, .reusable = false}}, SHIFT(315),
  [454] = {.entry = {.count = 1, .reusable = true}}, SHIFT(315),
  [456] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_primary_expression, 1, 0, 1), SHIFT(316),
  [459] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1809),
  [461] = {.entry = {.count = 1, .reusable = false}}, SHIFT(590),
  [463] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_program, 1, 0, 0),
  [465] = {.entry = {.count = 1, .reusable = true}}, SHIFT(369),
  [467] = {.entry = {.count = 1, .reusable = true}}, SHIFT(374),
  [469] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_program, 2, 0, 0),
  [471] = {.entry = {.count = 1, .reusable = true}}, SHIFT(88),
  [473] = {.entry = {.count = 1, .reusable = true}}, SHIFT(91),
  [475] = {.entry = {.count = 1, .reusable = true}}, SHIFT(586),
  [477] = {.entry = {.count = 1, .reusable = true}}, SHIFT(584),
  [479] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1042),
  [481] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1047),
  [483] = {.entry = {.count = 1, .reusable = true}}, SHIFT(394),
  [485] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_primary_expression, 1, 0, 1), SHIFT(119),
  [488] = {.entry = {.count = 1, .reusable = true}}, SHIFT(52),
  [490] = {.entry = {.count = 1, .reusable = false}}, SHIFT(225),
  [492] = {.entry = {.count = 1, .reusable = true}}, SHIFT(416),
  [494] = {.entry = {.count = 1, .reusable = true}}, SHIFT(68),
  [496] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_rest_pattern, 2, 0, 20),
  [498] = {.entry = {.count = 1, .reusable = true}}, SHIFT(194),
  [500] = {.entry = {.count = 1, .reusable = true}}, SHIFT(176),
  [502] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1203),
  [504] = {.entry = {.count = 1, .reusable = false}}, SHIFT(155),
  [506] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_primary_expression, 1, 0, 1), SHIFT(155),
  [509] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pattern, 1, -1, 1),
  [511] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_pattern, 1, -1, 1), SHIFT(317),
  [514] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_primary_expression, 1, 0, 1), REDUCE(sym_pattern, 1, -1, 1),
  [517] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_pattern, 1, -1, 1), SHIFT(246),
  [520] = {.entry = {.count = 1, .reusable = false}}, SHIFT(527),
  [522] = {.entry = {.count = 1, .reusable = false}}, SHIFT(544),
  [524] = {.entry = {.count = 1, .reusable = false}}, SHIFT(39),
  [526] = {.entry = {.count = 1, .reusable = false}}, SHIFT(82),
  [528] = {.entry = {.count = 1, .reusable = false}}, SHIFT(295),
  [530] = {.entry = {.count = 1, .reusable = false}}, SHIFT(377),
  [532] = {.entry = {.count = 1, .reusable = false}}, SHIFT(213),
  [534] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_primary_expression, 1, 0, 1), SHIFT(220),
  [537] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_primary_expression, 1, 0, 1), SHIFT(1829),
  [540] = {.entry = {.count = 1, .reusable = false}}, SHIFT(220),
  [542] = {.entry = {.count = 1, .reusable = true}}, SHIFT(220),
  [544] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_primary_expression, 1, 0, 1), SHIFT(294),
  [547] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1798),
  [549] = {.entry = {.count = 1, .reusable = false}}, SHIFT(793),
  [551] = {.entry = {.count = 1, .reusable = false}}, SHIFT(541),
  [553] = {.entry = {.count = 1, .reusable = false}}, SHIFT(172),
  [555] = {.entry = {.count = 1, .reusable = true}}, SHIFT(3),
  [557] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1517),
  [559] = {.entry = {.count = 1, .reusable = false}}, SHIFT(397),
  [561] = {.entry = {.count = 1, .reusable = false}}, SHIFT(29),
  [563] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1520),
  [565] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1440),
  [567] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1523),
  [569] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1261),
  [571] = {.entry = {.count = 1, .reusable = false}}, SHIFT(376),
  [573] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1374),
  [575] = {.entry = {.count = 1, .reusable = false}}, SHIFT(542),
  [577] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_primary_expression, 1, 0, 1), REDUCE(sym_rest_pattern, 2, 0, 20),
  [580] = {.entry = {.count = 1, .reusable = true}}, SHIFT(203),
  [582] = {.entry = {.count = 1, .reusable = false}}, SHIFT(577),
  [584] = {.entry = {.count = 1, .reusable = false}}, SHIFT(569),
  [586] = {.entry = {.count = 1, .reusable = false}}, SHIFT(77),
  [588] = {.entry = {.count = 1, .reusable = false}}, SHIFT(83),
  [590] = {.entry = {.count = 1, .reusable = false}}, SHIFT(337),
  [592] = {.entry = {.count = 1, .reusable = false}}, SHIFT(412),
  [594] = {.entry = {.count = 1, .reusable = false}}, SHIFT(216),
  [596] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_primary_expression, 1, 0, 1), SHIFT(335),
  [599] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_primary_expression, 1, 0, 1), SHIFT(1785),
  [602] = {.entry = {.count = 1, .reusable = false}}, SHIFT(335),
  [604] = {.entry = {.count = 1, .reusable = true}}, SHIFT(335),
  [606] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_primary_expression, 1, 0, 1), SHIFT(336),
  [609] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1777),
  [611] = {.entry = {.count = 1, .reusable = false}}, SHIFT(922),
  [613] = {.entry = {.count = 1, .reusable = false}}, SHIFT(568),
  [615] = {.entry = {.count = 1, .reusable = false}}, SHIFT(567),
  [617] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_primary_expression, 1, 0, 1), SHIFT(856),
  [620] = {.entry = {.count = 1, .reusable = false}}, SHIFT(78),
  [622] = {.entry = {.count = 1, .reusable = false}}, SHIFT(85),
  [624] = {.entry = {.count = 1, .reusable = false}}, SHIFT(276),
  [626] = {.entry = {.count = 1, .reusable = false}}, SHIFT(406),
  [628] = {.entry = {.count = 1, .reusable = true}}, SHIFT(209),
  [630] = {.entry = {.count = 1, .reusable = false}}, SHIFT(165),
  [632] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_primary_expression, 1, 0, 1), SHIFT(274),
  [635] = {.entry = {.count = 1, .reusable = false}}, SHIFT(274),
  [637] = {.entry = {.count = 1, .reusable = true}}, SHIFT(274),
  [639] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_primary_expression, 1, 0, 1), SHIFT(275),
  [642] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1771),
  [644] = {.entry = {.count = 1, .reusable = false}}, SHIFT(917),
  [646] = {.entry = {.count = 1, .reusable = false}}, SHIFT(248),
  [648] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_yield_expression, 1, 0, 0),
  [650] = {.entry = {.count = 1, .reusable = true}}, SHIFT(120),
  [652] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_yield_expression, 1, 0, 0),
  [654] = {.entry = {.count = 1, .reusable = true}}, SHIFT(105),
  [656] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1260),
  [658] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1832),
  [660] = {.entry = {.count = 1, .reusable = true}}, SHIFT(245),
  [662] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1242),
  [664] = {.entry = {.count = 1, .reusable = false}}, SHIFT(228),
  [666] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1318),
  [668] = {.entry = {.count = 1, .reusable = false}}, SHIFT(296),
  [670] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1829),
  [672] = {.entry = {.count = 1, .reusable = true}}, SHIFT(294),
  [674] = {.entry = {.count = 1, .reusable = false}}, SHIFT(338),
  [676] = {.entry = {.count = 1, .reusable = true}}, SHIFT(110),
  [678] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1785),
  [680] = {.entry = {.count = 1, .reusable = true}}, SHIFT(336),
  [682] = {.entry = {.count = 1, .reusable = false}}, SHIFT(318),
  [684] = {.entry = {.count = 1, .reusable = true}}, SHIFT(316),
  [686] = {.entry = {.count = 1, .reusable = false}}, SHIFT(356),
  [688] = {.entry = {.count = 1, .reusable = true}}, SHIFT(275),
  [690] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_body, 3, 0, 70),
  [692] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_body, 3, 0, 70),
  [694] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_statement_block, 3, 0, 0),
  [696] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_statement_block, 3, 0, 0),
  [698] = {.entry = {.count = 1, .reusable = true}}, SHIFT(89),
  [700] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_statement_block, 4, 0, 0),
  [702] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_statement_block, 4, 0, 0),
  [704] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_body, 2, 0, 0),
  [706] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_body, 2, 0, 0),
  [708] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_statement_block, 2, 0, 0),
  [710] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_statement_block, 2, 0, 0),
  [712] = {.entry = {.count = 1, .reusable = true}}, SHIFT(87),
  [714] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_object, 2, 0, 0),
  [716] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_object, 2, 0, 0),
  [718] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_object_pattern, 2, 0, 0),
  [720] = {.entry = {.count = 1, .reusable = true}}, SHIFT(385),
  [722] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_declaration, 5, 0, 93),
  [724] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_declaration, 5, 0, 93),
  [726] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class, 5, 0, 93),
  [728] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class, 5, 0, 93),
  [730] = {.entry = {.count = 1, .reusable = true}}, SHIFT(457),
  [732] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_declaration, 5, 0, 89),
  [734] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_declaration, 5, 0, 89),
  [736] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_expression, 5, 0, 89),
  [738] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_expression, 5, 0, 89),
  [740] = {.entry = {.count = 1, .reusable = true}}, SHIFT(455),
  [742] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_generator_function_declaration, 5, 0, 89),
  [744] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_generator_function_declaration, 5, 0, 89),
  [746] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_generator_function, 5, 0, 89),
  [748] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_generator_function, 5, 0, 89),
  [750] = {.entry = {.count = 1, .reusable = true}}, SHIFT(456),
  [752] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_generator_function_declaration, 6, 0, 99),
  [754] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_generator_function_declaration, 6, 0, 99),
  [756] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_generator_function, 6, 0, 99),
  [758] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_generator_function, 6, 0, 99),
  [760] = {.entry = {.count = 1, .reusable = true}}, SHIFT(470),
  [762] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_declaration, 3, 0, 37),
  [764] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_declaration, 3, 0, 37),
  [766] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class, 3, 0, 37),
  [768] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class, 3, 0, 37),
  [770] = {.entry = {.count = 1, .reusable = true}}, SHIFT(465),
  [772] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_declaration, 4, 0, 72),
  [774] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_declaration, 4, 0, 72),
  [776] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class, 4, 0, 72),
  [778] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class, 4, 0, 72),
  [780] = {.entry = {.count = 1, .reusable = true}}, SHIFT(431),
  [782] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_declaration, 4, 0, 77),
  [784] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_declaration, 4, 0, 77),
  [786] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_expression, 4, 0, 77),
  [788] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_expression, 4, 0, 77),
  [790] = {.entry = {.count = 1, .reusable = true}}, SHIFT(434),
  [792] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_declaration, 4, 0, 81),
  [794] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_declaration, 4, 0, 81),
  [796] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class, 4, 0, 81),
  [798] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class, 4, 0, 81),
  [800] = {.entry = {.count = 1, .reusable = true}}, SHIFT(442),
  [802] = {.entry = {.count = 1, .reusable = false}}, SHIFT(489),
  [804] = {.entry = {.count = 1, .reusable = false}}, SHIFT(72),
  [806] = {.entry = {.count = 1, .reusable = false}}, SHIFT(390),
  [808] = {.entry = {.count = 1, .reusable = false}}, SHIFT(388),
  [810] = {.entry = {.count = 1, .reusable = false}}, SHIFT(536),
  [812] = {.entry = {.count = 1, .reusable = false}}, SHIFT(549),
  [814] = {.entry = {.count = 1, .reusable = true}}, SHIFT(117),
  [816] = {.entry = {.count = 1, .reusable = false}}, SHIFT(38),
  [818] = {.entry = {.count = 1, .reusable = true}}, SHIFT(670),
  [820] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1260),
  [822] = {.entry = {.count = 1, .reusable = false}}, SHIFT(379),
  [824] = {.entry = {.count = 1, .reusable = false}}, SHIFT(804),
  [826] = {.entry = {.count = 1, .reusable = true}}, SHIFT(852),
  [828] = {.entry = {.count = 1, .reusable = true}}, SHIFT(683),
  [830] = {.entry = {.count = 1, .reusable = true}}, SHIFT(817),
  [832] = {.entry = {.count = 1, .reusable = true}}, SHIFT(730),
  [834] = {.entry = {.count = 1, .reusable = true}}, SHIFT(603),
  [836] = {.entry = {.count = 1, .reusable = false}}, SHIFT(555),
  [838] = {.entry = {.count = 1, .reusable = false}}, SHIFT(553),
  [840] = {.entry = {.count = 1, .reusable = true}}, SHIFT(870),
  [842] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1200),
  [844] = {.entry = {.count = 1, .reusable = false}}, SHIFT(488),
  [846] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1195),
  [848] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1213),
  [850] = {.entry = {.count = 1, .reusable = false}}, SHIFT(34),
  [852] = {.entry = {.count = 1, .reusable = true}}, SHIFT(108),
  [854] = {.entry = {.count = 1, .reusable = false}}, SHIFT(383),
  [856] = {.entry = {.count = 1, .reusable = false}}, SHIFT(827),
  [858] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_variable_declarator, 1, 0, 5),
  [860] = {.entry = {.count = 1, .reusable = false}}, SHIFT(218),
  [862] = {.entry = {.count = 1, .reusable = true}}, SHIFT(273),
  [864] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_variable_declarator, 1, 0, 5), SHIFT(1648),
  [867] = {.entry = {.count = 1, .reusable = false}}, SHIFT(533),
  [869] = {.entry = {.count = 1, .reusable = false}}, SHIFT(532),
  [871] = {.entry = {.count = 1, .reusable = true}}, SHIFT(868),
  [873] = {.entry = {.count = 1, .reusable = true}}, SHIFT(121),
  [875] = {.entry = {.count = 1, .reusable = false}}, SHIFT(36),
  [877] = {.entry = {.count = 1, .reusable = true}}, SHIFT(109),
  [879] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1346),
  [881] = {.entry = {.count = 1, .reusable = false}}, SHIFT(378),
  [883] = {.entry = {.count = 1, .reusable = true}}, SHIFT(223),
  [885] = {.entry = {.count = 1, .reusable = false}}, SHIFT(765),
  [887] = {.entry = {.count = 1, .reusable = false}}, SHIFT(563),
  [889] = {.entry = {.count = 1, .reusable = false}}, SHIFT(561),
  [891] = {.entry = {.count = 1, .reusable = true}}, SHIFT(871),
  [893] = {.entry = {.count = 1, .reusable = false}}, SHIFT(74),
  [895] = {.entry = {.count = 1, .reusable = true}}, SHIFT(106),
  [897] = {.entry = {.count = 1, .reusable = false}}, SHIFT(387),
  [899] = {.entry = {.count = 1, .reusable = false}}, SHIFT(851),
  [901] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_array_repeat1, 1, 0, 0), REDUCE(aux_sym_array_pattern_repeat1, 1, 0, 0),
  [904] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1227),
  [906] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1538),
  [908] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_array_pattern_repeat1, 1, 0, 0),
  [910] = {.entry = {.count = 1, .reusable = true}}, SHIFT(127),
  [912] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1118),
  [914] = {.entry = {.count = 1, .reusable = true}}, SHIFT(314),
  [916] = {.entry = {.count = 1, .reusable = true}}, SHIFT(728),
  [918] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1660),
  [920] = {.entry = {.count = 1, .reusable = true}}, SHIFT(601),
  [922] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_array_repeat1, 1, 0, 0),
  [924] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1669),
  [926] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1287),
  [928] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1319),
  [930] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1489),
  [932] = {.entry = {.count = 1, .reusable = true}}, SHIFT(103),
  [934] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1172),
  [936] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1197),
  [938] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1199),
  [940] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1196),
  [942] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1783),
  [944] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_primary_expression, 1, 0, 1), REDUCE(sym__property_name, 1, 0, 4),
  [947] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1836),
  [949] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1801),
  [951] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1655),
  [953] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1648),
  [955] = {.entry = {.count = 1, .reusable = false}}, SHIFT(575),
  [957] = {.entry = {.count = 1, .reusable = false}}, SHIFT(574),
  [959] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1223),
  [961] = {.entry = {.count = 1, .reusable = false}}, SHIFT(505),
  [963] = {.entry = {.count = 1, .reusable = false}}, SHIFT(73),
  [965] = {.entry = {.count = 1, .reusable = false}}, SHIFT(417),
  [967] = {.entry = {.count = 1, .reusable = false}}, SHIFT(965),
  [969] = {.entry = {.count = 1, .reusable = true}}, SHIFT(467),
  [971] = {.entry = {.count = 1, .reusable = true}}, SHIFT(43),
  [973] = {.entry = {.count = 1, .reusable = true}}, SHIFT(49),
  [975] = {.entry = {.count = 1, .reusable = true}}, SHIFT(62),
  [977] = {.entry = {.count = 1, .reusable = true}}, SHIFT(60),
  [979] = {.entry = {.count = 1, .reusable = true}}, SHIFT(66),
  [981] = {.entry = {.count = 1, .reusable = true}}, SHIFT(45),
  [983] = {.entry = {.count = 1, .reusable = true}}, SHIFT(46),
  [985] = {.entry = {.count = 1, .reusable = true}}, SHIFT(63),
  [987] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5),
  [989] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6),
  [991] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1865),
  [993] = {.entry = {.count = 1, .reusable = true}}, SHIFT(4),
  [995] = {.entry = {.count = 1, .reusable = false}}, SHIFT(570),
  [997] = {.entry = {.count = 1, .reusable = false}}, SHIFT(565),
  [999] = {.entry = {.count = 1, .reusable = false}}, SHIFT(79),
  [1001] = {.entry = {.count = 1, .reusable = false}}, SHIFT(393),
  [1003] = {.entry = {.count = 1, .reusable = true}}, SHIFT(102),
  [1005] = {.entry = {.count = 1, .reusable = false}}, SHIFT(529),
  [1007] = {.entry = {.count = 1, .reusable = false}}, SHIFT(530),
  [1009] = {.entry = {.count = 1, .reusable = false}}, SHIFT(35),
  [1011] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1754),
  [1013] = {.entry = {.count = 1, .reusable = false}}, SHIFT(372),
  [1015] = {.entry = {.count = 1, .reusable = false}}, SHIFT(688),
  [1017] = {.entry = {.count = 1, .reusable = false}}, SHIFT(503),
  [1019] = {.entry = {.count = 1, .reusable = false}}, SHIFT(504),
  [1021] = {.entry = {.count = 1, .reusable = false}}, SHIFT(16),
  [1023] = {.entry = {.count = 1, .reusable = false}}, SHIFT(367),
  [1025] = {.entry = {.count = 1, .reusable = false}}, SHIFT(550),
  [1027] = {.entry = {.count = 1, .reusable = false}}, SHIFT(546),
  [1029] = {.entry = {.count = 1, .reusable = false}}, SHIFT(33),
  [1031] = {.entry = {.count = 1, .reusable = false}}, SHIFT(373),
  [1033] = {.entry = {.count = 1, .reusable = false}}, SHIFT(566),
  [1035] = {.entry = {.count = 1, .reusable = false}}, SHIFT(573),
  [1037] = {.entry = {.count = 1, .reusable = false}}, SHIFT(75),
  [1039] = {.entry = {.count = 1, .reusable = false}}, SHIFT(414),
  [1041] = {.entry = {.count = 1, .reusable = false}}, SHIFT(540),
  [1043] = {.entry = {.count = 1, .reusable = false}}, SHIFT(539),
  [1045] = {.entry = {.count = 1, .reusable = false}}, SHIFT(32),
  [1047] = {.entry = {.count = 1, .reusable = false}}, SHIFT(382),
  [1049] = {.entry = {.count = 1, .reusable = false}}, SHIFT(761),
  [1051] = {.entry = {.count = 1, .reusable = false}}, SHIFT(554),
  [1053] = {.entry = {.count = 1, .reusable = false}}, SHIFT(559),
  [1055] = {.entry = {.count = 1, .reusable = false}}, SHIFT(51),
  [1057] = {.entry = {.count = 1, .reusable = false}}, SHIFT(386),
  [1059] = {.entry = {.count = 1, .reusable = false}}, SHIFT(819),
  [1061] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1528),
  [1063] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1525),
  [1065] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_primary_expression, 1, 0, 1), SHIFT(1122),
  [1068] = {.entry = {.count = 3, .reusable = true}}, REDUCE(sym_primary_expression, 1, 0, 1), REDUCE(sym__property_name, 1, 0, 4), SHIFT(126),
  [1072] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_primary_expression, 1, 0, 1), SHIFT(269),
  [1075] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1332),
  [1077] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1334),
  [1079] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1386),
  [1081] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1710),
  [1083] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1074),
  [1085] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1767),
  [1087] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1083),
  [1089] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1099),
  [1091] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1856),
  [1093] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1851),
  [1095] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_primary_expression, 1, 0, 1), SHIFT(126),
  [1098] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1323),
  [1100] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1311),
  [1102] = {.entry = {.count = 1, .reusable = true}}, SHIFT(986),
  [1104] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1304),
  [1106] = {.entry = {.count = 1, .reusable = true}}, SHIFT(118),
  [1108] = {.entry = {.count = 1, .reusable = true}}, SHIFT(384),
  [1110] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1827),
  [1112] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1768),
  [1114] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1393),
  [1116] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1869),
  [1118] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1810),
  [1120] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1378),
  [1122] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1823),
  [1124] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1818),
  [1126] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_try_statement, 2, 0, 7),
  [1128] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_try_statement, 2, 0, 7),
  [1130] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1412),
  [1132] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1688),
  [1134] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_try_statement, 3, 0, 30),
  [1136] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_try_statement, 3, 0, 30),
  [1138] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1820),
  [1140] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1819),
  [1142] = {.entry = {.count = 1, .reusable = true}}, SHIFT(391),
  [1144] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_do_statement, 4, 0, 63),
  [1146] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_do_statement, 4, 0, 63),
  [1148] = {.entry = {.count = 1, .reusable = true}}, SHIFT(430),
  [1150] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_catch_clause, 2, 0, 7),
  [1152] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_catch_clause, 2, 0, 7),
  [1154] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parenthesized_expression, 3, 0, 0),
  [1156] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parenthesized_expression, 3, 0, 0),
  [1158] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_catch_clause, 5, 0, 107),
  [1160] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_catch_clause, 5, 0, 107),
  [1162] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_if_statement, 3, 0, 26),
  [1164] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_if_statement, 3, 0, 26),
  [1166] = {.entry = {.count = 1, .reusable = false}}, SHIFT(42),
  [1168] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1781),
  [1170] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1780),
  [1172] = {.entry = {.count = 1, .reusable = true}}, SHIFT(400),
  [1174] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_empty_statement, 1, 0, 0),
  [1176] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_empty_statement, 1, 0, 0),
  [1178] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_declaration, 1, 0, 0),
  [1180] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_declaration, 1, 0, 0),
  [1182] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_using_declaration, 4, 0, 25),
  [1184] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_using_declaration, 4, 0, 25),
  [1186] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_import_statement, 5, 0, 54),
  [1188] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_import_statement, 5, 0, 54),
  [1190] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_using_declaration, 5, 0, 60),
  [1192] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_using_declaration, 5, 0, 60),
  [1194] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_else_clause, 2, 0, 0),
  [1196] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_else_clause, 2, 0, 0),
  [1198] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_switch_body, 3, 0, 0),
  [1200] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_switch_body, 3, 0, 0),
  [1202] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_using_declaration, 4, 0, 60),
  [1204] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_using_declaration, 4, 0, 60),
  [1206] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_export_statement, 4, 0, 54),
  [1208] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_export_statement, 4, 0, 54),
  [1210] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_if_statement, 4, 0, 61),
  [1212] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_if_statement, 4, 0, 61),
  [1214] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_in_statement, 3, 0, 28),
  [1216] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_in_statement, 3, 0, 28),
  [1218] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_do_statement, 5, 0, 63),
  [1220] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_do_statement, 5, 0, 63),
  [1222] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_declaration, 5, 0, 72),
  [1224] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_declaration, 5, 0, 72),
  [1226] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_while_statement, 3, 0, 29),
  [1228] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_while_statement, 3, 0, 29),
  [1230] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_export_statement, 2, 0, 3),
  [1232] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_export_statement, 2, 0, 3),
  [1234] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_declaration, 5, 0, 77),
  [1236] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_declaration, 5, 0, 77),
  [1238] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_try_statement, 3, 0, 31),
  [1240] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_try_statement, 3, 0, 31),
  [1242] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_break_statement, 3, 0, 32),
  [1244] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_break_statement, 3, 0, 32),
  [1246] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_continue_statement, 3, 0, 32),
  [1248] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_continue_statement, 3, 0, 32),
  [1250] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_return_statement, 3, 0, 0),
  [1252] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_return_statement, 3, 0, 0),
  [1254] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_throw_statement, 3, 0, 0),
  [1256] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_throw_statement, 3, 0, 0),
  [1258] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_export_statement, 5, 0, 92),
  [1260] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_export_statement, 5, 0, 92),
  [1262] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_declaration, 5, 0, 81),
  [1264] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_declaration, 5, 0, 81),
  [1266] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_switch_body, 2, 0, 0),
  [1268] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_switch_body, 2, 0, 0),
  [1270] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_in_statement, 4, 0, 62),
  [1272] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_in_statement, 4, 0, 62),
  [1274] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_export_statement, 4, 0, 55),
  [1276] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_export_statement, 4, 0, 55),
  [1278] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_expression_statement, 2, 0, 0),
  [1280] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_expression_statement, 2, 0, 0),
  [1282] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_statement, 6, 0, 97),
  [1284] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_statement, 6, 0, 97),
  [1286] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_break_statement, 2, 0, 0),
  [1288] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_break_statement, 2, 0, 0),
  [1290] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_export_statement, 3, 0, 14),
  [1292] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_export_statement, 3, 0, 14),
  [1294] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_labeled_statement, 3, -1, 15),
  [1296] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_labeled_statement, 3, -1, 15),
  [1298] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_import_statement, 4, 0, 54),
  [1300] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_import_statement, 4, 0, 54),
  [1302] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_declaration, 6, 0, 89),
  [1304] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_declaration, 6, 0, 89),
  [1306] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_generator_function_declaration, 6, 0, 89),
  [1308] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_generator_function_declaration, 6, 0, 89),
  [1310] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_declaration, 6, 0, 93),
  [1312] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_declaration, 6, 0, 93),
  [1314] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_finally_clause, 2, 0, 7),
  [1316] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_finally_clause, 2, 0, 7),
  [1318] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_export_statement, 3, 0, 0),
  [1320] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_export_statement, 3, 0, 0),
  [1322] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_try_statement, 4, 0, 64),
  [1324] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_try_statement, 4, 0, 64),
  [1326] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_continue_statement, 2, 0, 0),
  [1328] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_continue_statement, 2, 0, 0),
  [1330] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_debugger_statement, 2, 0, 0),
  [1332] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_debugger_statement, 2, 0, 0),
  [1334] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_statement, 7, 0, 104),
  [1336] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_statement, 7, 0, 104),
  [1338] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_statement, 7, 0, 105),
  [1340] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_statement, 7, 0, 105),
  [1342] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_declaration, 4, 0, 37),
  [1344] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_declaration, 4, 0, 37),
  [1346] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_statement, 7, 0, 106),
  [1348] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_statement, 7, 0, 106),
  [1350] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_return_statement, 2, 0, 0),
  [1352] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_return_statement, 2, 0, 0),
  [1354] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_import_statement, 4, 0, 21),
  [1356] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_import_statement, 4, 0, 21),
  [1358] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_export_statement, 4, 0, 80),
  [1360] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_export_statement, 4, 0, 80),
  [1362] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_generator_function_declaration, 7, 0, 99),
  [1364] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_generator_function_declaration, 7, 0, 99),
  [1366] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_statement, 1, 0, 0),
  [1368] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_statement, 1, 0, 0),
  [1370] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_import_statement, 3, 0, 21),
  [1372] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_import_statement, 3, 0, 21),
  [1374] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_with_statement, 3, 0, 22),
  [1376] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_with_statement, 3, 0, 22),
  [1378] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_variable_declaration, 3, 0, 0),
  [1380] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_variable_declaration, 3, 0, 0),
  [1382] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_lexical_declaration, 3, 0, 25),
  [1384] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_lexical_declaration, 3, 0, 25),
  [1386] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_using_declaration, 3, 0, 25),
  [1388] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_using_declaration, 3, 0, 25),
  [1390] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_switch_statement, 3, 0, 27),
  [1392] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_switch_statement, 3, 0, 27),
  [1394] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_statement, 8, 0, 115),
  [1396] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_statement, 8, 0, 115),
  [1398] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_statement, 8, 0, 116),
  [1400] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_statement, 8, 0, 116),
  [1402] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_statement, 8, 0, 117),
  [1404] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_statement, 8, 0, 117),
  [1406] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_statement, 9, 0, 120),
  [1408] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_statement, 9, 0, 120),
  [1410] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_variable_declaration, 4, 0, 0),
  [1412] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_variable_declaration, 4, 0, 0),
  [1414] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_lexical_declaration, 4, 0, 25),
  [1416] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_lexical_declaration, 4, 0, 25),
  [1418] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_export_statement, 3, 0, 52),
  [1420] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_export_statement, 3, 0, 52),
  [1422] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1202),
  [1424] = {.entry = {.count = 1, .reusable = false}}, SHIFT(113),
  [1426] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_primary_expression, 1, 0, 0),
  [1428] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_primary_expression, 1, 0, 0),
  [1430] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_primary_expression, 1, 0, 0), REDUCE(sym__property_name, 1, 0, 4),
  [1433] = {.entry = {.count = 1, .reusable = false}}, SHIFT(270),
  [1435] = {.entry = {.count = 1, .reusable = true}}, SHIFT(173),
  [1437] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__augmented_assignment_lhs, 1, 0, 0),
  [1439] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_subscript_expression, 5, 0, 91),
  [1441] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_subscript_expression, 5, 0, 91),
  [1443] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_member_expression, 3, 0, 45),
  [1445] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_member_expression, 3, 0, 45),
  [1447] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_member_expression, 3, 0, 46),
  [1449] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_member_expression, 3, 0, 46),
  [1451] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_member_expression, 3, 0, 47),
  [1453] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_member_expression, 3, 0, 47),
  [1455] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_member_expression, 3, 0, 48),
  [1457] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_member_expression, 3, 0, 48),
  [1459] = {.entry = {.count = 1, .reusable = false}}, SHIFT(319),
  [1461] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_subscript_expression, 4, 0, 79),
  [1463] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_subscript_expression, 4, 0, 79),
  [1465] = {.entry = {.count = 1, .reusable = false}}, SHIFT(249),
  [1467] = {.entry = {.count = 1, .reusable = true}}, SHIFT(168),
  [1469] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1469),
  [1471] = {.entry = {.count = 1, .reusable = false}}, SHIFT(136),
  [1473] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_primary_expression, 1, 0, 1), SHIFT(118),
  [1476] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__for_header, 7, 0, 109),
  [1478] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__for_header, 7, 0, 109),
  [1480] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__for_header, 7, 0, 110),
  [1482] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__for_header, 7, 0, 110),
  [1484] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__for_header, 7, 0, 111),
  [1486] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__for_header, 7, 0, 111),
  [1488] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__for_header, 7, 0, 112),
  [1490] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__for_header, 7, 0, 112),
  [1492] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__for_header, 7, 0, 113),
  [1494] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__for_header, 7, 0, 113),
  [1496] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__for_header, 7, 0, 114),
  [1498] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__for_header, 7, 0, 114),
  [1500] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__for_header, 6, 0, 102),
  [1502] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__for_header, 6, 0, 102),
  [1504] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__for_header, 6, 0, 103),
  [1506] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__for_header, 6, 0, 103),
  [1508] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__for_header, 8, 0, 118),
  [1510] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__for_header, 8, 0, 118),
  [1512] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__for_header, 5, 0, 95),
  [1514] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__for_header, 5, 0, 95),
  [1516] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__for_header, 8, 0, 119),
  [1518] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__for_header, 8, 0, 119),
  [1520] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__for_header, 5, 0, 96),
  [1522] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__for_header, 5, 0, 96),
  [1524] = {.entry = {.count = 1, .reusable = false}}, SHIFT(297),
  [1526] = {.entry = {.count = 1, .reusable = true}}, SHIFT(180),
  [1528] = {.entry = {.count = 1, .reusable = false}}, SHIFT(231),
  [1530] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pattern, 1, -1, 0),
  [1532] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_pattern, 1, -1, 0), SHIFT(319),
  [1535] = {.entry = {.count = 1, .reusable = true}}, SHIFT(196),
  [1537] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_primary_expression, 1, 0, 0), REDUCE(sym_pattern, 1, -1, 0),
  [1540] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_pattern, 1, -1, 0), SHIFT(249),
  [1543] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_rest_pattern, 2, 0, 0),
  [1545] = {.entry = {.count = 1, .reusable = true}}, SHIFT(155),
  [1547] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_primary_expression, 1, 0, 0), REDUCE(sym_rest_pattern, 2, 0, 0),
  [1550] = {.entry = {.count = 1, .reusable = true}}, SHIFT(160),
  [1552] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_primary_expression, 1, 0, 0), SHIFT(160),
  [1555] = {.entry = {.count = 1, .reusable = true}}, SHIFT(204),
  [1557] = {.entry = {.count = 1, .reusable = false}}, SHIFT(278),
  [1559] = {.entry = {.count = 1, .reusable = true}}, SHIFT(212),
  [1561] = {.entry = {.count = 1, .reusable = false}}, SHIFT(339),
  [1563] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_new_expression, 2, 0, 8),
  [1565] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_new_expression, 2, 0, 8),
  [1567] = {.entry = {.count = 1, .reusable = true}}, SHIFT(125),
  [1569] = {.entry = {.count = 1, .reusable = true}}, SHIFT(181),
  [1571] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1503),
  [1573] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1293),
  [1575] = {.entry = {.count = 1, .reusable = true}}, SHIFT(647),
  [1577] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_expression, 1, 0, 0),
  [1579] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_expression, 1, 0, 0),
  [1581] = {.entry = {.count = 1, .reusable = true}}, SHIFT(618),
  [1583] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_await_expression, 2, 0, 0),
  [1585] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_await_expression, 2, 0, 0),
  [1587] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1411),
  [1589] = {.entry = {.count = 1, .reusable = true}}, SHIFT(606),
  [1591] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_unary_expression, 2, 0, 9),
  [1593] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_unary_expression, 2, 0, 9),
  [1595] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_update_expression, 2, 0, 9),
  [1597] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_update_expression, 2, 0, 9),
  [1599] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_jsx_closing_element, 3, 0, 33),
  [1601] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_jsx_closing_element, 3, 0, 33),
  [1603] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_jsx_self_closing_element, 3, 0, 33),
  [1605] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_jsx_self_closing_element, 3, 0, 33),
  [1607] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class, 3, 0, 38),
  [1609] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class, 3, 0, 38),
  [1611] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_expression, 3, 0, 39),
  [1613] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_expression, 3, 0, 39),
  [1615] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_new_expression, 3, 1, 40),
  [1617] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_new_expression, 3, 1, 40),
  [1619] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_regex, 3, 0, 41),
  [1621] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_regex, 3, 0, 41),
  [1623] = {.entry = {.count = 1, .reusable = false}}, SHIFT(633),
  [1625] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_template_string, 3, 0, 0),
  [1627] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_template_string, 3, 0, 0),
  [1629] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_arrow_function, 3, 0, 43),
  [1631] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_arrow_function, 3, 0, 43),
  [1633] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_arguments, 2, 0, 0),
  [1635] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_arguments, 2, 0, 0),
  [1637] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array, 2, 0, 0),
  [1639] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array, 2, 0, 0),
  [1641] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_array, 2, 0, 0), REDUCE(sym_array_pattern, 2, 0, 0),
  [1644] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array_pattern, 2, 0, 0),
  [1646] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class, 2, 0, 7),
  [1648] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class, 2, 0, 7),
  [1650] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_update_expression, 2, 0, 11),
  [1652] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_update_expression, 2, 0, 11),
  [1654] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_call_expression, 2, 0, 10),
  [1656] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_call_expression, 2, 0, 10),
  [1658] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_jsx_element, 2, 0, 12),
  [1660] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_jsx_element, 2, 0, 12),
  [1662] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_call_expression, 3, 0, 49),
  [1664] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_call_expression, 3, 0, 49),
  [1666] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_jsx_closing_element, 2, 0, 0),
  [1668] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_jsx_closing_element, 2, 0, 0),
  [1670] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_jsx_element, 3, 0, 50),
  [1672] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_jsx_element, 3, 0, 50),
  [1674] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_arrow_function, 3, 0, 51),
  [1676] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_arrow_function, 3, 0, 51),
  [1678] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_string, 3, 0, 0),
  [1680] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_string, 3, 0, 0),
  [1682] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class, 3, 0, 53),
  [1684] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class, 3, 0, 53),
  [1686] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_object, 4, 0, 18),
  [1688] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_object, 4, 0, 18),
  [1690] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_object, 4, 0, 0),
  [1692] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_object, 4, 0, 0),
  [1694] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_jsx_self_closing_element, 4, 0, 65),
  [1696] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_jsx_self_closing_element, 4, 0, 65),
  [1698] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_arrow_function, 4, 0, 73),
  [1700] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_arrow_function, 4, 0, 73),
  [1702] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_expression, 4, 0, 74),
  [1704] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_expression, 4, 0, 74),
  [1706] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_arrow_function, 4, 0, 75),
  [1708] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_arrow_function, 4, 0, 75),
  [1710] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_arrow_function, 4, 0, 76),
  [1712] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_arrow_function, 4, 0, 76),
  [1714] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_arrow_function, 3, 0, 17),
  [1716] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_arrow_function, 3, 0, 17),
  [1718] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_generator_function, 4, 0, 74),
  [1720] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_generator_function, 4, 0, 74),
  [1722] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_object, 3, 0, 18),
  [1724] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_object, 3, 0, 18),
  [1726] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_object, 3, 0, 18), REDUCE(sym_object_pattern, 3, 0, 19),
  [1729] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_object_pattern, 3, 0, 19),
  [1731] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_regex, 4, 0, 78),
  [1733] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_regex, 4, 0, 78),
  [1735] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_arguments, 3, 0, 0),
  [1737] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_arguments, 3, 0, 0),
  [1739] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_template_string, 2, 0, 0),
  [1741] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_template_string, 2, 0, 0),
  [1743] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_arguments, 4, 0, 0),
  [1745] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_arguments, 4, 0, 0),
  [1747] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_object, 3, 0, 0),
  [1749] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_object, 3, 0, 0),
  [1751] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_string, 2, 0, 0),
  [1753] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_string, 2, 0, 0),
  [1755] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_generator_function, 5, 0, 88),
  [1757] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_generator_function, 5, 0, 88),
  [1759] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_meta_property, 3, 0, 0),
  [1761] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_meta_property, 3, 0, 0),
  [1763] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class, 4, 0, 82),
  [1765] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class, 4, 0, 82),
  [1767] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_object, 2, 0, 0), REDUCE(sym_object_pattern, 2, 0, 0),
  [1770] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array, 3, 0, 0),
  [1772] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array, 3, 0, 0),
  [1774] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array, 4, 0, 0),
  [1776] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array, 4, 0, 0),
  [1778] = {.entry = {.count = 1, .reusable = false}}, SHIFT(251),
  [1780] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_yield_expression, 3, 0, 0),
  [1782] = {.entry = {.count = 1, .reusable = false}}, SHIFT(250),
  [1784] = {.entry = {.count = 1, .reusable = true}}, SHIFT(253),
  [1786] = {.entry = {.count = 1, .reusable = true}}, SHIFT(254),
  [1788] = {.entry = {.count = 1, .reusable = false}}, SHIFT(255),
  [1790] = {.entry = {.count = 1, .reusable = true}}, SHIFT(255),
  [1792] = {.entry = {.count = 1, .reusable = false}}, SHIFT(256),
  [1794] = {.entry = {.count = 1, .reusable = true}}, SHIFT(257),
  [1796] = {.entry = {.count = 1, .reusable = false}}, SHIFT(258),
  [1798] = {.entry = {.count = 1, .reusable = false}}, SHIFT(259),
  [1800] = {.entry = {.count = 1, .reusable = true}}, SHIFT(251),
  [1802] = {.entry = {.count = 1, .reusable = true}}, SHIFT(260),
  [1804] = {.entry = {.count = 1, .reusable = true}}, SHIFT(250),
  [1806] = {.entry = {.count = 1, .reusable = false}}, SHIFT(261),
  [1808] = {.entry = {.count = 1, .reusable = true}}, SHIFT(261),
  [1810] = {.entry = {.count = 1, .reusable = true}}, SHIFT(262),
  [1812] = {.entry = {.count = 1, .reusable = true}}, SHIFT(299),
  [1814] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_ternary_expression, 5, 0, 90),
  [1816] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_assignment_expression, 3, 0, 16),
  [1818] = {.entry = {.count = 1, .reusable = true}}, SHIFT(123),
  [1820] = {.entry = {.count = 1, .reusable = true}}, SHIFT(189),
  [1822] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1507),
  [1824] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1310),
  [1826] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_assignment_expression, 3, 0, 42),
  [1828] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_binary_expression, 3, 0, 44),
  [1830] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_binary_expression, 3, 0, 44),
  [1832] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_yield_expression, 2, 0, 0),
  [1834] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_augmented_assignment_expression, 3, 0, 44),
  [1836] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_sequence_expression_repeat1, 2, 0, 0),
  [1838] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1494),
  [1840] = {.entry = {.count = 1, .reusable = true}}, SHIFT(744),
  [1842] = {.entry = {.count = 1, .reusable = false}}, SHIFT(300),
  [1844] = {.entry = {.count = 1, .reusable = false}}, SHIFT(298),
  [1846] = {.entry = {.count = 1, .reusable = true}}, SHIFT(301),
  [1848] = {.entry = {.count = 1, .reusable = true}}, SHIFT(302),
  [1850] = {.entry = {.count = 1, .reusable = false}}, SHIFT(303),
  [1852] = {.entry = {.count = 1, .reusable = true}}, SHIFT(303),
  [1854] = {.entry = {.count = 1, .reusable = false}}, SHIFT(304),
  [1856] = {.entry = {.count = 1, .reusable = true}}, SHIFT(305),
  [1858] = {.entry = {.count = 1, .reusable = false}}, SHIFT(306),
  [1860] = {.entry = {.count = 1, .reusable = false}}, SHIFT(307),
  [1862] = {.entry = {.count = 1, .reusable = true}}, SHIFT(300),
  [1864] = {.entry = {.count = 1, .reusable = true}}, SHIFT(308),
  [1866] = {.entry = {.count = 1, .reusable = true}}, SHIFT(298),
  [1868] = {.entry = {.count = 1, .reusable = false}}, SHIFT(309),
  [1870] = {.entry = {.count = 1, .reusable = true}}, SHIFT(309),
  [1872] = {.entry = {.count = 1, .reusable = true}}, SHIFT(310),
  [1874] = {.entry = {.count = 1, .reusable = true}}, SHIFT(277),
  [1876] = {.entry = {.count = 1, .reusable = false}}, SHIFT(810),
  [1878] = {.entry = {.count = 1, .reusable = false}}, SHIFT(233),
  [1880] = {.entry = {.count = 1, .reusable = false}}, SHIFT(219),
  [1882] = {.entry = {.count = 1, .reusable = true}}, SHIFT(235),
  [1884] = {.entry = {.count = 1, .reusable = true}}, SHIFT(236),
  [1886] = {.entry = {.count = 1, .reusable = false}}, SHIFT(237),
  [1888] = {.entry = {.count = 1, .reusable = true}}, SHIFT(237),
  [1890] = {.entry = {.count = 1, .reusable = false}}, SHIFT(238),
  [1892] = {.entry = {.count = 1, .reusable = true}}, SHIFT(239),
  [1894] = {.entry = {.count = 1, .reusable = false}}, SHIFT(240),
  [1896] = {.entry = {.count = 1, .reusable = false}}, SHIFT(241),
  [1898] = {.entry = {.count = 1, .reusable = true}}, SHIFT(233),
  [1900] = {.entry = {.count = 1, .reusable = true}}, SHIFT(242),
  [1902] = {.entry = {.count = 1, .reusable = true}}, SHIFT(219),
  [1904] = {.entry = {.count = 1, .reusable = false}}, SHIFT(243),
  [1906] = {.entry = {.count = 1, .reusable = true}}, SHIFT(243),
  [1908] = {.entry = {.count = 1, .reusable = true}}, SHIFT(247),
  [1910] = {.entry = {.count = 1, .reusable = true}}, SHIFT(232),
  [1912] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_pattern, 2, 0, 0),
  [1914] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_object_pattern, 3, 0, 19),
  [1916] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_object_pattern, 2, 0, 0),
  [1918] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_object, 2, 0, 0), REDUCE(sym_object_pattern, 2, 0, 0),
  [1921] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_object_assignment_pattern, 3, 0, 56), REDUCE(sym_assignment_expression, 3, 0, 42),
  [1924] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_object_assignment_pattern, 3, 0, 56),
  [1926] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_object_assignment_pattern, 3, 0, 56), REDUCE(sym_assignment_expression, 3, 0, 16),
  [1929] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_primary_expression, 1, 0, 0), REDUCE(sym__property_name, 1, 0, 0),
  [1932] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__property_name, 1, 0, 0),
  [1934] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_array, 3, 0, 0), REDUCE(sym_computed_property_name, 3, 0, 0),
  [1937] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_computed_property_name, 3, 0, 0),
  [1939] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_array, 2, 0, 0), REDUCE(sym_array_pattern, 2, 0, 0),
  [1942] = {.entry = {.count = 1, .reusable = true}}, SHIFT(234),
  [1944] = {.entry = {.count = 1, .reusable = true}}, SHIFT(439),
  [1946] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym_object, 3, 0, 18), REDUCE(sym_object_pattern, 3, 0, 19),
  [1949] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_object_assignment_pattern, 3, 0, 42), REDUCE(sym_assignment_expression, 3, 0, 42),
  [1952] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_object_assignment_pattern, 3, 0, 42),
  [1954] = {.entry = {.count = 1, .reusable = true}}, SHIFT(438),
  [1956] = {.entry = {.count = 1, .reusable = true}}, SHIFT(447),
  [1958] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_assignment_pattern, 3, 0, 42),
  [1960] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_spread_element, 2, 0, 0),
  [1962] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__initializer, 2, 0, 59),
  [1964] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__initializer, 2, 0, 59), SHIFT(298),
  [1967] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1217),
  [1969] = {.entry = {.count = 1, .reusable = true}}, SHIFT(987),
  [1971] = {.entry = {.count = 1, .reusable = true}}, SHIFT(694),
  [1973] = {.entry = {.count = 1, .reusable = true}}, SHIFT(116),
  [1975] = {.entry = {.count = 1, .reusable = false}}, SHIFT(994),
  [1977] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1407),
  [1979] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1003),
  [1981] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1007),
  [1983] = {.entry = {.count = 1, .reusable = true}}, SHIFT(252),
  [1985] = {.entry = {.count = 1, .reusable = true}}, SHIFT(515),
  [1987] = {.entry = {.count = 1, .reusable = true}}, SHIFT(516),
  [1989] = {.entry = {.count = 1, .reusable = true}}, SHIFT(47),
  [1991] = {.entry = {.count = 1, .reusable = true}}, SHIFT(48),
  [1993] = {.entry = {.count = 1, .reusable = true}}, SHIFT(506),
  [1995] = {.entry = {.count = 1, .reusable = true}}, SHIFT(518),
  [1997] = {.entry = {.count = 1, .reusable = true}}, SHIFT(520),
  [1999] = {.entry = {.count = 1, .reusable = true}}, SHIFT(509),
  [2001] = {.entry = {.count = 1, .reusable = true}}, SHIFT(512),
  [2003] = {.entry = {.count = 1, .reusable = true}}, SHIFT(513),
  [2005] = {.entry = {.count = 1, .reusable = true}}, SHIFT(514),
  [2007] = {.entry = {.count = 1, .reusable = true}}, SHIFT(50),
  [2009] = {.entry = {.count = 1, .reusable = true}}, SHIFT(767),
  [2011] = {.entry = {.count = 1, .reusable = true}}, SHIFT(139),
  [2013] = {.entry = {.count = 1, .reusable = true}}, SHIFT(517),
  [2015] = {.entry = {.count = 1, .reusable = true}}, SHIFT(522),
  [2017] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_array_repeat1, 2, 0, 0),
  [2019] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1375),
  [2021] = {.entry = {.count = 1, .reusable = true}}, SHIFT(747),
  [2023] = {.entry = {.count = 1, .reusable = true}}, SHIFT(534),
  [2025] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1266),
  [2027] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1220),
  [2029] = {.entry = {.count = 1, .reusable = true}}, SHIFT(648),
  [2031] = {.entry = {.count = 1, .reusable = false}}, SHIFT(998),
  [2033] = {.entry = {.count = 1, .reusable = false}}, SHIFT(999),
  [2035] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1010),
  [2037] = {.entry = {.count = 1, .reusable = true}}, SHIFT(524),
  [2039] = {.entry = {.count = 1, .reusable = true}}, SHIFT(649),
  [2041] = {.entry = {.count = 1, .reusable = true}}, SHIFT(634),
  [2043] = {.entry = {.count = 1, .reusable = true}}, SHIFT(500),
  [2045] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1314),
  [2047] = {.entry = {.count = 1, .reusable = true}}, SHIFT(493),
  [2049] = {.entry = {.count = 1, .reusable = true}}, SHIFT(149),
  [2051] = {.entry = {.count = 1, .reusable = true}}, SHIFT(138),
  [2053] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1212),
  [2055] = {.entry = {.count = 1, .reusable = true}}, SHIFT(665),
  [2057] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1004),
  [2059] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1002),
  [2061] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1005),
  [2063] = {.entry = {.count = 1, .reusable = true}}, SHIFT(491),
  [2065] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1100),
  [2067] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1211),
  [2069] = {.entry = {.count = 1, .reusable = true}}, SHIFT(733),
  [2071] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1000),
  [2073] = {.entry = {.count = 1, .reusable = false}}, SHIFT(995),
  [2075] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1008),
  [2077] = {.entry = {.count = 1, .reusable = true}}, SHIFT(404),
  [2079] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1214),
  [2081] = {.entry = {.count = 1, .reusable = true}}, SHIFT(812),
  [2083] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1001),
  [2085] = {.entry = {.count = 1, .reusable = false}}, SHIFT(993),
  [2087] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1011),
  [2089] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1210),
  [2091] = {.entry = {.count = 1, .reusable = true}}, SHIFT(854),
  [2093] = {.entry = {.count = 1, .reusable = false}}, SHIFT(996),
  [2095] = {.entry = {.count = 1, .reusable = false}}, SHIFT(997),
  [2097] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1006),
  [2099] = {.entry = {.count = 1, .reusable = true}}, SHIFT(816),
  [2101] = {.entry = {.count = 1, .reusable = true}}, SHIFT(551),
  [2103] = {.entry = {.count = 1, .reusable = true}}, SHIFT(61),
  [2105] = {.entry = {.count = 1, .reusable = true}}, SHIFT(538),
  [2107] = {.entry = {.count = 1, .reusable = true}}, SHIFT(64),
  [2109] = {.entry = {.count = 1, .reusable = true}}, SHIFT(65),
  [2111] = {.entry = {.count = 1, .reusable = true}}, SHIFT(67),
  [2113] = {.entry = {.count = 1, .reusable = true}}, SHIFT(14),
  [2115] = {.entry = {.count = 1, .reusable = true}}, SHIFT(519),
  [2117] = {.entry = {.count = 1, .reusable = true}}, SHIFT(523),
  [2119] = {.entry = {.count = 1, .reusable = true}}, SHIFT(44),
  [2121] = {.entry = {.count = 1, .reusable = true}}, SHIFT(145),
  [2123] = {.entry = {.count = 1, .reusable = true}}, SHIFT(146),
  [2125] = {.entry = {.count = 1, .reusable = true}}, SHIFT(148),
  [2127] = {.entry = {.count = 1, .reusable = true}}, SHIFT(511),
  [2129] = {.entry = {.count = 1, .reusable = false}}, SHIFT(321),
  [2131] = {.entry = {.count = 1, .reusable = false}}, SHIFT(320),
  [2133] = {.entry = {.count = 1, .reusable = true}}, SHIFT(322),
  [2135] = {.entry = {.count = 1, .reusable = true}}, SHIFT(323),
  [2137] = {.entry = {.count = 1, .reusable = false}}, SHIFT(324),
  [2139] = {.entry = {.count = 1, .reusable = true}}, SHIFT(324),
  [2141] = {.entry = {.count = 1, .reusable = false}}, SHIFT(325),
  [2143] = {.entry = {.count = 1, .reusable = true}}, SHIFT(326),
  [2145] = {.entry = {.count = 1, .reusable = false}}, SHIFT(327),
  [2147] = {.entry = {.count = 1, .reusable = false}}, SHIFT(328),
  [2149] = {.entry = {.count = 1, .reusable = true}}, SHIFT(321),
  [2151] = {.entry = {.count = 1, .reusable = true}}, SHIFT(329),
  [2153] = {.entry = {.count = 1, .reusable = true}}, SHIFT(320),
  [2155] = {.entry = {.count = 1, .reusable = false}}, SHIFT(330),
  [2157] = {.entry = {.count = 1, .reusable = true}}, SHIFT(330),
  [2159] = {.entry = {.count = 1, .reusable = true}}, SHIFT(331),
  [2161] = {.entry = {.count = 1, .reusable = true}}, SHIFT(334),
  [2163] = {.entry = {.count = 1, .reusable = true}}, SHIFT(445),
  [2165] = {.entry = {.count = 1, .reusable = false}}, SHIFT(280),
  [2167] = {.entry = {.count = 1, .reusable = false}}, SHIFT(279),
  [2169] = {.entry = {.count = 1, .reusable = true}}, SHIFT(281),
  [2171] = {.entry = {.count = 1, .reusable = true}}, SHIFT(282),
  [2173] = {.entry = {.count = 1, .reusable = false}}, SHIFT(283),
  [2175] = {.entry = {.count = 1, .reusable = true}}, SHIFT(283),
  [2177] = {.entry = {.count = 1, .reusable = false}}, SHIFT(284),
  [2179] = {.entry = {.count = 1, .reusable = true}}, SHIFT(285),
  [2181] = {.entry = {.count = 1, .reusable = false}}, SHIFT(286),
  [2183] = {.entry = {.count = 1, .reusable = false}}, SHIFT(287),
  [2185] = {.entry = {.count = 1, .reusable = true}}, SHIFT(280),
  [2187] = {.entry = {.count = 1, .reusable = true}}, SHIFT(288),
  [2189] = {.entry = {.count = 1, .reusable = true}}, SHIFT(279),
  [2191] = {.entry = {.count = 1, .reusable = false}}, SHIFT(289),
  [2193] = {.entry = {.count = 1, .reusable = true}}, SHIFT(289),
  [2195] = {.entry = {.count = 1, .reusable = true}}, SHIFT(290),
  [2197] = {.entry = {.count = 1, .reusable = true}}, SHIFT(354),
  [2199] = {.entry = {.count = 1, .reusable = false}}, SHIFT(341),
  [2201] = {.entry = {.count = 1, .reusable = false}}, SHIFT(340),
  [2203] = {.entry = {.count = 1, .reusable = true}}, SHIFT(342),
  [2205] = {.entry = {.count = 1, .reusable = true}}, SHIFT(343),
  [2207] = {.entry = {.count = 1, .reusable = false}}, SHIFT(344),
  [2209] = {.entry = {.count = 1, .reusable = true}}, SHIFT(344),
  [2211] = {.entry = {.count = 1, .reusable = false}}, SHIFT(345),
  [2213] = {.entry = {.count = 1, .reusable = true}}, SHIFT(346),
  [2215] = {.entry = {.count = 1, .reusable = false}}, SHIFT(347),
  [2217] = {.entry = {.count = 1, .reusable = false}}, SHIFT(348),
  [2219] = {.entry = {.count = 1, .reusable = true}}, SHIFT(341),
  [2221] = {.entry = {.count = 1, .reusable = true}}, SHIFT(349),
  [2223] = {.entry = {.count = 1, .reusable = true}}, SHIFT(340),
  [2225] = {.entry = {.count = 1, .reusable = false}}, SHIFT(350),
  [2227] = {.entry = {.count = 1, .reusable = true}}, SHIFT(350),
  [2229] = {.entry = {.count = 1, .reusable = true}}, SHIFT(351),
  [2231] = {.entry = {.count = 1, .reusable = true}}, SHIFT(355),
  [2233] = {.entry = {.count = 1, .reusable = true}}, SHIFT(441),
  [2235] = {.entry = {.count = 1, .reusable = true}}, SHIFT(440),
  [2237] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pair, 3, 0, 57),
  [2239] = {.entry = {.count = 1, .reusable = true}}, SHIFT(226),
  [2241] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1297),
  [2243] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_object_repeat1, 1, 0, 0), REDUCE(aux_sym_object_pattern_repeat1, 1, 0, 0),
  [2246] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1016),
  [2248] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1015),
  [2250] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1027),
  [2252] = {.entry = {.count = 1, .reusable = true}}, SHIFT(292),
  [2254] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_heritage, 2, 0, 0),
  [2256] = {.entry = {.count = 1, .reusable = true}}, SHIFT(267),
  [2258] = {.entry = {.count = 1, .reusable = true}}, SHIFT(333),
  [2260] = {.entry = {.count = 1, .reusable = true}}, SHIFT(313),
  [2262] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1548),
  [2264] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1389),
  [2266] = {.entry = {.count = 1, .reusable = true}}, SHIFT(353),
  [2268] = {.entry = {.count = 2, .reusable = false}}, REDUCE(sym__initializer, 2, 0, 59), SHIFT(340),
  [2271] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1277),
  [2273] = {.entry = {.count = 1, .reusable = true}}, SHIFT(989),
  [2275] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1620),
  [2277] = {.entry = {.count = 1, .reusable = true}}, SHIFT(269),
  [2279] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1012),
  [2281] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1448),
  [2283] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1013),
  [2285] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1020),
  [2287] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1342),
  [2289] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1130),
  [2291] = {.entry = {.count = 1, .reusable = true}}, SHIFT(402),
  [2293] = {.entry = {.count = 1, .reusable = true}}, SHIFT(985),
  [2295] = {.entry = {.count = 1, .reusable = true}}, SHIFT(312),
  [2297] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1330),
  [2299] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1331),
  [2301] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1023),
  [2303] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1228),
  [2305] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1009),
  [2307] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1101),
  [2309] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1028),
  [2311] = {.entry = {.count = 1, .reusable = true}}, SHIFT(594),
  [2313] = {.entry = {.count = 1, .reusable = true}}, SHIFT(984),
  [2315] = {.entry = {.count = 1, .reusable = true}}, SHIFT(625),
  [2317] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_body_repeat1, 2, 0, 71), SHIFT_REPEAT(1342),
  [2320] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_body_repeat1, 2, 0, 71), SHIFT_REPEAT(1130),
  [2323] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_class_body_repeat1, 2, 0, 71),
  [2325] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_body_repeat1, 2, 0, 71), SHIFT_REPEAT(985),
  [2328] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_body_repeat1, 2, 0, 71), SHIFT_REPEAT(312),
  [2331] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_body_repeat1, 2, 0, 71), SHIFT_REPEAT(1330),
  [2334] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_body_repeat1, 2, 0, 71), SHIFT_REPEAT(1331),
  [2337] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_body_repeat1, 2, 0, 71), SHIFT_REPEAT(1023),
  [2340] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_body_repeat1, 2, 0, 71), SHIFT_REPEAT(1228),
  [2343] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_body_repeat1, 2, 0, 71), SHIFT_REPEAT(1470),
  [2346] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_body_repeat1, 2, 0, 71), SHIFT_REPEAT(1009),
  [2349] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_body_repeat1, 2, 0, 71), SHIFT_REPEAT(1101),
  [2352] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_class_body_repeat1, 2, 0, 71), SHIFT_REPEAT(1028),
  [2355] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1262),
  [2357] = {.entry = {.count = 1, .reusable = true}}, SHIFT(992),
  [2359] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1251),
  [2361] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1796),
  [2363] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1296),
  [2365] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1340),
  [2367] = {.entry = {.count = 1, .reusable = true}}, SHIFT(86),
  [2369] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1397),
  [2371] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_object_repeat1, 1, 0, 0),
  [2373] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1021),
  [2375] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1024),
  [2377] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1035),
  [2379] = {.entry = {.count = 1, .reusable = true}}, SHIFT(90),
  [2381] = {.entry = {.count = 1, .reusable = true}}, SHIFT(988),
  [2383] = {.entry = {.count = 1, .reusable = true}}, SHIFT(399),
  [2385] = {.entry = {.count = 1, .reusable = true}}, SHIFT(982),
  [2387] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1394),
  [2389] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_object_pattern_repeat1, 1, 0, 0),
  [2391] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1122),
  [2393] = {.entry = {.count = 1, .reusable = true}}, SHIFT(820),
  [2395] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__property_name, 1, 0, 4),
  [2397] = {.entry = {.count = 1, .reusable = true}}, SHIFT(221),
  [2399] = {.entry = {.count = 1, .reusable = true}}, SHIFT(731),
  [2401] = {.entry = {.count = 1, .reusable = true}}, SHIFT(853),
  [2403] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1131),
  [2405] = {.entry = {.count = 1, .reusable = true}}, SHIFT(27),
  [2407] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1026),
  [2409] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1231),
  [2411] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1029),
  [2413] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__property_name, 1, 0, 4), SHIFT(1698),
  [2416] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1513),
  [2418] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1036),
  [2420] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1126),
  [2422] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1289),
  [2424] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1076),
  [2426] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1682),
  [2428] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1077),
  [2430] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1089),
  [2432] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1090),
  [2434] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1091),
  [2436] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_object_repeat1, 2, 0, 18), REDUCE(aux_sym_object_pattern_repeat1, 2, 0, 19),
  [2439] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1135),
  [2441] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1598),
  [2443] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1113),
  [2445] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1114),
  [2447] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1129),
  [2449] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1582),
  [2451] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1108),
  [2453] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1109),
  [2455] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_object_repeat1, 2, 0, 18),
  [2457] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1120),
  [2459] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1017),
  [2461] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1245),
  [2463] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1031),
  [2465] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1030),
  [2467] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1553),
  [2469] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1104),
  [2471] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1133),
  [2473] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1018),
  [2475] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1232),
  [2477] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1022),
  [2479] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1105),
  [2481] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1032),
  [2483] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1033),
  [2485] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1134),
  [2487] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1580),
  [2489] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1107),
  [2491] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1600),
  [2493] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1599),
  [2495] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1584),
  [2497] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1585),
  [2499] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_decorator, 2, 0, 0),
  [2501] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_decorator, 2, 0, 0),
  [2503] = {.entry = {.count = 1, .reusable = true}}, SHIFT(122),
  [2505] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1791),
  [2507] = {.entry = {.count = 1, .reusable = true}}, SHIFT(101),
  [2509] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_method_definition, 5, 0, 94),
  [2511] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 5, 0, 94),
  [2513] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_method_definition, 6, 0, 100),
  [2515] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 6, 0, 100),
  [2517] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_method_definition, 3, 0, 58),
  [2519] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 3, 0, 58),
  [2521] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_export_statement_repeat1, 2, 0, 13),
  [2523] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_export_statement_repeat1, 2, 0, 13),
  [2525] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_export_statement_repeat1, 2, 0, 13), SHIFT_REPEAT(1470),
  [2528] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1085),
  [2530] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_decorator_member_expression, 3, 0, 45),
  [2532] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_decorator_member_expression, 3, 0, 45),
  [2534] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_method_definition, 4, 0, 77),
  [2536] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 4, 0, 77),
  [2538] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_method_definition, 4, 0, 84),
  [2540] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 4, 0, 84),
  [2542] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_method_definition, 6, 0, 99),
  [2544] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 6, 0, 99),
  [2546] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1075),
  [2548] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_method_definition, 7, 0, 108),
  [2550] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 7, 0, 108),
  [2552] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_method_definition, 5, 0, 89),
  [2554] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_method_definition, 5, 0, 89),
  [2556] = {.entry = {.count = 1, .reusable = true}}, SHIFT(104),
  [2558] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_class_body_repeat1, 1, 0, 35),
  [2560] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_class_body_repeat1, 1, 0, 35),
  [2562] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_class_body_repeat1, 1, 0, 35), SHIFT_REPEAT(1073),
  [2565] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1128),
  [2567] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1575),
  [2569] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1095),
  [2571] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1096),
  [2573] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_class_body_repeat1, 2, 0, 35),
  [2575] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_class_body_repeat1, 2, 0, 35),
  [2577] = {.entry = {.count = 1, .reusable = true}}, SHIFT(217),
  [2579] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1694),
  [2581] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1123),
  [2583] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1730),
  [2585] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1103),
  [2587] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1098),
  [2589] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1124),
  [2591] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1079),
  [2593] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1508),
  [2595] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1102),
  [2597] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1088),
  [2599] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_static_block, 3, 0, 38),
  [2601] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_static_block, 3, 0, 38),
  [2603] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1127),
  [2605] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1568),
  [2607] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1092),
  [2609] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1093),
  [2611] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_class_static_block, 2, 0, 7),
  [2613] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_class_static_block, 2, 0, 7),
  [2615] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1121),
  [2617] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1117),
  [2619] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1136),
  [2621] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1602),
  [2623] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1115),
  [2625] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1116),
  [2627] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1604),
  [2629] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1577),
  [2631] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1570),
  [2633] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1137),
  [2635] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1072),
  [2637] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1529),
  [2639] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1112),
  [2641] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1087),
  [2643] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1539),
  [2645] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1540),
  [2647] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1715),
  [2649] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1718),
  [2651] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1740),
  [2653] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1741),
  [2655] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_export_statement_repeat1, 1, 0, 2),
  [2657] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_export_statement_repeat1, 1, 0, 2),
  [2659] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1562),
  [2661] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1082),
  [2663] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1569),
  [2665] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1561),
  [2667] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1132),
  [2669] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1084),
  [2671] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1583),
  [2673] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1111),
  [2675] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1086),
  [2677] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_decorator_call_expression, 2, 0, 10),
  [2679] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_decorator_call_expression, 2, 0, 10),
  [2681] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1593),
  [2683] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1595),
  [2685] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1596),
  [2687] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1603),
  [2689] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1576),
  [2691] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1610),
  [2693] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1611),
  [2695] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1613),
  [2697] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1614),
  [2699] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1552),
  [2701] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1597),
  [2703] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1559),
  [2705] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1565),
  [2707] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1726),
  [2709] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1712),
  [2711] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1739),
  [2713] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1594),
  [2715] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1601),
  [2717] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1581),
  [2719] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1609),
  [2721] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1612),
  [2723] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1572),
  [2725] = {.entry = {.count = 1, .reusable = true}}, SHIFT(332),
  [2727] = {.entry = {.count = 1, .reusable = true}}, SHIFT(263),
  [2729] = {.entry = {.count = 1, .reusable = true}}, SHIFT(264),
  [2731] = {.entry = {.count = 1, .reusable = true}}, SHIFT(352),
  [2733] = {.entry = {.count = 1, .reusable = true}}, SHIFT(291),
  [2735] = {.entry = {.count = 1, .reusable = true}}, SHIFT(311),
  [2737] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1606),
  [2739] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1793),
  [2741] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1169),
  [2743] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_import, 1, 0, 0),
  [2745] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1758),
  [2747] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1215),
  [2749] = {.entry = {.count = 1, .reusable = true}}, SHIFT(131),
  [2751] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1724),
  [2753] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1301),
  [2755] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1215),
  [2757] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1848),
  [2759] = {.entry = {.count = 1, .reusable = true}}, SHIFT(770),
  [2761] = {.entry = {.count = 1, .reusable = true}}, SHIFT(129),
  [2763] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1160),
  [2765] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1288),
  [2767] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1315),
  [2769] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1275),
  [2771] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1295),
  [2773] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1156),
  [2775] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1279),
  [2777] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1146),
  [2779] = {.entry = {.count = 1, .reusable = true}}, SHIFT(593),
  [2781] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1256),
  [2783] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1148),
  [2785] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1313),
  [2787] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1154),
  [2789] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_jsx_element_repeat1, 2, 0, 0), SHIFT_REPEAT(129),
  [2792] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_jsx_element_repeat1, 2, 0, 0), SHIFT_REPEAT(1160),
  [2795] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_jsx_element_repeat1, 2, 0, 0), SHIFT_REPEAT(1288),
  [2798] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_jsx_element_repeat1, 2, 0, 0),
  [2800] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1309),
  [2802] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1269),
  [2804] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1278),
  [2806] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1265),
  [2808] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1303),
  [2810] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1356),
  [2812] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1398),
  [2814] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1442),
  [2816] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1775),
  [2818] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1752),
  [2820] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1772),
  [2822] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_jsx_opening_element_repeat1, 2, 0, 66), SHIFT_REPEAT(1215),
  [2825] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_jsx_opening_element_repeat1, 2, 0, 66), SHIFT_REPEAT(131),
  [2828] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_jsx_opening_element_repeat1, 2, 0, 66),
  [2830] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_jsx_opening_element_repeat1, 2, 0, 66), SHIFT_REPEAT(1215),
  [2833] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1423),
  [2835] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1828),
  [2837] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1482),
  [2839] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1317),
  [2841] = {.entry = {.count = 1, .reusable = true}}, SHIFT(738),
  [2843] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1294),
  [2845] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1272),
  [2847] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1299),
  [2849] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1300),
  [2851] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1263),
  [2853] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1302),
  [2855] = {.entry = {.count = 1, .reusable = true}}, SHIFT(621),
  [2857] = {.entry = {.count = 1, .reusable = true}}, SHIFT(622),
  [2859] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1264),
  [2861] = {.entry = {.count = 1, .reusable = true}}, SHIFT(623),
  [2863] = {.entry = {.count = 1, .reusable = true}}, SHIFT(757),
  [2865] = {.entry = {.count = 1, .reusable = true}}, SHIFT(781),
  [2867] = {.entry = {.count = 1, .reusable = true}}, SHIFT(624),
  [2869] = {.entry = {.count = 1, .reusable = true}}, SHIFT(785),
  [2871] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1447),
  [2873] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1866),
  [2875] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1497),
  [2877] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1209),
  [2879] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1766),
  [2881] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1208),
  [2883] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1216),
  [2885] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_variable_declarator, 1, 0, 6),
  [2887] = {.entry = {.count = 1, .reusable = true}}, SHIFT(193),
  [2889] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_variable_declarator, 1, 0, 6), SHIFT(1649),
  [2892] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1495),
  [2894] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1493),
  [2896] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_sequence_expression_repeat1, 2, 0, 0), SHIFT_REPEAT(252),
  [2899] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_pattern, 3, 0, 0),
  [2901] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_object_pattern, 4, 0, 0),
  [2903] = {.entry = {.count = 1, .reusable = true}}, SHIFT(272),
  [2905] = {.entry = {.count = 1, .reusable = true}}, SHIFT(218),
  [2907] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1468),
  [2909] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_jsx_attribute, 1, 0, 4),
  [2911] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_jsx_attribute, 1, 0, 4),
  [2913] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1168),
  [2915] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_object_pattern, 4, 0, 19),
  [2917] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_pattern, 4, 0, 0),
  [2919] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_object_pattern, 3, 0, 0),
  [2921] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1338),
  [2923] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1337),
  [2925] = {.entry = {.count = 1, .reusable = true}}, SHIFT(115),
  [2927] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_sequence_expression, 2, 0, 0),
  [2929] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_jsx_namespace_name, 3, 0, 0),
  [2931] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_jsx_namespace_name, 3, 0, 0),
  [2933] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_jsx_attribute, 1, 0, 0),
  [2935] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_jsx_attribute, 1, 0, 0),
  [2937] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1170),
  [2939] = {.entry = {.count = 1, .reusable = true}}, SHIFT(126),
  [2941] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_definition, 1, 0, 36),
  [2943] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1240),
  [2945] = {.entry = {.count = 1, .reusable = true}}, SHIFT(739),
  [2947] = {.entry = {.count = 1, .reusable = true}}, SHIFT(211),
  [2949] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_nested_identifier, 3, 0, 45),
  [2951] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_nested_identifier, 3, 0, 45),
  [2953] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_definition, 2, 0, 67),
  [2955] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_definition, 2, 0, 69),
  [2957] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_template_string_repeat1, 2, 0, 0), SHIFT_REPEAT(1233),
  [2960] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_template_string_repeat1, 2, 0, 0),
  [2962] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_template_string_repeat1, 2, 0, 0), SHIFT_REPEAT(211),
  [2965] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1759),
  [2967] = {.entry = {.count = 1, .reusable = true}}, SHIFT(443),
  [2969] = {.entry = {.count = 1, .reusable = true}}, SHIFT(191),
  [2971] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1805),
  [2973] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1233),
  [2975] = {.entry = {.count = 1, .reusable = true}}, SHIFT(706),
  [2977] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1244),
  [2979] = {.entry = {.count = 1, .reusable = true}}, SHIFT(635),
  [2981] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_switch_body_repeat1, 2, 0, 0), SHIFT_REPEAT(1759),
  [2984] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_switch_body_repeat1, 2, 0, 0),
  [2986] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_switch_body_repeat1, 2, 0, 0), SHIFT_REPEAT(191),
  [2989] = {.entry = {.count = 1, .reusable = true}}, SHIFT(599),
  [2991] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_definition, 3, 0, 86),
  [2993] = {.entry = {.count = 1, .reusable = true}}, SHIFT(425),
  [2995] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1543),
  [2997] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1840),
  [2999] = {.entry = {.count = 1, .reusable = true}}, SHIFT(319),
  [3001] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_jsx_opening_element, 2, -1, 0),
  [3003] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_jsx_opening_element, 2, -1, 0),
  [3005] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__jsx_string, 2, 0, 0),
  [3007] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__jsx_string, 2, 0, 0),
  [3009] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1151),
  [3011] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1254),
  [3013] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1161),
  [3015] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1353),
  [3017] = {.entry = {.count = 1, .reusable = true}}, SHIFT(990),
  [3019] = {.entry = {.count = 1, .reusable = false}}, SHIFT(230),
  [3021] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1241),
  [3023] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_jsx_expression, 3, 0, 0),
  [3025] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_jsx_expression, 3, 0, 0),
  [3027] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1038),
  [3029] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1289),
  [3031] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1367),
  [3033] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1345),
  [3035] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_jsx_opening_element, 4, -1, 65),
  [3037] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_jsx_opening_element, 4, -1, 65),
  [3039] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1368),
  [3041] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__jsx_string, 3, 0, 0),
  [3043] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__jsx_string, 3, 0, 0),
  [3045] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1428),
  [3047] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1281),
  [3049] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1733),
  [3051] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_pattern, 1, -1, 0), SHIFT(249),
  [3054] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_variable_declarator, 2, 0, 23),
  [3056] = {.entry = {.count = 1, .reusable = true}}, SHIFT(171),
  [3058] = {.entry = {.count = 1, .reusable = true}}, SHIFT(183),
  [3060] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1392),
  [3062] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_variable_declarator, 2, 0, 24),
  [3064] = {.entry = {.count = 1, .reusable = true}}, SHIFT(190),
  [3066] = {.entry = {.count = 1, .reusable = true}}, SHIFT(156),
  [3068] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1360),
  [3070] = {.entry = {.count = 1, .reusable = true}}, SHIFT(983),
  [3072] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_jsx_expression, 2, 0, 0),
  [3074] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_jsx_expression, 2, 0, 0),
  [3076] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1147),
  [3078] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1165),
  [3080] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1362),
  [3082] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_jsx_attribute, 3, 0, 4),
  [3084] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_jsx_attribute, 3, 0, 4),
  [3086] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_jsx_attribute, 3, 0, 0),
  [3088] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_jsx_attribute, 3, 0, 0),
  [3090] = {.entry = {.count = 1, .reusable = true}}, SHIFT(496),
  [3092] = {.entry = {.count = 1, .reusable = true}}, SHIFT(187),
  [3094] = {.entry = {.count = 1, .reusable = true}}, SHIFT(497),
  [3096] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1444),
  [3098] = {.entry = {.count = 1, .reusable = true}}, SHIFT(610),
  [3100] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1549),
  [3102] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1366),
  [3104] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1373),
  [3106] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_jsx_opening_element, 3, -1, 33),
  [3108] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_jsx_opening_element, 3, -1, 33),
  [3110] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1152),
  [3112] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1164),
  [3114] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_jsx_opening_element_repeat1, 1, 0, 34),
  [3116] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_jsx_opening_element_repeat1, 1, 0, 34),
  [3118] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1067),
  [3120] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1273),
  [3122] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_pattern, 1, -1, 0), SHIFT(319),
  [3125] = {.entry = {.count = 1, .reusable = true}}, SHIFT(548),
  [3127] = {.entry = {.count = 1, .reusable = true}}, SHIFT(161),
  [3129] = {.entry = {.count = 1, .reusable = true}}, SHIFT(526),
  [3131] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1496),
  [3133] = {.entry = {.count = 1, .reusable = true}}, SHIFT(788),
  [3135] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1527),
  [3137] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1424),
  [3139] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1258),
  [3141] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1511),
  [3143] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1145),
  [3145] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1157),
  [3147] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__jsx_string_repeat1, 2, 0, 0), SHIFT_REPEAT(1320),
  [3150] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__jsx_string_repeat1, 2, 0, 0),
  [3152] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__jsx_string_repeat1, 2, 0, 0), SHIFT_REPEAT(1320),
  [3155] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1201),
  [3157] = {.entry = {.count = 1, .reusable = true}}, SHIFT(420),
  [3159] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1587),
  [3161] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1399),
  [3163] = {.entry = {.count = 1, .reusable = true}}, SHIFT(249),
  [3165] = {.entry = {.count = 1, .reusable = true}}, SHIFT(484),
  [3167] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1509),
  [3169] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1825),
  [3171] = {.entry = {.count = 1, .reusable = false}}, SHIFT(687),
  [3173] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1395),
  [3175] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1396),
  [3177] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1316),
  [3179] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1350),
  [3181] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1351),
  [3183] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1235),
  [3185] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1349),
  [3187] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1504),
  [3189] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1352),
  [3191] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_variable_declaration_repeat1, 2, 0, 0), SHIFT_REPEAT(1201),
  [3194] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_variable_declaration_repeat1, 2, 0, 0),
  [3196] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1675),
  [3198] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1479),
  [3200] = {.entry = {.count = 1, .reusable = true}}, SHIFT(266),
  [3202] = {.entry = {.count = 1, .reusable = false}}, SHIFT(689),
  [3204] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1328),
  [3206] = {.entry = {.count = 1, .reusable = true}}, SHIFT(423),
  [3208] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_sequence_expression_repeat1, 2, 0, 0), SHIFT_REPEAT(234),
  [3211] = {.entry = {.count = 1, .reusable = true}}, SHIFT(230),
  [3213] = {.entry = {.count = 1, .reusable = true}}, SHIFT(293),
  [3215] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1238),
  [3217] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym__jsx_string_repeat2, 2, 0, 0), SHIFT_REPEAT(1348),
  [3220] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym__jsx_string_repeat2, 2, 0, 0),
  [3222] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym__jsx_string_repeat2, 2, 0, 0), SHIFT_REPEAT(1348),
  [3225] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1237),
  [3227] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1274),
  [3229] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_array_repeat1, 2, 0, 0), SHIFT_REPEAT(127),
  [3232] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1348),
  [3234] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1276),
  [3236] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1348),
  [3238] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1384),
  [3240] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1257),
  [3242] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1384),
  [3244] = {.entry = {.count = 1, .reusable = false}}, SHIFT(640),
  [3246] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1361),
  [3248] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1427),
  [3250] = {.entry = {.count = 1, .reusable = true}}, SHIFT(459),
  [3252] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1703),
  [3254] = {.entry = {.count = 1, .reusable = true}}, SHIFT(472),
  [3256] = {.entry = {.count = 1, .reusable = false}}, SHIFT(614),
  [3258] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1329),
  [3260] = {.entry = {.count = 1, .reusable = true}}, SHIFT(485),
  [3262] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1363),
  [3264] = {.entry = {.count = 1, .reusable = true}}, SHIFT(124),
  [3266] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1659),
  [3268] = {.entry = {.count = 1, .reusable = true}}, SHIFT(454),
  [3270] = {.entry = {.count = 1, .reusable = true}}, SHIFT(453),
  [3272] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1536),
  [3274] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1462),
  [3276] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_template_substitution, 3, 0, 0),
  [3278] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1519),
  [3280] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1492),
  [3282] = {.entry = {.count = 1, .reusable = true}}, SHIFT(991),
  [3284] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1560),
  [3286] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1473),
  [3288] = {.entry = {.count = 1, .reusable = true}}, SHIFT(475),
  [3290] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1271),
  [3292] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1206),
  [3294] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1320),
  [3296] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1320),
  [3298] = {.entry = {.count = 1, .reusable = true}}, SHIFT(476),
  [3300] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1663),
  [3302] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1401),
  [3304] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1555),
  [3306] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1400),
  [3308] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1298),
  [3310] = {.entry = {.count = 1, .reusable = true}}, SHIFT(426),
  [3312] = {.entry = {.count = 1, .reusable = true}}, SHIFT(477),
  [3314] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1709),
  [3316] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1471),
  [3318] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_object_pattern_repeat1, 2, 0, 19),
  [3320] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_string_repeat1, 2, 0, 0),
  [3322] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_string_repeat1, 2, 0, 0), SHIFT_REPEAT(1395),
  [3325] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_string_repeat2, 2, 0, 0),
  [3327] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_string_repeat2, 2, 0, 0), SHIFT_REPEAT(1396),
  [3330] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1355),
  [3332] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1355),
  [3334] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1605),
  [3336] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1588),
  [3338] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1745),
  [3340] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1737),
  [3342] = {.entry = {.count = 1, .reusable = true}}, SHIFT(59),
  [3344] = {.entry = {.count = 1, .reusable = true}}, SHIFT(741),
  [3346] = {.entry = {.count = 1, .reusable = true}}, SHIFT(132),
  [3348] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_array_pattern_repeat1, 2, 0, 0),
  [3350] = {.entry = {.count = 1, .reusable = true}}, SHIFT(19),
  [3352] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1252),
  [3354] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1194),
  [3356] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1453),
  [3358] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1207),
  [3360] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_object_repeat1, 2, 0, 0), SHIFT_REPEAT(989),
  [3363] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_object_repeat1, 2, 0, 0),
  [3365] = {.entry = {.count = 1, .reusable = true}}, SHIFT(616),
  [3367] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1247),
  [3369] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1248),
  [3371] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_object_pattern_repeat1, 2, 0, 0), SHIFT_REPEAT(992),
  [3374] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_object_pattern_repeat1, 2, 0, 0),
  [3376] = {.entry = {.count = 1, .reusable = true}}, SHIFT(617),
  [3378] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1250),
  [3380] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_export_specifier, 1, 0, 6),
  [3382] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1270),
  [3384] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1308),
  [3386] = {.entry = {.count = 1, .reusable = true}}, SHIFT(651),
  [3388] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1249),
  [3390] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_array_pattern_repeat1, 2, 0, 0), SHIFT_REPEAT(121),
  [3393] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pair_pattern, 3, 0, 57),
  [3395] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1193),
  [3397] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1119),
  [3399] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__from_clause, 2, 0, 21),
  [3401] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1535),
  [3403] = {.entry = {.count = 1, .reusable = true}}, SHIFT(114),
  [3405] = {.entry = {.count = 1, .reusable = true}}, SHIFT(697),
  [3407] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__module_export_name, 1, 0, 0),
  [3409] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_import_specifier, 1, 0, 6),
  [3411] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1198),
  [3413] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1797),
  [3415] = {.entry = {.count = 1, .reusable = true}}, SHIFT(592),
  [3417] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1689),
  [3419] = {.entry = {.count = 1, .reusable = true}}, SHIFT(461),
  [3421] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_export_clause, 5, 0, 0),
  [3423] = {.entry = {.count = 1, .reusable = true}}, SHIFT(227),
  [3425] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1219),
  [3427] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1218),
  [3429] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_export_clause, 3, 0, 0),
  [3431] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1222),
  [3433] = {.entry = {.count = 1, .reusable = true}}, SHIFT(638),
  [3435] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1673),
  [3437] = {.entry = {.count = 1, .reusable = true}}, SHIFT(450),
  [3439] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1234),
  [3441] = {.entry = {.count = 1, .reusable = true}}, SHIFT(639),
  [3443] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_export_clause_repeat1, 2, 0, 0), SHIFT_REPEAT(1205),
  [3446] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_export_clause_repeat1, 2, 0, 0),
  [3448] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1563),
  [3450] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1192),
  [3452] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1521),
  [3454] = {.entry = {.count = 1, .reusable = true}}, SHIFT(800),
  [3456] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1522),
  [3458] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1649),
  [3460] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1034),
  [3462] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1627),
  [3464] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1590),
  [3466] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_formal_parameters_repeat1, 2, 0, 0),
  [3468] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_formal_parameters_repeat1, 2, 0, 0), SHIFT_REPEAT(135),
  [3471] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_named_imports_repeat1, 2, 0, 0), SHIFT_REPEAT(1221),
  [3474] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_named_imports_repeat1, 2, 0, 0),
  [3476] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1794),
  [3478] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1670),
  [3480] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_export_clause, 2, 0, 0),
  [3482] = {.entry = {.count = 1, .reusable = true}}, SHIFT(735),
  [3484] = {.entry = {.count = 1, .reusable = true}}, SHIFT(128),
  [3486] = {.entry = {.count = 1, .reusable = true}}, SHIFT(734),
  [3488] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1239),
  [3490] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1713),
  [3492] = {.entry = {.count = 1, .reusable = true}}, SHIFT(163),
  [3494] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1734),
  [3496] = {.entry = {.count = 1, .reusable = true}}, SHIFT(170),
  [3498] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1747),
  [3500] = {.entry = {.count = 1, .reusable = true}}, SHIFT(805),
  [3502] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_export_clause, 4, 0, 0),
  [3504] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1697),
  [3506] = {.entry = {.count = 1, .reusable = true}}, SHIFT(112),
  [3508] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_definition, 3, 0, 85),
  [3510] = {.entry = {.count = 1, .reusable = true}}, SHIFT(23),
  [3512] = {.entry = {.count = 1, .reusable = true}}, SHIFT(494),
  [3514] = {.entry = {.count = 1, .reusable = true}}, SHIFT(495),
  [3516] = {.entry = {.count = 1, .reusable = true}}, SHIFT(195),
  [3518] = {.entry = {.count = 1, .reusable = true}}, SHIFT(545),
  [3520] = {.entry = {.count = 1, .reusable = true}}, SHIFT(547),
  [3522] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1073),
  [3524] = {.entry = {.count = 1, .reusable = true}}, SHIFT(177),
  [3526] = {.entry = {.count = 1, .reusable = true}}, SHIFT(25),
  [3528] = {.entry = {.count = 1, .reusable = true}}, SHIFT(140),
  [3530] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_formal_parameters, 2, 0, 0),
  [3532] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_export_specifier, 3, 0, 83),
  [3534] = {.entry = {.count = 1, .reusable = true}}, SHIFT(28),
  [3536] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_import_specifier, 3, 0, 83),
  [3538] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1371),
  [3540] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_import_clause, 1, 0, 0),
  [3542] = {.entry = {.count = 1, .reusable = true}}, SHIFT(158),
  [3544] = {.entry = {.count = 1, .reusable = true}}, SHIFT(162),
  [3546] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1680),
  [3548] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1855),
  [3550] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1658),
  [3552] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1784),
  [3554] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_formal_parameters, 3, 0, 0),
  [3556] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_formal_parameters, 4, 0, 0),
  [3558] = {.entry = {.count = 1, .reusable = true}}, SHIFT(449),
  [3560] = {.entry = {.count = 1, .reusable = true}}, SHIFT(446),
  [3562] = {.entry = {.count = 1, .reusable = true}}, SHIFT(427),
  [3564] = {.entry = {.count = 1, .reusable = true}}, SHIFT(482),
  [3566] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1236),
  [3568] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_formal_parameters, 5, 0, 0),
  [3570] = {.entry = {.count = 1, .reusable = true}}, SHIFT(436),
  [3572] = {.entry = {.count = 1, .reusable = true}}, SHIFT(422),
  [3574] = {.entry = {.count = 1, .reusable = true}}, SHIFT(421),
  [3576] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_definition, 3, 0, 87),
  [3578] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_import_attribute, 2, 0, 0),
  [3580] = {.entry = {.count = 1, .reusable = true}}, SHIFT(487),
  [3582] = {.entry = {.count = 1, .reusable = true}}, SHIFT(474),
  [3584] = {.entry = {.count = 1, .reusable = true}}, SHIFT(178),
  [3586] = {.entry = {.count = 1, .reusable = true}}, SHIFT(437),
  [3588] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_definition, 2, 0, 68),
  [3590] = {.entry = {.count = 1, .reusable = true}}, SHIFT(981),
  [3592] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1225),
  [3594] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1225),
  [3596] = {.entry = {.count = 1, .reusable = true}}, SHIFT(462),
  [3598] = {.entry = {.count = 1, .reusable = true}}, SHIFT(468),
  [3600] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_definition, 4, 0, 98),
  [3602] = {.entry = {.count = 1, .reusable = true}}, SHIFT(201),
  [3604] = {.entry = {.count = 1, .reusable = true}}, SHIFT(202),
  [3606] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1800),
  [3608] = {.entry = {.count = 1, .reusable = true}}, SHIFT(750),
  [3610] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1687),
  [3612] = {.entry = {.count = 1, .reusable = true}}, SHIFT(11),
  [3614] = {.entry = {.count = 1, .reusable = true}}, SHIFT(169),
  [3616] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_named_imports, 4, 0, 0),
  [3618] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1808),
  [3620] = {.entry = {.count = 1, .reusable = true}}, SHIFT(888),
  [3622] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_named_imports, 2, 0, 0),
  [3624] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1860),
  [3626] = {.entry = {.count = 1, .reusable = true}}, SHIFT(278),
  [3628] = {.entry = {.count = 1, .reusable = true}}, SHIFT(214),
  [3630] = {.entry = {.count = 1, .reusable = true}}, SHIFT(205),
  [3632] = {.entry = {.count = 1, .reusable = true}}, SHIFT(206),
  [3634] = {.entry = {.count = 1, .reusable = true}}, SHIFT(207),
  [3636] = {.entry = {.count = 1, .reusable = true}}, SHIFT(208),
  [3638] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1191),
  [3640] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1681),
  [3642] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1770),
  [3644] = {.entry = {.count = 1, .reusable = true}},  ACCEPT_INPUT(),
  [3646] = {.entry = {.count = 1, .reusable = true}}, SHIFT(598),
  [3648] = {.entry = {.count = 1, .reusable = true}}, SHIFT(222),
  [3650] = {.entry = {.count = 1, .reusable = true}}, SHIFT(231),
  [3652] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1043),
  [3654] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1751),
  [3656] = {.entry = {.count = 1, .reusable = true}}, SHIFT(642),
  [3658] = {.entry = {.count = 1, .reusable = true}}, SHIFT(130),
  [3660] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_named_imports, 3, 0, 0),
  [3662] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_namespace_import, 3, 0, 0),
  [3664] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1654),
  [3666] = {.entry = {.count = 1, .reusable = true}}, SHIFT(297),
  [3668] = {.entry = {.count = 1, .reusable = true}}, SHIFT(182),
  [3670] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_namespace_export, 3, 0, 0),
  [3672] = {.entry = {.count = 1, .reusable = true}}, SHIFT(701),
  [3674] = {.entry = {.count = 1, .reusable = true}}, SHIFT(184),
  [3676] = {.entry = {.count = 1, .reusable = true}}, SHIFT(186),
  [3678] = {.entry = {.count = 1, .reusable = true}}, SHIFT(197),
  [3680] = {.entry = {.count = 1, .reusable = true}}, SHIFT(198),
  [3682] = {.entry = {.count = 1, .reusable = true}}, SHIFT(215),
  [3684] = {.entry = {.count = 1, .reusable = true}}, SHIFT(153),
  [3686] = {.entry = {.count = 1, .reusable = true}}, SHIFT(154),
  [3688] = {.entry = {.count = 1, .reusable = true}}, SHIFT(795),
  [3690] = {.entry = {.count = 1, .reusable = true}}, SHIFT(199),
  [3692] = {.entry = {.count = 1, .reusable = true}}, SHIFT(200),
  [3694] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1383),
  [3696] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1822),
  [3698] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1787),
  [3700] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_import_clause, 3, 0, 0),
  [3702] = {.entry = {.count = 1, .reusable = true}}, SHIFT(188),
  [3704] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1377),
  [3706] = {.entry = {.count = 1, .reusable = true}}, SHIFT(510),
  [3708] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1728),
  [3710] = {.entry = {.count = 1, .reusable = true}}, SHIFT(521),
  [3712] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1230),
  [3714] = {.entry = {.count = 1, .reusable = true}}, SHIFT(167),
  [3716] = {.entry = {.count = 1, .reusable = true}}, SHIFT(508),
  [3718] = {.entry = {.count = 1, .reusable = true}}, SHIFT(175),
  [3720] = {.entry = {.count = 1, .reusable = true}}, SHIFT(271),
  [3722] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1544),
  [3724] = {.entry = {.count = 1, .reusable = true}}, SHIFT(152),
  [3726] = {.entry = {.count = 1, .reusable = true}}, SHIFT(339),
  [3728] = {.entry = {.count = 1, .reusable = true}}, SHIFT(179),
  [3730] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1591),
  [3732] = {.entry = {.count = 1, .reusable = true}}, SHIFT(525),
  [3734] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_named_imports, 5, 0, 0),
  [3736] = {.entry = {.count = 1, .reusable = true}}, SHIFT(185),
};

enum ts_external_scanner_symbol_identifiers {
  ts_external_token__automatic_semicolon = 0,
  ts_external_token__template_chars = 1,
  ts_external_token__ternary_qmark = 2,
  ts_external_token_html_comment = 3,
  ts_external_token_PIPE_PIPE = 4,
  ts_external_token_escape_sequence = 5,
  ts_external_token_regex_pattern = 6,
  ts_external_token_jsx_text = 7,
};

static const TSSymbol ts_external_scanner_symbol_map[EXTERNAL_TOKEN_COUNT] = {
  [ts_external_token__automatic_semicolon] = sym__automatic_semicolon,
  [ts_external_token__template_chars] = sym__template_chars,
  [ts_external_token__ternary_qmark] = sym__ternary_qmark,
  [ts_external_token_html_comment] = sym_html_comment,
  [ts_external_token_PIPE_PIPE] = anon_sym_PIPE_PIPE,
  [ts_external_token_escape_sequence] = sym_escape_sequence,
  [ts_external_token_regex_pattern] = sym_regex_pattern,
  [ts_external_token_jsx_text] = sym_jsx_text,
};

static const bool ts_external_scanner_states[10][EXTERNAL_TOKEN_COUNT] = {
  [1] = {
    [ts_external_token__automatic_semicolon] = true,
    [ts_external_token__template_chars] = true,
    [ts_external_token__ternary_qmark] = true,
    [ts_external_token_html_comment] = true,
    [ts_external_token_PIPE_PIPE] = true,
    [ts_external_token_escape_sequence] = true,
    [ts_external_token_jsx_text] = true,
  },
  [2] = {
    [ts_external_token_html_comment] = true,
  },
  [3] = {
    [ts_external_token__automatic_semicolon] = true,
    [ts_external_token__ternary_qmark] = true,
    [ts_external_token_html_comment] = true,
    [ts_external_token_PIPE_PIPE] = true,
  },
  [4] = {
    [ts_external_token__ternary_qmark] = true,
    [ts_external_token_html_comment] = true,
    [ts_external_token_PIPE_PIPE] = true,
  },
  [5] = {
    [ts_external_token__automatic_semicolon] = true,
    [ts_external_token_html_comment] = true,
  },
  [6] = {
    [ts_external_token_html_comment] = true,
    [ts_external_token_jsx_text] = true,
  },
  [7] = {
    [ts_external_token__template_chars] = true,
    [ts_external_token_html_comment] = true,
    [ts_external_token_escape_sequence] = true,
  },
  [8] = {
    [ts_external_token_html_comment] = true,
    [ts_external_token_escape_sequence] = true,
  },
  [9] = {
    [ts_external_token_html_comment] = true,
    [ts_external_token_regex_pattern] = true,
  },
};

#ifdef __cplusplus
extern "C" {
#endif
void *tree_sitter_javascript_external_scanner_create(void);
void tree_sitter_javascript_external_scanner_destroy(void *);
bool tree_sitter_javascript_external_scanner_scan(void *, TSLexer *, const bool *);
unsigned tree_sitter_javascript_external_scanner_serialize(void *, char *);
void tree_sitter_javascript_external_scanner_deserialize(void *, const char *, unsigned);

#ifdef TREE_SITTER_HIDE_SYMBOLS
#define TS_PUBLIC
#elif defined(_WIN32)
#define TS_PUBLIC __declspec(dllexport)
#else
#define TS_PUBLIC __attribute__((visibility("default")))
#endif

TS_PUBLIC const TSLanguage *tree_sitter_javascript(void) {
  static const TSLanguage language = {
    .abi_version = LANGUAGE_VERSION,
    .symbol_count = SYMBOL_COUNT,
    .alias_count = ALIAS_COUNT,
    .token_count = TOKEN_COUNT,
    .external_token_count = EXTERNAL_TOKEN_COUNT,
    .state_count = STATE_COUNT,
    .large_state_count = LARGE_STATE_COUNT,
    .production_id_count = PRODUCTION_ID_COUNT,
    .supertype_count = SUPERTYPE_COUNT,
    .field_count = FIELD_COUNT,
    .max_alias_sequence_length = MAX_ALIAS_SEQUENCE_LENGTH,
    .parse_table = &ts_parse_table[0][0],
    .small_parse_table = ts_small_parse_table,
    .small_parse_table_map = ts_small_parse_table_map,
    .parse_actions = ts_parse_actions,
    .symbol_names = ts_symbol_names,
    .field_names = ts_field_names,
    .field_map_slices = ts_field_map_slices,
    .field_map_entries = ts_field_map_entries,
    .supertype_map_slices = ts_supertype_map_slices,
    .supertype_map_entries = ts_supertype_map_entries,
    .supertype_symbols = ts_supertype_symbols,
    .symbol_metadata = ts_symbol_metadata,
    .public_symbol_map = ts_symbol_map,
    .alias_map = ts_non_terminal_alias_map,
    .alias_sequences = &ts_alias_sequences[0][0],
    .lex_modes = (const void*)ts_lex_modes,
    .lex_fn = ts_lex,
    .keyword_lex_fn = ts_lex_keywords,
    .keyword_capture_token = sym_identifier,
    .external_scanner = {
      &ts_external_scanner_states[0][0],
      ts_external_scanner_symbol_map,
      tree_sitter_javascript_external_scanner_create,
      tree_sitter_javascript_external_scanner_destroy,
      tree_sitter_javascript_external_scanner_scan,
      tree_sitter_javascript_external_scanner_serialize,
      tree_sitter_javascript_external_scanner_deserialize,
    },
    .primary_state_ids = ts_primary_state_ids,
    .name = "javascript",
    .reserved_words = &ts_reserved_words[0][0],
    .max_reserved_word_set_size = 35,
    .metadata = {
      .major_version = 0,
      .minor_version = 25,
      .patch_version = 0,
    },
  };
  return &language;
}
#ifdef __cplusplus
}
#endif
